(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Sequence(S: sig
		       type 'a sequence 
		       type 'a elt
		       val fromArray: 'a elt array -> 'a sequence (* constant time *)
		       val unsafeSub: 'a sequence * int -> 'a elt
		       val length: 'a sequence -> int
		    end
		 ): SEQUENCE =
   struct
      open S

      structure Array = Primitive.Array

      open Primitive.Int

      val maxLen = 0x7FFFFFFE (* Lengths can't have the high bit set because
			       * the gc uses the high bit to mark header words
			       * for normal objects.
			       * They also can't be 31 one bits, because that
			       * is reserved for stack object headers.
			       *)

      fun sub(s, i) =
	 if safe andalso Primitive.Int.geu(i, length s)
	    then raise Subscript
	 else unsafeSub(s, i)

      fun update(a, i, x) =
	 if safe andalso Primitive.Int.geu(i, Array.length a)
	    then raise Subscript
	 else Array.unsafeUpdate(a, i, x)

      (* Depends on the fact that the runtime system fills in the array
       * with reasonable bogus values.
       *)
      fun tabulate(n, f) =
	 if safe andalso geu(n, maxLen)
	    then raise Size
	 else let
		 val a = Array.array n
		 fun loop i =
		    if i = n then ()
		    else (update(a, i, f i) ; loop(i + 1))
	      in loop 0 ;
		 fromArray a
	      end

      fun new(n, x) = tabulate(n, fn _ => x)

      fun fromList l =
	 let val n = List.length l
	    val a = Array.array n
	    val rec loop =
	       fn ([],     _) => ()
		| (x :: l, i) => (update(a, i, x) ;
				  loop(l, i + 1))
	 in loop(l, 0) ;
	    fromArray a
	 end
		
      type 'a slice = 'a sequence * int * int option
	 
      fun 'a wholeSlice(a: 'a sequence): 'a slice = (a, 0, NONE)

      fun checkSliceMax(start: int, num: int option, max: int): int =
	 case num of
	    NONE => if safe andalso (start < 0 orelse start > max)
		       then raise Subscript
		    else max
	  | SOME num =>
	       (* the arithmetic below is careful not to overflow *)
	       if safe
		  andalso (start < 0 orelse num < 0 orelse start > max - num)
		  then raise Subscript
	       else start + num

      fun checkSlice(s, i, opt) = checkSliceMax(i, opt, length s)

      fun foldli f b (slice as (s, min, _)) =
	 let val max = checkSlice slice
	    fun loop(i, b) =
	       if i = max then b
	       else loop(i + 1, f(i, sub(s, i), b))
	 in loop(min, b)
	 end

      fun appi f slice = foldli (fn (i, x, ()) => f(i, x)) () slice

      fun app f s = appi (f o #2) (wholeSlice s)

      fun foldri f b (slice as (s, min, _)) =
	 let val max = checkSlice slice
	    fun loop(i, b) =
	       if i < min
		  then b
	       else loop(i - 1, f(i, sub(s, i), b))
	 in loop(max - 1, b)
	 end

      local
	 fun make foldi f b s = foldi (fn (_, x, b) => f(x, b)) b (wholeSlice s)
      in fun foldl f = make foldli f
	 fun foldr f = make foldri f
      end

      fun mapi f (slice as (s, min, _)) =
	 let val max = checkSlice slice
	 in tabulate(max - min, fn i => let val j = i + min
					in f(j, sub(s, j))
					end)
	 end

      fun map f s = mapi (f o #2) (wholeSlice s)

      fun extract slice = mapi #2 slice

      fun concat sequences =
	 case sequences of
	    [s] => s
	  | _ => 
	       let
		  val size = List.foldl(fn (s, n) => n + length s) 0 sequences
		  val a = Array.array size
	       in List.foldl(fn (s, n) =>
			     let
				val imax = length s
				fun loop i =
				   if i = imax then ()
				   else (update(a, n + i, sub(s, i));
					 loop(i + 1))
			     in loop 0;
				n + imax
			     end)
		  0 sequences ;
		  fromArray a
	       end

      fun prefixToList(s, n) =
	 let
	    fun loop(i, l) =
	       if i < 0
		  then l
	       else loop(i - 1, sub(s, i) :: l)
	 in loop(n - 1, [])
	 end

      fun toList a = prefixToList(a, length a)
	 
      fun find(s, p) =
	 let
	    val max = length s
	    fun loop i =
	       if i = max
		  then NONE
	       else let val x = sub(s, i)
		    in if p x then SOME x else loop(i + 1)
		    end
	 in loop 0
	 end
   end
