(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure PosixProcEnv: POSIX_PROC_ENV =
   struct
      structure Prim = PosixPrimitive.ProcEnv
      open Prim
      structure Error = PosixError
      structure CS = C.CS

      open Prim

      local
	 val a = Array.array(MLTON_numgroups, 0w0)
      in
	 fun getgroups() =
	    let val n = Prim.getgroups(a)
	    in Error.checkResult n;
	       Array.prefixToList(a, n)
	    end
      end

      fun id x = x
      val uidToWord = id 
      val wordToUid = id
      val gidToWord = id
      val wordToGid = id

      fun getlogin() =
	 let val cs = Prim.getlogin()
	 in if CS.isNull cs
	       then raise(Error.SysErr("no login name", NONE))
	    else CS.toString cs
	 end

      fun setpgid{pid, pgid} =
	 Prim.setpgid(case pid of
			 NONE => 0
		       | SOME n => n,
		      case pgid of
			 NONE => 0w0
		       | SOME w => w)

      local
	 structure Uname = Prim.Uname
	 val u = Uname.buffer
      in
	 fun uname() =
	    (Error.checkResult(Uname.uname u);
	     [("sysname", CS.toString(Uname.sysname u)),
	      ("nodename", CS.toString(Uname.nodename u)),
	      ("release", CS.toString(Uname.release u)),
	      ("version", CS.toString(Uname.version u)),
	      ("machine", CS.toString(Uname.machine u))])
      end

      val time = Time.now

      fun sysconf s =
	 case List.find (fn (_, s') => s = s') sysconfNames of
	    NONE => Error.raiseSys Error.inval
	  | SOME(n, _) =>
	       let val res = Prim.sysconf n
	       in Error.checkResult res;
		  SysWord.fromInt res
	       end
	       
      local
	 structure Tms = Prim.Tms
	 val buf = Tms.buffer
	    
	 val ticksPerSec = Real.fromInt(SysWord.toIntX(sysconf "CLK_TCK"))
	 fun cvt ticks = Time.fromReal(Real.fromInt ticks / ticksPerSec)
      in
	 fun times() =
	    let val elapsed = Prim.times buf
	    in Error.checkResult elapsed;
	       {elapsed = cvt elapsed,
		utime = cvt(Tms.utime buf), 
		stime = cvt(Tms.stime buf), 
		cutime = cvt(Tms.cutime buf), 
		cstime = cvt(Tms.cstime buf)}
	    end
      end

      fun getenv name =
	 let val cs = Prim.getenv(String.nullTerm name)
	 in if CS.isNull cs
	       then NONE
	    else SOME(CS.toString cs)
	 end

      local
	 val e = C.CSS.toList Prim.environ
      in
	 fun environ() = e
      end

      fun ctermid() = CS.toString(Prim.ctermid())
	 
      fun isatty(FD n) = Prim.isatty n

      fun ttyname(FD n) =
	 let val cs = Prim.ttyname n
	 in if CS.isNull cs
	       then Error.error()
	    else CS.toString cs
	 end
   end
