(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* empty tuple is also a record *)

functor Record(S: RECORD_STRUCTS): RECORD = 
struct

open S

datatype 'a t =
   Tuple of 'a list
 | Record of (Field.t * 'a) list

val tuple = Tuple

val isTuple =
   fn Tuple _ => true
    | _ => false

fun toList r =
   case r of
      Tuple l => let
		    fun loop(n, l) =
		       case l of
			  [] => []
			| x :: xs => (Field.Int n, x) :: loop(n+1, xs)
		 in loop(1, l)
		 end
    | Record r => r
	 
fun detupleOpt(r: 'a t): 'a list option =
   case r of
      Tuple t => SOME t
    | Record _ => NONE

fun fromList l =
   let
      fun sort r =
	 #insertionSort(List.ordered{< = fn ((s, _), (s', _)) => Field.<(s, s')}) r
      fun isTuple fs: bool =
	 let
	    fun loop(n, fs) =
	       case fs of
		  [] => true
		| (f, _) :: fs =>
		     case f of
			Field.Int n' => Int.equals(n, n') andalso loop(n + 1, fs)
		      | _ => false
	 in loop(1, fs)
	 end
      val l = if isSorted then sort l else l
   in if isTuple l
	 then Tuple(List.map(l, #2))
      else Record l
   end

val peek: 'a t * Field.t -> 'a option =
   fn (Tuple [], _) => NONE
    | (Tuple t, Field.Int n) => let
				   val rec loop =
				      fn (1, x :: _) => SOME x
				       | (_, []) => NONE
				       | (0, _) => NONE
				       | (n, _ :: xs) => loop(n - 1, xs)
				in loop(n, t)
				end
    | (Record r, f) =>
	 (case List.peek(r, fn (f', _) => Field.equals(f, f')) of
	     NONE => NONE
	   | SOME(_, x) => SOME x)
    | _ => Error.bug "Record.peek"

fun range r =
   case r of
      Tuple xs => xs
    | Record r => List.map(r, #2)

fun exists(r, p) =
   case r of
      Tuple xs => List.exists(xs, p)
    | Record r => List.exists(r, fn (_, x) => p x)

fun fold(r, b, f) =
   case r of
      Tuple xs => List.fold(xs, b, f)
    | Record r => List.fold(r, b, fn ((_, x), b) => f(x, b))
	  
fun map(r: 'a t, f: 'a -> 'b): 'b t =
   case r of
      Tuple xs => Tuple(List.map(xs, f))
    | Record r => Record(List.map(r, fn (field, a) => (field, f a)))

fun foreach(r: 'a t, f: 'a -> unit): unit =
   case r of
      Tuple xs => List.foreach(xs, f)
    | Record r => List.foreach(r, fn (_, a) => f a)

fun change(r: 'a t, f: 'a list -> 'b list * 'c): 'b t * 'c =
   case r of
      Tuple xs => let val (ys, c) = f xs
		  in (Tuple ys, c)
		  end
    | Record r => let val (fs, xs) = List.unzip r
		      val (ys, c) = f xs
		  in (Record(List.zip(fs, ys)), c)
		  end

fun zip arg = fromList(List.zip arg)
fun unzip arg = List.unzip(toList arg)

fun layout{record, layoutTuple, separator, extra, layoutElt} =
   case (record, extra) of
      (Tuple xs, "") => layoutTuple xs
    | _ =>
	 let val r = toList record
	    open Layout
	 in seq[str "{",
		align(separateRight(List.map(r, fn (f, x) =>
					     align[seq[Field.layout f,
						       str(" " ^ separator)],
						   layoutElt x]),
				    ", ")),
		str extra,
		str "}"]
	 end

end
