(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor HashType(S: HASH_TYPE_STRUCTS): HASH_TYPE = 
struct

open S

structure Type =
   struct
      datatype t =
	 T of {
	       hash: word,
	       plist: PropertyList.t,
	       tree: tree
	      }
      and tree =
	 Var of Tyvar.t
	| Con of Tycon.t * t list
	| Variant of Tycon.t * Con.t

      fun hash(T{hash, ...}) = hash
      fun plist(T{plist, ...}) = plist
      fun tree(T{tree, ...}) = tree

      structure Dest =
	 struct
	    datatype dest = datatype tree
	    val dest = tree
	 end
      open Dest

      fun deconOpt t =
	 case dest t of
	    Con x => SOME x
	  | _ => NONE

      fun makeHomVariant{var, variant, con} =
	 let
	    val {get, destroy, ...} =
	       Property.newDest
	       (plist,
		Property.initRec
		(fn (t, get) =>
		 case dest t of
		    Var a => var(t, a)
		  | Con(c, ts) => con(t, c, List.map(ts, get))
		  | Variant x => variant x))
	 in {hom = get, destroy = destroy}
	 end

      fun makeHom{var, con} =
	 makeHomVariant{var = var,
			con = con,
			variant = fn _ => Error.bug "makeMonoHom saw variant"}

      fun hom{ty, var, con} =
	 let val {hom, destroy} = makeHom{var = var o #2,
					  con = fn (_, c, xs) => con(c, xs)}
	 in hom ty before destroy()
	 end

      fun makeMonoHom{con} =
	 makeHom{var = fn _ => Error.bug "makeMonoHom saw type variable",
		 con = con}

      fun monoHom{ty, con} =
	 let val {hom, destroy} = makeMonoHom{con = con}
	 in hom ty before destroy()
	 end

      local
	 structure Atype = Ast.Type
      in
	 fun toAst t =
	    hom{ty = t,
		var = Atype.var,
		con = fn (c, ts) =>
		if Tycon.equals(c, Tycon.tuple) then Atype.tuple ts
		else Atype.con(Tycon.toAst c, ts)}
      end
			 
      val layout = Ast.Type.layout o toAst

(*      val hash = Trace.trace("hash", layout, Layout.str o Word.toString) hash *)
(*	 let open Layout
	 in
	    case tree of
	       Var a => Tyvar.layout a
	     | Con(c, ts) => seq[Tycon.layout c, tuple(List.map(ts, layout))]
	 end *)

      fun newHash() = Word.fromInt(Int.random())

      fun equals(t, t'): bool = PropertyList.equals(plist t, plist t')

(*      val equals = Trace.trace2("Type.equals", layout, layout, Bool.layout) equals *)

      structure Table = HashTable

      local
	 val size = 10000
	 exception NotFound

	 val same: t * t -> bool =
	    fn (T{tree, ...}, T{tree = t, ...}) =>
	    case (tree, t) of
	       (Var a, Var a') => Tyvar.equals(a, a')
	     | (Con(c, ts), Con(c', ts')) =>
		  Tycon.equals(c, c')
		  andalso if Tycon.equals(c, Tycon.tuple)
			     then List.equals(ts, ts', equals)
			  else List.forall2(ts, ts', equals)
	     | (Variant(t, c), Variant(t', c')) =>
		  Tycon.equals(t, t') andalso Con.equals(c, c')
	     | _ => false

(*	 val same = Trace.trace2("same", layout, layout, Bool.layout) same *)
      in
	 val table: (t, t) Table.hash_table =
	    Table.mkTable (hash, same) (size, NotFound)
      end

      local
	 val next = ref 1
      in
	 fun msg() =
	    if Table.numItems table = !next
	       then (next := !next * 2
		     ; Control.messageLay(List.layout Int.layout
					  (List.keepAll
					   (Table.bucketSizes table,
					    fn n => n <> 0))))
					   
	    else ()
      end

      fun lookup t =
	 ((*msg()
	  ; *)(case Table.find table t of
		NONE => (Table.insert table (t, t); t)
	      | SOME t => t))

      fun var a =
	 lookup(T{hash = Tyvar.hash a,
		  plist = PropertyList.new(),
		  tree = Var a})

      val generator = 0wx5555
      fun con(c, ts) =
	 lookup(T{hash = List.fold(ts, Tycon.hash c,
				   fn (t, w) => Word.xorb(w * generator, hash t)),
		  plist = PropertyList.new(),
		  tree = Con(c, ts)})

      fun variant(tycon, con) =
	 lookup(T{hash = Word.xorb(Tycon.hash tycon, Con.hash con),
		  plist = PropertyList.new(),
		  tree = Variant(tycon, con)})

(*       val variant =
 * 	 Trace.trace2("variant", Tycon.layout, Con.layout,
 * 		      Layout.str o Word.toString o hash) variant
 *)
   end
structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
open Type Ops

local structure Type = Ast.Type
in
   fun toAst(t: t): Type.t =
      case dest t of
	 Var a => Type.var a
       | Con(c, ts) =>
	    let val ts = List.map(ts, toAst)
	    in if Tycon.equals(c, Tycon.tuple) then Type.tuple ts
	       else Type.con(Tycon.toAst c, ts)
	    end
       | _ => Error.bug "Type.toAst"
end

fun optionToAst z = Option.map(z, toAst)

fun stats() =
   let open Layout
   in align[seq[str "num types in table = ", Int.layout(Table.numItems table)],
	    Control.sizeMessage("hash table", table)]
   end

fun ofConst c = con(Const.tycon c, [])

fun isUnit t =
   case dest t of
      Con(c, []) => Tycon.equals(c, Tycon.tuple)
    | _ => false

structure Env = MonoEnv(structure Domain = Tyvar
			structure Range = Type)
   
fun substitute(ty, e) =
   case e of
      [] => ty (* This optimization is important so that monotypes
		* are not substituted inside of.
		*)
    | _ => let val env = Env.fromList e
	   in hom{ty = ty,
		  var = fn a => (case Env.peek(env, a) of
				    NONE => var a
				  | SOME ty => ty),
		  con = con}
	   end

(* val substitute =
 *    Trace.trace2("substitute", layout,
 * 		List.layout(Layout.tuple2(Tyvar.layout, Type.layout)),
 * 		layout) substitute		
 *)

fun equalss(ts: t list): t option =
   case ts of
      t :: ts =>
	 let fun loop[] = SOME t
	       | loop(t' :: ts) = if equals(t, t') then loop ts else NONE
	 in loop ts
	 end
    | [] => Error.error "equals"

local
   val out = Out.error
   val print = Out.outputc out
   exception TypeError
in
   fun error(msg, lay) =
      (print("Type error: " ^ msg ^ "\n") ;
       Layout.output(lay, out) ;
       print "\n" ;
       raise TypeError)
end

local
   structure Ptype = Prim.Type
in
   fun fromPrims ts = List.map(ts, fromPrim)
   and fromPrim t =
      case t of
	 Ptype.Var a => var a
       | Ptype.Con(c, ts) => con(c, fromPrims ts)
       | Ptype.Record r => con(Tycon.tuple, fromPrims(SortedRecord.range r))

   fun toPrim t = hom{ty = t,
		      var = Ptype.var,
		      con = Ptype.con}
end

fun tycon t =
   case dest t of
      Con(c, _) => c
    | _ => Error.bug "Type.tycon saw type variable"

fun containsTycon(ty, tycon) =
   hom{ty = ty,
       var = fn _ => false,
       con = fn (tycon', bs) => (Tycon.equals(tycon, tycon')
				 orelse List.exists(bs, fn b => b))}

end
