(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure IdCounter:
   sig
      val getCounter: string -> Counter.t
   end =
   struct
      structure Hash = HashTable
      val hash = HashString.hashString
      val initSize = 10000
      exception NotFound
      val table: (string, Counter.t) Hash.hash_table =
	 Hash.mkTable (hash, String.equals) (initSize, NotFound)

      fun getCounter s =
	 case Hash.find table s of
	    SOME c => c
	  | NONE => let val c = Counter.new 0
		    in Hash.insert table (s, c) ; c
		    end
   end

functor Id(S: ID_STRUCTS): ID =
struct

open S

structure Plist = PropertyList
   
datatype t = T of {originalName: string,
		   counter: Counter.t,
		   printName: string option ref,
		   plist: Plist.t}

fun plist(T{plist, ...}) = plist
fun originalName(T{originalName, ...}) = originalName

fun setName(T{printName, ...}, s) = printName := SOME s
   
fun toString(T{printName, originalName, counter, ...}) =
   case !printName of
      NONE =>
	 let
	    val s =
	       concat[String.translate (fn #"'" => "P" | c => str c) originalName,
		      "_",
		      Int.toString(Counter.next counter)]
	 in printName := SOME s;
	    s
	 end
    | SOME s => s

val layout = String.layout o toString
   
fun sameName(id, id') = 
   String.equals(originalName id, originalName id')

fun equals(id, id') = Plist.equals(plist id, plist id')

fun fromString s =
   T{originalName = s,
     counter = IdCounter.getCounter s,
     printName = ref(SOME s),
     plist = Plist.new()}
   
fun isOperator s =
   let val c = String.sub(s, 0)
   in not(Char.isAlpha c orelse Char.equals(c, #"'"))
   end

fun newString s =
   let val s = if isOperator s then "operator" else s
      val counter = IdCounter.getCounter s
   in fn () => T{originalName = s,
		 counter = counter,
		 printName = ref NONE,
		 plist = Plist.new()}
   end

val newNoname = newString noname

val newString = fn s => newString s ()
   
val bogus = newString "BOGUS"

fun new(T{originalName, counter, ...}) =
   T{originalName = originalName,
     counter = counter,
     printName = ref NONE,
     plist = Plist.new()}

val fromAst = newString o AstId.toString
fun fromAsts l = List.map(l, fromAst)
val toAst = AstId.fromString o toString
fun toAsts l = List.map(l, toAst)

val clear = Plist.clear o plist
   
end

functor HashId(S: ID_STRUCTS): HASH_ID =
struct

open S

structure Plist = PropertyList
   
datatype t = T of {originalName: string,
		   counter: Counter.t,
		   printName: string option ref,
		   hash: word,
		   plist: Plist.t}

fun hash(T{hash, ...}) = hash
fun plist(T{plist, ...}) = plist
fun originalName(T{originalName, ...}) = originalName

fun setName(T{printName, ...}, s) = printName := SOME s

fun toString(T{printName, originalName, counter, ...}) =
   case !printName of
      NONE =>
	 let
	    val s =
	       concat[String.translate (fn #"'" => "P" | c => str c) originalName,
		      "_",
		      Int.toString(Counter.next counter)]
	 in printName := SOME s;
	    s
	 end
    | SOME s => s

val layout = String.layout o toString
   
fun sameName(id, id') = 
   String.equals(originalName id, originalName id')

fun equals(id, id') = Plist.equals(plist id, plist id')

fun newHash() = Word.fromInt(Int.random())
   
fun fromString s =
   T{originalName = s,
     counter = IdCounter.getCounter s,
     printName = ref(SOME s),
     hash = newHash(),
     plist = Plist.new()}
   
fun isOperator s =
   let val c = String.sub(s, 0)
   in not(Char.isAlpha c orelse Char.equals(c, #"'"))
   end

fun newString s =
   let val s = if isOperator s then "operator" else s
      val counter = IdCounter.getCounter s
   in fn () => T{originalName = s,
		 counter = counter,
		 printName = ref NONE,
		 hash = newHash(),
		 plist = Plist.new()}
   end

val newNoname = newString noname

val newString = fn s => newString s ()
   
val bogus = newString "BOGUS"

fun new(T{originalName, counter, ...}) =
   T{originalName = originalName,
     counter = counter,
     printName = ref NONE,
     hash = newHash(),
     plist = Plist.new()}

val fromAst = newString o AstId.toString
fun fromAsts l = List.map(l, fromAst)
val toAst = AstId.fromString o toString
fun toAsts l = List.map(l, toAst)

val clear = Plist.clear o plist
   
end
