(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Prim(S: PRIM_STRUCTS): PRIM = 
struct

open S

local open Type
in
   structure Tycon = Tycon
   structure Tyvar = Tyvar
end

structure Name =
   struct
      datatype t =
	 Array | ArraySub | ArrayUpdate | ArrayLength
       | VectorSub | VectorLength | VectorFromArray
       | StringSub 
       | Ref | Deref | Assign
       | Bogus
       | SaveWorld
       | SaveStack
       | ExnName
       | Eq
       | Equal
       | Size
       | CPrim of string

      val isArrayAllocation = 
	 fn Array => true
	  | _ => false

      val isFunctional =
	 fn ArrayLength => true
	  | VectorSub => true
	  | VectorLength => true
	  | VectorFromArray => true
	  | StringSub => true
	  | ExnName => true
	  | Eq => true
	       (* Equal is commented out because I don't want globalization or
		* constant propagation to move it into the globals, because
		* polymorphic equality isn't implemented there.
		*)
(*	  | Equal => true  *)
	  | _ => false

      fun isMoveable p =
	 isFunctional p
	 orelse
	 case p of
	    Array => true
	  | Ref => true
	  | _ => false

      val maySideEffect =
	 fn ArrayUpdate => true
	  | Assign => true
	  | SaveWorld => true
	  | SaveStack => true
	  | Size => true
	  | CPrim _ => true
	  | _ => false

      val toString =
	 fn Array => "MLTON_Array_array"
	  | ArraySub => "MLTON_Array_sub"
	  | ArrayUpdate => "MLTON_Array_update"
	  | ArrayLength => "MLTON_Array_length"
	  | VectorSub => "MLTON_Vector_sub"
	  | VectorLength => "MLTON_Vector_length"
	  | VectorFromArray => "MLTON_Vector_fromArray"
	  | StringSub => "MLTON_String_sub"
	  | Ref => "ref"
	  | Deref => "!"
	  | Assign => ":="
	  | Bogus => "MLTON_MLton_bogus"
	  | SaveWorld => "MLTON_MLton_saveWorld"
	  | SaveStack => "MLTON_MLton_saveStack"
	  | ExnName => "MLTON_General_exnName"
	  | Equal => "="
	  | Eq => "MLTON_eq"
	  | Size => "MLTON_MLton_size"
	  | CPrim s => s

      (* Some of the values of these strings are important since they
       * are referred to in the basis library code.
       *)
      val fromString =
	 fn "MLTON_Array_array" => Array
	  | "MLTON_Array_sub" => ArraySub
	  | "MLTON_Array_update" => ArrayUpdate
	  | "MLTON_Array_length" => ArrayLength
	  | "MLTON_Vector_sub" => VectorSub
	  | "MLTON_Vector_length" => VectorLength
	  | "MLTON_Vector_fromArray" => VectorFromArray
	  | "MLTON_String_sub" => StringSub
	  | "ref" => Ref
	  | "!" => Deref
	  | ":=" => Assign
	  | "MLTON_MLton_saveWorld" => SaveWorld
	  | "MLTON_MLton_saveStack" => SaveStack
	  | "MLTON_MLton_size" => Size
	  | "MLTON_General_exnName" => ExnName
	  | "=" => Equal
	  | "MLTON_eq" => Eq
	  | s => CPrim s

      val layout = Layout.str o toString
   end

datatype t =
   T of {name: Name.t,
	 scheme: Scheme.t,
	 isMoveable: bool,
	 isFunctional: bool,
	 maySideEffect: bool,
	 numArgs: int option,
	 plist: PropertyList.t}

fun name(T{name, ...}) = name
fun isMoveable(T{isMoveable, ...}) = isMoveable
fun isFunctional(T{isFunctional, ...}) = isFunctional
fun maySideEffect(T{maySideEffect, ...}) = maySideEffect

val layout = Name.layout o name
val isArrayAllocation = Name.isArrayAllocation o name

fun scheme(T{scheme, ...}) = scheme
(*val scheme = Trace.trace("Prim.scheme", layout, Scheme.layout) scheme *)

fun numArgs(T{numArgs, ...}) = numArgs
(*val numArgs = Trace.trace("numArgs", layout, Option.layout Int.layout) numArgs *)
			  
fun plist(T{plist, ...}) = plist

val toString = Name.toString o name
val longid = Longid.short o Longid.Id.fromString o toString

fun equals(p, p') = PropertyList.equals(plist p, plist p')

fun new(name, scheme) =
   let
      val numArgs =
	 case Type.dearrowOpt(Scheme.ty scheme) of
	    NONE => NONE
	  | SOME(t, _) => (case Type.detupleOpt t of
			      NONE => SOME 1
			    | SOME ts => SOME(List.length ts))
   in T{plist = PropertyList.new(),
	name = name,
	scheme = scheme,
	isFunctional = (Name.isFunctional name
			orelse not(isSome numArgs)),
	isMoveable = Name.isMoveable name,
	maySideEffect = Name.maySideEffect name,
	numArgs = numArgs}
   end

local
   val newPrim = new
   open Type Scheme
   val new = newPrim
   val --> = arrow
   infix -->
in
   val deref = new(Name.Deref, make1(fn a => reff a --> a))
   val reff = new(Name.Ref, make1(fn a => a --> reff a))
   val vectorSub =
      new(Name.VectorSub, make1(fn a => tuple[vector a, int] --> a))
   val vectorLength =
      new(Name.VectorLength, make1(fn a => vector a --> int))
   val equal = new(Name.Equal, makeEqual1(fn a => tuple[a, a] --> bool))
   val eq = new(Name.Eq, makeEqual1(fn a => tuple[a, a] --> bool))
   val array = new(Name.Array, make1(fn a => int --> array a))
   val bogus = new(Name.Bogus, make1(fn a => a))
   val saveStack = new(Name.SaveStack, make0(unit --> stack))
   val saveWorld = new(Name.SaveWorld, make0(string --> bool))
   val size = new(Name.Size, make1(fn a => a --> int))

   val new = fn (str, scheme) => new(Name.fromString str, scheme)

   fun new0(name, ty) = new(name, make0 ty)

   val stringToIntInf =
      new0("MLTON_IntInf_fromString", tuple[string, Type.array word] --> intInf)
   val stringEqual = new0("MLTON_String_equal", tuple[string, string] --> bool)
   val intInfEqual = new0("MLTON_IntInf_equal", tuple[intInf, intInf] --> bool)
   val exit = new0("exit", int --> unit)
   val intPlus = new0("MLTON_Int_add", tuple[int, int] --> int)
   val print = new0("print", string --> unit)

   fun newNullary(name: string) = new0(name, unit --> unit)
end

fun 'a checkApp{prim, targs, args,
		con, detupleOpt, dearrowOpt, equals, isUnit}
   : 'a option =
   let
      val error = NONE
      val Scheme.T{tyvars, ty} = scheme prim
   in
      if List.length targs <> List.length tyvars
	 then error
      else
	 let
	    val env = List.zip(tyvars, targs)
	    fun var a =
	       case List.peek(env, fn (a', _) =>
			      Tyvar.equals(a, a')) of
		  NONE => Error.bug "prim scheme with free tyvar"
		| SOME(_, t) => t
	    val ty = Type.hom{ty = ty, var = var, con = con}
	 in
	    case numArgs prim of
	       NONE => (case args of
			   [] => SOME ty
			 | _ => error)
	     | SOME n =>
		  case dearrowOpt ty of
		     NONE => error
		   | SOME(argType, result) =>
			case (n, args) of
			   (0, []) => SOME result
			 | (1, [arg]) =>
			      if equals(arg, argType)
				 then SOME result
			      else error
			 | _ => 
			      case detupleOpt argType of
				 NONE => error
			       | SOME argTypes =>
				    if List.equals(args, argTypes, equals)
				       then SOME result
				    else error
	 end
   end

fun returnsBool p =
   case Type.dearrowOpt(Scheme.ty(scheme p)) of
      SOME(_, Type.Con(tycon, _)) => Tycon.equals(tycon, Tycon.bool)
    | _ => false

end
