(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AllocateRegisters(S: ALLOCATE_REGISTERS_STRUCTS): ALLOCATE_REGISTERS = 
struct

open S

local open Cps
in
   structure Dec = Dec
   structure Exp = Exp
   structure Func = Func
   structure Jump = Jump
   structure Prim = Prim
   structure PrimExp = PrimExp
   structure Program = Program
   structure Ctransfer = Transfer
   structure Var = Var
end

local open Machine
in
   structure Chunk = Chunk
   structure GCInfo = GCInfo
   structure Operand = Operand
   structure Mtype = Type
   structure Register = Register
   structure Statement = Statement
end

(* If a handler is stored in a stack frame, then we need both a uint for
 * the old handler and space for the handler itself
 *)
local
   open Mtype
in
   val labelSize = size label
   val handlerSize = labelSize + size uint
end

(* Compute live variables *)
structure Live = Live(structure Cps = Cps)

(*
 * Every variable that is ever live across a nontail call or in a handler
 * gets a stack slot.  Otherwise, it gets a register.
 * Only exception is for formals, which must be made registers.
 * For these, copy them into stack slot at beginning of block.
 *)

structure Info =
   struct
      datatype t =
	 T of {
	       varOperand: Cps.Var.t -> Machine.Operand.t,
	       gcInfo: GCInfo.t,
	       prelude: Machine.Statement.t list,
	       postlude: Machine.Statement.t list
	       }
   end 

(* ------------------------------------------------- *)
(*                     allocate                      *)
(* ------------------------------------------------- *)

fun allocate{program = Program.T{globals, functions, ...},
	     varType, jumpChunk, funcChunk, shouldAllocate, jumpHandlers} =
   let
      val {get = varOperand, set = setVarOperand} =
	 Property.new(Var.plist, Property.initRaise("operand", Var.layout))

(*       val varOperand =
 * 	 Trace.trace("varOperand", Var.layout, Operand.layout) varOperand
 * 	 
 *       val setVarOperand =
 * 	 Trace.trace2("setVarOperand",
 * 		      Var.layout, Operand.layout, Unit.layout) setVarOperand
 *)
      val {get = jumpInfo: Jump.t -> Info.t, set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("jump info", Jump.layout))

      val {get = contInfo: Jump.t ->  {
				       size: int,
				       liveOffsets: int list
				       },
	   set = setContInfo} =
	 Property.new(Jump.plist, Property.initRaise("cont info", Jump.layout))

      val {get = handlerInfo: Jump.t ->  {
					  size: int,
					  liveOffsets: int list
					  },
	   set = setHandlerInfo} =
	 Property.new(Jump.plist, Property.initRaise("handler info", Jump.layout))

      val {get = funcInfo, set = setFuncInfo} =
	 Property.new(Func.plist, Property.initRaise("func info", Func.layout))

      val {get = liveArray, set = setLiveArray} =
	 Property.new(Var.plist, Property.initRaise("liveArray", Var.layout))

      fun allocateFunc(name: Func.t,
		       args: (Var.t * Cps.Type.t) list,
		       body: Exp.t): unit = 
	 let
	    val {liveBegin, liveBeginNoFormals, liveArray = liveArrayVars,
		 destroy = destroyLive} =
	       Live.live{exp = body,
			 shouldConsider = shouldAllocate,
			 jumpHandlers = jumpHandlers}

	    (*
	     * Decide which variables will live in stack slots and which
	     * will live in registers.
	     * Initially, try to put everything in a register.
	     * Variables get moved to the stack if they are live in a handler
	     * or live across a nontail call.
	     *)
	    datatype place = Stack | Register
	    val {get = place: Var.t -> place,
		 set = setPlace, destroy = destroyPlace} =
	       Property.newDest(Var.plist, Property.initConst Register)

	    val {get = isCont: Jump.t -> bool,
		 set = setIsCont, destroy = destroyIsCont} =
	       Property.newDest(Jump.plist, Property.initConst false)

	    val {get = isHandler: Jump.t -> bool,
		 set = setIsHandler, destroy = destroyIsHandler} =
	       Property.newDest(Jump.plist, Property.initConst false)

	    (* Are handlers ever installed in this function? *)
	    val hasHandler: bool ref = ref false

	    fun forceStack(j: Jump.t): unit =
	       List.foreach(liveBegin j, fn x => setPlace(x, Stack))
	       
	    fun loopExp(e: Exp.t): unit =
	       let val {decs, transfer} = Exp.dest e
	       in List.foreach(decs, loopDec)
		  ; (case transfer of
			Ctransfer.Call{cont, ...} =>
			   (case cont of
			       NONE => ()
			     | SOME c => (forceStack c
					  ; setIsCont(c, true)))
		      | _ => ())
	       end
	    and loopDec(d: Dec.t) =
	       case d of
		  Dec.Fun{body, ...} => loopExp(body)
		| Dec.HandlerPush h => (forceStack h
					; setIsHandler(h, true)
					; hasHandler := true)
		| _ => ()

	    val _ = loopExp(body)

	    (* Do the allocation top down.
	     * Whenever entering a local function ...
	     *)

	    val frameSize = ref 0

	    (* The next available offset on the stack.
	     * It is 4 initially because the stack pointer always points
	     * at the return address in the previous frame.
	     *)
	    val nextOffset = 4

	    val (nextOffset, handlerOffset) =
	       if !hasHandler
		  then (nextOffset + handlerSize, SOME nextOffset)
	       else (nextOffset, NONE)

	    fun getMask(xs: Var.t list): {regs: Register.t list,
					  offsets: int list} =
	       List.fold
	       (xs, {regs = [], offsets = []}, fn (x, ros as {regs, offsets}) =>
		if shouldAllocate(x) andalso Mtype.isPointer(varType x)
		   then
		      let val opr = varOperand x
		      in case place(x) of
			 Register =>
			    (case Operand.deRegister opr of
				NONE => Error.bug "must be a register"
			      | SOME r => {regs = r :: regs, offsets = offsets})
		       | Stack =>
			    case Operand.deStackOffset opr of
			       NONE => Error.bug "must be a slot"
			     | SOME{offset, ...} => {offsets = offset :: offsets,
						     regs = regs}
		      end
		else ros)

	    val nextOffset = ref nextOffset

	    (* The next available register number for each type. *)
	    val nextReg = Mtype.memo(fn _ => ref 0)

	    fun nextRegister(ty: Mtype.t, c: Chunk.t): Register.t =
	       let val r = nextReg ty
		  val reg = Chunk.register(c, !r, ty)
	       in IntRef.inc(r)
		  ; reg
	       end

	    fun allocateVar(x: Var.t, c: Chunk.t): unit =
	       if shouldAllocate(x)
		  then (setVarOperand
			(x, 
			 let val ty = varType(x)
			 in case place(x) of
			    Stack =>
			       let val offset = Mtype.align(!nextOffset, ty)
			       in nextOffset := offset + Mtype.size(ty)
				  ; Operand.stackOffset{ty = ty,
							offset = offset}
			       end
			  | Register => Operand.register(nextRegister(ty, c))
			 end))
	       else ()

(* 	    val allocateVar =
 * 	       Trace.trace2("allocateVar", Var.layout, Layout.ignore, Unit.layout)
 * 	       allocateVar
 *)
	    val chunk = funcChunk(name)

	    val {get = externalVarOperand,
		 set = setExternalVarOperand} =
	       Property.new(Var.plist, Property.initConst NONE)
		 
	    val (prelude, liveRegs) =
	       List.fold
	       (args, ([], []), fn ((x, _), (prelude, liveRegs)) =>
		(allocateVar(x, chunk)
		 ; let
		      val opr = varOperand(x)
		      val (r, prelude) =
			 case Operand.deRegister(opr) of
			    SOME r => (r, prelude)
			  | NONE =>
			       let val r = nextRegister(varType x, chunk)
				  val opr' = Operand.register r
			       in setExternalVarOperand(x, SOME opr')
				  ; (r,
				     Statement.move{dst = opr, src = opr'}
				     :: prelude)
			       end
		      val liveRegs =
			 if Mtype.isPointer(Register.ty(r))
			    then r :: liveRegs
			 else liveRegs
		   in (prelude, liveRegs)
		   end))

	    fun allocate(e: Exp.t, c: Chunk.t): unit =
	       let
		  val {decs, ...} = Exp.dest e
		  fun getGCInfo(live: Var.t list) =
		     let val {regs, offsets} = getMask(live)
		     in GCInfo.T{frameSize = !nextOffset,
				 regs = regs,
				 offsets = offsets}
		     end
	       in List.foreach
		  (decs,
		   fn Dec.Bind{var, exp, ...} =>
		        (allocateVar(var, c)
			 ; (case exp of
			       PrimExp.PrimApp{prim, ...} =>
				  if Prim.isArrayAllocation(prim)
				     then
					setLiveArray
					(var, getGCInfo(liveArrayVars var))
				  else ()
			     | _ => ()))
		    | Dec.Fun{name, args, body, ...} =>
			 let
			    val saveReg = Mtype.memo(! o nextReg)
			    val saveOffset = !nextOffset
			    val _ = List.foreach(Mtype.all, ignore o saveReg)
			    val c' = jumpChunk(name)
			    val _ =
			       if Chunk.equals(c, c')
				  then ()
			       else (* We can reset all of the register counters
				     * because we know that no registers live
				     * across a chunk boundary.
				     *)
				  List.foreach(Mtype.all, fn t =>
					       nextReg t := 0)
			    val _ = List.foreach(args, fn (x, _) =>
						 allocateVar(x, c'))
			    val _ = allocate(body, c')
			    fun getLiveOffsets() =
			       #offsets(getMask(liveBeginNoFormals(name)))
			    val _ =
			       if isCont(name)
				  then
				     setContInfo
				     (name, {size = !nextOffset,
					     liveOffsets = getLiveOffsets()})
			       else ()
			    val _ =
			       if isHandler(name)
				  then
				     setHandlerInfo
				     (name, {size = valOf(handlerOffset),
					     liveOffsets = getLiveOffsets()})
			       else ()
			    val _ =
			       setJumpInfo
			       (name, Info.T{varOperand = varOperand,
					     gcInfo = getGCInfo(liveBegin name),
					     prelude = [],
					     postlude = []})

			    val _ = List.foreach(Mtype.all, fn t =>
						 nextReg(t) := saveReg(t))
			    val _ = nextOffset := saveOffset
			 in ()
			 end
		    | _ => ())
	       end

	    val _ = allocate(body, chunk)
	       
	    val varOperand =
	       fn x =>
	       case externalVarOperand(x) of
		  NONE => varOperand(x)
		| SOME z => z

	 in destroyPlace()
	    ; destroyIsCont()
	    ; destroyIsHandler()
	    ; (setFuncInfo
	       (name,
		{info =
		 Info.T{varOperand = varOperand,
			gcInfo = GCInfo.T{regs = liveRegs,
					  offsets = [],
					  frameSize = Mtype.size Mtype.label},
			prelude = prelude,
			postlude = []},
		 handlerOffset = handlerOffset}))
	    ; ()
	 end
   in List.foreach(functions, fn {name, args, body, ...} =>
		   allocateFunc(name, args, body))
      ; {arrayInfo = liveArray,
	 funcInfo = funcInfo,
	 contInfo = contInfo,
	 handlerInfo = handlerInfo,
	 jumpInfo = jumpInfo}
   end
end
