(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AllocateRegisters(S: ALLOCATE_REGISTERS_STRUCTS): ALLOCATE_REGISTERS = 
struct

open S

local open Cps
in
   structure Dec = Dec
   structure Exp = Exp
   structure Func = Func
   structure Jump = Jump
   structure Prim = Prim
   structure PrimExp = PrimExp
   structure Program = Program
   structure Ctransfer = Transfer
   structure Var = Var
end

local open Machine
in
   structure Chunk = Chunk
   structure Operand = Operand
   structure Mtype = Type
   structure Register = Register
   structure Statement = Statement
end

(* If a handler is stored in a stack frame, then we need both a uint for
 * the old handler and space for the handler itself
 *)
local
   open Mtype
in
   val labelSize = size label
   val handlerSize = labelSize + size uint
end

(* Compute live variables *)
structure Live = Live(structure Cps = Cps)

structure Dset = DisjointSet
   
(*
 * Simplest register allocator.
 * Every variable that is ever live across a nontail call or in a handler
 *  gets a stack slot.
 * Otherwise, make it a register.
 * Only exception is for formals, which must be made registers.
 * For these, copy them into stack slot at beginning of block.
 *
 * Initially, everybody requests a register.
 *  If somebody is forced to request a stack slot, it invalidates its register
 *  request, unless it is a formal.  A formal keeps the register request around
 *  and inserts code at the beginning of the block to save to the stack slot.
 * Stack slot requests are grouped by type.
 *)

structure RegRequest =
   struct
      datatype t = T of {(* all variables who share this register *)
			 vars: Var.t list ref,
			 chunk: Chunk.t,
			 valid: bool ref,
			 regRef: Register.t option ref}

      fun vars(T{vars, ...}) = !vars
   end

structure SlotRequest =
   struct
      datatype t = T of {(* all variables who share this slot *)
			 vars: Var.t list ref,
			 offsetRef: int option ref}
   end

structure Request =
   struct
      datatype t =
	 LocalReg of RegRequest.t
       | LocalSlot of SlotRequest.t
	 (* The initial request for (toplevel) function formals.
	  * This RegRequest is always valid, since arguments must be passed
	  * in registers, not on the stack.
	  *)
       | FormalReg of RegRequest.t
	 (* If the formal needs a stack slot, then its request looks like this.
	  * This RegRequest is always valid, since arguments must be passed
	  * in registers, not on the stack.
	  *)
       | FormalRegSlot of RegRequest.t * SlotRequest.t

      fun addVar(r: t, x: Var.t): unit =
	 let
	    fun reg(RegRequest.T{vars, ...}) = vars
	    fun slot(SlotRequest.T{vars, ...}) = vars
	    val vars =
	       case r of
		  LocalReg r => reg r
		| LocalSlot s => slot s
		| FormalReg r => reg r
		| FormalRegSlot(_, s) => slot s
	 in ListRef.push(vars, x)
	 end

      val toString =
	 fn LocalReg _ => "LocalReg"
	  | LocalSlot _ => "LocalSlot"
	  | FormalReg _ => "FormalReg"
	  | FormalRegSlot _ => "FormalRegSlot"

      val layout = Layout.str o toString
   end
datatype request = datatype Request.t

structure Info =
   struct
      datatype t =
	 T of {
	       varOperand: Cps.Var.t -> Machine.Operand.t,
	       liveBegin: {regs: Machine.Register.t list,
			   offsets: int list},
	       liveArray: Cps.Var.t -> {regs: Machine.Register.t list,
					offsets: int list},
	       prelude: Machine.Statement.t list,
	       postlude: Machine.Statement.t list
	       }
   end 

(* ------------------------------------------------- *)
(*                     allocate                      *)
(* ------------------------------------------------- *)

fun allocate{program = Program.T{globals, functions, ...},
	     varType, jumpChunk, funcChunk, shouldAllocate, jumpHandlers} =
   let
      val {get = isGlobal: Var.t -> bool, set = setGlobal,
	   destroy = destroyGlobal} =
	 Property.newDest(Var.plist, Property.initConst false)

      val _ = List.foreach(globals, fn {var, ...} => setGlobal(var, true))
	 
      val {get = varOperand, set = setVarOperand} =
	 Property.new(Var.plist, Property.initRaise("operand", Var.layout))

      val {get = jumpInfo: Jump.t -> Info.t, set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("jump info", Jump.layout))

      val {get = contInfo: Jump.t ->  {
				       size: int,
				       liveOffsets: int list
				       },
	   set = setContInfo} =
	 Property.new(Jump.plist, Property.initRaise("cont info", Jump.layout))

      val {get = handlerInfo: Jump.t ->  {
					  size: int,
					  liveOffsets: int list
					  },
	   set = setHandlerInfo} =
	 Property.new(Jump.plist, Property.initRaise("handler info", Jump.layout))

      val {get = funcInfo, set = setFuncInfo} =
	 Property.new(Func.plist, Property.initRaise("func info", Func.layout))

      val {get = liveArray, set = setLiveArray} =
	 Property.new(Var.plist, Property.initRaise("liveArray", Var.layout))

      fun allocateFunc(name, args, body) = 
	 let
	    val {liveBegin, liveBeginNoFormals, liveArray = liveArrayVars,
		 destroy = destroyLive} =
	       Live.live{exp = body,
			 shouldConsider = shouldAllocate,
			 jumpHandlers = jumpHandlers}

(*  	    val liveBegin =
 *  	       Trace.trace("liveBegin", Jump.layout, List.layout Var.layout)
 *  	       liveBegin
 *)

	    (* Are handlers ever installed in this function? *)
	    val hasHandler: bool ref = ref false
	    val handlerOffset: int option ref = ref NONE
	       
	    (* register requests *)
	    val regRequests: RegRequest.t list ref = ref []

	    fun newRegRequest(x: Var.t, c: Chunk.t): RegRequest.t =
	       let val r = RegRequest.T{vars = ref [x],
					chunk = c,
					valid = ref true,
					regRef = ref NONE}
	       in ListRef.push(regRequests, r)
		  ; r
	       end
	    
	    local
	       val {get: Var.t -> Request.t option Dset.t, destroy, ...} =
		  Property.newDest
		  (Var.plist, Property.initFun(fn _ => Dset.singleton NONE))
	    in
	       val varSet = get
	       val varRequest = valOf o Dset.value o get
	       val destroyVarRequest = destroy
	    end

	    (* Unify the requests for all case tests and branches.
	     * This implements a form of copy propagation so that there is no
	     * need to move from the test to the branch formal.
	     *)
	    val _ =
	       let
		  val {get = jumpArgs, set = setJumpArgs, destroy} =
		     Property.newDest(Jump.plist,
				      Property.initRaise("args", Jump.layout))
		  fun loopExp e =
		     let val {decs, transfer} = Exp.dest e
		     in List.foreach(decs,
				     fn Dec.Fun{name, args, body} =>
				          (setJumpArgs(name, args)
					   ; loopExp body)
				     | _ => ())
			; (case transfer of
			      Ctransfer.Case{test, cases, ...} =>
				 if isGlobal test
				    then ()
				 else
				    let val s = varSet test
				    in List.foreach
				       (cases, fn (_, j) =>
					case jumpArgs j of
					   [(x, _)] => Dset.union(s, varSet x)
					 | _ => ())
				    end
			    | _ => ())
		     end
	       in loopExp body
		  ; destroy()
	       end
	 
	    fun processRegRequests() =
	       let
		  val {get = hasBeenReset, set = setHasBeenReset, destroy} =
		     Property.newDest(Chunk.plist, Property.initConst false)
	       in List.foreach
		  (!regRequests, fn RegRequest.T{vars, regRef, chunk, valid, ...} =>
		   if !valid
		      then 
			 let
			    val _ =
			       if hasBeenReset chunk
				  then ()
			       else (setHasBeenReset(chunk, true)
				     ; Chunk.resetRegisterCounters(chunk))
			    val vars = !vars
			    val reg = Chunk.newRegister(chunk, varType(hd vars))
			    val oper = Operand.register reg
			 in regRef := SOME reg
			    ; List.foreach(vars, fn x => setVarOperand(x, oper))
			 end
		   else ())
	       end

	    val returnOffset = ref NONE

	    (* stack slot requests *)
	    local
	       val typeRequests = Mtype.memo(fn _ => ref [])
	    in
	       fun slotRequest(xs: Var.t list): SlotRequest.t =
		  let val ty = varType(hd xs)
		     val r = SlotRequest.T{vars = ref xs, offsetRef = ref NONE}
		  in ListRef.push(typeRequests ty, r)
		     ; r
		  end

	       fun processSlotRequests(): unit =
		  let
		     val offset = 4
		     val offset =
			if !hasHandler
			   then
			      (handlerOffset := SOME offset
			       ; (offset + handlerSize))
			else offset
		     val offset =
			List.fold
			(let open Mtype
			 in [char, double, int, pointer, uint]
			 end,
			    offset,
			    fn (ty, offset) =>
			    let
			       val size = Mtype.size ty
			       val offset = 
				  List.fold
				  (!(typeRequests ty), offset,
				   fn (SlotRequest.T{vars, offsetRef, ...},
				       offset) =>
				   let
				      val oper =
					 Operand.stackOffset{offset = offset,
							     ty = ty}
				   in List.foreach
				      (!vars, fn x =>setVarOperand(x, oper))
				      ; offsetRef := SOME offset
				      ; size + offset
				   end)
			    in if Mtype.equals(ty, Mtype.char)
				  then Mtype.align(offset, Mtype.int)
			       else offset
			    end)
		  in returnOffset := SOME offset
		  end
	    end

	    fun getMask(xs: Cps.Var.t list) =
	       List.fold
	       (xs, {regs = [], offsets = []}, fn (x, ros as {regs, offsets}) =>
		if Mtype.isPointer(varType x)
		   then let
			   val reg =
			      fn RegRequest.T{valid = ref true,
					      regRef = ref(SOME r), ...} =>
			      {regs = r :: regs, offsets = offsets}
					    | _ => Error.bug "strange RegRequest"
			   val slot =
			      fn SlotRequest.T{offsetRef = ref(SOME n), ...} =>
			      {regs = regs, offsets = n :: offsets}
			       | _ => Error.bug "strange SlotRequest"
			in case varRequest x of
			   LocalReg r => reg r
			 | LocalSlot s => slot s
			 | FormalReg r => reg r
			 | FormalRegSlot(_, s) => slot s
			end
		else ros)

	    val c = funcChunk name

	 in List.foreach
	    (args, fn (x, _) =>
	     let val s = varSet x
	     in case Dset.value s of
		NONE => Dset.setValue(s, SOME(FormalReg(newRegRequest(x, c))))
	      | SOME _ => Error.bug "var request already set"
	     end)
	    ;
	    (*
	     * Walk over the program, changing some regRequests to slotRequests.
	     *)
	    let
	       fun newLocal(x, c) =
		  let val s = varSet x
		  in case Dset.value s of
		     NONE => Dset.setValue(s, SOME(LocalReg(newRegRequest(x, c))))
		   | SOME r => Request.addVar(r, x)
		  end
	       fun forceStack l =
		  List.foreach
		  (liveBegin l, fn x =>
		   let val s = varSet x
		      fun change r = Dset.setValue(s, SOME r)
		   in case valOf(Dset.value s) of
		      LocalReg(RegRequest.T{valid, vars, ...}) =>
			 (valid := false
			  ; change(LocalSlot(slotRequest(!vars))))
		    | FormalReg r =>
			 change(FormalRegSlot(r, slotRequest(RegRequest.vars r)))
		    | _ => ()
		   end)

	       fun loopExp(e: Exp.t, c: Chunk.t): unit =
		  let val {decs, transfer} = Exp.dest e
		  in List.foreach
		     (decs,
		      fn Dec.Bind{var, ...} => newLocal(var, c)
		       | Dec.Fun{name, args, body} =>
			    let val c = jumpChunk name
			    in List.foreach(args, fn (x, _) => newLocal(x, c))
			       ; loopExp(body, c)
			    end
		       | Dec.HandlerPush h => (forceStack h; hasHandler := true)
		       | _ => ())
		     ; (case transfer of
			   Ctransfer.Call{cont, ...} =>
			      Option.map' forceStack cont
			 | _ => ())
		  end

	    in loopExp(body, c)
	    end
	 ;
	    
	(* Must process slotRequests after regRequests so that the varOperand
	 * for toplevel formals are the slot operands, not the reg ones.
	 * There is a patch in the varOperand for the toplevel function
	 * that does the right thing (i.e. returns the reg).
	 *)
      processRegRequests()
      ; processSlotRequests()
      ;
      let
	 val (argSlots, prelude, liveRegs) =
	    List.fold
	    (args, ([], [], []), fn ((x, _), (argSlots, prelude, liveRegs)) =>
	     let
		fun addLive' r =
		   if Mtype.isPointer(Register.ty r)
		      then r :: liveRegs
		   else liveRegs
		fun addLive(RegRequest.T{regRef, ...}) =
		   case !regRef of
		      SOME r => addLive' r
		    | NONE => Error.bug "invalid RegRequest"
	     in case varRequest x of
		FormalRegSlot(RegRequest.T{regRef, ...}, _) =>
		   (case !regRef of
		       SOME r =>
			  let val ro = Operand.register r
			  in ((x, ro) :: argSlots,
			      Statement.move{dst = varOperand x, src = ro}
			      :: prelude,
			      addLive' r)
			  end
		     | NONE => Error.bug "invalid RegRequest")
	      | FormalReg rr => (argSlots, prelude, addLive rr)
	      | _ => Error.bug "strange varRequest"
	     end)

	 val returnOffset = (case !returnOffset of
				NONE => Error.bug "return offset not set"
			      | SOME n => n)

	 val _ =
	    let
	       val varOperand =
		  fn x =>
		  case List.peek(argSlots, fn (x', _) => Var.equals(x, x')) of
		     NONE => varOperand x
		   | SOME(_, ro) => ro
	    in setFuncInfo
	       (name,
		{handlerOffset = !handlerOffset,
		 stackFrameSize = returnOffset,
		 info = Info.T{
			       varOperand = varOperand,
			       liveBegin = {regs = liveRegs, offsets = []},
			       liveArray = liveArray,
			       prelude = prelude,
			       postlude = []
			       }})
	    end

	 val handleTransfer =
	    fn Ctransfer.Call{cont = SOME c, ...} =>
	         (case getMask(liveBeginNoFormals c) of
		     {regs = [], offsets} =>
			setContInfo(c, {size = returnOffset,
					liveOffsets = offsets})
		   | _ => Error.bug "cont has live reg")
	     | _ => ()

	 val handleDec =
	    fn Dec.Fun{name, ...} =>
	        setJumpInfo
		(name,
		 Info.T{
			varOperand = varOperand,
			liveBegin = getMask(liveBegin name),
			liveArray = liveArray,
			prelude = [],
			postlude = []
			})
	     | Dec.HandlerPush h =>
		 (* Unlike Conts, it can happen that a handler that is never
		  * called has live regs, since handlers are only called by
		  * stack frames above them, and if there is no stack frame
		  * above me, my lives may not get moved into slots.  In this
		  * case it is OK to ignore the regs because we know the handler
		  * is never called.
		  *)
		 setHandlerInfo
		 (h, {size = valOf(!handlerOffset),
		      liveOffsets = #offsets(getMask(liveBeginNoFormals h))})
	    | Dec.Bind{var, exp = PrimExp.PrimApp{prim, ...}, ...} =>
		 if Prim.isArrayAllocation prim
		    then setLiveArray(var, getMask(liveArrayVars var))
		 else ()
	    | _ => ()


      in (* Set the jumpInfo for each local func *)
	 Exp.foreach'(body, {handleTransfer = handleTransfer,
			     handleDec = handleDec})
      end
      ; destroyVarRequest()
      ; destroyLive()
	 end

   in
      List.foreach(functions, fn {name, args, body, ...} =>
		   allocateFunc(name, args, body))
      ; destroyGlobal()
      ; {jumpInfo = jumpInfo,
	 handlerInfo = handlerInfo,
	 contInfo = contInfo,
	 funcInfo = funcInfo}
   end
   
end
