(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * In the Cps input program
 *   Const and ConApp{args = [], ...} must all be global
 *   so that the register allocator doesn't allocate a register or stack slot
 *   for them.
 *
 *)
functor Backend(S: BACKEND_STRUCTS): BACKEND = 
struct

open S

local open Cps
in structure Con = Con
   structure Const = Const
   structure Cdec = Dec
   structure Cexp = Exp
   structure Func = Func
   structure Function = Function
   structure Cprogram = Program
   structure Ctype = Type
   structure Jump = Jump
   structure Prim = Prim
   structure PrimExp = PrimExp
   structure Tycon = Tycon
   structure Ctransfer = Transfer
   structure Var = Var
end 

local open Machine
in
   structure Chunk = Chunk
   structure GCInfo = GCInfo
   structure Label = Label
   structure Mtype = Type
   structure Mprogram = Program
   structure Operand = Operand
   structure Register = Register
   structure Statement = Statement
   structure Mtransfer = Transfer
end

structure Chunkify = Chunkify(open Cps)

structure ParallelMove = ParallelMove()

structure LimitCheck = LimitCheck(open Cps)
   
structure Representation = Representation(structure Cps = Cps
					  structure Mtype = Mtype)

structure AllocateRegisters = AllocateRegisters(structure Cps = Cps
						structure Machine = Machine)
local open AllocateRegisters
in structure Info = Info
end

local open Representation
in structure TyconRep = TyconRep
   structure ConRep = ConRep
end

fun generate(program as Cprogram.T{datatypes, globals, functions, main})
   : Mprogram.t =
   let
      val insertLimitCheck = LimitCheck.limitCheck program
	 
       val {tyconRep, conRep, toMtype} =
	  Control.traceCall "compute representations"
	  Representation.compute program

       val _ =
	  Control.displays
	  ("rep", fn display =>
	   List.foreach
	   (datatypes, fn {tycon, cons} =>
	    let open Layout
	    in display(seq[Tycon.layout tycon,
			   str " ",
			   TyconRep.layout(tyconRep tycon)]);
	       display(indent(align(List.map(cons, fn {con, ...} =>
					     seq[Con.layout con,
						 str " ",
						 ConRep.layout(conRep con)])),
			      2))
	    end))

       fun toMtypes ts = List.map(ts, toMtype)

       val wordSize = 4
       val tagOffset = 0
       val tagType = Mtype.int

       val jumpHandlers = Cps.inferHandlers program

       val chunks =
	  Control.traceCall "chunkify" Chunkify.chunkify
	  {program = program,
	   jumpHandlers = jumpHandlers}

       val _ = 
	  Control.displays
	  ("chunks", fn display =>
	   List.foreach
	   (chunks, fn {funcs, jumps} =>
	    let open Layout
	    in display
	       (record([("funcs", List.layout Func.layout funcs),
			("jumps", List.layout Jump.layout jumps)]))
	    end))

       val {get = jumpInfo: Jump.t -> {chunk: Chunk.t,
				       formals: Operand.t list option ref},
	    set = setJumpInfo} =
	  Property.new(Jump.plist, Property.initRaise("jump info", Jump.layout))

       val jumpChunk = #chunk o jumpInfo

       val {get = funcInfo: Func.t -> {chunk: Chunk.t,
				       interChunk: Label.t,
				       params: {formals: Operand.t list,
						globalFormals: Operand.t list
						} option ref
				       },
	    set = setFuncInfo} =
	  Property.new(Func.plist, Property.initRaise("func info", Func.layout))

       val funcChunk = #chunk o funcInfo
 
       val mprogram = Mprogram.new()

       (* Lazily build global operands for an arbitrary number of instances of
	* each type.
	*)
       local
	  val typeGlobals: Mtype.t -> Operand.t list ref =
	     Mtype.memo(fn _ => ref [])
       in
	  fun typesGlobals(ts: Mtype.t list): Operand.t list =
	     let val index = Mtype.memo(fn _ => ref 0)
	     in List.foldr
		(ts, [], fn (t, accum) =>
		 let val gs = typeGlobals t
		    val rec loop =
		       fn (0, []) => let
					val g =
					   if Mtype.isPointer t
					      then
						 Mprogram.newGlobalPointerNonRoot
						 mprogram
					   else Mprogram.newGlobal(mprogram, t)
				     in gs := !gs @ [g];
					g
				     end
			| (0, g :: _) => g
			| (i, _ :: gs) => loop(i - 1, gs)
			| _ => Error.bug "typesGlobals"
		    val r = index t
		    val i = !r
		 in r := i + 1;
		    loop(i, !gs) :: accum
		 end)
	     end
       end

       fun typeGlobal(t: Mtype.t): Operand.t =
	  case typesGlobals[t] of
	     [p] => p
	   | _ => Error.bug "typeGlobal"

       val {get = jumpCont: Jump.t -> {label: Label.t,
				       formals: Operand.t list option ref,
				       generated: bool ref} option,
	    set = setJumpCont} =
	  Property.new(Jump.plist, Property.initConst NONE)

       val {get = jumpHandler: Jump.t -> {label: Label.t,
					  formals: Operand.t list option ref,
					  generated: bool ref} option,
	    set = setJumpHandler} =
	  Property.new(Jump.plist, Property.initConst NONE)

       fun getJumpHandler j =
	  case jumpHandler j of
	     NONE => Error.bug "getJumpHandler"
	   | SOME{label = h, ...} => h

       (* Create info for jumps used as conts. *)
       val conts =
	  let
	     val conts = ref []
	     local
		fun make (get, set) j =
		   case get j of
		      NONE =>
			 let val l = Label.new(jumpToLabel j)
			 in ListRef.push(conts, (l, j))
			    ; set(j, SOME{label = l,
					  formals = ref NONE,
					  generated = ref false})
			 end
		    | SOME _ => ()
	     in val newHandler = make(jumpHandler, setJumpHandler)
		val newCont = make(jumpCont, setJumpCont)
	     end
	     val handleDec =
		fn Cdec.HandlerPush h => newHandler h
		 | _ => ()
	     val handleTransfer =
		fn Ctransfer.Call{cont, ...} => Option.map'(cont, newCont)
		 | _ => ()
	  in List.foreach
	     (functions, fn {body, ...} =>
	      Cexp.foreach'(body, {handleDec = handleDec,
				   handleTransfer = handleTransfer}))
	     ; (List.foreach
		(functions, fn {body, ...} =>
		 Cexp.foreachDec
		 (body,
		  fn Cdec.Fun{name, args, ...} =>
		        let
			   val formals =
			      Promise.lazy
			      (fn () => typesGlobals(List.map(args, toMtype o #2)))
			in List.foreach
			   ([jumpCont, jumpHandler], fn f =>
			    case f name of
			       SOME{formals = r, ...} => r := SOME(formals())
			     | _ => ())
			end
		   | _ => ())))
	     ; !conts
	  end

       val machineChunks = ref []
	  
       val _ =
	  ((* Create the mprogram chunks. *)
	   List.foreach
	   (chunks, fn {funcs, jumps} =>
	    let val interChunkEntries = List.map(funcs, Label.new o funcToLabel)
	       val conts =
		  List.fold(jumps, [], fn (j, cs) =>
			    let
			       val cs = 
				  case jumpHandler j of
				     NONE => cs
				   | SOME{label = c, ...} => c :: cs
			    in case jumpCont j of
			       NONE => cs
			     | SOME{label = c, ...} => c :: cs
			    end)
	       val c = Mprogram.newChunk{program = mprogram,
					 entries = interChunkEntries @ conts}
	    in ListRef.push(machineChunks, c)
	       ; List.foreach2(funcs, interChunkEntries,
			     fn (f, l) =>
			     setFuncInfo(f, {chunk = c,
					     interChunk = l,
					     params = ref NONE}))
	       ; List.foreach(jumps, fn j =>
			      setJumpInfo(j, {chunk = c,
					      formals = ref NONE}))
	    end)
	  )

       val {get = varInfo: Var.t -> {ty: Ctype.t,
				     operand: Operand.t option ref,
				     isConstant: bool ref},
	    set = setVarInfo} =
	  Property.new(Var.plist, Property.initRaise("info", Var.layout))

       val varIsConstant = ! o #isConstant o varInfo
       val varType = #ty o varInfo
       val varOperand = ! o #operand o varInfo
	  
(*        val varOperand =
 * 	  Trace.trace("varOperand", Var.layout, Option.layout Operand.layout)
 * 	  varOperand
 *)

       val setVarOperand =
	  fn (x, p) =>
	  let val {operand, ...} = varInfo x
	  in case !operand of
	     NONE => operand := SOME p
	   | SOME _ => Error.bug("setVarOperand twice of " ^ Var.toString x)
	  end

(*        val setVarOperand = Trace.trace2("setVarOperand", Var.layout,
 * 					Operand.layout, Unit.layout) setVarOperand
 *)

       val _ =
	  let
	     val set = fn (x, t) => setVarInfo(x, {ty = t,
						   operand = ref NONE,
						   isConstant = ref false})
	     fun sets xts = List.foreach(xts, set)
	  in
	     List.foreach(globals, fn {var, ty, ...} => set(var, ty));
	     let
		fun loopExp e = List.foreach(#decs(Cexp.dest e), loopDec)
		and loopDec d =
		   case d of
		      Cdec.Bind{var, ty, ...} => set(var, ty)
		    | Cdec.Fun{args, body, ...} => (sets args; loopExp body)
		    | _ => ()
	     in List.foreach(functions, fn {name, args, body, ...} =>
			     (sets args; loopExp body))
	     end
	  end

       fun varTypes xs = List.map(xs, varType)
       val varMtype = toMtype o varType
       fun varMtypes xs = List.map(xs, varMtype)
	  
       fun sortTypes(initialOffset: int,
		     tys: Mtype.t list): {size: int,
					  offsets: int list,
					  numWordsNonPointers: int,
					  numPointers: int} =
	  let
	     val voids = ref []
	     val bytes = ref []
	     val doubleWords = ref []
	     val words = ref []
	     val pointers = ref []
	     val _ = List.foreachi(tys, fn (i, t) =>
				   ListRef.push
				   (if Mtype.isPointer t
				       then pointers
				    else (case Mtype.size t of
					     0 => voids
					   | 1 => bytes
					   | 4 => words
					   | 8 => doubleWords
					   | _ => Error.bug "strange size"),
				    (i, t)))
	     fun build(r, accum) =
		List.fold(!r, accum, fn ((index, ty), (res, offset)) =>
			  ({index = index, offset = offset, ty = ty} :: res,
			   offset + Mtype.size ty))
	     val (accum, offset) =
		build
		(voids,
		 build(bytes,
		       build(words,
			     build(doubleWords, ([], initialOffset)))))
	     val offset = Mtype.align(offset, Mtype.pointer)
	     val numWordsNonPointers = (offset - initialOffset) div wordSize
	     val (accum, size) = build(pointers, (accum, offset))
	     val components = List.reverse accum
	     fun loop(i, accum) =
		if i = ~1
		   then accum
		else loop(i - 1,
			  #offset(List.lookup(components, fn {index, ...} =>
					      i = index))
			  :: accum)
	     val offsets = loop(List.length components - 1, [])
	  in {size = size,
	      offsets = offsets,
	      numWordsNonPointers = numWordsNonPointers,
	      numPointers = List.length(!pointers)}
	  end

       (* Compute layout for each con and associate it with the con.
	*)
       local
	  val {get, set} =
	     Property.new(Con.plist, Property.initRaise("con info", Con.layout))
       in
	  val _ =
	     List.foreach
	     (datatypes, fn {cons, ...} =>
	      List.foreach(cons, fn {con, args} =>
			   let fun doit n = set(con, sortTypes(n, toMtypes args))
			   in case conRep con of
			      ConRep.Tuple => doit 0
			    | ConRep.TagTuple _ => doit 4
			    | _ => ()
			   end))

	  val conInfo = get

	  fun conOffset(c: Con.t, n: int): Operand.t =
	     Operand.int(List.nth(#offsets(get c), n))
       end

       (* Compute layout for each tuple type.
	*)
       local
	  val {get, set} =
	     Property.new
	     (Ctype.plist,
	      Property.initFun(fn t => sortTypes(0, toMtypes(Ctype.detuple t))))
       in
	  val tupleInfo = get
	  fun tupleOffset(t: Ctype.t, n: int): Operand.t =
	     Operand.int(List.nth(#offsets(get t), n))
       end

       fun parallelMove{srcs, dsts, chunk} =
	  let
	     val moves =
		List.map2(srcs, dsts, fn (src, dst) => {src = src, dst = dst})
	     fun temp r =
		Operand.register(Chunk.tempRegister(chunk, Operand.ty r))
	     val temp =
		Trace.trace("temp", Operand.layout, Operand.layout) temp
	  in
	     Trace.trace
	     ("parallelMove",
	      fn {moves, ...} =>
	      List.layout (fn {src, dst} =>
			   Layout.tuple
			   [Operand.layout src, Operand.layout dst])
	      moves,
	      Layout.ignore)
	     ParallelMove.move{
			       moves = moves,
			       equals = Operand.equals,
			       move = Statement.move,
			       temp = temp
			      }
	  end

       fun jumpFormals(j: Jump.t): Operand.t list =
	  case jumpInfo j of
	     {formals = ref(SOME os), ...} => os
	   | _ => Error.bug "jumpFormals"

       fun tail(to: Jump.t, args: Operand.t list): Statement.t list * Mtransfer.t =
	  (parallelMove{srcs = args,
			dsts = jumpFormals to,
			chunk = jumpChunk to},
	   Mtransfer.nearJump to)
	     
       fun genCase{chunk: Chunk.t,
		   profileName: string,
		   varOperand,
		   testVar: Var.t, testRep: TyconRep.t,
		   cases, default} =
	  let
	     val test = varOperand testVar
	     (* Creating this new block is OK because all it does is
	      * a few moves and then a transfer.  I.E. it does no allocations
	      * and can not trigger a gc.
	      *)
	     fun newBlock(statements, transfer): Label.t =
		let val l' = Label.newString "L"
		in Chunk.newBlock(chunk,
				  {profileName = profileName,
				   label = l',
				   gcInfo = NONE,
				   statements = statements,
				   transfer = transfer});
		   l'
		end
	     fun enum(test: Operand.t): Mtransfer.t =
		let
		   val cases =
		      List.fold(cases, [], fn ((c, l), cases) =>
				let fun keep n = (n, l) :: cases
				in case conRep c of
				   ConRep.Int n => keep n
				 | ConRep.IntCast n => keep n
				 | _ => cases
				end)
		in Mtransfer.switch{test = test, cases = cases, default = default}
		end
	     fun switchIP(pointer: Label.t): Mtransfer.t =
		Mtransfer.switchIP{test = test,
				   int = newBlock([], enum(Operand.castInt test)),
				   pointer = pointer}
	     fun enumAndOne(): Mtransfer.t =
		switchIP
		(newBlock
		 (case List.peek(cases, fn (c, _) =>
				 case conRep c of
				    ConRep.Transparent _ => true
				  | ConRep.Tuple => true
				  | _ => false)
		     of
			NONE => 
			   (case default of
			       NONE => Error.bug "enumAndOne: no default"
			     | SOME l => ([], Mtransfer.nearJump l))
		      | SOME(_, l) =>
			   (case jumpFormals l of
			       [] => []
			     | [z] =>
				  Statement.moves{srcs = [test], dsts = [z]}
			     | _ => Error.bug "case label must be unary or nullary",
			    Mtransfer.nearJump l)))
	     fun indirectTag(): Mtransfer.t =
		Mtransfer.switch
		{test = Operand.offset{base = test,
				       offset = Operand.int tagOffset,
				       ty = tagType},
		 default = default,
		 cases =
		 List.fold
		 (cases, [], fn ((c, l'), cases) =>
		  case conRep c of
		     ConRep.TagTuple n =>
			let 
			   val l =
			      case jumpFormals l' of
				 [] => l'
			       | [z] => newBlock([Statement.move{dst = z,
								 src = test}],
						 Mtransfer.nearJump l')
			       | _ => Error.bug "case label wrong number args"
			in (n, l) :: cases
			end 
		   | _ => cases)}

	  in case testRep of
	     TyconRep.Prim mtype =>
		(case (cases, default) of
		    ([(_, l')], _) =>
		       (* We use _ instead of NONE for the default becuase
			* there may be an unreachable default case
			*)
		       tail(l', (case jumpFormals l' of
				    [] => []
				  | _ => [test]))
		  | ([], SOME l') => tail(l', [])
		  | _ => Error.bug "prim datatype with more than one case")
	   | TyconRep.Enum => ([], enum test)
	   | TyconRep.EnumDirect => ([], enumAndOne())
	   | TyconRep.EnumIndirect => ([], enumAndOne())
	   | TyconRep.EnumIndirectTag =>
		([], switchIP(newBlock([], indirectTag())))
	   | TyconRep.IndirectTag => ([], indirectTag())
	  end

       fun genConstBind{var, ty, exp} =
	  let
	     fun set(oper: Operand.t): unit =
		let val {operand, isConstant, ...} = varInfo var
		in operand := SOME oper;
		   isConstant := true
		end
	     fun global(new, s) = set(new(mprogram, s))
	     fun bogus() = set Operand.void
	     fun nonExpansive() =
		if Mtype.isVoid(toMtype ty)
		   then bogus()
		else ()
	  in case exp of
	     PrimExp.Var x =>
		(case varOperand x of
		    NONE => ()
		  | SOME oper => set oper)
	   | PrimExp.Const c => 
		(case Const.node c of
		    Const.Int n =>
		       set
		       (Operand.int32
			(let val t = Const.tycon c
			 in if Tycon.equals(t, Tycon.int)
			       then n
			    else if Tycon.equals(t, Tycon.int31)
				    then n * 2
				 else Error.bug "strange Int"
			 end))
		  | Const.Word w =>
		       set
		       (let val t = Const.tycon c
			in if Tycon.equals(t, Tycon.word)
			      then Operand.uint w
			   else if Tycon.equals(t, Tycon.word8)
				   then Operand.char(Char.chr(Word32.toInt w))
				else if Tycon.equals(t, Tycon.word31)
					then Operand.uint(Word32.<<(w, 0w1))
				     else Error.bug "strange word"
			end)
		  | Const.Real f => set(Operand.float f)
		  | Const.Char c => set(Operand.char c)
		  | Const.String s => global(Mprogram.newString, s)
		  | Const.SmallIntInf i => set(Operand.intInf i)
		  | Const.IntInf s => global(Mprogram.newIntInf, s))
	   | PrimExp.Tuple [] => bogus()
	   | PrimExp.Tuple _ => ()
	   | PrimExp.ConApp{con, args} =>
		(case conRep con of
		    ConRep.Void => bogus()
		  | ConRep.Int n => set(Operand.int n)
		  | ConRep.IntCast n => set(Operand.pointer n)
		  | _ => ())
	   | PrimExp.ConSelect _ => nonExpansive()
	   | PrimExp.Select _ => nonExpansive()
	   | PrimExp.PrimApp{prim, targs, args} =>
		(case Prim.numArgs prim of
		    NONE => set(Operand.prim(prim, toMtype ty))
		  | _ => ())
	  end

(*        val genConstBind =
 * 	  Trace.trace("genConstBind", Var.layout o #var, Unit.layout)
 * 	  genConstBind
 *)

       (* Set the operands for constants.
	* This has to happen before register allocation so that RA doesn't
	* allocate registers for them.
	*)
       val _ =
	  (List.foreach(globals, genConstBind);
	   List.foreach
	   (globals, fn {var, ty, exp} => 
	    case varOperand var of
	       NONE =>
		  setVarOperand
		  (var,
		   case exp of
		      PrimExp.Var x =>
			 (case varOperand x of
			     NONE => Error.bug "global missing operand"
			   | SOME oper => oper)
		    | _ =>  Mprogram.newGlobal(mprogram, toMtype ty))
	     | SOME _ => ());
	   List.foreach(functions, fn {body, ...} =>
			Cexp.foreachBind(body, genConstBind)))

       val shouldAllocate = not o Option.isSome o varOperand
	  
       val {arrayInfo, contInfo, funcInfo = funcRegInfo, handlerInfo,
	    jumpInfo = jumpRegInfo} =
	  Control.pass
	  {name = "allocate registers",
	   suffix = "reg",
	   style = Control.No,
	   thunk = fn () =>
	   AllocateRegisters.allocate
	   {
	    program = program,
	    varType = varMtype,
	    jumpChunk = jumpChunk,
	    funcChunk = funcChunk,
	    shouldAllocate = shouldAllocate,
	    jumpHandlers = jumpHandlers
	    },
	   display =
	   Control.Layouts
	   (fn ({funcInfo, jumpInfo, ...}, layout) =>
	    let
	       val constantVarOperand = varOperand
		  
	       open Layout
	       fun layoutVar varOperand x =
		  layout(seq[Var.layout x, str " ",
			     Operand.layout(case constantVarOperand x of
					       NONE => varOperand x
					     | SOME z => z)])
		  
	       fun loopBind varOperand {var, ty, exp} = layoutVar varOperand var

	       fun loop(Info.T{varOperand, ...}, args, body) =
		  let val layoutVar = layoutVar varOperand
		     val loopBind = loopBind varOperand
		  in List.foreach(args, layoutVar o #1);
		     List.foreach
		     (#decs(Cexp.dest body),
		      fn Cdec.Bind b => loopBind b
		       | Cdec.Fun{name, args, body} =>
			    loop(jumpInfo name, args, body)
		       | _ => ())
		  end		     
	    in List.foreach(globals, loopBind(valOf o varOperand));
	       List.foreach
	       (functions, fn {name, args, body, ...} =>
		let val {info, handlerOffset, ...} = funcInfo name
		in layout(seq[str "function ", Func.layout name,
			      str " handlerOffset ",
			      Option.layout Int.layout handlerOffset])
		   ; loop(#info(funcInfo name), args, body)
		end)
	    end)}

       (* Set the params part of the label info for all labels.
	* This has to happen after register allocation, because it needs to
	* know the operands for formal parameters.
	*)
       val _ =
	  let
	     fun loopExp e = List.foreach(#decs(Cexp.dest e), loopDec)
	     and loopDec d =
		case d of
		   Cdec.Fun{name, args, body} =>
		      (let val Info.T{varOperand, ...} = jumpRegInfo name
			  val {formals, ...} = jumpInfo name
		       in formals := SOME(List.map(args, varOperand o #1))
		       end
		       ; loopExp body)
		 | _ => ()
	  in List.foreach
	     (functions, fn {name, args, body, ...} =>
	      (let val {chunk, interChunk, params} = funcInfo name
		   val {info = Info.T{varOperand, ...}, ...} =
		      funcRegInfo name
		   val formals = List.map(args, varOperand o #1)
		   val globalFormals =
		      typesGlobals(List.map(args, toMtype o #2))
	       in Chunk.newBlock
		  (chunk, {profileName = Func.toString name,
			   label = interChunk,
			   gcInfo = NONE,
			   statements =
			   Statement.moves{srcs = globalFormals,
					   dsts = formals},
			   transfer = Mtransfer.nearJump(funcToLabel name)})
		  ; params := SOME{
				   formals = formals,
				   globalFormals = globalFormals
				   }
	       end
	       ; loopExp body))
	  end

       (* ---------------------------------- *)
       (*             genPrimExp             *)
       (* ---------------------------------- *)

       fun genPrimExp(x: Var.t, ty: Ctype.t, e: PrimExp.t,
		      varOperand): Statement.t list =
	  if varIsConstant x
	     then []
	  else
	     let
		fun varOperands xs = List.map(xs, varOperand)

		fun move src =
		   [Statement.move{dst = varOperand x, src = src}]

		fun makeStores(ys, offsets) =
		   List.fold2(ys, offsets, [], fn (y, offset, stores) =>
			      if Mtype.isVoid(Operand.ty y)
				 then stores
			      else {offset = offset, value = y} :: stores)

		fun allocate(ys, {size, offsets,
				  numPointers, numWordsNonPointers}) =
		   [Statement.allocate
		    {dst = varOperand x,
		     size = size,
		     numPointers = numPointers,
		     numWordsNonPointers = numWordsNonPointers,
		     stores = makeStores(ys, offsets)}]

		fun allocateTagged(n: int, ys: Operand.t list,
				   {size, offsets,
				    numPointers, numWordsNonPointers}) =
		   [Statement.allocate
		    {dst = varOperand x,
		     size = size,
		     numPointers = numPointers,
		     numWordsNonPointers =
		     (* for the tag *) 1 + numWordsNonPointers,
		     stores = ({offset = tagOffset, value = Operand.int n}
			       :: makeStores(ys, offsets))}]


	     in case e of
		PrimExp.Var y => move(varOperand y)
	      | PrimExp.Tuple ys => allocate(varOperands ys, tupleInfo ty)
	      | PrimExp.Select{tuple, offset} =>
		   if Mtype.isVoid(Operand.ty(varOperand x))
		      then []
		   else 		
		      move(Operand.offset
			   {base = varOperand tuple,
			    offset = tupleOffset(varType tuple, offset),
			    ty = toMtype ty})
	      | PrimExp.ConSelect{variant, con, offset} =>
		   if Mtype.isVoid(Operand.ty(varOperand x))
		      then []
		   else let val variant = varOperand variant
			    fun doit() =
			       move(Operand.offset
				    {base = variant,
				     offset = conOffset(con, offset),
				     ty = toMtype ty})
			in case conRep con of
			   ConRep.Transparent _ => move variant
			 | ConRep.Tuple => doit()
			 | ConRep.TagTuple _ => doit()
			 | _ => Error.bug "strange conRep for ConSelect"
			end
	      | PrimExp.ConApp{con, args} =>
		   let val args = varOperands args
		      fun tuple() = allocate(args, conInfo con)
		   in case (conRep con, args) of
		      (ConRep.Transparent _, [y]) => move y
		    | (ConRep.Tuple, _) => tuple()
		    | (ConRep.TagTuple n, _) => allocateTagged(n, args, conInfo con)
		    | _ => Error.bug "strange ConApp"
		   end
	      | PrimExp.PrimApp{prim, targs, args} =>
		   let
		      fun unsafeSub(a, i, ty) =
			 let val ty = toMtype ty
			 in move(Operand.scaledOffset{base = a,
						      offset = i,
						      scale = Mtype.size ty,
						      ty = ty})
			 end
		      fun array(n: Operand.t, t: Mtype.t): Statement.t list =
			 let
			    val (nbnp, np) =
			       if Mtype.isPointer t
				  then (0, 1)
			       else (Mtype.size t, 0)
			 in [Statement.allocateArray{dst = varOperand x,
						     numElts = n,
						     numBytesNonPointers = nbnp,
						     numPointers = np,
						     gcInfo = arrayInfo x}]
			 end
		      val argOps = varOperands args
		      open Prim.Name
		   in case (Prim.name prim, targs, argOps) of
		      (VectorFromArray, _, [src]) =>
			 [Statement.move{dst = varOperand x, src = src}]
		    | (Array, [t], [n]) => array(n, toMtype t)
		    | (ArraySub, [t], [a, i]) => unsafeSub(a, i, t)
		    | (StringSub, [], [s, i]) => unsafeSub(s, i, Ctype.char)
		    | (VectorSub, [t], [v, i]) => unsafeSub(v, i, t)
		    | (ArrayUpdate, [t], [a, i, y]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ =>
			       [Statement.move
				{dst = Operand.scaledOffset{base = a,
							    offset = i,
							    scale = Mtype.size t,
							    ty = t},
				 src = y}]
			 end
		    | (Ref, [t], [y]) =>
			 let val t = toMtype t
			    val (ys, ts) = if Mtype.isVoid t
					      then ([], [])
					   else ([y], [t])
			 in allocate(ys, sortTypes(0, ts))
			 end
		    | (Deref, [t], [y]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ => [Statement.move{dst = varOperand x,
						 src = Operand.contents(y, t)}]
			 end
		    | (Assign, [t], [y, z]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ => [Statement.move{dst = Operand.contents(y, t),
						 src = z}]
			 end
		    | _ =>
			 (case Mtype.dest(toMtype ty) of
			     Mtype.Void => [Statement.sideEffect(prim, argOps)]
			   | _ => [Statement.assign{dst = varOperand x,
						    oper = prim,
						    args = argOps}])
		   end
	      | _ => Error.bug "genPrimExp saw strange primExp"
	     end

(*        val genPrimExp =
 * 	  Trace.trace("genPrimExp",
 * 		      fn (x, t, e, _, _) => Layout.tuple[Var.layout x,
 * 							 Ctype.layout t,
 * 							 PrimExp.layout e],
 * 		      Layout.ignore)
 * 	  genPrimExp
 *)

       local
	  fun make (labelGen, info) (j: Jump.t,
				     profileName: string,
				     f: Label.t -> unit): Label.t =
	     case labelGen j of
		NONE => Error.bug "missing labelGen"
	      | SOME{label = c, formals, generated} =>
		   (if !generated
		       then ()
		    else (generated := true
			  ; f c
			  ; let val {size, liveOffsets} = info j
			       val (statements, transfer) =
				  tail(j, valOf(!formals))
			       (* restore stack pointer *)
			       val statements = Statement.pop size :: statements
			    in Chunk.newBlock(jumpChunk j,
					      {profileName = profileName,
					       label = c,
					       gcInfo = NONE,
					       statements = statements,
					       transfer = transfer})
			    end)
		    ; c)
       in
	  val genCont = make(jumpCont, contInfo)
	  val genHandler = make(jumpHandler, handlerInfo)
       end

       (* ---------------------------------- *)
       (*            genTransfer             *)
       (* ---------------------------------- *)

       fun genTransfer(chunk: Chunk.t,
		       profileName: string,
		       t: Ctransfer.t,
		       handlers: Jump.t list,
		       varOperand): Statement.t list * Mtransfer.t =
	  case t of
	     Ctransfer.Jump{dst, args} => tail(dst, List.map(args, varOperand))
	   | Ctransfer.Call{func, args, cont} =>
		let val args = List.map(args, varOperand)
		   val {chunk = chunk', params, interChunk} = funcInfo func
		   val {formals, globalFormals} = valOf(!params)
		   val (setupArgs, transfer) =
		      if Chunk.equals(chunk, chunk')
			 then (parallelMove{srcs = args,
					    dsts = formals,
					    chunk = chunk},
			       Mtransfer.nearJump(funcToLabel func))
		      else (Statement.moves{srcs = args, dsts = globalFormals},
			    Mtransfer.farJump(chunk', interChunk))
		in case cont of
		   NONE => (setupArgs, transfer)
		 | SOME c =>
		      let val {size, liveOffsets} = contInfo c
			 val return =
			    genCont(c, profileName, fn l =>
				    Mprogram.newFrame
				    (mprogram, {return = l,
						chunk = jumpChunk c,
						size = size,
						liveOffsets = liveOffsets}))
		      in 
			 (setupArgs
			  @ [Statement.push size,
			     Statement.move
			     {dst = Operand.stackOffset{offset = 0,
							ty = Mtype.int},
			      src = Operand.label return}],
			  transfer)
		      end
		end
	   | Ctransfer.Return xs =>
		(Statement.moves{srcs = List.map(xs, varOperand),
				 dsts = typesGlobals(varMtypes xs)},
		 Mtransfer.return)
	   | Ctransfer.Case{test, cases, default, ...} =>
		(case (cases, default) of
		    ([], NONE) => ([], Mtransfer.bug)
		  | _ => 
		       case Ctype.tyconArgs(varType test) of
			  (tycon, []) =>
			     genCase{chunk = chunk,
				     profileName = profileName,
				     varOperand = varOperand,
				     testRep = tyconRep tycon,
				     testVar = test,
				     cases = cases,
				     default = default}
			| _ => Error.bug "strange type in case")
	   | Ctransfer.Raise xs =>
		let val xops = List.map(xs, varOperand)
		in case handlers of
		   [] =>
		      (* The stack move must occur after the operand moves
		       * because some of the operands may be on the stack.
		       *)
		      (Statement.moves{dsts = typesGlobals(varMtypes xs),
				       srcs = xops},
		       Mtransfer.raisee)
		  | h :: _ => tail(h, xops)
		end
	   | Ctransfer.Halt => ([], Mtransfer.halt(Operand.int 0))
	   | Ctransfer.Bug => ([], Mtransfer.bug)

       (*------------------------------------*)
       (*               genExp               *)
       (*------------------------------------*)

       fun genExp{exp = e: Cexp.t,
		  profileName: string,
		  insertLimitCheck: bool,
		  label: Label.t,
		  chunk: Chunk.t,
		  info = Info.T{gcInfo, prelude, postlude, varOperand = vo},
		  handlerOffset: int option,
		  handlers: Jump.t list,
		  prelude = prelude': Statement.t list}: unit =
	  let val {decs, transfer} = Cexp.dest e
	     val varOperand = (fn x =>
			       case varOperand x of
				  SOME p => p
				| NONE => vo x)
	     val (decs, handlers) =
		genDecs(decs, profileName, varOperand, handlerOffset, handlers)
	     val (preTransfer, transfer) =
		genTransfer(chunk, profileName, transfer, handlers, varOperand)
	     val statements =
		prelude' @ prelude @ decs @ postlude @ preTransfer
	  in
	     Chunk.newBlock(chunk,
			    {profileName = profileName,
			     label = label,
			     gcInfo = if insertLimitCheck
					 then SOME gcInfo
				      else NONE,
			     statements = statements,
			     transfer = transfer})
	  end

       and genDecs(ds: Cdec.t list,
		   profileName: string,
		   varOperand,
		   handlerOffset,
		   handlers: Jump.t list)
	  : Statement.t list * Jump.t list =
	  let
	     val (statements, handlers) =
		List.fold
		(ds, ([], handlers), fn (d , (statements, handlers)) =>
		 (case d of
		     Cdec.Bind{var, ty, exp} =>
			genPrimExp(var, ty, exp, varOperand) :: statements
		  | Cdec.Fun{name, body, ...} =>
		       (genExp{exp = body,
			       profileName = profileName,
			       insertLimitCheck = insertLimitCheck name, 
			       label = jumpToLabel name,
			       chunk = jumpChunk name,
			       info = jumpRegInfo name,
			       handlerOffset = handlerOffset,
			       handlers = jumpHandlers name,
			       prelude = []}
			; statements)
		  | Cdec.HandlerPush h =>
		       let
			  val c =
			     genHandler
			     (h, profileName, fn c =>
			      Mprogram.newHandler(mprogram, c, jumpChunk h))
			  val offset = valOf handlerOffset
			  val statements =
			     [Statement.move
			      {dst = Operand.stackOffset{offset = offset,
							 ty = Mtype.label},
			       src = Operand.label c}]
			     :: statements
		       in case handlers of
			  [] => ([Statement.saveExnStack{offset = offset}]
				 :: statements)
			| _ => statements
		       end
		  | Cdec.HandlerPop =>
		       let val offset = valOf handlerOffset
		       in case handlers of
			  [] => Error.bug "pop of empty handler stack"
			| _ :: handlers => 
			     (case handlers of
				 [] =>
				    [Statement.restoreExnStack{offset = offset}]
			       | h :: _ =>
				    [Statement.move
				     {dst =
				      Operand.stackOffset{offset = offset,
							  ty = Mtype.label},
				      src = Operand.label(getJumpHandler h)}])
			     :: statements
		       end,
		 Cps.deltaHandlers(d, handlers)))
	  in (List.fold(statements, [], op @), handlers)
	  end

   in
      if List.exists(functions, fn {name, ...} =>
		     Func.equals(name, Func.saveWorld))
	 then
	    let val {chunk, interChunk, ...} = funcInfo Func.saveWorld
	    in Mprogram.setSaveWorld(mprogram, chunk, interChunk)
	    end
      else ()

      (* Build the initGlobals chunk. *)
      ; let
	   val statements =
	      List.fold
	      (List.fold
	       (globals, [], fn ({var, ty, exp}, statements) =>
		(genPrimExp(var, ty, exp, valOf o varOperand)
		 :: statements)),
	       [], op @)
	   val initGlobals = Label.newString "initGlobals"
	   val chunk = Mprogram.newChunk{program = mprogram,
					 entries = [initGlobals]}
	   val {chunk = mainChunk, interChunk = main, ...} = funcInfo main
	in Mprogram.setMain(mprogram, chunk, initGlobals)
	   ; (Chunk.newBlock
	      (chunk, {profileName = "initGlobals",
		       label = initGlobals,
		       gcInfo =
		       SOME(GCInfo.T{regs = [], offsets = [],
				     frameSize = Mtype.size Mtype.label}),
		       statements = statements,
		       transfer = Mtransfer.farJump(mainChunk, main)}))
	end

      ; (List.foreach
	 (functions, fn {name, body, ...} =>
	  let val {info, handlerOffset, ...} = funcRegInfo name
	  in genExp{exp = body,
		    profileName = Func.toString name,
		    insertLimitCheck = true,
		    label = funcToLabel name,
		    chunk = funcChunk name,
		    info = info,
		    handlerOffset = handlerOffset,
		    handlers = [],
		    prelude = [Statement.stackOverflowCheck]}
	  end))
      ; Cprogram.clear program

      ; mprogram
   end
 
end
