(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Only compute lives where limit checks will be inserted:
 *  - function entry
 *  - continuation entry
 *  - array allocation
 *  - loop entry
 *
 * Use recursion for loops.
 *)
functor Live(S: LIVE_STRUCTS): LIVE = 
struct

open S
open Cps
open Dec PrimExp Transfer

structure VarStatus =
   struct
      datatype t = Unseen | Free | Bound

      val toString =
	 fn Unseen => "Unseen"
	  | Free => "Free"
	  | Bound => "Bound"

      val layout = Layout.str o toString
   end
datatype status = datatype VarStatus.t

structure JumpStatus =
   struct
      datatype t = Normal | InBody | Recursive

      val toString =
	 fn Normal => "Normal"
	  | InBody => "InBody"
	  | Recursive => "Recursive"

      val layout = Layout.str o toString
   end
datatype status = datatype JumpStatus.t

fun live{exp, shouldConsider, jumpHandlers} =
   let
      val {get = jumpInfo: Jump.t -> {frees: Var.t list ref,
				      freesNoFormals: Var.t list ref,
				      isCont: bool ref,
				      status: JumpStatus.t ref},
	   set = setJumpInfo,
	   destroy = destroyJumpInfo} =
	 Property.newDest
	 (Jump.plist, Property.initFun(fn _ => {frees = ref [],
						freesNoFormals = ref [],
						isCont = ref false,
						status = ref InBody}))

      fun getFreesNoFormals j = !(#freesNoFormals(jumpInfo j))

      val {get = liveArray, set = setLiveArray,
	   destroy = destroyLiveArray} =
	 Property.newDest(Var.plist, Property.initRaise("liveArray", Var.layout))

      val {get = varInfo: Var.t -> {status: VarStatus.t ref,
				    frees: Var.t list ref ref(* used as a unique id *)},
	   destroy = destroyVarInfo, ...} =
	 Property.newDest(Var.plist,
			  Property.initFun(let val bogus = ref []
					   in fn _ => {status = ref Unseen,
						       frees = ref bogus}
					   end))

      fun newScope(frees, new, th) =
	 let
	    val changed = ref false
	    val all = ref []
	    fun statusRef x =
	       let val {frees = frees', status} = varInfo x
	       in if frees = !frees'
		     then ()
		  else (ListRef.push
			(all, (frees', !frees', status, !status))
			; frees' := frees; status := Unseen)
		  ; status
	       end
	    fun bindVar x =
	       if shouldConsider x
		  then statusRef x := Bound
	       else ()
	    fun var x =
	       if shouldConsider x
		  then let val s = statusRef x
		       in case !s of
			  Unseen =>
			     (Assert.expensive
			      (fn () => not(List.exists(!frees, fn y =>
							Var.equals(x, y))))
			      ; changed := true
			      ; s := Free
			      ; new x
			      ; ListRef.push(frees, x))
			| _ => ()
		       end
	       else ()
	 in List.foreach(!frees, fn x => statusRef x := Free)
	    ; th(var, bindVar, changed)
	    ; List.foreach(!all, fn (r, v, r', v') => (r := v; r' := v'))
	 end
	 
      fun loopExp(e: Exp.t, handlers: Jump.t list, var, bindVar) =
	 let
	    val {decs, transfer} = Exp.dest e
	    fun transHandlers() = List.fold(decs, handlers, deltaHandlers)
	    fun loopDecs(decs, var, bindVar) =
	       let fun vars xs = List.foreach(xs, var)
		  fun jump j =
		     let val {status, freesNoFormals, ...} = jumpInfo j
		     in vars(!freesNoFormals)
			; (case !status of
			      InBody => status := Recursive
			    | _ => ())
		     end
	       in case decs of
		  [] => (case transfer of
			    Jump{dst, args, ...} => (vars args; jump dst)
			  | Call{args, cont, ...} =>
			       (vars args; Option.map'(cont, jump))
			  | Return xs => vars xs
			  | Raise xs =>
			       (vars xs
				; (case transHandlers() of
				      h :: _ => jump h
				    | _ => ()))
			  | Case{test, cases, default, ...} =>
			       (var test
				; Option.map'(default, jump)
				; List.foreach(cases, jump o #2))
			  | Bug => ()
			  | Halt => ())
		| dec :: decs =>
		     let fun continue() = loopDecs(decs, var, bindVar)
		     in case dec of
			HandlerPush _ => continue()
		      | HandlerPop => continue()
		      | Bind{var = x, exp, ...} =>
			   let fun normal() = (bindVar x; continue())
			   in case exp of
			      Const _ => normal()
			    | Var y => (var y; normal())
			    | Tuple xs => (vars xs; normal())
			    | Select{tuple, ...} => (var tuple; normal())
			    | ConSelect{variant, ...} => (var variant; normal())
			    | ConApp{args, ...} => (vars args; normal())
			    | PrimApp{prim, args, ...} =>
				 (vars args
				  ; (case Prim.name prim of
					Prim.Name.Array =>
					   let val frees = ref []
					      fun new _ = ()
					   in newScope
					      (frees, new, fn (var, bindVar, _) =>
					       (bindVar x
						; loopDecs(decs, var, bindVar)))
					      ; List.foreach(!frees, var)
					      ; setLiveArray(x, !frees)
					   end
				      | _ => normal()))
			   end
		      | Fun{name, args, body} =>
			   (let
			       val {frees, freesNoFormals, status, isCont, ...} =
				  jumpInfo name
			       val handlers = jumpHandlers name
			       val {get, set, destroy} =
				  Property.newDest
				  (Var.plist, Property.initConst false)
			       fun new x =
				  if get x then ()
				  else ListRef.push(freesNoFormals, x)
			    in List.foreach(args, fn (x, _) => set(x, true))
			       ; (newScope
				  (frees, new, fn (var, bindVar, changed) =>
				   (if !isCont
				       then (* in case there is a raise to here *)
					  (case handlers of
					      h :: _ =>
						 List.foreach
						 (getFreesNoFormals h, var)
					    | _ => ())
				    else ()
				    ; let
					 fun loop() =
					    (changed := false
					     ; loopExp(body, handlers, var, bindVar)
					     ; if !status = Recursive andalso !changed
						  then loop()
					       else ())
				      in status := InBody
					 ; loop()
					 ; status := Normal
				      end)))
			       ; destroy()
			    end
			    ; continue())
		     end
	       end
	 in loopDecs(decs, var, bindVar)
	 end
      fun ignore _ = ()
      val handleTransfer =
	 fn Call{cont, ...} => (case cont of
				   NONE => ()
				 | SOME c => #isCont(jumpInfo c) := true)
	  | _ => ()
   in Exp.foreachTransfer(exp, handleTransfer)
      ; loopExp(exp, [], ignore, ignore)
      ; {
	 liveBegin = ! o #frees o jumpInfo,
	 liveBeginNoFormals = getFreesNoFormals,
	 liveArray = liveArray,
	 destroy = fn () => (destroyJumpInfo()
			     ; destroyLiveArray()
			     ; destroyVarInfo())
	 }
   end

end
