(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Machine(S: MACHINE_STRUCTS): MACHINE = 
struct

open S

val wordSize = 4
val pointerSize = wordSize
val objectHeaderSize = wordSize
val arrayHeaderSize = 2 * wordSize
val intInfOverhead = arrayHeaderSize + wordSize (* for the sign *)

val overhead = "**C overhead**"

structure Label =
   struct
      open Label

      fun index l = toString l ^ "_index"
   end

structure Type = Mtype()

structure Type =
   struct
      open Type

      fun name t =
	 case dest t of
	    Char => "char"
	  | Double => "double"
	  | Int => "int"
	  | Uint => "uint"
	  | Pointer => "pointer"
	  | Void => "void"

      fun globalName ty = "global" ^ name ty
   end

structure C =
   struct
      val truee = "TRUE"
      val falsee = "FALSE"
	 
      fun call(f: string, xs: string list, print: string -> unit): unit =
	 (print f
	  ; print "("
	  ; (case xs of
		[] => ()
	      | x :: xs => (print x
			    ; List.foreach(xs, fn x => (print ","; print x))))
	  ; print ")")

      fun int32(n: Int32.int): string =
	 if n >= 0
	    then Int32.toString n
	 else if n = valOf Int32.minInt
		 then "(int)0x80000000" (* because of goofy gcc warning *)
	      else "-" ^ String.dropPrefix(Int32.toString n, 1)
      (* This overflows on Int32.minInt: Int32.toString(~ n) *)

      val int = int32 o Int32.fromInt

      fun char(c: char) =
	 concat[if Char.ord c >= 0x80 then "(uchar)" else "",
		"'", Char.toCString c, "'"]

      fun word(w: Word32.word) = "0x" ^ Word32.toString w

      (* the only difference between SML floats and C floats is that
       * SML uses "~" while C uses "-"
       *)
      val float = String.translate(fn #"~" => "-" | c => String.fromChar c)

      fun assign(lhs, rhs, print) =
	 (print lhs
	  ; print " = "
	  ; print rhs
	  ; print ";\n")

      fun string(s, print) =
	 let val quote = "\""
	 in print quote
	    ; print(String.toCString s)
	    ; print quote
	 end

      fun bug(s: string, print) =
	 (call("MLTON_bug",
	       [concat["\"", String.toCString s, "\""]],
	       print)
	  ; print ";\n")

      fun goto l = concat["goto ", l, ";"]
      fun gotoLabel l = goto(Label.toString l)

      local
	 val current = ref ""
      in
	 fun profile(detailed: string, nonDetailed: string,
		     print: string -> unit): unit =
	    if !Control.profile
	       then
		  if detailed <> !current
		     then
			(print "/* PROFILE: "
			 ; print detailed
			 ; print " & "
			 ; print nonDetailed
			 ; print " */\n"
			 ; current := detailed)
		  else ()
	    else ()
      end
   end

structure Register =
   struct
      datatype t =
	 T of {
	       (* register number.
	        * For pointer regs, this is also an index into an array of
		* locals used within the chunk to communicate with th GC.
		*)
	       index: int,
	       ty: Type.t
	      }

      fun ty(T{ty, ...}) = ty
      fun index(T{index, ...}) = index
	 
      fun equals(r1, r2) =
	 Type.equals(ty r1, ty r2)
	 andalso index r1 = index r2

      fun toString(T{index, ty}) = concat[Type.name ty, C.int index]

      val layout = Layout.str o toString
   end

structure Global =
   struct
      datatype t = T of {index: int, ty: Type.t}

      fun ty(T{ty, ...}) = ty
      fun index(T{index, ...}) = index

      fun toString(T{index, ty}) =
	 concat[Type.globalName ty, "[", C.int index, "]"]
   end

structure Operand =
   struct
      (* Be careful to update Operand.equals if you add new variants to this
       * datatype.
       *)
      datatype t =
	 Void
       | Int of int
       | Int32 of Int32.int
       | IntInf of SmallIntInf.t
       | Prim of string * Type.t
       | Pointer of int
       | Uint of Word32.word
       | Char of char
       | Float of string
       | Label of Label.t
       | CastInt of t
       | Register of Register.t
       | Global of Global.t
       | GlobalPointerNonRoot of int
       | Offset of {base: t, offset: t, scale: int, ty: Type.t}
       | StackOffset of {offset: int, ty: Type.t}
       | Contents of {oper: t, ty: Type.t}

      val rec toString =
	 fn Void => "void"
	  | Int n => C.int n
	  | Int32 n => C.int32 n
	  | IntInf s => concat["(pointer)", SmallIntInf.toCstring s]
	  | Prim(p, _) => p
	  | Pointer n => concat["(pointer)", C.int n]
	  | Uint w => C.word w
	  | Char c => C.char c
	  | Float s => C.float s
	  | Label l => Label.index l
	  | CastInt oper => concat["(int)", toString oper]
	  | Register r => Register.toString r
	  | Global g => Global.toString g
	  | GlobalPointerNonRoot n =>
	       concat["globalpointerNonRoot[", C.int n, "]"]
	  | Offset{base, offset, scale, ty} =>
	       concat["*(", Type.toString ty, "*)(",
		      toString base, " + ",
		      toString offset, " * ", C.int scale, ")"]
	  | StackOffset{offset, ty} =>
	       concat["*(", Type.toString ty, "*)(stackTop + ",
		      C.int offset, ")"]
	  | Contents{oper, ty} =>
	       concat["*(", Type.toString ty, "*)(", toString oper, ")"]

      val layout = Layout.str o toString

      val void = Void
      val int = Int
      val int32 = Int32
      val intInf = IntInf
      val pointer = Pointer
      val uint = Uint
      val char = Char
      val float = Float
      val label = Label
      val castInt = CastInt
      val register = Register
      val global = Global
      val stackOffset = StackOffset

      val deRegister =
	 fn Register r => SOME r
	  | _ => NONE

      val deStackOffset =
	 fn StackOffset s => SOME s
	  | _ => NONE

      fun bogus t =
	 case Type.dest t of
	    Type.Char => char #"\000"
	  | Type.Double => float "0.0"
	  | Type.Int => int 0
	  | Type.Uint => uint 0w0
	  | Type.Pointer => pointer 1
	  | _ => Error.bug "bogus not implemented for type"

      fun prim(p, t) =
	 case Prim.name p of
	    Prim.Name.CPrim s => Prim(s, t)
	  | Prim.Name.Bogus => bogus t
	  | _ => Error.bug("strange Prim operand "
			   ^ Layout.toString(Prim.layout p))

      fun offset{base, offset, ty} =
	 Offset{base = base, offset = offset, ty = ty, scale = 1}

      val scaledOffset = Offset
	 
      fun contents(z, t) = Contents{oper = z, ty = t}

      val ty =
	 fn Void => Type.void
	  | Int _ => Type.int
	  | Int32 _ => Type.int
	  | IntInf _ => Type.pointer
	  | Prim(_, ty) => ty
	  | Uint _ => Type.uint
	  | Char _ => Type.char
	  | Float _ => Type.double
	  | Pointer _ => Type.pointer
	  | CastInt _ => Type.int
	  | Label _ => Type.label
	  | Register r => Register.ty r
	  | Global g => Global.ty g
	  | GlobalPointerNonRoot _ => Type.pointer
	  | Offset{ty, ...} => ty
	  | Contents{ty, ...} => ty
	  | StackOffset{ty, ...} => ty

      val rec equals =
	 fn (Int n, Int n') => n = n'
	  | (Int32 n, Int32 n') => n = n'
	  | (IntInf n, IntInf n') => SmallIntInf.equals(n, n')
	  | (Pointer n, Pointer n') => n = n'
	  | (Uint w, Uint w') => w = w'
	  | (Char c, Char c') => c = c'
	  | (Float f, Float f') => f = f'
	  | (CastInt z, CastInt z') => equals(z, z')
	  | (Register r, Register r') => Register.equals(r, r')
	  | (Offset{base = b, offset = i, scale = s, ...},
	     Offset{base = b', offset = i', scale = s', ...}) =>
	    equals(b, b') andalso equals(i, i') andalso s = s'
	  | (StackOffset{offset = n, ...}, StackOffset{offset = n', ...}) =>
	       n = n'
	  | (Contents{oper = z, ...}, Contents{oper = z', ...}) => equals(z, z')
	  | _ => false
   end

structure GCInfo =
   struct
      datatype t =
	 T of {
	       (* Size of frame, including return address. *)
	       frameSize: int,
	       (* Live pointer valued registers. *)
	       regs: Register.t list,
	       (* Live pointer valued stack offsets. *)
	       offsets: int list
	       }
   end

(* ------------------------------------------------- *)
(*                    LimitCheck                     *)
(* ------------------------------------------------- *)

structure LimitCheck =
   struct
      fun output({bytes: string,
		  localIndices: string,
		  frameSize: int,
		  return: Label.t},
		 print) =
	 (C.call("MLTON_limitCheck",
		 [bytes, localIndices, C.int frameSize, Label.toString return],
		 print)
	  ; print ";\n")
   end

fun primApp(p: Prim.t, args: Operand.t list, print): unit =
   let open Prim.Name
      fun doit s = C.call(s, List.map(args, Operand.toString), print)
   in case Prim.name p of
      CPrim s => (case Prim.numArgs p of
		     NONE => print s
		   | SOME _ => doit s)
    | ArrayLength => doit "MLTON_Array_length"
    | VectorLength => doit "MLTON_Vector_length"
    | Eq => (if Type.isVoid(Operand.ty(hd args))
		then print "TRUE"
	     else doit "MLTON_eq")
    | Size => let val ty = Operand.ty(hd args)
	      in case Type.dest ty of
		 Type.Pointer => doit "MLTON_MLton_size"
	       | _ => print "0"
	      end
    | SaveStack => doit "MLTON_MLton_saveStack"
    | SaveWorld => doit "MLTON_MLton_saveWorld"
    | _ => Error.bug("unexpected primitive in machine: "
		     ^ Layout.toString(Prim.layout p))
   end

structure Statement =
   struct
      datatype t =
	 Noop
       | Move of {
		  dst: Operand.t,
		  src: Operand.t
		 }
       | Push of int
       | Assign of {dst: Operand.t,
		    oper: Prim.t,
		    args: Operand.t list}
       | SideEffect of Prim.t * Operand.t list
       | SaveExnStack of {offset: int}
       | RestoreExnStack of {offset: int}
       | StackOverflowCheck
       | Allocate of {dst: Operand.t,
		      size: int,
		      numPointers: int,
		      numWordsNonPointers: int,
		      stores: {offset: int,
			       value: Operand.t} list}
       | AllocateArray of {
			   user: {
				  dst: Operand.t,
				  numElts: Operand.t,
				  numPointers: int,
				  numBytesNonPointers: int,
				  gcInfo: GCInfo.t
				  },
			   limitCheck: {return: Label.t,
					bytes: string,
					localIndices: string} option ref
			  }

      val stackOverflowCheck = StackOverflowCheck
	 
      fun output(s, print) =
	 case s of
	    Noop => ()
	  | Move{dst, src} => (print(Operand.toString dst)
			       ; print " = "
			       ; print(Operand.toString src)
			       ; print ";\n")
	  | Push n => (print "stackTop += "; print(C.int n); print ";\n")
	  | Assign{dst, oper, args} => (print(Operand.toString dst)
					; print " = "
					; primApp(oper, args, print)
					; print ";\n")
	  | SideEffect(prim, args) => (primApp(prim, args, print); print ";\n")
	  | StackOverflowCheck => print "MLTON_stackOverflowCheck\n"
	  | SaveExnStack{offset} =>
	       (C.call("MLTON_saveExnStack", [C.int offset], print)
		; print "\n")
	  | RestoreExnStack{offset} =>
	       (C.call("MLTON_restoreExnStack", [C.int offset], print)
		; print "\n")
	  | Allocate{dst, size, numPointers, numWordsNonPointers, stores} =>
	       (print "*(word*)frontier = "
		; C.call("GC_objectHeader",
			 [C.int numWordsNonPointers, C.int numPointers],
			 print)
		; print ";\n"
		; C.assign(Operand.toString dst,
			   concat["frontier + ", C.int objectHeaderSize],
			   print)
		; (List.foreach
		   (stores, fn {offset, value} =>
		    if Type.isVoid(Operand.ty value)
		       then ()
		    else (print "*("
			  ; print(Type.toString(Operand.ty value))
			  ; print "*)(frontier + "
			  ; print(C.int(offset + objectHeaderSize))
			  ; print ") = "
			  ; print(Operand.toString value)
			  ; print ";\n")))
		; print "frontier += "
		; print(C.int(size + objectHeaderSize))
		; print ";\n")
	| AllocateArray{user = {dst, numElts, numPointers,
				numBytesNonPointers,
				gcInfo = GCInfo.T{frameSize, ...},
				...},
			limitCheck} =>
	  (case !limitCheck of
	      NONE => ()
	    | SOME{bytes, localIndices, return} =>
		 LimitCheck.output({bytes = bytes,
				    localIndices = localIndices,
				    frameSize = frameSize,
				    return = return},
				   print)
	   ; let val dst = Operand.toString dst
		 val numElts = Operand.toString numElts
	     in if numPointers = 0
		   then (print "MLTON_allocArrayNoPointers("
			 ; print dst
			 ; print ","
			 ; print numElts
			 ; print ","
			 ; print(C.int numBytesNonPointers)
			 ; print ");\n")
		else if numBytesNonPointers = 0
			then (print "MLTON_allocArrayPointers("
			      ; print dst
			      ; print ","
			      ; print numElts
			      ; print ","
			      ; print (C.int numPointers)
			      ; print ");\n")
		     else Error.unimplemented "tricky arrays"
	     end)

      val push = Push
      val pop = Push o ~

      fun move(arg as {dst, src}) =
	 if Type.isVoid(Operand.ty dst)
	    orelse Operand.equals(dst, src)
	    then Noop
	 else Move arg
	    
      fun assign(arg as {dst, ...}) =
	 if Type.isVoid(Operand.ty dst)
	    then Noop
	 else Assign arg

      val saveExnStack = SaveExnStack
      val restoreExnStack = RestoreExnStack
      val sideEffect = SideEffect

      fun allocate(arg as {dst, size, ...}) =
	 Allocate(if size = 0
		     then {dst = dst,
			   size = wordSize (* min size *),
			   numPointers = 0,
			   numWordsNonPointers = 1,
			   stores = []}
		  else arg)

      fun allocateArray user =
	 AllocateArray{user = user, limitCheck = ref NONE}

      fun moves{srcs, dsts} =
	 List.map2(srcs, dsts, fn (src, dst) =>
		   move{src = src, dst = dst})
   end

structure Offsets =
   struct
      structure Ints =
	 HashSet(val cacheSize = 1
		 val bits = 14
		 structure Element =
		    struct
		       open Int
		       fun hash n = Word.fromInt n
		    end)

      fun new name = 
	 let val all = ref []
	    val c = Counter.new 0
	    val {get, ...} =
	       Property.new
	       (Ints.plist,
		Property.initFun
		(fn offsets =>
		 let val name = concat[name, C.int(Counter.next c)]
		 in ListRef.push(all, {name = name,
				       offsets = Ints.toList offsets});
		    name
		 end))

	    fun declare(print: string -> unit): unit =
	       List.foreach
	       (!all, fn {name, offsets} =>
		(print "static ushort "
		 ; print name
		 ; print "[] = {"
		 ; print(C.int(List.length offsets))
		 ; List.foreach(offsets, fn i => (print ","; print(C.int i)))
		 ; print "};\n"))

	 in {name = get o Ints.fromList,
	     declare = declare}
	 end
   end

(* ------------------------------------------------- *)
(*                       Types                       *)
(* ------------------------------------------------- *)

structure ChunkName:>
   sig
      type t
      val new: unit -> t
      val toString: t -> string
   end =
   struct
      datatype t = T of string

      fun toString(T s) = s

      val counter = Counter.new 0

      fun new() = T("Chunk" ^ Int.toString(Counter.next counter))
   end

structure Types =
   struct
      datatype transfer =
	 Halt of Operand.t
       | Bug
       | Return
       | Raise
       | Switch of {
		    test: Operand.t,
		    cases: (int * Label.t) list,
		    default: Label.t option
		    }
       | SwitchIP of {
		      test: Operand.t,
		      int: Label.t,
		      pointer: Label.t
		      }
       | NearJump of Label.t
       | FarJump of chunk * Label.t

      and block = Block of {
			    profileName: string,
			    label: Label.t,
			    bytesNeeded: int option ref,
			    limitCheck: {
					 gcInfo: GCInfo.t,
					 extra: {bytes: int,
						 localMask: string,
						 return: Label.t} option ref
					 } option,
			    statements: Statement.t array,
			    transfer: transfer
			    }

      and chunk = Chunk of {
			    name: ChunkName.t,

			    (* where to start *)
			    entries: Label.t list,
			    
			    blocks: block list ref,

			    (* for each type, gives the max # registers used *)
			    regMax: Type.t -> int ref,

			    limitCheckInfo: {numLocals: int,
					     localUsed: int option array,
					     gcReturns: Label.t list
					     } option ref
			    }
   end

structure Chunk =
   struct
      datatype t = datatype Types.chunk

      fun name(Chunk{name, ...}) = name

      val nameToString = ChunkName.toString o name
   end

structure Frames =
   struct
      type t = {return: Label.t,
		chunk: Chunk.t,
		size: int,
		liveOffsets: int list} list ref
	 
      val maxFrameSize = Int.^(2, 16)

      fun add(frames: t, f as {size, ...}) =
	 if size > maxFrameSize
	    then Error.bug(concat["MLton cannot handle stack frames larger than ",
				  Int.toString maxFrameSize,
				  " bytes."])
	 else ListRef.push(frames, f)
   end

(* ------------------------------------------------- *)
(*                     Transfer                      *)
(* ------------------------------------------------- *)

structure Transfer =
   struct
      type chunk = Types.chunk
      datatype t = datatype Types.transfer
	 
      fun output(t, print, gotoLabel: Label.t -> unit): unit =
	 let 
	    fun iff(test, a, b): unit =
	       (print "if ("
		; print test
		; print ") {\n"
		; gotoLabel a
		; print "} else {\n"
		; gotoLabel b
		; print "}\n")
	 in case t of
	    Halt x => (C.call("MLTON_halt", [Operand.toString x], print)
		       ; print "\n")
	  | Bug => C.bug("cps machine", print)
	  | Return => print "MLTON_return\n"
	  | Raise => print "MLTON_raise\n"
	  | Switch{test, cases, default} =>
	       let val test = Operand.toString test
		  fun bool(t, f) = iff(test, t, f)
		  fun switch(cases, l) =
		     (print "switch ("
		      ; print test
		      ; print ") {\n"
		      ; (List.foreach
			 (cases, fn (n, l) => (print "case "
					       ; print(C.int n)
					       ; print ":\n"
					       ; gotoLabel l)))
		      ; print "default:\n"
		      ; gotoLabel l
		      ; print "\n}\n")
	       in case (cases,   default) of
		  ([],               NONE)   => Error.bug "switch: empty cases"
		| ([(_, l)],         NONE)   => gotoLabel l
		| ([],               SOME l) => gotoLabel l
		| ([(0, f), (1, t)], NONE)   => bool(t, f)
		| ([(1, t), (0, f)], NONE)   => bool(t, f)
		| ((_, l) :: cases', NONE)   => switch(cases', l)
		| (_,                SOME l) => switch(cases, l)
	       end
	  | SwitchIP{test, int, pointer} =>
	       iff(concat["0x3 & (int)", Operand.toString test], int, pointer)
	  | NearJump l => gotoLabel l
	  | FarJump(c, l) =>
	       (C.call("MLTON_farJump",
		       [Chunk.nameToString c, Label.toString l],
		       print)
		; print";\n")
	 end

      val halt = Halt
      val bug = Bug
      val return = Return
      val raisee = Raise
      val switch = Switch
      val switchIP = SwitchIP
      val nearJump = NearJump
      val farJump = FarJump
   end

(* ------------------------------------------------- *)
(*                       Block                       *)
(* ------------------------------------------------- *)

structure Block =
   struct
      datatype t = datatype Types.block
   end

val nonVoidTypes =
   let open Type
   in [char, double, int, uint, pointer]
   end

(* ------------------------------------------------- *)
(*                       Chunk                       *)
(* ------------------------------------------------- *)

structure Chunk =
   struct
      open Chunk

      (* ------------------------------------------------- *)
      (*                 insertLimitChecks                 *)
      (* ------------------------------------------------- *)

      fun numRegsOfType(Chunk{regMax, ...}, ty: Type.t): int = !(regMax ty)

      fun numPointers(c) = numRegsOfType(c, Type.pointer)
					 
      fun insertLimitChecks(self as Chunk{blocks, limitCheckInfo, ...},
			    frames, localMaskName, newIndex): unit =
	 let
	    val gcReturns: Label.t list ref = ref []
	       
	    (* For each pointer register index, tells if the register is used
	     * in some limit check mask, in which case it needs to be copied
	     * to/from the locals array for the GC.  If it is used, localUsed
	     * contains SOME(the index in the locals array).
	     *)
	    val localUsed: int option array = Array.new(numPointers self, NONE)

	    val nextLocal = Counter.new(0)

	    fun localMask(rs: Register.t list): string =
	       localMaskName
	       (List.map
		(rs, fn Register.T{index = i, ty} =>
		 if Type.equals(Type.pointer, ty)
		    then (case Array.sub(localUsed, i) of
			     NONE => 
				let val j = Counter.next nextLocal
				in Array.update(localUsed, i, SOME j)
				   ; j
				end
			   | SOME j => j)
		 else Error.bug "non-pointer in register mask"))
	       
	    fun newFrame{size, liveOffsets} =
	       let val return = Label.newNoname()
	       in newIndex return
		  ; ListRef.push(gcReturns, return)
		  ; Frames.add(frames, {return = return,
					chunk = self,
					size = size,
					liveOffsets = liveOffsets})
		  ; return
	       end

	    val {get = labelBlock, set = setLabelBlock, destroy} =
	       Property.newDest(Label.plist,
				Property.initRaise("block", Label.layout))

	    open Block Transfer
	       
	    val _ = List.foreach(!blocks, fn b as Block{label, ...} =>
				 setLabelBlock(label, b))
	       
	    fun memo(Block{bytesNeeded, statements, transfer, limitCheck, ...})
	       : int =
	       case !bytesNeeded of
		  SOME n => n
		| NONE =>
		     let
			val _ = bytesNeeded := SOME 0

			val goto = memo o labelBlock
			   
			val rest =
			   case transfer of
			      NearJump l => goto l
			    | SwitchIP{int, pointer, ...} =>
				 Int.max(goto int, goto pointer)
			    | Switch{cases, default, ...} =>
				 List.fold
				 (cases, (case default of
					     NONE => 0
					   | SOME l => goto l),
				  fn ((_, j), rest) => Int.max(goto j, rest))
			    | _ => 0
				 
			(* put in the limit check info at array allocations *)
			val bytesAllocated =
			   Array.foldr
			   (statements, rest, fn (statement, bytesAllocated) =>
			    case statement of
			       Statement.AllocateArray
			       {user = {gcInfo = GCInfo.T{frameSize,
							  regs, offsets},
					numElts, numPointers, numBytesNonPointers,
					...},
				limitCheck = r, ...} =>
			       let
				  val bytesPerElt =
				     if numPointers = 0
					then numBytesNonPointers
				     else if numBytesNonPointers = 0
					     then numPointers * pointerSize
					  else Error.unimplemented "tricky arrays"
				  val bytesAllocated =
				     bytesAllocated
				     (* space for array header *)
				     + arrayHeaderSize
				     (* space for forwarding pointer for zero
				      * length arrays *)
				     + pointerSize
			       in case numElts of
				  Operand.Int32 numElts =>
				     bytesAllocated
				     + Type.align(Int32.toInt numElts
						  * bytesPerElt,
						  Type.pointer)
				| _ => 
				     let
					val bytes =
					   concat[Operand.toString numElts,
						  " * ",
						  C.int bytesPerElt,
						  " + ",
						  C.int bytesAllocated]
					val lc =
					   {bytes = bytes,
					    return =
					    newFrame{size = frameSize,
						     liveOffsets = offsets},
					    localIndices = localMask regs}
				     in r := SOME lc
					; 0
				     end
			       end
			      | Statement.Allocate{size, ...} =>
				   size + bytesAllocated + objectHeaderSize
			      | _ => bytesAllocated
				   )
			   
			val n =
			   case limitCheck of
			      NONE => bytesAllocated
			    | SOME{extra,
				   gcInfo = GCInfo.T{regs,
						     offsets,
						     frameSize}} =>
				 (if bytesAllocated = 0
				     then ()
				  else let
					  val return =
					     newFrame{size = frameSize,
						      liveOffsets = offsets}
				       in extra
					  := SOME{bytes = bytesAllocated,
						  localMask = localMask regs,
						  return = return}
				       end
				  ; 0)
		     in bytesNeeded := SOME n
			; n
		     end
	 in List.foreach(!blocks, fn b => (memo b; ()))
	    ; limitCheckInfo := SOME{numLocals = Counter.value nextLocal,
				     localUsed = localUsed,
				     gcReturns = !gcReturns}
	    ; destroy()
	 end

      (* ------------------------------------------------- *)
      (*                      output                       *)
      (* ------------------------------------------------- *)
	 
      fun output(self as Chunk{name, entries, blocks, regMax,
			       limitCheckInfo, ...},
		 print) =
	 let
	    val {numLocals, gcReturns, localUsed} = valOf(! limitCheckInfo)
		     
	    val {get = labelInfo: Label.t -> {block: Block.t,
					      numJumps: int ref,
					      layedOut: bool ref} option,
		 set = setLabelInfo, destroy} =
	       Property.newDest(Label.plist, Property.initConst NONE)
		    
	    fun jump l =
	       case labelInfo l of
		  NONE => () (* When would this happen? *)
		| SOME{numJumps, ...} => IntRef.inc numJumps

	    val _ =
	       let open Block Transfer
	       in List.foreach
		  (!blocks, fn b as Block{label, ...} =>
		   setLabelInfo(label, SOME{block = b,
					    numJumps = ref 0,
					    layedOut = ref false}))
		  ; List.foreach(entries, jump)
		  ; (List.foreach
		     (!blocks, fn Block{transfer, ...} =>
		      case transfer of
			 Switch{cases, default, ...} =>
			    (List.foreach(cases, jump o #2);
			     case default of
				NONE => ()
			      | SOME l => jump l)
		       | SwitchIP{int, pointer, ...} => (jump int; jump pointer)
		       | NearJump l => jump l
		       | _ => ()))
	       end

	    fun gotoLabel l =
	       case labelInfo l of
		  NONE => (print(C.gotoLabel l); print "\n")
	      | SOME{numJumps,
		     block =
		     Block.Block{profileName, limitCheck, statements, transfer,
				 ...},
		     layedOut} =>
		if !layedOut
		   then (print(C.gotoLabel l); print "\n")
		else
		   (layedOut := true
		    ; C.profile(profileName, profileName, print)
		    ; if !numJumps > 1
			 then (print(Label.toString l)
			       ; print ":\n")
		      else ()
		    ; (case limitCheck of
			  NONE => ()
			| SOME{extra,
			       gcInfo = GCInfo.T{frameSize, ...}} =>
			     (case !extra of
				 NONE => ()
			       | SOME{bytes, localMask, return} =>
				    LimitCheck.output
				    ({bytes = C.int bytes,
				      localIndices = localMask,
				      frameSize = frameSize,
				      return = return},
				     print)))
		    ; Array.foreach(statements, fn s =>
				    Statement.output(s, print))
		    ; Transfer.output(transfer, print, gotoLabel))
	    val numPointers = numPointers(self)
	 in
	    C.profile("MLTON_chunk (magic)", overhead, print)
	    ; C.call("MLTON_chunk", [ChunkName.toString name], print)
	    (* standard declarations *)
	    ; print "\npointer locals["
	    ; print(C.int numLocals)
	    ; print "];\n"
	    ; (List.foreach
	       (nonVoidTypes, fn ty =>
		let val n = Type.name ty
		      val l = Type.toString ty
		   in Int.foreach
		      (0, !(regMax ty) - 1, fn i =>
		       (print l
			; print " "
			; print n
			; print(C.int i)
			; print ";\n"))
		   end))
	    (* end declarations *)

	    (* build the entry switch *)
	    ; C.profile("MLTON_chunkSwitch (magic)", overhead, print)
	    ; print "MLTON_chunkSwitch\n"
	    ; List.foreach(entries, fn l =>
			   (C.profile("MLTON_chunkSwitch (magic)", overhead, print)
			    ; print "case "
			    ; print(Label.index l)
			    ; print ":\n"
			    ; gotoLabel l))
	    ; C.profile("MLTON_endChunkSwitch (magic)", overhead, print)
	    ; print "MLTON_endChunkSwitch\n"
		   
	    (* gc *)
	    ; let val pointer = Type.name Type.pointer
		  fun saveOrRestore f =
		     Int.foreach
		     (0, numPointers - 1, fn i =>
		      case Array.sub(localUsed, i) of
			 NONE => ()
		       | SOME j => 
			    let val i = C.int i
			       val r = concat[pointer, i]
			       val l = concat["locals[", C.int j, "]"]
			       val (lhs, rhs) = f(l, r)
			    in C.assign(lhs, rhs, print)
			    end)
	      in C.profile("MLTON_gc (magic)", overhead, print)
		 ; print "MLTON_beforeGC\n"
		 ; saveOrRestore(fn lr => lr)
		 ; print "MLTON_GC\n"
		 ; saveOrRestore(fn (l, r) => (r, l))
		 ; print "switch (*(uint*)stackTop) {\n"
		 ; List.foreach(gcReturns, fn l =>
				(print "case "
				 ; print(Label.index l)
				 ; print ": goto "
				 ; print(Label.toString l)
				 ; print ";\n"))
		 ; print "default: die(\"invalid gcReturn\");\n}\n"
	      end
		  
	    ; print "MLTON_endChunk\n"

	    ; destroy()
	 end

      fun equals(Chunk{blocks = r, ...}, Chunk{blocks = r', ...}) = r = r'
	 
      fun new(entries: Label.t list): t =
	 Chunk{name = ChunkName.new(),
	       entries = entries,
	       blocks = ref [],
	       regMax = Type.memo(fn _ => ref 0),
	       limitCheckInfo = ref NONE}

      fun register(Chunk{regMax, ...}, n, ty) =
	 let val r = regMax ty
	 in r := Int.max(!r, n + 1)
	    ; Register.T{index = n, ty = ty}
	 end
	 
      fun tempRegister(c as Chunk{regMax, ...}, ty) =
	 register(c, !(regMax ty), ty)

      fun newBlock(Chunk{blocks, ...},
		   {profileName, label, statements, transfer, gcInfo}) =
	 ListRef.push
	 (blocks,
	  Block.Block{profileName = profileName,
		      label = label,
		      statements = Array.fromList statements,
		      transfer = transfer,
		      bytesNeeded = ref NONE,
		      limitCheck =
		      Option.map(gcInfo, fn gcInfo =>
				 {gcInfo = gcInfo, extra = ref NONE})})
   end

(* ------------------------------------------------- *)
(*                      Program                      *)
(* ------------------------------------------------- *)

structure Program =
   struct
      datatype t = T of {
			 main: (Chunk.t * Label.t) option ref,
			 saveWorld: (Chunk.t * Label.t) option ref,
			 chunks: Chunk.t list ref,
			 globalCounter: Type.t -> Counter.t,
			 globalPointerNonRootCounter: Counter.t,
			 strings: (Global.t * string) list ref,
			 intInfs: (Global.t * string) list ref,
			 frames: Frames.t,
			 handlers: (Label.t * Chunk.t) list ref,
			 frameOffsetsName: int list -> string,
			 declareFrameOffsets: (string -> unit) -> unit
			 }

      fun setMain(T{main, ...}, c, l) = main := SOME(c, l)
      fun setSaveWorld(T{saveWorld, ...}, c, l) = saveWorld := SOME(c, l)

      fun new() =
	 let
	    val globalCounter = Type.memo(fn _ => Counter.new 0)

	    val {name = frameOffsetsName,
		 declare = declareFrameOffsets} = Offsets.new "frameOffsets"
	 in
	    T{
	      main = ref NONE,
	      saveWorld = ref NONE,
	      chunks = ref [],
	      globalCounter = globalCounter,
	      globalPointerNonRootCounter = Counter.new 0,
	      strings = ref [],
	      intInfs = ref [],
	      frames = ref [],
	      handlers = ref [],
	      frameOffsetsName = frameOffsetsName,
	      declareFrameOffsets = declareFrameOffsets
	     }
	 end

      fun newFrame(T{frames, ...}, f) =
	 Frames.add(frames, f)

      fun newHandler(T{handlers, ...}, l, c) = ListRef.push(handlers, (l, c))

      fun output{program =
		 T{main, saveWorld, chunks, globalCounter, strings, intInfs,
		   handlers, frames, frameOffsetsName,
		   globalPointerNonRootCounter,
		   declareFrameOffsets, ...},
		 includes: string list,
		 print: string -> unit}: unit =
	 let
	    val {get = labelIndex: Label.t -> int, set = setLabelIndex} =
	       Property.new(Label.plist,
			    Property.initRaise("index", Label.layout))

	    val labelCounter = Counter.new 0
	    fun newIndex(l: Label.t) =
	       setLabelIndex(l, Counter.next labelCounter)

	    val {name = localMaskName, declare = declareLocalIndices} =
	       Offsets.new "localIndices"

	    val _ =
	       List.foreach
	       (!chunks, fn c =>
		Chunk.insertLimitChecks(c, frames, localMaskName, newIndex))

	    val maxFrameSize =
	       List.fold(!frames, 0, fn ({size, ...}, n) => Int.max(n, size))

	    (* Assign the entries of a chunk consecutive integers so that gcc
	     * will use a jump table.
	     *)
	    val _ =
	       List.foreach
	       (!chunks, fn Chunk.Chunk{entries, ...} =>
		List.foreach(entries, newIndex))
			     
	    val maxFrameIndex = Counter.value labelCounter

	    local
	       val nextChunks = Array.new(maxFrameIndex, NONE)
	       val frameLayouts = Array.new(maxFrameIndex, NONE)
	       val _ =
		  List.foreach(!handlers, fn (l, c) =>
			       Array.update(nextChunks, labelIndex l,
					    SOME c))
	       val _ =
		  List.foreach
		  (!frames, fn {return, chunk, size, liveOffsets} =>
		   let val index = labelIndex return
		   in Array.update(nextChunks, index, SOME chunk)
		      ; Array.update(frameLayouts, index,
				     SOME{size = size,
					  offsets = frameOffsetsName liveOffsets})
		   end)
	       fun make(name: string, a: 'a array, elt: 'a -> unit) () : unit =
		  (print "static "
		   ; print name
		   ; print " = {"
		   ; Array.foreachi(a, fn (i, x) =>
				    (if i > 0 then print "," else ()
				     ; elt x))
		   ; print "};\n")

	       val bogusFrame = {size = 0, offsets = "NULL"}
	    in
	       val declareNextChunks =
		  make("void ( *nextChunks[])()",
		       nextChunks,
		       fn NONE => print "NULL"
			| SOME chunk =>
			     (print "&";
			      print(ChunkName.toString(Chunk.name chunk))))
		  
	       val declareFrameLayouts =
		  make("GC_frameLayout frameLayouts[]",
		       frameLayouts,
		       fn fo => let
				   val {size, offsets} =
				      case fo of
					 NONE => bogusFrame
				       | SOME f => f
				in print "{"
				   ; print(C.int size)
				   ; print ","
				   ; print offsets
				   ; print "}"
				end)
	    end

	    fun declareIndices() =
	       let
		  fun doit(ls: Label.t list): unit =
		     List.foreach(ls, fn l =>
				  let val index = labelIndex l
				  in print "#define "
				     ; print(Label.index l)
				     ; print " "
				     ; print(C.int index)
				     ; print "\n"
				  end)
	       in List.foreach
		  (!chunks, fn Chunk.Chunk{entries, limitCheckInfo, ...} =>
		   (doit entries;
		    doit(#gcReturns(valOf(!limitCheckInfo)))))
	       end
	    
	    fun declareGlobals() =
	       (print "static pointer globalpointerNonRoot["
		; print(C.int(Counter.value globalPointerNonRootCounter))
		; print "];\n"
		; (List.foreach
		   (nonVoidTypes, fn ty =>
		    let val n = Counter.value(globalCounter ty)
		    in if n = 0
			  then ()
		       else (print "static "
			     ; print(Layout.toString(Type.layout ty))
			     ; print " "
			     ; print(Type.globalName ty)
			     ; print "["
			     ; print(C.int n)
			     ; print "];\n")
		    end)))

	    local
	       fun make(name: string, readOrWrite: string)() =
		  (print "static void "
		   ; print name
		   ; print "(FILE *file) {"
		   ; (List.foreach
		      (nonVoidTypes, fn ty =>
		       let val n = Counter.value(globalCounter ty)
		       in if n = 0
			     then ()
			  else
			     let val name = Type.globalName ty
			     in C.call(readOrWrite,
				       [name,
					concat["sizeof(", Type.toString ty, ")"],
					C.int n,
					"file"],
				       print)
				; print ";\n"
			     end
		       end))
		   ; print "}\n")
	    in val declareSaveGlobals = make("saveGlobals", "swrite")
	       val declareLoadGlobals = make("loadGlobals", "sread")
	    end
	       
            fun declareIntInfs() =
	       (print "static struct intInfRes_t *intInfRes;\n"
		; print "static struct intInfInit intInfInits[] =\n"
		; print "{"
		; List.foreach(!intInfs, fn (g, s) =>
			       (print "{"
				; print(C.int(Global.index g))
				; print ","
				; C.string(s, print)
				; print "},"))
		; print "{ 0, NULL }};\n")

	    fun declareStrings() =
	       (print "static struct GC_stringInit stringInits[] =\n"
		; print "{"
		; List.foreach(!strings, fn (g, s) =>
			       (print "{"
				; print(C.int(Global.index g))
				; print ","
				; C.string(s, print)
				; print ","
				; print(C.int(String.size s))
				; print "},"))
		; print "{ 0, NULL, 0 }};\n")

	    val numGlobals = Counter.value(globalCounter Type.pointer)
	       
	    fun declareMain() =
	       let
		  val stringSizes =
		     List.fold
		     (!strings, 0, fn ((_, s), n)  =>
		      n + arrayHeaderSize
		      + Type.align(size s, Type.pointer))
		  val intInfSizes =
		     List.fold
		     (!intInfs, 0, fn ((_, s), n) =>
		      n + intInfOverhead
		      + Type.align(size s, Type.pointer))
		  val liveSize = intInfSizes + stringSizes
		  val (usedFixedHeap, fromSize) =
		     case !Control.fixedHeap of
			NONE => (false, 0)
		      | SOME n =>
			   (* div 2 for semispace *)
			   (if n > 0 andalso liveSize >= n div 2 
			       then Out.output(Out.error,
					       "Warning: heap size used with -h is too small to hold static data.\n")
			    else ()
			    ; (true, n))
		  val globals = if numGlobals = 0 then "NULL" else "globalpointer"
		  val magic = C.word(MLton.random())
		  val (mainChunk, mainLabel) =
		     case !main of
			NONE => Error.bug "main not set"
		      | SOME(c, l) => (Chunk.nameToString c, Label.toString l)
		  val (saveWorldChunk, saveWorldLabel) =
		     case !saveWorld of
			NONE => ("NULL", mainLabel (* bogus *))
		      | SOME(c, l) => (Chunk.nameToString c, Label.toString l)
	       in C.profile("MLton_main (magic)", overhead, print)
		  ; C.call("MLTON_main",
			   [if usedFixedHeap then C.truee else C.falsee,
			    C.int fromSize,
			    C.int liveSize,
			    C.int maxFrameSize,
			    C.int numGlobals,
			    globals,
			    C.int maxFrameIndex,
			    magic,
			    if isSome(!saveWorld) then "&loadGlobals" else "NULL",
			    C.int(List.length(!intInfs)),
			    mainChunk,
			    mainLabel,
			    saveWorldChunk,
			    saveWorldLabel],
			   print)
	       end

	    fun declareChunks() =
	       List.foreach(!chunks, fn Chunk.Chunk{name, ...} =>
			    (print "static void "
			     ; print(ChunkName.toString name)
			     ; print "();\n"))

	    fun newline() = print "\n"

	 in List.foreach(includes, fn i =>
			 (print "#include \""
			  ; print i
			  ; print "\"\n"))
	    ; print "\nMLTON_globals\n"
	    ; declareIndices()
	    ; declareIntInfs()
	    ; declareStrings()
	    ; declareGlobals()
	    ; declareChunks()
	    ; declareNextChunks()
	    ; declareFrameOffsets print
	    ; declareFrameLayouts()
	    ; declareLocalIndices print
	    ; if isSome(!saveWorld)
		 then (declareSaveGlobals()
		       ; declareLoadGlobals())
	      else ()
	    ; newline()
	    ; List.foreach(!chunks, fn c => Chunk.output(c, print))
	    ; declareMain()
	    ; print "\n"
	 end

      fun newChunk{program = T{chunks, ...}, entries} =
	 let val c = Chunk.new(entries)
	 in ListRef.push(chunks, c)
	    ; c
	 end
	     
      fun newGlobal(T{globalCounter, ...}, ty) =
	 Global.T{index = Counter.next(globalCounter ty), ty = ty}

      fun newGlobalPointerNonRoot(T{globalPointerNonRootCounter, ...}) =
	 Operand.GlobalPointerNonRoot
	 (Counter.next globalPointerNonRootCounter)

      local
	 fun make(sel, ty) (p as T r, s) =
	     let val g = newGlobal(p, ty)
	     in ListRef.push(sel r, (g, s));
		Operand.global g
	     end
      in
	 val newIntInf = make(#intInfs, Type.pointer)
	 val newString = make(#strings, Type.pointer)
      end

      val newGlobal = Operand.global o newGlobal
   end

fun reset() = Offsets.Ints.reset()
   
end
