(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MACHINE_STRUCTS = 
   sig
      structure Label: ID
      structure Prim: PRIM
      structure SmallIntInf: SMALL_INT_INF
   end

signature MACHINE = 
   sig
      include MACHINE_STRUCTS

      structure Type: MTYPE

      structure Register:
	 sig
	    type t

	    val ty: t -> Type.t
	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      structure Operand:
	 sig
	    type t

	    val void: t

	    (* constants *)
	    val int: int -> t
	    val int32: Int32.int -> t
	    val intInf: SmallIntInf.t -> t

	    val prim: Prim.t * Type.t -> t (* Prim.numArgs prim = NONE *)

	    (* must be nonzero mod 4 *)
	    val pointer: int -> t

	    val uint: Word32.word -> t
	    val char: char -> t
	    val float: string -> t

	    val label: Label.t -> t

	    val bogus: Type.t -> t

	    (* takes an IntOrPointer and makes it an int *)
	    val castInt: t -> t

	    (* addresses *)
	    val register: Register.t -> t
	    val deRegister: t -> Register.t option
	       
	    (* *(base + (offset * scale)) *)
	    val scaledOffset: {base: t,
			       offset: t,
			       scale: int,
			       ty: Type.t} -> t

	    (* *(base + offset) *)
	    val offset: {base: t,
			 offset: t,
			 ty: Type.t} -> t

	    val stackOffset: {offset: int, ty: Type.t} -> t
	    val deStackOffset: t -> {offset: int, ty: Type.t} option
	       
	    val contents: t * Type.t -> t

	    val ty: t -> Type.t

	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      (* This is the information that the garbage collector needs to know
       * at a program point in order to do a collection.
       *)
      structure GCInfo:
	 sig
	    (* Regs and offsets should not contain duplicates. *)
	    datatype t =
	       T of {
		     (* Size of stack frame in bytes,
		      * including return address.
		      *)
		     frameSize: int,
		     (* Live pointer valued registers. *)
		     regs: Register.t list,
		     (* Live pointer valued stack offsets. *)
		     offsets: int list
		     }
	 end

      structure Statement:
	 sig
	    type t

	    (* when registers or offsets appear in operands, there is an
	     * implicit contents of.  when they appear as locations, there is not
	     *)
	    val move: {dst: Operand.t, src: Operand.t} -> t

	    (* error if dsts and srcs aren't of same length
	     *)
	    val moves: {
			dsts: Operand.t list,
			srcs: Operand.t list
		       } -> t list
	       
	    val assign: {dst: Operand.t,
			 oper: Prim.t,
			 args: Operand.t list} -> t

	    val sideEffect: Prim.t * Operand.t list -> t

	    val saveExnStack: {offset: int} -> t
	    val restoreExnStack: {offset: int} -> t

	    val stackOverflowCheck: t

	    (* push/pop number of bytes from stack *)
	    val push: int -> t
	    val pop: int -> t
	       
	    (* fixed size allocation *)
	    val allocate:
	       {dst: Operand.t,
		size: int,
		numPointers: int,
		numWordsNonPointers: int,
		stores: {offset: int,
			 value: Operand.t} list} -> t

	    val allocateArray:
	       {dst: Operand.t,
		numElts: Operand.t,
		numPointers: int,
		numBytesNonPointers: int,
		gcInfo: GCInfo.t
		} -> t
	 end

      structure Transfer:
	 sig
	    type t
	    type chunk

	    val halt: Operand.t -> t
	    val bug: t
	    val return: t
	    val raisee: t

	    val switch: {
			 test: Operand.t,
			 cases: (int * Label.t) list,
			 default: Label.t option
			} -> t

	    val switchIP: {
			   test: Operand.t,
			   int: Label.t,
			   pointer: Label.t
			  } -> t

	    val nearJump: Label.t -> t
	    val farJump: chunk * Label.t -> t
	 end

      structure Chunk:
	 sig
	    type t

	    val register: t * int * Type.t -> Register.t
	    val tempRegister: t * Type.t -> Register.t

	    val newBlock: 
	       t * {
		    profileName: string,
		    label: Label.t,
		    (* 
		     * If gcInfo is SOME, then a limit check will be inserted
		     * at the beginning of this block.
		     *)
		    gcInfo: GCInfo.t option,
		    statements: Statement.t list,
		    transfer: Transfer.t
		   } -> unit

	    val equals: t * t -> bool
	 end
      sharing type Transfer.chunk = Chunk.t

      structure Program:
	 sig
	    type t

	    val new: unit -> t

	    val setMain: t * Chunk.t * Label.t -> unit
	    val setSaveWorld: t * Chunk.t * Label.t -> unit

	    (* Create a new chunk *)
	    val newChunk: {program: t,
			   entries: Label.t list} -> Chunk.t

	    val newFrame:
	       t * {return: Label.t, (* where to return to *)
		    chunk: Chunk.t,
		    (* number of bytes in frame, including return address *)
		    size: int,
		    (* where are live pointers in the current stack frame
		     * relative to the stack pointer of the frame below
		     *)
		    liveOffsets: int list
		    } -> unit

	    val newHandler: t * Label.t * Chunk.t -> unit

	    val newGlobal: t * Type.t -> Operand.t

	    (* A global pointer that the GC doesn't use as a root *)
	    val newGlobalPointerNonRoot: t -> Operand.t
	       
	    val newString: t * string -> Operand.t
	    val newIntInf: t * string -> Operand.t

	    val output: {program: t,
			 includes: string list,
			 print: string -> unit} -> unit
	 end

      val reset: unit -> unit
   end
