(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Has a special case to make sure that true is represented as 1
 * and false is represented as 0.
 *)
functor Representation(S: REPRESENTATION_STRUCTS): REPRESENTATION = 
struct

open S
local open Cps
in
   structure Con = Con
   structure Tycon = Tycon
   structure Type = Type
end

structure TyconRep =
   struct
      datatype t =
	 (* Datatype contains a single variant, and hence constructor requires no
	  * additional representation.
	  *) 
	 Prim of Mtype.t
       (* All cons are non-value-carrying and are represented as ints. *)
       | Enum
       (* All cons except for one are non-value-carrying and are represented as
	* ints that are nonzero mod 4.  The value carrying con is represented
	* transparently, i.e. the value is known to be a pointer and is left
	* as such.
	*)
       | EnumDirect
       (* All cons except for one are non-value-carrying and are represented as
	* ints that are nonzero mod 4.  The value carrying con is represented
	* by boxing its arg.
	*)
       | EnumIndirect
       (* Non-value-carrying and are represented as ints that are nonzero mod 4.
	* Value carrying cons is represented by boxing the args and adding
	* an integer tag.
	*)
       | EnumIndirectTag
       (* All cons are value carrying and are represented by boxing the args
	* and adding an integer tag.
	*)
       | IndirectTag

      val void = Prim Mtype.void

      val toMtype =
	 fn Prim t => t
	  | Enum => Mtype.int
	  | EnumDirect => Mtype.pointer
	  | EnumIndirect => Mtype.pointer
	  | EnumIndirectTag => Mtype.pointer
	  | IndirectTag => Mtype.pointer

      val layout =
	 let open Layout
	 in fn Prim m => Mtype.layout m
	     | Enum => str "Enum"
	     | EnumDirect => str "EnumDirect"
	     | EnumIndirect => str "EnumIndirect"
	     | EnumIndirectTag => str "EnumIndirectTag"
	     | IndirectTag => str "IndirectTag"
	 end
      
      val equals =
	 fn (Prim t, Prim t') => Mtype.equals(t, t')
	  | (Enum, Enum) => true
	  | (EnumDirect, EnumDirect) => true
	  | (EnumIndirect, EnumIndirect) => true
	  | (EnumIndirectTag, EnumIndirectTag) => true
	  | (IndirectTag, IndirectTag) => true
	  | _ => false
   end

structure ConRep =
   struct
      datatype t =
	 Void
       | Int of int
       | IntCast of int
       | Transparent of Mtype.t
       | Tuple
       | TagTuple of int

      local
	 open Layout
      in
	 val layout =
	    fn Void => str "Void"
	     | Int n => seq[str "Int ", Int.layout n]
	     | IntCast n => seq[str "IntCast ", Int.layout n]
	     | Transparent t => seq[str "Transparent ", Mtype.layout t]
	     | Tuple => str "Tuple"
	     | TagTuple n => seq[str "TagTuple ", Int.layout n]
      end
   end

(* fixed-point.  Initially assume all datatype tycons are Void
 * Change them if they have more than one variant or contain
 * a useful component
 *)

fun compute(Cps.Program.T{datatypes, ...}) =
   let 
      val {get = tyconRep, set = setTyconRep} =
	 Property.new(Tycon.plist, Property.initRaise("rep", Tycon.layout))

(*       val tyconRep =
 * 	 Trace.trace("tyconRep", Tycon.layout, TyconRep.layout) tyconRep
 *)

      val {get = conRep, set = setConRep} =
	 Property.new(Con.plist, Property.initRaise("rep", Con.layout))

      val tyconMtype = TyconRep.toMtype o tyconRep

      fun toMtype t =
	 let open Type
	 in case Type.dest t of
	    Char => Mtype.char
	  | Int => Mtype.int
	  | IntInf => Mtype.pointer
	  | Pointer => Mtype.uint
	  | Word => Mtype.uint
	  | Word8 => Mtype.char
	  | Real => Mtype.double
	  | Stack => Mtype.pointer
	  | String => Mtype.pointer
	  | Array _ => Mtype.pointer
	  | Vector _ => Mtype.pointer
	  | Ref _ => Mtype.pointer
	  | Datatype c => tyconMtype c
	  | Tuple [] => Mtype.void
	  | Tuple _ => Mtype.pointer
	  | Variant(tycon, _) => tyconMtype tycon
	 end

      val _ =
	 List.foreach(datatypes, fn {tycon, ...} =>
		      setTyconRep(tycon, TyconRep.void))

      (* split list of constructors into those that have args and those that don't *)
      fun splitCons cons =
	 List.fold(cons, ([], []), fn ({con, args}, (no, have)) =>
		   case args of
		      [] => (con :: no, have)
		    | _ =>
			 if Mtype.isVoid(toMtype(Type.tuple args))
			    then (con :: no, have)
			 else (no, {con = con, args = args} :: have))
	 
   in
      FixedPoint.fixedPoint
      (fn continue =>
       List.foreach
       (datatypes, fn {tycon, cons} =>
	let
	   val (noArgs, haveArgs) = splitCons cons
	   val old = tyconRep tycon
	   val new =
	      case (noArgs, haveArgs) of
		   ([],     [])           => TyconRep.void
	         | ([_],    [])           => TyconRep.void
	         | (_,      [])           => TyconRep.Enum
	         | ([],     [{args, ...}]) => TyconRep.Prim(toMtype
							    (Type.tuple args))
		 | (_,      [{args, ...}]) =>
		      if (case args of
			     [a] => let open Type
				    in case Type.dest a of
				       Datatype c =>
					  (case tyconRep c of
					      TyconRep.IndirectTag => true
					    | _ => false)
				     | String => true
				     | Array _ => true
				     | Vector _ => true
				     | Ref _ => true
				     | Tuple _ => true
				     | _ => false
				    end
			   | _ => true)
			 then TyconRep.EnumDirect
		      else TyconRep.EnumIndirect
		 | ([],     _)            => TyconRep.IndirectTag
		 | _                      => TyconRep.EnumIndirectTag
	in if TyconRep.equals(old, new)
	      then ()
	   else (continue();
		 setTyconRep(tycon, new))
	end));

      (* set constructor representations *)
      let
	 fun direct(haveArgs, t) =
	    case haveArgs of
	       [{con, args}] => setConRep(con,
					  (case args of
					      [_] => ConRep.Transparent t
					    | _ => ConRep.Tuple))
	     | _ => ()
		  
	 (* choose tags that are not equal to 0 mod 4 *)
	 fun enum noArgs =
	    let
	       fun loop(i, cs) =
		  case cs of
		     [] => ()
		   | c :: cs => (setConRep(c, ConRep.IntCast i);
				 loop(i + 2, cs))
	    in loop(1, noArgs)
	    end
	 fun indirectTag haveArgs = 
	    List.foreachi(haveArgs, fn (i, {con, args}) =>
			  setConRep(con, ConRep.TagTuple i))
      in List.foreach
	 (datatypes, fn {tycon, cons} =>
	  let val (noArgs, haveArgs) = splitCons cons
	  in case tyconRep tycon of
	     TyconRep.Prim t =>
		(case (noArgs, haveArgs) of
		    ([], []) => ()
		  | ([con], []) => setConRep(con, ConRep.Void)
		  | ([], [_]) => direct(haveArgs, t)
		  | _ => Error.bug("strange TyconRep.Prim for "
				   ^ Layout.toString(Tycon.layout tycon)))
	   | TyconRep.Enum =>
		if Tycon.equals(tycon, Tycon.bool)
		   then (setConRep(Con.falsee, ConRep.Int 0) ;
			 setConRep(Con.truee, ConRep.Int 1))
		else List.foreachi(noArgs, fn (i, c) =>
				   setConRep(c, ConRep.Int i))
	   | TyconRep.EnumDirect =>
		(enum noArgs ; direct(haveArgs, Mtype.pointer))
	   | TyconRep.EnumIndirect =>
		(enum noArgs ;
		 List.foreach(haveArgs, fn {con, ...} =>
			      setConRep(con, ConRep.Tuple)))
	   | TyconRep.EnumIndirectTag => (enum noArgs ; indirectTag haveArgs)
	   | TyconRep.IndirectTag => indirectTag haveArgs
	  end)
      end ;

   {tyconRep = tyconRep,
    conRep = conRep,
    toMtype = toMtype}
   end

end
