(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * All local variables in the Sxml are renamed to new variables in Cps,
 * unless they are global, as determined by the Globalization pass.
 * Renaming must happen because an Sxml variable will be bound in the Cps
 * once for each lambda it occurs in.  The main trickiness is caused because a
 * property is used to implement the renamer map.  Hence, a variable binding
 * must always be visited with "newVar" or "newScope" before it is looked up
 * with "getNewVar".  "newScope" also handles resetting the variable to its
 * old value once the processing of the lambda is done.
 *)
functor ClosureConvert(S: CLOSURE_CONVERT_STRUCTS): CLOSURE_CONVERT = 
struct

open S

structure Globalize = Globalize(open Sxml)

local open Sxml
in structure Sexp = Exp
   structure Sdec = Dec
   structure Slambda = Lambda
   structure Spat = Pat
   structure SprimExp = PrimExp
   structure SvarExp = VarExp
   structure Stype = Type
   open Atoms
end

local open Cps
in
   structure Cause = Cause
   structure Ctype = Type
   structure Func = Func
   structure Cdec = Dec
   structure Cexp = DirectExp
   structure Cprogram = Program
end

structure ExtractTargs = ExtractTargs(structure Prim = Prim
				      structure Type = Ctype)

structure FlowAnalysis = FlowAnalysis(structure Sxml = Sxml)

local open FlowAnalysis
in structure Value = Value
   local open Value
   in structure Lambdas = Lambdas
   end
end

(* This is one of the results returned internally by the converter -- an
 * accumulation of toplevel Cps globals and function declarations.
 *)
structure Accum =
   struct
      structure AL = AppendList

      datatype t = T of {globals: {var: Var.t, ty: Ctype.t, exp: Cexp.t} AL.t,
			 functions: {name: Func.t,
				     args: (Var.t * Ctype.t) list,
				     body: Cps.Exp.t,
				     returns: Ctype.t list} list}

      val empty = T{globals = AL.empty, functions = []}

      fun addGlobals(T{globals, functions}, gs) =
	 T{globals = AL.append(globals, AL.fromList gs),
	   functions = functions}

      fun addGlobal(ac, g) = addGlobals(ac, [g])

      fun addFunc(T{globals, functions}, f) =
	 T{globals = globals, functions = f :: functions}

      fun done(T{globals, functions}) =
	 {functions = functions,
	  globals =
	  let
	     val exp =
		(* Must shrink because coercions may be inserted at constructor
		 * applications.  I'm pretty sure the shrinking will eliminate
		 * any case expressions/local functions.
		 * The "NoDelete" is because the shrinker is just processing
		 * globals and hence cannot safely delete a variable that
		 * has no occurrences, since there may still be occurrences in
		 * functions.
		 *)
		Cps.shrinkExpNoDelete
		(Cexp.toExp(Cexp.lett
			    {decs = AL.toList globals,
			     body = Cexp.tuple{exps = [], ty = Ctype.unit}}))
	     val decs = Cps.Exp.decs exp
	  in 
	     List.map(decs,
		      fn Cdec.Bind b => b
		       | _ => Error.bug "globalization produced a non binding")
	  end}
   end

(* val traceConvertPrimExp =
 *    Trace.trace2("convertPrimExp",
 * 		Var.layout, SprimExp.layout,
 * 		Layout.ignore: Cexp.t -> Layout.t)
 *)

(* val traceConvertExp =
 *    Trace.trace2("convertExp", Sexp.layout, Instance.layout, Cexp.layout)
 *)

structure LambdaFree = LambdaFree(Sxml)

val incKnownCount = Prim.newNullary "MLTON_incCpsKnown"
val incUnknownCount = Prim.newNullary "MLTON_incCpsUnknown"

fun closureConvert(program as Sxml.Program.T{datatypes, body}): Cps.Program.t =
   let
      val {value, conArg} =
	 Control.pass
	 {name = "flow analysis",
	  suffix = "flow",
	  thunk = fn () => FlowAnalysis.analyze program,
	  style = Control.No,
	  display =
	  Control.Layouts
	  (fn ({value, ...}, display) =>
	   Sexp.foreachBoundVar
	   (body, fn (x, _, _) => display(let open Layout
					  in seq[Var.layout x,
						 str " ",
						 Value.layout(value x)]
					  end)))}

      val {lambdaFree, lambdaRec} =
	 Control.traceCall "free variables" LambdaFree.lambdaFree program

      val {isGlobal} =
	 Control.traceCall "globalize"
	 Globalize.globalize{program = program, lambdaFree = lambdaFree}

(*      val isGlobal = Trace.trace("isGlobal", Var.layout, Bool.layout) isGlobal *)

      local
	 fun filter f l = List.keepAll(f l, not o isGlobal)
      in
	 val lambdaFree = filter lambdaFree
	 val lambdaRec = filter lambdaRec
      end
	 
      (* For each lambda, keep track of
       * - the "original" name in the source program, so the closure
       *   conversion output has some readability.
       * - the type of its environment record
       *)
      local
	 val {get: Slambda.t -> {
				 name: Func.t,
				 ty: Ctype.t option ref
				},
	      set} = Property.new(Slambda.plist,
				  Property.initRaise("info", Layout.ignore))
      in
	 val lambdaType = #ty o get
	 val lambdaName = #name o get

	 val _ =
	    let
	       fun doit(l, x: Var.t) =
		  set(l, {name = Func.newString(Var.originalName x),
			  ty = ref NONE})
	    in Sexp.foreachExp
	       (body, fn e =>
		let val {decs, ...} = Sexp.dest e
		in List.foreach
		   (decs, fn d =>
		    case d of
		       Sdec.MonoVal{var, exp, ...} =>
			  (case exp of
			      SprimExp.Lambda l => doit(l, var)
			    | _ => ())
		     | Sdec.Fun{decs, ...} =>
			  List.foreach(decs, fn {var, lambda, ...} =>
				       doit(lambda, var))
		     | _ => ())
		end)
	    end
      end
   
      val {get = lambdasInfoOpt, set = setLambdasInfo} =
	 Property.new(Lambdas.plist, Property.initConst NONE)

      val {hom = convertType, destroy = destroyConvertType} =
	 Stype.makeMonoHom{con = fn (_, c, ts) => Ctype.con(c, ts)}

      fun convertTypes ts = List.map(ts, convertType)

      val {get = valueTypeOpt, set = setValueType} =
	 Property.new(Value.plist, Property.initConst NONE)

      (* accumulates the new datatypes built for sets of lambdas *)
      val newDatatypes = ref []

      fun valueType(v: Value.t): Ctype.t =
	 case valueTypeOpt v of
	    SOME t => t
	  | NONE =>
	       let
		  val t = 
		     case Value.dest v of
			Value.Type t => convertType t
		      | Value.Ref v => Ctype.reff(valueType v)
		      | Value.Array v => Ctype.array(valueType v)
		      | Value.Vector v => Ctype.vector(valueType v)
		      | Value.Tuple vs => Ctype.tuple(List.map(vs, valueType))
		      | Value.Lambdas ls => #ty(lambdasInfo ls)
	       in setValueType(v, SOME t); t
	       end
      and lambdasInfo ls =
	 case lambdasInfoOpt ls of
	    SOME info => info
	  | NONE => 
	       let val tycon = Tycon.newString "lambdas"
		  val cons = List.map(Lambdas.toList ls, fn l =>
				      {lambda = Value.Lambda.dest l,
				       con = Con.newString "Env"})
		  val ty = Ctype.con(tycon, [])
		  val info = {ty = ty, cons = cons}
	       in setLambdasInfo(ls, SOME info)
		  ; (ListRef.push
		     (newDatatypes,
		      {tycon = tycon,
		       cons = List.map(cons, fn {lambda, con} =>
				       {con = con, args = [envType lambda]})}))
		  ; info
	       end

      and varType x = valueType(value x)

      and envType l =
	 let val ty = lambdaType l
	 in case !ty of
	    NONE => let val t = Ctype.tuple(List.map(lambdaFree l, varType))
		    in ty := SOME t; t
		    end
	  | SOME t => t
	 end

      fun valueLambdasInfo v =
 	 case Value.dest v of
	    Value.Lambdas l => lambdasInfo l
	  | _ => Error.bug "valueLambdasInfo of non-lambda"

      val varLambdasInfo = valueLambdasInfo o value

      val varExp = value o SvarExp.var
      val expValue = varExp o #result o Sexp.dest
	 
      val expType = valueType o expValue

      val datatypes =
	 List.map(datatypes, fn {tycon, cons, ...} =>
		  {tycon = tycon,
		   cons = List.fold(cons, [], fn ({con, ...}, cons) =>
				    {con = con,
				     args = (case conArg con of
						NONE => []
					      | SOME v => [valueType v])}
				    :: cons)})

      (* Variable renaming *)
      local
	 val {get, set} = Property.new(Var.plist, Property.initFun ref)
      in
	 fun getNewVar x = !(get x)

	 fun newVar(x: Var.t): Var.t =
	    if isGlobal x
	       then x
	    else let val x' = Var.new x
		 in get x := x'; x'
		 end

	 fun newScope(xs: Var.t list, f: Var.t list -> 'a): 'a =
	    let val refs = List.map(xs, get)
	       val old = List.map(refs, !)
	       val new = List.map(xs, newVar)
	       val res = f new
	    in List.foreach2(refs, old, op :=)
	       ; res
	    end
      end

(*      val newVar = Trace.trace("newVar", Var.layout, Var.layout) newVar *)

      (*------------------------------------*)
      (*               coerce               *)
      (*------------------------------------*)

(*       val traceCoerce =
 * 	 Trace.trace3("coerce", Cexp.layout, Value.layout, Value.layout,
 * 		      Cexp.layout)
 *)

(*       val traceCoerceTuple =
 * 	 let val layoutValues = List.layout(", ", Value.layout)
 * 	 in Trace.trace3("coerceTuple", Cexp.layout,
 * 			 layoutValues, layoutValues, Cexp.layout)
 * 	 end
 *)

      fun coerce arg: Cexp.t =
(*	 traceCoerce *)
	 (fn (e: Cexp.t, from: Value.t, to: Value.t) =>
	  if Value.equals(from, to)
	     then e
	  else 
	     case (Value.dest from, Value.dest to) of
		(Value.Tuple vs, Value.Tuple vs') =>
		   coerceTuple(e, valueType from, vs, valueType to, vs')
	      | (Value.Lambdas ls, Value.Lambdas ls') =>
		   if Lambdas.equals(ls, ls') then e
		   else
		      let val {cons, ...} = lambdasInfo ls
			 val {cons = cons', ty, ...} = lambdasInfo ls'
			 val {get = lambdaCon, set = setLambdaCon, destroy}=
			    Property.newDest(Slambda.plist, Property.initRaise("lambda target", Slambda.layout))
		      in List.foreach(cons', fn {lambda, con, ...} =>
				      setLambdaCon(lambda, con));
			 Cexp.casee
			 {cause = Cause.Coerce,
			  test = e, default = NONE,
			  ty = ty,
			  cases =
			  List.map
			  (cons, fn {lambda, con} =>
			   let val tuple = (Var.newNoname(), envType lambda)
			   in {con = con,
			       args = [tuple],
			       body = Cexp.conApp{con = lambdaCon lambda,
						  ty = ty,
						  args = [Cexp.var tuple]}}
			   end)}
			 before destroy()
		      end
	   | _ => Error.bug "impossible coercion") arg

      and coerceTuple arg =
(*	 traceCoerceTuple *)
	 (fn (e: Cexp.t,
	      ty: Ctype.t, vs: Value.t list,
	      ty': Ctype.t, vs': Value.t list) =>
	  if Ctype.equals(ty, ty')
	     then e
	  else 
	     Cexp.detuple
	     {tuple = e,
	      body =
	      fn components => 
	      Cexp.tuple
	      {exps = List.map3(components, vs, vs', fn (x, v, v') =>
				coerce(Cexp.var(x, valueType v), v, v')),
	       ty = ty'}}) arg
	 
      fun convertVar x : Cexp.t = Cexp.var(getNewVar x, varType x)
      
      fun convertVarExp(x: SvarExp.t) = convertVar(SvarExp.var x)
	 
      fun convertVarExps xs = List.map(xs, convertVarExp)

      (*------------------------------------*)      	       
      (*               apply                *)
      (*------------------------------------*)

      fun apply{func, arg, resultVar, resultVal}: Cexp.t =
	 let val funcVal = varExp func
	    val argVal = varExp arg
	    val argExp = convertVarExp arg
	    val ty = valueType resultVal
	    val {cons, ...} = valueLambdasInfo funcVal
	 in Cexp.casee
	    {cause = Cause.Dispatch,
	     test = convertVarExp func,
	     ty = ty,
	     default = NONE,
	     cases =
	     List.map
	     (cons, fn {lambda, con} =>
	      let val {arg = param, body, ...} = Slambda.dest lambda
		 val result = expValue body
		 val env = (Var.newString "env", envType lambda)
	      in {con = con,
		  args = [env],
		  body = (coerce(Cexp.call
				 {func = lambdaName lambda,
				  args = [Cexp.var env,
					  coerce(argExp, argVal, value param)],
				  ty = valueType result},
				 result, resultVal))}
	      end)}
	 end

      (*------------------------------------*)
      (*             convertExp             *)
      (*------------------------------------*)
	 
      fun envTuple(l: Slambda.t): Cexp.t =
	 Cexp.tuple{exps = List.map(lambdaFree l, convertVar),
		    ty = envType l}

      fun recursives(old: Var.t list, new: Var.t list, env) =
	 List.fold2
	 (old, new, [], fn (old, new, ac) =>
	  let val {cons, ty, ...} = varLambdasInfo old
	  in case cons of
	     [{con, ...}] =>
		{var = new,
		 ty = ty,
		 exp = Cexp.conApp{con = con, ty = ty, args = [Cexp.var env]}}
		:: ac
	   | _ => Error.bug "val rec lambda must be singleton"
	  end)

      (* Closure convert an expression, returning:
       *   - the target cps expression
       *   - a list of global declarations (in order)
       *   - a list of function declarations
       * Accumulate the globals onto the end of the given ones.
       *)
      fun convertExp(e: Sexp.t, ac: Accum.t): Cexp.t * Accum.t =
	 let val {decs, result} = Sexp.dest e
	    (* process decs left to right, since bindings of variables
	     * must be visited before uses
	     *)
	    val (decs, ac) =
	       List.fold
	       (decs, ([], ac), fn (d, (binds, ac)) =>
		case d of
		   Sdec.MonoVal{var, ty, exp} =>
		      let val (exp, ac) = convertPrimExp(var, exp, ac)
			 val bind = {var = newVar var,
				     ty = varType var,
				     exp = exp}
		      in if isGlobal var
			    then (binds, Accum.addGlobal(ac, bind))
			 else (bind :: binds, ac)
		      end
		 | Sdec.Fun{decs, ...} =>
		      (case decs of
			  [] => (binds, ac)
			| {var, ty, lambda} :: _ => 
			     let val tupleVar = Var.newString "tuple"
				val tupleTy = envType lambda
				val binds' =
				   {var = tupleVar,
				    ty = tupleTy,
				    exp = envTuple lambda}
				   :: recursives(List.map(decs, #var),
						 List.map(decs, newVar o #var),
						 (tupleVar, tupleTy))
				val (binds, ac) =
				   if isGlobal var
				      then (binds, Accum.addGlobals(ac, binds'))
				   else (List.fold(binds', binds, op ::), ac)
			     in (binds,
				 List.fold(decs, ac, fn ({lambda, ...}, ac) =>
					   convertLambda(lambda, ac)))
			     end)
		 | _ => Error.bug "closure convert saw strange dec")
	 in (Cexp.lett{decs = List.reverse decs,
		       body = convertVarExp result},
	     ac)
	 end

      and convertPrimExp(x: Var.t, e: SprimExp.t, ac: Accum.t): Cexp.t * Accum.t =
	 let
	    val v = value x
	    val ty = valueType v
	    fun convertJoin(e, ac) =
	       let val (e', ac) = convertExp(e, ac)
	       in (coerce(e', expValue e, v), ac)
	       end
	    fun simple e = (e, ac)
	 in case e of
	    SprimExp.Var x => simple(convertVarExp x)
	  | SprimExp.Const c => simple(Cexp.const c)
	  | SprimExp.PrimApp{prim, targs, args} =>
	       let open Prim.Name
		  val (targs, args) =
		     case (Prim.name prim, args) of
			(ArrayUpdate, [a, i, y]) =>
			   let val v = Value.dearray(varExp a)
			   in ([valueType v],
			       [convertVarExp a,
				convertVarExp i,
				coerce(convertVarExp y, varExp y, v)])
			   end
		      | (Assign, [r, y]) =>
			   let val v = Value.deref(varExp r)
			   in ([valueType v],
			       [convertVarExp r,
				coerce(convertVarExp y, varExp y, v)])
			   end
		      | (Ref, [y]) =>
			   let val v = Value.deref v
			   in ([valueType v],
			       [coerce(convertVarExp y, varExp y, v)])
			   end
		      | _ =>
			   (ExtractTargs.extract
			    {prim = prim,
			     args = List.map(args, varType o SvarExp.var),
			     result = ty},
			    convertVarExps args)
	       in simple(Cexp.primApp{prim = prim,
				      ty = ty,
				      targs = targs,
				      args = args})
	       end
	  | SprimExp.Tuple xs =>
	       simple(Cexp.tuple{exps = convertVarExps xs, ty = ty})
	  | SprimExp.Select{tuple, offset} =>
	       simple(Cexp.select{tuple = convertVarExp tuple,
				  offset = offset,
				  ty = ty})
	  | SprimExp.ConApp{con = con, arg, ...} =>
	       simple
	       (Cexp.conApp
		{con = con,
		 ty = ty,
		 args = (case (arg, conArg con) of
			    (NONE, NONE) => []
			  | (SOME arg, SOME conArg) =>
			       let val argVal = varExp arg
				  val arg = convertVarExp arg
			       in if Value.equals(argVal, conArg)
				     then [arg]
				  else [coerce(arg, argVal, conArg)]
			       end
			  | _ => Error.bug "constructor mismatch")})
	  | SprimExp.Raise x => simple(Cexp.raisee(convertVarExp x))
	  | SprimExp.Handle{try, catch = (catch, _), handler} =>
	       let val (try, ac) = convertJoin(try, ac)
		  val catch = (newVar catch, varType catch)
		  val (handler, ac) = convertJoin(handler, ac)
	       in (Cexp.handlee{try = try, ty = ty,
				catch = catch, handler = handler},
		   ac)
	       end
	  | SprimExp.Case{test, cases, default} =>
	       let
		  val (default, ac) =
		     case default of
			NONE => (NONE, ac)
		      | SOME e => let val (e, ac) =  convertJoin(e, ac)
				  in (SOME e, ac)
				  end
		  val (cases, ac) =
		     List.fold
		     (cases, ([], ac),
		      fn ((p as Spat.T{con, arg, ...}, e), (cases, ac)) =>
		      let
			 val args =
			    case (conArg con, arg) of
			       (NONE, NONE) => []
			     | (SOME v, SOME(arg, _)) =>
				  [(newVar arg, valueType v)]
			     | _ => Error.bug "constructor mismatch"
			 val (body, ac) = convertJoin(e, ac)
		      in ({con = con, args = args, body = body} :: cases, ac)
		      end)
	       in (Cexp.casee
		   {cause = Cause.User,
		    test = convertVarExp test,
		    ty = ty, cases = cases, default = default},
		   ac)
	       end
	  | SprimExp.Lambda l =>
	       let val ac = convertLambda(l, ac)
		  val {cons, ...} = valueLambdasInfo v
	       in case List.peek(cons, fn {con, lambda = l'} =>
				 Slambda.equals(l, l')) of
		  NONE => Error.bug "lambda must exist in its own set"
		| SOME{con, ...} =>
		     (Cexp.conApp{con = con, ty = ty, args = [envTuple l]}, ac)
	       end
	  | SprimExp.App{func, arg} =>
	       (apply{func = func, arg = arg, resultVar = x, resultVal = v},
		ac)
	 end

      and convertLambda(lambda, ac: Accum.t): Accum.t =
	 let val {arg = argVar, body, ...} = Slambda.dest lambda
	    val env = (Var.newString "env", envType lambda)
	    val args = [env, (newVar argVar, varType argVar)]
	    val returns = [expType body]
	    val recs = lambdaRec lambda
	 in newScope
	    (lambdaFree lambda, fn components =>
	     newScope
	     (recs, fn recs' =>
	      let val decs = recursives(recs, recs', env)
		 val (body, ac) = convertExp(body, ac)
	      in Accum.addFunc
		 (ac, {name = lambdaName lambda,
		       args = args, returns = returns,
		       body =
		       Cps.shrinkExp []
		       (Cexp.toExp
			(Cexp.lett
			 {decs = decs,
			  body = Cexp.detupleBind{tuple = Cexp.var env,
						  components = components,
						  body = body}}))})
	      end))
	 end

      (*------------------------------------*)
      (*    main body of closure convert    *)
      (*------------------------------------*)

      val main = Func.newString "main"

      val {functions, globals} =
	 Control.traceCall "convert"
	 (fn () =>
	  let val (body, ac) = convertExp(body, Accum.empty)
	  in Accum.done
	     (Accum.addFunc(ac, {name = main,
				 args = [],
				 returns = [Ctype.unit],
				 body = Cps.shrinkExp [] (Cexp.toExpHalt body)}))
	  end) ()

      val datatypes = datatypes @ (!newDatatypes)

      val program =
	 Cprogram.T{datatypes = datatypes,
		    globals = globals,
		    functions = functions,
		    main = main}
	 
   in destroyConvertType();
      Value.destroy();
      Cprogram.clear program;
      program
   end

end
