(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor LambdaFree(S: LAMBDA_FREE_STRUCTS): LAMBDA_FREE = 
struct

open S

open Dec PrimExp

structure Status =
   struct
      datatype t = Unseen | Free | Bound

      val toString =
	 fn Unseen => "Unseen"
	  | Free => "Free"
	  | Bound => "Bound"

      val layout = Layout.str o toString
   end
datatype status = datatype Status.t

fun lambdaFree(Program.T{body, ...}) =
   let
      val {get = getFree, set = setFree} =
	 Property.new(Lambda.plist, Property.initRaise("free", Lambda.layout))

      val {get = getRec, set = setRec} =
	 Property.new(Lambda.plist, Property.initConst [])

      type scope = {frees: Var.t list ref,
		    get: Var.t -> Status.t,
		    set: Var.t * Status.t -> unit}

      fun bind(x: Var.t, {set, ...}: scope) = set(x, Bound)

      fun var(x: Var.t, {get, set, frees}: scope) =
	 case get x of
	    Unseen => (set(x, Free); ListRef.push(frees, x))
	  | _ => ()

      fun vars(xs, s) = List.foreach(xs, fn x => var(x, s))
	 
      fun varExp(x: VarExp.t, s: scope) = var(VarExp.var x, s)
	 
      fun varExpOpt(x, s) =
	 case x of
	    NONE => ()
	  | SOME x => varExp(x, s)

      fun varExps(xs, s) = List.foreach(xs, fn x => varExp(x, s))

      val {get = varInfo: Var.t -> {frees: Var.t list ref ref,
				    status: Status.t ref}, ...} =
	 Property.new(Var.plist,
		      Property.initFun(let val bogusFrees = ref []
				       in fn _ => {frees = ref bogusFrees,
						   status = ref Unseen}
				       end))

      fun newScope th = 
	 let
	    val frees = ref []
	    val all = ref []
	    fun statusRef x =
	       let val {frees = frees', status} = varInfo x
	       in if frees = !frees'
		     then ()
		  else (ListRef.push(all, (frees', !frees', status, !status))
			; frees' := frees; status := Unseen)
		  ; status
	       end
		    
	    fun get x = !(statusRef x)

	    fun set(x, s) = statusRef x := s

	 in th{frees = frees, get = get, set = set}
	    ; List.foreach(!all, fn (r, v, r', v') => (r := v; r' := v'))
	    ; !frees
	 end
      
      fun exp(e, s) =
	 let val {decs, result} = Exp.dest e
	 in List.foreach
	    (decs,
	     fn Exception _ => ()
	      | MonoVal{var, exp, ...} => (primExp(exp, s); bind(var, s))
	      | PolyVal{var, exp = e, ...} => (exp(e, s); bind(var, s))
	      | Fun{decs, ...} =>
		   let
		      val {get = isBound, set, destroy} =
			 Property.newDest(Var.plist, Property.initConst false)
		      val _ =
			 List.foreach(decs, fn {var, ...} => set(var, true))
		      val xs =
			 newScope
			 (fn s =>
			  List.foreach
			  (decs, fn {lambda = l, ...} =>
			   setRec(l, List.fold(lambda l, [], fn (x, xs) =>
					       if isBound x
						  then x :: xs
					       else (var(x, s); xs)))))
		   in List.foreach(decs, fn {var, lambda, ...} =>
				   (setFree(lambda, xs)
				    ; bind(var, s)))
		      ; vars(xs, s)
		   end)
	    ; varExp(result, s)
	 end

      and primExp(e, s) = 
	 case e of
	    Const _ => ()
	  | Var x => varExp(x, s)
	  | Tuple xs => varExps(xs, s)
	  | Select{tuple, ...} => varExp(tuple, s)
	  | Lambda l =>
	       let val xs = lambda l
	       in setFree(l, xs); vars(xs, s)
	       end
	  | ConApp{arg, ...} => varExpOpt(arg, s)
	  | PrimApp{args, ...} => varExps(args, s)
	  | App{func, arg} => (varExp(func, s); varExp(arg, s))
	  | Raise x => varExp(x, s)
	  | Handle{try, catch, handler} =>
	       (exp(try, s); bind(#1 catch, s); exp(handler, s))
	  | Case{test, cases, default} =>
	       (varExp(test, s);
		expOpt(default, s);
		List.foreach(cases, fn (Pat.T{arg, ...}, e) =>
			     (case arg of
				 NONE => ()
			       | SOME(x, _) => bind(x, s);
				    exp(e, s))))
      and expOpt(e, s) =
	 case e of
	    NONE => ()
	  | SOME e => exp(e, s)

      and lambda l =
	 let val {arg, body, ...} = Lambda.dest l
	 in newScope(fn s => (bind(arg, s); exp(body, s)))
	 end

      val frees = newScope(fn s => exp(body, s))
	 
   in
      case frees of
	 [] => ()
       | _ => Error.bug "program has free variables";
	    
      {
       lambdaFree = getFree,
       lambdaRec = getRec
      }
   end

end
