(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Compile: COMPILE =
struct

(*---------------------------------------------------*)
(*              Intermediate Languages               *)
(*---------------------------------------------------*)

structure Ast = Ast()
structure Atoms = Atoms(structure Ast = Ast)
structure CoreML = CoreML(open Atoms
			  structure Type = Prim.Type)
structure Xml = Xml(open Atoms)
structure Sxml = Xml
structure Cps = Cps(open Atoms)
structure Machine = Machine(structure Label =
			       struct
				  open Cps.Func
				  fun newNoname() = newString "L"
			       end
			    structure Prim = Atoms.Prim
			    structure SmallIntInf = Atoms.SmallIntInf)

(*---------------------------------------------------*)
(*                  Compiler Passes                  *)
(*---------------------------------------------------*)

structure FrontEnd = FrontEnd(structure Ast = Ast)
structure DeadCode = DeadCode(structure CoreML = CoreML)
structure Elaborate = Elaborate(structure Ast = Ast
				structure CoreML = CoreML)
structure Infer = Infer(structure CoreML = CoreML
			structure Xml = Xml)
(*structure CallCount = CallCount(open Xml) *)
structure Monomorphise = Monomorphise(structure Xml = Xml
				      structure Sxml = Sxml)
structure ImplementExceptions = ImplementExceptions(open Sxml)
structure Polyvariance = Polyvariance(open Sxml)
structure ClosureConvert = ClosureConvert(structure Cps = Cps
					  structure Sxml = Sxml)
structure Backend = Backend(structure Cps = Cps
			    structure Machine = Machine
			    fun funcToLabel f = f
			    fun jumpToLabel j = j)

local open CoreML
in structure Constructors = Constructors
end

local open Elaborate
in structure Accum = Accum
   structure Decs = Decs
   structure Basis = Basis
   structure Env = Basis.Env
end

(* The first declaration declares the boolean variable "safe",
 * which we need to set according to Control.safe.
 * This is used to implement the -u option.
 * This is only used when Control.getMode() = Compile.
 *)

fun patchDecsSafe(ds: Decs.t): CoreML.Dec.t list =
   let fun error() = Error.bug "declarations don't start with \"val safe = true\""
   in case Decs.toList ds of
      d :: ds =>
	 let open CoreML
	 in
	    case d of
	       Dec.Val{pat = pat as Pat.Var x, tyvars = [], ...} =>
		  if Var.originalName x = "safe"
		     then
			Dec.Val{pat = pat, tyvars = [],
				exp = Exp.Con(if !Control.safe
						 then Con.truee
					      else Con.falsee)}
			:: ds
		  else error()
	     | _ => error()
	 end
    | _ => error()
   end

fun preprocess(ins: In.t, B: Basis.t, accum: Accum.t): Basis.t * Accum.t =
   let val _ = Control.reset()
      val ast =
	 Ref.fluidLet
	 (Ast.isInfix, fn _ => false,
	  fn () => Control.pass{name = "parse",
				suffix = "ast",
				style = Control.ML,
				display = Control.Layout Ast.Program.layout,
				thunk = fn () => FrontEnd.lexAndParse ins})
   in Control.message(fn () =>
		      let open Layout
		      in seq[str "size = ",
			     Int.layout(Ast.Program.size ast)]
		      end)
      ; (Control.pass
	 {name = "elaborate",
	  suffix = "core-ml",
	  style = Control.ML,
	  thunk = fn () => Elaborate.elaborateProgram(ast, B, accum),
	  display =
	  Control.Layout
	  (fn (_, A) =>
	   CoreML.Program.layout
	   (CoreML.Program.T{constructors = Constructors.empty,
			     decs = Decs.toList(Accum.decs A)}))})
   end

structure PrimEnv = PrimEnv(open Elaborate
			    val preprocess = preprocess)

(* raised to stop the compilation at some intermediate pass *)   
exception Done

fun process(ins: In.t, suffix: Decs.t, A: Accum.t, constructors,
	    out: Out.t): unit =
   let
      val decs = Accum.decs A
      fun partial() =
	 CoreML.Program.T
	 {decs = Decs.toList(Accum.decs(#2(preprocess(ins, Accum.basis A, A)))),
	  constructors = constructors}
      val coreml =
	 case Control.getMode() of
	    Control.Defunctorize => partial()
(*	  | Control.Monomorphise => partial() *)
	  | Control.Compile =>
	       let val basis = patchDecsSafe decs
		  val user =
		     Decs.toList
		     (Decs.append(Accum.decs
				  (#2(preprocess(ins, Accum.basis A, Accum.empty))),
				  suffix))
		  val basis = DeadCode.deadCode{basis = basis, user = user}
	       in CoreML.Program.T
		  {constructors = Constructors.empty,
		   decs = Constructors.toDecs constructors @ basis @ user}
	       end
      val _ = (Control.message(fn () => CoreML.Program.layoutStats coreml);
	       (case Control.getMode() of
		   Control.Defunctorize =>
		      (Layout.output(Layout.align
				     [Control.header Control.ML,
				      CoreML.Program.layout coreml],
				     out);
		       raise Done)
		 | _ => ()))
      val xml =
	 Control.passSimplify{name = "infer",
			      suffix = "xml",
			      style = Control.ML,
			      thunk = fn () => Infer.infer coreml,
			      display = Control.Layout Xml.Program.layout,
			      typeCheck = Xml.typeCheck,
			      simplify = Xml.simplify}
      val _ = Control.message(fn () => Xml.Program.layoutStats xml)
      val sxml =
	 Control.passSimplify{name = "mono",
			      suffix = "",
			      style = Control.ML,
			      thunk = fn () => Monomorphise.monomorphise xml,
			      display = Control.NoDisplay,
			      typeCheck = Sxml.typeCheck,
			      simplify = Sxml.simplify}
      val Sxml.Program.T{datatypes, ...} = sxml
   in
      Control.message(fn () => Sxml.Program.layoutStats sxml);
      case Control.getMode() of
	 Control.Defunctorize => Error.bug "shouldn't be here"
(*        | Control.Monomorphise =>
 * 	    (let open Layout
 * 	     in output(align
 * 		       [Control.header Control.ML,
 * 			str "local",
 * 			PrimEnv.getPatchRecordTypes(),
 * 			str "in val _ = ",
 * 			Sxml.Program.layout sxml,
 * 			str "end"], out)
 * 	     end;
 * 	     raise Done)
 *)
       | _ =>
	    let
	       val sxml =
		  Control.passSimplify
		  {name = "implement exceptions",
		   suffix = "sxml",
		   style = Control.ML,
		   thunk = fn () => ImplementExceptions.doit sxml,
		   typeCheck = Sxml.typeCheck,
		   display = Control.Layout Sxml.Program.layout,
		   simplify = Sxml.simplify}
	       val sxml =
		  Control.pass (* polyvariance has simplify built in *)
		  {name = "polyvariance",
		   suffix = "sxml.poly",
		   style = Control.ML,
		   thunk = fn () => Polyvariance.duplicate sxml,
		   display = Control.Layout Sxml.Program.layout}
 	       val _ = Control.message(fn () => Sxml.Program.layoutStats sxml)
	       val _ = Cps.Type.initialize()
	       val cps =
		  Control.passSimplify
		  {name = "closure convert",
		   suffix = "cps",
		   style = Control.No,
		   thunk = fn () => ClosureConvert.closureConvert sxml,
		   typeCheck = Cps.typeCheck,
		   simplify = Cps.simplify,
		   display = Control.Layouts Cps.Program.layouts}

	       val _ =
		  let open Control
		  in if !keepCps andalso not(!aux)
		     then Ref.fluidLet
			(aux, true, fn () =>
			 displays("cps", fn disp =>
				  Cps.Program.layouts(cps, disp)))
		     else ()
		  end

(* 	       val _ =
 * 		  Ref.fluidLet
 * 		  (Control.aux, true, fn () =>
 * 		   Control.displays("cps", fn disp =>
 * 				    Cps.Program.layouts(cps, disp)))
 *)

	       val machine =
		  Control.pass
		  {name = "backend",
		   suffix = "c",
		   style = Control.C,
		   display = Control.NoDisplay,
		   thunk = fn () => Backend.generate cps}
		  
	       val _ =
		  Control.traceCall
		  "emit C"
		  (fn () =>
		   (Layout.output(Control.header Control.C, out)
		    ; Out.newline out
		    ; Machine.Program.output{program = machine,
					     includes = !Control.includes,
					     print = fn s => Out.output(out, s)}
		    ))
		  ()
	    in Machine.reset()
	       ; Atoms.clearPrimitive()
	    end
   end

(*---------------------------------------------------*)
(*                    compileFile                    *)
(*---------------------------------------------------*)

fun compileFile{input: string, output: string}: unit =
   let val inputFile = File.fromString input
      fun partial accum =
	 let val (A, cs) = accum()
	 in Ast.isInfix := Env.toIsInfix(Basis.toEnv(Accum.basis A));
	    (A, Decs.empty, cs)
	 end
      (* need to do this before the Control.inputFile := *)
      val (accum, suffix, constructors) =
	 case Control.getMode() of
	    Control.Defunctorize => partial PrimEnv.defunctorize
(*	  | Control.Monomorphise => partial PrimEnv.monomorphise *)
	  | Control.Compile => PrimEnv.compile()
   in Control.inputFile := inputFile
      ; Control.messageLay(Control.header Control.No)
      ; (File.withIn
	 (inputFile, fn ins =>
	  (File.withOut
	   (File.fromString output, fn out =>
	    (process(ins, suffix, accum, constructors, out);
	     Control.message PropertyList.stats)
	    handle Done => ()))))
   end

val compileFile = Control.traceCall "compile" compileFile
   
end
