(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Control: CONTROL =
   struct
      local
	 val defaults = ref []
	 val settings = ref []
      in
	 fun setDefaults() = List.foreach(!defaults, fn f => f())

	 fun layoutSettings() =
	    let open Layout
	    in align(List.map(List.reverse(!settings),
			      fn (name, lay) => seq[name, lay()]))
	    end
	 
	 fun control{name, default, layout} =
	    let val r = ref default
	    in ListRef.push(settings, (Layout.str(name ^ ": "),
				       fn () => layout(! r)))
	       ; ListRef.push(defaults, fn () => r := default)
	       ; r
	    end
      end
      
      val aux = control{name = "aux",
			default = false,
			layout = Bool.layout}		     

      structure Chunk =
	 struct
	    datatype t =
	       OneChunk
	     | ChunkPerFunc
	     | Coalesce of {limit: int}

	    val toString =
	       fn OneChunk => "one chunk"
		| ChunkPerFunc => "chunk per function"
		| Coalesce{limit} => "coalesce " ^ Int.toString limit

	    val layout = Layout.str o toString
	 end

      datatype chunk = datatype Chunk.t
      val chunk = control{name = "chunk",
			  default = Coalesce{limit = 2000},
			  layout = Chunk.layout}

      val fixedHeap = control{name = "fixed heap",
			      default = NONE,
			      layout = Option.layout Int.layout}

      val flattenRounds = control{name = "flatten rounds",
				  default = 2,
				  layout = Int.layout}

      structure Inline =
	 struct
	    datatype t =
	       NonRecursive of {size: int option}
	     | Leaf of {size: int option}
	     | LeafNoLoop of {size: int option}
	     | Recursive of {size: int}

	    local open Layout
	       val iol = Option.layout Int.layout
	    in
	       val layout = 
		  fn NonRecursive{size} => seq[str "NonRecursive ", iol size]
		   | Leaf{size} => seq[str "Leaf ", iol size]
		   | LeafNoLoop{size} => seq[str "LeafNoLoop ", iol size]
		   | Recursive{size} =>
			seq[str "Recursive ", Int.layout size]
	    end
	 end

      val includes: string list ref =
	 control{name = "includes",
		 default = [],
		 layout = List.layout String.layout}
	 
      datatype inline = datatype Inline.t
      val layoutInline = Inline.layout

      val inline = control{name = "inline",
			   default = NonRecursive{size = SOME 50},
			   layout = Inline.layout}

      val inlineRounds = control{name = "inline rounds",
				 default = 1,
				 layout = Int.layout}

      fun setInlineSize(size: int): unit =
	 inline := (case !inline of
		       NonRecursive _ => NonRecursive{size = SOME size}
		     | Leaf _ => Leaf{size = SOME size}
		     | LeafNoLoop _ => LeafNoLoop{size = SOME size}
		     | Recursive _ => Recursive{size = size})
	 
      val inputFile = control{name = "input file",
			      default = File.fromString "bogus",
			      layout = File.layout}

      val instrument = control{name = "instrument",
			       default = false,
			       layout = Bool.layout}

      val instrumentSxml = control{name = "instrument Sxml",
				   default = false,
				   layout = Bool.layout}

      val keepCps = control{name = "keepCps",
			    default = false,
			    layout = Bool.layout}

      structure Match =
	 struct
	    datatype t = LeftToRight | Rotate

	    fun layout m =
	       Layout.str
	       (case m of
		   LeftToRight => "left to right"
		 | Rotate => "rotate")
	 end
      datatype match = datatype Match.t
      val match = control{name = "match",
			  default = LeftToRight,
			  layout = Match.layout}

      val messages = control{name = "messages",
			     default = false,
			     layout = Bool.layout}

      structure Mode =
	 struct
	    datatype t =
	       Defunctorize
(*	     | Monomorphise *)
	     | Compile
	       
	    val toString =
	       fn Defunctorize => "defunctorize"
(*		| Monomorphise => "monomorphise" *)
		| Compile => "compile"
		     
	    val layout = Layout.str o toString
	 end

      datatype mode = datatype Mode.t

      val mode = control{name = "mode",
			 default = Compile,
			 layout = Mode.layout}

      val polyvariance =
	 control{name = "polyvariance",
		 default = SOME{rounds = 2,
				small = 30,
				product = 300},
		 layout =
		 Option.layout
		 (fn {rounds, small, product} =>
		  Layout.record[("rounds", Int.layout rounds),
				("small", Int.layout small),
				("product", Int.layout product)])}

      val printAtFunEntry = control{name = "print at fun entry",
				    default = false,
				    layout = Bool.layout}

      val profile = control{name = "profile",
			    default = false,
			    layout = Bool.layout}

      val safe = control{name = "safe",
			 default = true,
			 layout = Bool.layout}

      val typeCheck = control{name = "type check",
			      default = true,
			      layout = Bool.layout}
	 
      fun setTypeCheck b =
	 case (b, !mode) of
(* 	    (true, Monomorphise) =>
 * 	       Error.bug "can't type check when monomorphising"
 *)
	    _ => typeCheck := b
      fun shouldTypeCheck() = !typeCheck

      fun setMode m =
 	 case m of
(* 	    Monomorphise => Error.unimplemented "Monomorphise"
 *)
	   _ => (mode := m
		  ; typeCheck := (case m of
				     Defunctorize => false
(*				   | Monomorphise => false *)
				   | Compile => !typeCheck))
      fun getMode() = !mode

      val version = Layout.str "1999-7-12"

      datatype style = No | C | ML

      fun preSuf style =
	 let
	    val (p, s) =
	       case style of
		  No => ("", "")
		| C => ("/*", "*/")
		| ML => ("(*", "*)")
	 in (Layout.str p, Layout.str s)
	 end
	 
      fun header commentStyle =
	 let
	    val (pre, suf) = preSuf commentStyle
	    open Layout
	 in align[pre,
		  seq[str "MLton ", version,
		      str " created this file on ", Date.layout(Date.now()),
		      str "."],
		  seq[str "Do not edit this file."],
		  str "Flag settings: ",
		  indent(layoutSettings(), 3),
		  suf]
	 end

      fun on name =
	 (Trace.Immediate.on[name]
	  ; Trace.Time.on[name])
	 
      fun traceCall name =
	 (on name
	  ; Trace.traceCall name)

      fun message(th: unit -> Layout.t): unit =
	 if !messages
	    then Trace.Immediate.message(th())
	 else ()

      fun messageLay(l: Layout.t): unit = message(fn () => l)

      fun messageStr(s: string): unit =
	 message(fn () => Layout.str s)

      fun messagesAreOn() = !messages
	 
      fun messagesOn() =
	 (messages := true
	  ; Trace.Time.flagged()
	  ; Trace.Immediate.flagged())

      fun messagesOff() =
	 (messages := false
	  ; Trace.Time.never()
	  ; Trace.Immediate.never())

      fun defaults() = (setDefaults(); messagesOff())

      val _ = defaults()
	 
      fun displays(name: string, thunk: (Layout.t -> unit) -> unit): unit =
	 traceCall "display"
	 File.withOut(File.fromString
		      (concat[String.dropSuffix(File.toString(!inputFile),
						4),
			      ".",
			      name]),
		      fn out =>
		      thunk(fn l => (Layout.output(l, out)
				     ; Out.newline out)))

      val displays = fn arg => if !aux then displays arg else ()
		      
      fun display(name: string, thunk: (unit -> Layout.t)): unit =
	 displays(name, fn disp => disp(traceCall "layout" thunk ()))

      (*------------------------------------*)
      (*               Errors               *)
      (*------------------------------------*)

      val numErrors = ref 0

      val errorThreshhold = ref 20
	 
      fun anyErrors() = !numErrors > 0

      fun error(r: Region.t, msg: Layout.t): unit =
	 (IntRef.inc numErrors
	  ; Out.output(Out.error,
		       Layout.toString(let open Layout
				       in seq[File.layout(!inputFile),
					      str ":",
					      Region.layout r,
					      str " Error: ",
					      msg]
				       end))
	  ; Out.newline Out.error
	  ; if !numErrors = !errorThreshhold
	       then Error.error "compilation aborted: too many errors"
	    else ())

      fun errorStr(r, msg) = error(r, Layout.str msg)

      fun reset() = (Region.reset(); numErrors := 0)

      (*---------------------------------------------------*)
      (*                  Compiler Passes                  *)
      (*---------------------------------------------------*)

      datatype 'a display =
	 NoDisplay
       | Layout of 'a -> Layout.t
       | Layouts of 'a * (Layout.t -> unit) -> unit

      fun 'a sizeMessage(name: string, a: 'a): Layout.t =
	 let open Layout
	 in str(concat[name, " size is ",
		       Int.toCommaString(MLton.size a), " bytes"])
	 end

      fun pass{name: string,
	       suffix: string,
	       style: style,
	       display = disp,
	       thunk: unit -> 'a}: 'a =
	 let
	    val result = traceCall name thunk ()
	 in
	    message(fn () => sizeMessage(suffix, result))
	       
	    ; if anyErrors()
		 then Error.error("compilation aborted: " ^ name
				  ^ " reported errors")
	      else ()

	    ; (case disp of
		  NoDisplay => ()
		| Layout layout =>
		     display(suffix, fn () => Layout.align[header style,
							   layout result])
		| Layouts layout =>
		     displays(suffix,
			      fn output =>
			      (output(header style)
			       ; layout(result, output))))
	    ; result
	 end

      fun passTypeCheck{name: string,
			suffix: string,
			style: style,
			display,
			thunk: unit -> 'a,
			typeCheck = tc: 'a -> unit}: 'a =
	 let val result = pass{name = name,
			       suffix = suffix,
			       display = display,
			       style = style,
			       thunk = thunk}
	 in if !typeCheck
	       then traceCall "typeCheck" tc result
	    else ()
	    ; result
	 end
      
      fun passSimplify{name, suffix, style, thunk, display, typeCheck, simplify} =
	 let
	    val result =
	       passTypeCheck{name = name,
			     suffix = suffix ^ ".unsimplified",
			     style = style,
			     thunk = thunk,
			     display = NoDisplay,
			     typeCheck = typeCheck}
	 in passTypeCheck{name = name ^ " simplify",
			  suffix = suffix,
			  style = style,
			  thunk = fn () => simplify result,
			  display = display,
			  typeCheck = typeCheck}
	 end
   end
