(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CORE_ML_STRUCTS = 
   sig
      include ATOMS
      structure Type: TYPE
      sharing Type = Prim.Type
   end

signature CORE_ML = 
   sig
      include CORE_ML_STRUCTS

      structure Pat :
	 sig
	    datatype t =
	       Wild
	     | Var of Var.t
	     | Const of Ast.Const.t
	     | Con of {
		       con: Con.t,
		       arg: t option
		      }
	     | Record of {
			  flexible: bool,
			  record: t Record.t
			 }
	     | Constraint of t * Type.t
	     | Layered of Var.t * t

	    (* constants *)
	    val unit: t
	       
	    (* constructors *)
	    val list: t list -> t
	    val tuple: t list -> t
	    val record: {
			 flexible: bool,
			 record: t Record.t
			} -> t

	    (* a list (without duplicates) of variables occurring in a pattern *)
	    val vars: t -> Var.t list

	    (* replace all variables with Wild *)
	    val removeVars: t -> t

	    (* removeOthersReplace(pat, old,new) replaces all variables in pat
	     * with Wild, except for old, which it replaces with new
	     *)
	    val removeOthersReplace: t * Var.t * Var.t -> t

	    (* true if the pattern contains a constant, constructor or variable *)
	    val isRefutable: t -> bool

	    val foreachVar: t * (Var.t -> unit) -> unit

	    (* conversion to Ast *)
	    val toAst: t -> Ast.Pat.t

	    val layout: t -> Layout.t
	 end

      structure Exp :
	 sig
	    type dec
	    datatype t =
	       Const of Ast.Const.t
	     | Var of Var.t
	     | Prim of Prim.t
	     | Con of Con.t
	     | Record of t Record.t
	     | Fn of (Pat.t * t) list
	     | App of t * t
	     | Let of dec list * t
	     | Constraint of t * Type.t
	     | Handle of t * (Pat.t * t) list
	     | Raise of t

	    (* macros *)
	    val selector: Record.Field.t -> t
	    val iff: t * t * t -> t
	    val andAlso: t * t -> t
	    val orElse: t * t -> t
	    val whilee: {test: t, expr: t} -> t
	    val casee: t * (Pat.t * t) list -> t

	    (* standard constructions *)
	    val unit: t
	    val seq: t list -> t
	    val list: t list -> t
	    val tuple: t list -> t
	    val delay: t -> t
	    val force: t -> t
	    val lambda: Var.t * t -> t
	    val compose: unit -> t

	    val foreachVar: t * (Var.t -> unit) -> unit
	       
	    (* true if the expression may side-effect. See p 19 of Definition *)
	    val isExpansive: t -> bool
	       
	    val layout: t -> Layout.t
	 end

      structure Dec :
	 sig
	    datatype t =
	       Val of {pat: Pat.t,
		       tyvars: Tyvar.t list,
		       exp: Exp.t}
	     | Fun of {
		       tyvars: Tyvar.t list,
		       decs: {
			      var: Var.t,
			      ty: Type.t option,
			      rules: (Pat.t * Exp.t) list
			     } list
		      }
	     | Datatype of {
			    tyvars: Tyvar.t list,
			    tycon: Tycon.t,
			    cons: {
				   con: Con.t,
				   arg: Type.t option
				  } list
			   } list
	     | Exception of {
			     con: Con.t,
			     arg: Type.t option
			    }
	     | Overload of {
			    var: Var.t,
			    scheme: Scheme.t,
			    ovlds: Var.t list
			   }

	    val toAst: t -> Ast.Dec.t
	    val layout: t -> Layout.t

	    val isExpansive: t -> bool
	 end
      where type t = Exp.dec

      structure Constructors:
	 sig
	    datatype t =
	       T of {
		     exceptions: {
				  con: Con.t,
				  arg: Type.t option
				  } list,
		     datatypes: {
				 tycon: Tycon.t,
				 tyvars: Tyvar.t list,
				 cons: {
					con: Con.t,
					arg: Type.t option
					} list
				 } list
		     }

	    val empty: t
	    val + : t * t -> t
	    val fromDecs: Dec.t list -> t
	    val toDecs: t -> Dec.t list
	    val layout: t -> Layout.t
	 end

      structure Program :
	 sig
	    datatype t = T of {
			       constructors: Constructors.t,
			       decs: Dec.t list
			      }

	    val layoutStats: t -> Layout.t

(*	    val empty: t *)
(*	    val append: t * t -> t *)

	    val layout: t -> Layout.t
	 end
   end
