(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Analyze(S: ANALYZE_STRUCTS): ANALYZE = 
struct

open S
open Dec PrimExp Transfer

fun 'a analyze
   {program = Program.T{main, datatypes, globals, functions},
    useFromTypeOnBinds,
    fromType, copy, coerce,
    const,
    tuple, select,
    constructors,
    primApp,
    filter} =
   let
      fun coerces(from, to) =
	 List.foreach2(from, to, fn (from, to) =>
		       coerce{from = from, to = to})

      val {conApp, conSelect, conArgs} =
	 case constructors of
	    SOME{conApp, conSelect} =>
	       {conApp = conApp,
		conSelect = conSelect,
		conArgs = fn _ => Error.bug "conArgs not defined"}
	  | NONE =>
	       let
		  val {get = conInfo, set = setConInfo} =
		     Property.new
		     (Con.plist, Property.initRaise("info", Con.layout))
		  val _ =
		     List.foreach
		     (datatypes, fn {tycon, cons} =>
		      let val v = fromType(Type.con(tycon, []))
		      in List.foreach
			 (cons, fn {con, args} =>
			  setConInfo(con, {value = v,
					   argTypes = args,
					   args = List.map(args, fromType)}))
		      end)
		  fun conApp{con, args} =
		     let val {args = args', value, ...} = conInfo con
		     in coerces(args, args');
			value
		     end
		  fun conArgs c = #args(conInfo c)
		  fun conSelect{variant, con, offset} =
		     List.nth(conArgs con, offset)
	       in {conApp = conApp,
		   conSelect = conSelect,
		   conArgs = conArgs}
	       end

      val boolVal = fromType Type.bool

      val {get = value: Var.t -> 'a, set = setValue} =
	 Property.new
	 (Var.plist, Property.initRaise("analyze var value", Var.layout))

      fun values xs = List.map(xs, value)

      val {get = func, set = setFunc} =
	 Property.new(Func.plist,
		      Property.initRaise("analyze func name", Func.layout))

      val {get = jump, set = setJump} =
	 Property.new(Jump.plist, Property.initRaise("analyze jump", Jump.layout))

      fun loopArgs args =
	 List.map(args, fn (x, t) =>
		  let val v = fromType t
		  in setValue(x, v);
		     v
		  end)

      val _ =
	 List.foreach
	 (functions, fn {name, args, returns, ...} =>
	  setFunc(name, {args = loopArgs args,
			 returns = List.map(returns, fromType)}))

      val exnVals: 'a list option ref = ref NONE

      fun getExnVals vs =
	 case !exnVals of
	    NONE => let val vs = List.map(vs, copy)
		    in exnVals := SOME vs; vs
		    end
	  | SOME vs => vs

      fun loopTransfer(t, shouldReturns): unit =
	 case t of
	    Call{func = f, args, cont = c} =>
	       let val {args = formals, returns} = func f
		  val shouldReturns =
		     case c of
			NONE => shouldReturns
		      | SOME c => jump c
	       in coerces(values args, formals);
		  coerces(returns, shouldReturns)
	       end
	  | Jump{dst, args} => coerces(values args, jump dst)
	  | Return xs => coerces(values xs, shouldReturns)
	  | Case{test, cases, default, ...} =>
	       let val test = value test 
	       in List.foreach
		  (cases, fn (c, j) =>
		   let val test = filter(test, c)
		   in case jump j of
		      [] => ()
		    | [v] => coerce{from = test, to = v}
		    | _ => Error.bug "case label must be nullary or unary"
		   end)
		  ; (case default of
			NONE => ()
		      | SOME j =>
			   case jump j of
			      [] => ()
			    | _ => Error.error "default label must be nullary")
	       end
	  | Raise xs => let val vs = values xs
			in coerces(vs, getExnVals vs)
			end
	  | Halt => ()
	  | Bug => ()

(*       val loopTransfer =
 * 	 Trace.trace2("Analyze.loopTransfer",
 * 		      Transfer.layout, Layout.ignore, Layout.ignore) loopTransfer
 *)

      fun loopPrimExp{var: Var.t, ty: Type.t, exp: PrimExp.t}: 'a =
	 case exp of
	    Const c => const c
	  | Var x => value x
	  | Tuple[_] => Error.bug "unary tuple"
	  | Tuple xs => tuple(values xs)
	  | Select{tuple, offset} =>
	       select{tuple = value tuple,
		      offset = offset,
		      resultType = ty}
	  | ConSelect{variant, con, offset} =>
	       conSelect{variant = value variant,
			 con = con,
			 offset = offset}
	  | ConApp{con, args} => conApp{con = con, args = values args}
	  | PrimApp{prim, targs, args} =>
	       primApp{prim = prim,
		       targs = targs,
		       args = values args,
		       resultVar = var,
		       resultType = ty}
      
(*       val loopPrimExp =
 * 	 Trace.trace("Analyze.loopPrimExp",
 * 		     fn {var, ty, exp, ...} =>
 * 		     Layout.record[("var", Var.layout var),
 * 				   ("ty", Type.layout ty),
 * 				   ("exp", PrimExp.layout exp)],
 * 		     Layout.ignore)
 * 	 loopPrimExp
 *)

      fun loopBind{var, exp, ty}: unit =
	 setValue(var,
		  let val v = loopPrimExp{var = var, ty = ty, exp = exp}
		  in if useFromTypeOnBinds
			then let val v' = fromType ty
			     in coerce{from = v, to = v'};
				v'
			     end
		     else v
		  end)
	       
      fun loopExp(e: Exp.t, returns): unit =
	 let val {decs, transfer, ...} = Exp.dest e
	 in loopDecs(decs, returns);
	    loopTransfer(transfer, returns)
	 end
      and loopDecs(ds, returns): unit =
	 List.foreach
	 (ds,
	  fn Bind b => loopBind b
	   | Fun{name, args, body} =>
		(setJump(name, loopArgs args)
		 ; loopExp(body, returns))
	   | HandlerPush h => let val vs = jump h
			      in coerces(getExnVals vs, vs)
			      end
	   | _ => ()
		)

      val _ = coerces([], #args(func main))
   in
      List.foreach(globals, loopBind);
      List.foreach(functions, fn {name, body, ...} =>
		   loopExp(body, #returns(func name)));
      
      {
       value = value,
       conArgs = conArgs,
       func = func,
       jump = jump,
       exnVals = !exnVals
      }
   end

end
