(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Invariant: Created globals only refer to other globals.
 *            Hence, the newly created globals may appear at the
 *            beginning of the program.
 *
 * Circular abstract values can arise as a result of programs like:
 *   datatype t = T of t
 *   fun f() = T(f())
 *   val _ = f()
 * There is special code in printing abstract values and in determining whether
 * they are global in order to avoid infinite loops.
 *)
functor ConstantPropagation(S: CONSTANT_PROPAGATION_STRUCTS)
   : CONSTANT_PROPAGATION = 
struct

open S

structure Once = Once(S)
structure Global = Global(S)

structure Type =
   struct
      open Type

      fun isSmall t =
	 case dest t of
	    Array _ => false
	  | Vector _ => false
	  | Datatype _ => false
	  | Variant _ => false
	  | Ref t => isSmall t
	  | Tuple ts => List.forall(ts, isSmall)
	  | _ => true
   end
   
open PrimExp Dec Transfer

structure Value =
   struct
      datatype global =
	 NotComputed
       | No
       | Yes of Var.t

      structure Const =
	 struct
	    datatype t = T of {const: const ref,
			       coercedTo: t list ref}
	    and const =
	       Undefined (* no possible value *)
	      | Const of Const.t
	      | Prim of {prim: Prim.t, targs: Type.t list, ty: Type.t}
	      | Unknown (* many possible values *)

	    fun new c = T{const = ref c,
			  coercedTo = ref []}

	    fun equals(T{const = r, ...}, T{const = r', ...}) = r = r'

	    val const = new o Const
	    val prim = new o Prim

	    fun undefined() = new Undefined
	    fun unknown() = new Unknown

	    fun layout(T{const, ...}) =
	       let open Layout
	       in case !const of
		  Undefined => str "undefined constant"
		| Const c => Const.layout c
		| Prim{prim, ...} => Prim.layout prim
		| Unknown => str "unknown constant"
	       end
	    
	    fun makeUnknown(T{const, coercedTo}): unit =
	       case !const of
		  Unknown => ()
		| _ => (const := Unknown;
			List.foreach(!coercedTo, makeUnknown);
			coercedTo := [])

	    fun send(c: t, c': const): unit =
	       let
		  fun loop(c as T{const, coercedTo}) =
		     case (c', !const) of
			(_, Unknown) => ()
		      | (_, Undefined) => (const := c';
					   List.foreach(!coercedTo, loop))
		      | (Const c', Const c'') =>
			   if Const.equals(c', c'')
			      then ()
			   else makeUnknown c
		      | (Prim{prim = p', targs = ts', ...},
			 Prim{prim = p'', targs = ts'', ...}) =>
			   if Prim.equals(p', p'')
			      andalso List.equals(ts', ts'', Type.equals)
			      then ()
			   else makeUnknown c
		      | _ => makeUnknown c
	       in loop c
	       end

	    fun coerce{from = from as T{const, coercedTo}, to: t}: unit =
	       if equals(from, to)
		  then ()
	       else
		  let fun push() = ListRef.push(coercedTo, to)
		  in case !const of
		     Unknown => makeUnknown to
		   | Undefined => push()
		   | c as Const _ => (push(); send(to, c))
		   | p as Prim _ => (push(); send(to, p))
		  end

	    fun unify(c, c') =
	       (coerce{from = c, to = c'};
		coerce{from = c', to = c})
	 end

      structure One =
	 struct
	    datatype 'a t = T of {global: Var.t option ref,
				  extra: 'a}

	    fun new(a: 'a): 'a t = T{global = ref NONE,
				     extra = a}

	    fun global(T{global = g, ...}) = g
	       
	    fun equals(n, n') = global n = global n'
	 end
      
      structure Place =
	 struct
	    datatype 'a t =
	       Undefined
	     | One of 'a One.t
	     | Unknown

	    val toString =
	       fn Undefined => "Undefined"
		| One _ => "One"
		| Unknown => "Unknown"

	    fun layout b = Layout.str(toString b)
	 end

      structure Birth =
	 struct
	    datatype 'a t = T of {place: 'a Place.t ref,
				  coercedTo: 'a t list ref}

	    fun layout(T{place, ...}) = Place.layout(!place)

	    fun equals(T{place = r, ...}, T{place = r', ...}) = r = r'
	    fun new p = T{place = ref p,
			  coercedTo = ref []}

	    fun undefined(): 'a t = new Place.Undefined
	    fun unknown(): 'a t = new Place.Unknown
	    fun here(a: 'a): 'a t = new(Place.One(One.new a))

	    fun makeUnknown(T{place, coercedTo, ...}) =
	       case !place of
		  Place.Unknown => ()
		| _ => (place := Place.Unknown;
			List.foreach(!coercedTo, makeUnknown);
			coercedTo := [])

	    fun send(b: 'a t, one): unit =
	       let
		  fun loop(b as T {place, coercedTo, ...}) =
		     case !place of
			Place.Undefined => (place := Place.One one;
					    List.foreach(!coercedTo, loop))
		      | Place.One one' => if One.equals(one, one')
					     then ()
					  else makeUnknown b
		      | Place.Unknown => ()
	       in loop b
	       end

	    fun coerce{from = from as T{place, coercedTo, ...}, to} =
	       if equals(from, to)
		  then ()
	       else
		  let fun push() = ListRef.push(coercedTo, to)
		  in case !place of
		     Place.Unknown => makeUnknown to
		   | Place.One one => (push(); send(to, one))
		   | Place.Undefined => push()
		  end

	    fun unify(c, c') =
	       (coerce{from = c, to = c'};
		coerce{from = c', to = c})
	 end

      structure Set = DisjointSet
	 
      datatype t =
	 T of {
	       value: value,
	       ty: Type.t,
	       global: global ref
	      } Set.t
      and value =
          Const of Const.t
	| Datatype of data
	| Vector of {length: t,
		     elt: t}
	| Ref of {birth: {init: t} Birth.t,
		  arg: t}
	| Array of {birth: unit Birth.t,
		    length: t,
		    elt: t}
	| Tuple of t list
      and data =
	 Data of {
		  value: dataVal ref,
		  coercedTo: data list ref,
		  selects: {
			    con: Con.t,
			    offset: int,
			    to: t
			   } list ref
		 }
      and dataVal =
	 Undefined
	| ConApp of {con: Con.t,
		     args: t list,
		     uniq: unit ref}
	| Unknown

      local
	 fun make sel (T s) = sel(Set.value s)
      in val value = make #value
	 val global = make #global
	 val ty = make #ty
      end

      fun equals(T s, T s') = Set.equals(s, s')

      fun layout v =
	 let open Layout
	 in case value v of
	    Const c => Const.layout c
	  | Datatype(Data{value, ...}) =>
	       (case !value of
		   Undefined => str "undefined datatype"
		 | ConApp{con, args, ...} => Con.layout con
		      (* Can't layout the args because there may be circularity *)
		      (* tuple(List.map(args, layout)) *)
		 | Unknown => str "unknown datatype")
	  | Ref{birth, arg, ...} => seq[str "ref", tuple[Birth.layout birth,
							 layout arg]]
	  | Array{birth, length, elt, ...} => seq[str "array",
						  tuple[Birth.layout birth,
							layout length,
							layout elt]]
	  | Vector{length, elt, ...} => seq[str "vector",
					   tuple[layout length,
						 layout elt]]
	  | Tuple vs => tuple(List.map(vs, layout))
	 end

      fun globals(vs: t list, newGlobal): (Var.t * Type.t) list option =
	 let
	    fun loop(vs, gs) =
	       case vs of
		  [] => SOME(List.reverse gs)
		| v :: vs =>
		     (case global(v, newGlobal) of
			 SOME g => loop(vs, g :: gs)
		       | NONE => NONE)
	 in loop(vs, [])
	 end

      and global(v as T s, newGlobal): (Var.t * Type.t) option =
	 let val {global = r, ty, value} = Set.value s
	 in case !r of
	    No => NONE
	  | Yes g => SOME(g, ty)
	  | NotComputed =>
	       let
		  (* avoid globalizing circular abstract values *)
		  val _ = r := No
		  fun yes e = Yes(newGlobal(ty, e))
		  fun unary(Birth.T{place, ...},
			    makeInit: 'a -> t,
			    prim: Prim.t,
			    targ: Type.t) =
		     case !place of
			Place.One(One.T{global = glob, extra}) =>
			   let val init = makeInit extra
			   in case global(init, newGlobal) of
			      SOME(x, _) =>
				 Yes(case !glob of
					NONE => 
					   let
					      val exp =
						 PrimExp.PrimApp{prim = prim,
								 targs = [targ],
								 args = [x]}
					      val g = newGlobal(ty, exp)
					   in glob := SOME g; g
					   end
				      | SOME g => g)
			    | _ => No
			   end
		      | _ => No
		  val g =
		     case value of
			Const(Const.T{const, ...}) =>
			   (case !const of
			       Const.Const c => yes(PrimExp.Const c)
			     | Const.Prim{prim, targs, ...} =>
				  yes(PrimExp.PrimApp{prim = prim,
						      targs = targs,
						      args = []})
			     | _ => No)
		      | Datatype(Data{value, ...}) =>
			   (case !value of
			       ConApp{con, args, ...} =>
				  (case globals(args, newGlobal) of
				      NONE => No
				    | SOME args =>
					 yes(PrimExp.ConApp
					     {con = con,
					      args = List.map(args, #1)}))
			     | _ => No)
		      | Ref{birth, ...} =>
			   unary(birth, fn {init} => init,
				 Prim.reff, Type.deref ty)
		      | Array{birth, length, ...} =>
			   unary(birth, fn _ => length,
				 Prim.array, Type.dearray ty)
		      | Vector _ => No
		      | Tuple vs =>
			   (case globals(vs, newGlobal) of
			       NONE => No
			     | SOME xts =>
				  yes(PrimExp.Tuple(List.map(xts, #1))))
	       in r := g; global(v, newGlobal)
	       end
	 end
      
      fun new(v: value, ty: Type.t): t =
	 T(Set.singleton{value = v,
			 ty = ty,
			 global = ref NotComputed})


      fun tuple vs = new(Tuple vs, Type.tuple(List.map(vs, ty)))

      fun const c = new(Const(Const.const c), Type.ofConst c)

      fun deconst v =
	 case value v of
	    Const c => c
	  | _ => Error.bug "deconst"
	       
      local
	 fun make (err, sel) v =
	    case value v of
	       Vector fs => sel fs
	     | _ => Error.bug err
      in val devector = make("devector", #elt)
	 val vectorLength = make("vectorLength", #length)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Array fs => sel fs
	     | _ => Error.bug err
      in val dearray = make("dearray", #elt)
	 val arrayLength = make("arrayLength", #length)
	 val arrayBirth = make("arrayBirth", #birth)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Ref fs => sel fs
	     | _ => Error.bug err
      in val deref = make("deref", #arg)
	 val refBirth = make("refBirth", #birth)
      end

      structure Data =
	 struct
	    datatype t = datatype data

	    local
	       fun make v () = Data{value = ref v,
				    coercedTo = ref [],
				    selects = ref []}
	    in val undefined = make Undefined
	       val unknown = make Unknown
	    end
	 end
      
      local
	 (* The extra birth is because of let-style polymorphism.
	  * arrayBirth is really the same as refBirth.
	  *)
	 fun make(const, data, refBirth, arrayBirth) =
	    let
	       fun loop(t: Type.t): t =
		  new
		  (case Type.dest t of
		      Type.Datatype _ => Datatype(data())
		    | Type.Variant _ => Datatype(data())
		    | Type.Ref t => Ref{birth = refBirth(),
					arg = loop t}
		    | Type.Array t => Array{birth = arrayBirth(),
					    length = loop Type.int,
					    elt = loop t}
		    | Type.Vector t => Vector{length = loop Type.int,
					      elt = loop t}
		    | Type.Tuple ts => Tuple(List.map(ts, loop))
		    | _ => Const(const()),
		   t)
	    in loop
	    end
      in
	 val fromType =
	    make(Const.undefined, Data.undefined, Birth.undefined, Birth.undefined)
	 val unknown =
	    make(Const.unknown, Data.unknown, Birth.unknown, Birth.unknown)
      end

      fun select{tuple, offset, resultType} =
	 case value tuple of
	   Tuple vs => List.nth(vs, offset)
	 | _ => Error.bug "select of non-tuple" 

      fun filter(v: t, c: Con.t): t = v

      fun unit() = tuple[]
   end

(* ------------------------------------------------- *)
(*                     simplify                      *)
(* ------------------------------------------------- *)

fun simplify(program as Program.T{datatypes, globals, functions, main})
   : Program.t =
   let
      val once = Once.once program

      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("args", Con.layout))

      val conArgTypes = #types o conInfo
      val conArgValues = #values o conInfo
      val conResult = #result o conInfo
	 
      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  let val result = Type.con(tycon, [])
	  in List.foreach
	     (cons, fn {con, args} =>
	      setConInfo(con, {result = result,
			       types = args,
			       values = List.map(args, Value.fromType)}))
	  end)
	 
      local open Value
      in
(* 	 val traceCoerce =
 * 	    Trace.trace("Value.coerce",
 * 			fn {from, to} => Layout.record[("from", layout from),
 * 						       ("to", layout to)],
 * 			Unit.layout)
 *)
	    
	 fun makeDataUnknown(Data{value, coercedTo, selects, ...}) =
	    let
	       fun doit() =
		  (value := Unknown;
		   List.foreach(!coercedTo, makeDataUnknown);
		   coercedTo := [];
		   List.foreach(!selects, fn {con, offset, to} =>
				coerce{to = to,
				       from = List.nth(conArgValues con, offset)}))
	    in case !value of
	       Unknown => ()
	     | Undefined => doit()
	     | ConApp _ => doit()
	    end

	 and sendConApp(d: data, ca as {con, args, uniq}): unit =
	    let
	       val v = ConApp ca
	       fun loop(Data{value, coercedTo, selects, ...}): unit =
		  case !value of
		     Unknown => ()
		   | Undefined =>
			(value := v;
			 List.foreach(!coercedTo, loop);
			 List.foreach
			 (!selects, fn {con = con', offset, to} =>
			  if Con.equals(con, con')
			     then coerce{from = List.nth(args, offset),
					 to = to}
			  else ()))
		   | ConApp{con = con', uniq = uniq', ...} =>
			 if uniq = uniq'
			    orelse (Con.equals(con, con')
				    andalso List.isEmpty args)
			    then ()
			 else makeDataUnknown d
	    in loop d
	    end
				
	 and coerces{froms, tos} =
	    List.foreach2(froms, tos, fn (from, to) =>
			  coerce{from = from, to = to})

	 and coerce arg =
(*	    traceCoerce *)
	    (fn {from, to} =>
	    if equals(from, to)
	       then ()
	    else
	       case (value from, value to) of
		  (Const from, Const to) => Const.coerce{from = from, to = to}
		| (Datatype from, Datatype to) => coerceData{from = from, to = to}
		| (Ref{birth, arg}, Ref{birth = b', arg = a'}) =>
		     (Birth.coerce{from = birth, to = b'};
		      unify(arg, a'))
		| (Array{birth = b, length = n, elt = x},
		   Array{birth = b', length = n', elt = x'}) =>
		     (Birth.coerce{from = b, to = b'};
		      coerce{from = n, to = n'};
		      unify(x, x'))
		| (Vector{length = n, elt = x}, Vector{length = n', elt = x'}) =>
		     (coerce{from = n, to = n'};
		      coerce{from = x, to = x'})
		| (Tuple vs, Tuple vs') => coerces{froms = vs, tos = vs'}
		| _ => Error.bug "strange coerce") arg

	 and unify(T s: t, T s': t): unit =
	    if Set.equals(s, s')
	       then ()
	    else
	       let val {value, ...} = Set.value s
		  val {value = value', ...} = Set.value s'
	       in Set.union(s, s');
		  case (value, value') of
		     (Const c, Const c') => Const.unify(c, c')
		   | (Datatype d, Datatype d') => unifyData(d, d')
		   | (Ref{birth, arg}, Ref{birth = b', arg = a'}) =>
			(Birth.unify(birth, b');
			 unify(arg, a'))
		   | (Array{birth = b, length = n, elt = x},
		      Array{birth = b', length = n', elt = x'}) =>
			(Birth.unify(b, b');
			 unify(n, n');
			 unify(x, x'))
		   | (Vector{length = n, elt = x}, Vector{length = n', elt = x'}) =>
			(unify(n, n');
			 unify(x, x'))
		   | (Tuple vs, Tuple vs') => List.foreach2(vs, vs', unify)
		   | _ => Error.bug "strange unify"
	       end

	 and unifyData(d, d') =
	    (coerceData{from = d, to = d'};
	     coerceData{from = d', to = d})

	 and coerceData{from = Data{value, coercedTo, ...}, to} =
	      case !value of
		 Undefined => ListRef.push(coercedTo, to)
	       | ConApp ca => (ListRef.push(coercedTo, to);
			       sendConApp(to, ca))
	       | Unknown => makeDataUnknown to

	 fun conApp{con: Con.t, args: t list}: t =
	    (coerces{froms = args, tos = conArgValues con};
	     new(Datatype
		 (Data{value = ref(ConApp{con = con, args = args, uniq = ref()}),
		       coercedTo = ref [],
		       selects = ref []}),
		 conResult con))

	 fun conSelect{variant, con, offset} =
	    case value variant of
	       Datatype(Data{value, selects, ...}) =>
		  let
		     fun new() =
			let val v = fromType(List.nth(conArgTypes con, offset))
			in ListRef.push(selects,
					{con = con, offset = offset, to = v});
			   v
			end
		  in case !value of
		     Undefined => new()
		   | Unknown => List.nth(conArgValues con, offset)
		   | ConApp{con = con', args, ...} =>
			let val v = new()
			in if Con.equals(con, con')
			      then coerce{from = List.nth(args, offset),
					  to = v}
			   else ();
			      v
			end
		  end
	     | _ => Error.bug "conSelect of non-datatype"

	 fun makeUnknown(v: t): unit =
	    case value v of
	       Const c => Const.makeUnknown c
	     | Datatype d => makeDataUnknown d
	     | Ref{arg, ...} => makeUnknown arg
	     | Array{length, elt, ...} => (makeUnknown length;
					   makeUnknown elt)
	     | Vector{length, elt} => (makeUnknown length;
				       makeUnknown elt)
	     | Tuple vs => List.foreach(vs, makeUnknown)

	 fun sideEffect(v: t): unit =
	    case value v of
	       Const _ => ()
	     | Datatype _ => ()
	     | Ref{arg, ...} => makeUnknown arg
	     | Array{elt, ...} => makeUnknown elt
	     | Vector{elt, ...} => makeUnknown elt
	     | Tuple vs => List.foreach(vs, sideEffect)
	    
	 fun primApp{prim, targs, args, resultVar, resultType}: t =
	    let
	       fun bear z =
		  if once resultVar andalso Type.isSmall resultType
		     then Birth.here z
		  else Birth.unknown()
	    in case Prim.numArgs prim of
	       NONE =>
		  (if (case Type.dest resultType of
			  Type.Datatype _ => true
			| Type.Array _ => true
			| Type.Vector _ => true
			| Type.Ref _ => true
			| Type.Tuple _ => true
			| Type.Variant _ => true
			| _ => false)
		      then unknown resultType
		   else new(Const(Const.prim{prim = prim,
					     targs = targs,
					     ty = resultType}),
			    resultType))
	     | SOME _ => 
		  let open Prim.Name
		  in case (Prim.name prim, targs, args) of
		     (Array, [_], [n]) =>
			let val a = fromType resultType
			in coerce{from = n, to = arrayLength a};
			   Birth.coerce{from = bear(), to = arrayBirth a};
			   a
			end
		   | (ArraySub, _, [a, _]) => dearray a
		   | (ArrayUpdate, _, [a, _, v]) =>
			(coerce{from = v, to = dearray a};
			 unit())
		   | (ArrayLength, _, [a]) => arrayLength a
		   | (VectorLength, _, [v]) => vectorLength v
		   | (VectorSub, _, [v, _]) => devector v
		   | (VectorFromArray, _, [a]) =>
			let val v = fromType resultType
			in coerce{from = dearray a, to = devector v};
			   v
			end
		   | (Ref, _, [v]) =>
			let val r = fromType resultType
			in coerce{from = v, to = deref r};
			   Birth.coerce{from = bear{init = v}, to = refBirth r};
			   r
			end
		   | (Deref, _, [r]) => deref r
		   | (Assign, _, [r, v]) =>
			(coerce{from = v, to = deref r}; unit())
		   | _ => (List.foreach(args, sideEffect);
			   unknown resultType)
		  end
	    end
      end

      val {value, ...} =
	 Control.traceCall "fixed point"
	 analyze{
		 program = program,
		 useFromTypeOnBinds = false,
		 fromType = Value.fromType,
		 copy = Value.fromType o Value.ty,
		 coerce = coerce,
		 const = Value.const,
		 tuple = Value.tuple,
		 select = Value.select,
		 constructors = SOME{conSelect = conSelect,
				     conApp = conApp},
		 primApp = primApp,
		 filter = Value.filter
		}

      val _ =
	 Control.displays
	 ("const", fn display =>
	  let open Layout
	  in display(str "Input program:\n")
	     ; Program.layouts(program, display)
	     ; display(str "\n\nConstants:")
	     ; (Program.foreachVar
		(program, fn (x, _) => display(seq[Var.layout x,
						   str " ",
						   Value.layout(value x)])))
	  end)

      (* Walk through the program
       *  - removing declarations whose rhs is constant
       *  - replacing variables whose value is constant with globals
       *  - building up the global decs
       *)

      val {new = newGlobal, all} = Global.make()

      fun replaceVar x =
	 case Value.global(value x, newGlobal) of
	    NONE => x
	  | SOME(g, _) => g

      fun replaceVars xs = List.map(xs, replaceVar)
	       	    
      fun loopBind{var, ty, exp} =
	 let
	    fun keep() =
	       SOME{var = var, ty = ty, exp = PrimExp.replaceVar(exp, replaceVar)}
	 in case (Value.global(value var, newGlobal), exp) of
	    (NONE, _) => keep()
	  | (_, PrimApp{prim, ...}) =>
	       if Prim.maySideEffect prim
		  then keep()
	       else NONE
	  | _ => NONE
	 end

      fun loopExp(e: Exp.t) : Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val transfer =
	       case transfer of
		  Call{func, args, cont} =>
		     Call{func = func, cont = cont, args = replaceVars args}
		| Jump{dst, args} => Jump{dst = dst, args = replaceVars args}
		| Return xs => Return(replaceVars xs)
		| Case{cause, test, cases, default} =>
		     Case{cause = cause, test = replaceVar test,
			  cases = cases, default = default}
		| Raise xs => Raise(replaceVars xs)
		| Halt => Halt
		| Bug => Bug
	    val decs = 
	       List.foldr
	       (decs, [], fn (d, ds) =>
		case d of
		   Bind b => (case loopBind b of
				 SOME b => Bind b :: ds
			       | NONE => ds)
		 | Fun{name, args, body} =>
		      Fun{name = name,
			  args = args,
			  body = loopExp body} :: ds
		 | d => d :: ds)
	 in Exp.make{transfer = transfer,
		     decs = decs}
	 end

      val shrinkExp = shrinkExp [] (* not sure what globals to put here *)
	 
      val functions =
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name, args = args, returns = returns,
		   body = shrinkExp(loopExp body)})

      val globals = List.reverse(List.fold(globals, [], fn (b, bs) =>
					   case loopBind b of
					      NONE => bs
					    | SOME b => b :: bs))
	 
      val globals = all() @ globals
	      	 
      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end

end
