(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * If a toplevel function is called
 *       1. from one place outside itself
 *   and 2. all calls to itself within itself are tail
 * Then that function can be made local at the point of its outer call.
 *)

functor Contify(S: CONTIFY_STRUCTS): CONTIFY = 
struct

open S
open Dec Transfer

structure Info =
   struct
      datatype t =
	 NoOuterCalls
       | OneOuterCall
       | NotCont

      val toString =
	 fn NoOuterCalls => "NoOuterCalls"
	  | OneOuterCall => "OneOuterCall"
	  | NotCont => "NotCont"

      val layout = Layout.str o toString
   end

fun contify(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = getCont, set = setCont} =
	 Property.new(Func.plist, Property.initConst Info.NoOuterCalls)

      fun outerCall f =
	 case getCont f of
	    Info.NoOuterCalls => setCont(f, Info.OneOuterCall)
	  | Info.OneOuterCall => setCont(f, Info.NotCont)
	  | Info.NotCont => ()

      fun notCont f = setCont(f, Info.NotCont)

      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  Exp.foreachCall
	  (body, fn {func, cont, ...} =>
	   if Func.equals(name, func)
	      then (case cont of
		       NONE => ()
		     | SOME _ => notCont name)
	   else outerCall func))
	 
      (* diagnostic *)
(*       val _ = List.foreach(functions, fn {name, ...} =>
 * 			   Control.message(let open Layout
 * 					   in seq[Func.layout name,
 * 						  str " => ",
 * 						  Info.layout(getCont name)]
 * 					   end))
 *)
      (* diagnostic *)
	 
      val {get, set} = Property.new(Func.plist, Property.initConst NONE)

      (* for functions turned into locals, record their
       * args, body, and new name
       *)
      val _ =
	 List.foreach
	 (functions, fn {name, args, body, ...} =>
	  case getCont name of
	     Info.OneOuterCall =>
		set(name,
		    SOME{args = args,
			 body = body,
			 jump = Jump.newString(Func.originalName name)})
	   | _ => ())

      (* Walk over all functions, removing those that aren't toplevel, and
       * descending those that are, inserting local functions
       * where necessary.
       *  - turn tail calls into nontail calls
       *  - turn returns into jumps
       *)
	 
      fun walkExp(f: Func.t, e: Exp.t, c: Jump.t option): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val decs = 
	       List.map
	       (decs, fn d =>
		case d of
		   Bind _ => d
		 | Fun{name, args, body} =>
		      Fun{name = name,
			  args = args,
			  body = walkExp(f, body, c)}
		 | HandlerPush h => HandlerPush h
		 | HandlerPop => HandlerPop)
	    fun make transfer = Exp.make{decs = decs,
					 transfer = transfer}
	 in
	    case transfer of
	       Call{func, args, cont} =>
		  let val newCont = (case cont of
					NONE => c
				      | SOME _ => cont)
		  in
		     case get func of
			NONE => make(Call{func = func,
					  args = args,
					  cont = newCont})
		      | SOME{jump, args = formals, body} =>
			 if Func.equals(f, func)
			    then (* cont must be tail -- turn into jump *)
			       (case cont of
				   SOME _ => Error.bug "nontail jump"
				 | NONE =>
				      make(Jump{dst = jump, args = args}))
			 else
			    Exp.make
			    {decs =
			     decs @ [Fun{name = jump,
					 args = formals,
					 body = walkExp(func, body, newCont)}],
			     transfer = Jump{dst = jump, args = args}}
		  end
	     | Return xs =>
		  make(case c of
			  NONE => transfer
			| SOME c => Jump{dst = c, args = xs})
	     | _ => make transfer
	 end

      val shrinkExp = shrinkExp globals

      val functions =
	 List.fold(functions, [], fn ({name, args, body, returns}, functions) =>
		   case get name of
		      NONE =>
			 {name = name, args = args, returns = returns,
			  body = shrinkExp(walkExp(name, body, NONE))}
			 :: functions
		    | SOME _ => functions)

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end

end
