(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor CallGraph(S: CPS_CALL_GRAPH_STRUCTS) : CPS_CALL_GRAPH =
   struct
      open S
      structure CpsTree = CpsTree
      structure Graph = DirectedGraph
      structure Node = Graph.Node

      open CpsTree
	 
      datatype t = T of {getNode: CpsTree.Func.t -> Graph.Node.t,
			 setNode: CpsTree.Func.t * Graph.Node.t -> unit,
			 isLeaf: Graph.Node.t -> bool,
			 getExpFromNode : Graph.Node.t -> { name : CpsTree.Func.t,
							   args : (CpsTree.Var.t * CpsTree.Type.t) list,
							   body : CpsTree.Exp.t },
			 getCallSites: Graph.Node.t -> CpsTree.Transfer.t list,
			 getSize: Graph.Node.t -> int,
			 setSize: Graph.Node.t * int -> unit}

      type function = { name : Func.t,
		        args : (Var.t * Type.t) list,
		        body : Exp.t } 

	 
      fun buildGraph(Program.T{datatypes,globals,functions,main}) =
	 let val graph = Graph.new()

	     val {get = getExpFromNode: Node.t -> function,
		  set = setNodeToExp} =
		Property.new(Graph.Node.plist, 
			     Property.initRaise("No Exp for this node", Node.layout))

	     fun createNode(id) = Graph.newNode(graph)

	     val {get = getNode: Func.t -> Node.t,
		  set = setNode} =
		Property.new(Func.plist, Property.initFun(createNode))

	     val {get = getCallSites: Node.t -> Transfer.t list,
		  set = setCallSites} =
		Property.new(Graph.Node.plist, Property.initConst [])

	     val getCallSites = Trace.trace("getCallSites", Node.layout, Layout.ignore) 
		getCallSites
	 
	     val {get = getSize: Node.t -> int,
		  set = setSize} =
		Property.new(Graph.Node.plist, Property.initConst 0)

	     val {get = isLeaf: Node.t -> bool,
		  set = setLeaf} =
		Property.new(Graph.Node.plist, Property.initConst true)

	     val isLeaf = Trace.trace("isLeaf",Node.layout,Bool.layout) isLeaf
	 
	     fun appendCallSites(node,callSite) =
		setCallSites(node, callSite::getCallSites(node))

	     fun traceFun(name,body,graph) =
		let val node = getNode(name)
		in case (Exp.dest body) of
		   {decs,transfer} =>
		      (List.foreach(decs, 
				    fn d => case d of
				    Dec.Fun cont =>
				       let val {name,args,body} = cont
					  val node = getNode(name)
					  val _ = setNodeToExp(node,cont)
				       in  traceFun(name,body,graph)
				       end
				  | _ => ());
		       case transfer of
			  Transfer.Call {func,args,cont} =>
			     let val funcNode = getNode(func)
			     in (Graph.addEdge(graph, {from=node, to=funcNode});
				 setCallSites(funcNode, transfer::getCallSites(funcNode));
				 case cont of
				    NONE => ()
				  | SOME l => let val contNode = getNode(l)
					      in Graph.addEdge(graph, {from=funcNode, to=contNode})
					      end)
			     end
			| Transfer.Jump {dst, args} =>
			     let val node = getNode(dst)
			     in  setCallSites(node, transfer::getCallSites(node))
			     end
			| Transfer.Case {test,cases,default, ...} =>
			     let val branches = List.map(cases, fn (c,l) =>
							 let val caseNode = getNode(l)
							 in Graph.addEdge(graph, 
									  {from=node, to=caseNode})
							 end)
			     in case default of
				NONE => ()
			      | SOME l => let val defaultNode = getNode(l)
					  in Graph.addEdge(graph, {from=node, to=defaultNode})
					  end
			     end
			| _ => ())
		end
	    
	     val _ = (List.foreach(functions, fn {name,args,body,returns} => 
				   let val newNode = getNode(name)
				   in (setNodeToExp(newNode, {name=name,args=args,body=body});
				       let exception E
				       in Exp.foreachTransfer(body, fn Transfer.Call c => (setLeaf(newNode,false);
											   raise E)
				                                    | _ =>  ())
					  handle E => ()
				       end)
				   end);
		      Graph.removeBackEdges(graph);
		      List.foreach(functions, fn {name,args,body,returns} => 
				   traceFun(name,body,graph)))
	 in T {getNode=getNode,
	       setNode=setNode,
	       isLeaf=isLeaf,
	       getExpFromNode=getExpFromNode,
	       getCallSites=getCallSites,
	       getSize=getSize,
	       setSize=setSize}
	    before Control.message(fn () =>
				   let open Layout
				   in seq[str"Graph size:  ",
					  Int.layout(List.length(Graph.nodes graph))]
				   end)
	 end

      val buildGraph = Control.traceCall "buildGraph" buildGraph
	
   end
 
functor CpsSize(S : CPS_SIZE_STRUCTS) : CPS_SIZE = 
   struct

      open S
      structure CpsTree = CpsTree
      structure CallGraph = CallGraph

      structure Program = CpsTree.Program
      structure Exp = CpsTree.Exp
      structure Type = CpsTree.Type
      structure Transfer = CpsTree.Transfer
      structure PrimExp = CpsTree.PrimExp
      structure Dec = CpsTree.Dec
      structure Exp = CpsTree.Exp
      structure Var = CpsTree.Var
      structure Func = CpsTree.Func

      open CallGraph
	 
      datatype known = K | U | Bot

      val maxSize = 100000
	 
      val {contains,add,...} = List.set {equal = Func.equals, layout = Func.layout}

      fun knownJoin (K,K) = K
	| knownJoin (K,U) = U
	| knownJoin (Bot,x) = x:known
	| knownJoin (x,Bot) = x:known
	| knownJoin (U,_) = U
				       
      fun computeSize(Program.T{datatypes,globals,functions,main},graph) =
	 let  val {get = getKnown: Var.t -> known,
		   set = setKnown,
		   destroy = destroyKnown} =  Property.newDest(Var.plist, Property.initConst Bot)

	    val CallGraph.T {getNode,setNode,isLeaf,getExpFromNode,getCallSites,getSize,setSize} = graph
	     fun computeArgProp(name,args) =  let val node = getNode(name)
      	                                          val callSites = getCallSites(node)
						  val length = List.length(args)
						  val argProps = 
						     List.fold(callSites,
							       List.duplicate(length,fn() => Bot),
							       fn (c, i) =>
							       case c of
								  Transfer.Call{func,args,cont} => 
								     List.map2(i,args,
									       fn(k,a) => (knownJoin(k,getKnown(a))))
								| Transfer.Jump {dst,args} => 
								     List.map2(i,args,
									       fn(k,a) => (knownJoin(k,getKnown(a))))
								| _ => i)
					      in List.foreach2(args, argProps, setKnown)
					      end
	  
	     fun computeKnownExp(exp) = 
		let val {decs,transfer} = Exp.dest(exp)
		in  List.foreach(decs,
				 fn d => (case d of
					     Dec.Bind {var,ty,exp} => 
						(case exp of
						    PrimExp.Var pvar => 
						       let val v = getKnown(pvar)
						       in setKnown(var,v)
						       end
						  | _ =>  setKnown(var,K))
					   | Dec.Fun {name,args,body} =>
						let val args = List.map(args, fn (var,varType) => var)
						in computeArgProp(name,args) 
						end
					   | _ => ()))
		end

	 
	     val computeKnownExp = Trace.trace("computeKnownExp", Exp.layout, Layout.ignore) 
		computeKnownExp
		
	     val {get = getVisited: Func.t -> bool,
		  set = setVisited,
		  destroy = destroyVisited} =
		Property.newDest(Func.plist, Property.initConst false)
		
	     fun computePrimExpSize(primExp) =
		(case primExp of
		    PrimExp.Const c => 0
		  | PrimExp.Var x => 1
		  | PrimExp.Select _ => 1
		  | PrimExp.Tuple vars => List.length(vars)
		  | PrimExp.ConSelect _ => 1
		  | PrimExp.ConApp {con, args} => List.length(args)
		  | PrimExp.PrimApp {prim, targs, args} => List.length(args))
	     and
		computeDecSize(dec) = 
		case dec of
		   Dec.Bind {var,ty,exp} => let val _ = (case exp of
							    PrimExp.Const c => setKnown(var,K)
							  | PrimExp.Var v => setKnown(var,getKnown(v))
							  | _ => ())
					    in case getKnown(var) of
					       K => 0
					     | _ => computePrimExpSize(exp)
					    end
		 |   Dec.Fun {name,args,body} => let val node = getNode(name)
						     val nodeSize = getSize(node)
						 in if (nodeSize > 0) orelse getVisited(name)
						       then nodeSize
						    else let val _ = setVisited(name,true)
							     val funSize = computeExpSize(body)
							 in funSize before 
							    setSize(node,funSize)
							 end
						 end
		 |   _ => 1 
	     and
		computeTransferSize(transfer) =
		case transfer of
		   Transfer.Call {func, args, cont} => 
		      let val node = getNode(func)
			 val funcSize = getSize(node)
		      in if funcSize = 0
			    then 
			       let val {name,args,body} =  getExpFromNode(node)
				  val size =  computeExpSize(body)
				  val _ = setSize(node,size)
			       in List.length(args)
			       end
			 else funcSize
		      end
		 |   Transfer.Jump {dst, args} => 1 + List.length(args)
		 |   Transfer.Case {test, cases, default, ...} => let val known = getKnown(test)
								 val defaultSize = 
								    case default of
								       NONE => 0
								     | SOME l => 1
							     in case known of
								K => defaultSize
							      | U => 2 + defaultSize +  
								   List.length(cases)
							      | _ => defaultSize
							     end
		 |   _  => 1
	     and
		computeExpSize(exp) =
		let val {decs,transfer} = Exp.dest(exp)
		   val _ = computeKnownExp(exp) 
		   val decSize = List.fold(decs,0,fn(d,i)=> if i > maxSize
							       then maxSize
							    else i + computeDecSize(d))
		   val transferSize = computeTransferSize(transfer)
		in decSize + transferSize
		end

	     val computeExpSize = Trace.trace("computeExpSize", Exp.layout, Layout.ignore) 
		computeExpSize
		

	 in (List.foreach(functions, fn {name,args,body,returns} =>
			      let val node = getNode(name)
			      in (setSize(node, List.length(args) + 
					  computeExpSize(body)))
			      end);
		 destroyVisited(); destroyKnown())
	 end

(*	       val computeSize = fn arg => ((computeSize arg)
	                           handle E => (print(General.exnName E); print("\n")))
*)

   end


functor CpsInline(S: CPS_INLINE_STRUCTS) : CPS_INLINE =
   struct
      structure CpsTree = S
      structure CallGraph = CallGraph(structure CpsTree = CpsTree)
      structure CpsSize = CpsSize(structure CpsTree = CpsTree
				  structure CallGraph = CallGraph)

      open CallGraph
      open CpsTree

      local open Control
      in val rounds = ref 1
	 val size = ref 2
      end

      val inlineSize = ref 0
      val maxInlineSize = 500
	 
      datatype CallType = REG | NON_TAIL of Jump.t
      datatype InlineState = INLINE of (Graph.Node.t * Exp.t) | NO_INLINE of Transfer.t

      val shrinkExp = S.shrinkExp

      val foo: Program.t option ref = ref NONE
	 
      fun computeInline(programz) =
	 let val program = (case !foo of
			       NONE => (print("there...");
					foo := SOME programz;
					programz)
			     | SOME program => (print("here...."); program))

	    fun loop(program,rounds) =
	     let val graph as T {getNode,setNode,isLeaf,getExpFromNode,getCallSites,getSize,setSize} =
		           buildGraph(program)
		 val _ = CpsSize.computeSize(program,graph)
		 val  CpsTree.Program.T{datatypes,globals,functions,main} = program
		 val shrinkExp = shrinkExp globals
		 fun small(node) = (getSize(node) < (!inlineSize))
		 fun alwaysInline(node) = let val callSites = getCallSites(node)
					     in if small(node) orelse List.length(callSites) = 1 
						orelse (isLeaf(node) andalso (getSize(node) < maxInlineSize))
						   then true
						else false
					     end

		 val {get = getVisited: Func.t -> int,
		      set = setVisited,
		      destroy = destroyVisited} =
		    Property.newDest(Func.plist, Property.initConst 0)

		 fun fixDecs(decs,callType) =
		    let open Dec
		       val newDecs = []
		    in List.fold(decs,
				 newDecs,
				 fn (d,newDecs) =>
				 List.append(newDecs,
					     case d of
						b as Bind {var,ty,exp} => [b]
					      | Fun {name,args,body} =>
						   [Fun {name = name,
							 args = args, 
							 body = fixReturns(body,callType)}]
					      | dec => [dec]))
		    end
		 
		 and fixTransfer(transfer,callType) = 
		    let open Transfer
		    in case transfer of
		       Call {func,args,cont} => Call {func = func,
						      args = args,
						      cont = case cont of
						      NONE => (case callType of
								  REG => NONE
								| NON_TAIL c => SOME c)
						    | SOME l => SOME l}
		     | r as Return vars => (case callType of
					       REG  => r
					     | NON_TAIL c => Jump{dst = c, args = vars})

		     | trans => trans
		    end
		 and fixReturns(exp,callType) = 
		    let open Exp
		       val {decs,transfer} = dest(exp)
		       val newDecs = fixDecs(decs,callType) 
		       val newTransfer = fixTransfer(transfer,callType)
		    in make({decs = newDecs, transfer = newTransfer})
		    end

		 fun inlineBody(node,e,actuals,callType) =
		    let val {name,args,body} = getExpFromNode(node)
		       val bindings = List.zip(args,actuals)
		       val substitution = List.map(bindings, fn ((formal,ty),actual) =>
						   {formal = formal, actual = actual})
		       val renamed = Exp.alphaRename {exp = body,
						      substitution = substitution,
						      handleJump = fn {old,new} => setNode(new,getNode(old))}
		       val {decs,transfer} = Exp.dest(renamed)
		       val new =  case transfer of
			  Transfer.Call {func,args,cont} => let val node = getNode(func)
								val body = #body(getExpFromNode(node))
							    in if alwaysInline(node)
								  then let val callType = (case cont of
											      NONE => REG
											    | SOME c => NON_TAIL c)  
									   val {decs=decs',transfer=transfer'} = 
									      inlineBody(node,body,args,callType)
								       in Exp.make({decs = List.append(decs,decs'),
										    transfer = transfer'})
								       end
							       else renamed
							    end
			| _ => renamed 	
		       val newBody = fixReturns(new,callType)
		       val newExp = Exp.dest(newBody)
		    in {decs = #decs newExp, transfer = #transfer newExp}
		    end

		 val inlineBody = Trace.trace4("inlineBody", Layout.ignore, Exp.layout, Layout.ignore, Layout.ignore, Layout.ignore)
		    inlineBody
		    
		 fun remove(Program.T{datatypes, globals, functions, main}) =
		    let
		       val {get: Func.t -> {body: Exp.t, visited: bool ref},
			    set, destroy} =
			  Property.newDest(Func.plist,Property.initRaise("info", Func.layout))
			  
		       val _ = List.foreach(functions, fn {name, body, ...} =>
					    set(name, {body = body, visited = ref false}))
		       fun visit(f: Func.t) =
			  let val {visited, body} = get f
			  in if !visited
				then ()
			     else (visited := true;
				   Exp.foreachTransfer(body,
						       fn Transfer.Call{func, ...} => visit func
							| _ => ()))
			  end
		       val _ = visit main
		       val functions = List.keepAll(functions, fn f => !(#visited(get(#name f))))
		    in destroy();
		       Program.T{datatypes = datatypes,
				 globals = globals,
				 functions = functions,
				 main = main}
		    end

		 fun inline(functions) =
		    let val {get: Func.t -> {body: Exp.t, visited: bool ref}, set, destroy} =
		       Property.newDest(Func.plist,
					Property.initRaise("info", Func.layout))
			val {get = init: Func.t -> bool, set = setInit, destroy = destroyInit} =
			   Property.newDest(Func.plist,Property.initConst false)
			   
			fun setDec(d) =
			   case d of
			      Dec.Fun {name,args,body} =>
				 (set(name,{body = body, visited = ref false});
				  if init(name)
				     then ()
				  else (setInit(name,true); setExp(body)))
			    | _ =>()
			and setExp(e) = Exp.foreach'(e,
						     {handleDec = setDec,
						      handleTransfer = fn _ => ()})

			fun buildInline(newDecs,newExp,transfer, operator) =
			   let fun build(func,args,cont) = 
			      let val {body,visited} = get(func)
				 val node = getNode(func)
			      in if operator(node) 
				    then let val newExp = inlineBody(node,body,args,
								     case cont of
									NONE => REG
								      | SOME c => NON_TAIL c)
					     val {decs=decs',transfer=transfer'} = newExp
					 in Exp.make( {decs = List.append(newDecs,decs'),
						       transfer = transfer'})				 
					 end
				 else newExp
			      end
			   in case transfer of
			      Transfer.Call {func,args,cont} => build(func,args,cont)
			    | Transfer.Jump {dst,args} => build(dst,args,NONE)  
			    | _ => newExp 
			   end

			fun visit(f : Func.t) =
			   let val {visited, body} = get f
			      fun inline(d) = case d of
				 Dec.Fun {name,args,body} =>
				    Dec.Fun {name=name,args=args,body=visit(name)}
			       | d => d

			   in  if !visited
				  then body
			       else (visited := true;
				     let val {decs,transfer} = Exp.dest(body)
					val newDecs = List.map(decs,inline)
					val newExp = Exp.make {decs = newDecs, transfer = transfer}
					val _ = set(f, {body = newExp, visited=visited})
				     in buildInline(newDecs,newExp,transfer,
						    case transfer of
						       Transfer.Call {func, args, cont} => alwaysInline
						     | Transfer.Jump {dst,args} => small   
						     | _ => (fn _ => false))
				     end)
			   end
		    in (List.foreach(functions, fn {name, args, body, returns}  =>
				     (set(name, {body = body, visited = ref false});
				      setExp(body)));
			List.map(functions, fn {name, args, body, returns}  =>
				 {name = name, args = args, body = visit(name), returns = returns}))
		    end
		    
		 val newFuns =  inline(functions)
		 val shrinkedExp = List.map(newFuns,
					    fn {name,args,body,returns} =>
					    {name = name,
					     args = args,
					     body =   shrinkExp(body), 
					     returns = returns})
		 val newProg = remove(CpsTree.Program.T {datatypes = datatypes,
							 globals = globals,
							 functions = shrinkedExp,
							 main = main })
		 val _ = (Program.clear(program); destroyVisited())
		 val _ = Control.message(fn () => Program.layoutStats(newProg))
	     in if rounds <= 1
		   then newProg
		else loop(newProg,rounds-1)
	     end
	  val newProg = loop(program,!rounds)
	in (Program.clear(program);
	    inlineSize := !size;
	    loop(newProg,0) before (Program.clear(newProg)))
	end
   end


 
(* (Layout.output(let open Layout
			    in align[seq[str "exp: ", Program.layout program,
					 str "newExp: ", Program.layout newProg]]
			    end, Out.standard); *)


(* Layout.output(let open Layout
			    in align[seq[str "exp: ", Program.layout program,
					 str "newExp: ", Program.layout newProg]]
			    end, Out.standard);

*)






(* 

             let val functions = List.map(functions, 
					  fn f as {name,args,body,returns} =>
					  let val {decs,transfer} = Exp.dest(body)
					  in case transfer of
                                                Transfer.Call c => {name=name,args=args,body=buildInline(decs,body,transfer,isLeaf),
								    returns=returns}
                                               | _ => f
					  end)

               Layout.output(let open Layout
									 in align[seq[str "func Name   ",
										      FuncName.layout func,
										      let val node = getNode(func)
										      in Bool.layout(isLeaf(node))
										      end,
										      str "\n"]]
									 end, Out.standard)
		      | _ => ()
		      end)
(List.foreach(functions, fn {name, ...} => let val n = getNode(name)
						       in Control.message(let open Layout
									  in seq[FuncName.layout name,
										 str " ",
										 Bool.layout(isLeaf(n))]
									  end)
						       end);
*)
(*

                                               (Layout.output(let open Layout
							      in align[seq[str "callee: ", Func.layout func],
									   str "cont", (case cont of
									                 SOME c => Jump.layout c
									                | NONE => Unit.layout () ),
									   str "\n"]
							      end, Out.standard);

						*)
