(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor CpsTree(S: CPS_TREE_STRUCTS): CPS_TREE = 
struct

open S

structure Type =
   struct
      local structure T = HashType(S)
      in open  T
      end

      fun tyconArgs t =
	 case Dest.dest t of
	    Dest.Con x => x
	  | _ => Error.bug "FirstOrderType.tyconArgs"
	       
      datatype dest =
	 Char
	| Int
	| IntInf
	| Pointer
	| Word
	| Word8
	| Real
	| Stack
	| String 
	| Array of t
	| Vector of t
	| Ref of t
	| Datatype of Tycon.t
	| Tuple of t list
	| Variant of Tycon.t * Con.t

      local
	 val {get, set} = Property.new(Tycon.plist, Property.initConst NONE)

	 fun nullary c =
	    fn [] => c
	     | _ => Error.bug "bogus application of nullary tycon"
		  
	 fun unary make =
	    fn [t] => make t
	     | _ => Error.bug "bogus application of unary tycon"

      in
	 fun initialize() =
	    List.foreach
	    ([(Tycon.tuple, Tuple),
	      (Tycon.char, nullary Char),
	      (Tycon.int, nullary Int),
	      (Tycon.intInf, nullary IntInf),
	      (Tycon.pointer, nullary Pointer),
	      (Tycon.real, nullary Real),
	      (Tycon.stack, nullary Stack),
	      (Tycon.string, nullary String),
	      (Tycon.word8, nullary Word8),
	      (Tycon.word, nullary Word),
	      (Tycon.array, unary Array),
	      (Tycon.vector, unary Vector),
	      (Tycon.reff, unary Ref)],
	     fn (tycon, f) => set(tycon, SOME f))

	 val _ = initialize()

	 fun dest t =
	    case Dest.dest t of
	       Dest.Con(tycon, ts) =>
		  (case get tycon of
		      NONE => Datatype tycon
		    | SOME f => f ts)
	     | Dest.Variant z => Variant z
	     | _ => Error.bug "dest"
      end

      local open Layout
      in
	 fun layout t =
	    case dest t of
	       Char => str "char"
	     | Int => str "int"
	     | IntInf => str "IntInf.int"
	     | Pointer => str "pointer"
	     | Word => str "word"
	     | Word8 => str "word8"
	     | Real => str "real"
	     | Stack => str "stack"
	     | String => str "string"
	     | Array t => seq[layout t, str " array"]
	     | Vector t => seq[layout t, str " vector"]
	     | Ref t => seq[layout t, str " ref"]
	     | Datatype t => Tycon.layout t
	     | Tuple [] => str "unit"
	     | Tuple ts => paren(seq(separate(List.map(ts, layout), " * ")))
	     | Variant(tycon, con) => seq[Tycon.layout tycon,
					  str "@",
					  Con.layout con]
      end
   end

structure Func =
   struct
      open Var (* Id(structure AstId = Ast.Var) *)

      fun newNoname() = newString "F"
      val saveWorld = newString "saveWorld"
   end

structure Jump =
   struct
      open Func
      fun newNoname() = newString "L"
   end

local open Layout
in
   fun layoutTuple xs = tuple(List.map(xs, Var.layout))
end

structure PrimExp =
   struct
      datatype t =
	 Const of Const.t
       | Var of Var.t
       | Tuple of Var.t list
       | Select of {tuple: Var.t,
		    offset: int}
       | ConSelect of {variant: Var.t,
		       con: Con.t,
		       offset: int}
       | ConApp of {con: Con.t,
		    args: Var.t list}
       | PrimApp of {prim: Prim.t,
		     targs: Type.t list,
		     args: Var.t list}

      fun foreachVar(e, f) =
	 let fun fs xs = List.foreach(xs, f)
	 in case e of
	    Const _ => ()
	  | Var x => f x
	  | Tuple xs => fs xs
	  | Select{tuple, ...} => f tuple
	  | ConSelect{variant, ...} => f variant
	  | ConApp{args, ...} => fs args
	  | PrimApp{args, ...} => fs args
	 end
      
      fun replaceVar(e, f) =
	 let fun fs xs = List.map(xs, f)
	 in case e of
	    Const _ => e
	  | Var x => Var(f x)
	  | Tuple xs => Tuple(fs xs)
	  | Select{tuple, offset} => Select{tuple = f tuple, offset = offset}
	  | ConSelect{variant, con, offset} =>
	       ConSelect{variant = f variant, con = con, offset = offset}
	  | ConApp{con, args} => ConApp{con = con, args = fs args}
	  | PrimApp{prim, targs, args} =>
	       PrimApp{prim = prim, targs = targs, args = fs args}
	 end

      local open Layout
      in
	 val layout =
	    fn Const c => Const.layout c
	     | Var x => Var.layout x
	     | Tuple xs => layoutTuple xs
	     | Select{tuple, offset} =>
		  seq[str "#", Int.layout(offset + 1), str " ", Var.layout tuple]
	     | ConSelect{variant, con, offset} =>
		  seq[Con.layout con, str "#",
		      Int.layout(offset + 1), str " ", Var.layout variant]
	     | ConApp{con, args} => seq[Con.layout con, layoutTuple args]
	     | PrimApp{prim, targs, args} =>
		  seq[Prim.layout prim,
		      (case targs of
			  [] => empty
			| _ => list(List.map(targs, Type.layout))),
		      layoutTuple args]
      end
   end

local
   open Layout
in
   fun layoutBind{var, ty, exp} =
      seq[str "val ", Var.layout var,
	  str ": ", Type.layout ty,
	  str " = ", PrimExp.layout exp]
end

fun clearBind{var, ty, exp} = Var.clear var

structure Cause =
   struct
      datatype t = User | Dispatch | PolyEqual | Coerce
   end

structure Transfer =
   struct
      datatype t =
	 Call of {func: Func.t,
		  args: Var.t list,
		  cont: Jump.t option}
       | Jump of {dst: Jump.t,
		  args: Var.t list}
       | Return of Var.t list
       | Case of {cause: Cause.t,
		  test: Var.t,
		  cases: (Con.t * Jump.t) list,
		  default: Jump.t option}
       | Raise of Var.t list
       | Halt
       | Bug

      fun foreachVar(t, f) =
	 let fun fs xs = List.foreach(xs, f)
	 in case t of
	    Call{args, ...} => fs args
	  | Jump{args, ...} => fs args
	  | Return xs => fs xs
	  | Case{test, ...} => f test
	  | Raise xs => fs xs
	  | Halt => ()
	  | Bug => ()
	 end

      fun replaceVar(t, f) =
	 let fun fs xs = List.map(xs, f)
	 in case t of
	    Call{func, args, cont} =>
	       Call{func = func, args = fs args, cont = cont}
	  | Jump{dst, args} => Jump{dst = dst, args = fs args}
	  | Return xs => Return(fs xs)
	  | Case{cause, test, cases, default} =>
	       Case{cause = cause,
		    test = f test, cases = cases, default = default}
	  | Raise xs => Raise(fs xs)
	  | Halt => Halt
	  | Bug => Bug
	 end

      local open Layout
      in
	 val layout =
	    fn Call{func, args, cont} =>
	    seq[Func.layout func,
		layoutTuple args,
		case cont of
		   NONE => str " Tail"
		 | SOME l => seq[str " NonTail ", Jump.layout l]]
	    | Jump{dst, args} =>
		 seq[Jump.layout dst, layoutTuple args]
	    | Return xs => seq[str "Return ", layoutTuple xs]
	    | Case{test, cases, default, ...} =>
		 align[seq[str "case ", Var.layout test, str " of"],
		       indent(align[alignPrefix
				    ("| ",
				     List.map(cases, fn (c, l) =>
					      seq[Con.layout c,
						  str " => ",
						     Jump.layout l])),
				    case default of
				       NONE => empty
				     | SOME l => seq[str "_ => ",
						     Jump.layout l]],
			      2)]
	    | Raise xs => seq[str "raise ", list(List.map(xs, Var.layout))]
	    | Halt => str "Halt"
	    | Bug => str "Bug"
      end
		
   end


structure DecExp =
   struct
      datatype dec =
	 Bind of {
		  var: Var.t,
		  ty: Type.t,
		  exp: PrimExp.t
		  }
       | Fun of {
		 name: Jump.t,
		 args: (Var.t * Type.t) list,
		 body: exp
		}
       | HandlerPush of Jump.t
       | HandlerPop

      and exp = Exp of {decs: dec list,
			transfer: Transfer.t}

      local
	 open Layout
      in
	 val rec layoutDec =
	    fn Bind b => layoutBind b
	     | Fun{name, args, body} =>
		  seq[str "fun ",
		      align[seq[Jump.layout name,
				tuple(List.map(args, fn (x, t) =>
					       seq[Var.layout x,
						   str ": ",
						   Type.layout t])),
				str " = "],
			    layoutExp body]]
	     | HandlerPush h => seq[str "val _ = HandlerPush ", Jump.layout h]
	     | HandlerPop => str "val _ = HandlerPop"
	 and layoutExp =
	    fn Exp{decs, transfer, ...} =>
	    let val transfer = Transfer.layout transfer
	    in case decs of
	       [] => transfer
	     | _ => align[seq[str "let ", align(List.map(decs, layoutDec))],
			  seq[str "in ", transfer],
			  str "end"]
	    end
      end
   end

structure Dec =
   struct
      datatype t = datatype DecExp.dec

      type exp = DecExp.exp

      val layout = DecExp.layoutDec
   end

structure Exp =
   struct
      datatype t = datatype DecExp.exp
	       
      fun dest(Exp r) = r
      val make = Exp

      val decs = #decs o dest
      val transfer = #transfer o dest

      val layout = DecExp.layoutExp
					     
      local
	 fun make f (Exp{decs, transfer}, x): t =
	    Exp{decs = f(x, decs), transfer = transfer}
      in val prefix = make(op ::)
	 val prefixs = make(op @)
      end

      fun foreach(e, {handleDec, handleTransfer}) =
	 let
	    fun exp(Exp{decs, transfer, ...}) =
	       (List.foreach(decs, dec);
		handleTransfer transfer)
	    and dec d =
	       let val after = handleDec d
	       in (case d of
		      Dec.Fun{body, ...} => exp body
		    | _ => ());
		  after()
	       end
	 in exp e
	 end

      fun ignore _ = ()
      fun ignore' _ = ignore

      fun foreach'(e, {handleDec, handleTransfer}) =
	 foreach(e, {handleTransfer = handleTransfer,
		     handleDec = fn d => (handleDec d; ignore)})

      fun foreachDec(e, handleDec) =
	 foreach'(e, {handleTransfer = ignore,
		      handleDec = handleDec})

      fun foreachVar(e, f) =
	 foreach'
	 (e, {handleTransfer = ignore,
	      handleDec = fn Dec.Bind{var, ty, ...} => f(var, ty)
			   | Dec.Fun{args, ...} => List.foreach(args, f)
			   | _ => ()})

      fun foreachBind(e, f) =
	 let
	    fun handleDec d =
	       (case d of
		   Dec.Bind r => f r
		 | _ => ();
	        ignore)
	 in foreach(e, {handleTransfer = ignore,
			handleDec = handleDec})
	 end

      fun foreachTransfer(e, f) =
	 foreach(e, {handleDec = ignore', handleTransfer = f})

      fun foreachCall(e, f) =
	 foreachTransfer(e, fn Transfer.Call r => f r | _ => ())

      fun clear(e: t): unit =
	 let open DecExp
	 in foreach'
	    (e, {handleTransfer = fn _ => (),
		 handleDec =
		 fn Bind b => clearBind b
		  | Fun{name, args, ...} => (Jump.clear name;
					     List.foreach(args, Var.clear o #1))
		  | _ => ()})
	 end

      fun alphaRename{exp: t,
		      substitution: {formal: Var.t, actual: Var.t} list,
		      handleJump: {old: Jump.t, new: Jump.t} -> unit} =
	 let
	    local
	       fun make(new, plist, layout) =
		  let
		     val {get, set, destroy} =
			Property.newDest(plist, Property.initConst NONE)
		     fun bind x = let val x' = new x
				  in set(x, SOME x'); x'
				  end
		     fun lookup x =
			case get x of
			   NONE => x
			 | SOME y => y
		  in (set, bind, lookup, destroy)
		  end
	    in
	       val (setVar, bindVar, lookupVar, destroyVar) =
		  make(Var.new, Var.plist, Var.layout)
	       val (_, bindJump, lookupJump, destroyJump) =
		  make(Jump.new, Jump.plist, Jump.layout)
	    end

	    val bindJump =
	       fn j => let val j' = bindJump j
		       in handleJump{old = j, new = j'}
			  ; j'
		       end

	    fun lookupVars xs = List.map(xs, lookupVar)

	    datatype dec = datatype Dec.t
	    datatype transfer = datatype Transfer.t
	       
	    fun loopTransfer(t: Transfer.t): Transfer.t =
	       case t of
		  Bug => Bug
		| Halt => Halt
		| Raise xs => Raise(lookupVars xs)
		| Case{cause = cause, test, cases, default} =>
		     Case{cause = cause,
			  test = lookupVar test,
			  cases = List.map(cases, fn (c, j) =>
					   (c, lookupJump j)),
			  default = (case default of
					NONE => NONE
				      | SOME j => SOME(lookupJump j))}
		| Return xs => Return(lookupVars xs)
		| Jump{dst, args} => Jump{dst = lookupJump dst,
					  args = lookupVars args}
		| Call{func, args, cont} =>
		     Call{func = func,
			  args = lookupVars args,
			  cont = Option.map(cont, lookupJump)}

	    fun loopExp e =
	       let val {decs, transfer} = dest e
		  val decs = List.map(decs, loopDec)
	       in make{decs = decs,
		       transfer = loopTransfer transfer}
	       end
	    and loopDec d =
	       case d of
		  HandlerPush h => HandlerPush(lookupJump h)
		| HandlerPop => HandlerPop
		| Fun{name, args, body} =>
		     Fun{name = bindJump name,
			 args = List.map(args, fn (x, t) => (bindVar x, t)),
			 body = loopExp body}
		| Bind{var, ty, exp} => 
		     Bind{exp = PrimExp.replaceVar(exp, lookupVar),
			  var = bindVar var,
			  ty = ty}
	 in List.foreach(substitution, fn {formal, actual} =>
			 setVar(formal, SOME actual))
	    ; (loopExp exp
	       before (destroyVar() ; destroyJump()))
	 end
   end

structure DirectExp =
   struct
      open Dec PrimExp Transfer
      structure Cont =
	 struct
	    type t = PrimExp.t * Type.t -> Exp.t

	    val layout = Layout.ignore

	    fun nameGen(k: Var.t * Type.t -> Exp.t): t =
	       fn (e, t) =>
	       case e of
		  Var x => k(x, t)
		| _ => let val x = Var.newNoname()
		       in Exp.prefix(k(x, t), Bind{var = x, ty = t, exp = e})
		       end
	    
	    fun name(k: Var.t -> Exp.t): t = nameGen(k o #1)

	    fun return(k: t, xt) = k xt

	    fun jump(j: Jump.t): t =
	       name(fn x =>
		    Exp.make{decs = [],
			     transfer = Jump{dst = j, args = [x]}})

	    val id: t = name(fn x => Exp.make{decs = [], transfer = Return[x]})

	    val halt: t = name(fn x => Exp.make{decs = [], transfer = Halt})

	    fun prefix(k: t, d: Dec.t): t =
	       fn z => Exp.prefix(return(k, z), d)
	 end
      
      type t = Cont.t -> Exp.t

      val layout = Layout.ignore

      fun send(e: t, k: Cont.t): Exp.t = e k

(*       val send =
 * 	 Trace.trace2("DirectExp.send", layout, Cont.layout, Exp.layout) send
 *)

      fun sendName(e, k) = send(e, Cont.name k)

      fun toExp e = send(e, Cont.id)
      fun toExpHalt e = send(e, Cont.halt)

      fun const c = fn k => Cont.return(k, (Const c, Type.ofConst c))

      fun var(x, t) = fn k => Cont.return(k, (Var x, t))

      fun convertsGen(es: t list, k: Var.t list -> Exp.t): Exp.t =
	 let
	    fun loop(es, xs) =
	       case es of
		  [] => k(List.reverse xs)
		| e :: es => sendName(e, fn x => loop(es, x :: xs))
	 in loop(es, [])
	 end

      fun converts(es: t list, make: Var.t list -> PrimExp.t * Type.t): t =
	 fn k => convertsGen(es, k o make)

      fun convert(e: t, make: Var.t -> PrimExp.t * Type.t): t =
	 fn k => send(e, Cont.name(k o make))

      fun name2(e1: t, e2: t, make: Var.t * Var.t -> t): t =
	 fn k =>
	 sendName(e1, fn x1 => sendName(e2, fn x2 => send(make(x1, x2), k)))

      fun seq(e1: t, e2: t): t = fn k => sendName(e1, fn _ => send(e2, k))

      fun tuple{exps: t list, ty: Type.t}: t =
	 case exps of
	    [e] => e
	  | _ => converts(exps, fn xs => (PrimExp.Tuple xs, ty))

      fun select{tuple, offset, ty} =
	 convert(tuple, fn tuple => (Select{tuple = tuple, offset = offset}, ty))

      fun conApp{con, args, ty} =
	 converts(args, fn args => (ConApp{con = con, args = args}, ty))

      fun reify(ty: Type.t, z: Jump.t option -> Exp.t): t =
	 fn k =>
	 let val result = Var.newNoname()
	    val rest = Cont.return(k, (Var result, ty))
	    fun nontail() =
	       let val j = Jump.newNoname()
	       in Exp.prefix(z(SOME j),
			     Fun{name = j,
				 args = [(result, ty)],
				 body = rest})
	       end
	 in case Exp.dest rest of
	    {decs = [], transfer = Return[result']} =>
	       (* should also do something for returning unit *)
	       if Var.equals(result, result')
		  then z NONE
	       else nontail()
	  | _ => nontail()
	 end

      fun call{func, args, ty} =
	 reify
	 (ty, fn jo =>
	  convertsGen
	  (args, fn args =>
	   Exp.make{decs = [],
		    transfer = Call{func = func, args = args, cont = jo}}))

      fun raisee e =
	 fn _ => sendName(e, fn x => Exp.make{decs = [], transfer = Raise[x]})

      fun lett{decs, body} =
	 fn k =>
	 let
	    val rec loop =
	       fn [] => body k
		| {var, ty, exp} :: ds =>
		     send(exp, fn (e, _) =>
			  Exp.prefix(loop ds, Bind{var = var, ty = ty, exp = e}))
	 in loop decs
	 end

      fun joinPoint(ty: Type.t, e: t): t =
	 reify(ty, fn jo => send(e, (case jo of
					NONE => Cont.id
				      | SOME j => Cont.jump j)))
	 
      fun primApp{prim, targs, args, ty} =
	 converts(args, fn args =>
		  (PrimApp{prim = prim, targs = targs, args = args}, ty))

      fun casee{cause, test, ty, cases, default} =
	 joinPoint
	 (ty, fn k =>
	  test
	  (Cont.nameGen
	   (fn (test, testTy) =>
	    let val tycon = Type.detycon testTy
	       val (cases, decs) =
		  List.fold
		  (cases, ([], []),
		   fn ({con, args, body}, (cases, decs)) =>
		   let val e = send(body, k)
		      val j = Jump.newNoname()
		      val x = Var.newNoname()
		   in ((con, j) :: cases,
		       Fun{name = j,
			   args = [(x, Type.variant(tycon, con))],
			   body =
			   Exp.prefixs
			   (e,
			    List.foldi
			    (args, [], fn (i, (y, t), ds) =>
			     Bind{var = y,
				  ty = t,
				  exp = ConSelect{variant = x,
						  con = con,
						  offset = i}}
			     :: ds))}
		       :: decs)
		   end)
	       val (default, decs) =
		  case default of
		     NONE => (NONE, decs)
		   | SOME e =>
			let val j = Jump.newNoname()
			in (SOME j,
			    Fun{name = j, args = [], body = send(e, k)}
			    :: decs)
			end
	    in Exp.make{decs = decs,
			transfer = Case{cause = cause,
					test = test,
					cases = cases,
					default = default}}
	    end)))
	  
      fun handlee{try, ty, catch, handler} =
	 joinPoint
	 (ty, fn k =>
	  let val h = Jump.newNoname()
	  in Exp.prefixs
	     (send(try, Cont.prefix(k, HandlerPop)),
	      [Fun{name = h,
		   args = [catch],
		   body = Exp.prefix(send(handler, k), HandlerPop)},
	       HandlerPush h])
	  end)

      fun detupleGen(e: PrimExp.t, t: Type.t, components: Var.t list,
		     body: Exp.t): Exp.t =
	 Exp.prefixs
	 (body,
	  case components of
	     [] => []
	   | [x] => [Bind{var = x, ty = t, exp = e}]
	   | _ =>
		let val ts = Type.detuple t
		   val tupleVar = Var.newNoname()
		in Bind{var = tupleVar, ty = t, exp = e}
		   ::
		   List.mapi(List.zip(components, ts), fn (i, (x, t)) =>
			     Bind{var = x, ty = t,
				  exp = Select{tuple = tupleVar,
					       offset = i}})
		end)
      
      fun detupleBind{tuple, components, body} =
	 fn k => send(tuple, fn (e, t) => detupleGen(e, t, components, body k))

      fun detuple{tuple, body}: t =
	 fn k =>
	 tuple
	 (fn (e, t) =>
	  case e of
	     Tuple xs => send(body xs, k)
	   | _ => let
		     val components =
			List.map(Type.detuple t, fn _ => Var.newNoname())
		  in detupleGen(e, t, components, send(body components, k))
		  end)

      local
	 fun make c = conApp{con = c, args = [], ty = Type.bool}
      in val truee = make Con.truee
	 val falsee = make Con.falsee
      end

      val int = const o Const.fromInt

      fun e1 + e2 =
	 primApp{prim = Prim.intPlus,
		 targs = [],
		 args = [e1, e2],
		 ty = Type.int}

      fun eq(e1, e2, ty) =
	 primApp{prim = Prim.eq,
		 targs = [ty],
		 args = [e1, e2],
		 ty = Type.bool}
   end

structure Datatype =
   struct
      type t = {tycon: Tycon.t,
		cons: {con: Con.t,
		       args: Type.t list} list}

      fun layout(ds: t list) =
	 let open Layout
	 in align
	    ((str "Datatypes:")
	     ::
	     List.map
	     (ds, fn {tycon, cons} =>
	      seq[Tycon.layout tycon,
		  str " = ",
		  align
		  (List.map(cons, fn {con, args} =>
			    seq[Con.layout con,
				case args of
				   [] => empty
				 | _ => 
				      seq[str " of ",
					  list(List.map(args, Type.layout))]]))]))
	 end
   end
		      
structure Function =
   struct
      type t = {name: Func.t,
		args: (Var.t * Type.t) list,
		body: Exp.t,
		returns: Type.t list}

      fun layout{name, args, body, returns} =
	 let open Layout
	 in align[seq[str "fun ",
		      Func.layout name,
		      tuple(List.map(args, fn (x, t) =>
				     seq[Var.layout x,
					 str ": ",
					 Type.layout t])),
		      str ": ",
		      tuple(List.map(returns, Type.layout)),
		      str " = "],
		  indent(Exp.layout body, 3)]
	 end
	 
      fun layouts(fs, output) =
	 List.foreach(fs, output o layout)
   end

structure Program =
   struct
      datatype t =
	 T of {
	       datatypes: Datatype.t list,
	       globals: {
			 var: Var.t,
			 ty: Type.t,
			 exp: PrimExp.t
			} list,
	       functions: Function.t list,
	       main: Func.t
	      }

      fun foreachVar(T{globals, functions, ...}, f) =
	 (List.foreach(globals, fn {var, ty, ...} => f(var, ty));
	  List.foreach
	  (functions, fn {args, body, ...} =>
	   (List.foreach(args, f);
	    Exp.foreachVar(body, f))))
	 
      fun layouts(T{datatypes, globals, functions, main},
		  output': Layout.t -> unit) =
	 let open Layout
	    val output = output'
	 in output(Datatype.layout datatypes)
	    ; output(str "\n\nGlobals:")
	    ; List.foreach(globals, output o layoutBind)
	    ; output(str "\n\nFunctions:")
	    ; Function.layouts(functions, output)
	    ; output(seq[str "\n\nMain: ", Func.layout main])
	 end

      fun layout(T{datatypes, globals, functions, main}) =
	 let open Layout
	 in align[Datatype.layout datatypes,
		  str "Globals:",
		  align(List.map(globals, layoutBind)),
		  str "Functions:",
		   align(List.map(functions, Function.layout)),
		   seq[str "Main: ", Func.layout main]]
	 end

      fun layoutStats(T{globals, functions, ...}) =
	 let
	    val numTypes = ref 0
	    fun inc _ = IntRef.inc numTypes
	    val {hom = countType, destroy} =
	       Type.makeHomVariant
	       {var = fn _ => Error.bug "cps-tree saw var",
		con = inc, variant = inc}
	    val numPrimExps = ref(List.length globals)
	    val numLocalFunctions = ref 0
	    val _ =
	       List.foreach
	       (functions, fn {body, ...} =>
		Exp.foreach'
		(body,
		 {handleDec = let open Dec
			      in fn Bind{ty, ...} => (countType ty;
						      IntRef.inc numPrimExps)
			          | Fun{args, ...} =>
				       (IntRef.inc numLocalFunctions;
					List.foreach(args, countType o #2))
				  | _ => ()
			      end,
		  handleTransfer = fn _ => ()}))
	    val numFunctions = List.length functions
	    open Layout
	 in destroy();

	    align
	    (List.map
	     ([("num functions", Int.layout numFunctions),
	       ("num local functions", Int.layout(!numLocalFunctions)),
	       ("num primExps", Int.layout(!numPrimExps)),
	       ("num types in program", Int.layout(!numTypes))],
	      fn (name, value) => seq[str(name ^ " "), value])
	     @ [Type.stats()])
	 end

      (* clear all property lists reachable from program *)
      fun clear(T{datatypes, globals, functions, main}) =
	 (List.foreach(datatypes, fn {tycon, cons} =>
		       (Tycon.clear tycon;
			List.foreach(cons, Con.clear o #con)));
	 List.foreach(globals, clearBind);
	 List.foreach
	 (functions, fn {name, args, body, returns} =>
	  (Func.clear name;
	   List.foreach(args, Var.clear o #1);
	   Exp.clear body)))
   end

end
