(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Here are the rules for changing control transfers within f
 * when f is inlined at a call within g.
 *
 *				   Call to f within g
 *
 *				Tail		NonTail C
 * inside f			---------	---------
 *	Return			Return		Jump L
 *	Tail			Tail		NonTail C
 *	NonTail C'		NonTail C'	NonTail C'
 *)

(* All raises must raise the same number of args and handlers must take this
 * same number of args.
 *)

(* For case statements, the type-checker allows a branch on a value carrying
 * constructor to go to a nullary jump, in this case the value is ignored.
 *)

signature CPS_TREE_STRUCTS = 
   sig
      include ATOMS
   end

signature CPS_TREE = 
   sig
      include CPS_TREE_STRUCTS

      structure Type:
	 sig
	    include HASH_TYPE
	       
	    val tyconArgs: t -> Tycon.t * t list
	       
	    datatype dest =
	       Char
	     | Int
	     | IntInf
	     | Pointer
	     | Word
	     | Word8
	     | Real
	     | Stack
	     | String
	     | Array of t
	     | Vector of t
	     | Ref of t
	     | Datatype of Tycon.t
	     | Tuple of t list
	     | Variant of Tycon.t * Con.t

	    (* must be called (once) before using dest *)
	    val initialize: unit -> unit
	       
	    val dest: t -> dest
	 end
      sharing Atoms = Type.Atoms

      structure Func:
	 sig
	    include ID
	    val saveWorld: t
	 end

      structure Jump: ID

      structure PrimExp:
	 sig
	    datatype t =
	       Const of Const.t
	     | Var of Var.t
	     | Tuple of Var.t list
	     | Select of {tuple: Var.t,
			  offset: int}
	     | ConSelect of {variant: Var.t,
			     con: Con.t,
			     offset: int}
	     | ConApp of {con: Con.t,
			  args: Var.t list}
	     | PrimApp of {prim: Prim.t,
			   targs: Type.t list,
			   args: Var.t list}

	    val replaceVar: t * (Var.t -> Var.t) -> t
	    val foreachVar: t * (Var.t -> unit) -> unit

	    val layout: t -> Layout.t
	 end

      structure Dec:
	 sig
	    type exp
	    datatype t =
	       Bind of {
			var: Var.t,
			ty: Type.t,
			exp: PrimExp.t
		       }
	     | Fun of {
		       name: Jump.t,
		       args: (Var.t * Type.t) list,
		       body: exp
		      }
	     | HandlerPush of Jump.t
	     | HandlerPop

	    val layout: t -> Layout.t
	 end

      structure Cause:
	 sig
	    datatype t = User | Dispatch | PolyEqual | Coerce
	 end
      
      structure Transfer:
	 sig
	    datatype t =
	       Call of {
			func: Func.t,
			args: Var.t list,
			cont: Jump.t option (* NONE means tail-call *)
		       }
	     | Jump of {
			dst: Jump.t,
			args: Var.t list
		       }
	     | Return of Var.t list
	     | Case of {
			cause: Cause.t,
			test: Var.t,
			cases: (Con.t * Jump.t) list, (* jumps are nullary or unary *)
			default: Jump.t option  (* jump is nullary *)
		       }
	     | Raise of Var.t list
	     | Halt (* successful termination *)
	     | Bug  (* MLton thought control couldn't reach here. *)

	    val replaceVar: t * (Var.t -> Var.t) -> t
	    val foreachVar: t * (Var.t -> unit) -> unit
	       
	    val layout: t -> Layout.t
	 end

      structure Exp:
	 sig
	    type t
	       
	    val dest: t -> {decs: Dec.t list,
			    transfer: Transfer.t}
	    val make: {decs: Dec.t list,
		       transfer: Transfer.t} -> t

	    val decs: t -> Dec.t list
	    val transfer: t -> Transfer.t

	    val prefix: t * Dec.t -> t
	    val prefixs: t * Dec.t list -> t

	    (* Visits binding occurrences before uses.
	     * handleDec is applied to a dec before visiting everything that
	     * is in the scope of the dec.  It produces a function f: unit -> unit
	     * that is applied after everything in the scope has been visited.
	     *)
	    val foreach: t * {handleDec: Dec.t -> unit -> unit,
			      handleTransfer: Transfer.t -> unit} -> unit
	    (* Different from foreach in that it only visits a dec before its
	     * scope and not after.
	     *)
	    val foreach': t * {handleDec: Dec.t -> unit,
			       handleTransfer: Transfer.t -> unit} -> unit
	    val foreachDec: t * (Dec.t -> unit) -> unit
	    (* foreach BOUND var *)
	    val foreachVar: t * (Var.t * Type.t -> unit) -> unit
	    val foreachBind:
	       t * ({var: Var.t, ty: Type.t, exp: PrimExp.t} -> unit) -> unit
	    val foreachTransfer: t * (Transfer.t -> unit) -> unit
	    val foreachCall: t * ({func: Func.t,
				   args: Var.t list,
				   cont: Jump.t option} -> unit) -> unit

	    (* Rename all bound variables and jumps in exp.
	     * Each free occurrence of a formal in the exp will be replaced
	     * by the corresponding actual.
	     * handleJump is applied to each jump that is renamed.
	     *)
	    val alphaRename:
	       {
		exp: t,
		substitution: {formal: Var.t, actual: Var.t} list,
		handleJump: {old: Jump.t, new: Jump.t} -> unit
	       } -> t

	    val clear: t -> unit

	    val layout: t -> Layout.t
	 end
      sharing type Dec.exp = Exp.t

      structure DirectExp:
	 sig
	    type t

	    val const: Const.t -> t
	    val var: Var.t * Type.t -> t
	    val tuple: {exps: t list, ty: Type.t} -> t
	    val select: {tuple: t,
			 offset: int,
			 ty: Type.t} -> t
	    val conApp: {con: Con.t,
			 args: t list,
			 ty: Type.t} -> t
	    val primApp: {prim: Prim.t,
			  targs: Type.t list,
			  args: t list,
			  ty: Type.t} -> t
	    val call: {func: Func.t,
		       args: t list,
		       ty: Type.t} -> t
	    val raisee: t -> t
	    val handlee: {try: t,
			  ty: Type.t,
			  catch: Var.t * Type.t,
			  handler: t} -> t
	    val casee: {
			cause: Cause.t,
			test: t,
			cases: {con: Con.t,
				args: (Var.t * Type.t) list,
				body: t} list,
			default: t option,
			ty: Type.t (* type of entire case expression *)
		       } -> t
	    val lett: {decs: {
			      var: Var.t,
			      ty: Type.t,
			      exp: t
			      } list,
		       body: t} -> t

	    val detuple: {tuple: t,
			  body: Var.t list -> t} -> t

	    val detupleBind: {tuple: t,
			      components: Var.t list,
			      body: t} -> t

	    val seq: t * t -> t
	    val name2: t * t * (Var.t * Var.t -> t) -> t
(* 	    val andd: t * t -> t
 * 	    val ands: t list -> t
 * 	    val or: t * t -> t
 *)
	    val truee: t
	    val falsee: t
	    val int: int -> t
	    val + : t * t -> t
	    val eq: t * t * Type.t -> t

	    val send: t * (PrimExp.t * Type.t -> Exp.t) -> Exp.t
	    val toExp: t -> Exp.t
	    val toExpHalt: t -> Exp.t
	 end

      structure Program:
	 sig
	    datatype t =
	       T of {
		     datatypes: {
				 tycon: Tycon.t,
				 cons: {
					con: Con.t,
					args: Type.t list
				       } list
				} list,
		     globals: {
			       var: Var.t,
			       ty: Type.t,
			       exp: PrimExp.t
			      } list,
		     functions: {
				 name: Func.t,
				 args: (Var.t * Type.t) list,
				 body: Exp.t,
				 returns: Type.t list
				} list,
		     main: Func.t
		    } (* must be nullary *)

	    val foreachVar: t * (Var.t * Type.t -> unit) -> unit

	    val layoutStats: t -> Layout.t

	    val layout: t -> Layout.t
	    val layouts: t * (Layout.t -> unit) -> unit

	    val clear: t -> unit
	 end
   end
