(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor DeepFlatten(S: FLATTEN_STRUCTS): FLATTEN = 
struct

open S

open Dec Function PrimExp

structure ExtractTargs = ExtractTargs(structure Prim = Prim
				      structure Type = Type)

structure Set = DisjointSet

structure TupleRep =
   struct
      datatype t = Flat | Tuple
	 
      val toString =
	 fn Flat => "Flat"
	  | Tuple => "Tuple"

      val isFlat =
	 fn Flat => true
	  | _ => false

      val layout = Layout.str o toString
   end

structure Rep =
   struct
      structure Count =
	 struct
	    datatype t =
	       Zero
	     | One
	     | Many

	    val toString =
	       fn Zero => "Zero"
		| One => "One"
		| Many => "Many"

	    val inc =
	       fn Zero => One
		| _ => Many

	    val isMany =
	       fn Many => true
		| _ => false

	    val layout = Layout.str o toString
	 end

      structure Set = DisjointSet

      structure Id = IntUniqueId()

      datatype tree =
	 Leaf of Type.t
       | Mutable of UnaryTycon.t * t
       | Tuple of {rep: TupleRep.t ref,
		   coercedFrom: t list ref,
		   components: t list}
      withtype t = {tree: tree,
		    count: Count.t ref,
		    (*			  id: Id.t, *)
		    tys: Type.t list option ref} Set.t


      val tree: t -> tree = #tree o Set.value

      fun layout s =
	 let open Layout
	 in
	    (*		  record[("id", Id.layout(#id(Set.value s))),
	     ("tree", )] *)
	    case tree s of
	       Leaf ty => paren(seq[str "Leaf ", Type.layout ty])
	     | Mutable(mt, r) =>
		  paren(seq[UnaryTycon.layout mt, str " ", layout r])
	     | Tuple{rep, components, ...} =>
		  seq[TupleRep.layout(!rep),
		      list(List.map(components, layout))]
	 end

      fun detuple r =
	 case tree r of
	    Tuple{components, ...} => components
	  | _ => Error.bug "Rep.detuple"

      fun isFlat s =
	 case tree s of
	    Tuple{rep, ...} => TupleRep.isFlat(! rep)
	  | _ => true

      fun tupleRep s =
	 case tree s of
	    Tuple{rep, ...} => !rep
	  | _ => Error.bug "Rep.tupleRep"

      local
	 fun make (mt, msg) s =
	    case tree s of
	       Mutable(mt', r) =>
		  if UnaryTycon.equals(mt, mt')
		     then r
		  else Error.bug msg
	     | _ => Error.bug msg
      in
	 val deref = make (UnaryTycon.Ref, "Rep.deref")
	 val dearray = make (UnaryTycon.Array, "Rep.dearray")
	 val devector = make (UnaryTycon.Vector, "Rep.devector")
      end

      val traceToType =
	 Trace.trace("Rep.toType", layout, Type.layout)
	 
      fun toType arg =
	 traceToType
	 (fn r =>
	  case toTypes r of
	     [] => Type.unit
	   | [t] => t
	   | _ => Error.bug "Rep.toType") arg
      and toTypes s: Type.t list =
	 let val {tree, tys, ...} = Set.value s
	 in
	    case !tys of
	       SOME tys => tys
	     | NONE =>
		  let
		     val tys' =
			case tree of
			   Leaf ty => [ty]
			 | Mutable(c, r) =>
			      [Type.con(UnaryTycon.toTycon c, [toType r])]
			 | Tuple{rep, components, ...} =>
			      let
				 val tys =
				    List.flattenMap(components, toTypes)
			      in
				 case !rep of
				    TupleRep.Tuple => [Type.tuple tys]
				  | TupleRep.Flat => tys
			      end
		  in tys := SOME tys';
		     tys'
		  end
	 end

      fun new tree = Set.singleton{tys = ref NONE,
				   (*					 id = Id.new(), *)
				   count = ref Count.Zero,
				   tree = tree}
	 

      fun copy(r: t): t = new(copyTree(tree r))
      and copyTree t =
	 case t of
	    Leaf _ => t
	  | Mutable(t, r) => Mutable(t, copy r)
	  | Tuple{components, ...} => Tuple{components = components,
					    rep = ref TupleRep.Flat,
					    coercedFrom = ref []}

      fun select{tuple: t, offset: int, resultType}: t =
	 case tree tuple of
	    Tuple{components, ...} => List.nth(components, offset)
	  | _ => Error.bug "Rep.select"
	       
      val traceUnify =
	 Trace.trace2("Rep.unify", layout, layout, Unit.layout)
	 
      fun tuplize(s: t): unit =
	 case tree s of
	    Tuple{rep, coercedFrom, ...} =>
	       (case !rep of
		   TupleRep.Tuple => ()
		 | TupleRep.Flat =>
		      (rep := TupleRep.Tuple;
		       List.foreach(!coercedFrom, fn s' => unify(s', s))))
	  | _ => ()
	       
      and unify arg: unit =
	 traceUnify
	 (fn (s: t, s': t) =>
	  if Set.equals(s, s')
	     then ()
	  else let val v = Set.value s
	       in
		  (* can't move the union before this, because the
		   * Flat-Tuple case that deals with s' *)
		  case (#tree v,     tree s')        of
		     (Leaf _,        Leaf _)         => ()
		   | (Mutable(_, s), Mutable(_, s')) => unify(s, s')
		   | (Tuple t,       Tuple t')       =>
			(case (!(#rep t),    !(#rep t'))     of
			    (TupleRep.Flat,  TupleRep.Flat)  =>
			       #coercedFrom t :=
			       !(#coercedFrom t) @ !(#coercedFrom t')
			  | (TupleRep.Tuple, TupleRep.Flat)  => tuplize s'
			  | (TupleRep.Flat,  TupleRep.Tuple) => tuplize s
			  | (TupleRep.Tuple, TupleRep.Tuple) => ();
			       List.foreach2(#components t, #components t',
					     unify))
		   | _ => Error.bug "Rep.unify";
			Set.union(s, s');
			Set.setValue(s, v)
	       end) arg

      fun reff(r: t): t =
	 (tuplize r; new(Mutable(UnaryTycon.Ref, r)))

      fun fromType(ty: Type.t): t =
	 let
	    fun mut(mt, ty) =
	       let val r = fromType ty
	       in tuplize r; Mutable(mt, r)
	       end
	 in new(case Type.dest ty of
		   Type.Ref ty => mut(UnaryTycon.Ref, ty)
		 | Type.Array ty => mut(UnaryTycon.Array, ty)
		 | Type.Vector ty => mut(UnaryTycon.Vector, ty)
		 | Type.Tuple ts => Tuple{rep = ref TupleRep.Flat,
					  coercedFrom = ref [],
					  components = List.map(ts, fromType)}
		 | _ => Leaf ty)
	 end

      val unit = fromType Type.unit

      fun incCount(s: t): unit =
	 let val {count, ...} = Set.value s
	    val c = Count.inc(!count)
	 in count := c;
	    if Count.isMany c
	       then tuplize s
	    else ()
	 end

      val const = fromType o Type.ofConst
      
      fun tuple(rs: t list): t =
	 (List.foreach(rs, incCount);
	  new(Tuple{rep = ref TupleRep.Flat,
		    coercedFrom = ref [],
		    components = rs}))

      val traceCoerce =
	 Trace.trace("Rep.coerce",
		     fn {from, to} =>
		     let open Layout
		     in record[("from", layout from),
			       ("to", layout to)]
		     end, Unit.layout)

      fun coerce arg: unit =
	 traceCoerce
	 (fn {from: t, to: t} =>
	  if Set.equals(from, to)
	     then ()
	  else (case (tree from, tree to)   of
		   (Leaf _,    Leaf _)    => ()
		 | (Mutable _, Mutable _) => unify(from, to)
		 | (Tuple t,   Tuple t') =>
		      (case !(#rep t') of
			  TupleRep.Flat =>
			     (ListRef.push(#coercedFrom t', from);
			      List.foreach2(#components t, #components t',
					    fn (r, r') =>
					    coerce{from = r, to = r'}))
			| TupleRep.Tuple => unify(from, to))
		 | _ => Error.bug "Representation.coerce")
	     ) arg

      fun primApp{prim, targs, args, resultVar, resultType} =
	 (List.foreach(args, tuplize);
	  let
	     open Prim.Name
	     val r =
		case (Prim.name prim, args) of
		   (ArraySub, [a, _]) => dearray a
		 | (VectorSub, [v, _]) => devector v
		 | (ArrayUpdate, [a, _, y]) => (unify(dearray a, y); unit)
		 | (Ref, [y]) => reff y
		 | (Deref, [r]) => deref r
		 | (Assign, [r, y]) => (unify(deref r, y); unit)
		 | _ => fromType resultType
	  in tuplize r; r
	  end)

      fun filter(v: t, c: Con.t): t = v
   end


fun flatten(program as Program.T{datatypes, globals, functions, main}) =
   let
      val {value, func, conArgs, ...} =
	 Control.traceCall "analyze"
	 analyze{program = program,
		 useFromTypeOnBinds = false,
		 fromType = Rep.fromType,
		 copy = Rep.copy,
		 coerce = Rep.coerce,
		 const = Rep.const,
		 tuple = Rep.tuple,
		 select = Rep.select,
		 constructors = NONE,
		 primApp = Rep.primApp,
		 filter = Rep.filter}

   (* diagnostic *)
      val _ =
	 Control.displays
	 ("flat", fn display =>
	  let open Layout
	     fun tup rs = tuple(List.map(rs, Rep.layout))
	  in
	     List.foreach
	     (functions, fn {name, ...} =>
	      display(seq[Func.layout name, str " ",
			  let val {args, returns} = func name
			  in seq[tup args, str " -> ", tup returns]
			  end]));
	     List.foreach
	     (datatypes, fn {cons, ...} =>
	      display(align(List.map
			    (cons, fn {con, ...} =>
			     seq[Con.layout con,
				 str " ",
				 tup(conArgs con)]))))
	  end)
   in
      Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

end
