(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Flatten arguments to jumps, constructors, and functions.
 * If a tuple is explicitly available at all uses of a jump (resp. function)
 * then
 *   - The formals and call sites are changed so that the components of the
 *     tuple are passed.
 *   - The tuple is reconstructed at the beginning of the body of the jump.
 *
 * Similarly, if a tuple is explicitly available at all uses of a constructor,
 *   - The constructor argument type is changed to flatten the tuple type.
 *   - The tuple is passed flat at each ConApp.
 *   - The tuple is reconstructed at each ConSelect.
 *
 * The trickiest bit is for constructors, because ConSelect offsets must be
 * changed and the tuple must be reconstructed.  For example,
 *
 * datatype t = T of [(u * v), (w * z)]
 * val x = ConSelect{con = T, offset = 2, variant = v}
 *
 * Suppose that both args to T are flattened.  Then this becomes:
 *
 * datatype t = T of [u, v, w, z]
 * val a = ConSelect{con = T, offset = 3, variant = v}
 * val b = ConSelect{con = T, offset = 4, variant = v}
 * val x = (a, b)
 *
 *)

functor Flatten(S: FLATTEN_STRUCTS): FLATTEN = 
struct

open S
open Dec PrimExp Transfer

structure Rep =
   struct
      structure Set = DisjointSet
	 
      datatype t =
	 Leaf
       | Tuple of bool Set.t (* true means keep it as a tuple *)

      local open Layout
      in
	 val layout =
	    fn Leaf => str "leaf"
	     | Tuple s => Bool.layout(Set.value s)
      end

      val isFlat =
	 fn Tuple s => not(Set.value s)
	  | _ => false
	       
      fun fromType t =
	 case Type.detupleOpt t of
	    NONE => Leaf
	  | SOME _ => Tuple(Set.singleton false)

      fun fromTypes ts = List.map(ts, fromType)

      fun fromFormals(xts: (Var.t * Type.t) list) = List.map(xts, fromType o #2)

      val tuplize: t -> unit =
	 fn Leaf => ()
	  | Tuple s => Set.setValue(s, true)

      fun tuplizes rs = List.foreach(rs, tuplize)

      val unify =
	 fn (Leaf, Leaf) => ()
	  | (Tuple s, Tuple s') =>
	       let val isTuple = Set.value s orelse Set.value s'
	       in Set.union(s, s');
		  Set.setValue(s, isTuple)
	       end
	  | _ => Error.bug "unify"

      fun unifys(rs, rs') = List.foreach2(rs, rs', unify)

      val layouts = List.layout layout
	 
(*      val unifys = Trace.trace2("unifys", layouts, layouts, Unit.layout) unifys *)
   end

fun flatten(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("args", Con.layout))

      val conArgs = #args o conInfo
      val conArgTypes = #argTypes o conInfo

      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach(cons, fn {con, args} =>
		       setConInfo(con, {argTypes = args,
					args = Rep.fromTypes args})))

      val {get = funcInfo: Func.t -> {args: Rep.t list,
					  returns: Rep.t list},
	   set = setFuncInfo} =
	 Property.new(Func.plist,
		      Property.initRaise("info", Func.layout))

      val funcArgs = #args o funcInfo

(*       val funcInfo =
 * 	 Trace.trace("funcInfo", Func.layout,
 * 		     fn {args, returns} =>
 * 		     Layout.record[("args", Rep.layouts args),
 * 				   ("returns", Rep.layouts returns)])
 * 	 funcInfo
 *)

      val _ =
	 List.foreach
	 (functions, fn {name, args, returns, ...} =>
	  setFuncInfo(name, {args = Rep.fromFormals args,
			     returns = Rep.fromTypes returns}))

      val {get = varTuple, set = setVarTuple} =
	 Property.new(Var.plist, Property.initConst NONE)

      val {get = jumpArgs: Jump.t -> Rep.t list, set = setJumpArgs} =
	 Property.new(Jump.plist, Property.initRaise("args", Jump.layout))

      fun coerces(xs: Var.t list, rs: Rep.t list) =
	 List.foreach2
	 (xs, rs, fn (x, r) => (case varTuple x of
				   NONE => Rep.tuplize r
				 | _ => ()))

(*       val coerces =
 * 	 Trace.trace2("coerces",
 * 		      List.layout Var.layout,
 * 		      List.layout Rep.layout,
 * 		      Unit.layout)
 * 	 coerces
 *)

      fun loopBind{var, ty, exp} =
	 case exp of
	    Tuple xs => setVarTuple(var, SOME xs)
	  | ConApp{con, args} => coerces(args, conArgs con)
	  | _ => ()

      val _ = List.foreach(globals, loopBind)
	       
      fun loopExp(e, returns) =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach
	    (decs,
	     fn Bind b => loopBind b
	      | Fun{name, args, body} =>
		   (setJumpArgs(name, Rep.fromFormals args)
		    ; loopExp(body, returns))
	      | HandlerPush j => Rep.tuplizes(jumpArgs j)
	      | _ => ());
	    case transfer of
	       Return xs => coerces(xs, returns)
	     | Call{func, args, cont} =>
		  let val {args = formals, returns = returns'} = funcInfo func
		  in coerces(args, funcArgs func);
		     Rep.unifys(returns',
				case cont of
				   NONE => returns
				 | SOME c => jumpArgs c)
		  end
	      | Jump{dst, args} => coerces(args, jumpArgs dst)
	      | _ => ()
	 end

      val _ =
	 Control.traceCall "analyze"
	 List.foreach(functions, fn {name, body, ...} =>
		      loopExp(body, #returns(funcInfo name)))

      val _ =
	 Control.displays
	 ("flat", fn display =>
	  List.foreach(functions, fn {name, ...} =>
		       let val {args, returns} = funcInfo name
			  open Layout
		       in display(seq[Func.layout name,
				      str " ",
				      align[List.layout Rep.layout args,
					    List.layout Rep.layout returns]])
		       end))

      fun flattenTypes(ts: Type.t list, rs: Rep.t list): Type.t list =
	 List.fold2(ts, rs, [], fn (t, r, ts) =>
		    if Rep.isFlat r
		       then List.fold(Type.detuple t, ts, op ::)
		    else t :: ts)

(*       val flattenTypes =
 * 	 Trace.trace2("flattenTypes",
 * 		      List.layout Type.layout,
 * 		      List.layout Rep.layout,
 * 		      List.layout Type.layout)
 * 	 flattenTypes
 *)

      val datatypes =
	 List.map
	 (datatypes, fn {tycon, cons} =>
	  {tycon = tycon,
	   cons =
	   List.map(cons, fn {con, args} =>
		    {con = con,
		     args = flattenTypes(args, conArgs con)})})
	 
      fun flattens(xs: Var.t list, rs: Rep.t list) =
	 List.fold2(xs, rs, [],
		    fn (x, r, xs) =>
		    if Rep.isFlat r
		       then (case varTuple x of
				SOME ys => List.fold(ys, xs, op ::)
			      | _ => Error.bug "tuple unavailable")
		     else x :: xs)

      fun loopBind{var: Var.t, ty: Type.t, exp} =
	 let fun single e = [{var = var, ty = ty, exp = e}]
	 in case exp of
	    ConApp{con, args} =>
	       single(ConApp{con = con, args = flattens(args, conArgs con)})
	  | ConSelect{variant, con, offset} =>
	       let val {args, argTypes} = conInfo con
		  val args = List.nthTail(args, offset)
		  val argTypes = List.nthTail(argTypes, offset)
	       in case (args, argTypes) of
		  (r :: rs, t :: ts) =>
		     let
			val offset =
			   List.fold2(rs, ts, 0, fn (r, t, offset) =>
				      offset
				      + (if Rep.isFlat r
					    then List.length(Type.detuple t)
					 else 1))
		     in if Rep.isFlat r
			   then
			      let val ts = Type.detuple t
				 val xs = List.map(ts, fn _ => Var.newNoname())
			      in #2(List.fold2
				    (List.reverse xs,
				     List.reverse ts,
				     (offset, single(Tuple xs)),
				     fn (x, t, (offset, bs)) =>
				     (offset + 1,
				      {var = x, ty = t,
				       exp = ConSelect{variant = variant,
						       con = con,
						       offset = offset}}
				      :: bs)))
			      end
			else single(ConSelect{variant = variant,
					      con = con,
					      offset = offset})
		     end
		| _ => Error.bug "bad ConSelect"
	       end
	  | _ => single exp
	 end

      val globals = List.foldr(globals, [], fn (b, bs) => loopBind b @ bs)

      fun loopExp(e: Exp.t, returns: Rep.t list): Exp.t =
	 let val {decs, transfer} = Exp.dest e

	    fun loopDecs ds =
	       List.foldr
	       (ds, [], fn (d, ds) =>
		case d of
		  Fun{name, args, body} =>
		     let val (args, body) = loopArgsBody(args, jumpArgs name,
							 body, returns)
		     in Fun{name = name, args = args, body = body}
		     end :: ds
		| Bind b => List.foldr(loopBind b, ds, fn (b, ds) =>
				       Bind b :: ds)
		| _ => d :: ds)

	    val transfer =
	       case transfer of
		  Return xs => Return(flattens(xs, returns))
		| Call{func, args, cont} =>
		     Call{func = func, cont = cont,
			  args = flattens(args, funcArgs func)}
		| Jump{dst, args} =>
		     Jump{dst = dst,
			  args = flattens(args, jumpArgs dst)}
		| _ => transfer
		     
	 in Exp.make{decs = loopDecs decs,
		     transfer = transfer}
	 end

      and loopArgsBody(args: (Var.t * Type.t) list, rs: Rep.t list,
		       body: Exp.t, returns: Rep.t list) =
	 let
	    val (args, prefix) =
	       List.fold2
	       (args, rs, ([], []),
		fn ((x, t), r, (args, prefix)) =>
		if Rep.isFlat r
		   then let val ts = Type.detuple t
			    val xs = List.map(ts, fn _ => Var.newNoname())
			    val args =
			       List.fold2
			       (xs, ts, args, fn (x, t, args) =>
				(x, t) :: args)
			in (args,
			    Bind{var = x, ty = t, exp = Tuple xs} :: prefix)
			end
		else ((x, t) :: args, prefix))
	 in (args, Exp.prefixs(loopExp(body, returns), prefix))
	 end

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let val {args = argReps, returns = returnReps} = funcInfo name
	     val (args, body) = loopArgsBody(args, argReps, body, returnReps)
	  in {name = name,
	      args = args,
	      body = shrinkExp body,
	      returns = flattenTypes(returns, returnReps)}
	  end)

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end
   
end
