(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor InferHandlers(S: INFER_HANDLERS_STRUCTS): INFER_HANDLERS = 
struct

open S

structure Set = DisjointSet
      
structure Handler' =
   struct
      datatype t = T of Jump.t option Set.t

      val new = T o Set.singleton o SOME
      fun unknown() = T(Set.singleton NONE)

      fun equals(T s, T s'): bool =
	 Set.canUnion
	 (s, s',
	  fn (NONE, h) => SOME h
	   | (h, NONE) => SOME h
	   | (SOME h, SOME h') =>
		if Jump.equals(h, h') then SOME(SOME h) else NONE)

      fun dest(T s) =
	 case Set.value s of
	    NONE => Error.bug "unknown handler"
	  | SOME h => h

      fun layout(T s) =
	 Option.layout Jump.layout (Set.value s)
   end

structure Handlers =
   struct
      datatype t = T of handlers Set.t
      and handlers =
	 Unknown
	| Empty
	| Push of {top: Handler'.t,
		   rest: t}

      val new = T o Set.singleton
	  
      val empty = new Empty
      fun unknown() = new Unknown
      fun push(rest, top) = new(Push{top = top,
				     rest = rest})

      fun pop(T s) =
	 case Set.value s of
	    Empty => Error.bug "pop of empty handler stack"
	  | Push{rest, ...} => rest
	  | Unknown =>
	       let val rest = unknown()
	       in Set.setValue(s, Push{top = Handler'.unknown(),
				       rest = rest})
		  ; rest
	       end

      fun isEmpty(T s) =
	 case Set.value s of
	    Empty => true
	  | _ => false

      fun toList hs =
	 let
	    fun loop(T s, ac) =
	       case Set.value s of
		  Unknown => Error.bug "toList of unknown handler stack"
		| Empty => rev ac
		| Push{top, rest} => loop(rest, Handler'.dest top :: ac)
	 in loop(hs, [])
	 end

      fun equals(T s, T s'): bool =
	 Set.canUnion
	 (s, s',
	  fn (Unknown, hs) => SOME hs
	   | (hs, Unknown) => SOME hs
	   | (Empty, Empty) => SOME Empty
	   | (hs as Push{top = t, rest = r}, Push{top = t', rest = r'}) =>
		if Handler'.equals(t, t') andalso equals(r, r')
		   then SOME hs
		else NONE
	   | _ => NONE)
		     
      fun layout(T s) =
	 let open Layout
	 in case Set.value s of
	    Unknown => str "Unknown"
	  | Empty => str "Empty"
	  | Push{top, rest} => seq[Handler'.layout top,
				   str " :: ",
				   layout rest]
	 end
   end

open Dec PrimExp Transfer

fun inferHandlers(Program.T{functions, ...}) =
   let
      val {get = jump, set = setJump} =
	 Property.new(Jump.plist, Property.initRaise("handlers", Jump.layout))
   in
      List.foreach
      (functions, fn {body, ...} =>
       let
	  fun loop(e: Exp.t, hs: Handlers.t): unit =
	     let
		fun check(msg, hs, hs') =
		   if Handlers.equals(hs, hs')
		      then ()
		   else (Control.message
			 (fn () => let open Layout
				   in align[seq[str "exp: ", Exp.layout e],
					    seq[str "hs: ", Handlers.layout hs],
					    seq[str "hs': ", Handlers.layout hs']]
				   end)
			 ; Error.bug("handler stack mismatch at " ^ msg))

		val {decs, transfer} = Exp.dest e
		val after =
		   List.fold
		   (decs, hs, fn (d, hs) =>
		    case d of
		       Fun{name, body, ...} =>
			  (let val hs = Handlers.unknown()
			   in setJump(name, hs)
			   end
			   ; loop(body, jump name)
			   ; hs)
		     | HandlerPush h =>
			  let val hs = Handlers.push(hs, Handler'.new h)
			  in check("push", hs, jump h)
			     ; hs
			  end
		     | HandlerPop => Handlers.pop hs
		     | _ => hs)
		fun checkJump j = check("jump", after, jump j)
		fun empty msg = check(msg, after, Handlers.empty)
	     in case transfer of
		Jump{dst, ...} => checkJump dst
	      | Case{cases, default, ...} =>
		   (List.foreach(cases, checkJump o #2)
		    ; Option.map'(default, checkJump))
	      | Halt => empty "halt"
	      | Return _ => empty "return"
	      | Bug => ()
	      | Raise _ => ()
	      | Call{cont = c, ...} =>
		   (case c of
		       NONE => empty "tail call"
		     | SOME c => check("nontail call", after, jump c))
	     end
       in loop(body, Handlers.empty)
       end)

(*      Trace.trace("jumpHandlers", Jump.layout, List.layout Handler.layout) *)
      ; (Handlers.toList o jump)
   end

val inferHandlers = Control.traceCall "inferHandlers" inferHandlers
   
fun deltaHandlers(d, hs) =
   case d of
      HandlerPush h => h :: hs
    | HandlerPop => (case hs of
			_ :: hs => hs
		      | _ => Error.bug "deltaHandlers")
    | _ => hs

end
