(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Inline(S: INLINE_STRUCTS): INLINE = 
struct

open S

open Dec PrimExp Transfer

(* Returns true if the size of the expression (measured somewhat arbitrarily)
 * is greater than SOME integer.  Returns true on NONE.
 *)
fun sizeGreaterThan(exp: Exp.t, size: int option): bool =
   case size of
      NONE => false
    | SOME size =>
	 let exception Big
	    val r = ref 0
	    fun incSize n =
	       let val new = !r + n
	       in if new > size
		     then raise Big
		  else r := new
	       end
	    val primExpSize =
	       fn Const _ => 0
		| Var _ => 0
		| Select _ => 1
		| ConSelect _ => 1
		| Tuple xs => List.length xs
		| ConApp{args, ...} => 1 + List.length args
		| PrimApp{args, ...} => 1 + List.length args
	    val transferSize =
	       fn Call{args, ...} => 1 + List.length args
		| Jump{args, ...} => 1 + List.length args
		| Return xs => 1 + List.length xs
		| Raise xs => 1 + List.length xs
		| Case{cases, ...} => 1 + List.length cases
		| Halt => 1
		| Bug => 1
	    val decSize =
	       fn Fun _ => 0
		| Bind{exp, ...} => primExpSize exp
		| _ => 1
	 in (Exp.foreach'(exp, {handleDec = incSize o decSize,
				handleTransfer = incSize o transferSize})
	     ; false)
	    handle Big => true
	 end

local
   fun 'a make (dontInlineBody: Exp.t * 'a -> bool)
      (Program.T{functions, ...}, a: 'a): Func.t -> bool =
   let
      val {get = shouldInline, set = setShouldInline} =
	 Property.new(Func.plist, Property.initConst false)
   in
      List.foreach(functions, fn {name, body, ...} =>
		   if dontInlineBody(body, a)
		      then ()
		   else setShouldInline(name, true))
      ; shouldInline
   end

   fun containsCall(exp: Exp.t): bool =
      let exception Yes
      in (Exp.foreachTransfer(exp, fn Call _ => raise Yes | _ => ())
	  ; false) handle Yes => true
      end

   fun containsLoop(exp: Exp.t): bool =
      let
	 exception Yes
	 val {get, set, destroy} =
	    Property.newDest(Jump.plist, Property.initConst false)
	 val loopTransfer =
	    fn Jump{dst, ...} => if get dst then raise Yes else ()
	     | _ => ()
	 fun loopExp(e: Exp.t): unit =
	    let val {decs, transfer} = Exp.dest e
	    in List.foreach(decs, loopDec)
	       ; loopTransfer transfer
	    end
	 and loopDec d =
	    case d of
	       Fun{name, body, ...} =>
		  (set(name, true); loopExp body; set(name, false))
	     | _ => ()
      in ((loopExp exp; false) handle Yes => true)
	 before destroy()
      end
in
   val leaf = make(fn (body, {size}) =>
		   sizeGreaterThan(body, size)
		   orelse containsCall body)

   val leafNoLoop = make(fn (body, {size}) =>
			 sizeGreaterThan(body, size)
			 orelse containsCall body
			 orelse containsLoop body)
end

structure Graph = DirectedGraph
structure Node = Graph.Node

fun nonRecursive(Program.T{functions, ...}, {size: int option}) =
   let
      val {get = funcInfo: Func.t -> {shouldInline: bool ref,
				      isBig: bool,
				      node: Node.t,
				      numCalls: int ref},
	   set = setFuncInfo} =
	 Property.new
	 (Func.plist, Property.initRaise("func info", Func.layout))

      val {set = setNodeFunc, get = nodeFunc} =
	 Property.new(Node.plist, Property.initRaise("func", Node.layout))

      val graph = Graph.new()

      (* initialize the info for each func *)
      val _ = 
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  let val n = Graph.newNode graph
	  in setNodeFunc(n, name)
	     ; setFuncInfo(name, {node = n,
				  isBig = sizeGreaterThan(body, size),
				  numCalls = ref 0,
				  shouldInline = ref true})
	  end)

      (* Update call counts. *)
      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  Exp.foreachTransfer
	  (body,
	   fn Call{func, ...} =>
	        let val {numCalls, shouldInline, isBig, ...} = funcInfo func
		   val n = !numCalls
		in numCalls := n + 1;
		   if ((n = 1 andalso isBig)
		       orelse Func.equals(name, func))
		      then shouldInline := false
		   else ()
		end
	    | _ => ()))
	 
      (* Build the call graph of potentially inlinable functions.
       * Don't add edges for functions that are not inlined, since they
       * can't cause infinite unrolling.
       *)
      val _ = 
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  let val {node = caller, shouldInline = si, ...} = funcInfo name
	  in Exp.foreachTransfer
	     (body,
	      fn Call{func, ...} =>
 	           let
		      val {numCalls, shouldInline = si', node = callee, ...} =
			 funcInfo func
		   in if !si andalso !si'
			 then Graph.addEdge(graph, {from = caller, to = callee})
		      else ()
		   end
	       | _ => ())
	  end)
      
      (* Compute strongly-connected components and set any function that is
       * in a nontrivial scc to be not inlined.
       *)
      val _ = 
	 List.foreach
	 (Graph.stronglyConnectedComponents graph,
	  fn [] => ()
	   | [_] => ()
	   | ns => List.foreach(ns, fn n =>
				#shouldInline(funcInfo(nodeFunc n)) := false))
   in ! o #shouldInline o funcInfo
   end

fun inline(p as Program.T{datatypes, globals, functions, main}) =
   let
      val shouldInline: Func.t -> bool =
	 case !Control.inline of
	    Control.NonRecursive r => nonRecursive(p, r)
	  | Control.Leaf r => leaf(p, r)
	  | Control.LeafNoLoop r => leafNoLoop(p, r)
	  | _ => Error.unimplemented "inline"
	    
      val {get = funInfo: Func.t -> {args: Var.t list,
				     body: Exp.t},
	   set = setFunInfo} =
	 Property.new(Func.plist, Property.initRaise("info", Func.layout))

      val _ = List.foreach(functions, fn {name, args, body, ...} =>
			   setFunInfo(name, {args = List.map(args, #1),
					     body = body}))

      fun inlineExp(exp: Exp.t, outerCont: Jump.t option): Exp.t =
	 let
	    fun loopExp e =
	       let val {decs, transfer} = Exp.dest e
		  val decs =
		     List.map
		     (decs,
		      fn Fun{name, args, body} =>
			    Fun{name = name, args = args, body = loopExp body}
		       | d => d)
		  fun simple t = Exp.make{decs = decs, transfer = t}
	       in case transfer of
		  Return xs =>
		     simple(case outerCont of
			       NONE => Return xs
			     | SOME j => Jump{dst = j, args = xs})
		| Call{func, args, cont} =>
		     let
			val cont =
			   case cont of
			      NONE => outerCont
			    | SOME c => SOME c
			val {args = formals, body} = funInfo func
		     in if shouldInline func
			   then
			      let
				 val body =
				    Exp.alphaRename
				    {exp = body,
				     handleJump = fn _ => (),
				     substitution =
				     List.map2(formals, args, fn (f, a) =>
					       {formal = f, actual = a})}
			      in Exp.prefixs(inlineExp(body, cont), decs)
			      end
			else simple(Call{func = func, args = args,
					 cont = cont})
		     end
		| _ => simple transfer
	       end

	 in loopExp exp
	 end

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.fold
	 (functions, [], fn ({name, args, body, returns}, functions) =>
	  if not(Func.equals(name, main))
	     andalso shouldInline name
	     then functions
	  else {name = name, args = args, returns = returns,
		body = shrinkExp(inlineExp(body, NONE))}
	     :: functions)

   in Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

end
