(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Once(S: ONCE_STRUCTS): ONCE = 
struct

open S
open Dec PrimExp Transfer

structure Graph = DirectedGraph
structure Node = Graph.Node

(*
 * Build control flow graph for main function.
 * Compute strongly-connected components and set the vars for any basic
 * block that is a trivial component and does not jump to itself to be once.
 *)
   
fun once(program as Program.T{globals, functions, main, ...}) =
   let
      val {get = once, set = setOnce} =
	 Property.new(Var.plist, Property.initConst false)

      val {set = setNode, get = jumpNode} =
	 Property.new(Jump.plist, Property.initRaise("info", Jump.layout))

      val {set = setNodeInfo, get = nodeInfo} =
	 Property.new(Node.plist, Property.initRaise("jump", Node.layout))

      val nodeVars = #vars o nodeInfo
      val nodeName = #name o nodeInfo

      val jumpHandlers = inferHandlers program

      val graph = Graph.new()

      fun loop(e: Exp.t, vars: Var.t list ref, node: Node.t,
	       handlers: Jump.t list) =
	 let val {decs, transfer} = Exp.dest e
	    fun jump j = Graph.addEdge(graph, {from = node, to = jumpNode j})
	    fun handler() =
	       case List.fold(decs, handlers, deltaHandlers) of
		  h :: _ => jump h
		| _ => ()
	 in List.foreach
	    (decs, 
	     fn Bind{var, ...} => ListRef.push(vars, var)
	      | Fun{name, body, ...} =>
		   let val n = Graph.newNode graph
		      val vars = ref []
		   in setNodeInfo(n, {name = Jump.layout name,
				      vars = vars});
		      setNode(name, n);
		      loop(body, vars, n, jumpHandlers name)
		   end
	      | _ => ());
	    
	    case transfer of
	       Call{cont, ...} => (case cont of
				      NONE => ()
				    | SOME c => (jump c; handler()))
	     | Jump{dst, ...} => jump dst
	     | Case{cases, default, ...} => (List.foreach(cases, jump o #2);
					     case default of
						NONE => ()
					      | SOME j => jump j)
	     | Raise _ => handler()
	     | _ => ()
	 end

      val {body, ...} =
	 case List.peek(functions, fn {name, ...} => Func.equals(name, main)) of
	    NONE => Error.bug "no main"
	  | SOME r => r

      val _ = 
	 let val n = Graph.newNode graph
	    val vars = ref []
	 in setNodeInfo(n, {name = Layout.str "main",
			    vars = vars});
	    loop(body, vars, n, [])
	 end
   in
      Control.displays
      ("once-graph", fn display =>
       (display(Exp.layout body);
	display(Layout.str "\n\n");
	Graph.display{graph = graph,
		      display = display,
		      layoutNode = nodeName}));
	
      List.foreach
      (Graph.stronglyConnectedComponents graph,
       fn [] => ()
	| [n] =>
	     if List.exists(Node.successors n, fn n' => Node.equals(n, n'))
		then ()
	     else List.foreach(!(nodeVars n), fn x => setOnce(x, true))
	   | _ => ());

      List.foreach(globals, fn {var, ...} => setOnce(var, true));
      
      Control.displays
      ("once-vars", fn display =>
       Exp.foreachVar(body, fn (x, _) => if once x then display(Var.layout x)
					 else ()));

      once
   end
   
end
