(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PolyEqual(S: POLY_EQUAL_STRUCTS): POLY_EQUAL = 
struct

open S

(*
 * This pass implements polymorphic equality.
 *
 * For each datatype tycon and vector type, it builds an equality function and
 * translates calls to = into calls to that function.
 *
 * Also generates calls to primitives stringEqual and intInfEqual.
 *
 * For tuples, it does the equality test inline.  I.E. it does not create
 * a separate equality function for each tuple type.
 *
 * All equality functions are only created if necessary, i.e. if equality
 * is actually used at a type.
 *
 * Optimizaions:
 *  - For datatype tycons that are enumerations, do not build a case dispatch,
 *    just use eq, since you know the backend will represent these as ints.
 *  - Deep equality always does an eq test first.
 *  - If one argument to = a constant int and the type will get translated
 *    to an IntOrPointer, then just use eq instead of the full equality.  This
 *    is important for implementing code like the following efficiently:
 *       if x = 0  ...    (where x is an IntInf.int)
 *)

open Dec PrimExp Transfer

structure Dexp =
   struct
      open DirectExp

      val casee =
	 fn {test, cases, default, ty} =>
	 casee{cause = Cause.PolyEqual,
	       test = test, cases = cases, default = default, ty = ty}
	 
      fun andd(e1, e2) =
	 casee{test = e1,
	       cases = [{con = Con.truee, args = [], body = e2},
			{con = Con.falsee, args = [], body = falsee}],
	       default = NONE,
	       ty = Type.bool}

      fun ands es =
	 case es of
	    [] => truee
	  | [e] => e
	  | e :: es => andd(e, ands es)

      fun or(e1, e2) =
	 casee{test = e1,
	       cases = [{con = Con.truee, args = [], body = truee},
			{con = Con.falsee, args = [], body = e2}],
	       default = NONE,
	       ty = Type.bool}
   end

val toExp = (shrinkExp []) o Dexp.toExp

fun polyEqual(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = info, set = setInfo} =
	 Property.new(Tycon.plist, Property.initRaise("info", Tycon.layout))

      val {get = isConst, set = setConst} =
	 Property.new(Var.plist, Property.initConst false)

      val isEnum = #isEnum o info
      val tyconCons = #cons o info

      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  setInfo(tycon,
		  {isEnum = List.forall(cons,
					fn {args = [], ...} => true
					 | _ => false),
		   cons = cons}))
	 
      val newFunctions = ref []
	 
      val {get = getEqualFunc, set = setEqualFunc} =
	 Property.new(Tycon.plist, Property.initConst NONE)

      val {get = getVectorEqualFunc,
	   set = setVectorEqualFunc,
	   destroy = destroyVectorEqualFunc, ...} =
	 Property.newDest(Type.plist, Property.initConst NONE)

      fun equalFunc(tycon: Tycon.t): Func.t =
	 case getEqualFunc tycon of
	    SOME f => f
	  | NONE =>
	       let val name = Func.newString "equal"
		  val _ = setEqualFunc(tycon, SOME name)
		  val arg1 = Var.newNoname()
		  val arg2 = Var.newNoname()
		  val ty = Type.con(tycon, [])
		  val args = [(arg1, ty), (arg2, ty)]
		  val body =
		     Dexp.casee
		     {test = Dexp.var(arg1, ty),
		      default = NONE, ty = Type.bool,
		      cases =
		      List.map
		      (tyconCons tycon, fn {con, args} =>
		       let
			  fun makeArgs() =
			     List.map(args, fn t => (Var.newNoname(), t))
			  val xs = makeArgs()
			  val ys = makeArgs()
		       in {con = con, args = xs,
			   body =
			   Dexp.casee
			   {test = Dexp.var(arg2, ty),
			    cases = [{con = con, args = ys,
				      body = Dexp.ands(List.map2(xs, ys,
								 fn ((x, t), (y, _)) =>
								 equal(x, y, t)))}],
			    default = SOME(Dexp.falsee),
			    ty = Type.bool}}
		       end)}
	       in ListRef.push(newFunctions,
			       {name = name,
				args = args,
				body = toExp body,
				returns = [Type.bool]});
		  name
	       end
      
      and vectorEqualFunc(t: Type.t): Func.t =
	 case getVectorEqualFunc t of
	    SOME f => f
	  | NONE =>
	       let
		  (* build two functions, one that checks the lengths and the
		   * other that loops
		   *)
		  val name = Func.newString "vectorEqual"
		  val _ = setVectorEqualFunc(t, SOME name)
		  val loop = Func.newString "vectorEqualLoop"
		  val vt = Type.vector t
		  val v1 = (Var.newNoname(), vt)
		  val v2 = (Var.newNoname(), vt)
		  val _ =
		     ListRef.push
		     (newFunctions,
		      {name = name, 
		       args = [v1, v2],
		       returns = [Type.bool],
		       body =
		       toExp
		       (let val v1 = Dexp.var v1
			    val v2 = Dexp.var v2
			    fun length e = Dexp.primApp{prim = Prim.vectorLength,
							targs = [t],
							args = [e],
							ty = Type.int}
			in Dexp.andd(Dexp.eq(length v1, length v2, Type.int),
				     Dexp.call
				     {func = loop,
				      args = [Dexp.int 0, length v1, v1, v2],
				      ty = Type.bool})
			end)})
		  val v1 = (Var.newNoname(), vt)
		  val v2 = (Var.newNoname(), vt)
		  val i = (Var.newNoname(), Type.int)
		  val len = (Var.newNoname(), Type.int)
		  val _ = 
		     ListRef.push
		     (newFunctions,
		      {name = loop,
		       args = [i, len, v1, v2],
		       returns = [Type.bool],
		       body =
		       toExp
		       (let val i = Dexp.var i
			    val len = Dexp.var len
			    val v1 = Dexp.var v1
			    val v2 = Dexp.var v2
			    fun sub(v, i) = Dexp.primApp{prim = Prim.vectorSub,
							 targs = [t],
							 args = [v, i],
							 ty = t}
			in Dexp.or(Dexp.eq(i, len, Type.int),
				   Dexp.name2
				   (sub(v1, i), sub(v2, i),
				    fn (x1, x2) =>
				    Dexp.andd(equal(x1, x2, t),
					      Dexp.call
					      {func = loop,
					       args = [Dexp.+(i, Dexp.int 1),
						       len, v1, v2],
					       ty = Type.bool})))
			end)})
	       in name
	       end

       and equal(x1: Var.t, x2: Var.t, ty: Type.t): Dexp.t =
	 let val e1 = Dexp.var(x1, ty)
	    val e2 = Dexp.var(x2, ty)
	    fun prim(p, targs) =
	       Dexp.primApp{prim = p, targs = targs, args = [e1, e2],
			    ty = Type.bool}
	    fun eq() = prim(Prim.eq, [ty])
	    fun hasConstArg() = isConst x1 orelse isConst x2
	 in case Type.dest ty of
	    Type.Char => eq()
	  | Type.Int => eq()
	  | Type.IntInf => if hasConstArg()
			      then eq()
			   else prim(Prim.intInfEqual, [])
	  | Type.Word => eq()
	  | Type.Word8 => eq()
	  | Type.String => prim(Prim.stringEqual, [])
	  | Type.Array _ => eq()
	  | Type.Vector t =>
	       Dexp.call{func = vectorEqualFunc t,
			 args = [e1, e2],
			 ty = Type.bool}
	  | Type.Ref _ => eq()
	  | Type.Datatype tycon =>
	       if isEnum tycon orelse hasConstArg()
		  then eq()
	       else Dexp.call{func = equalFunc tycon,
			      args = [e1, e2],
			      ty = Type.bool}
	  | Type.Tuple tys =>
	       let
		  (* test components i, i+1, ..., max-1 *)
		  fun loop(i, tys: Type.t list): Dexp.t =
		     case tys of
			[] => Dexp.truee
		      | ty :: tys =>
			   Dexp.name2
			   (Dexp.select{tuple = e1, offset = i, ty = ty},
			    Dexp.select{tuple = e2, offset = i, ty = ty},
			    fn (y1, y2) =>
			    Dexp.andd(equal(y1, y2, ty), loop(i + 1, tys)))
	       in loop(0, tys)
	       end
	  | _ => Error.bug "equal of strange type"
	 end

      fun loopBind{var, ty, exp} =
	 let fun const() = setConst(var, true)
	 in case exp of
	    Const c =>
	       (case Const.node c of
		   Const.SmallIntInf _ => const()
		 | _ => ())
	  | ConApp{args = [], ...} => const()
	  | _ => ()
	 end

      val _ = List.foreach(globals, loopBind)

      fun loopExp e =
      	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(decs,
			 fn Bind b => loopBind b
			  | _ => ())
	    ; (List.foldr
	       (decs, Exp.make{decs = [], transfer = transfer}, fn (d, e) =>
		let fun simple d = Exp.prefix(e, d)
		   fun normal() = simple d
		in case d of
		   Fun{name, args, body} =>
		      simple(Fun{name = name, args = args, body = loopExp body})
		 | Bind{var, exp, ...} =>
		      (case exp of
			  PrimApp{prim, targs, args} =>
			     (case (Prim.name prim, targs, args) of
				 (Prim.Name.Equal, [ty], [x1, x2]) =>
				    Dexp.send(equal(x1, x2, ty),
					      fn (e, _) => simple(Bind{var = var,
								       ty = Type.bool,
								       exp = e}))
			      | _ => normal())
			| _ => normal())
		 | _ => normal()
		end))
	 end

(*       (* compute poly-equal now for globals *)
 *       val globals =
 * 	 let
 * 	    val {get = value: Var.t -> unit ref * PrimExp.t, set} =
 * 	       Property.new(Var.plist, Property.initRaise("exp", Var.layout))
 * 	    fun equal(x: Var.t, y: Var.t, ty: Type.t): bool =
 * 	       if Var.equals(x, y)
 * 		  then true
 * 	       else
 * 		  let open Type
 * 		  in case dest ty of
 * 		     Char => (case (value x, value y) of
 * 				 (Const c, Const c') => Const.equals(c, c')
 * 			       | _ => Error.bug "poly equal expected constant"
 * 				 Type.Datatype
 * 
 * 				 
 * 	 in List.reverse
 * 	    (List.fold
 * 	     (globals, [], fn (bind as {var, exp, ...}, globals) =>
 * 	      (set(var, (case exp of
 * 			    Var y => value y
 * 			  | _ => (ref(), exp)))
 * 	       (case exp of
 * 		   PrimApp{prim, targs, args} =>
 * 		      (case (Prim.name prim, args) of
 * 			  (Prim.Name.Equal, [x, y]) =>
 * 			     {var = var, ty = ty,
 * 			      exp = ConApp{args = [],
 * 					   con = if equal(x, y)
 * 						    then Con.truee
 * 						 else Con.falsee}}
 * 			| _ => bind)
 * 		 | _ => bind) :: globals)))
 * 	 end
 *)
	   
      val functions =
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name,
		   args = args,
		   returns = returns,
		   body = loopExp body})

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = !newFunctions @ functions,
		   main = main}

   in Program.clear program;
      destroyVectorEqualFunc();
      program
   end

end
