(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Redundant(S: REDUNDANT_STRUCTS): REDUNDANT = 
struct

open S

structure Eqrel:>
   sig
      type t
	 
      val fromTypes: Type.t list -> t
	 
      val refine: t * (int * int -> bool) -> unit
      val unify: t * t -> unit
      val areEquivalent: t * int * int -> bool
      val classes: t -> int list list

      val fixedPoint: unit -> unit
	 
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end =
   struct
      structure R = EquivalenceRelation
      structure Set = DisjointSet

      datatype t = T of {me: R.t Set.t,
			 refinements: (int * int -> bool) list ref}

      val all: t list ref = ref []

      fun fromTypes ts =
	 let val er = T{me = Set.singleton(R.make(ts, Type.equals)),
			refinements = ref []}
	 in ListRef.push(all, er); er
	 end

      fun me(T{me, ...}) = Set.value me

      fun areEquivalent(er, i, j) = R.areEquivalent(me er, i, j)

      val classes = R.classes o me

      fun equals(T{me, ...}, T{me = me', ...}) = Set.equals(me, me')

      val layout = R.layout o me

      fun refine(T{refinements, ...}, f) = ListRef.push(refinements, f)

      (* Relies on the fact that all unifications happen before the fixed point.
       *)
      fun unify(T{me, ...}, T{me = me', ...}) = Set.union(me, me')

      fun fixedPoint() =
	 (FixedPoint.fixedPoint
	  (fn continue =>
	   List.foreach(!all, fn T{me, refinements} =>
			let val r = Set.value me
			   val r' = List.fold(!refinements, r,
					      fn (refinement, r) =>
					      R.refine(r, refinement))
			in if R.equals(r, r')
			      then ()
			   else (continue() ; Set.setValue(me, r'))
			end))
	  ; all := [])
   end


open Dec PrimExp Transfer

fun redundant(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = funcInfo: Func.t -> {
					  arg: Eqrel.t,
					  return: Eqrel.t
					 },
	   set = setFuncInfo} =
	 Property.new(Func.plist, Property.initRaise("info", Func.layout))

      val {get = jumpInfo: Jump.t -> Eqrel.t,
	   set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("info", Jump.layout))

      val {get = varInfo : Var.t -> {
				     index: int,
				     arg: Eqrel.t
				    } option,
	   set = setVarInfo} =
	 Property.new(Var.plist, Property.initConst NONE)

      fun varEquiv(x, y) =
	 Var.equals(x, y)
	 orelse (case (varInfo x, varInfo y) of
		    (SOME{index = i, arg = r},
		     SOME{index = i', arg = r'}) =>
		    Eqrel.equals(r, r') andalso Eqrel.areEquivalent(r, i, i')
		  | _ => false)

      (* compute the fixed point *)
      val _ =
	 let
	    fun makeFormalsRel(xs: (Var.t * Type.t) list): Eqrel.t =
	       let val eqrel = Eqrel.fromTypes(List.map(xs, #2))
	       in List.foreachi(xs, fn (i, (x, _)) =>
				setVarInfo(x, SOME{index = i, arg = eqrel}))
		  ; eqrel
	       end
	 in (* initialize all varInfo and funcInfo *)
	    List.foreach(functions, fn {name, args, returns, ...} =>
			 setFuncInfo(name, {
					    arg = makeFormalsRel args,
					    return = Eqrel.fromTypes returns
					   }))
	    (* Add the calls to all the funcInfos *)
	    ; (List.foreach
	       (functions, fn {name, body,...} =>
		let val {return, ...} = funcInfo name
		   val handleDec =
		      fn Fun{name, args, ...} =>
		            setJumpInfo(name, (makeFormalsRel args))
		       | _ => ()
		   val varEquiv =
		      fn vars => fn (i, j) => varEquiv(List.nth(vars, i),
						       List.nth(vars, j))
		   val handleTransfer =
		      fn Call{func, args, cont} =>
		           let val {arg, return = return'} = funcInfo func
			   in Eqrel.refine(arg, varEquiv args)
			      ; Eqrel.unify(return',
					    case cont of
					       NONE => return
					     | SOME c => jumpInfo c)
			   end
		       | Jump{dst, args, ...} =>
			    Eqrel.refine(jumpInfo dst, varEquiv args)
		       | Return xs => Eqrel.refine(return, varEquiv xs)
		       | _ => ()
		in Exp.foreach'
		   (body, {handleDec = handleDec,
			   handleTransfer = handleTransfer})
		end))
	    ; Eqrel.fixedPoint()
	 end

      val _ = 
	 Control.displays
	 ("redundant", fn display =>
	  List.foreach
	  (functions, fn {name, ...} =>
	   display
	   (let val {arg, return} = funcInfo name
	       open Layout
	    in seq[Func.layout name,
		   str "  ",
		   Eqrel.layout arg,
		   Eqrel.layout return]
	    end)))

      val {get = replacement : Var.t -> Var.t option, set = setReplacement} =
	 Property.new(Var.plist, Property.initConst NONE)

      datatype red =
	 Useful
	| Redundant of int (* the index it is the same as *)

      (* Turn an equivalence relation on 0 ... n -1 into a list of length n,
       * by choosing a representative of each class.
       *)
      fun 'a makeReds(xs: 'a list, r: Eqrel.t): red list =
	 let
	    val classes = Eqrel.classes r
	    fun getRed i =
	       let
		  val rec loop =
		     fn [] => Useful
		      | class :: classes =>
			   case class of
			      [] => Error.bug "empty class"
			    | [_] => Error.bug "trivial class"
			    | j :: js =>
				 if i = j
				    then Useful
				 else if List.exists(js, fn j => i = j)
					 then Redundant j
				      else loop classes
	       in loop classes
	       end
	 in List.mapi(xs, fn (i, _) => getRed i)
	 end
	  
      fun redundantFormals(r: Eqrel.t, xs: (Var.t * Type.t) list)
	 : red list * (Var.t * Type.t) list =
	 let val reds = makeReds(xs, r)
	    val xs =
	       List.fold2
	       (xs, reds, [], fn (x, red, ys) =>
		case red of
		   Useful => x :: ys
		 | Redundant i =>
		      (setReplacement(#1 x, SOME(#1(List.nth(xs, i))))
		       ; ys))
	 in (reds, List.reverse xs)
	 end

      fun keepUseful(reds: red list, xs: 'a list): 'a list =
	 List.reverse(List.fold2(reds, xs, [], fn (r, x, xs) =>
				 case r of
				    Useful => x :: xs
				  | _ => xs))

      val {get = funcReds : Func.t -> {args: red list,
					   returns: red list},
	   set = setFuncReds} =
	 Property.new(Func.plist, Property.initRaise("reds", Func.layout))

      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let val {arg, return} = funcInfo name
	     val return = makeReds(returns, return)
	     val (arg, args) = redundantFormals(arg, args)
	  in setFuncReds(name, {args = arg, returns = return})
	     ; {name = name, body = body,
		args = args,
		returns = keepUseful(return, returns)}
	  end)

      val {get = jumpReds : Jump.t -> red list, set = setJumpReds} =
	 Property.new(Jump.plist, Property.initRaise("red", Jump.layout))

      fun loopVar x =
	 case replacement x of
	    NONE => x
	  | SOME y => y

      fun loopVars xs = List.map(xs, loopVar)

      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let
	     val {returns = returnReds, ...} = funcReds name
	     fun loopTransfer t =
		case t of
		   Call{func, args, cont} =>
		      Call{func = func, cont = cont,
			   args = loopVars(keepUseful(#args(funcReds func),
						      args))}
		 | Jump{dst, args} =>
		      Jump{dst = dst,
			   args = loopVars(keepUseful(jumpReds dst, args))}
		 | Return xs => Return(loopVars(keepUseful(returnReds, xs)))
		 | Raise xs => Raise(loopVars xs)
		 | Case{cause, test, cases, default} =>
		      Case{cause = cause, test = loopVar test, cases = cases,
			   default = default}
		 | Halt => Halt
		 | Bug => Bug

	     fun loopExp e =
		let val {decs, transfer} = Exp.dest e
		in Exp.make{decs = List.map(decs, loopDec),
			    transfer = loopTransfer transfer}
		end
	     
	     and loopDec d =
		case d of
		   Bind{var, ty, exp} =>
		      Bind{var = var, ty = ty,
			   exp = PrimExp.replaceVar(exp, loopVar)}
		 | Fun{name, args, body} =>
		      let
			 val (reds, args) =
			    redundantFormals(jumpInfo name, args)
		      in setJumpReds(name, reds)
			 ; Fun{name = name, args = args, body = loopExp body}
		      end
		 | _ => d
	  in {name = name, args = args, returns = returns,
	      body = loopExp body}
	  end)

      val p = Program.T{datatypes = datatypes,
			globals = globals,
			functions = functions,
			main = main}
   in Program.clear p; p
   end

end
