(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* This pass must happen before polymorphic equality is implemented becuase
 * 1. it will make polymorphic equality faster because some types are simpler
 * 2. it removes uses of polymorphic equality that must return true
 *
 * This pass computes a "cardinality" of each datatype, which is an
 * abstraction of the number of values of the datatype.
 *   Zero means the datatype has no values (except for bottom).
 *   One means the datatype has one values (except for bottom).
 *   Many means the datatype has many values.
 *
 * This pass removes all datatypes whose cardinality is Zero or One
 * and removes
 *   components of tuples
 *   function args
 *   constructor args
 * which are such datatypes.
 *
 * This pass marks constructors as one of
 *   Useless: it never appears in a ConApp.
 *   Transparent: it is the only variant in its datatype
 *     and its argument type does not contain any uses of Tycon.array.
 *   Useful: otherwise
 * This pass also removes Useless and Transparent constructors.
 *
 * We must keep track of Transparent consturctors whose argument type
 * uses Tycon.array because of datatypes like the following:
 *   datatype t = T of t array
 * Such a datatype has Cardinality.Many, but we cannot eliminate
 * the datatype and replace the lhs by the rhs, i.e. we must keep the
 * circularity around .
 * Must do similar things for vectors.
 * 
 * Also, to eliminate as many Transparent constructors as possible, for
 * something like the following,
 *   datatype t = T of u array
 *   and u = U of t array
 * we (arbitrarily) expanding one of the datatypes first.
 * The result will be something like
 *   datatype u = U of u array array
 * where all uses of t are replaced by u array.
 *)

functor SimplifyTypes(S: SIMPLIFY_TYPES_STRUCTS): SIMPLIFY_TYPES = 
struct

open S

open S
open Dec PrimExp Transfer

structure Cardinality =
   struct
      datatype t = Zero | One | Many

      val isMany: t -> bool =
	 fn Many => true
	  | _ => false
	       
      fun layout c =
	 Layout.str(case c of
		       Zero => "zero"
		     | One => "one"
		     | Many => "many")

      val equals: t * t -> bool = op =

      val isZero: t -> bool =
	 fn Zero => true
	  | _ => false

      val rec add: t list -> t =
	 fn [] => Zero
	  | c :: cs =>
	       case c of
		  Many => Many
		| Zero => add cs
		| One =>
		     case add cs of
			Zero => One
		      | _ => Many

      val rec mul: t list -> t =
	 fn [] => One
	  | c :: cs =>
	       case c of
		  Zero => Zero
		| One => mul cs
		| Many =>
		     case mul cs of
			Zero => Zero
		      | _ => Many
   end

structure ConRep =
   struct
      datatype t =
	 Useless
       | Transparent
       | Useful

      val isUseful =
	 fn Useful => true
	  | _ => false

      val isUseless =
	 fn Useless => true
	  | _ => false

      val toString =
	 fn Useless => "useless"
	  | Transparent => "transparent"
	  | Useful => "useful"

      val layout = Layout.str o toString
   end

fun simplify(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conInfo: Con.t -> {rep: ConRep.t ref,
				    args: Type.t list},
	   set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("info", Con.layout))

      val conRep = ! o #rep o conInfo
      val conArgs = #args o conInfo
      fun setConRep(con, r) = #rep(conInfo con) := r
      val conIsUseful = ConRep.isUseful o conRep
      val conIsUseless = ConRep.isUseless o conRep

(*       val conIsUseful =
 * 	 Trace.trace("conIsUseful", Con.layout, Bool.layout) conIsUseful
 * 	 
 *       val setConRep =
 * 	 Trace.trace2("setConRep", Con.layout, ConRep.layout, Unit.layout)
 * 	 setConRep
 * 
 *)
      (* Initialize conInfo *)
      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach(cons, fn {con, args} =>
		       setConInfo(con, {rep = ref ConRep.Useless,
					args = args})))

      (* diagnostic *)
(*      val _ =
	 List.foreach
	 (datatypes, fn {tycon, ...} =>
	  Control.message
	  let open Layout
	  in seq[str "dependents of ",
		 Tycon.layout tycon,
		 str " = ",
		 List.layout (", ", Tycon.layout)
		 (!(#dependents(tyconInfo tycon)))]
	  end)
*)

      val {get = tyconReplacement: Tycon.t -> Type.t option,
	   set = setTyconReplacement} =
	 Property.new(Tycon.plist, Property.initConst NONE)

      val setTyconReplacement = fn (c, t) => setTyconReplacement(c, SOME t)

      val {get = tyconInfo: Tycon.t -> {
					cardinality: Cardinality.t ref,
					cons: {
					       con: Con.t,
					       args: Type.t list
					      } list,
					numCons: int ref,
					(* tycons whose cardinality depends on mine *)
					dependents: Tycon.t list ref,
					isOnWorklist: bool ref
				       },
	   set = setTyconInfo} =
	 Property.new(Tycon.plist, Property.initRaise("info", Tycon.layout))

      local
	 fun make sel = (! o sel o tyconInfo,
			 fn (t, x) => sel(tyconInfo t) := x)
      in
	 val (tyconNumCons, setTyconNumCons) = make #numCons
	 val (tyconCardinality, _) = make #cardinality
      end

      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  setTyconInfo(tycon, {
			       cardinality = ref Cardinality.Zero,
			       numCons = ref 0,
			       cons = cons,
			       dependents = ref [],
			       isOnWorklist = ref false
			      }))
	 
      (* Tentatively mark all constructors appearing in a ConApp as Useful
       * (some may later be marked as Transparent).
       *)
      val _ =
	 let
	    fun handleBind{var, ty, exp} =
	       case exp of
		  ConApp{con, ...} => setConRep(con, ConRep.Useful)
		| _ => ()
	    fun consUseful e = Exp.foreachBind(e, handleBind)
	 in 
	    (* Booleans are special because they are generated by primitives. *)
	    List.foreach([Con.truee, Con.falsee], fn c =>
			 setConRep(c, ConRep.Useful));
	    
	    List.foreach(globals, handleBind);
	    List.foreach(functions, fn {body, ...} => consUseful body)
	 end

      (* Remove useless constructors from datatypes.
       * Remove datatypes which have no cons.
       *)
      val datatypes =
	 List.fold
	 (datatypes, [], fn ({tycon, cons}, datatypes) =>
	  let val cons = List.keepAll(cons, conIsUseful o #con)
	  in case cons of
	     [] => (setTyconReplacement(tycon, Type.unit);
		    datatypes)
	   | _ => {tycon = tycon, cons = cons} :: datatypes
	  end)

      (* Build the dependents for each tycon. *)
      val _ =
	 let
	    val {get = isDatatype, set = setDatatype, destroy} =
	       Property.newDest(Tycon.plist, Property.initConst false)
	 in List.foreach
	    (datatypes, fn {tycon, ...} => setDatatype(tycon, true));

	    List.foreach
	    (datatypes, fn {tycon, cons} =>
	     let
		val {get = isDependent, set = setDependent, destroy} =
		   Property.newDest(Tycon.plist, Property.initConst false)
		fun setTypeDependents t =
		   let val (tycon', ts) = Type.tyconArgs t
		   in if isDatatype tycon'
			 then if isDependent tycon'
				 then ()
			      else (setDependent(tycon', true);
				    ListRef.push(#dependents(tyconInfo tycon'),
						 tycon))
		      else List.foreach(ts, setTypeDependents)
		   end
	     in List.foreach(cons, fn {args, ...} =>
			     List.foreach(args, setTypeDependents));
		destroy()
	     end);
	    
	    destroy()
	 end

      local open Type Cardinality
      in
	 fun typeCardinality t =
	    case dest t of
	       Ref t => (case typeCardinality t of
			    Zero => Zero
			  | _ => Many)
	     | Tuple ts => tupleCardinality ts
	     | Datatype tycon => tyconCardinality tycon
	     | _ => Many
	 and tupleCardinality ts = mul(List.map(ts, typeCardinality))
      end

(*       val typeCardinality =
 * 	 Trace.trace("typeCardinality", Type.layout, Cardinality.layout)
 * 	 typeCardinality
 *)
      fun conCardinality{con, args} = tupleCardinality args

(*       val conCardinality =
 * 	 Trace.trace("conCardinality", Con.layout o #con, Cardinality.layout)
 * 	 conCardinality
 *)
      (* Compute the tycon cardinalitues with a fixed point,
       * initially assuming every datatype tycon cardinality is Zero.
       *)
      val _ =
	 let
	    (* list of datatype tycons whose cardinality has not yet stabilized *)
	    val worklist = ref(List.map(datatypes, #tycon))
	    fun loop() =
	       case !worklist of
		  [] => ()
		| tycon :: tycons =>
		     (worklist := tycons;
		      let
			 val {cons, cardinality, dependents, isOnWorklist, ...} =
			    tyconInfo tycon
			 val c = Cardinality.add(List.map(cons, conCardinality))
		      in isOnWorklist := false;
			 if Cardinality.equals(c, !cardinality)
			    then ()
			 else (cardinality := c;
			       List.foreach
			       (!dependents, fn tycon =>
				let val {isOnWorklist, ...} = tyconInfo tycon
				in if !isOnWorklist
				      then ()
				   else (isOnWorklist := true;
					 ListRef.push(worklist, tycon))
				end))
		      end;
		      loop())
	 in Control.traceCall "fixed point" loop()
	 end

      (* diagnostic *)
(*       val _ =
 * 	 List.foreach(datatypes, fn {tycon, ...} =>
 * 		      Control.message
 * 		      let open Layout
 * 		      in seq[str "cardinality of ",
 * 			     Tycon.layout tycon,
 * 			     str " = ",
 * 			     Cardinality.layout(tyconCardinality tycon)]
 * 		      end)
 *)
	 	 
(*      val setTyconReplacement =
	 Trace.trace2("setTyconReplacement",
		      Tycon.layout, Type.layout, Layout.ignore)
	 setTyconReplacement *)

      fun transparent(tycon, con, args) =
	 (setTyconReplacement(tycon, Type.tuple args);
	  setConRep(con, ConRep.Transparent);
	  setTyconNumCons(tycon, 1))

(*       val transparent =
 * 	 Trace.trace3
 * 	 ("transparent",
 * 	  Tycon.layout, Con.layout, List.layout Type.layout, Unit.layout)
 * 	 transparent
 *)

      (* "unary" is datatatypes with one constructor whose rhs contains an
       * array (or vector) type.
       * For datatypes with one variant not containing an array type, eliminate
       * the datatype. 
       *)
      val (datatypes, unary) =
	 List.fold
	 (datatypes, ([], []), fn ({tycon, cons}, (datatypes, unary)) =>
	  let
	     (* remove all cons with zero cardinality and mark them as useless *)
	     val cons =
		List.fold
		(cons, [], fn (c as {con, ...}, cons) =>
		 case conCardinality c of
		    Cardinality.Zero => (setConRep(con, ConRep.Useless);
					 cons)
		  | _ => c :: cons)
	  in case cons of
	     [] => (setTyconNumCons(tycon, 0);
		    setTyconReplacement(tycon, Type.unit);
		    (datatypes, unary))
	   | [{con, args}] =>
		if List.exists(args, fn t =>
			       Type.containsTycon(t, Tycon.array)
			       orelse Type.containsTycon(t, Tycon.vector))
		   then (datatypes,
			 {tycon = tycon, con = con, args = args}
			 :: unary)
		else (transparent(tycon, con, args);
		      (datatypes, unary))
	   | _ => ({tycon = tycon, cons = cons} :: datatypes,
		   unary)
	  end)

      fun containsTycon(ty: Type.t, tyc: Tycon.t): bool =
	 let open Type
	    fun loop t =
	       case dest t of
		  Tuple ts => List.exists(ts, loop)
		| Array t => loop t
		| Vector t => loop t
		| Ref t => loop t
		| Datatype tyc' =>
		     (case tyconReplacement tyc' of
			 NONE => Tycon.equals(tyc, tyc')
		       | SOME t => loop t)
		| _ => false
	 in loop ty
	 end

      (* Keep the circular transparent cons, ditch the rest. *)
      val datatypes =
	 List.fold
	 (unary, datatypes, fn ({tycon, con, args}, accum) =>
	  if List.exists(args, fn arg => containsTycon(arg, tycon))
	     then {tycon = tycon, cons = [{con = con, args = args}]} :: accum
	  else (transparent(tycon, con, args);
		accum))

      val {get, set, destroy = destroySimplifyType} =
	 Property.newDest(Type.plist, Property.initConst NONE)

      fun simplifyType t =
	 case get t of
	    SOME t => t
	  | NONE => 
	       let
		  open Type
		  val t' =
		     case dest t of
			Tuple ts => Type.tuple(keepSimplifyTypes ts)
		      | Array t => array(simplifyType t)
		      | Vector t => vector(simplifyType t)
		      | Ref t => reff(simplifyType t)
		      | Datatype tycon => 
			   (case tyconReplacement tycon of
			       SOME t =>
				  let val t = simplifyType t
				  in setTyconReplacement(tycon, t);
				     t
				  end
			     | NONE => t)
		      | _ => t
	       in set(t, SOME t');
		  t'
	       end
      and simplifyTypes ts = List.map(ts, simplifyType)
      and keepSimplifyTypes ts =
	 List.keepAll(simplifyTypes ts, not o Type.isUnit)

       val simplifyType =
 	 Trace.trace("simplifyType", Type.layout, Type.layout)
 	 simplifyType

      (* Simplify constructor argument types. *)
      val datatypes =
	 List.map
	 (datatypes, fn {tycon, cons} =>
	  (setTyconNumCons(tycon, List.length cons) ;
	   {tycon = tycon,
	    cons = List.map(cons, fn {con, args} =>
			    {con = con,
			     args = keepSimplifyTypes args})}))

      val unitVar = Var.newNoname()

      val {get = varInfo: Var.t -> Type.t,
	   set = setVarInfo} =
	 Property.new(Var.plist, Property.initRaise("varInfo", Var.layout))

      fun simplifyVarType(x: Var.t, t: Type.t): Type.t =
	 (setVarInfo(x, t);
	  simplifyType t)

      val oldVarType = varInfo
      val newVarType = simplifyType o oldVarType

      fun simplifyVar(x: Var.t): Var.t =
	 if Type.isUnit(newVarType x)
	    then unitVar
	 else x

      fun simplifyVars xs = List.map(xs, simplifyVar)

      fun removeUselessVars xs =
	 List.keepAll(xs, not o Type.isUnit o newVarType)

      fun tuple xs =
	 case removeUselessVars xs of
	    [x] => Var x
	  | xs => Tuple xs

      fun simplifyFormals xts =
	 List.keepAll(List.map(xts, fn (x, t) => (x, simplifyVarType(x, t))),
		      fn (_, t) => not(Type.isUnit t))

      val typeIsUseful = not o Type.isUnit o simplifyType

      (* Compute the new offset for a select, given the list of types
       * comprising the original tuple.
       * Returns true if the component is the only component of the tuple
       * that exists.
       *)
      fun newOffset(ts: Type.t list, n: int): int * bool =
	 let
	    fun loop(ts, n, i) =
	       case ts of
		  [] => Error.bug "newOffset"
		| t :: ts =>
		     if n = 0
			then (i, (i = 0
				  andalso not(List.exists(ts, typeIsUseful))))
		     else loop(ts, n - 1,
			       if typeIsUseful t then i + 1 else i)
	 in loop(ts, n, 0)
	 end

      datatype 'a result =
	 Bugg
	| Delete
	| Keep of 'a

      fun simplifyPrimExp(e: PrimExp.t): PrimExp.t result =
	  case e of
	     Const _ => Keep e
	   | Var _ => Keep e
	   | Tuple xs => Keep(tuple xs)
	   | Select{tuple, offset} =>
		let
		  val (offset, isOne) =
		     newOffset(Type.detuple(oldVarType tuple), offset)
	       in if isOne
		     then Keep(Var tuple)
		  else Keep(Select{tuple = tuple, offset = offset})
	       end
	   | ConSelect{variant, offset, con} =>
		let val args = conArgs con
		   val offset = #1(newOffset(args, offset))
		in case conRep con of
		   ConRep.Useless => Bugg
		 | ConRep.Transparent =>
		      Keep(case args of
			      [_] => Var variant
			    | _ => Select{tuple = variant, offset = offset})
		 | ConRep.Useful =>
		      Keep(ConSelect{variant = variant, con = con,
				     offset = offset})
		end
	   | ConApp{con, args} =>
		(case conRep con of
		    ConRep.Transparent => Keep(tuple args)
		  | ConRep.Useful =>
		       Keep(ConApp{con = con, args = removeUselessVars args})
		  | ConRep.Useless => Bugg)
	   | PrimApp{prim, targs, args} =>
		Keep
		(let 
		    fun normal() =
		       PrimApp{prim = prim,
			       targs = simplifyTypes targs,
			       args = simplifyVars args}
		 in case Prim.name prim of
		    Prim.Name.Equal =>
		       (case args of
			   [a, _] => if Type.isUnit(newVarType a)
					then ConApp{con = Con.truee,
						    args = []}
				     else normal()
			 | _ => Error.bug "strange Equal PrimApp")
		  | _ => normal()
		 end)

      val {get = jumpIsNullary: Jump.t -> bool, set = setJumpNullary} =
	 Property.new(Jump.plist, Property.initConst false)

      fun simplifyTransfer(t : Transfer.t): Transfer.t =
	 case t of
	    Call{func, args, cont} =>
	       Call{func = func, cont = cont, args = removeUselessVars args}
	  | Jump{dst, args} => Jump{dst = dst, args = removeUselessVars args}
	  | Return xs => Return(removeUselessVars xs)
	  | Raise xs => Raise(removeUselessVars xs)
	  | Halt => Halt
	  | Bug => Bug
	  | Case{cause, test, cases, default} =>
	       let
		  val cases =
		     List.keepAll(cases, fn (con, _) =>
				  not(ConRep.isUseless(conRep con)))
		  val default =
		     case (cases, default) of
			(_,     NONE)    => NONE
		      | ([],    SOME j)  => SOME j
		      | (_,     SOME j)  =>
			   if List.length cases
			      = tyconNumCons(Type.tycon(oldVarType test))
			      then NONE
			   else SOME j
		  fun normal() =
		     Case{cause = cause, test = test, cases = cases,
			  default = default}
	       in case (cases, default) of
		  ([],         NONE)   => Bug
		| ([],         SOME j) => Jump{dst = j, args = []}
		| ([(con, j)], NONE)   =>
		     if ConRep.isUseful(conRep con)
			then normal() (* How can this occur? - from an array ??*)
		     else
			Jump{dst = j,
			     args = if jumpIsNullary j then [] else [test]}
		| _ => normal()
	       end

(*       val simplifyTransfer =
 * 	 Trace.trace("simplifyTransfer", Transfer.layout, Transfer.layout)
 * 	 simplifyTransfer
 *)

      fun simplifyBind{var, ty, exp} =
	 let val ty = simplifyVarType(var, ty)
	    fun normal() =
	       case simplifyPrimExp exp of
		  Bugg => Bugg
		| Delete => Delete
		| Keep e => Keep{var = var, ty = ty, exp = e}
	 in if Type.isUnit ty
	       then (case exp of
			PrimApp{prim, ...} =>
			   if Prim.maySideEffect prim
			      then normal()
			   else Delete
		      | _ => Delete)
	    else normal()
	 end

      fun simplifyExp e =
	 let val {decs, transfer} = Exp.dest e
	    fun loop(decs, accum) =
	       let fun done t = Exp.make{decs = List.reverse accum,
					 transfer = t}
	       in case decs of
		  [] => done(simplifyTransfer transfer)
		| d :: decs =>
		     case simplifyDec d of
			Bugg => done Bug
		      | Delete => loop(decs, accum)
		      | Keep d => loop(decs, d :: accum)
	       end
	 in loop(decs, [])
	 end
      and simplifyDec(d: Dec.t): Dec.t result =
	 case d of
	    Bind b => (case simplifyBind b of
			  Bugg => Bugg
			| Delete => Delete
			| Keep b => Keep(Bind b))
	  | Fun{name, args, body} =>
	       (case args of
		   [] => setJumpNullary(name, true)
		 | _ => ()
		; if List.exists(args, fn (_, t) =>
				 case Type.dest t of
				    Type.Variant(_, c) => conIsUseless c
				  | _ => false)
		     then Delete
		  else Keep(Fun{name = name,
				args = simplifyFormals args,
				body = simplifyExp body}))
	  | _ => Keep d

      val globals =
	 {var = unitVar,
	  ty = Type.unit,
	  exp = Tuple[]}
	 :: List.reverse(List.fold(globals, [], fn (b, bs) =>
				   case simplifyBind b of
				      Bugg => Error.bug "global bind can't fail"
				    | Delete => bs
				    | Keep b => b :: bs))

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  {name = name,
	   args = simplifyFormals args,
	   body = shrinkExp(simplifyExp body),
	   returns = keepSimplifyTypes returns})

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
	 
   in destroySimplifyType();
      Program.clear program;

      program
   end

end
