(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Simplify(S: sig
		       include SIMPLIFY_STRUCTS
		       sharing Func = Jump (* for CpsInline *)
		    end): SIMPLIFY = 
struct

open S

fun trace(name, pass) =
   Control.traceCall name
   (Trace.trace("Simplify." ^ name, Program.layout, Program.layout) pass)

structure ConstantPropagation = ConstantPropagation(S)
val constantPropagation =
   trace("constant propagation", ConstantPropagation.simplify)

structure Contify = Contify(S)
val contify = trace("contify", Contify.contify)

structure CpsInline = CpsInline(S)
val cpsInline = trace("cpsInline", CpsInline.computeInline)

(* structure DeepFlatten = DeepFlatten(S)
 * val deepFlatten = trace("deep flatten", DeepFlatten.flatten)
 *)

structure Flatten = Flatten(S)
val flatten = trace("flatten", Flatten.flatten)
val flatten =
   fn p =>
   let
      fun loop(i, p) =
	 if i = 0
	    then p
	 else loop(i - 1, flatten p)
   in loop(!Control.flattenRounds, p)
   end

structure Inline = Inline(S)
val inline = trace("inline", Inline.inline)

structure IntroduceLoops = IntroduceLoops(S)
val introduceLoops = trace("introduce loops", IntroduceLoops.introduceLoops)

(*structure LoopCount = LoopCount(S) *)

structure LoopInvariant = LoopInvariant(S)
val loopInvariant = trace("loop invariant", LoopInvariant.loopInvariant)
   
structure Useless = Useless(S)
val useless = trace("useless", Useless.useless)

structure PolyEqual = PolyEqual(S)
val polyEqual = trace("poly equal", PolyEqual.polyEqual)

structure RaiseToJump = RaiseToJump(S)
val raiseToJump = trace("raise to jump", RaiseToJump.raiseToJump)

structure Redundant = Redundant(S)
val redundant = trace("redundant", Redundant.redundant)

structure RemoveUnusedConstructors = RemoveUnusedConstructors(S)
val removeUnusedConstructors =
   trace("remove unused constructors", RemoveUnusedConstructors.remove)

structure RemoveUnusedFunctions = RemoveUnusedFunctions(S)
val removeUnusedFunctions =
   trace("remove unused functions", RemoveUnusedFunctions.remove)

structure RemoveUnusedGlobals = RemoveUnusedGlobals(S)
val removeUnusedGlobals =
   trace("remove unused globals", RemoveUnusedGlobals.remove)

structure SaveWorld = SaveWorld(S)
val saveWorld = trace("save world", SaveWorld.saveWorld)

structure SimplifyTypes = SimplifyTypes(S)
val simplifyTypes = trace("simplify types", SimplifyTypes.simplify)

fun stats p = (Control.message(fn () => Program.layoutStats p); p)

fun leafInline p =
   Ref.fluidLet
   (Control.inline, Control.Leaf{size = SOME 20 (* arbitrary *)}, fn () =>
    inline p)
val leafInline = trace("leaf inline", leafInline)

val inline =
   fn p =>
   case !Control.inline of
      Control.Recursive _ => cpsInline p
    | _ =>
	 let
	    fun loop(p, rounds) =
	       if rounds = 0
		  then p
	       else loop(inline p, rounds - 1)
	 in loop(p, !Control.inlineRounds)
	 end
			   
fun $ (f, g) x = g(f x)
infixr $

val typeCheck =
   stats
   $ (fn p => (if Control.shouldTypeCheck()
		  then Control.traceCall "typeCheck" typeCheck p
	       else ();
		  p))

fun save suffix p =
   (Ref.fluidLet
    (Control.aux, true, 
     fn () => Control.displays(suffix, fn layout => Program.layouts(p, layout)))
    ; p)

fun traces fs p = (Trace.Immediate.on fs; p)
   
(*
 * useless should run after constant propagation because const prop makes
 *  slots of tuples that are constant useless
 *
 * contify should be run before constant propagation because of the once pass
 *  that only looks at main -- hence want as much in main as possible.
 *
 * poly equal should run
 *   - after types are simplified so that many equals are turned into eqs
 *   - before inlining so that equality functions can be inlined
 *)
val simplify =
   stats
   $ removeUnusedFunctions $ stats
   $ leafInline $ stats
   $ constantPropagation $ typeCheck
   $ useless $ typeCheck
   $ simplifyTypes $ typeCheck
   $ polyEqual $ typeCheck
   $ contify $ typeCheck
   $ inline $ typeCheck
   $ raiseToJump $ typeCheck
   $ contify $ typeCheck
   $ introduceLoops $ loopInvariant $ typeCheck
   $ flatten $ typeCheck
   $ redundant $ typeCheck
   $ removeUnusedFunctions
   $ removeUnusedGlobals
   $ removeUnusedConstructors
   $ shrink $ typeCheck
   $ saveWorld (* must be last so that optimizations don't inline saveWorld *)
(*   $ LoopCount.instrument *)
   
val simplify = trace("simplify", simplify)

val typeCheck = S.typeCheck
   
end
