(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeCheck(S: TYPE_CHECK_STRUCTS): TYPE_CHECK = 
struct

open S
open Dec Exp PrimExp Transfer

fun equalss(ts, ts') = List.equals(ts, ts', Type.equals)

fun checkScopes(program as Program.T{datatypes, globals, functions, main}): unit =
   let
      datatype status =
	 Undefined
       | InScope
       | Defined

      fun make(layout, plist) =
	 let val {get, set} = Property.new(plist, Property.initConst Undefined)
	    fun bind x =
	       case get x of
		  Undefined => set(x, InScope)
		| _ => Error.bug("duplicate definition of "
				 ^ (Layout.toString(layout x)))
	    fun reference x =
	       case get x of
		  InScope => ()
		| _ => Error.bug(concat
				 ["reference to ",
				  Layout.toString(layout x),
				  " not in scope"])

	    fun unbind x = set(x, Defined)
	 in (bind, reference, unbind)
	 end

      val (bindTycon, getTycon, unbindTycon) = make(Tycon.layout, Tycon.plist)
      val (bindCon, getCon, unbindCon) = make(Con.layout, Con.plist)
      val (bindVar, getVar, unbindVar) = make(Var.layout, Var.plist)
      val (bindFunc, getFunc, unbindFunc) = make(Func.layout, Func.plist)
      val (bindJump, getJump, unbindJump) = make(Jump.layout, Jump.plist)

      fun getVars xs = List.foreach(xs, getVar)

      val loopPrimExp =
	 fn Const _ => ()
	  | Var x => getVar x
	  | Tuple xs => getVars xs
	  | Select{tuple, ...} => getVar tuple
	  | ConSelect{variant, ...} => getVar variant
	  | ConApp{args, ...} => getVars args
	  | PrimApp{args, ...} => getVars args

      val loopTransfer =
	 fn Call{func, args, cont} => (getFunc func;
				       getVars args;
				       case cont of
					  NONE => ()
					| SOME c => getJump c)
	  | Jump{dst, args} => (getJump dst; getVars args)
	  | Return xs => getVars xs
	  | Case{test, cases, default, ...} =>
	       (getVar test;
		List.foreach(cases, getJump o #2);
		Option.map'(default, getJump))
	  | Raise xs => getVars xs
	  | Halt => ()
	  | Bug => ()

      fun loopBind{var, ty, exp} = (loopPrimExp exp; bindVar var)
	 
      fun loopFunc(args, body) =
	 (List.foreach(args, bindVar o #1);
	  loopExp body;
	  List.foreach(args, unbindVar o #1))

      and loopExp e =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(decs, 
			 fn Bind b => loopBind b
			  | Fun{name, args, body} =>
			       (bindJump name; loopFunc(args, body))
			  | HandlerPush h => getJump h
			  | HandlerPop => ())
	    ; loopTransfer transfer
	    ; List.foreach(decs, 
			   fn Bind{var, ...} => unbindVar var
			    | Fun{name, ...} => unbindJump name
			    | HandlerPush _ => ()
			    | HandlerPop => ())
	 end

   in List.foreach(datatypes, fn {tycon, cons} =>
		   (bindTycon tycon
		    ; List.foreach(cons, bindCon o #con)))
      ; List.foreach(globals, loopBind)
      ; List.foreach(functions, bindFunc o #name)
      ; List.foreach(functions, fn {args, body, ...} => loopFunc(args, body))
      ; getFunc main
      ; Program.clear program
   end

val checkScopes = Control.traceCall "checkScopes" checkScopes
		     
fun typeCheck(program as Program.T{datatypes, functions, ...}): unit =
   let
      val _ = checkScopes program
      val _ = let val jumpHandlers = inferHandlers program
	      in List.foreach
		 (functions, fn {body, ...} =>
		  Exp.foreach'
		  (body, {handleTransfer = fn _ => (),
			  handleDec =
			  fn Fun{name, ...} => (jumpHandlers name; ())
			   | _ => ()}));
	       Program.clear program
	      end
      val out = Out.error
      val print = Out.outputc out

      exception TypeError

      fun error(msg, lay) =
	 (print("Type error: " ^ msg ^ "\n")
	  ; Layout.output(lay, out)
	  ; print "\n"
	  ; raise TypeError)

      fun coerce{from: Type.t, to: Type.t}: unit =
	 if Type.equals(from, to)
	    then ()
	 else error("Type.equals",
		    Layout.record[("from", Type.layout from),
				  ("to", Type.layout to)])

      val error = fn s => error(s, Layout.empty)
	 
(*       val coerce =
 * 	 Trace.trace("TypeCheck.coerce",
 * 		     fn {from, to} => let open Layout
 * 				      in record[("from", Type.layout from),
 * 						("to", Type.layout to)]
 * 				      end,
 * 				   Unit.layout) coerce
 *)

      fun select{tuple: Type.t, offset:int, resultType}: Type.t =
	 case Type.detupleOpt tuple of
	    NONE => error "select of non tuple"
	  | SOME ts => List.nth(ts, offset)

      fun primApp{prim, targs, args, resultVar, resultType} =
	 case Prim.checkApp{prim = prim,
			    targs = targs,
			    args = args,
			    con = Type.con,
			    equals = Type.equals,
			    dearrowOpt = Type.dearrowOpt,
			    detupleOpt = Type.detupleOpt,
			    isUnit = Type.isUnit
			    } of
	    NONE => error "bad primapp"
	  | SOME t => t

(*       val primApp =
 * 	 Trace.trace("checkPrimApp",
 * 		     fn {prim, args, ...} =>
 * 		     let open Layout
 * 		     in record[("prim", Prim.layout prim),
 * 			       ("args", List.layout Type.layout args)]
 * 		     end,
 * 		     Type.layout) primApp
 *)

      fun filter(t: Type.t, c: Con.t): Type.t =
	 case Type.dest t of
	    Type.Datatype tycon => Type.variant(tycon, c)
	  | _ => Error.bug "can't make variant of non datatype"
	     
      val {value = varType, ...} =
	 analyze{
		 program = program,
		 useFromTypeOnBinds = true,
		 fromType = fn x => x,
		 copy = fn x => x,
		 coerce = coerce,
		 const = Type.ofConst,
		 tuple = Type.tuple,
		 select = select,
		 constructors = NONE,
		 primApp = primApp,
		 filter = filter
		}

   in Program.clear program
   end

(* val typeCheck = Trace.trace("Cps.typeCheck", Program.layout,
 * 			    Layout.ignore) typeCheck
 *)
end
