(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Useless(S: USELESS_STRUCTS): USELESS = 
struct

open S

structure ExtractTargs = ExtractTargs(structure Prim = Prim
				      structure Type = Type)

(* useless thing elimination
 *  remove components of tuples that are constants (use unification)
 *  remove function arguments that are constants
 *  build some kind of dependence graph where 
 *    - a value of ground type is useful if it is an arg to a primitive
 *    - a tuple is useful if it contains a useful component
 *    - a conapp is useful if it contains a useful component
 *                            or is used in a case
 *
 * If a useful tuple is coerced to another useful tuple,
 *   then all of their components must agree (exactly).
 * It is trivial to convert a useful value to a useless one.
 *
 * It is also trivial to convert a useful tuple to one of its
 *  useful components -- but this seems hard
 *)

(* If something is useful, then it must exist.
 *)

(* Suppose that you have a ref/array/vector that is useful, but the
 * components aren't -- then the components are converted to type unit, and
 * any primapp args must be as well.
 *)

(* Weirdness with raise/handle.
 * There must be a uniform "calling convention" for raise and handle.
 * Hence, just because some of a handlers args are useless, that doesn't mean
 * that it can drop them, since they may be useful to another handler, and
 * hence every raise will pass them along.  The problem is that it is not
 * possible to tell solely from looking at a function declaration whether it is
 * a handler or not, and in fact, there is nothing preventing a jump being used
 * in both ways.  So, maybe the right thing is for the handler wrapper to
 * do
 * Another solution would be to unify all handler args!
 *)

   
structure Value =
   struct
      structure Set = DisjointSet

      local
	 structure Top =
	    struct
	       type t = unit
	       fun new(): t = ()
	    end
	 structure Dest =
	    struct
	       datatype t =
		  NotExists
		| Exists of Top.t
	       val bottom = NotExists
	       val top = Exists
	    end
	 structure L = TwoPointLattice(val bottom = "not exists"
				       val top = "exists"
				       structure Top = Top
				       structure Dest = Dest)
      in
	 structure Exists =
	    struct
	       open L
	       val mustExist = makeTop
	       val doesExist = isTop
	    end
	 datatype exists = datatype Dest.t
      end

      local
	 structure Top =
	    struct
	       type t = unit
	       fun new(): t = ()
	    end
	 structure Dest =
	    struct
	       datatype t =
		  Useless
		| Useful of Top.t
	       val bottom = Useless
	       val top = Useful
	    end
	 structure L = TwoPointLattice(val bottom = "useless"
				       val top = "useful"
				       structure Top = Top
				       structure Dest = Dest)
      in
	 structure Useful =
	    struct
	       open L
	       val makeUseful = makeTop
	       val isUseful = isTop
	    end
	 datatype useful = datatype Dest.t
      end

      datatype t =
	 T of {
	       ty: Type.t,
	       new: (Type.t * bool) option ref,
	       value: value
	      } Set.t
      and value =
	 Ground of Useful.t
	| Vector of {length: t,
		     elt: slot}
	| Array of {useful: Useful.t,
		    length: t,
		    elt: slot}
	| Ref of {useful: Useful.t,
		  arg: slot}
	| Tuple of slot list
      withtype slot = t * Exists.t

      local
	 fun make sel(T s) = sel(Set.value s)
      in
	 val value = make #value
	 val ty = make #ty
      end
   
      local
	 open Layout
      in
	 fun layout(T s) =
	    let val {value, ...} = Set.value s
	    in case value of
	       Ground g => seq[str "ground ", Useful.layout g]
	     | Tuple vs => tuple(List.map(vs, layoutSlot))
	     | Ref{arg,...} => paren(seq[str "ref ", layoutSlot arg])
	     | Vector{elt, length} =>
		  seq[str "vector", tuple[layout length, layoutSlot elt]]
	     | Array{elt, length, ...} =>
		  seq[str "array", tuple[layout length, layoutSlot elt]]
	    end
	 and layoutSlot(v, e) =
	    tuple[Exists.layout e, layout v]
      end

      fun unify(T s, T s') =
	 if Set.equals(s, s')
	    then ()
	 else let val {value = v, ...} = Set.value s
		  val {value = v', ...} = Set.value s'
	      in Set.union(s, s');
		 case (v, v') of
		    (Ground g, Ground g') => Useful.==(g, g')
		  | (Tuple vs, Tuple vs') => List.foreach2(vs, vs', unifySlot)
		  | (Ref{useful = u, arg = a}, Ref{useful = u', arg = a'}) =>
		       (Useful.==(u, u'); unifySlot(a, a'))
		  | (Array{length = n, elt = e, ...},
		     Array{length = n', elt = e', ...}) =>
		       (unify(n, n'); unifySlot(e, e'))
		  | (Vector{length = n, elt = e},
		     Vector{length = n', elt = e'}) =>
		    (unify(n, n'); unifySlot(e, e'))
		  | _ => Error.bug "strange unify"
	      end
      and unifySlot((v, e), (v', e')) = (unify(v, v'); Exists.==(e, e'))
	   
      fun coerce{from = from as T sfrom, to = to as T sto}: unit =
	 if Set.equals(sfrom, sto)
	    then ()
	 else
	    let
	       fun coerceSlot((v, e), (v', e')) =
		  (coerce{from = v, to = v'};
		   Exists.==(e, e'))
	    in case (value from, value to) of
	       (Ground to, Ground from) => Useful.<=(from, to)
	     | (Tuple vs, Tuple vs') =>
		  List.foreach2(vs, vs', coerceSlot)
	     | (Ref _, Ref _) => unify(from, to)
	     | (Array _, Array _) => unify(from, to)
	     | (Vector{length = n, elt = e}, Vector{length = n', elt = e'}) =>
		   (coerce{from = n, to = n'};
		    coerceSlot(e, e'))
	     | _ => Error.bug "strange coerce"
	    end

(*       val coerce =
 * 	 Trace.trace("Useless.coerce",
 * 		     fn {from, to} => let open Layout
 * 				      in record[("from", layout from),
 * 						("to", layout to)]
 * 				      end,
 * 		     Unit.layout) coerce
 *)

      fun coerces{from, to} =
	 List.foreach2(from, to, fn (from, to) => coerce{from = from, to = to})

      (* Coerce every ground value in v to u. *)
      fun deepCoerce(v: t, u: Useful.t): unit =
	 let
	    fun doit u' = Useful.<=(u', u)
	    fun loop(v: t): unit =
	       case value v of
		  Ground u' => doit u'
		| Vector{length, elt} => (loop length; slot elt)
		| Array{length, elt, useful} =>
		     (doit useful; loop length; slot elt)
		| Ref{useful, arg} => (doit useful; slot arg)
		| Tuple vs => List.foreach(vs, slot)
	    and slot(v, _) = loop v
	 in loop v
	 end

(*       val deepCoerce =
 * 	 Trace.trace2("Useless.deepCoerce", layout, Useful.layout, Unit.layout)
 * 	 deepCoerce
 *)

      fun deground(v: t): Useful.t =
	 case value v of
	    Ground g => g
	  | _ => Error.bug "deground"

      fun fromType(t: Type.t): t =
	 let
	    fun loop(t: Type.t, es: Exists.t list): t =
	       let
		  fun useful() =
		     let val u = Useful.new()
		     in Useful.addHandler
			(u, fn () => List.foreach(es, Exists.mustExist));
			u
		     end
		  fun slot t =
		     let val e = Exists.new()
		     in (loop(t, e :: es), e)
		     end
		  val loop = fn t => loop(t, es)
		  val value =
		     case Type.dest t of
			Type.Ref t => Ref{useful = useful(),
					  arg = slot t}
		      | Type.Array t =>
			   let val elt as (_, e) = slot t
			      val length = loop Type.int
			   in Exists.addHandler
			      (e, fn () => Useful.makeUseful(deground length));
			      Array{useful = useful(),
				    length = length,
				    elt = elt}
			   end
		      | Type.Vector t => Vector{length = loop Type.int,
						elt = slot t}
		      | Type.Tuple ts => Tuple(List.map(ts, slot))
		      | _ => Ground(useful())
	       in T(Set.singleton{ty = t,
				  new = ref NONE,
				  value = value})
	       end
	 in loop(t, [])
	 end

      val const = fromType o Type.ofConst

      fun detupleSlots(v: t): slot list =
	 case value v of
	    Tuple ss => ss
	  | _ => Error.bug "detuple"

      fun detuple v = List.map(detupleSlots v, #1)

      fun tuple(vs: t list): t =
	 let val t = Type.tuple(List.map(vs, ty))
	    val v = fromType t
	 in List.foreach2(vs, detuple v, fn (v, v') =>
			  coerce{from = v, to = v'});
	    v
	 end

      val unit = tuple[]

      fun select{tuple, offset, resultType} =
	 let val v = fromType resultType
	 in coerce{from = List.nth(detuple tuple, offset), to = v};
	    v
	 end

      local
	 fun make (err, sel) v =
	    case value v of
	       Vector fs => sel fs
	     | _ => Error.bug err
      in val devector = make("devector", #1 o #elt)
	 val vectorLength = make("vectorLength", #length)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Array fs => sel fs
	     | _ => Error.bug err
      in val dearray: t -> t = make("dearray", #1 o #elt)
	 val arrayLength = make("arrayLength", #length)
	 val arrayUseful = make("arrayUseful", #useful)
      end

      fun deref(r: t): t =
	 case value r of
	    Ref{arg, ...} => #1 arg
	  | _ => Error.bug "deref"

      fun filter(v, _) =
	 case value v of
	    Ground g => (Useful.makeUseful g; v)
	  | _ => Error.bug "filter of non ground"

      fun newType(v: t): Type.t = #1(getNew v)
      and isUseful(v: t): bool = #2(getNew v)
      and getNew(T s): Type.t * bool =
	 let val {value, ty, new, ...} = Set.value s
	 in case !new of
	    SOME z => z
	  | NONE =>
	       let 
		  fun slot(arg: t, e: Exists.t) =
		     let val (t, b) = getNew arg
		     in (if Exists.doesExist e then t else Type.unit, b)
		     end
		  fun wrap((t, b), f) = (f t, b)
		  fun or((t, b), b') = (t, b orelse b')
		  fun maybe(u: Useful.t, s: slot, make: Type.t -> Type.t) =
		     wrap(or(slot s, Useful.isUseful u), make)
		  val z =
		     case value of
			Ground u => (ty, Useful.isUseful u)
		      | Ref{useful, arg, ...} => maybe(useful, arg, Type.reff)
		      | Array{useful, elt, length, ...} =>
			   or(wrap(slot elt, Type.array),
			      Useful.isUseful useful orelse isUseful length)
		      | Vector{elt, length, ...} =>
			   or(wrap(slot elt, Type.vector), isUseful length)
		      | Tuple vs =>
			   wrap
			   (List.foldr
			    (vs, ([], false), fn ((v, e), (ts, useful)) =>
			     let val (t, u) = getNew v
			     in (if Exists.doesExist e then t :: ts else ts,
				 u orelse useful)
			     end),
			    Type.tuple)
	       in new := SOME z; z
	       end
	 end

(*       val getNew =
 * 	 Trace.trace("getNew", layout, Layout.tuple2(Type.layout, Bool.layout))
 * 	 getNew
 * 
 *       val newType = Trace.trace("newType", layout, Type.layout) newType
 *)
      fun newTypes vs =
	 List.foldr(vs, [], fn (v, ts) =>
		    let val (t, b) = getNew v
		    in if b then t :: ts else ts
		    end)
   end

structure Exists = Value.Exists

fun useless(program as Program.T{datatypes, functions, main, globals}) =
   let
      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("arg", Con.layout))

      val {get = tyconInfo: Tycon.t -> {useful: bool ref,
					cons: Con.t list},
	   set = setTyconInfo, destroy} =
	 Property.newDest(Tycon.plist, Property.initRaise("cons", Tycon.layout))
	   
      local open Value
      in
	 val _ =
	    List.foreach
	    (datatypes, fn {tycon, cons} =>
	     (setTyconInfo(tycon, {useful = ref false,
				   cons = List.map(cons, #con)})
	      ; let fun value() = fromType(Type.con(tycon, []))
		in List.foreach
		   (cons, fn {con, args} =>
		    setConInfo(con, {value = value,
				     argTypes = args,
				     args = List.map(args, fromType)}))
		end))

	 val conArgs = #args o conInfo

	 fun conApp{con, args} =
	    let val {args = args', value, ...} = conInfo con
	    in Value.coerces{from = args, to = args'};
	       value()
	    end

	 fun conSelect{variant, con, offset} =
	    let val {argTypes, args, ...} = conInfo con
	       val result = fromType(List.nth(argTypes, offset))
	    in coerce{from = List.nth(args, offset), to = result};
	       (* If any component of the result is useful,
		* then the variant is useful.
		*)
	       case value variant of
		  Ground g => deepCoerce(result, g)
		| _ => Error.bug "strange variant";
	       result
	    end
   
	 (* This is used for primitive args, since we have no idea what components
	  * of its args that a primitive will look at.
	  *)
	 fun deepMakeUseful v =
	    let val slot = deepMakeUseful o #1
	    in case value v of
	       Ground u =>
		  (Useful.makeUseful u;
		   (* Make all constructor args of this tycon useful *)
		   case Type.dest(ty v) of
		      Type.Datatype tycon =>
			 let val {useful, cons} = tyconInfo tycon
			 in if !useful
			       then ()
			    else (useful := true
				  ; List.foreach(cons, fn con =>
						 List.foreach(#args(conInfo con),
							      deepMakeUseful)))
			 end
		    | _ => ())
	     | Tuple vs => List.foreach(vs, slot)
	     | Ref{useful, arg} => (Useful.makeUseful useful; slot arg)
	     | Vector{length, elt} => (deepMakeUseful length; slot elt)
	     | Array{useful, length, elt} => (Useful.makeUseful useful;
					      deepMakeUseful length;
					      slot elt)
	    end

	 type value = t
	    
	 fun primApp{prim, targs, args, resultVar, resultType} =
	    let
	       val result = fromType resultType
	       fun return v = coerce{from = v, to = result}
	       infix dependsOn
	       fun v1 dependsOn v2 = deepCoerce(v2, deground v1)
	       open Prim.Name
	    in (case (Prim.name prim, args) of
		   (Array, [i]) => coerce{from = i, to = arrayLength result}
		 | (ArraySub, [a, i]) => (i dependsOn result;
					  return(dearray a))
		 | (ArrayUpdate, [a, i, v]) =>
		      (i dependsOn dearray a;
		       coerce{from = v, to = dearray a})
		 | (ArrayLength, [a]) => return(arrayLength a)
		 | (VectorLength, [v]) => return(vectorLength v)
		 | (VectorSub, [v, i]) => (i dependsOn result;
					   return(devector v))
		 | (VectorFromArray, [a]) =>
		      (case (value a, value result) of
			  (Value.Array{length = l, elt = e, ...},
			   Vector{length = l', elt = e', ...}) =>
			     (unify(l, l'); unifySlot(e, e'))
			 | _ => Error.bug "strange VectorFromArray")
		 | (Ref, [v]) => coerce{from = v, to = deref result}
		 | (Deref, [r]) => return(deref r)
		 | (Assign, [r, v]) => coerce{from = v, to = deref r}
		 (* could do slightly better for Eq and Equal
		  * But be careful when you do so, since it may make a t ref into
		  * a unit ref -- and then the translation will need to replace some
		  * vars with unit.
		  *)
		 | _ => (List.foreach(args, deepMakeUseful)
			 ; if (case Type.dest resultType of
				  Type.Array _ => true
				| Type.Vector _ => true
				| Type.Ref _ => true
				| _ => false)
			      then deepMakeUseful result
			   else ()))
	       ; result
	    end
      end

      val {value, func, jump, exnVals, ...} =
	 Control.traceCall "analyze"
	 analyze{
		 program = program,
		 useFromTypeOnBinds = true,
		 fromType = Value.fromType,
		 copy = Value.fromType o Value.ty,
		 coerce = Value.coerce,
		 const = Value.const,
		 tuple = Value.tuple,
		 select = Value.select,
		 constructors = SOME{conApp = conApp, conSelect = conSelect},
		 primApp = primApp,
		 filter = Value.filter
		}

      open Dec PrimExp Transfer

      (* Unify all handler args so that raise/handle has a consistent calling
       * convention.
       *)
      val exnVals = fn () => (case exnVals of
				 NONE => Error.bug "no exnVals"
			       | SOME vs => vs)
      val _ =
	 List.foreach
	 (functions, fn {body, ...} =>
	  Exp.foreach'
	  (body,
	   {handleTransfer = fn _ => (),
	    handleDec = fn HandlerPush j => List.foreach2(jump j, exnVals(),
							  Value.unify)
                         | _ => ()}))

      (* diagnostic *)
      val _ =
	 Control.displays
	 ("useless", fn display =>
	  let open Layout
	  in display(str "Input program:\n");
	     Program.layouts(program, display);
	     display(str "\n\nUseless:");
	     List.foreach
	     (datatypes, fn {tycon, cons} =>
	      display
	      (align[Tycon.layout tycon,
		     indent(align
			    (List.map(cons, fn {con, ...} =>
				      seq[Con.layout con, str " ",
					  tuple(List.map(conArgs con,
							 Value.layout))])),
			    2)]));
	     Program.foreachVar
	     (program, fn (x, _) => display(seq[Var.layout x,
						str " ",
						Value.layout(value x)]))
	  end)

      val varExists = Value.isUseful o value

      val unitVar = Var.newString "unit"

      val bogusGlobals: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
	 
      fun bogus(ty: Type.t): Var.t =
	 case List.peek(!bogusGlobals, fn {ty = ty', ...} =>
			Type.equals(ty, ty')) of
	    NONE => let val var = Var.newString "bogus"
		    in ListRef.push(bogusGlobals,
				    {var = var, ty = ty,
				     exp = PrimApp{prim = Prim.bogus,
						   targs = [ty],
						   args = []}});
		       var
		    end
	  | SOME{var, ...} => var

      fun keepUseful(xs, vs) =
	 List.foldr2(xs, vs, [], fn (x, v, xs) =>
		     let val (t, b) = Value.getNew v
		     in if b then (if varExists x then x else bogus t) :: xs
			else xs
		     end)

(*       val keepUseful =
 * 	 Trace.trace2("keepUseful",
 * 		      List.layout Var.layout,
 * 		      List.layout Value.layout,
 * 		      List.layout Var.layout)
 * 	 keepUseful
 *)
      fun keepUsefulArgs xts =
	 List.foldr(xts, [], fn ((x, _), xts) =>
		    let val (t, b) = Value.getNew(value x)
		    in if b then (x, t) :: xts
		       else xts
		    end)

(*       val keepUsefulArgs =
 * 	 Trace.trace2("keepUsefulArgs",
 * 		      List.layout(",", Layout.tuple2(Var.layout, Type.layout)),
 * 		      List.layout(",", Value.layout),
 * 		      List.layout(",", Layout.tuple2(Var.layout, Type.layout)))
 * 	 keepUsefulArgs
 *)

      (* Returns true if the component is the only component of the tuple
       * that exists.
       *)
      fun newOffset(bs: bool list, n: int): int * bool =
	 let
	    fun loop(bs, n, i) =
	       case bs of
		  [] => Error.bug "newOffset"
		| b :: bs =>
		     if n = 0
			then (i, (i = 0 andalso not(List.exists(bs, fn b => b))))
		     else loop(bs, n - 1, if b then i + 1 else i)
	 in loop(bs, n, 0)
	 end
	 
      fun loopPrimExp(e: PrimExp.t, resultType: Type.t, resultValue: Value.t) =
	 case e of
	    Const _ => e
	  | Var _ => e
	  | Tuple xs =>
	       let val slots = Value.detupleSlots resultValue
		  val xs =
		     List.fold2
		     (xs, slots, [], fn (x, (v, e), xs) =>
		      if Exists.doesExist e
			 then ((if varExists x then x else bogus(Value.newType v))
			       :: xs)
		      else xs)
	       in case xs of
		   [x] => Var x
		 | xs => Tuple(List.reverse xs)
	       end
	  | Select{tuple, offset} =>
	       let
		  val (offset, isOne) =
		     newOffset(List.map(Value.detupleSlots(value tuple),
					Exists.doesExist o #2),
			       offset)
	       in if isOne
		     then Var tuple
		  else Select{tuple = tuple,
			      offset = offset}
	       end
	  | ConSelect{variant, con, offset} =>
	       ConSelect
	       {variant = variant,
		con = con,
		offset = #1(newOffset(List.map(conArgs con, Value.isUseful),
				      offset))}
	  | ConApp{con, args} =>
	       ConApp{con = con,
		      args = keepUseful(args, conArgs con)}
	  | PrimApp{prim, args, ...} => 
	       let
		  val (args, argTypes) =
		     List.unzip
		     (List.map(args, fn x =>
			       let val (t, b) = Value.getNew(value x)
			       in if b then (x, t)
				  else (unitVar, Type.unit)
			       end))
	       in PrimApp
		  {prim = prim,
		   args = args,
		   targs = ExtractTargs.extract{prim = prim,
						args = argTypes,
						result = resultType}}
	       end

(*       val loopPrimExp =
 * 	 Trace.trace3("Useless.loopPrimExp",
 * 		      PrimExp.layout, Layout.ignore, Layout.ignore,
 * 		      PrimExp.layout) loopPrimExp
 *)
	       
      fun loopBind{var, exp, ty} =
	 let val v = value var
	    fun yes ty = SOME{var = var, ty = ty, exp = loopPrimExp(exp, ty, v)}
	    val (t, b) = Value.getNew v
	 in if b then yes t
	    else
	       case exp of
		  PrimApp{prim, args, ...} =>
		     if Prim.maySideEffect prim
			andalso let val check = Value.isUseful
				    open Prim.Name
				in case (Prim.name prim, args) of
				   (ArrayUpdate, [a, _, _]) =>
				      check(Value.dearray(value a))
				 | (Assign, [r, _]) =>
				      check(Value.deref(value r))
				 | _ => true
			       end
			then yes t
		     else NONE
	        | _ => NONE
	 end

      fun agree(vs, vs') =
	 List.forall2(vs, vs', fn (v, v') =>
		      not(Value.isUseful v) orelse Value.isUseful v')

      fun loopExp(e: Exp.t, returns): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val decs =
	       List.reverse
	       (List.fold
		(decs, [], fn (d, ds) =>
		 case d of
		    Fun{name, args, body} =>
		       Fun{name = name,
			   args = keepUsefulArgs args,
			   body = loopExp(body, returns)} :: ds
		  | Bind b => (case loopBind b of
				  NONE => ds
				| SOME b => Bind b :: ds)
		  | _ => d :: ds))
	    val (decs, transfer) =
	       case transfer of
		  (* May need to insert a wrapper around the cont
		   * to translate between the expected returns of
		   * the function and the useful args of the func.
		   * Yucko! - what about tail calls?
		   * It's entirely possible that my returns are less useful
		   * than my callees, in which case I need I need to drop
		   * some of them.
		   *)
		  Call{func = f, args, cont} =>
		     let val {args = fargs, returns = freturns} = func f
			fun wrapper(vs: Value.t list,
				    cont: Jump.t option,
				    makeTrans: Var.t list -> Transfer.t) =
			   if agree(freturns, vs)
			      then (decs, cont)
			   else
			      let
				 val j' = Jump.newNoname()
				 val (formals, actuals) =
				    List.foldr2
				    (freturns, vs, ([], []),
				     fn (v, v', (formals, actuals)) =>
				     if Value.isUseful v
					then let val x = Var.newNoname()
					     in ((x, Value.newType v)
						 :: formals,
						 if Value.isUseful v'
						    then x :: actuals
						 else actuals)
					     end
				     else (formals, actuals))
			      in (decs
				  @ [Fun
				     {name = j',
				      args = formals,
				      body =
				      Exp.make{decs = [],
					       transfer = makeTrans actuals}}],
				  SOME j')
			      end

			val (decs, cont) =
			   case cont of
			      NONE => wrapper(returns, NONE, Return)
			    | SOME c =>
				 wrapper(jump c, SOME c,
					 fn args => Jump{dst = c, args = args})
		     in (decs,
			 Call{func = f, cont = cont,
			      args = keepUseful(args, fargs)})
		     end
		| Jump{dst, args} =>
		     (decs,
		      Jump{dst = dst, args = keepUseful(args, jump dst)})
		| Return xs => (decs, Return(keepUseful(xs, returns)))
		| Raise xs => (decs, Raise(keepUseful(xs, exnVals())))
		| Case{cases = [], default = NONE, ...} =>
		     (* The test may be useless if there are no cases or default,
		      * in which case we must eliminate the case.
		      * However, can't just turn it into halt, because
		      * the handler stack won't be right.
		      *)
		     (decs, Bug)
		| t => (decs, t)

	 in Exp.make
	    {decs = decs,
	     transfer = transfer}
	 end

      val datatypes =
	 List.map(datatypes, fn {tycon, cons} =>
		  {tycon = tycon,
		   cons = List.map(cons, fn {con, args} =>
				   {con = con,
				    args = Value.newTypes(conArgs con)})})

      val globals =
	 List.foldr(globals, [], fn (b, bs) =>
		    case loopBind b of
		       NONE => bs
		     | SOME b => b :: bs)

      val globals = {var = unitVar, ty = Type.unit, exp = Tuple[]} :: globals

      val shrinkExp = shrinkExp globals

(*       val shrinkExp =
 * 	 Trace.trace("Useless.shrinkExp", Exp.layout, Exp.layout) shrinkExp
 *)
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, ...} =>
	  let val {args = argvs, returns = returnvs} = func name
	  in {name = name,
	      args = keepUsefulArgs args,
	      body = shrinkExp(loopExp(body, returnvs)),
	      returns = Value.newTypes returnvs}
	  end)

      val globals = !bogusGlobals @ globals
	     
      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
	 
   in Program.clear program
      ; destroy()
      ; program
   end

end
