(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateAtoms(S: ELABORATE_ATOMS_STRUCTS): ELABORATE_ATOMS = 
struct

open S

local open Ast
in structure Atype = Type
   structure DatatypeRhs = DatatypeRhs
   structure DatBind = DatBind
   structure Longtycon = Longtycon
   structure SortedRecord = SortedRecord
   structure TypBind = TypBind
   structure Tyvar = Tyvar
end

local open TypeStr
in structure Con = Con
   structure Cons = Cons
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
   structure TypeFcn = TypeFcn
end

structure Lookup =
   struct
      type t = Longtycon.t -> TypeStr.t

      fun fromFun f = f
      fun fromEnv(E: Env.t) longtycon = Env.lookupLongtycon(E, longtycon)
      fun plusEnv(lookup: t, E: Env.t): t =
	 fn longtycon =>
	 case Env.peekLongtycon(E, longtycon) of
	    SOME typeFcn => typeFcn
	  | NONE => lookup longtycon
      fun lookup(l: t, c: Longtycon.t) = l c
   end

(* make sure to replace all tyvars that have the same name with a single tyvar *)
fun elaborateType(ty: Atype.t, lookup: Lookup.t): Scheme.t =
   let 
      fun loop(ty: Atype.t, accum: Tyvar.t list) =
	 case Atype.node ty of
	    Atype.Var a => (* rule 44 *)
	       let
		  fun loop tyvars =
		     case tyvars of
			[] => (* new type variable, add it to the accum *)
			   (Type.var a, a :: accum)
		      | a' :: tyvars =>
			   if Tyvar.sameName(a, a')
			      then (Type.var a', accum)
			   else loop tyvars
	       in loop accum
	       end
	  | Atype.Con(c, ts) => (* rules 46, 47 *)
	       let val (ts, accum) = loops(ts, accum)
		  fun normal() = TypeFcn.apply(TypeStr.typeFcn(lookup c), ts)
	       in (case (Ast.Longtycon.split c, ts) of
		      (([], c), [t1, t2]) => if Ast.Tycon.equals(c, Ast.Tycon.arrow)
					     then Type.arrow(t1, t2)
					  else normal()
		    | _ => normal(),
		   accum)
	       end
	  | Atype.Record r => (* rules 45, 49 *)
	       let val (fs, ts) = SortedRecord.unzip r
		  val (ts, accum) = loops(ts, accum)
	       in (Type.record(SortedRecord.zip(fs, ts)), accum)
	       end

      and loops(ts, accum) =
	 List.fold(List.reverse ts, ([], accum), fn (t, (ts, accum)) =>
		   let val (t, accum) = loop(t, accum)
		   in (t :: ts, accum)
		   end)
	 
      val (ty, tyvars) = loop(ty, [])
	 
   in Scheme.T{tyvars = tyvars, ty = ty}
   end

(* Returns a scheme, plus the type variables that occured in the ty but
 * not in the tyvars.
 *)
fun elaborateScheme(tyvars, ty: Atype.t, lookup: Lookup.t)
  : Scheme.t * Tyvar.t list =
   let val Scheme.T{tyvars = tyvars', ty} = elaborateType(ty, lookup)
      (* Replace each tyvar with the corresponding tyvar'.
       * Keep track of any tyvars' that are left over.
       *)
      val (tyvars, tyvars') =
	 List.fold
	 (List.reverse tyvars, ([], tyvars'), fn (a, (tyvars, tyvars')) =>
	  let
	     fun loop(tyvars', remaining) =
		case tyvars' of
		   [] => (a, remaining)
		 | a' :: tyvars' =>
		      if Tyvar.sameName(a, a')
			 then (a', remaining @ tyvars')
		      else loop(tyvars', a' :: remaining)
	     val (a, tyvars') = loop(tyvars', [])
	  in (a :: tyvars, tyvars')
	  end)
   in (Scheme.T{tyvars = tyvars, ty = ty}, tyvars')
   end

fun elaborateClosedScheme arg: Scheme.t =
   let val (scheme, tyvars) = elaborateScheme arg
   in (case tyvars of
	  [] => ()
	| tyvar :: _ =>
	     Control.error
	     (Tyvar.region tyvar,
	      let open Layout
	      in seq[str "unbound type variables: ",
		     seq(separate(List.map(tyvars, Tyvar.layout), " "))]
	      end)) ;
      scheme
   end

fun elaborateTypeOpt(ty: Atype.t option, E: Env.t): Type.t option =
   case ty of
      NONE => NONE
    | SOME ty => SOME(Scheme.ty(elaborateType(ty, Lookup.fromEnv E)))

	    
fun elaborateTypBind(typBind, lookup: Lookup.t): Env.t =
   let val TypBind.T types = TypBind.node typBind
   in List.fold
      (types, Env.empty, fn ({tyvars, tycon, def}, E) =>
       let val tystr = TypeStr.def(elaborateClosedScheme(tyvars, def, lookup))
       in Env.extendTycon(E, tycon, tystr)
       end)
   end

fun elaborateCons{cons, tyvars, tycon: Tycon.t,
		  resultType, lookup}: Cons.t =
   let
      val cons =
	 List.map
	 (cons, fn (name, to) =>
	  let val con = Con.fromAst name
	     val scheme =
		elaborateClosedScheme(tyvars,
				      case to of
					 NONE => resultType
				       | SOME t => Atype.arrow(t, resultType),
				      lookup)
	  in (*CoreML.setConTycon(con, tycon) ;*)
	     ({name = name, con = con, scheme = scheme})
	  end)
	 
   in (*CoreML.setTyconCons(tycon, List.map(cons, #con)) ;*)
      Cons.T cons
   end
		
fun elaborateDatBind{datBind,
		     lookup: Lookup.t,
		     newTycon: Ast.Tycon.t -> Tycon.t}
  : {VE: Env.t, TE: Env.t, decs: Decs.t} =
   (* rules 28, 29, 81, 82 *)
   let
      val DatBind.T{datatypes, withtypes} = DatBind.node datBind
      (* build enough of an env so that that the withtypes
       * and the constructor argument types can be evaluated
       *)
      val (E, datatypes) =
	 List.fold
	 (datatypes, (Env.empty, []),
	  fn ({tyvars, tycon = name, cons}, (E, datatypes))=>
	  let val tycon = newTycon name
	  in (Env.extendTycon(E, name, TypeStr.tycon tycon),
	      {name = name, tycon = tycon: Tycon.t, tyvars = tyvars, cons = cons}
	      :: datatypes)
	  end)
      val lookup = Lookup.plusEnv(lookup, E)
      (* elaborate the withtypes *)
      val TE = elaborateTypBind(withtypes, lookup)
      val E = Env.+(E, TE)
      val lookup = Lookup.plusEnv(lookup, E)
      (* elaborate the datatypes, this time including the constructors *)
      val (VE, TE, datatypes) =
	 List.fold
	 (datatypes, (Env.empty, TE, []),
	  fn ({name, tyvars, tycon, cons}, (VE, TE, datatypes)) =>
	  let
	     val elabCons as Cons.T l =
		elaborateCons
		{cons = cons, tyvars = tyvars, tycon = tycon, lookup = lookup,
		 resultType = Atype.con(name, List.map(tyvars, Atype.var))}
	     val tyStr = TypeStr.data(tycon, elabCons)
	  in (Env.+(VE, Env.cons elabCons),
	      Env.extendTycon(TE, name, tyStr),
	      {tyvars = tyvars,
	       tycon = tycon,
	       cons =
	       List.map2(cons, l, fn ((_, to), {con, ...}) =>
			 {con = con,
			  arg = (case to of
				    NONE => NONE
				  | SOME t => SOME(Scheme.ty(elaborateType(t, lookup))))})}
	      :: datatypes)
	  end)
   in {VE = VE, TE = TE, decs = Decs.datatypes datatypes}
   end

fun elaborateDatatypeRhs{rhs, lookup: Lookup.t, newTycon}: Decs.t * Env.t =
   case DatatypeRhs.node rhs of
      DatatypeRhs.DatBind datBind => (* rule 17 *)
	 let val {VE, TE, decs} = elaborateDatBind{datBind = datBind,
						   lookup = lookup,
						   newTycon = newTycon}
	 in (decs, Env.+(TE, VE))
	 end
    | DatatypeRhs.Repl{lhs, rhs} => (* rule 18 *)
	 let val tyStr = lookup rhs
	    val E = Env.extendTycon(Env.empty, lhs, tyStr)
	 in (Decs.empty, Env.+(E, Env.cons(TypeStr.cons tyStr)))
	 end

end
