(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateCore(S: ELABORATE_CORE_STRUCTS): ELABORATE_CORE = 
struct

open S

local open Env
in structure Vid = Vid
end

local open Ast
in structure Record = Record
   structure Adec = Dec
   structure Aexp = Exp
   structure Atype = Type
   structure Avar = Var
   structure EbRhs = EbRhs
   structure Longvid = Longvid
   structure Longtycon = Longtycon
   structure Strid = Strid
   structure TypBind = TypBind
end

local open CoreML
in structure Cdec = Dec
   structure Cexp =
      struct
	 open Exp
	 val bogus = unit
      end
   structure Cpat = Pat
   structure Cprim = Prim
   structure Ctype = Type
   structure Cvar = Var
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Tyvar = Tyvar
end
      
local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
   val elaborateTypeOpt = elaborateTypeOpt
end

local open Env
in structure TypeStr = TypeStr
end

structure Parse = PrecedenceParse(structure Ast = Ast
				  structure Env = Env)

structure ElaboratePat = ElaboratePat(S)
   
fun constrain(e, tyOpt) =
   case tyOpt of
      NONE => e
    | SOME ty => CoreML.Exp.Constraint(e, ty)

(* val traceElaborateExp =
 *    Trace.trace2("elaborateExp", Ast.Exp.layout, Env.layout, CoreML.Exp.layout)
 *)
   
fun elaborateEbs(ebs, E: Env.t, decs, E'): Decs.t * Env.t =
   List.fold(ebs, (decs, E'), fn ((exn, rhs), (decs, E')) =>
	     let
		val (decs, exn') =
		   case EbRhs.node rhs of
		      EbRhs.Def c => (decs, Env.lookupLongcon(E, c))
		    | EbRhs.Gen to =>
			 let val exn' = CoreML.Con.fromAst exn
			 in (Decs.add(decs,
				      Cdec.Exception{con = exn',
						     arg = elaborateTypeOpt(to, E)}),
			     exn')
			 end
	     in (decs, Env.extendExn(E', exn, exn'))
	     end)
   
(*---------------------------------------------------*)
(*                   Declarations                    *)
(*---------------------------------------------------*)

local
(*    val traceElaborateDec =
 *       Trace.trace4("elaborateDec",
 * 		   Ast.Dec.layout, Env.layout, Layout.ignore, Layout.ignore,
 * 		   Layout.tuple2(Decs.layout, Env.layout))
 *)
in
   
fun elaborateDec arg =
(*   traceElaborateDec *)
   (fn (d, E, decs, E') =>
    case Adec.node d of
       Adec.Val(tyvars, vbs) => elaborateVbs(tyvars, vbs, E, decs, E')
     | Adec.Valrec(tyvars, rvbs) => elaborateRvbs(tyvars, rvbs, E, decs, E')
     | Adec.Fun(tyvars, fbs) => elaborateFbs(tyvars, fbs, E, decs, E')
     | Adec.Type typBind =>
	  (decs,
	   Env.+(E', ElaborateAtoms.elaborateTypBind(typBind, Lookup.fromEnv E)))
     | Adec.Datatype rhs =>
	  let
	     val (decs', E'') =
		ElaborateAtoms.elaborateDatatypeRhs
		{rhs = rhs,
		 lookup = Lookup.fromEnv E,
		 newTycon = CoreML.Tycon.fromAst}
	  in (Decs.append(decs, decs'), Env.+(E', E''))
	  end
     | Adec.Abstype{datBind, body} => (* rule 19 and p.57 *)
	  let
	     val {VE, TE, decs = decs'} =
		ElaborateAtoms.elaborateDatBind{datBind = datBind,
						lookup = Lookup.fromEnv E,
						newTycon = CoreML.Tycon.fromAst}
	  in elaborateDec(body, Env.plus[E, VE, TE], Decs.append(decs, decs'),
			  Env.+(E', Env.abs TE))
	  end
     | Adec.Exception ebs => elaborateEbs(ebs, E, decs, E')
     | Adec.Local(d, d') =>
	  let val (decs, E) = elaborateDec(d, E, decs, E)
	  in elaborateDec(d', E, decs, E')
	  end
     | Adec.SeqDec ds =>
	  let val (decs, _, E') =
	     List.fold(ds, (decs, E, E'), fn (d, (decs, E, E')) =>
		       let val (decs, E'') = elaborateDec(d, E, decs, Env.empty)
		       in (decs, Env.+(E, E''), Env.+(E', E''))
		       end)
	  in (decs, E')
	  end
     | Adec.Fix{ops, fixity} =>
	  (decs, Env.+(E', Env.fixs(List.map(ops, fn op' => (op', fixity)))))
     | Adec.Open paths =>
	  (decs,
	   Env.plus(E' :: (List.map(paths, fn p => Env.lookupLongstrid(E, p)))))
     | Adec.Overload(x, t, xs) =>
	  let val x' = Cvar.fromAst x
	     val scheme = elaborateType(t, Lookup.fromEnv E)
	     val E' = Env.+(E', Env.vars[(x, x', SOME scheme)])
	  in (Decs.add(decs,
		       Cdec.Overload
		       {var = x',
			scheme = scheme,
			ovlds = List.map(xs, fn x => Env.lookupLongvar(E, x))}),
	      E')
	  end) arg

and elaborateVbs(tyvars, vbs, E: Env.t, decs, E'): Decs.t * Env.t = 
   List.fold
   (vbs, (decs, E'), fn ((pat, exp), (decs, E')) =>
    let val exp = elaborateExp(exp, E)
        val (pat, E') = ElaboratePat.elaboratePat{pat = pat, env = E, accum = E'}
    in (Decs.add(decs, Cdec.Val{pat = pat,
				tyvars = tyvars,
				exp = exp}),
	E')
    end)

and elaborateFbs(tyvars, fbs, E: Env.t, decs, E'): Decs.t * Env.t =
   let
      val clausess =
	 List.map
	 (fbs, fn clauses =>
	  List.map(clauses, fn {pats, resultType, body} =>
		   let val {func, args} = Parse.parseClause(pats, E)
		      val print =
			 Aexp.longvid
			 (Ast.Longvid.long
			  ([Strid.fromString "Primitive",
			    Strid.fromString "Stdio"],
			   Ast.Vid.fromString "print"))
		   in {func = func, args = args,
		       resultType = resultType,
		       body =
		       (if !Control.printAtFunEntry
			   then Aexp.make
			      (Aexp.Seq
			       [Aexp.app(print,
					 Aexp.const(Ast.Const.String
						    ("Entering "
						     ^ Avar.toString func
						     ^ "\n"))),
				body])
			else body)}
		   end))
      val funcs =
	 List.map(clausess, fn {func, ...} :: _ => func
                             | [] => Error.error "no clauses in fundec")
      val newFuncs = List.map(funcs, Cvar.fromAst)
      val E'' = Env.vars(List.map2(funcs, newFuncs, fn (name, var) =>
				   (name, var, NONE)))
      val E = Env.+(E, E'')
   in (Decs.add
       (decs,
	Cdec.Fun
	{tyvars = tyvars,
	 decs =
	 List.map2
	 (clausess, newFuncs, fn (clauses,newFunc) =>
	  case clauses of
	     [] => Error.error "empty clauses in fundec"
	   | {args, ...} :: _ =>
		let val numVars = List.length args
		in {var = newFunc, ty = NONE,
		    rules =
		    let
		       val rs =
			  List.map
			  (clauses, fn {args, resultType, body, ...} =>
			   let
			      val (pats, E') =
				 ElaboratePat.elaboratePats{pats = args,
							    env = E,
							    accum = E}
			      val body = elaborateExp(body, E')
			   in (Cpat.tuple pats,
			       constrain(body,
					 elaborateTypeOpt(resultType, E)))
			   end)
		       fun make(0, xs) =
			  Cexp.casee(Cexp.tuple(List.fold(xs, [], fn (x, xs) =>
							  (Cexp.Var x) :: xs)),
				     rs)
			 | make(n, xs) =
			   let val x = Cvar.newNoname()
			   in Cexp.lambda(x, make(n-1, x :: xs))
			   end
		    in if numVars = 1 then rs
		       else (case make(numVars, []) of
				Cexp.Fn rules => rules
			      | _ => Error.bug "elaborateFbs")
		    end}
		end)}),
       Env.+(E', E''))
   end

and elaborateRvbs(tyvars, rvbs, E: Env.t, decs, E'): Decs.t * Env.t =
   let val vars = List.map(rvbs, #var)
      val newvars = List.map(vars, Cvar.fromAst)
      val E'' = Env.vars(List.map2(vars, newvars, fn (name, var) =>
				   (name, var, NONE)))
      val E = Env.+(E, E'')
   in (Decs.add
       (decs,
	Cdec.Fun
	{tyvars = tyvars,
	 decs = List.map2(rvbs, newvars, fn ({var, fixity, rules, ty},newvar) =>
			  {var = newvar,
			   ty = elaborateTypeOpt(ty, E),
			   rules = elaborateRules(rules, E)})}),
       Env.+(E', E''))
   end

(*---------------------------------------------------*)
(*                    Expressions                    *)
(*---------------------------------------------------*)

and elaborateExp arg: CoreML.Exp.t =
(*   traceElaborateExp *)
   (fn (e: Aexp.t, E: Env.t) =>
   case Aexp.node e of
      Aexp.Var{name=id, ...} =>
	 (case Env.lookupLongvid(E, id) of
	     vid => (case vid of
			Vid.Var x => Cexp.Var x
		      | Vid.ConAsVar c => Cexp.Con c
		      | Vid.Con c => Cexp.Con c
		      | Vid.Exn c => Cexp.Con c
		      | Vid.Prim p => Cexp.Prim p))
    | Aexp.Fn rs => Cexp.Fn(elaborateRules(rs, E))
    | Aexp.FlatApp items => elaborateExp(Parse.parseExp(items, E), E)
    | Aexp.App(e1, e2) => Cexp.App(elaborateExp(e1, E), elaborateExp(e2, E))
    | Aexp.Case(e, rs) => Cexp.casee(elaborateExp(e, E), elaborateRules(rs, E))
    | Aexp.Let(d, e) => let val (d, E) = elaborateDec(d, E, Decs.empty, E)
		       in Cexp.Let(Decs.toList d, elaborateExp(e, E))
		       end
    | Aexp.Seq es => Cexp.seq(elaborateExps(es, E))
    | Aexp.Const c => Cexp.Const c
    | Aexp.Record r => Cexp.Record(Record.map(r, fn e => elaborateExp(e, E)))
    | Aexp.List es => Cexp.list(elaborateExps(es, E))
    | Aexp.Selector f => Cexp.selector f
    | Aexp.Constraint(e, t) =>
	 Cexp.Constraint(elaborateExp(e, E),
			 Scheme.ty(elaborateType(t, Lookup.fromEnv E)))
    | Aexp.Handle(e, rs) => Cexp.Handle(elaborateExp(e, E), elaborateRules(rs, E))
    | Aexp.Raise e => Cexp.Raise(elaborateExp(e, E))
    | Aexp.If(a, b, c) =>
	 Cexp.iff(elaborateExp(a, E), elaborateExp(b, E), elaborateExp(c, E))
    | Aexp.Andalso(e, e') => Cexp.andAlso(elaborateExp(e, E), elaborateExp(e', E))
    | Aexp.Orelse(e, e') => Cexp.orElse(elaborateExp(e, E), elaborateExp(e', E))
    | Aexp.While{test, expr} =>
	 Cexp.whilee{test = elaborateExp(test, E),
		     expr = elaborateExp(expr, E)}
    | Aexp.Prim{name, ty} =>
	 Cexp.Prim
	 (Cprim.new(name, ElaborateAtoms.elaborateType(ty, Lookup.fromEnv E)))
	 ) arg
   
and elaborateExps(es: Ast.Exp.t list, E: Env.t): CoreML.Exp.t list =
   List.map(es, fn e => elaborateExp(e, E))

and elaborateRules(rs, E: Env.t) =
   List.map(rs, fn (pat, exp) => 
	    let
	       val (pat, E) =
		  ElaboratePat.elaboratePat{pat = pat,
					    env = E,
					    accum = E}
	    in (pat, elaborateExp(exp, E))
	    end)

end (* local *)

end
