(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateEnv(S: ELABORATE_ENV_STRUCTS): ELABORATE_ENV =
struct

open S

local open Ast
in structure Fixity = Fixity
   structure Strid = Strid
   structure Longcon = Longcon
   structure Longvid = Longvid
   structure Longstrid = Longstrid
   structure Longtycon = Longtycon
end

local open CoreML
in structure Con = Con
   structure Var = Var
   structure Prim = Prim
   structure Record = Record
   structure Srecord = SortedRecord
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
   structure Tyvar = Tyvar
   structure Var = Var
end

local open Interface
in structure Element = Element
   structure ShapeId = ShapeId
   structure Status = Status
end

local open TypeStr
in structure Cons = Cons
   structure TypeFcn = TypeFcn
end

structure SchemeOpt = MonoOption(Scheme)

structure Vid =
   struct
      open CoreML
	 
      datatype t =
	 Var of Var.t
       | ConAsVar of CoreML.Con.t
       | Con of Con.t
       | Exn of Con.t
       | Prim of Prim.t

      val status =
	 fn Var _ => Status.Var
	  | Prim _ => Status.Var
	  | ConAsVar _ => Status.Var
	  | Con _ => Status.Con
	  | Exn _ => Status.Exn

      val bogus = Var Var.bogus

      fun layout vid =
	 let open Layout
	    val (name, l) =
	       case vid of
		  Var v => ("Var", Var.layout v)
		| Con c => ("Con", Con.layout c)
		| ConAsVar c => ("ConAsVar", Con.layout c)
		| Exn c => ("Exn", Con.layout c)
		| Prim p => ("Prim", Prim.layout p)
	 in l(* paren(seq[str name, str " ", l])*)
	 end

      val deVar =
	 fn Var v => SOME v
	  | _ => NONE
		
      val deCon =
	 fn Con c => SOME c
	  | Exn c => SOME c
	  | _ => NONE

      val dePrim =
	 fn Prim p => SOME p
	  | _ => NONE
	  
      fun output(r, out) = Layout.output(layout r, out)
	 
      val equals = 
	 fn (Var x, Var x') => Var.equals(x, x')
	  | (Con c, Con c') => Con.equals(c, c')
	  | (Exn c, Exn c') => Con.equals(c, c')
	  | (Prim p, Prim p') => Prim.equals(p, p')
	  | _ => false
   end

(*------------------------------------*)
(*               Domain               *)
(*------------------------------------*)

structure Domain =
   struct
      open Ast
	 
      datatype t =
	 Fixity of Vid.t
	| Vid of Vid.t
	| Tycon of Tycon.t
	| Strid of Strid.t

      val Var = Vid o Ast.Vid.fromVar
      val Prim = Var
      val Con = Vid o Ast.Vid.fromCon

      fun hash x =
	 case x of
	    Fixity x => Vid.hash x
	  | Vid x => Vid.hash x
	  | Tycon x => Tycon.hash x
	  | Strid x => Strid.hash x

      val equals =
	 fn (Fixity x, Fixity y) => Vid.equals(x, y)
	  | (Vid x, Vid y) => Vid.equals(x, y)
	  | (Tycon x, Tycon y) => Tycon.equals(x, y)
	  | (Strid x, Strid y) => Strid.equals(x, y)
	  | _ => false

      fun layout d =
	 case d of
	    Fixity x => Vid.layout x
	  | Vid x => Vid.layout x
	  | Tycon x => Tycon.layout x
	  | Strid x => Strid.layout x
   end

structure Table = HashTable

(*---------------------------------------------------*)
(*                 Main Env Datatype                 *)
(*---------------------------------------------------*)

datatype t = T of {elements: elements,
		   shapeId: ShapeId.t option}
and elements =
   Empty
  | Table of table
  | Extend of (Domain.t * range) list * elements
  | Plus of elements list (* newer bindings in front *)
and range =
   Fixity of Ast.Fixity.t
  | Vid of {vid: Vid.t, scheme: SchemeOpt.t}
  | Tycon of TypeStr.t
  | Strid of t  (* always consolidated *)
withtype table = (Domain.t, range) Table.hash_table

type env = t

fun stamp(T{elements, ...}, shapeId) =
   T{elements = elements, shapeId = SOME shapeId}

fun unconstrained(es: elements): t =
   T{elements = es, shapeId = NONE}

val bogus = unconstrained Empty
   
fun elements(T{elements, ...}) = elements

(*val equals = fn _ => Error.unimplemented "ElaborateEnv.="*)

exception NotFound
fun consolidate'(E: elements): table =
   case E of
      Table t => t
    | _ => 
	 let 
	    (* isn't really accurate - just used to get an approximation *)
	    fun size(E, accum) = 
	       case E of
		  Empty => accum
		| Table t => accum + Table.numItems t
		| Extend(bs, E) => size(E, accum + List.length bs)
		| Plus Es => List.fold(Es, accum, size)
	    val size = size(E, 0)
	    val table = Table.mkTable(Domain.hash, Domain.equals) (size, NotFound)
	    fun insert(d, r) =
	       case Table.find table d of
		  NONE => Table.insert table (d, r)
		| SOME _ => ()
	    fun loop E =
	       case E of
		  Empty => ()
		| Table t => Table.appi insert t
		| Extend(drs, E) => (List.foreach(drs, insert) ; loop E)
		| Plus Es => List.foreach(Es, loop)
	 in loop E ; table
	 end

fun consolidate(T{elements, shapeId}): t =
   T{elements = Table(consolidate' elements),
     shapeId = shapeId}

local
   open Layout
in
   fun layout(E:env): Layout.t =
      let val t = consolidate'(elements E)
      in list(List.map(Table.listItemsi t, fn (d, r) =>
		       seq[Domain.layout d,
			   str " -> ",
			   layoutRange r]))
      end
   and layoutRange r =
      case r of
	 Fixity x => Fixity.layout x
       | Vid{scheme, vid} => tuple[Vid.layout vid,
				  case scheme of
				     NONE => empty
				   | SOME s => Scheme.layout s]
       | Tycon x => TypeStr.layout x
       | Strid x => layout x
   val layoutEnv = layout
end

val deFixity =
   fn Fixity x => SOME x
    | _ => NONE
val deVar =
   fn Vid{vid, ...} => Vid.deVar vid
    | _ => NONE
val deCon =
   fn Vid{vid, ...} => Vid.deCon vid
    | _ => NONE
val dePrim =
   fn Vid{vid, ...} => Vid.dePrim vid
    | _ => NONE
val deVid =
   fn Vid{vid, ...} => SOME vid
    | _ => NONE
val deTycon =
   fn Tycon x => SOME x
    | _ => NONE
val deStrid =
   fn Strid x => SOME x
    | _ => NONE

val empty = unconstrained Empty

fun plus(Es: t list): t =
   unconstrained(Plus(List.fold(Es, [], fn (E, Es) =>
				elements E :: Es)))

fun extend(E, d, r) =
   unconstrained
   (case elements E of
       Extend(drs, E) => Extend((d, r) :: drs, E)
     | E => Extend([(d, r)], E))

fun extends(E, drs) =
   case E of
      Extend(drs', E) => Extend(List.append(drs, drs'), E)
    | E => Extend(drs, E)

local
   val extend = fn (D, R) => fn (E, d, r) => extend(E, D d, R r)
   fun noScheme make vid = Vid{scheme = NONE, vid = make vid}
in val extendVar = extend(Domain.Var, noScheme Vid.Var)
   val extendPrim = extend(Domain.Prim, noScheme Vid.Prim)
   val extendCon = extend(Domain.Con, noScheme Vid.Con)
   val extendExn = extend(Domain.Con, noScheme Vid.Exn)
   val extendTycon = extend(Domain.Tycon, Tycon)
   val extendStrid = extend(Domain.Strid, Strid o consolidate)
end

local
   fun list(D: 'a -> Domain.t, R: 'b -> range) (drs: ('a * 'b) list): t =
      unconstrained(Extend(List.map(drs, fn (d, r) => (D d, R r)), Empty))
in val fixs = list(Domain.Fixity, Fixity)
   val strs = list(Domain.Strid, Strid o consolidate)
   val types = list(Domain.Tycon, Tycon)
end

fun cons(Cons.T l) =
   unconstrained
   (Extend
    (List.map(l, fn {name, con, scheme} =>
	      (Domain.Con name,
	       Vid{scheme = SOME scheme, vid = Vid.Con con})),
     Empty))

val fromCons = cons

fun vars l =
   unconstrained
   (Extend(List.map(l, fn (name, var, scheme) =>
		    (Domain.Var name,
		     Vid{vid = Vid.Var var, scheme = scheme})),
	   Empty))

type ('a, 'b) peek = t * 'a -> 'b option

local
   fun peek(D: 'a -> Domain.t, de: range -> 'b option): ('a, 'b) peek =
      fn (E: t, d :'a) =>
      let val d = D d
	 fun loop E =
	    case E of
	       Empty => NONE
	     | Extend(drs, E) =>
		  (case List.peek(drs, fn (d', _) => Domain.equals(d, d')) of
		      NONE => loop E
		    | SOME(_, r) => de r)
	     | Table t => (case Table.find t d of
			      NONE => NONE
			    | SOME r => de r)
	     | Plus Es =>
		  let fun loop' [] = NONE
			| loop'(E :: Es) = (case loop E of
					     NONE => loop' Es
					   | x => x)
		  in loop' Es
		  end
      in loop(elements E)
      end
in
   val peekFixity = peek(Domain.Fixity, deFixity)
   val peekVar = peek(Domain.Var, deVar)
   val peekCon = peek(Domain.Con, deCon)
   val peekPrim = peek(Domain.Prim, dePrim)
   val peekVid = peek(Domain.Vid, deVid)
   val peekTycon = peek(Domain.Tycon, deTycon)
   val peekStrid = peek(Domain.Strid, deStrid)
   val peek = peek(fn x => x, SOME)
end

fun lookupFixity(E, vid) =
   case peekFixity(E, vid) of
      NONE => Fixity.Nonfix
    | SOME fixity => fixity

fun toIsInfix (E:t) (vid:Ast.Vid.t): bool =
   case peekFixity(E, vid) of
      NONE => false
    | SOME f => Fixity.isInfix f

fun peekStrids(E: t, strids): t option =
   case strids of
      [] => SOME E
    | strid :: strids =>
	 case peekStrid(E, strid) of
	    SOME E => peekStrids(E, strids)
	  | NONE => NONE

fun peekLong(split, peek) (E, x) =
   let val (strids, x) = split x
   in case peekStrids(E, strids) of
      SOME E => peek(E, x)
    | NONE => NONE
   end

val peekLongvid = peekLong(Ast.Longvid.split, peekVid)
val peekLongvar = peekLong(Ast.Longvar.split, peekVar)
val peekLongcon = peekLong(Ast.Longcon.split, peekCon)
val peekLongtycon = peekLong(Ast.Longtycon.split, peekTycon)
val peekLongstrid = peekLong(Ast.Longstrid.split, peekStrid)

local
   fun lookup(peek: ('a, 'b) peek,
	      bogus: 'b,
	      unbound: 'a -> unit) (E: t, x: 'a): 'b =
      case peek(E, x) of
	 SOME y => y
       | NONE => (unbound x ; bogus)
in
   val lookupTycon =
      lookup(peekTycon, TypeStr.bogus, Ast.Tycon.unbound)
   val lookupStrid =
      lookup(peekStrid, bogus, Ast.Strid.unbound)
   val lookupLongvid =
      lookup(peekLongvid, Vid.bogus, Ast.Longvid.unbound)
   val lookupLongvar =
      lookup(peekLongvar, Var.bogus, Ast.Longvar.unbound)
   val lookupLongcon =
      lookup(peekLongcon, Con.bogus, Ast.Longcon.unbound)
   val lookupLongtycon =
      lookup(peekLongtycon, TypeStr.bogus, Ast.Longtycon.unbound)
   val lookupLongstrid =
      lookup(peekLongstrid, bogus, Ast.Longstrid.unbound)
(*   val lookup =
      lookup(peek, Domain.layout) *)
end

(*---------------------------------------------------*)
(*                        cut                        *)
(*---------------------------------------------------*)

(* section 5.3, 5.5, 5.6 and rules 52, 53 *)
(* transparent:
     look at the stamp within the env *)
fun cut{env, interface, opaque, region}: t =
   let
      fun error(name, l) =
	 Control.error
	 (region, let open Layout
		  in seq[str name, str " ", l,
			 str " in signature but not in structure"]
		  end)
      val interface =
	 Interface.realize
	 (interface, fn c =>
	  if opaque
	     then TypeFcn.Tycon(Tycon.newNoname())
	  else
	     TypeStr.typeFcn
	     (case peekLongtycon(env, c) of
		 NONE => (error(Longtycon.className, Longtycon.layout c) ;
			  TypeStr.bogus)
	       | SOME typeStr => typeStr))
      fun cut(E as T{elements, shapeId}, I, strids) =
	 let val shapeId' = Interface.shapeId I
	    val cutoff =
	       if opaque then NONE
	       else case shapeId of
		  NONE => NONE
		| SOME shapeId =>
		     if ShapeId.equals(shapeId, shapeId')
			then SOME E
		     else NONE
	 in case cutoff of
	    SOME E => E
	  | NONE =>
	       let
		  val E' =
		     Interface.fold
		     (I, empty, fn (elt, E') =>
		      case elt of
			 Element.Str{name, interface=I} =>
			    (case peekStrid(E,name) of
				NONE =>
				   (error(Longstrid.className,
					  Longstrid.layout	
					  (Longstrid.long(List.reverse strids,
							  name))) ;
				    E')
			       | SOME E =>
				    extendStrid(E', name, cut(E, I, name :: strids)))
			| Element.Val{name, status, scheme} =>
			     (case peekVid(E, name) of
				 NONE =>
				    (error(Longvid.className,
					   Longvid.layout
					   (Longvid.long(List.reverse strids,
							 name))) ;
				     E')
			       | SOME vid =>
				    let
				       val vid =
					  case (vid, status) of
					     (Vid.Con c, Status.Var) =>
						Vid.ConAsVar c
					   | (Vid.Exn c, Status.Var) =>
						Vid.ConAsVar c
					   | (_, Status.Var) => vid
					   | (Vid.Con _, Status.Con) => vid
					   | (Vid.Exn _, Status.Exn) => vid
					   | _ =>
						(Control.error
						 (region,
						  let open Layout
						  in seq
						     [str Longvid.className,
						      str " ",
						      Longvid.layout
						      (Longvid.long(List.reverse strids,
								    name)),
						      str " has status ",
						      Status.layout(Vid.status vid),
						      str " in structure but status ",
						      Status.layout status,
						      str " in signature "]
						  end)
						 ; vid)
				    in extend(E', Domain.Vid name,
					      Vid{vid = vid,
						  scheme = SOME scheme})
				    end)
			| Element.Type{name, typeStr, ...} => 
			     case peekTycon(E,name) of
				NONE =>
				   (error(Longtycon.className,
					  Longtycon.layout
					  (Longtycon.long(List.reverse strids,
							  name))) ;
				    E')
			      | SOME typeStr' =>
				   let val (typeFcn, cons) = TypeStr.dest typeStr
				   in extendTycon
				      (E', name,
				       TypeStr.make
				       (typeFcn,
					let
					   val cons' =
					      Cons.toList(TypeStr.cons typeStr')
					   fun lookup name =
					      case List.peek(cons',
							     fn {name=name', ...} =>
							     Ast.Con.equals(name,name')) of
						 NONE =>
						    (error(Longcon.className,
							   Longcon.layout
							   (Longcon.long
							    (List.reverse strids,
							     name))) ;
						     Con.bogus)
					       | SOME{con, ...} => con
					in Cons.T
					   (List.map
					    (Cons.toList cons, fn {name, scheme, ...} =>
					     {name = name, con = lookup name,
					      scheme = scheme}))
					end))
				   end)

		in stamp(E', shapeId')
		end
	 end
	 
   in cut(env, interface, [])
   end

fun abs(E: t) =
   unconstrained
   (Table
    (Table.map
     (fn Tycon typeStr => Tycon(TypeStr.fromTypeFcn(TypeStr.typeFcn typeStr))
   | r => r)
     (consolidate'(elements E))))

(*---------------------------------------------------*)
(*                      primEnv                      *)
(*---------------------------------------------------*)
fun addEquals E = extendPrim(E, Ast.Var.fromString "=", Prim.equal)

fun prim() =
   let 
      val E = empty

      val E =
	 List.fold
	 (Tycon.prims, E, fn (tycon, E) =>
	  extendTycon(E,
		      Ast.Tycon.fromString(CoreML.Tycon.originalName tycon),
		      TypeStr.tycon tycon))

      local
	 open Scheme Type
	 val --> = arrow
	 infixr -->
      in
	 val datatypes =
	    [{tycon = Tycon.bool,
	      tyvars = [],
	      cons = [{con = Con.falsee, arg = NONE},
		      {con = Con.truee, arg = NONE}]},
	     let val a = Tyvar.newNoname{equality = false}
	     in {tycon = Tycon.list,
		 tyvars = [a],
		 cons = [{con = Con.nill, arg = NONE},
			 {con = Con.cons, arg = SOME(tuple[var a, list(var a)])}]}
	     end,
	     let val a = Tyvar.newNoname{equality = false}
	     in {tycon = Tycon.reff,
		 tyvars = [a],
		 cons = [{con = Con.reff, arg = SOME(var a)}]}
	     end]
      end
   
      val E =
	 List.fold
	 (datatypes, E, fn ({tyvars, tycon, cons}, E) =>
	  let
	     val cons as Cons.T cs =
		Cons.T(List.map
		       (cons, fn {con, arg} =>
			let
			   val resultType =
			      Type.con(tycon, List.map(tyvars, Type.var))
			   val scheme =
			      Scheme.T
			      {tyvars = tyvars,
			       ty = (case arg of
					NONE => resultType
				      | SOME t => Type.arrow(t, resultType))}
			in {name = Con.toAst con,
			    con = con,
			    scheme = scheme}
			end))
	  in plus[fromCons cons,
		  extendTycon(E, Tycon.toAst tycon, TypeStr.data(tycon, cons))]
	  end)
	 
      val E = extendTycon(E, Ast.Tycon.fromString "unit",
			  TypeStr.def(Scheme.make0 Type.unit))
	 
      val E = addEquals E

      val exceptions =
	 List.map(CoreML.Con.excons, fn c => {con = c, arg = NONE})

      val E =
	 List.fold(CoreML.Con.excons, E, fn (c, E) =>
		   (extendCon(E, Con.toAst c, c)))
	 
   in (E, CoreML.Constructors.T{datatypes = datatypes,
				exceptions = exceptions})
   end

(*------------------------------------*)
(*              identify              *)
(*------------------------------------*)

fun containsRecord(ty: Type.t): bool =
   case ty of
      Type.Var _ => false
    | Type.Con(_, ts) => List.exists(ts, containsRecord)
    | Type.Record r =>
	 case Srecord.detupleOpt r of
	    NONE => true
	  | SOME ts => List.exists(ts, containsRecord)

local
   local open Ast
   in structure Acon = Con
      structure Dec = Dec
      structure Exp = Exp
      structure Longcon = Longcon
      structure Longvid = Longvid
      structure Pat = Pat
      structure Strdec = Strdec
      structure Strexp = Strexp
      structure Avid = Vid
      structure Topdec = Topdec
   end
   structure Field = Srecord.Field
in
   fun error() = Error.bug "wrap/unwrap"
   fun longvidFromStr s = Longvid.short(Avid.fromString s)
   fun longconFromStr s = Longcon.short(Acon.fromString s)
   local val c = Counter.new 0
   in fun newLongvid() = longvidFromStr("temp" ^ Int.toString(Counter.next c))
   end
   val none = longvidFromStr "NONE"
   val someCon = longconFromStr "SOME"
   val someVid = longvidFromStr "SOME"
      
   fun bind(e: Exp.t, f: Exp.t -> Exp.t): Exp.t =
      case Exp.node e of
	 Exp.Var _ => f e
       | _ => let val x = newLongvid()
	      in Exp.make(Exp.Let(Dec.make(Dec.Val([], [(Pat.longvid x, e)])),
				  f(Exp.longvid x)))
	      end

   fun wrap(ty: Type.t, e: Exp.t,
	    record as (makeRecord: Type.t Srecord.t * Exp.t list -> Exp.t Record.t,
		       fields: Type.t Srecord.t -> Field.t list),
	    tuple): Exp.t =
      let val unwrap = fn (ty, e) => wrap(ty, e, tuple, record)
	  val wrap = fn (ty, e) => wrap(ty, e, record, tuple)
      in if not(containsRecord ty)
	    then e
	 else
	    (case ty of
		Type.Con(c, [t]) =>
		   if String.equals("option", Layout.toString(Tycon.layout c))
		      then
			 let val x = newLongvid()
			 in Exp.casee
			    (e, [(Pat.longvid none, Exp.longvid none),
				 (Pat.make(Pat.App(someCon, Pat.longvid x)),
				  Exp.app(Exp.longvid someVid,
					  wrap(t, Exp.longvid x)))])
			 end
		   else error()
	      | Type.Con(c, [t1, t2]) =>
		   if Tycon.equals(c, Tycon.arrow)
		      then
			 let val x = newLongvid()
			 in Exp.fnn
			    [(Pat.longvid x,
			      wrap(t2, Exp.app(e, unwrap(t1, Exp.longvid x))))]
			 end
		   else error()
	      | Type.Record r =>
		   let val fs = fields r
		      val temps = List.map(fs, fn _ => newLongvid())
		   in Exp.make
		      (Exp.Let
		       (Dec.make(Dec.Val
				 ([], [(Pat.make
					(Pat.Record
					 {flexible = false,
					  items = 
					  List.map2(fs, temps, fn (f, t) =>
						    Pat.Item.Field
						    (f, Pat.longvid t))}),
					e)])),
			Exp.record
			(makeRecord
			 (r, List.map2(Srecord.toList r, temps,
				       fn ((_, t), x) =>
				       wrap(t, Exp.longvid x))))))
		   end
	      | _ => error())
      end

   val wrap = fn (ty, e) =>
      wrap(ty, e,
	   (fn (_, es) => Record.tuple es,
	    fn r => List.map(Srecord.toList r, #1)),
	   (fn (r, es) => Record.fromList(List.map2
					  (Srecord.toList r, es, fn ((f, _), e) =>
					   (f, e))),
	    fn r => List.mapi(Srecord.toList r, fn (i, _) => Field.Int(i + 1))))

(*    val wrap =
 *       Trace.trace2("wrap", Type.layout, Exp.layout, Exp.layout) wrap
 *)

   fun makeWrapper(ty: Type.t, x: Avid.t): Dec.t =
      let val x = Longvid.short x
      in Dec.make(Dec.Val([], [(Pat.longvid x, wrap(ty, Exp.longvid x))]))
      end

fun identify(E: t) =
   let
(*       val traceLoop = Trace.trace("identify", fn (E, _) => layout E, layout)
 *)
      (* keep track of what has been renamed so we can give things the
       * shortest possible name
       *)

      local
	 fun 'a rename(toString: 'a -> string,
		       setName: 'a * string -> unit,
		       plist: 'a -> PropertyList.t) =
	    let
	       val {get, set, destroy} =
		  Property.newDest(plist, Property.initConst false)
	    in {destroy = destroy,
		rename = fn (x, name) =>
		if get x
		   then if String.size name < String.size(toString x)
			   then setName(x, name)
			else ()
		else (set(x, true) ; setName(x, name))}
	    end
      in
	 val {destroy = destroyTycon, rename = renameTycon} =
	    rename(Tycon.toString, Tycon.setName, Tycon.plist)
	 val {destroy = destroyCon, rename = renameCon} =
	    rename(Con.toString, Con.setName, Con.plist)
      end

      val datatypes = ref []
      val exceptions = ref []

      fun loop arg: t * Ast.Strdec.t list =
	 (fn (E, strids: Strid.t list) =>
	  let
	     val wrappers: Ast.Strdec.t list ref = ref []
	     fun addWrapper(d: Ast.Strdec.t): unit = ListRef.push(wrappers, d)
		
	     fun longvid x = Longvid.long(strids, x)
	     fun ident(d, r) =
		case (d, r) of
		   (Domain.Fixity _, Fixity _) => r
		 | (Domain.Vid x, Vid{scheme, vid}) =>
		      Vid
		      {scheme = scheme,
		       vid = 
		       let 
			  fun con c =
			     (renameCon(c, Longvid.toString(longvid x)) ;
			      vid)
		       in case vid of
			  Vid.Var y =>
			     (case scheme of
				 NONE =>
				    (let open Layout
				     in output(Var.layout y, Out.error)
				     end ;
				     Error.error "identify encountered var without scheme")
			       | SOME s =>
				    let
				       val p =
					  Prim.new(Longvid.toString(longvid x), s)
				       val ty = Scheme.ty s
				    in if containsRecord ty
					  then addWrapper(Strdec.make
							  (Strdec.Core
							   (makeWrapper(ty, x))))
				       else () ;
				       Vid.Prim p
				    end)
			| Vid.Prim p =>
 			     (Control.messageStr
			      ("identify encountered primitive "
			       ^ Layout.toString(Prim.layout p))
			      ; Vid.Prim p)
			| Vid.ConAsVar c => con c
			| Vid.Con c => con c
			| Vid.Exn c =>
			     (case scheme of
				 SOME s =>
				    ListRef.push
				    (exceptions,
				     {con = c,
				      arg = (case (Type.dearrowOpt
						   (Scheme.apply(s, []))) of
						NONE => NONE
					      | SOME(t, _) => SOME t)})
			       | NONE => ()
			      ; con c)
		       end}
		 | (Domain.Tycon c, Tycon typeStr) =>
		      let val (typeFcn, cons) = TypeStr.dest typeStr
			 val cons = Cons.toList cons
		      in (case typeFcn of
			     TypeFcn.Fun _ => ()
			   | TypeFcn.Tycon tycon =>
				(case cons of
				    [] => ()
				  | {scheme = Scheme.T{tyvars, ...}, ...} :: _ =>
				       ListRef.push
				       (datatypes,
					let
					   val tyvars =
					      List.map(tyvars, fn a =>
						       Tyvar.newNoname
						       {equality =
							Tyvar.isEquality a})
					   val tys = List.map(tyvars, Type.var)
					   fun patchScheme s =
					      case Type.dearrowOpt(Scheme.apply(s, tys)) of
						 NONE => NONE
					       | SOME(t, _) => SOME t
					in {tycon = tycon,
					    tyvars = tyvars,
					    cons =
					    List.map(cons, fn {con, scheme, ...} =>
						     {con = con,
						      arg = patchScheme scheme})}
					end);
				 renameTycon
				 (tycon, (Longtycon.toString
					  (Longtycon.long(strids, c))))));
			 r
		      end
		 | (Domain.Strid s, Strid E) =>
		      let val (E, decs) = loop(E, strids @ [s])
		      in (case decs of
			     [] => ()
			   | _ =>
				addWrapper
				(Ast.Strdec.structuree
				 [{name = s,
				   def = Strexp.structt
				   (Strdec.seq
				    (Strdec.openn[Ast.Longstrid.short s]
				     :: decs)),
				   constraint = Ast.SigConst.None}])) ;
			 Strid E
		      end
		 | _ => Error.bug "strange (domain, range)"
	     val table = Table.mapi ident (consolidate'(elements E))
	  in (unconstrained(Table table), !wrappers)
	  end) arg

      val (E, strdecs) = loop(E, [])
	 
   in destroyCon();
      destroyTycon();
      
      (addEquals E,
       Ast.Program.T(List.map(strdecs, Topdec.strdec)),
       CoreML.Constructors.T{datatypes = !datatypes,
			     exceptions = !exceptions})
   end
end (* local *)

fun E + E' = plus[E, E']

end (* functor ElaborateEnv *)
