(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_ENV_STRUCTS =
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      structure Interface: INTERFACE
      sharing Ast = CoreML.Ast
      sharing CoreML = Interface.CoreML
      structure TypeStr: TYPE_STR
      sharing TypeStr = Interface.TypeStr
   end

signature ELABORATE_ENV =
   sig
      include ELABORATE_ENV_STRUCTS

      (* the value of a vid.  This is used to distinguish between vids whose
       * status cannot be determined at parse time
       *)
      structure Vid :
	 sig
	    datatype t =
	       Var of CoreML.Var.t
	     | ConAsVar of CoreML.Con.t (* a constructor, but it has status
					 * of a variable
					 *)
	     | Con of CoreML.Con.t
	     | Exn of CoreML.Con.t
	     | Prim of CoreML.Prim.t

	    val deVar: t -> CoreML.Var.t option
	    val deCon: t -> CoreML.Con.t option
	       
	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      type t

      val empty: t
	 
      (* environment with all the primitive (data)types and constructors *)
      val prim: unit -> t * CoreML.Constructors.t
	 
      (* overlay the bindings in the second arg on top of the first *)
      val + : t * t -> t
	 
      (* overlay a sequence -- later ones have precedence *)
      val plus: t list -> t
	 
      (* section 4.9 p. 20.  remove constructors from tyStrs *)
      val abs: t -> t
	 
      (* extract only the fixity portion of the env and return true if
       the argument vid is in the env and is infix.  Used for Ast layout. *)
      val toIsInfix: t -> Ast.Vid.t -> bool
	 
      (* filters out only those bindings in the env that also appear in
       the interface.  proceeds recursively on substructures *)
      val cut: {env: t,
		interface: Interface.t,
		opaque: bool,
		region: Region.t} -> t
	 
      (* Change all the variables to primitives, and make sure
       * everything has a type scheme.
       * Also returns an Ast that should be prepended to any program
       * that uses the primitives in order to patch the record types.
       *)
      val identify: t -> t * Ast.Program.t * CoreML.Constructors.t
	 
      (* This is extensionally the identity function.  But consolidates all of
       * the env into a single hash table so that things are faster.
       *)
      val consolidate: t -> t

      (* add a binding for the "=" primitive *)
      val addEquals: t -> t
	 
      (* constructors *)
      val fixs: (Ast.Vid.t * Ast.Fixity.t) list -> t
      val strs: (Ast.Strid.t * t) list -> t
      val vars: (Ast.Var.t * CoreML.Var.t * CoreML.Scheme.t option) list -> t
      val cons: TypeStr.Cons.t -> t
      val types: (Ast.Tycon.t * TypeStr.t) list -> t
	 
      (* add a new binding to an env *)
      val extendVar: t * Ast.Var.t * CoreML.Var.t -> t
      val extendCon: t * Ast.Con.t * CoreML.Con.t -> t
      val extendExn: t * Ast.Con.t * CoreML.Con.t -> t
      val extendPrim: t * Ast.Var.t * CoreML.Prim.t -> t
      val extendTycon: t * Ast.Tycon.t * TypeStr.t -> t
      val extendStrid: t * Ast.Strid.t * t -> t
	 
      (* check if a binding is in an env *)
      val peekLongvid: t * Ast.Longvid.t -> Vid.t option
      val peekLongvar: t * Ast.Longvar.t -> CoreML.Var.t option
      val peekLongcon: t * Ast.Longcon.t -> CoreML.Con.t option
      val peekLongtycon: t * Ast.Longtycon.t -> TypeStr.t option
      val peekLongstrid: t * Ast.Longstrid.t -> t option
      val peekVar: t * Ast.Var.t -> CoreML.Var.t option

      (* lookup a binding in an env *)
      val lookupFixity: t * Ast.Vid.t -> Ast.Fixity.t
      val lookupLongvid: t * Ast.Longvid.t -> Vid.t
      val lookupLongvar: t * Ast.Longvar.t -> CoreML.Var.t
      val lookupLongcon: t * Ast.Longcon.t -> CoreML.Con.t
      val lookupLongtycon: t * Ast.Longtycon.t -> TypeStr.t
      val lookupLongstrid: t * Ast.Longstrid.t -> t

      val layout: t -> Layout.t
   end

