(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaboratePat(S: ELABORATE_PAT_STRUCTS): ELABORATE_PAT = 
struct

open S

local open Ast
in structure Apat = Pat
   structure Field = Record.Field
   structure Fixop = Fixop
   structure Longvid = Longvid
   structure Record = Record
   structure Vid = Vid
end

local open CoreML
in structure Cpat = Pat
   structure Scheme = Scheme
end

local open Env
in structure TypeStr = TypeStr
end

structure Parse = PrecedenceParse(structure Ast = Ast
				  structure Env = Env)

local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
end

fun bind(x: Ast.Var.t, E': Env.t): CoreML.Var.t * Env.t = 
   let val x' = CoreML.Var.fromAst x
   in (x', Env.extendVar(E', x, x'))
   end

(* val traceElaboratePat =
 *    Trace.trace2("elaboratePat",
 * 		Apat.layout, Env.layout,
 * 		Layout.tuple2(Cpat.layout, Env.layout))
 *)

fun elaboratePat(p: Apat.t, E: Env.t, E'): Cpat.t * Env.t =
   let
      fun loop arg: Cpat.t * Env.t =
(*	 traceElaboratePat *)
	 (fn (p: Apat.t, E': Env.t) =>
	  case Apat.node p of
	    Apat.Wild => (Cpat.Wild, E')
	  | Apat.Var{name=x, ...} =>
	       (case Env.peekLongcon(E, Ast.Longvid.toLongcon x) of
		   SOME c => (Cpat.Con{con = c, arg = NONE}, E')
		 | NONE => (case Ast.Longvid.split x of
			       ([], x) => let val x = Ast.Vid.toVar x
					     val (x', E') = bind(x, E')
					 in (Cpat.Var x', E')
					 end
			     | _ => Error.error "longid in var pat"))
	  | Apat.Const c => (Cpat.Const c, E')
	  | Apat.Tuple ps => loopsCont(ps, E', Cpat.tuple)
	  | Apat.Record{items, flexible} =>
	       (* rules 36, 38, 39 and Appendix A, p.57 *)
	       let
		  val fps =
		     List.map
		     (items,
		      fn Apat.Item.Field fp => fp
		       | Apat.Item.Vid(vid, tyo, po) =>
			    (Field.String(Vid.toString vid),
			     let
				val p =
				   case po of
				      NONE => Apat.longvid(Longvid.short vid)
				    | SOME p =>
					 Apat.layered{fixop = Fixop.None,
						      var = Vid.toVar vid,
						      constraint = NONE,
						      pat = p}
			     in case tyo of
				NONE => p
			      | SOME ty => Apat.constraint(p, ty)
			     end))
		  val (fields, ps) = List.unzip fps
	       in loopsCont(ps, E', fn ps =>
			    Cpat.record
			    {record = Record.fromList(List.zip(fields, ps)),
			     flexible = flexible})
	       end
	  | Apat.List ps => loopsCont(ps, E', Cpat.list)
	  | Apat.FlatApp items => loop(Parse.parsePat(items, E), E')
	  | Apat.App(c, p) =>
	       let val (p, E') = loop(p, E')
	       in (Cpat.Con{con = Env.lookupLongcon(E, c),
			    arg = SOME p},
		   E')
	       end
	  | Apat.Constraint(p, t) =>
	       let val (p, E') = loop(p, E')
		  val t = Scheme.ty(elaborateType(t, Lookup.fromEnv E))
	       in (Cpat.Constraint(p, t), E')
	       end
	  | Apat.Layered{var=x, constraint, pat, ...} =>
	       let val (x', E') = bind(x, E')
		  val (pat, E') =
		     loop(case constraint of
			     NONE => pat
			   | SOME t => Apat.constraint(pat, t),
			  E')
	       in (Cpat.Layered(x', pat), E')
	       end) arg
      and loopsCont(ps: Apat.t list, E': Env.t, cont: Cpat.t list -> Cpat.t)
	: Cpat.t * Env.t =
	 let val (ps, E') = elaboratePats(ps, E, E')
	 in (cont ps, E')
	 end
   in loop(p, E')
   end
and elaboratePats(ps: Apat.t list, E, E'): Cpat.t list * Env.t =
   let val (ps, E') = List.fold(ps, ([], E'), fn (p, (ps, E')) =>
				let val (p, E') = elaboratePat(p, E, E')
				in (p :: ps, E')
				end)
   in (List.reverse ps, E')
   end

val elaboratePat =
   fn {pat, env, accum} => elaboratePat(pat, env, accum)

val elaboratePats =
   fn {pats, env, accum} => elaboratePats(pats, env, accum)

end
