(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateSigexp(S: ELABORATE_SIGEXP_STRUCTS): ELABORATE_SIGEXP = 
struct

open S

local open Ast
in structure Atype = Type
   structure DatBind = DatBind
   structure DatatypeRhs = DatatypeRhs
   structure Equation = Equation
   structure Longstrid = Longstrid
   structure Longtycon = Longtycon
   structure Sigexp = Sigexp
   structure Spec = Spec
   structure Strid = Strid
end

local open Basis
in structure Env = Env
end

local open Interface
in structure Status = Status
   structure TypeStr = FlexibleTypeStr
end

local open TypeStr
in structure Tycon = Tycon
   structure TypeFcn = TypeFcn
end

structure ElaborateAtoms =
   ElaborateAtoms(structure Ast = Ast
		  structure Decs =
		     struct
			type t = unit
			val empty = ()
			fun datatypes _ = ()
		     end
		  structure Env = Interface
		  structure TypeStr = TypeStr)
   
local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
   val elaborateTypeOpt = elaborateTypeOpt
   val elaborateClosedScheme = elaborateClosedScheme
   val elaborateCons = elaborateCons
   val elaborateDatatypeRhs = elaborateDatatypeRhs
end

structure Set = DisjointSet

(* rule 65 *)
fun elaborateSigexp(sigexp: Sigexp.t, B: Basis.t): Interface.t =
   case Sigexp.node sigexp of
      Sigexp.Var s => Basis.lookupSigid(B, s)
    | _ =>
	 let
	    local val E = Basis.toEnv B
	       val lk =
		  Lookup.fromFun
		  (fn c => TypeStr.fromStrict(Env.lookupLongtycon(E, c)))
	    in fun lookup I = Lookup.plusEnv(lk, I)
	    end

	    fun elaborateTypedescs(typedescs, I: Interface.t): Interface.t =
	       (* rule 80 *)
	       Interface.types
	       (List.map(typedescs, fn {tycon, tyvars, ty} =>
			 {name = tycon,
			  typeStr = 
			  case ty of
			     NONE => TypeStr.tycon(Tycon.newNoname())
			   | SOME ty =>
				TypeStr.def
				(elaborateClosedScheme
				 (tyvars, ty, lookup I))}))

	    fun elaborateSigexp(sigexp: Sigexp.t, I: Interface.t)
	      : Interface.t =
	       case Sigexp.node sigexp of
		  Sigexp.Spec spec => (* rule 62 *)
		     elaborateSpec(spec, I)
		| Sigexp.Var x => (* rule 63 *)
		     Interface.copy(Basis.lookupSigid(B, x))
		| Sigexp.Where(sigexp, wheres) => (* rule 64 *)
		     let val I' = elaborateSigexp(sigexp, I)
		     in Interface.wheres
			(I', List.map(wheres, fn {tyvars, longtycon, ty} =>
				      (longtycon,
				       TypeFcn.funn
				       (elaborateClosedScheme
					(tyvars, ty, lookup I))))) ;
			I'
		     end
		  
	    and elaborateSpec(spec: Ast.Spec.t, I: Interface.t): Interface.t =
	       case Spec.node spec of
		  Spec.Val xts => (* rules 68, 79 *)
		     Interface.vals
		     (List.map(xts, fn (x, t) =>
			       {name = Ast.Vid.fromVar x,
				status = Status.Var,
				scheme = elaborateType(t, lookup I)}))
		| Spec.Type typedescs => (* rule 69 *)
		     elaborateTypedescs(typedescs, I)
		| Spec.Eqtype typedescs => (* rule 70 *)
		     elaborateTypedescs(typedescs, I)
		| Spec.Datatype rhs => (* rules 71, 72 *)
		     #2(elaborateDatatypeRhs{rhs = rhs,
					     lookup = lookup I,
					     newTycon = fn _ => Tycon.newNoname()})
		| Spec.Exception cons => (* rule 73 *)
		     Interface.excons(elaborateCons{cons = cons, tyvars = [],
						    tycon = Tycon.exn,
						    lookup = lookup I,
						    resultType = Atype.exn})
		| Spec.Structure ss => (* rules 74, 84 *)
		     Interface.strs
		     (List.map(ss, fn (strid, sigexp) =>
			       {name = strid,
				interface = elaborateSigexp(sigexp, I)}))
		| Spec.IncludeSigexp sigexp => (* rule 75 *)
		     elaborateSigexp(sigexp, I)
		| Spec.IncludeSigids sigids => (* Appendix A, p.59 *)
		     elaborateSpec
		     (List.fold(sigids, Spec.make Spec.Empty, fn (sigid, spec) =>
				Spec.make
				(Spec.Seq(spec,
					  Spec.make
					  (Spec.IncludeSigexp
					   (Sigexp.var sigid))))),
		      I)
		| Spec.Empty => (* rule 76 *)
		     Interface.empty
		| Spec.Seq(s, s') => (* rule 77 *)
		     let val I' = elaborateSpec(s, I)
			val I'' = elaborateSpec(s', Interface.+(I, I'))
		     in Interface.+(I', I'')
		     end
		| Spec.Sharing{spec, equations} => (* rule 78 and section G.3.3 *)
		     let val I' = elaborateSpec(spec, I)
			fun share eqn =
			   case Equation.node eqn of
			      Equation.Type cs =>
			      (case cs of
				  [] => ()
				| c :: cs =>
				     List.foreach
				     (cs, fn c' =>
				      Interface.shareType(I', c, c')))
			    | Equation.Structure ss =>
			      let
				 fun loop ss =
				    case ss of
				       [] => ()
				     | s :: ss =>
					  (List.foreach
					   (ss, fn s' =>
					    Interface.share(I', s, s')) ;
					   loop ss)
			      in loop ss
			      end
		     in List.foreach(equations, share) ;
			I'
		     end
	 in elaborateSigexp(sigexp, Interface.empty)
	 end

(* val elaborateSigexp = 
 *    Trace.trace2("elaborateSigexp",
 * 		Sigexp.layout,
 * 		Layout.ignore (*Basis.layout*),
 * 		(*Layout.ignore*) Interface.layout)
 *    elaborateSigexp
 *)

end
