(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Interface(S: INTERFACE_STRUCTS): INTERFACE =
struct

open S

local open Ast
in structure Longtycon = Longtycon
   structure Longstrid = Longstrid
   structure Record = SortedRecord
   structure Strid = Strid
   structure Tyvar = Tyvar
end

local open CoreML
in structure Cscheme = Scheme
   structure Ctycon = Tycon
   structure Ctype = Type
end

structure CtypeStr = TypeStr
local open TypeStr
in structure CtypeFcn = TypeFcn
   structure Ccons = Cons
end

structure Set = DisjointSet

structure ShapeId = UniqueId()

structure Status =
   struct
      datatype t = Var | Con | Exn
      fun layout s =
	 Layout.str(case s of
		       Var => "Var"
		     | Con => "Con"
		     | Exn => "Exn")
   end

(* only needed for debugging *)
structure TyconId = IntUniqueId()
       
structure FlexibleTycon =
   struct
      structure TypeFcn =
	 struct
	    datatype t =
	       Tycon
	     | Fun
	     | Forced of CtypeFcn.t

	    fun layout f =
	       let open Layout
	       in case f of
		  Tycon => str "<flexible tycon>"
		| Fun => str "<flexible def>"
		| Forced f => paren(seq[str "forced ", CtypeFcn.layout f])
	       end

	    val strict =
	       fn Forced f => f
		| _ => Error.bug "impossible force of FlexibleTycon"
	 end
      
      datatype t = T of {copy: copy,
			 hasCons: bool,
			 typeFcn: TypeFcn.t,
			 id: TyconId.t} Set.t
      withtype copy = t option ref

      val equals = fn (T s, T s') => Set.equals(s, s')
	 
      fun layout(T s) =
	 let open Layout
	    val {hasCons, typeFcn, id, ...} = Set.value s
	 in record[("id", TyconId.layout id),
		   ("hasCons", Bool.layout hasCons),
		   ("typeFcn", TypeFcn.layout typeFcn)]
	 end

      fun setTypeFcn(T s, f: CtypeFcn.t): unit =
	 let val {copy, id, hasCons, ...} = Set.value s
	 in Set.setValue(s, {copy = copy, id = id, hasCons = hasCons,
			     typeFcn = TypeFcn.Forced f})
	 end

      fun new{hasCons: bool, typeFcn: TypeFcn.t}: t =
	 T(Set.singleton{copy = ref NONE,
			 hasCons = hasCons,
			 typeFcn = typeFcn,
			 id = TyconId.new()})

      fun newNoname() = new{hasCons = false, typeFcn = TypeFcn.Tycon}

      val bogus = newNoname()

      fun toTypeFcn(T s) = #typeFcn(Set.value s)

      val copies: copy list ref = ref []
	 
      fun copy(T s): t =
	 let val {copy, typeFcn, hasCons, ...} = Set.value s
	 in case !copy of
	    SOME c => c
	  | NONE => 
	       let val c = new{hasCons = hasCons,
			       typeFcn = typeFcn}
	       in ListRef.push(copies, copy) ;
		  copy := SOME c ;
		  c
	       end
	 end

      fun share(T s, reg, T s', reg'): unit =
	 let val {typeFcn, hasCons, id, ...} = Set.value s
	    val {typeFcn=typeFcn', hasCons=hasCons', ...} = Set.value s'
	    fun error reg =
	       Control.errorStr(reg, "type sharing on definitional type")
	 in case (typeFcn, typeFcn') of
	    (TypeFcn.Tycon, TypeFcn.Tycon) =>
	       (Set.union(s, s') ;
		Set.setValue(s, {copy = ref NONE, id = id,
				 hasCons = hasCons orelse hasCons',
				 typeFcn = TypeFcn.Tycon}))
	  | (TypeFcn.Fun, _) => error reg
	  | (_, TypeFcn.Fun) => error reg'
	  | _ => Error.bug "type sharing on Forced typeFcn"
	 end

      fun setAbbrev(T s, isTycon: bool, reg): unit =
	 let val {copy, id, hasCons, typeFcn} = Set.value s
	 in if hasCons andalso not isTycon
	       then Control.errorStr(reg, "type abbreviation on datatype")
	    else 
	       case typeFcn of
		  TypeFcn.Tycon =>
		     Set.setValue(s, {copy = copy,
				      id = id,
				      hasCons = hasCons,
				      typeFcn = if isTycon
						   then TypeFcn.Tycon
						else TypeFcn.Fun})
		| TypeFcn.Fun =>
		     Control.errorStr
		     (reg, "type abbreviation on type abbreviation")
		| TypeFcn.Forced _ =>
		     Control.errorStr
		     (reg, "type abbreviation on forced flexible tycon")
	 end

      fun strict(T s): CtypeFcn.t =
	 TypeFcn.strict(#typeFcn(Set.value s))
   end

structure Tycon =
   struct
      datatype t =
	 Flexible of FlexibleTycon.t
       | Rigid of Ctycon.t

      val arrow = Rigid Ctycon.arrow

      val newNoname = Flexible o FlexibleTycon.newNoname

      val toAst =
	 fn Flexible c =>
	      Ast.Tycon.fromString(Layout.toString(FlexibleTycon.layout c))
	  | Rigid c => Ctycon.toAst c
   
      val layout =
	 fn Flexible c => FlexibleTycon.layout c
	  | Rigid c => Ctycon.layout c

      fun copy(t: t): t =
	 case t of
	    Flexible c => Flexible(FlexibleTycon.copy c)
	  | Rigid _ => t

      fun strict(t: t): CtypeFcn.t =
	 case t of
	    Flexible c => FlexibleTycon.strict c
	  | Rigid c => CtypeFcn.Tycon c

      val fromStrict = Rigid

      val exn = fromStrict Ctycon.exn
   end

structure Type =
   struct
      datatype t =
	 Var of Tyvar.t
       | Con of Tycon.t * t list
       | Record of t Record.t

      fun hom(t, {var, con, record}) =
	 let
	    val rec loop =
	       fn Var a => var a
		| Con(c, ts) => con(c, List.map(ts, loop))
		| Record r => record(Record.map(r, loop))
	 in loop t
	 end
	       
      fun toAst t =
	 hom(t, {var = Ast.Type.var,
		 con = fn (c, ts) => Ast.Type.con(Tycon.toAst c, ts),
	         record = Ast.Type.record})

      val layout = Ast.Type.layout o toAst

(*      val traceStrict = Trace.trace("strict", layout, Ctype.layout) *)
	 
      fun strict t =
(*	 traceStrict*)
	 (fn t =>
	 case t of
	    Var a => Ctype.Var a
	  | Con(c, ts) => CtypeFcn.apply(Tycon.strict c, List.map(ts, strict))
	  | Record r => Ctype.Record(Record.map(r, strict))) t

      fun fromStrict t =
	 case t of
	    Ctype.Var a => Var a
	  | Ctype.Con(c, ts) => Con(Tycon.fromStrict c, List.map(ts, fromStrict))
	  | Ctype.Record r => Record(Record.map(r, fromStrict))
		 
      fun copy(t: t): t =
	 let
	    fun loop t =
	       case t of
		  Var _ => t
		| Con(c, ts) => Con(Tycon.copy c, List.map(ts, loop))
		| Record r => Record(Record.map(r, loop))
	 in loop t
	 end
	 
      val var = Var
      val record = Record
      fun arrow(t1, t2) = Con(Tycon.arrow, [t1, t2])

      local
	 structure Env = Env(structure Domain = Tyvar)
      in
	 fun substitute(t: t, sub: (Tyvar.t * t) list): t =
	    let val env = Env.fromList sub
	       fun loop t =
		  case t of
		     Var a => Env.lookup(env, a)
		   | Con(c, ts) => Con(c, List.map(ts, loop))
		   | Record r => Record(Record.map(r, loop))
	    in loop t
	    end
      end
   end

structure Scheme = GenericScheme(structure Tyvar = Tyvar
				 structure Type = Type)
structure Scheme =
   struct
      fun strict(Scheme.T{tyvars, ty}) =
	 Cscheme.T{tyvars = tyvars,
		   ty = Type.strict ty}

(*      val strict = Trace.trace("strict", Scheme.layout, Cscheme.layout) strict *)

      fun fromStrict(Cscheme.T{tyvars, ty}) =
	 Scheme.T{tyvars = tyvars, ty = Type.fromStrict ty}

      open Scheme

      fun copy(T{tyvars, ty}): t =
	 T{tyvars = tyvars, ty = Type.copy ty}
   end

structure TypeStr = TypeStr(structure Ast = Ast
			    structure Tycon = Tycon
			    structure Con = CoreML.Con
			    structure Type = Type
			    structure Scheme = Scheme)

structure TypeFcn =
   struct
      open TypeStr.TypeFcn

      fun copy(f: t): t =
	 case f of
	    Tycon c => Tycon(Tycon.copy c)
	  | Fun s => Fun(Scheme.copy s)

      fun strict(f: t) =
	 case f of
	    Tycon c => Tycon.strict c
	  | Fun s => CtypeFcn.Fun(Scheme.strict s)

      fun fromStrict f: t =
	 case f of
	    CtypeFcn.Tycon c => Tycon(Tycon.fromStrict c)
	  | CtypeFcn.Fun s => Fun(Scheme.fromStrict s)
   end

structure Cons =
   struct
      open TypeStr.Cons

      fun copy(T l): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.copy scheme}))

      fun strict(T l): Ccons.t =
	 Ccons.T(List.map(l, fn {name, con, scheme} =>
			  {name = name, con = con,
			   scheme = Scheme.strict scheme}))

      fun fromStrict(Ccons.T l): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.fromStrict scheme}))
end

structure TypeStr =
   struct
      fun toFlexible(s: TypeStr.t, reg, msg): FlexibleTycon.t =
	 case TypeStr.typeFcn s of
	    TypeFcn.Tycon(Tycon.Flexible c) => c
	  | _ => (Control.errorStr(reg, msg) ;
		  FlexibleTycon.bogus)

      fun copy(s: TypeStr.t): TypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in TypeStr.make(TypeFcn.copy f, Cons.copy cs)
	 end

      fun strict(s: TypeStr.t): CtypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in CtypeStr.make(TypeFcn.strict f, Cons.strict cs)
	 end

(*      val strict = Trace.trace("strict", TypeStr.layout, CtypeStr.layout) strict *)

      fun fromStrict(s: CtypeStr.t): TypeStr.t =
	 let val (f, cs) = CtypeStr.dest s
	 in TypeStr.make(TypeFcn.fromStrict f, Cons.fromStrict cs)
	 end
			 
      open TypeStr

      fun share(s: t, reg, s': t, reg'): unit =
	 let val msg = "type sharing on rigid tycon"
	 in FlexibleTycon.share(toFlexible(s, reg, msg), reg,
				toFlexible(s', reg', msg), reg')
	 end
   end
structure FlexibleTypeStr = TypeStr
	   
(*---------------------------------------------------*)
(*                   Main Datatype                   *)
(*---------------------------------------------------*)
(* Invariant: only ever union two envs if they have the same shape. *)
(* The shape of interface is the set of longtycons that are accessible in it. *)

datatype t = T of {copy: copy,
		   shapeId: ShapeId.t,
		   elements: element list,
		   wheres: (FlexibleTycon.t * TypeFcn.t) list ref} Set.t
and element =
   Val of {name: Ast.Vid.t,
	   scheme: Scheme.t,
	   status: Status.t}
  | Type of {name: Ast.Tycon.t,
	     typeStr: TypeStr.t}
  | Str of {name: Ast.Strid.t,
	    interface: t}
withtype copy = t option ref

type interface = t

fun equals(T s, T s') = Set.equals(s, s')

local open Layout
in
   fun layout(T s) =
      let val {elements, wheres, ...} = Set.value s
      in record[("wheres", list(List.map(!wheres, fn (c, f) =>
					 tuple[FlexibleTycon.layout c,
					       TypeFcn.layout f]))),
		("elements",
		 list(List.map(elements, layoutElement)))]
      end
   and layoutElement(e: element) =
      let
	 val (lhs, rhs) =
	    case e of
	       Val{name, scheme, status} =>
		  (Ast.Vid.layout name,
		   tuple[Status.layout status,
			 Scheme.layout scheme])
	     | Type{name, typeStr} =>
		  (Ast.Tycon.layout name,
		   TypeStr.layout typeStr)
	     | Str{name, interface} =>
		  (Ast.Strid.layout name, layout interface)
      in seq[lhs, str " -> ", rhs]
      end
end

(*val traceCheckType = Trace.trace("checkType", Type.layout, Unit.layout)
 *val traceCheckInterface = Trace.trace("checkInterface", layout, Unit.layout)
 *)
fun explicit elements: t =
   T(Set.singleton{copy = ref NONE,
		   shapeId = ShapeId.new(),
		   elements = elements,
		   wheres = ref []})

val empty = explicit []

val bogus = empty

fun vals l = explicit(List.map(l, Val))
fun strs l = explicit(List.map(l, Str))
fun types l = explicit(List.map(l, Type))
   
local
   fun make status (Cons.T l) =
      explicit(List.map(l, fn {name, scheme, ...} =>
			Val{name = Ast.Vid.fromCon name,
			    scheme = scheme,
			    status = status}))
in val cons = make Status.Con
   val excons = make Status.Exn
end

fun elements(T s): element list = #elements(Set.value s)
fun shapeId(T s) = #shapeId(Set.value s)
   
fun extendTycon(I, tycon, typeStr) =
   explicit(elements I @ [Type{name = tycon, typeStr = typeStr}])

val op + = fn (I, I') => explicit(elements I @ elements I')

fun peekTyconElements(elements: element list, tycon): TypeStr.t option =
   case List.peek(elements,
		  fn Type{name, ...} => Ast.Tycon.equals(tycon,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Type{typeStr, ...}) => SOME typeStr
    | _ => Error.bug "peekTyconElements"
	 
fun peekStridElements(elements, strid): t option =
   case List.peek(elements,
		  fn Str{name, ...} => Strid.equals(strid,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Str{interface, ...}) => SOME interface
    | _ => Error.bug "peekStridElements"

fun peekStrid(I: t, strid: Ast.Strid.t): t option =
   peekStridElements(elements I, strid)

fun peekStrids(I: t, strids: Ast.Strid.t list): t option =
   case strids of
      [] => SOME I
    | strid :: strids =>
	 case peekStrid(I, strid) of
	    NONE => NONE
	  | SOME I => peekStrids(I, strids)

fun lookupLongstrid(I: t, s: Longstrid.t): t =
   let val (strids, strid) = Longstrid.split s
   in case peekStrids(I, strids @ [strid]) of
      NONE => (Longstrid.unbound s ; bogus)
    | SOME I => I
   end

fun peekLongtycon(I: t, c: Longtycon.t): TypeStr.t option =
   let val (strids, c) = Longtycon.split c
   in case peekStrids(I, strids) of
      NONE => NONE
    | SOME I => peekTyconElements(elements I, c)
   end

fun lookupLongtycon(I: t, c: Longtycon.t): TypeStr.t =
   case peekLongtycon(I, c) of
      SOME typeStr => typeStr
    | NONE => (Longtycon.unbound c ;
	       TypeStr.bogus)

fun shareType(I: t, c: Longtycon.t, c': Longtycon.t) =
   TypeStr.share(lookupLongtycon(I, c), Longtycon.region c,
		 lookupLongtycon(I, c'), Longtycon.region c')

fun sameShape(m, m') = ShapeId.equals(shapeId m, shapeId m')

fun share(I as T s, reg, I' as T s', reg'): unit = 
   if Set.equals(s, s') then ()
   else
      if sameShape(I, I')
	 then
	    let
	       fun loop(T s, T s'): unit =
		  if Set.equals(s, s') then ()
		  else 
		     let
			val {elements=es, ...} = Set.value s
			val {elements=es', ...} = Set.value s'
		     in Set.union(s, s') ;
			List.foreach2
			(es, es',
			 fn (Type{typeStr=s, ...}, Type{typeStr=s', ...}) =>
			    TypeStr.share(s, reg, s', reg')
			  | (Str{interface=I, ...}, Str{interface=I', ...}) =>
			       loop(I, I')
			  | _ => ())
		     end
	    in loop(I, I')
	    end
      else (* different shapes -- need to share pointwise *)
	 let
	    val es = elements I
	    val es' = elements I'
	 in List.foreach
	    (es,
	     fn Type{name, typeStr = s} =>
	           (case peekTyconElements(es',name) of
		       NONE => ()
		     | SOME s' => TypeStr.share(s, reg, s', reg'))
	      | Str{name, interface = I} =>
		   (case peekStridElements(es', name) of
		       NONE => ()
		     | SOME I' => share(I, reg, I', reg'))
	      | _ => ())
	 end

val share =
   fn (m, s: Longstrid.t, s': Longstrid.t) =>
   share(lookupLongstrid(m, s),
	 Longstrid.region s,
	 lookupLongstrid(m, s'),
	 Longstrid.region s')

fun wheres(I as T s, l: (Longtycon.t * TypeFcn.t) list): unit =
   let val {wheres, ...} = Set.value s
   in List.foreach(l, fn (c, f) =>
		   let
		      val reg = Longtycon.region c
		      val flex =
			 TypeStr.toFlexible
			 (lookupLongtycon(I, c), reg, "where type on rigid tycon")
		   in FlexibleTycon.setAbbrev(flex, TypeFcn.isTycon f, reg) ;
		      ListRef.push(wheres, (flex, f))
		   end)
   end

structure Element =
   struct
      type interface = t
      datatype t =
	 Val of {name: Ast.Vid.t,
		 scheme: Cscheme.t,
		 status: Status.t}
       | Type of {name: Ast.Tycon.t,
		  typeStr: CtypeStr.t}
       | Str of {name: Ast.Strid.t,
		 interface: interface}
   end

fun copyAndRealize(I: t, getTypeFcnOpt): t =
   let
      (* keep track of all nodes that have forward pointers to copies, so
       that we can gc them when done *)
      val copies: copy list ref = ref []
	 
      fun loop(T s, strids: Ast.Strid.t list): t =
	 let val {copy, shapeId, elements, wheres, ...} = Set.value s
	 in case !copy of
	    SOME I => I
	  | NONE =>
	       let
		  val elements =
		     List.map
		     (elements,
		      fn Val{name, scheme, status} =>
		      Val{name = name, status = status,
			  scheme = Scheme.copy scheme}
		       | Type{name, typeStr} =>
			    let val typeStr = TypeStr.copy typeStr
			    in (case (TypeStr.typeFcn typeStr, getTypeFcnOpt) of
				   (TypeFcn.Tycon(Tycon.Flexible c), SOME f) =>
				      FlexibleTycon.setTypeFcn
				      (c, f(Longtycon.long(strids,name)))
				 | _ => ()) ;
			       Type{name = name,
				    typeStr = typeStr}
			    end
		       | Str{name, interface} =>
			    Str{name = name,
				interface = loop(interface, strids @ [name])})
		     
		  val wheres =
		     List.map(!wheres, fn (c, f) =>
			      let val c = FlexibleTycon.copy c
				 val f = TypeFcn.copy f
			      in (case getTypeFcnOpt of
				     NONE => ()
				   | SOME _ =>
					FlexibleTycon.setTypeFcn
					(c, TypeFcn.strict f)) ;
				 (c, f)
			      end)

		  val I = T(Set.singleton{copy = ref NONE,
					  shapeId = shapeId,
					  elements = elements,
					  wheres = ref wheres})
				     
	       in ListRef.push(copies, copy) ;
		  copy := SOME I ;
		  I
	       end
	 end
      
      val I = loop(I, [])

      fun clear copies =
	 (List.foreach(!copies, fn copy => copy := NONE) ;
	  copies := [])
	 
   in clear copies ;
      clear FlexibleTycon.copies ;
      I
   end

fun copy I = copyAndRealize(I, NONE)
fun realize(I, f) = copyAndRealize(I, SOME f)

(*val realize = Trace.trace2("realize", layout, Layout.ignore, layout) realize *)

val 'a fold =
   fn (T s, b: 'a, f: Element.t * 'a -> 'a) =>
   let val {elements, ...} = Set.value s
   in List.fold
      (elements, b, fn (elt, b) =>
       let
	  val elt =
	     case elt of
		Val{name, scheme, status} =>
		   Element.Val{name = name, status = status,
			       scheme = Scheme.strict scheme}
	      | Str r => Element.Str r
	      | Type{name, typeStr} =>
		   Element.Type{name = name,
				typeStr = TypeStr.strict typeStr}
       in f(elt, b)
       end)
   end

structure TypeStr = CtypeStr

end
