(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Interface(S: INTERFACE_STRUCTS): INTERFACE =
struct

open S

local open Ast
in structure Longtycon = Longtycon
   structure Longstrid = Longstrid
   structure Record = Record
   structure Strid = Strid
   structure Tyvar = Tyvar
end

local open CoreML
in structure Cscheme = Scheme
   structure Ctycon = Tycon
   structure Ctype = Type
end

structure CtypeStr = TypeStr
local open TypeStr
in structure CtypeFcn = TypeFcn
   structure Ccons = Cons
end

structure Set = DisjointSet

structure MapId = UniqueId()
structure ShapeId = UniqueId()

structure Status =
   struct
      datatype t = Var | Con | Exn
      fun layout s =
	 Layout.str(case s of
		       Var => "Var"
		     | Con => "Con"
		     | Exn => "Exn")
   end

fun 'a addCopy(copiess: 'a list ref list ref,
	       copies: 'a list ref,
	       copy: 'a) =
   (case !copies of
       [] => ListRef.push(copiess, copies)
     | _ => () ;
    ListRef.push(copies, copy))

(* only needed for debugging *)
structure TyconId = IntUniqueId()
       
structure FlexibleTycon =
   struct
      structure TypeFcn =
	 struct
	    datatype t =
	       Tycon
	     | Fun
	     | Forced of CtypeFcn.t

	    fun layout f =
	       let open Layout
	       in case f of
		  Tycon => str "<flexible tycon>"
		| Fun => str "<flexible def>"
		| Forced f => paren(seq[str "forced ", CtypeFcn.layout f])
	       end

	    val strict =
	       fn Forced f => f
		| _ => Error.bug "impossible force of FlexibleTycon"
	 end
      
      datatype t = T of {copies: copies,
			 hasCons: bool,
			 typeFcn: TypeFcn.t,
			 id: TyconId.t} Set.t
      withtype copies = (MapId.t * t) list ref

      val equals = fn (T s, T s') => Set.equals(s, s')
	 
      fun layout(T s) =
	 let open Layout
	    val {hasCons, typeFcn, id, ...} = Set.value s
	 in record[("id", TyconId.layout id),
		   ("hasCons", Bool.layout hasCons),
		   ("typeFcn", TypeFcn.layout typeFcn)]
	 end

      fun setTypeFcn(T s, f: CtypeFcn.t): unit =
	 let val {copies, id, hasCons, ...} = Set.value s
	 in Set.setValue(s, {copies = copies, id = id, hasCons = hasCons,
			     typeFcn = TypeFcn.Forced f})
	 end

      fun new{hasCons: bool, typeFcn: TypeFcn.t}: t =
	 T(Set.singleton{copies = ref [],
			 hasCons = hasCons,
			 typeFcn = typeFcn,
			 id = TyconId.new()})

      fun newNoname() = new{hasCons = false, typeFcn = TypeFcn.Tycon}

      fun toTypeFcn(T s) = #typeFcn(Set.value s)

      val copiess: copies list ref = ref []
	 
      fun copy(T s, id: MapId.t): t =
	 let val {copies, typeFcn, hasCons, ...} = Set.value s
	 in case List.peek(!copies, fn (id', _) => MapId.equals(id, id')) of
	    SOME(_, i) => i
	  | NONE => 
	       let val c = new{hasCons = hasCons,
			       typeFcn = typeFcn}
	       in addCopy(copiess, copies, (id, c)) ;
		  c
	       end
	 end

      fun share(T s, T s'): unit =
	 let val {typeFcn, hasCons, id, ...} = Set.value s
	    val {typeFcn=typeFcn', hasCons=hasCons', ...} = Set.value s'
	 in case (typeFcn, typeFcn') of
	    (TypeFcn.Tycon, TypeFcn.Tycon) =>
	       (Set.union(s, s') ;
		Set.setValue(s, {copies = ref [], id = id,
				 hasCons = hasCons orelse hasCons',
				 typeFcn = TypeFcn.Tycon}))
	  | _ => Error.error "type sharing on definitional type"
	 end

      fun setAbbrev(T s, isTycon: bool): unit =
	 let val {copies, id, hasCons, typeFcn} = Set.value s
	 in if hasCons andalso not isTycon
	       then Error.error "type abbreviation on datatype"
	    else 
	       case typeFcn of
		  TypeFcn.Tycon =>
		     Set.setValue(s, {copies = copies, id = id,
				      hasCons = hasCons,
				      typeFcn = if isTycon
						   then TypeFcn.Tycon
						else TypeFcn.Fun})
		| TypeFcn.Fun =>
		     Error.error "type abbreviation on type abbreviation"
		| TypeFcn.Forced _ =>
		     Error.bug "type abbreviation on forced flexible tycon"
	 end

      fun strict(T s): CtypeFcn.t =
	 TypeFcn.strict(#typeFcn(Set.value s))
   end

structure Tycon =
   struct
      datatype t =
	 Flexible of FlexibleTycon.t
       | Rigid of Ctycon.t

      val arrow = Rigid Ctycon.arrow

      val newNoname = Flexible o FlexibleTycon.newNoname

      val toAst =
	 fn Flexible c =>
	      Ast.Tycon.fromString(Layout.toString(FlexibleTycon.layout c))
	  | Rigid c => Ctycon.toAst c
   
      val layout =
	 fn Flexible c => FlexibleTycon.layout c
	  | Rigid c => Ctycon.layout c

      fun copy(t: t, id: MapId.t): t =
	 case t of
	    Flexible c => Flexible(FlexibleTycon.copy(c, id))
	  | Rigid _ => t

      fun strict(t: t): CtypeFcn.t =
	 case t of
	    Flexible c => FlexibleTycon.strict c
	  | Rigid c => CtypeFcn.Tycon c

      val fromStrict = Rigid

      val exn = fromStrict Ctycon.exn
   end

structure Type =
   struct
      datatype t =
	 Var of Tyvar.t
       | Con of Tycon.t * t list
       | Record of t Record.t

      fun toAst t =
	 case t of
	    Var a => Ast.Type.var a
	  | Con(c, ts) => Ast.Type.con(Tycon.toAst c, List.map(ts, toAst))
	  | Record r => Ast.Type.record(Record.map(r, toAst))

      val layout = Ast.Type.layout o toAst

(*      val traceStrict = Trace.trace("strict", layout, Ctype.layout) *)
	 
      fun strict t =
(*	 traceStrict *)
	 (fn t =>
	 case t of
	    Var a => Ctype.Var a
	  | Con(c, ts) => CtypeFcn.apply(Tycon.strict c, List.map(ts, strict))
	  | Record r => Ctype.Record(Record.map(r, strict))) t

      fun fromStrict t =
	 case t of
	    Ctype.Var a => Var a
	  | Ctype.Con(c, ts) => Con(Tycon.fromStrict c, List.map(ts, fromStrict))
	  | Ctype.Record r => Record(Record.map(r, fromStrict))
		 
      fun copy(t: t, id: MapId.t): t =
	 let
	    fun loop t =
	       case t of
		  Var _ => t
		| Con(c, ts) => Con(Tycon.copy(c, id), List.map(ts, loop))
		| Record r => Record(Record.map(r, loop))
	 in loop t
	 end
	 
      val var = Var
      val record = Record
      fun arrow(t1, t2) = Con(Tycon.arrow, [t1, t2])

      local
	 structure Env = Env(structure Domain = Tyvar)
      in
	 fun substitute(t: t, sub: (Tyvar.t * t) list): t =
	    let val env = Env.fromList sub
	       fun loop t =
		  case t of
		     Var a => Env.lookup(env, a)
		   | Con(c, ts) => Con(c, List.map(ts, loop))
		   | Record r => Record(Record.map(r, loop))
	    in loop t
	    end
      end
   end

structure Scheme = GenericScheme(structure Tyvar = Tyvar
				 structure Type = Type)
structure Scheme =
   struct
      fun strict(Scheme.T{tyvars, ty}) =
	 Cscheme.T{tyvars = tyvars,
		   ty = Type.strict ty}

(*      val strict = Trace.trace("strict", Scheme.layout, Cscheme.layout) strict *)

      fun fromStrict(Cscheme.T{tyvars, ty}) =
	 Scheme.T{tyvars = tyvars, ty = Type.fromStrict ty}

      open Scheme

      fun copy(T{tyvars, ty}, id: MapId.t): t =
	 T{tyvars = tyvars,
	   ty = Type.copy(ty, id)}
   end

structure TypeStr = TypeStr(structure Ast = Ast
			    structure Tycon = Tycon
			    structure Con = CoreML.Con
			    structure Type = Type
			    structure Scheme = Scheme)

structure TypeFcn =
   struct
      open TypeStr.TypeFcn

      fun copy(f: t, id: MapId.t): t =
	 case f of
	    Tycon c => Tycon(Tycon.copy(c, id))
	  | Fun s => Fun(Scheme.copy(s, id))

      fun strict(f: t) =
	 case f of
	    Tycon c => Tycon.strict c
	  | Fun s => CtypeFcn.Fun(Scheme.strict s)

      fun fromStrict f: t =
	 case f of
	    CtypeFcn.Tycon c => Tycon(Tycon.fromStrict c)
	  | CtypeFcn.Fun s => Fun(Scheme.fromStrict s)
   end

structure Cons =
   struct
      open TypeStr.Cons

      fun copy(T l, id: MapId.t): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.copy(scheme, id)}))

      fun strict(T l): Ccons.t =
	 Ccons.T(List.map(l, fn {name, con, scheme} =>
			  {name = name, con = con,
			   scheme = Scheme.strict scheme}))

      fun fromStrict(Ccons.T l): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.fromStrict scheme}))
end

structure TypeStr =
   struct
      fun toFlexible(s: TypeStr.t): FlexibleTycon.t =
	 case TypeStr.typeFcn s of
	    TypeFcn.Tycon(Tycon.Flexible c) => c
	  | _ => Error.error "TypeStr.toFlexible"

      fun copy(s: TypeStr.t, id: MapId.t): TypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in TypeStr.make(TypeFcn.copy(f, id), Cons.copy(cs, id))
	 end

      fun strict(s: TypeStr.t): CtypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in CtypeStr.make(TypeFcn.strict f, Cons.strict cs)
	 end

(*      val strict = Trace.trace("strict", TypeStr.layout, CtypeStr.layout) strict *)

      fun fromStrict(s: CtypeStr.t): TypeStr.t =
	 let val (f, cs) = CtypeStr.dest s
	 in TypeStr.make(TypeFcn.fromStrict f, Cons.fromStrict cs)
	 end
			 
      open TypeStr

      fun share(s: t, s': t): unit =
	 FlexibleTycon.share(toFlexible s, toFlexible s')
   end
structure LazyTypeStr = TypeStr
	   
(*---------------------------------------------------*)
(*                   Main Datatype                   *)
(*---------------------------------------------------*)
(* Invariant: only ever union two envs if they have the same shape. *)
(* The shape of interface is the set of longtycons that are accessible in it. *)

datatype t = T of promise Set.t
and promise =
   Delayed of {copyId: MapId.t,
	       original: env} (* back pointer *)
  | Forced of env
and env =
   Env of {visited: bool ref, (* used for depth-first-search *)
	   copies: copies, (* forward pointers *)
	   shapeId: ShapeId.t,
	   elements: element list,
	   wheres: (FlexibleTycon.t * TypeFcn.t) list ref}
and element =
   Val of {name: Ast.Vid.t,
	   scheme: Scheme.t,
	   status: Status.t}
  | Type of {name: Ast.Tycon.t,
	     typeStr: TypeStr.t}
  | Str of {name: Ast.Strid.t,
	    interface: t}
withtype copies = (MapId.t * t) list ref

type interface = t

val equals = fn (T s, T s') => Set.equals(s, s')
   
(* keep track of all nodes that have forward pointers to copies, so
 that we can gc them when done *)
val copiess: copies list ref = ref []

local
   fun clear copiess =
      (List.foreach(!copiess, fn copies => copies := []) ;
       copiess := [])
in
   fun copyingDone() =
      (clear copiess ; clear FlexibleTycon.copiess)
end

fun copyId(T s, id): t =
   case Set.value s of
      Forced(env as Env{visited, copies, ...}) =>
	 (case List.peek(!copies, fn (id', _) => MapId.equals(id, id')) of
	     SOME(_, m) => m
	   | NONE => 
		let val m = T(Set.singleton(Delayed{copyId = id,
						    original = env}))
		in addCopy(copiess, copies, (id, m)) ; m
		end)
    | _ => Error.bug "must copy a forced"

fun copy(I:t): t = copyId(I, MapId.new())

fun force(T s): env =
   case Set.value s of
      Forced e => e
    | Delayed{copyId = id, original = Env{copies, shapeId, elements, wheres, ...}} =>
	 let 
	    val e =
	       Env{visited = ref false,
		   copies = ref [],
		   shapeId = shapeId,
		   elements =
		   List.map
		   (elements,
		    fn Val{name, scheme, status} =>
		       Val{name = name, status = status,
			   scheme = Scheme.copy(scheme, id)}
		     | Type{name, typeStr} =>
			  Type{name = name,
			       typeStr = TypeStr.copy(typeStr, id)}
		     | Str{name, interface} =>
			  Str{name = name,
			      interface = copyId(interface, id)}),
		   wheres =
		   ref(List.map(!wheres, fn (c, f) =>
				(FlexibleTycon.copy(c, id),
				 TypeFcn.copy(f, id))))}
	 in Set.setValue(s, Forced e) ; e
	 end

local open Layout
in
   fun layout I =
      let val Env{elements, wheres, ...} = force I
      in record[("wheres", list(List.map(!wheres, fn (c, f) =>
					 tuple[FlexibleTycon.layout c,
					       TypeFcn.layout f]))),
		("elements",
		 list(List.map(elements, layoutElement)))]
      end
   and layoutElement(e: element) =
      let
	 val (lhs, rhs) =
	    case e of
	       Val{name, scheme, status} =>
		  (Ast.Vid.layout name,
		   tuple[Status.layout status,
			 Scheme.layout scheme])
	     | Type{name, typeStr} =>
		  (Ast.Tycon.layout name,
		   TypeStr.layout typeStr)
	     | Str{name, interface} =>
		  (Ast.Strid.layout name, layout interface)
      in seq[lhs, str " -> ", rhs]
      end
end

(* val traceCheckType = Trace.trace("checkType", Type.layout, Unit.layout)
 * val traceCheckInterface = Trace.trace("checkInterface", layout, Unit.layout)
 *)

fun completelyForce(I: t): unit = 
   let
      val visitedRefs = ref []
      val flexTycons = ref []
      fun addTycon c = ListRef.push(flexTycons, c)
      fun checkTycon c =
	 case c of
	    Tycon.Flexible c =>
	       (case List.peek(!flexTycons, fn c' => FlexibleTycon.equals(c, c')) of
		   NONE => Error.bug "unbound flexible tycon"
		 | SOME _ => ())
	  | Tycon.Rigid _ => ()
      fun checkType t =
(*	 traceCheckType *)
	 (fn t =>
	 case t of
	    Type.Var _ => ()
	  | Type.Con(c, ts) => (checkTycon c ; List.foreach(ts, checkType))
	  | Type.Record r => Record.foreach(r, checkType)) t
      fun checkScheme(Scheme.T{ty, ...}) = checkType ty
      fun checkCons(Cons.T l) =
	 List.foreach(l, fn {scheme, ...} => checkScheme scheme)
      fun checkTypeFcn f =
	 case f of
	    TypeFcn.Tycon c => checkTycon c
	  | TypeFcn.Fun s => checkScheme s
      fun checkTypeStr s =
	 let val (f, cs) = TypeStr.dest s
	 in checkTypeFcn f ; checkCons cs
	 end
      fun loop(I:t): unit =
(*	 traceCheckInterface *)
	 (fn I =>
	 let val env as Env{visited, elements, wheres, ...} = force I
	 in if not(!visited)
	       then ((* only need the following for debugging *)
		     (*List.foreach(!wheres, fn (c, f) => (addTycon c ;
							checkTypeFcn f)) ;
		     List.foreach
		     (elements,
		      fn Type{typeStr, ...} =>
		           (case TypeStr.typeFcn typeStr of
			       TypeFcn.Tycon(Tycon.Flexible c) => addTycon c
			     | _ => ())
		       | _ => ()) ; *)
		      (* only need the above for debugging *)
		     List.foreach
		     (elements,
		      fn Str{interface, ...} => loop interface
		       | Val{scheme, ...} => () (*checkScheme scheme*)
		       | Type{typeStr, ...} => () (*checkTypeStr typeStr*)) ;
		     ListRef.push(visitedRefs, visited) ;
		     visited := true)
	    else ()
	 end) I
   in loop I ;
      List.foreach(!visitedRefs, fn r => r := false) ;
      copyingDone()
   end

(* val completelyForce = Trace.trace("completelyForce", layout, Unit.layout)
 *    completelyForce
 *)

fun explicit elements: t =
   T(Set.singleton(Forced(Env{visited = ref false,
			      copies = ref [],
			      shapeId = ShapeId.new(),
			      elements = elements,
			      wheres = ref []})))

val empty = explicit []

fun vals l = explicit(List.map(l, Val))
fun strs l = explicit(List.map(l, Str))
fun types l = explicit(List.map(l, Type))
   
local
   fun make status (Cons.T l) =
      explicit(List.map(l, fn {name, scheme, ...} =>
			Val{name = Ast.Vid.fromCon name,
			    scheme = scheme,
			    status = status}))
in val cons = make Status.Con
   val excons = make Status.Exn
end

fun elements(I: t): element list =
   let val Env{elements, ...} = force I
   in elements
   end
   
fun extendTycon(I, tycon, typeStr) =
   explicit(elements I @ [Type{name = tycon, typeStr = typeStr}])

val op + = fn (I, I') => explicit(elements I @ elements I')

fun peekTyconElements(elements: element list, tycon): TypeStr.t option =
   case List.peek(elements,
		  fn Type{name, ...} => Ast.Tycon.equals(tycon,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Type{typeStr, ...}) => SOME typeStr
    | _ => Error.bug "peekTyconElements"
	 
fun peekStridElements(elements, strid): t option =
   case List.peek(elements,
		  fn Str{name, ...} => Strid.equals(strid,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Str{interface, ...}) => SOME interface
    | _ => Error.bug "peekStridElements"

fun peekStrid(I: t, strid: Ast.Strid.t): t option =
   peekStridElements(elements I, strid)

fun peekStrids(I: t, strids: Ast.Strid.t list): t option =
   case strids of
      [] => SOME I
    | strid :: strids =>
	 case peekStrid(I, strid) of
	    NONE => NONE
	  | SOME I => peekStrids(I, strids)

fun lookupLongstrid(I: t, s: Longstrid.t): t =
   let val (strids, strid) = Longstrid.split s
   in case peekStrids(I, strids @ [strid]) of
      NONE => Error.error "lookupLongstrid failed"
    | SOME I => I
   end

fun peekLongtycon(I: t, c: Longtycon.t): TypeStr.t option =
   let val (strids, c) = Longtycon.split c
   in case peekStrids(I, strids) of
      NONE => NONE
    | SOME I => peekTyconElements(elements I, c)
   end

fun lookupLongtycon(I: t, c: Longtycon.t): TypeStr.t =
   case peekLongtycon(I, c) of
      SOME typeStr => typeStr
    | NONE => Error.error "lookupLongtycon failed"

fun shareType(I: t, c: Longtycon.t, c': Longtycon.t) =
   TypeStr.share(lookupLongtycon(I, c), lookupLongtycon(I, c'))

fun shapeId(T s) =
   let val Env{shapeId, ...} = (case Set.value s of
				  Delayed{original, ...} => original
				| Forced env => env)
   in shapeId
   end

fun sameShape(m, m') = ShapeId.equals(shapeId m, shapeId m')
   
fun share(I as T s, I' as T s'): unit = 
   if Set.equals(s, s') then ()
   else
      if sameShape(I, I')
	 then
	    let
	       fun loop(I as T s, I' as T s'): unit =
		  if Set.equals(s, s') then ()
		  else 
		     let val Env{elements=es, ...} = force I
			val Env{elements=es', ...} = force I'
		     in Set.union(s, s') ;
			List.foreach2
			(es, es',
			 fn (Type{typeStr=s, ...}, Type{typeStr=s', ...}) =>
			    TypeStr.share(s, s')
			  | (Str{interface=I, ...}, Str{interface=I', ...}) =>
			       loop(I, I')
			  | _ => ())
		     end
	    in loop(I, I')
	    end
      else (* different shapes -- need to share pointwise *)
	 let val Env{elements=es, ...} = force I
	    val Env{elements=es', ...} = force I'
	 in List.foreach
	    (es,
	     fn Type{name, typeStr=s} =>
	           (case peekTyconElements(es',name) of
		       NONE => ()
		     | SOME s' => TypeStr.share(s, s'))
	      | Str{name, interface=I} =>
		   (case peekStridElements(es', name) of
		       NONE => ()
		     | SOME I' => share(I, I'))
	      | _ => ())
	 end

val share =
   fn (m, s: Longstrid.t, s': Longstrid.t) =>
   share(lookupLongstrid(m, s), lookupLongstrid(m, s'))

fun wheres(I: t, l: (Longtycon.t * TypeFcn.t) list): unit =
   let val Env{wheres, ...} = force I
   in List.foreach(l, fn (c, f) =>
		   let val flex = TypeStr.toFlexible(lookupLongtycon(I, c))
		   in FlexibleTycon.setAbbrev(flex, TypeFcn.isTycon f) ;
		      ListRef.push(wheres, (flex, f))
		   end)
   end

structure Element =
   struct
      type interface = t
      datatype t =
	 Val of {name: Ast.Vid.t,
		 scheme: Cscheme.t,
		 status: Status.t}
       | Type of {name: Ast.Tycon.t,
		  typeStr: CtypeStr.t}
       | Str of {name: Ast.Strid.t,
		 interface: interface}
   end

fun realize(I: t, getTypeFcn): t =
   let val I = copy I
      val visitedRefs = ref []
      fun loop(I:t, strids: Ast.Strid.t list): unit =
	 let val Env{visited, elements, wheres, ...} = force I
	 in if !visited then ()
	    else
	       (List.foreach(!wheres, fn (c, f) =>
			     FlexibleTycon.setTypeFcn(c, TypeFcn.strict f)) ;
		List.foreach(elements,
			     fn Type{name, typeStr} =>
			     (case TypeStr.typeFcn typeStr of
				 TypeFcn.Tycon(Tycon.Flexible c) =>
				    FlexibleTycon.setTypeFcn
				    (c, getTypeFcn(Longtycon.make(strids,name)))
			       | _ => ())
			      | Str{name, interface=I} => loop(I, strids@[name])
			      | _ => ()) ;
		ListRef.push(visitedRefs, visited) ;
		visited := true)
	 end
   in loop(I, []) ;
      copyingDone() ;
      List.foreach(!visitedRefs, fn r => r := false) ;
      I
   end

val 'a fold =
   fn (I: t, b: 'a, f: Element.t * 'a -> 'a) =>
   let val Env{elements, ...} = force I
   in List.fold
      (elements, b, fn (elt, b) =>
       let
	  val elt =
	     case elt of
		Val{name, scheme, status} =>
		   Element.Val{name = name, status = status,
			       scheme = Scheme.strict scheme}
	      | Str r => Element.Str r
	      | Type{name, typeStr} =>
		   Element.Type{name = name,
				typeStr = TypeStr.strict typeStr}
       in f(elt, b)
       end)
   end

structure TypeStr = CtypeStr

end
