(* Heavily modified from the SML/NJ sources by sweeks@research.nj.nec.com. *)

(* Copyright 1996 by AT&T Bell Laboratories *)
(* precedence.sml *)
 
functor PrecedenceParse(S: PRECEDENCE_PARSE_STRUCTS): PRECEDENCE_PARSE =
struct

open S
   
local open Ast
in structure Exp = Exp
   structure Fixity = Fixity
   structure Fixop = Fixop
   structure Longvid = Longvid
   structure Pat = Pat
   structure Vid = Vid
end
   
structure Fixval =
   struct
      datatype t = Nonfix | Infix of (int*int)

      fun eval(f: Fixity.t): t =
	 case f of
	    Fixity.Infix NONE => Infix(0, 1)
	  | Fixity.Infix(SOME n) => Infix(n+n, n+n+1)
	  | Fixity.Infixr NONE => Infix(1, 0)
	  | Fixity.Infixr(SOME n) => Infix(n+n+1, n+n)
	  | Fixity.Nonfix => Nonfix

      fun make({name: Longvid.t, fixop: Fixop.t}, E: Env.t): t =
	 case (fixop, Longvid.split name) of
	    (Fixop.None, ([], vid)) => eval(Env.lookupFixity(E, vid))
	  | _ => Nonfix

      fun makePat(p: Pat.t, E: Env.t): t =
	 case Pat.node p of
	    Pat.Var r => make(r, E)
	  | _ => Nonfix

      fun makeExp(e: Exp.t, E: Env.t): t =
	 case Exp.node e of
	    Exp.Var r => make(r, E)
	  | _ => Nonfix
   end

(*---------------------------------------------------*)
(*           from elaborate/precedence.sml           *)
(*---------------------------------------------------*)

datatype 'a precStack =
   INf of int * 'a * 'a precStack
 | NONf of 'a * 'a precStack
 | NILf
   
fun 'a parse{region: 'a -> Region.t,
	     items: 'a list,
	     fixval: 'a -> Fixval.t,
	     apply: 'a * 'a -> 'a,
	     tuple: 'a list -> 'a}: 'a =
   let
      fun error msg =
	 Control.errorStr(Region.list(items, region), msg)
	 
      fun ensureNONf((e, f), p) =
	 (case f of
	     Fixval.Nonfix => ()
	   | _ => Control.errorStr(region e,
				   "expression or pattern begins with infix identifier") ;
          NONf(e, p))
	  
      fun start token = ensureNONf(token, NILf)
	 
      (* parse an expression *)
      fun parse(stack: 'a precStack, (item: 'a, fixval: Fixval.t)) =
	 case (stack, (item, fixval)) of
	    (NONf(e, r), (e', Fixval.Nonfix)) => NONf(apply(e, e'), r)
	  | (p as INf _, token) => ensureNONf(token, p)
	  | (p as NONf(e1, INf(bp, e2, NONf(e3, r))),
	     (e4, f as Fixval.Infix(lbp, rbp))) =>
	    if lbp > bp then INf(rbp, e4, p)
	    else (if lbp = bp
		     then error "mixed left- and right-associative \
		      \operators of same precedence"
		  else ();
		  parse(NONf(apply(e2, tuple[e3, e1]), r), (e4, f)))
	   | (p as NONf _, (e', Fixval.Infix(_, rbp))) => INf(rbp, e', p)
	   | _ => Error.bug "Precedence.parse"

      (* clean up the stack *)
      fun finish stack =
	 case stack of
	    NONf(e1, INf(_, e2, NONf(e3, r))) =>
	       finish(NONf(apply(e2, tuple[e3, e1]), r))
	  | NONf(e1, NILf) => e1
	  | INf(_, e1, NONf(e2, p)) =>
	       (error "expression or pattern ends with infix identifier" ;
		finish(NONf(apply(e2, e1), p)))
	  | NILf => Error.bug "Corelang.finish NILf"
	  | _ => Error.bug "Corelang.finish"

      fun getfix x = (x, fixval x)
	 
   in case items of
      item :: items =>
	 finish(List.fold(items, start(getfix item), fn (item, state) =>
			  parse(state, getfix item)))
    | _ => Error.bug "parse"
   end

fun parsePat(ps, E) =
   parse{region = Pat.region,
	 items = ps,
	 fixval = fn p => Fixval.makePat(p, E),
	 tuple = Pat.tuple,
	 apply = fn (p1, p2) =>
	 case Pat.node p1 of
	    Pat.Var{name, ...} => Pat.make(Pat.App(Longvid.toLongcon name, p2))
	  | _ => Error.bug "applyPat got non-constructor"}

fun parseExp(es, E) =
   parse{region = Exp.region,
	 items = es,
	 fixval = fn e => Fixval.makeExp(e, E),
	 apply = Exp.app,
	 tuple = Exp.tuple}

(*---------------------------------------------------*)
(*                    parseClause                    *)
(*---------------------------------------------------*)

fun parseClause(pats: Pat.t list, E: Env.t) =
   let
      fun empty() = Error.bug "parseClause: empty clause"
      fun error msg =
	 (Control.errorStr(Region.list(pats, Pat.region), msg) ;
	  {func = Ast.Var.bogus,
	   args = []})
	    
      fun illegal() = error "illegal function symbol in clause"
      fun done(func: Pat.t, args: Pat.t list) =
	 case Pat.node func of
	    Pat.Var{name, ...} =>
	       (case Longvid.split name of
		   ([], x) => {func = Vid.toVar x, args = args}
		 | _ => illegal())
	  | _ => illegal()

      fun parse(ps : Pat.t list) =
	 case ps of
	    p :: rest =>
	       (case (Pat.node p, rest) of
		   (Pat.FlatApp[a, b, c], _) =>
		      done(b, Pat.tuple[a, c] :: rest)
		 | (_, []) => error "can't find function arguments in clause"
		 | _ => done(p, rest))
	  | _ => Error.bug "empty clause"
	     
   in case pats of
      [a, b, c] => (case Fixval.makePat(b, E) of
		     Fixval.Nonfix => parse pats
		   | _ => done(b, [Pat.tuple[a, c]]))
    | _ => parse pats
   end
    
end
