(*-------------------------------------------------------------------*)
(*                          Resizable Array                          *)
(*-------------------------------------------------------------------*)

functor ResizableArray1D(): RESIZABLE_ARRAY_1D =
struct

structure Array = Array1D
open Array.Index
   
structure A' =
   struct
      structure Index = Array.Index
      structure Shape = Index
      structure Index = Index
   
      datatype 'a t = T of {array: 'a option Array.t ref,
			    size: int ref}

      fun getArray(T{array, ...}) = !array
      fun sizeRef(T{size, ...}) = size
      fun size a = !(sizeRef a)
      val shape = size
	 
      fun maxSize a = Array.size(getArray a)
      fun minSize a = maxSize a div 4
   
      fun invariant a =
	 maxSize a >= 1
	 andalso minSize a <= size a
	 andalso size a <= maxSize a
   
      fun incSize a = IntRef.inc(sizeRef a)
      fun decSize a = IntRef.dec(sizeRef a)
      fun maxIndex a = size a - 1
   
      exception New = Array.New

      fun empty() =
	 T{array = ref(Array.new(1, NONE)),
	   size = ref 0}
	 
      fun new(s, x) =
	 if s = 0 then empty()
	 else T{array = ref(Array.new(1, SOME x)),
		size = ref s}
      val array = new
	    
      fun tabulate(s, f) =
	 if s = 0 then empty()
	 else T{array = ref(Array.tabulate(s, fn i => SOME(f i))),
		size = ref s}
	    
      fun subSafe(a, i) =
	 case Array.sub(getArray a, i) of
	    SOME x => x
	  | NONE => Error.bug "subSafe"
		 
      fun sub(a, i) =
	 if i < size a then subSafe(a, i)
	 else Error.error "sub"
	    
      fun updateSafe(a, i, x) =
	 Array.update(getArray a, i, SOME x)
   
      fun update(a, i, x) =
	 if i < size a then updateSafe(a, i, x)
	 else Error.error "update"

      fun fromList l =
	 let val a = Array.fromList(List.map(l, SOME))
	 in T{array = ref a,
	      size = ref(Array.size a)}
	 end
   end

structure A'' = Array1D(A')

open A' A''

fun subOption(a, i) =
   if i < size a
      then Array.sub(getArray a, i)
   else NONE

fun grow (a as T{array, ...}) =
   array := Array.tabulate(maxSize a * 2,
		       fn i => subOption(a, i))

fun shrink (a as T{array, ...}) =
   array := Array.tabulate(maxSize a div 2,
		       fn i => subOption(a, i))

fun addToEnd(a, x) =
   (if size a = maxSize a then grow a else ()
    ; updateSafe(a, size a, x)
    ; incSize a)
       
fun deleteLast a =
   if size a = 0
      then Error.error "deleteLast"
   else let val x = subSafe(a, maxIndex a)
	in (if size a = minSize a then shrink a else ()
	    ; decSize a
	    ; x)
	end

end

structure ResizableArray1D = ResizableArray1D()
