structure Layout: LAYOUT =
struct

val detailed = ref false

fun switch{detailed=d,normal=n} x =
   if !detailed then d x else n x
   
structure String = String0

datatype t = T of {length: int,
		   tree: tree}
and tree =
   Empty
 | String of string
 | Sequence of t list
 | Align of {force: bool, rows: t list}
 | Indent of t * int

fun length(T{length, ...}) = length

val empty = T{length = 0, tree = Empty}

fun isEmpty(T{length = 0, ...}) = true
  | isEmpty _ = false

fun str s =
   case s of
      "" => empty
    | _ => T{length = String.size s, tree = String s}

fun fold(l, b, f) = foldl f b l
	 
fun seq ts =
   let val len = fold(ts, 0, fn (t,n) => n + length t)
   in case len of
      0 => empty
    | _ => T{length = len, tree = Sequence ts}
   end

local
   fun make force ts =
      let
	 fun loop ts =
	    case ts of
	       [] => (ts, 0)
	     | t :: ts =>
		  let val (ts,n) = loop ts
		  in case length t of
		     0 => (ts,n)
		   | n' => (t :: ts, n + n' + 1)
		  end
	 val (ts, len) = loop ts
      in case len of
	 0 => empty
       | _ => T{length = len - 1, tree = Align{force = force, rows = ts}}
      end
in
   val align = make true
   val mayAlign = make false
end

fun indent(t,n) = T{length = length t, tree = Indent(t,n)}

val tabSize = 8
fun blanks n =
   String.make(n div tabSize, #"\t") ^ String.make(n mod tabSize, #" ")

fun outputTree(t, out) =
   let val print = Out.outputc out
      fun loop(T{tree, length}) =
	 (print "(length " ;
	  print(Int.toString length) ;
	  print ")" ;
	 case tree of
	    Empty => print "Empty"
	  | String s => (print "(String " ; print s ; print ")")
	  | Sequence ts => loops("Sequence", ts)
	  | Align{force, rows} => loops("Align", rows)
	  | Indent(t,n) => (print "(Indent " ;
			    print(Pervasive.Int.toString n) ;
			    print " " ;
			    loop t ;
			    print ")"))
      and loops(s, ts) = (print "(" ; print s ;
			 app (fn t => (print " " ; loop t)) ts ;
			 print ")")
   in loop t
   end

fun toString t =
   let
      fun loop(T{tree, ...}, accum) =
	 case tree of
	    Empty => accum
	  | String s => s :: accum
	  | Sequence ts => fold(ts, accum, loop)
	  | Align{force, rows} => fold(rows, accum, loop)
	  | Indent(t, _) => loop(t, accum)
   in String.concat(rev(loop(t, [])))
   end

fun print{tree: t,
	  print: string -> unit,
	  lineWidth: int} =
   let
      (*val _ = outputTree(t, out)*)
      fun newline() = print "\n"

      fun outputCompact(t, {pos, havePrinted}) =
	 let
	    fun loop(T{tree, ...}) =
	       case tree of
		  Empty => ()
		| String s => print s
		| Sequence ts => app loop ts
		| Indent(t, _) => loop t
		| Align{rows, ...} =>
		     case rows of
			[] => ()
		      | t :: ts => (loop t ;
				  app (fn t => (print " " ; loop t)) ts)
	 in loop t ; {pos = pos + length t, havePrinted = length t > 0}
	 end

      fun loop(t as T{length, tree}, state as {pos, havePrinted}) =
	 let
	    fun prePrint() =
	       if havePrinted then ()
	       else if pos >= 0 then print(blanks pos)
		    else (* Should print a warning.  Aborting
			  * is too annoying.
			  *)
		       (*Error.error "can't print negative space" *)
		       ()
	 in (*outputTree(t, Out.error) ;*)
	    case tree of
	       Empty => state
	     | Indent(t,n) =>
		  loop(t, if havePrinted then state
			  else {pos = pos + n, havePrinted = false})
	     | Sequence ts => fold(ts, state, loop)
	     | String s =>
		  (prePrint() ; print s ;
		   {pos = pos + length, havePrinted = true})
	     | Align{force, rows} =>
		  if not force andalso pos + length <= lineWidth
		     then (prePrint() ; outputCompact(t, state))
		  else (case rows of
			   [] => state
			 | t :: ts =>
			      fold
			      (ts, loop(t, state), fn (t, _) =>
			       (newline() ;
				loop(t, {pos = pos,
					 havePrinted = false}))))
	 end
   in loop(tree, {pos=0, havePrinted=false}) ; ()
   end

fun outputWidth(t, width, out) =
   print{tree = t,
	 lineWidth = width,
	 print = Out.outputc out}

local
   val defaultWidth = 80
in
   fun output(t, out) = outputWidth(t, defaultWidth, out)
   val print =
      fn (t, p) => print{tree = t, lineWidth = defaultWidth, print = p}
end

fun makeOutput layoutX (x, out) = output(layoutX x, out)

fun ignore _ = empty
   
fun separate(ts, s) =
   case ts of
      [] => []
    | t :: ts => t :: (let val s = str s
		       fun loop [] = []
			 | loop(t :: ts) = s :: t::(loop ts)
		   in loop ts
		   end)
   
fun separateLeft(ts, s) =
   case ts of
      [] => []
    | [t] => ts
    | t :: ts => t :: (map (fn t => seq[str s, t]) ts)

fun separateRight(ts, s) =
   rev(let val ts = rev ts
       in case ts of
	  [] => []
	| [t] => ts
	| t :: ts => t :: (map (fn t => seq[t, str s]) ts)
       end)

fun alignPrefix(prefix, ts) =
   case ts of
      [] => empty
    | t :: ts =>
	 mayAlign[t, indent(mayAlign(map (fn t => seq[str prefix, t]) ts),
			    ~(String.size prefix))]

local
   fun sequence(start, finish) ts =
      seq[str start, mayAlign(separateRight(ts, ",")), str finish]
in
   val list = sequence("[", "]")
   val tuple = sequence("(", ")")
   val vector = sequence("#[", "]")
   fun record fts = sequence("{", "}") (map (fn (f, t) => seq[str(f ^ " = "), t])
				       fts)
end

fun namedRecord(name, fields) = seq[str name, record fields]

fun paren t = seq[str "(", t, str ")"]

fun option layoutX opt =
   case opt of
      NONE => str "NONE"
    | SOME x => paren(seq[str "SOME ", layoutX x])

fun tuple2(l1, l2)(x1, x2) = tuple[l1 x1, l2 x2]
fun tuple3(l1, l2, l3)(x1, x2, x3) = tuple[l1 x1, l2 x2, l3 x3]
fun tuple4(l1, l2, l3, l4)(x1, x2, x3, x4) = tuple[l1 x1, l2 x2, l3 x3, l4 x4]
  
end
