functor DirectedGraph(S: DIRECTED_GRAPH_STRUCTS): DIRECTED_GRAPH = 
struct

open S

structure Node =
   struct
      datatype t = T of {successors: t list ref,
			 plist: PropertyList.t}

      fun layout _ = Layout.str "node"

      fun successors(T{successors, ...}) = !successors
      fun plist(T{plist, ...}) = plist

      fun equals(n,n') = PropertyList.equals(plist n, plist n')

      fun new() = T{successors = ref [],
		    plist = PropertyList.new()}

      fun addSuccessor(T{successors, ...}, n') = ListRef.push(successors, n')

      fun removeSuccessor(T{successors, ...}, n') =
	 successors := List.removeFirst(!successors, fn n => equals(n, n'))

      val equals = fn (n,n') => PropertyList.equals(plist n, plist n')
   end

structure Edge =
   struct
      type t = {from: Node.t, to: Node.t}
   end

structure DfsParam =
   struct
      type t = {startNode: Node.t -> unit,
		finishNode: Node.t -> unit,
		handleTreeEdge: Edge.t -> unit,
		handleNonTreeEdge: Edge.t -> unit,
		startTree: Node.t -> unit,
		finishTree: Node.t -> unit,
		finishDfs: unit -> unit}
	 
      fun ignore _ = ()

      fun finishNode f = {finishNode = f,
			  startNode = ignore,
			  handleTreeEdge = ignore,
			  handleNonTreeEdge = ignore,
			  startTree = ignore,
			  finishTree = ignore,
			  finishDfs = ignore}

      fun startNode f = {finishNode = ignore,
			 startNode = f,
			 handleTreeEdge = ignore,
			 handleNonTreeEdge = ignore,
			 startTree = ignore,
			 finishTree = ignore,
			 finishDfs = ignore}
	 
      fun seq f g a = (f a ; g a)
	 
      fun combine({startNode, finishNode,
		   handleTreeEdge, handleNonTreeEdge,
		   startTree, finishTree, finishDfs}: t,
		  {startNode = sn, finishNode = fin,
		   handleTreeEdge = ht, handleNonTreeEdge = hn,
		   startTree = st, finishTree = ft, finishDfs = fd}: t): t =
	 {startNode = seq startNode sn,
	  finishNode = seq finishNode fin,
	  handleTreeEdge = seq handleTreeEdge ht,
	  handleNonTreeEdge = seq handleNonTreeEdge hn,
	  startTree = seq startTree st,
	  finishTree = seq finishTree ft,
	  finishDfs = seq finishDfs fd}
   end

(*---------------------------------------------------*)
(*                  graph datatype                   *)
(*---------------------------------------------------*)

datatype t = T of {nodes: Node.t list ref}

fun nodes(T{nodes, ...}) = !nodes

fun new() = T{nodes = ref []}
   
fun newNode(T{nodes, ...}) =
   let val n = Node.new()
   in ListRef.push(nodes, n) ;
      n
   end

fun addEdge(_, {from, to}) = Node.addSuccessor(from, to)
fun removeEdge(_, {from, to}) = Node.removeSuccessor(from, to)

(*--------------------------------------------------------*)
(*                   Depth-First Search                   *)
(*--------------------------------------------------------*)

fun dfsNodes(g, ns, {startNode, finishNode,
		     handleTreeEdge, handleNonTreeEdge,
		     startTree, finishTree, finishDfs}) =
   let
      val {get = hasBeenVisited, set, destroy, ...} =
	 Property.newDest(Node.plist, Property.initConst false)

      fun visit n =
	 (startNode n ;
	  set(n, true) ;
	  List.foreach(Node.successors n, fn n' =>
		       let val edge = {from = n, to = n'}
		       in if hasBeenVisited n'
			     then handleNonTreeEdge edge
			  else (visit n' ; handleTreeEdge edge)
		       end) ;
	  finishNode n)
	       
   in List.foreach(ns, fn n =>
		   if hasBeenVisited n
		      then ()
		   else (startTree n ; visit n ; finishTree n)) ;
      destroy() ;
      finishDfs()
   end

fun dfs(g, p) = dfsNodes(g, nodes g, p)

fun display{graph, layoutNode, display} =
   dfs(graph,
       DfsParam.startNode(fn n =>
			  display(let open Layout
				  in seq[layoutNode n,
					 str " ",
					 list(List.map(Node.successors n,
						       layoutNode))]
				  end)))
			     
   
fun foreachDescendent(g, n, f) =
   dfsNodes(g, [n], DfsParam.finishNode f)

fun removeBackEdges g =
   let
      val discoverTime = Counter.new 0
      val {get, destroy, ...} =
	 Property.newDest
	 (Node.plist, Property.initFun(fn _ => {time = Counter.next discoverTime,
						alive = ref true}))
      val ignore = DfsParam.ignore
   in dfs
      (g, {startNode = fn n => (get n; ()),
	   finishNode = fn n => #alive(get n) := false,
	   handleNonTreeEdge =
	   fn e as {from, to} => let val {alive, time} = get to
				 in if !alive andalso time < #time(get from)
				       then removeEdge(g, e)
				    else ()
				 end,
	   handleTreeEdge = ignore,
	   startTree = ignore,
	   finishTree = ignore,
	   finishDfs = ignore})
   end

(*--------------------------------------------------------*)
(*                         Times                          *)
(*--------------------------------------------------------*)

fun discoverFinishTimes g =
    let val time = ref 0
	val {get = discover, set = setDiscover, destroy = destroyDiscover, ...} =
	   Property.newDest
	   (Node.plist, Property.initRaise("discover", Node.layout))
	val {get = finish, set = setFinish, destroy = destroyFinish, ...} =
	   Property.newDest(Node.plist, Property.initRaise("finish", Node.layout))
	   
    in {destroy = fn () => (destroyDiscover() ; destroyFinish()),
	discover = discover,
	finish = finish,
	param = {startNode = fn n => (time := !time + 1 ; setDiscover(n, !time)),
		 finishNode = fn n => (time := !time + 1 ; setFinish(n, !time)),
		 handleTreeEdge = DfsParam.ignore, handleNonTreeEdge = DfsParam.ignore,
		 startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
		 finishDfs = DfsParam.ignore}}
    end
(*
local
   fun undirected [] = []
     | undirected (n::ns) = 
       List.append(List.map(ns, fn n' => (n,n')),
		   undirected ns)
   fun directed ns =
      let val es = undirected ns
      in List.append(es, List.map(es, fn (n,n') => (n',n)))
      end
   val possible = if isDirected then directed else undirected
in val possibleEdges = possible o nodes
end
*)
(*--------------------------------------------------------*)
(*                        Foreach                         *)
(*--------------------------------------------------------*)

fun foreachNode(g, f) = List.foreach(nodes g, f)
   
fun foreachEdge(g, handleEdge) =
   dfs(g, {handleTreeEdge = handleEdge,
	  handleNonTreeEdge = handleEdge,
	  startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	  startNode = DfsParam.ignore, finishNode = DfsParam.ignore,
	  finishDfs = DfsParam.ignore})

(*--------------------------------------------------------*)
(*                         Input                          *)
(*--------------------------------------------------------*)
(*
exception Input
fun input(instream, inputInfo, inputEdgeInfo) =
   let val numNodes = Int.input instream
      val numEdges = Int.input instream
      val info = inputInfo instream
      val g = new numNodes
      val a = A.new(numEdges, NONE)
      fun inputEdge i =
	 (addEdge(g, Node.input instream, Node.input instream) ;
	  A.update(a, i, SOME(inputEdgeInfo instream)))
      val edgeInfo = Option.projector (fn e => A.sub(a, Edge.index e))
   in (Int.foreach(0, numEdges - 1, inputEdge) ;
       (g, info, edgeInfo))
   end handle _ => raise Input
*)
(*--------------------------------------------------------*)
(*                         Output                         *)
(*--------------------------------------------------------*)
(*
fun output(g, out) =
   let val print = Out.outputc out
   in (print "nodes = " ;
       Layout.output(Int.layout(numNodes g), out) ;
       print "\nedges =" ;
       Layout.output(Int.layout(numEdges g), out) ;
       (*outputInfo out ;*)
       foreachEdge(g, fn e => (print "\n" ;
			      Edge.output(e, out))))
			      (*print " " ;
			       outputEdgeInfo(e, out)*)
   end
	*)		      
(*--------------------------------------------------------*)
(*                         Random                         *)
(*--------------------------------------------------------*)
(*
fun maxNumEdges n = n * (n - 1)
   
fun random {numNodes,numEdges} =
   let val max = maxNumEdges numNodes
   in if numNodes < 0 orelse numEdges < 0 orelse numEdges > max
	 then Error.error "random"
      else let val g = new()
	       val needed = ref numEdges
	       val remaining = ref max
	       fun maybeAddEdge(n,n') =
		  (if Int.random(1, !remaining) <= !needed
		      then (addEdge(g, Node.fromInt n, Node.fromInt n') ;
			    IntRef.dec needed)
		   else () ;
		   IntRef.dec remaining)
	       val minNode = 0
	       val maxNode = numNodes - 1
	       fun directed n =
		  Int.foreach(0, maxNode, fn n' =>
			      if n = n' then () else maybeAddEdge(n,n'))
	       fun undirected n =
		  Int.foreach(n + 1, maxNode, fn n' => maybeAddEdge(n,n'))
	       val addEdges = if isDirected then directed
			      else undirected
	   in (Int.foreach(minNode, maxNode, addEdges) ;
	       g)
	   end
   end
*)
(*--------------------------------------------------------*)
(*                         Cycle                          *)
(*--------------------------------------------------------*)
(*
fun cycleParam g =
   let val {get = isActive, set = setActive} =
      nodeInfo(g, fn _ => false)
      val cycle = ref false
   in (cycle, {startNode = fn n => setActive(n, true),
	       finishNode = fn n => setActive(n, false),
	       handleNonTreeEdge =
	       fn (n, e) => let val n' = Edge.otherNode(e,n)
			   in if isActive n' then cycle := true
			      else ()
			   end,
			handleTreeEdge = DfsParam.ignore,
			startTree = DfsParam.ignore,
			finishTree = DfsParam.ignore,
			finishDfs = DfsParam.ignore})
   end

fun isCyclic g = let val (cycle, p) = cycleParam g
		 in (dfs(g, p) ; !cycle)
		 end
*)

(*--------------------------------------------------------*)
(*                    Topological Sort                    *)
(*--------------------------------------------------------*)

exception TopologicalSort

fun topSortParam g =
   let
      val {get = amVisiting, set, destroy, ...} =
	 Property.newDest(Node.plist, Property.initConst false)
      val ns = ref []
   in (ns, {startNode = fn n => set(n, true),
	    finishNode = fn n => (set(n, false) ; ListRef.push(ns,n)),
	    handleNonTreeEdge =
	    fn {from, to} => if amVisiting to
			       then raise TopologicalSort
			    else (),
	    startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	    handleTreeEdge = DfsParam.ignore,
	    finishDfs = fn () => destroy()})
   end

fun topologicalSort g = let val (ns, p) = topSortParam g
			in (dfs(g, p) ; !ns)
			end

(*--------------------------------------------------------*)
(*                       Transpose                        *)
(*--------------------------------------------------------*)
(*
fun transposeParam g =
   let val gt = new()
      fun handleEdge(n, e) = let val n' = Edge.otherNode(e,n)
			    in (addEdge(gt,n',n) ; ())
			    end
   in (gt, {handleTreeEdge = handleEdge,
	    handleNonTreeEdge = handleEdge,
	    finishDfs = DfsParam.ignore,
	    startNode = DfsParam.ignore, finishNode = DfsParam.ignore,
	    startTree = DfsParam.ignore, finishTree = DfsParam.ignore})
   end

fun transpose g = let val (gt, p) = transposeParam g
		  in (dfs(g, p) ; gt)
		  end
  *) 
(*--------------------------------------------------------*)
(*             Strongly Connected Components              *)
(*--------------------------------------------------------*)
   
(* from Cormen, Leiserson, and Rivest 23.5 *)
(*
fun sccCLR g =
   let
      val (gt, p) = transposeParam g
      val ns = ref []
      val p' = P.finishNode(fn n => ListRef.push(ns,n))
      val components = ref []
      val component = ref []
      fun startNode n = ListRef.push(component,n)
      fun startTree _ = component := []
      fun finishTree _ = ListRef.push(components, !component)
      val pt = {startNode = startNode,
		   startTree = startTree,
		   finishTree = finishTree,
		   finishNode = DfsParam.ignore,
		   finishDfs = DfsParam.ignore,
		   handleTreeEdge = DfsParam.ignore,
		   handleNonTreeEdge = DfsParam.ignore}
   in (dfs(g, P.combine(p, p')) ;
       dfsNodes(gt, !ns, pt) ;
       !components)
   end
*)

(* from Aho, Hopcroft, Ullman section 5.5 *)

fun stronglyConnectedComponents g =
   let
      val {get = discover, set = setDiscover, destroy = destroyDiscover, ...} =
	 Property.newDest(Node.plist, Property.initRaise("discover", Node.layout))
      val {get = low, set = setLow, destroy = destroyLow, ...} =
	 Property.newDest(Node.plist, Property.initRaise("low", Node.layout))
      val {get = isOnStack, set = setOnStack, destroy = destroyStack, ...} =
	 Property.newDest(Node.plist, Property.initConst false)
	 
      val stack = ref []
      val components = ref []
      val time = ref 0
	 
      fun pop() = let val n = ListRef.pop stack
		  in (setOnStack(n, false) ; n)
		  end
      fun popTo n = let fun popTo() = let val n' = pop()
				      in if Node.equals(n,n') then [n]
					 else n' :: (popTo())
				      end
		    in popTo()
		    end

      fun startNode n = (time := !time + 1 ;
			 setDiscover(n, !time) ;
			 setLow(n, !time) ;
			 setOnStack(n, true) ;
			 ListRef.push(stack,n))
	 
      fun finishNode n = if discover n = low n
			     then ListRef.push(components, popTo n)
			 else ()
			    
      fun updateLow{from, to} = if low to < low from then setLow(from, low to)
			       else ()
			       
      val handleTreeEdge = updateLow
	 
      fun handleNonTreeEdge e = 
	 if isOnStack(#to e)
	    then updateLow e
	 else ()
      
      val p = {startNode = startNode, finishNode = finishNode,
	       handleTreeEdge = handleTreeEdge,
	       handleNonTreeEdge = handleNonTreeEdge,
	       startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	       finishDfs = DfsParam.ignore}
	 
   in dfs(g, p) ;
      destroyLow() ;
      destroyStack() ;
      destroyDiscover() ;
      !components
   end

end
