(*-------------------------------------------------------------------*)
(*                              Forest                               *)
(*-------------------------------------------------------------------*)

(* Forest of heap ordered trees.
 Can be specialized to eager or lazy binomial heaps, fibonacci heaps. *)

functor ForestHeap (K: BOUNDED_ORDER): FOREST_HEAP =
struct

structure K = K
structure P = Pointer
   
(*--------------------------------------------------------*)
(*                      Heap Element                      *)
(*--------------------------------------------------------*)

structure E =
struct
   (* can't make child a circular list, because the elements
    aren't defined yet *)
   datatype 'a t = T of {value: 'a P.t,
			 mark: bool ref,
			 parent: 'a t P.t,
			 next: 'a t P.t,
			 prev: 'a t P.t,
			 child: 'a t P.t,
			 numChildren: int ref}

   fun destruct(T{prev, value,next, ...}) = (prev, P.! value,next)

   fun make p = T{value = p,
		  mark = ref false,
		  parent = P.null(),
		  next = P.null(),
		  prev = P.null(),
		  child = P.null(),
		  numChildren = ref 0}

   fun new v = make(P.new v)

   fun dummy v = make(P.null())
      
   fun valuePtr(T{value, ...}) = value
   fun value e = P.!(valuePtr e)
   fun mark(T{mark, ...}) = mark := true
   fun unMark(T{mark, ...}) = mark := false
   fun isMarked(T{mark, ...}) = !mark
   fun parent(T{parent, ...}) = P.! parent
   fun hasParent(T{parent, ...}) = not(P.isNull parent)
   fun setParent(T{parent, ...}, p) = P.:=(parent, p)
   fun clearParent(T{parent, ...}) = P.clear parent
   fun next(T{next, ...}) = next
   fun prev(T{prev, ...}) = prev
   fun child(T{child, ...}) = child
   fun numChildrenRef(T{numChildren, ...}) = numChildren
   fun numChildren e = !(numChildrenRef e)
   fun incNumChildren e = IntRef.inc(numChildrenRef e)
   fun decNumChildren e = IntRef.dec(numChildrenRef e)
   val equals = fn (e, e') => P.eq(valuePtr e, valuePtr e')
end
   
structure CL = CircularList(DoublyLinkedElement(E))

structure E =
   struct
      open E
      type 'a t = (K.t * 'a) E.t

      fun key(e: 'a t) = #1(value e)
      fun value(e: 'a t) = #2(E.value e)
      fun setKey(e, k) = P.:=(valuePtr e, (k, value e))

      fun siftUp e =
	 if hasParent e
	    then let val p = parent e
		 in if K.<(key e, key p)
		       then (P.swap(valuePtr e, valuePtr p) ;
			     siftUp p)
		    else ()
		 end
	 else ()
   end

(*--------------------------------------------------------*)
(*                     Heap Datatype                      *)
(*--------------------------------------------------------*)

datatype 'a t = T of {size: int ref,
		      roots: (K.t * 'a) CL.t,
		      min: 'a E.t P.t}

fun sizeRef (T{size, ...}) = size
fun size h = !(sizeRef h)
fun setSize(h,n) = sizeRef h := n
fun incSize h = IntRef.inc(sizeRef h)
fun decSize h = IntRef.inc(sizeRef h)

fun roots (T{roots, ...}) = roots
						       
fun min(T{min, ...}) = P.! min

fun clearMin(T{min, ...}) = P.clear min
 
fun updateMin(T{min, ...}, e) =
   if P.isNull min orelse K.<(E.key e, E.key(P.!min))
      then P.:=(min, e)
   else ()

fun addRoot(h, e) = (CL.insert(roots h, e) ; updateMin(h, e))
		 
fun isEmpty h = size h = 0

local
    fun linkPC(parent, child) =
	(E.incNumChildren parent ;
	 CL.insert(E.child parent, child) ;
	 E.setParent(E.parent child, parent) ;
	 E.unMark child ;
	 parent)
in fun link(e, e') =
    (* pre: numChildren e = numChildren e' *)
   if K.<(E.key e, E.key e')
      then linkPC(e, e')
   else linkPC(e', e)
end

fun unlink e = let val p = E.parent e
	       in (E.decNumChildren p ;
		   CL.delete(E.child p, e) ;
		   E.clearParent e ;
		   E.unMark e)
	       end
				
local
   structure A = Array1D
   structure I = Int
   structure R = Real
   val phi = (1.0 + (R.sqrt 5.0)) / 2.0
   fun maxNumChildren h = floor(R.log(phi, I.toReal(size h)))
in
   fun consolidate h =
      (clearMin h ;
       if size h = 0 then ()
       else let val a = A.new(maxNumChildren h + 1, NONE)
		fun insertIntoA e =
		    let val n = E.numChildren e
		    in case A.sub(a,n) of
			NONE => A.update(a,n, SOME e)
		      | SOME e' => (A.update(a,n, NONE) ;
				    insertIntoA(link(e, e')))
		    end
	    in (CL.deleteEach(roots h, insertIntoA) ;
		A.foreachIndex(a,
			       fn (_, NONE) => ()
				| (_, SOME e) => addRoot(h, e)))
	    end)
end

(*--------------------------------------------------------*)
(*            Constructors: empty, insert, new            *)
(*--------------------------------------------------------*)

fun empty() = T{size = ref 0,
		roots = CL.empty(),
		min = P.null()}

fun insertLazy(h, k, v) =
   let val e = E.new(k, v)
   in (incSize h ; addRoot(h, e) ; e)
   end

fun insertEager(h, k, v) =
    let val e = insertLazy(h, k, v)
    in (consolidate h ; e)
    end

fun newLazy kvs =
   let val h = empty()
   in (List.foreach(kvs, fn (k, v) => (insertLazy(h, k, v) ; ())) ;
       h)
   end

fun newEager kvs = let val h = newLazy kvs
		   in (consolidate h ; h)
		   end

(*--------------------------------------------------------*)
(*                       DeleteMin                        *)
(*--------------------------------------------------------*)
   
fun deleteMin h = let val m = min h
		      val c = E.child m
		      val rs = roots h
		  in (decSize h ;
		      CL.delete(rs, m) ;
		      CL.foreach(c, E.clearParent) ;
		      CL.splice(rs, c) ;
		      consolidate h ;
		      E.value m)
		  end

(*--------------------------------------------------------*)
(*                      DecreaseKey                       *)
(*--------------------------------------------------------*)

fun sift(_, e, _) = E.siftUp e

fun cut(h, e, k) =
   if E.hasParent e
      andalso K.<(k, E.key(E.parent e))
      then let val rs = roots h
	       fun cut e = if E.hasParent e
			      then let val p = E.parent e
				   in (unlink e ;
				       CL.insert(rs, e) ; 
				       if E.isMarked p
					  then cut p
				       else E.mark p)
				   end
			   else ()
	   in cut e
	   end
   else ()
      
fun decreaseKey(h, e, k) =
    if K.>(k, E.key e) then Error.error "decreaseKey"
    else (E.setKey(e, k) ; updateMin(h, e))

fun decreaseKeySift(h, e, k) = (decreaseKey(h, e, k) ;
			      sift(h, e, k))

fun decreaseKeyCut(h, e, k) = (decreaseKey(h, e, k) ;
			     cut(h, e, k))

(*--------------------------------------------------------*)
(*                         Delete                         *)
(*--------------------------------------------------------*)

fun delete(decreaseKey, h, e) =
   let val k = E.key e
   in (decreaseKey(h, e, K.smallest) ;
       deleteMin h ;
       E.setKey(e, k))
   end

fun deleteSift(h, e) = delete(decreaseKeySift, h, e)

fun deleteCut(h, e) = delete(decreaseKeyCut, h, e)

(*--------------------------------------------------------*)
(*                         Union                          *)
(*--------------------------------------------------------*)

fun union(h, h') =
    (setSize(h, size h + size h') ;
     CL.splice(roots h, roots h'))

fun unionEager(h, h') = (union(h, h') ; consolidate h)

fun unionLazy(h, h') = (union(h, h') ;
		       updateMin(h, min h') handle Min => ())

(*--------------------------------------------------------*)
(*                         Output                         *)
(*--------------------------------------------------------*)

fun output(heap, outputValue, out) = Error.unimplemented "output"

(*--------------------------------------------------------*)
(*                    Well-Formed Test                    *)
(*--------------------------------------------------------*)
(*
local
   fun sizeInTree e = 1 + sizeInTrees (CL.T (E.children e))
   and sizeInTrees l = CL.fold l 0 (fn (e,n) => n + sizeInTree e)
   fun sizeInHeap h = sizeInTrees (roots h)
   fun findMin h =
      let val min = ref NONE
	 fun updateMin e = (case !min of
			       NONE => min := SOME e
			     | SOME e' => if K.<(E.key e, E.key e')
					     then min := SOME e
					  else ())
      in (CL.foreach (roots h) updateMin ;
	  case !min of
	     SOME e => e
	   | NONE => bug "findMin") 
      end
   fun isTreeWellFormed e =
      let fun isChildWellFormed e' = (E.equals(e, E.parent e')
				      andalso K.<=(E.key e, E.key e')
				      andalso isTreeWellFormed e')
	 val cs = CL.T (E.children e)
      in E.numChildren e = CL.length cs
	 andalso CL.forall cs isChildWellFormed
      end
in
   fun isFibonacciHeap h =
      CL.forall (roots h) isTreeWellFormed
      andalso size h = sizeInHeap h
      andalso (isEmpty h
	       orelse K.equals(E.key (min h), E.key (findMin h)))
end
*)
end

