(*-------------------------------------------------------------------*)
(*                            Binary Heap                            *)
(*-------------------------------------------------------------------*)
functor BinaryHeap (Key: BOUNDED_TOTAL_ORDER): HEAP =
struct

val {bug, error,notImplemented, ...} = Error.errors("heap", "binary")

structure A = ResizableArray
structure AU = ResizableArrayUtil
structure Key = Key
structure K = Key
structure O = Option

(*--------------------------------------------------------*)
(*                        Element                         *)
(*--------------------------------------------------------*)
structure Elt =
   struct
      datatype 'a t = T of {key: K.t ref,
			    value: 'a,
			    index: int ref}
      fun new(k, v, i) = T{key = ref k,
			 value = v,
			 index = ref i}
      fun key(T{key, ...}) = !key
      fun setKey(T{key, ...}, k) = key := k
      fun value (T{value, ...}) = value
      fun index (T{index, ...}) = !index
      fun setIndex(T{index, ...}, i) = index := i
   end
structure E = Elt

(*--------------------------------------------------------*)
(*                          Heap                          *)
(*--------------------------------------------------------*)
   
datatype 'a t = T of 'a Elt.t A.t

fun empty() = T (A.listTo [])
   
fun fixIndex(a, i) = E.setIndex(A.sub(a, i), i)
fun swap(a, i, j) = (AU.swap(a, i, j) ;
		   fixIndex(a, i) ;
		   fixIndex(a, j))
   
fun isEmpty (T a) = A.size a = 0
   
fun parent i = Int.divide(i-1, 2)
fun left i = 2 * i + 1
fun right i = 2 * i + 2
fun key(a, i) = E.key(A.sub(a, i))
fun keyOption(a, i) = O.optionize E.key (A.subOption(a, i))

fun siftUp(a, i) =
   let fun siftUp i = if i = 0 then ()
		      else let val p = parent i
			   in if K.less(key(a, i), key(a, p))
				 then (swap(a, i, p) ; siftUp p)
			      else ()
			   end
   in siftUp i
   end
	
fun siftDown(a, i) =
   let
      fun siftDown i =
	 let val l = left i
	    val r = right i
	 in case keyOption(a, l) of
	    NONE => ()
	  | SOME kl =>
	       let val min = (case keyOption(a, r) of
				 NONE => l
			       | SOME kr => if K.less(kl, kr)
						then l else r)
	       in if K.less(key(a, i), key(a, min)) then ()
		  else (swap(a, i, min) ; siftDown min)
	       end
	 end
   in siftDown i
   end
      
fun new es =
   let val a = A.listTo(LU.tabulate(es, fn (i, (k, v)) => E.new(k, v, i)))
      val start = Int.divide(A.size a, 2)
   in (Iterate.forDown(start, 0, fn i => siftDown(a, i)) ;
       T a)
   end

fun isEmpty (T a) = A.size a = 0
   
fun foreach(T a, f) = AU.foreach(a, fn (_, x) => f x)

fun insert(T a, k, v) =
   let val i = A.size a
      val e = E.new(k, v, i)
   in (A.addToEnd(a, e) ; siftUp(a, i) ; e)
   end

fun min (h as (T a)) =
   if isEmpty h then error "min"
   else A.sub(a, 0)

fun deleteMin (h as (T a)) =
   if isEmpty h then error "deleteMin"
   else E.value (if A.size a = 1
		    then A.deleteLast a
		 else let val min = A.sub(a, 0)
		      in (A.update(a, 0, A.deleteLast a) ;
			  fixIndex(a, 0) ;
			  siftDown(a, 0) ;
			  min)
		      end)
 
fun decreaseKey(T a, e, k) =
   if K.less(E.key e, k) then error "decreaseKey"
   else (E.setKey(e, k) ; siftUp(a, E.index e))

fun delete(h, e) = (decreaseKey(h, e, K.min) ; deleteMin h ; ())

fun union(h, h') =
    foreach(h', fn e => (insert(h, E.key e, E.value e) ;
			()))

fun output _ = notImplemented "output"

end
