(*-------------------------------------------------------------------*)
(*                              Forest                               *)
(*-------------------------------------------------------------------*)

(* A general heap as forest of heap ordered trees data structure.
 Can be specialized to eager or lazy binomial heaps, fibonacci heaps. *)

functor ForestHeap (Key: BOUNDED_ORDER): FOREST_HEAP =
struct

val {bug, error, unimplemented, ...} = Error.errors("heap", "forest")

structure Key = Key
structure K = Key
   
(*--------------------------------------------------------*)
(*                      Heap Element                      *)
(*--------------------------------------------------------*)

structure Elt =
struct
   datatype 'a t = T of {key: Key.t ref,
			 value: 'a,
			 mark: bool ref,
			 parent: 'a t option ref,
			 prev: 'a t option ref,
			 next: 'a t option ref,
			 numChildren: int ref,
			 children: 'a t option ref}
      
   fun new(key, value) = T{key = ref key,
			  value = value,
			  numChildren = ref 0,
			  mark = ref false,
			  parent = ref NONE,
			  prev = ref NONE,
			  next = ref NONE,
			  children = ref NONE}

   fun key (T{key, ...}) = !key
   fun setKey (T{key, ...}, k) = key := k

   fun value(T{value, ...}) = value
      
   fun numChildrenRef (T{numChildren, ...}) = numChildren
   fun numChildren e = !(numChildrenRef e)
   fun incNumChildren e = numChildrenRef e := numChildren e + 1
   fun decNumChildren e = numChildrenRef e := numChildren e - 1

   fun mark (T{mark, ...}) = mark := true
   fun unMark (T{mark, ...}) = mark := false
   fun isMarked (T{mark, ...}) = !mark
      
   fun hasParent (T{parent = ref NONE, ...}) = false
     | hasParent _ = true
   fun parent (T{parent = ref (SOME parent), ...}) = parent
     | parent _ = bug "Element.parent"
   fun setParent(T{parent, ...}, e) = parent := SOME e
   fun clearParent (T{parent, ...}) = parent := NONE

   fun prevOpt(T{prev, ...}) = prev
   fun nextOpt(T{next, ...}) = next
   fun children(T{children, ...}) = children

   fun swap(T{mark, parent, prev,next,numChildren, children, ...},
	    T{mark=m, parent=p, prev=v,next=n,numChildren=u, children=c, ...}) =
      (Ref.swap(mark, m) ;
       Ref.swap(parent, p) ;
       Ref.swap(prev, v) ;
       Ref.swap(next,n) ;
       Ref.swap(numChildren, u) ;
       Ref.swap(children, c))
   
   fun siftUp e =
      let fun sift() = if hasParent e
			  then let val p = parent e
			       in if K.>=(key e, key p) then ()
				  else (swap(e, p) ; sift())
			       end
		       else ()
      in sift()
      end

   fun eq(e, e') = prev e = prev e'
end
structure E = Elt
   
(*--------------------------------------------------------*)
(*                     Heap Datatype                      *)
(*--------------------------------------------------------*)

structure CL = CircularList(DoublyLinkedElement(Elt))

datatype 'a t = T of {size: int ref,
		      roots: 'a CL.t,
		      min: 'a Elt.t option ref}

fun sizeRef (T{size, ...}) = size
fun size h = !(sizeRef h)
fun setSize(h,n) = sizeRef h := n
fun incSize h = IntRef.inc(sizeRef h)
fun decSize h = IntRef.inc(sizeRef h)

fun roots (T{roots, ...}) = roots
						       
fun min (T{min, ...}) = (case !min of
			   SOME e => e
			 | NONE => error "min")
fun clearMin (T{min, ...}) = min := NONE
 
fun updateMin(T{min, ...}, e) =
   case !min of
      NONE => min := SOME e
    | SOME e' => if K.>=(E.key e, E.key e') then ()
		 else min := SOME e

fun addRoot(h, e) = (CL.insert(roots h, e) ; updateMin(h, e))
		 
fun isEmpty h = size h = 0

local
    fun linkPC(parent, child) =
	(E.incNumChildren parent ;
	 CL.insert(CL.T(E.children parent), child) ;
	E.setParent(child, parent) ;
	 E.unMark child ;
	 parent)
in fun link(e, e') =
    (* pre: numChildren e = numChildren e' *)
    let
    in if Key.less(E.key e, E.key e')
	   then linkPC(e, e')
       else linkPC(e', e)
    end
end

fun unlink e = let val p = E.parent e
	       in (E.decNumChildren p ;
		   CL.delete(CL.T (E.children p), e) ;
		   E.clearParent e ;
		   E.unMark e)
	       end
				
local
   structure A = Array1D
   structure AU = Array1DUtil
   structure I = Int
   structure R = Real
   val phi = (1.0 + (sqrt 5.0)) / 2.0
   fun maxNumChildren h = floor (R.log{base = phi,
				       arg = I.toReal (size h)})
in
   fun consolidate h =
      (clearMin h ;
       if size h = 0 then ()
       else let val a = A.new(maxNumChildren h + 1, fn _ => NONE)
		fun insertIntoA e =
		    let val n = E.numChildren e
		    in case A.sub(a,n) of
			NONE => A.update(a,n, SOME e)
		      | SOME e' => (A.update(a,n, NONE) ;
				    insertIntoA(link(e, e')))
		    end
	    in (CL.deleteEach(roots h, insertIntoA) ;
		AU.foreach(a,
			   fn (_, NONE) => ()
			    | (_, SOME e) => addRoot(h, e)))
	    end)
end

(*--------------------------------------------------------*)
(*            Constructors: empty, insert, new            *)
(*--------------------------------------------------------*)

fun empty() = T{size = ref 0,
		roots = CL.empty(),
		min = ref NONE}

fun insertLazy(h, k, v) =
   let val e = E.new(k, v)
   in (incSize h ; addRoot(h, e) ; e)
   end

fun insertEager(h, k, v) =
    let val e = insertLazy(h, k, v)
    in (consolidate h ; e)
    end

fun newLazy kvs =
   let val h = empty()
   in (LU.foreach(kvs, fn (k, v) => (insertLazy(h, k, v) ; ())) ;
       h)
   end

fun newEager kvs = let val h = newLazy kvs
		   in (consolidate h ; h)
		   end

(*--------------------------------------------------------*)
(*                       DeleteMin                        *)
(*--------------------------------------------------------*)
   
fun deleteMin h = let val m = min h
		      val cs = CL.T (E.children m)
		      val rs = roots h
		  in (decSize h ;
		      CL.delete(rs, m) ;
		      CL.foreach(cs, E.clearParent) ;
		      CL.splice(rs, cs) ;
		      consolidate h ;
		      E.value m)
		  end

(*--------------------------------------------------------*)
(*                      DecreaseKey                       *)
(*--------------------------------------------------------*)

fun sift(_, e, _) = E.siftUp e

fun cut(h, e, k) =
   if not (E.hasParent e
	   andalso K.less(k, E.key(E.parent e)))
       then ()
   else let val rs = roots h
	    fun cut e = if E.hasParent e
			    then let val p = E.parent e
				 in (unlink e ;
				     CL.insert(rs, e) ; 
				     if E.isMarked p
					 then cut p
				     else E.mark p)
				 end
			else ()
	in cut e
	end

fun decreaseKey(h, e, k) =
    if Key.greater(k, E.key e) then error "decreaseKey"
    else (E.setKey(e, k) ; updateMin(h, e))

fun decreaseKeySift(h, e, k) = (decreaseKey(h, e, k) ;
			      sift(h, e, k))
fun decreaseKeyCut(h, e, k) = (decreaseKey(h, e, k) ;
			     cut(h, e, k))

(*--------------------------------------------------------*)
(*                         Delete                         *)
(*--------------------------------------------------------*)

fun delete(decreaseKey, h, e) =
   let val k = E.key e
   in (decreaseKey(h, e, K.min) ;
       deleteMin h ;
       E.setKey(e, k))
   end

fun deleteSift(h, e) = delete(decreaseKeySift, h, e)
fun deleteCut(h, e) = delete(decreaseKeyCut, h, e)

(*--------------------------------------------------------*)
(*                         Union                          *)
(*--------------------------------------------------------*)

fun union(h, h') =
    (setSize(h, size h + size h') ;
     CL.splice(roots h, roots h'))

fun unionEager(h, h') = (union(h, h') ; consolidate h)

fun unionLazy(h, h') = (union(h, h') ;
		       updateMin(h, min h') handle Min => ())

(*--------------------------------------------------------*)
(*                         Output                         *)
(*--------------------------------------------------------*)

fun output(heap, outputValue, out) = unimplemented "output"

(*--------------------------------------------------------*)
(*                    Well-Formed Test                    *)
(*--------------------------------------------------------*)
(*
local
   fun sizeInTree e = 1 + sizeInTrees (CL.T (E.children e))
   and sizeInTrees l = CL.fold l 0 (fn (e,n) => n + sizeInTree e)
   fun sizeInHeap h = sizeInTrees (roots h)
   fun findMin h =
      let val min = ref NONE
	 fun updateMin e = (case !min of
			       NONE => min := SOME e
			     | SOME e' => if Key.less (E.key e) (E.key e')
					     then min := SOME e
					  else ())
      in (CL.foreach (roots h) updateMin ;
	  case !min of
	     SOME e => e
	   | NONE => bug "findMin") 
      end
   fun isTreeWellFormed e =
      let fun isChildWellFormed e' = (E.equal e (E.parent e')
				      andalso Key.leq (E.key e) (E.key e')
				      andalso isTreeWellFormed e')
	 val cs = CL.T (E.children e)
      in E.numChildren e = CL.length cs
	 andalso CL.forall cs isChildWellFormed
      end
in
   fun isFibonacciHeap h =
      CL.forall (roots h) isTreeWellFormed
      andalso size h = sizeInHeap h
      andalso (isEmpty h
	       orelse Key.equal (E.key (min h)) (E.key (findMin h)))
end
*)
end

