(*-------------------------------------------------------------------*)
(*                               List                                *)
(*-------------------------------------------------------------------*)

functor List(structure L: BASIC_LIST_ITER): LIST =
struct

structure S = DestructableSequence(structure S = L)
   
open Int L S

val head = first
val tail = rest
val nthTail = nthRest

fun single x = cons(x, empty())
   
fun iterateDestruct(l, f) = iterate(l, f o destruct)

fun mapAppend(l, f, l') =
   iterateDestruct(l, fn NONE => Rest l'
                       | SOME(x, l) => First(f x, l))

fun map(l, f) = mapAppend(l, f, empty())

fun map2(s, s', f) =
   iterate((s, s'), fn (s, s') =>
	   case (destruct s, destruct s') of
	      (NONE, NONE) => Empty
	    | (SOME(x, s), SOME(x', s')) => First(f(x, x'), (s, s'))
	    | _ => Error.error "map2")

fun map3(s, s', s'', f) =
   iterate((s, s', s''), fn (s, s', s'') =>
	   case (destruct s, destruct s', destruct s'') of
	      (NONE, NONE, NONE) => Empty
	    | (SOME(x, s), SOME(x', s'), SOME(x'', s'')) =>
		 First(f(x, x', x''), (s, s', s''))
	    | _ => Error.error "map3")

fun map4(s1, s2, s3, s4, f) =
   iterate((s1, s2, s3, s4), fn (s1, s2, s3, s4) =>
	   case (destruct s1, destruct s2, destruct s3, destruct s4) of
	      (NONE, NONE, NONE, NONE) => Empty
	    | (SOME(x1, s1), SOME(x2, s2), SOME(x3, s3), SOME(x4, s4)) =>
		 First(f(x1, x2, x3, x4), (s1, s2, s3, s4))
	    | _ => Error.error "map4")

fun tabulate(s, f) =
   iterate((s, zero), fn (s, i) =>
	   case destruct s of
	      NONE => Empty
	    | SOME(x, s) => First(f(i, x), (s, add1 i)))
   
fun appendThunk(s, t) =
   iterateDestruct(s, fn NONE => Rest(t())
                       | SOME(x, s) => First(x, s))

fun append(s, s') = appendThunk(s, fn () => s')
   
fun flattenMap(l, f) =
   let
      fun fm l = case destruct l of
	 NONE => empty()
       | SOME(x, l) => appendThunk(f x, fn () => fm l)
   in fm l
   end

fun reverseAppend(l, l') = fold(l, l', cons)

fun reverse s = reverseAppend(s, empty())

fun foldr(l, b, f) = fold(reverse l, b, f)

fun mapi(l, f) = reverse(#1(fold(l, (empty(), 0), fn (x, (l, i)) =>
				(cons(f(i, x), l), i + 1))))

fun zip(l, l') =
   case (destruct l, destruct l') of
      (SOME(x, l), SOME(x', l')) => cons((x, x'), zip(l, l'))
    | (NONE, NONE) => empty()
    | _ => Error.error "zip"

fun zip3(l1, l2, l3) =
   case (destruct l1, destruct l2, destruct l3) of
      (SOME(x1, l1), SOME(x2, l2), SOME(x3, l3)) =>
	 cons((x1, x2, x3), zip3(l1, l2, l3))
    | (NONE, NONE, NONE) => empty()
    | _ => Error.error "zip3"

fun unzip l =
   foldr(l, (empty(), empty()), fn ((x , y), (xs, ys)) =>
	 (cons(x, xs), cons(y, ys)))

fun unzip3 l =
   foldr(l, (empty(), empty(), empty()), fn ((x, y, z), (xs, ys, zs)) =>
	 (cons(x, xs), cons(y, ys), cons(z, zs)))

fun flatten s =
   case destruct s of
      NONE => empty()
    | SOME(x, s) => appendThunk(x, fn () => flatten s)

fun remFst(s, p,notFound) = 
   iterateDestruct(s, fn NONE => notFound()
		       | SOME(x, s) => if p x then Rest s
				      else First(x, s))

fun removeFirst(s, p) =
   remFst(s, p, fn () => Error.error "removeFirst")

fun remove(s, p) = remFst(s, p, fn () => Empty)

fun peekRemove(s, p) =
   let
      fun loop(s, accum) =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) =>
	       if p x
		  then SOME(reverseAppend(accum, s), x)
	       else loop(s, cons(x, accum))
   in loop(s, empty())
   end

fun splitLast s =
   case destruct(reverse s) of
      NONE => Error.error "splitLast"
    | SOME(x, s) => (reverse s, x)
   
fun allButLast s = #1(splitLast s)
   
fun partition(l, p) =
   let
      fun loop(s, accum as {yes,no}) =
	 case destruct s of
	    NONE => accum
	  | SOME(x, xs) =>
	       loop(xs,
		    if p x then {yes = cons(x, yes), no = no}
		    else {yes = yes, no = cons(x,no)})
   in loop(reverse l, {yes = empty(), no = empty()})
   end

fun keepAll(l, p) = #yes(partition(l, p))
fun removeAll(l, p) = #no(partition(l, p))

fun keepAllMap(s, f) =
   let
      fun loop s =
	 case destruct s of
	    NONE => empty()
	  | SOME(x, s) =>
	       case f x of
		  NONE => loop s
		| SOME y => cons(y, loop s)
   in loop s
   end

fun separate(s, x) =
   case destruct s of
      NONE => empty()
    | SOME(y, s) =>
	 cons(y,
	      let
		 fun loop s =
		    case destruct s of
		       NONE => empty()
		     | SOME(y, s) => cons(x, cons(y, loop s))
	      in loop s
	      end)
	 
fun duplicate(n, f) =
   iterate(n, fn n => if isZero n then Empty
		     else First(f(), sub1 n))

fun firstN(s,n) =
   iterate((s,n), fn (s,n) =>
	   if isZero n then Empty
	   else case destruct s of
	      NONE => Error.error "firstN"
	    | SOME(x, s) => First(x, (s, sub1 n)))

fun firstAtMostN(s,n) =
   iterate((s,n), fn (s,n) =>
	   if isZero n then Empty
	   else case destruct s of
	      NONE => Empty
	    | SOME(x, s) => First(x, (s, sub1 n)))

fun splitAtMost(l,n) =
   if isZero n then SOME(empty(), l)
   else case destruct l of
      NONE => NONE
    | SOME(x, l) => (case splitAtMost(l, sub1 n) of
		       NONE => NONE
		     | SOME(l, l') => SOME(cons(x, l), l'))

fun prefixes l =
   cons(empty(), case destruct l of
	NONE => empty()
      | SOME(x, l) => map(prefixes l, fn p => cons(x, p)))

fun suffixes l = map(prefixes(reverse l), reverse)

fun each l =
   zip3(allButLast(prefixes l),
	l,
	tail(reverse(suffixes l)))

fun cross l =
   case destruct l of
      NONE => cons(empty(), empty())
    | SOME(x, l) =>
	 let val rest = cross l
	 in flattenMap(x, fn x => map(rest, fn t => cons(x, t)))
	 end
   
fun listTo [] = empty()
  | listTo(x :: l) = cons(x, listTo l)

fun 'a ordered{< : 'a * 'a -> bool} =
   let
      fun x > y = y < x

      fun insert(s, x) =
	 case destruct s of
	    NONE => single x
	  | SOME(y, s') =>
	       if x < y then cons(x, s)
	       else cons(y, insert(s', x))

      fun insertionSort s =
	 case destruct s of
	    NONE => empty()
	  | SOME(x, s) => insert(insertionSort s, x)

      fun partition (ns, x) =
	 fold(ns, (empty(), empty()),
	      fn (y, (left, right)) =>
	      if x < y then (left, cons(y, right))
	      else (cons(y, left), right))
    
      local  
	 val columnSize = fromInt 5
	 val sort = insertionSort
	 fun breakIntoColumns ns =
	    if Int.<(length ns, columnSize)
	       then cons(ns, empty())
	    else cons(firstN(ns, columnSize),
		      breakIntoColumns(nthRest(ns, columnSize)))
      in
	 fun median ns = orderStatistic(ns, Int.quot(length ns, two))
	 and orderStatistic(ns, i) =
	    if Int.<(length ns, columnSize)
	       then nth(sort ns, i)
	    else let
		    val medianOfMedians =
		       median(map(breakIntoColumns ns, columnMedian))
		    val (left, right) = partition(ns, medianOfMedians)
		    val numLeft = length left
		 in if Int.<(i,numLeft)
		       then orderStatistic(left, i)
		    else orderStatistic(right, i - numLeft)
		 end
	 and columnMedian ns =
	    nth(sort ns, Int.quot(length ns, two))
      end    

      fun choose (op <) (s,n) =
	 let fun insert(x, s) =
	    let fun insert(m, s) =
	       if Int.equals(m,n) then empty()
	       else case destruct s of
		  NONE => single x
		| SOME(y, s) => if x < y then cons(x, s)
			       else cons(y, insert(add1 m, s))
	    in insert(zero, s)
	    end
	 in firstN(fold(s, empty(), insert),n)
	 end
      
      val smallest = choose (op <)
      val largest = choose (op >)
    
      fun getFirst(l, extreme,name) =
	 case destruct(extreme(l, one)) of
	    SOME(x, _) => x
	  | _ => Error.error name

      fun min l = getFirst(l, smallest, "min")
      fun max l = getFirst(l, largest, "max")

   in {insert = insert,
       insertionSort = insertionSort,
       median = median,
       orderStatistic = orderStatistic,
       partition = partition,
       max = max,
       min = min,
       largest = largest,
       smallest = smallest}
   end

fun equivalence(xs, equal) =
   fold
   (xs, empty(), fn (x, xss) =>
    let
       fun insert l =
	  case destruct l of
	     NONE => single(single x)
	   | SOME(xs, xss) =>
		case destruct xs of
		   SOME(x', _) => if equal(x, x')
				    then cons(cons(x, xs), xss)
				 else cons(xs, insert xss)
		 | NONE => Error.bug "List.equivalence"
    in insert xss
    end)

(* subsets of exactly size n *)
fun subsets(s,n) =
   let
      fun subs(s, size, n, elts, accum) =
	 if Int.isZero n
	    then cons(elts, accum)
	 else case destruct s of
	    SOME(x, xs) =>
	       if Int.equals(size,n) then cons(cons(x, append(xs, elts)),
					  accum)
	       else subs(xs, Int.sub1 size, Int.sub1 n,
			 cons(x, elts),
			 subs(xs, Int.sub1 size, n, elts , accum))
	  | _ => Error.error "subsets"
   in subs(s, size s, n, empty(), empty())
   end

fun power s = subsets(s, size s)
   
fun 'a set{equal: 'a * 'a -> bool,
	   layout: 'a -> Layout.t} =
   let
      fun equalTo x x' = equal(x, x')
      fun contains(s, x) = exists(s, equalTo x)
      fun s <= s' = forall(s, fn x => contains(s', x))
      fun s' >= s = s <= s'
      val equals = fn (s, s') => s <= s' andalso s' <= s
      fun s < s' = s <= s' andalso exists(s', fn x => not(contains(s, x)))
      fun s' > s = s < s'
      fun singleton x = cons(x, empty())
      fun add(s, x) = if contains(s, x) then s else cons(x, s)
      fun areDisjoint(s, s') = forall(s, fn x => not(contains(s', x)))
      val subset = keepAll
      fun subsetSize(s, p) = fold(s, 0, fn (x,n) => if p x then n + 1 else n)
      fun s - s' = subset(s, fn x => not(contains(s', x)))
      fun s + s' = append(s - s', s')
      fun intersect(s, s') = subset(s, fn x => contains(s', x))
      fun unions ss = fold(ss, empty(), op +)
      val size = length
	 
      val layout = fn vs =>
	 let open Layout
	 in seq[str "{",
		align(separateRight(toList(map(vs, layout)), ", ")),
		str "}"]
	 end

      val remove = fn (s, x) => remove(s, equalTo x)
   
      fun replace(s, f) =
	 fold(s, empty(), fn (x, s) => (case f x of
					NONE => s
				      | SOME y => add(s, y)))
      fun map(s, f) = replace(s, fn x => SOME(f x))

   in {empty = empty(),
       singleton = singleton,
       size = size,
       equals = equals,
       <= = op <=,
       >= = op >=,
       < = op <,
       > = op >,
       + = op +,
       - = op -,
       intersect = intersect,
       unions = unions,
       add = add,
       remove = remove,
       contains = contains,
       power = power,
       subsets = subsets,
       areDisjoint = areDisjoint,
       subset = subset,
       subsetSize = subsetSize,
       map = map,
       replace = replace,
       layout = layout}
   end
      
fun toList l = fold(l, [], op ::)

fun compare(l, l', comp) =
   let
      fun compare(l, l') =
	 case (destruct l, destruct l')  of
	      (NONE,       NONE)         => EQUAL
	    | (_,          NONE)         => GREATER
	    | (NONE,       _)            => LESS
	    | (SOME(x, l), SOME(x', l')) => (case comp(x, x') of
						EQUAL => compare(l, l')
					      | r => r)
   in compare(l, l')
   end
	 
end
