(*-------------------------------------------------------------------*)
(*                            StrictList                             *)
(*-------------------------------------------------------------------*)

structure StrictList: STRICT_LIST =
   StrictList(structure List = ListIter(BasicStrictList()))

structure StrictList: sig
			 include STRICT_LIST
			 val revMap: 'a t * ('a -> 'b) -> 'b t
		      end =
struct

open StrictList

val isEmpty =
   fn [] => true
    | _ => false

fun fold(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    [] => b
	  | x :: l => loop(l, f(x, b))
   in loop(l, b)
   end

fun fold2(l1, l2, b, f) =
   let
      fun loop(l1, l2, b) =
	 case (l1,       l2)       of
  	      ([],       [])       => b
	    | (x1 :: l1, x2 :: l2) => loop(l1, l2, f(x1, x2, b))
	    | _                    => Error.bug "fold2"
   in loop(l1, l2, b)
   end

fun foldi(l, b, f) = #1(fold(l, (b, 0), fn (x, (b, i)) => (f(i, x, b), i + 1)))

fun foreachi(l, f) = foldi(l, (), fn (i, x, ()) => f(i, x))
							 
fun foreach(l, f) = fold(l, (), f o #1)

fun foreach2(l1, l2, f) = fold2(l1, l2, (), fn (x1, x2, ()) => f(x1, x2))

fun reverseAppend(l1, l2) = fold(l1, l2, op ::)

fun reverse l = reverseAppend(l, [])

fun foldr(l, b, f) = fold(reverse l, b, f)

fun foldr2(l1, l2, b, f) = fold2(reverse l1, reverse l2, b, f)

fun append(l1, l2) = foldr(l1, l2, op ::)

(* careful to apply the function left-to-right *)
fun revMap(l, f) = fold(l, [], fn (x, l) => f x :: l)
fun map(l, f) = reverse(revMap(l, f))

fun map2(l1, l2, f) = reverse(fold2(l1, l2, [], fn (x1, x2, l) => f(x1, x2) :: l))

fun forall(l, f) =
   let
      val rec loop =
	 fn [] => true
	  | x :: l => f x andalso loop l
   in loop l
   end

fun exists(l, f) = not(forall(l, not o f))

local
   fun make(l1, l2, f, unequal) =
      let
	 val rec loop =
	    fn ([],     [])         => true
	     | (x1 :: l1, x2 :: l2) => f(x1, x2) andalso loop(l1, l2)
	     | _                    => unequal()
      in loop(l1, l2)
      end
in
   fun forall2(l1, l2, f) = make(l1, l2, f, fn () => Error.bug "forall2")
   fun equals(l1, l2, f) = make(l1, l2, f, fn () => false)
end

val rec nth =
   fn (x :: _, 0) => x
    | (_ :: l, n) => nth(l, n - 1)
    | _ => Error.bug "nth"

fun empty() = []

val cons = op ::

fun compare(l, l', comp) =
   let
      val rec compare =
	 fn ([],     [])       => EQUAL
	  | (_,      [])       => GREATER
	  | ([],     _)        => LESS
	  | (x :: l, x' :: l') => (case comp(x, x') of
				      EQUAL => compare(l, l')
				    | r => r)
   in compare(l, l')
   end

fun allButLast l =
   case reverse l of
      _ :: l => reverse l
    | _ => Error.bug "allButLast"

fun last l =
   case l of
      [x] => x
    | _ :: l => last l
    | _ => Error.bug "last"

fun length l = fold(l, 0, fn (_, n) => n + 1)

fun zip(l1, l2) = foldr2(l1, l2, [], fn (x1, x2, l) => (x1, x2) :: l)

fun unzip l = foldr(l, ([], []), fn ((x1, x2), (l1, l2)) =>
		    (x1 :: l1, x2 :: l2))

fun peek(l, f) =
   let
      val rec loop =
	 fn [] => NONE
	  | x :: l => if f x then SOME x else loop l
   in loop l
   end

fun flatten l = foldr(l, [], append)

fun duplicate(n, f) =
   let
      fun loop(n, accum) =
	 if n = ~1
	    then accum
	 else loop(n - 1, f() :: accum)
   in loop(n - 1, [])
   end
   
(* TODO
equivalence
flattenMap
fromTo
index
keepAll
keepAllMap
keepFirst
layout
mapAppend
mapFromTo
mapi
ordered
remove
removeAll
separate
set
single
splitLast
tabulate
*)

end

structure List = StrictList
