functor FileSys(val tempPrefix: string) =
struct

structure OS = OS
structure FS = OS.FileSys
structure P = OS.Path
   
structure Directory =
   struct
      datatype t = T of string list

      fun arcs(T a) = a

      val root = T[]

      fun parent(T ss) =
	 case ss of
	    [] => Error.error "parent of root directory"
	  | _ => T(List.allButLast ss)
	 
      fun split s = List.removeAll(#arcs(P.fromString s),
				   fn s => String.equals(s, ""))

      fun current() = T(split(FS.getDir()))

      fun dirOf s =
	 if String.size s > 0 andalso Char.equals(#"/", String.sub(s, 0))
	    then root
	 else current()

      fun relative(T d, s) = T(List.append(d, split s))

      fun fromString s = relative(dirOf s, s)

      fun toString(T d) =
	 P.toString{arcs = d, isAbs = true, vol=""} ^ "/"

      val layout = Layout.str o toString

      local fun convert f = f o toString
      in val remove = convert FS.rmDir
	 val change = convert FS.chDir
	 val make = convert FS.mkDir
      end

      fun inDir(d, th) =
	 let val cur = current()
	 in change d ;
	    DynamicWind.dynamicWind(th, fn () => change cur)
	 end
   end
structure D = Directory
	 
structure File =
   struct
      datatype t = T of {dir: D.t, file: string}
      type dir = D.t

      fun appendSuffix(T{dir, file}, suf) = T{dir = dir, file = file ^ suf}
	 
      fun dir(T{dir, ...}) = dir
      fun name(T{file, ...}) = file
	 
      fun relative(d, s) =
	 let val {dir, file} = P.splitDirFile s
	 in T{dir = D.relative(d, dir), file = file}
	 end

      fun fromString s = relative(D.dirOf s, s)

      fun toString(T{dir, file}) = D.toString dir ^ file

      local fun unary f = f o toString
      in val layout = unary Layout.str
	 val size = unary FS.fileSize
	 val modTime = unary FS.modTime
      end

      fun isNewer(f1, f2) = Time.>=(modTime f1, modTime f2)

      fun withh(file, p, openn, close) =
	 let val f = toString file
	    val stream =
	       (openn f) handle IO.Io _ => Error.error("cannot open " ^ f)
	 in DynamicWind.dynamicWind(fn () => p stream,
				    fn () => close stream)
	 end 
      
      fun withOut(f, p) = withh(f, p, Out.openOut, Out.close)
      fun withAppend(f, p) = withh(f, p, Out.openAppend, Out.close)
      fun withIn(f, p) = withh(f, p, In.openIn, In.close)

      local
	 fun can withh f = withh(f, fn _ => true) handle _ => false
         fun ensure(pred, msg) f =
	    if pred f then ()
	    else Error.error("can not " ^ msg ^ " " ^ toString f)
      in
	 val canWrite = can withOut
	 val canRead = can withIn
	 val ensureWrite = ensure(canWrite, "write")
	 val ensureRead = ensure(canRead, "read")
      end

(*	 let val f = toString f
	 in if FS.access(f, [FS.A_WRITE])
	       then ()
	    else Error.error("can not write " ^ f)
	 end*)
      (*fun canRead f =
       (In.closeIn(In.openIn f) ; true)
       handle IO.Io _ => false
	  *)

      fun sameContents(f1, f2) =
	 size f1 = size f2
	 andalso withIn(f1, fn in1 =>
			withIn(f2, fn in2 =>
			       In.sameContents(in1, in2)))

      fun output(file, out) = Out.output(out, toString file)
	 
      fun outputContents(file, out) =
	 withIn(file, fn ins =>
		while not(In.endOf ins)
		   do Out.output(out, In.input ins))

      fun lines f =
	 withIn(f, fn ins =>
		let
		   fun loop() =
		      if In.endOf ins then []
		      else (String.dropLast(In.inputLine ins)) :: (loop())
		in loop()
		end)
   
      fun contents file = withIn(file, In.inputAll)

      fun move{from, to} =
	 Process.system("mv " ^ toString from ^ " " ^ toString to)
	 
      fun copy(source, dest) =
	 withOut(dest, fn out => outputContents(source, out))
	 
      fun concat(sources, dest) =
	 withOut(dest, fn out =>
		 List.foreach(sources, fn f => outputContents(f, out)))
	 
      val remove = FS.remove o toString

      val temp = fromString o OS.FileSys.tmpName

      fun withTemp p =
	 let val f = temp()
	 in DynamicWind.dynamicWind
	    (fn () => p f,
	     fn () => remove f)
	 end

      fun stringify processFile s =
	 withTemp(fn f =>
		  (withOut(f, fn out => Out.output(out,s)) ;
		   processFile f))

      fun withString(s, f) =
	 withTemp(fn tmp =>
		  (withOut(tmp, fn out => Out.output(out,s)) ;
		   f tmp))
		  
      fun withOutIn(fout, fin) =
	 withTemp(fn tmp =>
		  (withOut(tmp, fout),
		   withIn(tmp, fin)))

      fun withStringIn(s, fin) =
	 #2(withOutIn(fn out => Out.output(out, s),
		      fin))

      fun ignore(out, thunk) =
	 withOut(fromString "/dev/null", fn out' =>
		 Out.fluidLet(out, out', thunk))

      fun create f = withOut(f, fn _ => ())
   end
structure F = File

structure Directory =
   struct
      open D
      type file = File.t

      fun list d =
	 let val dstr = toString d
	    val stream = FS.openDir dstr
	    val _ = FS.readDir stream
	    val _ = FS.readDir stream
	    fun read() =
	       let val s = FS.readDir stream
	       in if s = "" then ([], [])
		  else let val (dirs, files) = read()
			   val x = dstr ^ s
		       in if FS.isLink x then (dirs, files)
			  else
			     if FS.isDir x
				then (relative(d, s) :: dirs, files)
			     else (dirs, F.relative(d, s) :: files)
		       end
	       end
	 in DynamicWind.dynamicWind(read, fn () => FS.closeDir stream)
	 end
      
      fun dirs d = #1(list d)
	 
      fun files d = #2(list d)
	 
      fun foreachFile(d, p) =
	 let
	    fun foreach d =
	       let val (ds, fs) = list d
	       in (List.foreach(fs, p) ;
		   List.foreach(ds, foreach))
	       end
	 in foreach d
	 end

      fun removeR d =
	 let val (ds, fs) = list d
	 in List.foreach(ds, removeR) ;
	    List.foreach(fs, File.remove) ;
	    remove d
	 end

      fun temp() = fromString(FS.tmpName())
	 
      fun inTemp thunk =
	 let val d = temp()
	 in make d ;
	    DynamicWind.dynamicWind
	    (fn () => inDir(d, fn _ => thunk()),
	     fn () => removeR d)
	 end
   end

type file = F.t
type dir = D.t
   
end

structure FileSys = FileSys(val tempPrefix = "/tmp/smltmp")
structure File: FILE = FileSys.File
structure Dir: DIR = FileSys.Directory
