(*-------------------------------------------------------------------*)
(*                       DestructableSequence                        *)
(*-------------------------------------------------------------------*)

functor DestructableSequence(structure S: BASIC_SEQUENCE)
  : DESTRUCTABLE_SEQUENCE =
struct

open Int S

fun isEmpty s =
   case destruct s of
      NONE => true
    | SOME _ => false

fun first s =
   case destruct s of
      SOME(x, _) => x
    | NONE => Error.error "first"

fun rest s =
   case destruct s of
      SOME(_, s) => s
    | NONE => Error.error "rest"

fun last s =
   case destruct s of
      NONE => Error.error "last"
    | SOME xs =>
	 let
	    fun last(x, s) =
	       case destruct s of
		  SOME xs => last xs
		| NONE => x
	 in last xs
	 end

fun fold(s, b, f) =
   let
      fun fold(s, b) =
	 case destruct s of
	    NONE => b
	  | SOME(x, s) => fold(s, f(x, b))
   in fold(s, b)
   end

fun foldi(s, b, f) =
   #2(fold(s, (0, b), fn (x, (i, b)) => (i + 1,  f(i, x, b))))

fun fold2(s, s', b, f) =
   let
      fun fold(s, s', b) =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => b
	  | (SOME(x, s), SOME(x', s')) => fold(s, s', f(x, x', b))
	  | _ => Error.error "fold2"
   in fold(s, s', b)
   end

fun fold3(s1, s2, s3, b, f) =
   let
      fun fold(s1, s2, s3, b) =
	 case (destruct s1, destruct s2, destruct s3) of
	    (NONE, NONE, NONE) => b
	  | (SOME(x1, s1), SOME(x2, s2), SOME(x3, s3)) => fold(s1, s2, s3, f(x1, x2, x3, b))
	  | _ => Error.error "fold3"
   in fold(s1, s2, s3, b)
   end

fun size s = fold(s, zero, fn (_,n) => add1 n)
val length = size

fun nthRest(s,n) = 
   if isZero n then s
   else case destruct s of
      NONE => Error.error "nthRest"
    | SOME(_, s) => nthRest(s, sub1 n)
      
fun nth(s,n) = first(nthRest(s,n))

fun index(s, p) =
   let
      fun index(s,n) =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) => if p x then SOME n else index(s, add1 n)
   in index(s, zero)
   end
   
fun foldr(s, b, f) =
   let
      fun fold s =
	 case destruct s of
	    NONE => b
	  | SOME(x, s) => f(x, fold s)
   in fold s
   end

fun toList s = foldr(s, [], fn (x, l) => x :: l)

fun foldr2(s, s', b, f) =
    let
       fun fold(s, s') =
	  case (destruct s, destruct s') of
	     (NONE, NONE) => b
	   | (SOME(x, s), SOME(x', s')) => 
		f(x, x', fold(s, s'))
	   | _ => Error.error "foldr2"
    in fold(s, s')
    end

fun foreach(s, f) =
   let
      fun foreach s =
	 case destruct s of
	    NONE => ()
	  | SOME(x, s) => (f x ; foreach s)
   in foreach s
   end

fun foralli(s, f) =
   let
      fun forall(s, i) =
	 case destruct s of
	    NONE => true
	  | SOME(x, s) => f(i, x) andalso forall(s, add1 i)
   in forall(s, zero)
   end
			  
fun foreachi(s, f) =
   let
      fun foreach(s, i) =
	 case destruct s of
	    NONE => ()
	  | SOME(x, s) => (f(i, x) ;
			  foreach(s, add1 i))
   in foreach(s, zero)
   end

(*
fun foldl2Cond(l, l', b, f) =
    let fun fold(l, l', b) =
	if isEmpty l then b
	else let val (x, l) = destruct l
		 val (x', l') = destruct l'
		 val (b, continue) = f(b, x, x')
	     in if continue then fold(l, l', b) else b
	     end
    in fold(l, l', b)
    end

fun foldl2(l, l', b, f) = foldl2Cond(l, l', b, fn a => (f a, true))
*)

fun foreach2(s, s', f) =
   let fun foreach(s, s') = (case (destruct s, destruct s') of
			       (NONE, NONE) => ()
			     | (SOME(x, s), SOME(x', s')) => 
				  (f(x, x') ; foreach(s, s'))
			     | _ => Error.error "foreach2")
   in foreach(s, s')
   end

fun foreach3(s, s', s'', f) =
   let fun foreach(s, s', s'') =
      (case (destruct s, destruct s', destruct s'') of
	  (NONE, NONE, NONE) => ()
	| (SOME(x, s), SOME(x', s'), SOME(x'', s'')) => 
	     (f(x, x', x'') ; foreach(s, s', s''))
	| _ => Error.error "foreach3")
   in foreach(s, s', s'')
   end

fun forall(s, f) =
   let
      fun forall s =
	 case destruct s of
	    NONE => true
	  | SOME(x, s) => f x andalso forall s
   in forall s
   end

fun forall2(s, s', f) =
   let
      fun forall(s, s') =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => true
	  | (SOME(x, s), SOME(x', s')) => f(x, x') andalso forall(s, s')
	  | _ => Error.error "forall2"
   in forall(s, s')
   end

fun peek(s, p) =
   let
      fun peek s =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) => if p x then SOME x
			 else peek s
   in peek s
   end

fun lookup arg = 
   case peek arg of
      SOME x => x
    | NONE => Error.error "lookup"

(*fun sameLength ss =
   case destruct s of
      NONE => true
    | SOME(s, ss) =>
	 let val n = length s
	 in forall(ss, fn s => I.=(n, length s))
	 end*)

fun exists(l, p) =
    case peek(l, p) of
	NONE => false
      | SOME _ => true

fun contains(l, x, eq) = exists(l, fn x' => eq(x, x'))

fun equals(s, s', eltEq) =
   let
      fun eq(s, s') =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => true
	  | (SOME(x, s), SOME(x', s')) => eltEq(x, x') andalso eq(s, s')
	  | _ => false
   in eq(s, s')
   end

fun toList xs =
   case destruct xs of
      NONE => []
    | SOME(x, xs) => x :: (toList xs)
      
local open Layout
in
   fun layoutSep(sep, layoutX) xs = list(map layoutX (toList xs))

   fun layout l = layoutSep(",", l)
(*   case destruct xs of
      NONE => empty
    | SOME(x, s) =>
	 align(layoutX x ::(let
			     fun loop s =
				case destruct s of
				   NONE => []
				 | SOME(x, s) => str sep :: layoutX x :: loop s
			  in loop s
			  end))
	 *)
end
			 
fun output(separator, outputElt) (s, out) =
   case destruct s of
      NONE => ()
    | SOME(x, s) => (outputElt(x, out) ;
		    foreach(s, fn x =>
			    (Out.output(out, separator) ;
			     outputElt(x, out))))

fun outputN(n, sep, outputElt) (s, out) =
   if isZero n then ()
   else case destruct s of
      NONE => ()
    | SOME (x, s) =>
	 (outputElt(x, out) ;
	  let
	     fun loop(s,n) =
		if isZero n then ()
		else case destruct s of
		   NONE => ()
		 | SOME(x, s) => (Out.output(out, sep) ;
				 outputElt(x, out) ;
				 loop(s, sub1 n))
	  in loop(s, sub1 n)
	  end)

end
