functor Control(structure StringMap: STRING_MAP
                datatype status = Always | Flagged | Never
		type flags
		val map: flags StringMap.t
		val getFlag: flags -> bool ref
		val default: bool ref
		val status: status) =
struct

val status = ref status

fun always() = status := Always
fun flagged() = status := Flagged
fun never() = status := Never

fun isOn b =
   case !status of
      Always => true
    | Flagged => !b
    | Never => false
   
val default = default

fun flag name = getFlag(StringMap.lookup(map,name))

fun whats b () =
   StringMap.keepAll(map, fn flags => b = !(getFlag flags))

val whatsOn = whats true
val whatsOff = whats false

fun sets b () =
   StringMap.foreach(map, fn flags => getFlag flags := b)
   
val all = sets true
val none = sets false

fun some b ss = List.foreach(ss, fn s => flag s := b)

val on = some true
val off = some false

end

(*-------------------------------------------------------------------*)
(*                               Trace                               *)
(*-------------------------------------------------------------------*)

functor Trace(structure IC: INTERMEDIATE_COMPUTATION
	      structure StringMap: STRING_MAP
	      structure Time: TIME
	      sharing type IC.Time.t = Time.t): TRACE =
struct

structure Timer = Time

val immediateDefault = ref false
val delayedDefault = ref false
val timeDefault = ref false

type flags = {immediate: bool ref,
	      delayed: bool ref,
	      time: bool ref}
   
val map = StringMap.new(fn () => {immediate = ref(!immediateDefault),
				  delayed = ref(!delayedDefault),
				  time = ref(!timeDefault)})
   
fun traceable() = StringMap.domain map

fun outputTraceable() =
   Layout.output(List.layout String.layout (traceable()),
		 Out.standard)

datatype status = Always | Flagged | Never

structure Immediate =
   Control(type flags = flags
	   val map = map
	   structure StringMap = StringMap
           datatype status = datatype status
	   fun getFlag({immediate, ...}: flags) = immediate
	   val default = immediateDefault
	   val status = Never)
structure Delayed =
   struct
      structure C = 
	 Control(type flags = flags
		 val map = map
		 structure StringMap = StringMap
		 datatype status = datatype status
		 fun getFlag({delayed, ...}: flags) = delayed
		 val default = delayedDefault
		 val status = Never)
      open C
	 
      val keepAll = ref true
   end

structure Time =
   Control(type flags = flags
	   val map = map
	   structure StringMap = StringMap
	   datatype status = datatype status
	   fun getFlag({time, ...}: flags) = time
	   val default = timeDefault
	   val status = Never)

fun never() = (Immediate.never() ;
	       Delayed.never() ;
	       Time.never())

fun always() = (Immediate.always() ;
		Delayed.always() ;
		Time.always())

fun flagged() = (Immediate.flagged() ;
		 Delayed.flagged() ;
		 Time.flagged())
   
fun reset() =
   StringMap.foreach(map, fn {immediate, delayed, time} =>
		     (immediate := false ;
		      delayed := false ;
		      time := false))
    (*StringMap.clear map*)

(*---------------------------------------------------*)
(*                 Delayed Feedback                  *)
(*---------------------------------------------------*)

structure Computation = IC.Computation

datatype comp =
   Working of IC.t
  | Finished of Computation.t

val emptyIc = IC.empty
fun empty() = Working(emptyIc())
   
val currentComputation = ref(empty())

fun clear() = currentComputation := empty()

fun finishedComputation() =
   case !currentComputation of
      Working ic => let val c = IC.finish ic
		    in currentComputation := Finished c ;
		       c
		    end
    | Finished c => c
val computation = finishedComputation

fun history() = Computation.output(finishedComputation(),
				   Out.error)
fun calls() = Computation.outputCalls(finishedComputation(),
				      Out.error)
fun times() = Computation.outputTimes(finishedComputation(),
				      Out.error)
fun inspect() = Computation.inspect(finishedComputation())

fun ic() =
   case !currentComputation of
      Finished _ => let val ic = emptyIc()
		    in currentComputation := Working ic ;
		       ic
		    end
    | Working ic => ic

fun delayedCall(name, layoutArg, layoutAns) =
   {call = fn () =>
    let val comp = ic()
       val comp = if !Delayed.keepAll orelse not(IC.atTopLevel comp)
		     then comp
		  else (clear() ; ic())
    in IC.call(comp, name, layoutArg)
    end,
    raisee = fn t => IC.raisee(ic(), t),
    return = fn (ans, t) => IC.return(ic(),
				     fn () => layoutAns ans,
				     t)}
	 
(*---------------------------------------------------*)
(*                Immediate Feedback                 *)
(*---------------------------------------------------*)

local
   val out = Out.error
   open Layout
      
   val indentation = ref 0
   val space = 3
   fun left() = indentation := !indentation - space
   fun right() = indentation := !indentation + space
   fun print l = (Layout.output(Layout.indent(l, !indentation), out) ;
		  Out.newline out)

   fun finish(t, res) =
      (left() ;
       print(case t of
		NONE => seq[str "==> ", res]
	      | SOME t =>
		 align[seq[str "==> time = ", Timer.layout t],
		       res]))
in
   fun immediateCall(name, outArg, layoutAns) =
      let fun call() = (print(seq[str name, str " ==> ",
				  outArg()
				  handle _ => str "<layout argument error>"]) ;
			right())
	 fun raisee t = finish(t, str "raise")
	 fun return(ans, t) =
	    finish(t, layoutAns ans handle _ => str "<layout answer error>") ;
      in {call = call,
	  raisee = raisee,
	  return = return}
      end

   structure Immediate =
      struct
	 open Immediate
	 fun message l =
	    (left() ;
	     indentation := 1 + !indentation; 
	     print l;
	     indentation := !indentation - 1;
	     right())
	 val messageStr = message o Layout.str
      end
end
	 
(*---------------------------------------------------*)
(*                  Instrumentation                  *)
(*---------------------------------------------------*)

val doTrace = true
   
fun trace(name, layoutArg, layoutAns) =
   if not doTrace
      then fn f => f
   else 
      let val {immediate, delayed, time} = StringMap.lookup(map,name)
      in fn f => fn arg =>
	 let
	    val immediate = Immediate.isOn immediate
	    val delayed = Delayed.isOn delayed
	    val time = Time.isOn time
	 in if not(immediate orelse delayed orelse time)
	       then f arg
	    else let val outArg = fn () => layoutArg arg
		     val noCall = {call = fn _ => (),
				   raisee = fn _ => (),
				   return = fn _ => ()}
		     val immed = if immediate
				    then immediateCall(name, outArg, layoutAns)
				 else noCall
		     val delay = if delayed
				    then delayedCall(name, outArg, layoutAns)
				 else noCall
		 in #call delay () ;
		    #call immed () ;
		    let
		       val startTime =
			  if time then SOME(Timer.times()) else NONE
		       fun getTime() =
			  case startTime of
			     NONE => NONE
			   | SOME t => SOME(let val {utime, stime, ...} = Timer.times()
					    in Timer.+(Timer.-(utime, #utime t),
						       Timer.-(stime, #stime t))
					    end)
		       val ans = f arg handle exn => let val t = getTime()
						     in #raisee delay t ;
							#raisee immed t ;
							raise exn
						     end
		       val t = getTime()
		    in #return delay (ans, t) ;
		       #return immed (ans, t) ;
		       ans
		    end
		 end
	 end
      end

fun ignore _ = Layout.empty
   
fun traceCall s f = trace(s, ignore, ignore) f
   
fun traceRec info =
   let val trace = trace info
   in fn f => let fun fix f a = trace (f (fix f)) a
	      in fix f
	      end
   end

fun trace0(name, layoutAns) =
   trace(name, Unit.layout, layoutAns)
	 
fun trace2(name, layout1, layout2, layoutAns) =
   trace(name, Layout.tuple2(layout1, layout2), layoutAns)

fun trace3(name, out1, out2, out3, outAns) =
   trace(name, Layout.tuple3(out1, out2, out3), outAns)

fun trace4(name, out1, out2, out3, out4, outAns) =
   trace(name, Layout.tuple4(out1, out2, out3, out4), outAns)

end
