(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Main: MAIN =
struct

structure PrimEnv = Compile.PrimEnv

fun message s = Out.output(Out.error, s)
   
exception Usage
fun 'a usage(s: string): 'a =
   (message s
    ; message "\nUsage: mlton [-cCdgmpSv] [-o file] [-k{a|c|s|o}*] [-O{0-9}]\n\
                \\t[-DDEFINE] [-iinclude] [-Idir] [-llibrary] [-Ldir]\n\
                \\t[-flatten n] [-inline n] [-no-polyvariance]\n\
		\\t[-unsafe] [-no-type-check]\n\
		\\t[-h heapSize[{k|m}]] [-s stackSize[{k|m}]]\n\
		\\tfile.{sml|c|s|o}\n"
    ; raise Usage)

structure Place =
   struct
      datatype t = SML | CoreML | C | S | O | OUT

      val toInt =
	 fn SML => 1
	  | CoreML => 2
	  | C => 3
	  | S => 4
	  | O => 5
	  | OUT => 6

      val toString =
	 fn SML => "sml"
	  | CoreML => "core-ml"
	  | C => "c"
	  | S => "s"
	  | O => "o"
	  | OUT => "out"

      val layout = Layout.str o toString

      fun compare(p, p') = Int.compare(toInt p, toInt p')

      fun start(s: string): t * string =
	 let
	    val rec loop =
	       fn [] => usage "invalid file suffix"
		| (suf, start) :: sufs =>
		     if String.isSuffix{string = s, suffix = suf}
			then (start, String.dropSuffix(s, size suf))
		     else loop sufs
	 in loop[(".sml", SML), (".c", C), (".s", S), (".o", O)]
	 end
   end

fun system(ss: string list): unit =
   let val s = concat(List.separate(ss, " "))
   in Control.messageStr s
      ; (case OS.Process.system s of
	    0 => ()
	  | _ => Error.error("call to system failed: " ^ s))
   end

fun commandLine(args: string list): unit =
   let
      fun error() = Error.bug "incorrect args from shell script"
      val (root, gcc, gccSwitches, args) =
	 case args of
	    root :: gcc :: args =>
	       let
		  fun loop(args, ac) =
		     case args of
			[] => error()
		      | arg :: args =>
			   if arg = "END"
			      then (rev ac, args)
			   else loop(args, arg :: ac)
		  val (gccSwitches, args) = loop(args, [])
	       in (Dir.fromString root, gcc, gccSwitches, args)
	       end
	  | _ => error()

      val debug = ref false
      val defines: string list ref = ref []
      val keepC = ref false
      val keepS = ref false
      val keepO = ref false
      val includes = ref []
      val includeDirs = ref []
      val lib = ref "mlton"
      val libs = ref []
      val libDirs = ref []
      val output = ref NONE
      val optimization = ref 1
      val stop = ref Place.OUT


      fun dooKeeps s =
	 Int.foreach(0, size s - 1, fn i =>
		     (case String.sub(s, i) of
			 #"a" => Control.aux
		       | #"c" => keepC
		       | #"s" => keepS
		       | #"o" => keepO
		       | _ => usage "invalid -k flag") := true)

      val rest =
	 let open Popt
	    fun push(r) = String(fn s => ListRef.push(r, s))
	 in parse{switches = args,
		  opts =
		  [
		   ("c", None(fn () => stop := Place.O)),
		   ("C", None(fn () => stop := Place.C)),
		   ("d", None(fn () => (Control.setMode Control.Defunctorize
					; stop := Place.CoreML))),
		   ("D", String(fn s => (ListRef.push(defines, s)
					 ; if s = "INSTRUMENT"
					      then Control.instrument := true
					   else ()))),
		   ("flatten", Int(fn n => Control.flattenRounds := n)),
		   ("g", None(fn () => (debug := true; lib := "mlton-gdb"))),
		   ("h", Mem(fn n => Control.fixedHeap := SOME n)),
		   ("i", push includes),
		   ("inline", Int Control.setInlineSize),
		   ("I", push includeDirs),
		   ("k", String dooKeeps),
		   ("l", push libs),
		   ("L", push libDirs),
		   ("no-polyvariance", None(fn () =>
					    Control.polyvariance := NONE)),
		   ("no-type-check", None(fn () => Control.setTypeCheck false)),
		   ("o", SpaceString(fn s => output := SOME(File.fromString s))),
		   ("O", Digit(fn i => optimization := i)),
		   ("p", None(fn () => (Control.profile := true
					; keepC := true
					; Control.keepCps := true))),
		   ("S", None(fn () => stop := Place.S)),
		   ("unsafe", None(fn () => Control.safe := false)),
		   ("v", None Control.messagesOn),
		   ("trace", String(fn s => let open Trace.Immediate
					    in flagged()
					       ; on[s]
					    end))
		   ]}
	 end

     in
	Control.includes := List.reverse(!includes)
	;
	case rest of
	   NONE => usage "error while parsing switches"
	 | SOME [] =>
	      let open Control
	      in if messagesAreOn()
		    then messageLay(header No)
		 else usage "must supply a file"
	      end
	 | SOME(input :: rest) =>
	      let val (start, base) = Place.start input
		 val stop = !stop
	      in case Place.compare(start, stop) of
		 GREATER => usage(concat["cannot go from ",
					 Place.toString start,
					 " to ",
					 Place.toString stop])
	       | EQUAL => usage "nothing to do"
	       | LESS =>
		    let
		       val tempFiles: File.t list ref = ref []
		       fun temp(suf: string): File.t =
			  let val f = File.fromString(OS.FileSys.tmpName() ^ suf)
			  in ListRef.push(tempFiles, f)
			     ; f
			  end
		       val input = File.fromString input
		       fun suffix s = File.fromString(base ^ s)
		       fun file(b, suf) =
			  if b then suffix suf
			  else temp suf
		       fun maybeOut suf =
			  case !output of
			     NONE => suffix suf
			   | SOME f => f
		       fun list(prefix: string, l: string list): string list =
			  List.map(l, fn s => prefix ^ s)
		       fun docc(input, output, switches, libs) =
			  system
			  (gcc :: switches
			   @ list("-D", !defines)
			   @ list("-I", rev(!includeDirs))
			   @ list("-L", rev(!libDirs))
			   @ [if !debug then "-g" else "",
				 "-O" ^ Int.toString(!optimization)]
			   @ gccSwitches
			   @ ["-o", File.toString output,
			      File.toString input]
			   @ libs)
		       fun compileO(input: File.t) =
			  case stop of
			     Place.O => ()
			   | Place.OUT =>
				docc(input, maybeOut "",
				     if !Control.profile
					then ["-p"]
				     else [],
				     list("-l", !lib :: !libs)
				     @ rest)
			   | _ => Error.bug "strange stop in compileO"
		       fun makeO(input: File.t) =
			  let
			     val output =
				case stop of
				   Place.O => maybeOut ".o"
				 | _ => if !keepO then suffix ".o"
					else temp ".o"
			  in docc(input, output,
				  "-c" :: (if !Control.profile
					      then ["-g"]
					   else []),
				  [])
			     ; compileO output
			  end
		       fun compileC(input: File.t) =
			  let
			     fun makeS output = docc(input, output, ["-S"], [])
			  in case stop of
			     Place.S => makeS(maybeOut ".s")
			   | _ => if !keepS
				     then let val output = suffix ".s"
					  in makeS output; makeO output
					  end
				  else makeO input
			  end
		       fun compile() =
			  case start of
			     Place.SML =>
				let
				   val output =
				      case stop of
					 Place.CoreML => maybeOut ".core-ml"
				       | Place.C => maybeOut ".c"
				       | _ => if !keepC then suffix ".c"
					      else temp ".c"
				in Compile.compileFile
				   {input = File.toString input,
				    output = File.toString output}
				   (* These collects are so that the heap
				    * is shrunk as much as possible before
				    * calling gcc
				    *)
				   ; MLton.GC.collect()
				   ; MLton.GC.collect()
				   ; (case stop of
					 Place.CoreML => ()
				       | Place.C => ()
				       | _ => compileC output)
				end
			   | Place.C => compileC input
			   | Place.S => makeO input
			   | Place.O => compileO input
			   | _ => Error.bug "invalid start"
		    in DynamicWind.dynamicWind
		       (compile, fn () => List.foreach(!tempFiles, File.remove))
		    end
	      end
   end 
   
fun exportNJ(root: Dir.t, file: File.t): unit =
   (PrimEnv.forceAll root
    ; Control.defaults()
    ; (SMLofNJ.exportFn
       (File.toString file,
	fn (_, args) => ((commandLine args; OS.Process.success)
			 handle _ => OS.Process.failure))))
   
fun exportMLton(): unit =
   case CommandLine.arguments() of
      [root, file] => exportNJ(Dir.fromString root, File.fromString file)
    | _ => Error.bug "usage: exportMLton root file"

end
