(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor InferType(S: INFER_TYPE_STRUCTS): INFER_TYPE =
struct

open S
local open Ast
in structure Aconst = Const
end
structure Field = Record.Field
structure Srecord = SortedRecord

structure Type =
   struct
      structure Set = DisjointSet

      (* tuples of length <> 1 are always represented as records.
       * there will never be tuples of length one
       *)
      datatype ty =
	 Unknown of unknown
       | Var of Tyvar.t
       | Con of Tycon.t * t list
       | Word (* an unresolved word type *)
       | Int (* an unresolved int type *)
       | Record of {flexible: bool, record: t Srecord.t}
      withtype t = {ty: ty,
		    plist: PropertyList.t} Set.t
      and unknown = {equality: bool, canGeneralize: bool}

      fun joinUnknown(r: unknown, r': unknown): unknown =
	 {equality = #equality r orelse #equality r',
	  canGeneralize = #canGeneralize r andalso #canGeneralize r'}

      val toType: t -> ty = #ty o Set.value
      val plist: t -> PropertyList.t = #plist o Set.value

      (* unknown is a bit strange because frees needs the reference to the
       * whole type (Set.t) and not just the unknown info
       *)
      fun makeHom{unknown, var, con, word, int, record} =
	 let
	    val {get, set, destroy} =
	       Property.newDest(plist, Property.initConst NONE)

	    fun loop t =
	       case get t of
		  SOME x => x
		| NONE =>
		     case toType t of
			Word => word
		      | Int => int
		      | Unknown u => unknown(u, t)
		      | Var a => var a
		      | Con(c, ts) => con(c, List.map(ts, loop))
		      | Record{flexible, record = r} =>
			   record{flexible = flexible,
				  record = Srecord.map(r, loop)}
	 in {hom = loop, destroy = destroy}
	 end

      fun hom{ty, unknown, var, con, word, int, record} =
	 let val {hom, destroy} = makeHom{unknown = unknown,
					  var = var,
					  con = con,
					  word = word,
					  int = int,
					  record = record}
	 in hom ty before destroy()
	 end

      fun deconOpt t =
	 case toType t of
	    Con x => SOME x
	  | _ => NONE

      fun layoutDetailed ty =
	 let open Layout
	 in hom{ty = ty,
		unknown = 
		fn ({equality, canGeneralize}, _) =>
		namedRecord("Unknown",
			    [("equality", Bool.layout equality),
			     ("canGeneralize", Bool.layout canGeneralize)]),
		var = fn a => paren(seq[str "Var ", Tyvar.layout a]),
		word = str "Word",
		int = str "Int",
		con = fn (c, ts) => paren(align[seq[str "Con ", Tycon.layout c],
						tuple ts]),
		record = fn {flexible, record} =>
		paren(seq([str "Record ",
			   Srecord.layout
			   {record = record,
			    separator = "=",
			    extra = if flexible then "..." else "",
			    layoutTuple = tuple,
			    layoutElt = fn x => x}]))}
      end

      local
	 structure T = Ast.Type
      in
	 fun toAst t =
	    case toType t of
	       Unknown _ => T.unit
	     | Var a => T.var a
	     | Word => T.con(Ast.Tycon.word, [])
	     | Int => T.con(Ast.Tycon.int, [])
	     | Con(c, ts) => T.con(Tycon.toAst c, List.map(ts, toAst))
	     | Record{flexible, record} => T.record(Srecord.map(record, toAst))
      end

      fun newTy(ty: ty): t =
	 Set.singleton{ty = ty,
		       plist = PropertyList.new()}

      val new = newTy o Unknown

      val record = newTy o Record

      val tuple =
	 fn [t] => t
	  | ts => record{flexible = false,
			 record = Srecord.tuple ts}

      val unit = tuple[]

      fun isUnit s =
	 case toType s of
	    Record{flexible = false, record} =>
	       (case Srecord.detupleOpt record of
		   SOME [] => true
		 | _ => false)
	  | _ => false

      val _ = if isUnit unit then () else Error.bug "unit isn't unit"
	  
      fun con(tycon, ts) =
	 if Tycon.equals(tycon, Tycon.tuple) then tuple ts
	 else newTy(Con(tycon, ts))

      val layout = Layout.switch{detailed = layoutDetailed,
				 normal = Ast.Type.layout o toAst}
   
      val equals: t * t -> bool = Set.equals
   end

structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
			   
open Ops Type (* order is important, since want specialized definitions
	       in Type to override general definitions in Ops *)

structure Free =
   struct
      structure F = Sum(structure X = Tyvar
			structure Y = Type)
      open F
      val fromTyvar = X
      val fromType = Y
   end

structure Frees =
   struct
      structure S = UnorderedSet(Free)
      open S

      fun fromTyvars tyvs = fromList(List.map(tyvs, Free.fromTyvar))
   end

local
   structure Env = MonoEnv(structure Domain = Tyvar
			   structure Range = Type)
in
   fun substitute(t: t, tyvarsAndTypes: (Tyvar.t * Type.t) list): t =
      let val env = Env.fromList tyvarsAndTypes
	 fun loop t =
	    case toType t of
	       Unknown _ => t
	     | Word => t
	     | Int => t
	     | Var a => (case Env.peek(env, a) of
				  NONE => t
				| SOME t => t)
	     | Con(c, ts) => newTy(Con(c, List.map(ts, loop)))
	     | Record{flexible, record} =>
		  if flexible
		     then Error.bug "substituting in flexible record type"
		  else newTy(Record{flexible = false,
				    record = Srecord.map(record, loop)})
      in loop t
      end
end

(* val substitute =
 *    Trace.trace2("substitute", layout, Layout.ignore, layout) substitute		
 *)

val var = newTy o Var

(*val con = Trace.trace2("con", Tycon.layout,
		       List.layout(", ", layout),
		       layout) con*)
fun ofConst c =
   case c of
      Aconst.Int _ => newTy Type.Int
    | Aconst.Word _ => newTy Type.Word
    | Aconst.Real _ => real
    | Aconst.String _ => string
    | Aconst.Char _ => char

fun derecord t =
   case toType t of
      Record r => r
    | _ => Error.error "Type.deRecord"

(* val traceFrees = Trace.trace("frees", layoutDetailed, Frees.layout) *)

local open Frees
in
   fun frees ty =
      hom{ty = ty,
	  unknown = (fn ({canGeneralize, ...}, t) =>
		     if canGeneralize
			then singleton(Free.fromType t)
		     else empty),
	  var = fn a => singleton(Free.fromTyvar a),
	  word = empty,
	  int = empty,
	  con = fn (_, fs) => unions fs,
	  record = fn {record, ...} => unions(Srecord.range record)}
end

fun close(t, getFs) =
   let val fs = frees t
   in if Frees.isEmpty fs
      then []
      else List.map(Frees.toList(Frees.-(fs, getFs())), fn free =>
		    case free of
		       Free.X a => a
		     | Free.Y s => 
			  case toType s of
			     Unknown{equality, ...} =>
				let val a = Tyvar.newNoname{equality = equality}
				in Set.setValue(s, {ty = Var a,
						    plist = PropertyList.new()});
				   a
				end
			   | _ => Error.bug "Type.close: not unknown")
   end

(* val close =
 *    Trace.trace2("close", layout, Function.layout, List.layout Tyvar.layout)
 *    close
 *)

(*val traceCanUnify = Trace.trace2("canUnify", layout, layout, Bool.layout) *)

fun canUnify arg = 
(*   traceCanUnify *)
   (fn (t, t') =>
    case (toType t,   toType t') of
         (Unknown _,  _)            => true
       | (_,          Unknown _)    => true
       | (Con(c, ts), Con(c', ts')) => (Tycon.equals(c, c')
					andalso
					List.forall2(ts, ts', canUnify))
       | (Con(c, []), Word)         => Tycon.isWordX c
       | (Word,       Con(c, []))   => Tycon.isWordX c
       | (Con(c, []), Int)          => Tycon.isIntX c
       | (Int,        Con(c, []))   => Tycon.isIntX c
       | (Var a,      Var a')       => Tyvar.equals(a, a')
       | (Word,       Word)         => true
       | (Int,        Int)          => true
       | (Record{flexible=f, record=r}, Record{flexible=f', record=r'}) =>
	    if f orelse f' then Error.bug "canUnify on flexible record"
	    else let val fs = Srecord.toList r
		     val fs' = Srecord.toList r'
		 in Int.equals(List.length fs, List.length fs')
		    andalso List.forall2(fs, fs', fn ((f, t), (f', t')) =>
					 Field.equals(f, f') andalso canUnify(t, t'))
		 end
       | _                          => false) arg

(*val traceUnify = Trace.trace2("unify", layout, layout, Unit.layout) *)

fun unify arg =
(*   traceUnify *)
   (fn (s, s') =>
    if Set.equals(s, s') then ()
    else
       let
	  val t = toType s
	  val t' = toType s'
	  val t =
	     case (t,          t')           of
		  (Unknown r,  Unknown r')   => Unknown(joinUnknown(r, r'))
		| (t,          Unknown _)    => t
		| (Unknown _,  t)            => t
		| (Var a,      Var a')       => if Tyvar.equals(a, a')
						   then t
						else Error.error "unify Var"
		| (Con(c, ts), Con(c', ts')) => if Tycon.equals(c, c')
						   then (unifys(ts, ts'); t)
						else Error.error "unify Con"
	        | (Con(c, []), Word)         => if Tycon.isWordX c
						   then t
						else Error.error "unify Word"
		| (Word,       Con(c, []))   => if Tycon.isWordX c
						 then t'
						else Error.error "unify Word"
		| (Con(c, []), Int)          => if Tycon.isIntX c
						   then t
						else Error.error "unify Int"
		| (Int,        Con(c, []))    => if Tycon.isIntX c
						   then t'
						else Error.error "unify Int"
		| (Word,       Word)         => t
		| (Int,        Int)          => t
		| (Record{flexible=f, record=r}, Record{flexible=f', record=r'}) =>
		     let val fs = Srecord.toList r
			val fs' = Srecord.toList r'
			val rec oneFlex =
			   fn ([], _) => ()
			    | (fs0 as (f, t) :: fs, (f', t') :: fs') =>
				 if Field.equals(f, f')
				    then (unify(t, t');
					  oneFlex(fs, fs'))
				 else oneFlex(fs0, fs')
			    | _ => Error.error "impossible unification of flexible record pattern"
				 
		     in if f
			   then
			      if f'
				 then
				    let
				       val rec loop =
					  fn ([], []) => []
					   | (fs, []) => fs
					   | ([], fs) => fs
					   | ((f, t) :: fs, (f', t') :: fs') =>
						if Field.equals(f, f')
						   then (unify(t, t');
							 (f, t) :: loop(fs, fs'))
						else if Field.<(f, f')
							then (f, t) :: loop(fs, (f', t') :: fs')
						     else (f', t') :: loop((f, t) :: fs, fs')
				    in Record{flexible = true,
					      record = Srecord.fromList(loop(fs, fs'))}
				    end
			      else (oneFlex(fs, fs'); t')
			      else if f' then (oneFlex(fs', fs); t)
			      else if Int.equals(List.length fs, List.length fs')
				      then (List.foreach2(fs, fs', fn ((_, t), (_, t')) =>
							  unify(t, t'));
					    t)
				   else Error.error "unify mismatch: different length records"
		     end
		| _ => Error.error "unify mismatch"
       in Set.union(s, s')
	  ; Set.setValue(s, {ty = t, plist = PropertyList.new()})
       end) arg

and unifys(ts: t list, ts': t list): unit =
   let
      val rec loop =
	 fn ([], []) => ()
	  | (t :: ts, t' :: ts') => (unify(t, t'); loop(ts, ts'))
	  | _ => Error.error "unify Con: different number of args"
   in loop(ts, ts')
   end

fun unifys ts =
   case ts of
      t :: ts => (List.foreach(ts, fn t' => unify(t, t')) ; t)
    | [] => new{equality = false, canGeneralize = true}

local
   structure X = XmlType
   val con = X.con
   val unknown = con(Tycon.tuple, [])
   val {hom, ...} =
      makeHom{unknown = fn _ => unknown,
	      var = X.var,
	      word = con(Tycon.defaultWord, []),
	      int = con(Tycon.defaultInt, []),
	      con = con,
	      record = fn {flexible, record} =>
	      if flexible
		 then Error.error "unresolved ... in flexible pattern"
	      else
		 case Srecord.range record of
		    [t] => t
		  | ts => con(Tycon.tuple, ts)}
in
   val toXml = hom

(*   val toXml = Trace.trace("toXml", layout, XmlType.layout) toXml *)
end

end
