(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Changes all uses of ref into primapps.
 *)

functor Infer(S: INFER_STRUCTS): INFER =
struct
      
open S

open CoreML.Atoms

structure Srecord = SortedRecord
structure Field = Record.Field
structure Scope = Scope(structure CoreML = CoreML)
structure Env = TypeEnv(open CoreML
			structure XmlType = Xml.Type)
structure Scheme = Env.InferScheme
structure Type = Env.InferType

structure PScheme = Prim.Scheme

local open Ast
in structure Aconst = Const
end
   
local open CoreML
in structure Cdec = Dec
   structure Cexp = Exp
   structure Ctype = Type
   structure Cpat = Pat
   structure Cprogram = Program
end

local open Xml
in structure Xdec = Dec
   structure Xexp = DirectExp
   structure Xtype = Type
   structure Xconst = Const
   structure Xlambda = Lambda
   structure Xpat = Pat
   structure XprimExp = PrimExp
   structure XvarExp = VarExp
end

structure Type =
   struct
      open Type

      fun fromCoreML t =
	 case t of
	    Ctype.Var a => var a
	  | Ctype.Con(c, ts) => con(c, List.map(ts, fromCoreML))
	  | Ctype.Record r => record{flexible = false,
				     record = Srecord.map(r, fromCoreML)}
   end

fun instantiate arg =
   let val {instance, args} = Scheme.instantiate arg
   in {instance = instance,
       args = fn () => List.map(args, Type.toXml)}
   end

fun instantiatePrim(PScheme.T{tyvars, ty}) =
   instantiate{scheme = Scheme.T{tyvars = tyvars, ty = Type.fromCoreML ty},
	       canGeneralize = true}

structure Frees = Type.Frees
structure VarRange = Env.VarRange

structure NestedPat = NestedPat(open Xml)
structure MatchCompile =
   MatchCompile(open CoreML
		structure Type = Xtype
		structure NestedPat = NestedPat
		structure Exp =
		   struct
		      open Xexp
		      val lett = let1
		      val var = monoVar

		      fun detuple{tuple, body} =
			 Xexp.detuple
			 {tuple = tuple,
			  body = fn xts => body(List.map(xts, fn (x, t) =>
							 (XvarExp.var x, t)))}
			 
		      fun casee{test, default, cases, ty} =
			 Xexp.casee
			 {test = test,
			  ty = ty,
			  default = default,
			  cases = List.map(cases, fn {con, targs, arg, rhs} =>
					   (Xpat.T{con = con,
						   targs = targs,
						   arg = arg},
					    rhs))}
		   end)

type patCode = (unit -> NestedPat.node) * Type.t
type expCode = (unit -> Xexp.t) * Type.t

(* decCode can't quite be of type  unit -> Xdec.t list
 * because expanding val p = e requires nesting the remaining declarations
 * inside a case expression
 *)
type decCode = Xexp.t * Xtype.t -> Xexp.t * Xtype.t

type ruleCode = patCode * expCode
type rulesCode = ruleCode list

fun finishPat((p, ty): patCode): NestedPat.t =
   NestedPat.new(p(), Type.toXml ty)
   
fun finishExp((e, ty): expCode): Xexp.t = e()
   
local open Layout
in
   fun layoutPatCode(_: patCode) = str "<patcode>"
   val layoutExpCode: expCode -> t = tuple2(fn _ => str"<exp>", Type.layout)
   fun layoutDecCode (_: decCode) = str "<dec>"
end

fun newType() = Type.new{equality = false,
			 canGeneralize = true}

val minInt31 = Int32.fromInt(valOf Int31.minInt)
val maxInt31 = Int32.fromInt(valOf Int31.maxInt)
   
fun makeXconst(c: Aconst.t, ty: Type.t): Xconst.t =
   let val tycon = Xtype.detycon(Type.toXml ty)
   in Xconst.make
      (case c of
	  Aconst.Char c => Xconst.Char c
	| Aconst.Int s =>
	     if Tycon.equals(tycon, Tycon.intInf)
		then (case SmallIntInf.fromString s of
			 NONE => Xconst.IntInf s
		       | SOME i => Xconst.SmallIntInf i)
	     else
		Xconst.Int
		(let
		    val radix =
		       if String.isPrefix{string = s, prefix = "0x"}
			  orelse String.isPrefix{string = s, prefix = "~0x"}
			  then StringCvt.HEX
		       else StringCvt.DEC
		 in case StringCvt.scanString(Pervasive.Int32.scan radix) s of
		    SOME n => if Tycon.equals(tycon, Tycon.int)
				 then n
			      else if Tycon.equals(tycon, Tycon.int31)
				      then if minInt31 <= n andalso n <= maxInt31
					      then n
					   else Error.error "int31 too big"
				   else Error.bug "strange int"
		  | NONE => Error.error "invalid integer constant"
		 end handle Overflow => Error.error "integer constant too big")
	| Aconst.Real r => Xconst.Real r
	| Aconst.String s => Xconst.String s
	| Aconst.Word w =>
	     Xconst.Word(if Tycon.equals(tycon, Tycon.word)
			    then w
			 else if Tycon.equals(tycon, Tycon.word8)
				 then if w = Word32.andb(w, 0wxFF)
					 then w
				      else Error.error "word8 too big"
			 else if Tycon.equals(tycon, Tycon.word31)
				 then if w = Word32.andb(w, 0wx7FFFFFFF)
					 then w
				      else Error.error "word31 too big"
			 else Error.bug("strange word " ^ Tycon.toString tycon)),
       tycon)
   end

fun 'a sortByField(l: (Field.t * 'a) list): (Field.t * 'a) list =
   #insertionSort
   (List.ordered{< = fn ((f, _), (f', _)) =>
		 Field.<(f, f')})
   l

(*---------------------------------------------------*)
(*                       infer                       *)
(*---------------------------------------------------*)

(* local
 *    structure Env =
 *       struct
 * 	 open Env
 * 	 val layout: t -> Layout.t = Layout.ignore
 *       end
 * in 
 *    val traceInferExp =
 *       Trace.trace2("inferExp", Cexp.layout, Env.layout, layoutExpCode)
 * 
 *    val traceInferDec =
 *       Trace.trace2("inferDec", Cdec.layout, Env.layout,
 * 		   Layout.tuple2(layoutDecCode, Env.layout))
 * 	 
 *    val traceInferPat =
 *       Trace.trace2("inferPat", Cpat.layout, Env.layout,
 * 		   Layout.tuple2(layoutPatCode, Env.layout))
 * end
 *)

fun infer(p: CoreML.Program.t): Xml.Program.t =
   let
      val {get = tyconCons: Tycon.t -> Con.t list,
	   set = setTyconCons} =
	 Property.new(Tycon.plist, Property.initRaise("tyconCons", Tycon.layout))
	 
      val {get = conInfo: Con.t -> {tycon: Tycon.t,
				    scheme: Prim.Scheme.t},
	   set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("conInfo", Con.layout))

      val conTycon = #tycon o conInfo
      val conScheme = #scheme o conInfo

(*      val conTycon = Trace.trace("conTycon", Con.layout, Tycon.layout) conTycon *)

(*       val conScheme =
 * 	 Trace.trace("conScheme", Con.layout, Prim.Scheme.layout) conScheme
 *)

      fun instCon c  = instantiatePrim(conScheme c)

      (*---------------------------------------------------*)
      (*                     inferPat                      *)
      (*---------------------------------------------------*)
      (*
       * inferPat yields
       * 1. thunk that can generate a typed nested pattern
       * 2. the type of the pattern
       * 3. a type environment extended with the variables declared in the pattern
       *)

      fun inferPat arg: patCode * Env.t =
(*	 traceInferPat *)
	 (fn (p: Cpat.t, env: Env.t) =>
	  case p of
	     Cpat.Wild => ((fn () => NestedPat.Wild, newType()), env)
	   | Cpat.Var x => let val t = newType()
			   in ((fn () => NestedPat.Var x, t),
			       Env.extendVar(env, x, Scheme.fromType t))
			   end
	   | Cpat.Const c =>
		let val ty = Type.ofConst c
		in ((fn () => NestedPat.Const(makeXconst(c, ty)), ty), env)
		end
	   | Cpat.Con{con, arg} =>
		let val {instance, args} = instCon con
		in case arg of
		   NONE => ((fn () => NestedPat.Con{con = con, targs = args(),
						    arg = NONE},
			     instance),
			    env)
		 | SOME p => 
		      let val (t1, t2) = Type.dearrow instance
			 val (p as (_, tp), env) = inferPat(p, env)
		      in Type.unify(t1, tp) ;
			 ((fn () => NestedPat.Con{con = con,
						  targs = args(),
						  arg = SOME(finishPat p)},
			   t2),
			  env)
		      end
		end
	   | Cpat.Record{flexible, record} =>
		let val (fs, ps) = Record.unzip record
		   val (ps, ts, env) = inferPats(ps, env)
		   val t = Type.record{flexible = flexible,
				       record = Srecord.zip(fs, ts)}
		in ((fn () =>
		     NestedPat.Tuple
		     (if not flexible
			 then List.map(sortByField(List.map2(fs, ps, fn (f, p) =>
							     (f, finishPat p))),
				       #2)
		      else let val {flexible, record=record'} = Type.derecord t
			   in if flexible
				 then (Control.messageLay(Cpat.layout p)
				       ; Error.error "unresolved flexible pattern")
			      else
				 let val record = Record.zip(fs, ps)
				 in List.map
				    (Srecord.toList record', fn (f, t) =>
				     case Record.peek(record, f) of
					NONE => NestedPat.new(NestedPat.Wild,
							      Type.toXml t)
				      | SOME p => finishPat p)
				 end
			   end),
			 t),
		    env)
		end
	   | Cpat.Constraint(p, t') => let val ((p, t), env) = inferPat(p, env)
				       in Type.unify(t, Type.fromCoreML t') ;
					  ((p, t), env)
				       end
	   | Cpat.Layered(x, p) => let val (p as (_, t), env) = inferPat(p, env)
				   in ((fn () => NestedPat.Layered(x, finishPat p),
					t),
				       Env.extendVar(env, x, Scheme.fromType t))
				   end) arg

      and inferPats(ps, env): patCode list * Type.t list * Env.t =
	 let
	    val (ps, ts, env) = 
	       List.fold(List.reverse ps, ([], [], env), fn (p, (ps, ts, env)) =>
			 let val (p as (_, t), env) = inferPat(p, env)
			 in (p :: ps, t :: ts, env)
			 end)
	 in (ps, ts, env)
	 end

      fun delayExp(e as (_, t): expCode): expCode =
	 (fn () => Xexp.lambda{arg = Var.newNoname(),
			       argType = Xtype.unit,
			       body = finishExp e,
			       bodyType = Type.toXml t},
	  Type.arrow(Type.unit, t))
	 
      fun letExp(d: decCode, e as (_, t): expCode): expCode =
	 (fn () => #1(d(finishExp e, Type.toXml t)), t)

      local
	 fun make(c: Con.t)(): Xexp.t =
	    Xexp.conApp{con = c, targs = [], arg = NONE, ty = Xtype.exn}
      in val match = make Con.match
	 val bind = make Con.bind
      end

      (* store all the overloads as promises in a list, so that we can
       * force them after performing the unifications.  This is necessary because
       * forcing an overload can cause further unifications to be done, and all
       * unifications must be done before converting to Xml
       *)
      val overloads: Xexp.t Promise.t list ref = ref []
	 
      fun varExp(x: Var.t, env: Env.t): expCode =
	 let val vr as VarRange.T{scheme, kind} = Env.lookupVarRange(env, x)
	 in case kind of
	    VarRange.Normal =>
	       let val {instance, args} = instantiate{scheme = scheme,
						      canGeneralize = true}
	       in (fn () => Xexp.var{var = x, targs = args(), ty = Type.toXml instance},
		   instance)
	       end
	  | VarRange.Recursive tyvars =>
	       let val ty = Scheme.ty scheme
	       in (fn () => Xexp.var{var = x,
				     targs = List.map(!tyvars, Xtype.var),
				     ty = Type.toXml ty},
		   ty)
	       end
	  | VarRange.Delayed =>
	       let val {instance, args} = instantiate{scheme = scheme,
						      canGeneralize = true}
		  val (_, t') = Type.dearrow instance
	       in (fn () =>
		   Xexp.app{func = Xexp.var{var = x,
					    targs = args(),
					    ty = Xtype.arrow(Xtype.unit,
							     Type.toXml t')},
			    arg = Xexp.unit(),
			    ty = Type.toXml t'},
		   t')
	       end
	  | VarRange.Overload yts =>
	       let
		  val {instance, args} = instantiate{scheme = scheme,
						     canGeneralize = false}
		  val promise =
		     Promise.delay
		     (fn () =>
		      let
			 val rec loop =
			    fn (y, t) :: yts =>
			    if Type.canUnify(instance, t)
			       then (Type.unify(instance, t) ;
				     Xexp.monoVar(y, Type.toXml t))
			    else loop yts
			     | [] => 
				  (Control.messageLay
				   let open Layout
				   in align[Var.layout x,
					    VarRange.layout vr]
				   end;
				   Error.error "impossible use of overloaded var")
		      in loop yts
		      end)
	       in ListRef.push(overloads, promise) ;
		  (fn () => Promise.force promise, instance)
	       end
	 end

      fun tyvarsDifference(t1, t2) =
	 List.keepAll(t1, fn a => not(List.exists(t2, fn a' => Tyvar.equals(a, a'))))

      (*------------------------------------*)
      (*               casee                *)
      (*------------------------------------*)

      fun casee{test: Xexp.t,
		testType: Xtype.t,
		caseType: Xtype.t,
		cases: (NestedPat.t * Xexp.t) list}: Xexp.t =
	 case cases of
	    (p, e) :: _ => 
	       let fun lett x = Xexp.let1{var = x, exp = test, body = e}
	       in case NestedPat.node p of
		  NestedPat.Wild => lett(Var.newNoname())
		| NestedPat.Var x => lett x
		(* for flat tuples, could do special case *)
		| _ => 
		     let
			val (decs, cases) =
			   List.unzip
			   (List.map
			    (cases, fn (p: NestedPat.t, e: Xexp.t) =>
			     let val args = NestedPat.varsAndTypes p
				val (vars, tys) = List.unzip args
				val func = Var.newNoname()
				val arg = Var.newNoname()
				val argType = Xtype.tuple tys
				val funcType = Xtype.arrow(argType, caseType)
				val dec =
				   Xdec.MonoVal
				   {var = func,
				    ty = funcType,
				    exp =
				    XprimExp.Lambda
				    (Xlambda.new
				     {arg = arg,
				      argType = argType,
				      body =
				      Xexp.toExp
				      (Xexp.detupleBind
				       {tuple = Xexp.monoVar(arg, argType),
					components = vars,
					body = e})})}
				fun finish rename =
				   Xexp.app
				   {func = Xexp.monoVar(func, funcType),
				    arg =
				    Xexp.tuple{exps = (List.map
						       (args, fn (x, t) =>
							Xexp.monoVar(rename x, t))),
					       ty = argType},
				    ty = caseType}
			     in (dec, (p, finish))
			     end))
			val testVar = Var.newNoname()
		     in Xexp.let1
			{var = testVar,
			 exp = test,
			 body = 
			 Xexp.lett
			 {decs = decs,
			  body = MatchCompile.matchCompile{conTycon = conTycon,
							   tyconCons = tyconCons,
							   test = testVar,
							   testType = testType,
							   caseType = caseType,
							   cases = cases}}}
		     end
	       end
	  | _ => Error.bug "case with no patterns"

      fun forceRules({argType, resultType, rules},
		     p: NestedPat.node,
		     e: Xexp.t): Xml.Lambda.t =
	 let val arg = Var.newNoname()
	    val argType = Type.toXml argType
	    val resultType = Type.toXml resultType
	 in Xml.Lambda.new
	    {arg = arg, argType = argType,
	     body = (Xexp.toExp
		     (casee{testType = argType,
			    test = Xexp.monoVar(arg, argType),
			    cases = (List.map(rules, fn (p, e) =>
					      (finishPat p, finishExp e))
				     @ [(NestedPat.new(p, argType),
					 Xexp.raisee(e, resultType))]),
			    caseType = resultType}))}
	 end

      fun forceRulesMatch rs =
	 forceRules(rs, NestedPat.Wild, match())

      fun forceRulesReraise rs =
	 let val x = Var.newNoname()
	 in forceRules(rs, NestedPat.Var x, Xexp.monoVar(x, Xtype.exn))
	 end

      val emptyDec: decCode = fn e => e

      fun appendDec(d: decCode, d': decCode): decCode = d o d'

      fun cons(thunk: unit -> Xdec.t list): decCode = 
	 fn (e, t) => (Xexp.lett{decs = thunk(), body = e}, t)

      fun valDec(tyvars: Tyvar.t list, var: Var.t, exp as (_, t): expCode,
		 kind: VarRange.kind, env: Env.t): decCode * Env.t =
	 (cons(fn () => [Xdec.PolyVal{tyvars = tyvars,
				      var = var,
				      ty = Type.toXml t,
				      exp = Xexp.toExp(finishExp exp)}]),
	  Env.extendVarRange(env, var,
			     VarRange.T
			     {scheme = Scheme.T{tyvars = tyvars, ty = t},
			      kind = kind}))

      fun patDec(p as (_, tp): patCode, e as (_, te): expCode): decCode =
	 (Type.unify(tp, te) ;
	  fn (body, ty) =>
	  let val p = finishPat p
	  in (casee{test = finishExp e,
		    testType = Type.toXml te,
		    caseType = ty,
		    cases = [(p, body),
			     (NestedPat.new(NestedPat.Wild, NestedPat.ty p),
			      Xexp.raisee(bind(), ty))]},
	      ty)
	  end)

      fun checkTyvars(tyvars, tyvarsClose, frees) =
	 case tyvarsDifference(tyvars, tyvarsClose) of
	    _ :: _ =>
	       if Frees.areDisjoint(Frees.fromTyvars tyvars, frees())
		  then ()
	       else Error.error "can't generalize tyvars"
	  | _ => ()

      fun inferValDec(tyvars: Tyvar.t list,
		      pat: Cpat.t,
		      exp: Cexp.t,
		      e: expCode as (_, te),
		      env: Env.t): decCode * Env.t =
	 let val (p as (_, tp), env') = inferPat(pat, env)
	 in Type.unify(te, tp) ;
	    if Cexp.isExpansive exp
	       then (case tyvars of
			[] => (patDec(p, e), env')
		      | _ => (Layout.output(CoreML.Exp.layout exp, Out.error) ;
			      Out.newline Out.error ;
			      Error.error "can't generalize an expansive exp"))
	    else
	       let val freesPromise = Promise.delay(fn () => Env.frees env)
		  fun frees() = Promise.force freesPromise
		  val tyvarsClose = Type.close(te, frees)
		  val _ = checkTyvars(tyvars, tyvarsClose, frees)
		  fun vd x = valDec(tyvarsClose, x, e, VarRange.Normal, env)
	       in case (tyvarsClose, pat) of
		  ([], _) => (patDec(p, e), env')
		| (_, Cpat.Wild) => vd(Var.newNoname())
		| (_, Cpat.Var x) => vd x
		| (_, Cpat.Constraint(Cpat.Var x, _)) => vd x
		| _ =>
		     let
			val _ =
			   Control.messageLay
			   (let open Layout
			    in align[str "expansion of polymorphic pattern",
				     Cdec.layout
				     (Cdec.Val{pat = pat, exp = exp,
					       tyvars = tyvars})]
			    end)
			val x = Var.newNoname()
			val (d, env) = vd x
			val d =
			   if Cpat.isRefutable pat
			      then appendDec
				 (d, patDec(#1(inferPat(Cpat.removeVars pat, env)),
					    varExp(x, env)))
			   else d
		     in List.fold
			(Cpat.vars pat, (d, env), fn (y, (d, env)) =>
			 let val y' = Var.new y
			    val p = Cpat.removeOthersReplace(pat, y, y')
			    val (p, env') = inferPat(p, env)
			    val e as (_, te) =
			       delayExp(letExp(patDec(p, varExp(x, env)),
					       varExp(y', env')))
			    val (d', env) =
			       valDec(Type.close(te, fn () =>
						 Env.frees env),
				      y, e, VarRange.Delayed, env)
			 in (appendDec(d, d'), env)
			 end)
		     end
	       end
	 end
(* accumulate the datatype declarations *)
      val dbsRef = ref []

      fun processConArg(tycon, tyvars, {con, arg}) =
	 let val result = Ctype.con(tycon, List.map(tyvars, Ctype.var))
	    val (ty, arg) =
	       case arg of
		  NONE => (result, NONE)
		| SOME arg => (Ctype.arrow(arg, result),
			       SOME(Type.toXml(Type.fromCoreML arg)))
	 in setConInfo(con, {tycon = tycon,
			     scheme = PScheme.T{tyvars = tyvars, ty = ty}});
	    {con = con, arg = arg}
	 end

      fun processException ca = processConArg(Tycon.exn, [], ca)
   
      fun processDatatypes(dbs, f): unit =
	 List.foreach
	 (dbs, fn {tyvars, tycon, cons} =>
	  let val cons = List.map(cons, fn ca => processConArg(tycon, tyvars, ca))
	  in setTyconCons(tycon, List.map(cons, #con));
	     f{tyvars = tyvars,
	       tycon = tycon,
	       cons = cons}
	  end)

      (*------------------------------------*)
      (*              inferDec              *)
      (*------------------------------------*)
      fun inferDec arg: decCode * Env.t =
(*	 traceInferDec *)
	 (fn (d: Cdec.t, env: Env.t) =>
	  case d of
	     Cdec.Val{tyvars, pat, exp} =>
		inferValDec(tyvars, pat, exp, inferExp(exp, env), env)
	   | Cdec.Datatype dbs =>
		(processDatatypes(dbs, fn db as {tycon, ...} =>
				  if Tycon.equals(tycon, Tycon.reff)
				     then ()
				  else ListRef.push(dbsRef, db));
		 (emptyDec, env))
	   | Cdec.Exception(e as {con, ...}) =>
		let val ca = processException e
		in (cons(fn () => [Xdec.Exception ca]),
		    env)
		end
	   | Cdec.Overload{var, scheme=CoreML.Scheme.T{tyvars, ty}, ovlds} =>
		(emptyDec,
		 let val ty = Type.fromCoreML ty
		 in Env.extendVarRange
		    (env, var,
		     VarRange.T{scheme = Scheme.T{tyvars = tyvars, ty = ty},
				kind =
				VarRange.Overload
				(List.map(ovlds, fn y =>
					  (y, Scheme.ty(Env.lookupVar(env, y)))))})
		 end)
	   | Cdec.Fun{tyvars, decs} =>
		let
		   (* type args to recursive calls *)
		   val args: Tyvar.t list ref = ref []
		   val (env', decs) =
		      List.fold
		      (decs, (env, []), fn ({var, ty, rules}, (env, decs)) =>
		       let val argType = newType()
			  val resultType = newType()
			  val t = Type.arrow(argType, resultType)
		       in (case ty of
			      NONE => ()
			    | SOME t' => Type.unify(t, Type.fromCoreML t')) ;
			  (Env.extendVarRange
			   (env, var, VarRange.T{scheme = Scheme.fromType t,
						 kind = VarRange.Recursive args}),
			   {var = var,
			    argType = argType,
			    resultType = resultType,
			    rules = rules} :: decs)
		       end)
		   val decs =
		      List.map(decs, fn {var, rules, argType, resultType} =>
			       let
				  val rs = inferRulesUnify(rules, env',
							   argType, resultType)
			       in {var = var,
				   ty = Type.arrow(argType, resultType),
				   rules = rs}
			       end)
		   val freesPromise = Promise.delay(fn () => Env.frees env)
		   fun frees() = Promise.force freesPromise
		   val tyvarsClose =
		      Tyvars.toList
		      (List.fold(decs, Tyvars.empty, fn ({ty, ...}, tyvars) =>
				 Tyvars.+(tyvars,
					  Tyvars.fromList(Type.close(ty, frees)))))
		   val _ = args := tyvarsClose
		   val env =
		      List.fold(decs, env, fn ({var, ty, ...}, env) =>
				Env.extendVar(env, var,
					      Scheme.T{tyvars = tyvarsClose,
						       ty = ty}))
		in checkTyvars(tyvars, tyvarsClose, frees) ;
		   (cons(fn () =>
			 [Xdec.Fun
			  {tyvars = tyvarsClose,
			   decs =
			   List.map(decs, fn {var, ty, rules} =>
				    {var = var,
				     ty = Type.toXml ty,
				     lambda = forceRulesMatch rules})}]),
		    env)
		end) arg

      and inferDecs(ds: Cdec.t list, env: Env.t): decCode * Env.t =
	 List.fold(ds, (emptyDec, env), fn (d, (d', env)) =>
		    let val (d, env) = inferDec(d, env)
		    in (appendDec(d', d), env)
		    end)

      (*------------------------------------*)
      (*              inferExp              *)
      (*------------------------------------*)

      and inferExps(es, env) =
	 List.map(es, fn e => inferExp(e, env))

      and inferExp arg: expCode =
(*	 traceInferExp *)
	 (fn (e, env) => 
	 case e of
	    Cexp.Var x => varExp(x, env)
	  | Cexp.Prim _ => apply(e, env, NONE)
	  | Cexp.Con _ => apply(e, env, NONE)
	  | Cexp.App(e1, e2) => apply(e1, env, SOME(inferExp(e2, env)))
	  | Cexp.Const c => let val ty = Type.ofConst c
			    in (fn () => Xexp.const(makeXconst(c, ty)),
				ty)
			    end
	  | Cexp.Record r =>
	       (* This code is messy because the components of the record have
		* to be evaluated left to right as they appeared in the source
		* program, but then ordered according to sorted field name within
		* the tuple
		*)
	       let
		  val fes = Record.toList r
		  val es = inferExps(List.map(fes, #2), env)
		  val ty = Type.record{flexible = false,
				       record =
				       Srecord.fromList
				       (List.map2(fes, es,
						  fn ((f, _), (_, t)) => (f, t)))}
	       in (fn () =>
		   Xexp.seq
		   (List.map(es, finishExp), fn es =>
		    Xexp.tuple{exps = (List.map
				       (sortByField
					(List.map2(fes, es, fn ((f, _), e) =>
						   (f, e))),
					#2)),
			       ty = Type.toXml ty}),
		   ty)
	       end
	  | Cexp.Fn rules =>
	       let val rs as {argType, resultType, rules} = inferRules(rules, env)
	       in (fn () => let
			       val {arg, argType, body} =
				  Xlambda.dest(forceRulesMatch rs)
			       val resultType = Type.toXml resultType
			    in Xexp.lambda
			       {arg = arg,
				argType = argType,
				body = Xexp.fromExp(body, resultType),
				bodyType = resultType}
			    end,
		   Type.arrow(argType, resultType))
	       end
	  | Cexp.Let(ds, e) =>
	       let val (ds, env) = inferDecs(ds, env)
	       in letExp(ds, inferExp(e, env))
	       end
	  | Cexp.Constraint(e, t') =>
	       let val ans as (_, t) = inferExp(e, env)
	       in Type.unify(t, Type.fromCoreML t') ;
		  ans
	       end
	  | Cexp.Handle(e, rs) =>
	       let val (e, t) = inferExp(e, env)
		  val rs as {argType, resultType, rules} = inferRules(rs, env)
	       in Type.unify(t, resultType) ;
		  Type.unify(argType, Type.exn) ;
		  (fn () => let
			       val {arg, body, ...} =
				  Xlambda.dest(forceRulesReraise rs)
			       val t' = Type.toXml t
			    in Xexp.handlee{try = finishExp(e, t),
					    catch = (arg, Type.toXml Type.exn),
					    handler = Xexp.fromExp(body, t'),
					    ty = t'}
			    end,
		   t)
	       end
	  | Cexp.Raise e =>
	       let val e as (_, t) = inferExp(e, env)
		  val resultType = newType()
	       in Type.unify(t, Type.exn) ;
		  (fn () => Xexp.raisee(finishExp e,
					Type.toXml resultType),
		   resultType)
	       end) arg

      and applyOne(e as (_, t): expCode, e' as (_, t'): expCode): expCode =
	 let val resultType = newType()
	 in Type.unify(t, Type.arrow(t', resultType)) ;
	    (fn () => Xexp.app{func = finishExp e,
			       arg = finishExp e',
			       ty = Type.toXml resultType},
	     resultType)
	 end

      and apply(e1: Cexp.t, env: Env.t, arg: expCode option): expCode =
	 let
	    fun eta(ty: Type.t,
		    make: (Xexp.t * Xtype.t) option * Xtype.t -> Xexp.t)
	       : expCode =
	       case (Type.dearrowOpt ty, arg)                 of
		    (NONE,               NONE)                =>
		       (fn () => make(NONE, Type.toXml ty), ty)
		  | (NONE,               SOME _)              =>
		       Error.error "application mismatch"
		  | (SOME(t1, t2),       SOME(argExp as (_, argTy))) =>
		       (Type.unify(t1, argTy) ;
			(fn () => make(SOME(finishExp argExp, Type.toXml t1),
				       Type.toXml t2),
			 t2))
		  | (SOME(t1, t2),       NONE)                =>
		       let val x = Var.newNoname()
		       in (fn () =>
			   let val t1 = Type.toXml t1
			      val t2 = Type.toXml t2
			   in Xexp.lambda
			      {arg = x,
			       argType = t1,
			       body = make(SOME(Xexp.monoVar(x, t1), t1), t2),
			       bodyType = t2}
			   end,
			   ty)
		       end
	 in
	    case e1 of
	       Cexp.Con con =>
		  let val {instance, args} = instCon con
		  in eta(instance, fn (arg, resultType) =>
			 if Con.equals(con, Con.reff)
			    then (case arg of
				     NONE => Error.bug "bad ref"
				   | SOME(arg, _) =>
					Xexp.primApp{prim = Prim.reff,
						     targs = args(),
						     args = [arg],
						     ty = resultType})
			 else Xexp.conApp{con = con,
					  targs = args(),
					  arg = (case arg of
						    NONE => NONE
						  | SOME(x, _) => SOME x),
					  ty = resultType})
		  end
	     | Cexp.Prim prim =>
		  let
		     val {instance, args = targs} =
			instantiatePrim(Prim.scheme prim)
		  in eta(instance, fn (arg, resultType) =>
			 let
			    fun make(args: Xexp.t list): Xexp.t =
			       Xexp.primApp{prim = prim,
					    targs = targs(),
					    args = args,
					    ty = resultType}
			 in
			    case (Prim.numArgs prim, arg) of
			         (NONE,              NONE)     => make[]
			       | (SOME n,            SOME(arg, argType)) =>
				    (case n of
					0 => make[]
				      | 1 => make[arg]
				      | _ => 
					   Xexp.detuple
					   {tuple = arg,
					    body =
					    fn components =>
					    make(List.map(components, Xexp.varExp))})
			       | _                             =>
				    Error.error "primApp mismatch"
			 end)
		  end
	     | Cexp.App(e1, e2) =>
		  let val e = apply(e1, env, SOME(inferExp(e2, env)))
		  in case arg of
		     NONE => e
		   | SOME e' => applyOne(e, e')
		  end
	     | _ => let val e1 = inferExp(e1, env)
		    in case arg of
		       NONE => e1
		     | SOME e2 => applyOne(e1, e2)
		    end
	 end
      and inferRules(rs, env): {argType: Type.t, resultType: Type.t,
				rules: (patCode * expCode) list} =
	 inferRulesUnify(rs, env, newType(), newType())
   
      and inferRulesUnify(rs, env, argType, resultType) =
	 {argType = argType,
	  resultType = resultType,
	  rules = List.map(rs, fn (p, e) =>
			   let val (p as (_, tp), env) = inferPat(p, env)
			      val e as (_, te)  = inferExp(e, env)
			   in Type.unify(tp, argType) ;
			      Type.unify(te, resultType) ;
			      (p, e)
			   end)}

      (*------------------------------------*)
      (*    main code for type inference    *)
      (*------------------------------------*)

      val Cprogram.T{decs,
		     constructors =
		     cs as CoreML.Constructors.T{datatypes, exceptions}}
	 = Scope.scope p

      val _ = (processDatatypes(datatypes, fn _ => ());
	       List.foreach(exceptions, fn ca => (processException ca; ())))
	    
      val (ds, env) = Control.traceCall "unification" inferDecs(decs, Env.empty)

      val _ = (List.foreach(!overloads, fn p => (Promise.force p ; ())) ;
	       overloads := [])
	 
      val (body, _) =
	 Control.traceCall "finish infer" ds (Xexp.unit(), Xtype.unit)

      val xml = Xml.Program.T{datatypes = !dbsRef,
			      body = Xexp.toExp body}

   in Xml.Program.clear xml;
      xml
   end

(*val infer = Trace.trace("infer", Layout.ignore, Xml.Program.layout) infer *)

structure Scheme = PScheme
   
end
