(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor MatchCompile(S: MATCH_COMPILE_STRUCTS): MATCH_COMPILE =
struct

open S

structure Env = MonoEnv(structure Domain = Var
			structure Range = Var)
   
structure FlatPat =
   struct
      datatype t =
	 Any
       | Const of Const.t
       | Con of {con: Con.t,
		 targs: Type.t list,
		 arg: NestedPat.t option}
       | Tuple of NestedPat.t list

      val isRefutable =
	 fn Any => false
	  | Const _ => true
	  | Con _ => true
	  | Tuple ps => List.exists(ps, NestedPat.isRefutable)

      val isAny =
	 fn Any => true
	  | _ => false

      (* get rid of Wild, Var, Layered, Or - also remove unary tuples *)
      fun flatten(var: Var.t, pat: NestedPat.t, env: Env.t): t list * Env.t =
	 let fun done p = ([p], env)
	    fun extend x = Env.extend(env, x, var)
	 in case NestedPat.node pat of
	    NestedPat.Wild => done Any
	  | NestedPat.Var x => ([Any], extend x)
	  | NestedPat.Layered(x, p) => flatten(var, p, extend x)
	  | NestedPat.Const c => done(Const c)
	  | NestedPat.Con x => done(Con x)
	  | NestedPat.Tuple[p] => flatten(var, p, env)
	  | NestedPat.Tuple ps => done(Tuple ps)
	 end

      fun flattens(vars, pats: NestedPat.t list, env): t list list * Env.t =
	 let
	    val (patss, env) =
	       List.fold2
	       (vars, pats, ([], env), fn (var, pat, (patss, env)) =>
		let val (pats, env) = flatten(var, pat, env)
		in (pats :: patss, env)
		end)
	 in (List.cross(List.reverse patss), env)
	 end
   end   

datatype continue =
   Finish of (Var.t -> Var.t) -> Exp.t
 | Matches of FlatPat.t list option * continue
type info = {accum: Env.t, continue: continue}
type rule = {pat: NestedPat.t, info: info}
type rules = rule list
type flatrule = {pat: FlatPat.t, info: info}
type flatrules = flatrule list
fun layoutFlatrules(_: flatrules) = Layout.str "<flatrules>"
type finish = info list -> Exp.t
fun layoutFinish(_: finish) = Layout.str "<finish>"

(* val traceMatchFlat =
 *    Trace.trace4
 *    ("matchFlat", Var.layout, Type.layout, layoutFlatrules, layoutFinish,
 *     Exp.layout)
 *)

(*---------------------------------------------------*)
(*                   matchCompile                    *)
(*---------------------------------------------------*)

fun matchCompile{test: Var.t,
		 conTycon: Con.t -> Tycon.t,
		 tyconCons: Tycon.t -> Con.t list,
		 testType: Type.t,
		 caseType: Type.t,
		 cases: (NestedPat.t * ((Var.t -> Var.t) -> Exp.t)) list}
  : Exp.t =
   let
      fun match(var: Var.t, ty: Type.t, rules: rules, finish: finish)
	: Exp.t =
	 let
	    val rules =
	       List.fold(List.reverse rules, [], fn (rule, rules) =>
			  flatten(var, rule, rules))
	 in matchFlat(var, ty, rules, finish)
	 end
      
      and flatten(var: Var.t,
		  {pat, info as {accum, continue}}: rule,
		  rules: flatrules): flatrules =
	 let val (pats, accum) = FlatPat.flatten(var, pat, accum)
	 in List.fold(List.reverse pats, rules,
		       fn (pat: FlatPat.t, rules: flatrules) =>
		       {pat = pat, info = {accum = accum,
					   continue = continue}}
		       :: rules)
	 end
      
      and matchFlat arg: Exp.t =
(*	 traceMatchFlat *)
	 (fn (var: Var.t, ty: Type.t, rules: flatrules, finish: finish) =>
	 let
	    val test = Exp.var(var, ty)
	    fun loop rules' =
	       case rules' of
		  [] => finish(List.map(rules, fn {pat, info} =>
					case pat of
					   FlatPat.Any => info
					 | _ => Error.bug "matchFlat"))
		| {pat, info} :: rs =>
		     case pat of
			FlatPat.Any => loop rs
		      | FlatPat.Const _ => const(test, rules, finish)
		      | FlatPat.Con _ => sum(test, rules, finish)
(*		      | FlatPat.Tuple[] =>
			   finish
			   (List.map(rules, fn {pat, info} =>
				     case pat of
					(FlatPat.Any | FlatPat.Tuple[]) =>
					   info
				      | _ => Error.bug "expected unit pattern"))*)
		      | FlatPat.Tuple ps => tuple(test, ty, rules, finish)
	 in loop rules
	 end) arg
	 
      and matches(vars: (Var.t * Type.t) list,
		  rules: {pats: NestedPat.t list option, info: info} list,
		  finish: finish): Exp.t =
	 let
	    val rules =
	       List.fold
	       (List.reverse rules, [],
		fn ({pats, info as {accum, continue}}, rules) =>
		case pats of
		   NONE => {pats = NONE, info = info} :: rules
		 | SOME pats =>
		      let
			 val (patss, accum) =
			    FlatPat.flattens(List.map(vars, #1), pats, accum)
		      in List.fold
			 (List.reverse patss, rules, fn (pats, rules) =>
			  {pats = SOME pats,
			   info = {accum = accum, continue = continue}}
			  :: rules)
		      end)
	 in matchesFlat(vars, rules, finish)
	 end
      
      and matchesFlat(vars: (Var.t * Type.t) list,
		      rules: {pats: FlatPat.t list option, info: info} list,
		      finish: finish): Exp.t =
	 let
	    (* bring the i'th element of the list to the front *)
	    fun rotate(l, i) =
	       let
		  fun loop(prev, i, rest) =
		     case (i, rest) of
			(0, x :: rest) => x :: (List.reverseAppend(prev, rest))
		      | (_, x :: rest) => loop(x :: prev, i-1, rest)
		      | _ => Error.error "rotate"
	       in loop([], i, l)
	       end
	    val normal = (vars, rules)
	    val (vars, rules) =
	       case (!Control.match, rules) of
		  (Control.LeftToRight, _) => normal
		| (_, []) => normal
		| (_, [_]) => normal
		| (Control.Rotate, {pats, info} :: _) =>
		  case pats of
		     NONE => normal
		   | SOME pats =>
			case List.index(pats, FlatPat.isRefutable) of
			   NONE => normal
			 | SOME 0 => normal
			 | SOME i =>
			      (rotate(vars, i),
			       List.map(rules, fn {pats, info} =>
					{pats = (case pats of
						    NONE => NONE
						  | SOME ps => SOME(rotate(ps, i))),
					 info = info}))
	 in case vars of
	    [] =>
	       finish(List.map(rules, fn {pats, info} =>
			       (case pats of
				   SOME(_ :: _) => Error.bug "matchesFlat1"
				 | _ => ()
			        ; info)))
	  | (var, ty) :: vars =>
	       let
		  val rules =
		     List.map
		     (rules, fn {pats, info as {accum, continue}} =>
		      case pats of
			 NONE => {pat = FlatPat.Any,
				  info = {accum = accum,
					  continue = Matches(NONE, continue)}}
		       | SOME(pat :: pats) =>
			    {pat = pat,
			     info = {accum = accum,
				     continue = Matches(SOME pats, continue)}}
		       | _ => Error.bug "wrong number of patterns")
	       in matchFlat
		  (var, ty, rules, fn rules =>
		   matchesFlat
		   (vars,
		    List.map(rules, fn {accum, continue} =>
			     case continue of
				Matches(pats, continue) =>
				   {pats = pats,
				    info = {accum = accum,
					    continue = continue}}
			      | _ => Error.bug "matchesFlat:"),
		    finish))
	       end
	 end

      (*------------------------------------*)
      (*               tuple                *)
      (*------------------------------------*)
      and tuple(test: Exp.t, ty: Type.t, rules: flatrules, finish: finish) =
	 let
	    val rules =
	       List.map
	       (rules, fn rule as {pat, info} =>
		case pat of
		   FlatPat.Any => {pats = NONE, info = info}
		 | FlatPat.Tuple pats => {pats = SOME pats, info = info}
		 | _ => Error.error "expected tuple pattern")
	 in Exp.detuple
	    {tuple = test,
	     body = fn vars => matches(vars, rules, finish)}
	 end

      (*------------------------------------*)
      (*                sum                 *)
      (*------------------------------------*)
      and sum(test, rules: flatrules, finish: finish) =
	 let
	    datatype arg = 
	       NoArg of info list
	     | Arg of {var: Var.t, ty: Type.t, rules: rules}
	    val (cases, defaults) =
	       List.fold
	       (List.reverse rules, ([], []), fn ({pat, info}, (cases, defaults)) =>
		case pat of
		   FlatPat.Any =>
		      (List.map
		       (cases, fn {con, tys, arg} =>
			{con = con, tys = tys,
			 arg = (case arg of
				   NoArg infos => NoArg(info :: infos)
				 | Arg{var, ty, rules} =>
				      Arg{var = var,
					  ty = ty,
					  rules = {pat = NestedPat.wild ty,
						   info = info} :: rules})}),
		       info :: defaults)
		 | FlatPat.Con{con=c, targs=tys, arg} => 
		      let
			 fun insert cases =
			    case cases of
			       [] =>
				  [{con = c, tys = tys,
				    arg =
				    (case arg of
					NONE => NoArg(info :: defaults)
				      | SOME p =>
					   let val ty = NestedPat.ty p
					   in Arg{var = Var.newNoname(),
						  ty = ty,
						  rules =
						  {pat = p, info = info}
						  :: (List.map
						      (defaults, fn info =>
						       {pat = NestedPat.wild ty,
							info = info}))}
					   end)}]
			     | (cas as {con, tys, arg=a}) :: cases =>
				  if Con.equals(c, con)
				     then {con = con, tys = tys,
					   arg = (case (a, arg) of
						     (NoArg infos, NONE) =>
							NoArg(info :: infos)
						   | (Arg{var, ty, rules},
						      SOME p) =>
						     Arg{var = var,
							 ty = ty,
							 rules = 
							 {pat = p, info = info}
							 :: rules}
						   | _ => Error.bug "use of constructor with and without arg in pattern match")}
					:: cases
				  else cas :: (insert cases)
		      in (insert cases, defaults)
		      end
		 | _ => Error.error "expected constructor pat")
	       
	    val default =
	       case cases of
		  [] => SOME(finish defaults)
		| {con, ...} :: _ =>
		     let val tycon = conTycon con
		     in if Tycon.equals(tycon, Tycon.exn)
			orelse List.length cases <> List.length(tyconCons tycon)
			   then SOME(finish defaults)
			else NONE
		     end
		  
	    fun normal() =
	       Exp.casee{test = test, default = default,
			 ty = caseType,
			 cases = 
			 List.map(cases, fn {con, tys, arg} =>
				  let
				     val (arg, rhs) =
					case arg of
					   NoArg infos => (NONE, finish infos)
					 | Arg{var, ty, rules} =>
					      (SOME(var, ty),
					       match(var, ty, rules, finish))
				  in {con = con, targs = tys, arg = arg,
				      rhs = rhs}
				  end)}
	 in case cases of
	    [{con, arg = Arg{var, ty, rules}, ...}] =>
	       if Con.equals(con, Con.reff)
		  then Exp.lett{var = var,
				exp = Exp.deref test,
				body = match(var, ty, rules, finish)}
	       else normal()
	  | _ => normal()
	 end

      (*------------------------------------*)
      (*               const                *)
      (*------------------------------------*)
      and const(test: Exp.t, rules: flatrules, finish: finish) =
	 let
	    val (cases, defaults) =
	       List.fold
	       (List.reverse rules, ([], []),
		fn ({pat, info}, (cases, defaults)) =>
		case pat of
		   FlatPat.Any =>
		      (List.map(cases, fn {const, infos} =>
				{const = const, infos = info :: infos}),
		       info :: defaults)
		 | FlatPat.Const c =>
		      let
			 fun insert cases =
			    case cases of
			       [] => [{const = c, infos = info :: defaults}]
			     | (casee as {const, infos}) :: cases =>
				  if Const.equals(c, const)
				     then {const = c, infos = info :: infos}
					    :: cases
				  else casee :: insert cases
		      in (insert cases, defaults)
		      end
		 | _ => Error.error "expected Const pat")
	 in List.fold(cases, finish defaults, fn ({const, infos}, rest) =>
		      Exp.iff{test = Exp.equal(test, Exp.const const),
			      thenn = finish infos,
			      elsee = rest,
			      ty = caseType})
	 end

   (*------------------------------------*)
   (*    main code for match compile     *)
   (*------------------------------------*)
      
   in match(test, testType,
	    List.map(cases, fn (p, f) => {pat = p, info = {accum = Env.empty,
							  continue = Finish f}}),
	    fn [] => Error.bug "matchRules: no default"
	     | {accum=env, continue} :: _ =>
		  (case continue of
		      Finish f => f (fn x => Env.lookup(env, x))
		    | _ => Error.bug "matchRules: expecting Finish"))
   end

(*val matchCase =
 *  Trace.trace3("matchCase", Var.layout, Type.layout,
 *		List.layout(", ", Layout.tuple2(NestedPat.layout, Exp.layout)),
 *		Exp.layout) matchCase
 *)
   
end
