(* cmcat takes a ".cm" file as input and produces an SML program as output.
 * It takes all of the files in the group (and subgroups) and outputs their
 * contents in the linear order deduced by CM. 
 *
 * cmcat does not understand the export filters used by CM.  Hence, the resulting
 * program may not be valid SML and may require manual renaming of some modules.
 *
 * Because of the problem with export filters, cmcat has a hack to move certain
 * modules first, based on their filename.  See the variable moveFirstHack.
 * This hack makes it more likely that programs that rebind basis library modules
 * will work.
 *
 * cmcat may be used from the SML/NJ repl or from the shell.
 *
 * To use from the repl, do something like the following:
 * Cmcat.cmcat{sources = "sources.cm",
 *	       out = TextIO.stdOut,
 *	       separate = false,
 *	       filenamesOnly = false}
 *
 * Before using from the shell, you must first call
 *   export{njDir = "<smlnj installation directory>"
 *
 * The usage is as follows:
 *   cmcat [-separate] [-filename] <file>
 *
 * cmcat writes its output on stdOut.
 * The -separate flag inserts a ; between each file.
 * The -filename flag outputs the filenames in order instead of their contents.
 *)

structure Cmcat:
   sig
      val cmcat: {(* The ".cm" filename *)
		  sources: string,
		  (* Where to write the output *)
		  out: TextIO.outstream,
 		  (* If true, then insert a ; between each file *)
		  separate: bool,
		  (* If true, then output filenames intstead of their contents *)
		  filenamesOnly: bool,
                  defines : string list
		 } -> unit

      (* Creates a (big) heap file. *)
      val export: {
		   (* The root of the SML/NJ implementation *)
		   njDir: string
		  } -> unit
   end =
   struct
      open TextIO OS
      open FileSys Path Process
	 
      fun 'a dynamicWind(thunk: unit -> 'a, cleanup: unit -> unit): 'a =
	 let val a = thunk()
	 in cleanup(); a
	 end handle exn => (cleanup(); raise exn)
	    
      local
	 fun ('stream, 'a) withh(file: string,
				 f: 'stream -> 'a,
				 openn: string -> 'stream,
				 close: 'stream -> unit): 'a =
	    let val stream = openn file
	    in dynamicWind(fn () => f stream,
			   fn () => close stream)
	    end
      in
	 fun 'a withIn(file: string, f: instream -> 'a): 'a =
	    withh(file, f, openIn, closeIn)

	 fun 'a withOut(file: string, f: outstream -> 'a): 'a =
	    withh(file, f, openOut, closeOut)
      end

      fun 'a fluidLet(out: outstream, out': outstream, thunk: unit -> 'a): 'a =
	 let val old = getOutstream out
	 in dynamicWind
	    (fn () => (setOutstream(out, getOutstream out'); thunk()),
	     fn () => setOutstream(out, old))
	 end
      
      fun 'a ignore(out: outstream, thunk: unit -> 'a): 'a =
	 withOut("/dev/null", fn out' => fluidLet(out, out', thunk))

      fun outputContents(file: string, out: outstream): unit =
	 withIn(file, fn ins =>
		while not(endOfStream ins)
		   do output(out, input ins))

      fun 'a inTemp(thunk: unit -> 'a): 'a =
	 let val cur = getDir()
	    val dir = tmpName()
	 in mkDir dir
	    ; (dynamicWind
	       (fn () => (chDir dir; thunk()),
		fn () => (system("/bin/rm -r " ^ dir); chDir cur)))
	 end

      local open Substring
      in
	 fun isSubstring{string, substring}: bool =
	    0 <> size(#2(position substring (all string)))
      end
	 
      fun cmcat{sources, out, separate, filenamesOnly, defines} =
	 inTemp
	 (fn () =>
	  let
            (* Define preprocessor symbols *)
            val _ = List.app (fn n => CM.SymVal.define (n, 0)) defines

       	    val files = ignore(stdErr, fn () => CM.names' sources)

	    val moveFirstHack = ["ml-yacc-lib", "smlnj-lib", "smltk"]

	    val (first, second) =
	       List.partition
	       (fn f =>
		List.exists
		(fn s => isSubstring {string = f, substring = s})
		moveFirstHack)
	       files

	    fun outputs fs =
	       List.app
	       (fn f => if filenamesOnly
			   then output(out, f ^ "\n")
			else (outputContents(f, out)
			      ; output(out, if separate then ";\n" else "\n")))
	       fs

	  in outputs first; outputs second
	  end)

      fun message s = output(stdErr, s ^ "\n")
   
      fun die msg =
	 (message "Usage: cmcat [-separate] [-filename] <file>"
	  ; message("Error: " ^ msg)
	  ; exit failure)

      fun export{njDir: string} =
	 (chDir njDir
	  ; chDir "bin"
	  ; if Posix.FileSys.access("cmcat", [])
	       then ()
	    else (system "ln -s .run-sml cmcat"; ())
	  ; chDir ".heap"
	  ; (SMLofNJ.exportFn
	     ("cmcat", fn (_, args) =>
	      let val separate = ref false
		 val filenamesOnly = ref false
                 val defines = ref nil : string list ref
		 fun loop args =
		    case args of
                       [file] => ((cmcat{sources = mkAbsolute{path =file,
							      relativeTo = getDir()},
		        		 out = stdOut,
					 separate = !separate,
					 filenamesOnly = !filenamesOnly,
                                         defines = !defines})
				  handle _ => die "cmcat failed"
				     ; 0)
		     | flag :: args =>
			  (case flag of
			      "-separate" => separate := true
			    | "-filename" => filenamesOnly := true
			    | _ =>
                                if String.isPrefix "-D" flag then
                                  defines := (String.substring (flag, 2, size flag - 2)) :: (!defines)
                                else
                                  die ("invalid flag" ^ flag)
				 ; loop args)
		     | _ => die "wrong number of arguments"
	      in loop args
	      end)))
   end
