(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* This simplifier is based on the following article:
 * Shrinking Lambda Expressions in Linear Time.
 * Journal of Functional Programming. Vol 7, no 5, 1997.
 *)

functor Simplify(S: SIMPLIFY_STRUCTS): SIMPLIFY = 
struct

open S
open XmlTree
open Dec PrimExp

(* val traceSimplifyExp =
 *    Trace.trace("simplifyExp", Exp.layout, Exp.layout)
 * 
 * val traceSimplifyDecs =
 *    Trace.trace("simplifyDecs", List.layout Dec.layout, List.layout Dec.layout)
 * 
 * val traceSimplifyLambda =
 *    Trace.trace("simplifyLambda", Lambda.layout, Lambda.layout)
 *)
fun simplifyOnce e =
   let
      (* counts the number of modifications made by the simplifier *)
      val numMods = ref 0
      fun incmod() = IntRef.inc numMods

      (*---------------------------------------------------*)
      (*                       sigma                       *)
      (*---------------------------------------------------*)
      (* sigma is a mapping from variables to variables.
       *  While descending the program all occurrences of x will be replaced by
       *  sigma(x).
       *  For example, after seeing let val x = y in e
       *  sigma(x) = y
       *  Then e will be processed and all occurrences of x will be replaced by y
       *)
	 
      local
	 val {get: Var.t -> VarExp.t option, set} =
	    Property.new(Var.plist, Property.initConst NONE)
      in
	 fun setSigma(x, y) = set(x, SOME y)

	 fun simplifyVarExp(x: VarExp.t): VarExp.t =
	    case get(VarExp.var x) of
	       NONE => x
	     | SOME y => y
		  
	 fun sigma(x: Var.t): Var.t =
	    case get x of
	       NONE => x
	     | SOME y => VarExp.var y
      end

      fun simplifyVarExps xs = List.map(xs, simplifyVarExp)

      (*---------------------------------------------------*)
      (*                      census                       *)
      (*---------------------------------------------------*)
      (* census walks over an expression and modifies variable occurrence counts,
       * which are stored on the property list associated with the variable
       *)

      (* only keep track of monomorphic variables, since the simplifier ignores
       * polymorphic ones
       *)
	 
      val {set = setNumOccurrences, get = numOccurrences} =
	 Property.new(Var.plist,
		      Property.initRaise("num occurrences", Var.layout))

      val setNumOccurrences =
	 fn (x, n) => (Assert.cheap(n >= 0); setNumOccurrences(x, n))
	    
(*        val numOccurrences =
 *  	 Trace.trace("numOccurrences", Var.layout, Int.layout) numOccurrences
 *)

      fun incNumOccurrences(x, n) =
	 setNumOccurrences(x, n + numOccurrences x)

(*       val incNumOccurrences =
 * 	 Trace.trace2("incNumOccurrences",
 *  		      Var.layout, Int.layout, Unit.layout) incNumOccurrences
 *)

      fun maybeIncNumOccurrences(VarExp.T{var, targs}, n) =
	 case targs of
	    [] => incNumOccurrences(var, n)
	  | _ => ()

      fun setVar(x: Var.t, y: VarExp.t): unit =
	 (maybeIncNumOccurrences(y, numOccurrences x)
	  ; setSigma(x, y))

      fun isUseless var = numOccurrences var = 0

      fun countExp(e: Exp.t): unit =
	 Exp.foreach{exp = e,
		     handleBoundVar = fn (x, [], _) => setNumOccurrences(x, 0)
				       | _ => (),
		     handleVarExp = fn x => maybeIncNumOccurrences(x, 1),
		     handleExp = fn _ => (), handlePrimExp = fn _ => ()}

      fun deleteVarExp(x: VarExp.t): unit =
	 maybeIncNumOccurrences(simplifyVarExp x, ~1)
	      
      fun deleteExp(e: Exp.t): unit =
	 Exp.foreachVarExp(e, deleteVarExp)

      fun deleteVarExps xs = List.foreach(xs, deleteVarExp)
      fun deleteLambda l = deleteExp(Lambda.body l)
	 
      (*---------------------------------------------------*)
      (*                    simplifyExp                    *)
      (*---------------------------------------------------*)

      local
	 val {get: Var.t -> bool, set} =
	    Property.new(Var.plist, Property.initConst false)
      in
	 val isInlined = get
	 fun setInlined x = set(x, true)
      end

      val {get = bind: Var.t -> PrimExp.t option, set} =
	 Property.new(Var.plist, Property.initConst NONE)

      fun removeBind x = set(x, NONE)
      fun setBind(x, b) = set(x, SOME b)

      val bind = bind o VarExp.var
	 
      fun simplifyExp arg: Exp.t =
(*	 traceSimplifyExp *)
	 (fn (e: Exp.t) =>
	  let val {decs, result} = Exp.dest e
	  in Exp.new{decs = simplifyDecs decs,
		     result = simplifyVarExp result}
	  end) arg
      and simplifyDecs(decs: Dec.t list): Dec.t list =
	 case decs of
	    [] => []
	  | dec :: decs =>
	       case dec of
		  Dec.Exception _ => dec :: (simplifyDecs decs)
		| Dec.PolyVal{var, tyvars, ty, exp} =>
		     Dec.PolyVal{var = var, tyvars = tyvars, ty = ty,
				 exp = simplifyExp exp}
		     :: simplifyDecs decs
		| Dec.Fun{tyvars, decs = decs'} =>
		     (case tyvars of
			 _ :: _ =>
			    Dec.Fun{tyvars = tyvars,
				    decs =
				    List.map(decs', fn {var, ty, lambda} =>
					     {var = var,
					      ty = ty,
					      lambda = simplifyLambda lambda})}
			    :: simplifyDecs decs
			| [] => 
			     let
				val decs' =
				   List.fold
				   (decs', [], fn (dec as {var, ty, lambda}, decs) =>
				    if isUseless var
				       then (deleteLambda lambda ; decs)
				    else (setBind(var, PrimExp.Lambda lambda) ;
					  dec :: decs))
				val decs = simplifyDecs decs
				(* need to walk over all the decs to removeBind before
				 * simplifying any of them because they are mutually
				 * recursive
				 *)
				val decs' =
				   List.fold
				   (decs', [], fn (dec as {var, lambda, ...}, decs) =>
				    if isInlined var
				       then decs
				    else if isUseless var
					    then (deleteLambda lambda ; decs)
					 else (removeBind var ; dec :: decs))
			     in case decs' of
				[] => decs
			      | _ => Dec.Fun{tyvars = tyvars,
					     decs =
					     List.map(decs', fn {var, ty, lambda} =>
						      {var = var,
						       ty = ty,
						       lambda = simplifyLambda lambda})}
				   :: decs
			     end)
		| Dec.MonoVal{var, ty, exp} =>
		     let
			fun finish(exp, decs) =
			   Dec.MonoVal{var = var, ty = ty, exp = exp} :: decs
			fun nonExpansive(delete: unit -> unit,
					 set: unit -> (unit -> PrimExp.t) option) =
			   if isUseless var
			      then (incmod(); delete(); simplifyDecs decs)
			   else let val s = set()
				    val decs = simplifyDecs decs
				in if isUseless var
				      then (incmod(); delete(); decs)
				   else (case s of
					    NONE => decs
					  | SOME n => finish(n(), decs))
				end
			fun expansive(e: PrimExp.t) = finish(e, simplifyDecs decs)
			fun nonExpansiveCon(delete, e: PrimExp.t) =
			   nonExpansive(delete,
					fn () => (setBind(var, e) ;
						  SOME(fn () => e)))
			fun expression(e: Exp.t): Dec.t list =
			   let val {decs=decs', result} = Exp.dest(simplifyExp e)
			      val _ = setVar(var, result)
			      val decs = simplifyDecs decs
			   in decs' @ decs
			   end
		     in case exp of
			Const _ => nonExpansive(fn () => (),
						fn () => SOME(fn () => exp))
		      | Var x =>
			   let val x = simplifyVarExp x
			   in setSigma(var, x)
			      ; maybeIncNumOccurrences(x, numOccurrences var - 1)
			      ; incmod()
			      ; simplifyDecs decs
			   end
		      | Tuple xs =>
			   let val xs = simplifyVarExps xs
			   in nonExpansiveCon(fn () => deleteVarExps xs,
					      Tuple xs)
			   end
		      | Select{tuple, offset} =>
			   let val tuple = simplifyVarExp tuple
			   in nonExpansive
			      (fn () => deleteVarExp tuple,
			       fn () =>
			       (case bind tuple of
				   NONE =>
				      SOME(fn () => Select{tuple = tuple,
							   offset = offset})
				 | SOME(Tuple xs) => 
				      (incmod()
				       ; deleteVarExp tuple
				       ; setVar(var, List.nth(xs, offset))
				       ; NONE)
				 | _ => Error.bug "simplifyPrimExp: Select"))
			   end
		      | ConApp{con, targs, arg} =>
			   let val arg = (case arg of
					     NONE => NONE
					   | SOME x => SOME(simplifyVarExp x))
			   in nonExpansiveCon
			      (fn () => (case arg of
					    NONE => ()
					  | SOME x => deleteVarExp x),
			       ConApp{con = con, targs = targs, arg = arg})
			   end			     
		      | PrimApp{prim, args, targs} =>
			   (case Prim.numArgs prim of
			       NONE =>
				  nonExpansive(fn () => (),
					       fn () => SOME(fn () =>
							     PrimApp{prim = prim,
								     targs = targs,
								     args = []}))
			     | SOME _ => 
				  expansive(PrimApp{prim = prim, targs = targs,
						    args = simplifyVarExps args}))
		      | Lambda l =>
			   nonExpansive
			   (fn () => if isInlined var
					then ()
				     else deleteLambda l,
					fn () => (setBind(var, exp)
						  ; SOME(fn () => Lambda(simplifyLambda l))))
		      | App{func, arg} =>
			   let val func = simplifyVarExp func
			      val arg = simplifyVarExp arg
			      fun normal() =
				 expansive(App{func = func, arg = arg})
			      val VarExp.T{var = funcVar, targs} = func
			   in case targs of
			      [] =>
				 if 1 = numOccurrences funcVar
				    then (case bind func of
					     NONE => normal()
					   | SOME(Lambda l) =>
						(incmod()
						 ; deleteVarExp arg
						 ; setVar(Lambda.arg l, arg)
						 ; setInlined funcVar
						 ; setNumOccurrences(funcVar, 0)
						 ; expression(Lambda.body l))
					   | _ =>
						Error.bug "simplifyPrimExp: App")
				 else normal()
			     | _ => normal()
			   end
		      | Raise x => expansive(Raise(simplifyVarExp x))
		      | Handle{try, catch, handler} =>
			   expansive(Handle{try = simplifyExp try,
					    catch = catch,
					    handler = simplifyExp handler})
		      | Case{test, cases, default} =>
			   let
			      fun match(f: Pat.t -> bool): Dec.t list =
				 let
				    fun loop cases =
				       case cases of
					  (pat, e) :: cases =>
					     if f pat
						then
						   (List.foreach(cases, fn (_, e) =>
								 deleteExp e) ;
						    case default of
						       NONE => ()
						     | SOME e => deleteExp e ;
							  expression e)
					     else (deleteExp e ; loop cases)
						     | []  => (case default of
								  SOME e => expression e
								| NONE => Error.bug "simplifyPrimExp: Case")
				 in incmod();
				    deleteVarExp test;
				    loop cases
				 end
			      val test = simplifyVarExp test
			   in case bind test of
			      SOME(ConApp{con = c, arg = NONE, ...}) =>
				 match(fn Pat.T{con = c', arg = NONE, ...} =>
				       Con.equals(c, c')
			       | _ => false)
			    | SOME(ConApp{con = c, arg = SOME x, ...}) =>
				 match(fn Pat.T{con = c', arg = SOME(x', _), ...} =>
				       if Con.equals(c, c')
					  then (setVar(x', x) ; true)
				       else false
			       | _ => false)
			    | NONE =>
				 expansive(Case{test = test,
						cases = List.map(cases, fn (p, e) =>
								 (p, simplifyExp e)),
						default = (case default of
							      NONE => NONE
							    | SOME e => SOME(simplifyExp e))})
			    | _ => Error.bug "simplifyPrimExp: Case2"
			   end
		     end

      and simplifyLambda l: Lambda.t =
(*	 traceSimplifyLambda *)
	 (fn l => 
	  let val {arg, argType, body} = Lambda.dest l
	  in Lambda.new{arg = arg,
			argType = argType,
			body = simplifyExp body}
	  end) l

      val _ = countExp e

      val e = simplifyExp e
   in
      Exp.clear e;

      Control.messageStr("simplify made " ^ Int.toString(!numMods)
			 ^ " modifications");
      
      (e, !numMods)
   end

(* val simplifyOnce =
 *    Trace.trace("simplifyOnce", Exp.layout,
 * 	       Layout.tuple2(Exp.layout, Int.layout)) simplifyOnce
 *)

fun simplify(Program.T{datatypes, body}): Program.t =
   Program.T
   {datatypes = datatypes,
    body = #1(simplifyOnce(#1(simplifyOnce body)))}

(* val simplify =
 *    Trace.trace("Xml.simplify", Program.layout, Program.layout) simplify
 *)

end
