(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor XmlTree(S: XML_TREE_STRUCTS): XML_TREE =
struct

open S
local open Ast
in structure Adec = Dec
   structure Aexp = Exp
   structure Apat = Pat
end

structure Type =
   struct
      structure T = HashType(S)
      open T

      datatype dest =
	 Var of Tyvar.t
       | Con of Tycon.t * t list

      fun dest t =
	 case Dest.dest t of
	    Dest.Var a => Var a
	  | Dest.Con x => Con x
	  | _ => Error.bug "Xml.Type.dest"
   end

structure Pat =
   struct
      structure Apat = Ast.Pat
	 
      datatype t = T of {con: Con.t,
			 targs: Type.t list,
			 arg: (Var.t * Type.t) option}
	 
      fun con(T{con, ...}) = con

      local fun make c = T{con = c, targs = [], arg = NONE}
      in val truee = make Con.truee
	 val falsee = make Con.falsee
      end

      fun toAst(T{con, arg, ...}) =
	 let val con = Con.toAst con
	 in case arg of
	    NONE => Apat.con con
	  | SOME (x, t) =>
	       Apat.app(con, Apat.var(Var.toAst x))
	 (*Apat.app(con, Apat.constraint(Apat.var(Var.toAst x), Type.toAst t)) *)
	 end

      val layout = Apat.layout o toAst
   end

(*---------------------------------------------------*)
(*                      VarExp                       *)
(*---------------------------------------------------*)

structure VarExp =
   struct
      datatype t = T of {var: Var.t,
			 targs: Type.t list}

      fun mono var = T{var = var, targs = []}

      fun var(T{var, ...}) = var
      fun targs(T{targs, ...}) = targs
	       
      val toAst = Aexp.var o Var.toAst o var

      fun toAsts(xs: t list): Aexp.t list = List.map(xs, toAst)
      val layout = Aexp.layout o toAst
   end

structure AppendList = AppendList
   
(*---------------------------------------------------*)
(*           Expressions and Declarations            *)
(*---------------------------------------------------*)

datatype exp =
   Exp of {decs: dec list,
	   result: VarExp.t}
and primExp =
   Const of Const.t
  | Var of VarExp.t
  | Tuple of VarExp.t list
  | Select of {tuple: VarExp.t,
	       offset: int}
  | Lambda of lambda
  | ConApp of {con: Con.t,
	       targs: Type.t list,
	       arg: VarExp.t option}
  | PrimApp of {prim: Prim.t,
		targs: Type.t list,
		args: VarExp.t list}
  | App of {func: VarExp.t,
	    arg: VarExp.t}
  | Case of {test: VarExp.t,
	     cases: (Pat.t * exp) list,
	     default: exp option}
  | Raise of VarExp.t
  | Handle of {try: exp,
	       catch: Var.t * Type.t,
	       handler: exp}
and dec =
   MonoVal of {var: Var.t,
	       ty: Type.t,
	       exp: primExp}
  | PolyVal of {tyvars: Tyvar.t list,
		ty: Type.t,
		var: Var.t,
		exp: exp}
  | Fun of {tyvars: Tyvar.t list,
	    decs: {var: Var.t,
		   ty: Type.t,
		   lambda: lambda} list}
  | Exception of {con: Con.t,
		  arg: Type.t option}
and lambda = Lam of {plist: PropertyList.t,
		     arg: Var.t,
		     argType: Type.t,
		     body: exp}

(*---------------------------------------------------*)
(*                 Conversion to Ast                 *)
(*---------------------------------------------------*)

fun expToAst(Exp{decs, result, ...}): Aexp.t =
   Aexp.lett(decsToAst decs, VarExp.toAst result)
and expsToAsts es = List.map(es, expToAst)
and decsToAst decs = List.map(decs, decToAst)
and decToAst d : Adec.t =
   case d of
      MonoVal{var, ty, exp} =>
	 Adec.make(Adec.Val
		   ([],
		    [(Apat.var(Var.toAst var),
(*		      Apat.constraint(Apat.var(Var.toAst var), Type.toAst ty), *)
		      primExpToAst exp)]))
    | PolyVal{tyvars, var, exp, ...} =>
	 Adec.vall(tyvars, Var.toAst var, expToAst exp)
    | Fun{tyvars, decs} =>
	 Adec.make
	 (Adec.Fun
	  (tyvars,
	   List.map(decs, fn {var, lambda = Lam{arg, argType, body, ...}, ...} =>
		    [{pats = [Apat.var(Var.toAst var),
			      Apat.constraint(Apat.var(Var.toAst arg),
					      Type.toAst argType)
(*			      Apat.var(Var.toAst arg) *)
			      ],
		      resultType = NONE,
		      body = expToAst body}])))
    | Exception{con, arg} =>
	 Adec.exceptionn(Con.toAst con, Type.optionToAst arg)
and primExpToAst e : Aexp.t =
   case e of
      Const c => Const.toAstExp c
    | Var x => VarExp.toAst x
    | Tuple xs => Aexp.tuple(VarExp.toAsts xs)
    | Select{tuple, offset} =>
	 Aexp.select{tuple = VarExp.toAst tuple,
		     offset = offset + 1}
    | Lambda lambda => Aexp.fnn(lambdaToAst lambda)
    | ConApp{con, arg, ...} =>
	 let val con = Aexp.con(Con.toAst con)
	 in case arg of
	    NONE => con
	  | SOME e => Aexp.app(con, VarExp.toAst e)
	 end
    | PrimApp{prim, args, ...} =>
	 let val p = Aexp.longvid(Prim.longid prim)
	 in
	    case Prim.numArgs prim of
	       NONE => p
	     | SOME _ => Aexp.app(p, Aexp.tuple(List.map(args, VarExp.toAst)))
	 end
    | App{func, arg} => Aexp.app(VarExp.toAst func, VarExp.toAst arg)
    | Raise x => Aexp.raisee(VarExp.toAst x)
    | Handle{try, catch, handler} =>
	 Aexp.handlee(expToAst try,
		      [(Apat.var(Var.toAst(#1 catch)),
			expToAst handler)])
    | Case{test, cases, default, ...} =>
	 Aexp.casee(VarExp.toAst test,
		    (List.map(cases, fn (pat, exp) =>
			      (Pat.toAst pat, expToAst exp))
		     @ (case default of
			   NONE => []
			 | SOME e => [(Ast.Pat.wild, expToAst e)])))
and lambdaToAst(Lam{arg, body, argType, ...}) =
   [(
     Apat.constraint(Apat.var(Var.toAst arg), Type.toAst argType),
(*     Apat.var(Var.toAst arg), *)
     expToAst body)]

fun layoutLambda f = Aexp.layout(Aexp.fnn(lambdaToAst f))

(*---------------------------------------------------*)
(*                   Declarations                    *)
(*---------------------------------------------------*)

structure Dec =
   struct
      type exp = exp
      datatype t = datatype dec

      val toAst = decToAst
      val layout = Ast.Dec.layout o toAst
   end

(*---------------------------------------------------*)
(*                    Expressions                    *)
(*---------------------------------------------------*)

structure PrimExp =
   struct
      type exp = exp
      datatype t = datatype primExp

      val toAst = primExpToAst
      val layout = Aexp.layout o toAst
   end

structure Exp =
   struct
      datatype t = datatype exp

      val new = Exp
      fun dest(Exp r) = r
      val decs = #decs o dest
      val result = #result o dest

      fun fromPrimExp(exp: PrimExp.t, ty: Type.t): t =
	 let val var = Var.newNoname()
	 in Exp{decs = [Dec.MonoVal{var = var, ty = ty, exp = exp}],
		result = VarExp.mono var}
	 end
	 
      local
	 fun make f (Exp{decs, result}, d) =
	    Exp{decs = f(d, decs),
		result = result}
      in val prefix = make(op ::)
	 val prefixs = make(op @)
      end

      val toAst = expToAst
      val layout = Ast.Exp.layout o toAst

      (*------------------------------------*)
      (*              foreach               *)
      (*------------------------------------*)
      fun foreach{exp: t,
		  handleExp: t -> unit,
		  handlePrimExp: PrimExp.t -> unit,
		  handleBoundVar: Var.t * Tyvar.t list * Type.t -> unit,
		  handleVarExp: VarExp.t -> unit}: unit =
	 let
	    fun monoVar(x, t) = handleBoundVar(x, [], t)
	    fun handleVarExps xs = List.foreach(xs, handleVarExp)
	    fun loopExp e =
	       let val {decs, result} = dest e
	       in List.foreach(decs, loopDec)
		  ; handleVarExp result
		  ; handleExp e
	       end
	    and loopPrimExp e =
	       (handlePrimExp e
		; (case e of
		      Const _ => ()
		    | Var x => handleVarExp x
		    | Tuple xs => handleVarExps xs
		    | Select{tuple, ...} => handleVarExp tuple
		    | Lambda lambda => loopLambda lambda
		    | PrimApp{args, ...} => handleVarExps args
		    | ConApp{arg, ...} => (case arg of
					      NONE => ()
					    | SOME x => handleVarExp x)
		    | App{func, arg} => (handleVarExp func
					 ; handleVarExp arg)
		    | Raise x => handleVarExp x
		    | Handle{try, catch, handler} =>
			 (loopExp try
			  ; monoVar catch
			  ; loopExp handler)
		    | Case{test, cases, default} =>
			 (handleVarExp test
			  ; List.foreach(cases, fn (Pat.T{arg, ...}, e) =>
					 (case arg of
					     NONE => ()
					   | SOME x => monoVar x
					  ; loopExp e))
			  ; (case default of
				NONE => ()
			      | SOME e => loopExp e))))
	    and loopDec d =
	       case d of
		  MonoVal{var, ty, exp} =>
		     (monoVar(var, ty); loopPrimExp exp)
		| PolyVal{var, tyvars, ty, exp} =>
		     (handleBoundVar(var, tyvars, ty)
		      ; loopExp exp)
		| Exception _ => ()
		| Fun{tyvars, decs, ...} =>
		     (List.foreach(decs, fn {var, ty, lambda} =>
				   handleBoundVar(var, tyvars, ty))
		      ; List.foreach(decs, fn {lambda, ...} =>
				     loopLambda lambda))
	    and loopLambda(Lam{arg, argType, body, ...}): unit =
	       (monoVar(arg, argType); loopExp body)
	 in loopExp exp
	 end

      fun ignore _ = ()

      fun foreachPrimExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = f,
		 handleExp = ignore,
		 handleBoundVar = ignore,
		 handleVarExp = ignore}

      fun foreachVarExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = ignore,
		 handleBoundVar = ignore,
		 handleVarExp = f}

      fun foreachBoundVar(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = ignore,
		 handleBoundVar = f,
		 handleVarExp = ignore}

      fun foreachExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = f,
		 handleBoundVar = ignore,
		 handleVarExp = ignore}

      fun size e =
	 let val n = ref 0
	    fun inc() = n := 1 + !n
	 in foreachPrimExp(e, fn _ => inc());
	    !n
	 end

(*      val size = Trace.trace("size", Layout.ignore, Int.layout) size *)

      fun clear(e: t): unit =
	 let open PrimExp
	    val clear = PropertyList.clear
	    fun clearTyvars ts = List.foreach(ts, Tyvar.clear)
	    fun clearPat(Pat.T{arg, ...}) =
	       case arg of
		  NONE => ()
		| SOME(x, _) => Var.clear x
	    fun clearExp e = clearDecs(decs e)
	    and clearDecs ds = List.foreach(ds, clearDec)
	    and clearDec d =
	       case d of
		  MonoVal{var, exp, ...} => (Var.clear var; clearPrimExp exp)
		| PolyVal{var, tyvars, exp, ...} =>
		     (Var.clear var;
		      clearTyvars tyvars;
		      clearExp exp)
		| Fun{tyvars, decs} =>
		     (clearTyvars tyvars;
		      List.foreach(decs, fn {var, ty, lambda} =>
				   (Var.clear var;
				    clearLambda lambda)))
		| Exception{con, ...} => Con.clear con
	    and clearPrimExp e =
	       case e of
		  Lambda l => clearLambda l
		| Case{cases, default, ...} =>
		     (List.foreach(cases, fn (p, e) =>
				   (clearPat p; clearExp e));
		      case default of
			 NONE => ()
		       | SOME e => clearExp e)
		| Handle{try, catch, handler} => 
		     (clearExp try;
		      Var.clear(#1 catch);
		      clearExp handler)
		| _ => ()
	    and clearLambda(Lam{arg, body, ...}) =
	       (Var.clear arg; clearExp body)
	 in clearExp e
	 end
   end

(*---------------------------------------------------*)
(*                      Lambda                       *)
(*---------------------------------------------------*)

structure Lambda =
   struct
      type exp = exp
      datatype t = datatype lambda

      fun arg(Lam{arg, ...}) = arg
      fun argType(Lam{argType, ...}) = argType 
      fun body(Lam{body, ...}) = body

      type node = {arg: Var.t,
		   argType: Type.t,
		   body: exp}

      fun new{arg, argType, body} =
	 Lam{plist = PropertyList.new(),
	     arg = arg, argType = argType, body = body}

      fun dest(Lam{arg, argType, body, ...}) =
	 {arg = arg, argType = argType, body = body}
	 
      fun plist(Lam{plist, ...}) = plist
	     
      val layout = layoutLambda
      fun equals(f:t, f':t) = PropertyList.equals(plist f, plist f')
   end

(* ------------------------------------------------- *)
(*                     DirectExp                     *)
(* ------------------------------------------------- *)
structure DirectExp =
   struct
      open Dec PrimExp

      structure Cont =
	 struct
	    type t = PrimExp.t * Type.t -> Exp.t

	    fun nameGen(k: VarExp.t * Type.t -> Exp.t): t =
	       fn (e, t) =>
	       case e of
		  Var x => k(x, t)
		| _ => let val x = Var.newNoname()
		       in Exp.prefix(k(VarExp.mono x, t),
				     MonoVal{var = x, ty = t, exp = e})
		       end
	    
	    fun name(k: VarExp.t * Type.t -> Exp.t): t = nameGen k

	    val id: t = name(fn (x, _) => Exp{decs = [], result = x})
	       
	    fun return(k: t, xt) = k xt
	 end

      type t = Cont.t -> Exp.t

      fun layout _ = Layout.str "<Xml.DirectExp>"

      fun send(e: t, k: Cont.t): Exp.t = e k

      fun toExp e = send(e, Cont.id)

      fun fromExp(Exp{decs, result}, ty): t =
	 fn k => Exp.prefixs(k(Var result, ty), decs)

      fun sendName(e, k) = send(e, Cont.name k)

      fun simple e k = Cont.return(k, e)
	 
      fun const c = simple(Const c, Type.ofConst c)

      fun varExp(x, t) = simple(Var x, t)

      fun var{var, targs, ty} = varExp(VarExp.T{var = var, targs = targs}, ty)

      fun monoVar(x, t) = var{var = x, targs = [], ty = t}

      fun convertsGen(es: t list, k: (VarExp.t * Type.t) list -> Exp.t): Exp.t =
	 let
	    fun loop(es, xs) =
	       case es of
		  [] => k(List.reverse xs)
		| e :: es => sendName(e, fn x => loop(es, x :: xs))
	 in loop(es, [])
	 end

      fun converts(es: t list,
		   make: (VarExp.t * Type.t) list -> PrimExp.t * Type.t): t =
	 fn k => convertsGen(es, k o make)

      fun convert(e: t, make: VarExp.t * Type.t -> PrimExp.t * Type.t): t =
	 fn k => send(e, Cont.name(k o make))

      fun convertOpt(e, make) =
	 case e of
	    NONE => simple(make NONE)
	  | SOME e => convert(e, make o SOME o #1)

      fun tuple{exps: t list, ty: Type.t}: t =
	 case exps of
	    [e] => e
	  | _ => converts(exps, fn xs => (PrimExp.Tuple(List.map(xs, #1)), ty))

      fun select{tuple, offset, ty} =
	 convert(tuple, fn (tuple, _) =>
		 (Select{tuple = tuple, offset = offset}, ty))

      fun conApp{con, targs, arg, ty} =
	 convertOpt(arg, fn arg =>
		    (ConApp{con = con, targs = targs, arg = arg}, ty))

      fun primApp{prim, targs, args, ty} =
	 converts(args, fn args =>
		  (PrimApp{prim = prim,
			   targs = targs,
			   args = List.map(args, #1)},
		   ty))

      fun convert2(e1, e2, make) =
	 converts([e1, e2],
		  fn [x1, x2] => make(x1, x2)
		   | _ => Error.bug "convert2")
				     
      fun app{func, arg, ty} =
	 convert2(func, arg, fn ((func, _), (arg, _)) =>
		  (App{func = func, arg = arg}, ty))

      fun casee{test, cases, default, ty} =
	 convert(test, fn (test, _) =>
		 (Case
		  {test = test,
		   cases = List.map(cases, fn (p, e) => (p, toExp e)),
		   default = (case default of
				 NONE => NONE
			       | SOME e => SOME(toExp e))},
		  ty))

      fun raisee(e: t, t: Type.t): t =
	 convert(e, fn (x, _) => (Raise x, t))
(* 	 fn k =>
 * 	 (Raise x, t)
 * 	 in sendName(e, fn (x, _) => Cont.return(Cont.id, (Raise x, t)))
 *)
	 
      fun handlee{try, catch, handler, ty} =
	 simple(Handle{try = toExp try,
		       catch = catch,
		       handler = toExp handler},
		ty)

      fun unit() = tuple{exps = [], ty = Type.unit}

      fun reff(e: t): t =
	 convert(e, fn (x, t) =>
		 (PrimApp{prim = Prim.reff,
			  targs = [t],
			  args = [x]},
		  Type.reff t))

      fun deref(e: t): t =
	 convert(e, fn (x, t) =>
		 let val t = Type.deref t
		 in (PrimApp{prim = Prim.deref, targs = [t], args = [x]},
		     t)
		 end)

      fun equal(e1, e2) =
	 convert2(e1, e2, fn ((x1, t), (x2, _)) =>
		  (PrimApp{prim = Prim.equal,
			   targs = [t],
			   args = [x1, x2]},
		   Type.bool))

      fun iff{test, thenn, elsee, ty} =
	 casee{test = test,
	       cases = [(Pat.truee, thenn), (Pat.falsee, elsee)],
	       default = NONE,
	       ty = ty}

      fun vall{var, exp}: Dec.t list =
	 let val t = ref Type.unit
	    val Exp{decs, result} =
	       sendName(exp, fn (x, t') => (t := t';			 
					    Exp{decs = [], result = x}))
	 in decs @ [MonoVal{var = var, ty = !t, exp = Var result}]
	 end

      fun sequence es =
    	 converts(es, fn xs => let val (x, t) = List.last xs
			       in (Var x, t)
			       end)

      fun seq(es, make) =
	 fn k => convertsGen(es, fn xts => send(make(List.map(xts, varExp)), k))

      fun lett{decs, body} = fn k => Exp.prefixs(send(body, k), decs)

      fun let1{var, exp, body} =
	 fn k => 
	 send(exp, fn (exp, ty) =>
	      Exp.prefix(send(body, k),
			 Dec.MonoVal{var = var, ty = ty, exp = exp}))
			  

      fun lambda{arg, argType, body, bodyType} =
	 simple(Lambda(Lambda.new{arg = arg, argType = argType,
				  body = toExp body}),
		Type.arrow(argType, bodyType))

      fun detupleGen(e: PrimExp.t, t: Type.t, components: Var.t list,
		     body: Exp.t): Exp.t =
	 Exp.prefixs
	 (body,
	  case components of
	     [] => []
	   | [x] => [MonoVal{var = x, ty = t, exp = e}]
	   | _ =>
		let val ts = Type.detuple t
		   val tupleVar = Var.newNoname()
		in MonoVal{var = tupleVar, ty = t, exp = e}
		   ::
		   List.mapi(List.zip(components, ts), fn (i, (x, t)) =>
			     MonoVal{var = x, ty = t,
				     exp = Select{tuple = VarExp.mono tupleVar,
						  offset = i}})
		end)
      
      fun detupleBind{tuple, components, body} =
	 fn k => send(tuple, fn (e, t) => detupleGen(e, t, components, body k))

      fun detuple{tuple: t, body}: t =
	 fn k =>
	 tuple
	 (fn (e, t) =>
	  let val ts = Type.detuple t
	  in case e of
	     Tuple xs => send(body(List.zip(xs, ts)), k)
	   | _ => let val components = List.map(ts, fn _ => Var.newNoname())
		  in detupleGen(e, t, components,
				send(body(List.map2(components, ts, fn (x, t) =>
						    (VarExp.mono x, t))),
				     k))
		  end
	  end)
   end
structure Exp =
   struct
      open Exp
	 
      fun unit() =
	 let open DirectExp
	 in toExp(tuple{exps = [], ty = Type.unit})
	 end
   end

(*---------------------------------------------------*)
(*                     Datatype                      *)
(*---------------------------------------------------*)

structure Datatype =
   struct
      type t = {tycon: Tycon.t,
		tyvars: Tyvar.t list,
		cons: {con: Con.t, arg: Type.t option} list}

      fun toAst({tyvars, tycon, cons}:t) =
	 {tyvars = tyvars,
	  tycon = Tycon.toAst tycon,
	  cons = List.map(cons, fn {con, arg} =>
			  (Con.toAst con, Type.optionToAst arg))}
   end

(*---------------------------------------------------*)
(*                      Program                      *)
(*---------------------------------------------------*)

structure Program =
   struct
      datatype t = T of {datatypes: Datatype.t list,
			 body: Exp.t}

      fun size(T{body, ...}) = Exp.size body

      fun toAst(T{datatypes, body}) =
	 let val body = Exp.toAst body
	 in case datatypes of
	    [] => body
	  | _ => Aexp.lett([Adec.datatypee(List.map(datatypes, Datatype.toAst))],
			   body)
	 end

      fun clear(T{datatypes, body}) =
	 (List.foreach(datatypes, fn {tycon, tyvars, cons} =>
		       (Tycon.clear tycon;
			List.foreach(tyvars, Tyvar.clear);
			List.foreach(cons, Con.clear o #con)));
	  Exp.clear body)

      fun layout p = Ast.Exp.layout(toAst p)

      val empty = T{datatypes = [],
		    body = Exp.unit()}

      fun layoutStats(T{datatypes, body}) =
	 let
	    val numTypes = ref 0
	    fun inc _ = numTypes := 1 + !numTypes
	    val {hom, destroy} = Type.makeHom{var = inc, con = inc}
	    val numPrimExps = ref 0
	    open Layout
	 in
	    List.foreach(datatypes, fn {cons, ...} =>
			 List.foreach(cons, fn {arg, ...} =>
				      case arg of
					 NONE => ()
				       | SOME t => hom t))
	    ; (Exp.foreach
	       {exp = body,
		handlePrimExp = fn _ => numPrimExps := 1 + !numPrimExps,
		handleVarExp = fn _ => (),
		handleBoundVar = hom o #3,
		handleExp = fn _ => ()})
	    ; destroy()
	    ; align[seq[str "size = ", Int.layout(!numPrimExps)],
		    seq[str "num types in program = ", Int.layout(!numTypes)],
		    Type.stats()]
	 end
   end

end
