(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* binding occurences:
 *   1. lambda arg
 *   2. pattern arg
 *   3. MonoVal dec
 *   4. PolyVal dec
 *   5. Fun dec
 *   6. Handle catch
 *)

signature XML_TREE_STRUCTS =
   sig
      include ATOMS
   end

signature XML_TREE =
   sig
      include XML_TREE_STRUCTS

      structure Type: XML_TYPE
      sharing Atoms = Type.Atoms
	 
      structure Pat:
	 sig
	    datatype t = T of {con: Con.t,
			       targs: Type.t list,
			       arg: (Var.t * Type.t) option}
	 
	    val falsee: t
	    val truee: t
	    val con: t -> Con.t
	    val toAst: t -> Ast.Pat.t
	    val layout: t -> Layout.t
	 end

      structure Lambda:
	 sig
	    type exp
	    include T

	    (* constructors *)
	    val new: {arg: Var.t,
		      argType: Type.t,
		      body: exp} -> t
	       
	    (* destructors *)
	    val dest: t -> {arg: Var.t,
			    argType: Type.t,
			    body: exp}
	    val arg: t -> Var.t
	    val argType: t -> Type.t
	    val body: t -> exp

	    val plist: t -> PropertyList.t
	 end

      (* VarExp is a type application.
       * variable (applied to type args) paired with its type.
       *)
      structure VarExp:
	 sig
	    datatype t = T of {var: Var.t, targs: Type.t list}

	    (* constructors *)
	    val mono: Var.t -> t

	    (* destructors *)
	    val var: t -> Var.t
	    val targs: t -> Type.t list

	    val layout: t -> Layout.t
	 end

      structure PrimExp:
	 sig
	    type exp = Lambda.exp
	    datatype t =
	       Const of Const.t
	     | Var of VarExp.t
	     | Tuple of VarExp.t list
	     | Select of {tuple: VarExp.t,
			  offset: int}
	     | Lambda of Lambda.t
	     | ConApp of {con: Con.t,
			  targs: Type.t list,
			  arg: VarExp.t option}
	     (* primitives can be curried, and we require that they are fully
	      * applied.  For example, "plus" takes one arg that is a tuple,
	      * but List.map takes two args.
	      *)
	     | PrimApp of {prim: Prim.t,
			   targs: Type.t list,
			   args: VarExp.t list}
	     | App of {func: VarExp.t,
		       arg: VarExp.t}
	     | Raise of VarExp.t
	     | Case of {test: VarExp.t,
			cases: (Pat.t * exp) list,
			default: exp option}
	     | Handle of
	       {try: exp,
		catch: (Var.t * Type.t), (* binds exception in handler *)
		handler: exp}

	    val layout: t -> Layout.t
	 end
	       
      structure Dec:
	 sig
	    type exp = Lambda.exp
	       
	    datatype t =
	       MonoVal of {var: Var.t,
			   ty: Type.t,
			   exp: PrimExp.t}
	     | PolyVal of {var: Var.t,
			   tyvars: Tyvar.t list,
			   ty: Type.t,
			   exp: exp}
	     | Fun of {tyvars: Tyvar.t list,
		       decs: {var: Var.t,
			      ty: Type.t,
			      lambda: Lambda.t} list}
	     | Exception of {con: Con.t,
			     arg: Type.t option}

	    val toAst: t -> Ast.Dec.t
	    val layout: t -> Layout.t
	 end

      structure Exp:
	 sig
	    type t = Lambda.exp

	    (* destructors *)
	    val dest: t -> {decs: Dec.t list,
			    result: VarExp.t}
	    val decs: t -> Dec.t list
	    val result: t -> VarExp.t
	       
	    (* constructors *)
	    val new: {decs: Dec.t list,
		      result: VarExp.t} -> t

	    val fromPrimExp: PrimExp.t * Type.t -> t

	    val prefix: t * Dec.t -> t
	       
	    (* foreach{exp, handleExp, handleBoundVar, handleVarExp}
	     * applies handleExp to each subexpresison of e (including e)
	     * applies handleBoundVar to each variable bound in e
	     * applies handleVarExp to each variable expression in e
	     * handleBoundVar will be called on a variable binding before
	       handleVarExp is called on any occurrences
	     * handleExp is called on an expression after it is called on
		all of its subexpressions *)
	    val foreach:
	       {exp: t,
		handleExp: t -> unit,
		handlePrimExp: PrimExp.t -> unit,
		handleBoundVar: Var.t * Tyvar.t list * Type.t -> unit,
		handleVarExp: VarExp.t -> unit} -> unit

	    val foreachExp: t * (t -> unit) -> unit
	    val foreachPrimExp: t * (PrimExp.t -> unit) -> unit
	    val foreachVarExp: t * (VarExp.t -> unit) -> unit
	    val foreachBoundVar:
	       t * (Var.t * Tyvar.t list * Type.t -> unit) -> unit

	    val size: t -> int
	       
	    val clear: t -> unit
	    val layout: t -> Layout.t
	 end

      structure DirectExp:
	 sig
	    type t

	    val const: Const.t -> t
	    val var: {var: Var.t, targs: Type.t list, ty: Type.t} -> t
	    val varExp: VarExp.t * Type.t -> t
	    val monoVar: Var.t * Type.t -> t
	    val tuple: {exps: t list, ty: Type.t} -> t
	    val select: {tuple: t,
			 offset: int,
			 ty: Type.t} -> t
	    val conApp: {con: Con.t,
			 targs: Type.t list,
			 arg: t option,
			 ty: Type.t} -> t
	    val app: {func: t,
		      arg: t,
		      ty: Type.t} -> t
	    val primApp: {prim: Prim.t,
			  targs: Type.t list,
			  args: t list,
			  ty: Type.t} -> t
	    val raisee: t * Type.t -> t
	    val handlee: {try: t,
			  ty: Type.t,
			  catch: Var.t * Type.t,
			  handler: t} -> t
	    val casee: {
			test: t,
			cases: (Pat.t * t) list,
			default: t option,
			ty: Type.t (* type of entire case expression *)
		       } -> t

	    val lett: {decs: Dec.t list, body: t} -> t
	    val let1: {var: Var.t, exp: t, body: t} -> t
	       
	    val lambda: {arg: Var.t,
			 argType: Type.t,
			 body: t,
			 bodyType: Type.t} -> t

	    val detuple: {tuple: t,
			  body: (VarExp.t * Type.t) list -> t} -> t

	    val detupleBind: {tuple: t,
			      components: Var.t list,
			      body: t} -> t
	       
	    val sequence: t list -> t
	    val seq: t list * (t list -> t) -> t

	    val unit: unit -> t
	    val reff: t -> t
	    val deref: t -> t
 	    val equal: t * t -> t
	    val iff: {test: t, thenn: t, elsee: t, ty: Type.t} -> t
	       
	    val vall: {var: Var.t, exp: t} -> Dec.t list

	    val toExp: t -> Exp.t
	    val fromExp: Exp.t * Type.t -> t

	    val layout: t -> Layout.t
	 end

      structure Program:
	 sig
	    datatype t =
	       T of {datatypes: {
				 tycon: Tycon.t,
				 tyvars: Tyvar.t list,
				 cons: {
					con: Con.t,
					arg: Type.t option
					} list
				 } list,
		     body: Exp.t}

	    val empty: t

	    (* Clear all property lists *)
	    val clear: t -> unit

	    val layoutStats: t -> Layout.t
	       
	    val layout: t -> Layout.t
	 end
   end
