structure Word31 = Word structure Int31 = Int
(* lib-base-sig.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *)

signature LIB_BASE =
  sig

    exception Unimplemented of string
	(* raised to report unimplemented features *)
    exception Impossible of string
	(* raised to report internal errors *)

    exception NotFound
	(* raised by searching operations *)

    val failure: {module: string, func: string, msg: string} -> 'a
	(* raise the exception Fail with a standard format message. *)

    val version: {date: string, system: string, version_id: int list}
    val banner: string

  end (* LIB_BASE *)

(* lib-base.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *)

structure LibBase: LIB_BASE =
  struct

  (* raised to report unimplemented features *)
    exception Unimplemented of string

  (* raised to report internal errors *)
    exception Impossible of string

  (* raised by searching operations *)
    exception NotFound

  (* raise the exception Fail with a standard format message. *)
    fun failure {module, func, msg} =
	  raise (Fail(concat[module, ".", func, ": ", msg]))

    val version = {
	    date = "June 1, 1996", 
	    system = "SML/NJ Library",
	    version_id = [1, 0]
	  }

    fun f ([], l) = l
      | f ([x: int], l) = (Int.toString x) :: l
      | f (x :: r, l) = (Int.toString x) :: "." :: f(r, l)

    val banner = concat (
	    #system version :: ", Version " ::
	    f (#version_id version, [", ", #date version]))

  end (* LibBase *)

(* random-sig.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *)

signature RANDOM =
  sig

    type rand
	(* the internal state of a random number generator *)

    val rand: (int * int) -> rand
	(* create rand from initial seed *)

    val toString: rand -> string
    val fromString: string -> rand
        (* convert state to and from string
         * fromString raises Fail if its argument
         * does not have the proper form.
         *)

    val randInt: rand -> int
	(* generate ints uniformly in [minInt, maxInt] *)

    val randNat: rand -> int
	(* generate ints uniformly in [0, maxInt] *)

    val randReal: rand -> real
	(* generate reals uniformly in [0.0, 1.0) *)

    val randRange: (int * int) -> rand -> int
	(* randRange (lo, hi) generates integers uniformly [lo, hi].
	 * Raises Fail if hi < lo.
	 *)

  end; (* RANDOM *)

(* random.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *
 * This package implements a random number generator using a subtract-with-borrow
 * (SWB) generator as described in Marsaglia and Zaman, "A New Class of Random Number
 * Generators, " Ann. Applied Prob. 1(3), 1991, pp. 462-480.
 * 
 * The SWB generator is a 31-bit generator with lags 48 and 8. It has period 
 * (2^1487 - 2^247)/105 or about 10^445. In general, these generators are
 * excellent. However, they act locally like a lagged Fibonacci generator
 * and thus have troubles with the birthday test. Thus, we combine this SWB
 * generator with the linear congruential generator (48271*a)mod(2^31-1).
 *
 * Although the interface is fairly abstract, the implementation uses 
 * 31-bit ML words. At some point, it might be good to use 32-bit words.
 *)

structure Random: RANDOM =
  struct
    structure A   = Array
    structure LW  = LargeWord
    structure W8A = Word8Array
    structure W8V = Word8Vector
    structure P   = Pack32Big

    val << = Word31.<<
    val >> = Word31.>>
    val & = Word31.andb
    val ++ = Word31.orb
    val xorb = Word31.xorb
    infix << >> & ++

    val nbits = 31                                      (* bits per word *)
    val maxWord: Word31.word = 0wx7FFFFFFF             (* largest word *)
    val bit30: Word31.word   = 0wx40000000
    val lo30: Word31.word    = 0wx3FFFFFFF

    val N = 48
    val lag = 8
    val offset = N-lag

    fun error (f, msg) = LibBase.failure {module="Random", func=f, msg=msg}

    val two2neg30 = 1.0/((real 0x8000)*(real 0x8000))   (* 2^~30 *)

    fun minus(x, y, false) = (x - y, y > x)
      | minus(x, y, true) = (x - y - 0w1, y >= x)

    datatype rand = RND of {
        vals  : Word31.word A.array, (* seed array *)
        borrow: bool ref,           (* last borrow *)
        congx : Word31.word ref,    (* congruential seed *)
        index : int ref             (* index of next available value in vals *)
      }

      (* We represent state as a string, starting with an initial
       * word acting as an magic cookie (with bit 0 determining the
       * value of borrow), followed by a word containing index and a word
       * containing congx, followed by the seed array.
       *)
    val numWords = 3 + N
    val magic: LW.word = 0wx72646e64
    fun toString (RND{vals, borrow, congx, index}) = let
          val arr = W8A.array (4*numWords, 0w0)
          val word0 = if !borrow then LW.orb (magic, 0w1) else magic
          fun fill (src, dst) =
                if src = N then ()
                else (
                  P.update (arr, dst, Word31.toLargeWord (A.sub (vals, src)));
                  fill (src+1, dst+1)
                )
          in
            P.update (arr, 0, word0);
            P.update (arr, 1, LW.fromInt (!index));
            P.update (arr, 2, Word31.toLargeWord (!congx));
            fill (0, 3);
            Byte.bytesToString (W8A.extract (arr, 0, NONE))
          end

    fun fromString s = let
          val bytes = Byte.stringToBytes s
          val _ = if W8V.length bytes = 4 * numWords then ()
                  else error ("fromString", "invalid state string")
          val word0 = P.subVec (bytes, 0)
          val _ = if LW.andb(word0, 0wxFFFFFFFE) = magic then ()
                  else error ("fromString", "invalid state string")
          fun subVec i = P.subVec (bytes, i)
          val borrow = ref (LW.andb(word0, 0w1) = 0w1)
          val index = ref (LW.toInt (subVec 1))
          val congx = ref (Word31.fromLargeWord (subVec 2))
          val arr = A.array (N, 0w0: Word31.word)
          fun fill (src, dst) =
                if dst = N then ()
                else (
                  A.update (arr, dst, Word31.fromLargeWord (subVec src));
                  fill (src+1, dst+1)
                )
          in
            fill (3, 0);
            RND{vals = arr,
                index = index, 
                congx = congx, 
                borrow = borrow}
          end

      (* linear congruential generator:
       * multiplication by 48271 mod (2^31 - 1) 
       *)
    val a: Word31.word = 0w48271
    val m: Word31.word = 0w2147483647
    val q = m div a
    val r = m mod a
    fun lcg seed = let
          val left = a * (seed mod q)
          val right = r * (seed div q)
          in
            if left > right then left - right
            else (m - right) + left
          end

      (* Fill seed array using subtract-with-borrow generator:
       *  x[n] = x[n-lag] - x[n-N] - borrow
       * Sets index to 1 and returns 0th value.
       *)
    fun fill (RND{vals, index, congx, borrow}) = let
          fun update (ix, iy, b) = let
                val (z, b') = minus(A.sub(vals, ix), A.sub(vals, iy), b)
                in
                  A.update(vals, iy, z); b'
                end
          fun fillup (i, b) =
                if i = lag then b
                else fillup(i+1, update(i+offset, i, b))
          fun fillup' (i, b) =
                if i = N then b
                else fillup'(i+1, update(i-lag, i, b))
          in
            borrow := fillup' (lag, fillup (0, !borrow));
            index := 1;
            A.sub(vals, 0)
          end

      (* Create initial seed array and state of generator.
       * Fills the seed array one bit at a time by taking the leading 
       * bit of the xor of a shift register and a congruential sequence. 
       * The congruential generator is (c*48271) mod (2^31 - 1).
       * The shift register generator is c(I + L18)(I + R13).
       * The same congruential generator continues to be used as a 
       * mixing generator with the SWB generator.
       *)
    fun rand (congy, shrgx) = let
          fun mki (i, c, s) = let
                val c' = lcg c
                val s' = xorb(s, s << 0w18)
                val s'' = xorb(s', s' >> 0w13)
                val i' = (lo30 & (i >> 0w1)) ++ (bit30 & xorb(c', s''))
                in (i', c', s'') end
	  fun iterate (0, v) = v
	    | iterate (n, v) = iterate(n-1, mki v)
          fun mkseed (congx, shrgx) = iterate (nbits, (0w0, congx, shrgx))
          fun genseed (0, seeds, congx, _) = (seeds, congx)
            | genseed (n, seeds, congx, shrgx) = let
                val (seed, congx', shrgx') = mkseed (congx, shrgx)
                in genseed(n-1, seed :: seeds, congx', shrgx') end
          val congx = ((Word31.fromInt congy & maxWord) << 0w1)+0w1
          val (seeds, congx) = genseed(N, [], congx, Word31.fromInt shrgx)
          in
            RND{vals = A.fromList seeds, 
                index = ref 0, 
                congx = ref congx, 
                borrow = ref false}
          end

      (* Get next random number. The tweak function combines
       * the number from the SWB generator with a number from
       * the linear congruential generator.
       *)
    fun randWord (r as RND{vals, index, congx, ...}) = let
         val idx = !index
         fun tweak i = let
               val c = lcg (!congx)
               in
                 congx := c;
                 xorb(i, c)
               end
         in
           if idx = N then tweak(fill r)
           else tweak(A.sub(vals, idx)) before index := idx+1
         end

    fun randInt state = Word31.toIntX(randWord state)
    fun randNat state = Word31.toIntX(randWord state & lo30)
    fun randReal state =
      (real(randNat state) + real(randNat state) * two2neg30) * two2neg30

    fun randRange (i, j) = 
          if j < i 
            then error ("randRange", "hi < lo")
            else
(* sweeks commented out the following because MLton doesn't handle overflow. *)
(* 		  
 * 		  let val R = two2neg30*real(j - i + 1)
 *               in
 *                 fn s => i + trunc(R*real(randNat s))
 *               end handle _ =>
 *)
	       let val ri = real i
		  val R = (real j)-ri+1.0
	       in
                  fn s => trunc(ri + R*(randReal s))
	       end

  end; (* Random *)

(* hash-string.sml
 *
 * COPYRIGHT (c) 1992 by AT&T Bell Laboratories
 *)

structure HashString: sig

    val hashString: string -> word

  end = struct

    fun charToWord c = Word.fromInt(Char.ord c)

  (* A function to hash a character.  The computation is:
   *
   *   h = 33 * h + 720 + c
   *)
    fun hashChar (c, h) = Word.<<(h, 0w5) + h + 0w720 + (charToWord c)

    fun hashString s = CharVector.foldl hashChar 0w0 s
	  
  end (* HashString *)
(* hash-table-rep.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 * COPYRIGHT (c) 1996 AT&T Research.
 *
 * This is the internal representation of hash tables, along with some
 * utility functions.  It is used in both the polymorphic and functor
 * hash table implementations.
 *
 * AUTHOR:  John Reppy
 *	    AT&T Bell Laboratories
 *	    Murray Hill, NJ 07974
 *	    jhr@research.att.com
 *)

structure HashTableRep: sig

    datatype ('a, 'b) bucket
      = NIL
      | B of (word * 'a * 'b * ('a, 'b) bucket)

    type ('a, 'b) table = ('a, 'b) bucket array

    val alloc: int -> ('a, 'b) table
	(* allocate a table of at least the given size *)

    val growTable: (('a, 'b) table * int) -> ('a, 'b) table
	(* grow a table to the specified size *)

    val growTableIfNeeded: (('a, 'b) table ref * int) -> bool
	(* conditionally grow a table; the second argument is the number
	 * of items currently in the table.
	 *)

    val clear: ('a, 'b) table -> unit
	(* remove all items *)

    val listItems : (('a, 'b) table * int ref) -> 'b list
    val listItemsi: (('a, 'b) table * int ref) -> ('a * 'b) list


    val appi: ('a * 'b -> 'c) -> ('a, 'b) table -> unit
    val app: ('a -> 'b) -> ('c, 'a) table -> unit

    val mapi: ('a * 'b -> 'c) -> ('a, 'b) table -> ('a, 'c) table
    val map: ('a -> 'b) -> ('c, 'a) table -> ('c, 'b) table

    val foldi: ('a * 'b * 'c -> 'c) -> 'c -> ('a, 'b) table -> 'c
    val fold: ('a * 'b -> 'b) -> 'b -> ('c, 'a) table -> 'b

    val filteri: ('a * 'b -> bool) -> ('a, 'b) table -> unit
    val filter: ('a -> bool) -> ('b, 'a) table -> unit

    val copy: ('a, 'b) table -> ('a, 'b) table

    val bucketSizes: ('a, 'b) table -> int list

  end = struct

    datatype ('a, 'b) bucket
      = NIL
      | B of (word * 'a * 'b * ('a, 'b) bucket)

    type ('a, 'b) table = ('a, 'b) bucket array

    fun index (i, sz) = Word.toIntX(Word.andb(i, Word.fromInt sz - 0w1))

  (* find smallest power of 2 (>= 32) that is >= n *)
    fun roundUp n = let
	  fun f i = if (i >= n) then i else f(i * 2)
	  in
	    f 32
	  end

  (* Create a new table; the int is a size hint and the exception
   * is to be raised by find.
   *)
    fun alloc sizeHint = Array.array(roundUp sizeHint, NIL)

  (* grow a table to the specified size *)
    fun growTable (table, newSz) = let
	  val newArr = Array.array (newSz, NIL)
	  fun copy NIL = ()
	    | copy (B(h, key, v, rest)) = let
		val indx = index (h, newSz)
		in
		  Array.update (newArr, indx,
		    B(h, key, v, Array.sub(newArr, indx)));
		  copy rest
		end
	  in
	    Array.app copy table;
	    newArr
	  end

  (* conditionally grow a table; return true if it grew. *)
    fun growTableIfNeeded (table, nItems) = let
	    val arr = !table
	    val sz = Array.length arr
	    in
	      if (nItems >= sz)
		then (table := growTable (arr, sz+sz); true)
		else false
	    end

  (* remove all items *)
    fun clear table = Array.modify (fn _ => NIL) table

  (* return a list of the items in the table *)
    fun listItems (table, nItems) = let
	  fun f (_, l, 0) = l
	    | f (~1, l, _) = l
	    | f (i, l, n) = let
		fun g (NIL, l, n) = f (i-1, l, n)
		  | g (B(_, k, v, r), l, n) = g(r, v :: l, n-1)
		in
		  g (Array.sub(table, i), l, n)
		end
	  in
	    f ((Array.length table) - 1, [], !nItems)
	  end (* listItems *)
    fun listItemsi (table, nItems) = let
	  fun f (_, l, 0) = l
	    | f (~1, l, _) = l
	    | f (i, l, n) = let
		fun g (NIL, l, n) = f (i-1, l, n)
		  | g (B(_, k, v, r), l, n) = g(r, (k, v) :: l, n-1)
		in
		  g (Array.sub(table, i), l, n)
		end
	  in
	    f ((Array.length table) - 1, [], !nItems)
	  end (* listItems *)

  (* Apply a function to the entries of the table *)
    fun appi f table = let
	  fun appF NIL = ()
	    | appF (B(_, key, item, rest)) = (f (key, item); appF rest)
	  in
	    Array.app appF table
	  end (* appi *)
    fun app f table = let
	  fun appF NIL = ()
	    | appF (B(_, key, item, rest)) = (f item; appF rest)
	  in
	    Array.app appF table
	  end (* app *)

  (* Map a table to a new table that has the same keys *)
    fun mapi f table = let
	  fun mapF NIL = NIL
	    | mapF (B(hash, key, item, rest)) =
		B(hash, key, f (key, item), mapF rest)
	  val newTbl = Array.tabulate (
		Array.length table,
		fn i => mapF (Array.sub(table, i)))
	  in
	    newTbl
	  end (* transform *)

  (* Map a table to a new table that has the same keys *)
    fun map f table = let
	  fun mapF NIL = NIL
	    | mapF (B(hash, key, item, rest)) = B(hash, key, f item, mapF rest)
	  val newTbl = Array.tabulate (
		Array.length table,
		fn i => mapF (Array.sub(table, i)))
	  in
	    newTbl
	  end (* map *)

    fun foldi f init table = let
	  fun foldF (NIL, accum) = accum
	    | foldF (B(hash, key, item, rest), accum) =
		foldF(rest, f(key, item, accum))
	  in
	    Array.foldl foldF init table
	  end
    fun fold f init table = let
	  fun foldF (NIL, accum) = accum
	    | foldF (B(hash, key, item, rest), accum) =
		foldF(rest, f(item, accum))
	  in
	    Array.foldl foldF init table
	  end

  (* remove any hash table items that do not satisfy the given
   * predicate.
   *)
    fun filteri pred table = let
	  fun filterP NIL = NIL
	    | filterP (B(hash, key, item, rest)) = if (pred(key, item))
		then B(hash, key, item, filterP rest)
		else filterP rest
	  in
	    Array.modify filterP table
	  end (* filteri *)
    fun filter pred table = let
	  fun filterP NIL = NIL
	    | filterP (B(hash, key, item, rest)) = if (pred item)
		then B(hash, key, item, filterP rest)
		else filterP rest
	  in
	    Array.modify filterP table
	  end (* filter *)

  (* Create a copy of a hash table *)
    fun copy table =
	  Array.tabulate (Array.length table, fn i => Array.sub(table, i));

  (* returns a list of the sizes of the various buckets.  This is to
   * allow users to gauge the quality of their hashing function.
   *)
    fun bucketSizes table = let
	  fun len (NIL, n) = n
	    | len (B(_, _, _, r), n) = len(r, n+1)
	  in
	    Array.foldr (fn (b, l) => len(b, 0) :: l) [] table
	  end

  end (* HashTableRep *)
(* hash-table-sig.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 *
 * The signature of the polymorphic hash table structure.
 *
 * AUTHOR:  John Reppy
 *	    AT&T Bell Laboratories
 *	    Murray Hill, NJ 07974
 *	    jhr@research.att.com
 *)

signature HASH_TABLE =
  sig

    type ('a, 'b) hash_table
	(* type of hash table mapping 'a to 'b *)

    val mkTable: (('a -> word) * (('a * 'a) -> bool)) -> (int * exn)
	  -> ('a, 'b) hash_table
	(* Given a hashing function and an equality predicate, create a new table;
	 * the int is a size hint and the exception is to be raised by find.
	 *)

    val clear: ('a, 'b) hash_table -> unit
	(* remove all elements from the table *)

    val insert: ('a, 'b) hash_table -> ('a * 'b) -> unit
	(* Insert an item.  If the key already has an item associated with it,
	 * then the old item is discarded.
	 *)

    val lookup: ('a, 'b) hash_table -> 'a -> 'b
	(* Find an item, the table's exception is raised if the item doesn't exist *)

    val find: ('a, 'b) hash_table -> 'a -> 'b option
	(* Look for an item, return NONE if the item doesn't exist *)

    val remove: ('a, 'b) hash_table -> 'a -> 'b
	(* Remove an item, returning the item.  The table's exception is raised if
	 * the item doesn't exist.
	 *)

    val numItems: ('a, 'b) hash_table ->  int
	(* Return the number of items in the table *)

    val listItems : ('a, 'b) hash_table -> 'b list
    val listItemsi: ('a, 'b) hash_table -> ('a * 'b) list
	(* Return a list of the items (and their keys) in the table *)

    val app : ('b -> unit) -> ('a, 'b) hash_table -> unit
    val appi: (('a * 'b) -> unit) -> ('a, 'b) hash_table -> unit
	(* Apply a function to the entries of the table *)

    val map : ('b -> 'c) -> ('a, 'b) hash_table -> ('a, 'c) hash_table
    val mapi: (('a * 'b) -> 'c) -> ('a, 'b) hash_table -> ('a, 'c) hash_table
	(* Map a table to a new table that has the same keys *)

    val fold : (('b *'c) -> 'c) -> 'c -> ('a, 'b) hash_table -> 'c
    val foldi: (('a * 'b * 'c) -> 'c) -> 'c -> ('a, 'b) hash_table -> 'c
	(* Fold a function over the elements of a table *)

    val filter : ('b -> bool) -> ('a, 'b) hash_table -> unit
    val filteri: (('a * 'b) -> bool) -> ('a, 'b) hash_table -> unit
	(* remove any hash table items that do not satisfy the given
	 * predicate.
	 *)

    val copy: ('a, 'b) hash_table -> ('a, 'b) hash_table
	(* Create a copy of a hash table *)

    val bucketSizes: ('a, 'b) hash_table -> int list
	(* returns a list of the sizes of the various buckets.  This is to
	 * allow users to gauge the quality of their hashing function.
	 *)

  end (* HASH_TABLE *)
(* hash-table.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 *
 * Polymorphic hash tables.
 *
 * AUTHOR:  John Reppy
 *	    AT&T Bell Laboratories
 *	    Murray Hill, NJ 07974
 *	    jhr@research.att.com
 *)

structure HashTable: HASH_TABLE =
  struct

    structure HTRep = HashTableRep

    datatype ('a, 'b) hash_table = HT of {
	hash_fn: 'a -> word,
	eq_pred: ('a * 'a) -> bool,
	not_found: exn,
	table: ('a, 'b) HTRep.table ref,
	n_items: int ref
      }

    fun index (i, sz) = Word.toIntX(Word.andb(i, Word.fromInt sz - 0w1))

  (* find smallest power of 2 (>= 32) that is >= n *)
    fun roundUp n = let
	  fun f i = if (i >= n) then i else f(i * 2)
	  in
	    f 32
	  end

  (* Create a new table; the int is a size hint and the exception
   * is to be raised by find.
   *)
    fun mkTable (hash, eq) (sizeHint, notFound) = HT{
	    hash_fn = hash,
	    eq_pred = eq,
	    not_found = notFound,
	    table = ref (HTRep.alloc sizeHint),
	    n_items = ref 0
	  }

  (* remove all elements from the table *)
    fun clear (HT{table, n_items, ...}) = (HTRep.clear(!table); n_items := 0)

  (* Insert an item.  If the key already has an item associated with it,
   * then the old item is discarded.
   *)
    fun insert (tbl as HT{hash_fn, eq_pred, table, n_items, ...}) (key, item) = let
	  val arr = !table
	  val sz = Array.length arr
	  val hash = hash_fn key
	  val indx = index (hash, sz)
	  fun look HTRep.NIL = (
		Array.update(arr, indx,
		  HTRep.B(hash, key, item, Array.sub(arr, indx)));
		n_items := !n_items + 1;
		HTRep.growTableIfNeeded (table, !n_items);
		HTRep.NIL)
	    | look (HTRep.B(h, k, v, r)) = if ((hash = h) andalso eq_pred(key, k))
		then HTRep.B(hash, key, item, r)
		else (case (look r)
		   of HTRep.NIL => HTRep.NIL
		    | rest => HTRep.B(h, k, v, rest)
		  (* end case *))
	  in
	    case (look (Array.sub (arr, indx)))
	     of HTRep.NIL => ()
	      | b => Array.update(arr, indx, b)
	  end

  (* find an item, the table's exception is raised if the item doesn't exist *)
    fun lookup (HT{hash_fn, eq_pred, table, not_found, ...}) key = let
	  val arr = !table
	  val sz = Array.length arr
	  val hash = hash_fn key
	  val indx = index (hash, sz)
	  fun look HTRep.NIL = raise not_found
	    | look (HTRep.B(h, k, v, r)) = if ((hash = h) andalso eq_pred(key, k))
		then v
		else look r
	  in
	    look (Array.sub (arr, indx))
	  end

  (* look for an item, return NONE if the item doesn't exist *)
    fun find (HT{hash_fn, eq_pred, table, ...}) key = let
	  val arr = !table
	  val sz = Array.length arr
	  val hash = hash_fn key
	  val indx = index (hash, sz)
	  fun look HTRep.NIL = NONE
	    | look (HTRep.B(h, k, v, r)) = if ((hash = h) andalso eq_pred(key, k))
		then SOME v
		else look r
	  in
	    look (Array.sub (arr, indx))
	  end

  (* Remove an item.  The table's exception is raised if
   * the item doesn't exist.
   *)
    fun remove (HT{hash_fn, eq_pred, not_found, table, n_items}) key = let
	  val arr = !table
	  val sz = Array.length arr
	  val hash = hash_fn key
	  val indx = index (hash, sz)
	  fun look HTRep.NIL = raise not_found
	    | look (HTRep.B(h, k, v, r)) = if ((hash = h) andalso eq_pred(key, k))
		then (v, r)
		else let val (item, r') = look r in (item, HTRep.B(h, k, v, r')) end
	  val (item, bucket) = look (Array.sub (arr, indx))
	  in
	    Array.update (arr, indx, bucket);
	    n_items := !n_items - 1;
	    item
	  end (* remove *)

  (* Return the number of items in the table *)
   fun numItems (HT{n_items, ...}) = !n_items

  (* return a list of the items in the table *)
    fun listItems (HT{table = ref arr, n_items, ...}) =
	  HTRep.listItems (arr, n_items)
    fun listItemsi (HT{table = ref arr, n_items, ...}) =
	  HTRep.listItemsi (arr, n_items)

  (* Apply a function to the entries of the table *)
    fun appi f (HT{table, ...}) = HTRep.appi f (! table)
    fun app f (HT{table, ...}) = HTRep.app f (! table)

  (* Map a table to a new table that has the same keys and exception *)
    fun mapi f (HT{hash_fn, eq_pred, table, n_items, not_found}) = HT{
	    hash_fn = hash_fn, eq_pred = eq_pred,
	    table = ref(HTRep.mapi f (! table)),
	    n_items = ref(!n_items),
	    not_found = not_found
	  }

  (* Map a table to a new table that has the same keys and exception *)
    fun map f (HT{hash_fn, eq_pred, table, n_items, not_found}) = HT{
	    hash_fn = hash_fn, eq_pred = eq_pred,
	    table = ref(HTRep.map f (! table)),
	    n_items = ref(!n_items),
	    not_found = not_found
	  }

  (* Fold a function over the entries of the table *)
    fun foldi f init (HT{table, ...}) = HTRep.foldi f init (! table)
    fun fold f init (HT{table, ...}) = HTRep.fold f init (! table)

  (* remove any hash table items that do not satisfy the given
   * predicate.
   *)
    fun filteri pred (HT{table, ...}) = HTRep.filteri pred (! table)
    fun filter pred (HT{table, ...}) = HTRep.filter pred (! table)

  (* Create a copy of a hash table *)
    fun copy (HT{hash_fn, eq_pred, table, n_items, not_found}) =HT{
	    hash_fn = hash_fn, eq_pred = eq_pred,
	    table = ref(HTRep.copy (! table)), n_items = ref(!n_items),
	    not_found = not_found
	  }

  (* returns a list of the sizes of the various buckets.  This is to
   * allow users to gauge the quality of their hashing function.
   *)
    fun bucketSizes (HT{table, ...}) = HTRep.bucketSizes(! table)

  end (* HashTable *)
(* splaytree-sig.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *
 * Signature for a splay tree data structure.
 *
 *)

signature SPLAY_TREE = 
  sig
    datatype 'a splay = 
      SplayObj of {
        value: 'a,
        right: 'a splay,
        left: 'a splay
      }
    | SplayNil


    val splay: (('a -> order) * 'a splay) -> (order * 'a splay)
      (* (r, tree') = splay (cmp, tree) 
       * where tree' is tree adjusted using the comparison function cmp
       * and, if tree' = SplayObj{value, ...}, r = cmp value.
       * tree' = SplayNil iff tree = SplayNil, in which case r is undefined.
       *)

    val join: 'a splay * 'a splay -> 'a splay
      (* join(t, t') returns a new splay tree formed of t and t'
       *)

  end (* SPLAY_TREE *)

(* splaytree.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *
 * Splay tree structure.
 *
 *)

structure SplayTree: SPLAY_TREE = 
  struct

    datatype 'a splay = 
      SplayObj of {
        value: 'a,
        right: 'a splay,
        left: 'a splay
      }
    | SplayNil

    datatype 'a ans_t = No | Eq of 'a | Lt of 'a | Gt of 'a

    fun splay (compf, root) = let
        fun adj SplayNil = (No, SplayNil, SplayNil)
          | adj (arg as SplayObj{value, left, right}) =
              (case compf value of
                EQUAL => (Eq value, left, right)
              | GREATER =>
                  (case left of
                    SplayNil => (Gt value, SplayNil, right)
                  | SplayObj{value=value', left=left', right=right'} =>
                      (case compf value' of
                        EQUAL => (Eq value', left',
                                    SplayObj{value=value, left=right', right=right})
                      | GREATER =>
                          (case left' of 
                            SplayNil => (Gt value', left', SplayObj{value=value, left=right', right=right})
                          | _ => 
                            let val (V, L, R) = adj left'
                                val rchild = SplayObj{value=value, left=right', right=right}
                            in
                              (V, L, SplayObj{value=value', left=R, right=rchild})
                            end
                          ) (* end case *)
                      | _ =>
                          (case right' of 
                            SplayNil => (Lt value', left', SplayObj{value=value, left=right', right=right})
                          | _ =>
                            let val (V, L, R) = adj right'
                                 val rchild = SplayObj{value=value, left=R, right=right}
                                 val lchild = SplayObj{value=value', left=left', right=L}
                            in
                              (V, lchild, rchild)
                            end
                          ) (* end case *)
                      ) (* end case *)
                  ) (* end case *)
              | _ =>
                 (case right of
                   SplayNil => (Lt value, left, SplayNil)
                 | SplayObj{value=value', left=left', right=right'} =>
                     (case compf value' of
                       EQUAL =>
                         (Eq value', SplayObj{value=value, left=left, right=left'}, right')
                     | LESS =>
                         (case right' of
                           SplayNil => (Lt value', SplayObj{value=value, left=left, right=left'}, right')
                         | _ =>
                           let val (V, L, R) = adj right'
                               val lchild = SplayObj{value=value, left=left, right=left'}
                           in
                             (V, SplayObj{value=value', left=lchild, right=L}, R)
                           end
                         ) (* end case *)
                     | _ =>
                         (case left' of
                           SplayNil => (Gt value', SplayObj{value=value, left=left, right=left'}, right')
                         | _ =>
                           let val (V, L, R) = adj left'
                               val rchild = SplayObj{value=value', left=R, right=right'}
                               val lchild = SplayObj{value=value, left=left, right=L}
                           in
                             (V, lchild, rchild)
                           end
                         ) (* end case *)
                     ) (* end case *)
                 ) (* end case *)
              ) (* end case *)
      in
        case adj root of
          (No, _, _) => (GREATER, SplayNil)
        | (Eq v, l, r) => (EQUAL, SplayObj{value=v, left=l, right=r})
        | (Lt v, l, r) => (LESS, SplayObj{value=v, left=l, right=r})
        | (Gt v, l, r) => (GREATER, SplayObj{value=v, left=l, right=r})
      end

    fun lrotate SplayNil = SplayNil
      | lrotate (arg as SplayObj{value, left, right=SplayNil}) = arg
      | lrotate (SplayObj{value, left, right=SplayObj{value=v, left=l, right=r}}) = 
          lrotate (SplayObj{value=v, left=SplayObj{value=value, left=left, right=l}, right=r})

    fun join (SplayNil, SplayNil) = SplayNil
      | join (SplayNil, t) = t
      | join (t, SplayNil) = t
      | join (l, r) =
          case lrotate l of
            SplayNil => r      (* impossible as l is not SplayNil *)
          | SplayObj{value, left, right} => SplayObj{value=value, left=left, right=r}

  end (* SplayTree *)

(* ord-key-sig.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *
 * Abstract linearly ordered keys.
 *
 *)

signature ORD_KEY =
  sig
    type ord_key

    val compare: ord_key * ord_key -> order

  end (* ORD_KEY *)
(* ord-map-sig.sml
 *
 * COPYRIGHT (c) 1996 by AT&T Research.  See COPYRIGHT file for details.
 *
 * Abstract signature of an applicative-style finite maps (dictionaries)
 * structure over ordered monomorphic keys.
 *)

signature ORD_MAP =
  sig

    structure Key: ORD_KEY

    type 'a map

    val empty: 'a map
	(* The empty map *)

    val insert : 'a map * Key.ord_key * 'a -> 'a map
    val insert': ((Key.ord_key * 'a) * 'a map) -> 'a map
	(* Insert an item. *)

    val find: 'a map * Key.ord_key -> 'a option
	(* Look for an item, return NONE if the item doesn't exist *)

    val remove: 'a map * Key.ord_key -> 'a map * 'a
	(* Remove an item, returning new map and value removed.
         * Raises LibBase.NotFound if not found.
	 *)

    val numItems: 'a map ->  int
	(* Return the number of items in the map *)

    val listItems : 'a map -> 'a list
    val listItemsi: 'a map -> (Key.ord_key * 'a) list
	(* Return an ordered list of the items (and their keys) in the map.
         *)

    val collate: ('a * 'a -> order) -> ('a map * 'a map) -> order
	(* given an ordering on the map's range, return an ordering
	 * on the map.
	 *)

    val unionWith : ('a * 'a -> 'a) -> ('a map * 'a map) -> 'a map
    val unionWithi: (Key.ord_key * 'a * 'a -> 'a) -> ('a map * 'a map) -> 'a map
	(* return a map whose domain is the union of the domains of the two input
	 * maps, using the supplied function to define the map on elements that
	 * are in both domains.
	 *)

    val intersectWith : ('a * 'b -> 'c) -> ('a map * 'b map) -> 'c map
    val intersectWithi: (Key.ord_key * 'a * 'b -> 'c) -> ('a map * 'b map) -> 'c map
	(* return a map whose domain is the intersection of the domains of the
	 * two input maps, using the supplied function to define the range.
	 *)

    val app : ('a -> unit) -> 'a map -> unit
    val appi: ((Key.ord_key * 'a) -> unit) -> 'a map -> unit
	(* Apply a function to the entries of the map in map order. *)

    val map : ('a -> 'b) -> 'a map -> 'b map
    val mapi: (Key.ord_key * 'a -> 'b) -> 'a map -> 'b map
	(* Create a new map by applying a map function to the
         * name/value pairs in the map.
         *)

    val foldl : ('a * 'b -> 'b) -> 'b -> 'a map -> 'b
    val foldli: (Key.ord_key * 'a * 'b -> 'b) -> 'b -> 'a map -> 'b
	(* Apply a folding function to the entries of the map
         * in increasing map order.
         *)

    val foldr : ('a * 'b -> 'b) -> 'b -> 'a map -> 'b
    val foldri: (Key.ord_key * 'a * 'b -> 'b) -> 'b -> 'a map -> 'b
	(* Apply a folding function to the entries of the map
         * in decreasing map order.
         *)

    val filter : ('a -> bool) -> 'a map -> 'a map
    val filteri: (Key.ord_key * 'a -> bool) -> 'a map -> 'a map
	(* Filter out those elements of the map that do not satisfy the
	 * predicate.  The filtering is done in increasing map order.
	 *)

    val mapPartial : ('a -> 'b option) -> 'a map -> 'b map
    val mapPartiali: (Key.ord_key * 'a -> 'b option) -> 'a map -> 'b map
	(* map a partial function over the elements of a map in increasing
	 * map order.
	 *)

  end (* ORD_MAP *)
(* splay-map-fn.sml
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.  See COPYRIGHT file for details.
 *
 * Functor implementing dictionaries using splay trees.
 *
 *)

functor SplayMapFn (K: ORD_KEY): ORD_MAP =
  struct
    structure Key = K
    open SplayTree

    datatype 'a map = 
        EMPTY
      | MAP of {
        root: (K.ord_key * 'a) splay ref,
        nobj: int
      }

    fun cmpf k (k', _) = K.compare(k', k)

    val empty = EMPTY
    
	(* Insert an item.  
	 *)
    fun insert (EMPTY, key, v) =
          MAP{nobj=1, root=ref(SplayObj{value=(key, v), left=SplayNil, right=SplayNil})}
      | insert (MAP{root,nobj}, key, v) =
          case splay (cmpf key, !root) of
            (EQUAL, SplayObj{value, left, right}) => 
              MAP{nobj=nobj, root=ref(SplayObj{value=(key, v), left=left, right=right})}
          | (LESS, SplayObj{value, left, right}) => 
              MAP{
                nobj=nobj+1,
                root=ref(SplayObj{value=(key, v), left=SplayObj{value=value, left=left, right=SplayNil}, right=right})
              }
          | (GREATER, SplayObj{value, left, right}) => 
              MAP{
                nobj=nobj+1,
                root=ref(SplayObj{
                  value=(key, v),
                  left=left,
                  right=SplayObj{value=value, left=SplayNil, right=right}
                })
              }
          | (_, SplayNil) => raise LibBase.Impossible "SplayMapFn.insert SplayNil"
    fun insert' ((k, x), m) = insert(m, k, x)

  (* Look for an item, return NONE if the item doesn't exist *)
    fun find (EMPTY, _) = NONE
      | find (MAP{root,nobj}, key) = (case splay (cmpf key, !root)
	   of (EQUAL, r as SplayObj{value, ...}) => (root := r; SOME(#2 value))
	    | (_, r) => (root := r; NONE))

	(* Remove an item.
         * Raise LibBase.NotFound if not found
	 *)
    fun remove (EMPTY, _) = raise LibBase.NotFound
      | remove (MAP{root,nobj}, key) = (case (splay (cmpf key, !root))
	 of (EQUAL, SplayObj{value, left, right}) => 
	      if nobj = 1
		then (EMPTY, #2 value)
		else (MAP{root=ref(join(left, right)),nobj=nobj-1}, #2 value)
	    | (_, r) => (root := r; raise LibBase.NotFound)
	  (* end case *))

	(* Return the number of items in the table *)
    fun numItems EMPTY = 0
      | numItems (MAP{nobj, ...}) = nobj

	(* Return a list of the items (and their keys) in the dictionary *)
    fun listItems EMPTY = []
      | listItems (MAP{root, ...}) = let
	  fun apply (SplayNil, l) = l
            | apply (SplayObj{value=(_, v), left, right}, l) =
                apply(left, v :: (apply (right, l)))
        in
          apply (!root, [])
        end
    fun listItemsi EMPTY = []
      | listItemsi (MAP{root, ...}) = let
	  fun apply (SplayNil, l) = l
            | apply (SplayObj{value, left, right}, l) =
                apply(left, value :: (apply (right, l)))
        in
          apply (!root, [])
        end

    local
      fun next ((t as SplayObj{right, ...}) :: rest) = (t, left(right, rest))
	| next _ = (SplayNil, [])
      and left (SplayNil, rest) = rest
	| left (t as SplayObj{left=l, ...}, rest) = left(l, t :: rest)
    in
    fun collate cmpRng (EMPTY, EMPTY) = EQUAL
      | collate cmpRng (EMPTY, _) = LESS
      | collate cmpRng (_, EMPTY) = GREATER
      | collate cmpRng (MAP{root=s1, ...}, MAP{root=s2, ...}) = let
	  fun cmp (t1, t2) = (case (next t1, next t2)
		 of ((SplayNil, _), (SplayNil, _)) => EQUAL
		  | ((SplayNil, _), _) => LESS
		  | (_, (SplayNil, _)) => GREATER
		  | ((SplayObj{value=(x1, y1), ...}, r1),
		     (SplayObj{value=(x2, y2), ...}, r2)
		    ) => (
		      case Key.compare(x1, x2)
		       of EQUAL => (case cmpRng (y1, y2)
			     of EQUAL => cmp (r1, r2)
			      | order => order
			    (* end case *))
			| order => order
		      (* end case *))
		(* end case *))
	  in
	    cmp (left(!s1, []), left(!s2, []))
	  end
    end (* local *)

	(* Apply a function to the entries of the dictionary *)
    fun appi af EMPTY = ()
      | appi af (MAP{root, ...}) =
          let fun apply SplayNil = ()
                | apply (SplayObj{value, left, right}) = 
                    (apply left; af value; apply right)
        in
          apply (!root)
        end

    fun app af EMPTY = ()
      | app af (MAP{root, ...}) =
          let fun apply SplayNil = ()
                | apply (SplayObj{value=(_, value), left, right}) = 
                    (apply left; af value; apply right)
        in
          apply (!root)
        end
(*
    fun revapp af (MAP{root, ...}) =
          let fun apply SplayNil = ()
                | apply (SplayObj{value, left, right}) = 
                    (apply right; af value; apply left)
        in
          apply (!root)
        end
*)

	(* Fold function *)
    fun foldri (abf: K.ord_key * 'a * 'b -> 'b) b EMPTY = b
      | foldri (abf: K.ord_key * 'a * 'b -> 'b) b (MAP{root, ...}) =
          let fun apply (SplayNil: (K.ord_key * 'a) splay, b) = b
                | apply (SplayObj{value, left, right}, b) =
                    apply(left, abf(#1 value, #2 value, apply(right, b)))
        in
          apply (!root, b)
        end

    fun foldr (abf: 'a * 'b -> 'b) b EMPTY = b
      | foldr (abf: 'a * 'b -> 'b) b (MAP{root, ...}) =
          let fun apply (SplayNil: (K.ord_key * 'a) splay, b) = b
                | apply (SplayObj{value=(_, value), left, right}, b) =
                    apply(left, abf(value, apply(right, b)))
        in
          apply (!root, b)
        end

    fun foldli (abf: K.ord_key * 'a * 'b -> 'b) b EMPTY = b
      | foldli (abf: K.ord_key * 'a * 'b -> 'b) b (MAP{root, ...}) =
          let fun apply (SplayNil: (K.ord_key * 'a) splay, b) = b
                | apply (SplayObj{value, left, right}, b) =
                    apply(right, abf(#1 value, #2 value, apply(left, b)))
        in
          apply (!root, b)
        end

    fun foldl (abf: 'a * 'b -> 'b) b EMPTY = b
      | foldl (abf: 'a * 'b -> 'b) b (MAP{root, ...}) =
          let fun apply (SplayNil: (K.ord_key * 'a) splay, b) = b
                | apply (SplayObj{value=(_, value), left, right}, b) =
                    apply(right, abf(value, apply(left, b)))
        in
          apply (!root, b)
        end

	(* Map a table to a new table that has the same keys*)
    fun mapi (af: K.ord_key * 'a -> 'b) EMPTY = EMPTY
      | mapi (af: K.ord_key * 'a -> 'b) (MAP{root,nobj}) =
          let fun ap (SplayNil: (K.ord_key * 'a) splay) = SplayNil
                | ap (SplayObj{value, left, right}) = let
                    val left' = ap left
                    val value' = (#1 value, af value)
                    in
                      SplayObj{value = value', left = left', right = ap right}
                    end
        in
          MAP{root = ref(ap (!root)), nobj = nobj}
        end

    fun map (af: 'a -> 'b) EMPTY = EMPTY
      | map (af: 'a -> 'b) (MAP{root,nobj}) =
          let fun ap (SplayNil: (K.ord_key * 'a) splay) = SplayNil
                | ap (SplayObj{value, left, right}) = let
                    val left' = ap left
                    val value' = (#1 value, af (#2 value))
                    in
                      SplayObj{value = value', left = left', right = ap right}
                    end
        in
          MAP{root = ref(ap (!root)), nobj = nobj}
        end

(* the following are generic implementations of the unionWith and intersectWith
 * operetions.  These should be specialized for the internal representations
 * at some point.
 *)
    fun unionWith f (m1, m2) = let
	  fun ins f (key, x, m) = (case find(m, key)
		 of NONE => insert(m, key, x)
		  | (SOME x') => insert(m, key, f(x, x'))
		(* end case *))
	  in
	    if (numItems m1 > numItems m2)
	      then foldli (ins (fn (a, b) => f(b, a))) m1 m2
	      else foldli (ins f) m2 m1
	  end
    fun unionWithi f (m1, m2) = let
	  fun ins f (key, x, m) = (case find(m, key)
		 of NONE => insert(m, key, x)
		  | (SOME x') => insert(m, key, f(key, x, x'))
		(* end case *))
	  in
	    if (numItems m1 > numItems m2)
	      then foldli (ins (fn (k, a, b) => f(k, b, a))) m1 m2
	      else foldli (ins f) m2 m1
	  end

    fun intersectWith f (m1, m2) = let
	(* iterate over the elements of m1, checking for membership in m2 *)
	  fun intersect f (m1, m2) = let
		fun ins (key, x, m) = (case find(m2, key)
		       of NONE => m
			| (SOME x') => insert(m, key, f(x, x'))
		      (* end case *))
		in
		  foldli ins empty m1
		end
	  in
	    if (numItems m1 > numItems m2)
	      then intersect f (m1, m2)
	      else intersect (fn (a, b) => f(b, a)) (m2, m1)
	  end

    fun intersectWithi f (m1, m2) = let
	(* iterate over the elements of m1, checking for membership in m2 *)
	  fun intersect f (m1, m2) = let
		fun ins (key, x, m) = (case find(m2, key)
		       of NONE => m
			| (SOME x') => insert(m, key, f(key, x, x'))
		      (* end case *))
		in
		  foldli ins empty m1
		end
	  in
	    if (numItems m1 > numItems m2)
	      then intersect f (m1, m2)
	      else intersect (fn (k, a, b) => f(k, b, a)) (m2, m1)
	  end

  (* this is a generic implementation of mapPartial.  It should
   * be specialized to the data-structure at some point.
   *)
    fun mapPartial f m = let
	  fun g (key, item, m) = (case f item
		 of NONE => m
		  | (SOME item') => insert(m, key, item')
		(* end case *))
	  in
	    foldli g empty m
	  end
    fun mapPartiali f m = let
	  fun g (key, item, m) = (case f(key, item)
		 of NONE => m
		  | (SOME item') => insert(m, key, item')
		(* end case *))
	  in
	    foldli g empty m
	  end

  (* this is a generic implementation of filter.  It should
   * be specialized to the data-structure at some point.
   *)
    fun filter predFn m = let
	  fun f (key, item, m) = if predFn item
		then insert(m, key, item)
		else m
	  in
	    foldli f empty m
	  end
    fun filteri predFn m = let
	  fun f (key, item, m) = if predFn(key, item)
		then insert(m, key, item)
		else m
	  in
	    foldli f empty m
	  end

  end (* SplayDictFn *)
(* ML-Yacc Parser Generator (c) 1989 Andrew W. Appel, David R. Tarditi 
 *
 * $Log: base.sig, v $
 * Revision 1.1.1.1  1997/01/14 01:38:04  george
 *   Version 109.24
 *
 * Revision 1.1.1.1  1996/01/31  16:01:42  george
 * Version 109
 * 
 *)

(* base.sig: Base signature file for SML-Yacc.  This file contains signatures
   that must be loaded before any of the files produced by ML-Yacc are loaded
*)

(* STREAM: signature for a lazy stream.*)

signature STREAM =
 sig type 'xa stream
     val streamify: (unit -> 'a) -> 'a stream
     val cons: 'a * 'a stream -> 'a stream
     val get: 'a stream -> 'a * 'a stream
 end

(* LR_TABLE: signature for an LR Table.

   The list of actions and gotos passed to mkLrTable must be ordered by state
   number. The values for state 0 are the first in the list, the values for
    state 1 are next, etc.
*)

signature LR_TABLE =
    sig
        datatype ('a, 'b) pairlist = EMPTY | PAIR of 'a * 'b * ('a, 'b) pairlist
	datatype state = STATE of int
	datatype term = T of int
	datatype nonterm = NT of int
	datatype action = SHIFT of state
			| REDUCE of int
			| ACCEPT
			| ERROR
	type table
	
	val numStates: table -> int
	val numRules: table -> int
	val describeActions: table -> state ->
				(term, action) pairlist * action
	val describeGoto: table -> state -> (nonterm, state) pairlist
	val action: table -> state * term -> action
	val goto: table -> state * nonterm -> state
	val initialState: table -> state
	exception Goto of state * nonterm

	val mkLrTable: {actions: ((term, action) pairlist * action) array,
			 gotos: (nonterm, state) pairlist array,
			 numStates: int, numRules: int,
			 initialState: state} -> table
    end

(* TOKEN: signature revealing the internal structure of a token. This signature
   TOKEN distinct from the signature {parser name}_TOKENS produced by ML-Yacc.
   The {parser name}_TOKENS structures contain some types and functions to
    construct tokens from values and positions.

   The representation of token was very carefully chosen here to allow the
   polymorphic parser to work without knowing the types of semantic values
   or line numbers.

   This has had an impact on the TOKENS structure produced by SML-Yacc, which
   is a structure parameter to lexer functors.  We would like to have some
   type 'a token which functions to construct tokens would create.  A
   constructor function for a integer token might be

	  INT: int * 'a * 'a -> 'a token.
 
   This is not possible because we need to have tokens with the representation
   given below for the polymorphic parser.

   Thus our constructur functions for tokens have the form:

	  INT: int * 'a * 'a -> (svalue, 'a) token

   This in turn has had an impact on the signature that lexers for SML-Yacc
   must match and the types that a user must declare in the user declarations
   section of lexers.
*)

signature TOKEN =
    sig
	structure LrTable: LR_TABLE
        datatype ('a, 'b) token = TOKEN of LrTable.term * ('a * 'b * 'b)
	val sameToken: ('a, 'b) token * ('a, 'b) token -> bool
    end

(* LR_PARSER: signature for a polymorphic LR parser *)

signature LR_PARSER =
    sig
	structure Stream: STREAM
	structure LrTable: LR_TABLE
	structure Token: TOKEN

	sharing LrTable = Token.LrTable

	exception ParseError

	val parse: {table: LrTable.table,
		     lexer: ('b, 'c) Token.token Stream.stream,
		     arg: 'arg,
		     saction: int *
			       'c *
				(LrTable.state * ('b * 'c * 'c)) list * 
				'arg ->
				     LrTable.nonterm *
				     ('b * 'c * 'c) *
				     ((LrTable.state *('b * 'c * 'c)) list),
		     void: 'b,
		     ec: { is_keyword: LrTable.term -> bool,
			    noShift: LrTable.term -> bool,
			    preferred_change: (LrTable.term list * LrTable.term list) list,
			    errtermvalue: LrTable.term -> 'b,
			    showTerminal: LrTable.term -> string,
			    terms: LrTable.term list,
			    error: string * 'c * 'c -> unit
			   },
		     lookahead: int  (* max amount of lookahead used in *)
				      (* error correction *)
			} -> 'b *
			     (('b, 'c) Token.token Stream.stream)
    end

(* LEXER: a signature that most lexers produced for use with SML-Yacc's
   output will match.  The user is responsible for declaring type token,
   type pos, and type svalue in the UserDeclarations section of a lexer.

   Note that type token is abstract in the lexer.  This allows SML-Yacc to
   create a TOKENS signature for use with lexers produced by ML-Lex that
   treats the type token abstractly.  Lexers that are functors parametrized by
   a Tokens structure matching a TOKENS signature cannot examine the structure
   of tokens.
*)

signature LEXER =
   sig
       structure UserDeclarations :
	   sig
	        type ('a, 'b) token
		type pos
		type svalue
	   end
	val makeLexer: (int -> string) -> unit -> 
         (UserDeclarations.svalue, UserDeclarations.pos) UserDeclarations.token
   end

(* ARG_LEXER: the %arg option of ML-Lex allows users to produce lexers which
   also take an argument before yielding a function from unit to a token
*)

signature ARG_LEXER =
   sig
       structure UserDeclarations :
	   sig
	        type ('a, 'b) token
		type pos
		type svalue
		type arg
	   end
	val makeLexer: (int -> string) -> UserDeclarations.arg -> unit -> 
         (UserDeclarations.svalue, UserDeclarations.pos) UserDeclarations.token
   end

(* PARSER_DATA: the signature of ParserData structures in {parser name}LrValsFun
   produced by  SML-Yacc.  All such structures match this signature.  

   The {parser name}LrValsFun produces a structure which contains all the values
   except for the lexer needed to call the polymorphic parser mentioned
   before.

*)

signature PARSER_DATA =
   sig
        (* the type of line numbers *)

	type pos

	(* the type of semantic values *)

	type svalue

         (* the type of the user-supplied argument to the parser *)
 	type arg
 
	(* the intended type of the result of the parser.  This value is
	   produced by applying extract from the structure Actions to the
	   final semantic value resultiing from a parse.
	 *)

	type result

	structure LrTable: LR_TABLE
	structure Token: TOKEN
	sharing Token.LrTable = LrTable

	(* structure Actions contains the functions which mantain the
	   semantic values stack in the parser.  Void is used to provide
	   a default value for the semantic stack.
	 *)

	structure Actions: 
	  sig
	      val actions: int * pos *
		   (LrTable.state * (svalue * pos * pos)) list * arg->
		         LrTable.nonterm * (svalue * pos * pos) *
			 ((LrTable.state *(svalue * pos * pos)) list)
	      val void: svalue
	      val extract: svalue -> result
	  end

	(* structure EC contains information used to improve error
	   recovery in an error-correcting parser *)

	structure EC :
	   sig
	     val is_keyword: LrTable.term -> bool
	     val noShift: LrTable.term -> bool
 	     val preferred_change: (LrTable.term list * LrTable.term list) list
	     val errtermvalue: LrTable.term -> svalue
	     val showTerminal: LrTable.term -> string
	     val terms: LrTable.term list
	   end

	(* table is the LR table for the parser *)

	val table: LrTable.table
    end

(* signature PARSER is the signature that most user parsers created by 
   SML-Yacc will match.
*)

signature PARSER =
    sig
        structure Token: TOKEN
	structure Stream: STREAM
	exception ParseError

	(* type pos is the type of line numbers *)

	type pos

	(* type result is the type of the result from the parser *)

	type result

         (* the type of the user-supplied argument to the parser *)
 	type arg
	
	(* type svalue is the type of semantic values for the semantic value
	   stack
	 *)

	type svalue

	(* val makeLexer is used to create a stream of tokens for the parser *)

	val makeLexer: (int -> string) ->
			 (svalue, pos) Token.token Stream.stream

	(* val parse takes a stream of tokens and a function to print
	   errors and returns a value of type result and a stream containing
	   the unused tokens
	 *)

	val parse: int * ((svalue, pos) Token.token Stream.stream) *
		    (string * pos * pos -> unit) * arg ->
				result * (svalue, pos) Token.token Stream.stream

	val sameToken: (svalue, pos) Token.token * (svalue, pos) Token.token ->
				bool
     end

(* signature ARG_PARSER is the signature that will be matched by parsers whose
    lexer takes an additional argument.
*)

signature ARG_PARSER = 
    sig
        structure Token: TOKEN
	structure Stream: STREAM
	exception ParseError

	type arg
	type lexarg
	type pos
	type result
	type svalue

	val makeLexer: (int -> string) -> lexarg ->
			 (svalue, pos) Token.token Stream.stream
	val parse: int * ((svalue, pos) Token.token Stream.stream) *
		    (string * pos * pos -> unit) * arg ->
				result * (svalue, pos) Token.token Stream.stream

	val sameToken: (svalue, pos) Token.token * (svalue, pos) Token.token ->
				bool
     end

(* ML-Yacc Parser Generator (c) 1989 Andrew W. Appel, David R. Tarditi 
 *
 * $Log: lrtable.sml, v $
 * Revision 1.1.1.1  1997/01/14 01:38:04  george
 *   Version 109.24
 *
 * Revision 1.1.1.1  1996/01/31  16:01:42  george
 * Version 109
 * 
 *)

structure LrTable: LR_TABLE = 
    struct
	open Array List
	infix 9 sub
	datatype ('a, 'b) pairlist = EMPTY
				  | PAIR of 'a * 'b * ('a, 'b) pairlist
	datatype term = T of int
	datatype nonterm = NT of int
	datatype state = STATE of int
	datatype action = SHIFT of state
			| REDUCE of int (* rulenum from grammar *)
			| ACCEPT
			| ERROR
	exception Goto of state * nonterm
	type table = {states: int, rules: int, initialState: state,
		      action: ((term, action) pairlist * action) array,
		      goto:  (nonterm, state) pairlist array}
	val numStates = fn ({states, ...}: table) => states
	val numRules = fn ({rules, ...}: table) => rules
	val describeActions =
	   fn ({action, ...}: table) => 
	           fn (STATE s) => action sub s
	val describeGoto =
	   fn ({goto, ...}: table) =>
	           fn (STATE s) => goto sub s
	fun findTerm (T term, row, default) =
	    let fun find (PAIR (T key, data, r)) =
		       if key < term then find r
		       else if key=term then data
		       else default
		   | find EMPTY = default
	    in find row
	    end
	fun findNonterm (NT nt, row) =
	    let fun find (PAIR (NT key, data, r)) =
		       if key < nt then find r
		       else if key=nt then SOME data
		       else NONE
		   | find EMPTY = NONE
	    in find row
	    end
	val action = fn ({action, ...}: table) =>
		fn (STATE state, term) =>
		  let val (row, default) = action sub state
		  in findTerm(term, row, default)
		  end
	val goto = fn ({goto, ...}: table) =>
			fn (a as (STATE state,nonterm)) =>
			  case findNonterm(nonterm, goto sub state)
			  of SOME state => state
			   | NONE => raise (Goto a)
	val initialState = fn ({initialState, ...}: table) => initialState
	val mkLrTable = fn {actions, gotos, initialState,numStates,numRules} =>
	     ({action=actions, goto=gotos,
	       states=numStates,
	       rules=numRules,
               initialState=initialState}: table)
end;
(* ML-Yacc Parser Generator (c) 1989 Andrew W. Appel, David R. Tarditi 
 *
 * $Log: stream.sml, v $
 * Revision 1.2  1997/08/26 19:18:55  jhr
 *   Replaced used of "abstraction" with ":>".
 *
# Revision 1.1.1.1  1997/01/14  01:38:04  george
#   Version 109.24
#
 * Revision 1.1.1.1  1996/01/31  16:01:43  george
 * Version 109
 * 
 *)

(* Stream: a structure implementing a lazy stream.  The signature STREAM
   is found in base.sig *)

structure Stream :> STREAM =
struct
   datatype 'a str = EVAL of 'a * 'a str ref | UNEVAL of (unit->'a)

   type 'a stream = 'a str ref

   fun get(ref(EVAL t)) = t
     | get(s as ref(UNEVAL f)) = 
	    let val t = (f(), ref(UNEVAL f)) in s := EVAL t; t end

   fun streamify f = ref(UNEVAL f)
   fun cons(a, s) = ref(EVAL(a, s))

end;
(* ML-Yacc Parser Generator (c) 1989 Andrew W. Appel, David R. Tarditi 
 *
 * $Log: parser2.sml, v $
 * Revision 1.2  1997/08/26 19:18:54  jhr
 *   Replaced used of "abstraction" with ":>".
 *
# Revision 1.1.1.1  1997/01/14  01:38:04  george
#   Version 109.24
#
 * Revision 1.3  1996/10/03  03:36:58  jhr
 * Qualified identifiers that are no-longer top-level (quot, rem, min, max).
 *
 * Revision 1.2  1996/02/26  15:02:29  george
 *    print no longer overloaded.
 *    use of makestring has been removed and replaced with Int.toString ..
 *    use of IO replaced with TextIO
 *
 * Revision 1.1.1.1  1996/01/31  16:01:42  george
 * Version 109
 * 
 *)

(* parser.sml:  This is a parser driver for LR tables with an error-recovery
   routine added to it.  The routine used is described in detail in this
   article:

	'A Practical Method for LR and LL Syntactic Error Diagnosis and
	 Recovery', by M. Burke and G. Fisher, ACM Transactions on
	 Programming Langauges and Systems, Vol. 9, No. 2, April 1987,
	 pp. 164-197.

    This program is an implementation is the partial, deferred method discussed
    in the article.  The algorithm and data structures used in the program
    are described below.  

    This program assumes that all semantic actions are delayed.  A semantic
    action should produce a function from unit -> value instead of producing the
    normal value.  The parser returns the semantic value on the top of the
    stack when accept is encountered.  The user can deconstruct this value
    and apply the unit -> value function in it to get the answer.

    It also assumes that the lexer is a lazy stream.

    Data Structures:
    ----------------
	
	* The parser:

	   The state stack has the type

		 (state * (semantic value * line # * line #)) list

	   The parser keeps a queue of (state stack * lexer pair).  A lexer pair
	 consists of a terminal * value pair and a lexer.  This allows the 
	 parser to reconstruct the states for terminals to the left of a
	 syntax error, and attempt to make error corrections there.

	   The queue consists of a pair of lists (x, y).  New additions to
	 the queue are cons'ed onto y.  The first element of x is the top
	 of the queue.  If x is nil, then y is reversed and used
	 in place of x.

    Algorithm:
    ----------

	* The steady-state parser:  

	    This parser keeps the length of the queue of state stacks at
	a steady state by always removing an element from the front when
	another element is placed on the end.

	    It has these arguments:

	   stack: current stack
	   queue: value of the queue
	   lexPair ((terminal, value), lex stream)

	When SHIFT is encountered, the state to shift to and the value are
	are pushed onto the state stack.  The state stack and lexPair are
	placed on the queue.  The front element of the queue is removed.

	When REDUCTION is encountered, the rule is applied to the current
	stack to yield a triple (nonterm, value,new stack).  A new
	stack is formed by adding (goto(top state of stack,nonterm), value)
	to the stack.

	When ACCEPT is encountered, the top value from the stack and the
	lexer are returned.

	When an ERROR is encountered, fixError is called.  FixError
	takes the arguments to the parser, fixes the error if possible and
        returns a new set of arguments.

	* The distance-parser:

	This parser includes an additional argument distance.  It pushes
	elements on the queue until it has parsed distance tokens, or an
	ACCEPT or ERROR occurs.  It returns a stack, lexer, the number of
	tokens left unparsed, a queue, and an action option.
*)

signature FIFO = 
  sig type 'a queue
      val empty: 'a queue
      exception Empty
      val get: 'a queue -> 'a * 'a queue
      val put: 'a * 'a queue -> 'a queue
  end

(* drt (12/15/89) -- the functor should be used in development work, but
   it wastes space in the release version.

functor ParserGen(structure LrTable: LR_TABLE
		  structure Stream: STREAM): LR_PARSER =
*)

structure LrParser :> LR_PARSER =
   struct
      structure LrTable = LrTable
      structure Stream = Stream

      structure Token: TOKEN =
	struct
	    structure LrTable = LrTable
	    datatype ('a, 'b) token = TOKEN of LrTable.term * ('a * 'b * 'b)
	    val sameToken = fn (TOKEN(t, _), TOKEN(t', _)) => t=t'
        end

      open LrTable
      open Token

      val DEBUG1 = false
      val DEBUG2 = false
      exception ParseError
      exception ParseImpossible of int

      structure Fifo :> FIFO =
        struct
	  type 'a queue = ('a list * 'a list)
	  val empty = (nil,nil)
	  exception Empty
	  fun get(a :: x, y) = (a, (x, y))
	    | get(nil, nil) = raise Empty
	    | get(nil, y) = get(rev y, nil)
 	  fun put(a, (x, y)) = (x, a :: y)
        end

      type ('a, 'b) elem = (state * ('a * 'b * 'b))
      type ('a, 'b) stack = ('a, 'b) elem list
      type ('a, 'b) lexv = ('a, 'b) token
      type ('a, 'b) lexpair = ('a, 'b) lexv * (('a, 'b) lexv Stream.stream)
      type ('a, 'b) distanceParse =
		 ('a, 'b) lexpair *
		 ('a, 'b) stack * 
		 (('a, 'b) stack * ('a, 'b) lexpair) Fifo.queue *
		 int ->
		   ('a, 'b) lexpair *
		   ('a, 'b) stack * 
		   (('a, 'b) stack * ('a, 'b) lexpair) Fifo.queue *
		   int *
		   action option

      type ('a, 'b) ecRecord =
	 {is_keyword: term -> bool,
          preferred_change: (term list * term list) list,
	  error: string * 'b * 'b -> unit,
	  errtermvalue: term -> 'a,
	  terms: term list,
	  showTerminal: term -> string,
	  noShift: term -> bool}

      local 
	 val print = fn s => TextIO.output(TextIO.stdOut, s)
	 val println = fn s => (print s; print "\n")
	 val showState = fn (STATE s) => "STATE " ^ (Int.toString s)
      in
        fun printStack(stack: ('a, 'b) stack, n: int) =
         case stack
           of (state, _) :: rest =>
                 (print("\t" ^ Int.toString n ^ ": ");
                  println(showState state);
                  printStack(rest, n+1))
            | nil => ()
                
        fun prAction showTerminal
		 (stack as (state, _) :: _, next as (TOKEN (term, _), _), action) =
             (println "Parse: state stack:";
              printStack(stack, 0);
              print("       state="
                         ^ showState state	
                         ^ " next="
                         ^ showTerminal term
                         ^ " action="
                        );
              case action
                of SHIFT state => println ("SHIFT " ^ (showState state))
                 | REDUCE i => println ("REDUCE " ^ (Int.toString i))
                 | ERROR => println "ERROR"
		 | ACCEPT => println "ACCEPT")
        | prAction _ (_, _, action) = ()
     end

    (* ssParse: parser which maintains the queue of (state * lexvalues) in a
	steady-state.  It takes a table, showTerminal function, saction
	function, and fixError function.  It parses until an ACCEPT is
	encountered, or an exception is raised.  When an error is encountered,
	fixError is called with the arguments of parseStep (lexv, stack, and
	queue).  It returns the lexv, and a new stack and queue adjusted so
	that the lexv can be parsed *)
	
    val ssParse =
      fn (table, showTerminal, saction, fixError, arg) =>
	let val prAction = prAction showTerminal
	    val action = LrTable.action table
	    val goto = LrTable.goto table
	    fun parseStep(args as
			 (lexPair as (TOKEN (terminal, value as (_, leftPos, _)),
				      lexer
				      ),
			  stack as (state, _) :: _,
			  queue)) =
	      let val nextAction = action (state, terminal)
	          val _ = if DEBUG1 then prAction(stack, lexPair,nextAction)
			  else ()
	      in case nextAction
		 of SHIFT s =>
		  let val newStack = (s, value) :: stack
		      val newLexPair = Stream.get lexer
		      val (_,newQueue) =Fifo.get(Fifo.put((newStack,newLexPair),
							    queue))
		  in parseStep(newLexPair, (s, value) :: stack,newQueue)
		  end
		 | REDUCE i =>
		     (case saction(i, leftPos, stack, arg)
		      of (nonterm, value, stack as (state, _) :: _) =>
		          parseStep(lexPair, (goto(state,nonterm), value) :: stack,
				    queue)
		       | _ => raise (ParseImpossible 197))
		 | ERROR => parseStep(fixError args)
		 | ACCEPT => 
			(case stack
			 of (_, (topvalue, _, _)) :: _ =>
				let val (token, restLexer) = lexPair
				in (topvalue, Stream.cons(token, restLexer))
				end
			  | _ => raise (ParseImpossible 202))
	      end
	    | parseStep _ = raise (ParseImpossible 204)
	in parseStep
	end

    (*  distanceParse: parse until n tokens are shifted, or accept or
	error are encountered.  Takes a table, showTerminal function, and
	semantic action function.  Returns a parser which takes a lexPair
	(lex result * lexer), a state stack, a queue, and a distance
	(must be > 0) to parse.  The parser returns a new lex-value, a stack
	with the nth token shifted on top, a queue, a distance, and action
	option. *)

    val distanceParse =
      fn (table, showTerminal, saction, arg) =>
	let val prAction = prAction showTerminal
	    val action = LrTable.action table
	    val goto = LrTable.goto table
	    fun parseStep(lexPair, stack, queue, 0) = (lexPair, stack, queue, 0, NONE)
	      | parseStep(lexPair as (TOKEN (terminal, value as (_, leftPos, _)),
				      lexer
				     ),
			  stack as (state, _) :: _,
			  queue, distance) =
	      let val nextAction = action(state, terminal)
	          val _ = if DEBUG1 then prAction(stack, lexPair,nextAction)
			  else ()
	      in case nextAction
		 of SHIFT s =>
		  let val newStack = (s, value) :: stack
		      val newLexPair = Stream.get lexer
		  in parseStep(newLexPair, (s, value) :: stack,
			       Fifo.put((newStack,newLexPair), queue), distance-1)
		  end
		 | REDUCE i =>
		    (case saction(i, leftPos, stack, arg)
		      of (nonterm, value, stack as (state, _) :: _) =>
		         parseStep(lexPair, (goto(state,nonterm), value) :: stack,
				 queue, distance)
		      | _ => raise (ParseImpossible 240))
		 | ERROR => (lexPair, stack, queue, distance, SOME nextAction)
		 | ACCEPT => (lexPair, stack, queue, distance, SOME nextAction)
	      end
	   | parseStep _ = raise (ParseImpossible 242)
	in parseStep: ('a, 'b) distanceParse 
	end

(* mkFixError: function to create fixError function which adjusts parser state
   so that parse may continue in the presence of an error *)

fun mkFixError({is_keyword, terms, errtermvalue,
	      preferred_change,noShift,
	      showTerminal, error, ...}: ('a, 'b) ecRecord,
	     distanceParse: ('a, 'b) distanceParse,
	     minAdvance, maxAdvance) 

            (lexv as (TOKEN (term, value as (_, leftPos, _)), _), stack, queue) =
    let val _ = if DEBUG2 then
			error("syntax error found at " ^ (showTerminal term),
			      leftPos, leftPos)
		else ()

        fun tokAt(t, p) = TOKEN(t, (errtermvalue t, p, p))

	val minDelta = 3

	(* pull all the state * lexv elements from the queue *)

	val stateList = 
	   let fun f q = let val (elem,newQueue) = Fifo.get q
			 in elem :: (f newQueue)
			 end handle Fifo.Empty => nil
	   in f queue
	   end

	(* now number elements of stateList, giving distance from
	   error token *)

	val (_, numStateList) =
	      List.foldr (fn (a, (num, r)) => (num+1, (a,num) :: r)) (0, []) stateList

	(* Represent the set of potential changes as a linked list.

	   Values of datatype Change hold information about a potential change.

	   oper = oper to be applied
	   pos = the # of the element in stateList that would be altered.
	   distance = the number of tokens beyond the error token which the
	     change allows us to parse.
	   new = new terminal * value pair at that point
	   orig = original terminal * value pair at the point being changed.
	 *)

	datatype ('a, 'b) change = CHANGE of
	   {pos: int, distance: int, leftPos: 'b, rightPos: 'b,
	    new: ('a, 'b) lexv list, orig: ('a, 'b) lexv list}


         val showTerms = concat o map (fn TOKEN(t, _) => " " ^ showTerminal t)

	 val printChange = fn c =>
	  let val CHANGE {distance,new, orig, pos, ...} = c
	  in (print ("{distance= " ^ (Int.toString distance));
	      print (", orig ="); print(showTerms orig);
	      print (",new ="); print(showTerms new);
	      print (", pos= " ^ (Int.toString pos));
	      print "}\n")
	  end

	val printChangeList = app printChange

(* parse: given a lexPair, a stack, and the distance from the error
   token, return the distance past the error token that we are able to parse.*)

	fun parse (lexPair, stack, queuePos: int) =
	    case distanceParse(lexPair, stack, Fifo.empty, queuePos+maxAdvance+1)
             of (_, _, _, distance, SOME ACCEPT) => 
		        if maxAdvance-distance-1 >= 0 
			    then maxAdvance 
			    else maxAdvance-distance-1
	      | (_, _, _, distance, _) => maxAdvance - distance - 1

(* catList: concatenate results of scanning list *)

	fun catList l f = List.foldr (fn(a, r)=> f a @ r) [] l

        fun keywordsDelta new = if List.exists (fn(TOKEN(t, _))=>is_keyword t) new
	               then minDelta else 0

        fun tryChange{lex, stack, pos, leftPos, rightPos, orig,new} =
	     let val lex' = List.foldr (fn (t', p)=>(t', Stream.cons p)) lex new
		 val distance = parse(lex', stack, pos+length new-length orig)
	      in if distance >= minAdvance + keywordsDelta new 
		   then [CHANGE{pos=pos, leftPos=leftPos, rightPos=rightPos,
				distance=distance, orig=orig,new=new}] 
		   else []
	     end


(* tryDelete: Try to delete n terminals.
              Return single-element [success] or nil.
	      Do not delete unshiftable terminals. *)


    fun tryDelete n ((stack, lexPair as (TOKEN(term, (_, l, r)), _)), qPos) =
	let fun del(0, accum, left, right, lexPair) =
	          tryChange{lex=lexPair, stack=stack,
			    pos=qPos, leftPos=left, rightPos=right,
			    orig=rev accum, new=[]}
	      | del(n, accum, left, right, (tok as TOKEN(term, (_, _, r)), lexer)) =
		   if noShift term then []
		   else del(n-1, tok :: accum, left, r, Stream.get lexer)
         in del(n, [], l, r, lexPair)
        end

(* tryInsert: try to insert tokens before the current terminal;
       return a list of the successes  *)

        fun tryInsert((stack, lexPair as (TOKEN(_, (_, l, _)), _)), queuePos) =
	       catList terms (fn t =>
		 tryChange{lex=lexPair, stack=stack,
			   pos=queuePos, orig=[],new=[tokAt(t, l)],
			   leftPos=l, rightPos=l})
			      
(* trySubst: try to substitute tokens for the current terminal;
       return a list of the successes  *)

        fun trySubst ((stack, lexPair as (orig as TOKEN (term, (_, l, r)), lexer)),
		      queuePos) =
	      if noShift term then []
	      else
		  catList terms (fn t =>
		      tryChange{lex=Stream.get lexer, stack=stack,
				pos=queuePos,
				leftPos=l, rightPos=r, orig=[orig],
				new=[tokAt(t, r)]})

     (* do_delete(toks, lexPair) tries to delete tokens "toks" from "lexPair".
         If it succeeds, returns SOME(toks', l, r, lp), where
	     toks' is the actual tokens (with positions and values) deleted,
	     (l, r) are the (leftmost, rightmost) position of toks', 
	     lp is what remains of the stream after deletion 
     *)
        fun do_delete(nil, lp as (TOKEN(_, (_, l, _)), _)) = SOME(nil, l, l, lp)
          | do_delete([t], (tok as TOKEN(t', (_, l, r)), lp')) =
	       if t=t'
		   then SOME([tok], l, r, Stream.get lp')
                   else NONE
          | do_delete(t :: rest, (tok as TOKEN(t', (_, l, r)), lp')) =
	       if t=t'
		   then case do_delete(rest, Stream.get lp')
                         of SOME(deleted, l', r', lp'') =>
			       SOME(tok :: deleted, l, r', lp'')
			  | NONE => NONE
		   else NONE
			     
        fun tryPreferred((stack, lexPair), queuePos) =
	    catList preferred_change (fn (delete, insert) =>
	       if List.exists noShift delete then [] (* should give warning at
						 parser-generation time *)
               else case do_delete(delete, lexPair)
                     of SOME(deleted, l, r, lp) => 
			    tryChange{lex=lp, stack=stack, pos=queuePos,
				      leftPos=l, rightPos=r, orig=deleted,
				      new=map (fn t=>(tokAt(t, r))) insert}
		      | NONE => [])

	val changes = catList numStateList tryPreferred @
	                catList numStateList tryInsert @
			  catList numStateList trySubst @
			    catList numStateList (tryDelete 1) @
			      catList numStateList (tryDelete 2) @
			        catList numStateList (tryDelete 3)

	val findMaxDist = fn l => 
	  foldr (fn (CHANGE {distance, ...}, high) => Int.max(distance, high)) 0 l

(* maxDist: max distance past error taken that we could parse *)

	val maxDist = findMaxDist changes

(* remove changes which did not parse maxDist tokens past the error token *)

        val changes = catList changes 
	      (fn(c as CHANGE{distance, ...}) => 
		  if distance=maxDist then [c] else [])

      in case changes 
	  of (l as change :: _) =>
	      let fun print_msg (CHANGE {new, orig, leftPos, rightPos, ...}) =
		  let val s = 
		      case (orig,new)
			  of (_ :: _, []) => "deleting " ^ (showTerms orig)
	                   | ([], _ :: _) => "inserting " ^ (showTerms new)
			   | _ => "replacing " ^ (showTerms orig) ^
				 " with " ^ (showTerms new)
		  in error ("syntax error: " ^ s, leftPos, rightPos)
		  end
		   
		  val _ = 
		      (if length l > 1 andalso DEBUG2 then
			   (print "multiple fixes possible; could fix it by:\n";
			    app print_msg l;
			    print "chosen correction:\n")
		       else ();
		       print_msg change)

		  (* findNth: find nth queue entry from the error
		   entry.  Returns the Nth queue entry and the  portion of
		   the queue from the beginning to the nth-1 entry.  The
		   error entry is at the end of the queue.

		   Examples:

		   queue = a b c d e
		   findNth 0 = (e, a b c d)
		   findNth 1 =  (d, a b c)
		   *)

		  val findNth = fn n =>
		      let fun f (h :: t, 0) = (h, rev t)
			    | f (h :: t,n) = f(t,n-1)
			    | f (nil, _) = let exception FindNth
					  in raise FindNth
					  end
		      in f (rev stateList,n)
		      end
		
		  val CHANGE {pos, orig,new, ...} = change
		  val (last, queueFront) = findNth pos
		  val (stack, lexPair) = last

		  val lp1 = foldl(fn (_, (_, r)) => Stream.get r) lexPair orig
		  val lp2 = foldr(fn(t, r)=>(t, Stream.cons r)) lp1 new

		  val restQueue = 
		      Fifo.put((stack, lp2),
			       foldl Fifo.put Fifo.empty queueFront)

		  val (lexPair, stack, queue, _, _) =
		      distanceParse(lp2, stack, restQueue, pos)

	      in (lexPair, stack, queue)
	      end
	| nil => (error("syntax error found at " ^ (showTerminal term),
			leftPos, leftPos); raise ParseError)
    end

   val parse = fn {arg, table, lexer, saction, void, lookahead,
		   ec=ec as {showTerminal, ...}: ('a, 'b) ecRecord} =>
	let val distance = 15   (* defer distance tokens *)
	    val minAdvance = 1  (* must parse at least 1 token past error *)
	    val maxAdvance = Int.max(lookahead, 0)(* max distance for parse check *)
	    val lexPair = Stream.get lexer
	    val (TOKEN (_, (_, leftPos, _)), _) = lexPair
	    val startStack = [(initialState table, (void, leftPos, leftPos))]
	    val startQueue = Fifo.put((startStack, lexPair), Fifo.empty)
	    val distanceParse = distanceParse(table, showTerminal, saction, arg)
	    val fixError = mkFixError(ec, distanceParse, minAdvance, maxAdvance)
	    val ssParse = ssParse(table, showTerminal, saction, fixError, arg)
	    fun loop (lexPair, stack, queue, _, SOME ACCEPT) =
		   ssParse(lexPair, stack, queue)
	      | loop (lexPair, stack, queue, 0, _) = ssParse(lexPair, stack, queue)
	      | loop (lexPair, stack, queue, distance, SOME ERROR) =
		 let val (lexPair, stack, queue) = fixError(lexPair, stack, queue)
		 in loop (distanceParse(lexPair, stack, queue, distance))
		 end
	      | loop _ = let exception ParseInternal
			 in raise ParseInternal
			 end
	in loop (distanceParse(lexPair, startStack, startQueue, distance))
	end
 end;

(* drt (12/15/89) -- needed only when the code above is functorized

structure LrParser = ParserGen(structure LrTable=LrTable
			     structure Stream=Stream);
*)
(* ML-Yacc Parser Generator (c) 1989 Andrew W. Appel, David R. Tarditi 
 *
 * $Log: join.sml, v $
 * Revision 1.1.1.1  1997/01/14 01:38:04  george
 *   Version 109.24
 *
 * Revision 1.1.1.1  1996/01/31  16:01:42  george
 * Version 109
 * 
 *)

(* functor Join creates a user parser by putting together a Lexer structure,
   an LrValues structure, and a polymorphic parser structure.  Note that
   the Lexer and LrValues structure must share the type pos (i.e. the type
   of line numbers), the type svalues for semantic values, and the type
   of tokens.
*)

functor Join(structure Lex: LEXER
	     structure ParserData: PARSER_DATA
	     structure LrParser: LR_PARSER
	     sharing ParserData.LrTable = LrParser.LrTable
	     sharing ParserData.Token = LrParser.Token
	     sharing type Lex.UserDeclarations.svalue = ParserData.svalue
	     sharing type Lex.UserDeclarations.pos = ParserData.pos
	     sharing type Lex.UserDeclarations.token = ParserData.Token.token)
		: PARSER =
struct
    structure Token = ParserData.Token
    structure Stream = LrParser.Stream
 
    exception ParseError = LrParser.ParseError

    type arg = ParserData.arg
    type pos = ParserData.pos
    type result = ParserData.result
    type svalue = ParserData.svalue
    val makeLexer = LrParser.Stream.streamify o Lex.makeLexer
    val parse = fn (lookahead, lexer, error, arg) =>
	(fn (a, b) => (ParserData.Actions.extract a, b))
     (LrParser.parse {table = ParserData.table,
	        lexer=lexer,
		lookahead=lookahead,
		saction = ParserData.Actions.actions,
		arg=arg,
		void= ParserData.Actions.void,
	        ec = {is_keyword = ParserData.EC.is_keyword,
		      noShift = ParserData.EC.noShift,
		      preferred_change = ParserData.EC.preferred_change,
		      errtermvalue = ParserData.EC.errtermvalue,
		      error=error,
		      showTerminal = ParserData.EC.showTerminal,
		      terms = ParserData.EC.terms}}
      )
     val sameToken = Token.sameToken
end

(* functor JoinWithArg creates a variant of the parser structure produced 
   above.  In this case, the makeLexer take an additional argument before
   yielding a value of type unit -> (svalue, pos) token
 *)

functor JoinWithArg(structure Lex: ARG_LEXER
	     structure ParserData: PARSER_DATA
	     structure LrParser: LR_PARSER
	     sharing ParserData.LrTable = LrParser.LrTable
	     sharing ParserData.Token = LrParser.Token
	     sharing type Lex.UserDeclarations.svalue = ParserData.svalue
	     sharing type Lex.UserDeclarations.pos = ParserData.pos
	     sharing type Lex.UserDeclarations.token = ParserData.Token.token)
		: ARG_PARSER  =
struct
    structure Token = ParserData.Token
    structure Stream = LrParser.Stream

    exception ParseError = LrParser.ParseError

    type arg = ParserData.arg
    type lexarg = Lex.UserDeclarations.arg
    type pos = ParserData.pos
    type result = ParserData.result
    type svalue = ParserData.svalue

    val makeLexer = fn s => fn arg =>
		 LrParser.Stream.streamify (Lex.makeLexer s arg)
    val parse = fn (lookahead, lexer, error, arg) =>
	(fn (a, b) => (ParserData.Actions.extract a, b))
     (LrParser.parse {table = ParserData.table,
	        lexer=lexer,
		lookahead=lookahead,
		saction = ParserData.Actions.actions,
		arg=arg,
		void= ParserData.Actions.void,
	        ec = {is_keyword = ParserData.EC.is_keyword,
		      noShift = ParserData.EC.noShift,
		      preferred_change = ParserData.EC.preferred_change,
		      errtermvalue = ParserData.EC.errtermvalue,
		      error=error,
		      showTerminal = ParserData.EC.showTerminal,
		      terms = ParserData.EC.terms}}
      )
    val sameToken = Token.sameToken
end;
signature ERROR =
    sig
       val bug: string -> 'a
       val unimplemented: string -> 'a
       val error: string -> 'a
   end
structure Error: ERROR =
struct

fun print str = TextIO.output(TextIO.stdErr, str)

fun err(name, msg, exn) =
   (print name
    ; print ": "
    ; print msg
    ; print "\n"
    ; raise exn)
   
exception Unimplemented
fun unimplemented msg = err("Unimplemented", msg, Unimplemented)
    
exception Bug
fun bug msg = err("Bug", msg, Bug)

exception Error
fun error msg = err("Error", msg, Error)
	       
end

signature DYNAMIC_WIND =
   sig
      val dynamicWind: (unit -> 'a) * (unit -> unit) -> 'a
   end
structure DynamicWind: DYNAMIC_WIND =
struct

fun dynamicWind(thunk, cleanup) =
   let val a = thunk()
   in cleanup() ; a
   end handle exn => (cleanup() ; raise exn)

end

structure Pervasive =
   struct
      structure Array = Array
      structure Vector = Vector
      structure Bool = Bool
      structure Char = Char
      structure Date = Date
      structure General = General
      structure Int = Int
      structure Int32 = Int32
      structure IntInf = IntInf
      structure IO = TextIO
      structure List = List
      structure Option = Option
      structure Real = Real
      structure String = String
      structure SMLofNJ = SMLofNJ
      structure Time = Time
   end
signature OUTSTREAM =
   sig
      type t
	 
      val standard: t
      val error: t

      val openOut: string -> t
      val openAppend: string -> t
      val close: t -> unit
      val output: t * string -> unit
      val outputc: t -> string -> unit
      val flush: t -> unit
      val newline: t -> unit
      val fluidLet: t * t * (unit -> 'a) -> 'a

      (*val forward: {from: t, to: t} -> unit*)

      val print: string -> unit

      val ignore: 'a * t -> unit

      val output2: ('a * t -> unit) * ('b * t -> unit)
	 -> (('a * 'b) * t -> unit)

      val output3 :
	 ('a * t -> unit) * ('b * t -> unit) * ('c * t -> unit)
	 -> (('a * 'b * 'c) * t -> unit)
   end
(*-------------------------------------------------------------------*)
(*                             Outstream                             *)
(*-------------------------------------------------------------------*)

structure Outstream: OUTSTREAM =
struct

structure O = Pervasive.IO
open O
   
type t = outstream
   
val standard = stdOut
val error = stdErr

fun ignore _ = ()

val close = closeOut
(*val outputc = O.outputc*)
fun outputc stream string = output(stream, string)
val flush = flushOut

fun newline s = output(s, "\n")
   
fun print s = output(standard, s)
   
fun outputNothing _ = ()

fun fluidLet(s1, s2, thunk) =
   let val old = O.getOutstream s1
   in O.setOutstream(s1, O.getOutstream s2)
      ; DynamicWind.dynamicWind(thunk, fn () => O.setOutstream(s1, old))
   end

fun output2 (outX, outY) ((x, y), out) =
   (output(out, "(") ;
    outX(x, out) ;
    output(out, ", ") ;
    outY(y, out) ;
    output(out, ")"))

fun output3 (outX, outY, outZ) ((x, y, z), out) =
   (output(out, "(") ;
    outX(x, out) ;
    output(out, ", ") ;
    outY(y, out) ;
    output(out, ", ") ;
    outZ(z, out) ;
    output(out, ")"))

fun tripleOutput(f, g, h) ((a, b, c), out) =
   let val print = outputc out
   in print "(" ;
      f(a, out) ;
      print ", " ;
      g(b, out) ;
      print ", " ;
      h(c, out) ;
      print ")"
   end

end

structure Out = Outstream
signature RELATION =
   sig
      datatype t = datatype order (* from the pervasive environment *)
	 
      val lessEqual: {< : 'a * 'a -> bool,
		       equals: 'a * 'a -> bool}
	 -> {> : 'a * 'a -> bool,
	     >= : 'a * 'a -> bool,
	     <= : 'a * 'a -> bool,
	     min: 'a * 'a -> 'a,
	     max: 'a * 'a -> 'a,
	     compare: 'a * 'a -> t}

      val compare: ('a * 'a -> t)
	 -> {equals: 'a * 'a -> bool,
	     < : 'a * 'a -> bool,
	     > : 'a * 'a -> bool,
	     >= : 'a * 'a -> bool,
	     <= : 'a * 'a -> bool,
	     min: 'a * 'a -> 'a,
	     max: 'a * 'a -> 'a}
   end
structure Relation: RELATION =
struct

datatype t = datatype order
(*   Less
 | Equal
 | Greater*)

fun lessEqual{<, equals} = 
   let
      fun a > b = b < a
      fun a <= b = a < b orelse equals(a, b)
      fun a >= b = b < a orelse equals(b, a)
      fun compare(a, b) = if a < b then LESS
			 else if equals(a, b) then EQUAL
			      else GREATER
      fun min(x, y) = if x < y then x else y
      fun max(x, y) = if x < y then y else x
   in {> = op >, <= = op <=, >= = op >=,
       compare = compare, min = min, max = max}
   end

fun compare c =
   let fun equals(x, y) = (case c(x, y) of
			     EQUAL => true
			   | _ => false)
      fun x <  y = (case c(x, y) of
		       LESS => true
		     | _ => false)
      fun x <= y = (case c(x, y) of
			LESS => true
		      | EQUAL => true
		      | _ => false)
      fun x > y = (case c(x, y) of
		      GREATER => true
		    | _ => false)
      fun x >= y = (case c(x, y) of
		       GREATER => true
		     | EQUAL => true
		     | _ => false)
      fun max(x, y) = (case c(x, y) of
			 GREATER => x
		       | _ => y)
      fun min(x, y) = (case c(x, y) of
			 GREATER => y
		       | _ => x)
   in {equals = equals,
       < = op <, > = op >, <= = op <=, >= = op >=,
       min = min, max = max}
   end
   
end
signature CHAR0 =
   sig
      type t

      val equals: t * t -> bool 
      val output: t * TextIO.outstream -> unit

      val compare: t * t -> Relation.t
      val < : t * t -> bool
      val <= : t * t -> bool
      val > : t * t -> bool
      val >= : t * t -> bool
      val max: t * t -> t
      val min: t * t -> t

      val space: t
      val dash: t
      val newline: t
      val minChar: t
      val maxChar: t

      val isPrint: t -> bool
      val isLower: t -> bool
      val isAscii: t -> bool
      val isUpper: t -> bool
      val isSpace: t -> bool
      val isAlphaNum: t -> bool
      val isGraph: t -> bool
      val isHexDigit: t -> bool
      val isDigit: t -> bool
      val isAlpha: t -> bool

      val succ: t -> t
      val pred: t -> t
	 
      val toInt: t -> int
      val fromInt: int -> t
      val toPrintable: t -> string

      val ord: t -> int
      val chr: int -> t

      val digitToInt: t -> int option

      val toString: t -> string
      val fromString: string -> t option

      val toCString: t -> string
      val fromCString: string -> t option
   end where type t = char
structure Char0: CHAR0 =
struct

open Pervasive.Char

type t = char

val space = #" "
val dash = #"-"
val newline = #"\n"

val maxChar = #"\255"
val minChar = #"\000"
   
val toInt = ord
val fromInt = chr

val equals = op =

val {max, min, compare, ...} = Relation.lessEqual{< = op <, equals = equals}

val toString = Pervasive.String.str

fun digitToInt(c: t): int option =
   if isDigit c
      then SOME(ord c - ord #"0")
   else NONE

fun toPrintable c =
   if isPrint c
      then (case c of
	       #"\"" => "\\\""
	     | #"\\" => "\\\\"
	     | _ => toString c)
   else (case c of
	    #"\n" => "\\n"
	  | #"\t" => "\\t"
	  | _ => "\\" ^ (let val n = ord c
			    val d1 = n div 100
			    val n = n - d1 * 100
			    val d2 = n div 10
			    val d3 = n - d2 * 10
			    val t = Pervasive.Int.toString
			 in t d1 ^ t d2 ^ t d3
			 end))

val toCString =
   fn #"\000" => "\\000"
    | c => toCString c
	 
fun output(c, out) = Pervasive.IO.output(out, toString c)

end
signature ORDER0 =
   sig
      type t
      val compare: t * t -> Relation.t
      val equals: t * t -> bool
      val < : t * t -> bool
      val <= : t * t -> bool
      val > : t * t -> bool
      val >= : t * t -> bool
      val max: t * t -> t
      val min: t * t -> t
   end
signature STRING0 =
   sig
      structure Char: CHAR0

      include ORDER0 where type t = string

      val empty: t
      val newline: t
	 
      val isEmpty: t -> bool

      val contains: t * Char.t -> bool
		    
      val fromChar: Char.t -> t
      val toChar: t -> Char.t

      val make: int * Char.t -> t
      val last: t -> Char.t

      (* beginning at start, with length chars *)
      val substring1: t * {start: int, length: int} -> t

      (* inclusive of start, exclusive of finish *)
      val substring2: t * {start: int, finish: int} -> t

      val isSubstring: {string: string, substring: string} -> bool
	 
      val prefix: t * int -> t
      val suffix: t * int -> t
      val dropPrefix: t * int -> t
      val dropSuffix: t * int -> t
      val dropLast: t -> t

      val isPrefix: {string: t, prefix: t} -> bool
      val isSuffix: {string: t, suffix: t} -> bool

      (* splits the string into substrings broken at char,
       e.g. split("foo$bar$baz", #"$") = ["foo", "bar", "baz"] *)
      val split: t * Char.t -> t list
	 
      val output: t * TextIO.outstream -> unit

      (* Basis Library *)
      val maxSize: int
      val size: t -> int
      val sub: t * int -> Char.t
      val concat: t list -> t
      val ^ : t * t -> t
      val implode: Char.t list -> t
      val explode: t -> Char.t list
      val tokens: (Char.t -> bool) -> t -> t list 
      val fields: (Char.t -> bool) -> t -> t list 
      val translate: (Char.t -> string) -> string -> string

      val fromString: t -> t option
      val toString: t -> t
      val fromCString: t -> t option
      val toCString: t -> t
   end
structure String0: STRING0 =
struct

structure PC = Pervasive.Char
structure PS = Pervasive.String   
open PS
structure Char = Char0

type t = string

val empty = ""

val newline = "\n"

fun isEmpty "" = true
  | isEmpty _ = false
    
val length = size

fun last s = sub(s, length s - 1)

fun append(x, y) = x ^ y

exception ToChar
fun toChar s = if length s = 1 then sub(s, 0) else raise ToChar

val fromChar = str

fun contains(s, c) = PC.contains s c

val equals = op =
   
val {compare, ...} = Relation.lessEqual{< = op <, equals = equals}

fun output(s, out) = Pervasive.IO.output(out, s)

val maxLength = PS.maxSize  (* PS.maxLen *)

fun make(n, c) =
    let fun loop n = if n = 0 then []
		     else c :: (loop(n - 1))
    in implode(loop n)
    end
   
fun substring1(s, {start, length}) =
   PS.substring(s, start, length)

fun substring2(s, {start, finish}) =
   PS.substring(s, start, finish-start)

fun prefix(s, len) =
   substring1(s, {start = 0,
		 length = len})

fun suffix(s, len) =
   substring1(s, {start = length s - len,
		  length = len})
   
fun dropSuffix(s,n) =
   substring1(s, {start=0, length = length s - n})

fun dropLast s = dropSuffix(s, 1)

fun dropPrefix(s,n) =
   substring2(s, {start=n, finish = length s})

fun isPrefix{string, prefix} = PS.isPrefix prefix string

val fromChar = PS.str

(*fun toPrintable s =
   concat(List.map Char.toPrintable (explode s))*)

fun split(s, c) =
   let
      val max = length s
      fun loop(start, cur) =
	 if cur = max
	    then [substring2(s, {start = start, finish = max})]
	 else if Char.equals(sub(s, cur), c)
		 then substring2(s, {start = start, finish = cur})
		    :: loop(cur + 1, cur + 1)
	      else loop(start, cur + 1)
   in loop(0, 0)
   end

fun max(s1, s2) = if s1 > s2 then s1 else s2
fun min(s1, s2) = if s1 < s2 then s1 else s2

local open Int
in
   fun isSuffix{string, suffix} =
      let val n = length string
	 val n' = length suffix
	 fun loop(i, j) =
	    i = n orelse (Char.equals(sub(string, i), sub(suffix, j))
			  andalso loop(i + 1, j + 1))
      in n' <= n andalso loop(n - n', 0)
      end

   fun isSubstring{string, substring} =
      let val n = length substring
	 val maxIndex = length string - n
	 fun loop i =
	    i < maxIndex
	    andalso (let
			fun loop(i, j) =
			   j = n
			   orelse (Char.equals(sub(string, i), sub(substring, j))
				   andalso loop(i + 1, j + 1))
		     in loop(i, 0)
		     end
			orelse loop(i + 1))
      in loop 0
      end
end

val toCString = translate Char.toCString

end
signature LAYOUT =
   sig
      type t

      val empty: t

      val isEmpty: t -> bool
	 
      (* convert a string to a layout object *)
      val str: string -> t

      (* layout the objects on the same line *)
      val seq: t list -> t

      (* layout the objects on separate lines*)
      val align: t list -> t

      (* layout the objects on separate lines, if necessary *)
      val mayAlign: t list -> t

      (* indent the entire object *)
      val indent: t * int -> t

      (* put string between elements *)
      val separate: t list * string -> t list

      (* adds string at beginning of all objects except first *)
      val separateLeft: t list * string -> t list

      (* adds string at the end of all objects except last *) 
      val separateRight: t list * string -> t list

      val alignPrefix: string * t list -> t

      val ignore: 'a -> t
	 
      (* semi-standard layouts *)
      val option: ('a -> t) -> 'a option -> t
      val tuple2: ('a -> t) * ('b -> t) -> 'a * 'b -> t
      val tuple3: ('a -> t) * ('b -> t) * ('c -> t) -> 'a * 'b * 'c -> t
      val tuple4: ('a -> t) * ('b -> t) * ('c -> t) * ('d -> t)
	 -> 'a * 'b * 'c * 'd -> t
      val list: t list -> t
      val tuple: t list -> t
      val vector: t list -> t
      val paren: t -> t
      val record: (string * t) list -> t
      val namedRecord: string * (string * t) list -> t

      val toString: t -> string
	 
      (* print the object *)
      val print: t * (string -> unit) -> unit
      val outputWidth: t * int * Out.t -> unit
      val output: t * Out.t -> unit
      val outputTree: t * Out.t -> unit

      val makeOutput: ('a -> t) -> 'a * Out.t -> unit

      (* whether or not to print things in detail -
       * routines that create layouts should use this flag to decide
       * how detailed to print
       *)
      val detailed: bool ref
      val switch: {detailed: 'a -> t,
		    normal: 'a -> t} -> 'a -> t
   end
structure Layout: LAYOUT =
struct

val detailed = ref false

fun switch{detailed=d,normal=n} x =
   if !detailed then d x else n x
   
structure String = String0

datatype t = T of {length: int,
		   tree: tree}
and tree =
   Empty
 | String of string
 | Sequence of t list
 | Align of {force: bool, rows: t list}
 | Indent of t * int

fun length(T{length, ...}) = length

val empty = T{length = 0, tree = Empty}

fun isEmpty(T{length = 0, ...}) = true
  | isEmpty _ = false

fun str s =
   case s of
      "" => empty
    | _ => T{length = String.size s, tree = String s}

fun fold(l, b, f) = foldl f b l
	 
fun seq ts =
   let val len = fold(ts, 0, fn (t,n) => n + length t)
   in case len of
      0 => empty
    | _ => T{length = len, tree = Sequence ts}
   end

local
   fun make force ts =
      let
	 fun loop ts =
	    case ts of
	       [] => (ts, 0)
	     | t :: ts =>
		  let val (ts,n) = loop ts
		  in case length t of
		     0 => (ts,n)
		   | n' => (t :: ts, n + n' + 1)
		  end
	 val (ts, len) = loop ts
      in case len of
	 0 => empty
       | _ => T{length = len - 1, tree = Align{force = force, rows = ts}}
      end
in
   val align = make true
   val mayAlign = make false
end

fun indent(t,n) = T{length = length t, tree = Indent(t,n)}

val tabSize = 8
fun blanks n =
   String.make(n div tabSize, #"\t") ^ String.make(n mod tabSize, #" ")

fun outputTree(t, out) =
   let val print = Out.outputc out
      fun loop(T{tree, length}) =
	 (print "(length " ;
	  print(Int.toString length) ;
	  print ")" ;
	 case tree of
	    Empty => print "Empty"
	  | String s => (print "(String " ; print s ; print ")")
	  | Sequence ts => loops("Sequence", ts)
	  | Align{force, rows} => loops("Align", rows)
	  | Indent(t,n) => (print "(Indent " ;
			    print(Pervasive.Int.toString n) ;
			    print " " ;
			    loop t ;
			    print ")"))
      and loops(s, ts) = (print "(" ; print s ;
			 app (fn t => (print " " ; loop t)) ts ;
			 print ")")
   in loop t
   end

fun toString t =
   let
      fun loop(T{tree, ...}, accum) =
	 case tree of
	    Empty => accum
	  | String s => s :: accum
	  | Sequence ts => fold(ts, accum, loop)
	  | Align{force, rows} => fold(rows, accum, loop)
	  | Indent(t, _) => loop(t, accum)
   in String.concat(rev(loop(t, [])))
   end

fun print{tree: t,
	  print: string -> unit,
	  lineWidth: int} =
   let
      (*val _ = outputTree(t, out)*)
      fun newline() = print "\n"

      fun outputCompact(t, {pos, havePrinted}) =
	 let
	    fun loop(T{tree, ...}) =
	       case tree of
		  Empty => ()
		| String s => print s
		| Sequence ts => app loop ts
		| Indent(t, _) => loop t
		| Align{rows, ...} =>
		     case rows of
			[] => ()
		      | t :: ts => (loop t ;
				  app (fn t => (print " " ; loop t)) ts)
	 in loop t ; {pos = pos + length t, havePrinted = length t > 0}
	 end

      fun loop(t as T{length, tree}, state as {pos, havePrinted}) =
	 let
	    fun prePrint() =
	       if havePrinted then ()
	       else if pos >= 0 then print(blanks pos)
		    else (* Should print a warning.  Aborting
			  * is too annoying.
			  *)
		       (*Error.error "can't print negative space" *)
		       ()
	 in (*outputTree(t, Out.error) ;*)
	    case tree of
	       Empty => state
	     | Indent(t,n) =>
		  loop(t, if havePrinted then state
			  else {pos = pos + n, havePrinted = false})
	     | Sequence ts => fold(ts, state, loop)
	     | String s =>
		  (prePrint() ; print s ;
		   {pos = pos + length, havePrinted = true})
	     | Align{force, rows} =>
		  if not force andalso pos + length <= lineWidth
		     then (prePrint() ; outputCompact(t, state))
		  else (case rows of
			   [] => state
			 | t :: ts =>
			      fold
			      (ts, loop(t, state), fn (t, _) =>
			       (newline() ;
				loop(t, {pos = pos,
					 havePrinted = false}))))
	 end
   in loop(tree, {pos=0, havePrinted=false}) ; ()
   end

fun outputWidth(t, width, out) =
   print{tree = t,
	 lineWidth = width,
	 print = Out.outputc out}

local
   val defaultWidth = 80
in
   fun output(t, out) = outputWidth(t, defaultWidth, out)
   val print =
      fn (t, p) => print{tree = t, lineWidth = defaultWidth, print = p}
end

fun makeOutput layoutX (x, out) = output(layoutX x, out)

fun ignore _ = empty
   
fun separate(ts, s) =
   case ts of
      [] => []
    | t :: ts => t :: (let val s = str s
		       fun loop [] = []
			 | loop(t :: ts) = s :: t::(loop ts)
		   in loop ts
		   end)
   
fun separateLeft(ts, s) =
   case ts of
      [] => []
    | [t] => ts
    | t :: ts => t :: (map (fn t => seq[str s, t]) ts)

fun separateRight(ts, s) =
   rev(let val ts = rev ts
       in case ts of
	  [] => []
	| [t] => ts
	| t :: ts => t :: (map (fn t => seq[t, str s]) ts)
       end)

fun alignPrefix(prefix, ts) =
   case ts of
      [] => empty
    | t :: ts =>
	 mayAlign[t, indent(mayAlign(map (fn t => seq[str prefix, t]) ts),
			    ~(String.size prefix))]

local
   fun sequence(start, finish) ts =
      seq[str start, mayAlign(separateRight(ts, ",")), str finish]
in
   val list = sequence("[", "]")
   val tuple = sequence("(", ")")
   val vector = sequence("#[", "]")
   fun record fts = sequence("{", "}") (map (fn (f, t) => seq[str(f ^ " = "), t])
				       fts)
end

fun namedRecord(name, fields) = seq[str name, record fields]

fun paren t = seq[str "(", t, str ")"]

fun option layoutX opt =
   case opt of
      NONE => str "NONE"
    | SOME x => paren(seq[str "SOME ", layoutX x])

fun tuple2(l1, l2)(x1, x2) = tuple[l1 x1, l2 x2]
fun tuple3(l1, l2, l3)(x1, x2, x3) = tuple[l1 x1, l2 x2, l3 x3]
fun tuple4(l1, l2, l3, l4)(x1, x2, x3, x4) = tuple[l1 x1, l2 x2, l3 x3, l4 x4]
  
end
signature STRING =
   sig
      include STRING0

      val layout: t -> Layout.t
   end
structure String: STRING =
   struct
      open String0
	 
      val layout = Layout.str
   end
signature INSTREAM =
   sig
      type t

      val standard: t
      val openIn: string -> t
      val close: t -> unit

      val endOf: t -> bool

      val sameContents: t * t -> bool

      val peekChar: t -> char option

      val input: t -> string
      val inputN: t * int -> string
      val inputChar: t -> char option
      val inputLine: t -> string

      val ignoreSpaces: t -> unit

      (* inputTo(ins, pred) inputs up to but not including the first char
       that satisfies pred *)
      val inputTo: t * (char -> bool) -> string

      val inputToSpace: t -> string
      val inputToChar: t * char -> string
	 
      val inputAll: t -> string
      val outputAll: t * Out.t -> unit

      val inputNothing: t -> unit

      val layout: t -> Layout.t
   end
(*-------------------------------------------------------------------*)
(*                             Instream                              *)
(*-------------------------------------------------------------------*)

structure Instream: INSTREAM =
struct

structure String = String0
structure Char = String.Char
structure I = Pervasive.IO
open I
   
type t = I.instream

val standard = stdIn
val close = closeIn
val inputChar = input1
val peekChar = lookahead   
val endOf = endOfStream

(*fun inputAll i =
   let
      fun loop accum = 
	 if endOf i then accum
	 else (input i) :: accum
   in String.concat(rev (loop []))
   end
*)
fun outputAll(i, out) =
   let
      fun loop() = if endOf i then ()
		   else (Out.output(out, input i) ; loop())
   in loop()
   end

fun inputTo(i, p) =
   let
      fun loop accum =
	 case peekChar i of
	    NONE => accum
	  | SOME c => if p c then accum else (inputChar i ; loop(c :: accum))
   in String.implode(rev(loop []))
   end

fun sameContents(in1, in2) =
   let
      fun loop() =
	 case (input1 in1, input1 in2) of
	    (NONE, NONE) => true
	  | (SOME c1, SOME c2) => Char.equals(c1, c2) andalso loop()
	  | _ => false
   in loop()
   end

fun inputToSpace i = inputTo(i, Char.isSpace)
fun inputToChar(i, c) = inputTo(i, fn c' => Char.equals(c, c'))
fun ignoreSpaces i = (inputTo(i,not o Char.isSpace) ; ())

(*
fun duplicate i n =
   let val fis = LU.duplicate n I.openChanIn
      val fs = LU.map fis #1
      val is = LU.map fis #2
      fun dup() = (case inputChar i of
		      NONE => LU.foreach is close
		    | SOME c => let val s = C.toString c
				in (LU.foreach fs (fn f => f s) ;
				    dup())
				end)
   in (CML.spawn dup ; is)
   end
*)
fun inputNothing _ = ()

fun layout _ = Layout.str "<instream>"
   
end

structure In = Instream
signature CHAR =
   sig
      include CHAR0

      val layout: t -> Layout.t
   end
   
(*-------------------------------------------------------------------*)
(*                               Char                                *)
(*-------------------------------------------------------------------*)

structure Char: CHAR =
   struct
      open Char0
      fun layout c = Layout.str(toString c)
   end
signature ITERATE =
   sig
      val iterate: 'a * ('a -> bool) * ('a -> 'a) -> 'a
      (* iterate(s, p, f) = f(...f(f(s))) until satisfies p *)

      val whileDo: (unit -> bool) * (unit -> unit) -> unit

      val repeatUntil: (unit -> unit) * (unit -> bool) -> unit
   end
(*-------------------------------------------------------------------*)
(*                              Iterate                              *)
(*-------------------------------------------------------------------*)

structure Iterate: ITERATE =
struct
   
fun iterate(start, isTerm,next) =
   let fun loop s = if isTerm s then s else loop(next s)
   in loop start
   end
   
fun whileDo(test, f) = iterate((),not o test, f)
   
fun repeatUntil(f, test) = (f() ; iterate((), test, f))
   
end
signature T =
   sig
      type t
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end

signature BASIC_RING =
   sig
      include T

      val zero: t
      val one: t
      val + : t * t -> t
      val ~ : t -> t
      val * : t * t -> t
   end
signature RING =
   sig
      include BASIC_RING

      val - : t * t -> t
      val isZero: t -> bool
	 
      val negOne: t
      val two: t
      val three: t
      val add1: t -> t
      val sub1: t -> t

      val double: t -> t
      val square: t -> t
   end

signature ORDERED_RING =
   sig
      include RING

      (* include BASIC_ORDERED_RING *)
      val compare: t * t -> Relation.t
      val < : t * t -> bool
      val <= : t * t -> bool
      val > : t * t -> bool
      val >= : t * t -> bool
      (* *)

      val isPositive: t -> bool
      val isNegative: t -> bool
      val abs: t -> t

      val foldl: t * t * 'a * ('a * t -> 'a) -> 'a
	 
      val sum: {from: t,
		 to: t,
		 term: t -> t} -> t
	 
      val prod: {from: t,
		  to: t,
		  term: t -> t} -> t
      val prodFromTo: t * t -> t

      val factorial: t -> t
      
      val max: t * t -> t
      val min: t * t -> t
   end

signature INTEGER =
   sig
      include ORDERED_RING

      (* include BASIC_INTEGER *)
      val div: t * t -> t
      val mod: t * t -> t
      val quot: t * t -> t
      val rem: t * t -> t
      val randomRange: t * t -> unit -> t
      val toReal: t -> real
      val toInt: t -> int
      val fromInt: int -> t
      exception Input
      val input: In.t -> t
      val toString: t -> string
      val fromString: string -> t option
      (* *)

      val toCommaString: t -> string

      val output: t * Out.t -> unit

      val inputBetween :
	 {ins: In.t,
	  error: unit -> unit,
	  min: t,
	  max: t} -> t
	 
      val isEven: t -> bool
      val isOdd: t -> bool

      val choose: t * t -> t

      val ^ : t * t -> t

      val gcd: t * t -> t
      val lcm: t * t -> t

      (* Iteration *)
      val fromTo: {start:t, stop:t, step:t} * 'a * (t * 'a -> 'a) -> 'a
	    
      val fold: t * t * 'a * (t * 'a -> 'a) -> 'a

      val foldDown: t * t * 'a * (t * 'a -> 'a) -> 'a

      val forall: t * t * (t -> bool) -> bool
	 
      val forStep:
	 {start:t, stop:t, step:t} * (t -> unit) -> unit

      val foreach: t * t * (t -> unit) -> unit

      val forDown: t * t * (t -> unit) -> unit
   end



(*-------------------------------------------------------------------*)
(*                               Ring                                *)
(*-------------------------------------------------------------------*)

functor Ring (R: BASIC_RING): RING =
struct

open R

val (op -) = fn (n, m) => n + ~m

fun add1 n = n + one

fun sub1 n = n - one

val negOne = sub1 zero
   
val two = add1 one

val three = add1 two

fun double n = n + n
fun square n = n * n

fun isZero n = equals(n, zero)

end
signature BASIC_ORDERED_RING =
   sig
      include BASIC_RING
      
      (* include ORDER *)
      val compare: t * t -> Relation.t
      val < : t * t -> bool
      val <= : t * t -> bool
      val > : t * t -> bool
      val >= : t * t -> bool
      (* *)
   end

(*-------------------------------------------------------------------*)
(*                            OrderedRing                            *)
(*-------------------------------------------------------------------*)
  
functor OrderedRing(R: BASIC_ORDERED_RING): ORDERED_RING =
struct

structure U = Ring(R)
open R U

fun isPositive n = n > zero

fun isNegative n = n < zero

fun abs n = if isPositive n then n else ~ n

fun foldl(from, to, b, f) =
   let fun fold(n, a) = if n > to then a
		       else fold(add1 n, f(a,n))
   in fold(from, b)
   end

local 
   fun abs (combine, base) {from, to, term} =
      foldl(from, to, base, fn (a, i) => combine(a, term i))
in
   val sum = abs((op +), zero)
   val prod = abs((op * ), one)
end

fun prodFromTo(f, t) = prod{from = f, to = t, term = fn n => n}

fun factorial n = prodFromTo(one,n)

fun max(m,n) = if m > n then m else n

fun min(m,n) = if m < n then m else n
   
end
signature BASIC_INTEGER =
   sig
      include BASIC_ORDERED_RING

      val div: t * t -> t
      val mod: t * t -> t
      val quot: t * t -> t
      val rem: t * t -> t

      (* randomRange(i, j)() is uniform in [i, j]
       * It is curried because it does some computation with i and j.
       *)
      val randomRange: t * t -> unit -> t
	 
      val toReal: t -> real
      val toInt: t -> int
      val fromInt: int -> t

      exception Input
      val input: In.t -> t

      val toString: t -> string
      val fromString: string -> t option
   end
(*-------------------------------------------------------------------*)
(*                              Integer                              *)
(*-------------------------------------------------------------------*)

functor Integer(I: BASIC_INTEGER): INTEGER =
struct

structure R = OrderedRing(I)
open I R

fun toCommaString n =
   let
      fun loop(chars, accum) =
	 case chars of
	    x1 :: x2 :: x3 :: (chars as _ :: _) =>
	       loop(chars, #"," :: x3 :: x2 :: x1 :: accum)
	  | _ => implode(rev chars @ accum)
   in loop(rev(explode(toString n)), [])
   end

fun choose(n, k) =
   let val k = max(k,n-k)
   in (prodFromTo(add1 k,n)) div (factorial(n-k))
   end

fun output(n, out) = Out.output(out, toString n)
   
fun isEven n = isZero(n mod two)
val isOdd = not o isEven

val (op ^) = fn (base, exp) =>
   let fun power exp = if isZero exp then one
		       else let val n = square(power(quot(exp, two)))
			    in if isEven exp then n
			       else base*n
			    end
   in power exp
   end
   
fun powerMod{base, exp, modulus} =
   let val m = modulus
      fun loop n =
	 if isZero n then one
	 else let val r = square(loop(n div two)) mod m
	      in if isEven n then r else (base*r) mod m
	      end
   in loop exp
   end			 

fun gcd(a, b) = if isZero b then a else gcd(b, a mod b)

fun lcm(a, b) = (a*b) div (gcd(a, b))

fun inputBetween{ins, error, min, max} =
   let
      fun loop() =
	 let fun continue() = (error() ; loop())
	 in let val n = input ins
	    in if min <= n andalso n <= max
		  then n
	       else continue()
	    end handle Input => continue()
	 end
   in loop()
   end

fun fromTo({start, stop, step}, s,nextS) =
    let val cmp = if isPositive step
		     then (op >)
		  else (op <)
    in #1(Iterate.iterate((s, start),
		  fn (_, i) => cmp(i, stop),
		  fn (s, i) => (nextS(i, s), i + step)))
    end

fun fold(start, stop, s,nextS) =
   fromTo({start = start, stop = stop, step = one}, s, nextS)

fun foldDown(start, stop, s,nextS) =
   fromTo({start = start,
	   stop = stop,
	   step = negOne},
	  s,
	  nextS)

fun forall(m,n, f) =
   let
      fun forall m = m > n
	 orelse (f m andalso forall(add1 m))
   in forall m
   end

fun forStep(sss, f) = fromTo(sss, (), fn (k, ()) => f k)
   
fun foreach(start, stop, f) = forStep({start = start,
				       stop = stop,
				       step = one},
				    f)
   
fun forDown(start, stop, f) =
   forStep({start = start,
	    stop = stop,
	    step = negOne},
	   f)

end

signature FIXED_INTEGER =
   sig
      include INTEGER
	 
      val maxInt: t
      val minInt: t
      val random: unit -> t
   end where type t = int
(*-------------------------------------------------------------------*)
(*                           FixedInteger                            *)
(*-------------------------------------------------------------------*)

structure FixedInteger: FIXED_INTEGER =
struct

structure I =
   struct
      structure I = Pervasive.Int
      open I
	 
      type t = int
	 
      val zero = 0
      val one = 1
	 
      fun fromInt n = n
      fun toInt n = n
      val toReal = Pervasive.Real.fromInt
      val fromString = I.fromString
      val toString = I.toString
      val layout = Layout.str o toString
	     
      local
	 structure R = Random
	 val seeds = (13, 14)
	 val r = R.rand seeds
      in
	 fun randomRange(lo, hi) =
	    let val f = R.randRange (lo, hi)
	    in fn () => f r
	    end
      end

      val equals = op =
	 
      val {compare, ...} = Relation.lessEqual{< = op <, equals = equals}

      fun output(i, out) = Out.output(out, toString i)

      exception Input
      fun input i = (In.ignoreSpaces i ;
		     case fromString(In.inputToSpace i) of
			NONE => raise Input
		      | SOME n => n)
	    
      val maxInt = valOf maxInt
      val minInt = valOf minInt
   end

structure I' = Integer(I)
open I I'

val random = randomRange(minInt, maxInt)

end

structure Integer = FixedInteger
structure Int = Integer
signature CONTAINER =
   sig
      type 'a t

      val size: 'a t -> Int.t
      val foreach: 'a t * ('a -> unit) -> unit
      val forall: 'a t * ('a -> bool) -> bool
      val exists: 'a t * ('a -> bool) -> bool
      val contains: 'a t * 'a * ('a * 'a -> bool) -> bool
      val equals: 'a t * 'b t * ('a * 'b -> bool) -> bool
      val layout: ('a -> Layout.t) -> 'a t -> Layout.t

      (* Specify a string to separate the elements *)
      val layoutSep: string * ('a -> Layout.t) -> 'a t -> Layout.t
   end
signature SEQUENCE =
   sig
      include CONTAINER
	 
      val isEmpty: 'a t -> bool
      val length: 'a t -> Int.t

      val forall2: 'a t * 'b t * ('a * 'b -> bool) -> bool
      val foreach2: 'a t * 'b t * ('a * 'b -> unit) -> unit
      val foreach3: 'a t * 'b t * 'c t * ('a * 'b * 'c -> unit) -> unit
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldi: 'a t * 'b * (int * 'a * 'b -> 'b) -> 'b
      val fold2: 'a t * 'b t * 'c * ('a * 'b * 'c -> 'c) -> 'c
      val fold3: 'a t * 'b t * 'c t * 'd * ('a * 'b * 'c * 'd -> 'd) -> 'd
      val foldr: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldr2: 'a t * 'b t * 'c *
	 ('a * 'b * 'c -> 'c) -> 'c
      val foralli: 'a t * (Int.t * 'a -> bool) -> bool
      val foreachi: 'a t * (Int.t * 'a -> unit) -> unit
      val nth: 'a t * Int.t -> 'a
      val peek: 'a t * ('a -> bool) -> 'a option
      val lookup: 'a t * ('a -> bool) -> 'a
      (*val sameLength: 'a t t -> bool*)
      val outputN :
	 Int.t (* how many *)
	 * string (* separator *)
	 * ('a * Out.t -> unit) (* output element *)
	 -> 'a t * Out.t -> unit
      val toList: 'a t -> 'a list
   end

signature DESTRUCTABLE_SEQUENCE =
   sig
      include SEQUENCE
      
      (* include BASIC_SEQUENCE *)
      val destruct: 'a t -> ('a * 'a t) option
      (* *)
	 
      val first: 'a t -> 'a
      val rest: 'a t -> 'a t
      val nthRest: 'a t * Int.t -> 'a t
      val last: 'a t -> 'a
      val index: 'a t * ('a -> bool) -> Int.t option
   end
signature LIST =
   sig
      include DESTRUCTABLE_SEQUENCE

      (* include BASIC_LIST_ITER *)
      val empty: unit -> 'a t
      val cons: 'a * 'a t -> 'a t
      datatype ('a, 'b) iter =
	 Empty
       | First of 'a * 'b
       | Rest of 'a t
      val iterate: 'b * ('b -> ('a, 'b) iter) -> 'a t
      (* *)

      val head: 'a t -> 'a
      val tail: 'a t -> 'a t
      val nthTail: 'a t * Int.t -> 'a t
      val single: 'a -> 'a t
      val appendThunk: 'a t * (unit -> 'a t) -> 'a t
      val append: 'a t * 'a t -> 'a t
	 
      val flatten: 'a t t -> 'a t
	 
      val reverse: 'a t -> 'a t

      (* reverseAppend(l1, l2) = append(reverse l1, l2) *)
      val reverseAppend: 'a t * 'a t -> 'a t

      val zip: 'a t * 'b t -> ('a * 'b) t
      val zip3: 'a t * 'b t * 'c t -> ('a * 'b * 'c) t
      val unzip: ('a * 'b) t -> 'a t * 'b t
      val unzip3: ('a * 'b * 'c) t -> 'a t * 'b t * 'c t

      (* mapAppend(l1, f, l2) = append(map(l1, f), l2) *)
      val mapAppend: 'a t * ('a -> 'b) * 'b t -> 'b t
      val map: 'a t * ('a -> 'b) -> 'b t
      val mapi: 'a t * (int * 'a -> 'b) -> 'b t
      val map2: 'a t * 'b t * ('a * 'b -> 'c) -> 'c t
      val map3: 'a t * 'b t * 'c t * ('a * 'b * 'c -> 'd) -> 'd t
      val map4: 'a t * 'b t * 'c t * 'd t * ('a * 'b * 'c * 'd -> 'e) -> 'e t
      val tabulate: 'a t * (Int.t * 'a -> 'b) -> 'b t
      val partition: 'a t * ('a -> bool) -> {yes: 'a t, no: 'a t}
      val keepAll: 'a t * ('a -> bool) -> 'a t
      val keepAllMap: 'a t * ('a -> 'b option) -> 'b t
      val removeAll: 'a t * ('a -> bool) -> 'a t

      (* group according to equivalence relation *)
      val equivalence: 'a t * ('a * 'a -> bool) -> 'a t t

      val separate: 'a t * 'a -> 'a t

      (* removes firsts element if it exists
       returns NONE if pred is false on everything *)
      val peekRemove: 'a t * ('a -> bool) -> ('a t * 'a) option
	 
      (* error if predicate isn't true on some element *)
      val removeFirst: 'a t * ('a -> bool) -> 'a t

      (* removes first element on which predicate is true.
       returns original list if predicate is false on every elt. *)
      val remove: 'a t * ('a -> bool) -> 'a t

      val splitLast: 'a t -> 'a t * 'a
      val allButLast: 'a t -> 'a t
	  
      val flattenMap: 'a t * ('a -> 'b t) -> 'b t
      val duplicate: Int.t * (unit -> 'a) -> 'a t
      val firstN: 'a t * Int.t -> 'a t
      val splitAtMost: 'a t * Int.t -> ('a t * 'a t) option

      (*   cross [[1, 2, 3], [4, 5], [6, 7]] =
       [[1, 4, 6], [1, 4, 7], [1, 5, 6], [1, 5, 7],
       [2, 4, 6], [2, 4, 7], [2, 5, 6], [2, 5, 7],
       [3, 4, 6], [3, 4, 7], [3, 5, 6], [3, 5, 7]] *)
      val cross: 'a t t -> 'a t t
	 
      val prefixes: 'a t -> 'a t t
      val suffixes: 'a t -> 'a t t
	 
      (* extract each element paired with the elements before it
       and after it in the same order as in the list *)
      val each: 'a t -> ('a t * 'a * 'a t) t

      val listTo: 'a list -> 'a t

      val compare: 'a t * 'a t * ('a * 'a -> order) -> order

      val ordered :
	 {< : 'a * 'a -> bool}
	 -> {insert: 'a t * 'a -> 'a t,
	     insertionSort: 'a t  -> 'a t,
	     median: 'a t -> 'a,
	     orderStatistic: 'a t * Int.t -> 'a,
	     partition: 'a t * 'a -> 'a t * 'a t,
	     max: 'a t -> 'a,
	     min: 'a t -> 'a,
	     largest: 'a t * Int.t -> 'a t,
	     smallest: 'a t * Int.t -> 'a t}

      val power: 'a t -> 'a t t
      val subsets: 'a t * int -> 'a t t
	 
      val set: {equal: 'a * 'a -> bool,
		 layout: 'a -> Layout.t}
	 -> {empty: 'a t,
	     singleton: 'a -> 'a t,
	     size: 'a t -> int,
	     equals: 'a t * 'a t -> bool,
	     <= : 'a t * 'a t -> bool,
	     >= : 'a t * 'a t -> bool,
	     < : 'a t * 'a t -> bool,
	     > : 'a t * 'a t -> bool,
	     + : 'a t * 'a t -> 'a t,
	     - : 'a t * 'a t -> 'a t,
	     intersect: 'a t * 'a t -> 'a t,
	     unions: 'a t t -> 'a t,
	     add: 'a t * 'a -> 'a t,
	     remove: 'a t * 'a -> 'a t,
	     contains: 'a t * 'a -> bool,
	     power: 'a t -> 'a t t,
	     subsets: 'a t * int -> 'a t t,
	     areDisjoint: 'a t * 'a t -> bool,
	     subset: 'a t * ('a -> bool) -> 'a t, (* aka keepAll *)
	     subsetSize: 'a t * ('a -> bool) -> int,
	     replace: 'a t * ('a -> 'a option) -> 'a t,
	     map: 'a t * ('a -> 'a) -> 'a t,
	     layout: 'a t -> Layout.t}
   end 
signature STRICT_LIST =
   sig
      include LIST
   end
signature BASIC_SEQUENCE =
   sig
      type 'a t
      val destruct: 'a t -> ('a * 'a t) option
   end
signature BASIC_LIST =
   sig
      include BASIC_SEQUENCE

      val empty: unit -> 'a t
      val cons: 'a * 'a t -> 'a t
   end
signature BASIC_LIST_ITER =
   sig
      include BASIC_LIST
      
      datatype ('a, 'b) iter =
	 Empty
       | First of 'a * 'b
       | Rest of 'a t

      val iterate: 'b * ('b -> ('a, 'b) iter) -> 'a t
   end
(*-------------------------------------------------------------------*)
(*                       DestructableSequence                        *)
(*-------------------------------------------------------------------*)

functor DestructableSequence(structure S: BASIC_SEQUENCE)
  : DESTRUCTABLE_SEQUENCE =
struct

open Int S

fun isEmpty s =
   case destruct s of
      NONE => true
    | SOME _ => false

fun first s =
   case destruct s of
      SOME(x, _) => x
    | NONE => Error.error "first"

fun rest s =
   case destruct s of
      SOME(_, s) => s
    | NONE => Error.error "rest"

fun last s =
   case destruct s of
      NONE => Error.error "last"
    | SOME xs =>
	 let
	    fun last(x, s) =
	       case destruct s of
		  SOME xs => last xs
		| NONE => x
	 in last xs
	 end

fun fold(s, b, f) =
   let
      fun fold(s, b) =
	 case destruct s of
	    NONE => b
	  | SOME(x, s) => fold(s, f(x, b))
   in fold(s, b)
   end

fun foldi(s, b, f) =
   #2(fold(s, (0, b), fn (x, (i, b)) => (i + 1,  f(i, x, b))))

fun fold2(s, s', b, f) =
   let
      fun fold(s, s', b) =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => b
	  | (SOME(x, s), SOME(x', s')) => fold(s, s', f(x, x', b))
	  | _ => Error.error "fold2"
   in fold(s, s', b)
   end

fun fold3(s1, s2, s3, b, f) =
   let
      fun fold(s1, s2, s3, b) =
	 case (destruct s1, destruct s2, destruct s3) of
	    (NONE, NONE, NONE) => b
	  | (SOME(x1, s1), SOME(x2, s2), SOME(x3, s3)) => fold(s1, s2, s3, f(x1, x2, x3, b))
	  | _ => Error.error "fold3"
   in fold(s1, s2, s3, b)
   end

fun size s = fold(s, zero, fn (_,n) => add1 n)
val length = size

fun nthRest(s,n) = 
   if isZero n then s
   else case destruct s of
      NONE => Error.error "nthRest"
    | SOME(_, s) => nthRest(s, sub1 n)
      
fun nth(s,n) = first(nthRest(s,n))

fun index(s, p) =
   let
      fun index(s,n) =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) => if p x then SOME n else index(s, add1 n)
   in index(s, zero)
   end
   
fun foldr(s, b, f) =
   let
      fun fold s =
	 case destruct s of
	    NONE => b
	  | SOME(x, s) => f(x, fold s)
   in fold s
   end

fun toList s = foldr(s, [], fn (x, l) => x :: l)

fun foldr2(s, s', b, f) =
    let
       fun fold(s, s') =
	  case (destruct s, destruct s') of
	     (NONE, NONE) => b
	   | (SOME(x, s), SOME(x', s')) => 
		f(x, x', fold(s, s'))
	   | _ => Error.error "foldr2"
    in fold(s, s')
    end

fun foreach(s, f) =
   let
      fun foreach s =
	 case destruct s of
	    NONE => ()
	  | SOME(x, s) => (f x ; foreach s)
   in foreach s
   end

fun foralli(s, f) =
   let
      fun forall(s, i) =
	 case destruct s of
	    NONE => true
	  | SOME(x, s) => f(i, x) andalso forall(s, add1 i)
   in forall(s, zero)
   end
			  
fun foreachi(s, f) =
   let
      fun foreach(s, i) =
	 case destruct s of
	    NONE => ()
	  | SOME(x, s) => (f(i, x) ;
			  foreach(s, add1 i))
   in foreach(s, zero)
   end

(*
fun foldl2Cond(l, l', b, f) =
    let fun fold(l, l', b) =
	if isEmpty l then b
	else let val (x, l) = destruct l
		 val (x', l') = destruct l'
		 val (b, continue) = f(b, x, x')
	     in if continue then fold(l, l', b) else b
	     end
    in fold(l, l', b)
    end

fun foldl2(l, l', b, f) = foldl2Cond(l, l', b, fn a => (f a, true))
*)

fun foreach2(s, s', f) =
   let fun foreach(s, s') = (case (destruct s, destruct s') of
			       (NONE, NONE) => ()
			     | (SOME(x, s), SOME(x', s')) => 
				  (f(x, x') ; foreach(s, s'))
			     | _ => Error.error "foreach2")
   in foreach(s, s')
   end

fun foreach3(s, s', s'', f) =
   let fun foreach(s, s', s'') =
      (case (destruct s, destruct s', destruct s'') of
	  (NONE, NONE, NONE) => ()
	| (SOME(x, s), SOME(x', s'), SOME(x'', s'')) => 
	     (f(x, x', x'') ; foreach(s, s', s''))
	| _ => Error.error "foreach3")
   in foreach(s, s', s'')
   end

fun forall(s, f) =
   let
      fun forall s =
	 case destruct s of
	    NONE => true
	  | SOME(x, s) => f x andalso forall s
   in forall s
   end

fun forall2(s, s', f) =
   let
      fun forall(s, s') =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => true
	  | (SOME(x, s), SOME(x', s')) => f(x, x') andalso forall(s, s')
	  | _ => Error.error "forall2"
   in forall(s, s')
   end

fun peek(s, p) =
   let
      fun peek s =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) => if p x then SOME x
			 else peek s
   in peek s
   end

fun lookup arg = 
   case peek arg of
      SOME x => x
    | NONE => Error.error "lookup"

(*fun sameLength ss =
   case destruct s of
      NONE => true
    | SOME(s, ss) =>
	 let val n = length s
	 in forall(ss, fn s => I.=(n, length s))
	 end*)

fun exists(l, p) =
    case peek(l, p) of
	NONE => false
      | SOME _ => true

fun contains(l, x, eq) = exists(l, fn x' => eq(x, x'))

fun equals(s, s', eltEq) =
   let
      fun eq(s, s') =
	 case (destruct s, destruct s') of
	    (NONE, NONE) => true
	  | (SOME(x, s), SOME(x', s')) => eltEq(x, x') andalso eq(s, s')
	  | _ => false
   in eq(s, s')
   end

fun toList xs =
   case destruct xs of
      NONE => []
    | SOME(x, xs) => x :: (toList xs)
      
local open Layout
in
   fun layoutSep(sep, layoutX) xs = list(map layoutX (toList xs))

   fun layout l = layoutSep(",", l)
(*   case destruct xs of
      NONE => empty
    | SOME(x, s) =>
	 align(layoutX x ::(let
			     fun loop s =
				case destruct s of
				   NONE => []
				 | SOME(x, s) => str sep :: layoutX x :: loop s
			  in loop s
			  end))
	 *)
end
			 
fun output(separator, outputElt) (s, out) =
   case destruct s of
      NONE => ()
    | SOME(x, s) => (outputElt(x, out) ;
		    foreach(s, fn x =>
			    (Out.output(out, separator) ;
			     outputElt(x, out))))

fun outputN(n, sep, outputElt) (s, out) =
   if isZero n then ()
   else case destruct s of
      NONE => ()
    | SOME (x, s) =>
	 (outputElt(x, out) ;
	  let
	     fun loop(s,n) =
		if isZero n then ()
		else case destruct s of
		   NONE => ()
		 | SOME(x, s) => (Out.output(out, sep) ;
				 outputElt(x, out) ;
				 loop(s, sub1 n))
	  in loop(s, sub1 n)
	  end)

end
(*-------------------------------------------------------------------*)
(*                               List                                *)
(*-------------------------------------------------------------------*)

functor List(structure L: BASIC_LIST_ITER): LIST =
struct

structure S = DestructableSequence(structure S = L)
   
open Int L S

val head = first
val tail = rest
val nthTail = nthRest

fun single x = cons(x, empty())
   
fun iterateDestruct(l, f) = iterate(l, f o destruct)

fun mapAppend(l, f, l') =
   iterateDestruct(l, fn NONE => Rest l'
                       | SOME(x, l) => First(f x, l))

fun map(l, f) = mapAppend(l, f, empty())

fun map2(s, s', f) =
   iterate((s, s'), fn (s, s') =>
	   case (destruct s, destruct s') of
	      (NONE, NONE) => Empty
	    | (SOME(x, s), SOME(x', s')) => First(f(x, x'), (s, s'))
	    | _ => Error.error "map2")

fun map3(s, s', s'', f) =
   iterate((s, s', s''), fn (s, s', s'') =>
	   case (destruct s, destruct s', destruct s'') of
	      (NONE, NONE, NONE) => Empty
	    | (SOME(x, s), SOME(x', s'), SOME(x'', s'')) =>
		 First(f(x, x', x''), (s, s', s''))
	    | _ => Error.error "map3")

fun map4(s1, s2, s3, s4, f) =
   iterate((s1, s2, s3, s4), fn (s1, s2, s3, s4) =>
	   case (destruct s1, destruct s2, destruct s3, destruct s4) of
	      (NONE, NONE, NONE, NONE) => Empty
	    | (SOME(x1, s1), SOME(x2, s2), SOME(x3, s3), SOME(x4, s4)) =>
		 First(f(x1, x2, x3, x4), (s1, s2, s3, s4))
	    | _ => Error.error "map4")

fun tabulate(s, f) =
   iterate((s, zero), fn (s, i) =>
	   case destruct s of
	      NONE => Empty
	    | SOME(x, s) => First(f(i, x), (s, add1 i)))
   
fun appendThunk(s, t) =
   iterateDestruct(s, fn NONE => Rest(t())
                       | SOME(x, s) => First(x, s))

fun append(s, s') = appendThunk(s, fn () => s')
   
fun flattenMap(l, f) =
   let
      fun fm l = case destruct l of
	 NONE => empty()
       | SOME(x, l) => appendThunk(f x, fn () => fm l)
   in fm l
   end

fun reverseAppend(l, l') = fold(l, l', cons)

fun reverse s = reverseAppend(s, empty())

fun foldr(l, b, f) = fold(reverse l, b, f)

fun mapi(l, f) = reverse(#1(fold(l, (empty(), 0), fn (x, (l, i)) =>
				(cons(f(i, x), l), i + 1))))

fun zip(l, l') =
   case (destruct l, destruct l') of
      (SOME(x, l), SOME(x', l')) => cons((x, x'), zip(l, l'))
    | (NONE, NONE) => empty()
    | _ => Error.error "zip"

fun zip3(l1, l2, l3) =
   case (destruct l1, destruct l2, destruct l3) of
      (SOME(x1, l1), SOME(x2, l2), SOME(x3, l3)) =>
	 cons((x1, x2, x3), zip3(l1, l2, l3))
    | (NONE, NONE, NONE) => empty()
    | _ => Error.error "zip3"

fun unzip l =
   foldr(l, (empty(), empty()), fn ((x , y), (xs, ys)) =>
	 (cons(x, xs), cons(y, ys)))

fun unzip3 l =
   foldr(l, (empty(), empty(), empty()), fn ((x, y, z), (xs, ys, zs)) =>
	 (cons(x, xs), cons(y, ys), cons(z, zs)))

fun flatten s =
   case destruct s of
      NONE => empty()
    | SOME(x, s) => appendThunk(x, fn () => flatten s)

fun remFst(s, p,notFound) = 
   iterateDestruct(s, fn NONE => notFound()
		       | SOME(x, s) => if p x then Rest s
				      else First(x, s))

fun removeFirst(s, p) =
   remFst(s, p, fn () => Error.error "removeFirst")

fun remove(s, p) = remFst(s, p, fn () => Empty)

fun peekRemove(s, p) =
   let
      fun loop(s, accum) =
	 case destruct s of
	    NONE => NONE
	  | SOME(x, s) =>
	       if p x
		  then SOME(reverseAppend(accum, s), x)
	       else loop(s, cons(x, accum))
   in loop(s, empty())
   end

fun splitLast s =
   case destruct(reverse s) of
      NONE => Error.error "splitLast"
    | SOME(x, s) => (reverse s, x)
   
fun allButLast s = #1(splitLast s)
   
fun partition(l, p) =
   let
      fun loop(s, accum as {yes,no}) =
	 case destruct s of
	    NONE => accum
	  | SOME(x, xs) =>
	       loop(xs,
		    if p x then {yes = cons(x, yes), no = no}
		    else {yes = yes, no = cons(x,no)})
   in loop(reverse l, {yes = empty(), no = empty()})
   end

fun keepAll(l, p) = #yes(partition(l, p))
fun removeAll(l, p) = #no(partition(l, p))

fun keepAllMap(s, f) =
   let
      fun loop s =
	 case destruct s of
	    NONE => empty()
	  | SOME(x, s) =>
	       case f x of
		  NONE => loop s
		| SOME y => cons(y, loop s)
   in loop s
   end

fun separate(s, x) =
   case destruct s of
      NONE => empty()
    | SOME(y, s) =>
	 cons(y,
	      let
		 fun loop s =
		    case destruct s of
		       NONE => empty()
		     | SOME(y, s) => cons(x, cons(y, loop s))
	      in loop s
	      end)
	 
fun duplicate(n, f) =
   iterate(n, fn n => if isZero n then Empty
		     else First(f(), sub1 n))

fun firstN(s,n) =
   iterate((s,n), fn (s,n) =>
	   if isZero n then Empty
	   else case destruct s of
	      NONE => Error.error "firstN"
	    | SOME(x, s) => First(x, (s, sub1 n)))

fun firstAtMostN(s,n) =
   iterate((s,n), fn (s,n) =>
	   if isZero n then Empty
	   else case destruct s of
	      NONE => Empty
	    | SOME(x, s) => First(x, (s, sub1 n)))

fun splitAtMost(l,n) =
   if isZero n then SOME(empty(), l)
   else case destruct l of
      NONE => NONE
    | SOME(x, l) => (case splitAtMost(l, sub1 n) of
		       NONE => NONE
		     | SOME(l, l') => SOME(cons(x, l), l'))

fun prefixes l =
   cons(empty(), case destruct l of
	NONE => empty()
      | SOME(x, l) => map(prefixes l, fn p => cons(x, p)))

fun suffixes l = map(prefixes(reverse l), reverse)

fun each l =
   zip3(allButLast(prefixes l),
	l,
	tail(reverse(suffixes l)))

fun cross l =
   case destruct l of
      NONE => cons(empty(), empty())
    | SOME(x, l) =>
	 let val rest = cross l
	 in flattenMap(x, fn x => map(rest, fn t => cons(x, t)))
	 end
   
fun listTo [] = empty()
  | listTo(x :: l) = cons(x, listTo l)

fun 'a ordered{< : 'a * 'a -> bool} =
   let
      fun x > y = y < x

      fun insert(s, x) =
	 case destruct s of
	    NONE => single x
	  | SOME(y, s') =>
	       if x < y then cons(x, s)
	       else cons(y, insert(s', x))

      fun insertionSort s =
	 case destruct s of
	    NONE => empty()
	  | SOME(x, s) => insert(insertionSort s, x)

      fun partition (ns, x) =
	 fold(ns, (empty(), empty()),
	      fn (y, (left, right)) =>
	      if x < y then (left, cons(y, right))
	      else (cons(y, left), right))
    
      local  
	 val columnSize = fromInt 5
	 val sort = insertionSort
	 fun breakIntoColumns ns =
	    if Int.<(length ns, columnSize)
	       then cons(ns, empty())
	    else cons(firstN(ns, columnSize),
		      breakIntoColumns(nthRest(ns, columnSize)))
      in
	 fun median ns = orderStatistic(ns, Int.quot(length ns, two))
	 and orderStatistic(ns, i) =
	    if Int.<(length ns, columnSize)
	       then nth(sort ns, i)
	    else let
		    val medianOfMedians =
		       median(map(breakIntoColumns ns, columnMedian))
		    val (left, right) = partition(ns, medianOfMedians)
		    val numLeft = length left
		 in if Int.<(i,numLeft)
		       then orderStatistic(left, i)
		    else orderStatistic(right, i - numLeft)
		 end
	 and columnMedian ns =
	    nth(sort ns, Int.quot(length ns, two))
      end    

      fun choose (op <) (s,n) =
	 let fun insert(x, s) =
	    let fun insert(m, s) =
	       if Int.equals(m,n) then empty()
	       else case destruct s of
		  NONE => single x
		| SOME(y, s) => if x < y then cons(x, s)
			       else cons(y, insert(add1 m, s))
	    in insert(zero, s)
	    end
	 in firstN(fold(s, empty(), insert),n)
	 end
      
      val smallest = choose (op <)
      val largest = choose (op >)
    
      fun getFirst(l, extreme,name) =
	 case destruct(extreme(l, one)) of
	    SOME(x, _) => x
	  | _ => Error.error name

      fun min l = getFirst(l, smallest, "min")
      fun max l = getFirst(l, largest, "max")

   in {insert = insert,
       insertionSort = insertionSort,
       median = median,
       orderStatistic = orderStatistic,
       partition = partition,
       max = max,
       min = min,
       largest = largest,
       smallest = smallest}
   end

fun equivalence(xs, equal) =
   fold
   (xs, empty(), fn (x, xss) =>
    let
       fun insert l =
	  case destruct l of
	     NONE => single(single x)
	   | SOME(xs, xss) =>
		case destruct xs of
		   SOME(x', _) => if equal(x, x')
				    then cons(cons(x, xs), xss)
				 else cons(xs, insert xss)
		 | NONE => Error.bug "List.equivalence"
    in insert xss
    end)

(* subsets of exactly size n *)
fun subsets(s,n) =
   let
      fun subs(s, size, n, elts, accum) =
	 if Int.isZero n
	    then cons(elts, accum)
	 else case destruct s of
	    SOME(x, xs) =>
	       if Int.equals(size,n) then cons(cons(x, append(xs, elts)),
					  accum)
	       else subs(xs, Int.sub1 size, Int.sub1 n,
			 cons(x, elts),
			 subs(xs, Int.sub1 size, n, elts , accum))
	  | _ => Error.error "subsets"
   in subs(s, size s, n, empty(), empty())
   end

fun power s = subsets(s, size s)
   
fun 'a set{equal: 'a * 'a -> bool,
	   layout: 'a -> Layout.t} =
   let
      fun equalTo x x' = equal(x, x')
      fun contains(s, x) = exists(s, equalTo x)
      fun s <= s' = forall(s, fn x => contains(s', x))
      fun s' >= s = s <= s'
      val equals = fn (s, s') => s <= s' andalso s' <= s
      fun s < s' = s <= s' andalso exists(s', fn x => not(contains(s, x)))
      fun s' > s = s < s'
      fun singleton x = cons(x, empty())
      fun add(s, x) = if contains(s, x) then s else cons(x, s)
      fun areDisjoint(s, s') = forall(s, fn x => not(contains(s', x)))
      val subset = keepAll
      fun subsetSize(s, p) = fold(s, 0, fn (x,n) => if p x then n + 1 else n)
      fun s - s' = subset(s, fn x => not(contains(s', x)))
      fun s + s' = append(s - s', s')
      fun intersect(s, s') = subset(s, fn x => contains(s', x))
      fun unions ss = fold(ss, empty(), op +)
      val size = length
	 
      val layout = fn vs =>
	 let open Layout
	 in seq[str "{",
		align(separateRight(toList(map(vs, layout)), ", ")),
		str "}"]
	 end

      val remove = fn (s, x) => remove(s, equalTo x)
   
      fun replace(s, f) =
	 fold(s, empty(), fn (x, s) => (case f x of
					NONE => s
				      | SOME y => add(s, y)))
      fun map(s, f) = replace(s, fn x => SOME(f x))

   in {empty = empty(),
       singleton = singleton,
       size = size,
       equals = equals,
       <= = op <=,
       >= = op >=,
       < = op <,
       > = op >,
       + = op +,
       - = op -,
       intersect = intersect,
       unions = unions,
       add = add,
       remove = remove,
       contains = contains,
       power = power,
       subsets = subsets,
       areDisjoint = areDisjoint,
       subset = subset,
       subsetSize = subsetSize,
       map = map,
       replace = replace,
       layout = layout}
   end
      
fun toList l = fold(l, [], op ::)

fun compare(l, l', comp) =
   let
      fun compare(l, l') =
	 case (destruct l, destruct l')  of
	      (NONE,       NONE)         => EQUAL
	    | (_,          NONE)         => GREATER
	    | (NONE,       _)            => LESS
	    | (SOME(x, l), SOME(x', l')) => (case comp(x, x') of
						EQUAL => compare(l, l')
					      | r => r)
   in compare(l, l')
   end
	 
end
functor StrictList(structure List: BASIC_LIST_ITER): STRICT_LIST =
   let structure L' = List(structure I = Int
                           structure L = List)
   in struct open List L' end
   end
(*-------------------------------------------------------------------*)
(*                             ListIter                              *)
(*-------------------------------------------------------------------*)

functor ListIter(L: BASIC_LIST): BASIC_LIST_ITER = 
struct

open L

datatype ('a, 'b) iter =
   Empty
 | First of 'a * 'b
 | Rest of 'a t
   
fun iterate(b, f) =
   let
      fun iter b =
	 case f b of
	    Empty => empty()
	  | First(x, b) => cons(x, iter b)
	  | Rest s => s
   in iter b
   end

end

signature BASIC_STRICT_LIST =
   sig
      include BASIC_LIST
   end
(*-------------------------------------------------------------------*)
(*                          BasicStrictList                          *)
(*-------------------------------------------------------------------*)

functor BasicStrictList(): BASIC_STRICT_LIST =
struct
      
type 'a t = 'a list
   
fun empty() = []
      
fun cons(x, l) = x :: l
      
fun isEmpty [] = true
  | isEmpty _ = false
    
fun destruct [] = NONE
  | destruct(x :: xs) = SOME(x, xs)
   
end
(*-------------------------------------------------------------------*)
(*                            StrictList                             *)
(*-------------------------------------------------------------------*)

structure StrictList: STRICT_LIST =
   StrictList(structure List = ListIter(BasicStrictList()))

structure StrictList: sig
			 include STRICT_LIST
			 val revMap: 'a t * ('a -> 'b) -> 'b t
		      end =
struct

open StrictList

val isEmpty =
   fn [] => true
    | _ => false

fun fold(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    [] => b
	  | x :: l => loop(l, f(x, b))
   in loop(l, b)
   end

fun fold2(l1, l2, b, f) =
   let
      fun loop(l1, l2, b) =
	 case (l1,       l2)       of
  	      ([],       [])       => b
	    | (x1 :: l1, x2 :: l2) => loop(l1, l2, f(x1, x2, b))
	    | _                    => Error.bug "fold2"
   in loop(l1, l2, b)
   end

fun foldi(l, b, f) = #1(fold(l, (b, 0), fn (x, (b, i)) => (f(i, x, b), i + 1)))

fun foreachi(l, f) = foldi(l, (), fn (i, x, ()) => f(i, x))
							 
fun foreach(l, f) = fold(l, (), f o #1)

fun foreach2(l1, l2, f) = fold2(l1, l2, (), fn (x1, x2, ()) => f(x1, x2))

fun reverseAppend(l1, l2) = fold(l1, l2, op ::)

fun reverse l = reverseAppend(l, [])

fun foldr(l, b, f) = fold(reverse l, b, f)

fun foldr2(l1, l2, b, f) = fold2(reverse l1, reverse l2, b, f)

fun append(l1, l2) = foldr(l1, l2, op ::)

(* careful to apply the function left-to-right *)
fun revMap(l, f) = fold(l, [], fn (x, l) => f x :: l)
fun map(l, f) = reverse(revMap(l, f))

fun map2(l1, l2, f) = reverse(fold2(l1, l2, [], fn (x1, x2, l) => f(x1, x2) :: l))

fun forall(l, f) =
   let
      val rec loop =
	 fn [] => true
	  | x :: l => f x andalso loop l
   in loop l
   end

fun exists(l, f) = not(forall(l, not o f))

local
   fun make(l1, l2, f, unequal) =
      let
	 val rec loop =
	    fn ([],     [])         => true
	     | (x1 :: l1, x2 :: l2) => f(x1, x2) andalso loop(l1, l2)
	     | _                    => unequal()
      in loop(l1, l2)
      end
in
   fun forall2(l1, l2, f) = make(l1, l2, f, fn () => Error.bug "forall2")
   fun equals(l1, l2, f) = make(l1, l2, f, fn () => false)
end

val rec nth =
   fn (x :: _, 0) => x
    | (_ :: l, n) => nth(l, n - 1)
    | _ => Error.bug "nth"

fun empty() = []

val cons = op ::

fun compare(l, l', comp) =
   let
      val rec compare =
	 fn ([],     [])       => EQUAL
	  | (_,      [])       => GREATER
	  | ([],     _)        => LESS
	  | (x :: l, x' :: l') => (case comp(x, x') of
				      EQUAL => compare(l, l')
				    | r => r)
   in compare(l, l')
   end

fun allButLast l =
   case reverse l of
      _ :: l => reverse l
    | _ => Error.bug "allButLast"

fun last l =
   case l of
      [x] => x
    | _ :: l => last l
    | _ => Error.bug "last"

fun length l = fold(l, 0, fn (_, n) => n + 1)

fun zip(l1, l2) = foldr2(l1, l2, [], fn (x1, x2, l) => (x1, x2) :: l)

fun unzip l = foldr(l, ([], []), fn ((x1, x2), (l1, l2)) =>
		    (x1 :: l1, x2 :: l2))

fun peek(l, f) =
   let
      val rec loop =
	 fn [] => NONE
	  | x :: l => if f x then SOME x else loop l
   in loop l
   end

fun flatten l = foldr(l, [], append)

fun duplicate(n, f) =
   let
      fun loop(n, accum) =
	 if n = ~1
	    then accum
	 else loop(n - 1, f() :: accum)
   in loop(n - 1, [])
   end
   
(* TODO
equivalence
flattenMap
fromTo
index
keepAll
keepAllMap
keepFirst
layout
mapAppend
mapFromTo
mapi
ordered
remove
removeAll
separate
set
single
splitLast
tabulate
*)

end

structure List = StrictList
signature ORDER =
   sig
      include ORDER0
      val layout: t -> Layout.t
   end
signature TIME = 
   sig
      type int = Int32.int

      include ORDER

      val zero: t
      val now: unit -> t
      type times =
	  {
            elapsed: t,  (* elapsed system time *)
            utime  : t,  (* user time of process *)
            stime  : t,  (* system time of process *)
            cutime : t,  (* user time of terminated child processes *)
            cstime : t   (* system time of terminated child processes *)
          }
      val times: unit -> times
      val fromReal: real -> t
      val toReal: t -> real
      val toSeconds: t -> int
      val fromSeconds: int -> t
      val toMilliseconds: t -> int
      val fromMilliseconds: int -> t
      val toMicroseconds: t -> int
      val fromMicroseconds: int -> t
      val + : t * t -> t
      val - : t * t -> t

      val toString: t -> string
(*      val fromString: string -> t option *)

      val output: t * Out.t -> unit
   end
functor Time(): TIME = 
struct

open Pervasive.Time

type int = Int32.int
type t = time

type times =
   {
    elapsed: t,  (* elapsed system time *)
    utime  : t,  (* user time of process *)
    stime  : t,  (* system time of process *)
    cutime : t,  (* user time of terminated child processes *)
    cstime : t   (* system time of terminated child processes *)
    }

val times = Posix.ProcEnv.times
   
val zero = fromReal 0.0
   
val equals = op =

val {min, max, ...} = Relation.compare compare

val layout = Layout.str o toString

fun output(t, out) = Out.output(out, toString t)

end

structure Time = Time()
signature PROCESS =
   sig
      val system: string -> unit
   end
structure Process: PROCESS =
struct

structure Process = OS.Process

fun system str =
   case Process.system str of
      0 => ()
    | _ => Error.error("call to system failed: " ^ str)
   
end
signature FILE =
   sig
      type t
      type dir

      val appendSuffix: t * string -> t
      val canRead: t -> bool
      val canWrite: t -> bool
      val concat: t list * t -> unit
      val contents: t -> string
      val copy: t * t -> unit
      val create: t -> unit (* make an empty file *)
      val dir: t -> dir
      val ensureRead: t -> unit
      val ensureWrite: t -> unit
      val fromString: string -> t
      val ignore : Out.t * (unit -> 'a) -> 'a
      val isNewer: t * t -> bool
      val layout: t -> Layout.t
      val lines: t -> string list
      val modTime: t -> Time.t
      val move: {from: t, to: t} -> unit
      val name: t -> string
      val output: t * Out.t -> unit
      val outputContents: t * Out.t -> unit
      val relative: dir * string -> t
      val remove: t -> unit
      val sameContents: t * t -> bool
      val size: t -> int
      val stringify : (t -> 'a) -> (string -> 'a)
      val temp: unit -> t
      val toString: t -> string
      val withAppend: t * (Out.t -> 'a) -> 'a
      val withIn: t * (In.t -> 'a) -> 'a
      val withOut: t * (Out.t -> 'a) -> 'a
      val withOutIn: (Out.t -> 'a) * (In.t -> 'b) -> 'a * 'b
      val withString : string * (t -> 'a) -> 'a
      val withStringIn : string * (In.t -> 'a) -> 'a
      val withTemp: (t -> 'a) -> 'a
   end
signature DIR =
   sig
      type t
      type file
	 
      val root: t

      val temp: unit -> t
      val inTemp: (unit -> 'a) -> 'a
	 
      val current: unit -> t
      val change: t -> unit
      val parent: t -> t
      val inDir: t * (unit -> 'a) -> 'a
      val remove: t -> unit
      val removeR: t -> unit (* remove recursively, i.e. all contents *)
      val make: t -> unit
	 
      val fromString: string -> t
      val relative: t * string -> t
      val toString: t -> string
	 
      val list: t -> t list * file list
      val dirs: t -> t list
      val files: t -> file list
	 
      (* descends subdirectories *)
      val foreachFile: t * (file -> unit) -> unit

      val layout: t -> Layout.t
   end
functor FileSys(val tempPrefix: string) =
struct

structure OS = OS
structure FS = OS.FileSys
structure P = OS.Path
   
structure Directory =
   struct
      datatype t = T of string list

      fun arcs(T a) = a

      val root = T[]

      fun parent(T ss) =
	 case ss of
	    [] => Error.error "parent of root directory"
	  | _ => T(List.allButLast ss)
	 
      fun split s = List.removeAll(#arcs(P.fromString s),
				   fn s => String.equals(s, ""))

      fun current() = T(split(FS.getDir()))

      fun dirOf s =
	 if String.size s > 0 andalso Char.equals(#"/", String.sub(s, 0))
	    then root
	 else current()

      fun relative(T d, s) = T(List.append(d, split s))

      fun fromString s = relative(dirOf s, s)

      fun toString(T d) =
	 P.toString{arcs = d, isAbs = true, vol=""} ^ "/"

      val layout = Layout.str o toString

      local fun convert f = f o toString
      in val remove = convert FS.rmDir
	 val change = convert FS.chDir
	 val make = convert FS.mkDir
      end

      fun inDir(d, th) =
	 let val cur = current()
	 in change d ;
	    DynamicWind.dynamicWind(th, fn () => change cur)
	 end
   end
structure D = Directory
	 
structure File =
   struct
      datatype t = T of {dir: D.t, file: string}
      type dir = D.t

      fun appendSuffix(T{dir, file}, suf) = T{dir = dir, file = file ^ suf}
	 
      fun dir(T{dir, ...}) = dir
      fun name(T{file, ...}) = file
	 
      fun relative(d, s) =
	 let val {dir, file} = P.splitDirFile s
	 in T{dir = D.relative(d, dir), file = file}
	 end

      fun fromString s = relative(D.dirOf s, s)

      fun toString(T{dir, file}) = D.toString dir ^ file

      local fun unary f = f o toString
      in val layout = unary Layout.str
	 val size = unary FS.fileSize
	 val modTime = unary FS.modTime
      end

      fun isNewer(f1, f2) = Time.>=(modTime f1, modTime f2)

      fun withh(file, p, openn, close) =
	 let val f = toString file
	    val stream =
	       (openn f) handle IO.Io _ => Error.error("cannot open " ^ f)
	 in DynamicWind.dynamicWind(fn () => p stream,
				    fn () => close stream)
	 end 
      
      fun withOut(f, p) = withh(f, p, Out.openOut, Out.close)
      fun withAppend(f, p) = withh(f, p, Out.openAppend, Out.close)
      fun withIn(f, p) = withh(f, p, In.openIn, In.close)

      local
	 fun can withh f = withh(f, fn _ => true) handle _ => false
         fun ensure(pred, msg) f =
	    if pred f then ()
	    else Error.error("can not " ^ msg ^ " " ^ toString f)
      in
	 val canWrite = can withOut
	 val canRead = can withIn
	 val ensureWrite = ensure(canWrite, "write")
	 val ensureRead = ensure(canRead, "read")
      end

(*	 let val f = toString f
	 in if FS.access(f, [FS.A_WRITE])
	       then ()
	    else Error.error("can not write " ^ f)
	 end*)
      (*fun canRead f =
       (In.closeIn(In.openIn f) ; true)
       handle IO.Io _ => false
	  *)

      fun sameContents(f1, f2) =
	 size f1 = size f2
	 andalso withIn(f1, fn in1 =>
			withIn(f2, fn in2 =>
			       In.sameContents(in1, in2)))

      fun output(file, out) = Out.output(out, toString file)
	 
      fun outputContents(file, out) =
	 withIn(file, fn ins =>
		while not(In.endOf ins)
		   do Out.output(out, In.input ins))

      fun lines f =
	 withIn(f, fn ins =>
		let
		   fun loop() =
		      if In.endOf ins then []
		      else (String.dropLast(In.inputLine ins)) :: (loop())
		in loop()
		end)
   
      fun contents file = withIn(file, In.inputAll)

      fun move{from, to} =
	 Process.system("mv " ^ toString from ^ " " ^ toString to)
	 
      fun copy(source, dest) =
	 withOut(dest, fn out => outputContents(source, out))
	 
      fun concat(sources, dest) =
	 withOut(dest, fn out =>
		 List.foreach(sources, fn f => outputContents(f, out)))
	 
      val remove = FS.remove o toString

      val temp = fromString o OS.FileSys.tmpName

      fun withTemp p =
	 let val f = temp()
	 in DynamicWind.dynamicWind
	    (fn () => p f,
	     fn () => remove f)
	 end

      fun stringify processFile s =
	 withTemp(fn f =>
		  (withOut(f, fn out => Out.output(out,s)) ;
		   processFile f))

      fun withString(s, f) =
	 withTemp(fn tmp =>
		  (withOut(tmp, fn out => Out.output(out,s)) ;
		   f tmp))
		  
      fun withOutIn(fout, fin) =
	 withTemp(fn tmp =>
		  (withOut(tmp, fout),
		   withIn(tmp, fin)))

      fun withStringIn(s, fin) =
	 #2(withOutIn(fn out => Out.output(out, s),
		      fin))

      fun ignore(out, thunk) =
	 withOut(fromString "/dev/null", fn out' =>
		 Out.fluidLet(out, out', thunk))

      fun create f = withOut(f, fn _ => ())
   end
structure F = File

structure Directory =
   struct
      open D
      type file = File.t

      fun list d =
	 let val dstr = toString d
	    val stream = FS.openDir dstr
	    val _ = FS.readDir stream
	    val _ = FS.readDir stream
	    fun read() =
	       let val s = FS.readDir stream
	       in if s = "" then ([], [])
		  else let val (dirs, files) = read()
			   val x = dstr ^ s
		       in if FS.isLink x then (dirs, files)
			  else
			     if FS.isDir x
				then (relative(d, s) :: dirs, files)
			     else (dirs, F.relative(d, s) :: files)
		       end
	       end
	 in DynamicWind.dynamicWind(read, fn () => FS.closeDir stream)
	 end
      
      fun dirs d = #1(list d)
	 
      fun files d = #2(list d)
	 
      fun foreachFile(d, p) =
	 let
	    fun foreach d =
	       let val (ds, fs) = list d
	       in (List.foreach(fs, p) ;
		   List.foreach(ds, foreach))
	       end
	 in foreach d
	 end

      fun removeR d =
	 let val (ds, fs) = list d
	 in List.foreach(ds, removeR) ;
	    List.foreach(fs, File.remove) ;
	    remove d
	 end

      fun temp() = fromString(FS.tmpName())
	 
      fun inTemp thunk =
	 let val d = temp()
	 in make d ;
	    DynamicWind.dynamicWind
	    (fn () => inDir(d, fn _ => thunk()),
	     fn () => removeR d)
	 end
   end

type file = F.t
type dir = D.t
   
end

structure FileSys = FileSys(val tempPrefix = "/tmp/smltmp")
structure File: FILE = FileSys.File
structure Dir: DIR = FileSys.Directory
signature MY_DIRS =
   sig
      val dirs: unit -> {home: Dir.t,
			  sml: Dir.t,
			  smlnj: Dir.t,
			  bin: Dir.t,
			  binFiles: Dir.t,
			  heap: Dir.t,
			  src: Dir.t,
			  compiler: Dir.t}
      val exportFn: string * (string * string list -> int) -> unit
      val exportML: string -> bool
   end
structure MyDirs: MY_DIRS =
   struct
      open Dir
	 
      structure Host =
	 struct
	    datatype t = Eponym | Samson | Snake
	       
	    fun current() = 
	       case OS.Process.getEnv "HOST" of
		  NONE => Error.error "can't figure out host"
		| SOME s =>
		     case s of
			"eponym.nj.nec.com" => Eponym
		      | "samson.nj.nec.com" => Samson
		      | "snake" => Snake
		      | "snake.nj.nec.com" => Snake
		      | _ => Error.error "unrecognized host"
	 end
      open Host
	 
      fun dirs() =
	 let
	    val home =
	       fromString(case current() of
			     Eponym => "/home/sweeks/"
			   | Snake => "/home/sweeks/"
			   | Samson => "/usr/tmp/sweeks/")
	    val sml = relative(home, "sml")
	    val smlnj = relative(home, "sml/sml-nj-root")
	    val bin = relative(smlnj, "bin")
	    val binFiles = relative(smlnj, "bin.x86-unix")
	    val heap = relative(bin, ".heap")
	    val src = relative(smlnj, "src")
	    val compiler = relative(src, "sml-nj")
	 in {home = home, sml = sml, smlnj = smlnj, bin = bin,
	     binFiles = binFiles, heap = heap, src = src, compiler = compiler}
	 end
      
      fun exportFn(name, f) =
	 SMLofNJ.exportFn(File.toString(File.relative(#heap(dirs()),name)),
			  f)
      fun exportML name =
	 SMLofNJ.exportML(File.toString(File.relative(#heap(dirs()),name)))
   end
signature REF =
   sig
      type 'a t

      val ! : 'a t -> 'a
      val := : 'a t * 'a -> unit

      val equals: 'a t * 'a t -> bool

      val fluidLet: 'a t * 'a * (unit -> 'b) -> 'b

      val swap: 'a t * 'a t -> unit
      val getAndSet: ('a -> 'b ref) -> ('a -> 'b) * ('a * 'b -> unit)
      val getSet: ('a -> Layout.t) -> {get: unit -> 'a,
					set: 'a -> unit,
					clear: unit -> unit,
					layout: unit -> Layout.t,
					output: Out.t -> unit,
					print: unit -> unit}
   end
(*-------------------------------------------------------------------*)
(*                                Ref                                *)
(*-------------------------------------------------------------------*)
   
structure Ref: REF =
struct

type 'a t = 'a ref

val (op !) = (op !)

val (op :=) = op :=

fun equals(r: 'a t, r') = r = r'
   
fun swap(r, r') = let val v = !r
		  in r := !r'; r' := v
		  end

fun getAndSet sel = (! o sel, fn (x, v) => sel x := v)

fun ('a, 'b) fluidLet(r: 'a t, x: 'a, th: unit -> 'b): 'b =
   let val old = !r
   in r := x
      ; DynamicWind.dynamicWind(th, fn () => r := old)
   end

fun getSet layout = 
   let val r = ref NONE
      fun get() =
	 case !r of
	    NONE => Error.error "not available"
	  | SOME v => v
      fun set v = r := SOME v
      fun clear() = r := NONE
      val layout = fn () => layout(get())
      fun output out = Layout.output(layout(), out)
      fun print() = output Out.standard
   in {get = get,
       set = set,
       clear = clear,
       layout = layout,
       output = output,
       print = print}
   end


end

signature LIST_REF =
   sig
      structure L: LIST
	 
      type 'a t = 'a L.t Ref.t
	 
      val pop: 'a t -> 'a
      val top: 'a t -> 'a
      val push: 'a t * 'a -> unit
   end
(*-------------------------------------------------------------------*)
(*                              ListRef                              *)
(*-------------------------------------------------------------------*)

functor ListRef(L: LIST): LIST_REF =
struct

structure L = L
   
type 'a t = 'a L.t Ref.t
   
fun push(r, x) = r := L.cons(x, !r)

fun top(ref l) =
   case L.destruct l of
      SOME(x, _) => x
    | NONE => Error.error "top"
	 
fun pop(r as ref l) =
   case L.destruct l of
      SOME(x, l) => (r := l ; x)
    | NONE => Error.error "pop"

end

structure ListRef = ListRef(List)
signature UNIT = T where type t = unit
structure Unit: UNIT =
struct

type t = unit

val equals = fn ((), ()) => true

fun layout() = Layout.str"()"

end
signature STRING_MAP = 
   sig
      type 'a t

      val new: (unit -> 'a) -> 'a t
      val clear: 'a t -> unit
      val lookup: 'a t * string -> 'a
      val domain: 'a t -> string list
      val keepAll: 'a t * ('a -> bool) -> string list
      val foreach: 'a t * ('a -> unit) -> unit
   end
signature COMPUTATION = 
   sig
      structure Time: TIME
	 
      type t

      val keepAll: t * (string -> bool) -> t
      val inspect: t -> unit
      val output: t * Out.t -> unit
      val outputCalls: t * Out.t -> unit
      val outputTimes: t * Out.t -> unit
      val time: t -> Time.t
   end
signature INTERMEDIATE_COMPUTATION_STRUCTS =
   sig
      structure Time: TIME
   end

signature INTERMEDIATE_COMPUTATION = 
   sig
      include INTERMEDIATE_COMPUTATION_STRUCTS
      
      structure Computation: COMPUTATION
      
      type t

      val empty: unit -> t
      val call: t * string * (unit -> Layout.t) -> unit
      val raisee: t * Time.t option -> unit
      val return: t * (unit -> Layout.t) * Time.t option -> unit
      val finish: t -> Computation.t
      val atTopLevel: t -> bool
   end
signature TRACE_CONTROL = 
   sig
      (* controls what tracing info is gathered *)
      val always: unit -> unit      (* always gather info *)
      val flagged: unit -> unit     (* only on flagged functions *)
      val never: unit -> unit       (* never gather info *)
	 
      (* value of newly created flag *)
      val default: bool ref

      (* turn on/off tracing for given functions *)
      val on: string list -> unit
      val off: string list -> unit

      (* turn on/off tracing for all functions *)
      val all: unit -> unit
      val none: unit -> unit

      (* returns functions that are on/off *)
      val whatsOn: unit -> string list
      val whatsOff: unit -> string list
   end

signature TRACE =
   sig
      structure Immediate :
	 sig
	    include TRACE_CONTROL

	    val message: Layout.t -> unit
	    val messageStr: string -> unit
	 end
      structure Delayed :
	 sig
	    include TRACE_CONTROL
	    val keepAll: bool ref
	 end
      structure Time: TRACE_CONTROL

      val never: unit -> unit
      val always: unit -> unit
      val flagged: unit -> unit
      val traceable: unit -> string list
      val outputTraceable: unit -> unit
      val reset: unit -> unit
	 
      (*---------- Delayed Feedback ----------*)
      structure Computation: COMPUTATION

      (* clear computation history *)
      val clear: unit -> unit

      (* get computation history *)
      val computation: unit -> Computation.t

      (* show computation history *)
      val history: unit -> unit

      (* show computation history, without arguments or results *)
      val calls: unit -> unit

      val times: unit -> unit

      (* enter the inspector *)
      val inspect: unit -> unit

      (*---------- Instrumentation ----------*)

      val traceCall: string -> ('a -> 'b) -> 'a -> 'b
	 
      val trace: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 -> ('a -> 'b)
	 -> ('a -> 'b)

      val trace0: string * ('a -> Layout.t) -> (unit -> 'a) -> (unit -> 'a)

      val trace2: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 -> ('a * 'b -> 'c)
	 -> ('a * 'b -> 'c)

      val trace3: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 * ('d -> Layout.t)
	 -> ('a * 'b * 'c -> 'd)
	 -> ('a * 'b * 'c -> 'd)

      val trace4: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 * ('d -> Layout.t)
	 * ('e -> Layout.t)
	 -> ('a * 'b * 'c * 'd -> 'e)
	 -> ('a * 'b * 'c * 'd -> 'e)

      val traceRec :string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 -> (('a -> 'b) -> ('a -> 'b))
	 -> 'a -> 'b 
   end
functor Control(structure StringMap: STRING_MAP
                datatype status = Always | Flagged | Never
		type flags
		val map: flags StringMap.t
		val getFlag: flags -> bool ref
		val default: bool ref
		val status: status) =
struct

val status = ref status

fun always() = status := Always
fun flagged() = status := Flagged
fun never() = status := Never

fun isOn b =
   case !status of
      Always => true
    | Flagged => !b
    | Never => false
   
val default = default

fun flag name = getFlag(StringMap.lookup(map,name))

fun whats b () =
   StringMap.keepAll(map, fn flags => b = !(getFlag flags))

val whatsOn = whats true
val whatsOff = whats false

fun sets b () =
   StringMap.foreach(map, fn flags => getFlag flags := b)
   
val all = sets true
val none = sets false

fun some b ss = List.foreach(ss, fn s => flag s := b)

val on = some true
val off = some false

end

(*-------------------------------------------------------------------*)
(*                               Trace                               *)
(*-------------------------------------------------------------------*)

functor Trace(structure IC: INTERMEDIATE_COMPUTATION
	      structure StringMap: STRING_MAP
	      structure Time: TIME
	      sharing type IC.Time.t = Time.t): TRACE =
struct

structure Timer = Time

val immediateDefault = ref false
val delayedDefault = ref false
val timeDefault = ref false

type flags = {immediate: bool ref,
	      delayed: bool ref,
	      time: bool ref}
   
val map = StringMap.new(fn () => {immediate = ref(!immediateDefault),
				  delayed = ref(!delayedDefault),
				  time = ref(!timeDefault)})
   
fun traceable() = StringMap.domain map

fun outputTraceable() =
   Layout.output(List.layout String.layout (traceable()),
		 Out.standard)

datatype status = Always | Flagged | Never

structure Immediate =
   Control(type flags = flags
	   val map = map
	   structure StringMap = StringMap
           datatype status = datatype status
	   fun getFlag({immediate, ...}: flags) = immediate
	   val default = immediateDefault
	   val status = Never)
structure Delayed =
   struct
      structure C = 
	 Control(type flags = flags
		 val map = map
		 structure StringMap = StringMap
		 datatype status = datatype status
		 fun getFlag({delayed, ...}: flags) = delayed
		 val default = delayedDefault
		 val status = Never)
      open C
	 
      val keepAll = ref true
   end

structure Time =
   Control(type flags = flags
	   val map = map
	   structure StringMap = StringMap
	   datatype status = datatype status
	   fun getFlag({time, ...}: flags) = time
	   val default = timeDefault
	   val status = Never)

fun never() = (Immediate.never() ;
	       Delayed.never() ;
	       Time.never())

fun always() = (Immediate.always() ;
		Delayed.always() ;
		Time.always())

fun flagged() = (Immediate.flagged() ;
		 Delayed.flagged() ;
		 Time.flagged())
   
fun reset() =
   StringMap.foreach(map, fn {immediate, delayed, time} =>
		     (immediate := false ;
		      delayed := false ;
		      time := false))
    (*StringMap.clear map*)

(*---------------------------------------------------*)
(*                 Delayed Feedback                  *)
(*---------------------------------------------------*)

structure Computation = IC.Computation

datatype comp =
   Working of IC.t
  | Finished of Computation.t

val emptyIc = IC.empty
fun empty() = Working(emptyIc())
   
val currentComputation = ref(empty())

fun clear() = currentComputation := empty()

fun finishedComputation() =
   case !currentComputation of
      Working ic => let val c = IC.finish ic
		    in currentComputation := Finished c ;
		       c
		    end
    | Finished c => c
val computation = finishedComputation

fun history() = Computation.output(finishedComputation(),
				   Out.error)
fun calls() = Computation.outputCalls(finishedComputation(),
				      Out.error)
fun times() = Computation.outputTimes(finishedComputation(),
				      Out.error)
fun inspect() = Computation.inspect(finishedComputation())

fun ic() =
   case !currentComputation of
      Finished _ => let val ic = emptyIc()
		    in currentComputation := Working ic ;
		       ic
		    end
    | Working ic => ic

fun delayedCall(name, layoutArg, layoutAns) =
   {call = fn () =>
    let val comp = ic()
       val comp = if !Delayed.keepAll orelse not(IC.atTopLevel comp)
		     then comp
		  else (clear() ; ic())
    in IC.call(comp, name, layoutArg)
    end,
    raisee = fn t => IC.raisee(ic(), t),
    return = fn (ans, t) => IC.return(ic(),
				     fn () => layoutAns ans,
				     t)}
	 
(*---------------------------------------------------*)
(*                Immediate Feedback                 *)
(*---------------------------------------------------*)

local
   val out = Out.error
   open Layout
      
   val indentation = ref 0
   val space = 3
   fun left() = indentation := !indentation - space
   fun right() = indentation := !indentation + space
   fun print l = (Layout.output(Layout.indent(l, !indentation), out) ;
		  Out.newline out)

   fun finish(t, res) =
      (left() ;
       print(case t of
		NONE => seq[str "==> ", res]
	      | SOME t =>
		 align[seq[str "==> time = ", Timer.layout t],
		       res]))
in
   fun immediateCall(name, outArg, layoutAns) =
      let fun call() = (print(seq[str name, str " ==> ",
				  outArg()
				  handle _ => str "<layout argument error>"]) ;
			right())
	 fun raisee t = finish(t, str "raise")
	 fun return(ans, t) =
	    finish(t, layoutAns ans handle _ => str "<layout answer error>") ;
      in {call = call,
	  raisee = raisee,
	  return = return}
      end

   structure Immediate =
      struct
	 open Immediate
	 fun message l =
	    (left() ;
	     indentation := 1 + !indentation; 
	     print l;
	     indentation := !indentation - 1;
	     right())
	 val messageStr = message o Layout.str
      end
end
	 
(*---------------------------------------------------*)
(*                  Instrumentation                  *)
(*---------------------------------------------------*)

val doTrace = true
   
fun trace(name, layoutArg, layoutAns) =
   if not doTrace
      then fn f => f
   else 
      let val {immediate, delayed, time} = StringMap.lookup(map,name)
      in fn f => fn arg =>
	 let
	    val immediate = Immediate.isOn immediate
	    val delayed = Delayed.isOn delayed
	    val time = Time.isOn time
	 in if not(immediate orelse delayed orelse time)
	       then f arg
	    else let val outArg = fn () => layoutArg arg
		     val noCall = {call = fn _ => (),
				   raisee = fn _ => (),
				   return = fn _ => ()}
		     val immed = if immediate
				    then immediateCall(name, outArg, layoutAns)
				 else noCall
		     val delay = if delayed
				    then delayedCall(name, outArg, layoutAns)
				 else noCall
		 in #call delay () ;
		    #call immed () ;
		    let
		       val startTime =
			  if time then SOME(Timer.times()) else NONE
		       fun getTime() =
			  case startTime of
			     NONE => NONE
			   | SOME t => SOME(let val {utime, stime, ...} = Timer.times()
					    in Timer.+(Timer.-(utime, #utime t),
						       Timer.-(stime, #stime t))
					    end)
		       val ans = f arg handle exn => let val t = getTime()
						     in #raisee delay t ;
							#raisee immed t ;
							raise exn
						     end
		       val t = getTime()
		    in #return delay (ans, t) ;
		       #return immed (ans, t) ;
		       ans
		    end
		 end
	 end
      end

fun ignore _ = Layout.empty
   
fun traceCall s f = trace(s, ignore, ignore) f
   
fun traceRec info =
   let val trace = trace info
   in fn f => let fun fix f a = trace (f (fix f)) a
	      in fix f
	      end
   end

fun trace0(name, layoutAns) =
   trace(name, Unit.layout, layoutAns)
	 
fun trace2(name, layout1, layout2, layoutAns) =
   trace(name, Layout.tuple2(layout1, layout2), layoutAns)

fun trace3(name, out1, out2, out3, outAns) =
   trace(name, Layout.tuple3(out1, out2, out3), outAns)

fun trace4(name, out1, out2, out3, out4, outAns) =
   trace(name, Layout.tuple4(out1, out2, out3, out4), outAns)

end
functor StringMap(): STRING_MAP = 
struct

datatype 'a t = T of {map: {name: string, value: 'a} list ref,
		      default: unit -> 'a}

fun new default = T{map = ref [], default = default}

fun clear(T{map, ...}) = map := []
   
fun lookup(T{map, default},name) =
   case List.peek(!map, fn {name=name', ...} => name = name') of
      NONE => let val value = default()
	      in ListRef.push(map, {name = name, value = value}) ;
		 value
	      end
    | SOME{value, ...} => value

fun domain(T{map, ...}) = List.map(!map, fn {name, ...} => name)

fun keepAll(T{map, ...}, pred) = 
   List.keepAllMap(!map, fn {name, value} =>
		   if pred value then SOME name else NONE)

fun foreach(T{map, ...}, f) =
   List.foreach(!map, fn {value, ...} => f value)

end
functor IntermediateComputation(S: INTERMEDIATE_COMPUTATION_STRUCTS) :
   INTERMEDIATE_COMPUTATION = 
struct

open S

structure Result =
   struct
      datatype t =
	 Raise
       | Return of unit -> Layout.t

      local open Layout
      in
	 fun layout r =
	    case r of
	       Raise => str "raise"
	     | Return f => f()
      end
   end

structure Computation =
   struct
      structure Time = Time
	 
      datatype t = T of callResult list
      withtype callResult = {name: string,
			     layoutArg: unit -> Layout.t,
			     time: Time.t option,
			     body: t,
			     result: Result.t}

      structure CR =
	 struct
	    open Layout

	    type t = callResult

	    fun layoutName({name, ...}:t) = str(name ^ " ")
	       
	    fun layoutCall(cr as {layoutArg, ...}:t) =
	       seq[layoutName cr, layoutArg()]

	    val darrow = str "==> "
	       
	    fun layoutDarrow _ = darrow
	       
	    fun layoutResult({result, ...}:t) =
	       Result.layout result
	       
	    fun layoutTime({time, ...}:t) =
	       case time of
		  SOME t => Time.layout t
		| NONE => empty
		     
	    fun layoutReturn({result, ...}:t) =
	       seq[darrow, Result.layout result]
	 end

      fun time(T crs) = List.fold(crs, Time.zero, fn ({time, ...}, t) =>
				  case time of
				     NONE => t
				   | SOME t' => Time.+(t, t'))

      fun keepAll(c, pred) =
	 let
	    fun keepAll(T crs) =
	       T(List.foldr(crs, [],
			    fn ({name, body, layoutArg, time, result}, crs) =>
			    let val body as T crs' = keepAll body
			    in if pred name
				  then {name = name, body = body,
					layoutArg = layoutArg, time = time,
					result = result} :: crs
			       else crs' @ crs
			    end))
	 in keepAll c
	 end

      fun makeOutputs(pre, post, filter) out =
	 let val indentation = ref 0
	    val space = 3
	    fun left() = indentation := !indentation - space
	    fun right() = indentation := !indentation + space
	    fun print l = (Layout.output(Layout.indent(l, !indentation), out) ;
			   Out.newline out)
	    fun output(T crs) = List.foreach(crs, outputCr)
	    and outputCr(cr as {body, ...}) =
	       let val printCr = filter cr
	       in if printCr then (print(pre cr) ; right())
		  else () ;
		  output body ;
		  if printCr then (left() ; print(post cr)) else ()
	       end
	 in (output, outputCr)
	 end
      
      val makeOutput =
	 makeOutputs(CR.layoutCall, CR.layoutReturn, fn _ => true)
      fun output(c, out) = #1(makeOutput out) c
      fun outputCr(cr, out) = #2(makeOutput out) cr

      val makeOutputCalls =
	 makeOutputs(CR.layoutName, CR.layoutDarrow, fn _ => true)
      fun outputCalls(c, out) = #1(makeOutputCalls out) c
      fun outputCrCalls(cr, out) = #2(makeOutputCalls out) cr

      fun outputTimes(c, out) =
	 #1(makeOutputs(CR.layoutName, CR.layoutTime,
			fn {time, ...} =>
			case time of
			   NONE => false
			 | SOME _ => true) out) c

      local
	 val out = Out.error
	 val print = Out.outputc out

	 fun inputBetween(min, max) =
	   Int.inputBetween{ins = In.standard,
			    error = fn () => Out.output(out, "? "),
			    min = min, max = max}

	 fun choose (choices: (string * 'a) list): 'a =
	    let
	       val n =
		  List.fold(choices, 0, fn ((name, th),n) =>
			    (Layout.output(Int.layout n, out) ;
			     print ". " ; print name ; print "\n" ;
			     n+1))
	       val _ = Out.output(out, "? ")
	       val m = inputBetween(0,n-1)
	    in #2(List.nth(choices, m))
	    end

	 fun chooseThunk cs = choose cs ()

	 exception Quit
	 exception Back
	 
	 val standardChoices = [("quit", fn () => raise Quit),
				("back", fn () => raise Back)]
	 
	 fun inspect(c as T crs) =
	    case crs of
	       [cr] => inspectCr cr
	     | _ =>
		  let
		     fun loop() =
			(chooseThunk
			 (standardChoices
			  @ [("skip raises", fn () =>
			      (skipRaises c ; raise Back)),
			     ("output", fn () => output(c, out)),
			     ("output calls",
			      fn () => outputCalls(c, out))]
			  @ choices c) ;
			 loop())
		  in loop() handle Back => ()
		  end
	 and choices(c as T crs) =
	    List.map(crs, fn cr as {name, ...} =>
		     ("  " ^ name, fn () => inspectCr cr))
	 and inspectCr(cr as {name, layoutArg, body, result, time}) =
	    (print "Call to " ; print name ; print "\n" ;
	     let
		fun loop() =
		   (chooseThunk
		    (standardChoices
		     @ [("skip raises", fn () =>
			 (skipRaisesCr cr ; raise Back)),
			("output", fn () =>
			 outputCr(cr, out)),
			("output calls", fn () =>
			 outputCrCalls(cr, out)),
			("argument", fn () =>
			 (Layout.output(layoutArg(), out) ;
			  Out.newline out)),
			("result", fn () =>
			 (Layout.output(Result.layout result, out) ;
			  Out.newline out))]
		     @ choices body) ;
		    loop())
	     in loop() handle Back => ()
	     end)
	 and skipRaises(c as T crs) =
	    case crs of
	       [] => ()
	     | _ => (skipRaisesCr(List.last crs) ;
		     inspect c)
	 and skipRaisesCr(cr as {result, body, ...}) =
	    (case result of
		Result.Raise => skipRaises body
	      | Result.Return _ => () ;
             inspectCr cr)
      in val inspect = fn c => inspect c handle Quit => ()
      end      
   
   end

(*---------------------------------------------------*)
(*                   Main Datatype                   *)
(*---------------------------------------------------*)

datatype t =
   T of {calls: {name: string,
		  layoutArg: unit -> Layout.t,
		  prev: Computation.callResult list} list ref,
	 after: Computation.callResult list ref}

fun empty() = T{calls = ref [],
		after = ref []}

fun atTopLevel(T{calls, ...}) = List.isEmpty(!calls)
	    
fun call(T{calls, after},name, layoutArg) =
   (ListRef.push(calls, {name = name, layoutArg = layoutArg, prev = !after}) ;
    after := [])
   
fun return(T{calls, after}, result, time) =
   case !calls of
      [] => Error.error "return without a call"
    | {name, layoutArg, prev} :: cs =>
	 (calls := cs ;
	  after :=
	  {name = name, layoutArg = layoutArg,
	   result = result, time = time,
	   body = Computation.T(List.reverse(!after))} :: prev)

fun raisee(c, t) = return(c, Result.Raise, t)
val return = fn (c, r, t) => return(c, Result.Return r, t)

fun finish(c as T{calls, after}) =
   let
      fun loop() =
	 case !calls of
	    [] => Computation.T(List.reverse(!after))
	  | _ => (raisee(c, NONE) ; loop())
   in loop()
   end

end
structure StringMap = StringMap()
structure Trace =
   Trace(structure IC = IntermediateComputation(structure Time = Time)
	 structure StringMap = StringMap
	 structure Time = Time)
structure Computation = Trace.Computation
signature OPTION =
   sig
      type 'a t = 'a option

      val equals: ('a * 'a -> bool) -> 'a t * 'a t -> bool
      val isSome: 'a t -> bool
      val map: 'a t * ('a -> 'b) -> 'b t
      val map': 'a t * ('a -> unit) -> unit
      val valOf: 'a t -> 'a
      val layout: ('a -> Layout.t) -> 'a t -> Layout.t
   end
(*-------------------------------------------------------------------*)
(*                              Option                               *)
(*-------------------------------------------------------------------*)

structure Option: OPTION =
struct

type 'a t = 'a option

open Pervasive.Option

fun map'(opt, f) =
   case opt of
      NONE => ()
    | SOME x => f x

fun map(opt, f) =
   case opt of
      NONE => NONE
    | SOME x => SOME(f x)

fun equals eq =
   fn (NONE, NONE) => true
    | (SOME x, SOME y) => eq(x, y)
    | _ => false

local open Layout
in fun layout layoutX opt =
   case opt of
      NONE => str "None"
    | SOME x => seq[str "Some(", layoutX x, str ")"]
end

end
signature BOOL =
   sig
      include T

      val toString: t -> string
      val fromString: string -> t option
   end where type t = bool
(*-------------------------------------------------------------------*)
(*                               Bool                                *)
(*-------------------------------------------------------------------*)

structure Bool: BOOL =
struct
   
open Pervasive.Bool
   
type t = bool

val equals =
   fn (true, true) => true
    | (false, false) => true
    | _ => false

val layout = Layout.str o toString
(*fun output(b, out) = Pervasive.IO.output(out, toString b)*)
   
end
signature RING_ELT_REF =
   sig
      structure R: RING

      type t = R.t Ref.t
	 
      val + : t * R.t -> unit
      val - : t * R.t -> unit
      val * : t * R.t -> unit
      val inc: t -> unit
      val dec: t -> unit
      val counter: R.t -> {next: unit -> R.t, cur: unit -> R.t}
   end
(*-------------------------------------------------------------------*)
(*                            RingEltRef                             *)
(*-------------------------------------------------------------------*)

functor RingEltRef(R: RING): RING_ELT_REF =
struct

structure R = R
open R
   
type t = R.t Ref.t

val (op +) = fn (r, x) => r := !r + x
   
fun inc r = r + one

val (op -) = fn (r, x) => r := !r - x

fun dec r = r - one
   
val (op * ) = fn (r, x) => r := !r * x

fun counter n = let val r = ref n
		in {next = fn () => let val n = !r
				    in inc r ; n
				    end,
                    cur = fn () => !r}
		end

end
structure IntRef = RingEltRef(Int)
signature DATE_STRUCTS = 
   sig
   end

signature MY_DATE = 
   sig
      include DATE (* the pervasive one *)

      type t = date

      val monthToInt: month -> int
      val now: unit -> t
      val layout: t -> Layout.t
   end
functor Date(S: DATE_STRUCTS): MY_DATE = 
struct

open S

open Pervasive.Date
structure Time = Time

type t = date

val monthToInt =
   fn Jan => 1
    | Feb => 2
    | Mar => 3
    | Apr => 4
    | May => 5
    | Jun => 6
    | Jul => 7
    | Aug => 8
    | Sep => 9
    | Oct => 10
    | Nov => 11
    | Dec => 12
   
fun now() = fromTimeLocal(Time.now())

val layout = Layout.str o toString
   
end

structure Date = Date()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REGION_STRUCTS = 
   sig
   end

signature REGION = 
   sig
      include REGION_STRUCTS
      
      datatype t = T of {left: int, right: int}

      val bogus: t

      val left: t -> int
      val right: t -> int

      val list: 'a list * ('a -> t) -> t

      (* for maintaining the current line number while lexing and parsing *)
      val current: unit -> t
      val newline: int -> unit
      val reset: unit -> unit
	 
      val layout: t -> Layout.t

      structure Wrap :
	 sig
	    type region
	    type 'a t
	    val region: 'a t -> region
	    val node: 'a t -> 'a
	    val makeRegion: 'a * region -> 'a t
	    val makeRegion':  'a * int * int -> 'a t
	    val make: 'a -> 'a t
	    val dest: 'a t -> 'a * region
	    val left: 'a t -> int
	    val right: 'a t -> int
	 end
      sharing type Wrap.region = t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Region(S: REGION_STRUCTS): REGION = 
struct

open S

datatype t = T of {left: int, right: int}

fun left(T{left, ...}) = left
fun right(T{right, ...}) = right
   
val bogus = T{left = ~1, right = ~1}

fun list(xs, reg) =
   case xs of	       
      [] => bogus
    | x :: _ => T{left = left(reg x),
		right = right(reg(List.last xs))}

val isBogus =
   fn (T{left = ~1, ...}) => true
    | _ => false

val lineNum = ref 1
val lineIndexes = ref[0]

fun current() =
   case !lineIndexes of
      index :: _ => T{left = index, right = index}
    | _ => Error.bug "Region.current"

fun newline index =
   (IntRef.inc lineNum ;
    ListRef.push(lineIndexes, index))

fun indexToPos index =
   let
      fun loop(indexes, lineNum) =
	 case indexes of
	    i :: indexes =>
	       if index >= i
		  then {line = lineNum,
			column = index - i}
	       else loop(indexes, lineNum-1)
	  | _ => {line = ~1, column = ~1} (*Error.bug "indexToPos"*)
   in loop(!lineIndexes, !lineNum)
   end

local open Layout
   fun layoutIndex index =
      let val {line, column} = indexToPos index
	 open Layout
      in seq[Int.layout line, str ".", Int.layout column]
      end
in   
   fun layout(T{left, right}) =
      if left = ~1
	 then str "<bogus>"
      else seq[layoutIndex left, str "-", layoutIndex right]
end

fun reset() = (lineNum := 1 ;
	       lineIndexes := [0])

val make = T

structure Wrap =
   struct
      type region = t
      datatype 'a t = T of {node: 'a, region: region}

      fun node(T{node, ...}) = node
      fun region(T{region, ...}) = region
      fun makeRegion(node, region) = T{node = node, region = region}
      fun makeRegion'(node, left, right) = T{node = node,
					   region = make{left = left,
							 right = right}}
      fun make node = makeRegion(node, bogus)
      fun dest(T{node, region}) = (node, region)
      val left = fn T{region, ...} => left region
      val right = fn T{region, ...} => right region
   end
   
end

structure Region = Region()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CONTROL =
   sig
      val instrumentSxml: bool ref
	 
      (* set all flags to their default values *)
      val defaults: unit -> unit

      (*------------------------------------*)
      (*            Begin Flags             *)
      (*------------------------------------*)

      (* whether or not to print auxiliary files *)
      val aux: bool ref

      datatype chunk =
	 OneChunk
       | ChunkPerFunc
       | Coalesce of {limit: int}

      val chunk: chunk ref

      (* SOME n means that the executable should use a fixed heap of size n *)
      val fixedHeap: int option ref

      val flattenRounds: int ref

      (* The .h files that should be #include'd in the .c file. *)
      val includes: string list ref
	 
      datatype inline =
	 NonRecursive of {size: int option}
       | Leaf of {size: int option}
       | LeafNoLoop of {size: int option}
       | Recursive of {size: int}

      val inline: inline ref
      val layoutInline: inline -> Layout.t
      val setInlineSize: int -> unit

      val inlineRounds: int ref
	 
      (* *)
      val inputFile: File.t ref

      (* call count instrumentation *)
      val instrument: bool ref

      val keepCps: bool ref

      (* used by the match compiler *)
      datatype match = LeftToRight | Rotate
      val match: match ref

      (* how to use MLton *)
      datatype mode =
	 Defunctorize
(*       | Monomorphise *)
       | Compile
      val setMode: mode -> unit
      val getMode: unit -> mode

      val polyvariance:
	 {
	  rounds: int,
	  (* Only duplicate big functions when
	   * (size - small) * (number of occurrences - 1) <= product
	   *)
	  small: int,
	  product: int
	 } option ref

      (* elaborator inserts code to print a message on entry to each function *)
      val printAtFunEntry: bool ref

      (* profiling of the C code (at the granularity of Cps function). *)
      val profile: bool ref

      (* overflow, bounds, div by zero checking.
       * controls the value of the variable MLton.safe available in the Basis
       *)
      val safe: bool ref
	 
      (* run typeChecker on ILs *)
      val setTypeCheck: bool -> unit
      val shouldTypeCheck: unit -> bool

      (*------------------------------------*)
      (*             End Flags              *)
      (*------------------------------------*)

      (* tracing and other informative messages*)
      val messagesAreOn: unit -> bool
      val messagesOn: unit -> unit
      val messagesOff: unit -> unit
      val message: (unit -> Layout.t) -> unit
      val messageLay: Layout.t -> unit
      val messageStr: string -> unit
      val traceCall: string -> ('a -> 'b) -> 'a -> 'b
      val sizeMessage: string * 'a -> Layout.t

      val displays: string * ((Layout.t -> unit) -> unit) -> unit
      val display: string * (unit -> Layout.t) -> unit

      (*------------------------------------*)
      (*          Error Reporting           *)
      (*------------------------------------*)
      val reset: unit -> unit
      val anyErrors: unit -> bool

      (* abort compilation once this many errors reached *)
      val errorThreshhold: int ref

      val error: Region.t * Layout.t -> unit
      val errorStr: Region.t * string -> unit

      (*------------------------------------*)
      (*          Compiler Passes           *)
      (*------------------------------------*)
      datatype style = No | C | ML

      datatype 'a display =
	 NoDisplay
       | Layout of 'a -> Layout.t
       | Layouts of 'a * (Layout.t -> unit) -> unit

      val header: style -> Layout.t

      val pass: {name: string,
		 suffix: string,
		 style: style,
		 thunk: unit -> 'a,
		 display: 'a display} ->'a
	 
      val passTypeCheck: {name: string,
			   suffix: string,
			   style: style,
			   thunk: unit -> 'a,
			   display: 'a display,
			   typeCheck: 'a -> unit} -> 'a
	 
      val passSimplify: {name: string,
			  suffix: string,
			  style: style,
			  thunk: unit -> 'a,
			  display: 'a display,
			  simplify: 'a -> 'a,
			  typeCheck: 'a -> unit} -> 'a
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Control: CONTROL =
   struct
      local
	 val defaults = ref []
	 val settings = ref []
      in
	 fun setDefaults() = List.foreach(!defaults, fn f => f())

	 fun layoutSettings() =
	    let open Layout
	    in align(List.map(List.reverse(!settings),
			      fn (name, lay) => seq[name, lay()]))
	    end
	 
	 fun control{name, default, layout} =
	    let val r = ref default
	    in ListRef.push(settings, (Layout.str(name ^ ": "),
				       fn () => layout(! r)))
	       ; ListRef.push(defaults, fn () => r := default)
	       ; r
	    end
      end
      
      val aux = control{name = "aux",
			default = false,
			layout = Bool.layout}		     

      structure Chunk =
	 struct
	    datatype t =
	       OneChunk
	     | ChunkPerFunc
	     | Coalesce of {limit: int}

	    val toString =
	       fn OneChunk => "one chunk"
		| ChunkPerFunc => "chunk per function"
		| Coalesce{limit} => "coalesce " ^ Int.toString limit

	    val layout = Layout.str o toString
	 end

      datatype chunk = datatype Chunk.t
      val chunk = control{name = "chunk",
			  default = Coalesce{limit = 2000},
			  layout = Chunk.layout}

      val fixedHeap = control{name = "fixed heap",
			      default = NONE,
			      layout = Option.layout Int.layout}

      val flattenRounds = control{name = "flatten rounds",
				  default = 2,
				  layout = Int.layout}

      structure Inline =
	 struct
	    datatype t =
	       NonRecursive of {size: int option}
	     | Leaf of {size: int option}
	     | LeafNoLoop of {size: int option}
	     | Recursive of {size: int}

	    local open Layout
	       val iol = Option.layout Int.layout
	    in
	       val layout = 
		  fn NonRecursive{size} => seq[str "NonRecursive ", iol size]
		   | Leaf{size} => seq[str "Leaf ", iol size]
		   | LeafNoLoop{size} => seq[str "LeafNoLoop ", iol size]
		   | Recursive{size} =>
			seq[str "Recursive ", Int.layout size]
	    end
	 end

      val includes: string list ref =
	 control{name = "includes",
		 default = [],
		 layout = List.layout String.layout}
	 
      datatype inline = datatype Inline.t
      val layoutInline = Inline.layout

      val inline = control{name = "inline",
			   default = NonRecursive{size = SOME 50},
			   layout = Inline.layout}

      val inlineRounds = control{name = "inline rounds",
				 default = 1,
				 layout = Int.layout}

      fun setInlineSize(size: int): unit =
	 inline := (case !inline of
		       NonRecursive _ => NonRecursive{size = SOME size}
		     | Leaf _ => Leaf{size = SOME size}
		     | LeafNoLoop _ => LeafNoLoop{size = SOME size}
		     | Recursive _ => Recursive{size = size})
	 
      val inputFile = control{name = "input file",
			      default = File.fromString "bogus",
			      layout = File.layout}

      val instrument = control{name = "instrument",
			       default = false,
			       layout = Bool.layout}

      val instrumentSxml = control{name = "instrument Sxml",
				   default = false,
				   layout = Bool.layout}

      val keepCps = control{name = "keepCps",
			    default = false,
			    layout = Bool.layout}

      structure Match =
	 struct
	    datatype t = LeftToRight | Rotate

	    fun layout m =
	       Layout.str
	       (case m of
		   LeftToRight => "left to right"
		 | Rotate => "rotate")
	 end
      datatype match = datatype Match.t
      val match = control{name = "match",
			  default = LeftToRight,
			  layout = Match.layout}

      val messages = control{name = "messages",
			     default = false,
			     layout = Bool.layout}

      structure Mode =
	 struct
	    datatype t =
	       Defunctorize
(*	     | Monomorphise *)
	     | Compile
	       
	    val toString =
	       fn Defunctorize => "defunctorize"
(*		| Monomorphise => "monomorphise" *)
		| Compile => "compile"
		     
	    val layout = Layout.str o toString
	 end

      datatype mode = datatype Mode.t

      val mode = control{name = "mode",
			 default = Compile,
			 layout = Mode.layout}

      val polyvariance =
	 control{name = "polyvariance",
		 default = SOME{rounds = 2,
				small = 30,
				product = 300},
		 layout =
		 Option.layout
		 (fn {rounds, small, product} =>
		  Layout.record[("rounds", Int.layout rounds),
				("small", Int.layout small),
				("product", Int.layout product)])}

      val printAtFunEntry = control{name = "print at fun entry",
				    default = false,
				    layout = Bool.layout}

      val profile = control{name = "profile",
			    default = false,
			    layout = Bool.layout}

      val safe = control{name = "safe",
			 default = true,
			 layout = Bool.layout}

      val typeCheck = control{name = "type check",
			      default = true,
			      layout = Bool.layout}
	 
      fun setTypeCheck b =
	 case (b, !mode) of
(* 	    (true, Monomorphise) =>
 * 	       Error.bug "can't type check when monomorphising"
 *)
	    _ => typeCheck := b
      fun shouldTypeCheck() = !typeCheck

      fun setMode m =
 	 case m of
(* 	    Monomorphise => Error.unimplemented "Monomorphise"
 *)
	   _ => (mode := m
		  ; typeCheck := (case m of
				     Defunctorize => false
(*				   | Monomorphise => false *)
				   | Compile => !typeCheck))
      fun getMode() = !mode

      val version = Layout.str "1999-7-12"

      datatype style = No | C | ML

      fun preSuf style =
	 let
	    val (p, s) =
	       case style of
		  No => ("", "")
		| C => ("/*", "*/")
		| ML => ("(*", "*)")
	 in (Layout.str p, Layout.str s)
	 end
	 
      fun header commentStyle =
	 let
	    val (pre, suf) = preSuf commentStyle
	    open Layout
	 in align[pre,
		  seq[str "MLton ", version,
		      str " created this file on ", Date.layout(Date.now()),
		      str "."],
		  seq[str "Do not edit this file."],
		  str "Flag settings: ",
		  indent(layoutSettings(), 3),
		  suf]
	 end

      fun on name =
	 (Trace.Immediate.on[name]
	  ; Trace.Time.on[name])
	 
      fun traceCall name =
	 (on name
	  ; Trace.traceCall name)

      fun message(th: unit -> Layout.t): unit =
	 if !messages
	    then Trace.Immediate.message(th())
	 else ()

      fun messageLay(l: Layout.t): unit = message(fn () => l)

      fun messageStr(s: string): unit =
	 message(fn () => Layout.str s)

      fun messagesAreOn() = !messages
	 
      fun messagesOn() =
	 (messages := true
	  ; Trace.Time.flagged()
	  ; Trace.Immediate.flagged())

      fun messagesOff() =
	 (messages := false
	  ; Trace.Time.never()
	  ; Trace.Immediate.never())

      fun defaults() = (setDefaults(); messagesOff())

      val _ = defaults()
	 
      fun displays(name: string, thunk: (Layout.t -> unit) -> unit): unit =
	 traceCall "display"
	 File.withOut(File.fromString
		      (concat[String.dropSuffix(File.toString(!inputFile),
						4),
			      ".",
			      name]),
		      fn out =>
		      thunk(fn l => (Layout.output(l, out)
				     ; Out.newline out)))

      val displays = fn arg => if !aux then displays arg else ()
		      
      fun display(name: string, thunk: (unit -> Layout.t)): unit =
	 displays(name, fn disp => disp(traceCall "layout" thunk ()))

      (*------------------------------------*)
      (*               Errors               *)
      (*------------------------------------*)

      val numErrors = ref 0

      val errorThreshhold = ref 20
	 
      fun anyErrors() = !numErrors > 0

      fun error(r: Region.t, msg: Layout.t): unit =
	 (IntRef.inc numErrors
	  ; Out.output(Out.error,
		       Layout.toString(let open Layout
				       in seq[File.layout(!inputFile),
					      str ":",
					      Region.layout r,
					      str " Error: ",
					      msg]
				       end))
	  ; Out.newline Out.error
	  ; if !numErrors = !errorThreshhold
	       then Error.error "compilation aborted: too many errors"
	    else ())

      fun errorStr(r, msg) = error(r, Layout.str msg)

      fun reset() = (Region.reset(); numErrors := 0)

      (*---------------------------------------------------*)
      (*                  Compiler Passes                  *)
      (*---------------------------------------------------*)

      datatype 'a display =
	 NoDisplay
       | Layout of 'a -> Layout.t
       | Layouts of 'a * (Layout.t -> unit) -> unit

      fun 'a sizeMessage(name: string, a: 'a): Layout.t =
	 let open Layout
	 in str(concat[name, " size is ",
		       Int.toCommaString(MLton.size a), " bytes"])
	 end

      fun pass{name: string,
	       suffix: string,
	       style: style,
	       display = disp,
	       thunk: unit -> 'a}: 'a =
	 let
	    val result = traceCall name thunk ()
	 in
	    message(fn () => sizeMessage(suffix, result))
	       
	    ; if anyErrors()
		 then Error.error("compilation aborted: " ^ name
				  ^ " reported errors")
	      else ()

	    ; (case disp of
		  NoDisplay => ()
		| Layout layout =>
		     display(suffix, fn () => Layout.align[header style,
							   layout result])
		| Layouts layout =>
		     displays(suffix,
			      fn output =>
			      (output(header style)
			       ; layout(result, output))))
	    ; result
	 end

      fun passTypeCheck{name: string,
			suffix: string,
			style: style,
			display,
			thunk: unit -> 'a,
			typeCheck = tc: 'a -> unit}: 'a =
	 let val result = pass{name = name,
			       suffix = suffix,
			       display = display,
			       style = style,
			       thunk = thunk}
	 in if !typeCheck
	       then traceCall "typeCheck" tc result
	    else ()
	    ; result
	 end
      
      fun passSimplify{name, suffix, style, thunk, display, typeCheck, simplify} =
	 let
	    val result =
	       passTypeCheck{name = name,
			     suffix = suffix ^ ".unsimplified",
			     style = style,
			     thunk = thunk,
			     display = NoDisplay,
			     typeCheck = typeCheck}
	 in passTypeCheck{name = name ^ " simplify",
			  suffix = suffix,
			  style = style,
			  thunk = fn () => simplify result,
			  display = display,
			  typeCheck = typeCheck}
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature POPT =
   sig
      (* This type specifies what kind of arguments a switch expects
       * and provides the function to be applied to the argument.
       *)
      datatype opt =
	 (* no args *)
	 None of unit -> unit
	 (* one arg: an intege. *)
       | Int of int -> unit
	 (* one arg: a single digit. *)
       | Digit of int -> unit
	 (* one arg: an integer followed by optional k or m. *)
       | Mem of int -> unit
	 (* Any string immediately follows the switch. *)
       | String of string -> unit
	 (* one arg: any string *)
       | SpaceString of string -> unit

      (* Parse the switches, applying the first matching opt to each switch,
       * and return any remaining args.
       * Returns NONE if it encounters an error.
       * For example, if opts is:
       *  [("foo", None f)]
       * and the switches are:
       *  ["-foo", "bar"]
       * then parse will call f() and return "bar".
       *)
      val parse:
	 {
	  switches: string list,
	  opts: (string * opt) list
	 }
	 -> string list option
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Popt: POPT =
struct

datatype opt =
   None of unit -> unit
 | Int of int -> unit
 | Digit of int -> unit
 | Mem of int -> unit
 | String of string -> unit
 | SpaceString of string -> unit
   
fun memString(s: string): int option =
   let
      val n = size s
      fun loop(i, ac) =
	 if i = size s
	    then SOME ac
	 else
	    let val c = String.sub(s, i)
	       fun done n = SOME(n * ac)
	    in case c of
	       #"m" => done 1048576
	     | #"k" => done 1024
	     | _ =>
		  case Char.digitToInt c of
		     NONE => NONE
		   | SOME c => loop(i + 1, ac * 10 + c)
	    end
   in loop(0, 0)
   end

(* Parse the command line opts and return any remaining args. *)
fun parse{switches: string list,
	  opts: (string * opt) list}: string list option =
   let
      exception Error
      fun error() = raise Error
      val rec loop =
	 fn [] => []
	  | switch :: switches =>
	       (case String.sub(switch, 0) of
		   #"-" =>
		   let val switch = String.dropPrefix(switch, 1)
		   in case List.peek(opts, fn (switch', arg) =>
				     switch = switch') of
		      NONE =>
			 let
			    val rec loop' =
			       fn [] => error()
				| (switch', arg) :: opts =>
				     let
					fun doit f =
					   if String.isPrefix
					      {string = switch, prefix = switch'}
					      then f(String.dropPrefix
						     (switch, size switch'))
					   else loop' opts
				     in case arg of
					Digit f =>
					   doit(fn s =>
						if size s = 1
						   then (case Char.digitToInt
							    (String.sub(s, 0)) of
							    NONE => error()
							  | SOME i => f i)
						else error())
				      | String f => doit f
				      | _ => loop' opts
				     end
			 in loop' opts
			    ; loop switches
			 end
		    | SOME(_, arg) =>
			 let
			    fun next(f, get) =
			       case switches of
				  switch :: switches =>
				     (case get switch of
					 SOME n => (f n; loop switches)
				       | _ => error())
				| _ => error()
			 in (case arg of
				None f => (f(); loop switches)
			      | Int f => next(f, Int.fromString)
			      | SpaceString f => next(f, SOME)
			      | Mem f => next(f, memString)
			      | _ => error())
			 end
		   end
		 | _ => switch :: switches)
   in SOME(loop switches) handle Error => NONE
   end
end
signature PROPERTY_LIST = 
   sig
      type t

      (* create an empty property list *)
      val new: unit -> t

      (* pointer equality of property lists *)
      val equals: t * t -> bool
	 
      (* remove all properties from the list *)
      val clear: t -> unit

      (* create a new property *)
      val newProperty:
	 unit -> {
                  (* See if a property is in a property list.
		   * NONE if it isn't.
		   *)
		  peek: t -> 'a option,
		  (* get a property from a property list.
		   * Raises error if the property isn't there.
		   *)
		  get: t -> 'a,
		  (* set(p, v, f)
		   * Change the value of property.
		   * Create a new property and call f if it isn't already there.
		   *)
		  set: t * 'a * (unit -> unit) -> unit,
		  (* Add the value of the property -- must not already exist *)
		  add: t * 'a -> unit,
		  (* remove a property from a property list.
		   * noop if the property isn't there.
		   *)
		  remove: t -> unit
		  }

      val stats: unit -> Layout.t
   end
(*  PROPERTY
 *  hides the underlying property and property lists and gives just
 *  the functions for manipulating "symbols" and "values"
 *  (as they are called in LISP property lists.
 * 
 *  To associate arbitrary information with a data structure, add an extra field
 *  to the data structure that holds a property list.  Then call 'wrap', passing
 *  the field selector and a property, and it will return the property
 *  operations that act directly on the data structure.   Or, call 'wrapNew' and
 *  it will create a new property and return the same operations.
 * 
 *  'pair' associates arbitrary information with pairs of values,
 *  where the first kind of value holds a property list and the second has an
 *  equality function.  'pair' stores an alist based on the equality function
 *  in the property.
 *)
							
signature PROPERTY =
   sig
      structure Plist: PROPERTY_LIST

      type ('sym, 'val) init

      val initConst: 'val -> ('sym, 'val) init
      val initFun: ('sym -> 'val) -> ('sym, 'val) init
      val initRec: ('sym * ('sym -> 'val) -> 'val) -> ('sym, 'val) init
      val initRaise: string * ('sym -> Layout.t) -> ('sym, 'val) init
	 
      val new:
	 ('sym -> Plist.t) * ('sym, 'val) init
	 -> {
	     get: 'sym -> 'val,
	     set: 'sym * 'val -> unit
	    }

      val newDest:
	 ('sym -> Plist.t) * ('sym, 'val) init
	 -> {
	     get: 'sym -> 'val,
	     set: 'sym * 'val -> unit,
	     destroy: unit -> unit
	    }

      (* associate a property with pairs of symbols
       * uses the property of the first and an alist of the second
       *)
(*       val pair :
 * 	 {property: ('sym1, ('sym2 * 'val) list) t,
 * 	  equals: 'sym2 * 'sym2 -> bool}
 * 	 -> ('sym1 * 'sym2, 'val) t
 *)
   end
functor Property(Plist: PROPERTY_LIST): PROPERTY =
struct

structure Plist = Plist
   
datatype ('sym, 'val) init =
   Const of 'val
 | Fun of 'sym -> 'val
 | Raise of string * ('sym -> Layout.t)
 | Rec of 'sym * ('sym -> 'val) -> 'val

val initConst = Const
val initFun = Fun
val initRaise = Raise
val initRec = Rec
   
fun print s = TextIO.output(TextIO.stdErr, s)

fun ('sym, 'val) make(plist: 'sym -> Plist.t,
		      i: ('sym, 'val) init,
		      destroy) =
   let
      val {add, peek, set = set', remove, ...} = Plist.newProperty()

      val (new, destroy) = destroy remove

      fun set(x: 'sym, v: 'val): unit =
	 let val pl = plist x
	 in set'(pl, v, fn () => new pl)
	 end

      fun get(x: 'sym): 'val =
	 let val pl = plist x
	    fun make v = (new pl; add(pl, v); v)
	 in case peek(plist x) of
	    SOME v => v
	  | NONE =>
	       case i of
		  Const v => v
		| Raise(name, layout) =>
		     Error.bug
		     (let open Layout
		      in toString(seq[layout x,
				      str " has no ", str name, str " property"])
		      end)
		| Fun f => make(f x)
		| Rec f => make(f(x, get))
	 end
			 
   in {get = get, set = set, destroy = destroy}
   end

fun new(plist, init) =
   let
      val {get, set, ...} =
	 make(plist, init,
	      fn _ =>
	      (fn _ => (),
	       fn () => Error.bug "attempt to destroy non-destructible property"))
   in {get = get, set = set}
   end

fun newDest(plist, init) =
   make(plist, init,
	fn remove => let val plists = ref []
		     in (fn plist => ListRef.push(plists, plist),
			 fn () => (List.foreach(!plists, remove);
				   plists := []))
		     end)

(* fun ('sym1, 'sym2, 'val) pair
 *    {property: ('sym1, ('sym2 * 'val) list) t,
 *     equals: 'sym2 * 'sym2 -> bool}: ('sym1 * 'sym2, 'val) t =
 *    fn i =>
 *    let
 *       fun equalTo b (b', _) = equals(b, b')
 *       val p = property(initConst [])
 * 	 
 *       fun get(s1: 'sym1, s2: 'sym2): 'val =
 * 	 let val alist = #get p s1
 * 	 in case List.peek(alist, equalTo s2) of
 * 	    NONE => init((s1, s2), i, get, 
 * 			 fn v => #set p (s1, (s2, v) :: alist))
 * 	  | SOME(_, v) => v
 * 	 end
 * 
 *       fun set((s1: 'sym1, s2: 'sym2), v: 'val): unit =
 * 	 #set p
 * 	 (s1, 
 * 	  let
 * 	     fun loop(s2vs: ('sym2 * 'val) list): ('sym2 * 'val) list =
 * 		case s2vs of
 * 		   [] => [(s2, v)]
 * 		 | ((s2v as (s2', _)) :: s2vs) =>
 * 		      if equals(s2, s2') then (s2, v) :: s2vs
 * 		      else s2v :: loop s2vs
 * 	  in loop(#get p s1)
 * 	  end)
 * 
 *    in {get = get,
 *        set = set,
 *        destroy = #destroy p}
 *    end
 *)
end
signature BASIC_ORDERED_FIELD =
   sig
      include BASIC_ORDERED_RING

      (* include BASIC_FIELD *)
      val inverse: t -> t
   end

signature BASIC_REAL =
   sig
      include BASIC_ORDERED_FIELD

      structure Format :
	 sig
	    type t
	    val sci: int option -> t
	    val fix: int option -> t
	    val gen: int option -> t
	 end
      structure I: INTEGER

      val random: t * t -> t
      val floor: t -> I.t
      val ceiling: t -> I.t
      val sin: real -> real
      val cos: real -> real
      val atan: real -> real
      val atan2: real * real -> real
      val ln: t -> t
      val exp: t -> t
      val sqrt: t -> t
      val fromInt: I.t -> t
      val fromReal: real -> t
      val toReal: t -> real
      val toString: t -> string
      val fromString: string -> t option
      exception Input
      val input: In.t -> t
      val format: t * Format.t -> string
   end

signature REAL =
   sig
      include BASIC_REAL

      val abs: t -> t
      val / : t * t -> t
      val log: t * t -> t
      val log2: t -> t
      val ^ : t * t -> t
      val min: t * t -> t
      val max: t * t -> t
      val choose: t * t -> t
   end
signature BASIC_FIELD =
   sig
      include BASIC_RING
      
      val inverse: t -> t
   end

signature FIELD =
   sig
      include RING

      (* include BASIC_FIELD *)
      val inverse: t -> t
      (* *)
	 
      val / : t * t -> t
   end

(*-------------------------------------------------------------------*)
(*                               Field                               *)
(*-------------------------------------------------------------------*)

functor Field (F: BASIC_FIELD): FIELD =
struct

structure U = Ring(F)

open F U

val (op /) = fn (x, y) => x * inverse y
    
end
signature ORDERED_FIELD =
   sig
      include ORDERED_RING

      (* include BASIC_ORDERED_FIELD *)
      (* include FIELD *)
      val inverse: t -> t
      val / : t * t -> t
      (* *)
   end

(*-------------------------------------------------------------------*)
(*                           OrderedField                            *)
(*-------------------------------------------------------------------*)

functor OrderedField(F: BASIC_ORDERED_FIELD): ORDERED_FIELD =
struct

structure U = Field(F)
structure U' = OrderedRing(F)
open F U U'

end
(*-------------------------------------------------------------------*)
(*                               Real                                *)
(*-------------------------------------------------------------------*)

functor Real(R: BASIC_REAL): REAL =
struct

structure R' = OrderedField(R)
open R R'

fun choose(n, k) =
   let val k = max(k,n-k)
   in prodFromTo(add1 k,n) / factorial(n-k)
   end

fun abs x = if x >= zero then x else ~ x
   
fun log(base, arg) = (ln arg) / (ln base)
   
val ln2 = ln two
   
fun log2 x = ln x / ln2
   
val (op ^) = fn (m,n) => exp(n * ln m)

fun min(x, y) = if x > y then y else x

fun max(x, y) = if x > y then x else y
  
end
signature FLOAT_REAL =
   sig
      include REAL
   end where type t = real

(*-------------------------------------------------------------------*)
(*                             FloatReal                             *)
(*-------------------------------------------------------------------*)

structure FloatReal: FLOAT_REAL =
struct

structure R =
   struct
      structure I = Int
	 
      open Pervasive.Real
      open Math
	 
      type t = real
	 
      val zero = 0.0
      val one = 1.0

      val ceiling = ceil
	 
      fun inverse x = 1.0/x
	 
      val equals = Pervasive.Real.==
(*      val equals = fn _ => Error.unimplemented "FloatReal.equals" *)
      val {compare, ...} = Relation.lessEqual{< = op <, equals = equals}

      val layout = Layout.str o toString
      fun output(x, out) = Out.output(out, toString x)

      exception Input 
      fun input i =
	 case fromString(In.inputToSpace i) of
	    SOME x => x
	  | NONE => raise Input
	    
      local
	 structure R = Random
	 val seeds = (13, 13)
	 val r = R.rand seeds
      in
	 fun random(lo, hi) = lo + ((hi - lo) * (R.randReal r))
      end

      val fromInt = real
      fun fromReal x = x
      fun toReal x = x

      structure Format =
	 struct
	    open StringCvt
	    type t = realfmt
	    val sci = SCI
	    val fix = FIX
	    val gen = GEN
	 end
      fun format(x, f) = Pervasive.Real.fmt f x
   end

structure R' = Real(R)
open R R'

end

structure Real = FloatReal
signature HET_CONTAINER =
   sig
      type t

      val new: unit -> {make: 'a -> t,
			pred: t -> bool,
			peek: t -> 'a option}
   end
functor PropertyList(H: HET_CONTAINER):> PROPERTY_LIST =
struct

datatype t = T of H.t list ref

fun new(): t = T(ref [])

val equals = fn (T r, T r') => Ref.equals(r, r')

fun clear(T hs) = hs := []

val numPeeks = ref 0
val numLinks = ref 0
val maxLength = ref 0
   
fun stats() =
   let open Layout
   in align
      [seq[str "numPeeks = ", Int.layout(!numPeeks)],
       seq[str "average position in property list = ",
	   str(Real.format(real(!numLinks) / real(!numPeeks),
			   Real.Format.fix(SOME 3)))]]
   end

fun print s = TextIO.output(TextIO.stdErr, s)

fun 'a newProperty() =
   let val {make, pred, peek = peekH} = H.new()
      fun peek(T hs) =
	 let
	    fun loop(l, n) =
	       let
		  fun update() =
		     numLinks := n + !numLinks
(* 		      if n > !maxLength
 * 			 then (maxLength := n;
 * 			       print(concat["peek followed ",
 * 					    Int.toString n,
 * 					    " links \n"]))
 * 		      else ())
 *)
	       in case l of
		  [] => (update(); NONE)
		| e :: l =>
		     case peekH e of
			r as SOME _ => (update(); r)
		      | NONE => loop(l, n + 1)
	       end
	 in numPeeks := 1 + !numPeeks;
	    loop(!hs, 0)
	 end

      fun get(plist: t) =
	 case peek plist of
	    SOME x => x
	  | NONE => Error.error "get"

      fun set(T hs, v: 'a, f: unit -> unit): unit =
	 let val h = make v
	    val rec loop =
	       fn [] => (f(); [h])
		| h' :: hs => if pred h'
				 then h :: hs
			      else h' :: loop hs
	 in hs := loop(!hs)
	 end

      fun add(T hs, v: 'a): unit = hs := make v :: !hs

      fun remove(T hs) = hs := List.remove(!hs, pred)
   in
      {add = add, peek = peek, get = get, set = set, remove = remove}
   end

end
functor ExnHetContainer():> HET_CONTAINER =
   struct
      type t = exn

      fun 'a new() =
	 let exception E of 'a
	 in {make = E,
	     pred = fn E _ => true | _ => false,
	     peek = fn E x => SOME x | _ => NONE}
	 end
   end

functor RefHetContainer():> HET_CONTAINER =
   struct
      type t = unit ref * (unit -> unit)

      fun 'a new() =
	 let
	    val id = ref()
	    val r: 'a option ref = ref NONE
	    fun make v = (id, fn () => r := SOME v)
	    fun peek(id', f) =
	       if id = id' then (f(); !r before r := NONE)
	       else NONE
	    fun pred(id', _) = id = id'
	 in {make = make, pred = pred, peek = peek}
	 end
   end
structure HetContainer = RefHetContainer()
structure HetContainer = ExnHetContainer()
structure PropertyList = PropertyList(ExnHetContainer())
structure Property = Property(PropertyList)
signature COUNTER =
   sig
      type t
      val new: int -> t
      val next: t -> int
      val tick: t -> unit
      val reset: t * int -> unit
      val value: t -> int
      val equals: t * t -> bool
   end
(*-------------------------------------------------------------------*)
(*                              Counter                              *)
(*-------------------------------------------------------------------*)

functor Counter(): COUNTER =
struct

structure I = Int
structure R = RingEltRef(I)
   
datatype t = T of R.t

fun new n = T(ref n)

fun reset(T r, n) = r := n

fun tick(T r) = R.inc r

fun value(T r) = !r

fun next c = value c before tick c

val equals = fn (T r, T r') => r = r'
   
end

structure Counter = Counter(Int)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature WRAPPED =
   sig
      type node'
      type obj
      val dest: obj -> node' * Region.t
      val region: obj -> Region.t
      val node: obj -> node'
      val make: node' -> obj
      val makeRegion: node' * Region.t -> obj
      val makeRegion': node' * int * int -> obj
      val left: obj -> int
      val right: obj -> int
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYVAR_STRUCTS = 
   sig
   end

signature TYVAR = 
   sig
      include TYVAR_STRUCTS
      include T
      include WRAPPED sharing type obj = t

      (* create a new type variable *)
      val new: {name: string, equality: bool} -> t

      (* create a new type variable named a_n, where n is a counter *)
      val newNoname: {equality: bool} -> t

      (* reset the counter for new type variables *)
      val reset: unit -> unit 

      (* newString "'a" creates a type variable named a
       * newString "''a" creates an equality type variable named a
       *)
      val newString: string * int * int -> t

      (* predicates *)
      val isEquality: t -> bool
      val sameName: t * t -> bool

      val hash: t -> word

      val plist: t -> PropertyList.t
      val clear: t -> unit

      (* layout a list of type variables *)
      val layouts: t list -> Layout.t
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Tyvar(S: TYVAR_STRUCTS) :> TYVAR = 
struct

open S

structure Wrap = Region.Wrap
open Wrap
type node' = {name: string,
	      equality: bool,
	      hash: word,
	      plist: PropertyList.t}
type t = node' Wrap.t
type obj = t

fun newHash() = Word.fromInt(Int.random())

fun toString(tyvar: t) =
   let val {name, equality, ...} = node tyvar
   in (if equality then "''" else "'") ^ name
   end
       
val layout = Layout.str o toString

local
   fun make sel (tyvar:t) = sel(node tyvar)
in
   val name = make #name
   val hash = make #hash
   val plist = make #plist
   val isEquality = make #equality
end

val clear = PropertyList.clear o plist
fun equals(a, a') = PropertyList.equals(plist a, plist a')
fun sameName(a, a') = String.equals(name a, name a')
   
fun newRegion({name, equality}, region) =
   makeRegion({name = name,
	       equality = equality,
	       hash = newHash(),
	       plist = PropertyList.new()},
	      region)

fun new f = newRegion(f, Region.bogus)

fun newString(s, left, right) =
   newRegion(if Char.equals(#"'", String.sub(s, 1))
		then {name = String.dropPrefix(s, 2),
		      equality = true}
	     else {name = String.dropPrefix(s, 1),
		   equality = false},
	     Region.T{left = left, right = right})

(*val make = Trace.trace2("Tyvar.make", String.layout, Bool.layout,
 *			layout) make
 *)

local val c = Counter.new 0
in fun reset() = Counter.reset(c, 0)
   fun newNoname{equality} =
      new{name = "a_" ^ Int.toString(Counter.next c),
	   equality = equality}
end

local open Layout
in
   fun layouts ts =
      case ts of
	 [] => empty
       | [t] => layout t
       | _ => tuple(List.map(ts, layout))
end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature FIELD_STRUCTS = 
   sig
   end

signature FIELD = 
   sig
      include FIELD_STRUCTS
      
      datatype t =
	 String of string
       | Int of int

      val layout: t -> Layout.t
      val equals: t * t -> bool
	 
      (* ordering used for sorting *)
      val < : t * t -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature RECORD_STRUCTS = 
   sig
      val isSorted: bool
      structure Field: FIELD
   end

signature RECORD = 
   sig
      include RECORD_STRUCTS
	 
      type 'a t

      (* constructor *)
      val fromList: (Field.t * 'a) list -> 'a t

      (* destructor *)
      val toList: 'a t -> (Field.t * 'a) list
	 
      (* tuple[a, b, c] creates {1 = a, 2 = b, 3 = c} *)
      val tuple: 'a list -> 'a t

      val isTuple: 'a t -> bool

      (* detuple r returns the components, if r is a tuple *)
      val detupleOpt: 'a t -> 'a list option

      (* range{1 = a, 2 = b, 3 = c} returns [a, b, c] *)
      val range: 'a t -> 'a list
	 
      (* usual list operations *)
      val map: 'a t * ('a -> 'b) -> 'b t
      val foreach: 'a t * ('a -> unit) -> unit
      val exists: 'a t * ('a -> bool) -> bool
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val peek: 'a t * Field.t -> 'a option
      val unzip: 'a t -> Field.t list * 'a list
      val zip: Field.t list * 'a list -> 'a t
	 
      (* create a record with the same fields but a new range.
       also return some additional info *)
      val change: 'a t * ('a list -> 'b list * 'c) -> 'b t * 'c
	 
      val layout: {record: 'a t,
		    separator: string,
		    extra: string,
		    layoutTuple: 'a list -> Layout.t,
		    layoutElt: 'a -> Layout.t} -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* empty tuple is also a record *)

functor Record(S: RECORD_STRUCTS): RECORD = 
struct

open S

datatype 'a t =
   Tuple of 'a list
 | Record of (Field.t * 'a) list

val tuple = Tuple

val isTuple =
   fn Tuple _ => true
    | _ => false

fun toList r =
   case r of
      Tuple l => let
		    fun loop(n, l) =
		       case l of
			  [] => []
			| x :: xs => (Field.Int n, x) :: loop(n+1, xs)
		 in loop(1, l)
		 end
    | Record r => r
	 
fun detupleOpt(r: 'a t): 'a list option =
   case r of
      Tuple t => SOME t
    | Record _ => NONE

fun fromList l =
   let
      fun sort r =
	 #insertionSort(List.ordered{< = fn ((s, _), (s', _)) => Field.<(s, s')}) r
      fun isTuple fs: bool =
	 let
	    fun loop(n, fs) =
	       case fs of
		  [] => true
		| (f, _) :: fs =>
		     case f of
			Field.Int n' => Int.equals(n, n') andalso loop(n + 1, fs)
		      | _ => false
	 in loop(1, fs)
	 end
      val l = if isSorted then sort l else l
   in if isTuple l
	 then Tuple(List.map(l, #2))
      else Record l
   end

val peek: 'a t * Field.t -> 'a option =
   fn (Tuple [], _) => NONE
    | (Tuple t, Field.Int n) => let
				   val rec loop =
				      fn (1, x :: _) => SOME x
				       | (_, []) => NONE
				       | (0, _) => NONE
				       | (n, _ :: xs) => loop(n - 1, xs)
				in loop(n, t)
				end
    | (Record r, f) =>
	 (case List.peek(r, fn (f', _) => Field.equals(f, f')) of
	     NONE => NONE
	   | SOME(_, x) => SOME x)
    | _ => Error.bug "Record.peek"

fun range r =
   case r of
      Tuple xs => xs
    | Record r => List.map(r, #2)

fun exists(r, p) =
   case r of
      Tuple xs => List.exists(xs, p)
    | Record r => List.exists(r, fn (_, x) => p x)

fun fold(r, b, f) =
   case r of
      Tuple xs => List.fold(xs, b, f)
    | Record r => List.fold(r, b, fn ((_, x), b) => f(x, b))
	  
fun map(r: 'a t, f: 'a -> 'b): 'b t =
   case r of
      Tuple xs => Tuple(List.map(xs, f))
    | Record r => Record(List.map(r, fn (field, a) => (field, f a)))

fun foreach(r: 'a t, f: 'a -> unit): unit =
   case r of
      Tuple xs => List.foreach(xs, f)
    | Record r => List.foreach(r, fn (_, a) => f a)

fun change(r: 'a t, f: 'a list -> 'b list * 'c): 'b t * 'c =
   case r of
      Tuple xs => let val (ys, c) = f xs
		  in (Tuple ys, c)
		  end
    | Record r => let val (fs, xs) = List.unzip r
		      val (ys, c) = f xs
		  in (Record(List.zip(fs, ys)), c)
		  end

fun zip arg = fromList(List.zip arg)
fun unzip arg = List.unzip(toList arg)

fun layout{record, layoutTuple, separator, extra, layoutElt} =
   case (record, extra) of
      (Tuple xs, "") => layoutTuple xs
    | _ =>
	 let val r = toList record
	    open Layout
	 in seq[str "{",
		align(separateRight(List.map(r, fn (f, x) =>
					     align[seq[Field.layout f,
						       str(" " ^ separator)],
						   layoutElt x]),
				    ", ")),
		str extra,
		str "}"]
	 end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Field(S: FIELD_STRUCTS): FIELD = 
struct

open S

datatype t =
   String of string
 | Int of int
   
val fromString = String
val fromInt = Int

val equals =
   fn (String s, String s') => String.equals(s, s')
    | (Int n, Int n') => Int.equals(n,n')
    | _ => false
	 
fun layout f =
   case f of
      String s => Layout.str s
    | Int n => Int.layout n
	 
val op < =
   fn (String s, String s') => String.<(s, s')
    | (Int n, Int n') => Int.<(n,n')
    | (String _, Int _) => false
    | (Int _, String _) => true

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_CONST_STRUCTS =
   sig
   end

signature AST_CONST =
   sig
      include AST_CONST_STRUCTS
	 
      datatype t =
	 Int of string
       | Word of Word32.word
       | Real of string
       | String of string
       | Char of char

      val fromInt: int -> t

      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AstConst(S: AST_CONST_STRUCTS) :> AST_CONST =
struct
	 
datatype t =
   Int of string
 | Word of Word32.word
 | Real of string
 | String of string
 | Char of char

val equals = fn _ => Error.unimplemented "Ast.Const.equals"

val fromInt = Int o Int.toString

local
   open Layout
   fun wrap(pre, post, s) = str(pre ^ String.toString s ^ post)
in
   fun layout c =
      case c of
	 Int s => str s
       | Real l => String.layout l
       | Word w => seq[str "0wx", str(Word32.toString w)]
       | Char c => wrap("#\"", "\"", String.implode[c])
       | String s => wrap("\"", "\"", s)
end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRIM_TYCONS_STRUCTS =
   sig
      type t
      val fromString: string -> t
      val equals: t * t -> bool
   end

signature PRIM_TYCONS =
   sig
      type tycon

      val array: tycon
      val arrow: tycon
      val bool: tycon
      val char: tycon
      val exn: tycon
      val int: tycon
      val int31: tycon
      val intInf: tycon
      val list: tycon
      val pointer: tycon
      val real: tycon
      val reff: tycon
      val stack: tycon
      val string: tycon
      val tuple: tycon
      val vector: tycon
      val word: tycon
      val word8: tycon
      val word31: tycon

      val prims: tycon list

      val defaultInt: tycon
      val defaultWord: tycon
	 
      val isWordX: tycon -> bool
      val isIntX: tycon -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRIM_CONS_STRUCTS =
   sig
      type t
      val equals: t * t -> bool
      val fromString: string -> t
   end

signature PRIM_CONS =
   sig
      type con

      (* built-in constructors *)
      val cons: con
      val falsee: con
      val nill: con
      val reff: con
      val truee: con

      (* built-in exception constructors *)
      val bind: con
      val match: con

      val excons: con list
      val isPrimExcon: con -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_ID_STRUCTS =
   sig
      (* the name to print out for error messages *)
      val className: string
   end

signature AST_ID =
   sig
      include AST_ID_STRUCTS
      include T
      include WRAPPED
	 (* where type obj = t *)
	 sharing type obj = t

      val fromString: string -> t
      val toString: t -> string
      val hash: t -> word
      val bogus: t

      (* report an unbound error message *)
      val unbound: t -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LONGID_STRUCTS =
   sig
      structure Strid: AST_ID
      structure Id: AST_ID
   end

signature LONGID =
   sig
      include LONGID_STRUCTS
      include T
      datatype node = T of {strids: Strid.t list, id: Id.t}

      include WRAPPED sharing type node' = node
		      sharing type obj = t

      (* constructors *)
      val fromString: string * Region.t -> t
      val long: Strid.t list * Id.t -> t
      val short: Id.t -> t

      (* destructors *)
      val split: t -> Strid.t list * Id.t
      val toId: t -> Id.t
	 
      (* prepend with a path: 
       prepend(([B, C], x), A) = ([A, B, C], x)
       prepends(([C, D], x), [A, B]) = ([A, B, C, D], x) *)
      val prepend: t * Strid.t -> t
      val prepends: t * Strid.t list -> t

      (* isLong returns true if the list of strids is nonempty *)
      val isLong: t -> bool

      (* the name to print out for error messages *)
      val className: string

      val bogus: t
	 
      val toString: t -> string

      (* report an unbound error message *)
      val unbound: t -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_ATOMS_STRUCTS = 
   sig
      structure Const: AST_CONST
      structure Record: RECORD
      structure SortedRecord: RECORD
      sharing Record.Field = SortedRecord.Field
      structure Tyvar: TYVAR
   end

signature AST_ATOMS =
   sig
      include AST_ATOMS_STRUCTS
      
      structure Tycon:
	 sig
	    include AST_ID
	    include PRIM_TYCONS sharing type tycon = t
	 end

      structure Var: AST_ID

      structure Con:
	 sig
	    include AST_ID
	    include PRIM_CONS sharing type con = t
	 end

      structure Sigid: AST_ID
      structure Strid: AST_ID
      structure Fctid: AST_ID

      structure Vid:
	 sig
	    include AST_ID

	    (* conversions to and from variables and constructors *)
            val fromVar: Var.t -> t
	    val fromCon: Con.t -> t
	    val toVar: t -> Var.t
	    val toCon: t -> Con.t
	    val toFctid: t -> Fctid.t
	 end

      structure Longtycon:
	 sig
	    include LONGID
            val arrow: t
	    val exn: t
	 end sharing Longtycon.Id = Tycon

      structure Longvar: LONGID sharing Longvar.Id = Var

      structure Longcon: LONGID sharing Longcon.Id = Con

      structure Longstrid: LONGID sharing Longstrid.Id = Strid

      structure Longvid:
	 sig
	    include LONGID

	    val bind: t
	    val cons: t
	    val falsee: t
	    val match: t
	    val nill: t
	    val reff: t
	    val truee: t

	    val fromLongcon: Longcon.t -> t
	    val toLongvar: t -> Longvar.t
	    val toLongcon: t -> Longcon.t
	    val toLongstrid: t -> Longstrid.t
	 end sharing Longvid.Id = Vid

      sharing Strid = Longtycon.Strid = Longvar.Strid = Longcon.Strid
	 = Longvid.Strid = Longstrid.Strid

      structure Type :
	 sig
	    type t
	    datatype node =
	       Var of Tyvar.t
	     | Con of Longtycon.t * t list
	     | Record of t SortedRecord.t

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

	    val var: Tyvar.t -> t
	    val con: Tycon.t * t list -> t
	    val record: t SortedRecord.t -> t
	    val arrow: t * t -> t
	    val exn:  t
	    val tuple: t list -> t
	    val unit: t
	    val layout: t -> Layout.t
	    val layoutOption: t option -> Layout.t
	    val layoutApp: Layout.t * 'a list * ('a -> Layout.t) -> Layout.t
	 end

      structure TypBind :
	 sig
	    type t
	    datatype node = T of {tyvars: Tyvar.t list,
				  tycon: Tycon.t,
				  def: Type.t} list
	    include WRAPPED sharing type node' = node
			    sharing type obj = t

            val empty: t

	    val layout: t -> Layout.t
	 end

      structure DatBind :
	 sig
	    type t
	    datatype node =
	       T of {datatypes: {tyvars: Tyvar.t list,
				 tycon: Tycon.t,
				 cons: (Con.t * Type.t option) list} list,
		     withtypes: TypBind.t}
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	    val layout: string * t -> Layout.t
	 end

      structure DatatypeRhs :
	 sig
	    type t
	    datatype node =
	       DatBind of DatBind.t
	     | Repl of {lhs: Tycon.t, rhs: Longtycon.t}
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	    val layout: t -> Layout.t
	 end

      val bind: Layout.t * Layout.t -> Layout.t
      val layoutAnds: string * 'a list * (Layout.t * 'a -> Layout.t) -> Layout.t
      datatype bindStyle = OneLine | Split of int
      val layoutAndsBind :
	 string * string * 'a list * ('a -> bindStyle * Layout.t * Layout.t)
	 -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_CORE_STRUCTS = 
   sig
      include AST_ATOMS
   end

signature AST_CORE = 
   sig
      include AST_CORE_STRUCTS
      
      (* controls layout of applications of Vid's *)
      val isInfix: (Vid.t -> bool) ref
      
      structure Fixity:
	 sig
	    datatype t =
	       Infix of int option
	     | Infixr of int option
	     | Nonfix

	    val isInfix: t -> bool
	    val layout: t -> Layout.t
	 end
	 
      structure Fixop:
	 sig
	    datatype t = Op | None
	 end

      structure Pat:
	 sig
	    type t
	       
	    structure Item:
	       sig
		  type pat
		  datatype t =
		     Field of Record.Field.t * pat
		   | Vid of Vid.t * Type.t option * pat option 
		     (* vid <:ty> <as pat> *)
	       end
	    sharing type Item.pat = t

	    datatype node =
	       Wild
	     | Var of {fixop: Fixop.t, name: Longvid.t}
	     | Const of Const.t
	     | Tuple of t list
	     | Record of {items: Item.t list, flexible: bool}
	     | List of t list
	     | FlatApp of t list
	     | App of Longcon.t * t
	     | Constraint of t * Type.t
	     | Layered of {fixop: Fixop.t,
			   var: Var.t,
			   constraint: Type.t option,
			   pat: t}
	       
	    include WRAPPED sharing type node' = node
			    sharing type obj = t

	    val wild: t
	    val longvid: Longvid.t -> t
	    val var: Var.t -> t
	    val const: Const.t -> t
	    val tuple: t list -> t
	    (*	    val record: {record: t Record.t, flexible:bool} -> t*)
	    val con: Con.t -> t
	    val app: Con.t * t -> t
	    val constraint: t * Type.t -> t
	    val layered: {fixop: Fixop.t,
			   var: Var.t,
			   constraint: Type.t option,
			   pat: t} -> t

	    val layout: t -> Layout.t
	 end
      
      structure Exp:
	 sig
	    type dec
	    type t
	    datatype node =
	       Const of Const.t
	     | Var of {name: Longvid.t, fixop: Fixop.t}
	     | Fn of (Pat.t * t) list
	     | FlatApp of t list
	     | App of t * t
	     | Case of t * (Pat.t * t) list
	     | Let of dec * t
	     | Seq of t list
	     | Record of t Record.t
	     | List of t list
	     | Selector of Record.Field.t
	     | Constraint of t * Type.t
	     | Handle of t * (Pat.t * t) list
	     | Raise of t
	     | If of t * t * t
	     | Andalso of t * t
	     | Orelse of t * t
	     | While of {test: t, expr: t}
	     | Prim of {name: string, ty: Type.t}

	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	       
	    val const: Const.t -> t
	    val var: Var.t -> t
	    val longvid: Longvid.t -> t
	    val fnn: (Pat.t * t) list -> t
	    val app: t * t -> t
	    val casee: t * (Pat.t * t) list -> t
	    val constraint: t * Type.t -> t
	    val con: Con.t -> t
	    val handlee: t * (Pat.t * t) list -> t
	    val raisee: t -> t
	    val tuple: t list -> t
	    val select: {tuple: t, offset: int} -> t
	    val record: t Record.t -> t
	    val lett: dec list * t -> t

	    val iff: t * t * t -> t
	    val unit: t

	    val layout: t -> Layout.t
	 end

      structure EbRhs:
	 sig
	    type t
	    datatype node =
	       Gen of Type.t option
	     | Def of Longcon.t
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	 end

      structure Dec:
	 sig
	    type t
	    datatype node =
	       Val of Tyvar.t list * (Pat.t * Exp.t) list
	     | Valrec of Tyvar.t list * {var: Var.t,
					 ty: Type.t option,
					 fixity: Vid.t option,
					 rules: (Pat.t * Exp.t) list} list
	     | Fun of Tyvar.t list * {pats: Pat.t list,
				      resultType: Type.t option,
				      body: Exp.t} list list
	     | Type of TypBind.t
	     | Datatype of DatatypeRhs.t
	     | Abstype of {datBind: DatBind.t, body: t}
	     | Exception of (Con.t * EbRhs.t) list
	     | SeqDec of t list
	     | Local of t * t
	     | Open of Longstrid.t list
	     | Overload of Var.t * Type.t * Longvar.t list
	     | Fix of {fixity: Fixity.t, ops: Vid.t list}
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	       
            val empty: t
	    val vall: Tyvar.t list * Var.t * Exp.t -> t
	    val funn:
	       Tyvar.t list
	       * (Var.t * (Pat.t * Exp.t) list * Type.t option) list -> t
	    val exceptionn: Con.t * Type.t option -> t
	    val openn: Longstrid.t list -> t
	    val datatypee:
	        {tyvars: Tyvar.t list,
		 tycon: Tycon.t,
		 cons: (Con.t * Type.t option) list} list -> t
	       
	    val layout: t -> Layout.t
	 end
      sharing type Dec.t = Exp.dec

      val layoutLet: Layout.t * Layout.t -> Layout.t
      val layoutLocal: Layout.t * Layout.t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AstCore(S: AST_CORE_STRUCTS): AST_CORE = 
struct

open S Layout
structure Field = Record.Field
structure Wrap = Region.Wrap

val layouts = List.map

val isInfix': (Vid.t -> bool) ref = ref(fn _ => false)

fun isInfix id = !isInfix' id

structure Fixity =
   struct
      datatype t =
	 Infix of int option
       | Infixr of int option
       | Nonfix

      val isInfix =
	 fn Nonfix => false
	  | _ => true
	  
      val toString =
	 fn Infix NONE => "infix"
	  | Infix(SOME n) => "infix " ^ Int.toString n
	  | Infixr NONE => "infixr"
	  | Infixr(SOME n) => "infixr " ^ Int.toString n
	  | Nonfix => "nonfix"
	       
      val layout = Layout.str o toString
   end

structure Fixop =
   struct
      datatype t = Op | None

      val layout =
	 fn Op => str "op "
	  | None => empty
   end

fun layoutApp(func: 'a,
	      getLongvid: 'a -> Longvid.t option,
	      layoutFunc: 'a -> Layout.t,
	      arg: 'b,
	      getPair: 'b -> ('b * 'b) option,
	      layoutArg: 'b -> Layout.t) =
   let fun nonfixx() = mayAlign[layoutFunc func, layoutArg arg]
   in case getLongvid func of
      SOME x =>
	 if Longvid.isLong x orelse not(isInfix(Longvid.toId x))
	    then nonfixx()
	 else (case getPair arg of
		  SOME(e1, e2) => mayAlign[layoutArg e1,
				       seq[Longvid.layout x,
					   str " ",
					   layoutArg e2]]
		| NONE => seq[str "op ", Longvid.layout x,
			      str " ", layoutArg arg])
    | _ => nonfixx()
   end
	 
fun layoutConstraint(t, ty) =
   mayAlign[seq[t, str " :"], Type.layout ty]
   
fun maybeConstrain(e, tyo) =
   case tyo of
      NONE => e
    | SOME ty => layoutConstraint(e, ty)

fun nest(prefix, x, y) =
   align[seq[str prefix, x],
	      seq[str "in ", y],
	      str "end"]

fun layoutLet(d, e) = nest("let ", d, e)
fun layoutLocal(d, d') = nest("local ", d, d')

fun layoutLongvid x =
   seq[if not(Longvid.isLong x) andalso isInfix(Longvid.toId x)
(*	  orelse Longvid.toString x = "=" *)
	  then str "op "
       else empty,
       str(let val s = Longvid.toString x
	   in if s = "*" then " * "
	      else if String.isSuffix{string = s, suffix = "*"}
		      then s ^ " "
		   else s
	   end)]

(*---------------------------------------------------*)
(*                     Patterns                      *)
(*---------------------------------------------------*)

structure Pat =
   struct
      open Wrap
      datatype node =
	 Wild
       | Var of {fixop: Fixop.t, name: Longvid.t}
       | Const of Const.t
       | Tuple of t list
       | Record of {items: item list, flexible: bool}
       | List of t list
       | FlatApp of t list
       | App of Longcon.t * t
       | Constraint of t * Type.t
       | Layered of {fixop: Fixop.t,
		     var: Var.t,
		     constraint: Type.t option,
		     pat: t}
      and item =
	 Field of Record.Field.t * t
	| Vid of Vid.t * Type.t option * t option 
      withtype t = node Wrap.t
      type node' = node
      type obj = t

      structure Item =
	 struct
	    type pat = t
	    datatype t = datatype item
	 end

      val wild = make Wild
      val const = make o Const
      val record = make o Record
      val constraint = make o Constraint
      val layered = make o Layered

      val emptyList = make(List[])

      fun longvid x = make(Var{name = x, fixop = Fixop.None})
      val var = longvid o Longvid.short o Vid.fromVar

      fun con c = if Con.equals(c, Con.nill) then emptyList
		  else longvid(Longvid.short(Vid.fromCon c))
		     
      fun app(c, p) =
	 let val default = make(App(Longcon.short c, p))
	 in if Con.equals(c, Con.cons)
	       then (case node p of
			Tuple[p1, p2] =>
			   (case node p2 of
			       List ps => make(List(p1 :: ps))
			     | _ => default)
		      | _ => default)
	    else default
	 end
      
      fun isWild p =
	 case node p of
	    Wild => true
	  | _ => false

      local
	 fun isLongvid x p =
	    case node p of
	       Var{name=x', ...} => Longvid.equals(x, x')
	     | _ => false
      in val isTrue = isLongvid Longvid.truee
	 val isFalse = isLongvid Longvid.falsee
      end
   
      val tuple =
	 fn [p] => p
	  | ps => make(Tuple ps)
      
      fun layout(p, isDelimited) =
	 let fun delimit t = if isDelimited then t else paren t
	 in case node p of
	    Wild => str "_"
	  | Var{name, fixop} => seq[Fixop.layout fixop, layoutLongvid name]
	  | Const c => Const.layout c
	  | Tuple ps => Layout.tuple(layoutPats ps)
	  | Record{items, flexible} =>
	       seq[str "{",
		   mayAlign(separateRight(List.map(items, layoutItem), ", ")),
		   if flexible
		      then str(case items of
				  [] => "..."
				| _ => ", ...")
		   else empty,
		   str "}"]
	  | List ps => Layout.list(layoutPats ps)
	  | FlatApp ps => delimit(layoutFlatApp ps)
	  | App(c, p) =>
	       delimit(layoutApp(c, SOME o Longvid.fromLongcon, Longcon.layout,
				 p,
				 fn p => (case node p of
					     Tuple[p1, p2] => SOME(p1, p2)
					   | _ => NONE),
				 layoutF))
	  | Constraint(p, t) => delimit(layoutConstraint(layoutF p, t))
	  | Layered{fixop, var, constraint, pat} =>
	       delimit
	       (mayAlign[maybeConstrain(seq[Fixop.layout fixop, Var.layout var],
				     constraint),
		      seq[str "as ", layoutT pat]])
	 end
      and layoutPats ps = layouts(ps, layoutT)
      and layoutF p = layout(p, false)
      and layoutT p = layout(p, true)
      and layoutFlatApp ps = seq(separate(List.map(ps, layoutF), " "))
      and layoutItem i =
	 case i of
	    Field(f, p) => seq[Field.layout f, str " = ", layoutT p]
	  | Vid(vid, tyo, po) =>
	       seq[Vid.layout vid,
		   case tyo of
		      NONE => empty
		    | SOME ty => seq[str ": ", Type.layout ty],
		   case po of
		      NONE => empty
		    | SOME p => seq[str " as ", layoutT p]]

      val layout = layoutT
   
      val unit = tuple []
   end

structure Eb =
   struct
      structure Rhs =
	 struct
	    open Wrap
	    datatype node =
	       Gen of Type.t option
	     | Def of Longcon.t
	    type t = node Wrap.t
	    type node' = node
	    type obj = t
	 
	    fun layout rhs =
	       case node rhs of
		  Gen to => Type.layoutOption to
		| Def c => seq[str " = ", Longcon.layout c]
	 end
      
      type t = Con.t * Rhs.t

      fun gen(c, to) = {exn = c, rhs = Rhs.Gen to}

      fun layout(exn, rhs) =
	 seq[Con.layout exn, Rhs.layout rhs]
   end

structure EbRhs = Eb.Rhs

(*---------------------------------------------------*)
(*           Expressions and Declarations            *)
(*---------------------------------------------------*)

datatype expNode =
   Var of {name: Longvid.t, fixop: Fixop.t}
  | Fn of rules
  | FlatApp of exp list
  | App of exp * exp
  | Case of exp * rules
  | Let of dec * exp
  | Seq of exp list
  | Const of Const.t
  | Record of exp Record.t
  | List of exp list
  | Selector of Field.t
  | Constraint of exp * Type.t
  | Handle of exp * rules
  | Raise of exp
  | If of exp * exp * exp
  | Andalso of exp * exp
  | Orelse of exp * exp
  | While of {test: exp, expr: exp}
  | Prim of {name: string, ty: Type.t}
and decNode =
   Val of Tyvar.t list * (Pat.t * exp) list
  | Valrec of Tyvar.t list * {var: Var.t,
			      ty: Type.t option,
			      fixity: Vid.t option,
			      rules: rules} list
  | Fun of Tyvar.t list *  {pats: Pat.t list,
			    resultType: Type.t option,
			    body: exp} list list
  | Type of TypBind.t
  | Datatype of DatatypeRhs.t
  | Abstype of {datBind: DatBind.t, body: dec}
  | Exception of Eb.t list
  | Local of dec * dec
  | SeqDec of dec list
  | Open of Longstrid.t list
  | Overload of Var.t * Type.t * Longvar.t list
  | Fix of {fixity: Fixity.t, ops: Vid.t list}
withtype
   exp = expNode Wrap.t
and dec = decNode Wrap.t
and rules = (Pat.t * expNode Wrap.t) list

open Wrap

fun layoutAndsTyvars(prefix, (tyvars, xs), layoutX) =
   layoutAnds(prefix,
	      case List.map(xs, layoutX) of
		 x :: xs =>
		    (case tyvars of
			[] => x
		      | _ => seq[Tyvar.layouts tyvars, str " ", x]) :: xs
	       | [] => [],
	      fn (prefix, x) => seq[prefix, x])

fun layoutExp(e, isDelimited) =
   let fun delimit t = if isDelimited then t else paren t
   in case node e of
      Var{name, fixop} => seq[Fixop.layout fixop, layoutLongvid name]
    | Fn rs => delimit(seq[str "fn ", layoutRules rs])
    | FlatApp es => seq(separate(List.map(es, layoutExpF), " "))
    | App(function, argument) =>
	 delimit
	 (layoutApp(function,
		    fn e => (case node e of
				Var{name, ...} => SOME name
			      | _ => NONE),
		    layoutExpF,
		    argument,
		    fn e => (case node e of
				Record r =>
				   (case Record.detupleOpt r of
				       SOME[e1, e2] => SOME(e1, e2)
				     | _ => NONE)
			      | _ => NONE),
		    layoutExpF))
    | Case(expr, rules) =>
	 delimit(align[seq[str "case ", layoutExpT expr,
				str " of"],
			    indent(layoutRules rules, 2)])
    | Let(dec, expr) => layoutLet(layoutDec dec, layoutExpT expr)
    | Seq es => paren(align(separateRight(layoutExpsT es, " ;")))
    | Const c => Const.layout c
    | Record r =>
	 Record.layout{record = r,
		       separator = "=",
		       extra = "",
		       layoutTuple = (fn [e] => layoutExp(e, isDelimited)
		     | es => tuple(layoutExpsT es)),
		       layoutElt = layoutExpT}
    | List es => list(layoutExpsT es)
    | Selector f => seq[str "#", Field.layout f]
    | Constraint(expr, constraint) =>
	 delimit(layoutConstraint(layoutExpF expr, constraint))
    | Handle(expr, rules) =>
	 delimit(align[layoutExpF expr,
		       seq[str "handle ", layoutRules rules]])
    | Raise e => delimit(seq[str "raise ", layoutExpF e])
    | If(test, thenCase, elseCase) =>
	 delimit(mayAlign[seq[str "if ", layoutExpT test],
		       seq[str "then ", layoutExpT thenCase],
		       seq[str "else ", layoutExpT elseCase]])
    | Andalso(e, e') =>
	 delimit(mayAlign[layoutExpF e,
		       seq[str "andalso ", layoutExpF e']])
    | Orelse(e, e') =>
	 delimit(mayAlign[layoutExpF e,
		       seq[str "orelse ", layoutExpF e']])
    | While{test, expr} =>
	 delimit(align[seq[str "while ", layoutExpT test],
		       seq[str "do ", layoutExpT expr]])
    | Prim{name, ...} => str name
   end
and layoutExpsT es = layouts(es, layoutExpT)
and layoutExpT e = layoutExp(e, true)
and layoutExpF e = layoutExp(e, false)

and layoutRules rs = alignPrefix("| ", layouts(rs, layoutRule))
   
and layoutRule(pat, exp) =
   mayAlign[seq[Pat.layoutF pat, str " =>"],
	 layoutExpF exp]
      
and layoutDec d =
   case node d of
      Local(d, d') => layoutLocal(layoutDec d, layoutDec d')
    | SeqDec ds => align(layoutDecs ds)
    | Val vbs => layoutAndsTyvars("val", vbs, layoutVb)
    | Valrec rvbs => layoutAndsTyvars("val rec", rvbs, layoutRvb)
    | Fun fbs => layoutAndsTyvars("fun", fbs, layoutFb)
    | Type typBind => TypBind.layout typBind
    | Datatype rhs => DatatypeRhs.layout rhs
    | Abstype{datBind, body} =>
	 align[DatBind.layout("abstype", datBind),
	       seq[str "with ", layoutDec body],
	       str "end"]
    | Exception ebs =>
	 layoutAnds("exception", ebs, fn (prefix, eb) => seq[prefix, Eb.layout eb])
    | Open ss => seq[str "open ",
		     seq(separate(List.map(ss, Longstrid.layout), " "))]
    | Overload(x, t, xs) =>
	 seq[str "_overload ",
	     align[layoutConstraint(Var.layout x, t),
		   layoutAnds("as", xs, fn (prefix, x) =>
			      seq[prefix, Longvar.layout x])]]
    | Fix{fixity, ops} =>
	 seq[Fixity.layout fixity, str " ",
	     seq(separate(List.map(ops, Vid.layout), " "))]
and layoutDecs ds = layouts(ds, layoutDec)
and layoutVb(pat, exp) =
   bind(Pat.layoutT pat, layoutExpT exp)

and layoutRvb{var, rules, ty, ...} =
   bind(maybeConstrain(Var.layout var, ty),
	seq[str "fn ", layoutRules rules])
   
and layoutFb clauses =
   alignPrefix("| ", List.map(clauses, layoutClause))
   
and layoutClause({pats, resultType, body}) =
   mayAlign[seq[maybeConstrain(Pat.layoutFlatApp pats,
			    resultType),
	     str " ="],
	 layoutExpF body] (* this has to be layoutExpF in case body
			   is a case expression *)

structure Rules =
   struct
      type t = (Pat.t * exp) list
   end

structure Exp =
   struct
      open Wrap
      type dec = dec
      type t = exp
      datatype node = datatype expNode
      type node' = node
      type obj = t

      val const = make o Const
      val constraint = make o Constraint
      val fnn = make o Fn
      val handlee = make o Handle
      val raisee = make o Raise
      val record = make o Record

      fun longvid name = make(Var{name = name, fixop = Fixop.None})
      val var = longvid o Longvid.short o Vid.fromVar

      fun select{tuple, offset}: t =
	 (* Can't do this because it generates a flex record pattern, which
	  * sometimes can't be resolved.
	  *)
	 make(App(make(Selector(Field.Int offset)),
		  tuple))
(* 	 let val x = Longvid.short(Vid.fromString "x") *)
(* 	 in make(Let(make *)
(* 		     (Val([], [(Pat.tuple(IntList.mapFromTo(1, size, fn n => *)
(* 							   if n = offset *)
(* 							      then Pat.longvid x *)
(* 							   else Pat.wild)), *)
(* 			       tuple)])), *)
(* 		     longvid x)) *)
(* 	 end *)

      local
	 fun isLongvid x e =
	    case node e of
	       Var{name=x', ...} => Longvid.equals(x, x')
	     | _ => false
      in val isTrue = isLongvid Longvid.truee
	 val isFalse = isLongvid Longvid.falsee
      end
   
      fun iff(a: t, b: t, c: t): t =
	 make(if isTrue b then Orelse(a, c)
	      else if isFalse c then Andalso(a, b)
		   else If(a, b, c))
		 
      fun casee(e: t, rs: (Pat.t * t) list) =
	 let val default = make(Case(e, rs))
	 in case rs of
	    [(p1, e1), (p2, e2)] =>
	       if Pat.isTrue p1 andalso Pat.isFalse p2
		  then iff(e, e1, e2)
	       else if Pat.isFalse p1 andalso Pat.isTrue p2
		       then iff(e, e2, e1)
		    else default
	  | _ => default
	 end

      val emptyList: t = make(List[])
	 
      fun con c: t = if Con.equals(c, Con.nill) then emptyList
		      else longvid(Longvid.short(Vid.fromCon c))

      fun app(e1: t, e2: t): t =
	 let val e = make(App(e1, e2))
	 in case node e1 of
	    Fn rs => casee(e2, rs)
	  | Var{name=x, ...} =>
	       if Longvid.equals(x, Longvid.cons)
		  then case node e2 of
		     Record r =>
			(case Record.detupleOpt r of
			    SOME[e1, es] =>
			       (case node es of
				   List es => make(List(e1 :: es))
				 | _ => e)
			  | _ => e)
		   | _ => e
	       else e
		   | _ => e
	 end

      fun lett(ds: dec list, e: t): t =
	 let
	    fun loop(ds, es) =
	       case ds of
		  [] => SOME(List.reverse es)
		| d :: ds =>
		     case node d of
			Val([], [(p, e)]) =>
			   if Pat.isWild p
			      then loop(ds, e :: es)
			   else NONE
		      | _ => NONE
	 in make(case loop(ds, []) of
		    SOME es => Seq(es @ [e])
		  | NONE => 
		       case node e of
			  Let(d, e) => Let(make(SeqDec(ds@[d])), e)
			| _ => Let(make(SeqDec ds), e))
	 end

      val tuple: t list -> t =
	 fn [e] => e
	  | es => make(Record(Record.tuple es))

      val unit: t = tuple[]

      fun delay(e: t): t = fnn[(Pat.tuple[], e)]
(*	 
      val handleFunc =
	 let val e = Var.fromString "e"
	    val f = Var.fromString "f"
	    val x = Var.fromString "x"
	 in fnn(rules[(Pat.tuple[Pat.var e, Pat.var f],
		       make(Handle(app(var e, unit),
				   rules[(Pat.var x, app(var f, var x))])))])
	 end
*)
      val layout = layoutExpT
   end

structure Dec =
   struct
      open Wrap
      type t = dec
      datatype node = datatype decNode
      type node' = node
      type obj = t

      val openn = make o Open
	 
      fun funn(tyvars, rvbs): t =
	 make
	 (if List.forall(rvbs, fn (_, _, resultTy) =>
			 case resultTy of
			    NONE => true
			  | _ => false)
	     then Fun(tyvars,
		      List.map
		      (rvbs, fn (var, rules, _) =>
		       let val vp = Pat.longvid(Longvid.short(Vid.fromVar var))
		       in List.map
			  (rules, fn (pat, exp) =>
			   {pats = [vp, pat],
			    body = exp,
			    resultType = NONE})
		       end))
	  else Valrec(tyvars,
		      List.map(rvbs, fn (var, rules, _) =>
			       {var = var,
				fixity = NONE,
				rules = rules,
				ty = NONE}))) (* can't use the resultty *)
	     
      fun exceptionn(exn: Con.t, to: Type.t option): t =
	 make(Exception[(exn, make(Eb.Rhs.Gen to))])


      fun datatypee datatypes: t =
	 make
	 (Datatype
	  (DatatypeRhs.make
	   (DatatypeRhs.DatBind
	    (DatBind.make(DatBind.T{withtypes = TypBind.empty,
				    datatypes = datatypes})))))

      val seq = make o SeqDec
      val empty = seq[]

      fun vall(tyvars, var, exp): t =
	 make(Val(tyvars, [(Pat.var var, exp)]))
	 
      (*
      fun typee(tyvars, tycon, typ) =
	 Type(TypBind.make
	      (TypBind.T[{tycon = tycon, def = typ, tyvars = tyvars}]))

*)
      val layout = layoutDec
   end

val isInfix = isInfix'
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PrimTycons(S: PRIM_TYCONS_STRUCTS)
   :> PRIM_TYCONS where type tycon = S.t =
   struct
      open S

      type tycon = t

      val array = fromString "array"
      val arrow = fromString "->"
      val bool = fromString "bool"
      val char = fromString "char"
      val exn = fromString "exn"
      val int = fromString "int"
      val int31 = fromString "int31"
      val intInf = fromString "intInf"
      val list = fromString "list"
      val pointer = fromString "pointer"
      val real = fromString "real"
      val reff = fromString "ref"
      val stack = fromString "stack"
      val string = fromString "string"
      val tuple = fromString "*"
      val vector = fromString "vector"
      val word = fromString "word"
      val word8 = fromString "word8"
      val word31 = fromString "word31"

      val prims =
	 [array, arrow, bool, char, exn, int, int31, intInf, list, pointer,
	  real, reff, stack, string, tuple, vector, word, word8, word31]

      val defaultInt = int
      val defaultWord = word

      fun equalTo t t' = equals(t, t')

      local
	 fun is l t = List.exists(l, equalTo t)
      in
	 val isWordX = is [word, word8, word31]
	 val isIntX = is [int, int31, intInf]
      end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PrimCons(S: PRIM_CONS_STRUCTS) :> PRIM_CONS where type con = S.t =
   struct
      open S

      type con = t
	 
      val cons = fromString "::"
      val falsee = fromString "false"
      val nill = fromString "nil"
      val reff = fromString "ref"
      val truee = fromString "true"

      (* exception constructors *)
      val bind = fromString "Bind"
      val match = fromString "Match"

      val excons = [bind, match]
	 
      fun isPrimExcon c = List.exists(excons, fn c' => equals(c, c'))
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Longid(S: LONGID_STRUCTS): LONGID =
struct

open S

datatype node = T of {strids: Strid.t list,
		      id: Id.t}
type node' = node
structure Wrap = Region.Wrap
open Wrap
type t = node Wrap.t
type obj = t

fun split id =
   let val T{strids, id, ...} = node id
   in (strids, id)
   end

fun prepend(id, strid) =
   let val (T{strids, id}, region) = dest id
   in makeRegion(T{strids = strid :: strids, id = id},
		 region)
   end

fun prepends(id, strids') =
   let val (T{strids, id}, region) = dest id
   in makeRegion(T{strids = strids' @ strids, id = id},
		 region)
   end

fun isLong id =
   let val T{strids, ...} = node id
   in not(List.isEmpty strids)
   end

fun toId id =
   let val T{id, ...} = node id
   in id
   end
   
val equals =
   fn (id, id') =>
   let val T{strids=ss, id=i} = node id
      val T{strids=ss', id=i'} = node id'
   in List.equals(ss, ss', Strid.equals) andalso Id.equals(i, i')
   end
   
fun long(strids, id) = make(T{strids = strids, id = id})
fun short id = long([], id)

local open Layout
in
   fun layout id =
      let val T{strids, id} = node id
      in seq[case strids of
		[] => empty
	      | _ => seq[seq(separate(List.map(strids, Strid.layout), ".")),
			 str "."],
		   Id.layout id]
      end
end

val toString = Layout.toString o layout

fun fromString(longid: string, region: Region.t): t =
   let val (strids, id) = List.splitLast(String.split(longid, #"."))
   in makeRegion(T{strids = List.map(strids, Strid.fromString),
		   id = Id.fromString id},
		 region)
   end

val className = Id.className

val bogus = short Id.bogus
   
fun unbound(x: t): unit =
	 Control.error
	 (region x,
	  let open Layout
	  in seq[str "unbound ", str className, str " ", layout x]
	  end)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AstId(S: AST_ID_STRUCTS): AST_ID =
   struct
      open S

      datatype t = T of {string: String.t,
			 hash: word,
			 region: Region.t}
      type obj = t
      type node' = string
      type region = Region.t

      fun makeRegion(s, r) = T{string = s,
			      hash = HashString.hashString s,
			      region = r}
      fun makeRegion'(s, x, y) =
	 makeRegion(s, Region.T{left = x, right = y})
      fun make s = makeRegion(s, Region.bogus)
      val fromString = make
      fun dest(T{string, region, ...}) = (string, region)

      val bogus = fromString "<bogus>"

      fun toString(T{string, ...}) = string
      val node = toString
      fun region(T{region, ...}) = region
      fun hash(T{hash, ...}) = hash

      val left = Region.left o region
      val right = Region.left o region

      local fun unary (f: string -> 'a): t -> 'a = f o toString
      in val layout = unary String.layout
      end
	    
      local
	 fun binary (f: string * string -> 'a) (x :t, y: t): 'a =
	    f(toString x, toString y)
      in val op < = binary String.<
	 val op > = binary String.>
	 val op >= = binary String.>=
	 val op <= = binary String.<=
	 fun max(x, y) = if x > y then x else y
	 fun min(x, y) = if x < y then x else y
	 val compare = binary String.compare
      end

      val equals = fn (T{string=s, hash=h, ...}, T{string=s', hash=h', ...}) =>
	 h = h' andalso String.equals(s, s')

      fun unbound(x: t): unit =
	 Control.error
	 (region x,
	  let open Layout
	  in seq[str "unbound ", str className, str " ", layout x]
	  end)
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AstAtoms(S: AST_ATOMS_STRUCTS) :> AST_ATOMS = 
struct

open S
structure Wrap = Region.Wrap

structure Tycon =
   struct
      structure Id = AstId(val className = "tycon")
      open Id

      structure P = PrimTycons(open Id)
      open P
   end

structure Var = AstId(val className = "variable")
   
structure Con =
   struct
      structure Id = AstId(val className = "constructor")
      open Id

      structure P = PrimCons(Id)
      open P
   end

structure Sigid = AstId(val className = "signature")
structure Strid = AstId(val className = "structure")
structure Fctid = AstId(val className = "functor")

structure Vid =
   struct
      structure I = AstId(val className = "variable")
      open I
      val fromCon = fromString o Con.toString
      val fromVar = fromString o Var.toString
      val toCon = Con.fromString o toString
      val toVar = Var.fromString o toString
      val toFctid = Fctid.fromString o toString
      val toStrid = Strid.fromString o toString
      val bind = fromCon Con.bind
      val cons = fromCon Con.cons
      val falsee = fromCon Con.falsee
      val match = fromCon Con.match
      val nill = fromCon Con.nill
      val reff = fromCon Con.reff
      val truee = fromCon Con.truee
   end

structure Longtycon =
   struct
      structure T = Longid(structure Strid = Strid
			   structure Id = Tycon)
      open T
      val arrow = short Tycon.arrow
      val exn = short Tycon.exn
      fun isArrow t =
	 case split t of
	    ([], tycon) => Tycon.equals(tycon, Tycon.arrow)
	  | _ => false
   end
structure Longvar = Longid(structure Strid = Strid
			   structure Id = Var)
structure Longcon =
   struct
      structure L = Longid(structure Strid = Strid
			   structure Id = Con)
      open L
      val nill = short Con.nill
      val cons = short Con.cons
   end

structure Longstrid = Longid(structure Strid = Strid
			     structure Id = Strid)
structure Longfctid = Longid(structure Strid = Strid
			     structure Id = Fctid)

structure Longvid =
   struct
      structure L = Longid(structure Strid = Strid
			   structure Id = Vid)
      open L
      fun fromLongcon(c: Longcon.t): t =
	 let val (strids, id) = Longcon.split c
	 in make(T{strids = strids, id = Vid.fromCon id})
	 end
      local
	 fun to (make,node, conv) x =
	    let val (T{strids, id}, region) = dest x
	    in make(node{strids = strids, id =  conv id}, region)
	    end
      in val toLongvar = to(Longvar.makeRegion, Longvar.T, Vid.toVar)
	 val toLongcon = to(Longcon.makeRegion, Longcon.T, Vid.toCon)
	 val toLongstrid = to(Longstrid.makeRegion, Longstrid.T, Vid.toStrid)
	 val toLongfctid = to(Longfctid.makeRegion, Longfctid.T, Vid.toFctid)
      end

      val bind = short Vid.bind
      val cons = short Vid.cons
      val falsee = short Vid.falsee
      val match = short Vid.match
      val nill = short Vid.nill
      val reff = short Vid.reff
      val truee = short Vid.truee
   end

open Layout

structure Type =
   struct
      structure Record = SortedRecord
      open Wrap
      datatype node =
	 Var of Tyvar.t
       | Con of Longtycon.t * t list
       | Record of t Record.t
      withtype t = node Wrap.t
      type node' = node
      type obj = t

      val var = make o Var
	 
      val record = make o Record
      val tuple = record o Record.tuple
      val unit = tuple[]

      fun con(c: Tycon.t, ts: t list): t =
	 if Tycon.equals(c, Tycon.tuple)
	    then tuple ts
	 else make(Con(Longtycon.short c, ts))

      fun arrow(t1, t2) = con(Tycon.arrow, [t1, t2])

      val exn = con(Tycon.exn, [])

      fun layoutApp(tycon, args, layoutArg) =
	 case args of
	    [] => tycon
	  | [arg] => seq[layoutArg arg, str " ", tycon]
	  | _ => seq[Layout.tuple(List.map(args, layoutArg)), str " ", tycon]
	       
      fun layout ty =
	 case node ty of
	    Var v => Tyvar.layout v
	  | Con(c, tys) =>
	       if Longtycon.equals(c, Longtycon.arrow)
		  then (case tys of
			   [t1, t2] => paren(mayAlign[layout t1,
						      seq[str "-> ", layout t2]])
			 | _ => Error.error "non-binary -> tyc")
	       else layoutApp(Longtycon.layout c, tys, layout)
	  | Record r => Record.layout{record = r,
				      separator = ":", extra = "",
				      layoutElt = layout,
				      layoutTuple = layoutTupleTy}
      and layoutTupleTy tys =
	 case tys of
	    [] => str "unit"
	  | [t] => layout t
	  | _ => paren(mayAlign(separateLeft(List.map(tys, layout), "* ")))
	       
      fun layoutOption ty =
	 case ty of
	    NONE => empty
	  | SOME ty => seq[str " of ", layout ty]
   end

fun bind(x, y) = mayAlign[seq[x, str " ="], y]

val layouts = List.map
   
fun 'a layoutAnds(prefix: string,
		  xs: 'a list, 
		  layoutX: Layout.t * 'a -> Layout.t): Layout.t =
   case xs of
      x :: xs => align(layoutX(str(prefix ^ " "), x)
		       :: List.map(xs, fn x => layoutX(str "and ", x)))
    | [] => empty

datatype bindStyle = OneLine | Split of int

fun 'a layoutBind(bind: string,
		  layout: 'a -> bindStyle * Layout.t * Layout.t)
   (prefix: Layout.t, x: 'a): Layout.t =
   let val (style, lhs, rhs) = layout x
      val lhs = seq[prefix, lhs, str " " , str bind]
   in case style of
      OneLine => seq[lhs, str " ", rhs]
    | Split indentation => align[lhs, indent(rhs, indentation)]
   end

fun layoutAndsBind(prefix, bind, xs, layout) =
   layoutAnds(prefix, xs, layoutBind(bind, layout))
   
(*---------------------------------------------------*)
(*                      TypBind                      *)
(*---------------------------------------------------*)

structure TypBind =
   struct
      datatype node =
	 T of {tycon: Tycon.t,
	       def: Type.t,
	       tyvars: Tyvar.t list} list
      open Wrap
      type t = node Wrap.t
      type node' = node
      type obj = t

      fun layout t =
	 let val T l = node t
	 in layoutAndsBind
	    ("type", "=", l, fn {tycon, def, tyvars} =>
	     (OneLine,
	      Type.layoutApp(Tycon.layout tycon, tyvars, Tyvar.layout),
	      Type.layout def))
	 end
      val empty = make(T[])
   end

(*---------------------------------------------------*)
(*                      DatBind                      *)
(*---------------------------------------------------*)

structure DatBind =
   struct
      datatype node =
	 T of {datatypes: {tyvars: Tyvar.t list,
			    tycon: Tycon.t,
			    cons: (Con.t * Type.t option) list} list,
	       withtypes: TypBind.t}

      open Wrap
      type t = node Wrap.t
      type node' = node
      type obj = t
	 
      fun layout(prefix, d) =
	 let val T{datatypes, withtypes} = node d
	 in align
	    [layoutAndsBind
	     (prefix, "=", datatypes, fn {tyvars, tycon, cons} =>
	      (OneLine,
	       Type.layoutApp(Tycon.layout tycon, tyvars, Tyvar.layout),
	       alignPrefix("| ", List.map(cons, fn (c, to) =>
					  seq[Con.layout c,
					      Type.layoutOption to])))),
	     case TypBind.node withtypes of
		TypBind.T[] => empty
	      | _ => seq[str "with", TypBind.layout withtypes]]
	 end
   end

structure DatatypeRhs =
   struct
      datatype node =
	 DatBind of DatBind.t
       | Repl of {lhs: Tycon.t, rhs: Longtycon.t}

      open Wrap
      type t = node Wrap.t
      type node' = node
      type obj = t
	 
      fun layout d =
	 case node d of
	    DatBind d => DatBind.layout("datatype", d)
	  | Repl{lhs, rhs} =>
	       seq[str "datatype ", Tycon.layout lhs,
		   str " = datatype ", Longtycon.layout rhs]
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_STRUCTS =
   sig
   end

signature AST =
   sig
      include AST_CORE

      structure Sigexp :
	 sig
	    type spec

	    type t
	    datatype node =
	       Var of Sigid.t
             | Where of t * {tyvars: Tyvar.t list,
			     longtycon: Longtycon.t,
			     ty: Type.t} list
	     | Spec of spec

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

            val var: Sigid.t -> t
	    val wheree: t * {tyvars: Tyvar.t list,
			     longtycon: Longtycon.t,
			     ty: Type.t} list -> t
	    val spec: spec -> t

	    val layout: t -> Layout.t
	 end

      structure SigConst :
	 sig
	    datatype t =
	       None
	     | Transparent of Sigexp.t
	     | Opaque of Sigexp.t
	 end

      structure Equation :
	 sig
	    type t
	    datatype node =
	       Type of Longtycon.t list
	     | Structure of Longstrid.t list
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	 end

      structure Spec :
	 sig
	    type t
	    datatype node =
	      Val of (Var.t * Type.t) list
	     | Type of {tyvars: Tyvar.t list,
			tycon: Tycon.t,
			ty: Type.t option} list
	     | Eqtype of {tyvars: Tyvar.t list,
			  tycon: Tycon.t,
			  ty: Type.t option} list
	     | Datatype of DatatypeRhs.t
	     | Exception of (Con.t * Type.t option) list
	     | Structure of (Strid.t * Sigexp.t) list
	     | IncludeSigexp of Sigexp.t
	     | IncludeSigids of Sigid.t list
	     | Empty
	     | Seq of t * t
	     | Sharing of {spec: t,
			   equations: Equation.t list}

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

	    val layout: t -> Layout.t
	 end
      sharing type Spec.t = Sigexp.spec

      structure Strexp :
	 sig
	    type strdec

	    type t
	    datatype node =
	       Var of Longstrid.t
	     | Struct of strdec
             | Constrained of t * SigConst.t
	     | App of Fctid.t * t
	     | Let of strdec * t

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

            val var: Longstrid.t -> t
	    val structt: strdec -> t
	    val constrained: t * SigConst.t -> t
	    val app: Fctid.t * t -> t
	    val lett: strdec * t -> t
	       
	    val layout: t -> Layout.t
	 end

      structure Strdec :
	 sig
	    type t
	    datatype node =
	       Structure of {name: Strid.t,
			     def: Strexp.t,
			     constraint: SigConst.t} list
	     | Seq of t list
	     | Local of t * t
	     | Core of Dec.t

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

            val structuree: {name: Strid.t,
			      def: Strexp.t,
			      constraint: SigConst.t} list -> t
	    val seq: t list -> t
	    val locall: t * t -> t
            val core: Dec.t -> t
	    val openn: Longstrid.t list -> t
	       
	    val layout: t -> Layout.t
	 end
      sharing type Strdec.t = Strexp.strdec

      structure FctArg :
	 sig
	    type t
	    datatype node =
	       Structure of Strid.t * Sigexp.t
	     | Spec of Spec.t
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	 end
      
      structure Topdec :
	 sig
	    type t
	    datatype node =
	       Strdec of Strdec.t
	     | Signature of (Sigid.t * Sigexp.t) list
	     | Functor of {name: Fctid.t,
			   arg: FctArg.t,
			   result: SigConst.t,
			   body: Strexp.t} list

	    include WRAPPED sharing type node' = node
			    sharing type obj = t

            val strdec: Strdec.t -> t
	    val signaturee: (Sigid.t * Sigexp.t) list -> t
	    val functorr: {name: Fctid.t,
			    arg: FctArg.t,
			    result: SigConst.t,
			    body: Strexp.t} list -> t
	       
	    val layout: t -> Layout.t
	 end

      structure Program :
	 sig
	    datatype t = T of Topdec.t list

	    val size: t -> int
	    val layout: t -> Layout.t
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Ast(S: AST_STRUCTS): AST = 
struct

open S

structure Const = AstConst()
structure Field = Field()
structure Record = Record(val isSorted = false
			  structure Field = Field)
structure SortedRecord = Record(val isSorted = true
				structure Field = Field)
structure Tyvar = Tyvar()
   
structure AstAtoms = AstAtoms(structure Const = Const
			      structure Record = Record
			      structure SortedRecord = SortedRecord
			      structure Tyvar = Tyvar)

structure AstCore = AstCore(open AstAtoms)

open AstCore Layout
   
val layouts = List.map
structure Wrap = Region.Wrap
val node = Wrap.node
   
structure Equation =
   struct
      open Wrap
      datatype node =
	 Type of Longtycon.t list
       | Structure of Longstrid.t list
      type t = node Wrap.t
      type node' = node
      type obj = t

      fun layout eq =
	 case node eq of
	    Type longtycons =>
	       seq(str "sharing type "
		   :: separate(List.map(longtycons, Longtycon.layout), " = "))
	  | Structure longstrids =>
	       seq(str "sharing "
		   :: separate(List.map(longstrids, Longstrid.layout), " = "))
   end

type typedescs = {tyvars: Tyvar.t list,
		  tycon: Tycon.t,
		  ty: Type.t option} list

datatype sigexpNode =
   Var of Sigid.t
 | Where of sigexp * {tyvars: Tyvar.t list,
		      longtycon: Longtycon.t,
		      ty: Type.t} list
 | Spec of spec
and sigConst =
   None
  | Transparent of sigexp
  | Opaque of sigexp
and specNode =
   Empty
  | Seq of spec * spec
  | Structure of (Strid.t * sigexp) list
  | Type of typedescs
  | Eqtype of typedescs
  | Val of (Var.t * Type.t) list
  | Datatype of DatatypeRhs.t
  | Exception of (Con.t * Type.t option) list
  | IncludeSigexp of sigexp
  | IncludeSigids of Sigid.t list
  | Sharing of {spec: spec, equations: Equation.t list}
withtype spec = specNode Wrap.t
and sigexp = sigexpNode Wrap.t

fun layoutTypedescs(prefix, typedescs) =
   layoutAnds(prefix, typedescs, fn (prefix, {tyvars, tycon, ty}) =>
	      seq[prefix,
		  Type.layoutApp(Tycon.layout tycon, tyvars, Tyvar.layout),
		  case ty of
		     NONE => empty
		   | SOME ty => seq[str " = ", Type.layout ty]])

fun layoutSigexp(e: sigexp): Layout.t =
   case node e of
      Var s => Sigid.layout s
    | Where(e, ws) =>
	 let val e = layoutSigexp e
	 in case ws of
	    [] => e
	  | _ => 
	       seq[e, 
		   layoutAndsBind
		   (" where", "=", ws, fn {tyvars, longtycon, ty} =>
		    (OneLine,
		     seq[str "type ",
			 Type.layoutApp
			 (Longtycon.layout longtycon, tyvars,
			  Tyvar.layout)],
		     Type.layout ty))]
	 end
    | Spec s => align[str "sig",
		      indent(layoutSpec s, 3),
		      str "end"]

and layoutSigConst sigConst =
   case sigConst of
      None => empty
    | Transparent s => seq[str ": ", layoutSigexp s]
    | Opaque s => seq[str " :> ", layoutSigexp s]

and layoutSpec(s: spec): t =
   case node s of
      Empty => empty
    | Seq(s, s') => align[layoutSpec s, layoutSpec s']
    | Structure l =>
	 layoutAndsBind("structure", ":", l, fn (strid, sigexp) =>
			(case node sigexp of
			    Var _ => OneLine
			  | _ => Split 3,
			 Strid.layout strid,
			 layoutSigexp sigexp))
    | Type typedescs => layoutTypedescs("type", typedescs)
    | Eqtype typedescs => layoutTypedescs("eqtype", typedescs)
    | Val sts =>
	 layoutAndsBind
	 ("val", ":", sts, fn (x, t) => (OneLine, Var.layout x, Type.layout t))
    | Datatype rhs => DatatypeRhs.layout rhs
    | Exception sts =>
	 layoutAnds
	 ("exception", sts, fn (prefix, (c, to)) => seq[prefix,
						      Con.layout c,
						      Type.layoutOption to])
    | IncludeSigexp s => seq[str "include ", layoutSigexp s]
    | IncludeSigids sigids =>
	 seq(str "include "
	     :: separate(List.map(sigids, Sigid.layout), " "))
    | Sharing{spec, equations} =>
	 align[layoutSpec spec,
	       align(List.map(equations, Equation.layout))]

structure Sigexp =
   struct
      open Wrap
      type spec = spec
      type t = sigexp
      datatype node = datatype sigexpNode
      type node' = node
      type obj = t
	 
      fun wheree(sigexp: t, wherespecs): t =
	 case wherespecs of
	    [] => sigexp
	  | _ => makeRegion(Where(sigexp, wherespecs),
			    region sigexp)

      val spec = make o Spec
      val var = make o Var
	 
      fun varString(s: string): t = make(Var(Sigid.fromString s))
						 
      val layout = layoutSigexp
   end

structure SigConst =
   struct
      datatype t = datatype sigConst
      val layout = layoutSigConst
   end

structure Spec =
   struct
      open Wrap
      datatype node = datatype specNode
      type t = spec
      type node' = node
      type obj = t
	 
      val empty = make Empty
	 	 
      val layout = layoutSpec
   end

(*---------------------------------------------------*)
(*                Strdecs and Strexps                *)
(*---------------------------------------------------*)

datatype strdecNode =
   Structure of {name: Strid.t,
		 def: strexp,
		 constraint: SigConst.t} list
  | Local of strdec * strdec
  | Seq of strdec list
  | Core of Dec.t
and strexpNode =
   Var of Longstrid.t
  | Struct of strdec
  | Constrained of strexp * SigConst.t
  | App of Fctid.t * strexp
  | Let of strdec * strexp
withtype strexp = strexpNode Wrap.t
and strdec = strdecNode Wrap.t

fun layoutStrdec d =
   case node d of
      Structure strbs =>
	 layoutAndsBind("structure", "=", strbs, fn {name, def, constraint} =>
			(case node def of
			    Var _ => OneLine
			  | _ => Split 3,
			 seq[Strid.layout name, SigConst.layout constraint],
			 layoutStrexp def))
    | Local(d, d') => layoutLocal(layoutStrdec d, layoutStrdec d')
    | Seq ds => align(layoutStrdecs ds)
    | Core d => Dec.layout d

and layoutStrdecs ds = layouts(ds, layoutStrdec)
	 
and layoutStrexp exp =
   case node exp of
      Var s => Longstrid.layout s
    | Struct d => align[str "struct",
			indent(layoutStrdec d, 3),
			str "end"]
    | Constrained(e, c) => mayAlign[layoutStrexp e, SigConst.layout c]
    | App(f, e) =>
	 seq[Fctid.layout f, str "(", layoutStrexp e, str ")"]
    | Let(dec, strexp) => layoutLet(layoutStrdec dec, layoutStrexp strexp)
	 
structure Strexp =
   struct
      open Wrap
      type strdec = strdec
      type t = strexp
      datatype node = datatype strexpNode
      type node' = node
      type obj = t
      val var = make o Var
      val structt = make o Struct
      val constrained = make o Constrained
      val app = make o App
      val lett = make o Let
      val layout = layoutStrexp
   end

structure Strdec =
   struct
      open Wrap
      type t = strdec
      datatype node = datatype strdecNode
      type node' = node
      type obj = t

      val structuree = make o Structure
      val locall = make o Local
      val core = make o Core
      val seq = make o Seq

      val openn = core o Dec.openn

      val layout = layoutStrdec
   end

structure FctArg =
   struct
      open Wrap
      datatype node =
	 Structure of Strid.t * Sigexp.t
       | Spec of Spec.t
      type t = node Wrap.t
      type node' = node
      type obj = t

      fun layout a =
	 case node a of
	    Structure(strid, sigexp) =>
	       seq[Strid.layout strid, str ": ", Sigexp.layout sigexp]
	  | Spec spec => Spec.layout spec
   end

(*---------------------------------------------------*)
(*                      Topdec                       *)
(*---------------------------------------------------*)

structure Topdec =
   struct
      open Wrap
      datatype node =
	 Strdec of Strdec.t
       | Signature of (Sigid.t * Sigexp.t) list
       | Functor of {name: Fctid.t,
		     arg: FctArg.t,
		     result: SigConst.t,
		     body: Strexp.t} list
      type t = node Wrap.t
      type node' = node
      type obj = t
	 
      fun layout d =
	 case node d of
	    Strdec d => Strdec.layout d
	  | Signature sigbs =>
	       layoutAndsBind("signature", "=", sigbs, fn (name, def) =>
			      (case Sigexp.node def of
				  Sigexp.Var _ => OneLine
				| _ => Split 3,
			       Sigid.layout name,
			       Sigexp.layout def))
	  | Functor fctbs =>
	       layoutAndsBind("functor", "=", fctbs,
			      fn {name, arg, result, body} =>
	                      (Split 0,
			       seq[Fctid.layout name,
				   paren(FctArg.layout arg),
				   layoutSigConst result],
			       layoutStrexp body))

      val strdec = make o Strdec
      val signaturee = make o Signature
      val functorr = make o Functor
   end			
      
structure Program =
   struct
      datatype t = T of Topdec.t list

      fun layout(T ds) = Layout.align(List.map(ds, Topdec.layout))

      fun size(T ds): int =
	 let
	    open Dec Exp Strexp Strdec Topdec
	    val n = ref 0
	    fun inc() = n := 1 + !n

	    fun dec(d: Dec.t): unit =
	       case Dec.node d of
		  Val(_, pes) => List.foreach(pes, exp o #2)
		| Valrec(_, fs) => List.foreach(fs, fn {rules, ...} =>
						List.foreach(rules, exp o #2))
		| Fun(_, ds) => List.foreach(ds, fn cs =>
					     List.foreach(cs, exp o #body))
		| Abstype{body, ...} => dec body
		| Exception cs => List.foreach(cs, fn _ => inc())
		| SeqDec ds => List.foreach(ds, dec)
		| Dec.Local(d, d') => (dec d; dec d')
		| _ => ()
		     
	    and exp(e: Exp.t): unit =
	       (inc();
		case Exp.node e of
		   Fn rs => rules rs
		 | FlatApp es => exps es
		 | Exp.App(e, e') => (exp e; exp e')
		 | Case(e, rs) => (exp e; rules rs)
		 | Exp.Let(d, e) => (dec d; exp e)
		 | Exp.Seq es => exps es
		 | Record r => Record.foreach(r, exp)
		 | List es => exps es
		 | Constraint(e, _) => exp e
		 | Handle(e, rs) => (exp e; rules rs)
		 | Raise e => exp e
		 | If(e1, e2, e3) => (exp e1; exp e2; exp e3)
		 | Andalso(e1, e2) => (exp e1; exp e2)
		 | Orelse(e1, e2) => (exp e1; exp e2)
		 | While{test, expr} => (exp test; exp expr)
		 | _ => ())

	    and exps es = List.foreach(es, exp)
		      
	    and rules rs = List.foreach(rs, exp o #2)

	    fun strdec d =
	       case Strdec.node d of
		  Structure ds => List.foreach(ds, fn {def, ...} => strexp def)
		| Seq ds => List.foreach(ds, strdec)
		| Local(d, d') => (strdec d; strdec d')
		| Core d => dec d
	    and strexp e =
	       case Strexp.node e of
		  Struct d => strdec d
		| Constrained(e, _) => strexp e
		| App(_, e) => strexp e
		| Let(d, e) => (strdec d; strexp e)
		| _ => ()

	    fun topdec d =
	       case Topdec.node d of
		  Strdec d => strdec d
		| Functor ds => List.foreach(ds, fn {body, ...} => strexp body)
		| _ => ()
	 in List.foreach(ds, topdec);
	    !n
	 end
   end

end
signature MONO_CONTAINER =
   sig
      structure Element: T

      include T

      val size: t -> Int.t
      val foreach: t * (Element.t -> unit) -> unit
      val forall: t * (Element.t -> bool) -> bool
      val exists: t * (Element.t -> bool) -> bool
   end
signature SET_STRUCTS =
   sig
      structure Element: T
   end

signature SET =
   sig
      include MONO_CONTAINER

      (* include ORDER *)
      val < : t * t -> bool
      val <= : t * t -> bool
      val > : t * t -> bool
      val >= : t * t -> bool
      (* *)

      val + : t * t -> t
      val - : t * t -> t
      val intersect: t * t -> t

      val unions: t list -> t

      val subset: t * (Element.t -> bool) -> t
      val subsetSize: t * (Element.t -> bool) -> int
      val map: t * (Element.t -> Element.t) -> t
      val replace: t * (Element.t -> Element.t option) -> t

      val fold: t * 'a * (Element.t * 'a -> 'a) -> 'a

      val empty: t
      val singleton: Element.t -> t
      val add: t * Element.t -> t
      val remove: t * Element.t -> t
	 
      val contains: t * Element.t -> bool
      val isEmpty: t -> bool

      val power: t -> t list
      val subsets: t * Int.t -> t list

      val areDisjoint: t * t -> bool

      (* list must contain no duplicates *)
      val fromList: Element.t list -> t
      val toList: t -> Element.t list
   end
functor UnorderedSet(Element: T): SET =
struct

structure Element = Element
   
open List

type t = Element.t List.t
    
val {empty, singleton, size, equals, <=, >=, <, >, +, -, intersect, unions,
     add, remove, contains, power, subsets, areDisjoint, subset, subsetSize,
     map, replace, layout} =
   List.set{equal = Element.equals,
	    layout = Element.layout}
   
val fromList = fn l => List.fold(l, empty, fn (x, s) => add(s, x))
val toList = fn x => x

end
signature MONO_ENV_STRUCTS =
   sig
      structure Domain: T
      structure Range: T
   end

signature BASIC_MONO_ENV =
   sig
      include MONO_ENV_STRUCTS
	 
      type t
      val fromList: (Domain.t * Range.t) list -> t
      val toList: t -> (Domain.t * Range.t) list
      val peek: t * Domain.t -> Range.t option
      val extend: t * Domain.t * Range.t -> t
   end

signature MONO_ENV = 
   sig
      include BASIC_MONO_ENV
      
      val size: t -> int
      val lookup: t * Domain.t -> Range.t
      val new: Domain.t list * (Domain.t -> Range.t) -> t
      val empty: t
      val single: Domain.t * Range.t -> t
      val isEmpty: t -> bool
      val domain: t -> Domain.t list
      val + : t * t -> t
      val plus: t list -> t
      val multiExtend: t * Domain.t list * Range.t list -> t
      val restrict: t * Domain.t list -> t
      val remove: t * Domain.t -> t
      val map: t * (Range.t -> Range.t) -> t
      val mapi: t * (Domain.t * Range.t -> Range.t) -> t
      val fold: t * 'a * (Range.t * 'a -> 'a) -> 'a
      val foldi: t * 'a * (Domain.t * Range.t * 'a -> 'a) -> 'a
      val foreach: t * (Range.t -> unit) -> unit
      val foreachi: t * (Domain.t * Range.t -> unit) -> unit
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
functor BasicEnvToEnv(S: BASIC_MONO_ENV): MONO_ENV =
struct

open S

val isEmpty = List.isEmpty o toList

fun layout e =
   let open Layout
   in seq[str "[",
	  align(List.map(toList e, fn (d, r) =>
			 seq[Domain.layout d, str " -> ", Range.layout r])),
	  str"]"]
   end

fun maybeLayout(env,name) =
   if isEmpty env then Layout.empty
   else let open Layout
	in seq[str name, str " = ", layout env]
	end

val size = List.length o toList
   
val empty = fromList []

fun domain e = List.map(toList e, #1)

fun single(d, r) = extend(empty, d, r)

fun new(ds, f) = fromList(List.map(ds, fn d => (d, f d)))

fun map(e, f) = fromList(List.map(toList e, fn (d, r) => (d, f r)))

fun mapi(e, f) = fromList(List.map(toList e, fn (d, r) => (d, f(d, r))))
   
fun env + env' =
   let fun f() = List.fold(toList env', env, fn ((d, r), env) => extend(env, d, r))
   in f() ; f()
   end

fun plus es = List.fold(es, empty, fn (e, accum) => accum + e)

val plus = Trace.trace("plus", List.layout layout, layout) plus

fun remove(env, d) =
   fromList(List.remove(toList env, fn (d', _) => Domain.equals(d, d')))

fun lookup(env, d) = case peek(env, d) of
   SOME r => r
 | NONE => (Layout.output(Domain.layout d, Out.error) ;
	    Out.newline Out.error ;
	    Error.error "lookup")

fun restrict(env, ds) = new(ds, fn d => lookup(env, d))

fun multiExtend(env, [], []) = env
  | multiExtend(env, d :: ds, r :: rs) = multiExtend(extend(env, d, r), ds, rs)
  | multiExtend _ = Error.error "multiExtend"

fun fold(e, b, f) = List.fold(toList e, b, fn ((_, r), b) => f(r, b))
fun foldi(e, b, f) = List.fold(toList e, b, fn ((d, r), b) => f(d, r, b))

fun foreach(e, f) = List.foreach(toList e, f o #2)
fun foreachi(e, f) = List.foreach(toList e, f)

fun forall(e, f) = List.forall(toList e, f o #2)
fun foralli(e, f) = List.forall(toList e, f)
   
val equals =
   fn (e1, e2) =>
   Int.equals(size e1, size e2)
   andalso foralli(e1, fn (d, r) =>
		   case peek(e2, d) of
		      NONE => false
		    | SOME r' => Range.equals(r, r'))
   
end
signature ENV_STRUCTS =
   sig
      structure Domain: T
   end

signature ENV =
   sig
      include ENV_STRUCTS
      
      type 'a t

      val size: 'a t -> int
      val new: Domain.t list * (Domain.t -> 'a) -> 'a t
      val fromList: (Domain.t * 'a) list -> 'a t
      val toList: 'a t -> (Domain.t * 'a) list
      val empty: unit -> 'a t
      val single: Domain.t * 'a -> 'a t
      val isEmpty: 'a t -> bool

      val domain: 'a t -> Domain.t list
	 
      val singleton: Domain.t * 'a -> 'a t
      val extend: 'a t * Domain.t * 'a -> 'a t
      val + : 'a t * 'a t -> 'a t
      val plus: 'a t list -> 'a t
      val multiExtend: 'a t * Domain.t list * 'a list -> 'a t

      val peek: 'a t * Domain.t -> 'a option
      val lookup: 'a t * Domain.t -> 'a

      val restrict: 'a t * Domain.t list -> 'a t
      val remove: 'a t * Domain.t -> 'a t

      val map: 'a t * ('a -> 'b) -> 'b t
      val mapi: 'a t * (Domain.t * 'a -> 'b) -> 'b t
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldi: 'a t * 'b * (Domain.t * 'a * 'b -> 'b) -> 'b
      val foreach: 'a t * ('a -> unit) -> unit
      val foreachi: 'a t * (Domain.t * 'a -> unit) -> unit
      val forall: 'a t * ('a -> bool) -> bool
      val foralli: 'a t * (Domain.t * 'a -> bool) -> bool

      val equals: ('a * 'a -> bool) -> 'a t * 'a t -> bool
      val layout: ('a -> Layout.t) -> 'a t -> Layout.t
      val maybeLayout: string * ('a -> Layout.t) -> 'a t -> Layout.t
   end
functor MakeMonoEnv(S: sig
			   structure Env: ENV
			   structure Range: T
			end): MONO_ENV =
struct

open S
open Env

type t = Range.t t

val empty: t = empty()

val equals = equals Range.equals

val layout = layout Range.layout
val maybeLayout = fn (env,name) => maybeLayout (name, Range.layout) env

end

(* THIS ISN'T HERE BECAUSE (no surprise) there is an NJ BUG

functor MonoEnv(S: MONO_ENV_STRUCTS): MONO_ENV =
   MakeMonoEnv(structure Env = Env(S)
	       structure Range = S.Range)
*)

functor MonoEnv(S: MONO_ENV_STRUCTS): MONO_ENV =
BasicEnvToEnv
(open S

 datatype t = T of (Domain.t * Range.t) List.t
 type range = Range.t

 fun eq d d' = Domain.equals(d, d')
 fun member(d, ds) = List.exists(ds, eq d)
    
 fun invariant(T drs) =
    let fun loop([], _) = true
	  | loop((d, _) :: drs, seen) =
	    not(member(d, seen)) andalso loop(drs, d :: seen)
    in loop(drs, [])
    end

 val fromList = T
    
 fun toList(T l) = l
    
 fun equalTo d (d', _) = Domain.equals(d, d')

 fun extend(T drs, d, r) =
    T(List.cons((d, r), List.remove(drs, equalTo d)))

 fun peek(T l, d) =
    case List.peek(l, equalTo d) of
       NONE => NONE
     | SOME (_, r) => SOME r
	  )

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ID_STRUCTS =
   sig
      structure AstId: AST_ID
      val noname: string
   end

signature ID =
   sig
      include ID_STRUCTS
      include T

      (* raw constructor - doesn't add uniquefying suffix *)
      val fromString: string -> t

      (* raw destructor *)
      val originalName: t -> string
	 
      (* constructors *)
      val bogus: t
      val new: t -> t            (* with the same prefix *)
      val newString: string -> t (* given prefix *)
      val newNoname: unit -> t   (* prefix is "noName" *)

      (* conversions *)
      val fromAst: AstId.t -> t
      val fromAsts: AstId.t list -> t list
      val toAst: t -> AstId.t
      val toAsts: t list -> AstId.t list

      val toString: t -> string

      val sameName: t * t -> bool

      val setName: t * string -> unit

      val plist: t -> PropertyList.t
      val clear: t -> unit
   end

signature HASH_ID =
   sig
      include ID

      val hash: t -> word 
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature VAR_STRUCTS =
   sig
      structure AstId: AST_ID
   end

signature VAR =
   sig
      (*include VAR_STRUCTS*)
      include HASH_ID
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure IdCounter:
   sig
      val getCounter: string -> Counter.t
   end =
   struct
      structure Hash = HashTable
      val hash = HashString.hashString
      val initSize = 10000
      exception NotFound
      val table: (string, Counter.t) Hash.hash_table =
	 Hash.mkTable (hash, String.equals) (initSize, NotFound)

      fun getCounter s =
	 case Hash.find table s of
	    SOME c => c
	  | NONE => let val c = Counter.new 0
		    in Hash.insert table (s, c) ; c
		    end
   end

functor Id(S: ID_STRUCTS): ID =
struct

open S

structure Plist = PropertyList
   
datatype t = T of {originalName: string,
		   counter: Counter.t,
		   printName: string option ref,
		   plist: Plist.t}

fun plist(T{plist, ...}) = plist
fun originalName(T{originalName, ...}) = originalName

fun setName(T{printName, ...}, s) = printName := SOME s
   
fun toString(T{printName, originalName, counter, ...}) =
   case !printName of
      NONE =>
	 let
	    val s =
	       concat[String.translate (fn #"'" => "P" | c => str c) originalName,
		      "_",
		      Int.toString(Counter.next counter)]
	 in printName := SOME s;
	    s
	 end
    | SOME s => s

val layout = String.layout o toString
   
fun sameName(id, id') = 
   String.equals(originalName id, originalName id')

fun equals(id, id') = Plist.equals(plist id, plist id')

fun fromString s =
   T{originalName = s,
     counter = IdCounter.getCounter s,
     printName = ref(SOME s),
     plist = Plist.new()}
   
fun isOperator s =
   let val c = String.sub(s, 0)
   in not(Char.isAlpha c orelse Char.equals(c, #"'"))
   end

fun newString s =
   let val s = if isOperator s then "operator" else s
      val counter = IdCounter.getCounter s
   in fn () => T{originalName = s,
		 counter = counter,
		 printName = ref NONE,
		 plist = Plist.new()}
   end

val newNoname = newString noname

val newString = fn s => newString s ()
   
val bogus = newString "BOGUS"

fun new(T{originalName, counter, ...}) =
   T{originalName = originalName,
     counter = counter,
     printName = ref NONE,
     plist = Plist.new()}

val fromAst = newString o AstId.toString
fun fromAsts l = List.map(l, fromAst)
val toAst = AstId.fromString o toString
fun toAsts l = List.map(l, toAst)

val clear = Plist.clear o plist
   
end

functor HashId(S: ID_STRUCTS): HASH_ID =
struct

open S

structure Plist = PropertyList
   
datatype t = T of {originalName: string,
		   counter: Counter.t,
		   printName: string option ref,
		   hash: word,
		   plist: Plist.t}

fun hash(T{hash, ...}) = hash
fun plist(T{plist, ...}) = plist
fun originalName(T{originalName, ...}) = originalName

fun setName(T{printName, ...}, s) = printName := SOME s

fun toString(T{printName, originalName, counter, ...}) =
   case !printName of
      NONE =>
	 let
	    val s =
	       concat[String.translate (fn #"'" => "P" | c => str c) originalName,
		      "_",
		      Int.toString(Counter.next counter)]
	 in printName := SOME s;
	    s
	 end
    | SOME s => s

val layout = String.layout o toString
   
fun sameName(id, id') = 
   String.equals(originalName id, originalName id')

fun equals(id, id') = Plist.equals(plist id, plist id')

fun newHash() = Word.fromInt(Int.random())
   
fun fromString s =
   T{originalName = s,
     counter = IdCounter.getCounter s,
     printName = ref(SOME s),
     hash = newHash(),
     plist = Plist.new()}
   
fun isOperator s =
   let val c = String.sub(s, 0)
   in not(Char.isAlpha c orelse Char.equals(c, #"'"))
   end

fun newString s =
   let val s = if isOperator s then "operator" else s
      val counter = IdCounter.getCounter s
   in fn () => T{originalName = s,
		 counter = counter,
		 printName = ref NONE,
		 hash = newHash(),
		 plist = Plist.new()}
   end

val newNoname = newString noname

val newString = fn s => newString s ()
   
val bogus = newString "BOGUS"

fun new(T{originalName, counter, ...}) =
   T{originalName = originalName,
     counter = counter,
     printName = ref NONE,
     hash = newHash(),
     plist = Plist.new()}

val fromAst = newString o AstId.toString
fun fromAsts l = List.map(l, fromAst)
val toAst = AstId.fromString o toString
fun toAsts l = List.map(l, toAst)

val clear = Plist.clear o plist
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Var(S: VAR_STRUCTS): VAR = 
struct

open S

structure V = HashId(structure AstId = AstId
		     val noname = "x")
open V
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYCON_STRUCTS = 
   sig
      structure AstId: AST_ID
   end

signature TYCON =
   sig
      include HASH_ID
      include PRIM_TYCONS where type tycon = t

      (* delete any info on tycon property lists *)
      val clearPrimitive: unit -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature UNARY_TYCON_STRUCTS = 
   sig
      structure Tycon: TYCON
   end

signature UNARY_TYCON = 
   sig
      include UNARY_TYCON_STRUCTS
      
      datatype t = Ref | Array | Vector

      val toTycon: t -> Tycon.t
      val toString: t -> string
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor UnaryTycon(S: UNARY_TYCON_STRUCTS): UNARY_TYCON = 
struct

open S

datatype t = Ref | Array | Vector
   
val toTycon =
   fn Ref => Tycon.reff
    | Array => Tycon.array
    | Vector => Tycon.vector
	 
val toString =
   fn Ref => "Ref"
    | Array => "Array"
    | Vector => "Vector"
	 
val equals: t * t -> bool = op =
   
val layout = Layout.str o toString

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor UseName(S: sig
		       include T
		       val sameName: t * t -> bool
		    end): T =
   struct
      open S
      val equals = sameName
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Tycon(S: TYCON_STRUCTS): TYCON = 
struct

open S

structure Id = HashId(structure AstId = AstId
		      val noname = "t")
open Id

structure P = PrimTycons(Id)
open P

fun clearPrimitive() = List.foreach(prims, clear)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_OPS_STRUCTS =
   sig
      structure Tycon: TYCON

      type t
      val con: Tycon.t * t list -> t
      val deconOpt: t -> (Tycon.t * t list) option
   end

signature TYPE_OPS =
   sig
      (* Don't want to include TYPE_OPS_STRUCTS because don't want to propagate
       * the Tycon structure, which will cause duplicate specifications later on.
       *)
      type tycon
      type t
      val con: tycon * t list -> t
      val deconOpt: t -> (tycon * t list) option

      (* constants *)
      val bool: t
      val char: t
      val exn: t
      val int: t
      val intInf: t
      val real: t
      val stack: t
      val string: t
      val unit: t
      val unitRef: t
      val word8: t
      val word: t

      val defaultInt: t
      val defaultWord: t

      (* constructors *)
      val arrow: t * t -> t
      val array: t -> t
      val vector: t -> t
      val list: t -> t
      val reff: t -> t
      val tuple: t list -> t

      (* destructors *)
      val detycon: t -> tycon
      val dearray: t -> t
      val devector: t -> t
      val dearrow: t -> t * t
      val arg: t -> t    (* arg = #1 o dearrow *)
      val result: t -> t (* result = #2 o dearrow *)
      val dearrowOpt: t -> (t * t) option
      val dearrayOpt: t -> t option
      val deref: t -> t
      val derefOpt: t -> t option
      val detuple: t -> t list
      val detupleOpt: t -> t list option
      val nth: t * int -> t 

      (* predicates *)
      val isTuple: t -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeOps(S: TYPE_OPS_STRUCTS): TYPE_OPS =
struct

open S

type tycon = Tycon.t
   
local
   fun nullary tycon = con(tycon, [])
in
   val bool = nullary Tycon.bool
   val char = nullary Tycon.char
   val exn = nullary Tycon.exn
   val int = nullary Tycon.int
   val intInf = nullary Tycon.intInf
   val real = nullary Tycon.real
   val stack = nullary Tycon.stack
   val string = nullary Tycon.string
   val word = nullary Tycon.word
   val word8 = nullary Tycon.word8

   val defaultInt = nullary Tycon.defaultInt
   val defaultWord = nullary Tycon.defaultWord
end

local
   fun unary tycon t = con(tycon, [t])
in
   val array = unary Tycon.array
   val vector = unary Tycon.vector
   val list = unary Tycon.list
   val reff = unary Tycon.reff
end

local
   fun binary tycon(t1, t2) = con(tycon, [t1, t2])
in
   val arrow = binary Tycon.arrow
end


fun deUnaryOpt tycon t =
   case deconOpt t of
      SOME(c, [t]) => if Tycon.equals(c, tycon) then SOME t else NONE
    | _ => NONE

val dearrayOpt = deUnaryOpt Tycon.array
val derefOpt = deUnaryOpt Tycon.reff

fun deUnary tycon t =
   case deUnaryOpt tycon t of
      SOME t => t
    | NONE => Error.bug "deUnary"

val dearray = deUnary Tycon.array
val devector = deUnary Tycon.vector
val deref = deUnary Tycon.reff
   
fun tuple[t] = t
  | tuple ts = con(Tycon.tuple, ts)

val unit = tuple[]

fun detupleOpt t =
   case deconOpt t of
      SOME(c, ts) => if Tycon.equals(c, Tycon.tuple) then SOME ts else NONE
    | NONE => NONE

val isTuple = Option.isSome o detupleOpt

fun detuple t =
   case detupleOpt t of
      SOME t => t
    | NONE => Error.error "detuple"

fun nth(t, n) = List.nth(detuple t, n)

val unitRef = reff unit

fun detycon t =
   case deconOpt t of
      SOME(c, _) => c
    | NONE => Error.error "detycon"

fun dearrowOpt t =
   case deconOpt t of
      SOME(c, [t1, t2]) => if Tycon.equals(c, Tycon.arrow)
			    then SOME(t1, t2)
			 else NONE
    | _ => NONE

fun dearrow t =
   case dearrowOpt t of
      SOME x => x
    | NONE => Error.error "Type.dearrow"

val arg = #1 o dearrow
val result = #2 o dearrow

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_STRUCTS =
   sig
      structure Ast: AST
      structure Record: RECORD
      structure Tycon: TYCON
      structure Tyvar: TYVAR
      sharing Tyvar = Ast.Tyvar
      sharing Ast.Tycon = Tycon.AstId
      sharing Record = Ast.SortedRecord
   end

signature TYPE = 
   sig
      include TYPE_STRUCTS
      include TYPE_OPS where type tycon = Tycon.t
	    
      datatype t' =
	 Var of Tyvar.t
       | Con of Tycon.t * t' list
       | Record of t' Record.t
      sharing type t = t'
	 
      (* constructors *)
      val var: Tyvar.t -> t
      val record: t Record.t -> t

      (* a list (without duplicates) of all the type variables in a type *)
      val tyvars: t -> Tyvar.t list

      (* substitute(t, [(a1, t1), ..., (an, tn)]) performs simultaneous
       substitution of the ti for ai in t *)
      val substitute: t * (Tyvar.t * t) list -> t

      (* conversion to Ast *)
      val toAst: t -> Ast.Type.t
      val optionToAst: t option -> Ast.Type.t option

      val hom: {ty: t,
		var: Tyvar.t -> 'a,
		con: Tycon.t * 'a list -> 'a} -> 'a

      val layout: t -> Layout.t
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Type(S: TYPE_STRUCTS): TYPE =
struct

open S

structure Type =
   struct
      datatype t =
	 Var of var
       | Con of con
       | Record of record
      withtype var = Tyvar.t
      and con = Tycon.t * t list
      and record = t Record.t
      datatype t' = datatype t

      val var = Var
      fun con(c, ts) =
	 if Tycon.equals(c, Tycon.tuple)
	    then (case ts of
		     [t] => t
		   | ts => Record(Record.tuple ts))
	 else Con(c, ts)
      val record = Record

      val deconOpt =
	 fn Con(c, ts) => SOME(c, ts)
	  | Record r => (case Record.detupleOpt r of
			    NONE => NONE
			  | SOME ts => SOME(Tycon.tuple, ts))
	  | _ => NONE
   end

structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
open Ops Type

structure Tyvars = UnorderedSet(Tyvar)
   
fun tyvars t =
   case t of
      Var a => Tyvars.singleton a
    | Con(_, ts) => Tyvars.unions(List.map(ts, tyvars))
    | Record r => Tyvars.unions(List.map(Record.range r, tyvars))
	 
val tyvars = Tyvars.toList o tyvars

fun toAst t =
   case t of
      Var a => Ast.Type.var a
    | Con(c, ts) => Ast.Type.con(Tycon.toAst c, List.map(ts, toAst))
    | Record r => Ast.Type.record(Record.map(r, toAst))

val layout = Ast.Type.layout o toAst

fun optionToAst z = Option.map(z, toAst)

fun substitute(t, sub) =
   let
      fun loop t =
	 case t of
	    Var a => let fun lookup [] = Error.error "substitute"
			   | lookup((a', t) ::s) =
			     if Tyvar.equals(a, a') then t else lookup s
		     in lookup sub
		     end
	  | Con(c, ts) => Con(c, List.map(ts, loop))
	  | Record r => Record(Record.map(r, loop))
   in loop t
   end

fun hom{ty, var, con} =
   let
      val rec loop =
	 fn Var a => var a
	  | Con(c, ts) => con(c, loops ts)
	  | Record r =>
	       (case Record.detupleOpt r of
		   SOME ts => con(Tycon.tuple, loops ts)
		 | NONE => Error.bug "Type.hom")
      and loops = fn ts => List.map(ts, loop)
   in loop ty
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SMALL_INT_INF_STRUCTS = 
   sig
   end

signature SMALL_INT_INF = 
   sig
      include SMALL_INT_INF_STRUCTS
      
      type t

      (* The argument is a SML language integer literal.
       * Given a string which consists of
       *	an optional ~
       *	either one or more of [0-9]
       *            or 0x followed by one or more of [0-9a-fA-F]
       * if it fits in a fixnum IntInf.int, return SOME of the corresponding
       * word.  If it doesn't, return NONE.
       *)
      val fromString: string -> t option
	 
      val toMLstring: t -> string
      val toCstring: t -> string

      val hash: t -> word

      val equals: t * t -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor SmallIntInf(S: SMALL_INT_INF_STRUCTS): SMALL_INT_INF = 
struct

open S

val hash = Word.fromLargeWord

structure Word = Word32
   
type t = Word.word
   
val equals = op =

fun toCstring w = "0x" ^ Word.toString w

val layout = Layout.str o toCstring

fun toMLstring w = Int.toString(Word.toIntX(Word.~>>(w, 0w1)))

(*
 * The IntInf.fromInt calls are just because SML/NJ doesn't
 * overload integer constants for IntInf.int's.
 * This code relies on the language that MLton is implemented in using at
 * least 31 bits for integers.
 *)
val minSmall: IntInf.int = IntInf.fromInt ~0x40000000
val maxSmall: IntInf.int = IntInf.fromInt 0x3FFFFFFF

(* Warning: fromString will raise the Subscript exception on inputs "" and "~".
 * This code is messier than it needs to be because it works around a bug in the 
 * SML/NJ implementation of IntInf.scan StringCvt.HEX.  In particular, it handles
 * the ~ sign and possible 0x itself.
 *)
fun fromString (str: string): t option =
       let val size = String.size str
	   fun reader offset =
		  if offset = size
		     then NONE
		     else SOME (String.sub (str, offset), offset + 1)
	   val (start, negate) =
	      if String.sub (str, 0) = #"~"
		 then (1, true)
	      else (0, false)
	   (* We carefully avoid calling IntInf.scan on the 0x in a hex constant
	    * because of an SML/NJ bug.
	    *)
	   val (base, start) =
	      if (String.sub (str, start) = #"0"
		  andalso (case reader (start + 1) of
			      SOME (#"x", next) => true
			    | _ => false))
		 then (StringCvt.HEX, start + 2)
	      else (StringCvt.DEC, start)
	   val (v, _) = valOf (IntInf.scan base reader start)
	   val v = if negate then IntInf.~ v else v
       in if IntInf.<= (minSmall, v) andalso IntInf.<= (v, maxSmall)
	     then let val w = Word.fromInt (IntInf.toInt v)
		      val res = Word.orb (0w1, Word.<< (w, 0w1))
		  in SOME res
		  end
	     else NONE
       end

(* val fromString =
 *    Trace.trace("SmallIntInf.fromString",
 * 	       String.layout,
 * 	       Option.layout layout) fromString
 *)

(*
 * If you want to compile MLton using an ML implementation which does not
 * have IntInf, then use the following instead.  Note, in this case Overflow
 * MUST be raised.
 *
 * fun fromString (str: string): t option =
 *        let val size = String.size str
 *            fun reader offset =
 *                   if offset = size
 *                      then NONE
 *                      else SOME (String.sub (str, offset), offset + 1)
 *            val start = if String.sub (str, 0) = #"~"
 *                           then 1
 *                           else 0
 *            val base = if String.sub (str, start) = #"0"
 *                           then case reader (start + 1) of
 *                                SOME (#"x", next) => StringCvt.HEX
 *                                | _ => StringCvt.DEC
 *                           else StringCvt.DEC
 *        in (case Pervasive.Int.scan base reader 0 of
 *            SOME (resv, _) =>
 *               let val resw = Word.fromInt resv
 *                   val res = Word.orb (0w1, Word.<< (resw, 0w1))
 *               in if Word.toLargeIntX (Word.xorb (resw, res)) < 0
 *                  then NONE
 *                  else SOME res
 *               end
 *              | _ => Error.bug "SmallIntInf.fromString")
 *           handle Overflow => NONE
 *        end
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature GENERIC_SCHEME_STRUCTS = 
   sig
      structure Tyvar: TYVAR
      structure Type: sig
			  type t
			  val var: Tyvar.t -> t
			  val substitute: t * (Tyvar.t * t) list -> t
			  val layout: t -> Layout.t
		       end
   end

signature GENERIC_SCHEME = 
   sig
      type tyvar
      type ty

      datatype t = T of {tyvars: tyvar list, ty: ty}

      (* constructors *)
      val fromType: ty -> t
      val make0: ty -> t
      val make1: (ty -> ty) -> t
      val make2: (ty * ty -> ty) -> t
      val makeEqual1: (ty -> ty) -> t

      (* destructors *)
      val tyvars: t -> tyvar list
      val ty: t -> ty

      val apply: t * ty list -> ty

      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SCHEME = 
   sig
      include GENERIC_SCHEME

      structure Type: TYPE
      structure Tyvar: TYVAR
      structure Tycon: TYCON

      sharing type ty = Type.t
      sharing type tyvar = Type.Tyvar.t
      sharing Tycon = Type.Tycon
      sharing Tyvar = Type.Tyvar
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRIM_STRUCTS = 
   sig
      structure Longid: LONGID
      structure Scheme: SCHEME
      structure Type: TYPE
      sharing Type = Scheme.Type
   end

signature PRIM = 
   sig
      include PRIM_STRUCTS

      structure Name:
	 sig
	    datatype t =
	       (* implemented in backend.fun *)
	       Array | ArraySub | ArrayUpdate
	     | VectorSub | VectorFromArray
	     | StringSub 
	     | Ref | Deref | Assign
	     | SaveWorld
	     | SaveStack
	     (* implemented in implement-exceptions.fun *)
	     | ExnName
	     (* implemented in poly-equal.fun *)
	     | Equal
	     (* implemented in machine.fun *)
	     | Bogus (*: unit -> 'a.  Make a bogus value of any type *)
	     | ArrayLength
	     | VectorLength
	     | Eq
	     | Size
	     | CPrim of string
	 end

      type t
	 
      (* these are essential primitives that must appear in the basis *)
      val array: t
      val bogus: t
      val deref: t
      val eq: t    (* pointer equality *)
      val equal: t (* polymorphic equality *)
      val exit: t
      val intInfEqual: t
      val intPlus: t
      val print: t
      val reff: t
      val saveStack: t
      val saveWorld: t
      val size: t
      val stringEqual: t
      val stringToIntInf: t
      val vectorLength: t
      val vectorSub: t

      (* constructors *)
      val new: string * Scheme.t -> t
      val newNullary: string -> t (* new of type unit -> unit *)

      (* destructors *)
      val longid: t -> Longid.t
      val name: t -> Name.t
      val numArgs: t -> int option
      val scheme: t -> Scheme.t

      val checkApp: {
		     prim: t,
		     targs: 'a list,
		     args: 'a list,
		     con: Type.Tycon.t * 'a list -> 'a,
		     equals: 'a * 'a -> bool,
		     dearrowOpt: 'a -> ('a * 'a) option,
		     detupleOpt: 'a -> 'a list option,
		     isUnit: 'a -> bool
		     } -> 'a option

      (* example: Update, + (because of exceptions)
       * not example: Ref, Array, !, Subscript
       *)
      val maySideEffect: t -> bool

      (* examples: Ref, Array
       * not example: Update, Subscript, :=, !, +
       *)
      val isMoveable: t -> bool

      (*
       * Always returns same result when given same args.
       * isFunctional implies isMoveable.
       * isFuntional implies not maySideEffect.
       * examples: ArrayLength, Equal, +
       * not examples: Ref, !, Array, Subscript
       *)
      val isFunctional: t -> bool
	 
      val isArrayAllocation: t -> bool

      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Prim(S: PRIM_STRUCTS): PRIM = 
struct

open S

local open Type
in
   structure Tycon = Tycon
   structure Tyvar = Tyvar
end

structure Name =
   struct
      datatype t =
	 Array | ArraySub | ArrayUpdate | ArrayLength
       | VectorSub | VectorLength | VectorFromArray
       | StringSub 
       | Ref | Deref | Assign
       | Bogus
       | SaveWorld
       | SaveStack
       | ExnName
       | Eq
       | Equal
       | Size
       | CPrim of string

      val isArrayAllocation = 
	 fn Array => true
	  | _ => false

      val isFunctional =
	 fn ArrayLength => true
	  | VectorSub => true
	  | VectorLength => true
	  | VectorFromArray => true
	  | StringSub => true
	  | ExnName => true
	  | Eq => true
	       (* Equal is commented out because I don't want globalization or
		* constant propagation to move it into the globals, because
		* polymorphic equality isn't implemented there.
		*)
(*	  | Equal => true  *)
	  | _ => false

      fun isMoveable p =
	 isFunctional p
	 orelse
	 case p of
	    Array => true
	  | Ref => true
	  | _ => false

      val maySideEffect =
	 fn ArrayUpdate => true
	  | Assign => true
	  | SaveWorld => true
	  | SaveStack => true
	  | Size => true
	  | CPrim _ => true
	  | _ => false

      val toString =
	 fn Array => "MLTON_Array_array"
	  | ArraySub => "MLTON_Array_sub"
	  | ArrayUpdate => "MLTON_Array_update"
	  | ArrayLength => "MLTON_Array_length"
	  | VectorSub => "MLTON_Vector_sub"
	  | VectorLength => "MLTON_Vector_length"
	  | VectorFromArray => "MLTON_Vector_fromArray"
	  | StringSub => "MLTON_String_sub"
	  | Ref => "ref"
	  | Deref => "!"
	  | Assign => ":="
	  | Bogus => "MLTON_MLton_bogus"
	  | SaveWorld => "MLTON_MLton_saveWorld"
	  | SaveStack => "MLTON_MLton_saveStack"
	  | ExnName => "MLTON_General_exnName"
	  | Equal => "="
	  | Eq => "MLTON_eq"
	  | Size => "MLTON_MLton_size"
	  | CPrim s => s

      (* Some of the values of these strings are important since they
       * are referred to in the basis library code.
       *)
      val fromString =
	 fn "MLTON_Array_array" => Array
	  | "MLTON_Array_sub" => ArraySub
	  | "MLTON_Array_update" => ArrayUpdate
	  | "MLTON_Array_length" => ArrayLength
	  | "MLTON_Vector_sub" => VectorSub
	  | "MLTON_Vector_length" => VectorLength
	  | "MLTON_Vector_fromArray" => VectorFromArray
	  | "MLTON_String_sub" => StringSub
	  | "ref" => Ref
	  | "!" => Deref
	  | ":=" => Assign
	  | "MLTON_MLton_saveWorld" => SaveWorld
	  | "MLTON_MLton_saveStack" => SaveStack
	  | "MLTON_MLton_size" => Size
	  | "MLTON_General_exnName" => ExnName
	  | "=" => Equal
	  | "MLTON_eq" => Eq
	  | s => CPrim s

      val layout = Layout.str o toString
   end

datatype t =
   T of {name: Name.t,
	 scheme: Scheme.t,
	 isMoveable: bool,
	 isFunctional: bool,
	 maySideEffect: bool,
	 numArgs: int option,
	 plist: PropertyList.t}

fun name(T{name, ...}) = name
fun isMoveable(T{isMoveable, ...}) = isMoveable
fun isFunctional(T{isFunctional, ...}) = isFunctional
fun maySideEffect(T{maySideEffect, ...}) = maySideEffect

val layout = Name.layout o name
val isArrayAllocation = Name.isArrayAllocation o name

fun scheme(T{scheme, ...}) = scheme
(*val scheme = Trace.trace("Prim.scheme", layout, Scheme.layout) scheme *)

fun numArgs(T{numArgs, ...}) = numArgs
(*val numArgs = Trace.trace("numArgs", layout, Option.layout Int.layout) numArgs *)
			  
fun plist(T{plist, ...}) = plist

val toString = Name.toString o name
val longid = Longid.short o Longid.Id.fromString o toString

fun equals(p, p') = PropertyList.equals(plist p, plist p')

fun new(name, scheme) =
   let
      val numArgs =
	 case Type.dearrowOpt(Scheme.ty scheme) of
	    NONE => NONE
	  | SOME(t, _) => (case Type.detupleOpt t of
			      NONE => SOME 1
			    | SOME ts => SOME(List.length ts))
   in T{plist = PropertyList.new(),
	name = name,
	scheme = scheme,
	isFunctional = (Name.isFunctional name
			orelse not(isSome numArgs)),
	isMoveable = Name.isMoveable name,
	maySideEffect = Name.maySideEffect name,
	numArgs = numArgs}
   end

local
   val newPrim = new
   open Type Scheme
   val new = newPrim
   val --> = arrow
   infix -->
in
   val deref = new(Name.Deref, make1(fn a => reff a --> a))
   val reff = new(Name.Ref, make1(fn a => a --> reff a))
   val vectorSub =
      new(Name.VectorSub, make1(fn a => tuple[vector a, int] --> a))
   val vectorLength =
      new(Name.VectorLength, make1(fn a => vector a --> int))
   val equal = new(Name.Equal, makeEqual1(fn a => tuple[a, a] --> bool))
   val eq = new(Name.Eq, makeEqual1(fn a => tuple[a, a] --> bool))
   val array = new(Name.Array, make1(fn a => int --> array a))
   val bogus = new(Name.Bogus, make1(fn a => a))
   val saveStack = new(Name.SaveStack, make0(unit --> stack))
   val saveWorld = new(Name.SaveWorld, make0(string --> bool))
   val size = new(Name.Size, make1(fn a => a --> int))

   val new = fn (str, scheme) => new(Name.fromString str, scheme)

   fun new0(name, ty) = new(name, make0 ty)

   val stringToIntInf =
      new0("MLTON_IntInf_fromString", tuple[string, Type.array word] --> intInf)
   val stringEqual = new0("MLTON_String_equal", tuple[string, string] --> bool)
   val intInfEqual = new0("MLTON_IntInf_equal", tuple[intInf, intInf] --> bool)
   val exit = new0("exit", int --> unit)
   val intPlus = new0("MLTON_Int_add", tuple[int, int] --> int)
   val print = new0("print", string --> unit)

   fun newNullary(name: string) = new0(name, unit --> unit)
end

fun 'a checkApp{prim, targs, args,
		con, detupleOpt, dearrowOpt, equals, isUnit}
   : 'a option =
   let
      val error = NONE
      val Scheme.T{tyvars, ty} = scheme prim
   in
      if List.length targs <> List.length tyvars
	 then error
      else
	 let
	    val env = List.zip(tyvars, targs)
	    fun var a =
	       case List.peek(env, fn (a', _) =>
			      Tyvar.equals(a, a')) of
		  NONE => Error.bug "prim scheme with free tyvar"
		| SOME(_, t) => t
	    val ty = Type.hom{ty = ty, var = var, con = con}
	 in
	    case numArgs prim of
	       NONE => (case args of
			   [] => SOME ty
			 | _ => error)
	     | SOME n =>
		  case dearrowOpt ty of
		     NONE => error
		   | SOME(argType, result) =>
			case (n, args) of
			   (0, []) => SOME result
			 | (1, [arg]) =>
			      if equals(arg, argType)
				 then SOME result
			      else error
			 | _ => 
			      case detupleOpt argType of
				 NONE => error
			       | SOME argTypes =>
				    if List.equals(args, argTypes, equals)
				       then SOME result
				    else error
	 end
   end

fun returnsBool p =
   case Type.dearrowOpt(Scheme.ty(scheme p)) of
      SOME(_, Type.Con(tycon, _)) => Tycon.equals(tycon, Tycon.bool)
    | _ => false

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CON_STRUCTS = 
   sig
      structure AstId: AST_ID
   end

signature CON = 
   sig
      (*include CON_STRUCTS*)
      include HASH_ID
      include PRIM_CONS where type con = t

      (* delete any info on con property lists *)
      val clearPrimitive: unit -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CONST_STRUCTS = 
   sig
      structure Ast: AST
      structure Tycon: TYCON
      sharing Tycon.AstId = Ast.Tycon
      structure SmallIntInf: SMALL_INT_INF
   end

signature CONST = 
   sig
      include CONST_STRUCTS

      type t

      datatype node =
	 Char of char
       | Int of Int32.int
       | SmallIntInf of SmallIntInf.t
       | IntInf of string
       | Real of string
       | String of string
       | Word of Word32.word

      val make: node * Tycon.t -> t
	 
      val node: t -> node
      val tycon: t -> Tycon.t

      val fromInt: int -> t
      val fromString: string -> t

      val toAstExp: t -> Ast.Exp.t
      val toAstPat: t -> Ast.Pat.t

      val hash: t -> word
	 
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ATOMS_STRUCTS =
   sig
      structure Ast: AST
   end

signature ATOMS' =
   sig
      include ATOMS_STRUCTS

      structure Con: CON sharing Con.AstId = Ast.Con
      structure Prim: PRIM sharing Prim.Longid = Ast.Longvid
      structure Tycon: TYCON sharing Tycon.AstId = Ast.Tycon
      structure UnaryTycon: UNARY_TYCON sharing Tycon = UnaryTycon.Tycon
      structure Scheme: SCHEME
      structure Var: VAR sharing Var.AstId = Ast.Var
      structure Const: CONST
      structure SmallIntInf: SMALL_INT_INF
      sharing SmallIntInf = Const.SmallIntInf
   
      sharing Tycon = Const.Tycon

      sharing Ast = Const.Ast = Prim.Type.Ast
      sharing Tycon = Scheme.Tycon
      sharing Ast.Tyvar = Scheme.Tyvar
      sharing Scheme = Prim.Scheme

      structure Record: RECORD
      sharing Record = Ast.Record
      structure SortedRecord: RECORD
      sharing SortedRecord = Ast.SortedRecord

      structure Tyvar: TYVAR
      sharing Tyvar = Ast.Tyvar

      structure Tyvars: SET sharing type Tyvars.Element.t = Tyvar.t
      structure Cons: SET sharing type Cons.Element.t = Con.t
      structure Vars: SET sharing type Vars.Element.t = Var.t
      structure Tycons: SET sharing type Tycons.Element.t = Tycon.t

      structure TyvarEnv :
	 sig
	    include MONO_ENV 

	    (* rename(env, tyvars) extends env by mapping each tyvar to
	     * a new tyvar (with the same equality property).  It returns
	     * the extended environment and the list of new tyvars
	     *)
            val rename: t * Tyvar.t list -> t * Tyvar.t list
	 end
      sharing type TyvarEnv.Domain.t = Tyvar.t
      sharing type TyvarEnv.Range.t = Tyvar.t

      (* delete any info on property lists of primitive cons, tycons, etc.*)
      val clearPrimitive: unit -> unit
   end

signature ATOMS =
   sig
      structure Atoms: ATOMS'
	 
      include ATOMS'

      sharing Ast = Atoms.Ast
      sharing Const = Atoms.Const
      sharing Var = Atoms.Var
      sharing Con = Atoms.Con
      sharing Prim = Atoms.Prim
      sharing Tycon = Atoms.Tycon
      sharing Tyvar = Atoms.Tyvar
      sharing Record = Atoms.Record
      sharing Vars = Atoms.Vars
      sharing Cons = Atoms.Cons
      sharing Tycons = Atoms.Tycons
      sharing Tyvars = Atoms.Tyvars
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor GenericScheme(S: GENERIC_SCHEME_STRUCTS): GENERIC_SCHEME =
struct

open S

type ty = Type.t
type tyvar = Tyvar.t
   
datatype t = T of {tyvars: Tyvar.t list, ty: Type.t}

fun ty(T{ty, ...}) = ty
fun tyvars(T{tyvars, ...}) = tyvars

fun isPoly(T{tyvars, ...}) = not(List.isEmpty tyvars)

fun fromType t = T{tyvars = [], ty = t}

val equals = fn _ => Error.unimplemented "Scheme.equals"

fun layout(T{tyvars, ty}) =
   let open Layout
      val ty = Type.layout ty
   in case tyvars of
      [] => ty
    | _ => align[seq[str "Forall ",
		     tuple(List.map(tyvars, Tyvar.layout)),
		     str "."],
		 ty]
   end

fun apply(T{tyvars, ty}, args) =
   case (tyvars, args) of
      ([], []) => ty  (* must special case this, since don't want to substitute
		       in monotypes *)
    | _ => Type.substitute(ty, List.zip(tyvars, args))

fun makeGen(numTyvars, equality, makeType): t =
   let val tyvars = List.duplicate(numTyvars, fn () =>
				   Tyvar.newNoname{equality = equality})
   in T{tyvars = tyvars,
	ty = makeType(List.map(tyvars, Type.var))}
   end

val make0 = fromType
fun make1 makeType = makeGen(1, false,
			     fn [a] => makeType a
			      | _ => Error.bug "make1")
fun make2 makeType = makeGen(2, false,
			     fn [a, b] => makeType(a, b)
			      | _ => Error.bug "make2")
fun makeEqual1 makeType = makeGen(1, true,
				 fn [a] => makeType a
				  | _ => Error.bug "makeEqual1")

end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Const(S: CONST_STRUCTS): CONST = 
struct

open S

local open Ast
in structure Aconst = Const
end

datatype node =
   Char of char
 | Int of Int32.int
 | SmallIntInf of SmallIntInf.t
 | IntInf of string
 | Real of string
 | String of string
 | Word of Word32.word

datatype t = T of {node: node,
		   tycon: Tycon.t}

fun make(n, t) = T{node = n, tycon = t}

fun node(T{node, ...}) = node
fun tycon(T{tycon, ...}) = tycon

local
   fun newHash() = Word.fromInt(Int.random())
   val char = newHash()
in
   fun hash(c: t): word =
      case node c of
	 Char c =>
	    Word.xorb(char,
		      Word.fromLargeWord(Word8.toLargeWord(Byte.charToByte c)))
       | Int i =>
	    (* the div is to make sure it fits in NJ's 31 bit word *)
	    Word.fromInt(Int32.toInt(i div 2))
       | SmallIntInf s => SmallIntInf.hash s
       | IntInf s => HashString.hashString s
       | Real s => HashString.hashString s
       | String s => HashString.hashString s
       | Word w => Word.fromLargeWord w
end
   
fun 'a toAst(make: Ast.Const.t -> 'a, constrain: 'a * Ast.Type.t -> 'a) c =
   let
      fun maybeConstrain(defaultTycon, aconst) =
	 let val t = tycon c
	 in if Tycon.equals(t, defaultTycon)
	       then make aconst
	    else constrain(make aconst, Ast.Type.con(Tycon.toAst t, []))
	 end
      fun int s = maybeConstrain(Tycon.defaultInt, Aconst.Int s)
   in case node c of
      Char c => make(Aconst.Char c)
    | Int n => int(Int32.toString n)
    | SmallIntInf i => int(SmallIntInf.toMLstring i)
    | IntInf s => int s
    | Real s => make(Aconst.Real s)
    | String s => make(Aconst.String s)
    | Word w => maybeConstrain(Tycon.defaultWord, Aconst.Word w)
   end

val toAstExp = toAst(Ast.Exp.const, Ast.Exp.constraint)
val toAstPat = toAst(Ast.Pat.const, Ast.Pat.constraint)

fun equals(c, c') =
   Tycon.equals(tycon c, tycon c')
   andalso
   case (node c, node c') of
      (Char c, Char c') => c = c'
    | (Int n, Int n') => n = n'
    | (SmallIntInf n, SmallIntInf n') => SmallIntInf.equals(n, n')
    | (IntInf s, IntInf s') => String.equals(s, s')
    | (Real s, Real s') => String.equals(s, s')
    | (String s, String s') => String.equals(s, s')
    | (Word w, Word w') => w = w'
    | _ => false

fun fromInt n = T{node = Int(Int32.fromInt n),
		  tycon = Tycon.defaultInt}

fun fromString s = T{node = String s,
		     tycon = Tycon.string}

local
   open Layout
   fun wrap(pre, post, s) = str(pre ^ String.toString s ^ post)
in
   fun layout c =
      case node c of
	 Int n => str(Int32.toString n)
       | SmallIntInf i => str(SmallIntInf.toMLstring i)
       | IntInf s => str s
       | Real l => String.layout l
       | Word w => seq[str "0wx", str(Word32.toString w)]
       | Char c => wrap("#\"", "\"", String.implode[c])
       | String s => wrap("\"", "\"", s)
end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Con(S: CON_STRUCTS): CON = 
struct

open S

structure C = HashId(structure AstId = AstId
		     val noname = "C")
open C
structure P = PrimCons(C)
open P

val fromAst = newString o AstId.toString
   
fun fromAsts l = List.map(l, fromAst)
   
val toAst = AstId.fromString o toString
   
fun toAsts l = List.map(l, toAst)

val all = [cons, falsee, nill, reff, truee, bind, match]
   
fun clearPrimitive() = List.foreach(all, clear)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Atoms(S: ATOMS_STRUCTS): ATOMS =
struct

structure Atoms =
   struct
      open S

      structure SmallIntInf = SmallIntInf()

      structure Var = Var(structure AstId = Ast.Var)

      structure Tycon = Tycon(structure AstId = Ast.Tycon)

      structure UnaryTycon = UnaryTycon(structure Tycon = Tycon)
	 
      structure Type =
	 Type(structure Ast = Ast
	      structure Record = Ast.SortedRecord
	      structure Tyvar = Ast.Tyvar
	      structure Tycon = Tycon)
      structure Scheme: SCHEME =
	 struct
	    structure Arg =
	       struct
		  structure Tycon = Tycon
		  structure Tyvar = Ast.Tyvar
		  structure Type = Type
	       end
	    structure S = GenericScheme(Arg)
	    open S Arg
	 end
      structure Con = Con(structure AstId = Ast.Con
			  structure Var = Var)
      structure Prim = Prim(structure Longid = Ast.Longvid
			    structure Type = Type
			    structure Scheme = Scheme)
      structure Const = Const(structure Ast = Ast
			      structure SmallIntInf = SmallIntInf
			      structure Tycon = Tycon)

      structure Record = Ast.Record
      structure SortedRecord = Ast.SortedRecord
      structure Tyvar = Ast.Tyvar

      structure Tyvars = UnorderedSet(Tyvar)
      structure Vars = UnorderedSet(Var)
      structure Cons = UnorderedSet(Con)
      structure Tycons = UnorderedSet(Tycon)

      structure TyvarEnv =
	 struct
	    structure Env = MonoEnv(structure Domain = UseName(Tyvar)
				    structure Range = Tyvar)
	    open Env

	    fun rename(env: t, tyvars: Tyvar.t list): t * Tyvar.t list =
	       let
		  val (env, tyvars) =
		     List.fold
		     (tyvars, (env, []), fn (tyv, (env, tyvs)) =>
		      let val tyv' = Tyvar.newNoname{equality = Tyvar.isEquality tyv}
		      in (extend(env, tyv, tyv'), tyv' :: tyvs)
		      end)
	       in (env, List.reverse tyvars)
	       end
	 end

      fun clearPrimitive() = (Tycon.clearPrimitive(); Con.clearPrimitive())
   end

open Atoms

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature HASH_TYPE_STRUCTS = 
   sig
      include ATOMS
   end

signature HASH_TYPE = 
   sig
      include HASH_TYPE_STRUCTS
      include TYPE_OPS sharing type tycon = Tycon.t

      val hash: t -> word
	 
      val var: Tyvar.t -> t

      val tycon: t -> Tycon.t
      val isUnit: t -> bool
      val ofConst: Const.t -> t
      val fromPrim: Prim.Type.t -> t
      val toPrim: t -> Prim.Type.t
      val variant: Tycon.t * Con.t -> t

      structure Dest :
	 sig
	    datatype dest =
	       Var of Tyvar.t
	     | Con of Tycon.t * t list
	     | Variant of Tycon.t * Con.t (* datatype variant *)
	    val dest: t -> dest
	 end

      (* substitute(t, [(a1, t1), ..., (an, tn)]) performs simultaneous
       * substitution of the ti for ai in t.
       * The ai's are not required to contain every free variable in t
       *)
      val substitute: t * (Tyvar.t * t) list -> t
	 
      (* conversion to Ast *)
      val toAst: t -> Ast.Type.t
      val optionToAst: t option -> Ast.Type.t option

      (* SOME if they are all =, else NONE.  Error for empty list. *)
      val equalss: t list -> t option
	 
      (* for reporting type errors *)
      val error: string * Layout.t -> 'a

      val containsTycon: t * Tycon.t -> bool

      val plist: t -> PropertyList.t

      (* O(1) time *)
      val equals: t * t -> bool

      val makeHomVariant:
	 {var: t * Tyvar.t -> 'a,
	  con: t * Tycon.t * 'a list -> 'a,
	  variant: Tycon.t * Con.t -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}
	 
      val makeHom:
	 {var: t * Tyvar.t -> 'a,
	  con: t * Tycon.t * 'a list -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}

      val makeMonoHom :
	 {con: t * Tycon.t * 'a list -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}
	 
      val hom: {ty: t,
		var: Tyvar.t -> 'a,
		con: Tycon.t * 'a list -> 'a} -> 'a

      val layout: t -> Layout.t
      val stats: unit -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Here are the rules for changing control transfers within f
 * when f is inlined at a call within g.
 *
 *				   Call to f within g
 *
 *				Tail		NonTail C
 * inside f			---------	---------
 *	Return			Return		Jump L
 *	Tail			Tail		NonTail C
 *	NonTail C'		NonTail C'	NonTail C'
 *)

(* All raises must raise the same number of args and handlers must take this
 * same number of args.
 *)

(* For case statements, the type-checker allows a branch on a value carrying
 * constructor to go to a nullary jump, in this case the value is ignored.
 *)

signature CPS_TREE_STRUCTS = 
   sig
      include ATOMS
   end

signature CPS_TREE = 
   sig
      include CPS_TREE_STRUCTS

      structure Type:
	 sig
	    include HASH_TYPE
	       
	    val tyconArgs: t -> Tycon.t * t list
	       
	    datatype dest =
	       Char
	     | Int
	     | IntInf
	     | Pointer
	     | Word
	     | Word8
	     | Real
	     | Stack
	     | String
	     | Array of t
	     | Vector of t
	     | Ref of t
	     | Datatype of Tycon.t
	     | Tuple of t list
	     | Variant of Tycon.t * Con.t

	    (* must be called (once) before using dest *)
	    val initialize: unit -> unit
	       
	    val dest: t -> dest
	 end
      sharing Atoms = Type.Atoms

      structure Func:
	 sig
	    include ID
	    val saveWorld: t
	 end

      structure Jump: ID

      structure PrimExp:
	 sig
	    datatype t =
	       Const of Const.t
	     | Var of Var.t
	     | Tuple of Var.t list
	     | Select of {tuple: Var.t,
			  offset: int}
	     | ConSelect of {variant: Var.t,
			     con: Con.t,
			     offset: int}
	     | ConApp of {con: Con.t,
			  args: Var.t list}
	     | PrimApp of {prim: Prim.t,
			   targs: Type.t list,
			   args: Var.t list}

	    val replaceVar: t * (Var.t -> Var.t) -> t
	    val foreachVar: t * (Var.t -> unit) -> unit

	    val layout: t -> Layout.t
	 end

      structure Dec:
	 sig
	    type exp
	    datatype t =
	       Bind of {
			var: Var.t,
			ty: Type.t,
			exp: PrimExp.t
		       }
	     | Fun of {
		       name: Jump.t,
		       args: (Var.t * Type.t) list,
		       body: exp
		      }
	     | HandlerPush of Jump.t
	     | HandlerPop

	    val layout: t -> Layout.t
	 end

      structure Cause:
	 sig
	    datatype t = User | Dispatch | PolyEqual | Coerce
	 end
      
      structure Transfer:
	 sig
	    datatype t =
	       Call of {
			func: Func.t,
			args: Var.t list,
			cont: Jump.t option (* NONE means tail-call *)
		       }
	     | Jump of {
			dst: Jump.t,
			args: Var.t list
		       }
	     | Return of Var.t list
	     | Case of {
			cause: Cause.t,
			test: Var.t,
			cases: (Con.t * Jump.t) list, (* jumps are nullary or unary *)
			default: Jump.t option  (* jump is nullary *)
		       }
	     | Raise of Var.t list
	     | Halt (* successful termination *)
	     | Bug  (* MLton thought control couldn't reach here. *)

	    val replaceVar: t * (Var.t -> Var.t) -> t
	    val foreachVar: t * (Var.t -> unit) -> unit
	       
	    val layout: t -> Layout.t
	 end

      structure Exp:
	 sig
	    type t
	       
	    val dest: t -> {decs: Dec.t list,
			    transfer: Transfer.t}
	    val make: {decs: Dec.t list,
		       transfer: Transfer.t} -> t

	    val decs: t -> Dec.t list
	    val transfer: t -> Transfer.t

	    val prefix: t * Dec.t -> t
	    val prefixs: t * Dec.t list -> t

	    (* Visits binding occurrences before uses.
	     * handleDec is applied to a dec before visiting everything that
	     * is in the scope of the dec.  It produces a function f: unit -> unit
	     * that is applied after everything in the scope has been visited.
	     *)
	    val foreach: t * {handleDec: Dec.t -> unit -> unit,
			      handleTransfer: Transfer.t -> unit} -> unit
	    (* Different from foreach in that it only visits a dec before its
	     * scope and not after.
	     *)
	    val foreach': t * {handleDec: Dec.t -> unit,
			       handleTransfer: Transfer.t -> unit} -> unit
	    val foreachDec: t * (Dec.t -> unit) -> unit
	    (* foreach BOUND var *)
	    val foreachVar: t * (Var.t * Type.t -> unit) -> unit
	    val foreachBind:
	       t * ({var: Var.t, ty: Type.t, exp: PrimExp.t} -> unit) -> unit
	    val foreachTransfer: t * (Transfer.t -> unit) -> unit
	    val foreachCall: t * ({func: Func.t,
				   args: Var.t list,
				   cont: Jump.t option} -> unit) -> unit

	    (* Rename all bound variables and jumps in exp.
	     * Each free occurrence of a formal in the exp will be replaced
	     * by the corresponding actual.
	     * handleJump is applied to each jump that is renamed.
	     *)
	    val alphaRename:
	       {
		exp: t,
		substitution: {formal: Var.t, actual: Var.t} list,
		handleJump: {old: Jump.t, new: Jump.t} -> unit
	       } -> t

	    val clear: t -> unit

	    val layout: t -> Layout.t
	 end
      sharing type Dec.exp = Exp.t

      structure DirectExp:
	 sig
	    type t

	    val const: Const.t -> t
	    val var: Var.t * Type.t -> t
	    val tuple: {exps: t list, ty: Type.t} -> t
	    val select: {tuple: t,
			 offset: int,
			 ty: Type.t} -> t
	    val conApp: {con: Con.t,
			 args: t list,
			 ty: Type.t} -> t
	    val primApp: {prim: Prim.t,
			  targs: Type.t list,
			  args: t list,
			  ty: Type.t} -> t
	    val call: {func: Func.t,
		       args: t list,
		       ty: Type.t} -> t
	    val raisee: t -> t
	    val handlee: {try: t,
			  ty: Type.t,
			  catch: Var.t * Type.t,
			  handler: t} -> t
	    val casee: {
			cause: Cause.t,
			test: t,
			cases: {con: Con.t,
				args: (Var.t * Type.t) list,
				body: t} list,
			default: t option,
			ty: Type.t (* type of entire case expression *)
		       } -> t
	    val lett: {decs: {
			      var: Var.t,
			      ty: Type.t,
			      exp: t
			      } list,
		       body: t} -> t

	    val detuple: {tuple: t,
			  body: Var.t list -> t} -> t

	    val detupleBind: {tuple: t,
			      components: Var.t list,
			      body: t} -> t

	    val seq: t * t -> t
	    val name2: t * t * (Var.t * Var.t -> t) -> t
(* 	    val andd: t * t -> t
 * 	    val ands: t list -> t
 * 	    val or: t * t -> t
 *)
	    val truee: t
	    val falsee: t
	    val int: int -> t
	    val + : t * t -> t
	    val eq: t * t * Type.t -> t

	    val send: t * (PrimExp.t * Type.t -> Exp.t) -> Exp.t
	    val toExp: t -> Exp.t
	    val toExpHalt: t -> Exp.t
	 end

      structure Program:
	 sig
	    datatype t =
	       T of {
		     datatypes: {
				 tycon: Tycon.t,
				 cons: {
					con: Con.t,
					args: Type.t list
				       } list
				} list,
		     globals: {
			       var: Var.t,
			       ty: Type.t,
			       exp: PrimExp.t
			      } list,
		     functions: {
				 name: Func.t,
				 args: (Var.t * Type.t) list,
				 body: Exp.t,
				 returns: Type.t list
				} list,
		     main: Func.t
		    } (* must be nullary *)

	    val foreachVar: t * (Var.t * Type.t -> unit) -> unit

	    val layoutStats: t -> Layout.t

	    val layout: t -> Layout.t
	    val layouts: t * (Layout.t -> unit) -> unit

	    val clear: t -> unit
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ANALYZE_STRUCTS = 
   sig
      include CPS_TREE
   end

signature ANALYZE = 
   sig
      include ANALYZE_STRUCTS
      
      val analyze:
	 {
	  program: Program.t,

	  useFromTypeOnBinds: bool,
	  
	  fromType: Type.t -> 'a,

	  copy: 'a -> 'a,

	  coerce: {
		   from: 'a,
		   to: 'a
		  } -> unit,
	  const: Const.t -> 'a,
	  tuple: 'a list -> 'a,
	  select: {
		   tuple: 'a,
		   offset: int,
		   resultType: Type.t
		  } -> 'a,

	  (* NONE means merge all uses of same constructor *)
	  constructors:
	  {
	   conApp: {
		    con: Con.t,
		    args: 'a list
		   } -> 'a,
	   conSelect: {
		       variant: 'a,
		       con: Con.t,
		       offset: int
		      } -> 'a
	  } option,
	  
	  primApp: {
		    prim: Prim.t,
		    targs: Type.t list,
		    args: 'a list,
		    resultVar: Var.t,
		    resultType: Type.t
		   } -> 'a,

	  (* The value is part of a datatype -- intersect it so it must be
	   * of the specified variant
	   *)
	  filter: 'a * Con.t -> 'a
	 }
	 -> {
	     value: Var.t -> 'a,
	     conArgs: Con.t -> 'a list,
	     func: Func.t -> {args: 'a list,
			      returns: 'a list},
	     jump: Jump.t -> 'a list,
	     exnVals: 'a list option
	    }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INFER_HANDLERS_STRUCTS = 
   sig
      include ANALYZE
   end

signature INFER_HANDLERS = 
   sig
      include INFER_HANDLERS_STRUCTS
      
      val deltaHandlers: Dec.t * Jump.t list -> Jump.t list
	 
      val inferHandlers: Program.t -> Jump.t -> Jump.t list
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_CHECK_STRUCTS = 
   sig
      include INFER_HANDLERS
   end

signature TYPE_CHECK = 
   sig
      include TYPE_CHECK_STRUCTS
      
      val typeCheck: Program.t -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeCheck(S: TYPE_CHECK_STRUCTS): TYPE_CHECK = 
struct

open S
open Dec Exp PrimExp Transfer

fun equalss(ts, ts') = List.equals(ts, ts', Type.equals)

fun checkScopes(program as Program.T{datatypes, globals, functions, main}): unit =
   let
      datatype status =
	 Undefined
       | InScope
       | Defined

      fun make(layout, plist) =
	 let val {get, set} = Property.new(plist, Property.initConst Undefined)
	    fun bind x =
	       case get x of
		  Undefined => set(x, InScope)
		| _ => Error.bug("duplicate definition of "
				 ^ (Layout.toString(layout x)))
	    fun reference x =
	       case get x of
		  InScope => ()
		| _ => Error.bug(concat
				 ["reference to ",
				  Layout.toString(layout x),
				  " not in scope"])

	    fun unbind x = set(x, Defined)
	 in (bind, reference, unbind)
	 end

      val (bindTycon, getTycon, unbindTycon) = make(Tycon.layout, Tycon.plist)
      val (bindCon, getCon, unbindCon) = make(Con.layout, Con.plist)
      val (bindVar, getVar, unbindVar) = make(Var.layout, Var.plist)
      val (bindFunc, getFunc, unbindFunc) = make(Func.layout, Func.plist)
      val (bindJump, getJump, unbindJump) = make(Jump.layout, Jump.plist)

      fun getVars xs = List.foreach(xs, getVar)

      val loopPrimExp =
	 fn Const _ => ()
	  | Var x => getVar x
	  | Tuple xs => getVars xs
	  | Select{tuple, ...} => getVar tuple
	  | ConSelect{variant, ...} => getVar variant
	  | ConApp{args, ...} => getVars args
	  | PrimApp{args, ...} => getVars args

      val loopTransfer =
	 fn Call{func, args, cont} => (getFunc func;
				       getVars args;
				       case cont of
					  NONE => ()
					| SOME c => getJump c)
	  | Jump{dst, args} => (getJump dst; getVars args)
	  | Return xs => getVars xs
	  | Case{test, cases, default, ...} =>
	       (getVar test;
		List.foreach(cases, getJump o #2);
		Option.map'(default, getJump))
	  | Raise xs => getVars xs
	  | Halt => ()
	  | Bug => ()

      fun loopBind{var, ty, exp} = (loopPrimExp exp; bindVar var)
	 
      fun loopFunc(args, body) =
	 (List.foreach(args, bindVar o #1);
	  loopExp body;
	  List.foreach(args, unbindVar o #1))

      and loopExp e =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(decs, 
			 fn Bind b => loopBind b
			  | Fun{name, args, body} =>
			       (bindJump name; loopFunc(args, body))
			  | HandlerPush h => getJump h
			  | HandlerPop => ())
	    ; loopTransfer transfer
	    ; List.foreach(decs, 
			   fn Bind{var, ...} => unbindVar var
			    | Fun{name, ...} => unbindJump name
			    | HandlerPush _ => ()
			    | HandlerPop => ())
	 end

   in List.foreach(datatypes, fn {tycon, cons} =>
		   (bindTycon tycon
		    ; List.foreach(cons, bindCon o #con)))
      ; List.foreach(globals, loopBind)
      ; List.foreach(functions, bindFunc o #name)
      ; List.foreach(functions, fn {args, body, ...} => loopFunc(args, body))
      ; getFunc main
      ; Program.clear program
   end

val checkScopes = Control.traceCall "checkScopes" checkScopes
		     
fun typeCheck(program as Program.T{datatypes, functions, ...}): unit =
   let
      val _ = checkScopes program
      val _ = let val jumpHandlers = inferHandlers program
	      in List.foreach
		 (functions, fn {body, ...} =>
		  Exp.foreach'
		  (body, {handleTransfer = fn _ => (),
			  handleDec =
			  fn Fun{name, ...} => (jumpHandlers name; ())
			   | _ => ()}));
	       Program.clear program
	      end
      val out = Out.error
      val print = Out.outputc out

      exception TypeError

      fun error(msg, lay) =
	 (print("Type error: " ^ msg ^ "\n")
	  ; Layout.output(lay, out)
	  ; print "\n"
	  ; raise TypeError)

      fun coerce{from: Type.t, to: Type.t}: unit =
	 if Type.equals(from, to)
	    then ()
	 else error("Type.equals",
		    Layout.record[("from", Type.layout from),
				  ("to", Type.layout to)])

      val error = fn s => error(s, Layout.empty)
	 
(*       val coerce =
 * 	 Trace.trace("TypeCheck.coerce",
 * 		     fn {from, to} => let open Layout
 * 				      in record[("from", Type.layout from),
 * 						("to", Type.layout to)]
 * 				      end,
 * 				   Unit.layout) coerce
 *)

      fun select{tuple: Type.t, offset:int, resultType}: Type.t =
	 case Type.detupleOpt tuple of
	    NONE => error "select of non tuple"
	  | SOME ts => List.nth(ts, offset)

      fun primApp{prim, targs, args, resultVar, resultType} =
	 case Prim.checkApp{prim = prim,
			    targs = targs,
			    args = args,
			    con = Type.con,
			    equals = Type.equals,
			    dearrowOpt = Type.dearrowOpt,
			    detupleOpt = Type.detupleOpt,
			    isUnit = Type.isUnit
			    } of
	    NONE => error "bad primapp"
	  | SOME t => t

(*       val primApp =
 * 	 Trace.trace("checkPrimApp",
 * 		     fn {prim, args, ...} =>
 * 		     let open Layout
 * 		     in record[("prim", Prim.layout prim),
 * 			       ("args", List.layout Type.layout args)]
 * 		     end,
 * 		     Type.layout) primApp
 *)

      fun filter(t: Type.t, c: Con.t): Type.t =
	 case Type.dest t of
	    Type.Datatype tycon => Type.variant(tycon, c)
	  | _ => Error.bug "can't make variant of non datatype"
	     
      val {value = varType, ...} =
	 analyze{
		 program = program,
		 useFromTypeOnBinds = true,
		 fromType = fn x => x,
		 copy = fn x => x,
		 coerce = coerce,
		 const = Type.ofConst,
		 tuple = Type.tuple,
		 select = select,
		 constructors = NONE,
		 primApp = primApp,
		 filter = filter
		}

   in Program.clear program
   end

(* val typeCheck = Trace.trace("Cps.typeCheck", Program.layout,
 * 			    Layout.ignore) typeCheck
 *)
end
signature DISJOINT_SET =
   sig
      type 'a t
	 
      val singleton: 'a -> 'a t

      (* Each set is associated with a single value.
       * When unions occur, one of the values is chosen.
       *)
      val value: 'a t -> 'a
      val setValue: 'a t * 'a -> unit
	 
      val equals: 'a t * 'a t -> bool

      val representative: 'a t -> 'a t
      val isRepresentative: 'a t -> bool

      val union: 'a t * 'a t -> unit

      val canUnion: 'a t * 'a t * ('a * 'a -> 'a option) -> bool
   end
(*-------------------------------------------------------------------*)
(*                           Disjoint Set                            *)
(*-------------------------------------------------------------------*)

functor DisjointSet (): DISJOINT_SET =
struct
   
datatype 'a t = T of 'a parent ref
and 'a parent =
   Parent of 'a t
  | Root of {value: 'a, rank: int}

fun singleton v = T (ref (Root{value = v, rank = 0}))

val rank =
   fn T(ref (Root{rank, ...})) => rank
    | _ => Error.bug "rank"

val setRank =
   fn (T(r as ref (Root{value, ...})), rank) =>
        r := Root{value = value, rank = rank}
    | _ => Error.bug "setRootValue"

fun incrementRank r = setRank(r, rank r + 1)
   
val parent =
   fn T(ref (Parent p)) => p
    | _ => Error.bug "parent"    

fun setParent(T r, p) = r := Parent p

val rootValue =
   fn T(ref (Root{value, ...})) => value
    | _ => Error.bug "rootValue"

val setRootValue =
   fn (T (r as ref (Root{rank, ...})), v) => r := Root{value = v, rank = rank}
    | _ => Error.bug "setRootValue"
    
fun equal(T r, T r') = r = r'
   
val isRoot  =
   fn T(ref (Root _)) => true
    | _ => false
    
val isRepresentative = isRoot
			
fun root s = if isRoot s then s
	     else let val r = root (parent s)
		  in setParent(s, r);
		     r
		  end
	       
val representative = root

fun value s = rootValue(root s)
   
fun setValue(s, v) = setRootValue(root s, v)

val equals = fn (s1, s2) => equal(root s1, root s2)

fun union(s, s') =
   let val r = root s
      val r' = root s'
   in if equal(r, r') then ()
      else let val n = rank r
	       val n' = rank r'
	   in if n < n' then setParent(r, r')
	      else (setParent(r', r)
		    ; if Int.equals(n,n') then incrementRank r else ())
	   end
   end

fun canUnion(s, s', f) =
   equals(s, s')
   orelse (case f(value s, value s') of
	      NONE => false
	    | SOME v => (union(s, s')
			 ; setValue(s, v)
			 ; true))

end

structure DisjointSet = DisjointSet()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature EXTRACT_TARGS_STRUCTS = 
   sig
      structure Prim: PRIM
      structure Type: TYPE_OPS
   end

signature EXTRACT_TARGS = 
   sig
      include EXTRACT_TARGS_STRUCTS
      
      val extract: {prim: Prim.t,
		    args: Type.t list,
		    result: Type.t} -> Type.t list
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ExtractTargs(S: EXTRACT_TARGS_STRUCTS): EXTRACT_TARGS = 
struct

open S

fun extract{prim, args, result} =
   let open Prim.Name
   in case (Prim.name prim, args) of
      (Array, _) => [Type.dearray result]
    | (ArraySub, _) => [result]
    | (VectorSub, _) => [result]
    | (ArrayUpdate, [_, _, z]) => [z]
    | (ArrayLength, [z]) => [Type.dearray z]
    | (VectorLength, [z]) => [Type.devector z]
    | (VectorFromArray, [z]) => [Type.dearray z]
    | (Ref, [z]) => [z]
    | (Deref, _) => [result]
    | (Assign, [_, z]) => [z]
    | (Equal, [z, _]) => [z]
    | (Eq, [z, _]) => [z]
    | (Size, [z]) => [z]
    | _ => []
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SHRINK_STRUCTS = 
   sig
      include TYPE_CHECK
   end

signature SHRINK = 
   sig
      include SHRINK_STRUCTS
      
      val shrinkExp:
	 {var: Var.t, ty: Type.t, exp: PrimExp.t} list
	 -> Exp.t -> Exp.t
      val shrinkExpNoDelete: Exp.t -> Exp.t
      val shrink: Program.t -> Program.t

      val simplifyProgram: (Exp.t -> Exp.t) -> Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature USELESS_STRUCTS = 
   sig
      include SHRINK
   end

signature USELESS = 
   sig
      include USELESS_STRUCTS
      
      val useless: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TWO_POINT_LATTICE_STRUCTS = 
   sig
      structure Top: sig
			type t
			val new: unit -> t
		     end
      structure Dest:
	 sig
	    type t
	    val bottom: t
	    val top: Top.t -> t
	 end
      (* pretty print names *)
      val bottom: string
      val top:string
   end

signature TWO_POINT_LATTICE = 
   sig
      include TWO_POINT_LATTICE_STRUCTS
      
      type t

      val dest: t -> Dest.t

      val new: unit -> t

      val isTop: t -> bool
      val isBottom: t -> bool

      (* handler will be run once iff value becomes top *)
      val addHandler: t * (unit -> unit) -> unit

      val makeTop: t -> unit

      (* force rhs to be top if lhs is *)
      val <= : t * t -> unit

      (* force lhs and rhs to be the same *)
      val == : t * t -> unit

      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TwoPointLattice(S: TWO_POINT_LATTICE_STRUCTS): TWO_POINT_LATTICE = 
struct

open S

structure Set = DisjointSet
   
datatype t = T of value Set.t
and value =
   Bottom of (unit -> unit) list ref  (* If I become Top, then run these. *)
  | Top of Top.t

fun value(T s) = Set.value s

fun dest v =
   case value v of
      Bottom _ => Dest.bottom
    | Top t => Dest.top t

fun toString e =
   case value e of
      Bottom _ => bottom
    | Top _ => top

val layout = Layout.str o toString

fun new(): t = T(Set.singleton(Bottom(ref [])))

fun equals(T s, T s') = Set.equals(s, s')

fun addHandler(e, h) =
   case value e of
      Bottom hs => ListRef.push(hs, h)
    | Top _ => h()

fun isTop s =
   case value s of
      Top _ => true
    | _ => false

fun isBottom s =
   case value s of
      Bottom _ => true
    | _ => false

fun runHandlers hs = List.foreach(!hs, fn h => h())

fun makeTop(T s) =
   case Set.value s of
      Top _ => ()
    | Bottom hs => (Set.setValue(s, Top(Top.new()));
		    runHandlers hs)

fun from <= to =
   if equals(from, to)
      then ()
   else
      case (value from, value to) of
	 (_, Top _) => ()
       | (Top _, _) => makeTop to
       | (Bottom hs, _) => ListRef.push(hs, fn () => makeTop to)

fun ==(T s, T s') =
   if Set.equals(s, s')
      then ()
   else
      let val e = Set.value s
	 val e' = Set.value s'
      in Set.union(s, s');
	 case (e, e') of
	    (Top _, Top _) => ()
	  | (Bottom hs, Top _) => (Set.setValue(s, e'); runHandlers hs)
	  | (Top _, Bottom hs) => (Set.setValue(s, e); runHandlers hs)
	  | (Bottom hs, Bottom hs') => Set.setValue(s, Bottom(ref(!hs @ !hs')))
      end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Useless(S: USELESS_STRUCTS): USELESS = 
struct

open S

structure ExtractTargs = ExtractTargs(structure Prim = Prim
				      structure Type = Type)

(* useless thing elimination
 *  remove components of tuples that are constants (use unification)
 *  remove function arguments that are constants
 *  build some kind of dependence graph where 
 *    - a value of ground type is useful if it is an arg to a primitive
 *    - a tuple is useful if it contains a useful component
 *    - a conapp is useful if it contains a useful component
 *                            or is used in a case
 *
 * If a useful tuple is coerced to another useful tuple,
 *   then all of their components must agree (exactly).
 * It is trivial to convert a useful value to a useless one.
 *
 * It is also trivial to convert a useful tuple to one of its
 *  useful components -- but this seems hard
 *)

(* If something is useful, then it must exist.
 *)

(* Suppose that you have a ref/array/vector that is useful, but the
 * components aren't -- then the components are converted to type unit, and
 * any primapp args must be as well.
 *)

(* Weirdness with raise/handle.
 * There must be a uniform "calling convention" for raise and handle.
 * Hence, just because some of a handlers args are useless, that doesn't mean
 * that it can drop them, since they may be useful to another handler, and
 * hence every raise will pass them along.  The problem is that it is not
 * possible to tell solely from looking at a function declaration whether it is
 * a handler or not, and in fact, there is nothing preventing a jump being used
 * in both ways.  So, maybe the right thing is for the handler wrapper to
 * do
 * Another solution would be to unify all handler args!
 *)

   
structure Value =
   struct
      structure Set = DisjointSet

      local
	 structure Top =
	    struct
	       type t = unit
	       fun new(): t = ()
	    end
	 structure Dest =
	    struct
	       datatype t =
		  NotExists
		| Exists of Top.t
	       val bottom = NotExists
	       val top = Exists
	    end
	 structure L = TwoPointLattice(val bottom = "not exists"
				       val top = "exists"
				       structure Top = Top
				       structure Dest = Dest)
      in
	 structure Exists =
	    struct
	       open L
	       val mustExist = makeTop
	       val doesExist = isTop
	    end
	 datatype exists = datatype Dest.t
      end

      local
	 structure Top =
	    struct
	       type t = unit
	       fun new(): t = ()
	    end
	 structure Dest =
	    struct
	       datatype t =
		  Useless
		| Useful of Top.t
	       val bottom = Useless
	       val top = Useful
	    end
	 structure L = TwoPointLattice(val bottom = "useless"
				       val top = "useful"
				       structure Top = Top
				       structure Dest = Dest)
      in
	 structure Useful =
	    struct
	       open L
	       val makeUseful = makeTop
	       val isUseful = isTop
	    end
	 datatype useful = datatype Dest.t
      end

      datatype t =
	 T of {
	       ty: Type.t,
	       new: (Type.t * bool) option ref,
	       value: value
	      } Set.t
      and value =
	 Ground of Useful.t
	| Vector of {length: t,
		     elt: slot}
	| Array of {useful: Useful.t,
		    length: t,
		    elt: slot}
	| Ref of {useful: Useful.t,
		  arg: slot}
	| Tuple of slot list
      withtype slot = t * Exists.t

      local
	 fun make sel(T s) = sel(Set.value s)
      in
	 val value = make #value
	 val ty = make #ty
      end
   
      local
	 open Layout
      in
	 fun layout(T s) =
	    let val {value, ...} = Set.value s
	    in case value of
	       Ground g => seq[str "ground ", Useful.layout g]
	     | Tuple vs => tuple(List.map(vs, layoutSlot))
	     | Ref{arg,...} => paren(seq[str "ref ", layoutSlot arg])
	     | Vector{elt, length} =>
		  seq[str "vector", tuple[layout length, layoutSlot elt]]
	     | Array{elt, length, ...} =>
		  seq[str "array", tuple[layout length, layoutSlot elt]]
	    end
	 and layoutSlot(v, e) =
	    tuple[Exists.layout e, layout v]
      end

      fun unify(T s, T s') =
	 if Set.equals(s, s')
	    then ()
	 else let val {value = v, ...} = Set.value s
		  val {value = v', ...} = Set.value s'
	      in Set.union(s, s');
		 case (v, v') of
		    (Ground g, Ground g') => Useful.==(g, g')
		  | (Tuple vs, Tuple vs') => List.foreach2(vs, vs', unifySlot)
		  | (Ref{useful = u, arg = a}, Ref{useful = u', arg = a'}) =>
		       (Useful.==(u, u'); unifySlot(a, a'))
		  | (Array{length = n, elt = e, ...},
		     Array{length = n', elt = e', ...}) =>
		       (unify(n, n'); unifySlot(e, e'))
		  | (Vector{length = n, elt = e},
		     Vector{length = n', elt = e'}) =>
		    (unify(n, n'); unifySlot(e, e'))
		  | _ => Error.bug "strange unify"
	      end
      and unifySlot((v, e), (v', e')) = (unify(v, v'); Exists.==(e, e'))
	   
      fun coerce{from = from as T sfrom, to = to as T sto}: unit =
	 if Set.equals(sfrom, sto)
	    then ()
	 else
	    let
	       fun coerceSlot((v, e), (v', e')) =
		  (coerce{from = v, to = v'};
		   Exists.==(e, e'))
	    in case (value from, value to) of
	       (Ground to, Ground from) => Useful.<=(from, to)
	     | (Tuple vs, Tuple vs') =>
		  List.foreach2(vs, vs', coerceSlot)
	     | (Ref _, Ref _) => unify(from, to)
	     | (Array _, Array _) => unify(from, to)
	     | (Vector{length = n, elt = e}, Vector{length = n', elt = e'}) =>
		   (coerce{from = n, to = n'};
		    coerceSlot(e, e'))
	     | _ => Error.bug "strange coerce"
	    end

(*       val coerce =
 * 	 Trace.trace("Useless.coerce",
 * 		     fn {from, to} => let open Layout
 * 				      in record[("from", layout from),
 * 						("to", layout to)]
 * 				      end,
 * 		     Unit.layout) coerce
 *)

      fun coerces{from, to} =
	 List.foreach2(from, to, fn (from, to) => coerce{from = from, to = to})

      (* Coerce every ground value in v to u. *)
      fun deepCoerce(v: t, u: Useful.t): unit =
	 let
	    fun doit u' = Useful.<=(u', u)
	    fun loop(v: t): unit =
	       case value v of
		  Ground u' => doit u'
		| Vector{length, elt} => (loop length; slot elt)
		| Array{length, elt, useful} =>
		     (doit useful; loop length; slot elt)
		| Ref{useful, arg} => (doit useful; slot arg)
		| Tuple vs => List.foreach(vs, slot)
	    and slot(v, _) = loop v
	 in loop v
	 end

(*       val deepCoerce =
 * 	 Trace.trace2("Useless.deepCoerce", layout, Useful.layout, Unit.layout)
 * 	 deepCoerce
 *)

      fun deground(v: t): Useful.t =
	 case value v of
	    Ground g => g
	  | _ => Error.bug "deground"

      fun fromType(t: Type.t): t =
	 let
	    fun loop(t: Type.t, es: Exists.t list): t =
	       let
		  fun useful() =
		     let val u = Useful.new()
		     in Useful.addHandler
			(u, fn () => List.foreach(es, Exists.mustExist));
			u
		     end
		  fun slot t =
		     let val e = Exists.new()
		     in (loop(t, e :: es), e)
		     end
		  val loop = fn t => loop(t, es)
		  val value =
		     case Type.dest t of
			Type.Ref t => Ref{useful = useful(),
					  arg = slot t}
		      | Type.Array t =>
			   let val elt as (_, e) = slot t
			      val length = loop Type.int
			   in Exists.addHandler
			      (e, fn () => Useful.makeUseful(deground length));
			      Array{useful = useful(),
				    length = length,
				    elt = elt}
			   end
		      | Type.Vector t => Vector{length = loop Type.int,
						elt = slot t}
		      | Type.Tuple ts => Tuple(List.map(ts, slot))
		      | _ => Ground(useful())
	       in T(Set.singleton{ty = t,
				  new = ref NONE,
				  value = value})
	       end
	 in loop(t, [])
	 end

      val const = fromType o Type.ofConst

      fun detupleSlots(v: t): slot list =
	 case value v of
	    Tuple ss => ss
	  | _ => Error.bug "detuple"

      fun detuple v = List.map(detupleSlots v, #1)

      fun tuple(vs: t list): t =
	 let val t = Type.tuple(List.map(vs, ty))
	    val v = fromType t
	 in List.foreach2(vs, detuple v, fn (v, v') =>
			  coerce{from = v, to = v'});
	    v
	 end

      val unit = tuple[]

      fun select{tuple, offset, resultType} =
	 let val v = fromType resultType
	 in coerce{from = List.nth(detuple tuple, offset), to = v};
	    v
	 end

      local
	 fun make (err, sel) v =
	    case value v of
	       Vector fs => sel fs
	     | _ => Error.bug err
      in val devector = make("devector", #1 o #elt)
	 val vectorLength = make("vectorLength", #length)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Array fs => sel fs
	     | _ => Error.bug err
      in val dearray: t -> t = make("dearray", #1 o #elt)
	 val arrayLength = make("arrayLength", #length)
	 val arrayUseful = make("arrayUseful", #useful)
      end

      fun deref(r: t): t =
	 case value r of
	    Ref{arg, ...} => #1 arg
	  | _ => Error.bug "deref"

      fun filter(v, _) =
	 case value v of
	    Ground g => (Useful.makeUseful g; v)
	  | _ => Error.bug "filter of non ground"

      fun newType(v: t): Type.t = #1(getNew v)
      and isUseful(v: t): bool = #2(getNew v)
      and getNew(T s): Type.t * bool =
	 let val {value, ty, new, ...} = Set.value s
	 in case !new of
	    SOME z => z
	  | NONE =>
	       let 
		  fun slot(arg: t, e: Exists.t) =
		     let val (t, b) = getNew arg
		     in (if Exists.doesExist e then t else Type.unit, b)
		     end
		  fun wrap((t, b), f) = (f t, b)
		  fun or((t, b), b') = (t, b orelse b')
		  fun maybe(u: Useful.t, s: slot, make: Type.t -> Type.t) =
		     wrap(or(slot s, Useful.isUseful u), make)
		  val z =
		     case value of
			Ground u => (ty, Useful.isUseful u)
		      | Ref{useful, arg, ...} => maybe(useful, arg, Type.reff)
		      | Array{useful, elt, length, ...} =>
			   or(wrap(slot elt, Type.array),
			      Useful.isUseful useful orelse isUseful length)
		      | Vector{elt, length, ...} =>
			   or(wrap(slot elt, Type.vector), isUseful length)
		      | Tuple vs =>
			   wrap
			   (List.foldr
			    (vs, ([], false), fn ((v, e), (ts, useful)) =>
			     let val (t, u) = getNew v
			     in (if Exists.doesExist e then t :: ts else ts,
				 u orelse useful)
			     end),
			    Type.tuple)
	       in new := SOME z; z
	       end
	 end

(*       val getNew =
 * 	 Trace.trace("getNew", layout, Layout.tuple2(Type.layout, Bool.layout))
 * 	 getNew
 * 
 *       val newType = Trace.trace("newType", layout, Type.layout) newType
 *)
      fun newTypes vs =
	 List.foldr(vs, [], fn (v, ts) =>
		    let val (t, b) = getNew v
		    in if b then t :: ts else ts
		    end)
   end

structure Exists = Value.Exists

fun useless(program as Program.T{datatypes, functions, main, globals}) =
   let
      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("arg", Con.layout))

      val {get = tyconInfo: Tycon.t -> {useful: bool ref,
					cons: Con.t list},
	   set = setTyconInfo, destroy} =
	 Property.newDest(Tycon.plist, Property.initRaise("cons", Tycon.layout))
	   
      local open Value
      in
	 val _ =
	    List.foreach
	    (datatypes, fn {tycon, cons} =>
	     (setTyconInfo(tycon, {useful = ref false,
				   cons = List.map(cons, #con)})
	      ; let fun value() = fromType(Type.con(tycon, []))
		in List.foreach
		   (cons, fn {con, args} =>
		    setConInfo(con, {value = value,
				     argTypes = args,
				     args = List.map(args, fromType)}))
		end))

	 val conArgs = #args o conInfo

	 fun conApp{con, args} =
	    let val {args = args', value, ...} = conInfo con
	    in Value.coerces{from = args, to = args'};
	       value()
	    end

	 fun conSelect{variant, con, offset} =
	    let val {argTypes, args, ...} = conInfo con
	       val result = fromType(List.nth(argTypes, offset))
	    in coerce{from = List.nth(args, offset), to = result};
	       (* If any component of the result is useful,
		* then the variant is useful.
		*)
	       case value variant of
		  Ground g => deepCoerce(result, g)
		| _ => Error.bug "strange variant";
	       result
	    end
   
	 (* This is used for primitive args, since we have no idea what components
	  * of its args that a primitive will look at.
	  *)
	 fun deepMakeUseful v =
	    let val slot = deepMakeUseful o #1
	    in case value v of
	       Ground u =>
		  (Useful.makeUseful u;
		   (* Make all constructor args of this tycon useful *)
		   case Type.dest(ty v) of
		      Type.Datatype tycon =>
			 let val {useful, cons} = tyconInfo tycon
			 in if !useful
			       then ()
			    else (useful := true
				  ; List.foreach(cons, fn con =>
						 List.foreach(#args(conInfo con),
							      deepMakeUseful)))
			 end
		    | _ => ())
	     | Tuple vs => List.foreach(vs, slot)
	     | Ref{useful, arg} => (Useful.makeUseful useful; slot arg)
	     | Vector{length, elt} => (deepMakeUseful length; slot elt)
	     | Array{useful, length, elt} => (Useful.makeUseful useful;
					      deepMakeUseful length;
					      slot elt)
	    end

	 type value = t
	    
	 fun primApp{prim, targs, args, resultVar, resultType} =
	    let
	       val result = fromType resultType
	       fun return v = coerce{from = v, to = result}
	       infix dependsOn
	       fun v1 dependsOn v2 = deepCoerce(v2, deground v1)
	       open Prim.Name
	    in (case (Prim.name prim, args) of
		   (Array, [i]) => coerce{from = i, to = arrayLength result}
		 | (ArraySub, [a, i]) => (i dependsOn result;
					  return(dearray a))
		 | (ArrayUpdate, [a, i, v]) =>
		      (i dependsOn dearray a;
		       coerce{from = v, to = dearray a})
		 | (ArrayLength, [a]) => return(arrayLength a)
		 | (VectorLength, [v]) => return(vectorLength v)
		 | (VectorSub, [v, i]) => (i dependsOn result;
					   return(devector v))
		 | (VectorFromArray, [a]) =>
		      (case (value a, value result) of
			  (Value.Array{length = l, elt = e, ...},
			   Vector{length = l', elt = e', ...}) =>
			     (unify(l, l'); unifySlot(e, e'))
			 | _ => Error.bug "strange VectorFromArray")
		 | (Ref, [v]) => coerce{from = v, to = deref result}
		 | (Deref, [r]) => return(deref r)
		 | (Assign, [r, v]) => coerce{from = v, to = deref r}
		 (* could do slightly better for Eq and Equal
		  * But be careful when you do so, since it may make a t ref into
		  * a unit ref -- and then the translation will need to replace some
		  * vars with unit.
		  *)
		 | _ => (List.foreach(args, deepMakeUseful)
			 ; if (case Type.dest resultType of
				  Type.Array _ => true
				| Type.Vector _ => true
				| Type.Ref _ => true
				| _ => false)
			      then deepMakeUseful result
			   else ()))
	       ; result
	    end
      end

      val {value, func, jump, exnVals, ...} =
	 Control.traceCall "analyze"
	 analyze{
		 program = program,
		 useFromTypeOnBinds = true,
		 fromType = Value.fromType,
		 copy = Value.fromType o Value.ty,
		 coerce = Value.coerce,
		 const = Value.const,
		 tuple = Value.tuple,
		 select = Value.select,
		 constructors = SOME{conApp = conApp, conSelect = conSelect},
		 primApp = primApp,
		 filter = Value.filter
		}

      open Dec PrimExp Transfer

      (* Unify all handler args so that raise/handle has a consistent calling
       * convention.
       *)
      val exnVals = fn () => (case exnVals of
				 NONE => Error.bug "no exnVals"
			       | SOME vs => vs)
      val _ =
	 List.foreach
	 (functions, fn {body, ...} =>
	  Exp.foreach'
	  (body,
	   {handleTransfer = fn _ => (),
	    handleDec = fn HandlerPush j => List.foreach2(jump j, exnVals(),
							  Value.unify)
                         | _ => ()}))

      (* diagnostic *)
      val _ =
	 Control.displays
	 ("useless", fn display =>
	  let open Layout
	  in display(str "Input program:\n");
	     Program.layouts(program, display);
	     display(str "\n\nUseless:");
	     List.foreach
	     (datatypes, fn {tycon, cons} =>
	      display
	      (align[Tycon.layout tycon,
		     indent(align
			    (List.map(cons, fn {con, ...} =>
				      seq[Con.layout con, str " ",
					  tuple(List.map(conArgs con,
							 Value.layout))])),
			    2)]));
	     Program.foreachVar
	     (program, fn (x, _) => display(seq[Var.layout x,
						str " ",
						Value.layout(value x)]))
	  end)

      val varExists = Value.isUseful o value

      val unitVar = Var.newString "unit"

      val bogusGlobals: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
	 
      fun bogus(ty: Type.t): Var.t =
	 case List.peek(!bogusGlobals, fn {ty = ty', ...} =>
			Type.equals(ty, ty')) of
	    NONE => let val var = Var.newString "bogus"
		    in ListRef.push(bogusGlobals,
				    {var = var, ty = ty,
				     exp = PrimApp{prim = Prim.bogus,
						   targs = [ty],
						   args = []}});
		       var
		    end
	  | SOME{var, ...} => var

      fun keepUseful(xs, vs) =
	 List.foldr2(xs, vs, [], fn (x, v, xs) =>
		     let val (t, b) = Value.getNew v
		     in if b then (if varExists x then x else bogus t) :: xs
			else xs
		     end)

(*       val keepUseful =
 * 	 Trace.trace2("keepUseful",
 * 		      List.layout Var.layout,
 * 		      List.layout Value.layout,
 * 		      List.layout Var.layout)
 * 	 keepUseful
 *)
      fun keepUsefulArgs xts =
	 List.foldr(xts, [], fn ((x, _), xts) =>
		    let val (t, b) = Value.getNew(value x)
		    in if b then (x, t) :: xts
		       else xts
		    end)

(*       val keepUsefulArgs =
 * 	 Trace.trace2("keepUsefulArgs",
 * 		      List.layout(",", Layout.tuple2(Var.layout, Type.layout)),
 * 		      List.layout(",", Value.layout),
 * 		      List.layout(",", Layout.tuple2(Var.layout, Type.layout)))
 * 	 keepUsefulArgs
 *)

      (* Returns true if the component is the only component of the tuple
       * that exists.
       *)
      fun newOffset(bs: bool list, n: int): int * bool =
	 let
	    fun loop(bs, n, i) =
	       case bs of
		  [] => Error.bug "newOffset"
		| b :: bs =>
		     if n = 0
			then (i, (i = 0 andalso not(List.exists(bs, fn b => b))))
		     else loop(bs, n - 1, if b then i + 1 else i)
	 in loop(bs, n, 0)
	 end
	 
      fun loopPrimExp(e: PrimExp.t, resultType: Type.t, resultValue: Value.t) =
	 case e of
	    Const _ => e
	  | Var _ => e
	  | Tuple xs =>
	       let val slots = Value.detupleSlots resultValue
		  val xs =
		     List.fold2
		     (xs, slots, [], fn (x, (v, e), xs) =>
		      if Exists.doesExist e
			 then ((if varExists x then x else bogus(Value.newType v))
			       :: xs)
		      else xs)
	       in case xs of
		   [x] => Var x
		 | xs => Tuple(List.reverse xs)
	       end
	  | Select{tuple, offset} =>
	       let
		  val (offset, isOne) =
		     newOffset(List.map(Value.detupleSlots(value tuple),
					Exists.doesExist o #2),
			       offset)
	       in if isOne
		     then Var tuple
		  else Select{tuple = tuple,
			      offset = offset}
	       end
	  | ConSelect{variant, con, offset} =>
	       ConSelect
	       {variant = variant,
		con = con,
		offset = #1(newOffset(List.map(conArgs con, Value.isUseful),
				      offset))}
	  | ConApp{con, args} =>
	       ConApp{con = con,
		      args = keepUseful(args, conArgs con)}
	  | PrimApp{prim, args, ...} => 
	       let
		  val (args, argTypes) =
		     List.unzip
		     (List.map(args, fn x =>
			       let val (t, b) = Value.getNew(value x)
			       in if b then (x, t)
				  else (unitVar, Type.unit)
			       end))
	       in PrimApp
		  {prim = prim,
		   args = args,
		   targs = ExtractTargs.extract{prim = prim,
						args = argTypes,
						result = resultType}}
	       end

(*       val loopPrimExp =
 * 	 Trace.trace3("Useless.loopPrimExp",
 * 		      PrimExp.layout, Layout.ignore, Layout.ignore,
 * 		      PrimExp.layout) loopPrimExp
 *)
	       
      fun loopBind{var, exp, ty} =
	 let val v = value var
	    fun yes ty = SOME{var = var, ty = ty, exp = loopPrimExp(exp, ty, v)}
	    val (t, b) = Value.getNew v
	 in if b then yes t
	    else
	       case exp of
		  PrimApp{prim, args, ...} =>
		     if Prim.maySideEffect prim
			andalso let val check = Value.isUseful
				    open Prim.Name
				in case (Prim.name prim, args) of
				   (ArrayUpdate, [a, _, _]) =>
				      check(Value.dearray(value a))
				 | (Assign, [r, _]) =>
				      check(Value.deref(value r))
				 | _ => true
			       end
			then yes t
		     else NONE
	        | _ => NONE
	 end

      fun agree(vs, vs') =
	 List.forall2(vs, vs', fn (v, v') =>
		      not(Value.isUseful v) orelse Value.isUseful v')

      fun loopExp(e: Exp.t, returns): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val decs =
	       List.reverse
	       (List.fold
		(decs, [], fn (d, ds) =>
		 case d of
		    Fun{name, args, body} =>
		       Fun{name = name,
			   args = keepUsefulArgs args,
			   body = loopExp(body, returns)} :: ds
		  | Bind b => (case loopBind b of
				  NONE => ds
				| SOME b => Bind b :: ds)
		  | _ => d :: ds))
	    val (decs, transfer) =
	       case transfer of
		  (* May need to insert a wrapper around the cont
		   * to translate between the expected returns of
		   * the function and the useful args of the func.
		   * Yucko! - what about tail calls?
		   * It's entirely possible that my returns are less useful
		   * than my callees, in which case I need I need to drop
		   * some of them.
		   *)
		  Call{func = f, args, cont} =>
		     let val {args = fargs, returns = freturns} = func f
			fun wrapper(vs: Value.t list,
				    cont: Jump.t option,
				    makeTrans: Var.t list -> Transfer.t) =
			   if agree(freturns, vs)
			      then (decs, cont)
			   else
			      let
				 val j' = Jump.newNoname()
				 val (formals, actuals) =
				    List.foldr2
				    (freturns, vs, ([], []),
				     fn (v, v', (formals, actuals)) =>
				     if Value.isUseful v
					then let val x = Var.newNoname()
					     in ((x, Value.newType v)
						 :: formals,
						 if Value.isUseful v'
						    then x :: actuals
						 else actuals)
					     end
				     else (formals, actuals))
			      in (decs
				  @ [Fun
				     {name = j',
				      args = formals,
				      body =
				      Exp.make{decs = [],
					       transfer = makeTrans actuals}}],
				  SOME j')
			      end

			val (decs, cont) =
			   case cont of
			      NONE => wrapper(returns, NONE, Return)
			    | SOME c =>
				 wrapper(jump c, SOME c,
					 fn args => Jump{dst = c, args = args})
		     in (decs,
			 Call{func = f, cont = cont,
			      args = keepUseful(args, fargs)})
		     end
		| Jump{dst, args} =>
		     (decs,
		      Jump{dst = dst, args = keepUseful(args, jump dst)})
		| Return xs => (decs, Return(keepUseful(xs, returns)))
		| Raise xs => (decs, Raise(keepUseful(xs, exnVals())))
		| Case{cases = [], default = NONE, ...} =>
		     (* The test may be useless if there are no cases or default,
		      * in which case we must eliminate the case.
		      * However, can't just turn it into halt, because
		      * the handler stack won't be right.
		      *)
		     (decs, Bug)
		| t => (decs, t)

	 in Exp.make
	    {decs = decs,
	     transfer = transfer}
	 end

      val datatypes =
	 List.map(datatypes, fn {tycon, cons} =>
		  {tycon = tycon,
		   cons = List.map(cons, fn {con, args} =>
				   {con = con,
				    args = Value.newTypes(conArgs con)})})

      val globals =
	 List.foldr(globals, [], fn (b, bs) =>
		    case loopBind b of
		       NONE => bs
		     | SOME b => b :: bs)

      val globals = {var = unitVar, ty = Type.unit, exp = Tuple[]} :: globals

      val shrinkExp = shrinkExp globals

(*       val shrinkExp =
 * 	 Trace.trace("Useless.shrinkExp", Exp.layout, Exp.layout) shrinkExp
 *)
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, ...} =>
	  let val {args = argvs, returns = returnvs} = func name
	  in {name = name,
	      args = keepUsefulArgs args,
	      body = shrinkExp(loopExp(body, returnvs)),
	      returns = Value.newTypes returnvs}
	  end)

      val globals = !bogusGlobals @ globals
	     
      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
	 
   in Program.clear program
      ; destroy()
      ; program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SIMPLIFY_TYPES_STRUCTS = 
   sig
      include SHRINK
   end

signature SIMPLIFY_TYPES = 
   sig
      include SIMPLIFY_TYPES_STRUCTS
      
      val simplify: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* This pass must happen before polymorphic equality is implemented becuase
 * 1. it will make polymorphic equality faster because some types are simpler
 * 2. it removes uses of polymorphic equality that must return true
 *
 * This pass computes a "cardinality" of each datatype, which is an
 * abstraction of the number of values of the datatype.
 *   Zero means the datatype has no values (except for bottom).
 *   One means the datatype has one values (except for bottom).
 *   Many means the datatype has many values.
 *
 * This pass removes all datatypes whose cardinality is Zero or One
 * and removes
 *   components of tuples
 *   function args
 *   constructor args
 * which are such datatypes.
 *
 * This pass marks constructors as one of
 *   Useless: it never appears in a ConApp.
 *   Transparent: it is the only variant in its datatype
 *     and its argument type does not contain any uses of Tycon.array.
 *   Useful: otherwise
 * This pass also removes Useless and Transparent constructors.
 *
 * We must keep track of Transparent consturctors whose argument type
 * uses Tycon.array because of datatypes like the following:
 *   datatype t = T of t array
 * Such a datatype has Cardinality.Many, but we cannot eliminate
 * the datatype and replace the lhs by the rhs, i.e. we must keep the
 * circularity around .
 * Must do similar things for vectors.
 * 
 * Also, to eliminate as many Transparent constructors as possible, for
 * something like the following,
 *   datatype t = T of u array
 *   and u = U of t array
 * we (arbitrarily) expanding one of the datatypes first.
 * The result will be something like
 *   datatype u = U of u array array
 * where all uses of t are replaced by u array.
 *)

functor SimplifyTypes(S: SIMPLIFY_TYPES_STRUCTS): SIMPLIFY_TYPES = 
struct

open S

open S
open Dec PrimExp Transfer

structure Cardinality =
   struct
      datatype t = Zero | One | Many

      val isMany: t -> bool =
	 fn Many => true
	  | _ => false
	       
      fun layout c =
	 Layout.str(case c of
		       Zero => "zero"
		     | One => "one"
		     | Many => "many")

      val equals: t * t -> bool = op =

      val isZero: t -> bool =
	 fn Zero => true
	  | _ => false

      val rec add: t list -> t =
	 fn [] => Zero
	  | c :: cs =>
	       case c of
		  Many => Many
		| Zero => add cs
		| One =>
		     case add cs of
			Zero => One
		      | _ => Many

      val rec mul: t list -> t =
	 fn [] => One
	  | c :: cs =>
	       case c of
		  Zero => Zero
		| One => mul cs
		| Many =>
		     case mul cs of
			Zero => Zero
		      | _ => Many
   end

structure ConRep =
   struct
      datatype t =
	 Useless
       | Transparent
       | Useful

      val isUseful =
	 fn Useful => true
	  | _ => false

      val isUseless =
	 fn Useless => true
	  | _ => false

      val toString =
	 fn Useless => "useless"
	  | Transparent => "transparent"
	  | Useful => "useful"

      val layout = Layout.str o toString
   end

fun simplify(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conInfo: Con.t -> {rep: ConRep.t ref,
				    args: Type.t list},
	   set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("info", Con.layout))

      val conRep = ! o #rep o conInfo
      val conArgs = #args o conInfo
      fun setConRep(con, r) = #rep(conInfo con) := r
      val conIsUseful = ConRep.isUseful o conRep
      val conIsUseless = ConRep.isUseless o conRep

(*       val conIsUseful =
 * 	 Trace.trace("conIsUseful", Con.layout, Bool.layout) conIsUseful
 * 	 
 *       val setConRep =
 * 	 Trace.trace2("setConRep", Con.layout, ConRep.layout, Unit.layout)
 * 	 setConRep
 * 
 *)
      (* Initialize conInfo *)
      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach(cons, fn {con, args} =>
		       setConInfo(con, {rep = ref ConRep.Useless,
					args = args})))

      (* diagnostic *)
(*      val _ =
	 List.foreach
	 (datatypes, fn {tycon, ...} =>
	  Control.message
	  let open Layout
	  in seq[str "dependents of ",
		 Tycon.layout tycon,
		 str " = ",
		 List.layout (", ", Tycon.layout)
		 (!(#dependents(tyconInfo tycon)))]
	  end)
*)

      val {get = tyconReplacement: Tycon.t -> Type.t option,
	   set = setTyconReplacement} =
	 Property.new(Tycon.plist, Property.initConst NONE)

      val setTyconReplacement = fn (c, t) => setTyconReplacement(c, SOME t)

      val {get = tyconInfo: Tycon.t -> {
					cardinality: Cardinality.t ref,
					cons: {
					       con: Con.t,
					       args: Type.t list
					      } list,
					numCons: int ref,
					(* tycons whose cardinality depends on mine *)
					dependents: Tycon.t list ref,
					isOnWorklist: bool ref
				       },
	   set = setTyconInfo} =
	 Property.new(Tycon.plist, Property.initRaise("info", Tycon.layout))

      local
	 fun make sel = (! o sel o tyconInfo,
			 fn (t, x) => sel(tyconInfo t) := x)
      in
	 val (tyconNumCons, setTyconNumCons) = make #numCons
	 val (tyconCardinality, _) = make #cardinality
      end

      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  setTyconInfo(tycon, {
			       cardinality = ref Cardinality.Zero,
			       numCons = ref 0,
			       cons = cons,
			       dependents = ref [],
			       isOnWorklist = ref false
			      }))
	 
      (* Tentatively mark all constructors appearing in a ConApp as Useful
       * (some may later be marked as Transparent).
       *)
      val _ =
	 let
	    fun handleBind{var, ty, exp} =
	       case exp of
		  ConApp{con, ...} => setConRep(con, ConRep.Useful)
		| _ => ()
	    fun consUseful e = Exp.foreachBind(e, handleBind)
	 in 
	    (* Booleans are special because they are generated by primitives. *)
	    List.foreach([Con.truee, Con.falsee], fn c =>
			 setConRep(c, ConRep.Useful));
	    
	    List.foreach(globals, handleBind);
	    List.foreach(functions, fn {body, ...} => consUseful body)
	 end

      (* Remove useless constructors from datatypes.
       * Remove datatypes which have no cons.
       *)
      val datatypes =
	 List.fold
	 (datatypes, [], fn ({tycon, cons}, datatypes) =>
	  let val cons = List.keepAll(cons, conIsUseful o #con)
	  in case cons of
	     [] => (setTyconReplacement(tycon, Type.unit);
		    datatypes)
	   | _ => {tycon = tycon, cons = cons} :: datatypes
	  end)

      (* Build the dependents for each tycon. *)
      val _ =
	 let
	    val {get = isDatatype, set = setDatatype, destroy} =
	       Property.newDest(Tycon.plist, Property.initConst false)
	 in List.foreach
	    (datatypes, fn {tycon, ...} => setDatatype(tycon, true));

	    List.foreach
	    (datatypes, fn {tycon, cons} =>
	     let
		val {get = isDependent, set = setDependent, destroy} =
		   Property.newDest(Tycon.plist, Property.initConst false)
		fun setTypeDependents t =
		   let val (tycon', ts) = Type.tyconArgs t
		   in if isDatatype tycon'
			 then if isDependent tycon'
				 then ()
			      else (setDependent(tycon', true);
				    ListRef.push(#dependents(tyconInfo tycon'),
						 tycon))
		      else List.foreach(ts, setTypeDependents)
		   end
	     in List.foreach(cons, fn {args, ...} =>
			     List.foreach(args, setTypeDependents));
		destroy()
	     end);
	    
	    destroy()
	 end

      local open Type Cardinality
      in
	 fun typeCardinality t =
	    case dest t of
	       Ref t => (case typeCardinality t of
			    Zero => Zero
			  | _ => Many)
	     | Tuple ts => tupleCardinality ts
	     | Datatype tycon => tyconCardinality tycon
	     | _ => Many
	 and tupleCardinality ts = mul(List.map(ts, typeCardinality))
      end

(*       val typeCardinality =
 * 	 Trace.trace("typeCardinality", Type.layout, Cardinality.layout)
 * 	 typeCardinality
 *)
      fun conCardinality{con, args} = tupleCardinality args

(*       val conCardinality =
 * 	 Trace.trace("conCardinality", Con.layout o #con, Cardinality.layout)
 * 	 conCardinality
 *)
      (* Compute the tycon cardinalitues with a fixed point,
       * initially assuming every datatype tycon cardinality is Zero.
       *)
      val _ =
	 let
	    (* list of datatype tycons whose cardinality has not yet stabilized *)
	    val worklist = ref(List.map(datatypes, #tycon))
	    fun loop() =
	       case !worklist of
		  [] => ()
		| tycon :: tycons =>
		     (worklist := tycons;
		      let
			 val {cons, cardinality, dependents, isOnWorklist, ...} =
			    tyconInfo tycon
			 val c = Cardinality.add(List.map(cons, conCardinality))
		      in isOnWorklist := false;
			 if Cardinality.equals(c, !cardinality)
			    then ()
			 else (cardinality := c;
			       List.foreach
			       (!dependents, fn tycon =>
				let val {isOnWorklist, ...} = tyconInfo tycon
				in if !isOnWorklist
				      then ()
				   else (isOnWorklist := true;
					 ListRef.push(worklist, tycon))
				end))
		      end;
		      loop())
	 in Control.traceCall "fixed point" loop()
	 end

      (* diagnostic *)
(*       val _ =
 * 	 List.foreach(datatypes, fn {tycon, ...} =>
 * 		      Control.message
 * 		      let open Layout
 * 		      in seq[str "cardinality of ",
 * 			     Tycon.layout tycon,
 * 			     str " = ",
 * 			     Cardinality.layout(tyconCardinality tycon)]
 * 		      end)
 *)
	 	 
(*      val setTyconReplacement =
	 Trace.trace2("setTyconReplacement",
		      Tycon.layout, Type.layout, Layout.ignore)
	 setTyconReplacement *)

      fun transparent(tycon, con, args) =
	 (setTyconReplacement(tycon, Type.tuple args);
	  setConRep(con, ConRep.Transparent);
	  setTyconNumCons(tycon, 1))

(*       val transparent =
 * 	 Trace.trace3
 * 	 ("transparent",
 * 	  Tycon.layout, Con.layout, List.layout Type.layout, Unit.layout)
 * 	 transparent
 *)

      (* "unary" is datatatypes with one constructor whose rhs contains an
       * array (or vector) type.
       * For datatypes with one variant not containing an array type, eliminate
       * the datatype. 
       *)
      val (datatypes, unary) =
	 List.fold
	 (datatypes, ([], []), fn ({tycon, cons}, (datatypes, unary)) =>
	  let
	     (* remove all cons with zero cardinality and mark them as useless *)
	     val cons =
		List.fold
		(cons, [], fn (c as {con, ...}, cons) =>
		 case conCardinality c of
		    Cardinality.Zero => (setConRep(con, ConRep.Useless);
					 cons)
		  | _ => c :: cons)
	  in case cons of
	     [] => (setTyconNumCons(tycon, 0);
		    setTyconReplacement(tycon, Type.unit);
		    (datatypes, unary))
	   | [{con, args}] =>
		if List.exists(args, fn t =>
			       Type.containsTycon(t, Tycon.array)
			       orelse Type.containsTycon(t, Tycon.vector))
		   then (datatypes,
			 {tycon = tycon, con = con, args = args}
			 :: unary)
		else (transparent(tycon, con, args);
		      (datatypes, unary))
	   | _ => ({tycon = tycon, cons = cons} :: datatypes,
		   unary)
	  end)

      fun containsTycon(ty: Type.t, tyc: Tycon.t): bool =
	 let open Type
	    fun loop t =
	       case dest t of
		  Tuple ts => List.exists(ts, loop)
		| Array t => loop t
		| Vector t => loop t
		| Ref t => loop t
		| Datatype tyc' =>
		     (case tyconReplacement tyc' of
			 NONE => Tycon.equals(tyc, tyc')
		       | SOME t => loop t)
		| _ => false
	 in loop ty
	 end

      (* Keep the circular transparent cons, ditch the rest. *)
      val datatypes =
	 List.fold
	 (unary, datatypes, fn ({tycon, con, args}, accum) =>
	  if List.exists(args, fn arg => containsTycon(arg, tycon))
	     then {tycon = tycon, cons = [{con = con, args = args}]} :: accum
	  else (transparent(tycon, con, args);
		accum))

      val {get, set, destroy = destroySimplifyType} =
	 Property.newDest(Type.plist, Property.initConst NONE)

      fun simplifyType t =
	 case get t of
	    SOME t => t
	  | NONE => 
	       let
		  open Type
		  val t' =
		     case dest t of
			Tuple ts => Type.tuple(keepSimplifyTypes ts)
		      | Array t => array(simplifyType t)
		      | Vector t => vector(simplifyType t)
		      | Ref t => reff(simplifyType t)
		      | Datatype tycon => 
			   (case tyconReplacement tycon of
			       SOME t =>
				  let val t = simplifyType t
				  in setTyconReplacement(tycon, t);
				     t
				  end
			     | NONE => t)
		      | _ => t
	       in set(t, SOME t');
		  t'
	       end
      and simplifyTypes ts = List.map(ts, simplifyType)
      and keepSimplifyTypes ts =
	 List.keepAll(simplifyTypes ts, not o Type.isUnit)

       val simplifyType =
 	 Trace.trace("simplifyType", Type.layout, Type.layout)
 	 simplifyType

      (* Simplify constructor argument types. *)
      val datatypes =
	 List.map
	 (datatypes, fn {tycon, cons} =>
	  (setTyconNumCons(tycon, List.length cons) ;
	   {tycon = tycon,
	    cons = List.map(cons, fn {con, args} =>
			    {con = con,
			     args = keepSimplifyTypes args})}))

      val unitVar = Var.newNoname()

      val {get = varInfo: Var.t -> Type.t,
	   set = setVarInfo} =
	 Property.new(Var.plist, Property.initRaise("varInfo", Var.layout))

      fun simplifyVarType(x: Var.t, t: Type.t): Type.t =
	 (setVarInfo(x, t);
	  simplifyType t)

      val oldVarType = varInfo
      val newVarType = simplifyType o oldVarType

      fun simplifyVar(x: Var.t): Var.t =
	 if Type.isUnit(newVarType x)
	    then unitVar
	 else x

      fun simplifyVars xs = List.map(xs, simplifyVar)

      fun removeUselessVars xs =
	 List.keepAll(xs, not o Type.isUnit o newVarType)

      fun tuple xs =
	 case removeUselessVars xs of
	    [x] => Var x
	  | xs => Tuple xs

      fun simplifyFormals xts =
	 List.keepAll(List.map(xts, fn (x, t) => (x, simplifyVarType(x, t))),
		      fn (_, t) => not(Type.isUnit t))

      val typeIsUseful = not o Type.isUnit o simplifyType

      (* Compute the new offset for a select, given the list of types
       * comprising the original tuple.
       * Returns true if the component is the only component of the tuple
       * that exists.
       *)
      fun newOffset(ts: Type.t list, n: int): int * bool =
	 let
	    fun loop(ts, n, i) =
	       case ts of
		  [] => Error.bug "newOffset"
		| t :: ts =>
		     if n = 0
			then (i, (i = 0
				  andalso not(List.exists(ts, typeIsUseful))))
		     else loop(ts, n - 1,
			       if typeIsUseful t then i + 1 else i)
	 in loop(ts, n, 0)
	 end

      datatype 'a result =
	 Bugg
	| Delete
	| Keep of 'a

      fun simplifyPrimExp(e: PrimExp.t): PrimExp.t result =
	  case e of
	     Const _ => Keep e
	   | Var _ => Keep e
	   | Tuple xs => Keep(tuple xs)
	   | Select{tuple, offset} =>
		let
		  val (offset, isOne) =
		     newOffset(Type.detuple(oldVarType tuple), offset)
	       in if isOne
		     then Keep(Var tuple)
		  else Keep(Select{tuple = tuple, offset = offset})
	       end
	   | ConSelect{variant, offset, con} =>
		let val args = conArgs con
		   val offset = #1(newOffset(args, offset))
		in case conRep con of
		   ConRep.Useless => Bugg
		 | ConRep.Transparent =>
		      Keep(case args of
			      [_] => Var variant
			    | _ => Select{tuple = variant, offset = offset})
		 | ConRep.Useful =>
		      Keep(ConSelect{variant = variant, con = con,
				     offset = offset})
		end
	   | ConApp{con, args} =>
		(case conRep con of
		    ConRep.Transparent => Keep(tuple args)
		  | ConRep.Useful =>
		       Keep(ConApp{con = con, args = removeUselessVars args})
		  | ConRep.Useless => Bugg)
	   | PrimApp{prim, targs, args} =>
		Keep
		(let 
		    fun normal() =
		       PrimApp{prim = prim,
			       targs = simplifyTypes targs,
			       args = simplifyVars args}
		 in case Prim.name prim of
		    Prim.Name.Equal =>
		       (case args of
			   [a, _] => if Type.isUnit(newVarType a)
					then ConApp{con = Con.truee,
						    args = []}
				     else normal()
			 | _ => Error.bug "strange Equal PrimApp")
		  | _ => normal()
		 end)

      val {get = jumpIsNullary: Jump.t -> bool, set = setJumpNullary} =
	 Property.new(Jump.plist, Property.initConst false)

      fun simplifyTransfer(t : Transfer.t): Transfer.t =
	 case t of
	    Call{func, args, cont} =>
	       Call{func = func, cont = cont, args = removeUselessVars args}
	  | Jump{dst, args} => Jump{dst = dst, args = removeUselessVars args}
	  | Return xs => Return(removeUselessVars xs)
	  | Raise xs => Raise(removeUselessVars xs)
	  | Halt => Halt
	  | Bug => Bug
	  | Case{cause, test, cases, default} =>
	       let
		  val cases =
		     List.keepAll(cases, fn (con, _) =>
				  not(ConRep.isUseless(conRep con)))
		  val default =
		     case (cases, default) of
			(_,     NONE)    => NONE
		      | ([],    SOME j)  => SOME j
		      | (_,     SOME j)  =>
			   if List.length cases
			      = tyconNumCons(Type.tycon(oldVarType test))
			      then NONE
			   else SOME j
		  fun normal() =
		     Case{cause = cause, test = test, cases = cases,
			  default = default}
	       in case (cases, default) of
		  ([],         NONE)   => Bug
		| ([],         SOME j) => Jump{dst = j, args = []}
		| ([(con, j)], NONE)   =>
		     if ConRep.isUseful(conRep con)
			then normal() (* How can this occur? - from an array ??*)
		     else
			Jump{dst = j,
			     args = if jumpIsNullary j then [] else [test]}
		| _ => normal()
	       end

(*       val simplifyTransfer =
 * 	 Trace.trace("simplifyTransfer", Transfer.layout, Transfer.layout)
 * 	 simplifyTransfer
 *)

      fun simplifyBind{var, ty, exp} =
	 let val ty = simplifyVarType(var, ty)
	    fun normal() =
	       case simplifyPrimExp exp of
		  Bugg => Bugg
		| Delete => Delete
		| Keep e => Keep{var = var, ty = ty, exp = e}
	 in if Type.isUnit ty
	       then (case exp of
			PrimApp{prim, ...} =>
			   if Prim.maySideEffect prim
			      then normal()
			   else Delete
		      | _ => Delete)
	    else normal()
	 end

      fun simplifyExp e =
	 let val {decs, transfer} = Exp.dest e
	    fun loop(decs, accum) =
	       let fun done t = Exp.make{decs = List.reverse accum,
					 transfer = t}
	       in case decs of
		  [] => done(simplifyTransfer transfer)
		| d :: decs =>
		     case simplifyDec d of
			Bugg => done Bug
		      | Delete => loop(decs, accum)
		      | Keep d => loop(decs, d :: accum)
	       end
	 in loop(decs, [])
	 end
      and simplifyDec(d: Dec.t): Dec.t result =
	 case d of
	    Bind b => (case simplifyBind b of
			  Bugg => Bugg
			| Delete => Delete
			| Keep b => Keep(Bind b))
	  | Fun{name, args, body} =>
	       (case args of
		   [] => setJumpNullary(name, true)
		 | _ => ()
		; if List.exists(args, fn (_, t) =>
				 case Type.dest t of
				    Type.Variant(_, c) => conIsUseless c
				  | _ => false)
		     then Delete
		  else Keep(Fun{name = name,
				args = simplifyFormals args,
				body = simplifyExp body}))
	  | _ => Keep d

      val globals =
	 {var = unitVar,
	  ty = Type.unit,
	  exp = Tuple[]}
	 :: List.reverse(List.fold(globals, [], fn (b, bs) =>
				   case simplifyBind b of
				      Bugg => Error.bug "global bind can't fail"
				    | Delete => bs
				    | Keep b => b :: bs))

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  {name = name,
	   args = simplifyFormals args,
	   body = shrinkExp(simplifyExp body),
	   returns = keepSimplifyTypes returns})

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
	 
   in destroySimplifyType();
      Program.clear program;

      program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SIMPLIFY_STRUCTS = 
   sig
      include SHRINK
   end

signature SIMPLIFY = 
   sig
      include SIMPLIFY_STRUCTS
      
      val simplify: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CONSTANT_PROPAGATION_STRUCTS = 
   sig
      include SHRINK
   end

signature CONSTANT_PROPAGATION = 
   sig
      include CONSTANT_PROPAGATION_STRUCTS
      
      val simplify: Program.t -> Program.t
   end
signature DIRECTED_GRAPH_STRUCTS = 
   sig
   end

signature DIRECTED_GRAPH = 
   sig
      include DIRECTED_GRAPH_STRUCTS
      
      structure Node :
	 sig
	    include T

	    val successors: t -> t list
	    val plist: t -> PropertyList.t
	 end
      
      structure Edge :
	 sig
	    type t = {from: Node.t, to: Node.t}
	 end

      (* the main graph type *)
      type t

      val new: unit -> t
      val newNode: t -> Node.t
      val addEdge: t * Edge.t -> unit

      (* error if edge isn't there *)
      val removeEdge: t * Edge.t -> unit

      val nodes: t -> Node.t list

      val foreachNode: t * (Node.t -> unit) -> unit
	       
      val foreachEdge: t * (Edge.t -> unit) -> unit

(*      val isCyclic: t -> bool*)

(*      val random: {numNodes: int, numEdges: int} -> t*)

      (* depth first search *)
	 
      structure DfsParam :
	 sig
	    type t = {startNode: Node.t -> unit,
		      finishNode: Node.t -> unit,
		      handleTreeEdge: Edge.t -> unit,
		      handleNonTreeEdge: Edge.t -> unit,
		      startTree: Node.t -> unit,
		      finishTree: Node.t -> unit,
		      finishDfs: unit -> unit}
	    val finishNode: (Node.t -> unit) -> t
	    val ignore: 'a -> unit
	    val combine: t * t -> t
	 end

      val dfsNodes: t * Node.t list * DfsParam.t -> unit
      val dfs: t * DfsParam.t -> unit

      val removeBackEdges: t -> unit

      val foreachDescendent: t * Node.t * (Node.t -> unit) -> unit

      val discoverFinishTimes: t -> {discover: Node.t -> int,
				     finish: Node.t -> int,
				     destroy: unit -> unit,
				     param: DfsParam.t}

      val display:
	 {graph: t,
	  layoutNode: Node.t -> Layout.t,
	  display: Layout.t -> unit} -> unit

      (* strongly-connected components.
       * Each component is given as a list of nodes.
       * The components are returned topologically sorted.
       *)
      val stronglyConnectedComponents: t -> Node.t list list
	 
      exception TopologicalSort
      val topologicalSort: t -> Node.t list
(*
      val transpose: t -> t
*)
      (* input/output *)
(*
      exception Input
      val input: In.t * (In.t -> 'a)* (In.t -> 'b)
	  -> t * 'a * (Edge.t -> 'b)

      val output: t * Out.t -> unit*)
   end
functor DirectedGraph(S: DIRECTED_GRAPH_STRUCTS): DIRECTED_GRAPH = 
struct

open S

structure Node =
   struct
      datatype t = T of {successors: t list ref,
			 plist: PropertyList.t}

      fun layout _ = Layout.str "node"

      fun successors(T{successors, ...}) = !successors
      fun plist(T{plist, ...}) = plist

      fun equals(n,n') = PropertyList.equals(plist n, plist n')

      fun new() = T{successors = ref [],
		    plist = PropertyList.new()}

      fun addSuccessor(T{successors, ...}, n') = ListRef.push(successors, n')

      fun removeSuccessor(T{successors, ...}, n') =
	 successors := List.removeFirst(!successors, fn n => equals(n, n'))

      val equals = fn (n,n') => PropertyList.equals(plist n, plist n')
   end

structure Edge =
   struct
      type t = {from: Node.t, to: Node.t}
   end

structure DfsParam =
   struct
      type t = {startNode: Node.t -> unit,
		finishNode: Node.t -> unit,
		handleTreeEdge: Edge.t -> unit,
		handleNonTreeEdge: Edge.t -> unit,
		startTree: Node.t -> unit,
		finishTree: Node.t -> unit,
		finishDfs: unit -> unit}
	 
      fun ignore _ = ()

      fun finishNode f = {finishNode = f,
			  startNode = ignore,
			  handleTreeEdge = ignore,
			  handleNonTreeEdge = ignore,
			  startTree = ignore,
			  finishTree = ignore,
			  finishDfs = ignore}

      fun startNode f = {finishNode = ignore,
			 startNode = f,
			 handleTreeEdge = ignore,
			 handleNonTreeEdge = ignore,
			 startTree = ignore,
			 finishTree = ignore,
			 finishDfs = ignore}
	 
      fun seq f g a = (f a ; g a)
	 
      fun combine({startNode, finishNode,
		   handleTreeEdge, handleNonTreeEdge,
		   startTree, finishTree, finishDfs}: t,
		  {startNode = sn, finishNode = fin,
		   handleTreeEdge = ht, handleNonTreeEdge = hn,
		   startTree = st, finishTree = ft, finishDfs = fd}: t): t =
	 {startNode = seq startNode sn,
	  finishNode = seq finishNode fin,
	  handleTreeEdge = seq handleTreeEdge ht,
	  handleNonTreeEdge = seq handleNonTreeEdge hn,
	  startTree = seq startTree st,
	  finishTree = seq finishTree ft,
	  finishDfs = seq finishDfs fd}
   end

(*---------------------------------------------------*)
(*                  graph datatype                   *)
(*---------------------------------------------------*)

datatype t = T of {nodes: Node.t list ref}

fun nodes(T{nodes, ...}) = !nodes

fun new() = T{nodes = ref []}
   
fun newNode(T{nodes, ...}) =
   let val n = Node.new()
   in ListRef.push(nodes, n) ;
      n
   end

fun addEdge(_, {from, to}) = Node.addSuccessor(from, to)
fun removeEdge(_, {from, to}) = Node.removeSuccessor(from, to)

(*--------------------------------------------------------*)
(*                   Depth-First Search                   *)
(*--------------------------------------------------------*)

fun dfsNodes(g, ns, {startNode, finishNode,
		     handleTreeEdge, handleNonTreeEdge,
		     startTree, finishTree, finishDfs}) =
   let
      val {get = hasBeenVisited, set, destroy, ...} =
	 Property.newDest(Node.plist, Property.initConst false)

      fun visit n =
	 (startNode n ;
	  set(n, true) ;
	  List.foreach(Node.successors n, fn n' =>
		       let val edge = {from = n, to = n'}
		       in if hasBeenVisited n'
			     then handleNonTreeEdge edge
			  else (visit n' ; handleTreeEdge edge)
		       end) ;
	  finishNode n)
	       
   in List.foreach(ns, fn n =>
		   if hasBeenVisited n
		      then ()
		   else (startTree n ; visit n ; finishTree n)) ;
      destroy() ;
      finishDfs()
   end

fun dfs(g, p) = dfsNodes(g, nodes g, p)

fun display{graph, layoutNode, display} =
   dfs(graph,
       DfsParam.startNode(fn n =>
			  display(let open Layout
				  in seq[layoutNode n,
					 str " ",
					 list(List.map(Node.successors n,
						       layoutNode))]
				  end)))
			     
   
fun foreachDescendent(g, n, f) =
   dfsNodes(g, [n], DfsParam.finishNode f)

fun removeBackEdges g =
   let
      val discoverTime = Counter.new 0
      val {get, destroy, ...} =
	 Property.newDest
	 (Node.plist, Property.initFun(fn _ => {time = Counter.next discoverTime,
						alive = ref true}))
      val ignore = DfsParam.ignore
   in dfs
      (g, {startNode = fn n => (get n; ()),
	   finishNode = fn n => #alive(get n) := false,
	   handleNonTreeEdge =
	   fn e as {from, to} => let val {alive, time} = get to
				 in if !alive andalso time < #time(get from)
				       then removeEdge(g, e)
				    else ()
				 end,
	   handleTreeEdge = ignore,
	   startTree = ignore,
	   finishTree = ignore,
	   finishDfs = ignore})
   end

(*--------------------------------------------------------*)
(*                         Times                          *)
(*--------------------------------------------------------*)

fun discoverFinishTimes g =
    let val time = ref 0
	val {get = discover, set = setDiscover, destroy = destroyDiscover, ...} =
	   Property.newDest
	   (Node.plist, Property.initRaise("discover", Node.layout))
	val {get = finish, set = setFinish, destroy = destroyFinish, ...} =
	   Property.newDest(Node.plist, Property.initRaise("finish", Node.layout))
	   
    in {destroy = fn () => (destroyDiscover() ; destroyFinish()),
	discover = discover,
	finish = finish,
	param = {startNode = fn n => (time := !time + 1 ; setDiscover(n, !time)),
		 finishNode = fn n => (time := !time + 1 ; setFinish(n, !time)),
		 handleTreeEdge = DfsParam.ignore, handleNonTreeEdge = DfsParam.ignore,
		 startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
		 finishDfs = DfsParam.ignore}}
    end
(*
local
   fun undirected [] = []
     | undirected (n::ns) = 
       List.append(List.map(ns, fn n' => (n,n')),
		   undirected ns)
   fun directed ns =
      let val es = undirected ns
      in List.append(es, List.map(es, fn (n,n') => (n',n)))
      end
   val possible = if isDirected then directed else undirected
in val possibleEdges = possible o nodes
end
*)
(*--------------------------------------------------------*)
(*                        Foreach                         *)
(*--------------------------------------------------------*)

fun foreachNode(g, f) = List.foreach(nodes g, f)
   
fun foreachEdge(g, handleEdge) =
   dfs(g, {handleTreeEdge = handleEdge,
	  handleNonTreeEdge = handleEdge,
	  startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	  startNode = DfsParam.ignore, finishNode = DfsParam.ignore,
	  finishDfs = DfsParam.ignore})

(*--------------------------------------------------------*)
(*                         Input                          *)
(*--------------------------------------------------------*)
(*
exception Input
fun input(instream, inputInfo, inputEdgeInfo) =
   let val numNodes = Int.input instream
      val numEdges = Int.input instream
      val info = inputInfo instream
      val g = new numNodes
      val a = A.new(numEdges, NONE)
      fun inputEdge i =
	 (addEdge(g, Node.input instream, Node.input instream) ;
	  A.update(a, i, SOME(inputEdgeInfo instream)))
      val edgeInfo = Option.projector (fn e => A.sub(a, Edge.index e))
   in (Int.foreach(0, numEdges - 1, inputEdge) ;
       (g, info, edgeInfo))
   end handle _ => raise Input
*)
(*--------------------------------------------------------*)
(*                         Output                         *)
(*--------------------------------------------------------*)
(*
fun output(g, out) =
   let val print = Out.outputc out
   in (print "nodes = " ;
       Layout.output(Int.layout(numNodes g), out) ;
       print "\nedges =" ;
       Layout.output(Int.layout(numEdges g), out) ;
       (*outputInfo out ;*)
       foreachEdge(g, fn e => (print "\n" ;
			      Edge.output(e, out))))
			      (*print " " ;
			       outputEdgeInfo(e, out)*)
   end
	*)		      
(*--------------------------------------------------------*)
(*                         Random                         *)
(*--------------------------------------------------------*)
(*
fun maxNumEdges n = n * (n - 1)
   
fun random {numNodes,numEdges} =
   let val max = maxNumEdges numNodes
   in if numNodes < 0 orelse numEdges < 0 orelse numEdges > max
	 then Error.error "random"
      else let val g = new()
	       val needed = ref numEdges
	       val remaining = ref max
	       fun maybeAddEdge(n,n') =
		  (if Int.random(1, !remaining) <= !needed
		      then (addEdge(g, Node.fromInt n, Node.fromInt n') ;
			    IntRef.dec needed)
		   else () ;
		   IntRef.dec remaining)
	       val minNode = 0
	       val maxNode = numNodes - 1
	       fun directed n =
		  Int.foreach(0, maxNode, fn n' =>
			      if n = n' then () else maybeAddEdge(n,n'))
	       fun undirected n =
		  Int.foreach(n + 1, maxNode, fn n' => maybeAddEdge(n,n'))
	       val addEdges = if isDirected then directed
			      else undirected
	   in (Int.foreach(minNode, maxNode, addEdges) ;
	       g)
	   end
   end
*)
(*--------------------------------------------------------*)
(*                         Cycle                          *)
(*--------------------------------------------------------*)
(*
fun cycleParam g =
   let val {get = isActive, set = setActive} =
      nodeInfo(g, fn _ => false)
      val cycle = ref false
   in (cycle, {startNode = fn n => setActive(n, true),
	       finishNode = fn n => setActive(n, false),
	       handleNonTreeEdge =
	       fn (n, e) => let val n' = Edge.otherNode(e,n)
			   in if isActive n' then cycle := true
			      else ()
			   end,
			handleTreeEdge = DfsParam.ignore,
			startTree = DfsParam.ignore,
			finishTree = DfsParam.ignore,
			finishDfs = DfsParam.ignore})
   end

fun isCyclic g = let val (cycle, p) = cycleParam g
		 in (dfs(g, p) ; !cycle)
		 end
*)

(*--------------------------------------------------------*)
(*                    Topological Sort                    *)
(*--------------------------------------------------------*)

exception TopologicalSort

fun topSortParam g =
   let
      val {get = amVisiting, set, destroy, ...} =
	 Property.newDest(Node.plist, Property.initConst false)
      val ns = ref []
   in (ns, {startNode = fn n => set(n, true),
	    finishNode = fn n => (set(n, false) ; ListRef.push(ns,n)),
	    handleNonTreeEdge =
	    fn {from, to} => if amVisiting to
			       then raise TopologicalSort
			    else (),
	    startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	    handleTreeEdge = DfsParam.ignore,
	    finishDfs = fn () => destroy()})
   end

fun topologicalSort g = let val (ns, p) = topSortParam g
			in (dfs(g, p) ; !ns)
			end

(*--------------------------------------------------------*)
(*                       Transpose                        *)
(*--------------------------------------------------------*)
(*
fun transposeParam g =
   let val gt = new()
      fun handleEdge(n, e) = let val n' = Edge.otherNode(e,n)
			    in (addEdge(gt,n',n) ; ())
			    end
   in (gt, {handleTreeEdge = handleEdge,
	    handleNonTreeEdge = handleEdge,
	    finishDfs = DfsParam.ignore,
	    startNode = DfsParam.ignore, finishNode = DfsParam.ignore,
	    startTree = DfsParam.ignore, finishTree = DfsParam.ignore})
   end

fun transpose g = let val (gt, p) = transposeParam g
		  in (dfs(g, p) ; gt)
		  end
  *) 
(*--------------------------------------------------------*)
(*             Strongly Connected Components              *)
(*--------------------------------------------------------*)
   
(* from Cormen, Leiserson, and Rivest 23.5 *)
(*
fun sccCLR g =
   let
      val (gt, p) = transposeParam g
      val ns = ref []
      val p' = P.finishNode(fn n => ListRef.push(ns,n))
      val components = ref []
      val component = ref []
      fun startNode n = ListRef.push(component,n)
      fun startTree _ = component := []
      fun finishTree _ = ListRef.push(components, !component)
      val pt = {startNode = startNode,
		   startTree = startTree,
		   finishTree = finishTree,
		   finishNode = DfsParam.ignore,
		   finishDfs = DfsParam.ignore,
		   handleTreeEdge = DfsParam.ignore,
		   handleNonTreeEdge = DfsParam.ignore}
   in (dfs(g, P.combine(p, p')) ;
       dfsNodes(gt, !ns, pt) ;
       !components)
   end
*)

(* from Aho, Hopcroft, Ullman section 5.5 *)

fun stronglyConnectedComponents g =
   let
      val {get = discover, set = setDiscover, destroy = destroyDiscover, ...} =
	 Property.newDest(Node.plist, Property.initRaise("discover", Node.layout))
      val {get = low, set = setLow, destroy = destroyLow, ...} =
	 Property.newDest(Node.plist, Property.initRaise("low", Node.layout))
      val {get = isOnStack, set = setOnStack, destroy = destroyStack, ...} =
	 Property.newDest(Node.plist, Property.initConst false)
	 
      val stack = ref []
      val components = ref []
      val time = ref 0
	 
      fun pop() = let val n = ListRef.pop stack
		  in (setOnStack(n, false) ; n)
		  end
      fun popTo n = let fun popTo() = let val n' = pop()
				      in if Node.equals(n,n') then [n]
					 else n' :: (popTo())
				      end
		    in popTo()
		    end

      fun startNode n = (time := !time + 1 ;
			 setDiscover(n, !time) ;
			 setLow(n, !time) ;
			 setOnStack(n, true) ;
			 ListRef.push(stack,n))
	 
      fun finishNode n = if discover n = low n
			     then ListRef.push(components, popTo n)
			 else ()
			    
      fun updateLow{from, to} = if low to < low from then setLow(from, low to)
			       else ()
			       
      val handleTreeEdge = updateLow
	 
      fun handleNonTreeEdge e = 
	 if isOnStack(#to e)
	    then updateLow e
	 else ()
      
      val p = {startNode = startNode, finishNode = finishNode,
	       handleTreeEdge = handleTreeEdge,
	       handleNonTreeEdge = handleNonTreeEdge,
	       startTree = DfsParam.ignore, finishTree = DfsParam.ignore,
	       finishDfs = DfsParam.ignore}
	 
   in dfs(g, p) ;
      destroyLow() ;
      destroyStack() ;
      destroyDiscover() ;
      !components
   end

end
structure DirectedGraph = DirectedGraph()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ONCE_STRUCTS = 
   sig
      include INFER_HANDLERS
   end

signature ONCE = 
   sig
      include ONCE_STRUCTS

      (* Returns true for let bound variables that are guaranteed to
       * be bound no more than once during the entire program.
       *)
      val once: Program.t -> Var.t -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Once(S: ONCE_STRUCTS): ONCE = 
struct

open S
open Dec PrimExp Transfer

structure Graph = DirectedGraph
structure Node = Graph.Node

(*
 * Build control flow graph for main function.
 * Compute strongly-connected components and set the vars for any basic
 * block that is a trivial component and does not jump to itself to be once.
 *)
   
fun once(program as Program.T{globals, functions, main, ...}) =
   let
      val {get = once, set = setOnce} =
	 Property.new(Var.plist, Property.initConst false)

      val {set = setNode, get = jumpNode} =
	 Property.new(Jump.plist, Property.initRaise("info", Jump.layout))

      val {set = setNodeInfo, get = nodeInfo} =
	 Property.new(Node.plist, Property.initRaise("jump", Node.layout))

      val nodeVars = #vars o nodeInfo
      val nodeName = #name o nodeInfo

      val jumpHandlers = inferHandlers program

      val graph = Graph.new()

      fun loop(e: Exp.t, vars: Var.t list ref, node: Node.t,
	       handlers: Jump.t list) =
	 let val {decs, transfer} = Exp.dest e
	    fun jump j = Graph.addEdge(graph, {from = node, to = jumpNode j})
	    fun handler() =
	       case List.fold(decs, handlers, deltaHandlers) of
		  h :: _ => jump h
		| _ => ()
	 in List.foreach
	    (decs, 
	     fn Bind{var, ...} => ListRef.push(vars, var)
	      | Fun{name, body, ...} =>
		   let val n = Graph.newNode graph
		      val vars = ref []
		   in setNodeInfo(n, {name = Jump.layout name,
				      vars = vars});
		      setNode(name, n);
		      loop(body, vars, n, jumpHandlers name)
		   end
	      | _ => ());
	    
	    case transfer of
	       Call{cont, ...} => (case cont of
				      NONE => ()
				    | SOME c => (jump c; handler()))
	     | Jump{dst, ...} => jump dst
	     | Case{cases, default, ...} => (List.foreach(cases, jump o #2);
					     case default of
						NONE => ()
					      | SOME j => jump j)
	     | Raise _ => handler()
	     | _ => ()
	 end

      val {body, ...} =
	 case List.peek(functions, fn {name, ...} => Func.equals(name, main)) of
	    NONE => Error.bug "no main"
	  | SOME r => r

      val _ = 
	 let val n = Graph.newNode graph
	    val vars = ref []
	 in setNodeInfo(n, {name = Layout.str "main",
			    vars = vars});
	    loop(body, vars, n, [])
	 end
   in
      Control.displays
      ("once-graph", fn display =>
       (display(Exp.layout body);
	display(Layout.str "\n\n");
	Graph.display{graph = graph,
		      display = display,
		      layoutNode = nodeName}));
	
      List.foreach
      (Graph.stronglyConnectedComponents graph,
       fn [] => ()
	| [n] =>
	     if List.exists(Node.successors n, fn n' => Node.equals(n, n'))
		then ()
	     else List.foreach(!(nodeVars n), fn x => setOnce(x, true))
	   | _ => ());

      List.foreach(globals, fn {var, ...} => setOnce(var, true));
      
      Control.displays
      ("once-vars", fn display =>
       Exp.foreachVar(body, fn (x, _) => if once x then display(Var.layout x)
					 else ()));

      once
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature GLOBAL_STRUCTS = 
   sig
      include CPS_TREE
   end

signature GLOBAL = 
   sig
      include GLOBAL_STRUCTS
      
      val make:
	 unit -> {
		  new: Type.t * PrimExp.t -> Var.t,
		  all: unit -> {var: Var.t, ty: Type.t, exp: PrimExp.t} list
		 }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Global(S: GLOBAL_STRUCTS): GLOBAL = 
struct

open S
open PrimExp

fun equalss(cs, cs') = List.equals(cs, cs', Var.equals)

val primExpEquals =
   fn (Const c, Const c') => Const.equals(c, c')
    | (PrimApp{prim = p, targs = ts, ...}, PrimApp{prim = p', targs = ts', ...}) =>
	 (case Prim.numArgs p of
	     NONE => Prim.equals(p, p') andalso List.equals(ts, ts', Type.equals)
	   | _ => false)
    | (ConApp{con = c, args}, ConApp{con = c', args = args'}) =>
	 Con.equals(c, c') andalso equalss(args, args')
    | (Tuple xs, Tuple xs') => equalss(xs, xs')
    | _ => false

(* A really simple implementation using a list of all globals to do
 * the interning.
 *)
fun make() =
   let
      val binds: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
	 	 
      fun all() = List.reverse(!binds) before binds := []

      fun new(ty: Type.t, exp: PrimExp.t): Var.t =
	 case List.peek(!binds, fn {exp = exp', ...} =>
			primExpEquals(exp, exp')) of
	    SOME{var, ...} => var
	  | NONE =>
	       let val var = Var.newString "global"
	       in ListRef.push(binds, {var = var, ty = ty, exp = exp})
		  ; var
	       end

(*       val new = Trace.trace2("Global.new", Type.layout, PrimExp.layout,
 * 			     Var.layout) new
 *)
	 
   in {new = new, all = all}
   end

(* A trickier implementation using hash consing.
 *)

structure Table = HashTable

local
   fun newHash() = Word.fromInt(Int.random())
   val tuple = newHash()
   val conApp = newHash()
   val primApp = newHash()
   fun hashVars(xs: Var.t list, w: word): word =
      List.fold(xs, tuple, fn (x, w) => Word.xorb(w, Var.hash x))
in
   val hash: PrimExp.t -> word =
      fn Const c => Const.hash c
       | Tuple xs => hashVars(xs, tuple)
       | ConApp{args, ...} => hashVars(args, conApp)
       | PrimApp{args, ...} => hashVars(args, primApp)
       | _ => Error.bug "hash not implemented for prim exp"
end

fun make() =
   let
      val binds: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
      fun all() = List.reverse(!binds) before binds := []
	 
      val size = 10000
      exception NotFound
      val table = Table.mkTable (hash, primExpEquals) (size, NotFound)

      fun new(ty: Type.t, exp: PrimExp.t): Var.t =
	 case Table.find table exp of
	    SOME x => x
	  | NONE => let val x = Var.newString "global"
		    in ListRef.push(binds, {var = x, ty = ty, exp = exp})
		       ; Table.insert table (exp, x)
		       ; x
		    end	   
   in {new = new, all = all}
   end
	    
end


(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Invariant: Created globals only refer to other globals.
 *            Hence, the newly created globals may appear at the
 *            beginning of the program.
 *
 * Circular abstract values can arise as a result of programs like:
 *   datatype t = T of t
 *   fun f() = T(f())
 *   val _ = f()
 * There is special code in printing abstract values and in determining whether
 * they are global in order to avoid infinite loops.
 *)
functor ConstantPropagation(S: CONSTANT_PROPAGATION_STRUCTS)
   : CONSTANT_PROPAGATION = 
struct

open S

structure Once = Once(S)
structure Global = Global(S)

structure Type =
   struct
      open Type

      fun isSmall t =
	 case dest t of
	    Array _ => false
	  | Vector _ => false
	  | Datatype _ => false
	  | Variant _ => false
	  | Ref t => isSmall t
	  | Tuple ts => List.forall(ts, isSmall)
	  | _ => true
   end
   
open PrimExp Dec Transfer

structure Value =
   struct
      datatype global =
	 NotComputed
       | No
       | Yes of Var.t

      structure Const =
	 struct
	    datatype t = T of {const: const ref,
			       coercedTo: t list ref}
	    and const =
	       Undefined (* no possible value *)
	      | Const of Const.t
	      | Prim of {prim: Prim.t, targs: Type.t list, ty: Type.t}
	      | Unknown (* many possible values *)

	    fun new c = T{const = ref c,
			  coercedTo = ref []}

	    fun equals(T{const = r, ...}, T{const = r', ...}) = r = r'

	    val const = new o Const
	    val prim = new o Prim

	    fun undefined() = new Undefined
	    fun unknown() = new Unknown

	    fun layout(T{const, ...}) =
	       let open Layout
	       in case !const of
		  Undefined => str "undefined constant"
		| Const c => Const.layout c
		| Prim{prim, ...} => Prim.layout prim
		| Unknown => str "unknown constant"
	       end
	    
	    fun makeUnknown(T{const, coercedTo}): unit =
	       case !const of
		  Unknown => ()
		| _ => (const := Unknown;
			List.foreach(!coercedTo, makeUnknown);
			coercedTo := [])

	    fun send(c: t, c': const): unit =
	       let
		  fun loop(c as T{const, coercedTo}) =
		     case (c', !const) of
			(_, Unknown) => ()
		      | (_, Undefined) => (const := c';
					   List.foreach(!coercedTo, loop))
		      | (Const c', Const c'') =>
			   if Const.equals(c', c'')
			      then ()
			   else makeUnknown c
		      | (Prim{prim = p', targs = ts', ...},
			 Prim{prim = p'', targs = ts'', ...}) =>
			   if Prim.equals(p', p'')
			      andalso List.equals(ts', ts'', Type.equals)
			      then ()
			   else makeUnknown c
		      | _ => makeUnknown c
	       in loop c
	       end

	    fun coerce{from = from as T{const, coercedTo}, to: t}: unit =
	       if equals(from, to)
		  then ()
	       else
		  let fun push() = ListRef.push(coercedTo, to)
		  in case !const of
		     Unknown => makeUnknown to
		   | Undefined => push()
		   | c as Const _ => (push(); send(to, c))
		   | p as Prim _ => (push(); send(to, p))
		  end

	    fun unify(c, c') =
	       (coerce{from = c, to = c'};
		coerce{from = c', to = c})
	 end

      structure One =
	 struct
	    datatype 'a t = T of {global: Var.t option ref,
				  extra: 'a}

	    fun new(a: 'a): 'a t = T{global = ref NONE,
				     extra = a}

	    fun global(T{global = g, ...}) = g
	       
	    fun equals(n, n') = global n = global n'
	 end
      
      structure Place =
	 struct
	    datatype 'a t =
	       Undefined
	     | One of 'a One.t
	     | Unknown

	    val toString =
	       fn Undefined => "Undefined"
		| One _ => "One"
		| Unknown => "Unknown"

	    fun layout b = Layout.str(toString b)
	 end

      structure Birth =
	 struct
	    datatype 'a t = T of {place: 'a Place.t ref,
				  coercedTo: 'a t list ref}

	    fun layout(T{place, ...}) = Place.layout(!place)

	    fun equals(T{place = r, ...}, T{place = r', ...}) = r = r'
	    fun new p = T{place = ref p,
			  coercedTo = ref []}

	    fun undefined(): 'a t = new Place.Undefined
	    fun unknown(): 'a t = new Place.Unknown
	    fun here(a: 'a): 'a t = new(Place.One(One.new a))

	    fun makeUnknown(T{place, coercedTo, ...}) =
	       case !place of
		  Place.Unknown => ()
		| _ => (place := Place.Unknown;
			List.foreach(!coercedTo, makeUnknown);
			coercedTo := [])

	    fun send(b: 'a t, one): unit =
	       let
		  fun loop(b as T {place, coercedTo, ...}) =
		     case !place of
			Place.Undefined => (place := Place.One one;
					    List.foreach(!coercedTo, loop))
		      | Place.One one' => if One.equals(one, one')
					     then ()
					  else makeUnknown b
		      | Place.Unknown => ()
	       in loop b
	       end

	    fun coerce{from = from as T{place, coercedTo, ...}, to} =
	       if equals(from, to)
		  then ()
	       else
		  let fun push() = ListRef.push(coercedTo, to)
		  in case !place of
		     Place.Unknown => makeUnknown to
		   | Place.One one => (push(); send(to, one))
		   | Place.Undefined => push()
		  end

	    fun unify(c, c') =
	       (coerce{from = c, to = c'};
		coerce{from = c', to = c})
	 end

      structure Set = DisjointSet
	 
      datatype t =
	 T of {
	       value: value,
	       ty: Type.t,
	       global: global ref
	      } Set.t
      and value =
          Const of Const.t
	| Datatype of data
	| Vector of {length: t,
		     elt: t}
	| Ref of {birth: {init: t} Birth.t,
		  arg: t}
	| Array of {birth: unit Birth.t,
		    length: t,
		    elt: t}
	| Tuple of t list
      and data =
	 Data of {
		  value: dataVal ref,
		  coercedTo: data list ref,
		  selects: {
			    con: Con.t,
			    offset: int,
			    to: t
			   } list ref
		 }
      and dataVal =
	 Undefined
	| ConApp of {con: Con.t,
		     args: t list,
		     uniq: unit ref}
	| Unknown

      local
	 fun make sel (T s) = sel(Set.value s)
      in val value = make #value
	 val global = make #global
	 val ty = make #ty
      end

      fun equals(T s, T s') = Set.equals(s, s')

      fun layout v =
	 let open Layout
	 in case value v of
	    Const c => Const.layout c
	  | Datatype(Data{value, ...}) =>
	       (case !value of
		   Undefined => str "undefined datatype"
		 | ConApp{con, args, ...} => Con.layout con
		      (* Can't layout the args because there may be circularity *)
		      (* tuple(List.map(args, layout)) *)
		 | Unknown => str "unknown datatype")
	  | Ref{birth, arg, ...} => seq[str "ref", tuple[Birth.layout birth,
							 layout arg]]
	  | Array{birth, length, elt, ...} => seq[str "array",
						  tuple[Birth.layout birth,
							layout length,
							layout elt]]
	  | Vector{length, elt, ...} => seq[str "vector",
					   tuple[layout length,
						 layout elt]]
	  | Tuple vs => tuple(List.map(vs, layout))
	 end

      fun globals(vs: t list, newGlobal): (Var.t * Type.t) list option =
	 let
	    fun loop(vs, gs) =
	       case vs of
		  [] => SOME(List.reverse gs)
		| v :: vs =>
		     (case global(v, newGlobal) of
			 SOME g => loop(vs, g :: gs)
		       | NONE => NONE)
	 in loop(vs, [])
	 end

      and global(v as T s, newGlobal): (Var.t * Type.t) option =
	 let val {global = r, ty, value} = Set.value s
	 in case !r of
	    No => NONE
	  | Yes g => SOME(g, ty)
	  | NotComputed =>
	       let
		  (* avoid globalizing circular abstract values *)
		  val _ = r := No
		  fun yes e = Yes(newGlobal(ty, e))
		  fun unary(Birth.T{place, ...},
			    makeInit: 'a -> t,
			    prim: Prim.t,
			    targ: Type.t) =
		     case !place of
			Place.One(One.T{global = glob, extra}) =>
			   let val init = makeInit extra
			   in case global(init, newGlobal) of
			      SOME(x, _) =>
				 Yes(case !glob of
					NONE => 
					   let
					      val exp =
						 PrimExp.PrimApp{prim = prim,
								 targs = [targ],
								 args = [x]}
					      val g = newGlobal(ty, exp)
					   in glob := SOME g; g
					   end
				      | SOME g => g)
			    | _ => No
			   end
		      | _ => No
		  val g =
		     case value of
			Const(Const.T{const, ...}) =>
			   (case !const of
			       Const.Const c => yes(PrimExp.Const c)
			     | Const.Prim{prim, targs, ...} =>
				  yes(PrimExp.PrimApp{prim = prim,
						      targs = targs,
						      args = []})
			     | _ => No)
		      | Datatype(Data{value, ...}) =>
			   (case !value of
			       ConApp{con, args, ...} =>
				  (case globals(args, newGlobal) of
				      NONE => No
				    | SOME args =>
					 yes(PrimExp.ConApp
					     {con = con,
					      args = List.map(args, #1)}))
			     | _ => No)
		      | Ref{birth, ...} =>
			   unary(birth, fn {init} => init,
				 Prim.reff, Type.deref ty)
		      | Array{birth, length, ...} =>
			   unary(birth, fn _ => length,
				 Prim.array, Type.dearray ty)
		      | Vector _ => No
		      | Tuple vs =>
			   (case globals(vs, newGlobal) of
			       NONE => No
			     | SOME xts =>
				  yes(PrimExp.Tuple(List.map(xts, #1))))
	       in r := g; global(v, newGlobal)
	       end
	 end
      
      fun new(v: value, ty: Type.t): t =
	 T(Set.singleton{value = v,
			 ty = ty,
			 global = ref NotComputed})


      fun tuple vs = new(Tuple vs, Type.tuple(List.map(vs, ty)))

      fun const c = new(Const(Const.const c), Type.ofConst c)

      fun deconst v =
	 case value v of
	    Const c => c
	  | _ => Error.bug "deconst"
	       
      local
	 fun make (err, sel) v =
	    case value v of
	       Vector fs => sel fs
	     | _ => Error.bug err
      in val devector = make("devector", #elt)
	 val vectorLength = make("vectorLength", #length)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Array fs => sel fs
	     | _ => Error.bug err
      in val dearray = make("dearray", #elt)
	 val arrayLength = make("arrayLength", #length)
	 val arrayBirth = make("arrayBirth", #birth)
      end

      local
	 fun make (err, sel) v =
	    case value v of
	       Ref fs => sel fs
	     | _ => Error.bug err
      in val deref = make("deref", #arg)
	 val refBirth = make("refBirth", #birth)
      end

      structure Data =
	 struct
	    datatype t = datatype data

	    local
	       fun make v () = Data{value = ref v,
				    coercedTo = ref [],
				    selects = ref []}
	    in val undefined = make Undefined
	       val unknown = make Unknown
	    end
	 end
      
      local
	 (* The extra birth is because of let-style polymorphism.
	  * arrayBirth is really the same as refBirth.
	  *)
	 fun make(const, data, refBirth, arrayBirth) =
	    let
	       fun loop(t: Type.t): t =
		  new
		  (case Type.dest t of
		      Type.Datatype _ => Datatype(data())
		    | Type.Variant _ => Datatype(data())
		    | Type.Ref t => Ref{birth = refBirth(),
					arg = loop t}
		    | Type.Array t => Array{birth = arrayBirth(),
					    length = loop Type.int,
					    elt = loop t}
		    | Type.Vector t => Vector{length = loop Type.int,
					      elt = loop t}
		    | Type.Tuple ts => Tuple(List.map(ts, loop))
		    | _ => Const(const()),
		   t)
	    in loop
	    end
      in
	 val fromType =
	    make(Const.undefined, Data.undefined, Birth.undefined, Birth.undefined)
	 val unknown =
	    make(Const.unknown, Data.unknown, Birth.unknown, Birth.unknown)
      end

      fun select{tuple, offset, resultType} =
	 case value tuple of
	   Tuple vs => List.nth(vs, offset)
	 | _ => Error.bug "select of non-tuple" 

      fun filter(v: t, c: Con.t): t = v

      fun unit() = tuple[]
   end

(* ------------------------------------------------- *)
(*                     simplify                      *)
(* ------------------------------------------------- *)

fun simplify(program as Program.T{datatypes, globals, functions, main})
   : Program.t =
   let
      val once = Once.once program

      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("args", Con.layout))

      val conArgTypes = #types o conInfo
      val conArgValues = #values o conInfo
      val conResult = #result o conInfo
	 
      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  let val result = Type.con(tycon, [])
	  in List.foreach
	     (cons, fn {con, args} =>
	      setConInfo(con, {result = result,
			       types = args,
			       values = List.map(args, Value.fromType)}))
	  end)
	 
      local open Value
      in
(* 	 val traceCoerce =
 * 	    Trace.trace("Value.coerce",
 * 			fn {from, to} => Layout.record[("from", layout from),
 * 						       ("to", layout to)],
 * 			Unit.layout)
 *)
	    
	 fun makeDataUnknown(Data{value, coercedTo, selects, ...}) =
	    let
	       fun doit() =
		  (value := Unknown;
		   List.foreach(!coercedTo, makeDataUnknown);
		   coercedTo := [];
		   List.foreach(!selects, fn {con, offset, to} =>
				coerce{to = to,
				       from = List.nth(conArgValues con, offset)}))
	    in case !value of
	       Unknown => ()
	     | Undefined => doit()
	     | ConApp _ => doit()
	    end

	 and sendConApp(d: data, ca as {con, args, uniq}): unit =
	    let
	       val v = ConApp ca
	       fun loop(Data{value, coercedTo, selects, ...}): unit =
		  case !value of
		     Unknown => ()
		   | Undefined =>
			(value := v;
			 List.foreach(!coercedTo, loop);
			 List.foreach
			 (!selects, fn {con = con', offset, to} =>
			  if Con.equals(con, con')
			     then coerce{from = List.nth(args, offset),
					 to = to}
			  else ()))
		   | ConApp{con = con', uniq = uniq', ...} =>
			 if uniq = uniq'
			    orelse (Con.equals(con, con')
				    andalso List.isEmpty args)
			    then ()
			 else makeDataUnknown d
	    in loop d
	    end
				
	 and coerces{froms, tos} =
	    List.foreach2(froms, tos, fn (from, to) =>
			  coerce{from = from, to = to})

	 and coerce arg =
(*	    traceCoerce *)
	    (fn {from, to} =>
	    if equals(from, to)
	       then ()
	    else
	       case (value from, value to) of
		  (Const from, Const to) => Const.coerce{from = from, to = to}
		| (Datatype from, Datatype to) => coerceData{from = from, to = to}
		| (Ref{birth, arg}, Ref{birth = b', arg = a'}) =>
		     (Birth.coerce{from = birth, to = b'};
		      unify(arg, a'))
		| (Array{birth = b, length = n, elt = x},
		   Array{birth = b', length = n', elt = x'}) =>
		     (Birth.coerce{from = b, to = b'};
		      coerce{from = n, to = n'};
		      unify(x, x'))
		| (Vector{length = n, elt = x}, Vector{length = n', elt = x'}) =>
		     (coerce{from = n, to = n'};
		      coerce{from = x, to = x'})
		| (Tuple vs, Tuple vs') => coerces{froms = vs, tos = vs'}
		| _ => Error.bug "strange coerce") arg

	 and unify(T s: t, T s': t): unit =
	    if Set.equals(s, s')
	       then ()
	    else
	       let val {value, ...} = Set.value s
		  val {value = value', ...} = Set.value s'
	       in Set.union(s, s');
		  case (value, value') of
		     (Const c, Const c') => Const.unify(c, c')
		   | (Datatype d, Datatype d') => unifyData(d, d')
		   | (Ref{birth, arg}, Ref{birth = b', arg = a'}) =>
			(Birth.unify(birth, b');
			 unify(arg, a'))
		   | (Array{birth = b, length = n, elt = x},
		      Array{birth = b', length = n', elt = x'}) =>
			(Birth.unify(b, b');
			 unify(n, n');
			 unify(x, x'))
		   | (Vector{length = n, elt = x}, Vector{length = n', elt = x'}) =>
			(unify(n, n');
			 unify(x, x'))
		   | (Tuple vs, Tuple vs') => List.foreach2(vs, vs', unify)
		   | _ => Error.bug "strange unify"
	       end

	 and unifyData(d, d') =
	    (coerceData{from = d, to = d'};
	     coerceData{from = d', to = d})

	 and coerceData{from = Data{value, coercedTo, ...}, to} =
	      case !value of
		 Undefined => ListRef.push(coercedTo, to)
	       | ConApp ca => (ListRef.push(coercedTo, to);
			       sendConApp(to, ca))
	       | Unknown => makeDataUnknown to

	 fun conApp{con: Con.t, args: t list}: t =
	    (coerces{froms = args, tos = conArgValues con};
	     new(Datatype
		 (Data{value = ref(ConApp{con = con, args = args, uniq = ref()}),
		       coercedTo = ref [],
		       selects = ref []}),
		 conResult con))

	 fun conSelect{variant, con, offset} =
	    case value variant of
	       Datatype(Data{value, selects, ...}) =>
		  let
		     fun new() =
			let val v = fromType(List.nth(conArgTypes con, offset))
			in ListRef.push(selects,
					{con = con, offset = offset, to = v});
			   v
			end
		  in case !value of
		     Undefined => new()
		   | Unknown => List.nth(conArgValues con, offset)
		   | ConApp{con = con', args, ...} =>
			let val v = new()
			in if Con.equals(con, con')
			      then coerce{from = List.nth(args, offset),
					  to = v}
			   else ();
			      v
			end
		  end
	     | _ => Error.bug "conSelect of non-datatype"

	 fun makeUnknown(v: t): unit =
	    case value v of
	       Const c => Const.makeUnknown c
	     | Datatype d => makeDataUnknown d
	     | Ref{arg, ...} => makeUnknown arg
	     | Array{length, elt, ...} => (makeUnknown length;
					   makeUnknown elt)
	     | Vector{length, elt} => (makeUnknown length;
				       makeUnknown elt)
	     | Tuple vs => List.foreach(vs, makeUnknown)

	 fun sideEffect(v: t): unit =
	    case value v of
	       Const _ => ()
	     | Datatype _ => ()
	     | Ref{arg, ...} => makeUnknown arg
	     | Array{elt, ...} => makeUnknown elt
	     | Vector{elt, ...} => makeUnknown elt
	     | Tuple vs => List.foreach(vs, sideEffect)
	    
	 fun primApp{prim, targs, args, resultVar, resultType}: t =
	    let
	       fun bear z =
		  if once resultVar andalso Type.isSmall resultType
		     then Birth.here z
		  else Birth.unknown()
	    in case Prim.numArgs prim of
	       NONE =>
		  (if (case Type.dest resultType of
			  Type.Datatype _ => true
			| Type.Array _ => true
			| Type.Vector _ => true
			| Type.Ref _ => true
			| Type.Tuple _ => true
			| Type.Variant _ => true
			| _ => false)
		      then unknown resultType
		   else new(Const(Const.prim{prim = prim,
					     targs = targs,
					     ty = resultType}),
			    resultType))
	     | SOME _ => 
		  let open Prim.Name
		  in case (Prim.name prim, targs, args) of
		     (Array, [_], [n]) =>
			let val a = fromType resultType
			in coerce{from = n, to = arrayLength a};
			   Birth.coerce{from = bear(), to = arrayBirth a};
			   a
			end
		   | (ArraySub, _, [a, _]) => dearray a
		   | (ArrayUpdate, _, [a, _, v]) =>
			(coerce{from = v, to = dearray a};
			 unit())
		   | (ArrayLength, _, [a]) => arrayLength a
		   | (VectorLength, _, [v]) => vectorLength v
		   | (VectorSub, _, [v, _]) => devector v
		   | (VectorFromArray, _, [a]) =>
			let val v = fromType resultType
			in coerce{from = dearray a, to = devector v};
			   v
			end
		   | (Ref, _, [v]) =>
			let val r = fromType resultType
			in coerce{from = v, to = deref r};
			   Birth.coerce{from = bear{init = v}, to = refBirth r};
			   r
			end
		   | (Deref, _, [r]) => deref r
		   | (Assign, _, [r, v]) =>
			(coerce{from = v, to = deref r}; unit())
		   | _ => (List.foreach(args, sideEffect);
			   unknown resultType)
		  end
	    end
      end

      val {value, ...} =
	 Control.traceCall "fixed point"
	 analyze{
		 program = program,
		 useFromTypeOnBinds = false,
		 fromType = Value.fromType,
		 copy = Value.fromType o Value.ty,
		 coerce = coerce,
		 const = Value.const,
		 tuple = Value.tuple,
		 select = Value.select,
		 constructors = SOME{conSelect = conSelect,
				     conApp = conApp},
		 primApp = primApp,
		 filter = Value.filter
		}

      val _ =
	 Control.displays
	 ("const", fn display =>
	  let open Layout
	  in display(str "Input program:\n")
	     ; Program.layouts(program, display)
	     ; display(str "\n\nConstants:")
	     ; (Program.foreachVar
		(program, fn (x, _) => display(seq[Var.layout x,
						   str " ",
						   Value.layout(value x)])))
	  end)

      (* Walk through the program
       *  - removing declarations whose rhs is constant
       *  - replacing variables whose value is constant with globals
       *  - building up the global decs
       *)

      val {new = newGlobal, all} = Global.make()

      fun replaceVar x =
	 case Value.global(value x, newGlobal) of
	    NONE => x
	  | SOME(g, _) => g

      fun replaceVars xs = List.map(xs, replaceVar)
	       	    
      fun loopBind{var, ty, exp} =
	 let
	    fun keep() =
	       SOME{var = var, ty = ty, exp = PrimExp.replaceVar(exp, replaceVar)}
	 in case (Value.global(value var, newGlobal), exp) of
	    (NONE, _) => keep()
	  | (_, PrimApp{prim, ...}) =>
	       if Prim.maySideEffect prim
		  then keep()
	       else NONE
	  | _ => NONE
	 end

      fun loopExp(e: Exp.t) : Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val transfer =
	       case transfer of
		  Call{func, args, cont} =>
		     Call{func = func, cont = cont, args = replaceVars args}
		| Jump{dst, args} => Jump{dst = dst, args = replaceVars args}
		| Return xs => Return(replaceVars xs)
		| Case{cause, test, cases, default} =>
		     Case{cause = cause, test = replaceVar test,
			  cases = cases, default = default}
		| Raise xs => Raise(replaceVars xs)
		| Halt => Halt
		| Bug => Bug
	    val decs = 
	       List.foldr
	       (decs, [], fn (d, ds) =>
		case d of
		   Bind b => (case loopBind b of
				 SOME b => Bind b :: ds
			       | NONE => ds)
		 | Fun{name, args, body} =>
		      Fun{name = name,
			  args = args,
			  body = loopExp body} :: ds
		 | d => d :: ds)
	 in Exp.make{transfer = transfer,
		     decs = decs}
	 end

      val shrinkExp = shrinkExp [] (* not sure what globals to put here *)
	 
      val functions =
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name, args = args, returns = returns,
		   body = shrinkExp(loopExp body)})

      val globals = List.reverse(List.fold(globals, [], fn (b, bs) =>
					   case loopBind b of
					      NONE => bs
					    | SOME b => b :: bs))
	 
      val globals = all() @ globals
	      	 
      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CONTIFY_STRUCTS = 
   sig
      include SHRINK
   end

signature CONTIFY = 
   sig
      include CONTIFY_STRUCTS
      
      val contify: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * If a toplevel function is called
 *       1. from one place outside itself
 *   and 2. all calls to itself within itself are tail
 * Then that function can be made local at the point of its outer call.
 *)

functor Contify(S: CONTIFY_STRUCTS): CONTIFY = 
struct

open S
open Dec Transfer

structure Info =
   struct
      datatype t =
	 NoOuterCalls
       | OneOuterCall
       | NotCont

      val toString =
	 fn NoOuterCalls => "NoOuterCalls"
	  | OneOuterCall => "OneOuterCall"
	  | NotCont => "NotCont"

      val layout = Layout.str o toString
   end

fun contify(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = getCont, set = setCont} =
	 Property.new(Func.plist, Property.initConst Info.NoOuterCalls)

      fun outerCall f =
	 case getCont f of
	    Info.NoOuterCalls => setCont(f, Info.OneOuterCall)
	  | Info.OneOuterCall => setCont(f, Info.NotCont)
	  | Info.NotCont => ()

      fun notCont f = setCont(f, Info.NotCont)

      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  Exp.foreachCall
	  (body, fn {func, cont, ...} =>
	   if Func.equals(name, func)
	      then (case cont of
		       NONE => ()
		     | SOME _ => notCont name)
	   else outerCall func))
	 
      (* diagnostic *)
(*       val _ = List.foreach(functions, fn {name, ...} =>
 * 			   Control.message(let open Layout
 * 					   in seq[Func.layout name,
 * 						  str " => ",
 * 						  Info.layout(getCont name)]
 * 					   end))
 *)
      (* diagnostic *)
	 
      val {get, set} = Property.new(Func.plist, Property.initConst NONE)

      (* for functions turned into locals, record their
       * args, body, and new name
       *)
      val _ =
	 List.foreach
	 (functions, fn {name, args, body, ...} =>
	  case getCont name of
	     Info.OneOuterCall =>
		set(name,
		    SOME{args = args,
			 body = body,
			 jump = Jump.newString(Func.originalName name)})
	   | _ => ())

      (* Walk over all functions, removing those that aren't toplevel, and
       * descending those that are, inserting local functions
       * where necessary.
       *  - turn tail calls into nontail calls
       *  - turn returns into jumps
       *)
	 
      fun walkExp(f: Func.t, e: Exp.t, c: Jump.t option): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	    val decs = 
	       List.map
	       (decs, fn d =>
		case d of
		   Bind _ => d
		 | Fun{name, args, body} =>
		      Fun{name = name,
			  args = args,
			  body = walkExp(f, body, c)}
		 | HandlerPush h => HandlerPush h
		 | HandlerPop => HandlerPop)
	    fun make transfer = Exp.make{decs = decs,
					 transfer = transfer}
	 in
	    case transfer of
	       Call{func, args, cont} =>
		  let val newCont = (case cont of
					NONE => c
				      | SOME _ => cont)
		  in
		     case get func of
			NONE => make(Call{func = func,
					  args = args,
					  cont = newCont})
		      | SOME{jump, args = formals, body} =>
			 if Func.equals(f, func)
			    then (* cont must be tail -- turn into jump *)
			       (case cont of
				   SOME _ => Error.bug "nontail jump"
				 | NONE =>
				      make(Jump{dst = jump, args = args}))
			 else
			    Exp.make
			    {decs =
			     decs @ [Fun{name = jump,
					 args = formals,
					 body = walkExp(func, body, newCont)}],
			     transfer = Jump{dst = jump, args = args}}
		  end
	     | Return xs =>
		  make(case c of
			  NONE => transfer
			| SOME c => Jump{dst = c, args = xs})
	     | _ => make transfer
	 end

      val shrinkExp = shrinkExp globals

      val functions =
	 List.fold(functions, [], fn ({name, args, body, returns}, functions) =>
		   case get name of
		      NONE =>
			 {name = name, args = args, returns = returns,
			  body = shrinkExp(walkExp(name, body, NONE))}
			 :: functions
		    | SOME _ => functions)

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CPS_CALL_GRAPH_STRUCTS =
   sig
      structure CpsTree : CPS_TREE
      sharing CpsTree.Func = CpsTree.Jump
   end
      
signature CPS_CALL_GRAPH =
   sig
      include CPS_CALL_GRAPH_STRUCTS
      structure Graph : DIRECTED_GRAPH
     
      datatype t = T of {getNode: CpsTree.Func.t -> Graph.Node.t,
			 setNode: CpsTree.Func.t * Graph.Node.t -> unit,
			 isLeaf: Graph.Node.t -> bool,
			 getExpFromNode : Graph.Node.t -> { name : CpsTree.Func.t,
							   args : (CpsTree.Var.t * CpsTree.Type.t) list,
							   body : CpsTree.Exp.t },
			 getCallSites: Graph.Node.t -> CpsTree.Transfer.t list,
			 getSize: Graph.Node.t -> int,
			 setSize: Graph.Node.t * int -> unit}
	 
      val buildGraph : CpsTree.Program.t -> t
   end

signature CPS_SIZE_STRUCTS =
   sig
      structure CpsTree : CPS_TREE
      structure CallGraph : CPS_CALL_GRAPH
      sharing CpsTree = CallGraph.CpsTree
      sharing CpsTree.Func = CpsTree.Jump	
   end
	 
signature CPS_SIZE =
   sig
      structure CpsTree : CPS_TREE
      structure CallGraph : CPS_CALL_GRAPH
      val computeSize : CpsTree.Program.t * CallGraph.t -> unit
   end	       

signature CPS_INLINE_STRUCTS =
    sig
       include SHRINK
       sharing Func = Jump
      end

signature CPS_INLINE = 
    sig
        include CPS_TREE
  	val computeInline : Program.t -> Program.t
    end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor CallGraph(S: CPS_CALL_GRAPH_STRUCTS) : CPS_CALL_GRAPH =
   struct
      open S
      structure CpsTree = CpsTree
      structure Graph = DirectedGraph
      structure Node = Graph.Node

      open CpsTree
	 
      datatype t = T of {getNode: CpsTree.Func.t -> Graph.Node.t,
			 setNode: CpsTree.Func.t * Graph.Node.t -> unit,
			 isLeaf: Graph.Node.t -> bool,
			 getExpFromNode : Graph.Node.t -> { name : CpsTree.Func.t,
							   args : (CpsTree.Var.t * CpsTree.Type.t) list,
							   body : CpsTree.Exp.t },
			 getCallSites: Graph.Node.t -> CpsTree.Transfer.t list,
			 getSize: Graph.Node.t -> int,
			 setSize: Graph.Node.t * int -> unit}

      type function = { name : Func.t,
		        args : (Var.t * Type.t) list,
		        body : Exp.t } 

	 
      fun buildGraph(Program.T{datatypes,globals,functions,main}) =
	 let val graph = Graph.new()

	     val {get = getExpFromNode: Node.t -> function,
		  set = setNodeToExp} =
		Property.new(Graph.Node.plist, 
			     Property.initRaise("No Exp for this node", Node.layout))

	     fun createNode(id) = Graph.newNode(graph)

	     val {get = getNode: Func.t -> Node.t,
		  set = setNode} =
		Property.new(Func.plist, Property.initFun(createNode))

	     val {get = getCallSites: Node.t -> Transfer.t list,
		  set = setCallSites} =
		Property.new(Graph.Node.plist, Property.initConst [])

	     val getCallSites = Trace.trace("getCallSites", Node.layout, Layout.ignore) 
		getCallSites
	 
	     val {get = getSize: Node.t -> int,
		  set = setSize} =
		Property.new(Graph.Node.plist, Property.initConst 0)

	     val {get = isLeaf: Node.t -> bool,
		  set = setLeaf} =
		Property.new(Graph.Node.plist, Property.initConst true)

	     val isLeaf = Trace.trace("isLeaf",Node.layout,Bool.layout) isLeaf
	 
	     fun appendCallSites(node,callSite) =
		setCallSites(node, callSite::getCallSites(node))

	     fun traceFun(name,body,graph) =
		let val node = getNode(name)
		in case (Exp.dest body) of
		   {decs,transfer} =>
		      (List.foreach(decs, 
				    fn d => case d of
				    Dec.Fun cont =>
				       let val {name,args,body} = cont
					  val node = getNode(name)
					  val _ = setNodeToExp(node,cont)
				       in  traceFun(name,body,graph)
				       end
				  | _ => ());
		       case transfer of
			  Transfer.Call {func,args,cont} =>
			     let val funcNode = getNode(func)
			     in (Graph.addEdge(graph, {from=node, to=funcNode});
				 setCallSites(funcNode, transfer::getCallSites(funcNode));
				 case cont of
				    NONE => ()
				  | SOME l => let val contNode = getNode(l)
					      in Graph.addEdge(graph, {from=funcNode, to=contNode})
					      end)
			     end
			| Transfer.Jump {dst, args} =>
			     let val node = getNode(dst)
			     in  setCallSites(node, transfer::getCallSites(node))
			     end
			| Transfer.Case {test,cases,default, ...} =>
			     let val branches = List.map(cases, fn (c,l) =>
							 let val caseNode = getNode(l)
							 in Graph.addEdge(graph, 
									  {from=node, to=caseNode})
							 end)
			     in case default of
				NONE => ()
			      | SOME l => let val defaultNode = getNode(l)
					  in Graph.addEdge(graph, {from=node, to=defaultNode})
					  end
			     end
			| _ => ())
		end
	    
	     val _ = (List.foreach(functions, fn {name,args,body,returns} => 
				   let val newNode = getNode(name)
				   in (setNodeToExp(newNode, {name=name,args=args,body=body});
				       let exception E
				       in Exp.foreachTransfer(body, fn Transfer.Call c => (setLeaf(newNode,false);
											   raise E)
				                                    | _ =>  ())
					  handle E => ()
				       end)
				   end);
		      Graph.removeBackEdges(graph);
		      List.foreach(functions, fn {name,args,body,returns} => 
				   traceFun(name,body,graph)))
	 in T {getNode=getNode,
	       setNode=setNode,
	       isLeaf=isLeaf,
	       getExpFromNode=getExpFromNode,
	       getCallSites=getCallSites,
	       getSize=getSize,
	       setSize=setSize}
	    before Control.message(fn () =>
				   let open Layout
				   in seq[str"Graph size:  ",
					  Int.layout(List.length(Graph.nodes graph))]
				   end)
	 end

      val buildGraph = Control.traceCall "buildGraph" buildGraph
	
   end
 
functor CpsSize(S : CPS_SIZE_STRUCTS) : CPS_SIZE = 
   struct

      open S
      structure CpsTree = CpsTree
      structure CallGraph = CallGraph

      structure Program = CpsTree.Program
      structure Exp = CpsTree.Exp
      structure Type = CpsTree.Type
      structure Transfer = CpsTree.Transfer
      structure PrimExp = CpsTree.PrimExp
      structure Dec = CpsTree.Dec
      structure Exp = CpsTree.Exp
      structure Var = CpsTree.Var
      structure Func = CpsTree.Func

      open CallGraph
	 
      datatype known = K | U | Bot

      val maxSize = 100000
	 
      val {contains,add,...} = List.set {equal = Func.equals, layout = Func.layout}

      fun knownJoin (K,K) = K
	| knownJoin (K,U) = U
	| knownJoin (Bot,x) = x:known
	| knownJoin (x,Bot) = x:known
	| knownJoin (U,_) = U
				       
      fun computeSize(Program.T{datatypes,globals,functions,main},graph) =
	 let  val {get = getKnown: Var.t -> known,
		   set = setKnown,
		   destroy = destroyKnown} =  Property.newDest(Var.plist, Property.initConst Bot)

	    val CallGraph.T {getNode,setNode,isLeaf,getExpFromNode,getCallSites,getSize,setSize} = graph
	     fun computeArgProp(name,args) =  let val node = getNode(name)
      	                                          val callSites = getCallSites(node)
						  val length = List.length(args)
						  val argProps = 
						     List.fold(callSites,
							       List.duplicate(length,fn() => Bot),
							       fn (c, i) =>
							       case c of
								  Transfer.Call{func,args,cont} => 
								     List.map2(i,args,
									       fn(k,a) => (knownJoin(k,getKnown(a))))
								| Transfer.Jump {dst,args} => 
								     List.map2(i,args,
									       fn(k,a) => (knownJoin(k,getKnown(a))))
								| _ => i)
					      in List.foreach2(args, argProps, setKnown)
					      end
	  
	     fun computeKnownExp(exp) = 
		let val {decs,transfer} = Exp.dest(exp)
		in  List.foreach(decs,
				 fn d => (case d of
					     Dec.Bind {var,ty,exp} => 
						(case exp of
						    PrimExp.Var pvar => 
						       let val v = getKnown(pvar)
						       in setKnown(var,v)
						       end
						  | _ =>  setKnown(var,K))
					   | Dec.Fun {name,args,body} =>
						let val args = List.map(args, fn (var,varType) => var)
						in computeArgProp(name,args) 
						end
					   | _ => ()))
		end

	 
	     val computeKnownExp = Trace.trace("computeKnownExp", Exp.layout, Layout.ignore) 
		computeKnownExp
		
	     val {get = getVisited: Func.t -> bool,
		  set = setVisited,
		  destroy = destroyVisited} =
		Property.newDest(Func.plist, Property.initConst false)
		
	     fun computePrimExpSize(primExp) =
		(case primExp of
		    PrimExp.Const c => 0
		  | PrimExp.Var x => 1
		  | PrimExp.Select _ => 1
		  | PrimExp.Tuple vars => List.length(vars)
		  | PrimExp.ConSelect _ => 1
		  | PrimExp.ConApp {con, args} => List.length(args)
		  | PrimExp.PrimApp {prim, targs, args} => List.length(args))
	     and
		computeDecSize(dec) = 
		case dec of
		   Dec.Bind {var,ty,exp} => let val _ = (case exp of
							    PrimExp.Const c => setKnown(var,K)
							  | PrimExp.Var v => setKnown(var,getKnown(v))
							  | _ => ())
					    in case getKnown(var) of
					       K => 0
					     | _ => computePrimExpSize(exp)
					    end
		 |   Dec.Fun {name,args,body} => let val node = getNode(name)
						     val nodeSize = getSize(node)
						 in if (nodeSize > 0) orelse getVisited(name)
						       then nodeSize
						    else let val _ = setVisited(name,true)
							     val funSize = computeExpSize(body)
							 in funSize before 
							    setSize(node,funSize)
							 end
						 end
		 |   _ => 1 
	     and
		computeTransferSize(transfer) =
		case transfer of
		   Transfer.Call {func, args, cont} => 
		      let val node = getNode(func)
			 val funcSize = getSize(node)
		      in if funcSize = 0
			    then 
			       let val {name,args,body} =  getExpFromNode(node)
				  val size =  computeExpSize(body)
				  val _ = setSize(node,size)
			       in List.length(args)
			       end
			 else funcSize
		      end
		 |   Transfer.Jump {dst, args} => 1 + List.length(args)
		 |   Transfer.Case {test, cases, default, ...} => let val known = getKnown(test)
								 val defaultSize = 
								    case default of
								       NONE => 0
								     | SOME l => 1
							     in case known of
								K => defaultSize
							      | U => 2 + defaultSize +  
								   List.length(cases)
							      | _ => defaultSize
							     end
		 |   _  => 1
	     and
		computeExpSize(exp) =
		let val {decs,transfer} = Exp.dest(exp)
		   val _ = computeKnownExp(exp) 
		   val decSize = List.fold(decs,0,fn(d,i)=> if i > maxSize
							       then maxSize
							    else i + computeDecSize(d))
		   val transferSize = computeTransferSize(transfer)
		in decSize + transferSize
		end

	     val computeExpSize = Trace.trace("computeExpSize", Exp.layout, Layout.ignore) 
		computeExpSize
		

	 in (List.foreach(functions, fn {name,args,body,returns} =>
			      let val node = getNode(name)
			      in (setSize(node, List.length(args) + 
					  computeExpSize(body)))
			      end);
		 destroyVisited(); destroyKnown())
	 end

(*	       val computeSize = fn arg => ((computeSize arg)
	                           handle E => (print(General.exnName E); print("\n")))
*)

   end


functor CpsInline(S: CPS_INLINE_STRUCTS) : CPS_INLINE =
   struct
      structure CpsTree = S
      structure CallGraph = CallGraph(structure CpsTree = CpsTree)
      structure CpsSize = CpsSize(structure CpsTree = CpsTree
				  structure CallGraph = CallGraph)

      open CallGraph
      open CpsTree

      local open Control
      in val rounds = ref 1
	 val size = ref 2
      end

      val inlineSize = ref 0
      val maxInlineSize = 500
	 
      datatype CallType = REG | NON_TAIL of Jump.t
      datatype InlineState = INLINE of (Graph.Node.t * Exp.t) | NO_INLINE of Transfer.t

      val shrinkExp = S.shrinkExp

      val foo: Program.t option ref = ref NONE
	 
      fun computeInline(programz) =
	 let val program = (case !foo of
			       NONE => (print("there...");
					foo := SOME programz;
					programz)
			     | SOME program => (print("here...."); program))

	    fun loop(program,rounds) =
	     let val graph as T {getNode,setNode,isLeaf,getExpFromNode,getCallSites,getSize,setSize} =
		           buildGraph(program)
		 val _ = CpsSize.computeSize(program,graph)
		 val  CpsTree.Program.T{datatypes,globals,functions,main} = program
		 val shrinkExp = shrinkExp globals
		 fun small(node) = (getSize(node) < (!inlineSize))
		 fun alwaysInline(node) = let val callSites = getCallSites(node)
					     in if small(node) orelse List.length(callSites) = 1 
						orelse (isLeaf(node) andalso (getSize(node) < maxInlineSize))
						   then true
						else false
					     end

		 val {get = getVisited: Func.t -> int,
		      set = setVisited,
		      destroy = destroyVisited} =
		    Property.newDest(Func.plist, Property.initConst 0)

		 fun fixDecs(decs,callType) =
		    let open Dec
		       val newDecs = []
		    in List.fold(decs,
				 newDecs,
				 fn (d,newDecs) =>
				 List.append(newDecs,
					     case d of
						b as Bind {var,ty,exp} => [b]
					      | Fun {name,args,body} =>
						   [Fun {name = name,
							 args = args, 
							 body = fixReturns(body,callType)}]
					      | dec => [dec]))
		    end
		 
		 and fixTransfer(transfer,callType) = 
		    let open Transfer
		    in case transfer of
		       Call {func,args,cont} => Call {func = func,
						      args = args,
						      cont = case cont of
						      NONE => (case callType of
								  REG => NONE
								| NON_TAIL c => SOME c)
						    | SOME l => SOME l}
		     | r as Return vars => (case callType of
					       REG  => r
					     | NON_TAIL c => Jump{dst = c, args = vars})

		     | trans => trans
		    end
		 and fixReturns(exp,callType) = 
		    let open Exp
		       val {decs,transfer} = dest(exp)
		       val newDecs = fixDecs(decs,callType) 
		       val newTransfer = fixTransfer(transfer,callType)
		    in make({decs = newDecs, transfer = newTransfer})
		    end

		 fun inlineBody(node,e,actuals,callType) =
		    let val {name,args,body} = getExpFromNode(node)
		       val bindings = List.zip(args,actuals)
		       val substitution = List.map(bindings, fn ((formal,ty),actual) =>
						   {formal = formal, actual = actual})
		       val renamed = Exp.alphaRename {exp = body,
						      substitution = substitution,
						      handleJump = fn {old,new} => setNode(new,getNode(old))}
		       val {decs,transfer} = Exp.dest(renamed)
		       val new =  case transfer of
			  Transfer.Call {func,args,cont} => let val node = getNode(func)
								val body = #body(getExpFromNode(node))
							    in if alwaysInline(node)
								  then let val callType = (case cont of
											      NONE => REG
											    | SOME c => NON_TAIL c)  
									   val {decs=decs',transfer=transfer'} = 
									      inlineBody(node,body,args,callType)
								       in Exp.make({decs = List.append(decs,decs'),
										    transfer = transfer'})
								       end
							       else renamed
							    end
			| _ => renamed 	
		       val newBody = fixReturns(new,callType)
		       val newExp = Exp.dest(newBody)
		    in {decs = #decs newExp, transfer = #transfer newExp}
		    end

		 val inlineBody = Trace.trace4("inlineBody", Layout.ignore, Exp.layout, Layout.ignore, Layout.ignore, Layout.ignore)
		    inlineBody
		    
		 fun remove(Program.T{datatypes, globals, functions, main}) =
		    let
		       val {get: Func.t -> {body: Exp.t, visited: bool ref},
			    set, destroy} =
			  Property.newDest(Func.plist,Property.initRaise("info", Func.layout))
			  
		       val _ = List.foreach(functions, fn {name, body, ...} =>
					    set(name, {body = body, visited = ref false}))
		       fun visit(f: Func.t) =
			  let val {visited, body} = get f
			  in if !visited
				then ()
			     else (visited := true;
				   Exp.foreachTransfer(body,
						       fn Transfer.Call{func, ...} => visit func
							| _ => ()))
			  end
		       val _ = visit main
		       val functions = List.keepAll(functions, fn f => !(#visited(get(#name f))))
		    in destroy();
		       Program.T{datatypes = datatypes,
				 globals = globals,
				 functions = functions,
				 main = main}
		    end

		 fun inline(functions) =
		    let val {get: Func.t -> {body: Exp.t, visited: bool ref}, set, destroy} =
		       Property.newDest(Func.plist,
					Property.initRaise("info", Func.layout))
			val {get = init: Func.t -> bool, set = setInit, destroy = destroyInit} =
			   Property.newDest(Func.plist,Property.initConst false)
			   
			fun setDec(d) =
			   case d of
			      Dec.Fun {name,args,body} =>
				 (set(name,{body = body, visited = ref false});
				  if init(name)
				     then ()
				  else (setInit(name,true); setExp(body)))
			    | _ =>()
			and setExp(e) = Exp.foreach'(e,
						     {handleDec = setDec,
						      handleTransfer = fn _ => ()})

			fun buildInline(newDecs,newExp,transfer, operator) =
			   let fun build(func,args,cont) = 
			      let val {body,visited} = get(func)
				 val node = getNode(func)
			      in if operator(node) 
				    then let val newExp = inlineBody(node,body,args,
								     case cont of
									NONE => REG
								      | SOME c => NON_TAIL c)
					     val {decs=decs',transfer=transfer'} = newExp
					 in Exp.make( {decs = List.append(newDecs,decs'),
						       transfer = transfer'})				 
					 end
				 else newExp
			      end
			   in case transfer of
			      Transfer.Call {func,args,cont} => build(func,args,cont)
			    | Transfer.Jump {dst,args} => build(dst,args,NONE)  
			    | _ => newExp 
			   end

			fun visit(f : Func.t) =
			   let val {visited, body} = get f
			      fun inline(d) = case d of
				 Dec.Fun {name,args,body} =>
				    Dec.Fun {name=name,args=args,body=visit(name)}
			       | d => d

			   in  if !visited
				  then body
			       else (visited := true;
				     let val {decs,transfer} = Exp.dest(body)
					val newDecs = List.map(decs,inline)
					val newExp = Exp.make {decs = newDecs, transfer = transfer}
					val _ = set(f, {body = newExp, visited=visited})
				     in buildInline(newDecs,newExp,transfer,
						    case transfer of
						       Transfer.Call {func, args, cont} => alwaysInline
						     | Transfer.Jump {dst,args} => small   
						     | _ => (fn _ => false))
				     end)
			   end
		    in (List.foreach(functions, fn {name, args, body, returns}  =>
				     (set(name, {body = body, visited = ref false});
				      setExp(body)));
			List.map(functions, fn {name, args, body, returns}  =>
				 {name = name, args = args, body = visit(name), returns = returns}))
		    end
		    
		 val newFuns =  inline(functions)
		 val shrinkedExp = List.map(newFuns,
					    fn {name,args,body,returns} =>
					    {name = name,
					     args = args,
					     body =   shrinkExp(body), 
					     returns = returns})
		 val newProg = remove(CpsTree.Program.T {datatypes = datatypes,
							 globals = globals,
							 functions = shrinkedExp,
							 main = main })
		 val _ = (Program.clear(program); destroyVisited())
		 val _ = Control.message(fn () => Program.layoutStats(newProg))
	     in if rounds <= 1
		   then newProg
		else loop(newProg,rounds-1)
	     end
	  val newProg = loop(program,!rounds)
	in (Program.clear(program);
	    inlineSize := !size;
	    loop(newProg,0) before (Program.clear(newProg)))
	end
   end


 
(* (Layout.output(let open Layout
			    in align[seq[str "exp: ", Program.layout program,
					 str "newExp: ", Program.layout newProg]]
			    end, Out.standard); *)


(* Layout.output(let open Layout
			    in align[seq[str "exp: ", Program.layout program,
					 str "newExp: ", Program.layout newProg]]
			    end, Out.standard);

*)






(* 

             let val functions = List.map(functions, 
					  fn f as {name,args,body,returns} =>
					  let val {decs,transfer} = Exp.dest(body)
					  in case transfer of
                                                Transfer.Call c => {name=name,args=args,body=buildInline(decs,body,transfer,isLeaf),
								    returns=returns}
                                               | _ => f
					  end)

               Layout.output(let open Layout
									 in align[seq[str "func Name   ",
										      FuncName.layout func,
										      let val node = getNode(func)
										      in Bool.layout(isLeaf(node))
										      end,
										      str "\n"]]
									 end, Out.standard)
		      | _ => ()
		      end)
(List.foreach(functions, fn {name, ...} => let val n = getNode(name)
						       in Control.message(let open Layout
									  in seq[FuncName.layout name,
										 str " ",
										 Bool.layout(isLeaf(n))]
									  end)
						       end);
*)
(*

                                               (Layout.output(let open Layout
							      in align[seq[str "callee: ", Func.layout func],
									   str "cont", (case cont of
									                 SOME c => Jump.layout c
									                | NONE => Unit.layout () ),
									   str "\n"]
							      end, Out.standard);

						*)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature FLATTEN_STRUCTS = 
   sig
      include SHRINK
   end

signature FLATTEN = 
   sig
      include FLATTEN_STRUCTS
      
      val flatten: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Flatten arguments to jumps, constructors, and functions.
 * If a tuple is explicitly available at all uses of a jump (resp. function)
 * then
 *   - The formals and call sites are changed so that the components of the
 *     tuple are passed.
 *   - The tuple is reconstructed at the beginning of the body of the jump.
 *
 * Similarly, if a tuple is explicitly available at all uses of a constructor,
 *   - The constructor argument type is changed to flatten the tuple type.
 *   - The tuple is passed flat at each ConApp.
 *   - The tuple is reconstructed at each ConSelect.
 *
 * The trickiest bit is for constructors, because ConSelect offsets must be
 * changed and the tuple must be reconstructed.  For example,
 *
 * datatype t = T of [(u * v), (w * z)]
 * val x = ConSelect{con = T, offset = 2, variant = v}
 *
 * Suppose that both args to T are flattened.  Then this becomes:
 *
 * datatype t = T of [u, v, w, z]
 * val a = ConSelect{con = T, offset = 3, variant = v}
 * val b = ConSelect{con = T, offset = 4, variant = v}
 * val x = (a, b)
 *
 *)

functor Flatten(S: FLATTEN_STRUCTS): FLATTEN = 
struct

open S
open Dec PrimExp Transfer

structure Rep =
   struct
      structure Set = DisjointSet
	 
      datatype t =
	 Leaf
       | Tuple of bool Set.t (* true means keep it as a tuple *)

      local open Layout
      in
	 val layout =
	    fn Leaf => str "leaf"
	     | Tuple s => Bool.layout(Set.value s)
      end

      val isFlat =
	 fn Tuple s => not(Set.value s)
	  | _ => false
	       
      fun fromType t =
	 case Type.detupleOpt t of
	    NONE => Leaf
	  | SOME _ => Tuple(Set.singleton false)

      fun fromTypes ts = List.map(ts, fromType)

      fun fromFormals(xts: (Var.t * Type.t) list) = List.map(xts, fromType o #2)

      val tuplize: t -> unit =
	 fn Leaf => ()
	  | Tuple s => Set.setValue(s, true)

      fun tuplizes rs = List.foreach(rs, tuplize)

      val unify =
	 fn (Leaf, Leaf) => ()
	  | (Tuple s, Tuple s') =>
	       let val isTuple = Set.value s orelse Set.value s'
	       in Set.union(s, s');
		  Set.setValue(s, isTuple)
	       end
	  | _ => Error.bug "unify"

      fun unifys(rs, rs') = List.foreach2(rs, rs', unify)

      val layouts = List.layout layout
	 
(*      val unifys = Trace.trace2("unifys", layouts, layouts, Unit.layout) unifys *)
   end

fun flatten(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conInfo, set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("args", Con.layout))

      val conArgs = #args o conInfo
      val conArgTypes = #argTypes o conInfo

      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach(cons, fn {con, args} =>
		       setConInfo(con, {argTypes = args,
					args = Rep.fromTypes args})))

      val {get = funcInfo: Func.t -> {args: Rep.t list,
					  returns: Rep.t list},
	   set = setFuncInfo} =
	 Property.new(Func.plist,
		      Property.initRaise("info", Func.layout))

      val funcArgs = #args o funcInfo

(*       val funcInfo =
 * 	 Trace.trace("funcInfo", Func.layout,
 * 		     fn {args, returns} =>
 * 		     Layout.record[("args", Rep.layouts args),
 * 				   ("returns", Rep.layouts returns)])
 * 	 funcInfo
 *)

      val _ =
	 List.foreach
	 (functions, fn {name, args, returns, ...} =>
	  setFuncInfo(name, {args = Rep.fromFormals args,
			     returns = Rep.fromTypes returns}))

      val {get = varTuple, set = setVarTuple} =
	 Property.new(Var.plist, Property.initConst NONE)

      val {get = jumpArgs: Jump.t -> Rep.t list, set = setJumpArgs} =
	 Property.new(Jump.plist, Property.initRaise("args", Jump.layout))

      fun coerces(xs: Var.t list, rs: Rep.t list) =
	 List.foreach2
	 (xs, rs, fn (x, r) => (case varTuple x of
				   NONE => Rep.tuplize r
				 | _ => ()))

(*       val coerces =
 * 	 Trace.trace2("coerces",
 * 		      List.layout Var.layout,
 * 		      List.layout Rep.layout,
 * 		      Unit.layout)
 * 	 coerces
 *)

      fun loopBind{var, ty, exp} =
	 case exp of
	    Tuple xs => setVarTuple(var, SOME xs)
	  | ConApp{con, args} => coerces(args, conArgs con)
	  | _ => ()

      val _ = List.foreach(globals, loopBind)
	       
      fun loopExp(e, returns) =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach
	    (decs,
	     fn Bind b => loopBind b
	      | Fun{name, args, body} =>
		   (setJumpArgs(name, Rep.fromFormals args)
		    ; loopExp(body, returns))
	      | HandlerPush j => Rep.tuplizes(jumpArgs j)
	      | _ => ());
	    case transfer of
	       Return xs => coerces(xs, returns)
	     | Call{func, args, cont} =>
		  let val {args = formals, returns = returns'} = funcInfo func
		  in coerces(args, funcArgs func);
		     Rep.unifys(returns',
				case cont of
				   NONE => returns
				 | SOME c => jumpArgs c)
		  end
	      | Jump{dst, args} => coerces(args, jumpArgs dst)
	      | _ => ()
	 end

      val _ =
	 Control.traceCall "analyze"
	 List.foreach(functions, fn {name, body, ...} =>
		      loopExp(body, #returns(funcInfo name)))

      val _ =
	 Control.displays
	 ("flat", fn display =>
	  List.foreach(functions, fn {name, ...} =>
		       let val {args, returns} = funcInfo name
			  open Layout
		       in display(seq[Func.layout name,
				      str " ",
				      align[List.layout Rep.layout args,
					    List.layout Rep.layout returns]])
		       end))

      fun flattenTypes(ts: Type.t list, rs: Rep.t list): Type.t list =
	 List.fold2(ts, rs, [], fn (t, r, ts) =>
		    if Rep.isFlat r
		       then List.fold(Type.detuple t, ts, op ::)
		    else t :: ts)

(*       val flattenTypes =
 * 	 Trace.trace2("flattenTypes",
 * 		      List.layout Type.layout,
 * 		      List.layout Rep.layout,
 * 		      List.layout Type.layout)
 * 	 flattenTypes
 *)

      val datatypes =
	 List.map
	 (datatypes, fn {tycon, cons} =>
	  {tycon = tycon,
	   cons =
	   List.map(cons, fn {con, args} =>
		    {con = con,
		     args = flattenTypes(args, conArgs con)})})
	 
      fun flattens(xs: Var.t list, rs: Rep.t list) =
	 List.fold2(xs, rs, [],
		    fn (x, r, xs) =>
		    if Rep.isFlat r
		       then (case varTuple x of
				SOME ys => List.fold(ys, xs, op ::)
			      | _ => Error.bug "tuple unavailable")
		     else x :: xs)

      fun loopBind{var: Var.t, ty: Type.t, exp} =
	 let fun single e = [{var = var, ty = ty, exp = e}]
	 in case exp of
	    ConApp{con, args} =>
	       single(ConApp{con = con, args = flattens(args, conArgs con)})
	  | ConSelect{variant, con, offset} =>
	       let val {args, argTypes} = conInfo con
		  val args = List.nthTail(args, offset)
		  val argTypes = List.nthTail(argTypes, offset)
	       in case (args, argTypes) of
		  (r :: rs, t :: ts) =>
		     let
			val offset =
			   List.fold2(rs, ts, 0, fn (r, t, offset) =>
				      offset
				      + (if Rep.isFlat r
					    then List.length(Type.detuple t)
					 else 1))
		     in if Rep.isFlat r
			   then
			      let val ts = Type.detuple t
				 val xs = List.map(ts, fn _ => Var.newNoname())
			      in #2(List.fold2
				    (List.reverse xs,
				     List.reverse ts,
				     (offset, single(Tuple xs)),
				     fn (x, t, (offset, bs)) =>
				     (offset + 1,
				      {var = x, ty = t,
				       exp = ConSelect{variant = variant,
						       con = con,
						       offset = offset}}
				      :: bs)))
			      end
			else single(ConSelect{variant = variant,
					      con = con,
					      offset = offset})
		     end
		| _ => Error.bug "bad ConSelect"
	       end
	  | _ => single exp
	 end

      val globals = List.foldr(globals, [], fn (b, bs) => loopBind b @ bs)

      fun loopExp(e: Exp.t, returns: Rep.t list): Exp.t =
	 let val {decs, transfer} = Exp.dest e

	    fun loopDecs ds =
	       List.foldr
	       (ds, [], fn (d, ds) =>
		case d of
		  Fun{name, args, body} =>
		     let val (args, body) = loopArgsBody(args, jumpArgs name,
							 body, returns)
		     in Fun{name = name, args = args, body = body}
		     end :: ds
		| Bind b => List.foldr(loopBind b, ds, fn (b, ds) =>
				       Bind b :: ds)
		| _ => d :: ds)

	    val transfer =
	       case transfer of
		  Return xs => Return(flattens(xs, returns))
		| Call{func, args, cont} =>
		     Call{func = func, cont = cont,
			  args = flattens(args, funcArgs func)}
		| Jump{dst, args} =>
		     Jump{dst = dst,
			  args = flattens(args, jumpArgs dst)}
		| _ => transfer
		     
	 in Exp.make{decs = loopDecs decs,
		     transfer = transfer}
	 end

      and loopArgsBody(args: (Var.t * Type.t) list, rs: Rep.t list,
		       body: Exp.t, returns: Rep.t list) =
	 let
	    val (args, prefix) =
	       List.fold2
	       (args, rs, ([], []),
		fn ((x, t), r, (args, prefix)) =>
		if Rep.isFlat r
		   then let val ts = Type.detuple t
			    val xs = List.map(ts, fn _ => Var.newNoname())
			    val args =
			       List.fold2
			       (xs, ts, args, fn (x, t, args) =>
				(x, t) :: args)
			in (args,
			    Bind{var = x, ty = t, exp = Tuple xs} :: prefix)
			end
		else ((x, t) :: args, prefix))
	 in (args, Exp.prefixs(loopExp(body, returns), prefix))
	 end

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let val {args = argReps, returns = returnReps} = funcInfo name
	     val (args, body) = loopArgsBody(args, argReps, body, returnReps)
	  in {name = name,
	      args = args,
	      body = shrinkExp body,
	      returns = flattenTypes(returns, returnReps)}
	  end)

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = functions,
		   main = main}
   in Program.clear program;
      program
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INLINE_STRUCTS = 
   sig
      include SHRINK
   end

signature INLINE = 
   sig
      include INLINE_STRUCTS
      
      val inline: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Inline(S: INLINE_STRUCTS): INLINE = 
struct

open S

open Dec PrimExp Transfer

(* Returns true if the size of the expression (measured somewhat arbitrarily)
 * is greater than SOME integer.  Returns true on NONE.
 *)
fun sizeGreaterThan(exp: Exp.t, size: int option): bool =
   case size of
      NONE => false
    | SOME size =>
	 let exception Big
	    val r = ref 0
	    fun incSize n =
	       let val new = !r + n
	       in if new > size
		     then raise Big
		  else r := new
	       end
	    val primExpSize =
	       fn Const _ => 0
		| Var _ => 0
		| Select _ => 1
		| ConSelect _ => 1
		| Tuple xs => List.length xs
		| ConApp{args, ...} => 1 + List.length args
		| PrimApp{args, ...} => 1 + List.length args
	    val transferSize =
	       fn Call{args, ...} => 1 + List.length args
		| Jump{args, ...} => 1 + List.length args
		| Return xs => 1 + List.length xs
		| Raise xs => 1 + List.length xs
		| Case{cases, ...} => 1 + List.length cases
		| Halt => 1
		| Bug => 1
	    val decSize =
	       fn Fun _ => 0
		| Bind{exp, ...} => primExpSize exp
		| _ => 1
	 in (Exp.foreach'(exp, {handleDec = incSize o decSize,
				handleTransfer = incSize o transferSize})
	     ; false)
	    handle Big => true
	 end

local
   fun 'a make (dontInlineBody: Exp.t * 'a -> bool)
      (Program.T{functions, ...}, a: 'a): Func.t -> bool =
   let
      val {get = shouldInline, set = setShouldInline} =
	 Property.new(Func.plist, Property.initConst false)
   in
      List.foreach(functions, fn {name, body, ...} =>
		   if dontInlineBody(body, a)
		      then ()
		   else setShouldInline(name, true))
      ; shouldInline
   end

   fun containsCall(exp: Exp.t): bool =
      let exception Yes
      in (Exp.foreachTransfer(exp, fn Call _ => raise Yes | _ => ())
	  ; false) handle Yes => true
      end

   fun containsLoop(exp: Exp.t): bool =
      let
	 exception Yes
	 val {get, set, destroy} =
	    Property.newDest(Jump.plist, Property.initConst false)
	 val loopTransfer =
	    fn Jump{dst, ...} => if get dst then raise Yes else ()
	     | _ => ()
	 fun loopExp(e: Exp.t): unit =
	    let val {decs, transfer} = Exp.dest e
	    in List.foreach(decs, loopDec)
	       ; loopTransfer transfer
	    end
	 and loopDec d =
	    case d of
	       Fun{name, body, ...} =>
		  (set(name, true); loopExp body; set(name, false))
	     | _ => ()
      in ((loopExp exp; false) handle Yes => true)
	 before destroy()
      end
in
   val leaf = make(fn (body, {size}) =>
		   sizeGreaterThan(body, size)
		   orelse containsCall body)

   val leafNoLoop = make(fn (body, {size}) =>
			 sizeGreaterThan(body, size)
			 orelse containsCall body
			 orelse containsLoop body)
end

structure Graph = DirectedGraph
structure Node = Graph.Node

fun nonRecursive(Program.T{functions, ...}, {size: int option}) =
   let
      val {get = funcInfo: Func.t -> {shouldInline: bool ref,
				      isBig: bool,
				      node: Node.t,
				      numCalls: int ref},
	   set = setFuncInfo} =
	 Property.new
	 (Func.plist, Property.initRaise("func info", Func.layout))

      val {set = setNodeFunc, get = nodeFunc} =
	 Property.new(Node.plist, Property.initRaise("func", Node.layout))

      val graph = Graph.new()

      (* initialize the info for each func *)
      val _ = 
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  let val n = Graph.newNode graph
	  in setNodeFunc(n, name)
	     ; setFuncInfo(name, {node = n,
				  isBig = sizeGreaterThan(body, size),
				  numCalls = ref 0,
				  shouldInline = ref true})
	  end)

      (* Update call counts. *)
      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  Exp.foreachTransfer
	  (body,
	   fn Call{func, ...} =>
	        let val {numCalls, shouldInline, isBig, ...} = funcInfo func
		   val n = !numCalls
		in numCalls := n + 1;
		   if ((n = 1 andalso isBig)
		       orelse Func.equals(name, func))
		      then shouldInline := false
		   else ()
		end
	    | _ => ()))
	 
      (* Build the call graph of potentially inlinable functions.
       * Don't add edges for functions that are not inlined, since they
       * can't cause infinite unrolling.
       *)
      val _ = 
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  let val {node = caller, shouldInline = si, ...} = funcInfo name
	  in Exp.foreachTransfer
	     (body,
	      fn Call{func, ...} =>
 	           let
		      val {numCalls, shouldInline = si', node = callee, ...} =
			 funcInfo func
		   in if !si andalso !si'
			 then Graph.addEdge(graph, {from = caller, to = callee})
		      else ()
		   end
	       | _ => ())
	  end)
      
      (* Compute strongly-connected components and set any function that is
       * in a nontrivial scc to be not inlined.
       *)
      val _ = 
	 List.foreach
	 (Graph.stronglyConnectedComponents graph,
	  fn [] => ()
	   | [_] => ()
	   | ns => List.foreach(ns, fn n =>
				#shouldInline(funcInfo(nodeFunc n)) := false))
   in ! o #shouldInline o funcInfo
   end

fun inline(p as Program.T{datatypes, globals, functions, main}) =
   let
      val shouldInline: Func.t -> bool =
	 case !Control.inline of
	    Control.NonRecursive r => nonRecursive(p, r)
	  | Control.Leaf r => leaf(p, r)
	  | Control.LeafNoLoop r => leafNoLoop(p, r)
	  | _ => Error.unimplemented "inline"
	    
      val {get = funInfo: Func.t -> {args: Var.t list,
				     body: Exp.t},
	   set = setFunInfo} =
	 Property.new(Func.plist, Property.initRaise("info", Func.layout))

      val _ = List.foreach(functions, fn {name, args, body, ...} =>
			   setFunInfo(name, {args = List.map(args, #1),
					     body = body}))

      fun inlineExp(exp: Exp.t, outerCont: Jump.t option): Exp.t =
	 let
	    fun loopExp e =
	       let val {decs, transfer} = Exp.dest e
		  val decs =
		     List.map
		     (decs,
		      fn Fun{name, args, body} =>
			    Fun{name = name, args = args, body = loopExp body}
		       | d => d)
		  fun simple t = Exp.make{decs = decs, transfer = t}
	       in case transfer of
		  Return xs =>
		     simple(case outerCont of
			       NONE => Return xs
			     | SOME j => Jump{dst = j, args = xs})
		| Call{func, args, cont} =>
		     let
			val cont =
			   case cont of
			      NONE => outerCont
			    | SOME c => SOME c
			val {args = formals, body} = funInfo func
		     in if shouldInline func
			   then
			      let
				 val body =
				    Exp.alphaRename
				    {exp = body,
				     handleJump = fn _ => (),
				     substitution =
				     List.map2(formals, args, fn (f, a) =>
					       {formal = f, actual = a})}
			      in Exp.prefixs(inlineExp(body, cont), decs)
			      end
			else simple(Call{func = func, args = args,
					 cont = cont})
		     end
		| _ => simple transfer
	       end

	 in loopExp exp
	 end

      val shrinkExp = shrinkExp globals
	 
      val functions =
	 List.fold
	 (functions, [], fn ({name, args, body, returns}, functions) =>
	  if not(Func.equals(name, main))
	     andalso shouldInline name
	     then functions
	  else {name = name, args = args, returns = returns,
		body = shrinkExp(inlineExp(body, NONE))}
	     :: functions)

   in Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INTRODUCE_LOOPS_STRUCTS = 
   sig
      include CPS_TREE
   end

signature INTRODUCE_LOOPS = 
   sig
      include INTRODUCE_LOOPS_STRUCTS
      
      val introduceLoops: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Change any toplevel function that only calls itself in tail position
 * into one with a local loop and no self calls.
 *)
functor IntroduceLoops(S: INTRODUCE_LOOPS_STRUCTS): INTRODUCE_LOOPS = 
struct

open S
open Dec PrimExp Transfer

fun introduceLoops(Program.T{datatypes, globals, functions, main}) =
   let
      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let
	     val callsItself = ref false
	     exception NonTail
	     val noChange = (args, body)
	     val (args, body) =
		(Exp.foreachTransfer
		 (body,
		  fn Call{func, cont, ...} =>
		       if Func.equals(name, func)
			  then (callsItself := true;
				case cont of
				   NONE => ()
				 | SOME _ => raise NonTail)
		       else ()
		    | _ => ());
		 if !callsItself
		    then
		       let
			  val loopName = Jump.newString "loop"
			  val newArgs =
			     List.map(args, fn (x, t) => (Var.new x, t))
			  fun loopTransfer t =
			     case t of
				Call{func, args, cont} =>
				   if Func.equals(name, func)
				      then Jump{dst = loopName,
						args = args}
				   else t
			      | _ => t
			  fun loopExp e =
			     let val {decs, transfer} = Exp.dest e
			     in Exp.make{decs = List.map(decs, loopDec),
					 transfer = loopTransfer transfer}
			     end
			  and loopDec d =
			     case d of
				Fun{name, args, body} =>
				   Fun{name = name,
				       args = args,
				       body = loopExp body}
			      | _ => d
		       in (newArgs,
			   Exp.make{decs = [Fun{name = loopName,
						args = args,
						body = loopExp body}],
				    transfer =
				    Jump{dst = loopName,
					 args = List.map(newArgs, #1)}})
		       end
		 else noChange)
		handle NonTail => noChange
	  in {name = name,
	      args = args,
	      body = body,
	      returns = returns}
	  end)
   in Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LOOP_INVARIANT_STRUCTS = 
   sig
      include SHRINK
   end

signature LOOP_INVARIANT = 
   sig
      include LOOP_INVARIANT_STRUCTS
      
      val loopInvariant: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Remove loop invariant args to local loops.
 * fun loop(x, y) = ... loop(x, z) ...
 *
 * becomes
 *
 * fun loop(x, y') =
 *     let fun loop'(y) = ... loop'(z) ...
 *     in loop'(y')
 *     end
 *)

functor LoopInvariant(S: LOOP_INVARIANT_STRUCTS): LOOP_INVARIANT = 
struct

open S
open Dec PrimExp Transfer

fun introExp(e: Exp.t): Exp.t =
   let
      val {get = jumpInfo: Jump.t -> {
				      callsSelf: bool ref,
				      inScope: bool ref,
				      invariant: (Var.t * bool ref) list,
				      newName: Jump.t option ref
				     },
	   set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("info", Jump.layout))

      fun loopTransfer t =
	 case t of
	    Jump{dst, args} =>
	       let val {callsSelf, inScope, invariant, ...} = jumpInfo dst
	       in if !inScope
		     then (callsSelf := true;
			   List.foreach2(args, invariant, fn (x, (y, b)) =>
					 if !b andalso not(Var.equals(x, y))
					    then b := false
					 else ()))
		  else ()
	       end
	  | _ => ()

      fun loopExp e =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(Exp.decs e, loopDec);
	    loopTransfer transfer
	 end
      and loopDec d =
	 case d of
	    Fun{name, args, body} =>
	       let val inScope = ref false
	       in setJumpInfo(name,
			      {callsSelf = ref false,
			       invariant = List.map(args, fn (x, _) =>
						    (x, ref true)),
			       inScope = inScope,
			       newName = ref NONE})
		  ; inScope := true
		  ; loopExp body
		  ; inScope := false
	       end
	  | _ => ()

      val _ = loopExp e

      fun remove(xs: 'a list, inv: ('b * bool ref) list): 'a list =
	 List.reverse
	 (List.fold2(xs, inv, [], fn (x, (_, b), xs) =>
		     if !b then xs else x :: xs))

      fun loopTransfer t =
	 case t of
	    Jump{dst, args} =>
	       let val {newName, invariant, ...} = jumpInfo dst
	       in case !newName of
		  NONE => t
		| SOME dst =>
		     Jump{dst = dst,
			  args = remove(args, invariant)}
	       end
	  | _ => t
	       
      fun loopExp e =
	 let val {decs, transfer} = Exp.dest e
	 in Exp.make{decs = List.map(decs, loopDec),
		     transfer = loopTransfer transfer}
	 end
      and loopDec d =
	 case d of
	    Fun{name, args, body} =>
	       let
		  val {newName, callsSelf, invariant, ...} =
		     jumpInfo name
		  val (args, body) =
		     if !callsSelf
			andalso List.exists(invariant, ! o #2)
			then let val name' = Jump.new name
				 val (outerFormals,
				      innerFormals,
				      innerActuals) =
				    List.foldr2
				    (args, invariant, ([], [], []),
				     fn ((x, t), (_, b), (ofs, ifs, ias)) =>
				     if !b
					then ((x, t) :: ofs, ifs, ias)
				     else let val x' = Var.new x
					  in ((x', t) :: ofs,
					      (x, t) :: ifs,
					      x' :: ias)
					  end)
				 val _ = newName := SOME name'
				 val body =
				    Exp.make
				    {decs = [Fun{name = name',
						 args = innerFormals,
						 body = loopExp body}],
				     transfer = Jump{dst = name',
						     args = innerActuals}}
			     in newName := NONE;
				(outerFormals, body)
			     end
		     else (args, loopExp body)
	       in Fun{name = name, args = args, body = body}
	       end
	  | _ => d

      val e = loopExp e
	 
   in e
   end

val loopInvariant = simplifyProgram introExp

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature POLY_EQUAL_STRUCTS = 
   sig
      include SHRINK
   end

signature POLY_EQUAL = 
   sig
      include POLY_EQUAL_STRUCTS
      
      val polyEqual: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PolyEqual(S: POLY_EQUAL_STRUCTS): POLY_EQUAL = 
struct

open S

(*
 * This pass implements polymorphic equality.
 *
 * For each datatype tycon and vector type, it builds an equality function and
 * translates calls to = into calls to that function.
 *
 * Also generates calls to primitives stringEqual and intInfEqual.
 *
 * For tuples, it does the equality test inline.  I.E. it does not create
 * a separate equality function for each tuple type.
 *
 * All equality functions are only created if necessary, i.e. if equality
 * is actually used at a type.
 *
 * Optimizaions:
 *  - For datatype tycons that are enumerations, do not build a case dispatch,
 *    just use eq, since you know the backend will represent these as ints.
 *  - Deep equality always does an eq test first.
 *  - If one argument to = a constant int and the type will get translated
 *    to an IntOrPointer, then just use eq instead of the full equality.  This
 *    is important for implementing code like the following efficiently:
 *       if x = 0  ...    (where x is an IntInf.int)
 *)

open Dec PrimExp Transfer

structure Dexp =
   struct
      open DirectExp

      val casee =
	 fn {test, cases, default, ty} =>
	 casee{cause = Cause.PolyEqual,
	       test = test, cases = cases, default = default, ty = ty}
	 
      fun andd(e1, e2) =
	 casee{test = e1,
	       cases = [{con = Con.truee, args = [], body = e2},
			{con = Con.falsee, args = [], body = falsee}],
	       default = NONE,
	       ty = Type.bool}

      fun ands es =
	 case es of
	    [] => truee
	  | [e] => e
	  | e :: es => andd(e, ands es)

      fun or(e1, e2) =
	 casee{test = e1,
	       cases = [{con = Con.truee, args = [], body = truee},
			{con = Con.falsee, args = [], body = e2}],
	       default = NONE,
	       ty = Type.bool}
   end

val toExp = (shrinkExp []) o Dexp.toExp

fun polyEqual(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = info, set = setInfo} =
	 Property.new(Tycon.plist, Property.initRaise("info", Tycon.layout))

      val {get = isConst, set = setConst} =
	 Property.new(Var.plist, Property.initConst false)

      val isEnum = #isEnum o info
      val tyconCons = #cons o info

      val _ =
	 List.foreach
	 (datatypes, fn {tycon, cons} =>
	  setInfo(tycon,
		  {isEnum = List.forall(cons,
					fn {args = [], ...} => true
					 | _ => false),
		   cons = cons}))
	 
      val newFunctions = ref []
	 
      val {get = getEqualFunc, set = setEqualFunc} =
	 Property.new(Tycon.plist, Property.initConst NONE)

      val {get = getVectorEqualFunc,
	   set = setVectorEqualFunc,
	   destroy = destroyVectorEqualFunc, ...} =
	 Property.newDest(Type.plist, Property.initConst NONE)

      fun equalFunc(tycon: Tycon.t): Func.t =
	 case getEqualFunc tycon of
	    SOME f => f
	  | NONE =>
	       let val name = Func.newString "equal"
		  val _ = setEqualFunc(tycon, SOME name)
		  val arg1 = Var.newNoname()
		  val arg2 = Var.newNoname()
		  val ty = Type.con(tycon, [])
		  val args = [(arg1, ty), (arg2, ty)]
		  val body =
		     Dexp.casee
		     {test = Dexp.var(arg1, ty),
		      default = NONE, ty = Type.bool,
		      cases =
		      List.map
		      (tyconCons tycon, fn {con, args} =>
		       let
			  fun makeArgs() =
			     List.map(args, fn t => (Var.newNoname(), t))
			  val xs = makeArgs()
			  val ys = makeArgs()
		       in {con = con, args = xs,
			   body =
			   Dexp.casee
			   {test = Dexp.var(arg2, ty),
			    cases = [{con = con, args = ys,
				      body = Dexp.ands(List.map2(xs, ys,
								 fn ((x, t), (y, _)) =>
								 equal(x, y, t)))}],
			    default = SOME(Dexp.falsee),
			    ty = Type.bool}}
		       end)}
	       in ListRef.push(newFunctions,
			       {name = name,
				args = args,
				body = toExp body,
				returns = [Type.bool]});
		  name
	       end
      
      and vectorEqualFunc(t: Type.t): Func.t =
	 case getVectorEqualFunc t of
	    SOME f => f
	  | NONE =>
	       let
		  (* build two functions, one that checks the lengths and the
		   * other that loops
		   *)
		  val name = Func.newString "vectorEqual"
		  val _ = setVectorEqualFunc(t, SOME name)
		  val loop = Func.newString "vectorEqualLoop"
		  val vt = Type.vector t
		  val v1 = (Var.newNoname(), vt)
		  val v2 = (Var.newNoname(), vt)
		  val _ =
		     ListRef.push
		     (newFunctions,
		      {name = name, 
		       args = [v1, v2],
		       returns = [Type.bool],
		       body =
		       toExp
		       (let val v1 = Dexp.var v1
			    val v2 = Dexp.var v2
			    fun length e = Dexp.primApp{prim = Prim.vectorLength,
							targs = [t],
							args = [e],
							ty = Type.int}
			in Dexp.andd(Dexp.eq(length v1, length v2, Type.int),
				     Dexp.call
				     {func = loop,
				      args = [Dexp.int 0, length v1, v1, v2],
				      ty = Type.bool})
			end)})
		  val v1 = (Var.newNoname(), vt)
		  val v2 = (Var.newNoname(), vt)
		  val i = (Var.newNoname(), Type.int)
		  val len = (Var.newNoname(), Type.int)
		  val _ = 
		     ListRef.push
		     (newFunctions,
		      {name = loop,
		       args = [i, len, v1, v2],
		       returns = [Type.bool],
		       body =
		       toExp
		       (let val i = Dexp.var i
			    val len = Dexp.var len
			    val v1 = Dexp.var v1
			    val v2 = Dexp.var v2
			    fun sub(v, i) = Dexp.primApp{prim = Prim.vectorSub,
							 targs = [t],
							 args = [v, i],
							 ty = t}
			in Dexp.or(Dexp.eq(i, len, Type.int),
				   Dexp.name2
				   (sub(v1, i), sub(v2, i),
				    fn (x1, x2) =>
				    Dexp.andd(equal(x1, x2, t),
					      Dexp.call
					      {func = loop,
					       args = [Dexp.+(i, Dexp.int 1),
						       len, v1, v2],
					       ty = Type.bool})))
			end)})
	       in name
	       end

       and equal(x1: Var.t, x2: Var.t, ty: Type.t): Dexp.t =
	 let val e1 = Dexp.var(x1, ty)
	    val e2 = Dexp.var(x2, ty)
	    fun prim(p, targs) =
	       Dexp.primApp{prim = p, targs = targs, args = [e1, e2],
			    ty = Type.bool}
	    fun eq() = prim(Prim.eq, [ty])
	    fun hasConstArg() = isConst x1 orelse isConst x2
	 in case Type.dest ty of
	    Type.Char => eq()
	  | Type.Int => eq()
	  | Type.IntInf => if hasConstArg()
			      then eq()
			   else prim(Prim.intInfEqual, [])
	  | Type.Word => eq()
	  | Type.Word8 => eq()
	  | Type.String => prim(Prim.stringEqual, [])
	  | Type.Array _ => eq()
	  | Type.Vector t =>
	       Dexp.call{func = vectorEqualFunc t,
			 args = [e1, e2],
			 ty = Type.bool}
	  | Type.Ref _ => eq()
	  | Type.Datatype tycon =>
	       if isEnum tycon orelse hasConstArg()
		  then eq()
	       else Dexp.call{func = equalFunc tycon,
			      args = [e1, e2],
			      ty = Type.bool}
	  | Type.Tuple tys =>
	       let
		  (* test components i, i+1, ..., max-1 *)
		  fun loop(i, tys: Type.t list): Dexp.t =
		     case tys of
			[] => Dexp.truee
		      | ty :: tys =>
			   Dexp.name2
			   (Dexp.select{tuple = e1, offset = i, ty = ty},
			    Dexp.select{tuple = e2, offset = i, ty = ty},
			    fn (y1, y2) =>
			    Dexp.andd(equal(y1, y2, ty), loop(i + 1, tys)))
	       in loop(0, tys)
	       end
	  | _ => Error.bug "equal of strange type"
	 end

      fun loopBind{var, ty, exp} =
	 let fun const() = setConst(var, true)
	 in case exp of
	    Const c =>
	       (case Const.node c of
		   Const.SmallIntInf _ => const()
		 | _ => ())
	  | ConApp{args = [], ...} => const()
	  | _ => ()
	 end

      val _ = List.foreach(globals, loopBind)

      fun loopExp e =
      	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(decs,
			 fn Bind b => loopBind b
			  | _ => ())
	    ; (List.foldr
	       (decs, Exp.make{decs = [], transfer = transfer}, fn (d, e) =>
		let fun simple d = Exp.prefix(e, d)
		   fun normal() = simple d
		in case d of
		   Fun{name, args, body} =>
		      simple(Fun{name = name, args = args, body = loopExp body})
		 | Bind{var, exp, ...} =>
		      (case exp of
			  PrimApp{prim, targs, args} =>
			     (case (Prim.name prim, targs, args) of
				 (Prim.Name.Equal, [ty], [x1, x2]) =>
				    Dexp.send(equal(x1, x2, ty),
					      fn (e, _) => simple(Bind{var = var,
								       ty = Type.bool,
								       exp = e}))
			      | _ => normal())
			| _ => normal())
		 | _ => normal()
		end))
	 end

(*       (* compute poly-equal now for globals *)
 *       val globals =
 * 	 let
 * 	    val {get = value: Var.t -> unit ref * PrimExp.t, set} =
 * 	       Property.new(Var.plist, Property.initRaise("exp", Var.layout))
 * 	    fun equal(x: Var.t, y: Var.t, ty: Type.t): bool =
 * 	       if Var.equals(x, y)
 * 		  then true
 * 	       else
 * 		  let open Type
 * 		  in case dest ty of
 * 		     Char => (case (value x, value y) of
 * 				 (Const c, Const c') => Const.equals(c, c')
 * 			       | _ => Error.bug "poly equal expected constant"
 * 				 Type.Datatype
 * 
 * 				 
 * 	 in List.reverse
 * 	    (List.fold
 * 	     (globals, [], fn (bind as {var, exp, ...}, globals) =>
 * 	      (set(var, (case exp of
 * 			    Var y => value y
 * 			  | _ => (ref(), exp)))
 * 	       (case exp of
 * 		   PrimApp{prim, targs, args} =>
 * 		      (case (Prim.name prim, args) of
 * 			  (Prim.Name.Equal, [x, y]) =>
 * 			     {var = var, ty = ty,
 * 			      exp = ConApp{args = [],
 * 					   con = if equal(x, y)
 * 						    then Con.truee
 * 						 else Con.falsee}}
 * 			| _ => bind)
 * 		 | _ => bind) :: globals)))
 * 	 end
 *)
	   
      val functions =
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name,
		   args = args,
		   returns = returns,
		   body = loopExp body})

      val program =
	 Program.T{datatypes = datatypes,
		   globals = globals,
		   functions = !newFunctions @ functions,
		   main = main}

   in Program.clear program;
      destroyVectorEqualFunc();
      program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature RAISE_TO_JUMP_STRUCTS = 
   sig
      include INFER_HANDLERS
   end

signature RAISE_TO_JUMP = 
   sig
      include RAISE_TO_JUMP_STRUCTS

      (* Replaces raises where the target handler is known with a jump to that
       * handler.
       *)
      val raiseToJump: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RaiseToJump(S: RAISE_TO_JUMP_STRUCTS): RAISE_TO_JUMP = 
struct

open S
open Dec Transfer

fun raiseToJump(p as Program.T{datatypes, globals, functions, main}) =
   let
      val jumpHandlers = inferHandlers p
      fun loopExp(e: Exp.t, hs: Jump.t list): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	 in Exp.make{decs = List.map(decs, loopDec),
		     transfer = (case transfer of
				    Raise args =>
				       (case List.fold(decs, hs, deltaHandlers) of
					   [] => transfer
					 | h :: _ =>
					      Jump{dst = h, args = args})
				  | _ => transfer)}
	 end
      and loopDec d =
	 case d of
	    Fun{name, args, body} =>
	       Fun{name = name, args = args,
		   body = loopExp(body, jumpHandlers name)}
	  | _ => d
   in Program.T{datatypes = datatypes,
		globals = globals,
		functions = List.map(functions, fn {name, args, body, returns} =>
				     {name = name,
				      args = args,
				      body = loopExp(body, []),
				      returns = returns}),
		main = main}
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature FIXED_POINT = 
   sig
      val fixedPoint: ((unit -> unit) -> unit) -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure FixedPoint: FIXED_POINT =
   struct
      fun fixedPoint f =
	 let
	    fun loop() =
	       let val changed = ref false
	       in f(fn () => changed := true);
		  if !changed
		     then loop()
		  else ()
	       end
	 in loop()
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REDUNDANT_STRUCTS = 
   sig
      include CPS_TREE
   end

signature REDUNDANT = 
   sig
      include REDUNDANT_STRUCTS
      
      val redundant: Program.t -> Program.t
   end
signature RING_ELT_LIST =
   sig
      structure R: RING

      include LIST

      val + : R.t t -> R.t
      val * : R.t t -> R.t

      val mapFromTo: R.t * R.t * (R.t -> 'a) -> 'a t
      val fromTo: R.t * R.t -> R.t t
   end
(*-------------------------------------------------------------------*)
(*                            RingEltList                            *)
(*-------------------------------------------------------------------*)

functor RingEltList(structure L: LIST
		    structure R: ORDERED_RING): RING_ELT_LIST =
struct

structure R = R
open L
   
val op + = fn l => fold(l, R.zero, R.+)

val op * = fn l => fold(l, R.one, R.* )

fun mapFromTo(n, m, f) =
   iterate(n, fn n => if R.>(n, m) then Empty
		      else First(f n, R.add1 n))
   
fun fromTo(n, m) = mapFromTo(n, m, fn x => x)

(*val equals = fn (l, l') => L.equals(l, l', R.equals)
	 
fun output(l, out) = L.output(l, ", ", R.output, out)
  *) 
end

structure IntList = RingEltList(structure L = List
				structure R = Int)
(*-------------------------------------------------------------------*)
(*                         ConvertToSequence                         *)
(*-------------------------------------------------------------------*)

functor ConvertToSequence(structure S: SEQUENCE
			  type 'a t
			  val convert: 'a t -> 'a S.t): SEQUENCE =
struct


type 'a t = 'a t

fun size s = S.size(convert s)

fun foldi(s, b, f) = S.foldi(convert s, b, f)
   
fun fold(s, b, f) = S.fold(convert s, b, f)

fun fold2(s, s', b, f) = S.fold2(convert s, convert s', b, f)

fun fold3(s1, s2, s3, b, f) = S.fold3(convert s1, convert s2, convert s3, b, f)

fun foreach(a, f) = S.foreach(convert a, f)

fun foreach2(s, s', f) = S.foreach2(convert s, convert s', f)

fun foreach3(s, s', s'', f) = S.foreach3(convert s, convert s', convert s'', f)
   
fun forall(s, f) = S.forall(convert s, f)

fun forall2(s, s', f) = S.forall2(convert s, convert s', f)

fun exists(s, f) = S.exists(convert s, f)

fun contains(s, x, eq) = S.contains(convert s, x, eq)
   
val equals = fn (s, s', eq) => S.equals(convert s, convert s', eq)

fun layout p s = S.layout p (convert s)
fun layoutSep p s = S.layoutSep p (convert s)
(*fun output(sep, outElt) (a, out) = S.output(sep, outElt) (convert a, out)*)

fun isEmpty s = S.isEmpty(convert s)

val length = size
   
fun foldr(s, b, f) = S.foldr(convert s, b, f)
   
fun foldr2(s, s', b, f) = S.foldr2(convert s, convert s', b, f)
   
fun foralli(a, f) = S.foralli(convert a, f)
fun foreachi(a, f) = S.foreachi(convert a, f)

fun nth(s,n) = S.nth(convert s,n)

fun lookup(s, p) = S.lookup(convert s, p)
fun peek(s, p) = S.peek(convert s, p)

fun outputN(n, sep, outElt) (s, out) =
   S.outputN(n, sep, outElt) (convert s, out)

fun toList s = S.toList(convert s)
   
end
signature PROMISE =
   sig
      type 'a t
	 
      val delay: (unit -> 'a) -> 'a t
	 
      val force: 'a t -> 'a
   end
(*-------------------------------------------------------------------*)
(*                              Promise                              *)
(*-------------------------------------------------------------------*)

functor PromiseUnmemo(): PROMISE =
struct
    
datatype 'a t = T of unit -> 'a

val delay = T

fun force(T th) = th()

end

structure PromiseUnmemo = PromiseUnmemo()
functor Promise() :
   sig
      include PROMISE
      exception Force
      val lazy: (unit -> 'a) -> (unit -> 'a)
   end =
struct

datatype 'a t = T of 'a state ref
and 'a state =
   Unevaluated of unit -> 'a
  | Evaluating
  | Evaluated of 'a

fun delay th = T(ref(Unevaluated th))

exception Force
fun force(T r) =
   case !r of
      Evaluated x => x
    | Unevaluated th => ((r := Evaluating ;
			  let val x = th()
			  in r := Evaluated x ; x
			  end) handle exn => (r := Unevaluated th ;
					      raise exn))
    | Evaluating => raise Force

fun lazy th =
   let val p = delay th
   in fn () => force p
   end

end

structure Promise = Promise()
signature LAZY_LIST =
   sig
      include LIST

      (* include LIST_BASIC_LAZY *)
      val lazyCons: (unit -> ('a * 'a t) option) -> 'a t
      (* *)

      val appendReverse: 'a t * 'a t -> 'a t
      val appendReverseIncremental: 'a t * 'a t -> 'a t
	 (* performs one step of the reverse for
	  each step of the append *)

   (* appendReverseNthTail c (l, l',n) =
    L.append(l, L.reverse(L.nthTail(l',n))),
    where the computation is performed incrementally, with c steps
    of the tail and reverse being performed for each element of l.
       pre: length l*)
      val appendReverseNthTail :
	 Int.t -> 'a t * 'a t * Int.t -> 'a t
   end
signature BASIC_LAZY_LIST =
   sig
      include BASIC_LIST_ITER

      val lazyCons: (unit -> ('a * 'a t) option) -> 'a t
   end
(*-------------------------------------------------------------------*)
(*                           BasicLazyList                           *)
(*-------------------------------------------------------------------*)

functor BasicLazyList(P: PROMISE): BASIC_LAZY_LIST =
struct

datatype 'a t = T of ('a * 'a t) option P.t
   
fun destruct(T p) = P.force p
      
datatype ('a, 'b) iter =
   Empty
  | First of 'a * 'b
  | Rest of 'a t
       
fun empty() = T(P.delay(fn () => NONE))

fun lazyCons th = T(P.delay th)

fun cons(x, l) = T(P.delay(fn () => SOME(x, l)))
   
fun iterate(b, f) =
   let
      fun iter b =
	 T(P.delay(fn () =>
		 case f b of
		    Empty => NONE
		  | First(a, b) => SOME(a, iter b)
		  | Rest s => destruct s))
   in iter b
   end

end
(*-------------------------------------------------------------------*)
(*                             LazyList                              *)
(*-------------------------------------------------------------------*)

functor LazyList(structure L: BASIC_LAZY_LIST): LAZY_LIST =
struct

structure L' = List(structure L = L)
   
open Int L L'
   
fun appendReverse(l, r) =
   appendThunk(l, fn () => reverse r)

fun appendReverseIncremental(l, l') =
   iterate((l, l', empty()), fn (l, l', a) =>
	   case (destruct l, destruct l') of
		(NONE, NONE) => Rest(empty())
	      | (NONE, SOME(x, _)) => Rest(cons(x, a))
	      | (SOME(x, l), NONE) => Rest(cons(x, append(l, a)))
	      | (SOME(x, l), SOME(x', l')) => First(x, (l, l', cons(x', a))))
   
fun appendAppendReverse c =
   let fun rot(L, R, A) =
	 let fun finish() = append(L, append(reverse R, A))
	 in case destruct L of
	    NONE => finish()
	  | SOME(x, L) =>
	       case splitAtMost(R, c) of
		  NONE => finish()
		| SOME(R, R') =>
		     lazyCons(fn () =>
			      SOME(x,
				   rot(L, R', append(reverse R, A))))
	 end
   in rot
   end
      
fun appendReverseNthTail c =
   let val aar = appendAppendReverse c
      fun rot(L, R,n) = 
	 if n >= c
	    then L.lazyCons(fn () =>
			    SOME(head L,
				 rot(tail L,
				     nthTail(R, c),
				     n - c)))
	 else aar(L, nthTail(R,n), empty())
   in rot
   end

end

functor StandardLazyList(P: PROMISE) =
   LazyList(structure L = BasicLazyList(P)
	    structure I = Int)

functor LazyListUnmemo(I: INTEGER) =
   LazyList(structure L = BasicLazyList(PromiseUnmemo)
	    structure I = I)

structure LazyListUnmemo = StandardLazyList(PromiseUnmemo)

structure LazyList = StandardLazyList(Promise)
signature BASIC_ARRAY =
   sig
      structure Shape: T
      structure Index: T

      type 'a t
	 
      exception New
      val new: Shape.t * 'a -> 'a t
      val array: Shape.t * 'a -> 'a t (* synonym for new *)
      val tabulate: Shape.t * (Index.t -> 'a) -> 'a t

      val shape: 'a t -> Shape.t

      val sub: 'a t * Index.t -> 'a
      val update: 'a t * Index.t * 'a -> unit
   end
signature BASIC_ARRAY_1D =
   sig
      include BASIC_ARRAY where type Shape.t = int where type Index.t = int

      val fromList: 'a list -> 'a t
   end
signature ARRAY =
   sig
      include BASIC_ARRAY
      
      val copy: 'a t -> 'a t
      val equals: 'a t * 'a t * ('a * 'a -> bool) -> bool
      val exists: 'a t * ('a -> bool) -> bool
      val foldi: 'a t * 'b * (Index.t * 'a * 'b -> 'b) -> 'b
      val forall: 'a t * ('a -> bool) -> bool
      val foralli: 'a t * (Index.t * 'a -> bool) -> bool
      val foreach: 'a t * ('a -> unit) -> unit
      val foreachi: 'a t * (Index.t * 'a -> unit) -> unit
      val foreach2: 'a t * 'b t * ('a * 'b -> unit) -> unit
      val getAndSet: 'a t -> (Index.t -> 'a) * (Index.t * 'a -> unit)
      val layout: ('a -> Layout.t) -> 'a t -> Layout.t
      val shuffle: 'a t -> unit
      val size: 'a t -> int
      val swap: 'a t * Index.t * Index.t -> unit
   end
signature ARRAY_1D =
   sig
      include ARRAY
      
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldr: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldr2: 'a t * 'b t * 'c * ('a * 'b * 'c -> 'c) -> 'c
      val peek: 'a t * ('a -> bool) -> 'a option
      val outputN: Int.t * string * ('a * Out.t -> unit) -> 'a t * Out.t -> unit

      val fromList: 'a list -> 'a t
      val toList: 'a t -> 'a list
   end
functor BasicArray1D(): BASIC_ARRAY_1D =
struct

structure A = Pervasive.Array
structure I = FixedInteger
structure Shape = I
structure Index = I
structure L = StrictList

open A
   
type 'a t = 'a array
   
exception New = Size

val new = array
val size = length
val shape = size

end
functor Array1D(A: BASIC_ARRAY_1D): ARRAY_1D =
struct

open Int A
structure Index = Int
structure Shape = Int

local
   structure L = LazyListUnmemo(Int)

   fun convert a =
      let val maxIndex = A.shape a
      in L.iterate(Int.zero, fn i =>
		   if Int.equals(i, maxIndex) then L.Empty
		   else L.First(sub(a, i), Int.add1 i))
      end
   
   structure S =
      ConvertToSequence(structure S = L
			type 'a t = 'a A.t
			val convert = convert)
in
   open S
end

val size = A.shape
val length = size
val nth = sub

fun foldi(a, b, f) =
   Int.fold(0, size a - 1, b, fn (i, b) => f(i, sub(a, i), b))

fun foreach(a, f) = fold(a, (), f o #1)

fun foldr(a, b, f) =
   Int.foldDown(size a - 1, 0, b, fn (i, b) => f(sub(a, i), b))
   
fun copy a = tabulate(size a, fn i => sub(a, i))

fun toList a = foldr(a, List.empty(), List.cons)
   
fun swap(a, i, j) =
   let val t = sub(a, i)
   in (update(a, i, sub(a, j)) ;
       update(a, j, t))
   end

fun maxIndex a = sub1(size a)

fun foreachDown(a, f) =
   Int.forDown(maxIndex a, zero, fn i => f(i, sub(a, i)))
   
fun shuffle a =
   foreachDown(a, fn (i, _) => swap(a, i, randomRange(zero, i)()))

fun getAndSet a = (fn i => sub(a, i),
		   fn (i, x) => update(a, i, x))
   
end

structure Array1D = Array1D(BasicArray1D())
structure Array = Array1D
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature EQUIVALENCE_RELATION_STRUCTS = 
   sig
   end

signature EQUIVALENCE_RELATION = 
   sig
      include EQUIVALENCE_RELATION_STRUCTS
      
      (* An equivalence relation on integers. *)
      type t

      (* r(x, y) iff 0 <= x < n andalso 0 <= y < n *)
      val universal : int -> t

      (* make(l, f) is the equivalence relation r such that
       * r(x, y) iff f(List.nth(l, x), List.nth(l, y))
       * f must be an equivalence relation on 'a (this is not checked).
       *)
      val make : 'a list * ('a * 'a -> bool) -> t

      val areEquivalent : t * int * int -> bool

      (* Return the equivalence classes in the relation *)
      val classes : t -> int list list

      (* refine(r, f) is the equivalence relation r' such that
       * r'(x, y) iff r(x, y) andalso f(x, y).
       * f must be an equivalence relation (this is not checked).
       *)
      val refine : t * (int * int -> bool) -> t
	 
      val equals : t * t -> bool
      val layout : t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor EquivalenceRelation(S: EQUIVALENCE_RELATION_STRUCTS):> EQUIVALENCE_RELATION = 
struct

open S

(* An equivalence relation is represented as a list of pairs of equivalent
 * integers, except that reflexive pairs (e.g. (1, 1)) are not listed, and
 * of the pairs (a, b) (b, a) where a < b, only (a, b) is listed
 * The list of pairs is kept sorted in increasing lexicographic order.
 *)

datatype t = T of (int * int) list

fun layout(T ps) =
   List.layout (fn (x, y) => Layout.tuple[Int.layout x, Int.layout y]) ps

fun make(xs, equals) =
   let
      val xs = List.mapi(xs, fn ix => ix)
      val rec loop =
	 fn (xs, accum) =>
	 case xs of
	    [] => List.reverse accum
	  | (i,x)::xs =>
	       loop(xs,
		    case List.keepAll(xs, fn (_,x') => equals(x,x')) of
		       [] => accum
		     | yes => List.fold(yes, accum, fn ((i',_),accum) =>
					(i,i')::accum))
   in T(loop(xs, []))
   end

fun universal n =
   let
      val rec loop =
	 fn [] => []
	  | x::l => List.mapAppend(l, fn y => (x,y), loop l)
   in T(loop(IntList.fromTo(0, n - 1)))
   end

fun refine(T r, f) = T(List.keepAll(r, f))

fun areEquivalent(T r, x, y) =
   let val x' = Int.min(x,y)
      val y' = Int.max(x,y)
   in List.exists(r, fn (a,b) => a = x' andalso b = y')
   end

val equals = op =
   
fun classes(T pairs) =
   let val max = List.fold(pairs, ~1, fn ((_,y),m) =>
			   Int.max(y,m))
      val done = Array.new(max + 1, false)
      fun loop(pairs, i, class, classes) =
	 case pairs of
	    [] => class::classes
	  | (j,k)::pairs =>
	       if i = j 
		  then let val class = if Array.sub(done,k)
					  then class
				       else (Array.update(done,k,true) ;
					     k::class)
		       in loop(pairs, i, class, classes)
		       end
	       else
		  start((j,k)::pairs,
			case class of
			   [_] => classes
			 | _ => class::classes)
      and start(pairs,classes) =
	 case pairs of
	    [] => classes
	  | (i,j)::pairs => if Array.sub(done,i)
			       then start(pairs,classes)
			    else (Array.update(done,j,true) ;
				  loop(pairs, i, [j,i], classes))
   in if max = ~1
	 then []
      else start(pairs, [])
   end

(* val classes = Trace.trace("classes", layout,
 * 			  fn cs =>
 * 			  let open Layout
 * 			  in list(List.map(cs, fn c =>
 * 					   list(List.map(c,Int.layout))))
 * 			  end) classes
 *)
end

structure EquivalenceRelation = EquivalenceRelation()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Redundant(S: REDUNDANT_STRUCTS): REDUNDANT = 
struct

open S

structure Eqrel:>
   sig
      type t
	 
      val fromTypes: Type.t list -> t
	 
      val refine: t * (int * int -> bool) -> unit
      val unify: t * t -> unit
      val areEquivalent: t * int * int -> bool
      val classes: t -> int list list

      val fixedPoint: unit -> unit
	 
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end =
   struct
      structure R = EquivalenceRelation
      structure Set = DisjointSet

      datatype t = T of {me: R.t Set.t,
			 refinements: (int * int -> bool) list ref}

      val all: t list ref = ref []

      fun fromTypes ts =
	 let val er = T{me = Set.singleton(R.make(ts, Type.equals)),
			refinements = ref []}
	 in ListRef.push(all, er); er
	 end

      fun me(T{me, ...}) = Set.value me

      fun areEquivalent(er, i, j) = R.areEquivalent(me er, i, j)

      val classes = R.classes o me

      fun equals(T{me, ...}, T{me = me', ...}) = Set.equals(me, me')

      val layout = R.layout o me

      fun refine(T{refinements, ...}, f) = ListRef.push(refinements, f)

      (* Relies on the fact that all unifications happen before the fixed point.
       *)
      fun unify(T{me, ...}, T{me = me', ...}) = Set.union(me, me')

      fun fixedPoint() =
	 (FixedPoint.fixedPoint
	  (fn continue =>
	   List.foreach(!all, fn T{me, refinements} =>
			let val r = Set.value me
			   val r' = List.fold(!refinements, r,
					      fn (refinement, r) =>
					      R.refine(r, refinement))
			in if R.equals(r, r')
			      then ()
			   else (continue() ; Set.setValue(me, r'))
			end))
	  ; all := [])
   end


open Dec PrimExp Transfer

fun redundant(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = funcInfo: Func.t -> {
					  arg: Eqrel.t,
					  return: Eqrel.t
					 },
	   set = setFuncInfo} =
	 Property.new(Func.plist, Property.initRaise("info", Func.layout))

      val {get = jumpInfo: Jump.t -> Eqrel.t,
	   set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("info", Jump.layout))

      val {get = varInfo : Var.t -> {
				     index: int,
				     arg: Eqrel.t
				    } option,
	   set = setVarInfo} =
	 Property.new(Var.plist, Property.initConst NONE)

      fun varEquiv(x, y) =
	 Var.equals(x, y)
	 orelse (case (varInfo x, varInfo y) of
		    (SOME{index = i, arg = r},
		     SOME{index = i', arg = r'}) =>
		    Eqrel.equals(r, r') andalso Eqrel.areEquivalent(r, i, i')
		  | _ => false)

      (* compute the fixed point *)
      val _ =
	 let
	    fun makeFormalsRel(xs: (Var.t * Type.t) list): Eqrel.t =
	       let val eqrel = Eqrel.fromTypes(List.map(xs, #2))
	       in List.foreachi(xs, fn (i, (x, _)) =>
				setVarInfo(x, SOME{index = i, arg = eqrel}))
		  ; eqrel
	       end
	 in (* initialize all varInfo and funcInfo *)
	    List.foreach(functions, fn {name, args, returns, ...} =>
			 setFuncInfo(name, {
					    arg = makeFormalsRel args,
					    return = Eqrel.fromTypes returns
					   }))
	    (* Add the calls to all the funcInfos *)
	    ; (List.foreach
	       (functions, fn {name, body,...} =>
		let val {return, ...} = funcInfo name
		   val handleDec =
		      fn Fun{name, args, ...} =>
		            setJumpInfo(name, (makeFormalsRel args))
		       | _ => ()
		   val varEquiv =
		      fn vars => fn (i, j) => varEquiv(List.nth(vars, i),
						       List.nth(vars, j))
		   val handleTransfer =
		      fn Call{func, args, cont} =>
		           let val {arg, return = return'} = funcInfo func
			   in Eqrel.refine(arg, varEquiv args)
			      ; Eqrel.unify(return',
					    case cont of
					       NONE => return
					     | SOME c => jumpInfo c)
			   end
		       | Jump{dst, args, ...} =>
			    Eqrel.refine(jumpInfo dst, varEquiv args)
		       | Return xs => Eqrel.refine(return, varEquiv xs)
		       | _ => ()
		in Exp.foreach'
		   (body, {handleDec = handleDec,
			   handleTransfer = handleTransfer})
		end))
	    ; Eqrel.fixedPoint()
	 end

      val _ = 
	 Control.displays
	 ("redundant", fn display =>
	  List.foreach
	  (functions, fn {name, ...} =>
	   display
	   (let val {arg, return} = funcInfo name
	       open Layout
	    in seq[Func.layout name,
		   str "  ",
		   Eqrel.layout arg,
		   Eqrel.layout return]
	    end)))

      val {get = replacement : Var.t -> Var.t option, set = setReplacement} =
	 Property.new(Var.plist, Property.initConst NONE)

      datatype red =
	 Useful
	| Redundant of int (* the index it is the same as *)

      (* Turn an equivalence relation on 0 ... n -1 into a list of length n,
       * by choosing a representative of each class.
       *)
      fun 'a makeReds(xs: 'a list, r: Eqrel.t): red list =
	 let
	    val classes = Eqrel.classes r
	    fun getRed i =
	       let
		  val rec loop =
		     fn [] => Useful
		      | class :: classes =>
			   case class of
			      [] => Error.bug "empty class"
			    | [_] => Error.bug "trivial class"
			    | j :: js =>
				 if i = j
				    then Useful
				 else if List.exists(js, fn j => i = j)
					 then Redundant j
				      else loop classes
	       in loop classes
	       end
	 in List.mapi(xs, fn (i, _) => getRed i)
	 end
	  
      fun redundantFormals(r: Eqrel.t, xs: (Var.t * Type.t) list)
	 : red list * (Var.t * Type.t) list =
	 let val reds = makeReds(xs, r)
	    val xs =
	       List.fold2
	       (xs, reds, [], fn (x, red, ys) =>
		case red of
		   Useful => x :: ys
		 | Redundant i =>
		      (setReplacement(#1 x, SOME(#1(List.nth(xs, i))))
		       ; ys))
	 in (reds, List.reverse xs)
	 end

      fun keepUseful(reds: red list, xs: 'a list): 'a list =
	 List.reverse(List.fold2(reds, xs, [], fn (r, x, xs) =>
				 case r of
				    Useful => x :: xs
				  | _ => xs))

      val {get = funcReds : Func.t -> {args: red list,
					   returns: red list},
	   set = setFuncReds} =
	 Property.new(Func.plist, Property.initRaise("reds", Func.layout))

      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let val {arg, return} = funcInfo name
	     val return = makeReds(returns, return)
	     val (arg, args) = redundantFormals(arg, args)
	  in setFuncReds(name, {args = arg, returns = return})
	     ; {name = name, body = body,
		args = args,
		returns = keepUseful(return, returns)}
	  end)

      val {get = jumpReds : Jump.t -> red list, set = setJumpReds} =
	 Property.new(Jump.plist, Property.initRaise("red", Jump.layout))

      fun loopVar x =
	 case replacement x of
	    NONE => x
	  | SOME y => y

      fun loopVars xs = List.map(xs, loopVar)

      val functions =
	 List.map
	 (functions, fn {name, args, body, returns} =>
	  let
	     val {returns = returnReds, ...} = funcReds name
	     fun loopTransfer t =
		case t of
		   Call{func, args, cont} =>
		      Call{func = func, cont = cont,
			   args = loopVars(keepUseful(#args(funcReds func),
						      args))}
		 | Jump{dst, args} =>
		      Jump{dst = dst,
			   args = loopVars(keepUseful(jumpReds dst, args))}
		 | Return xs => Return(loopVars(keepUseful(returnReds, xs)))
		 | Raise xs => Raise(loopVars xs)
		 | Case{cause, test, cases, default} =>
		      Case{cause = cause, test = loopVar test, cases = cases,
			   default = default}
		 | Halt => Halt
		 | Bug => Bug

	     fun loopExp e =
		let val {decs, transfer} = Exp.dest e
		in Exp.make{decs = List.map(decs, loopDec),
			    transfer = loopTransfer transfer}
		end
	     
	     and loopDec d =
		case d of
		   Bind{var, ty, exp} =>
		      Bind{var = var, ty = ty,
			   exp = PrimExp.replaceVar(exp, loopVar)}
		 | Fun{name, args, body} =>
		      let
			 val (reds, args) =
			    redundantFormals(jumpInfo name, args)
		      in setJumpReds(name, reds)
			 ; Fun{name = name, args = args, body = loopExp body}
		      end
		 | _ => d
	  in {name = name, args = args, returns = returns,
	      body = loopExp body}
	  end)

      val p = Program.T{datatypes = datatypes,
			globals = globals,
			functions = functions,
			main = main}
   in Program.clear p; p
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REMOVE_UNUSED_CONSTRUCTORS_STRUCTS = 
   sig
      include CPS_TREE
   end

signature REMOVE_UNUSED_CONSTRUCTORS = 
   sig
      include REMOVE_UNUSED_CONSTRUCTORS_STRUCTS
      
      val remove: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RemoveUnusedConstructors(S: REMOVE_UNUSED_CONSTRUCTORS_STRUCTS)
   : REMOVE_UNUSED_CONSTRUCTORS = 
struct

open S
open Dec PrimExp Transfer

fun remove(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conIsUseful: Con.t -> bool,
	   set = setConUseful,
	   destroy = destroyConUseful} =
	 Property.newDest(Con.plist, Property.initConst false)

      fun loopBind{var, ty, exp} =
	 case exp of
	    ConApp{con, ...} => setConUseful(con, true)
	  | _ => ()

      val _ =
	 (setConUseful(Con.truee, true)
	  ; setConUseful(Con.falsee, true)
	  ; List.foreach(globals, loopBind)
	  ; List.foreach(functions, fn {body, ...} =>
			 Exp.foreachBind(body, loopBind)))

      val {get = conNumCons: Con.t -> int,
	   set = setConNumCons, destroy = destroyConNumCons} =
	 Property.newDest(Con.plist,
			  Property.initRaise("num cons", Con.layout))
	 
      val datatypes =
	 List.revMap
	 (datatypes, fn {tycon, cons} =>
	  let val cons = List.keepAll(cons, conIsUseful o #con)
	     val n = List.length cons
	  in List.foreach(cons, fn {con, ...} => setConNumCons(con, n))
	     ; {tycon = tycon,
		cons = cons}
	  end)
	 
      fun loopTransfer(t: Transfer.t): Transfer.t =
	 case t of
	    Case{cause, test, cases, default} =>
	       let val cases = List.keepAll(cases, conIsUseful o #1)
		  fun keep default = Case{cause = cause,
					  test = test,
					  cases = cases,
					  default = default}
		  fun none() = keep NONE
	       in case default of
		  NONE => none()
		| SOME j => 
		     case cases of
			[] => Jump{dst = j, args = []}
		      | (c, _) :: _ =>
			   if conNumCons c = List.length cases
			      then none()
			   else keep(SOME j)
	       end
	  | _ => t
	 
      fun loopExp(e: Exp.t): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	 in Exp.make{decs = List.map(decs, loopDec),
		     transfer = loopTransfer transfer}
	 end
      and loopDec(d: Dec.t): Dec.t =
	 case d of
	    Fun{name, args, body} => Fun{name = name,
					 args = args,
					 body = loopExp body}
	  | _ => d

      val functions =
	 List.revMap
	 (functions, fn {name, args, body, returns} =>
	  {name = name, args = args, returns = returns,
	   body = loopExp body})
	 
   in destroyConUseful()
      ; destroyConNumCons()
      ; Program.T{datatypes = datatypes,
		  globals = globals,
		  functions = functions,
		  main = main}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REMOVE_UNUSED_FUNCTIONS_STRUCTS = 
   sig
      include CPS_TREE
   end

signature REMOVE_UNUSED_FUNCTIONS = 
   sig
      include REMOVE_UNUSED_FUNCTIONS_STRUCTS
      
      val remove: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RemoveUnusedFunctions(S: REMOVE_UNUSED_FUNCTIONS_STRUCTS)
   : REMOVE_UNUSED_FUNCTIONS = 
struct

open S

fun remove(Program.T{datatypes, globals, functions, main}) =
   let
      val {get: Func.t -> {body: Exp.t,
			   visited: bool ref},
	   set, destroy} =
	 Property.newDest(Func.plist,
			  Property.initRaise("info", Func.layout))
      val _ = List.foreach(functions, fn {name, body, ...} =>
			   set(name, {body = body, visited = ref false}))
      fun visit(f: Func.t) =
	 let val {visited, body} = get f
	 in if !visited
	       then ()
	    else (visited := true
		  ; Exp.foreachTransfer(body,
					fn Transfer.Call{func, ...} => visit func
					 | _ => ()))
	 end
      val _ = visit main
      val functions = List.keepAll(functions, fn f => !(#visited(get(#name f))))
   in destroy()
      ; Program.T{datatypes = datatypes,
		  globals = globals,
		  functions = functions,
		  main = main}
   end
	 

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REMOVE_UNUSED_GLOBALS_STRUCTS = 
   sig
      include CPS_TREE
   end

signature REMOVE_UNUSED_GLOBALS = 
   sig
      include REMOVE_UNUSED_GLOBALS_STRUCTS
      
      val remove: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RemoveUnusedGlobals(S: REMOVE_UNUSED_GLOBALS_STRUCTS): REMOVE_UNUSED_GLOBALS = 
struct

open S
open Dec PrimExp

fun remove(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = isGlobal: Var.t -> bool, set = setGlobal,
	   destroy = destroyGlobal} =
	 Property.newDest(Var.plist, Property.initConst false)
	 
      val {get = isUseful: Var.t -> bool, set = setUseful,
	   destroy = destroyUseful} =
	 Property.newDest(Var.plist, Property.initConst false)

      fun var x = if isGlobal x then setUseful(x, true) else ()
      fun handlePrimExp e = PrimExp.foreachVar(e, var)
      val handleDec =
	 fn Bind{exp, ...} => handlePrimExp exp
	  | _ => ()
      fun handleTransfer t = Transfer.foreachVar(t, var)
   in
      List.foreach(globals, fn {var, ...} => setGlobal(var, true))
      ; List.foreach(functions, fn {body, ...} =>
		     Exp.foreach'(body, {handleDec = handleDec,
					 handleTransfer = handleTransfer}))
      ; List.foreach(List.reverse globals, fn {var, exp, ...} =>
		     if isUseful var
			then handlePrimExp exp
		     else ())
      ; (Program.T{datatypes = datatypes,
		   globals = List.keepAll(globals, isUseful o #var),
		   functions = functions,
		   main = main}
	 before (destroyGlobal()
		 ; destroyUseful()))
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SAVE_WORLD_STRUCTS = 
   sig
      include CPS_TREE
   end

signature SAVE_WORLD = 
   sig
      include SAVE_WORLD_STRUCTS
      
      val saveWorld: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor SaveWorld(S: SAVE_WORLD_STRUCTS): SAVE_WORLD = 
struct

open S

open Dec PrimExp Transfer

fun usesSaveWorld(Program.T{functions, ...}): bool =
   let exception Yes
   in (List.foreach
       (functions, fn {body, ...} =>
	Exp.foreachBind(body, fn {exp, ...} =>
			case exp of
			   PrimApp{prim, ...} =>
			      if (case Prim.name prim of
				     Prim.Name.SaveWorld => true
				   | Prim.Name.SaveStack => true
				   | Prim.Name.Size => true
				   | _ => false)
				 then raise Yes
			      else ()
			 | _ => ()));
       false) handle Yes => true
   end

fun saveWorld(Program.T{datatypes, globals, functions, main}) =
   let
      val sizeFunctions = ref []
      val {get = sizeName, destroy, ...} =
	 Property.newDest
	 (Type.plist,
	  Property.initFun
	  (fn t =>
	   let val f = Func.newString "size"
	      val arg = Var.newNoname()
	      val s = Var.newNoname()
	   in ListRef.push
	      (sizeFunctions,
	       {name = f,
		args = [(arg, t)],
		body = Exp.make{decs =
				[Dec.Bind{var = s,
					  ty = Type.int,
					  exp = PrimApp{prim = Prim.size,
							targs = [t],
							args = [arg]}}],
				transfer = Return[s]},
		returns = [Type.int]});
	      f
	   end))

      val saveStack = Func.newNoname()
	   
      fun replacePrimApps(e: Exp.t): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	 in List.foldr
	    (decs, Exp.make{decs = [], transfer = transfer}, fn (d, e) =>
	     let fun normal() = Exp.prefix(e, d)
	     in case d of
		Fun{name, args, body} =>
		   Exp.prefix(e, Fun{name = name,
				     args = args,
				     body = replacePrimApps body})
	      | Bind{var, exp = PrimExp.PrimApp{prim, targs, args}, ...} =>
		   let
		      fun doit(name, formals) =
			 let val j = Jump.newNoname()
			 in Exp.make
			    {decs = [Dec.Fun{name = j, args = formals, body = e}],
			     transfer = Transfer.Call{func = name, args = args,
						      cont = SOME j}}
			 end
		   in case (Prim.name prim, targs) of
		      (Prim.Name.SaveWorld, []) =>
			 doit(Func.saveWorld, [(var, Type.bool)])
		    | (Prim.Name.SaveStack, []) =>
			 doit(saveStack, [(var, Type.stack)])
		    | (Prim.Name.Size, [t]) =>
			 doit(sizeName t, [(var, Type.int)])
		    | _ => normal()
		   end
	      | _ => normal()
	     end)
	 end

      val functions = 
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name,
		   args = args,
		   body = replacePrimApps body,
		   returns = returns})
	 
      fun make(func: Func.t, formals, return, prim, actuals) =
	 {name = func,
	  args = formals,
	  body = let val b = Var.newNoname()
		 in Exp.make{decs = [Bind{var = b,
					  ty = return,
					  exp = PrimApp{prim = prim,
							targs = [],
							args = actuals}}],
			     transfer = Return[b]}
		 end,
	  returns = [return]}
      val functions =
	 let val arg = Var.newNoname()
	 in make(Func.saveWorld,
		 [(arg, Type.string)],
		 Type.bool,
		 Prim.saveWorld,
		 [arg])
	 end
	 :: make(saveStack,
		 [],
		 Type.stack,
		 Prim.saveStack,
		 [])
	 :: (!sizeFunctions @ functions)
	 
   in destroy();
      Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

val saveWorld =
   fn p => if usesSaveWorld p then saveWorld p else p

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Simplify(S: sig
		       include SIMPLIFY_STRUCTS
		       sharing Func = Jump (* for CpsInline *)
		    end): SIMPLIFY = 
struct

open S

fun trace(name, pass) =
   Control.traceCall name
   (Trace.trace("Simplify." ^ name, Program.layout, Program.layout) pass)

structure ConstantPropagation = ConstantPropagation(S)
val constantPropagation =
   trace("constant propagation", ConstantPropagation.simplify)

structure Contify = Contify(S)
val contify = trace("contify", Contify.contify)

structure CpsInline = CpsInline(S)
val cpsInline = trace("cpsInline", CpsInline.computeInline)

(* structure DeepFlatten = DeepFlatten(S)
 * val deepFlatten = trace("deep flatten", DeepFlatten.flatten)
 *)

structure Flatten = Flatten(S)
val flatten = trace("flatten", Flatten.flatten)
val flatten =
   fn p =>
   let
      fun loop(i, p) =
	 if i = 0
	    then p
	 else loop(i - 1, flatten p)
   in loop(!Control.flattenRounds, p)
   end

structure Inline = Inline(S)
val inline = trace("inline", Inline.inline)

structure IntroduceLoops = IntroduceLoops(S)
val introduceLoops = trace("introduce loops", IntroduceLoops.introduceLoops)

(*structure LoopCount = LoopCount(S) *)

structure LoopInvariant = LoopInvariant(S)
val loopInvariant = trace("loop invariant", LoopInvariant.loopInvariant)
   
structure Useless = Useless(S)
val useless = trace("useless", Useless.useless)

structure PolyEqual = PolyEqual(S)
val polyEqual = trace("poly equal", PolyEqual.polyEqual)

structure RaiseToJump = RaiseToJump(S)
val raiseToJump = trace("raise to jump", RaiseToJump.raiseToJump)

structure Redundant = Redundant(S)
val redundant = trace("redundant", Redundant.redundant)

structure RemoveUnusedConstructors = RemoveUnusedConstructors(S)
val removeUnusedConstructors =
   trace("remove unused constructors", RemoveUnusedConstructors.remove)

structure RemoveUnusedFunctions = RemoveUnusedFunctions(S)
val removeUnusedFunctions =
   trace("remove unused functions", RemoveUnusedFunctions.remove)

structure RemoveUnusedGlobals = RemoveUnusedGlobals(S)
val removeUnusedGlobals =
   trace("remove unused globals", RemoveUnusedGlobals.remove)

structure SaveWorld = SaveWorld(S)
val saveWorld = trace("save world", SaveWorld.saveWorld)

structure SimplifyTypes = SimplifyTypes(S)
val simplifyTypes = trace("simplify types", SimplifyTypes.simplify)

fun stats p = (Control.message(fn () => Program.layoutStats p); p)

fun leafInline p =
   Ref.fluidLet
   (Control.inline, Control.Leaf{size = SOME 20 (* arbitrary *)}, fn () =>
    inline p)
val leafInline = trace("leaf inline", leafInline)

val inline =
   fn p =>
   case !Control.inline of
      Control.Recursive _ => cpsInline p
    | _ =>
	 let
	    fun loop(p, rounds) =
	       if rounds = 0
		  then p
	       else loop(inline p, rounds - 1)
	 in loop(p, !Control.inlineRounds)
	 end
			   
fun $ (f, g) x = g(f x)
infixr $

val typeCheck =
   stats
   $ (fn p => (if Control.shouldTypeCheck()
		  then Control.traceCall "typeCheck" typeCheck p
	       else ();
		  p))

fun save suffix p =
   (Ref.fluidLet
    (Control.aux, true, 
     fn () => Control.displays(suffix, fn layout => Program.layouts(p, layout)))
    ; p)

fun traces fs p = (Trace.Immediate.on fs; p)
   
(*
 * useless should run after constant propagation because const prop makes
 *  slots of tuples that are constant useless
 *
 * contify should be run before constant propagation because of the once pass
 *  that only looks at main -- hence want as much in main as possible.
 *
 * poly equal should run
 *   - after types are simplified so that many equals are turned into eqs
 *   - before inlining so that equality functions can be inlined
 *)
val simplify =
   stats
   $ removeUnusedFunctions $ stats
   $ leafInline $ stats
   $ constantPropagation $ typeCheck
   $ useless $ typeCheck
   $ simplifyTypes $ typeCheck
   $ polyEqual $ typeCheck
   $ contify $ typeCheck
   $ inline $ typeCheck
   $ raiseToJump $ typeCheck
   $ contify $ typeCheck
   $ introduceLoops $ loopInvariant $ typeCheck
   $ flatten $ typeCheck
   $ redundant $ typeCheck
   $ removeUnusedFunctions
   $ removeUnusedGlobals
   $ removeUnusedConstructors
   $ shrink $ typeCheck
   $ saveWorld (* must be last so that optimizations don't inline saveWorld *)
(*   $ LoopCount.instrument *)
   
val simplify = trace("simplify", simplify)

val typeCheck = S.typeCheck
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Shrink(S: SHRINK_STRUCTS): SHRINK = 
struct

open S
open Dec PrimExp Transfer

(* val traceSimplifyExp = Trace.trace("simplifyExp", Exp.layout, Exp.layout)
 * val traceSimplifyTransfer =
 *    Trace.trace("simplifyTransfer", Transfer.layout, Exp.layout)
 *)

fun shrinkExp globals =
   let
      val {get = globalBind: Var.t -> PrimExp.t option, set} =
	 Property.new(Var.plist, Property.initConst NONE)
   in List.foreach(globals, fn {var, exp, ty} =>
		   set(var,
		       case exp of
			  Var y => globalBind y
			| _ => SOME exp))
      ;		   
      fn (exp: Exp.t, mayDelete: bool) =>
      let
	 (* ---------------------------------- *)
	 (*             variables              *)
	 (* ---------------------------------- *)

	 val {get = varInfo: Var.t -> {
				       numOccurrences: int ref,
				       replacement: Var.t ref,
				       bind: PrimExp.t option ref
				       } option,
	      set = setVarInfo} =
	    Property.new(Var.plist, Property.initConst NONE)

	 fun simplifyVar x =
	    case varInfo x of
	       NONE => x
	     | SOME{replacement, ...} => !replacement

	 (*       val simplifyVar =
	  * 	 Trace.trace("simplifyVar", Var.layout, Var.layout) simplifyVar
	  *)

	 fun simplifyVars xs = List.map(xs, simplifyVar)

	 fun bind(x: Var.t): PrimExp.t option =
	    case varInfo x of
	       NONE => globalBind x
	     | SOME{bind, ...} => !bind

	 (*       val bind =
	  * 	 Trace.trace("bind", Var.layout, Option.layout PrimExp.layout) bind
	  *)

	 fun incNumOccurrences(x: Var.t, n: int): unit =
	    case varInfo x of
	       NONE => ()
	     | SOME{numOccurrences, ...} =>
		  let val new = n + !numOccurrences
		  in if new < 0
			then (Control.message
			      (fn () =>
			       let open Layout
			       in tuple[Var.layout x, Int.layout new]
			       end)
			      ; Error.bug "incNumOccurrences")
		     else ()
		     ; numOccurrences := new
		  end

	 (*       val incNumOccurrences =
	  * 	 Trace.trace2("incNumOccurrences", Var.layout, Int.layout, Unit.layout)
	  * 	 incNumOccurrences
	  *)

	 fun deleteVar x = incNumOccurrences(x, ~1)

	 (*       val deleteVar = Trace.trace("deleteVar", Var.layout, Unit.layout) deleteVar
	  *)
	 fun deleteVars xs = List.foreach(xs, deleteVar)

	 local
	    fun doit(x: Var.t, f) =
			case varInfo x of
			   NONE => Error.bug "attempt to use varInfo of free variable"
			 | SOME r => f r
	 in
	    fun setReplacement(x: Var.t, y: Var.t): unit =
	       doit(x, fn {replacement, bind = b, numOccurrences} =>
		    (replacement := y
		     ; b := bind y
		     ; incNumOccurrences(y, !numOccurrences)))

	    fun setBind(x: Var.t, e: PrimExp.t): unit =
	       doit(x, fn {bind, ...} => bind := SOME e)
	       
	    fun isUseless(x: Var.t): bool =
	       doit(x, fn {numOccurrences, ...} =>
		    mayDelete andalso !numOccurrences = 0)
	 end

	 (*       val isUseless = Trace.trace("isUseless", Var.layout, Bool.layout) isUseless
	  *)

	 (* ---------------------------------- *)
	 (*               jumps                *)
	 (* ---------------------------------- *)

	 datatype jumpInfo =
	    Unknown
	   | Eta of Jump.t
	   | Useful of  {
			 numOccurrences: int ref,
			 numBodyOccurrences: int ref,
			 inlined: bool ref,
			 params: Var.t list,
			 body: Exp.t,
			 isTail: bool
			 }
	     
	 val {get = jumpInfo: Jump.t -> jumpInfo, set = setJumpInfo} =
	    Property.new(Jump.plist, Property.initConst Unknown)

	 fun replaceJump j =
	    case jumpInfo j of
	       Eta j' => replaceJump j'
	     | _ => j

	 fun deleteJumpN(j, n) =
	    let
	       fun loop j =
		  case jumpInfo j of
		     Eta j' => loop j'
		   | Unknown => ()
		   | Useful{numOccurrences, ...} =>
			numOccurrences := !numOccurrences - n
	    in loop j
	    end

	 fun deleteJump j = deleteJumpN(j, 1)

	 (* ---------------------------------- *)
	 (*          occurence counts          *)
	 (* ---------------------------------- *)

	 fun walkExp{
		     exp: Exp.t,
		     delta: int,
		     bind: Var.t -> unit,
		     walkFun: {name: Jump.t,
			       args: (Var.t * Type.t) list,
			       body: Exp.t} -> unit -> unit
		     } =
	    let
	       fun var x =
		  case varInfo x of
		     NONE => ()
		   | SOME{replacement, ...} => incNumOccurrences(!replacement, delta)
	       fun vars xs = List.foreach(xs, var)
	       fun jump j =
		  case jumpInfo j of
		     Unknown => ()
		   | Eta j' => jump j'
		   | Useful{numOccurrences, ...} =>
			numOccurrences := delta + !numOccurrences
	       val loopPrimExp =
		  fn Const _ => ()
		   | Var x => var x
		   | Tuple xs => vars xs
		   | Select{tuple, ...} => var tuple
		   | ConSelect{variant, ...} => var variant
		   | ConApp{args, ...} => vars args
		   | PrimApp{args, ...} => vars args
	       val loopTransfer =
		  fn Call{args, cont, ...} =>
		       (vars args; Option.map'(cont, jump))
		   | Jump{dst, args, ...} => (jump dst; vars args)
		   | Return xs => vars xs
		   | Case{test, cases, default, ...} =>
			(var test
			 ; Option.map'(default, jump)
			 ; List.foreach(cases, jump o #2))
		   | Raise xs => vars xs
		   | Halt => ()
		   | Bug => ()
	       fun loopDec d =
		  case d of
		     Bind{var, exp, ...} => (bind var; loopPrimExp exp)
		   | Fun(f as {args, body, ...}) =>
			let val walkFun = walkFun f
			in List.foreach(args, bind o #1)
			   ; loopExp body
			   ; walkFun()
			end
		   | HandlerPush h => jump h
		   | HandlerPop => ()
	       and loopExp e =
		  let val {decs, transfer} = Exp.dest e
		  in List.foreach(decs, loopDec)
		     ; loopTransfer transfer
		  end
	    in loopExp exp
	    end

	 (* Compute occurrence counts *)
	 val _ =
	    let
	       fun bind x = setVarInfo(x, SOME{replacement = ref x,
					       numOccurrences = ref 0,
					       bind = ref NONE})

	       fun walkFun{name, args, body} =
		  let
		     fun set info = setJumpInfo(name, info)
		     fun normal() =
			let val params = List.map(args, fn (x, _) => x)
			   val numOccurrences = ref 0
			   val numBodyOccurrences = ref 0
			in set(Useful
			       {numOccurrences = numOccurrences,
				numBodyOccurrences = numBodyOccurrences,
				inlined = ref false,
				params = params,
				body = body,
				isTail = (case Exp.dest body of
					     {decs = [], transfer = Return xs} =>
						List.equals(params, xs, Var.equals)
					   | _ => false)})
			   ; fn () => numBodyOccurrences := !numOccurrences
			end
		  in case Exp.dest body of
		     {decs = [], transfer = Jump{dst, args = args'}} =>
			if List.equals(args, args', fn ((x, _), x') =>
				       Var.equals(x, x'))
			   andalso not(Jump.equals(dst, name))
			   then (set(Eta(replaceJump dst))
				 ; fn () => ())
			else normal()
		      | _ => normal()
		  end
		  
	    in walkExp{exp = exp,
		       delta = 1,
		       bind = bind,
		       walkFun = walkFun}
	    end

	 fun deleteExp(e: Exp.t): unit =
	    let fun ignore _ = ()
	    in walkExp{exp = e, delta = ~1, bind = ignore,
		       walkFun = fn _ => ignore}
	    end

	 (* ---------------------------------- *)
	 (*               shrink               *)
	 (* ---------------------------------- *)

	 fun jump{dst, args}: Exp.t =
	    let
	       fun normal args =
		  Exp.make{decs = [],
			   transfer = Jump{dst = dst, args = args}}
	    in case jumpInfo dst of
	       Unknown => normal args
	     | Eta j => jump{dst = j, args = args}
	     | Useful{inlined, numOccurrences, params, body, ...} =>
		  if !numOccurrences = 1
		     then (numOccurrences := 0
			   ; inlined := true
			   ;
			   (* The special case for [] is here because a case may
			    * get turned into a jump if the test is known, and
			    * jumps for case branches are allowed to ignore their
			    * argument.
			    *)
			   case params of
			      [] => ()
			    | _ => List.foreach2(params, args, setReplacement)
			   ; deleteVars args
			   ; simplifyExp body)
		  else (case params of
			   [] => normal []
			 | _ => normal args)
	    end
	 
	 and simplifyTransfer arg : Exp.t =
(*	    traceSimplifyTransfer *)
	    (fn (t: Transfer.t) =>
	     let fun trans t = Exp.make{decs = [], transfer = t}
	     in case t of
		Call{func, args, cont} =>
		   trans(Call{func = func, args = simplifyVars args,
			      cont = (case cont of
					 NONE => NONE
				       | SOME c =>
					    (case jumpInfo c of
						Useful{isTail = true, ...} => NONE
					      | Eta c => SOME c
					      | _ => SOME c))})
	      | Jump{dst, args} => jump{dst = dst, args = simplifyVars args}
	      | Return xs => trans(Return(simplifyVars xs))
	      | Raise xs => trans(Raise(simplifyVars xs))
	      | Halt => trans Halt
	      | Bug => trans Bug
	      | Case{cause, test, cases, default} =>
		   let val test = simplifyVar test
		      val cases =
			 List.revMap(cases, fn (c, j) => (c, replaceJump j))
		      val default = Option.map(default, replaceJump)
		   in case (cases, default) of
		      ([], NONE) => (deleteVar test; trans Bug)
		    | ([], SOME j) => (deleteVar test; jump{dst = j, args = []})
		    | ((_, j) :: cases', _) =>
			 case bind test of
			    NONE =>
			       let val numOccs = ref 0
				  fun equals j' =
				     (IntRef.inc numOccs
				      ; Jump.equals(j, j'))
			       in if List.forall(cases', fn (_, j') =>
						 equals j')
				  andalso (case default of
					      NONE => true
					    | SOME j' => equals j')
				  andalso !numOccs >= 1
				     then (deleteVar test
					   ; deleteJumpN(j, !numOccs)
					   ; jump{dst = j, args = []})
				  else trans(Case{cause = cause,
						  test = test,
						  cases = cases,
						  default = default})
			       end
			 | SOME(ConApp{con, ...}) =>
			      let
				 val rec loop =
				    fn [] =>
				    (case default of
					NONE => trans Bug
				      | SOME dst =>
					   (deleteVar test
					    ; jump{dst = dst, args = []}))
				     | (c, j) :: cases => 
					  if Con.equals(con, c)
					     then (List.foreach
						   (cases, deleteJump o #2)
						   ; (case default of
							 NONE => ()
						       | SOME j => deleteJump j)
						   ; jump{dst = j, args = [test]})
					  else (deleteJump j; loop cases)
			      in loop cases
			      end
			  | _ => Error.bug "strange bind for case test"
		   end
	     end) arg

	 and simplifyExp arg : Exp.t =
(*	    traceSimplifyExp *)
	    (fn (e: Exp.t) => 
	     let val {decs, transfer} = Exp.dest e
		fun simplifyDecs(decs: Dec.t list): Exp.t =
		   case decs of
		      [] => simplifyTransfer transfer
		    | dec :: decs =>
			 let fun keep d = Exp.prefix(simplifyDecs decs, d)
			 in case dec of
			    Fun{name, args, body} =>
			       let
				  fun filter() =
				     case jumpInfo name of
					Unknown => Error.bug "missing jumpInfo"
				      | Eta j => (deleteJump j; true)
				      | Useful{numBodyOccurrences,
					       numOccurrences, inlined, ...} =>
					   if !numOccurrences = !numBodyOccurrences
					      then (if !inlined
						       then ()
						    else deleteExp body
						    ; true)
					   else false
				  val isGone = filter()
				  val rest = simplifyDecs decs
			       in if isGone orelse filter()
				     then rest
				  else
				     Exp.prefix(rest, Fun{name = name, args = args,
							  body = simplifyExp body})
			       end
			  | Bind{var, ty, exp} =>
			       let
				  fun finish(exp, rest): Exp.t =
				     Exp.prefix(rest,
						Bind{var = var, ty = ty, exp = exp})
				  fun nonExpansive
				     (delete: unit -> unit,
				      set: unit -> (unit -> PrimExp.t) option): Exp.t =
				     if isUseless var
					then (delete(); simplifyDecs decs)
				     else let val s = set()
					      val rest = simplifyDecs decs
					  in if isUseless var
						then (delete(); rest)
					     else (case s of
						      NONE => rest
						    | SOME e => finish(e(), rest))
					  end
				  fun nonExpansiveCon(delete, exp: PrimExp.t) =
				     nonExpansive(delete,
						  fn () => (setBind(var, exp)
							    ; SOME(fn () => exp)))
			       in case exp of
				  Const _ => nonExpansive(fn () => (),
							  fn () => SOME(fn () => exp))
				| Var x =>
				     let val x = simplifyVar x
				     in setReplacement(var, x)
					; deleteVar x
					; simplifyDecs decs
				     end
				| Tuple xs =>
				     let val xs = simplifyVars xs
				     in nonExpansiveCon(fn () => deleteVars xs,
							Tuple xs)
				     end
				| ConApp{con, args} =>
				     let val args = simplifyVars args
				     in nonExpansiveCon
					(fn () => deleteVars args,
					 ConApp{con = con, args = args})
				     end
				| Select{tuple, offset} =>
				     let val tuple = simplifyVar tuple
				     in nonExpansive
					(fn () => deleteVar tuple,
					 fn () =>
					 case bind tuple of
					    NONE => SOME(fn () =>
							 Select{tuple = tuple,
								offset = offset})
					  | SOME(Tuple xs) =>
					       (deleteVar tuple
						; (setReplacement
						   (var, List.nth(xs, offset)))
						; NONE)
					  | _ => Error.bug "select of non-Tuple")
				     end
				| ConSelect{variant, con, offset} =>
				     let val variant = simplifyVar variant
				     in nonExpansive
					(fn () => deleteVar variant,
					 fn () =>
					 case bind variant of
					    NONE => SOME(fn () =>
							 ConSelect{variant = variant,
								   con = con,
								   offset = offset})
					  | SOME(ConApp{con, args}) =>
					       (deleteVar variant
						; (setReplacement
						   (var, List.nth(args, offset)))
						; NONE)
					  | _ => Error.bug "conSelect of non-ConApp")
				     end
				| PrimApp{prim, targs, args} =>
				     let val args = simplifyVars args
					val e =
					   PrimApp{prim = prim, targs = targs, args = args}
				     in if Prim.maySideEffect prim
					   then finish(e, simplifyDecs decs)
					else nonExpansive(fn () => deleteVars args,
							  fn () => SOME(fn () => e))
				     end
			       end
			  | _ => keep dec
			 end
	     in simplifyDecs decs
	     end) arg

	 val exp = simplifyExp exp

      in Exp.clear exp
	 ; exp
      end
   end

val shrinkExpNoDelete = fn e => shrinkExp [] (e, false)

(*val traceShrinkExp = Trace.trace("shrinkExp", Exp.layout, Exp.layout) *)

val shrinkExp = fn globals => let val shrinkExp = shrinkExp globals
			      in (*traceShrinkExp*)(fn e => shrinkExp(e, true))
			      end
   
fun simplifyProgram simplifyExp (Program.T{datatypes, globals, functions, main}) =
   let
      val shrinkExp = shrinkExp globals
      val functions =
	 List.revMap
	 (functions, fn {name, args, body, returns} =>
	  {name = name, args = args,
	   body = shrinkExp(simplifyExp body),
	   returns = returns})

   in Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

fun shrink p = simplifyProgram (fn x => x) p

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor InferHandlers(S: INFER_HANDLERS_STRUCTS): INFER_HANDLERS = 
struct

open S

structure Set = DisjointSet
      
structure Handler' =
   struct
      datatype t = T of Jump.t option Set.t

      val new = T o Set.singleton o SOME
      fun unknown() = T(Set.singleton NONE)

      fun equals(T s, T s'): bool =
	 Set.canUnion
	 (s, s',
	  fn (NONE, h) => SOME h
	   | (h, NONE) => SOME h
	   | (SOME h, SOME h') =>
		if Jump.equals(h, h') then SOME(SOME h) else NONE)

      fun dest(T s) =
	 case Set.value s of
	    NONE => Error.bug "unknown handler"
	  | SOME h => h

      fun layout(T s) =
	 Option.layout Jump.layout (Set.value s)
   end

structure Handlers =
   struct
      datatype t = T of handlers Set.t
      and handlers =
	 Unknown
	| Empty
	| Push of {top: Handler'.t,
		   rest: t}

      val new = T o Set.singleton
	  
      val empty = new Empty
      fun unknown() = new Unknown
      fun push(rest, top) = new(Push{top = top,
				     rest = rest})

      fun pop(T s) =
	 case Set.value s of
	    Empty => Error.bug "pop of empty handler stack"
	  | Push{rest, ...} => rest
	  | Unknown =>
	       let val rest = unknown()
	       in Set.setValue(s, Push{top = Handler'.unknown(),
				       rest = rest})
		  ; rest
	       end

      fun isEmpty(T s) =
	 case Set.value s of
	    Empty => true
	  | _ => false

      fun toList hs =
	 let
	    fun loop(T s, ac) =
	       case Set.value s of
		  Unknown => Error.bug "toList of unknown handler stack"
		| Empty => rev ac
		| Push{top, rest} => loop(rest, Handler'.dest top :: ac)
	 in loop(hs, [])
	 end

      fun equals(T s, T s'): bool =
	 Set.canUnion
	 (s, s',
	  fn (Unknown, hs) => SOME hs
	   | (hs, Unknown) => SOME hs
	   | (Empty, Empty) => SOME Empty
	   | (hs as Push{top = t, rest = r}, Push{top = t', rest = r'}) =>
		if Handler'.equals(t, t') andalso equals(r, r')
		   then SOME hs
		else NONE
	   | _ => NONE)
		     
      fun layout(T s) =
	 let open Layout
	 in case Set.value s of
	    Unknown => str "Unknown"
	  | Empty => str "Empty"
	  | Push{top, rest} => seq[Handler'.layout top,
				   str " :: ",
				   layout rest]
	 end
   end

open Dec PrimExp Transfer

fun inferHandlers(Program.T{functions, ...}) =
   let
      val {get = jump, set = setJump} =
	 Property.new(Jump.plist, Property.initRaise("handlers", Jump.layout))
   in
      List.foreach
      (functions, fn {body, ...} =>
       let
	  fun loop(e: Exp.t, hs: Handlers.t): unit =
	     let
		fun check(msg, hs, hs') =
		   if Handlers.equals(hs, hs')
		      then ()
		   else (Control.message
			 (fn () => let open Layout
				   in align[seq[str "exp: ", Exp.layout e],
					    seq[str "hs: ", Handlers.layout hs],
					    seq[str "hs': ", Handlers.layout hs']]
				   end)
			 ; Error.bug("handler stack mismatch at " ^ msg))

		val {decs, transfer} = Exp.dest e
		val after =
		   List.fold
		   (decs, hs, fn (d, hs) =>
		    case d of
		       Fun{name, body, ...} =>
			  (let val hs = Handlers.unknown()
			   in setJump(name, hs)
			   end
			   ; loop(body, jump name)
			   ; hs)
		     | HandlerPush h =>
			  let val hs = Handlers.push(hs, Handler'.new h)
			  in check("push", hs, jump h)
			     ; hs
			  end
		     | HandlerPop => Handlers.pop hs
		     | _ => hs)
		fun checkJump j = check("jump", after, jump j)
		fun empty msg = check(msg, after, Handlers.empty)
	     in case transfer of
		Jump{dst, ...} => checkJump dst
	      | Case{cases, default, ...} =>
		   (List.foreach(cases, checkJump o #2)
		    ; Option.map'(default, checkJump))
	      | Halt => empty "halt"
	      | Return _ => empty "return"
	      | Bug => ()
	      | Raise _ => ()
	      | Call{cont = c, ...} =>
		   (case c of
		       NONE => empty "tail call"
		     | SOME c => check("nontail call", after, jump c))
	     end
       in loop(body, Handlers.empty)
       end)

(*      Trace.trace("jumpHandlers", Jump.layout, List.layout Handler.layout) *)
      ; (Handlers.toList o jump)
   end

val inferHandlers = Control.traceCall "inferHandlers" inferHandlers
   
fun deltaHandlers(d, hs) =
   case d of
      HandlerPush h => h :: hs
    | HandlerPop => (case hs of
			_ :: hs => hs
		      | _ => Error.bug "deltaHandlers")
    | _ => hs

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CPS_STRUCTS = 
   sig
      include CPS_TREE_STRUCTS
   end

signature CPS = 
   sig
      include SIMPLIFY
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Analyze(S: ANALYZE_STRUCTS): ANALYZE = 
struct

open S
open Dec PrimExp Transfer

fun 'a analyze
   {program = Program.T{main, datatypes, globals, functions},
    useFromTypeOnBinds,
    fromType, copy, coerce,
    const,
    tuple, select,
    constructors,
    primApp,
    filter} =
   let
      fun coerces(from, to) =
	 List.foreach2(from, to, fn (from, to) =>
		       coerce{from = from, to = to})

      val {conApp, conSelect, conArgs} =
	 case constructors of
	    SOME{conApp, conSelect} =>
	       {conApp = conApp,
		conSelect = conSelect,
		conArgs = fn _ => Error.bug "conArgs not defined"}
	  | NONE =>
	       let
		  val {get = conInfo, set = setConInfo} =
		     Property.new
		     (Con.plist, Property.initRaise("info", Con.layout))
		  val _ =
		     List.foreach
		     (datatypes, fn {tycon, cons} =>
		      let val v = fromType(Type.con(tycon, []))
		      in List.foreach
			 (cons, fn {con, args} =>
			  setConInfo(con, {value = v,
					   argTypes = args,
					   args = List.map(args, fromType)}))
		      end)
		  fun conApp{con, args} =
		     let val {args = args', value, ...} = conInfo con
		     in coerces(args, args');
			value
		     end
		  fun conArgs c = #args(conInfo c)
		  fun conSelect{variant, con, offset} =
		     List.nth(conArgs con, offset)
	       in {conApp = conApp,
		   conSelect = conSelect,
		   conArgs = conArgs}
	       end

      val boolVal = fromType Type.bool

      val {get = value: Var.t -> 'a, set = setValue} =
	 Property.new
	 (Var.plist, Property.initRaise("analyze var value", Var.layout))

      fun values xs = List.map(xs, value)

      val {get = func, set = setFunc} =
	 Property.new(Func.plist,
		      Property.initRaise("analyze func name", Func.layout))

      val {get = jump, set = setJump} =
	 Property.new(Jump.plist, Property.initRaise("analyze jump", Jump.layout))

      fun loopArgs args =
	 List.map(args, fn (x, t) =>
		  let val v = fromType t
		  in setValue(x, v);
		     v
		  end)

      val _ =
	 List.foreach
	 (functions, fn {name, args, returns, ...} =>
	  setFunc(name, {args = loopArgs args,
			 returns = List.map(returns, fromType)}))

      val exnVals: 'a list option ref = ref NONE

      fun getExnVals vs =
	 case !exnVals of
	    NONE => let val vs = List.map(vs, copy)
		    in exnVals := SOME vs; vs
		    end
	  | SOME vs => vs

      fun loopTransfer(t, shouldReturns): unit =
	 case t of
	    Call{func = f, args, cont = c} =>
	       let val {args = formals, returns} = func f
		  val shouldReturns =
		     case c of
			NONE => shouldReturns
		      | SOME c => jump c
	       in coerces(values args, formals);
		  coerces(returns, shouldReturns)
	       end
	  | Jump{dst, args} => coerces(values args, jump dst)
	  | Return xs => coerces(values xs, shouldReturns)
	  | Case{test, cases, default, ...} =>
	       let val test = value test 
	       in List.foreach
		  (cases, fn (c, j) =>
		   let val test = filter(test, c)
		   in case jump j of
		      [] => ()
		    | [v] => coerce{from = test, to = v}
		    | _ => Error.bug "case label must be nullary or unary"
		   end)
		  ; (case default of
			NONE => ()
		      | SOME j =>
			   case jump j of
			      [] => ()
			    | _ => Error.error "default label must be nullary")
	       end
	  | Raise xs => let val vs = values xs
			in coerces(vs, getExnVals vs)
			end
	  | Halt => ()
	  | Bug => ()

(*       val loopTransfer =
 * 	 Trace.trace2("Analyze.loopTransfer",
 * 		      Transfer.layout, Layout.ignore, Layout.ignore) loopTransfer
 *)

      fun loopPrimExp{var: Var.t, ty: Type.t, exp: PrimExp.t}: 'a =
	 case exp of
	    Const c => const c
	  | Var x => value x
	  | Tuple[_] => Error.bug "unary tuple"
	  | Tuple xs => tuple(values xs)
	  | Select{tuple, offset} =>
	       select{tuple = value tuple,
		      offset = offset,
		      resultType = ty}
	  | ConSelect{variant, con, offset} =>
	       conSelect{variant = value variant,
			 con = con,
			 offset = offset}
	  | ConApp{con, args} => conApp{con = con, args = values args}
	  | PrimApp{prim, targs, args} =>
	       primApp{prim = prim,
		       targs = targs,
		       args = values args,
		       resultVar = var,
		       resultType = ty}
      
(*       val loopPrimExp =
 * 	 Trace.trace("Analyze.loopPrimExp",
 * 		     fn {var, ty, exp, ...} =>
 * 		     Layout.record[("var", Var.layout var),
 * 				   ("ty", Type.layout ty),
 * 				   ("exp", PrimExp.layout exp)],
 * 		     Layout.ignore)
 * 	 loopPrimExp
 *)

      fun loopBind{var, exp, ty}: unit =
	 setValue(var,
		  let val v = loopPrimExp{var = var, ty = ty, exp = exp}
		  in if useFromTypeOnBinds
			then let val v' = fromType ty
			     in coerce{from = v, to = v'};
				v'
			     end
		     else v
		  end)
	       
      fun loopExp(e: Exp.t, returns): unit =
	 let val {decs, transfer, ...} = Exp.dest e
	 in loopDecs(decs, returns);
	    loopTransfer(transfer, returns)
	 end
      and loopDecs(ds, returns): unit =
	 List.foreach
	 (ds,
	  fn Bind b => loopBind b
	   | Fun{name, args, body} =>
		(setJump(name, loopArgs args)
		 ; loopExp(body, returns))
	   | HandlerPush h => let val vs = jump h
			      in coerces(getExnVals vs, vs)
			      end
	   | _ => ()
		)

      val _ = coerces([], #args(func main))
   in
      List.foreach(globals, loopBind);
      List.foreach(functions, fn {name, body, ...} =>
		   loopExp(body, #returns(func name)));
      
      {
       value = value,
       conArgs = conArgs,
       func = func,
       jump = jump,
       exnVals = !exnVals
      }
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor HashType(S: HASH_TYPE_STRUCTS): HASH_TYPE = 
struct

open S

structure Type =
   struct
      datatype t =
	 T of {
	       hash: word,
	       plist: PropertyList.t,
	       tree: tree
	      }
      and tree =
	 Var of Tyvar.t
	| Con of Tycon.t * t list
	| Variant of Tycon.t * Con.t

      fun hash(T{hash, ...}) = hash
      fun plist(T{plist, ...}) = plist
      fun tree(T{tree, ...}) = tree

      structure Dest =
	 struct
	    datatype dest = datatype tree
	    val dest = tree
	 end
      open Dest

      fun deconOpt t =
	 case dest t of
	    Con x => SOME x
	  | _ => NONE

      fun makeHomVariant{var, variant, con} =
	 let
	    val {get, destroy, ...} =
	       Property.newDest
	       (plist,
		Property.initRec
		(fn (t, get) =>
		 case dest t of
		    Var a => var(t, a)
		  | Con(c, ts) => con(t, c, List.map(ts, get))
		  | Variant x => variant x))
	 in {hom = get, destroy = destroy}
	 end

      fun makeHom{var, con} =
	 makeHomVariant{var = var,
			con = con,
			variant = fn _ => Error.bug "makeMonoHom saw variant"}

      fun hom{ty, var, con} =
	 let val {hom, destroy} = makeHom{var = var o #2,
					  con = fn (_, c, xs) => con(c, xs)}
	 in hom ty before destroy()
	 end

      fun makeMonoHom{con} =
	 makeHom{var = fn _ => Error.bug "makeMonoHom saw type variable",
		 con = con}

      fun monoHom{ty, con} =
	 let val {hom, destroy} = makeMonoHom{con = con}
	 in hom ty before destroy()
	 end

      local
	 structure Atype = Ast.Type
      in
	 fun toAst t =
	    hom{ty = t,
		var = Atype.var,
		con = fn (c, ts) =>
		if Tycon.equals(c, Tycon.tuple) then Atype.tuple ts
		else Atype.con(Tycon.toAst c, ts)}
      end
			 
      val layout = Ast.Type.layout o toAst

(*      val hash = Trace.trace("hash", layout, Layout.str o Word.toString) hash *)
(*	 let open Layout
	 in
	    case tree of
	       Var a => Tyvar.layout a
	     | Con(c, ts) => seq[Tycon.layout c, tuple(List.map(ts, layout))]
	 end *)

      fun newHash() = Word.fromInt(Int.random())

      fun equals(t, t'): bool = PropertyList.equals(plist t, plist t')

(*      val equals = Trace.trace2("Type.equals", layout, layout, Bool.layout) equals *)

      structure Table = HashTable

      local
	 val size = 10000
	 exception NotFound

	 val same: t * t -> bool =
	    fn (T{tree, ...}, T{tree = t, ...}) =>
	    case (tree, t) of
	       (Var a, Var a') => Tyvar.equals(a, a')
	     | (Con(c, ts), Con(c', ts')) =>
		  Tycon.equals(c, c')
		  andalso if Tycon.equals(c, Tycon.tuple)
			     then List.equals(ts, ts', equals)
			  else List.forall2(ts, ts', equals)
	     | (Variant(t, c), Variant(t', c')) =>
		  Tycon.equals(t, t') andalso Con.equals(c, c')
	     | _ => false

(*	 val same = Trace.trace2("same", layout, layout, Bool.layout) same *)
      in
	 val table: (t, t) Table.hash_table =
	    Table.mkTable (hash, same) (size, NotFound)
      end

      local
	 val next = ref 1
      in
	 fun msg() =
	    if Table.numItems table = !next
	       then (next := !next * 2
		     ; Control.messageLay(List.layout Int.layout
					  (List.keepAll
					   (Table.bucketSizes table,
					    fn n => n <> 0))))
					   
	    else ()
      end

      fun lookup t =
	 ((*msg()
	  ; *)(case Table.find table t of
		NONE => (Table.insert table (t, t); t)
	      | SOME t => t))

      fun var a =
	 lookup(T{hash = Tyvar.hash a,
		  plist = PropertyList.new(),
		  tree = Var a})

      val generator = 0wx5555
      fun con(c, ts) =
	 lookup(T{hash = List.fold(ts, Tycon.hash c,
				   fn (t, w) => Word.xorb(w * generator, hash t)),
		  plist = PropertyList.new(),
		  tree = Con(c, ts)})

      fun variant(tycon, con) =
	 lookup(T{hash = Word.xorb(Tycon.hash tycon, Con.hash con),
		  plist = PropertyList.new(),
		  tree = Variant(tycon, con)})

(*       val variant =
 * 	 Trace.trace2("variant", Tycon.layout, Con.layout,
 * 		      Layout.str o Word.toString o hash) variant
 *)
   end
structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
open Type Ops

local structure Type = Ast.Type
in
   fun toAst(t: t): Type.t =
      case dest t of
	 Var a => Type.var a
       | Con(c, ts) =>
	    let val ts = List.map(ts, toAst)
	    in if Tycon.equals(c, Tycon.tuple) then Type.tuple ts
	       else Type.con(Tycon.toAst c, ts)
	    end
       | _ => Error.bug "Type.toAst"
end

fun optionToAst z = Option.map(z, toAst)

fun stats() =
   let open Layout
   in align[seq[str "num types in table = ", Int.layout(Table.numItems table)],
	    Control.sizeMessage("hash table", table)]
   end

fun ofConst c = con(Const.tycon c, [])

fun isUnit t =
   case dest t of
      Con(c, []) => Tycon.equals(c, Tycon.tuple)
    | _ => false

structure Env = MonoEnv(structure Domain = Tyvar
			structure Range = Type)
   
fun substitute(ty, e) =
   case e of
      [] => ty (* This optimization is important so that monotypes
		* are not substituted inside of.
		*)
    | _ => let val env = Env.fromList e
	   in hom{ty = ty,
		  var = fn a => (case Env.peek(env, a) of
				    NONE => var a
				  | SOME ty => ty),
		  con = con}
	   end

(* val substitute =
 *    Trace.trace2("substitute", layout,
 * 		List.layout(Layout.tuple2(Tyvar.layout, Type.layout)),
 * 		layout) substitute		
 *)

fun equalss(ts: t list): t option =
   case ts of
      t :: ts =>
	 let fun loop[] = SOME t
	       | loop(t' :: ts) = if equals(t, t') then loop ts else NONE
	 in loop ts
	 end
    | [] => Error.error "equals"

local
   val out = Out.error
   val print = Out.outputc out
   exception TypeError
in
   fun error(msg, lay) =
      (print("Type error: " ^ msg ^ "\n") ;
       Layout.output(lay, out) ;
       print "\n" ;
       raise TypeError)
end

local
   structure Ptype = Prim.Type
in
   fun fromPrims ts = List.map(ts, fromPrim)
   and fromPrim t =
      case t of
	 Ptype.Var a => var a
       | Ptype.Con(c, ts) => con(c, fromPrims ts)
       | Ptype.Record r => con(Tycon.tuple, fromPrims(SortedRecord.range r))

   fun toPrim t = hom{ty = t,
		      var = Ptype.var,
		      con = Ptype.con}
end

fun tycon t =
   case dest t of
      Con(c, _) => c
    | _ => Error.bug "Type.tycon saw type variable"

fun containsTycon(ty, tycon) =
   hom{ty = ty,
       var = fn _ => false,
       con = fn (tycon', bs) => (Tycon.equals(tycon, tycon')
				 orelse List.exists(bs, fn b => b))}

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor CpsTree(S: CPS_TREE_STRUCTS): CPS_TREE = 
struct

open S

structure Type =
   struct
      local structure T = HashType(S)
      in open  T
      end

      fun tyconArgs t =
	 case Dest.dest t of
	    Dest.Con x => x
	  | _ => Error.bug "FirstOrderType.tyconArgs"
	       
      datatype dest =
	 Char
	| Int
	| IntInf
	| Pointer
	| Word
	| Word8
	| Real
	| Stack
	| String 
	| Array of t
	| Vector of t
	| Ref of t
	| Datatype of Tycon.t
	| Tuple of t list
	| Variant of Tycon.t * Con.t

      local
	 val {get, set} = Property.new(Tycon.plist, Property.initConst NONE)

	 fun nullary c =
	    fn [] => c
	     | _ => Error.bug "bogus application of nullary tycon"
		  
	 fun unary make =
	    fn [t] => make t
	     | _ => Error.bug "bogus application of unary tycon"

      in
	 fun initialize() =
	    List.foreach
	    ([(Tycon.tuple, Tuple),
	      (Tycon.char, nullary Char),
	      (Tycon.int, nullary Int),
	      (Tycon.intInf, nullary IntInf),
	      (Tycon.pointer, nullary Pointer),
	      (Tycon.real, nullary Real),
	      (Tycon.stack, nullary Stack),
	      (Tycon.string, nullary String),
	      (Tycon.word8, nullary Word8),
	      (Tycon.word, nullary Word),
	      (Tycon.array, unary Array),
	      (Tycon.vector, unary Vector),
	      (Tycon.reff, unary Ref)],
	     fn (tycon, f) => set(tycon, SOME f))

	 val _ = initialize()

	 fun dest t =
	    case Dest.dest t of
	       Dest.Con(tycon, ts) =>
		  (case get tycon of
		      NONE => Datatype tycon
		    | SOME f => f ts)
	     | Dest.Variant z => Variant z
	     | _ => Error.bug "dest"
      end

      local open Layout
      in
	 fun layout t =
	    case dest t of
	       Char => str "char"
	     | Int => str "int"
	     | IntInf => str "IntInf.int"
	     | Pointer => str "pointer"
	     | Word => str "word"
	     | Word8 => str "word8"
	     | Real => str "real"
	     | Stack => str "stack"
	     | String => str "string"
	     | Array t => seq[layout t, str " array"]
	     | Vector t => seq[layout t, str " vector"]
	     | Ref t => seq[layout t, str " ref"]
	     | Datatype t => Tycon.layout t
	     | Tuple [] => str "unit"
	     | Tuple ts => paren(seq(separate(List.map(ts, layout), " * ")))
	     | Variant(tycon, con) => seq[Tycon.layout tycon,
					  str "@",
					  Con.layout con]
      end
   end

structure Func =
   struct
      open Var (* Id(structure AstId = Ast.Var) *)

      fun newNoname() = newString "F"
      val saveWorld = newString "saveWorld"
   end

structure Jump =
   struct
      open Func
      fun newNoname() = newString "L"
   end

local open Layout
in
   fun layoutTuple xs = tuple(List.map(xs, Var.layout))
end

structure PrimExp =
   struct
      datatype t =
	 Const of Const.t
       | Var of Var.t
       | Tuple of Var.t list
       | Select of {tuple: Var.t,
		    offset: int}
       | ConSelect of {variant: Var.t,
		       con: Con.t,
		       offset: int}
       | ConApp of {con: Con.t,
		    args: Var.t list}
       | PrimApp of {prim: Prim.t,
		     targs: Type.t list,
		     args: Var.t list}

      fun foreachVar(e, f) =
	 let fun fs xs = List.foreach(xs, f)
	 in case e of
	    Const _ => ()
	  | Var x => f x
	  | Tuple xs => fs xs
	  | Select{tuple, ...} => f tuple
	  | ConSelect{variant, ...} => f variant
	  | ConApp{args, ...} => fs args
	  | PrimApp{args, ...} => fs args
	 end
      
      fun replaceVar(e, f) =
	 let fun fs xs = List.map(xs, f)
	 in case e of
	    Const _ => e
	  | Var x => Var(f x)
	  | Tuple xs => Tuple(fs xs)
	  | Select{tuple, offset} => Select{tuple = f tuple, offset = offset}
	  | ConSelect{variant, con, offset} =>
	       ConSelect{variant = f variant, con = con, offset = offset}
	  | ConApp{con, args} => ConApp{con = con, args = fs args}
	  | PrimApp{prim, targs, args} =>
	       PrimApp{prim = prim, targs = targs, args = fs args}
	 end

      local open Layout
      in
	 val layout =
	    fn Const c => Const.layout c
	     | Var x => Var.layout x
	     | Tuple xs => layoutTuple xs
	     | Select{tuple, offset} =>
		  seq[str "#", Int.layout(offset + 1), str " ", Var.layout tuple]
	     | ConSelect{variant, con, offset} =>
		  seq[Con.layout con, str "#",
		      Int.layout(offset + 1), str " ", Var.layout variant]
	     | ConApp{con, args} => seq[Con.layout con, layoutTuple args]
	     | PrimApp{prim, targs, args} =>
		  seq[Prim.layout prim,
		      (case targs of
			  [] => empty
			| _ => list(List.map(targs, Type.layout))),
		      layoutTuple args]
      end
   end

local
   open Layout
in
   fun layoutBind{var, ty, exp} =
      seq[str "val ", Var.layout var,
	  str ": ", Type.layout ty,
	  str " = ", PrimExp.layout exp]
end

fun clearBind{var, ty, exp} = Var.clear var

structure Cause =
   struct
      datatype t = User | Dispatch | PolyEqual | Coerce
   end

structure Transfer =
   struct
      datatype t =
	 Call of {func: Func.t,
		  args: Var.t list,
		  cont: Jump.t option}
       | Jump of {dst: Jump.t,
		  args: Var.t list}
       | Return of Var.t list
       | Case of {cause: Cause.t,
		  test: Var.t,
		  cases: (Con.t * Jump.t) list,
		  default: Jump.t option}
       | Raise of Var.t list
       | Halt
       | Bug

      fun foreachVar(t, f) =
	 let fun fs xs = List.foreach(xs, f)
	 in case t of
	    Call{args, ...} => fs args
	  | Jump{args, ...} => fs args
	  | Return xs => fs xs
	  | Case{test, ...} => f test
	  | Raise xs => fs xs
	  | Halt => ()
	  | Bug => ()
	 end

      fun replaceVar(t, f) =
	 let fun fs xs = List.map(xs, f)
	 in case t of
	    Call{func, args, cont} =>
	       Call{func = func, args = fs args, cont = cont}
	  | Jump{dst, args} => Jump{dst = dst, args = fs args}
	  | Return xs => Return(fs xs)
	  | Case{cause, test, cases, default} =>
	       Case{cause = cause,
		    test = f test, cases = cases, default = default}
	  | Raise xs => Raise(fs xs)
	  | Halt => Halt
	  | Bug => Bug
	 end

      local open Layout
      in
	 val layout =
	    fn Call{func, args, cont} =>
	    seq[Func.layout func,
		layoutTuple args,
		case cont of
		   NONE => str " Tail"
		 | SOME l => seq[str " NonTail ", Jump.layout l]]
	    | Jump{dst, args} =>
		 seq[Jump.layout dst, layoutTuple args]
	    | Return xs => seq[str "Return ", layoutTuple xs]
	    | Case{test, cases, default, ...} =>
		 align[seq[str "case ", Var.layout test, str " of"],
		       indent(align[alignPrefix
				    ("| ",
				     List.map(cases, fn (c, l) =>
					      seq[Con.layout c,
						  str " => ",
						     Jump.layout l])),
				    case default of
				       NONE => empty
				     | SOME l => seq[str "_ => ",
						     Jump.layout l]],
			      2)]
	    | Raise xs => seq[str "raise ", list(List.map(xs, Var.layout))]
	    | Halt => str "Halt"
	    | Bug => str "Bug"
      end
		
   end


structure DecExp =
   struct
      datatype dec =
	 Bind of {
		  var: Var.t,
		  ty: Type.t,
		  exp: PrimExp.t
		  }
       | Fun of {
		 name: Jump.t,
		 args: (Var.t * Type.t) list,
		 body: exp
		}
       | HandlerPush of Jump.t
       | HandlerPop

      and exp = Exp of {decs: dec list,
			transfer: Transfer.t}

      local
	 open Layout
      in
	 val rec layoutDec =
	    fn Bind b => layoutBind b
	     | Fun{name, args, body} =>
		  seq[str "fun ",
		      align[seq[Jump.layout name,
				tuple(List.map(args, fn (x, t) =>
					       seq[Var.layout x,
						   str ": ",
						   Type.layout t])),
				str " = "],
			    layoutExp body]]
	     | HandlerPush h => seq[str "val _ = HandlerPush ", Jump.layout h]
	     | HandlerPop => str "val _ = HandlerPop"
	 and layoutExp =
	    fn Exp{decs, transfer, ...} =>
	    let val transfer = Transfer.layout transfer
	    in case decs of
	       [] => transfer
	     | _ => align[seq[str "let ", align(List.map(decs, layoutDec))],
			  seq[str "in ", transfer],
			  str "end"]
	    end
      end
   end

structure Dec =
   struct
      datatype t = datatype DecExp.dec

      type exp = DecExp.exp

      val layout = DecExp.layoutDec
   end

structure Exp =
   struct
      datatype t = datatype DecExp.exp
	       
      fun dest(Exp r) = r
      val make = Exp

      val decs = #decs o dest
      val transfer = #transfer o dest

      val layout = DecExp.layoutExp
					     
      local
	 fun make f (Exp{decs, transfer}, x): t =
	    Exp{decs = f(x, decs), transfer = transfer}
      in val prefix = make(op ::)
	 val prefixs = make(op @)
      end

      fun foreach(e, {handleDec, handleTransfer}) =
	 let
	    fun exp(Exp{decs, transfer, ...}) =
	       (List.foreach(decs, dec);
		handleTransfer transfer)
	    and dec d =
	       let val after = handleDec d
	       in (case d of
		      Dec.Fun{body, ...} => exp body
		    | _ => ());
		  after()
	       end
	 in exp e
	 end

      fun ignore _ = ()
      fun ignore' _ = ignore

      fun foreach'(e, {handleDec, handleTransfer}) =
	 foreach(e, {handleTransfer = handleTransfer,
		     handleDec = fn d => (handleDec d; ignore)})

      fun foreachDec(e, handleDec) =
	 foreach'(e, {handleTransfer = ignore,
		      handleDec = handleDec})

      fun foreachVar(e, f) =
	 foreach'
	 (e, {handleTransfer = ignore,
	      handleDec = fn Dec.Bind{var, ty, ...} => f(var, ty)
			   | Dec.Fun{args, ...} => List.foreach(args, f)
			   | _ => ()})

      fun foreachBind(e, f) =
	 let
	    fun handleDec d =
	       (case d of
		   Dec.Bind r => f r
		 | _ => ();
	        ignore)
	 in foreach(e, {handleTransfer = ignore,
			handleDec = handleDec})
	 end

      fun foreachTransfer(e, f) =
	 foreach(e, {handleDec = ignore', handleTransfer = f})

      fun foreachCall(e, f) =
	 foreachTransfer(e, fn Transfer.Call r => f r | _ => ())

      fun clear(e: t): unit =
	 let open DecExp
	 in foreach'
	    (e, {handleTransfer = fn _ => (),
		 handleDec =
		 fn Bind b => clearBind b
		  | Fun{name, args, ...} => (Jump.clear name;
					     List.foreach(args, Var.clear o #1))
		  | _ => ()})
	 end

      fun alphaRename{exp: t,
		      substitution: {formal: Var.t, actual: Var.t} list,
		      handleJump: {old: Jump.t, new: Jump.t} -> unit} =
	 let
	    local
	       fun make(new, plist, layout) =
		  let
		     val {get, set, destroy} =
			Property.newDest(plist, Property.initConst NONE)
		     fun bind x = let val x' = new x
				  in set(x, SOME x'); x'
				  end
		     fun lookup x =
			case get x of
			   NONE => x
			 | SOME y => y
		  in (set, bind, lookup, destroy)
		  end
	    in
	       val (setVar, bindVar, lookupVar, destroyVar) =
		  make(Var.new, Var.plist, Var.layout)
	       val (_, bindJump, lookupJump, destroyJump) =
		  make(Jump.new, Jump.plist, Jump.layout)
	    end

	    val bindJump =
	       fn j => let val j' = bindJump j
		       in handleJump{old = j, new = j'}
			  ; j'
		       end

	    fun lookupVars xs = List.map(xs, lookupVar)

	    datatype dec = datatype Dec.t
	    datatype transfer = datatype Transfer.t
	       
	    fun loopTransfer(t: Transfer.t): Transfer.t =
	       case t of
		  Bug => Bug
		| Halt => Halt
		| Raise xs => Raise(lookupVars xs)
		| Case{cause = cause, test, cases, default} =>
		     Case{cause = cause,
			  test = lookupVar test,
			  cases = List.map(cases, fn (c, j) =>
					   (c, lookupJump j)),
			  default = (case default of
					NONE => NONE
				      | SOME j => SOME(lookupJump j))}
		| Return xs => Return(lookupVars xs)
		| Jump{dst, args} => Jump{dst = lookupJump dst,
					  args = lookupVars args}
		| Call{func, args, cont} =>
		     Call{func = func,
			  args = lookupVars args,
			  cont = Option.map(cont, lookupJump)}

	    fun loopExp e =
	       let val {decs, transfer} = dest e
		  val decs = List.map(decs, loopDec)
	       in make{decs = decs,
		       transfer = loopTransfer transfer}
	       end
	    and loopDec d =
	       case d of
		  HandlerPush h => HandlerPush(lookupJump h)
		| HandlerPop => HandlerPop
		| Fun{name, args, body} =>
		     Fun{name = bindJump name,
			 args = List.map(args, fn (x, t) => (bindVar x, t)),
			 body = loopExp body}
		| Bind{var, ty, exp} => 
		     Bind{exp = PrimExp.replaceVar(exp, lookupVar),
			  var = bindVar var,
			  ty = ty}
	 in List.foreach(substitution, fn {formal, actual} =>
			 setVar(formal, SOME actual))
	    ; (loopExp exp
	       before (destroyVar() ; destroyJump()))
	 end
   end

structure DirectExp =
   struct
      open Dec PrimExp Transfer
      structure Cont =
	 struct
	    type t = PrimExp.t * Type.t -> Exp.t

	    val layout = Layout.ignore

	    fun nameGen(k: Var.t * Type.t -> Exp.t): t =
	       fn (e, t) =>
	       case e of
		  Var x => k(x, t)
		| _ => let val x = Var.newNoname()
		       in Exp.prefix(k(x, t), Bind{var = x, ty = t, exp = e})
		       end
	    
	    fun name(k: Var.t -> Exp.t): t = nameGen(k o #1)

	    fun return(k: t, xt) = k xt

	    fun jump(j: Jump.t): t =
	       name(fn x =>
		    Exp.make{decs = [],
			     transfer = Jump{dst = j, args = [x]}})

	    val id: t = name(fn x => Exp.make{decs = [], transfer = Return[x]})

	    val halt: t = name(fn x => Exp.make{decs = [], transfer = Halt})

	    fun prefix(k: t, d: Dec.t): t =
	       fn z => Exp.prefix(return(k, z), d)
	 end
      
      type t = Cont.t -> Exp.t

      val layout = Layout.ignore

      fun send(e: t, k: Cont.t): Exp.t = e k

(*       val send =
 * 	 Trace.trace2("DirectExp.send", layout, Cont.layout, Exp.layout) send
 *)

      fun sendName(e, k) = send(e, Cont.name k)

      fun toExp e = send(e, Cont.id)
      fun toExpHalt e = send(e, Cont.halt)

      fun const c = fn k => Cont.return(k, (Const c, Type.ofConst c))

      fun var(x, t) = fn k => Cont.return(k, (Var x, t))

      fun convertsGen(es: t list, k: Var.t list -> Exp.t): Exp.t =
	 let
	    fun loop(es, xs) =
	       case es of
		  [] => k(List.reverse xs)
		| e :: es => sendName(e, fn x => loop(es, x :: xs))
	 in loop(es, [])
	 end

      fun converts(es: t list, make: Var.t list -> PrimExp.t * Type.t): t =
	 fn k => convertsGen(es, k o make)

      fun convert(e: t, make: Var.t -> PrimExp.t * Type.t): t =
	 fn k => send(e, Cont.name(k o make))

      fun name2(e1: t, e2: t, make: Var.t * Var.t -> t): t =
	 fn k =>
	 sendName(e1, fn x1 => sendName(e2, fn x2 => send(make(x1, x2), k)))

      fun seq(e1: t, e2: t): t = fn k => sendName(e1, fn _ => send(e2, k))

      fun tuple{exps: t list, ty: Type.t}: t =
	 case exps of
	    [e] => e
	  | _ => converts(exps, fn xs => (PrimExp.Tuple xs, ty))

      fun select{tuple, offset, ty} =
	 convert(tuple, fn tuple => (Select{tuple = tuple, offset = offset}, ty))

      fun conApp{con, args, ty} =
	 converts(args, fn args => (ConApp{con = con, args = args}, ty))

      fun reify(ty: Type.t, z: Jump.t option -> Exp.t): t =
	 fn k =>
	 let val result = Var.newNoname()
	    val rest = Cont.return(k, (Var result, ty))
	    fun nontail() =
	       let val j = Jump.newNoname()
	       in Exp.prefix(z(SOME j),
			     Fun{name = j,
				 args = [(result, ty)],
				 body = rest})
	       end
	 in case Exp.dest rest of
	    {decs = [], transfer = Return[result']} =>
	       (* should also do something for returning unit *)
	       if Var.equals(result, result')
		  then z NONE
	       else nontail()
	  | _ => nontail()
	 end

      fun call{func, args, ty} =
	 reify
	 (ty, fn jo =>
	  convertsGen
	  (args, fn args =>
	   Exp.make{decs = [],
		    transfer = Call{func = func, args = args, cont = jo}}))

      fun raisee e =
	 fn _ => sendName(e, fn x => Exp.make{decs = [], transfer = Raise[x]})

      fun lett{decs, body} =
	 fn k =>
	 let
	    val rec loop =
	       fn [] => body k
		| {var, ty, exp} :: ds =>
		     send(exp, fn (e, _) =>
			  Exp.prefix(loop ds, Bind{var = var, ty = ty, exp = e}))
	 in loop decs
	 end

      fun joinPoint(ty: Type.t, e: t): t =
	 reify(ty, fn jo => send(e, (case jo of
					NONE => Cont.id
				      | SOME j => Cont.jump j)))
	 
      fun primApp{prim, targs, args, ty} =
	 converts(args, fn args =>
		  (PrimApp{prim = prim, targs = targs, args = args}, ty))

      fun casee{cause, test, ty, cases, default} =
	 joinPoint
	 (ty, fn k =>
	  test
	  (Cont.nameGen
	   (fn (test, testTy) =>
	    let val tycon = Type.detycon testTy
	       val (cases, decs) =
		  List.fold
		  (cases, ([], []),
		   fn ({con, args, body}, (cases, decs)) =>
		   let val e = send(body, k)
		      val j = Jump.newNoname()
		      val x = Var.newNoname()
		   in ((con, j) :: cases,
		       Fun{name = j,
			   args = [(x, Type.variant(tycon, con))],
			   body =
			   Exp.prefixs
			   (e,
			    List.foldi
			    (args, [], fn (i, (y, t), ds) =>
			     Bind{var = y,
				  ty = t,
				  exp = ConSelect{variant = x,
						  con = con,
						  offset = i}}
			     :: ds))}
		       :: decs)
		   end)
	       val (default, decs) =
		  case default of
		     NONE => (NONE, decs)
		   | SOME e =>
			let val j = Jump.newNoname()
			in (SOME j,
			    Fun{name = j, args = [], body = send(e, k)}
			    :: decs)
			end
	    in Exp.make{decs = decs,
			transfer = Case{cause = cause,
					test = test,
					cases = cases,
					default = default}}
	    end)))
	  
      fun handlee{try, ty, catch, handler} =
	 joinPoint
	 (ty, fn k =>
	  let val h = Jump.newNoname()
	  in Exp.prefixs
	     (send(try, Cont.prefix(k, HandlerPop)),
	      [Fun{name = h,
		   args = [catch],
		   body = Exp.prefix(send(handler, k), HandlerPop)},
	       HandlerPush h])
	  end)

      fun detupleGen(e: PrimExp.t, t: Type.t, components: Var.t list,
		     body: Exp.t): Exp.t =
	 Exp.prefixs
	 (body,
	  case components of
	     [] => []
	   | [x] => [Bind{var = x, ty = t, exp = e}]
	   | _ =>
		let val ts = Type.detuple t
		   val tupleVar = Var.newNoname()
		in Bind{var = tupleVar, ty = t, exp = e}
		   ::
		   List.mapi(List.zip(components, ts), fn (i, (x, t)) =>
			     Bind{var = x, ty = t,
				  exp = Select{tuple = tupleVar,
					       offset = i}})
		end)
      
      fun detupleBind{tuple, components, body} =
	 fn k => send(tuple, fn (e, t) => detupleGen(e, t, components, body k))

      fun detuple{tuple, body}: t =
	 fn k =>
	 tuple
	 (fn (e, t) =>
	  case e of
	     Tuple xs => send(body xs, k)
	   | _ => let
		     val components =
			List.map(Type.detuple t, fn _ => Var.newNoname())
		  in detupleGen(e, t, components, send(body components, k))
		  end)

      local
	 fun make c = conApp{con = c, args = [], ty = Type.bool}
      in val truee = make Con.truee
	 val falsee = make Con.falsee
      end

      val int = const o Const.fromInt

      fun e1 + e2 =
	 primApp{prim = Prim.intPlus,
		 targs = [],
		 args = [e1, e2],
		 ty = Type.int}

      fun eq(e1, e2, ty) =
	 primApp{prim = Prim.eq,
		 targs = [ty],
		 args = [e1, e2],
		 ty = Type.bool}
   end

structure Datatype =
   struct
      type t = {tycon: Tycon.t,
		cons: {con: Con.t,
		       args: Type.t list} list}

      fun layout(ds: t list) =
	 let open Layout
	 in align
	    ((str "Datatypes:")
	     ::
	     List.map
	     (ds, fn {tycon, cons} =>
	      seq[Tycon.layout tycon,
		  str " = ",
		  align
		  (List.map(cons, fn {con, args} =>
			    seq[Con.layout con,
				case args of
				   [] => empty
				 | _ => 
				      seq[str " of ",
					  list(List.map(args, Type.layout))]]))]))
	 end
   end
		      
structure Function =
   struct
      type t = {name: Func.t,
		args: (Var.t * Type.t) list,
		body: Exp.t,
		returns: Type.t list}

      fun layout{name, args, body, returns} =
	 let open Layout
	 in align[seq[str "fun ",
		      Func.layout name,
		      tuple(List.map(args, fn (x, t) =>
				     seq[Var.layout x,
					 str ": ",
					 Type.layout t])),
		      str ": ",
		      tuple(List.map(returns, Type.layout)),
		      str " = "],
		  indent(Exp.layout body, 3)]
	 end
	 
      fun layouts(fs, output) =
	 List.foreach(fs, output o layout)
   end

structure Program =
   struct
      datatype t =
	 T of {
	       datatypes: Datatype.t list,
	       globals: {
			 var: Var.t,
			 ty: Type.t,
			 exp: PrimExp.t
			} list,
	       functions: Function.t list,
	       main: Func.t
	      }

      fun foreachVar(T{globals, functions, ...}, f) =
	 (List.foreach(globals, fn {var, ty, ...} => f(var, ty));
	  List.foreach
	  (functions, fn {args, body, ...} =>
	   (List.foreach(args, f);
	    Exp.foreachVar(body, f))))
	 
      fun layouts(T{datatypes, globals, functions, main},
		  output': Layout.t -> unit) =
	 let open Layout
	    val output = output'
	 in output(Datatype.layout datatypes)
	    ; output(str "\n\nGlobals:")
	    ; List.foreach(globals, output o layoutBind)
	    ; output(str "\n\nFunctions:")
	    ; Function.layouts(functions, output)
	    ; output(seq[str "\n\nMain: ", Func.layout main])
	 end

      fun layout(T{datatypes, globals, functions, main}) =
	 let open Layout
	 in align[Datatype.layout datatypes,
		  str "Globals:",
		  align(List.map(globals, layoutBind)),
		  str "Functions:",
		   align(List.map(functions, Function.layout)),
		   seq[str "Main: ", Func.layout main]]
	 end

      fun layoutStats(T{globals, functions, ...}) =
	 let
	    val numTypes = ref 0
	    fun inc _ = IntRef.inc numTypes
	    val {hom = countType, destroy} =
	       Type.makeHomVariant
	       {var = fn _ => Error.bug "cps-tree saw var",
		con = inc, variant = inc}
	    val numPrimExps = ref(List.length globals)
	    val numLocalFunctions = ref 0
	    val _ =
	       List.foreach
	       (functions, fn {body, ...} =>
		Exp.foreach'
		(body,
		 {handleDec = let open Dec
			      in fn Bind{ty, ...} => (countType ty;
						      IntRef.inc numPrimExps)
			          | Fun{args, ...} =>
				       (IntRef.inc numLocalFunctions;
					List.foreach(args, countType o #2))
				  | _ => ()
			      end,
		  handleTransfer = fn _ => ()}))
	    val numFunctions = List.length functions
	    open Layout
	 in destroy();

	    align
	    (List.map
	     ([("num functions", Int.layout numFunctions),
	       ("num local functions", Int.layout(!numLocalFunctions)),
	       ("num primExps", Int.layout(!numPrimExps)),
	       ("num types in program", Int.layout(!numTypes))],
	      fn (name, value) => seq[str(name ^ " "), value])
	     @ [Type.stats()])
	 end

      (* clear all property lists reachable from program *)
      fun clear(T{datatypes, globals, functions, main}) =
	 (List.foreach(datatypes, fn {tycon, cons} =>
		       (Tycon.clear tycon;
			List.foreach(cons, Con.clear o #con)));
	 List.foreach(globals, clearBind);
	 List.foreach
	 (functions, fn {name, args, body, returns} =>
	  (Func.clear name;
	   List.foreach(args, Var.clear o #1);
	   Exp.clear body)))
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Cps(S: CPS_STRUCTS): CPS = 
   Simplify(Shrink(TypeCheck(InferHandlers(Analyze(CpsTree(S))))))
signature ASSERT =
   sig
      (* will always be checked *)
      val cheap: bool -> unit
	 
      val checkExpensive: bool
      (* will only be checked if checkExpensive = true *)
      val expensive: (unit -> bool) -> unit
   end
structure Assert: ASSERT =
   struct
      fun cheap(b: bool): unit =
	 if b then () else Error.bug "cheap assertion failed"

      val checkExpensive = false
      
      fun expensive(th: unit -> bool): unit =
	 if checkExpensive andalso not(th())
	    then Error.bug "expensive assertion failed"
	 else ()
   end
signature HASH_SET_STRUCTS = 
   sig
      structure Element :
	 sig
	    include T
	    val hash: t -> word
	 end

      (* how many binary operations to cache *)
      val cacheSize: int

      (* 2^bits buckets in the hash table *)
      val bits: int
   end

signature HASH_SET =
   sig
      include HASH_SET_STRUCTS
	 
      type t
	       
      val empty: t
      val singleton: Element.t -> t
      val isEmpty: t -> bool
      val foreach: t * (Element.t -> unit) -> unit
      val - : t * t -> t
      val + : t * t -> t
      val intersect: t * t -> t
      val fromList: Element.t list -> t
      val toList: t -> Element.t list
	 
      val plist: t -> PropertyList.t

      val equals: t * t -> bool
      val layout: t -> Layout.t

      val reset: unit -> unit

      val stats: unit -> {hits: int, misses: int}
   end
structure HashSetRep =
   struct
      datatype 'a t = T of {elements: 'a list,
			    plist: PropertyList.t}

   end

functor HashSet(S: HASH_SET_STRUCTS): HASH_SET =
struct

open S

val _ = Assert.cheap(cacheSize >= 1 andalso bits >= 1)

type elements = Element.t list
   
structure Tree: sig
		   structure Set:
		      sig
			 type t

			 val equals: t * t -> bool
			 val toList: t -> elements
			 val plist: t -> PropertyList.t
		      end
		   
		   type t
		      
		   val new: unit -> t
		   val insert: t * elements -> Set.t
		   val size: t -> int
		end =
   struct
      structure Set  =
	 struct
	    open HashSetRep
	    type t = Element.t t
	       
	    fun new elements = T{elements = elements,
				 plist = PropertyList.new()}

	    fun elements(T{elements, ...}) = elements
	    fun plist(T{plist, ...}) = plist

	    val toList = elements

	    fun equals(s, s') = PropertyList.equals(plist s, plist s')
	 end

      datatype node =
	 Node of {element: Element.t,
		  isIn: t,
		  isNotIn: t}
       | Leaf of Set.t
      withtype t = node option ref

      fun new(): t = ref NONE

      fun size(t: t): int =
	 case !t of
	    NONE => 0
	  | SOME(Leaf _) => 1
	  | SOME(Node{isIn, isNotIn, ...}) => size isIn + size isNotIn
	 
      fun contains(es, e) = List.exists(es, fn e' => Element.equals(e, e'))
	 
      fun insert(tree, elements) =
	 let
	    fun loop tree =
	       case !tree of
		  NONE => let val s = Set.new elements
			  in tree := SOME(Leaf s) ; s
			  end
		| SOME(Node{element, isIn, isNotIn}) =>
		  if contains(elements, element)
		     then loop isIn
		  else loop isNotIn
		| SOME(Leaf s') =>
		     let
			fun loop arg =
			   case arg of
			      ([], []) => s' (* same set *)
			    | ([], x' :: _) =>
				 let val s = Set.new elements
				 in tree :=
				    SOME(Node{element = x',
					      isIn = ref(SOME(Leaf s')),
					      isNotIn = ref(SOME(Leaf s))}) ;
				    s
				 end
			    | (x :: xs, xs') =>
				 let
				    fun loop2(xs', accum) =
				       case xs' of
					  [] =>
					     let val s = Set.new elements
					     in tree :=
						SOME(Node{element = x,
							  isIn = ref(SOME(Leaf s)),
							  isNotIn =
							  ref(SOME(Leaf s'))}) ;
						s
					     end
					| x' :: xs' =>
					     if Element.equals(x, x')
						then loop(xs, accum @ xs')
					     else loop2(xs', x' :: accum)
				 in loop2(xs', [])
				 end
		     in loop(elements, Set.elements s')
		     end
	 in loop tree
	 end
		     
   end

open Tree.Set

val tableSize = Int.^(2, bits)

val maxIndex = tableSize - 1

val mask = Word.fromInt maxIndex

val table = Array.tabulate(tableSize, fn _ => Tree.new())
   
fun hashToIndex(w: word): int = Word.toInt(Word.andb(w, mask))

fun intern(l: Element.t list, h: word) =
   Tree.insert(Array.sub(table, hashToIndex h), l)

(* the hash of a set is the xorb of the hash of its members *)
fun hash(l: Element.t list) =
   List.fold(l, 0w0, fn (e, w) => Word.xorb(w, Element.hash e))
   
fun fromList l =
   let val l = List.fold(l, [], fn (x, l) =>
			 if List.exists(l, fn x' => Element.equals(x, x'))
			    then l
			 else x :: l)
   in intern(l, hash l)
   end

val empty = fromList []

fun isEmpty s = equals(s, empty)

fun foreach(s, f) = List.foreach(toList s, f)
   
fun singleton x = fromList [x]

val cacheHits = ref 0
val cacheMisses = ref 0

val numOps = ref 0
val next = ref 1

fun stats() = {hits = !cacheHits, misses = !cacheMisses}
fun reset() =
   (* need to clear out and reset the tables *)
   (cacheHits := 0;
    cacheMisses := 0;
    Int.foreach(0, maxIndex, fn i => Array.update(table, i, Tree.new())))

(*      Int.foreach(0, maxIndex, fn i =>
		  let val n = Tree.size(Vector.sub(table, i))
		  in if n > 0
			then Control.message(seq[Int.layout i,
						 str " -> ",
						 Int.layout n])
		     else ()
		  end)*)

local
   fun binary (oper: elements * elements -> elements) =
      let
	 val cache = Array.new(cacheSize, NONE)
      in
	 fn (s: t, s': t) =>
	 let
	    fun loop i =
	       if i = cacheSize
		  then
		     let val s'' = fromList(oper(toList s, toList s'))
		     in IntRef.inc cacheMisses;
			Array.update(cache, Int.randomRange(0, cacheSize - 1)(),
				     SOME(s, s', s''));
			s''
		     end
	       else case Array.sub(cache, i) of
		  NONE => loop(i + 1)
		| SOME(s1, s1', s'') =>
		     if equals(s, s1) andalso equals(s', s1')
			then (IntRef.inc cacheHits ; s'')
		     else loop(i + 1)
	 in loop 0
	 end
      end

   val {+, -, intersect, layout, ...} =
      List.set{equal = Element.equals,
	       layout = Element.layout}
in
   val op + = binary op +
   val op - = binary op -
   val op intersect = binary intersect

   val layout = layout o toList
end

(* val fromList = Trace.trace("fromList", List.layout Element.layout, layout) fromList *)
   
(*
 * fun traceBinary(name, f) = Trace.trace2(name, layout, layout, layout) f
 * val op + = traceBinary("+", op +)
 * val op - = traceBinary("-", op -)
 * val op intersect = traceBinary("intersect", intersect)
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MTYPE_STRUCTS = 
   sig
   end

signature MTYPE = 
   sig
      include MTYPE_STRUCTS

      type t
	 
      datatype dest =
	 Char
       | Double
       | Int
       | Pointer
       | Uint
       | Void

      val all: t list

      val dest: t -> dest

      val char: t
      val double: t
      val int: t
      val pointer: t
      val uint: t
      val void: t

      val label: t (* same as uint *)
	 
      val isVoid: t -> bool
      val isPointer: t -> bool

      (* align an address *)	 
      val align: int * t -> int

      val toString: t -> string
	 
      val size: t -> int (* bytes *)
      val equals: t * t -> bool
      val layout: t -> Layout.t

      val memo: (t -> 'a) -> (t -> 'a)
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Mtype(S: MTYPE_STRUCTS): MTYPE = 
struct

open S

datatype dest =
   Char
 | Double
 | Int
 | Pointer
 | Uint
 | Void

datatype t = T of {dest: dest}

fun dest(T{dest, ...}) = dest

fun equals(t, t') = dest t = dest t'

local
   fun new dest = T{dest = dest}
in
   val char = new Char
   val double = new Double
   val int = new Int
   val pointer = new Pointer
   val uint = new Uint
   val void = new Void
end

val all = [char, double, int, pointer, uint, void]

fun memo f =
   let val all = List.map(all, fn t => (t, f t))
   in fn t => #2(List.lookup(all, fn (t', _) => equals(t, t')))
   end

val label = uint
  
fun isVoid t =
   case dest t of
      Void => true
    | _ => false

fun isPointer t =
   case dest t of
      Pointer => true
    | _ => false
	 
local
   val byte = 1
   val word = 4
   val double = 8
in
   fun size t =
      case dest t of
	 Char => byte
       | Double => double
       | Int => word
       | Pointer => word
       | Uint => word
       | Void => 0
end

fun toString t =
   case dest t of
      Char => "uchar"
    | Double => "double"
    | Int => "int"
    | Pointer => "pointer"
    | Uint => "uint"
    | Void => "void"

val layout = Layout.str o toString

fun align(n: int, ty: t): int =
   let val s = size ty
   in if n mod s = 0
	 then n
      else n - (n mod s) + s
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MACHINE_STRUCTS = 
   sig
      structure Label: ID
      structure Prim: PRIM
      structure SmallIntInf: SMALL_INT_INF
   end

signature MACHINE = 
   sig
      include MACHINE_STRUCTS

      structure Type: MTYPE

      structure Register:
	 sig
	    type t

	    val ty: t -> Type.t
	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      structure Operand:
	 sig
	    type t

	    val void: t

	    (* constants *)
	    val int: int -> t
	    val int32: Int32.int -> t
	    val intInf: SmallIntInf.t -> t

	    val prim: Prim.t * Type.t -> t (* Prim.numArgs prim = NONE *)

	    (* must be nonzero mod 4 *)
	    val pointer: int -> t

	    val uint: Word32.word -> t
	    val char: char -> t
	    val float: string -> t

	    val label: Label.t -> t

	    val bogus: Type.t -> t

	    (* takes an IntOrPointer and makes it an int *)
	    val castInt: t -> t

	    (* addresses *)
	    val register: Register.t -> t
	    val deRegister: t -> Register.t option
	       
	    (* *(base + (offset * scale)) *)
	    val scaledOffset: {base: t,
			       offset: t,
			       scale: int,
			       ty: Type.t} -> t

	    (* *(base + offset) *)
	    val offset: {base: t,
			 offset: t,
			 ty: Type.t} -> t

	    val stackOffset: {offset: int, ty: Type.t} -> t
	    val deStackOffset: t -> {offset: int, ty: Type.t} option
	       
	    val contents: t * Type.t -> t

	    val ty: t -> Type.t

	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      (* This is the information that the garbage collector needs to know
       * at a program point in order to do a collection.
       *)
      structure GCInfo:
	 sig
	    (* Regs and offsets should not contain duplicates. *)
	    datatype t =
	       T of {
		     (* Size of stack frame in bytes,
		      * including return address.
		      *)
		     frameSize: int,
		     (* Live pointer valued registers. *)
		     regs: Register.t list,
		     (* Live pointer valued stack offsets. *)
		     offsets: int list
		     }
	 end

      structure Statement:
	 sig
	    type t

	    (* when registers or offsets appear in operands, there is an
	     * implicit contents of.  when they appear as locations, there is not
	     *)
	    val move: {dst: Operand.t, src: Operand.t} -> t

	    (* error if dsts and srcs aren't of same length
	     *)
	    val moves: {
			dsts: Operand.t list,
			srcs: Operand.t list
		       } -> t list
	       
	    val assign: {dst: Operand.t,
			 oper: Prim.t,
			 args: Operand.t list} -> t

	    val sideEffect: Prim.t * Operand.t list -> t

	    val saveExnStack: {offset: int} -> t
	    val restoreExnStack: {offset: int} -> t

	    val stackOverflowCheck: t

	    (* push/pop number of bytes from stack *)
	    val push: int -> t
	    val pop: int -> t
	       
	    (* fixed size allocation *)
	    val allocate:
	       {dst: Operand.t,
		size: int,
		numPointers: int,
		numWordsNonPointers: int,
		stores: {offset: int,
			 value: Operand.t} list} -> t

	    val allocateArray:
	       {dst: Operand.t,
		numElts: Operand.t,
		numPointers: int,
		numBytesNonPointers: int,
		gcInfo: GCInfo.t
		} -> t
	 end

      structure Transfer:
	 sig
	    type t
	    type chunk

	    val halt: Operand.t -> t
	    val bug: t
	    val return: t
	    val raisee: t

	    val switch: {
			 test: Operand.t,
			 cases: (int * Label.t) list,
			 default: Label.t option
			} -> t

	    val switchIP: {
			   test: Operand.t,
			   int: Label.t,
			   pointer: Label.t
			  } -> t

	    val nearJump: Label.t -> t
	    val farJump: chunk * Label.t -> t
	 end

      structure Chunk:
	 sig
	    type t

	    val register: t * int * Type.t -> Register.t
	    val tempRegister: t * Type.t -> Register.t

	    val newBlock: 
	       t * {
		    profileName: string,
		    label: Label.t,
		    (* 
		     * If gcInfo is SOME, then a limit check will be inserted
		     * at the beginning of this block.
		     *)
		    gcInfo: GCInfo.t option,
		    statements: Statement.t list,
		    transfer: Transfer.t
		   } -> unit

	    val equals: t * t -> bool
	 end
      sharing type Transfer.chunk = Chunk.t

      structure Program:
	 sig
	    type t

	    val new: unit -> t

	    val setMain: t * Chunk.t * Label.t -> unit
	    val setSaveWorld: t * Chunk.t * Label.t -> unit

	    (* Create a new chunk *)
	    val newChunk: {program: t,
			   entries: Label.t list} -> Chunk.t

	    val newFrame:
	       t * {return: Label.t, (* where to return to *)
		    chunk: Chunk.t,
		    (* number of bytes in frame, including return address *)
		    size: int,
		    (* where are live pointers in the current stack frame
		     * relative to the stack pointer of the frame below
		     *)
		    liveOffsets: int list
		    } -> unit

	    val newHandler: t * Label.t * Chunk.t -> unit

	    val newGlobal: t * Type.t -> Operand.t

	    (* A global pointer that the GC doesn't use as a root *)
	    val newGlobalPointerNonRoot: t -> Operand.t
	       
	    val newString: t * string -> Operand.t
	    val newIntInf: t * string -> Operand.t

	    val output: {program: t,
			 includes: string list,
			 print: string -> unit} -> unit
	 end

      val reset: unit -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Machine(S: MACHINE_STRUCTS): MACHINE = 
struct

open S

val wordSize = 4
val pointerSize = wordSize
val objectHeaderSize = wordSize
val arrayHeaderSize = 2 * wordSize
val intInfOverhead = arrayHeaderSize + wordSize (* for the sign *)

val overhead = "**C overhead**"

structure Label =
   struct
      open Label

      fun index l = toString l ^ "_index"
   end

structure Type = Mtype()

structure Type =
   struct
      open Type

      fun name t =
	 case dest t of
	    Char => "char"
	  | Double => "double"
	  | Int => "int"
	  | Uint => "uint"
	  | Pointer => "pointer"
	  | Void => "void"

      fun globalName ty = "global" ^ name ty
   end

structure C =
   struct
      val truee = "TRUE"
      val falsee = "FALSE"
	 
      fun call(f: string, xs: string list, print: string -> unit): unit =
	 (print f
	  ; print "("
	  ; (case xs of
		[] => ()
	      | x :: xs => (print x
			    ; List.foreach(xs, fn x => (print ","; print x))))
	  ; print ")")

      fun int32(n: Int32.int): string =
	 if n >= 0
	    then Int32.toString n
	 else if n = valOf Int32.minInt
		 then "(int)0x80000000" (* because of goofy gcc warning *)
	      else "-" ^ String.dropPrefix(Int32.toString n, 1)
      (* This overflows on Int32.minInt: Int32.toString(~ n) *)

      val int = int32 o Int32.fromInt

      fun char(c: char) =
	 concat[if Char.ord c >= 0x80 then "(uchar)" else "",
		"'", Char.toCString c, "'"]

      fun word(w: Word32.word) = "0x" ^ Word32.toString w

      (* the only difference between SML floats and C floats is that
       * SML uses "~" while C uses "-"
       *)
      val float = String.translate(fn #"~" => "-" | c => String.fromChar c)

      fun assign(lhs, rhs, print) =
	 (print lhs
	  ; print " = "
	  ; print rhs
	  ; print ";\n")

      fun string(s, print) =
	 let val quote = "\""
	 in print quote
	    ; print(String.toCString s)
	    ; print quote
	 end

      fun bug(s: string, print) =
	 (call("MLTON_bug",
	       [concat["\"", String.toCString s, "\""]],
	       print)
	  ; print ";\n")

      fun goto l = concat["goto ", l, ";"]
      fun gotoLabel l = goto(Label.toString l)

      local
	 val current = ref ""
      in
	 fun profile(detailed: string, nonDetailed: string,
		     print: string -> unit): unit =
	    if !Control.profile
	       then
		  if detailed <> !current
		     then
			(print "/* PROFILE: "
			 ; print detailed
			 ; print " & "
			 ; print nonDetailed
			 ; print " */\n"
			 ; current := detailed)
		  else ()
	    else ()
      end
   end

structure Register =
   struct
      datatype t =
	 T of {
	       (* register number.
	        * For pointer regs, this is also an index into an array of
		* locals used within the chunk to communicate with th GC.
		*)
	       index: int,
	       ty: Type.t
	      }

      fun ty(T{ty, ...}) = ty
      fun index(T{index, ...}) = index
	 
      fun equals(r1, r2) =
	 Type.equals(ty r1, ty r2)
	 andalso index r1 = index r2

      fun toString(T{index, ty}) = concat[Type.name ty, C.int index]

      val layout = Layout.str o toString
   end

structure Global =
   struct
      datatype t = T of {index: int, ty: Type.t}

      fun ty(T{ty, ...}) = ty
      fun index(T{index, ...}) = index

      fun toString(T{index, ty}) =
	 concat[Type.globalName ty, "[", C.int index, "]"]
   end

structure Operand =
   struct
      (* Be careful to update Operand.equals if you add new variants to this
       * datatype.
       *)
      datatype t =
	 Void
       | Int of int
       | Int32 of Int32.int
       | IntInf of SmallIntInf.t
       | Prim of string * Type.t
       | Pointer of int
       | Uint of Word32.word
       | Char of char
       | Float of string
       | Label of Label.t
       | CastInt of t
       | Register of Register.t
       | Global of Global.t
       | GlobalPointerNonRoot of int
       | Offset of {base: t, offset: t, scale: int, ty: Type.t}
       | StackOffset of {offset: int, ty: Type.t}
       | Contents of {oper: t, ty: Type.t}

      val rec toString =
	 fn Void => "void"
	  | Int n => C.int n
	  | Int32 n => C.int32 n
	  | IntInf s => concat["(pointer)", SmallIntInf.toCstring s]
	  | Prim(p, _) => p
	  | Pointer n => concat["(pointer)", C.int n]
	  | Uint w => C.word w
	  | Char c => C.char c
	  | Float s => C.float s
	  | Label l => Label.index l
	  | CastInt oper => concat["(int)", toString oper]
	  | Register r => Register.toString r
	  | Global g => Global.toString g
	  | GlobalPointerNonRoot n =>
	       concat["globalpointerNonRoot[", C.int n, "]"]
	  | Offset{base, offset, scale, ty} =>
	       concat["*(", Type.toString ty, "*)(",
		      toString base, " + ",
		      toString offset, " * ", C.int scale, ")"]
	  | StackOffset{offset, ty} =>
	       concat["*(", Type.toString ty, "*)(stackTop + ",
		      C.int offset, ")"]
	  | Contents{oper, ty} =>
	       concat["*(", Type.toString ty, "*)(", toString oper, ")"]

      val layout = Layout.str o toString

      val void = Void
      val int = Int
      val int32 = Int32
      val intInf = IntInf
      val pointer = Pointer
      val uint = Uint
      val char = Char
      val float = Float
      val label = Label
      val castInt = CastInt
      val register = Register
      val global = Global
      val stackOffset = StackOffset

      val deRegister =
	 fn Register r => SOME r
	  | _ => NONE

      val deStackOffset =
	 fn StackOffset s => SOME s
	  | _ => NONE

      fun bogus t =
	 case Type.dest t of
	    Type.Char => char #"\000"
	  | Type.Double => float "0.0"
	  | Type.Int => int 0
	  | Type.Uint => uint 0w0
	  | Type.Pointer => pointer 1
	  | _ => Error.bug "bogus not implemented for type"

      fun prim(p, t) =
	 case Prim.name p of
	    Prim.Name.CPrim s => Prim(s, t)
	  | Prim.Name.Bogus => bogus t
	  | _ => Error.bug("strange Prim operand "
			   ^ Layout.toString(Prim.layout p))

      fun offset{base, offset, ty} =
	 Offset{base = base, offset = offset, ty = ty, scale = 1}

      val scaledOffset = Offset
	 
      fun contents(z, t) = Contents{oper = z, ty = t}

      val ty =
	 fn Void => Type.void
	  | Int _ => Type.int
	  | Int32 _ => Type.int
	  | IntInf _ => Type.pointer
	  | Prim(_, ty) => ty
	  | Uint _ => Type.uint
	  | Char _ => Type.char
	  | Float _ => Type.double
	  | Pointer _ => Type.pointer
	  | CastInt _ => Type.int
	  | Label _ => Type.label
	  | Register r => Register.ty r
	  | Global g => Global.ty g
	  | GlobalPointerNonRoot _ => Type.pointer
	  | Offset{ty, ...} => ty
	  | Contents{ty, ...} => ty
	  | StackOffset{ty, ...} => ty

      val rec equals =
	 fn (Int n, Int n') => n = n'
	  | (Int32 n, Int32 n') => n = n'
	  | (IntInf n, IntInf n') => SmallIntInf.equals(n, n')
	  | (Pointer n, Pointer n') => n = n'
	  | (Uint w, Uint w') => w = w'
	  | (Char c, Char c') => c = c'
	  | (Float f, Float f') => f = f'
	  | (CastInt z, CastInt z') => equals(z, z')
	  | (Register r, Register r') => Register.equals(r, r')
	  | (Offset{base = b, offset = i, scale = s, ...},
	     Offset{base = b', offset = i', scale = s', ...}) =>
	    equals(b, b') andalso equals(i, i') andalso s = s'
	  | (StackOffset{offset = n, ...}, StackOffset{offset = n', ...}) =>
	       n = n'
	  | (Contents{oper = z, ...}, Contents{oper = z', ...}) => equals(z, z')
	  | _ => false
   end

structure GCInfo =
   struct
      datatype t =
	 T of {
	       (* Size of frame, including return address. *)
	       frameSize: int,
	       (* Live pointer valued registers. *)
	       regs: Register.t list,
	       (* Live pointer valued stack offsets. *)
	       offsets: int list
	       }
   end

(* ------------------------------------------------- *)
(*                    LimitCheck                     *)
(* ------------------------------------------------- *)

structure LimitCheck =
   struct
      fun output({bytes: string,
		  localIndices: string,
		  frameSize: int,
		  return: Label.t},
		 print) =
	 (C.call("MLTON_limitCheck",
		 [bytes, localIndices, C.int frameSize, Label.toString return],
		 print)
	  ; print ";\n")
   end

fun primApp(p: Prim.t, args: Operand.t list, print): unit =
   let open Prim.Name
      fun doit s = C.call(s, List.map(args, Operand.toString), print)
   in case Prim.name p of
      CPrim s => (case Prim.numArgs p of
		     NONE => print s
		   | SOME _ => doit s)
    | ArrayLength => doit "MLTON_Array_length"
    | VectorLength => doit "MLTON_Vector_length"
    | Eq => (if Type.isVoid(Operand.ty(hd args))
		then print "TRUE"
	     else doit "MLTON_eq")
    | Size => let val ty = Operand.ty(hd args)
	      in case Type.dest ty of
		 Type.Pointer => doit "MLTON_MLton_size"
	       | _ => print "0"
	      end
    | SaveStack => doit "MLTON_MLton_saveStack"
    | SaveWorld => doit "MLTON_MLton_saveWorld"
    | _ => Error.bug("unexpected primitive in machine: "
		     ^ Layout.toString(Prim.layout p))
   end

structure Statement =
   struct
      datatype t =
	 Noop
       | Move of {
		  dst: Operand.t,
		  src: Operand.t
		 }
       | Push of int
       | Assign of {dst: Operand.t,
		    oper: Prim.t,
		    args: Operand.t list}
       | SideEffect of Prim.t * Operand.t list
       | SaveExnStack of {offset: int}
       | RestoreExnStack of {offset: int}
       | StackOverflowCheck
       | Allocate of {dst: Operand.t,
		      size: int,
		      numPointers: int,
		      numWordsNonPointers: int,
		      stores: {offset: int,
			       value: Operand.t} list}
       | AllocateArray of {
			   user: {
				  dst: Operand.t,
				  numElts: Operand.t,
				  numPointers: int,
				  numBytesNonPointers: int,
				  gcInfo: GCInfo.t
				  },
			   limitCheck: {return: Label.t,
					bytes: string,
					localIndices: string} option ref
			  }

      val stackOverflowCheck = StackOverflowCheck
	 
      fun output(s, print) =
	 case s of
	    Noop => ()
	  | Move{dst, src} => (print(Operand.toString dst)
			       ; print " = "
			       ; print(Operand.toString src)
			       ; print ";\n")
	  | Push n => (print "stackTop += "; print(C.int n); print ";\n")
	  | Assign{dst, oper, args} => (print(Operand.toString dst)
					; print " = "
					; primApp(oper, args, print)
					; print ";\n")
	  | SideEffect(prim, args) => (primApp(prim, args, print); print ";\n")
	  | StackOverflowCheck => print "MLTON_stackOverflowCheck\n"
	  | SaveExnStack{offset} =>
	       (C.call("MLTON_saveExnStack", [C.int offset], print)
		; print "\n")
	  | RestoreExnStack{offset} =>
	       (C.call("MLTON_restoreExnStack", [C.int offset], print)
		; print "\n")
	  | Allocate{dst, size, numPointers, numWordsNonPointers, stores} =>
	       (print "*(word*)frontier = "
		; C.call("GC_objectHeader",
			 [C.int numWordsNonPointers, C.int numPointers],
			 print)
		; print ";\n"
		; C.assign(Operand.toString dst,
			   concat["frontier + ", C.int objectHeaderSize],
			   print)
		; (List.foreach
		   (stores, fn {offset, value} =>
		    if Type.isVoid(Operand.ty value)
		       then ()
		    else (print "*("
			  ; print(Type.toString(Operand.ty value))
			  ; print "*)(frontier + "
			  ; print(C.int(offset + objectHeaderSize))
			  ; print ") = "
			  ; print(Operand.toString value)
			  ; print ";\n")))
		; print "frontier += "
		; print(C.int(size + objectHeaderSize))
		; print ";\n")
	| AllocateArray{user = {dst, numElts, numPointers,
				numBytesNonPointers,
				gcInfo = GCInfo.T{frameSize, ...},
				...},
			limitCheck} =>
	  (case !limitCheck of
	      NONE => ()
	    | SOME{bytes, localIndices, return} =>
		 LimitCheck.output({bytes = bytes,
				    localIndices = localIndices,
				    frameSize = frameSize,
				    return = return},
				   print)
	   ; let val dst = Operand.toString dst
		 val numElts = Operand.toString numElts
	     in if numPointers = 0
		   then (print "MLTON_allocArrayNoPointers("
			 ; print dst
			 ; print ","
			 ; print numElts
			 ; print ","
			 ; print(C.int numBytesNonPointers)
			 ; print ");\n")
		else if numBytesNonPointers = 0
			then (print "MLTON_allocArrayPointers("
			      ; print dst
			      ; print ","
			      ; print numElts
			      ; print ","
			      ; print (C.int numPointers)
			      ; print ");\n")
		     else Error.unimplemented "tricky arrays"
	     end)

      val push = Push
      val pop = Push o ~

      fun move(arg as {dst, src}) =
	 if Type.isVoid(Operand.ty dst)
	    orelse Operand.equals(dst, src)
	    then Noop
	 else Move arg
	    
      fun assign(arg as {dst, ...}) =
	 if Type.isVoid(Operand.ty dst)
	    then Noop
	 else Assign arg

      val saveExnStack = SaveExnStack
      val restoreExnStack = RestoreExnStack
      val sideEffect = SideEffect

      fun allocate(arg as {dst, size, ...}) =
	 Allocate(if size = 0
		     then {dst = dst,
			   size = wordSize (* min size *),
			   numPointers = 0,
			   numWordsNonPointers = 1,
			   stores = []}
		  else arg)

      fun allocateArray user =
	 AllocateArray{user = user, limitCheck = ref NONE}

      fun moves{srcs, dsts} =
	 List.map2(srcs, dsts, fn (src, dst) =>
		   move{src = src, dst = dst})
   end

structure Offsets =
   struct
      structure Ints =
	 HashSet(val cacheSize = 1
		 val bits = 14
		 structure Element =
		    struct
		       open Int
		       fun hash n = Word.fromInt n
		    end)

      fun new name = 
	 let val all = ref []
	    val c = Counter.new 0
	    val {get, ...} =
	       Property.new
	       (Ints.plist,
		Property.initFun
		(fn offsets =>
		 let val name = concat[name, C.int(Counter.next c)]
		 in ListRef.push(all, {name = name,
				       offsets = Ints.toList offsets});
		    name
		 end))

	    fun declare(print: string -> unit): unit =
	       List.foreach
	       (!all, fn {name, offsets} =>
		(print "static ushort "
		 ; print name
		 ; print "[] = {"
		 ; print(C.int(List.length offsets))
		 ; List.foreach(offsets, fn i => (print ","; print(C.int i)))
		 ; print "};\n"))

	 in {name = get o Ints.fromList,
	     declare = declare}
	 end
   end

(* ------------------------------------------------- *)
(*                       Types                       *)
(* ------------------------------------------------- *)

structure ChunkName:>
   sig
      type t
      val new: unit -> t
      val toString: t -> string
   end =
   struct
      datatype t = T of string

      fun toString(T s) = s

      val counter = Counter.new 0

      fun new() = T("Chunk" ^ Int.toString(Counter.next counter))
   end

structure Types =
   struct
      datatype transfer =
	 Halt of Operand.t
       | Bug
       | Return
       | Raise
       | Switch of {
		    test: Operand.t,
		    cases: (int * Label.t) list,
		    default: Label.t option
		    }
       | SwitchIP of {
		      test: Operand.t,
		      int: Label.t,
		      pointer: Label.t
		      }
       | NearJump of Label.t
       | FarJump of chunk * Label.t

      and block = Block of {
			    profileName: string,
			    label: Label.t,
			    bytesNeeded: int option ref,
			    limitCheck: {
					 gcInfo: GCInfo.t,
					 extra: {bytes: int,
						 localMask: string,
						 return: Label.t} option ref
					 } option,
			    statements: Statement.t array,
			    transfer: transfer
			    }

      and chunk = Chunk of {
			    name: ChunkName.t,

			    (* where to start *)
			    entries: Label.t list,
			    
			    blocks: block list ref,

			    (* for each type, gives the max # registers used *)
			    regMax: Type.t -> int ref,

			    limitCheckInfo: {numLocals: int,
					     localUsed: int option array,
					     gcReturns: Label.t list
					     } option ref
			    }
   end

structure Chunk =
   struct
      datatype t = datatype Types.chunk

      fun name(Chunk{name, ...}) = name

      val nameToString = ChunkName.toString o name
   end

structure Frames =
   struct
      type t = {return: Label.t,
		chunk: Chunk.t,
		size: int,
		liveOffsets: int list} list ref
	 
      val maxFrameSize = Int.^(2, 16)

      fun add(frames: t, f as {size, ...}) =
	 if size > maxFrameSize
	    then Error.bug(concat["MLton cannot handle stack frames larger than ",
				  Int.toString maxFrameSize,
				  " bytes."])
	 else ListRef.push(frames, f)
   end

(* ------------------------------------------------- *)
(*                     Transfer                      *)
(* ------------------------------------------------- *)

structure Transfer =
   struct
      type chunk = Types.chunk
      datatype t = datatype Types.transfer
	 
      fun output(t, print, gotoLabel: Label.t -> unit): unit =
	 let 
	    fun iff(test, a, b): unit =
	       (print "if ("
		; print test
		; print ") {\n"
		; gotoLabel a
		; print "} else {\n"
		; gotoLabel b
		; print "}\n")
	 in case t of
	    Halt x => (C.call("MLTON_halt", [Operand.toString x], print)
		       ; print "\n")
	  | Bug => C.bug("cps machine", print)
	  | Return => print "MLTON_return\n"
	  | Raise => print "MLTON_raise\n"
	  | Switch{test, cases, default} =>
	       let val test = Operand.toString test
		  fun bool(t, f) = iff(test, t, f)
		  fun switch(cases, l) =
		     (print "switch ("
		      ; print test
		      ; print ") {\n"
		      ; (List.foreach
			 (cases, fn (n, l) => (print "case "
					       ; print(C.int n)
					       ; print ":\n"
					       ; gotoLabel l)))
		      ; print "default:\n"
		      ; gotoLabel l
		      ; print "\n}\n")
	       in case (cases,   default) of
		  ([],               NONE)   => Error.bug "switch: empty cases"
		| ([(_, l)],         NONE)   => gotoLabel l
		| ([],               SOME l) => gotoLabel l
		| ([(0, f), (1, t)], NONE)   => bool(t, f)
		| ([(1, t), (0, f)], NONE)   => bool(t, f)
		| ((_, l) :: cases', NONE)   => switch(cases', l)
		| (_,                SOME l) => switch(cases, l)
	       end
	  | SwitchIP{test, int, pointer} =>
	       iff(concat["0x3 & (int)", Operand.toString test], int, pointer)
	  | NearJump l => gotoLabel l
	  | FarJump(c, l) =>
	       (C.call("MLTON_farJump",
		       [Chunk.nameToString c, Label.toString l],
		       print)
		; print";\n")
	 end

      val halt = Halt
      val bug = Bug
      val return = Return
      val raisee = Raise
      val switch = Switch
      val switchIP = SwitchIP
      val nearJump = NearJump
      val farJump = FarJump
   end

(* ------------------------------------------------- *)
(*                       Block                       *)
(* ------------------------------------------------- *)

structure Block =
   struct
      datatype t = datatype Types.block
   end

val nonVoidTypes =
   let open Type
   in [char, double, int, uint, pointer]
   end

(* ------------------------------------------------- *)
(*                       Chunk                       *)
(* ------------------------------------------------- *)

structure Chunk =
   struct
      open Chunk

      (* ------------------------------------------------- *)
      (*                 insertLimitChecks                 *)
      (* ------------------------------------------------- *)

      fun numRegsOfType(Chunk{regMax, ...}, ty: Type.t): int = !(regMax ty)

      fun numPointers(c) = numRegsOfType(c, Type.pointer)
					 
      fun insertLimitChecks(self as Chunk{blocks, limitCheckInfo, ...},
			    frames, localMaskName, newIndex): unit =
	 let
	    val gcReturns: Label.t list ref = ref []
	       
	    (* For each pointer register index, tells if the register is used
	     * in some limit check mask, in which case it needs to be copied
	     * to/from the locals array for the GC.  If it is used, localUsed
	     * contains SOME(the index in the locals array).
	     *)
	    val localUsed: int option array = Array.new(numPointers self, NONE)

	    val nextLocal = Counter.new(0)

	    fun localMask(rs: Register.t list): string =
	       localMaskName
	       (List.map
		(rs, fn Register.T{index = i, ty} =>
		 if Type.equals(Type.pointer, ty)
		    then (case Array.sub(localUsed, i) of
			     NONE => 
				let val j = Counter.next nextLocal
				in Array.update(localUsed, i, SOME j)
				   ; j
				end
			   | SOME j => j)
		 else Error.bug "non-pointer in register mask"))
	       
	    fun newFrame{size, liveOffsets} =
	       let val return = Label.newNoname()
	       in newIndex return
		  ; ListRef.push(gcReturns, return)
		  ; Frames.add(frames, {return = return,
					chunk = self,
					size = size,
					liveOffsets = liveOffsets})
		  ; return
	       end

	    val {get = labelBlock, set = setLabelBlock, destroy} =
	       Property.newDest(Label.plist,
				Property.initRaise("block", Label.layout))

	    open Block Transfer
	       
	    val _ = List.foreach(!blocks, fn b as Block{label, ...} =>
				 setLabelBlock(label, b))
	       
	    fun memo(Block{bytesNeeded, statements, transfer, limitCheck, ...})
	       : int =
	       case !bytesNeeded of
		  SOME n => n
		| NONE =>
		     let
			val _ = bytesNeeded := SOME 0

			val goto = memo o labelBlock
			   
			val rest =
			   case transfer of
			      NearJump l => goto l
			    | SwitchIP{int, pointer, ...} =>
				 Int.max(goto int, goto pointer)
			    | Switch{cases, default, ...} =>
				 List.fold
				 (cases, (case default of
					     NONE => 0
					   | SOME l => goto l),
				  fn ((_, j), rest) => Int.max(goto j, rest))
			    | _ => 0
				 
			(* put in the limit check info at array allocations *)
			val bytesAllocated =
			   Array.foldr
			   (statements, rest, fn (statement, bytesAllocated) =>
			    case statement of
			       Statement.AllocateArray
			       {user = {gcInfo = GCInfo.T{frameSize,
							  regs, offsets},
					numElts, numPointers, numBytesNonPointers,
					...},
				limitCheck = r, ...} =>
			       let
				  val bytesPerElt =
				     if numPointers = 0
					then numBytesNonPointers
				     else if numBytesNonPointers = 0
					     then numPointers * pointerSize
					  else Error.unimplemented "tricky arrays"
				  val bytesAllocated =
				     bytesAllocated
				     (* space for array header *)
				     + arrayHeaderSize
				     (* space for forwarding pointer for zero
				      * length arrays *)
				     + pointerSize
			       in case numElts of
				  Operand.Int32 numElts =>
				     bytesAllocated
				     + Type.align(Int32.toInt numElts
						  * bytesPerElt,
						  Type.pointer)
				| _ => 
				     let
					val bytes =
					   concat[Operand.toString numElts,
						  " * ",
						  C.int bytesPerElt,
						  " + ",
						  C.int bytesAllocated]
					val lc =
					   {bytes = bytes,
					    return =
					    newFrame{size = frameSize,
						     liveOffsets = offsets},
					    localIndices = localMask regs}
				     in r := SOME lc
					; 0
				     end
			       end
			      | Statement.Allocate{size, ...} =>
				   size + bytesAllocated + objectHeaderSize
			      | _ => bytesAllocated
				   )
			   
			val n =
			   case limitCheck of
			      NONE => bytesAllocated
			    | SOME{extra,
				   gcInfo = GCInfo.T{regs,
						     offsets,
						     frameSize}} =>
				 (if bytesAllocated = 0
				     then ()
				  else let
					  val return =
					     newFrame{size = frameSize,
						      liveOffsets = offsets}
				       in extra
					  := SOME{bytes = bytesAllocated,
						  localMask = localMask regs,
						  return = return}
				       end
				  ; 0)
		     in bytesNeeded := SOME n
			; n
		     end
	 in List.foreach(!blocks, fn b => (memo b; ()))
	    ; limitCheckInfo := SOME{numLocals = Counter.value nextLocal,
				     localUsed = localUsed,
				     gcReturns = !gcReturns}
	    ; destroy()
	 end

      (* ------------------------------------------------- *)
      (*                      output                       *)
      (* ------------------------------------------------- *)
	 
      fun output(self as Chunk{name, entries, blocks, regMax,
			       limitCheckInfo, ...},
		 print) =
	 let
	    val {numLocals, gcReturns, localUsed} = valOf(! limitCheckInfo)
		     
	    val {get = labelInfo: Label.t -> {block: Block.t,
					      numJumps: int ref,
					      layedOut: bool ref} option,
		 set = setLabelInfo, destroy} =
	       Property.newDest(Label.plist, Property.initConst NONE)
		    
	    fun jump l =
	       case labelInfo l of
		  NONE => () (* When would this happen? *)
		| SOME{numJumps, ...} => IntRef.inc numJumps

	    val _ =
	       let open Block Transfer
	       in List.foreach
		  (!blocks, fn b as Block{label, ...} =>
		   setLabelInfo(label, SOME{block = b,
					    numJumps = ref 0,
					    layedOut = ref false}))
		  ; List.foreach(entries, jump)
		  ; (List.foreach
		     (!blocks, fn Block{transfer, ...} =>
		      case transfer of
			 Switch{cases, default, ...} =>
			    (List.foreach(cases, jump o #2);
			     case default of
				NONE => ()
			      | SOME l => jump l)
		       | SwitchIP{int, pointer, ...} => (jump int; jump pointer)
		       | NearJump l => jump l
		       | _ => ()))
	       end

	    fun gotoLabel l =
	       case labelInfo l of
		  NONE => (print(C.gotoLabel l); print "\n")
	      | SOME{numJumps,
		     block =
		     Block.Block{profileName, limitCheck, statements, transfer,
				 ...},
		     layedOut} =>
		if !layedOut
		   then (print(C.gotoLabel l); print "\n")
		else
		   (layedOut := true
		    ; C.profile(profileName, profileName, print)
		    ; if !numJumps > 1
			 then (print(Label.toString l)
			       ; print ":\n")
		      else ()
		    ; (case limitCheck of
			  NONE => ()
			| SOME{extra,
			       gcInfo = GCInfo.T{frameSize, ...}} =>
			     (case !extra of
				 NONE => ()
			       | SOME{bytes, localMask, return} =>
				    LimitCheck.output
				    ({bytes = C.int bytes,
				      localIndices = localMask,
				      frameSize = frameSize,
				      return = return},
				     print)))
		    ; Array.foreach(statements, fn s =>
				    Statement.output(s, print))
		    ; Transfer.output(transfer, print, gotoLabel))
	    val numPointers = numPointers(self)
	 in
	    C.profile("MLTON_chunk (magic)", overhead, print)
	    ; C.call("MLTON_chunk", [ChunkName.toString name], print)
	    (* standard declarations *)
	    ; print "\npointer locals["
	    ; print(C.int numLocals)
	    ; print "];\n"
	    ; (List.foreach
	       (nonVoidTypes, fn ty =>
		let val n = Type.name ty
		      val l = Type.toString ty
		   in Int.foreach
		      (0, !(regMax ty) - 1, fn i =>
		       (print l
			; print " "
			; print n
			; print(C.int i)
			; print ";\n"))
		   end))
	    (* end declarations *)

	    (* build the entry switch *)
	    ; C.profile("MLTON_chunkSwitch (magic)", overhead, print)
	    ; print "MLTON_chunkSwitch\n"
	    ; List.foreach(entries, fn l =>
			   (C.profile("MLTON_chunkSwitch (magic)", overhead, print)
			    ; print "case "
			    ; print(Label.index l)
			    ; print ":\n"
			    ; gotoLabel l))
	    ; C.profile("MLTON_endChunkSwitch (magic)", overhead, print)
	    ; print "MLTON_endChunkSwitch\n"
		   
	    (* gc *)
	    ; let val pointer = Type.name Type.pointer
		  fun saveOrRestore f =
		     Int.foreach
		     (0, numPointers - 1, fn i =>
		      case Array.sub(localUsed, i) of
			 NONE => ()
		       | SOME j => 
			    let val i = C.int i
			       val r = concat[pointer, i]
			       val l = concat["locals[", C.int j, "]"]
			       val (lhs, rhs) = f(l, r)
			    in C.assign(lhs, rhs, print)
			    end)
	      in C.profile("MLTON_gc (magic)", overhead, print)
		 ; print "MLTON_beforeGC\n"
		 ; saveOrRestore(fn lr => lr)
		 ; print "MLTON_GC\n"
		 ; saveOrRestore(fn (l, r) => (r, l))
		 ; print "switch (*(uint*)stackTop) {\n"
		 ; List.foreach(gcReturns, fn l =>
				(print "case "
				 ; print(Label.index l)
				 ; print ": goto "
				 ; print(Label.toString l)
				 ; print ";\n"))
		 ; print "default: die(\"invalid gcReturn\");\n}\n"
	      end
		  
	    ; print "MLTON_endChunk\n"

	    ; destroy()
	 end

      fun equals(Chunk{blocks = r, ...}, Chunk{blocks = r', ...}) = r = r'
	 
      fun new(entries: Label.t list): t =
	 Chunk{name = ChunkName.new(),
	       entries = entries,
	       blocks = ref [],
	       regMax = Type.memo(fn _ => ref 0),
	       limitCheckInfo = ref NONE}

      fun register(Chunk{regMax, ...}, n, ty) =
	 let val r = regMax ty
	 in r := Int.max(!r, n + 1)
	    ; Register.T{index = n, ty = ty}
	 end
	 
      fun tempRegister(c as Chunk{regMax, ...}, ty) =
	 register(c, !(regMax ty), ty)

      fun newBlock(Chunk{blocks, ...},
		   {profileName, label, statements, transfer, gcInfo}) =
	 ListRef.push
	 (blocks,
	  Block.Block{profileName = profileName,
		      label = label,
		      statements = Array.fromList statements,
		      transfer = transfer,
		      bytesNeeded = ref NONE,
		      limitCheck =
		      Option.map(gcInfo, fn gcInfo =>
				 {gcInfo = gcInfo, extra = ref NONE})})
   end

(* ------------------------------------------------- *)
(*                      Program                      *)
(* ------------------------------------------------- *)

structure Program =
   struct
      datatype t = T of {
			 main: (Chunk.t * Label.t) option ref,
			 saveWorld: (Chunk.t * Label.t) option ref,
			 chunks: Chunk.t list ref,
			 globalCounter: Type.t -> Counter.t,
			 globalPointerNonRootCounter: Counter.t,
			 strings: (Global.t * string) list ref,
			 intInfs: (Global.t * string) list ref,
			 frames: Frames.t,
			 handlers: (Label.t * Chunk.t) list ref,
			 frameOffsetsName: int list -> string,
			 declareFrameOffsets: (string -> unit) -> unit
			 }

      fun setMain(T{main, ...}, c, l) = main := SOME(c, l)
      fun setSaveWorld(T{saveWorld, ...}, c, l) = saveWorld := SOME(c, l)

      fun new() =
	 let
	    val globalCounter = Type.memo(fn _ => Counter.new 0)

	    val {name = frameOffsetsName,
		 declare = declareFrameOffsets} = Offsets.new "frameOffsets"
	 in
	    T{
	      main = ref NONE,
	      saveWorld = ref NONE,
	      chunks = ref [],
	      globalCounter = globalCounter,
	      globalPointerNonRootCounter = Counter.new 0,
	      strings = ref [],
	      intInfs = ref [],
	      frames = ref [],
	      handlers = ref [],
	      frameOffsetsName = frameOffsetsName,
	      declareFrameOffsets = declareFrameOffsets
	     }
	 end

      fun newFrame(T{frames, ...}, f) =
	 Frames.add(frames, f)

      fun newHandler(T{handlers, ...}, l, c) = ListRef.push(handlers, (l, c))

      fun output{program =
		 T{main, saveWorld, chunks, globalCounter, strings, intInfs,
		   handlers, frames, frameOffsetsName,
		   globalPointerNonRootCounter,
		   declareFrameOffsets, ...},
		 includes: string list,
		 print: string -> unit}: unit =
	 let
	    val {get = labelIndex: Label.t -> int, set = setLabelIndex} =
	       Property.new(Label.plist,
			    Property.initRaise("index", Label.layout))

	    val labelCounter = Counter.new 0
	    fun newIndex(l: Label.t) =
	       setLabelIndex(l, Counter.next labelCounter)

	    val {name = localMaskName, declare = declareLocalIndices} =
	       Offsets.new "localIndices"

	    val _ =
	       List.foreach
	       (!chunks, fn c =>
		Chunk.insertLimitChecks(c, frames, localMaskName, newIndex))

	    val maxFrameSize =
	       List.fold(!frames, 0, fn ({size, ...}, n) => Int.max(n, size))

	    (* Assign the entries of a chunk consecutive integers so that gcc
	     * will use a jump table.
	     *)
	    val _ =
	       List.foreach
	       (!chunks, fn Chunk.Chunk{entries, ...} =>
		List.foreach(entries, newIndex))
			     
	    val maxFrameIndex = Counter.value labelCounter

	    local
	       val nextChunks = Array.new(maxFrameIndex, NONE)
	       val frameLayouts = Array.new(maxFrameIndex, NONE)
	       val _ =
		  List.foreach(!handlers, fn (l, c) =>
			       Array.update(nextChunks, labelIndex l,
					    SOME c))
	       val _ =
		  List.foreach
		  (!frames, fn {return, chunk, size, liveOffsets} =>
		   let val index = labelIndex return
		   in Array.update(nextChunks, index, SOME chunk)
		      ; Array.update(frameLayouts, index,
				     SOME{size = size,
					  offsets = frameOffsetsName liveOffsets})
		   end)
	       fun make(name: string, a: 'a array, elt: 'a -> unit) () : unit =
		  (print "static "
		   ; print name
		   ; print " = {"
		   ; Array.foreachi(a, fn (i, x) =>
				    (if i > 0 then print "," else ()
				     ; elt x))
		   ; print "};\n")

	       val bogusFrame = {size = 0, offsets = "NULL"}
	    in
	       val declareNextChunks =
		  make("void ( *nextChunks[])()",
		       nextChunks,
		       fn NONE => print "NULL"
			| SOME chunk =>
			     (print "&";
			      print(ChunkName.toString(Chunk.name chunk))))
		  
	       val declareFrameLayouts =
		  make("GC_frameLayout frameLayouts[]",
		       frameLayouts,
		       fn fo => let
				   val {size, offsets} =
				      case fo of
					 NONE => bogusFrame
				       | SOME f => f
				in print "{"
				   ; print(C.int size)
				   ; print ","
				   ; print offsets
				   ; print "}"
				end)
	    end

	    fun declareIndices() =
	       let
		  fun doit(ls: Label.t list): unit =
		     List.foreach(ls, fn l =>
				  let val index = labelIndex l
				  in print "#define "
				     ; print(Label.index l)
				     ; print " "
				     ; print(C.int index)
				     ; print "\n"
				  end)
	       in List.foreach
		  (!chunks, fn Chunk.Chunk{entries, limitCheckInfo, ...} =>
		   (doit entries;
		    doit(#gcReturns(valOf(!limitCheckInfo)))))
	       end
	    
	    fun declareGlobals() =
	       (print "static pointer globalpointerNonRoot["
		; print(C.int(Counter.value globalPointerNonRootCounter))
		; print "];\n"
		; (List.foreach
		   (nonVoidTypes, fn ty =>
		    let val n = Counter.value(globalCounter ty)
		    in if n = 0
			  then ()
		       else (print "static "
			     ; print(Layout.toString(Type.layout ty))
			     ; print " "
			     ; print(Type.globalName ty)
			     ; print "["
			     ; print(C.int n)
			     ; print "];\n")
		    end)))

	    local
	       fun make(name: string, readOrWrite: string)() =
		  (print "static void "
		   ; print name
		   ; print "(FILE *file) {"
		   ; (List.foreach
		      (nonVoidTypes, fn ty =>
		       let val n = Counter.value(globalCounter ty)
		       in if n = 0
			     then ()
			  else
			     let val name = Type.globalName ty
			     in C.call(readOrWrite,
				       [name,
					concat["sizeof(", Type.toString ty, ")"],
					C.int n,
					"file"],
				       print)
				; print ";\n"
			     end
		       end))
		   ; print "}\n")
	    in val declareSaveGlobals = make("saveGlobals", "swrite")
	       val declareLoadGlobals = make("loadGlobals", "sread")
	    end
	       
            fun declareIntInfs() =
	       (print "static struct intInfRes_t *intInfRes;\n"
		; print "static struct intInfInit intInfInits[] =\n"
		; print "{"
		; List.foreach(!intInfs, fn (g, s) =>
			       (print "{"
				; print(C.int(Global.index g))
				; print ","
				; C.string(s, print)
				; print "},"))
		; print "{ 0, NULL }};\n")

	    fun declareStrings() =
	       (print "static struct GC_stringInit stringInits[] =\n"
		; print "{"
		; List.foreach(!strings, fn (g, s) =>
			       (print "{"
				; print(C.int(Global.index g))
				; print ","
				; C.string(s, print)
				; print ","
				; print(C.int(String.size s))
				; print "},"))
		; print "{ 0, NULL, 0 }};\n")

	    val numGlobals = Counter.value(globalCounter Type.pointer)
	       
	    fun declareMain() =
	       let
		  val stringSizes =
		     List.fold
		     (!strings, 0, fn ((_, s), n)  =>
		      n + arrayHeaderSize
		      + Type.align(size s, Type.pointer))
		  val intInfSizes =
		     List.fold
		     (!intInfs, 0, fn ((_, s), n) =>
		      n + intInfOverhead
		      + Type.align(size s, Type.pointer))
		  val liveSize = intInfSizes + stringSizes
		  val (usedFixedHeap, fromSize) =
		     case !Control.fixedHeap of
			NONE => (false, 0)
		      | SOME n =>
			   (* div 2 for semispace *)
			   (if n > 0 andalso liveSize >= n div 2 
			       then Out.output(Out.error,
					       "Warning: heap size used with -h is too small to hold static data.\n")
			    else ()
			    ; (true, n))
		  val globals = if numGlobals = 0 then "NULL" else "globalpointer"
		  val magic = C.word(MLton.random())
		  val (mainChunk, mainLabel) =
		     case !main of
			NONE => Error.bug "main not set"
		      | SOME(c, l) => (Chunk.nameToString c, Label.toString l)
		  val (saveWorldChunk, saveWorldLabel) =
		     case !saveWorld of
			NONE => ("NULL", mainLabel (* bogus *))
		      | SOME(c, l) => (Chunk.nameToString c, Label.toString l)
	       in C.profile("MLton_main (magic)", overhead, print)
		  ; C.call("MLTON_main",
			   [if usedFixedHeap then C.truee else C.falsee,
			    C.int fromSize,
			    C.int liveSize,
			    C.int maxFrameSize,
			    C.int numGlobals,
			    globals,
			    C.int maxFrameIndex,
			    magic,
			    if isSome(!saveWorld) then "&loadGlobals" else "NULL",
			    C.int(List.length(!intInfs)),
			    mainChunk,
			    mainLabel,
			    saveWorldChunk,
			    saveWorldLabel],
			   print)
	       end

	    fun declareChunks() =
	       List.foreach(!chunks, fn Chunk.Chunk{name, ...} =>
			    (print "static void "
			     ; print(ChunkName.toString name)
			     ; print "();\n"))

	    fun newline() = print "\n"

	 in List.foreach(includes, fn i =>
			 (print "#include \""
			  ; print i
			  ; print "\"\n"))
	    ; print "\nMLTON_globals\n"
	    ; declareIndices()
	    ; declareIntInfs()
	    ; declareStrings()
	    ; declareGlobals()
	    ; declareChunks()
	    ; declareNextChunks()
	    ; declareFrameOffsets print
	    ; declareFrameLayouts()
	    ; declareLocalIndices print
	    ; if isSome(!saveWorld)
		 then (declareSaveGlobals()
		       ; declareLoadGlobals())
	      else ()
	    ; newline()
	    ; List.foreach(!chunks, fn c => Chunk.output(c, print))
	    ; declareMain()
	    ; print "\n"
	 end

      fun newChunk{program = T{chunks, ...}, entries} =
	 let val c = Chunk.new(entries)
	 in ListRef.push(chunks, c)
	    ; c
	 end
	     
      fun newGlobal(T{globalCounter, ...}, ty) =
	 Global.T{index = Counter.next(globalCounter ty), ty = ty}

      fun newGlobalPointerNonRoot(T{globalPointerNonRootCounter, ...}) =
	 Operand.GlobalPointerNonRoot
	 (Counter.next globalPointerNonRootCounter)

      local
	 fun make(sel, ty) (p as T r, s) =
	     let val g = newGlobal(p, ty)
	     in ListRef.push(sel r, (g, s));
		Operand.global g
	     end
      in
	 val newIntInf = make(#intInfs, Type.pointer)
	 val newString = make(#strings, Type.pointer)
      end

      val newGlobal = Operand.global o newGlobal
   end

fun reset() = Offsets.Ints.reset()
   
end
signature FUNCTION =
   sig
      val curry: ('a * 'b -> 'c) -> ('a -> 'b -> 'c)
      val uncurry: ('a -> 'b -> 'c) -> ('a * 'b -> 'c)
      val compose: ('b -> 'c) * ('a -> 'b) -> ('a -> 'c)
      val seq: ('a -> 'b) * ('b -> 'c) -> ('a -> 'c)
      val seq3: ('a -> 'b) * ('b -> 'c) * ('c -> 'd) -> ('a -> 'd)
      val layout: ('a -> 'b) -> Layout.t
      val output: ('a -> 'b) * TextIO.outstream -> unit
   end
structure Function: FUNCTION =
struct

fun curry f x y = f(x, y)

fun uncurry f (x, y) = f x y
   
fun compose(f, g) x = f(g(x))

fun seq(f, g) x = g(f(x))

fun seq3(f, g, h) x = h(g(f(x)))

fun layout _ = Layout.str "<function>"

fun output(_, out) = Out.output(out, "<function>")
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REPRESENTATION_STRUCTS = 
   sig
      structure Cps: CPS
      structure Mtype: MTYPE
   end

signature REPRESENTATION = 
   sig
      include REPRESENTATION_STRUCTS

      structure TyconRep:
	 sig
	    datatype t =
	       Prim of Mtype.t
	     | Enum
	     | EnumDirect
	     | EnumIndirect
	     | EnumIndirectTag
	     | IndirectTag

	    val void: t

	    val toMtype: t -> Mtype.t

	    val layout: t -> Layout.t
	    val equals: t * t -> bool
	 end

      (* How a constructor variant of a datatype is represented. *)
      structure ConRep:
	 sig
	    datatype t =
	       (* need no representation *)
	       Void
	     (* an integer *)
	     | Int of int
	     (* an integer, but of Pointer type *)
	     | IntCast of int
	     (* just keep the value itself *)
	     | Transparent of Mtype.t
	     (* box the arg(s) *)
	     | Tuple
	     (* box the arg(s) and add the integer tag as the first word *)
	     | TagTuple of int

	    val layout: t -> Layout.t
	 end
      
      val compute:
	 Cps.Program.t
	 -> {
	     tyconRep: Cps.Tycon.t -> TyconRep.t,
	     conRep: Cps.Con.t -> ConRep.t,
	     toMtype: Cps.Type.t -> Mtype.t
	    }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Has a special case to make sure that true is represented as 1
 * and false is represented as 0.
 *)
functor Representation(S: REPRESENTATION_STRUCTS): REPRESENTATION = 
struct

open S
local open Cps
in
   structure Con = Con
   structure Tycon = Tycon
   structure Type = Type
end

structure TyconRep =
   struct
      datatype t =
	 (* Datatype contains a single variant, and hence constructor requires no
	  * additional representation.
	  *) 
	 Prim of Mtype.t
       (* All cons are non-value-carrying and are represented as ints. *)
       | Enum
       (* All cons except for one are non-value-carrying and are represented as
	* ints that are nonzero mod 4.  The value carrying con is represented
	* transparently, i.e. the value is known to be a pointer and is left
	* as such.
	*)
       | EnumDirect
       (* All cons except for one are non-value-carrying and are represented as
	* ints that are nonzero mod 4.  The value carrying con is represented
	* by boxing its arg.
	*)
       | EnumIndirect
       (* Non-value-carrying and are represented as ints that are nonzero mod 4.
	* Value carrying cons is represented by boxing the args and adding
	* an integer tag.
	*)
       | EnumIndirectTag
       (* All cons are value carrying and are represented by boxing the args
	* and adding an integer tag.
	*)
       | IndirectTag

      val void = Prim Mtype.void

      val toMtype =
	 fn Prim t => t
	  | Enum => Mtype.int
	  | EnumDirect => Mtype.pointer
	  | EnumIndirect => Mtype.pointer
	  | EnumIndirectTag => Mtype.pointer
	  | IndirectTag => Mtype.pointer

      val layout =
	 let open Layout
	 in fn Prim m => Mtype.layout m
	     | Enum => str "Enum"
	     | EnumDirect => str "EnumDirect"
	     | EnumIndirect => str "EnumIndirect"
	     | EnumIndirectTag => str "EnumIndirectTag"
	     | IndirectTag => str "IndirectTag"
	 end
      
      val equals =
	 fn (Prim t, Prim t') => Mtype.equals(t, t')
	  | (Enum, Enum) => true
	  | (EnumDirect, EnumDirect) => true
	  | (EnumIndirect, EnumIndirect) => true
	  | (EnumIndirectTag, EnumIndirectTag) => true
	  | (IndirectTag, IndirectTag) => true
	  | _ => false
   end

structure ConRep =
   struct
      datatype t =
	 Void
       | Int of int
       | IntCast of int
       | Transparent of Mtype.t
       | Tuple
       | TagTuple of int

      local
	 open Layout
      in
	 val layout =
	    fn Void => str "Void"
	     | Int n => seq[str "Int ", Int.layout n]
	     | IntCast n => seq[str "IntCast ", Int.layout n]
	     | Transparent t => seq[str "Transparent ", Mtype.layout t]
	     | Tuple => str "Tuple"
	     | TagTuple n => seq[str "TagTuple ", Int.layout n]
      end
   end

(* fixed-point.  Initially assume all datatype tycons are Void
 * Change them if they have more than one variant or contain
 * a useful component
 *)

fun compute(Cps.Program.T{datatypes, ...}) =
   let 
      val {get = tyconRep, set = setTyconRep} =
	 Property.new(Tycon.plist, Property.initRaise("rep", Tycon.layout))

(*       val tyconRep =
 * 	 Trace.trace("tyconRep", Tycon.layout, TyconRep.layout) tyconRep
 *)

      val {get = conRep, set = setConRep} =
	 Property.new(Con.plist, Property.initRaise("rep", Con.layout))

      val tyconMtype = TyconRep.toMtype o tyconRep

      fun toMtype t =
	 let open Type
	 in case Type.dest t of
	    Char => Mtype.char
	  | Int => Mtype.int
	  | IntInf => Mtype.pointer
	  | Pointer => Mtype.uint
	  | Word => Mtype.uint
	  | Word8 => Mtype.char
	  | Real => Mtype.double
	  | Stack => Mtype.pointer
	  | String => Mtype.pointer
	  | Array _ => Mtype.pointer
	  | Vector _ => Mtype.pointer
	  | Ref _ => Mtype.pointer
	  | Datatype c => tyconMtype c
	  | Tuple [] => Mtype.void
	  | Tuple _ => Mtype.pointer
	  | Variant(tycon, _) => tyconMtype tycon
	 end

      val _ =
	 List.foreach(datatypes, fn {tycon, ...} =>
		      setTyconRep(tycon, TyconRep.void))

      (* split list of constructors into those that have args and those that don't *)
      fun splitCons cons =
	 List.fold(cons, ([], []), fn ({con, args}, (no, have)) =>
		   case args of
		      [] => (con :: no, have)
		    | _ =>
			 if Mtype.isVoid(toMtype(Type.tuple args))
			    then (con :: no, have)
			 else (no, {con = con, args = args} :: have))
	 
   in
      FixedPoint.fixedPoint
      (fn continue =>
       List.foreach
       (datatypes, fn {tycon, cons} =>
	let
	   val (noArgs, haveArgs) = splitCons cons
	   val old = tyconRep tycon
	   val new =
	      case (noArgs, haveArgs) of
		   ([],     [])           => TyconRep.void
	         | ([_],    [])           => TyconRep.void
	         | (_,      [])           => TyconRep.Enum
	         | ([],     [{args, ...}]) => TyconRep.Prim(toMtype
							    (Type.tuple args))
		 | (_,      [{args, ...}]) =>
		      if (case args of
			     [a] => let open Type
				    in case Type.dest a of
				       Datatype c =>
					  (case tyconRep c of
					      TyconRep.IndirectTag => true
					    | _ => false)
				     | String => true
				     | Array _ => true
				     | Vector _ => true
				     | Ref _ => true
				     | Tuple _ => true
				     | _ => false
				    end
			   | _ => true)
			 then TyconRep.EnumDirect
		      else TyconRep.EnumIndirect
		 | ([],     _)            => TyconRep.IndirectTag
		 | _                      => TyconRep.EnumIndirectTag
	in if TyconRep.equals(old, new)
	      then ()
	   else (continue();
		 setTyconRep(tycon, new))
	end));

      (* set constructor representations *)
      let
	 fun direct(haveArgs, t) =
	    case haveArgs of
	       [{con, args}] => setConRep(con,
					  (case args of
					      [_] => ConRep.Transparent t
					    | _ => ConRep.Tuple))
	     | _ => ()
		  
	 (* choose tags that are not equal to 0 mod 4 *)
	 fun enum noArgs =
	    let
	       fun loop(i, cs) =
		  case cs of
		     [] => ()
		   | c :: cs => (setConRep(c, ConRep.IntCast i);
				 loop(i + 2, cs))
	    in loop(1, noArgs)
	    end
	 fun indirectTag haveArgs = 
	    List.foreachi(haveArgs, fn (i, {con, args}) =>
			  setConRep(con, ConRep.TagTuple i))
      in List.foreach
	 (datatypes, fn {tycon, cons} =>
	  let val (noArgs, haveArgs) = splitCons cons
	  in case tyconRep tycon of
	     TyconRep.Prim t =>
		(case (noArgs, haveArgs) of
		    ([], []) => ()
		  | ([con], []) => setConRep(con, ConRep.Void)
		  | ([], [_]) => direct(haveArgs, t)
		  | _ => Error.bug("strange TyconRep.Prim for "
				   ^ Layout.toString(Tycon.layout tycon)))
	   | TyconRep.Enum =>
		if Tycon.equals(tycon, Tycon.bool)
		   then (setConRep(Con.falsee, ConRep.Int 0) ;
			 setConRep(Con.truee, ConRep.Int 1))
		else List.foreachi(noArgs, fn (i, c) =>
				   setConRep(c, ConRep.Int i))
	   | TyconRep.EnumDirect =>
		(enum noArgs ; direct(haveArgs, Mtype.pointer))
	   | TyconRep.EnumIndirect =>
		(enum noArgs ;
		 List.foreach(haveArgs, fn {con, ...} =>
			      setConRep(con, ConRep.Tuple)))
	   | TyconRep.EnumIndirectTag => (enum noArgs ; indirectTag haveArgs)
	   | TyconRep.IndirectTag => indirectTag haveArgs
	  end)
      end ;

   {tyconRep = tyconRep,
    conRep = conRep,
    toMtype = toMtype}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PARALLEL_MOVE_STRUCTS = 
   sig
   end

signature PARALLEL_MOVE = 
   sig
      include PARALLEL_MOVE_STRUCTS

      (* Allows overlapping froms and tos.
       * Hence, has to be careful to use temps.
       * 'a is the type of "registers"
       * 'b is the type of "statements"
       *)
      val move:
	 {
	  moves: {src: 'a, dst: 'a} list,
	  (* are two registers the same *)
	  equals: 'a * 'a -> bool,
	  (* create a move statement *)
	  move: {src: 'a, dst: 'a} -> 'b,
	  (* return a new temporary register like input reg *)
	  temp: 'a -> 'a
	 } -> 'b list
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ParallelMove(S: PARALLEL_MOVE_STRUCTS): PARALLEL_MOVE = 
struct

open S

fun ('reg, 'statement) move{moves, equals, move, temp} : 'statement list =
   let
      val mvs =
	 List.fold(moves, [], fn (mv as {src, dst}, mvs) =>
		   if equals(src, dst)
		      then mvs
		   else mv :: mvs)
      fun loopTop(mvs, moves) = loop(mvs, [], moves, false)
      and loop(mvs, hard, moves, changed) =
	 case mvs of
	    [] =>
	       (case hard of
		   [] => List.reverse moves
		 | {src, dst} :: hard' => 
		      if changed
			 then loopTop(hard, moves)
		      else
			 let
			    val temp = temp src
			    val moves =
			       move{src = src, dst = dst}
			       :: move{dst = temp, src = dst}
			       :: moves
			    val hard =
			       List.map(hard', fn mv as {src = s, dst = d} =>
					if equals(dst, s)
					   then {src = temp, dst = d}
					else mv)
			 in loopTop(hard, moves)
			 end)
	  | (mv as {src, dst}) :: mvs =>
	       let fun isHard l = List.exists(l, fn {src, dst = _} =>
					      equals(dst, src))
	       in if isHard mvs orelse isHard hard
		     then loop(mvs, mv :: hard, moves, changed)
		  else loop(mvs, hard,
			    move{src = src, dst = dst} :: moves,
			    true)
	       end
   in loopTop(mvs, [])
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LIMIT_CHECK_STRUCTS = 
   sig
      include CPS
   end

signature LIMIT_CHECK = 
   sig
      include LIMIT_CHECK_STRUCTS

      (* Determine at which jumps limit checks should be inserted. *)
      val limitCheck: Program.t -> Jump.t -> bool
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor LimitCheck(S: LIMIT_CHECK_STRUCTS): LIMIT_CHECK = 
struct

open S
open Dec Transfer
   
fun limitCheck(Program.T{functions, ...}) =
   let
      val {get = limitCheck, set = setLimitCheck} =
	 Property.new(Jump.plist, Property.initConst false)

      val {get = inBody, set = setInBody, destroy} =
	 Property.newDest(Jump.plist, Property.initConst false)

      fun loopExp e =
	 let val {decs, transfer} = Exp.dest e
	 in List.foreach(Exp.decs e,
			 fn Fun{name, body, ...} => (setInBody(name, true)
						     ; loopExp body
						     ; setInBody(name, false))
			  | HandlerPush h => setLimitCheck(h, true)
			  | _ => ())
	    ; (case transfer of
		  Jump{dst, ...} => (if inBody dst
					then setLimitCheck(dst, true)
				     else ())
		| Call{cont, ...} => (case cont of
					 SOME c => setLimitCheck(c, true)
				       | _ => ())
		| _ => ())
	 end
   in
      List.foreach(functions, loopExp o #body)
      ; destroy()
      ; limitCheck
   end
   
end


(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature EQUIVALENCE_GRAPH_STRUCTS = 
   sig
   end

signature EQUIVALENCE_GRAPH = 
   sig
      include EQUIVALENCE_GRAPH_STRUCTS
      
      structure Class:
	 sig
	    (* The type of equivalence classes. *)
	    type t

	    val plist: t -> PropertyList.t
	 end

      (* The type of directed graphs with equivalence relations on nodes. *)
      type t

      (* Return a new graph. *)
      val new: unit -> t

      (* newNode(g, i) adds a new node to graph g, where the size of the node
       * is i.  The new node is not equivalent to any other node.  Return the
       * class of the node.
       *)
      val newClass: t * int -> Class.t

      (* Add a new edge between two classes.
       * Increment the weight of the edge if it's already there.
       *)
      val addEdge: t * {from: Class.t, to: Class.t} -> unit

      (* Make two classes equivalent.
       * The size of the resulting class is the sum of the sizes of the original
       * two classes.  This is a noop if the classes are already equal.
       *)
      val == : t * Class.t * Class.t -> unit

      (* Make the equivalence relation as coarse as possible so that the
       * number of edges between classes in minimized, subject to the constraint
       * that the sum of the node sizes in an equivalence class is <= maxNodeSize.
       * Classes for which this constraint was violated by previous calls to ==
       * should not be made coarser.
       *)
      val greedy: {graph: t, maxClassSize: int} -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor EquivalenceGraph(S: EQUIVALENCE_GRAPH_STRUCTS): EQUIVALENCE_GRAPH = 
struct

open S

structure Set = DisjointSet
structure Plist = PropertyList

(* A simple implementation where greedy doesn't do anything *)
structure Class =
   struct
      type t = Plist.t Set.t
      val plist = Set.value
   end
type t = unit
fun new() = ()
fun newClass _ = Set.singleton(PropertyList.new())
fun addEdge _ = ()
fun ==(_, c, c') = Set.union(c, c')
fun greedy _ = ()

(* A slightly trickier implementation where greedy just walks over the edges
 * in order.
 *)
structure Class =
   struct
      datatype t = T of {size: int ref,
			 plist: Plist.t} Set.t

      local
	 fun make sel (T s) = sel(Set.value s)
      in val plist = make #plist
	 val size = make (! o #size)
      end

      fun setSize(T s, n) = #size(Set.value s) := n

      fun new(size: int): t = T(Set.singleton{size = ref size,
					      plist = Plist.new()})

      fun ==(c as T s, T s') =
	 if Set.equals(s, s')
	    then ()
	 else let val {size = ref n, plist} = Set.value s
		  val {size = ref n', plist} = Set.value s'
	      in Set.union(s, s')
		 ; setSize(c, n + n')
	      end
   end

datatype t = T of {edges: (Class.t * Class.t) list ref}

fun new() = T{edges = ref []}

fun newClass(_, n) = Class.new n

fun addEdge(T{edges, ...}, {from, to}) =
   ListRef.push(edges, (from, to))

fun ==(_, c, c') = Class.==(c, c')

fun greedy{graph = T{edges, ...}, maxClassSize} =
   List.foreach(!edges, fn (c, c') =>
		if Class.size c + Class.size c' <= maxClassSize
		   then Class.==(c, c')
		else ())

(*
 * Given an edge, return how desirable it is to merge the endpoints
 * of the edge.  The result is an int option because we return
 * NONE if they are not mergable.
 * Note, it looks at the details inside Class, but the whole thing
 * is just a hack.
 *)
fun goodness (Class.T lhs: Class.t, Class.T rhs: Class.t): int option =
       if Set.equals (lhs, rhs)
	  then NONE
	  else let val {size = ref lsize, ...} = Set.value lhs
		   val {size = ref rsize, ...} = Set.value rhs
	       in SOME (~ (lsize + rsize))
	       end

fun findBest (edges: (Class.t * Class.t) list)
	     : (Class.t * Class.t) option =
       let fun folder (e: Class.t * Class.t,
                       ac: (int * (Class.t * Class.t)) option) =
                  case goodness e of
                     NONE => ac
                     | SOME g =>
                          case ac of
                             NONE => SOME (g, e)
                             | SOME (g', _) =>
                                  if g > g'
                                     then SOME (g, e)
                                     else ac
       in case List.fold (edges, NONE, folder) of
	     NONE => NONE
	     | SOME (goodness, e) => (
(* 					print ("\nHCC:\tgoodness " ^
 * 					       Int.toString goodness ^
 * 					       "\n");
 *)
					SOME e
				     )
       end

fun greedy' {graph = T {edges, ...}, maxClassSize} =
       let fun loop () =
		  case findBest (! edges) of
		     NONE => ()
		     | SOME (lhs, rhs) =>
			  if Class.size lhs + Class.size rhs <= maxClassSize
			     then (
				     Class.== (lhs, rhs);
				     loop ()
				  )
			     else ()
       in loop ()
       end

end

structure EquivalenceGraph = EquivalenceGraph()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CHUNKIFY_STRUCTS = 
   sig
      include CPS
   end

signature CHUNKIFY = 
   sig
      include CHUNKIFY_STRUCTS

      (* Partitions all the labels declared into disjoint sets, referred
       * to as chunks.  Returns the list of chunks.
       * All funcs, conts, and handlers are assumed to be entry points.
       * All conts and handlers are assumed to be return points.
       *)
      val chunkify:
	 {program: Program.t,
	  jumpHandlers: Jump.t -> Jump.t list}
	 -> {
	     funcs: Func.t list,
	     jumps: Jump.t list
	     } list
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Chunkify(S: CHUNKIFY_STRUCTS): CHUNKIFY = 
struct

open S
open Dec PrimExp Transfer

(* A chunkifier that puts each function in its own chunk.
 *)
fun chunkPerFunc(Program.T{functions, ...}) =
   List.map
   (functions, fn {name, body, ...} =>
    let val jumps = ref []
    in Exp.foreachDec(body,
		      fn Fun{name, ...}=> ListRef.push(jumps, name)
		       | _ => ())
       ; {funcs = [name],
	  jumps = !jumps}
    end)

(* A simple chunkifier that puts all code in the same chunk.
 *)
fun oneChunk(Program.T{functions, main, ...}) =
   let val funcs = ref []
      val jumps = ref []
   in List.foreach(functions, fn {name, body, ...} =>
		   (ListRef.push(funcs, name)
		    ; (Exp.foreachDec
		       (body,
			fn Fun{name, ...} => ListRef.push(jumps, name)
			 | _ => ()))))
      ; [{funcs = !funcs, jumps = !jumps}]
   end

structure Label = Jump
structure Set = DisjointSet

(* The size of an expression, not including nested functions. *)

fun size(e: Exp.t) =
   let
      val {decs, transfer} = Exp.dest e
      val decsSize =
	 List.fold(decs, 0, fn (d, n) => case d of Fun _ => n | _ => n + 1)
      val transferSize =
	 case transfer of
	    Case{cases, ...} => 1 + List.length cases
	  | _ => 1
   in decsSize + transferSize
   end

(* Compute the list of functions that each function returns to *)
fun returnsTo(Program.T{functions, ...}) =
   let
      val {get: Func.t -> {returnsTo: Jump.t list ref,
			   tailCalls: Func.t list ref},
	   destroy, ...} =
	 Property.newDest(Func.plist, Property.initFun(fn _ =>
						       {returnsTo = ref [],
							tailCalls = ref []}))

      fun returnTo(f: Func.t, j: Jump.t): unit =
	 let val {returnsTo, tailCalls} = get f
	 in if List.exists(!returnsTo, fn j' => Jump.equals(j, j'))
	       then ()
	    else (ListRef.push(returnsTo, j)
		  ; List.foreach(!tailCalls, fn f => returnTo(f, j)))
	 end
      
      fun tailCall(from: Func.t, to: Func.t): unit =
	 let val {returnsTo, tailCalls} = get from
	 in if List.exists(!tailCalls, fn f => Func.equals(to, f))
	       then ()
	    else (ListRef.push(tailCalls, to)
		  ; List.foreach(!returnsTo, fn j => returnTo(to, j)))
	 end

      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  Exp.foreachTransfer
	  (body,
	   fn Call{func, cont, ...} => (case cont of
					   NONE => tailCall(name, func)
					 | SOME j => returnTo(func, j))
	    | _ => ()))
	 
   in {returnsTo = ! o #returnsTo o get,
       destroy = fn () => ()}
   end

structure Graph = EquivalenceGraph
structure Class = Graph.Class
fun coalesce(program as Program.T{functions, ...}, jumpHandlers, limit) =
   let
      val graph = Graph.new()
	 
      val {get = funcClass: Func.t -> Class.t, set = setFuncClass,
	   destroy = destroyFuncClass} =
	 Property.newDest(Func.plist, Property.initRaise("class", Func.layout))

      val {get = jumpClass: Jump.t -> Class.t, set = setJumpClass,
	   destroy = destroyJumpClass} =
	 Property.newDest(Jump.plist, Property.initRaise("class", Jump.layout))

      fun 'a newClass(name: 'a, exp: Exp.t, setClass: 'a * Class.t -> unit)
	 : Class.t =
	 let val class = Graph.newClass(graph, size exp)
	 in setClass(name, class)
	    ; class
	 end

      (* Build the initial partition.
       * Ensure that all Cps jumps are in the same equivalence class.
       *)
      fun loopExp(e: Exp.t, n: Class.t, hs: Jump.t list): unit =
	 let val {decs, transfer} = Exp.dest e
	    fun same(j: Jump.t): unit = Graph.==(graph, n, jumpClass j)
	 in List.foreach
	    (decs,
	     fn Fun{name, body, ...} =>
	          loopExp(body, newClass(name, body, setJumpClass),
			  jumpHandlers name)
	      | _ => ())
	    ; (case transfer of
		  Jump{dst, ...} => same dst
		| Case{cases, default, ...} =>
		     (List.foreach(cases, same o #2)
		      ; Option.map'(default, same))
		| Raise _ => (case List.fold(decs, hs, deltaHandlers) of
				 [] => ()
			       | h :: _ => same h)
		| _ => ())
	 end

      val _ =
	 List.foreach(functions, fn {name, body, ...} =>
		      loopExp(body, newClass(name, body, setFuncClass), []))

      val {returnsTo, destroy = destroyReturnsTo} = returnsTo program

      (* Add edges, and then coalesce the graph. *)
      val _ =
	 List.foreach
	 (functions, fn {name, body, ...} =>
	  let
	     val returnsTo = List.map(returnsTo name, jumpClass)
	     fun loopExp(e: Exp.t, class: Class.t): unit =
		let val {decs, transfer} = Exp.dest e
		in List.foreach
		   (decs,
		    fn Fun{name, body, ...} => loopExp(body, jumpClass name)
		     | _ => ())
		   ; (case transfer of
			 Call{func, cont, ...} =>
			    Graph.addEdge(graph, {from = class,
						  to = funcClass func})
		       | Return _ =>
			     List.foreach(returnsTo, fn c =>
					  Graph.addEdge(graph, {from = class,
								to = c}))
		       | _ => ())
		end
	  in loopExp(body, funcClass name)
	  end)
	 
      val _ =
	 if limit = 0
	    then ()
	 else Graph.greedy{graph = graph, maxClassSize = limit}

      type chunk = {funcs: Func.t list ref,
		    jumps: Jump.t list ref}

      val chunks: chunk list ref = ref []

      val {get = classChunk: Class.t -> chunk, destroy = destroyClassChunk, ...} =
	 Property.newDest
	 (Class.plist,
	  Property.initFun(fn _ =>
			   let val c = {funcs = ref [],
					jumps = ref []}
			   in ListRef.push(chunks, c)
			      ; c
			   end))

      val _ =
	 let
	    fun 'a new(l: 'a, get: 'a -> Class.t, sel: chunk -> 'a list ref): unit =
	       ListRef.push(sel(classChunk(get l)), l)
	 in List.foreach
	    (functions, fn {name, body, ...} =>
	     (new(name, funcClass, #funcs)
	      ; Exp.foreachDec(body,
			       fn Fun{name, ...} => new(name, jumpClass, #jumps)
				| _ => ())))
	 end

   in destroyFuncClass()
      ; destroyJumpClass()
      ; destroyClassChunk()
      ; destroyReturnsTo()
      ; List.map(!chunks, fn {funcs, jumps} =>
		 {funcs = !funcs,
		  jumps = !jumps})
   end
   
fun chunkify{program, jumpHandlers} =
   case !Control.chunk of
       Control.ChunkPerFunc => chunkPerFunc program
     | Control.OneChunk => oneChunk program
     | Control.Coalesce{limit} => coalesce(program, jumpHandlers, limit)
       
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature BACKEND_STRUCTS = 
   sig
      structure Cps: CPS
      structure Machine: MACHINE

      val funcToLabel: Cps.Func.t -> Machine.Label.t
      val jumpToLabel: Cps.Jump.t -> Machine.Label.t

      sharing Cps.Jump = Machine.Label
      sharing Cps.Const.SmallIntInf = Machine.SmallIntInf
      sharing Cps.Prim = Machine.Prim
   end

signature BACKEND = 
   sig
      include BACKEND_STRUCTS
      
      val generate: Cps.Program.t -> Machine.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LIVE_STRUCTS = 
   sig
      structure Cps: CPS
   end

signature LIVE = 
   sig
      include LIVE_STRUCTS
      
      type t

      val live:
	 {
	  exp: Cps.Exp.t,
	  shouldConsider: Cps.Var.t -> bool,
	  jumpHandlers: Cps.Jump.t -> Cps.Jump.t list
	 }
	 -> {
	     (* live variables at beginning of block *)
	     liveBegin: Cps.Jump.t -> Cps.Var.t list,

	     (* live variables at the beginning of a block, excepting its formals
	      *)
	     liveBeginNoFormals: Cps.Jump.t -> Cps.Var.t list,
	     
	     (* live variables at array allocation *)
	     liveArray: Cps.Var.t -> Cps.Var.t list,

	     destroy: unit -> unit
	     }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Only compute lives where limit checks will be inserted:
 *  - function entry
 *  - continuation entry
 *  - array allocation
 *  - loop entry
 *
 * Use recursion for loops.
 *)
functor Live(S: LIVE_STRUCTS): LIVE = 
struct

open S
open Cps
open Dec PrimExp Transfer

structure VarStatus =
   struct
      datatype t = Unseen | Free | Bound

      val toString =
	 fn Unseen => "Unseen"
	  | Free => "Free"
	  | Bound => "Bound"

      val layout = Layout.str o toString
   end
datatype status = datatype VarStatus.t

structure JumpStatus =
   struct
      datatype t = Normal | InBody | Recursive

      val toString =
	 fn Normal => "Normal"
	  | InBody => "InBody"
	  | Recursive => "Recursive"

      val layout = Layout.str o toString
   end
datatype status = datatype JumpStatus.t

fun live{exp, shouldConsider, jumpHandlers} =
   let
      val {get = jumpInfo: Jump.t -> {frees: Var.t list ref,
				      freesNoFormals: Var.t list ref,
				      isCont: bool ref,
				      status: JumpStatus.t ref},
	   set = setJumpInfo,
	   destroy = destroyJumpInfo} =
	 Property.newDest
	 (Jump.plist, Property.initFun(fn _ => {frees = ref [],
						freesNoFormals = ref [],
						isCont = ref false,
						status = ref InBody}))

      fun getFreesNoFormals j = !(#freesNoFormals(jumpInfo j))

      val {get = liveArray, set = setLiveArray,
	   destroy = destroyLiveArray} =
	 Property.newDest(Var.plist, Property.initRaise("liveArray", Var.layout))

      val {get = varInfo: Var.t -> {status: VarStatus.t ref,
				    frees: Var.t list ref ref(* used as a unique id *)},
	   destroy = destroyVarInfo, ...} =
	 Property.newDest(Var.plist,
			  Property.initFun(let val bogus = ref []
					   in fn _ => {status = ref Unseen,
						       frees = ref bogus}
					   end))

      fun newScope(frees, new, th) =
	 let
	    val changed = ref false
	    val all = ref []
	    fun statusRef x =
	       let val {frees = frees', status} = varInfo x
	       in if frees = !frees'
		     then ()
		  else (ListRef.push
			(all, (frees', !frees', status, !status))
			; frees' := frees; status := Unseen)
		  ; status
	       end
	    fun bindVar x =
	       if shouldConsider x
		  then statusRef x := Bound
	       else ()
	    fun var x =
	       if shouldConsider x
		  then let val s = statusRef x
		       in case !s of
			  Unseen =>
			     (Assert.expensive
			      (fn () => not(List.exists(!frees, fn y =>
							Var.equals(x, y))))
			      ; changed := true
			      ; s := Free
			      ; new x
			      ; ListRef.push(frees, x))
			| _ => ()
		       end
	       else ()
	 in List.foreach(!frees, fn x => statusRef x := Free)
	    ; th(var, bindVar, changed)
	    ; List.foreach(!all, fn (r, v, r', v') => (r := v; r' := v'))
	 end
	 
      fun loopExp(e: Exp.t, handlers: Jump.t list, var, bindVar) =
	 let
	    val {decs, transfer} = Exp.dest e
	    fun transHandlers() = List.fold(decs, handlers, deltaHandlers)
	    fun loopDecs(decs, var, bindVar) =
	       let fun vars xs = List.foreach(xs, var)
		  fun jump j =
		     let val {status, freesNoFormals, ...} = jumpInfo j
		     in vars(!freesNoFormals)
			; (case !status of
			      InBody => status := Recursive
			    | _ => ())
		     end
	       in case decs of
		  [] => (case transfer of
			    Jump{dst, args, ...} => (vars args; jump dst)
			  | Call{args, cont, ...} =>
			       (vars args; Option.map'(cont, jump))
			  | Return xs => vars xs
			  | Raise xs =>
			       (vars xs
				; (case transHandlers() of
				      h :: _ => jump h
				    | _ => ()))
			  | Case{test, cases, default, ...} =>
			       (var test
				; Option.map'(default, jump)
				; List.foreach(cases, jump o #2))
			  | Bug => ()
			  | Halt => ())
		| dec :: decs =>
		     let fun continue() = loopDecs(decs, var, bindVar)
		     in case dec of
			HandlerPush _ => continue()
		      | HandlerPop => continue()
		      | Bind{var = x, exp, ...} =>
			   let fun normal() = (bindVar x; continue())
			   in case exp of
			      Const _ => normal()
			    | Var y => (var y; normal())
			    | Tuple xs => (vars xs; normal())
			    | Select{tuple, ...} => (var tuple; normal())
			    | ConSelect{variant, ...} => (var variant; normal())
			    | ConApp{args, ...} => (vars args; normal())
			    | PrimApp{prim, args, ...} =>
				 (vars args
				  ; (case Prim.name prim of
					Prim.Name.Array =>
					   let val frees = ref []
					      fun new _ = ()
					   in newScope
					      (frees, new, fn (var, bindVar, _) =>
					       (bindVar x
						; loopDecs(decs, var, bindVar)))
					      ; List.foreach(!frees, var)
					      ; setLiveArray(x, !frees)
					   end
				      | _ => normal()))
			   end
		      | Fun{name, args, body} =>
			   (let
			       val {frees, freesNoFormals, status, isCont, ...} =
				  jumpInfo name
			       val handlers = jumpHandlers name
			       val {get, set, destroy} =
				  Property.newDest
				  (Var.plist, Property.initConst false)
			       fun new x =
				  if get x then ()
				  else ListRef.push(freesNoFormals, x)
			    in List.foreach(args, fn (x, _) => set(x, true))
			       ; (newScope
				  (frees, new, fn (var, bindVar, changed) =>
				   (if !isCont
				       then (* in case there is a raise to here *)
					  (case handlers of
					      h :: _ =>
						 List.foreach
						 (getFreesNoFormals h, var)
					    | _ => ())
				    else ()
				    ; let
					 fun loop() =
					    (changed := false
					     ; loopExp(body, handlers, var, bindVar)
					     ; if !status = Recursive andalso !changed
						  then loop()
					       else ())
				      in status := InBody
					 ; loop()
					 ; status := Normal
				      end)))
			       ; destroy()
			    end
			    ; continue())
		     end
	       end
	 in loopDecs(decs, var, bindVar)
	 end
      fun ignore _ = ()
      val handleTransfer =
	 fn Call{cont, ...} => (case cont of
				   NONE => ()
				 | SOME c => #isCont(jumpInfo c) := true)
	  | _ => ()
   in Exp.foreachTransfer(exp, handleTransfer)
      ; loopExp(exp, [], ignore, ignore)
      ; {
	 liveBegin = ! o #frees o jumpInfo,
	 liveBeginNoFormals = getFreesNoFormals,
	 liveArray = liveArray,
	 destroy = fn () => (destroyJumpInfo()
			     ; destroyLiveArray()
			     ; destroyVarInfo())
	 }
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 *
 *)
signature ALLOCATE_REGISTERS_STRUCTS = 
   sig
      structure Cps: CPS
      structure Machine: MACHINE
   end

signature ALLOCATE_REGISTERS = 
   sig
      include ALLOCATE_REGISTERS_STRUCTS

      structure Info:
	 sig
	    datatype t =
	       T of {
		     (* Where do variables defined and/or used in this block
		      * reside. Will only be called on variables x such that
		      * shouldAllocate x. 
		      *
		      * Funcnames have the restriction that you can't assign
		      * stack offset operands to formal parameters since the
		      * stack pointer won't be set until after the parameter
		      * passing. Must use a register.
		      *)
		      varOperand: Cps.Var.t -> Machine.Operand.t,

		      gcInfo: Machine.GCInfo.t,
		      
		      (* Appear at the beginning of the block. *)
		      prelude: Machine.Statement.t list,
		      
		      (* Appear before the transfer at the end of the block. *)
		      postlude: Machine.Statement.t list
		      }
	 end
      
      (* Use Machine.Chunk.newRegister with the appropriate chunk to get a new
       * register.
       *)
      val allocate:
	 {
	  program: Cps.Program.t,

	  varType: Cps.Var.t -> Machine.Type.t,

	  funcChunk: Cps.Func.t -> Machine.Chunk.t,
	  jumpChunk: Cps.Jump.t -> Machine.Chunk.t,

	  jumpHandlers: Cps.Jump.t -> Cps.Jump.t list,

	  (* Whether or not a stack slot/register needs to be allocated for a
	   * given variable
	   *)
	  shouldAllocate: Cps.Var.t -> bool
	 }
	 ->
	 {
          (* GCInfo at the binding of a var to an array allocation. *)
	  arrayInfo: Cps.Var.t -> Machine.GCInfo.t,

	  funcInfo:
	  Cps.Func.t -> {
			 info: Info.t,
			 (* If handlers are used, this gives the stack offset
			  * where the old exnStack and handler should be stored
			  *)
			 handlerOffset: int option
			 },
	  
	  contInfo:
	  Cps.Jump.t -> {
			 (* number of bytes in frame, including return address *)
			 size: int,
			 (* offsets of live pointers in frame *)
			 liveOffsets: int list
			 },

	  handlerInfo:
	  Cps.Jump.t -> {
			 (* amount to subtract to get to the next frame *)
			 size: int,
			 (* offsets of live pointers in frame *)
			 liveOffsets: int list
			 },

	  jumpInfo: Cps.Jump.t -> Info.t
	  }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AllocateRegisters(S: ALLOCATE_REGISTERS_STRUCTS): ALLOCATE_REGISTERS = 
struct

open S

local open Cps
in
   structure Dec = Dec
   structure Exp = Exp
   structure Func = Func
   structure Jump = Jump
   structure Prim = Prim
   structure PrimExp = PrimExp
   structure Program = Program
   structure Ctransfer = Transfer
   structure Var = Var
end

local open Machine
in
   structure Chunk = Chunk
   structure GCInfo = GCInfo
   structure Operand = Operand
   structure Mtype = Type
   structure Register = Register
   structure Statement = Statement
end

(* If a handler is stored in a stack frame, then we need both a uint for
 * the old handler and space for the handler itself
 *)
local
   open Mtype
in
   val labelSize = size label
   val handlerSize = labelSize + size uint
end

(* Compute live variables *)
structure Live = Live(structure Cps = Cps)

(*
 * Every variable that is ever live across a nontail call or in a handler
 * gets a stack slot.  Otherwise, it gets a register.
 * Only exception is for formals, which must be made registers.
 * For these, copy them into stack slot at beginning of block.
 *)

structure Info =
   struct
      datatype t =
	 T of {
	       varOperand: Cps.Var.t -> Machine.Operand.t,
	       gcInfo: GCInfo.t,
	       prelude: Machine.Statement.t list,
	       postlude: Machine.Statement.t list
	       }
   end 

(* ------------------------------------------------- *)
(*                     allocate                      *)
(* ------------------------------------------------- *)

fun allocate{program = Program.T{globals, functions, ...},
	     varType, jumpChunk, funcChunk, shouldAllocate, jumpHandlers} =
   let
      val {get = varOperand, set = setVarOperand} =
	 Property.new(Var.plist, Property.initRaise("operand", Var.layout))

(*       val varOperand =
 * 	 Trace.trace("varOperand", Var.layout, Operand.layout) varOperand
 * 	 
 *       val setVarOperand =
 * 	 Trace.trace2("setVarOperand",
 * 		      Var.layout, Operand.layout, Unit.layout) setVarOperand
 *)
      val {get = jumpInfo: Jump.t -> Info.t, set = setJumpInfo} =
	 Property.new(Jump.plist, Property.initRaise("jump info", Jump.layout))

      val {get = contInfo: Jump.t ->  {
				       size: int,
				       liveOffsets: int list
				       },
	   set = setContInfo} =
	 Property.new(Jump.plist, Property.initRaise("cont info", Jump.layout))

      val {get = handlerInfo: Jump.t ->  {
					  size: int,
					  liveOffsets: int list
					  },
	   set = setHandlerInfo} =
	 Property.new(Jump.plist, Property.initRaise("handler info", Jump.layout))

      val {get = funcInfo, set = setFuncInfo} =
	 Property.new(Func.plist, Property.initRaise("func info", Func.layout))

      val {get = liveArray, set = setLiveArray} =
	 Property.new(Var.plist, Property.initRaise("liveArray", Var.layout))

      fun allocateFunc(name: Func.t,
		       args: (Var.t * Cps.Type.t) list,
		       body: Exp.t): unit = 
	 let
	    val {liveBegin, liveBeginNoFormals, liveArray = liveArrayVars,
		 destroy = destroyLive} =
	       Live.live{exp = body,
			 shouldConsider = shouldAllocate,
			 jumpHandlers = jumpHandlers}

	    (*
	     * Decide which variables will live in stack slots and which
	     * will live in registers.
	     * Initially, try to put everything in a register.
	     * Variables get moved to the stack if they are live in a handler
	     * or live across a nontail call.
	     *)
	    datatype place = Stack | Register
	    val {get = place: Var.t -> place,
		 set = setPlace, destroy = destroyPlace} =
	       Property.newDest(Var.plist, Property.initConst Register)

	    val {get = isCont: Jump.t -> bool,
		 set = setIsCont, destroy = destroyIsCont} =
	       Property.newDest(Jump.plist, Property.initConst false)

	    val {get = isHandler: Jump.t -> bool,
		 set = setIsHandler, destroy = destroyIsHandler} =
	       Property.newDest(Jump.plist, Property.initConst false)

	    (* Are handlers ever installed in this function? *)
	    val hasHandler: bool ref = ref false

	    fun forceStack(j: Jump.t): unit =
	       List.foreach(liveBegin j, fn x => setPlace(x, Stack))
	       
	    fun loopExp(e: Exp.t): unit =
	       let val {decs, transfer} = Exp.dest e
	       in List.foreach(decs, loopDec)
		  ; (case transfer of
			Ctransfer.Call{cont, ...} =>
			   (case cont of
			       NONE => ()
			     | SOME c => (forceStack c
					  ; setIsCont(c, true)))
		      | _ => ())
	       end
	    and loopDec(d: Dec.t) =
	       case d of
		  Dec.Fun{body, ...} => loopExp(body)
		| Dec.HandlerPush h => (forceStack h
					; setIsHandler(h, true)
					; hasHandler := true)
		| _ => ()

	    val _ = loopExp(body)

	    (* Do the allocation top down.
	     * Whenever entering a local function ...
	     *)

	    val frameSize = ref 0

	    (* The next available offset on the stack.
	     * It is 4 initially because the stack pointer always points
	     * at the return address in the previous frame.
	     *)
	    val nextOffset = 4

	    val (nextOffset, handlerOffset) =
	       if !hasHandler
		  then (nextOffset + handlerSize, SOME nextOffset)
	       else (nextOffset, NONE)

	    fun getMask(xs: Var.t list): {regs: Register.t list,
					  offsets: int list} =
	       List.fold
	       (xs, {regs = [], offsets = []}, fn (x, ros as {regs, offsets}) =>
		if shouldAllocate(x) andalso Mtype.isPointer(varType x)
		   then
		      let val opr = varOperand x
		      in case place(x) of
			 Register =>
			    (case Operand.deRegister opr of
				NONE => Error.bug "must be a register"
			      | SOME r => {regs = r :: regs, offsets = offsets})
		       | Stack =>
			    case Operand.deStackOffset opr of
			       NONE => Error.bug "must be a slot"
			     | SOME{offset, ...} => {offsets = offset :: offsets,
						     regs = regs}
		      end
		else ros)

	    val nextOffset = ref nextOffset

	    (* The next available register number for each type. *)
	    val nextReg = Mtype.memo(fn _ => ref 0)

	    fun nextRegister(ty: Mtype.t, c: Chunk.t): Register.t =
	       let val r = nextReg ty
		  val reg = Chunk.register(c, !r, ty)
	       in IntRef.inc(r)
		  ; reg
	       end

	    fun allocateVar(x: Var.t, c: Chunk.t): unit =
	       if shouldAllocate(x)
		  then (setVarOperand
			(x, 
			 let val ty = varType(x)
			 in case place(x) of
			    Stack =>
			       let val offset = Mtype.align(!nextOffset, ty)
			       in nextOffset := offset + Mtype.size(ty)
				  ; Operand.stackOffset{ty = ty,
							offset = offset}
			       end
			  | Register => Operand.register(nextRegister(ty, c))
			 end))
	       else ()

(* 	    val allocateVar =
 * 	       Trace.trace2("allocateVar", Var.layout, Layout.ignore, Unit.layout)
 * 	       allocateVar
 *)
	    val chunk = funcChunk(name)

	    val {get = externalVarOperand,
		 set = setExternalVarOperand} =
	       Property.new(Var.plist, Property.initConst NONE)
		 
	    val (prelude, liveRegs) =
	       List.fold
	       (args, ([], []), fn ((x, _), (prelude, liveRegs)) =>
		(allocateVar(x, chunk)
		 ; let
		      val opr = varOperand(x)
		      val (r, prelude) =
			 case Operand.deRegister(opr) of
			    SOME r => (r, prelude)
			  | NONE =>
			       let val r = nextRegister(varType x, chunk)
				  val opr' = Operand.register r
			       in setExternalVarOperand(x, SOME opr')
				  ; (r,
				     Statement.move{dst = opr, src = opr'}
				     :: prelude)
			       end
		      val liveRegs =
			 if Mtype.isPointer(Register.ty(r))
			    then r :: liveRegs
			 else liveRegs
		   in (prelude, liveRegs)
		   end))

	    fun allocate(e: Exp.t, c: Chunk.t): unit =
	       let
		  val {decs, ...} = Exp.dest e
		  fun getGCInfo(live: Var.t list) =
		     let val {regs, offsets} = getMask(live)
		     in GCInfo.T{frameSize = !nextOffset,
				 regs = regs,
				 offsets = offsets}
		     end
	       in List.foreach
		  (decs,
		   fn Dec.Bind{var, exp, ...} =>
		        (allocateVar(var, c)
			 ; (case exp of
			       PrimExp.PrimApp{prim, ...} =>
				  if Prim.isArrayAllocation(prim)
				     then
					setLiveArray
					(var, getGCInfo(liveArrayVars var))
				  else ()
			     | _ => ()))
		    | Dec.Fun{name, args, body, ...} =>
			 let
			    val saveReg = Mtype.memo(! o nextReg)
			    val saveOffset = !nextOffset
			    val _ = List.foreach(Mtype.all, ignore o saveReg)
			    val c' = jumpChunk(name)
			    val _ =
			       if Chunk.equals(c, c')
				  then ()
			       else (* We can reset all of the register counters
				     * because we know that no registers live
				     * across a chunk boundary.
				     *)
				  List.foreach(Mtype.all, fn t =>
					       nextReg t := 0)
			    val _ = List.foreach(args, fn (x, _) =>
						 allocateVar(x, c'))
			    val _ = allocate(body, c')
			    fun getLiveOffsets() =
			       #offsets(getMask(liveBeginNoFormals(name)))
			    val _ =
			       if isCont(name)
				  then
				     setContInfo
				     (name, {size = !nextOffset,
					     liveOffsets = getLiveOffsets()})
			       else ()
			    val _ =
			       if isHandler(name)
				  then
				     setHandlerInfo
				     (name, {size = valOf(handlerOffset),
					     liveOffsets = getLiveOffsets()})
			       else ()
			    val _ =
			       setJumpInfo
			       (name, Info.T{varOperand = varOperand,
					     gcInfo = getGCInfo(liveBegin name),
					     prelude = [],
					     postlude = []})

			    val _ = List.foreach(Mtype.all, fn t =>
						 nextReg(t) := saveReg(t))
			    val _ = nextOffset := saveOffset
			 in ()
			 end
		    | _ => ())
	       end

	    val _ = allocate(body, chunk)
	       
	    val varOperand =
	       fn x =>
	       case externalVarOperand(x) of
		  NONE => varOperand(x)
		| SOME z => z

	 in destroyPlace()
	    ; destroyIsCont()
	    ; destroyIsHandler()
	    ; (setFuncInfo
	       (name,
		{info =
		 Info.T{varOperand = varOperand,
			gcInfo = GCInfo.T{regs = liveRegs,
					  offsets = [],
					  frameSize = Mtype.size Mtype.label},
			prelude = prelude,
			postlude = []},
		 handlerOffset = handlerOffset}))
	    ; ()
	 end
   in List.foreach(functions, fn {name, args, body, ...} =>
		   allocateFunc(name, args, body))
      ; {arrayInfo = liveArray,
	 funcInfo = funcInfo,
	 contInfo = contInfo,
	 handlerInfo = handlerInfo,
	 jumpInfo = jumpInfo}
   end
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * In the Cps input program
 *   Const and ConApp{args = [], ...} must all be global
 *   so that the register allocator doesn't allocate a register or stack slot
 *   for them.
 *
 *)
functor Backend(S: BACKEND_STRUCTS): BACKEND = 
struct

open S

local open Cps
in structure Con = Con
   structure Const = Const
   structure Cdec = Dec
   structure Cexp = Exp
   structure Func = Func
   structure Function = Function
   structure Cprogram = Program
   structure Ctype = Type
   structure Jump = Jump
   structure Prim = Prim
   structure PrimExp = PrimExp
   structure Tycon = Tycon
   structure Ctransfer = Transfer
   structure Var = Var
end 

local open Machine
in
   structure Chunk = Chunk
   structure GCInfo = GCInfo
   structure Label = Label
   structure Mtype = Type
   structure Mprogram = Program
   structure Operand = Operand
   structure Register = Register
   structure Statement = Statement
   structure Mtransfer = Transfer
end

structure Chunkify = Chunkify(open Cps)

structure ParallelMove = ParallelMove()

structure LimitCheck = LimitCheck(open Cps)
   
structure Representation = Representation(structure Cps = Cps
					  structure Mtype = Mtype)

structure AllocateRegisters = AllocateRegisters(structure Cps = Cps
						structure Machine = Machine)
local open AllocateRegisters
in structure Info = Info
end

local open Representation
in structure TyconRep = TyconRep
   structure ConRep = ConRep
end

fun generate(program as Cprogram.T{datatypes, globals, functions, main})
   : Mprogram.t =
   let
      val insertLimitCheck = LimitCheck.limitCheck program
	 
       val {tyconRep, conRep, toMtype} =
	  Control.traceCall "compute representations"
	  Representation.compute program

       val _ =
	  Control.displays
	  ("rep", fn display =>
	   List.foreach
	   (datatypes, fn {tycon, cons} =>
	    let open Layout
	    in display(seq[Tycon.layout tycon,
			   str " ",
			   TyconRep.layout(tyconRep tycon)]);
	       display(indent(align(List.map(cons, fn {con, ...} =>
					     seq[Con.layout con,
						 str " ",
						 ConRep.layout(conRep con)])),
			      2))
	    end))

       fun toMtypes ts = List.map(ts, toMtype)

       val wordSize = 4
       val tagOffset = 0
       val tagType = Mtype.int

       val jumpHandlers = Cps.inferHandlers program

       val chunks =
	  Control.traceCall "chunkify" Chunkify.chunkify
	  {program = program,
	   jumpHandlers = jumpHandlers}

       val _ = 
	  Control.displays
	  ("chunks", fn display =>
	   List.foreach
	   (chunks, fn {funcs, jumps} =>
	    let open Layout
	    in display
	       (record([("funcs", List.layout Func.layout funcs),
			("jumps", List.layout Jump.layout jumps)]))
	    end))

       val {get = jumpInfo: Jump.t -> {chunk: Chunk.t,
				       formals: Operand.t list option ref},
	    set = setJumpInfo} =
	  Property.new(Jump.plist, Property.initRaise("jump info", Jump.layout))

       val jumpChunk = #chunk o jumpInfo

       val {get = funcInfo: Func.t -> {chunk: Chunk.t,
				       interChunk: Label.t,
				       params: {formals: Operand.t list,
						globalFormals: Operand.t list
						} option ref
				       },
	    set = setFuncInfo} =
	  Property.new(Func.plist, Property.initRaise("func info", Func.layout))

       val funcChunk = #chunk o funcInfo
 
       val mprogram = Mprogram.new()

       (* Lazily build global operands for an arbitrary number of instances of
	* each type.
	*)
       local
	  val typeGlobals: Mtype.t -> Operand.t list ref =
	     Mtype.memo(fn _ => ref [])
       in
	  fun typesGlobals(ts: Mtype.t list): Operand.t list =
	     let val index = Mtype.memo(fn _ => ref 0)
	     in List.foldr
		(ts, [], fn (t, accum) =>
		 let val gs = typeGlobals t
		    val rec loop =
		       fn (0, []) => let
					val g =
					   if Mtype.isPointer t
					      then
						 Mprogram.newGlobalPointerNonRoot
						 mprogram
					   else Mprogram.newGlobal(mprogram, t)
				     in gs := !gs @ [g];
					g
				     end
			| (0, g :: _) => g
			| (i, _ :: gs) => loop(i - 1, gs)
			| _ => Error.bug "typesGlobals"
		    val r = index t
		    val i = !r
		 in r := i + 1;
		    loop(i, !gs) :: accum
		 end)
	     end
       end

       fun typeGlobal(t: Mtype.t): Operand.t =
	  case typesGlobals[t] of
	     [p] => p
	   | _ => Error.bug "typeGlobal"

       val {get = jumpCont: Jump.t -> {label: Label.t,
				       formals: Operand.t list option ref,
				       generated: bool ref} option,
	    set = setJumpCont} =
	  Property.new(Jump.plist, Property.initConst NONE)

       val {get = jumpHandler: Jump.t -> {label: Label.t,
					  formals: Operand.t list option ref,
					  generated: bool ref} option,
	    set = setJumpHandler} =
	  Property.new(Jump.plist, Property.initConst NONE)

       fun getJumpHandler j =
	  case jumpHandler j of
	     NONE => Error.bug "getJumpHandler"
	   | SOME{label = h, ...} => h

       (* Create info for jumps used as conts. *)
       val conts =
	  let
	     val conts = ref []
	     local
		fun make (get, set) j =
		   case get j of
		      NONE =>
			 let val l = Label.new(jumpToLabel j)
			 in ListRef.push(conts, (l, j))
			    ; set(j, SOME{label = l,
					  formals = ref NONE,
					  generated = ref false})
			 end
		    | SOME _ => ()
	     in val newHandler = make(jumpHandler, setJumpHandler)
		val newCont = make(jumpCont, setJumpCont)
	     end
	     val handleDec =
		fn Cdec.HandlerPush h => newHandler h
		 | _ => ()
	     val handleTransfer =
		fn Ctransfer.Call{cont, ...} => Option.map'(cont, newCont)
		 | _ => ()
	  in List.foreach
	     (functions, fn {body, ...} =>
	      Cexp.foreach'(body, {handleDec = handleDec,
				   handleTransfer = handleTransfer}))
	     ; (List.foreach
		(functions, fn {body, ...} =>
		 Cexp.foreachDec
		 (body,
		  fn Cdec.Fun{name, args, ...} =>
		        let
			   val formals =
			      Promise.lazy
			      (fn () => typesGlobals(List.map(args, toMtype o #2)))
			in List.foreach
			   ([jumpCont, jumpHandler], fn f =>
			    case f name of
			       SOME{formals = r, ...} => r := SOME(formals())
			     | _ => ())
			end
		   | _ => ())))
	     ; !conts
	  end

       val machineChunks = ref []
	  
       val _ =
	  ((* Create the mprogram chunks. *)
	   List.foreach
	   (chunks, fn {funcs, jumps} =>
	    let val interChunkEntries = List.map(funcs, Label.new o funcToLabel)
	       val conts =
		  List.fold(jumps, [], fn (j, cs) =>
			    let
			       val cs = 
				  case jumpHandler j of
				     NONE => cs
				   | SOME{label = c, ...} => c :: cs
			    in case jumpCont j of
			       NONE => cs
			     | SOME{label = c, ...} => c :: cs
			    end)
	       val c = Mprogram.newChunk{program = mprogram,
					 entries = interChunkEntries @ conts}
	    in ListRef.push(machineChunks, c)
	       ; List.foreach2(funcs, interChunkEntries,
			     fn (f, l) =>
			     setFuncInfo(f, {chunk = c,
					     interChunk = l,
					     params = ref NONE}))
	       ; List.foreach(jumps, fn j =>
			      setJumpInfo(j, {chunk = c,
					      formals = ref NONE}))
	    end)
	  )

       val {get = varInfo: Var.t -> {ty: Ctype.t,
				     operand: Operand.t option ref,
				     isConstant: bool ref},
	    set = setVarInfo} =
	  Property.new(Var.plist, Property.initRaise("info", Var.layout))

       val varIsConstant = ! o #isConstant o varInfo
       val varType = #ty o varInfo
       val varOperand = ! o #operand o varInfo
	  
(*        val varOperand =
 * 	  Trace.trace("varOperand", Var.layout, Option.layout Operand.layout)
 * 	  varOperand
 *)

       val setVarOperand =
	  fn (x, p) =>
	  let val {operand, ...} = varInfo x
	  in case !operand of
	     NONE => operand := SOME p
	   | SOME _ => Error.bug("setVarOperand twice of " ^ Var.toString x)
	  end

(*        val setVarOperand = Trace.trace2("setVarOperand", Var.layout,
 * 					Operand.layout, Unit.layout) setVarOperand
 *)

       val _ =
	  let
	     val set = fn (x, t) => setVarInfo(x, {ty = t,
						   operand = ref NONE,
						   isConstant = ref false})
	     fun sets xts = List.foreach(xts, set)
	  in
	     List.foreach(globals, fn {var, ty, ...} => set(var, ty));
	     let
		fun loopExp e = List.foreach(#decs(Cexp.dest e), loopDec)
		and loopDec d =
		   case d of
		      Cdec.Bind{var, ty, ...} => set(var, ty)
		    | Cdec.Fun{args, body, ...} => (sets args; loopExp body)
		    | _ => ()
	     in List.foreach(functions, fn {name, args, body, ...} =>
			     (sets args; loopExp body))
	     end
	  end

       fun varTypes xs = List.map(xs, varType)
       val varMtype = toMtype o varType
       fun varMtypes xs = List.map(xs, varMtype)
	  
       fun sortTypes(initialOffset: int,
		     tys: Mtype.t list): {size: int,
					  offsets: int list,
					  numWordsNonPointers: int,
					  numPointers: int} =
	  let
	     val voids = ref []
	     val bytes = ref []
	     val doubleWords = ref []
	     val words = ref []
	     val pointers = ref []
	     val _ = List.foreachi(tys, fn (i, t) =>
				   ListRef.push
				   (if Mtype.isPointer t
				       then pointers
				    else (case Mtype.size t of
					     0 => voids
					   | 1 => bytes
					   | 4 => words
					   | 8 => doubleWords
					   | _ => Error.bug "strange size"),
				    (i, t)))
	     fun build(r, accum) =
		List.fold(!r, accum, fn ((index, ty), (res, offset)) =>
			  ({index = index, offset = offset, ty = ty} :: res,
			   offset + Mtype.size ty))
	     val (accum, offset) =
		build
		(voids,
		 build(bytes,
		       build(words,
			     build(doubleWords, ([], initialOffset)))))
	     val offset = Mtype.align(offset, Mtype.pointer)
	     val numWordsNonPointers = (offset - initialOffset) div wordSize
	     val (accum, size) = build(pointers, (accum, offset))
	     val components = List.reverse accum
	     fun loop(i, accum) =
		if i = ~1
		   then accum
		else loop(i - 1,
			  #offset(List.lookup(components, fn {index, ...} =>
					      i = index))
			  :: accum)
	     val offsets = loop(List.length components - 1, [])
	  in {size = size,
	      offsets = offsets,
	      numWordsNonPointers = numWordsNonPointers,
	      numPointers = List.length(!pointers)}
	  end

       (* Compute layout for each con and associate it with the con.
	*)
       local
	  val {get, set} =
	     Property.new(Con.plist, Property.initRaise("con info", Con.layout))
       in
	  val _ =
	     List.foreach
	     (datatypes, fn {cons, ...} =>
	      List.foreach(cons, fn {con, args} =>
			   let fun doit n = set(con, sortTypes(n, toMtypes args))
			   in case conRep con of
			      ConRep.Tuple => doit 0
			    | ConRep.TagTuple _ => doit 4
			    | _ => ()
			   end))

	  val conInfo = get

	  fun conOffset(c: Con.t, n: int): Operand.t =
	     Operand.int(List.nth(#offsets(get c), n))
       end

       (* Compute layout for each tuple type.
	*)
       local
	  val {get, set} =
	     Property.new
	     (Ctype.plist,
	      Property.initFun(fn t => sortTypes(0, toMtypes(Ctype.detuple t))))
       in
	  val tupleInfo = get
	  fun tupleOffset(t: Ctype.t, n: int): Operand.t =
	     Operand.int(List.nth(#offsets(get t), n))
       end

       fun parallelMove{srcs, dsts, chunk} =
	  let
	     val moves =
		List.map2(srcs, dsts, fn (src, dst) => {src = src, dst = dst})
	     fun temp r =
		Operand.register(Chunk.tempRegister(chunk, Operand.ty r))
	     val temp =
		Trace.trace("temp", Operand.layout, Operand.layout) temp
	  in
	     Trace.trace
	     ("parallelMove",
	      fn {moves, ...} =>
	      List.layout (fn {src, dst} =>
			   Layout.tuple
			   [Operand.layout src, Operand.layout dst])
	      moves,
	      Layout.ignore)
	     ParallelMove.move{
			       moves = moves,
			       equals = Operand.equals,
			       move = Statement.move,
			       temp = temp
			      }
	  end

       fun jumpFormals(j: Jump.t): Operand.t list =
	  case jumpInfo j of
	     {formals = ref(SOME os), ...} => os
	   | _ => Error.bug "jumpFormals"

       fun tail(to: Jump.t, args: Operand.t list): Statement.t list * Mtransfer.t =
	  (parallelMove{srcs = args,
			dsts = jumpFormals to,
			chunk = jumpChunk to},
	   Mtransfer.nearJump to)
	     
       fun genCase{chunk: Chunk.t,
		   profileName: string,
		   varOperand,
		   testVar: Var.t, testRep: TyconRep.t,
		   cases, default} =
	  let
	     val test = varOperand testVar
	     (* Creating this new block is OK because all it does is
	      * a few moves and then a transfer.  I.E. it does no allocations
	      * and can not trigger a gc.
	      *)
	     fun newBlock(statements, transfer): Label.t =
		let val l' = Label.newString "L"
		in Chunk.newBlock(chunk,
				  {profileName = profileName,
				   label = l',
				   gcInfo = NONE,
				   statements = statements,
				   transfer = transfer});
		   l'
		end
	     fun enum(test: Operand.t): Mtransfer.t =
		let
		   val cases =
		      List.fold(cases, [], fn ((c, l), cases) =>
				let fun keep n = (n, l) :: cases
				in case conRep c of
				   ConRep.Int n => keep n
				 | ConRep.IntCast n => keep n
				 | _ => cases
				end)
		in Mtransfer.switch{test = test, cases = cases, default = default}
		end
	     fun switchIP(pointer: Label.t): Mtransfer.t =
		Mtransfer.switchIP{test = test,
				   int = newBlock([], enum(Operand.castInt test)),
				   pointer = pointer}
	     fun enumAndOne(): Mtransfer.t =
		switchIP
		(newBlock
		 (case List.peek(cases, fn (c, _) =>
				 case conRep c of
				    ConRep.Transparent _ => true
				  | ConRep.Tuple => true
				  | _ => false)
		     of
			NONE => 
			   (case default of
			       NONE => Error.bug "enumAndOne: no default"
			     | SOME l => ([], Mtransfer.nearJump l))
		      | SOME(_, l) =>
			   (case jumpFormals l of
			       [] => []
			     | [z] =>
				  Statement.moves{srcs = [test], dsts = [z]}
			     | _ => Error.bug "case label must be unary or nullary",
			    Mtransfer.nearJump l)))
	     fun indirectTag(): Mtransfer.t =
		Mtransfer.switch
		{test = Operand.offset{base = test,
				       offset = Operand.int tagOffset,
				       ty = tagType},
		 default = default,
		 cases =
		 List.fold
		 (cases, [], fn ((c, l'), cases) =>
		  case conRep c of
		     ConRep.TagTuple n =>
			let 
			   val l =
			      case jumpFormals l' of
				 [] => l'
			       | [z] => newBlock([Statement.move{dst = z,
								 src = test}],
						 Mtransfer.nearJump l')
			       | _ => Error.bug "case label wrong number args"
			in (n, l) :: cases
			end 
		   | _ => cases)}

	  in case testRep of
	     TyconRep.Prim mtype =>
		(case (cases, default) of
		    ([(_, l')], _) =>
		       (* We use _ instead of NONE for the default becuase
			* there may be an unreachable default case
			*)
		       tail(l', (case jumpFormals l' of
				    [] => []
				  | _ => [test]))
		  | ([], SOME l') => tail(l', [])
		  | _ => Error.bug "prim datatype with more than one case")
	   | TyconRep.Enum => ([], enum test)
	   | TyconRep.EnumDirect => ([], enumAndOne())
	   | TyconRep.EnumIndirect => ([], enumAndOne())
	   | TyconRep.EnumIndirectTag =>
		([], switchIP(newBlock([], indirectTag())))
	   | TyconRep.IndirectTag => ([], indirectTag())
	  end

       fun genConstBind{var, ty, exp} =
	  let
	     fun set(oper: Operand.t): unit =
		let val {operand, isConstant, ...} = varInfo var
		in operand := SOME oper;
		   isConstant := true
		end
	     fun global(new, s) = set(new(mprogram, s))
	     fun bogus() = set Operand.void
	     fun nonExpansive() =
		if Mtype.isVoid(toMtype ty)
		   then bogus()
		else ()
	  in case exp of
	     PrimExp.Var x =>
		(case varOperand x of
		    NONE => ()
		  | SOME oper => set oper)
	   | PrimExp.Const c => 
		(case Const.node c of
		    Const.Int n =>
		       set
		       (Operand.int32
			(let val t = Const.tycon c
			 in if Tycon.equals(t, Tycon.int)
			       then n
			    else if Tycon.equals(t, Tycon.int31)
				    then n * 2
				 else Error.bug "strange Int"
			 end))
		  | Const.Word w =>
		       set
		       (let val t = Const.tycon c
			in if Tycon.equals(t, Tycon.word)
			      then Operand.uint w
			   else if Tycon.equals(t, Tycon.word8)
				   then Operand.char(Char.chr(Word32.toInt w))
				else if Tycon.equals(t, Tycon.word31)
					then Operand.uint(Word32.<<(w, 0w1))
				     else Error.bug "strange word"
			end)
		  | Const.Real f => set(Operand.float f)
		  | Const.Char c => set(Operand.char c)
		  | Const.String s => global(Mprogram.newString, s)
		  | Const.SmallIntInf i => set(Operand.intInf i)
		  | Const.IntInf s => global(Mprogram.newIntInf, s))
	   | PrimExp.Tuple [] => bogus()
	   | PrimExp.Tuple _ => ()
	   | PrimExp.ConApp{con, args} =>
		(case conRep con of
		    ConRep.Void => bogus()
		  | ConRep.Int n => set(Operand.int n)
		  | ConRep.IntCast n => set(Operand.pointer n)
		  | _ => ())
	   | PrimExp.ConSelect _ => nonExpansive()
	   | PrimExp.Select _ => nonExpansive()
	   | PrimExp.PrimApp{prim, targs, args} =>
		(case Prim.numArgs prim of
		    NONE => set(Operand.prim(prim, toMtype ty))
		  | _ => ())
	  end

(*        val genConstBind =
 * 	  Trace.trace("genConstBind", Var.layout o #var, Unit.layout)
 * 	  genConstBind
 *)

       (* Set the operands for constants.
	* This has to happen before register allocation so that RA doesn't
	* allocate registers for them.
	*)
       val _ =
	  (List.foreach(globals, genConstBind);
	   List.foreach
	   (globals, fn {var, ty, exp} => 
	    case varOperand var of
	       NONE =>
		  setVarOperand
		  (var,
		   case exp of
		      PrimExp.Var x =>
			 (case varOperand x of
			     NONE => Error.bug "global missing operand"
			   | SOME oper => oper)
		    | _ =>  Mprogram.newGlobal(mprogram, toMtype ty))
	     | SOME _ => ());
	   List.foreach(functions, fn {body, ...} =>
			Cexp.foreachBind(body, genConstBind)))

       val shouldAllocate = not o Option.isSome o varOperand
	  
       val {arrayInfo, contInfo, funcInfo = funcRegInfo, handlerInfo,
	    jumpInfo = jumpRegInfo} =
	  Control.pass
	  {name = "allocate registers",
	   suffix = "reg",
	   style = Control.No,
	   thunk = fn () =>
	   AllocateRegisters.allocate
	   {
	    program = program,
	    varType = varMtype,
	    jumpChunk = jumpChunk,
	    funcChunk = funcChunk,
	    shouldAllocate = shouldAllocate,
	    jumpHandlers = jumpHandlers
	    },
	   display =
	   Control.Layouts
	   (fn ({funcInfo, jumpInfo, ...}, layout) =>
	    let
	       val constantVarOperand = varOperand
		  
	       open Layout
	       fun layoutVar varOperand x =
		  layout(seq[Var.layout x, str " ",
			     Operand.layout(case constantVarOperand x of
					       NONE => varOperand x
					     | SOME z => z)])
		  
	       fun loopBind varOperand {var, ty, exp} = layoutVar varOperand var

	       fun loop(Info.T{varOperand, ...}, args, body) =
		  let val layoutVar = layoutVar varOperand
		     val loopBind = loopBind varOperand
		  in List.foreach(args, layoutVar o #1);
		     List.foreach
		     (#decs(Cexp.dest body),
		      fn Cdec.Bind b => loopBind b
		       | Cdec.Fun{name, args, body} =>
			    loop(jumpInfo name, args, body)
		       | _ => ())
		  end		     
	    in List.foreach(globals, loopBind(valOf o varOperand));
	       List.foreach
	       (functions, fn {name, args, body, ...} =>
		let val {info, handlerOffset, ...} = funcInfo name
		in layout(seq[str "function ", Func.layout name,
			      str " handlerOffset ",
			      Option.layout Int.layout handlerOffset])
		   ; loop(#info(funcInfo name), args, body)
		end)
	    end)}

       (* Set the params part of the label info for all labels.
	* This has to happen after register allocation, because it needs to
	* know the operands for formal parameters.
	*)
       val _ =
	  let
	     fun loopExp e = List.foreach(#decs(Cexp.dest e), loopDec)
	     and loopDec d =
		case d of
		   Cdec.Fun{name, args, body} =>
		      (let val Info.T{varOperand, ...} = jumpRegInfo name
			  val {formals, ...} = jumpInfo name
		       in formals := SOME(List.map(args, varOperand o #1))
		       end
		       ; loopExp body)
		 | _ => ()
	  in List.foreach
	     (functions, fn {name, args, body, ...} =>
	      (let val {chunk, interChunk, params} = funcInfo name
		   val {info = Info.T{varOperand, ...}, ...} =
		      funcRegInfo name
		   val formals = List.map(args, varOperand o #1)
		   val globalFormals =
		      typesGlobals(List.map(args, toMtype o #2))
	       in Chunk.newBlock
		  (chunk, {profileName = Func.toString name,
			   label = interChunk,
			   gcInfo = NONE,
			   statements =
			   Statement.moves{srcs = globalFormals,
					   dsts = formals},
			   transfer = Mtransfer.nearJump(funcToLabel name)})
		  ; params := SOME{
				   formals = formals,
				   globalFormals = globalFormals
				   }
	       end
	       ; loopExp body))
	  end

       (* ---------------------------------- *)
       (*             genPrimExp             *)
       (* ---------------------------------- *)

       fun genPrimExp(x: Var.t, ty: Ctype.t, e: PrimExp.t,
		      varOperand): Statement.t list =
	  if varIsConstant x
	     then []
	  else
	     let
		fun varOperands xs = List.map(xs, varOperand)

		fun move src =
		   [Statement.move{dst = varOperand x, src = src}]

		fun makeStores(ys, offsets) =
		   List.fold2(ys, offsets, [], fn (y, offset, stores) =>
			      if Mtype.isVoid(Operand.ty y)
				 then stores
			      else {offset = offset, value = y} :: stores)

		fun allocate(ys, {size, offsets,
				  numPointers, numWordsNonPointers}) =
		   [Statement.allocate
		    {dst = varOperand x,
		     size = size,
		     numPointers = numPointers,
		     numWordsNonPointers = numWordsNonPointers,
		     stores = makeStores(ys, offsets)}]

		fun allocateTagged(n: int, ys: Operand.t list,
				   {size, offsets,
				    numPointers, numWordsNonPointers}) =
		   [Statement.allocate
		    {dst = varOperand x,
		     size = size,
		     numPointers = numPointers,
		     numWordsNonPointers =
		     (* for the tag *) 1 + numWordsNonPointers,
		     stores = ({offset = tagOffset, value = Operand.int n}
			       :: makeStores(ys, offsets))}]


	     in case e of
		PrimExp.Var y => move(varOperand y)
	      | PrimExp.Tuple ys => allocate(varOperands ys, tupleInfo ty)
	      | PrimExp.Select{tuple, offset} =>
		   if Mtype.isVoid(Operand.ty(varOperand x))
		      then []
		   else 		
		      move(Operand.offset
			   {base = varOperand tuple,
			    offset = tupleOffset(varType tuple, offset),
			    ty = toMtype ty})
	      | PrimExp.ConSelect{variant, con, offset} =>
		   if Mtype.isVoid(Operand.ty(varOperand x))
		      then []
		   else let val variant = varOperand variant
			    fun doit() =
			       move(Operand.offset
				    {base = variant,
				     offset = conOffset(con, offset),
				     ty = toMtype ty})
			in case conRep con of
			   ConRep.Transparent _ => move variant
			 | ConRep.Tuple => doit()
			 | ConRep.TagTuple _ => doit()
			 | _ => Error.bug "strange conRep for ConSelect"
			end
	      | PrimExp.ConApp{con, args} =>
		   let val args = varOperands args
		      fun tuple() = allocate(args, conInfo con)
		   in case (conRep con, args) of
		      (ConRep.Transparent _, [y]) => move y
		    | (ConRep.Tuple, _) => tuple()
		    | (ConRep.TagTuple n, _) => allocateTagged(n, args, conInfo con)
		    | _ => Error.bug "strange ConApp"
		   end
	      | PrimExp.PrimApp{prim, targs, args} =>
		   let
		      fun unsafeSub(a, i, ty) =
			 let val ty = toMtype ty
			 in move(Operand.scaledOffset{base = a,
						      offset = i,
						      scale = Mtype.size ty,
						      ty = ty})
			 end
		      fun array(n: Operand.t, t: Mtype.t): Statement.t list =
			 let
			    val (nbnp, np) =
			       if Mtype.isPointer t
				  then (0, 1)
			       else (Mtype.size t, 0)
			 in [Statement.allocateArray{dst = varOperand x,
						     numElts = n,
						     numBytesNonPointers = nbnp,
						     numPointers = np,
						     gcInfo = arrayInfo x}]
			 end
		      val argOps = varOperands args
		      open Prim.Name
		   in case (Prim.name prim, targs, argOps) of
		      (VectorFromArray, _, [src]) =>
			 [Statement.move{dst = varOperand x, src = src}]
		    | (Array, [t], [n]) => array(n, toMtype t)
		    | (ArraySub, [t], [a, i]) => unsafeSub(a, i, t)
		    | (StringSub, [], [s, i]) => unsafeSub(s, i, Ctype.char)
		    | (VectorSub, [t], [v, i]) => unsafeSub(v, i, t)
		    | (ArrayUpdate, [t], [a, i, y]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ =>
			       [Statement.move
				{dst = Operand.scaledOffset{base = a,
							    offset = i,
							    scale = Mtype.size t,
							    ty = t},
				 src = y}]
			 end
		    | (Ref, [t], [y]) =>
			 let val t = toMtype t
			    val (ys, ts) = if Mtype.isVoid t
					      then ([], [])
					   else ([y], [t])
			 in allocate(ys, sortTypes(0, ts))
			 end
		    | (Deref, [t], [y]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ => [Statement.move{dst = varOperand x,
						 src = Operand.contents(y, t)}]
			 end
		    | (Assign, [t], [y, z]) =>
			 let val t = toMtype t
			 in case Mtype.dest t of
			    Mtype.Void => []
			  | _ => [Statement.move{dst = Operand.contents(y, t),
						 src = z}]
			 end
		    | _ =>
			 (case Mtype.dest(toMtype ty) of
			     Mtype.Void => [Statement.sideEffect(prim, argOps)]
			   | _ => [Statement.assign{dst = varOperand x,
						    oper = prim,
						    args = argOps}])
		   end
	      | _ => Error.bug "genPrimExp saw strange primExp"
	     end

(*        val genPrimExp =
 * 	  Trace.trace("genPrimExp",
 * 		      fn (x, t, e, _, _) => Layout.tuple[Var.layout x,
 * 							 Ctype.layout t,
 * 							 PrimExp.layout e],
 * 		      Layout.ignore)
 * 	  genPrimExp
 *)

       local
	  fun make (labelGen, info) (j: Jump.t,
				     profileName: string,
				     f: Label.t -> unit): Label.t =
	     case labelGen j of
		NONE => Error.bug "missing labelGen"
	      | SOME{label = c, formals, generated} =>
		   (if !generated
		       then ()
		    else (generated := true
			  ; f c
			  ; let val {size, liveOffsets} = info j
			       val (statements, transfer) =
				  tail(j, valOf(!formals))
			       (* restore stack pointer *)
			       val statements = Statement.pop size :: statements
			    in Chunk.newBlock(jumpChunk j,
					      {profileName = profileName,
					       label = c,
					       gcInfo = NONE,
					       statements = statements,
					       transfer = transfer})
			    end)
		    ; c)
       in
	  val genCont = make(jumpCont, contInfo)
	  val genHandler = make(jumpHandler, handlerInfo)
       end

       (* ---------------------------------- *)
       (*            genTransfer             *)
       (* ---------------------------------- *)

       fun genTransfer(chunk: Chunk.t,
		       profileName: string,
		       t: Ctransfer.t,
		       handlers: Jump.t list,
		       varOperand): Statement.t list * Mtransfer.t =
	  case t of
	     Ctransfer.Jump{dst, args} => tail(dst, List.map(args, varOperand))
	   | Ctransfer.Call{func, args, cont} =>
		let val args = List.map(args, varOperand)
		   val {chunk = chunk', params, interChunk} = funcInfo func
		   val {formals, globalFormals} = valOf(!params)
		   val (setupArgs, transfer) =
		      if Chunk.equals(chunk, chunk')
			 then (parallelMove{srcs = args,
					    dsts = formals,
					    chunk = chunk},
			       Mtransfer.nearJump(funcToLabel func))
		      else (Statement.moves{srcs = args, dsts = globalFormals},
			    Mtransfer.farJump(chunk', interChunk))
		in case cont of
		   NONE => (setupArgs, transfer)
		 | SOME c =>
		      let val {size, liveOffsets} = contInfo c
			 val return =
			    genCont(c, profileName, fn l =>
				    Mprogram.newFrame
				    (mprogram, {return = l,
						chunk = jumpChunk c,
						size = size,
						liveOffsets = liveOffsets}))
		      in 
			 (setupArgs
			  @ [Statement.push size,
			     Statement.move
			     {dst = Operand.stackOffset{offset = 0,
							ty = Mtype.int},
			      src = Operand.label return}],
			  transfer)
		      end
		end
	   | Ctransfer.Return xs =>
		(Statement.moves{srcs = List.map(xs, varOperand),
				 dsts = typesGlobals(varMtypes xs)},
		 Mtransfer.return)
	   | Ctransfer.Case{test, cases, default, ...} =>
		(case (cases, default) of
		    ([], NONE) => ([], Mtransfer.bug)
		  | _ => 
		       case Ctype.tyconArgs(varType test) of
			  (tycon, []) =>
			     genCase{chunk = chunk,
				     profileName = profileName,
				     varOperand = varOperand,
				     testRep = tyconRep tycon,
				     testVar = test,
				     cases = cases,
				     default = default}
			| _ => Error.bug "strange type in case")
	   | Ctransfer.Raise xs =>
		let val xops = List.map(xs, varOperand)
		in case handlers of
		   [] =>
		      (* The stack move must occur after the operand moves
		       * because some of the operands may be on the stack.
		       *)
		      (Statement.moves{dsts = typesGlobals(varMtypes xs),
				       srcs = xops},
		       Mtransfer.raisee)
		  | h :: _ => tail(h, xops)
		end
	   | Ctransfer.Halt => ([], Mtransfer.halt(Operand.int 0))
	   | Ctransfer.Bug => ([], Mtransfer.bug)

       (*------------------------------------*)
       (*               genExp               *)
       (*------------------------------------*)

       fun genExp{exp = e: Cexp.t,
		  profileName: string,
		  insertLimitCheck: bool,
		  label: Label.t,
		  chunk: Chunk.t,
		  info = Info.T{gcInfo, prelude, postlude, varOperand = vo},
		  handlerOffset: int option,
		  handlers: Jump.t list,
		  prelude = prelude': Statement.t list}: unit =
	  let val {decs, transfer} = Cexp.dest e
	     val varOperand = (fn x =>
			       case varOperand x of
				  SOME p => p
				| NONE => vo x)
	     val (decs, handlers) =
		genDecs(decs, profileName, varOperand, handlerOffset, handlers)
	     val (preTransfer, transfer) =
		genTransfer(chunk, profileName, transfer, handlers, varOperand)
	     val statements =
		prelude' @ prelude @ decs @ postlude @ preTransfer
	  in
	     Chunk.newBlock(chunk,
			    {profileName = profileName,
			     label = label,
			     gcInfo = if insertLimitCheck
					 then SOME gcInfo
				      else NONE,
			     statements = statements,
			     transfer = transfer})
	  end

       and genDecs(ds: Cdec.t list,
		   profileName: string,
		   varOperand,
		   handlerOffset,
		   handlers: Jump.t list)
	  : Statement.t list * Jump.t list =
	  let
	     val (statements, handlers) =
		List.fold
		(ds, ([], handlers), fn (d , (statements, handlers)) =>
		 (case d of
		     Cdec.Bind{var, ty, exp} =>
			genPrimExp(var, ty, exp, varOperand) :: statements
		  | Cdec.Fun{name, body, ...} =>
		       (genExp{exp = body,
			       profileName = profileName,
			       insertLimitCheck = insertLimitCheck name, 
			       label = jumpToLabel name,
			       chunk = jumpChunk name,
			       info = jumpRegInfo name,
			       handlerOffset = handlerOffset,
			       handlers = jumpHandlers name,
			       prelude = []}
			; statements)
		  | Cdec.HandlerPush h =>
		       let
			  val c =
			     genHandler
			     (h, profileName, fn c =>
			      Mprogram.newHandler(mprogram, c, jumpChunk h))
			  val offset = valOf handlerOffset
			  val statements =
			     [Statement.move
			      {dst = Operand.stackOffset{offset = offset,
							 ty = Mtype.label},
			       src = Operand.label c}]
			     :: statements
		       in case handlers of
			  [] => ([Statement.saveExnStack{offset = offset}]
				 :: statements)
			| _ => statements
		       end
		  | Cdec.HandlerPop =>
		       let val offset = valOf handlerOffset
		       in case handlers of
			  [] => Error.bug "pop of empty handler stack"
			| _ :: handlers => 
			     (case handlers of
				 [] =>
				    [Statement.restoreExnStack{offset = offset}]
			       | h :: _ =>
				    [Statement.move
				     {dst =
				      Operand.stackOffset{offset = offset,
							  ty = Mtype.label},
				      src = Operand.label(getJumpHandler h)}])
			     :: statements
		       end,
		 Cps.deltaHandlers(d, handlers)))
	  in (List.fold(statements, [], op @), handlers)
	  end

   in
      if List.exists(functions, fn {name, ...} =>
		     Func.equals(name, Func.saveWorld))
	 then
	    let val {chunk, interChunk, ...} = funcInfo Func.saveWorld
	    in Mprogram.setSaveWorld(mprogram, chunk, interChunk)
	    end
      else ()

      (* Build the initGlobals chunk. *)
      ; let
	   val statements =
	      List.fold
	      (List.fold
	       (globals, [], fn ({var, ty, exp}, statements) =>
		(genPrimExp(var, ty, exp, valOf o varOperand)
		 :: statements)),
	       [], op @)
	   val initGlobals = Label.newString "initGlobals"
	   val chunk = Mprogram.newChunk{program = mprogram,
					 entries = [initGlobals]}
	   val {chunk = mainChunk, interChunk = main, ...} = funcInfo main
	in Mprogram.setMain(mprogram, chunk, initGlobals)
	   ; (Chunk.newBlock
	      (chunk, {profileName = "initGlobals",
		       label = initGlobals,
		       gcInfo =
		       SOME(GCInfo.T{regs = [], offsets = [],
				     frameSize = Mtype.size Mtype.label}),
		       statements = statements,
		       transfer = Mtransfer.farJump(mainChunk, main)}))
	end

      ; (List.foreach
	 (functions, fn {name, body, ...} =>
	  let val {info, handlerOffset, ...} = funcRegInfo name
	  in genExp{exp = body,
		    profileName = Func.toString name,
		    insertLimitCheck = true,
		    label = funcToLabel name,
		    chunk = funcChunk name,
		    info = info,
		    handlerOffset = handlerOffset,
		    handlers = [],
		    prelude = [Statement.stackOverflowCheck]}
	  end))
      ; Cprogram.clear program

      ; mprogram
   end
 
end
signature APPEND_LIST = 
   sig
      type 'a t
      val empty: 'a t
      val cons: 'a * 'a t -> 'a t
      val map: 'a t * ('a -> 'b) -> 'b t
      val single: 'a -> 'a t
      val append: 'a t * 'a t -> 'a t
      val appends: 'a t list -> 'a t
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foldr: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val foreach: 'a t * ('a -> unit) -> unit
      val fromList: 'a list -> 'a t
      val toList: 'a t -> 'a list
      val layout: string * ('a -> Layout.t) -> 'a t -> Layout.t
   end
functor AppendList(): APPEND_LIST =
struct

datatype 'a t =
   Base of 'a list
 | Append of 'a t list

val empty = Base[]
fun single x = Base[x]
val fromList = Base
fun append(t1, t2) = Append[t1, t2]
val appends = Append
fun cons(x, l) = append(single x, l)

fun fold(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    Base l => List.fold(l, b, f)
	  | Append l => List.fold(l, b, loop)
   in loop(l, b)
   end

fun foreach(l, f) = fold(l, (), fn (x, ()) => f x)

fun foldr(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    Base l => List.foldr(l, b, f)
	  | Append l => List.foldr(l, b, loop)
   in loop(l, b)
   end

fun map(l, f) =
   let
      val rec loop =
	 fn Base l => Base(List.map(l, f))
	  | Append l => Append(List.map(l, loop))
   in loop l
   end

fun toList ds =
   let fun loop(Base xs, rest) = List.append(xs, rest)
	 | loop(Append ts, rest) = List.foldr(ts, rest, loop)
   in loop(ds, [])
   end

local open Layout
in fun layout (sep, layoutX) xs =
   let fun loop(Base xs) = seq[List.layoutSep (sep, layoutX) xs, str sep]
	 | loop(Append ts) = seq(List.map(ts, loop))
   in loop xs
   end
end

end

structure AppendList = AppendList()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature XML_TYPE =
   sig
      include HASH_TYPE
	 
      datatype dest =
	 Var of Tyvar.t
       | Con of Tycon.t * t list
	 
      val dest: t -> dest
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* binding occurences:
 *   1. lambda arg
 *   2. pattern arg
 *   3. MonoVal dec
 *   4. PolyVal dec
 *   5. Fun dec
 *   6. Handle catch
 *)

signature XML_TREE_STRUCTS =
   sig
      include ATOMS
   end

signature XML_TREE =
   sig
      include XML_TREE_STRUCTS

      structure Type: XML_TYPE
      sharing Atoms = Type.Atoms
	 
      structure Pat:
	 sig
	    datatype t = T of {con: Con.t,
			       targs: Type.t list,
			       arg: (Var.t * Type.t) option}
	 
	    val falsee: t
	    val truee: t
	    val con: t -> Con.t
	    val toAst: t -> Ast.Pat.t
	    val layout: t -> Layout.t
	 end

      structure Lambda:
	 sig
	    type exp
	    include T

	    (* constructors *)
	    val new: {arg: Var.t,
		      argType: Type.t,
		      body: exp} -> t
	       
	    (* destructors *)
	    val dest: t -> {arg: Var.t,
			    argType: Type.t,
			    body: exp}
	    val arg: t -> Var.t
	    val argType: t -> Type.t
	    val body: t -> exp

	    val plist: t -> PropertyList.t
	 end

      (* VarExp is a type application.
       * variable (applied to type args) paired with its type.
       *)
      structure VarExp:
	 sig
	    datatype t = T of {var: Var.t, targs: Type.t list}

	    (* constructors *)
	    val mono: Var.t -> t

	    (* destructors *)
	    val var: t -> Var.t
	    val targs: t -> Type.t list

	    val layout: t -> Layout.t
	 end

      structure PrimExp:
	 sig
	    type exp = Lambda.exp
	    datatype t =
	       Const of Const.t
	     | Var of VarExp.t
	     | Tuple of VarExp.t list
	     | Select of {tuple: VarExp.t,
			  offset: int}
	     | Lambda of Lambda.t
	     | ConApp of {con: Con.t,
			  targs: Type.t list,
			  arg: VarExp.t option}
	     (* primitives can be curried, and we require that they are fully
	      * applied.  For example, "plus" takes one arg that is a tuple,
	      * but List.map takes two args.
	      *)
	     | PrimApp of {prim: Prim.t,
			   targs: Type.t list,
			   args: VarExp.t list}
	     | App of {func: VarExp.t,
		       arg: VarExp.t}
	     | Raise of VarExp.t
	     | Case of {test: VarExp.t,
			cases: (Pat.t * exp) list,
			default: exp option}
	     | Handle of
	       {try: exp,
		catch: (Var.t * Type.t), (* binds exception in handler *)
		handler: exp}

	    val layout: t -> Layout.t
	 end
	       
      structure Dec:
	 sig
	    type exp = Lambda.exp
	       
	    datatype t =
	       MonoVal of {var: Var.t,
			   ty: Type.t,
			   exp: PrimExp.t}
	     | PolyVal of {var: Var.t,
			   tyvars: Tyvar.t list,
			   ty: Type.t,
			   exp: exp}
	     | Fun of {tyvars: Tyvar.t list,
		       decs: {var: Var.t,
			      ty: Type.t,
			      lambda: Lambda.t} list}
	     | Exception of {con: Con.t,
			     arg: Type.t option}

	    val toAst: t -> Ast.Dec.t
	    val layout: t -> Layout.t
	 end

      structure Exp:
	 sig
	    type t = Lambda.exp

	    (* destructors *)
	    val dest: t -> {decs: Dec.t list,
			    result: VarExp.t}
	    val decs: t -> Dec.t list
	    val result: t -> VarExp.t
	       
	    (* constructors *)
	    val new: {decs: Dec.t list,
		      result: VarExp.t} -> t

	    val fromPrimExp: PrimExp.t * Type.t -> t

	    val prefix: t * Dec.t -> t
	       
	    (* foreach{exp, handleExp, handleBoundVar, handleVarExp}
	     * applies handleExp to each subexpresison of e (including e)
	     * applies handleBoundVar to each variable bound in e
	     * applies handleVarExp to each variable expression in e
	     * handleBoundVar will be called on a variable binding before
	       handleVarExp is called on any occurrences
	     * handleExp is called on an expression after it is called on
		all of its subexpressions *)
	    val foreach:
	       {exp: t,
		handleExp: t -> unit,
		handlePrimExp: PrimExp.t -> unit,
		handleBoundVar: Var.t * Tyvar.t list * Type.t -> unit,
		handleVarExp: VarExp.t -> unit} -> unit

	    val foreachExp: t * (t -> unit) -> unit
	    val foreachPrimExp: t * (PrimExp.t -> unit) -> unit
	    val foreachVarExp: t * (VarExp.t -> unit) -> unit
	    val foreachBoundVar:
	       t * (Var.t * Tyvar.t list * Type.t -> unit) -> unit

	    val size: t -> int
	       
	    val clear: t -> unit
	    val layout: t -> Layout.t
	 end

      structure DirectExp:
	 sig
	    type t

	    val const: Const.t -> t
	    val var: {var: Var.t, targs: Type.t list, ty: Type.t} -> t
	    val varExp: VarExp.t * Type.t -> t
	    val monoVar: Var.t * Type.t -> t
	    val tuple: {exps: t list, ty: Type.t} -> t
	    val select: {tuple: t,
			 offset: int,
			 ty: Type.t} -> t
	    val conApp: {con: Con.t,
			 targs: Type.t list,
			 arg: t option,
			 ty: Type.t} -> t
	    val app: {func: t,
		      arg: t,
		      ty: Type.t} -> t
	    val primApp: {prim: Prim.t,
			  targs: Type.t list,
			  args: t list,
			  ty: Type.t} -> t
	    val raisee: t * Type.t -> t
	    val handlee: {try: t,
			  ty: Type.t,
			  catch: Var.t * Type.t,
			  handler: t} -> t
	    val casee: {
			test: t,
			cases: (Pat.t * t) list,
			default: t option,
			ty: Type.t (* type of entire case expression *)
		       } -> t

	    val lett: {decs: Dec.t list, body: t} -> t
	    val let1: {var: Var.t, exp: t, body: t} -> t
	       
	    val lambda: {arg: Var.t,
			 argType: Type.t,
			 body: t,
			 bodyType: Type.t} -> t

	    val detuple: {tuple: t,
			  body: (VarExp.t * Type.t) list -> t} -> t

	    val detupleBind: {tuple: t,
			      components: Var.t list,
			      body: t} -> t
	       
	    val sequence: t list -> t
	    val seq: t list * (t list -> t) -> t

	    val unit: unit -> t
	    val reff: t -> t
	    val deref: t -> t
 	    val equal: t * t -> t
	    val iff: {test: t, thenn: t, elsee: t, ty: Type.t} -> t
	       
	    val vall: {var: Var.t, exp: t} -> Dec.t list

	    val toExp: t -> Exp.t
	    val fromExp: Exp.t * Type.t -> t

	    val layout: t -> Layout.t
	 end

      structure Program:
	 sig
	    datatype t =
	       T of {datatypes: {
				 tycon: Tycon.t,
				 tyvars: Tyvar.t list,
				 cons: {
					con: Con.t,
					arg: Type.t option
					} list
				 } list,
		     body: Exp.t}

	    val empty: t

	    (* Clear all property lists *)
	    val clear: t -> unit

	    val layoutStats: t -> Layout.t
	       
	    val layout: t -> Layout.t
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor XmlTree(S: XML_TREE_STRUCTS): XML_TREE =
struct

open S
local open Ast
in structure Adec = Dec
   structure Aexp = Exp
   structure Apat = Pat
end

structure Type =
   struct
      structure T = HashType(S)
      open T

      datatype dest =
	 Var of Tyvar.t
       | Con of Tycon.t * t list

      fun dest t =
	 case Dest.dest t of
	    Dest.Var a => Var a
	  | Dest.Con x => Con x
	  | _ => Error.bug "Xml.Type.dest"
   end

structure Pat =
   struct
      structure Apat = Ast.Pat
	 
      datatype t = T of {con: Con.t,
			 targs: Type.t list,
			 arg: (Var.t * Type.t) option}
	 
      fun con(T{con, ...}) = con

      local fun make c = T{con = c, targs = [], arg = NONE}
      in val truee = make Con.truee
	 val falsee = make Con.falsee
      end

      fun toAst(T{con, arg, ...}) =
	 let val con = Con.toAst con
	 in case arg of
	    NONE => Apat.con con
	  | SOME (x, t) =>
	       Apat.app(con, Apat.var(Var.toAst x))
	 (*Apat.app(con, Apat.constraint(Apat.var(Var.toAst x), Type.toAst t)) *)
	 end

      val layout = Apat.layout o toAst
   end

(*---------------------------------------------------*)
(*                      VarExp                       *)
(*---------------------------------------------------*)

structure VarExp =
   struct
      datatype t = T of {var: Var.t,
			 targs: Type.t list}

      fun mono var = T{var = var, targs = []}

      fun var(T{var, ...}) = var
      fun targs(T{targs, ...}) = targs
	       
      val toAst = Aexp.var o Var.toAst o var

      fun toAsts(xs: t list): Aexp.t list = List.map(xs, toAst)
      val layout = Aexp.layout o toAst
   end

structure AppendList = AppendList
   
(*---------------------------------------------------*)
(*           Expressions and Declarations            *)
(*---------------------------------------------------*)

datatype exp =
   Exp of {decs: dec list,
	   result: VarExp.t}
and primExp =
   Const of Const.t
  | Var of VarExp.t
  | Tuple of VarExp.t list
  | Select of {tuple: VarExp.t,
	       offset: int}
  | Lambda of lambda
  | ConApp of {con: Con.t,
	       targs: Type.t list,
	       arg: VarExp.t option}
  | PrimApp of {prim: Prim.t,
		targs: Type.t list,
		args: VarExp.t list}
  | App of {func: VarExp.t,
	    arg: VarExp.t}
  | Case of {test: VarExp.t,
	     cases: (Pat.t * exp) list,
	     default: exp option}
  | Raise of VarExp.t
  | Handle of {try: exp,
	       catch: Var.t * Type.t,
	       handler: exp}
and dec =
   MonoVal of {var: Var.t,
	       ty: Type.t,
	       exp: primExp}
  | PolyVal of {tyvars: Tyvar.t list,
		ty: Type.t,
		var: Var.t,
		exp: exp}
  | Fun of {tyvars: Tyvar.t list,
	    decs: {var: Var.t,
		   ty: Type.t,
		   lambda: lambda} list}
  | Exception of {con: Con.t,
		  arg: Type.t option}
and lambda = Lam of {plist: PropertyList.t,
		     arg: Var.t,
		     argType: Type.t,
		     body: exp}

(*---------------------------------------------------*)
(*                 Conversion to Ast                 *)
(*---------------------------------------------------*)

fun expToAst(Exp{decs, result, ...}): Aexp.t =
   Aexp.lett(decsToAst decs, VarExp.toAst result)
and expsToAsts es = List.map(es, expToAst)
and decsToAst decs = List.map(decs, decToAst)
and decToAst d : Adec.t =
   case d of
      MonoVal{var, ty, exp} =>
	 Adec.make(Adec.Val
		   ([],
		    [(Apat.var(Var.toAst var),
(*		      Apat.constraint(Apat.var(Var.toAst var), Type.toAst ty), *)
		      primExpToAst exp)]))
    | PolyVal{tyvars, var, exp, ...} =>
	 Adec.vall(tyvars, Var.toAst var, expToAst exp)
    | Fun{tyvars, decs} =>
	 Adec.make
	 (Adec.Fun
	  (tyvars,
	   List.map(decs, fn {var, lambda = Lam{arg, argType, body, ...}, ...} =>
		    [{pats = [Apat.var(Var.toAst var),
			      Apat.constraint(Apat.var(Var.toAst arg),
					      Type.toAst argType)
(*			      Apat.var(Var.toAst arg) *)
			      ],
		      resultType = NONE,
		      body = expToAst body}])))
    | Exception{con, arg} =>
	 Adec.exceptionn(Con.toAst con, Type.optionToAst arg)
and primExpToAst e : Aexp.t =
   case e of
      Const c => Const.toAstExp c
    | Var x => VarExp.toAst x
    | Tuple xs => Aexp.tuple(VarExp.toAsts xs)
    | Select{tuple, offset} =>
	 Aexp.select{tuple = VarExp.toAst tuple,
		     offset = offset + 1}
    | Lambda lambda => Aexp.fnn(lambdaToAst lambda)
    | ConApp{con, arg, ...} =>
	 let val con = Aexp.con(Con.toAst con)
	 in case arg of
	    NONE => con
	  | SOME e => Aexp.app(con, VarExp.toAst e)
	 end
    | PrimApp{prim, args, ...} =>
	 let val p = Aexp.longvid(Prim.longid prim)
	 in
	    case Prim.numArgs prim of
	       NONE => p
	     | SOME _ => Aexp.app(p, Aexp.tuple(List.map(args, VarExp.toAst)))
	 end
    | App{func, arg} => Aexp.app(VarExp.toAst func, VarExp.toAst arg)
    | Raise x => Aexp.raisee(VarExp.toAst x)
    | Handle{try, catch, handler} =>
	 Aexp.handlee(expToAst try,
		      [(Apat.var(Var.toAst(#1 catch)),
			expToAst handler)])
    | Case{test, cases, default, ...} =>
	 Aexp.casee(VarExp.toAst test,
		    (List.map(cases, fn (pat, exp) =>
			      (Pat.toAst pat, expToAst exp))
		     @ (case default of
			   NONE => []
			 | SOME e => [(Ast.Pat.wild, expToAst e)])))
and lambdaToAst(Lam{arg, body, argType, ...}) =
   [(
     Apat.constraint(Apat.var(Var.toAst arg), Type.toAst argType),
(*     Apat.var(Var.toAst arg), *)
     expToAst body)]

fun layoutLambda f = Aexp.layout(Aexp.fnn(lambdaToAst f))

(*---------------------------------------------------*)
(*                   Declarations                    *)
(*---------------------------------------------------*)

structure Dec =
   struct
      type exp = exp
      datatype t = datatype dec

      val toAst = decToAst
      val layout = Ast.Dec.layout o toAst
   end

(*---------------------------------------------------*)
(*                    Expressions                    *)
(*---------------------------------------------------*)

structure PrimExp =
   struct
      type exp = exp
      datatype t = datatype primExp

      val toAst = primExpToAst
      val layout = Aexp.layout o toAst
   end

structure Exp =
   struct
      datatype t = datatype exp

      val new = Exp
      fun dest(Exp r) = r
      val decs = #decs o dest
      val result = #result o dest

      fun fromPrimExp(exp: PrimExp.t, ty: Type.t): t =
	 let val var = Var.newNoname()
	 in Exp{decs = [Dec.MonoVal{var = var, ty = ty, exp = exp}],
		result = VarExp.mono var}
	 end
	 
      local
	 fun make f (Exp{decs, result}, d) =
	    Exp{decs = f(d, decs),
		result = result}
      in val prefix = make(op ::)
	 val prefixs = make(op @)
      end

      val toAst = expToAst
      val layout = Ast.Exp.layout o toAst

      (*------------------------------------*)
      (*              foreach               *)
      (*------------------------------------*)
      fun foreach{exp: t,
		  handleExp: t -> unit,
		  handlePrimExp: PrimExp.t -> unit,
		  handleBoundVar: Var.t * Tyvar.t list * Type.t -> unit,
		  handleVarExp: VarExp.t -> unit}: unit =
	 let
	    fun monoVar(x, t) = handleBoundVar(x, [], t)
	    fun handleVarExps xs = List.foreach(xs, handleVarExp)
	    fun loopExp e =
	       let val {decs, result} = dest e
	       in List.foreach(decs, loopDec)
		  ; handleVarExp result
		  ; handleExp e
	       end
	    and loopPrimExp e =
	       (handlePrimExp e
		; (case e of
		      Const _ => ()
		    | Var x => handleVarExp x
		    | Tuple xs => handleVarExps xs
		    | Select{tuple, ...} => handleVarExp tuple
		    | Lambda lambda => loopLambda lambda
		    | PrimApp{args, ...} => handleVarExps args
		    | ConApp{arg, ...} => (case arg of
					      NONE => ()
					    | SOME x => handleVarExp x)
		    | App{func, arg} => (handleVarExp func
					 ; handleVarExp arg)
		    | Raise x => handleVarExp x
		    | Handle{try, catch, handler} =>
			 (loopExp try
			  ; monoVar catch
			  ; loopExp handler)
		    | Case{test, cases, default} =>
			 (handleVarExp test
			  ; List.foreach(cases, fn (Pat.T{arg, ...}, e) =>
					 (case arg of
					     NONE => ()
					   | SOME x => monoVar x
					  ; loopExp e))
			  ; (case default of
				NONE => ()
			      | SOME e => loopExp e))))
	    and loopDec d =
	       case d of
		  MonoVal{var, ty, exp} =>
		     (monoVar(var, ty); loopPrimExp exp)
		| PolyVal{var, tyvars, ty, exp} =>
		     (handleBoundVar(var, tyvars, ty)
		      ; loopExp exp)
		| Exception _ => ()
		| Fun{tyvars, decs, ...} =>
		     (List.foreach(decs, fn {var, ty, lambda} =>
				   handleBoundVar(var, tyvars, ty))
		      ; List.foreach(decs, fn {lambda, ...} =>
				     loopLambda lambda))
	    and loopLambda(Lam{arg, argType, body, ...}): unit =
	       (monoVar(arg, argType); loopExp body)
	 in loopExp exp
	 end

      fun ignore _ = ()

      fun foreachPrimExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = f,
		 handleExp = ignore,
		 handleBoundVar = ignore,
		 handleVarExp = ignore}

      fun foreachVarExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = ignore,
		 handleBoundVar = ignore,
		 handleVarExp = f}

      fun foreachBoundVar(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = ignore,
		 handleBoundVar = f,
		 handleVarExp = ignore}

      fun foreachExp(e, f) =
	 foreach{exp = e,
		 handlePrimExp = ignore,
		 handleExp = f,
		 handleBoundVar = ignore,
		 handleVarExp = ignore}

      fun size e =
	 let val n = ref 0
	    fun inc() = n := 1 + !n
	 in foreachPrimExp(e, fn _ => inc());
	    !n
	 end

(*      val size = Trace.trace("size", Layout.ignore, Int.layout) size *)

      fun clear(e: t): unit =
	 let open PrimExp
	    val clear = PropertyList.clear
	    fun clearTyvars ts = List.foreach(ts, Tyvar.clear)
	    fun clearPat(Pat.T{arg, ...}) =
	       case arg of
		  NONE => ()
		| SOME(x, _) => Var.clear x
	    fun clearExp e = clearDecs(decs e)
	    and clearDecs ds = List.foreach(ds, clearDec)
	    and clearDec d =
	       case d of
		  MonoVal{var, exp, ...} => (Var.clear var; clearPrimExp exp)
		| PolyVal{var, tyvars, exp, ...} =>
		     (Var.clear var;
		      clearTyvars tyvars;
		      clearExp exp)
		| Fun{tyvars, decs} =>
		     (clearTyvars tyvars;
		      List.foreach(decs, fn {var, ty, lambda} =>
				   (Var.clear var;
				    clearLambda lambda)))
		| Exception{con, ...} => Con.clear con
	    and clearPrimExp e =
	       case e of
		  Lambda l => clearLambda l
		| Case{cases, default, ...} =>
		     (List.foreach(cases, fn (p, e) =>
				   (clearPat p; clearExp e));
		      case default of
			 NONE => ()
		       | SOME e => clearExp e)
		| Handle{try, catch, handler} => 
		     (clearExp try;
		      Var.clear(#1 catch);
		      clearExp handler)
		| _ => ()
	    and clearLambda(Lam{arg, body, ...}) =
	       (Var.clear arg; clearExp body)
	 in clearExp e
	 end
   end

(*---------------------------------------------------*)
(*                      Lambda                       *)
(*---------------------------------------------------*)

structure Lambda =
   struct
      type exp = exp
      datatype t = datatype lambda

      fun arg(Lam{arg, ...}) = arg
      fun argType(Lam{argType, ...}) = argType 
      fun body(Lam{body, ...}) = body

      type node = {arg: Var.t,
		   argType: Type.t,
		   body: exp}

      fun new{arg, argType, body} =
	 Lam{plist = PropertyList.new(),
	     arg = arg, argType = argType, body = body}

      fun dest(Lam{arg, argType, body, ...}) =
	 {arg = arg, argType = argType, body = body}
	 
      fun plist(Lam{plist, ...}) = plist
	     
      val layout = layoutLambda
      fun equals(f:t, f':t) = PropertyList.equals(plist f, plist f')
   end

(* ------------------------------------------------- *)
(*                     DirectExp                     *)
(* ------------------------------------------------- *)
structure DirectExp =
   struct
      open Dec PrimExp

      structure Cont =
	 struct
	    type t = PrimExp.t * Type.t -> Exp.t

	    fun nameGen(k: VarExp.t * Type.t -> Exp.t): t =
	       fn (e, t) =>
	       case e of
		  Var x => k(x, t)
		| _ => let val x = Var.newNoname()
		       in Exp.prefix(k(VarExp.mono x, t),
				     MonoVal{var = x, ty = t, exp = e})
		       end
	    
	    fun name(k: VarExp.t * Type.t -> Exp.t): t = nameGen k

	    val id: t = name(fn (x, _) => Exp{decs = [], result = x})
	       
	    fun return(k: t, xt) = k xt
	 end

      type t = Cont.t -> Exp.t

      fun layout _ = Layout.str "<Xml.DirectExp>"

      fun send(e: t, k: Cont.t): Exp.t = e k

      fun toExp e = send(e, Cont.id)

      fun fromExp(Exp{decs, result}, ty): t =
	 fn k => Exp.prefixs(k(Var result, ty), decs)

      fun sendName(e, k) = send(e, Cont.name k)

      fun simple e k = Cont.return(k, e)
	 
      fun const c = simple(Const c, Type.ofConst c)

      fun varExp(x, t) = simple(Var x, t)

      fun var{var, targs, ty} = varExp(VarExp.T{var = var, targs = targs}, ty)

      fun monoVar(x, t) = var{var = x, targs = [], ty = t}

      fun convertsGen(es: t list, k: (VarExp.t * Type.t) list -> Exp.t): Exp.t =
	 let
	    fun loop(es, xs) =
	       case es of
		  [] => k(List.reverse xs)
		| e :: es => sendName(e, fn x => loop(es, x :: xs))
	 in loop(es, [])
	 end

      fun converts(es: t list,
		   make: (VarExp.t * Type.t) list -> PrimExp.t * Type.t): t =
	 fn k => convertsGen(es, k o make)

      fun convert(e: t, make: VarExp.t * Type.t -> PrimExp.t * Type.t): t =
	 fn k => send(e, Cont.name(k o make))

      fun convertOpt(e, make) =
	 case e of
	    NONE => simple(make NONE)
	  | SOME e => convert(e, make o SOME o #1)

      fun tuple{exps: t list, ty: Type.t}: t =
	 case exps of
	    [e] => e
	  | _ => converts(exps, fn xs => (PrimExp.Tuple(List.map(xs, #1)), ty))

      fun select{tuple, offset, ty} =
	 convert(tuple, fn (tuple, _) =>
		 (Select{tuple = tuple, offset = offset}, ty))

      fun conApp{con, targs, arg, ty} =
	 convertOpt(arg, fn arg =>
		    (ConApp{con = con, targs = targs, arg = arg}, ty))

      fun primApp{prim, targs, args, ty} =
	 converts(args, fn args =>
		  (PrimApp{prim = prim,
			   targs = targs,
			   args = List.map(args, #1)},
		   ty))

      fun convert2(e1, e2, make) =
	 converts([e1, e2],
		  fn [x1, x2] => make(x1, x2)
		   | _ => Error.bug "convert2")
				     
      fun app{func, arg, ty} =
	 convert2(func, arg, fn ((func, _), (arg, _)) =>
		  (App{func = func, arg = arg}, ty))

      fun casee{test, cases, default, ty} =
	 convert(test, fn (test, _) =>
		 (Case
		  {test = test,
		   cases = List.map(cases, fn (p, e) => (p, toExp e)),
		   default = (case default of
				 NONE => NONE
			       | SOME e => SOME(toExp e))},
		  ty))

      fun raisee(e: t, t: Type.t): t =
	 convert(e, fn (x, _) => (Raise x, t))
(* 	 fn k =>
 * 	 (Raise x, t)
 * 	 in sendName(e, fn (x, _) => Cont.return(Cont.id, (Raise x, t)))
 *)
	 
      fun handlee{try, catch, handler, ty} =
	 simple(Handle{try = toExp try,
		       catch = catch,
		       handler = toExp handler},
		ty)

      fun unit() = tuple{exps = [], ty = Type.unit}

      fun reff(e: t): t =
	 convert(e, fn (x, t) =>
		 (PrimApp{prim = Prim.reff,
			  targs = [t],
			  args = [x]},
		  Type.reff t))

      fun deref(e: t): t =
	 convert(e, fn (x, t) =>
		 let val t = Type.deref t
		 in (PrimApp{prim = Prim.deref, targs = [t], args = [x]},
		     t)
		 end)

      fun equal(e1, e2) =
	 convert2(e1, e2, fn ((x1, t), (x2, _)) =>
		  (PrimApp{prim = Prim.equal,
			   targs = [t],
			   args = [x1, x2]},
		   Type.bool))

      fun iff{test, thenn, elsee, ty} =
	 casee{test = test,
	       cases = [(Pat.truee, thenn), (Pat.falsee, elsee)],
	       default = NONE,
	       ty = ty}

      fun vall{var, exp}: Dec.t list =
	 let val t = ref Type.unit
	    val Exp{decs, result} =
	       sendName(exp, fn (x, t') => (t := t';			 
					    Exp{decs = [], result = x}))
	 in decs @ [MonoVal{var = var, ty = !t, exp = Var result}]
	 end

      fun sequence es =
    	 converts(es, fn xs => let val (x, t) = List.last xs
			       in (Var x, t)
			       end)

      fun seq(es, make) =
	 fn k => convertsGen(es, fn xts => send(make(List.map(xts, varExp)), k))

      fun lett{decs, body} = fn k => Exp.prefixs(send(body, k), decs)

      fun let1{var, exp, body} =
	 fn k => 
	 send(exp, fn (exp, ty) =>
	      Exp.prefix(send(body, k),
			 Dec.MonoVal{var = var, ty = ty, exp = exp}))
			  

      fun lambda{arg, argType, body, bodyType} =
	 simple(Lambda(Lambda.new{arg = arg, argType = argType,
				  body = toExp body}),
		Type.arrow(argType, bodyType))

      fun detupleGen(e: PrimExp.t, t: Type.t, components: Var.t list,
		     body: Exp.t): Exp.t =
	 Exp.prefixs
	 (body,
	  case components of
	     [] => []
	   | [x] => [MonoVal{var = x, ty = t, exp = e}]
	   | _ =>
		let val ts = Type.detuple t
		   val tupleVar = Var.newNoname()
		in MonoVal{var = tupleVar, ty = t, exp = e}
		   ::
		   List.mapi(List.zip(components, ts), fn (i, (x, t)) =>
			     MonoVal{var = x, ty = t,
				     exp = Select{tuple = VarExp.mono tupleVar,
						  offset = i}})
		end)
      
      fun detupleBind{tuple, components, body} =
	 fn k => send(tuple, fn (e, t) => detupleGen(e, t, components, body k))

      fun detuple{tuple: t, body}: t =
	 fn k =>
	 tuple
	 (fn (e, t) =>
	  let val ts = Type.detuple t
	  in case e of
	     Tuple xs => send(body(List.zip(xs, ts)), k)
	   | _ => let val components = List.map(ts, fn _ => Var.newNoname())
		  in detupleGen(e, t, components,
				send(body(List.map2(components, ts, fn (x, t) =>
						    (VarExp.mono x, t))),
				     k))
		  end
	  end)
   end
structure Exp =
   struct
      open Exp
	 
      fun unit() =
	 let open DirectExp
	 in toExp(tuple{exps = [], ty = Type.unit})
	 end
   end

(*---------------------------------------------------*)
(*                     Datatype                      *)
(*---------------------------------------------------*)

structure Datatype =
   struct
      type t = {tycon: Tycon.t,
		tyvars: Tyvar.t list,
		cons: {con: Con.t, arg: Type.t option} list}

      fun toAst({tyvars, tycon, cons}:t) =
	 {tyvars = tyvars,
	  tycon = Tycon.toAst tycon,
	  cons = List.map(cons, fn {con, arg} =>
			  (Con.toAst con, Type.optionToAst arg))}
   end

(*---------------------------------------------------*)
(*                      Program                      *)
(*---------------------------------------------------*)

structure Program =
   struct
      datatype t = T of {datatypes: Datatype.t list,
			 body: Exp.t}

      fun size(T{body, ...}) = Exp.size body

      fun toAst(T{datatypes, body}) =
	 let val body = Exp.toAst body
	 in case datatypes of
	    [] => body
	  | _ => Aexp.lett([Adec.datatypee(List.map(datatypes, Datatype.toAst))],
			   body)
	 end

      fun clear(T{datatypes, body}) =
	 (List.foreach(datatypes, fn {tycon, tyvars, cons} =>
		       (Tycon.clear tycon;
			List.foreach(tyvars, Tyvar.clear);
			List.foreach(cons, Con.clear o #con)));
	  Exp.clear body)

      fun layout p = Ast.Exp.layout(toAst p)

      val empty = T{datatypes = [],
		    body = Exp.unit()}

      fun layoutStats(T{datatypes, body}) =
	 let
	    val numTypes = ref 0
	    fun inc _ = numTypes := 1 + !numTypes
	    val {hom, destroy} = Type.makeHom{var = inc, con = inc}
	    val numPrimExps = ref 0
	    open Layout
	 in
	    List.foreach(datatypes, fn {cons, ...} =>
			 List.foreach(cons, fn {arg, ...} =>
				      case arg of
					 NONE => ()
				       | SOME t => hom t))
	    ; (Exp.foreach
	       {exp = body,
		handlePrimExp = fn _ => numPrimExps := 1 + !numPrimExps,
		handleVarExp = fn _ => (),
		handleBoundVar = hom o #3,
		handleExp = fn _ => ()})
	    ; destroy()
	    ; align[seq[str "size = ", Int.layout(!numPrimExps)],
		    seq[str "num types in program = ", Int.layout(!numTypes)],
		    Type.stats()]
	 end
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature XML_STRUCTS =
   sig
      include XML_TREE_STRUCTS
   end

signature XML =
   sig
      include XML_TREE

      val simplify: Program.t -> Program.t
      val typeCheck: Program.t -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_CHECK_STRUCTS = 
   sig
      structure XmlTree: XML_TREE
   end

signature TYPE_CHECK = 
   sig
      include TYPE_CHECK_STRUCTS
      
      val typeCheck: XmlTree.Program.t -> unit
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeCheck(S: TYPE_CHECK_STRUCTS): TYPE_CHECK = 
struct

open S
open XmlTree   
open Dec PrimExp

fun typeCheck(program as Program.T{datatypes, body}): unit =
   let
      val {get = getCon: Con.t -> {tyvars: Tyvar.t list, ty: Type.t}, set} =
	 Property.new(Con.plist, Property.initRaise("scheme", Con.layout))

      fun setCon({con, arg}, tyvars, result) =
	 set(con, {tyvars = tyvars,
		   ty = (case arg of
			    NONE => result
			  | SOME ty => Type.arrow(ty, result))})

      fun checkConExp(c: Con.t, ts: Type.t list): Type.t =
	 let val {tyvars, ty} = getCon c
	 in Type.substitute(ty, List.zip(tyvars, ts))
	 end
	 
      val {get = getVar: Var.t -> {tyvars: Tyvar.t list, ty: Type.t},
	   set = setVar} =
	 Property.new(Var.plist, Property.initRaise("var scheme", Var.layout))
	 
      (* val getVar = Trace.trace("getVar", Var.layout, Layout.ignore) getVar *)
      (* val setVar = Trace.trace2("setVar", Var.layout, Layout.ignore, Layout.ignore) setVar *)

      fun checkVarExp(VarExp.T{var, targs}): Type.t =
	 let val {tyvars, ty} = getVar var
	 in if length targs = length tyvars
	       then Type.substitute(ty, List.zip(tyvars, targs))
	    else
	       Type.error("variable applied to wrong number of type args",
			  let open Layout
			  in align[Var.layout var,
				   seq[str "tyvars: ",
				       List.layout Tyvar.layout tyvars],
				   seq[str "targs: ",
				       List.layout Type.layout targs]]
			    end)
	 end

      fun checkVarExps xs = List.map(xs, checkVarExp)

      fun checkPat(p as Pat.T{con, targs, arg}): Type.t =
	 let val t = checkConExp(con, targs)
	 in
	    case (arg,         Type.dearrowOpt t) of
	         (NONE,        NONE)              => t
	       | (SOME(x, ty), SOME(t1, t2))      => 
		    if Type.equals(t1, ty)
		       then (setVar(x, {tyvars = [], ty = t1}) ; t2)
		    else Type.error("argument constraint of wrong type",
				    let open Layout
				    in align[seq[str "t1: ", Type.layout t1],
					     seq[str "ty: ", Type.layout ty],
					     seq[str "p: ", Pat.layout p]]
				    end)
	       | _ => Type.error("constructor pattern mismatch", Pat.layout p)
	 end
	 
(*       val traceCheckExp =
 * 	 Trace.trace("Xml.checkExp", Exp.layout, Type.layout)
 * 
 *       val traceCheckPrimExp = 
 * 	 Trace.trace2
 * 	 ("Xml.checkPrimExp", PrimExp.layout, Type.layout, Type.layout)
 *)

      local
	 val exnType = ref NONE
      in
	 fun isExnType t =
	    case !exnType of
	       NONE => (exnType := SOME t; true)
	     | SOME t' => Type.equals(t, t')
      end

      fun check(t: Type.t, t': Type.t, layout: unit -> Layout.t): unit =
	 if Type.equals(t, t')
	    then ()
	 else Type.error("type mismatch",
			 Layout.align[Type.layout t,
				      Type.layout t'])

      fun checkExp arg: Type.t =
(*	 traceCheckExp *)
	 (fn (exp: Exp.t) =>
	 let val {decs, result} = Exp.dest exp
	 in List.foreach(decs, checkDec) ; checkVarExp result
	 end) arg
      and checkPrimExp arg: Type.t =
(*	 traceCheckPrimExp *)
	 (fn (e: PrimExp.t, ty: Type.t) => 
	 let
	    fun error msg =
	       Type.error(msg, let open Layout
			       in seq[str "exp: ", PrimExp.layout e]
			       end)
	    fun checkApp(t1, x) =
	       let val t2 = checkVarExp x
	       in case Type.dearrowOpt t1 of
		  SOME(t2', t3) =>
		     if Type.equals(t2, t2') then t3
		     else
			Type.error("actual and formal not of same type",
				   let open Layout
				   in align[seq[str "actual: ", Type.layout t2],
					    seq[str "formal: ", Type.layout t2']]
				   end)
		| NONE => error "function not of arrow type"
	       end
	    
	    fun checkApps(t, es) = List.fold(es, t, fn (e, t) => checkApp(t, e))
	 in case e of
	    Var x => checkVarExp x
	  | Const c => Type.ofConst c
	  | Tuple[_] => error "unary tuple"
	  | Tuple xs => Type.tuple(checkVarExps xs)
	  | Select{tuple, offset} =>
	       (case Type.detupleOpt(checkVarExp tuple) of
		   SOME ts => List.nth(ts, offset)
		 | NONE => error "selection from nontuple")
	  | Lambda l => checkLambda l
	  | PrimApp{prim, targs, args} =>
	       (case Prim.checkApp{prim = prim,
				   targs = targs,
				   args = checkVarExps args,
				   con = Type.con,
				   equals = Type.equals,
				   dearrowOpt = Type.dearrowOpt,
				   detupleOpt = Type.detupleOpt,
				   isUnit = Type.isUnit
				   } of
		   NONE => error "bad primapp"
		 | SOME t => t)
	  | ConApp{con, targs, arg} =>
	       let val t = checkConExp(con, targs)
	       in case arg of
		  NONE => t
		| SOME e => checkApp(t, e)
	       end
	  | App{func, arg} => checkApp(checkVarExp func, arg)
	  | Raise x => if isExnType(checkVarExp x)
			  then ty
		       else error "bad raise"
	  | Handle{try, catch = (catch, catchType), handler} =>
	       let val ty = checkExp try
		  val _ = setVar(catch, {tyvars = [],
					 ty = catchType})
		  val ty' = checkExp handler
	       in if Type.equals(ty, ty') then ty
		  else error "bad handle"
	       end
	  | Case{test, cases, default} =>
	       let val ty = checkVarExp test
		  val (ptys, etys) =
		     List.unzip
		     (List.map(cases, fn (p, e) =>
			       (checkPat p, checkExp e)))
	       in case(Type.equalss ptys, Type.equalss etys) of
		  (NONE, _) => error "patterns not of same type"
		| (_, NONE) => error "branches not of same type"
		| (SOME pty, SOME ety) =>
		     if Type.equals(ty, pty)
			then
			   case default of
			      NONE => ety
			    | SOME e =>
				 if Type.equals(ety, checkExp e)
				    then ety
				 else error "default of wrong type"
		     else error "test and patterns of different types"
	       end
	 end) arg
      
      and checkLambda l: Type.t =
	 let val {arg, argType, body} = Lambda.dest l
	 in setVar(arg, {tyvars = [], ty = argType}) ;
	    Type.arrow(argType, checkExp body)
	 end
	    
      and checkDec d =
	 let val check = fn (t, t') => check(t, t', fn () => Dec.layout d)
	 in case d of
	    Exception c => setCon(c, [], Type.exn)
	  | MonoVal{var, ty, exp} =>
	       (check(ty, checkPrimExp(exp, ty));
		setVar(var, {tyvars = [], ty = ty}))
	  | PolyVal{tyvars, var, ty, exp} =>
	       (case tyvars of
		   [] => Error.bug "empty tyvars in PolyVal dec"
		 | _ => ();
		check(ty, checkExp exp);
		setVar(var, {tyvars = tyvars, ty = ty}))
	  | Fun{tyvars, decs} =>
	       (List.foreach(decs, fn {var, ty, lambda} =>
			     setVar(var, {tyvars = tyvars, ty = ty}));
		List.foreach
		(decs, fn {ty, lambda, ...} =>
		 let val {arg, argType, body} = Lambda.dest lambda
		 in setVar(arg, {tyvars = [], ty = argType});
		    check(ty, Type.arrow(argType, checkExp body))
		 end))
	 end
	       
   in List.foreach(datatypes, fn {tycon, tyvars, cons} =>
		   let val ty = Type.con(tycon, List.map(tyvars, Type.var))
		   in List.foreach(cons, fn c => setCon(c, tyvars, ty))
		   end)
      ; if Type.equals(checkExp body, Type.unit)
	   then ()
	else Error.bug "program must be of type unit"
      ; Program.clear program
   end

(* val typeCheck =
 *    Trace.trace("Xml.typeCheck", Program.layout, Unit.layout) typeCheck
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SIMPLIFY_TYPES_STRUCTS = 
   sig
      include XML_TREE
   end

signature SIMPLIFY_TYPES = 
   sig
      include SIMPLIFY_TYPES_STRUCTS
      
      val simplifyTypes: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor SimplifyTypes(S: SIMPLIFY_TYPES_STRUCTS): SIMPLIFY_TYPES = 
struct

open S

structure Graph = DirectedGraph
structure Node = Graph.Node

fun simplifyTypes(p as Program.T{datatypes, body}) =
   let
      val g = Graph.new()
	 
      val {get = tyconInfo: Tycon.t -> {node: Node.t,
					isOneVariantArrow: bool ref,
					cons: {con: Con.t,
					       arg: Type.t option
					       } list
					} option,
	   set = setTyconInfo, destroy = destroyTycon} =
	 Property.newDest(Tycon.plist, Property.initConst NONE)

      val {get = nodeTycon, set = setNodeTycon} =
	 Property.new(Node.plist, Property.initRaise("tycon", Node.layout))

      val _ =
	 (List.foreach
	  (datatypes, fn {tycon, cons, ...} =>
	   let val node = Graph.newNode g
	   in setTyconInfo(tycon, SOME{node = node,
				       isOneVariantArrow = ref false,
				       cons = cons})
	      ; setNodeTycon(node, tycon)
	   end)
	  ; (List.foreach
	     (datatypes, fn {tycon, cons, ...} =>
	      let
		 val {node = from, ...} = valOf(tyconInfo tycon)
		 fun loop(t: Type.t): unit =
		    case Type.dest t of
		       Type.Var _ => ()
		     | Type.Con(tycon', ts) =>
			  (if Tycon.equals(tycon, tycon')
			      then (case tyconInfo tycon' of
				       NONE => ()
				     | SOME{node = to, ...} =>
					     Graph.addEdge
					     (g, {from = from, to = to}))
			   else ()
			   ; List.foreach(ts, loop))
	      in List.foreach(cons, fn {arg, ...} =>
			      case arg of
				 NONE => ()
			       | SOME t => loop t)
	      end)))

      fun num(datatypes, p) =
	 List.fold(datatypes, 0, fn (d, n) => if p d then n + 1 else n)

      val numDatatypes = List.length datatypes

      val arrowDatatypes =
	 List.keepAll(datatypes, fn {cons, ...} =>
		      List.exists(cons, fn {arg, ...} =>
				  case arg of
				     NONE => false
				   | SOME t =>
					Type.containsTycon(t, Atoms.Tycon.arrow)))

      val numArrowDatatypes = List.length arrowDatatypes

      val oneVariantArrows = List.keepAll(arrowDatatypes, fn {cons, ...} =>
					  1 = List.length cons)

      val numOneVariantArrows = List.length oneVariantArrows

      val _ =
	 List.foreach(oneVariantArrows, fn {tycon, ...} =>
		      let val {isOneVariantArrow, ...} = valOf(tyconInfo tycon)
		      in isOneVariantArrow := true
		      end)
	 
      val components =
	 Graph.stronglyConnectedComponents g
	 
      val numEliminable =
	 List.fold
	 (components, 0, fn (nodes, n) =>
	  case nodes of
	     [node] =>
		if List.exists(Node.successors node, fn node' =>
			       Node.equals(node, node'))
		   then n
		else
		   let
		      val {isOneVariantArrow, ...} =
			 valOf(tyconInfo(nodeTycon node))
		   in if !isOneVariantArrow
			 then n + 1
		      else n
		   end
	      | _ => n)

   in Control.message
      (fn () => let open Layout
		in align[seq[str "datatypes: ", Int.layout numDatatypes],
			 seq[str "-> datatypes: ", Int.layout numArrowDatatypes],
			 seq[str "one variants: ", Int.layout numOneVariantArrows],
			 seq[str "eliminable: ", Int.layout numEliminable]]
		end)
      ; destroyTycon()
      ; p
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Compute strongly connected components on fun decs to make them
 * as small as possible.
 *)

signature SCC_FUNS_STRUCTS = 
   sig
      include XML_TREE
   end

signature SCC_FUNS = 
   sig
      include SCC_FUNS_STRUCTS
      
      val sccFuns: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor SccFuns(S: SCC_FUNS_STRUCTS): SCC_FUNS = 
struct

open S
open Dec PrimExp

structure Graph = DirectedGraph
structure Node = Graph.Node

fun sccFuns(Program.T{datatypes, body}) =
   let
      (* For each function appearing in a fun dec record its node, which will
       * have edges to the nodes of other functions declared in the same dec
       * if they appear in it's body.
       *)
      val {get = funInfo: Var.t -> {
				    node: Node.t,
				    visit: (unit -> unit) ref
				    } option,
	   set = setFunInfo} =
	 Property.new(Var.plist, Property.initConst NONE)

      val {get = nodeLambda, set = setNodeLambda} =
	 Property.new(Node.plist, Property.initRaise("lambda", Node.layout))

      fun loopVar x =
	 case funInfo x of
	    NONE => ()
	  | SOME{visit, ...} => !visit()

      val loopVarExp = loopVar o VarExp.var

      fun loopVarExps xs = List.foreach(xs, loopVarExp)

      fun loopLambda(l: Lambda.t): Lambda.t =
	 let val {arg, argType, body} = Lambda.dest l
	 in Lambda.new{arg = arg, argType = argType, body = loopExp body}
	 end

      and loopPrimExp(e: PrimExp.t): PrimExp.t =
	 case e of
	    Const _ => e
	  | Var x => (loopVarExp x; e)
	  | Tuple xs => (loopVarExps xs; e)
	  | Select{tuple, ...} => (loopVarExp tuple; e)
	  | Lambda l => Lambda(loopLambda l)
	  | ConApp{arg, ...} => (Option.map'(arg, loopVarExp); e)
	  | PrimApp{args, ...} => (loopVarExps args; e)
	  | App{func, arg} => (loopVarExp func; loopVarExp arg; e)
	  | Raise x => (loopVarExp x; e)
	  | Case{test, cases, default} =>
	       (loopVarExp test;
		Case{test = test,
		     cases = List.map(cases, fn (p, e) => (p, loopExp e)),
		     default = Option.map(default, loopExp)})
	  | Handle{try, catch, handler} =>
	       Handle{try = loopExp try,
		      catch = catch,
		      handler = loopExp handler}
	       
      and loopExp(e: Exp.t): Exp.t =
	 let val {decs, result} = Exp.dest e
	    val decs =
	       List.reverse
	       (List.fold
		(decs, [], fn (dec, decs) =>
		 case dec of
		    MonoVal{var, ty, exp} =>
		       MonoVal{var = var, ty = ty,
			       exp = loopPrimExp exp} :: decs
		  | PolyVal{var, tyvars, ty, exp} =>
		       PolyVal{var = var, tyvars = tyvars, ty = ty,
			       exp = loopExp exp} :: decs
		  | Exception _ => dec :: decs
		  | Fun{tyvars, decs = lambdas} =>
		       let val g = Graph.new()
			  val _ =
			     List.foreach
			     (lambdas, fn {var, ty, lambda} =>
			      setFunInfo(var, SOME{node = Graph.newNode g,
						   visit = ref ignore}))
			  val _ = 
			     List.foreach
			     (lambdas, fn {var, ty, lambda} =>
			      let val {node = from, ...} = valOf(funInfo var)
			      in List.foreach
				 (lambdas, fn {var = x, ...} =>
				  let val {visit, node = to} = valOf(funInfo x)
				  in visit := (fn () =>
					       (Graph.addEdge(g, {from = from,
								  to = to})
						; visit := ignore))
				  end)
				 ; (setNodeLambda
				    (from, {var = var,
					    ty = ty,
					    lambda = loopLambda lambda}))
				 ; (List.foreach
				    (lambdas, fn {var, ...} =>
				     let val {visit, ...} = valOf(funInfo var)
				     in visit := ignore
				     end))
			      end)
		       in List.map
			  (Graph.stronglyConnectedComponents g, fn nodes =>
			   Fun{tyvars = tyvars,
			       decs = List.map(nodes, nodeLambda)})
			  @ decs
		       end))
	 in loopVarExp result
	    ; Exp.new{decs = decs, result = result}
	 end
   in Program.T{datatypes = datatypes,
		body = loopExp body}
   end
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SIMPLIFY_STRUCTS = 
   sig
      structure XmlTree: XML_TREE
   end

signature SIMPLIFY = 
   sig
      include SIMPLIFY_STRUCTS
      
      val simplify: XmlTree.Program.t -> XmlTree.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* This simplifier is based on the following article:
 * Shrinking Lambda Expressions in Linear Time.
 * Journal of Functional Programming. Vol 7, no 5, 1997.
 *)

functor Simplify(S: SIMPLIFY_STRUCTS): SIMPLIFY = 
struct

open S
open XmlTree
open Dec PrimExp

(* val traceSimplifyExp =
 *    Trace.trace("simplifyExp", Exp.layout, Exp.layout)
 * 
 * val traceSimplifyDecs =
 *    Trace.trace("simplifyDecs", List.layout Dec.layout, List.layout Dec.layout)
 * 
 * val traceSimplifyLambda =
 *    Trace.trace("simplifyLambda", Lambda.layout, Lambda.layout)
 *)
fun simplifyOnce e =
   let
      (* counts the number of modifications made by the simplifier *)
      val numMods = ref 0
      fun incmod() = IntRef.inc numMods

      (*---------------------------------------------------*)
      (*                       sigma                       *)
      (*---------------------------------------------------*)
      (* sigma is a mapping from variables to variables.
       *  While descending the program all occurrences of x will be replaced by
       *  sigma(x).
       *  For example, after seeing let val x = y in e
       *  sigma(x) = y
       *  Then e will be processed and all occurrences of x will be replaced by y
       *)
	 
      local
	 val {get: Var.t -> VarExp.t option, set} =
	    Property.new(Var.plist, Property.initConst NONE)
      in
	 fun setSigma(x, y) = set(x, SOME y)

	 fun simplifyVarExp(x: VarExp.t): VarExp.t =
	    case get(VarExp.var x) of
	       NONE => x
	     | SOME y => y
		  
	 fun sigma(x: Var.t): Var.t =
	    case get x of
	       NONE => x
	     | SOME y => VarExp.var y
      end

      fun simplifyVarExps xs = List.map(xs, simplifyVarExp)

      (*---------------------------------------------------*)
      (*                      census                       *)
      (*---------------------------------------------------*)
      (* census walks over an expression and modifies variable occurrence counts,
       * which are stored on the property list associated with the variable
       *)

      (* only keep track of monomorphic variables, since the simplifier ignores
       * polymorphic ones
       *)
	 
      val {set = setNumOccurrences, get = numOccurrences} =
	 Property.new(Var.plist,
		      Property.initRaise("num occurrences", Var.layout))

      val setNumOccurrences =
	 fn (x, n) => (Assert.cheap(n >= 0); setNumOccurrences(x, n))
	    
(*        val numOccurrences =
 *  	 Trace.trace("numOccurrences", Var.layout, Int.layout) numOccurrences
 *)

      fun incNumOccurrences(x, n) =
	 setNumOccurrences(x, n + numOccurrences x)

(*       val incNumOccurrences =
 * 	 Trace.trace2("incNumOccurrences",
 *  		      Var.layout, Int.layout, Unit.layout) incNumOccurrences
 *)

      fun maybeIncNumOccurrences(VarExp.T{var, targs}, n) =
	 case targs of
	    [] => incNumOccurrences(var, n)
	  | _ => ()

      fun setVar(x: Var.t, y: VarExp.t): unit =
	 (maybeIncNumOccurrences(y, numOccurrences x)
	  ; setSigma(x, y))

      fun isUseless var = numOccurrences var = 0

      fun countExp(e: Exp.t): unit =
	 Exp.foreach{exp = e,
		     handleBoundVar = fn (x, [], _) => setNumOccurrences(x, 0)
				       | _ => (),
		     handleVarExp = fn x => maybeIncNumOccurrences(x, 1),
		     handleExp = fn _ => (), handlePrimExp = fn _ => ()}

      fun deleteVarExp(x: VarExp.t): unit =
	 maybeIncNumOccurrences(simplifyVarExp x, ~1)
	      
      fun deleteExp(e: Exp.t): unit =
	 Exp.foreachVarExp(e, deleteVarExp)

      fun deleteVarExps xs = List.foreach(xs, deleteVarExp)
      fun deleteLambda l = deleteExp(Lambda.body l)
	 
      (*---------------------------------------------------*)
      (*                    simplifyExp                    *)
      (*---------------------------------------------------*)

      local
	 val {get: Var.t -> bool, set} =
	    Property.new(Var.plist, Property.initConst false)
      in
	 val isInlined = get
	 fun setInlined x = set(x, true)
      end

      val {get = bind: Var.t -> PrimExp.t option, set} =
	 Property.new(Var.plist, Property.initConst NONE)

      fun removeBind x = set(x, NONE)
      fun setBind(x, b) = set(x, SOME b)

      val bind = bind o VarExp.var
	 
      fun simplifyExp arg: Exp.t =
(*	 traceSimplifyExp *)
	 (fn (e: Exp.t) =>
	  let val {decs, result} = Exp.dest e
	  in Exp.new{decs = simplifyDecs decs,
		     result = simplifyVarExp result}
	  end) arg
      and simplifyDecs(decs: Dec.t list): Dec.t list =
	 case decs of
	    [] => []
	  | dec :: decs =>
	       case dec of
		  Dec.Exception _ => dec :: (simplifyDecs decs)
		| Dec.PolyVal{var, tyvars, ty, exp} =>
		     Dec.PolyVal{var = var, tyvars = tyvars, ty = ty,
				 exp = simplifyExp exp}
		     :: simplifyDecs decs
		| Dec.Fun{tyvars, decs = decs'} =>
		     (case tyvars of
			 _ :: _ =>
			    Dec.Fun{tyvars = tyvars,
				    decs =
				    List.map(decs', fn {var, ty, lambda} =>
					     {var = var,
					      ty = ty,
					      lambda = simplifyLambda lambda})}
			    :: simplifyDecs decs
			| [] => 
			     let
				val decs' =
				   List.fold
				   (decs', [], fn (dec as {var, ty, lambda}, decs) =>
				    if isUseless var
				       then (deleteLambda lambda ; decs)
				    else (setBind(var, PrimExp.Lambda lambda) ;
					  dec :: decs))
				val decs = simplifyDecs decs
				(* need to walk over all the decs to removeBind before
				 * simplifying any of them because they are mutually
				 * recursive
				 *)
				val decs' =
				   List.fold
				   (decs', [], fn (dec as {var, lambda, ...}, decs) =>
				    if isInlined var
				       then decs
				    else if isUseless var
					    then (deleteLambda lambda ; decs)
					 else (removeBind var ; dec :: decs))
			     in case decs' of
				[] => decs
			      | _ => Dec.Fun{tyvars = tyvars,
					     decs =
					     List.map(decs', fn {var, ty, lambda} =>
						      {var = var,
						       ty = ty,
						       lambda = simplifyLambda lambda})}
				   :: decs
			     end)
		| Dec.MonoVal{var, ty, exp} =>
		     let
			fun finish(exp, decs) =
			   Dec.MonoVal{var = var, ty = ty, exp = exp} :: decs
			fun nonExpansive(delete: unit -> unit,
					 set: unit -> (unit -> PrimExp.t) option) =
			   if isUseless var
			      then (incmod(); delete(); simplifyDecs decs)
			   else let val s = set()
				    val decs = simplifyDecs decs
				in if isUseless var
				      then (incmod(); delete(); decs)
				   else (case s of
					    NONE => decs
					  | SOME n => finish(n(), decs))
				end
			fun expansive(e: PrimExp.t) = finish(e, simplifyDecs decs)
			fun nonExpansiveCon(delete, e: PrimExp.t) =
			   nonExpansive(delete,
					fn () => (setBind(var, e) ;
						  SOME(fn () => e)))
			fun expression(e: Exp.t): Dec.t list =
			   let val {decs=decs', result} = Exp.dest(simplifyExp e)
			      val _ = setVar(var, result)
			      val decs = simplifyDecs decs
			   in decs' @ decs
			   end
		     in case exp of
			Const _ => nonExpansive(fn () => (),
						fn () => SOME(fn () => exp))
		      | Var x =>
			   let val x = simplifyVarExp x
			   in setSigma(var, x)
			      ; maybeIncNumOccurrences(x, numOccurrences var - 1)
			      ; incmod()
			      ; simplifyDecs decs
			   end
		      | Tuple xs =>
			   let val xs = simplifyVarExps xs
			   in nonExpansiveCon(fn () => deleteVarExps xs,
					      Tuple xs)
			   end
		      | Select{tuple, offset} =>
			   let val tuple = simplifyVarExp tuple
			   in nonExpansive
			      (fn () => deleteVarExp tuple,
			       fn () =>
			       (case bind tuple of
				   NONE =>
				      SOME(fn () => Select{tuple = tuple,
							   offset = offset})
				 | SOME(Tuple xs) => 
				      (incmod()
				       ; deleteVarExp tuple
				       ; setVar(var, List.nth(xs, offset))
				       ; NONE)
				 | _ => Error.bug "simplifyPrimExp: Select"))
			   end
		      | ConApp{con, targs, arg} =>
			   let val arg = (case arg of
					     NONE => NONE
					   | SOME x => SOME(simplifyVarExp x))
			   in nonExpansiveCon
			      (fn () => (case arg of
					    NONE => ()
					  | SOME x => deleteVarExp x),
			       ConApp{con = con, targs = targs, arg = arg})
			   end			     
		      | PrimApp{prim, args, targs} =>
			   (case Prim.numArgs prim of
			       NONE =>
				  nonExpansive(fn () => (),
					       fn () => SOME(fn () =>
							     PrimApp{prim = prim,
								     targs = targs,
								     args = []}))
			     | SOME _ => 
				  expansive(PrimApp{prim = prim, targs = targs,
						    args = simplifyVarExps args}))
		      | Lambda l =>
			   nonExpansive
			   (fn () => if isInlined var
					then ()
				     else deleteLambda l,
					fn () => (setBind(var, exp)
						  ; SOME(fn () => Lambda(simplifyLambda l))))
		      | App{func, arg} =>
			   let val func = simplifyVarExp func
			      val arg = simplifyVarExp arg
			      fun normal() =
				 expansive(App{func = func, arg = arg})
			      val VarExp.T{var = funcVar, targs} = func
			   in case targs of
			      [] =>
				 if 1 = numOccurrences funcVar
				    then (case bind func of
					     NONE => normal()
					   | SOME(Lambda l) =>
						(incmod()
						 ; deleteVarExp arg
						 ; setVar(Lambda.arg l, arg)
						 ; setInlined funcVar
						 ; setNumOccurrences(funcVar, 0)
						 ; expression(Lambda.body l))
					   | _ =>
						Error.bug "simplifyPrimExp: App")
				 else normal()
			     | _ => normal()
			   end
		      | Raise x => expansive(Raise(simplifyVarExp x))
		      | Handle{try, catch, handler} =>
			   expansive(Handle{try = simplifyExp try,
					    catch = catch,
					    handler = simplifyExp handler})
		      | Case{test, cases, default} =>
			   let
			      fun match(f: Pat.t -> bool): Dec.t list =
				 let
				    fun loop cases =
				       case cases of
					  (pat, e) :: cases =>
					     if f pat
						then
						   (List.foreach(cases, fn (_, e) =>
								 deleteExp e) ;
						    case default of
						       NONE => ()
						     | SOME e => deleteExp e ;
							  expression e)
					     else (deleteExp e ; loop cases)
						     | []  => (case default of
								  SOME e => expression e
								| NONE => Error.bug "simplifyPrimExp: Case")
				 in incmod();
				    deleteVarExp test;
				    loop cases
				 end
			      val test = simplifyVarExp test
			   in case bind test of
			      SOME(ConApp{con = c, arg = NONE, ...}) =>
				 match(fn Pat.T{con = c', arg = NONE, ...} =>
				       Con.equals(c, c')
			       | _ => false)
			    | SOME(ConApp{con = c, arg = SOME x, ...}) =>
				 match(fn Pat.T{con = c', arg = SOME(x', _), ...} =>
				       if Con.equals(c, c')
					  then (setVar(x', x) ; true)
				       else false
			       | _ => false)
			    | NONE =>
				 expansive(Case{test = test,
						cases = List.map(cases, fn (p, e) =>
								 (p, simplifyExp e)),
						default = (case default of
							      NONE => NONE
							    | SOME e => SOME(simplifyExp e))})
			    | _ => Error.bug "simplifyPrimExp: Case2"
			   end
		     end

      and simplifyLambda l: Lambda.t =
(*	 traceSimplifyLambda *)
	 (fn l => 
	  let val {arg, argType, body} = Lambda.dest l
	  in Lambda.new{arg = arg,
			argType = argType,
			body = simplifyExp body}
	  end) l

      val _ = countExp e

      val e = simplifyExp e
   in
      Exp.clear e;

      Control.messageStr("simplify made " ^ Int.toString(!numMods)
			 ^ " modifications");
      
      (e, !numMods)
   end

(* val simplifyOnce =
 *    Trace.trace("simplifyOnce", Exp.layout,
 * 	       Layout.tuple2(Exp.layout, Int.layout)) simplifyOnce
 *)

fun simplify(Program.T{datatypes, body}): Program.t =
   Program.T
   {datatypes = datatypes,
    body = #1(simplifyOnce(#1(simplifyOnce body)))}

(* val simplify =
 *    Trace.trace("Xml.simplify", Program.layout, Program.layout) simplify
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Xml(S: XML_STRUCTS): XML =
   struct
      structure XmlTree = XmlTree(S)

      open XmlTree
      structure TypeCheck = TypeCheck(structure XmlTree = XmlTree)
      val typeCheck = TypeCheck.typeCheck

      structure SimplifyTypes = SimplifyTypes(open XmlTree)

      structure SccFuns = SccFuns(open XmlTree)

      structure Simplify = Simplify(structure XmlTree = XmlTree)
      val simplify =
	 Simplify.simplify
(* SimplifyTypes doesn't do anything yet.
 *       o SimplifyTypes.simplifyTypes
 *)
	 o SccFuns.sccFuns
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SXML = XML

(*
 invariants:
 * no Exception decs
 * no PolyVal decs
 * hasPrimExns = false
 * tyvar lists in datatype, val, and fun decs are always empty
 * type lists are empty for variable args and for non-built-in-constructor args
 * no tyvars in types
 *)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature POLYVARIANCE_STRUCTS = 
   sig
      include SXML
   end

signature POLYVARIANCE = 
   sig
      include POLYVARIANCE_STRUCTS
      
      val duplicate: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * Duplicate a let bound function at each variable reference
 * if cost is smaller than threshold.
 * 
 *)
functor Polyvariance(S: POLYVARIANCE_STRUCTS): POLYVARIANCE = 
struct

open S
open Dec PrimExp

structure Type =
   struct
      open Type

      fun containsArrow t = containsTycon(t, Tycon.arrow)

      fun isHigherOrder t =
	 case dearrowOpt t of
	    NONE => false
	  | SOME(t1, t2) => containsArrow t1 orelse isHigherOrder t2

(*       val isHigherOrder =
 * 	 Trace.trace("isHigherOrder", layout, Bool.layout) isHigherOrder
 *)
   end

fun lambdaSize(Program.T{body, ...}): Lambda.t -> int =
   let
      val {get = size: Lambda.t -> int, set} =
	 Property.new(Lambda.plist, Property.initRaise("size", Lambda.layout))
      fun loopExp(e: Exp.t, n: int): int =
	 List.fold
	 (Exp.decs e, n, fn (d, n) =>
	  case d of
	     MonoVal{exp, ...} => loopPrimExp(exp, n + 1)
	   | PolyVal{exp, ...} => loopExp(exp, n + 1)
	   | Fun{decs, ...} => List.fold(decs, n, fn ({lambda, ...}, n) =>
					 loopLambda(lambda, n))
	   | Exception _ => n + 1)
      and loopLambda(l: Lambda.t, n): int =
	 let val m = loopExp(Lambda.body l, 0)
	 in set(l, m); m + n
	 end
      and loopPrimExp(e: PrimExp.t, n: int): int =
	 case e of
	    Case{cases, default, ...} =>
	       List.fold
	       (cases,
		(case default of
		    NONE => n
		  | SOME e => loopExp(e, n)),
		fn ((_, e), n) => loopExp(e, n))
	  | Handle{try, handler, ...} =>
	      loopExp(try, loopExp(handler, n))
	  | Lambda l => loopLambda(l, n)
	  | _ => n
   in loopExp(body, 0)
      ; size
   end

val counter = Counter.new 0
	   
fun shouldDuplicate(program as Program.T{body, ...}, small, product)
   : Var.t -> bool =
   let
      val costs: (Var.t * int * int * int) list ref = ref []

      val lambdaSize = lambdaSize program
	 
      fun isOK(var: Var.t, size: int, numOccurrences: int): bool =
	 let val cost = (numOccurrences - 1) * (size - small)
	 in ListRef.push(costs, (var, size, numOccurrences, cost))
	    ; cost <= product
	 end
	 
      type info = {numOccurrences: int ref,
		   shouldDuplicate: bool ref}

      val {get = varInfo: Var.t -> info option, set = setVarInfo} =
	 Property.new(Var.plist, Property.initConst NONE)

      fun new{var, ty, lambda}: unit =
	 if Type.isHigherOrder ty
	    then setVarInfo(var, SOME{numOccurrences = ref 0,
				      shouldDuplicate = ref false})
	 else ()

      fun loopExp(e: Exp.t, numDuplicates: int): unit =
	 let
	    fun loopVar(x: VarExp.t): unit =
	       case varInfo(VarExp.var x) of
		  NONE => ()
		| SOME{numOccurrences, ...} =>
		     numOccurrences := !numOccurrences + numDuplicates
	    fun loopVars xs = List.foreach(xs, loopVar)
	    val {decs, result} = Exp.dest e
	    val rec loopDecs =
	       fn [] => loopVar result
		| dec :: decs =>
		     case dec of
			MonoVal{var, ty, exp} =>
			   (case exp of
			       Lambda l =>
				  (new{var = var, ty = ty, lambda = l}
				   ; loopDecs decs
				   ; let
					val body = Lambda.body l
					val numDuplicates =
					   case varInfo var of
					      NONE => numDuplicates
					    | SOME{numOccurrences,
						   shouldDuplicate} =>
					      if isOK(var, lambdaSize l,
						      !numOccurrences)
						 then (shouldDuplicate := true
						       ; !numOccurrences)
					      else numDuplicates
				     in loopExp(body, numDuplicates)
				     end)
			     | _ =>
				  let
				     val loopExp =
					fn e => loopExp(e, numDuplicates)
				  in (case exp of
					 Const _ => ()
				       | Var x => loopVar x
				       | Tuple xs => loopVars xs
				       | Select{tuple, ...} => loopVar tuple
				       | ConApp{arg, ...} =>
					    Option.map'(arg, loopVar)
				       | PrimApp{args, ...} => loopVars args
				       | App{func, arg} =>
					    (loopVar func; loopVar arg)
				       | Raise x => loopVar x
				       | Case{test, cases, default} =>
					    (loopVar test
					     ; (List.foreach(cases, loopExp o #2))
					     ; Option.map'(default, loopExp))
				       | Handle{try, handler, ...} =>
					    (loopExp try; loopExp handler)
				       | _ => Error.bug "unexpected primExp")
				    ; loopDecs decs
				  end)
		      | Fun{decs = lambdas, ...} =>
			   let
			      val _ = (List.foreach(lambdas, new)
				       ; loopDecs decs)
			      val dups =
				 List.fold
				 (lambdas, [], fn ({var, lambda, ...}, dups) =>
				  let val body = Lambda.body lambda
				  in case varInfo var of
				     NONE => (loopExp(body, numDuplicates); dups)
				   | SOME info =>
					{body = body,
					 size = lambdaSize lambda,
					 info = info} :: dups
				  end)
			   in case dups of
			      [] => ()
			    | _ => 
				 let
				    val size =
				       List.fold
				       (dups, 0, fn ({body, size, ...}, n) =>
					n + size)
				    val numOccurrences =
				       List.fold
				       (dups, 0,
					fn ({info = {numOccurrences, ...}, ...},
					    n) => n + !numOccurrences)
				 in if isOK((case lambdas of
						{var, ...} :: _ => var
					      | _ => Error.bug "empty lambdas"),
					    size, numOccurrences)
				       then (List.foreach
					     (dups,
					      fn {body,
						  info = {shouldDuplicate, ...},
						  ...} =>
					      (shouldDuplicate := true
					       ; loopExp(body, numOccurrences))))
				    else
				       List.foreach
				       (dups, fn {body, ...} =>
					loopExp(body, numDuplicates))
				 end
			   end
		      | _ => Error.bug "strange dec"
	 in loopDecs decs
	 end
   in loopExp(body, 1)
      ; let
	   val {insertionSort, ...} =
	      List.ordered{< = fn ((_, _, _, c), (_, _, _, c')) => c < c'}
	in Control.displays
	   ("threshold" ^ Int.toString(Counter.next counter),
	    fn layout => 
	    List.foreach
	    (insertionSort(!costs), fn (x, size, numOcc, c) =>
	     layout(let open Layout
		    in seq[Var.layout x,
			   str " ", Int.layout size,
			   str " ", Int.layout numOcc,
			   str " ", Int.layout c]
		    end)))
	end
      ; (fn x =>
	 case varInfo x of
	    NONE => false
	  | SOME{shouldDuplicate, ...} => !shouldDuplicate)
   end

fun duplicate(program as Program.T{datatypes, body}, small: int, product: int) =
   let
      val shouldDuplicate = shouldDuplicate(program, small, product)
      datatype info =
	 Replace of Var.t
       | Dup of {
		 duplicates: Var.t list ref
		 }
	 
      val {get = varInfo: Var.t -> info, set = setVarInfo} =
	 Property.new(Var.plist, Property.initRaise("info", Var.layout))

      fun loopVar(x: VarExp.t): VarExp.t =
	 VarExp.mono
	 (let val x = VarExp.var x
	  in case varInfo x of
	     Replace y => y
	   | Dup{duplicates, ...} =>
		let val x' = Var.new x
		in ListRef.push(duplicates, x')
		   ; x'
		end
	  end)

      fun loopVars xs = List.map(xs, loopVar)

      fun bind(x: Var.t): Var.t =
	 let val x' = Var.new x
	 in setVarInfo(x, Replace x')
	    ; x'
	 end

      fun bindVarType(x, t) = (bind x, t)

      fun bindPat(Pat.T{con, targs, arg}) =
	 Pat.T{con = con,
	       targs = targs,
	       arg = Option.map(arg, bindVarType)}
      
      fun new{var, ty, lambda}: unit =
	 if shouldDuplicate var
	    then setVarInfo(var, Dup{duplicates = ref []})
	 else (bind var; ())
	 
      fun loopExp(e: Exp.t): Exp.t =
	 let val {decs, result} = Exp.dest e
	 in Exp.new(loopDecs(decs, result))
	 end

      and loopLambda(l: Lambda.t): Lambda.t =
	 let val {arg, argType, body} = Lambda.dest l
	 in Lambda.new{arg = bind arg,
		       argType = argType,
		       body = loopExp body}
	 end

      and loopDecs(ds: Dec.t list, result): {decs: Dec.t list,
					     result: VarExp.t} =
	 case ds of
	    [] => {decs = [], result = loopVar result}
	  | d :: ds =>
	       case d of
		  MonoVal{var, ty, exp} =>
		     (case exp of
			 Lambda l =>
			    let
			       val _ = new{var = var, ty = ty, lambda = l}
			       val {decs, result} = loopDecs(ds, result)
			       val decs =
				  case varInfo var of
				     Replace var =>
					MonoVal{var = var, ty = ty,
						exp = Lambda(loopLambda l)}
					:: decs
				   | Dup{duplicates, ...} =>
					List.fold
					(!duplicates, decs, fn (var, decs) =>
					 MonoVal{var = var, ty = ty,
						 exp = Lambda(loopLambda l)}
					 :: decs)
			    in {decs = decs, result = result}
			    end
		       | _ => 
			    let
			       val exp =
				  case exp of
				     Const _ => exp
				   | Var x => Var(loopVar x)
				   | Tuple xs => Tuple(loopVars xs)
				   | Select{tuple, offset} =>
					Select{tuple = loopVar tuple,
					       offset = offset}
				   | ConApp{con, targs, arg} =>
					ConApp{con = con,
					       targs = targs,
					       arg = Option.map(arg, loopVar)}
				   | PrimApp{prim, targs, args} =>
					PrimApp{prim = prim,
						targs = targs,
						args = loopVars args}
				   | App{func, arg} =>
					App{func = loopVar func,
					    arg = loopVar arg}
				   | Raise x => Raise(loopVar x)
				   | Case{test, cases, default} =>
					Case{test = loopVar test,
					     cases = (List.map
						      (cases, fn (p, e) =>
						       (bindPat p, loopExp e))),
					     default = Option.map(default, loopExp)}
				   | Handle{try, catch, handler} =>
					Handle{try = loopExp try,
					       catch = bindVarType catch,
					       handler = loopExp handler}
				   | _ => Error.bug "unexpected primExp"
			       val var = bind var
			       val {decs, result} = loopDecs(ds, result)
			    in {decs = (MonoVal{var = var, ty = ty, exp = exp}
					:: decs),
				result = result}
			    end)
		| Fun{decs, ...} =>
		     let
			val _ = List.foreach(decs, new)
			val {decs = ds, result} = loopDecs(ds, result)
			val infos = List.map(decs, varInfo o #var)
			val ac =
			   List.fold2
			   (decs, infos, [], fn ({var, ty, lambda}, info, ac) =>
			    case info of
			       Replace var =>
				  {var = var, ty = ty,
				   lambda = loopLambda lambda} :: ac
			     | Dup _ => ac)
			val dups =
			   List.fold2(decs, infos, [], fn (dec, info, ac) =>
				      case info of
					 Replace _ => ac
				       | Dup{duplicates, ...} =>
					    (dec, !duplicates) :: ac)
			val decs =
			   List.fold
			   (dups, ac,
			    fn (({var, ty, lambda}, duplicates), ac) =>
			    List.fold
			    (duplicates, ac, fn (var', ac) =>
			     let
				val vars =
				   List.map(dups, fn ({var = var'', ...}, _) =>
					    if Var.equals(var, var'')
					       then (setVarInfo(var, Replace var')
						     ; var')
					    else bind var'')
			     in List.fold2
				(dups, vars, ac,
				 fn (({ty, lambda, ...}, _), var, ac) =>
				 {var = var, ty = ty, lambda = loopLambda lambda}
				 :: ac)
			     end))
		     in {decs = Fun{tyvars = [], decs = decs} :: ds,
			 result = result}
		     end
		| _ => Error.bug "polyvariance saw bogus dec"

      val program =
	 Program.T{datatypes = datatypes,
		   body = loopExp body}
	 
   in Program.clear program; program
   end

val duplicate =
   fn p =>
   case !Control.polyvariance of
      NONE => p
    | SOME{rounds, small, product} =>
	 let
	    fun loop(p, n) =
	       if n = 0
		  then p
	       else let val p = simplify(duplicate(p, small, product))
		    in Control.message(fn () => Program.layoutStats p)
		       ; loop(p, n - 1)
		    end
	 in loop(p, rounds)
	 end
   
end
functor MonoList(X: T): T =
struct

type t = X.t list

val equals = fn (l, l') => List.equals(l, l', X.equals)

val layout = List.layout X.layout
			     
end
signature FINITE_FUNCTION = 
   sig
      type ('a, 'b) t
	 
      val size: ('a, 'b) t -> int
      val lookup: ('a, 'b) t * 'a -> 'b
      val foreach: ('a, 'b) t * ('a * 'b -> unit) -> unit
      val toFunction: ('a, 'b) t -> 'a -> 'b
   end

signature POLY_CACHE =
   sig
      include FINITE_FUNCTION
      
      val new: {equal: 'a * 'a -> bool} -> ('a, 'b) t
      val fromList: {equal: 'a * 'a -> bool,
		      elements: ('a * 'b) list} -> ('a, 'b) t
      val peek: ('a, 'b) t * 'a -> 'b option
      val addNew: ('a, 'b) t * 'a * 'b -> unit
      val getOrAdd: ('a, 'b) t * 'a * (unit -> 'b) -> 'b
      val eq: ('a, 'b) t * ('a, 'b) t -> bool
      val toList: ('a, 'b) t -> ('a * 'b) list
   end 
functor PolyCache(): POLY_CACHE =
struct

datatype ('a, 'b) t = T of {equal: ('a * 'a) -> bool,
			   elts: ('a * 'b) list ref}

fun fromList{equal, elements} = T{equal = equal, elts = ref elements}
   
fun new{equal} = T{equal = equal, elts = ref []}

fun toList(T{elts, ...}) = !elts

fun size c = List.length(toList c)
fun foreach(c, f) = List.foreach(toList c, f)
   
fun peek(T{equal, elts = ref l}, x) =
   case List.peek(l, fn (x', _) => equal(x, x')) of
      NONE => NONE
    | SOME(_, y) => SOME y

fun lookup cx = valOf(peek cx)

fun toFunction c a = lookup(c, a)
   
fun addNew(T{elts = r as ref l, ...}, x, y) = r := (x, y) :: l

fun getOrAdd(c, x, th) =
   case peek(c, x) of
      NONE => let val y = th()
	      in addNew(c, x, y) ; y
	      end
    | SOME y => y

fun eq(T{elts=r, ...}, T{elts=r', ...}) = r = r'
   
end 

structure PolyCache = PolyCache()
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SXML_EXNS = XML

(*
 invariants:
 * no PolyVal decs
 * hasExns = true
 * tyvar lists in datatype, val, and fun decs are always empty
 * type lists are empty for variable args and for non-built-in-constructor args
 * no tyvars in types
 *)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MONOMORPHISE_STRUCTS = 
   sig
      structure Xml: XML
      structure Sxml: SXML_EXNS
      sharing Xml.Atoms = Sxml.Atoms
   end

signature MONOMORPHISE = 
   sig
      include MONOMORPHISE_STRUCTS
      
      val monomorphise: Xml.Program.t -> Sxml.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Monomorphise(S: MONOMORPHISE_STRUCTS): MONOMORPHISE =
struct

open S

open Xml.Atoms
local open Xml
in structure Xpat = Pat
   structure Xdec = Dec
   structure Xexp = Exp
   structure Xlambda = Lambda
   structure XprimExp = PrimExp
   structure Xprogram = Program
   structure Xtype = Type
   structure XvarExp = VarExp
end
local open Sxml
in structure Spat = Pat
   structure Sdec = Dec
   structure Sexp = Exp
   structure Slambda = Lambda
   structure SprimExp = PrimExp
   structure Sprogram = Program
   structure Stype = Type
   structure SvarExp = VarExp
end

structure Cache :
   sig
      type 'a t
      val new: unit -> 'a t
      val getOrAdd: 'a t * Stype.t list * (unit -> 'a) -> 'a
      val toList: 'a t -> (Stype.t list * 'a) list
   end =
   struct
      structure Stype =
	 struct
	    open Stype
	    val layout =
	       fn t =>
	       let open Layout
	       in record[("hash", str(Word.toString(hash t))),
			 ("type", layout t)]
	       end
	 end

      structure Stypes = MonoList(Stype)

(*       structure C1 =
 * 	 struct
 * 	    structure P = PolyCache
 * 	       
 * 	    type 'a t = (Stype.t list, 'a) P.t
 * 	    fun new() = P.new{equal = Stypes.equals}
 * 	    val getOrAdd = P.getOrAdd
 * 
 * 	    val toList = P.toList
 * 	 end
 *)

      structure C2 =
      (* use a splay tree based on lexicographic ordering of lists of hash values of
       * types.  Use an alist (i.e. polycache) within each bucket of the splay tree *)
	 struct
	    structure Stype =
	       struct
		  open Stype
		  val op < = fn (t, t') => hash t < hash t'
		  fun equals(t, t') = hash t = hash t'
		  val {compare, ...} =
		     Relation.lessEqual{< = op <, equals = equals}
	       end

	    structure Cache = PolyCache

	    structure S =
	       SplayMapFn
	       (type ord_key = Stype.t list
		val compare = fn (ts, ts') => List.compare(ts, ts',
							   Stype.compare))

	    type 'a t = (Stypes.t, 'a) Cache.t S.map ref

	    fun new() = ref S.empty

	    fun getOrAdd(m, k, th) =
	       case S.find(!m, k) of
		  NONE => let
			     val x = th()
			     val cache = Cache.fromList{equal = Stypes.equals,
							elements = [(k, x)]}
			  in m := S.insert(!m, k, cache); x
			  end
		| SOME cache => Cache.getOrAdd(cache, k, th)
		     
	    fun toList c =
	       List.fold(S.listItems(! c), [], fn (cache, items) =>
			 Cache.toList cache @ items)
	 end
      open C2

(* This commented out stuff compares C1 and C2 to make sure they are behaving
 * the same way.  Useful to find bugs in C2.
 *)
(* 
 *       type 'a t = 'a C1.t * 'a C2.t
 * 
 *       fun new() = (C1.new(), C2.new())
 * 
 *       fun layout(c1, c2) =
 * 	 let
 * 	    fun layout c =
 * 	       Layout.list(List.map(c, fn (ts, _) => Stypes.layout ts))
 * 	 in Layout.align[layout(C1.toList c1), layout(C2.toList c2)]
 * 	 end
 * 
 *       fun getOrAdd(c as (c1, c2), ts, th) =
 * 	 let val th = Promise.lazy th
 * 	    val b1 = ref false
 * 	    val b2 = ref false
 * 	    fun check(g, c, b) = g(c, ts, fn () => (b := true; th()))
 * 	    val x = check(C1.getOrAdd, c1, b1)
 * 	 in check(C2.getOrAdd, c2, b2);
 * 	    if !b1 = !b2
 * 	       then x
 * 	    else (Control.message
 * 		  let open Layout
 * 		  in seq[str "getOrAdd",
 * 			 tuple[layout c, Stypes.layout ts]]
 * 		  end;
 * 		  Error.bug "getOrAdd")
 * 	 end
 *       
 *       fun toList(c as (c1, c2)) =
 * 	 let val l1 = C1.toList c1
 * 	    val l2 = C2.toList c2
 * 	    fun sub(l1, l2) =
 * 	       List.forall(l1, fn (ts1, _) =>
 * 			   List.exists(l2, fn (ts2, _) =>
 * 				       Stypes.equals(ts1, ts2)))
 * 	 in if (List.length l1 = List.length l2
 * 		andalso sub(l1, l2) andalso sub(l2, l1))
 * 	       then l1
 * 	    else (Control.message
 * 		  let open Layout
 * 		  in seq[str "toList ", layout c]
 * 		  end;
 * 		  Error.bug "toList")
 * 	 end
 *)
   end
      
fun monomorphise(Xprogram.T{datatypes, body}): Sprogram.t =
   let
      val {set = setVar,
	   get = getVar: Var.t -> Stype.t list -> SvarExp.t} =
	 Property.new(Var.plist, Property.initRaise("var", Var.layout))

(*       val setVar =
 * 	 Trace.trace2("setVar", Var.layout, Layout.ignore, Unit.layout) setVar
 *       val getVar =
 * 	 Trace.trace("getVar", Var.layout, Layout.ignore) getVar
 *)
	 
      val {set = setCon, get = peekCon: Con.t -> (Stype.t list -> Con.t) option} =
	 Property.new(Con.plist, Property.initConst NONE)
	 
      val {set = setTycon,
	   get = peekTycon: Tycon.t -> (Stype.t list -> Tycon.t) option} =
	 Property.new(Tycon.plist, Property.initConst NONE)
	 
      val {set = setTyvar, get = getTyvar: Tyvar.t -> Stype.t} =
	 Property.new(Tyvar.plist, Property.initRaise("tyvar", Tyvar.layout))

      (* val getTyvar = Trace.trace("getTyvar", Tyvar.layout, Stype.layout) getTyvar *)

      (* val setTyvars = Trace.trace2("setTyvar", Tyvar.layout, Stype.layout, Unit.layout) setTyvar *)

      fun setTyvars(tyvs, tys) = List.foreach2(tyvs, tys, setTyvar)
	 
      fun monoType(t: Xtype.t): Stype.t =
	 Xtype.hom{ty = t,
		   var = getTyvar,
		   con = fn (c, ts) => Stype.con(case peekTycon c of
						    NONE => (c, ts)
						  | SOME f => (f ts, []))}

(*       val monoType = Trace.trace("monoType", Xtype.layout, Stype.layout) monoType
 *)
      fun monoTypeOpt(to: Xtype.t option): Stype.t option =
	 case to of
	    NONE => NONE
	  | SOME t => SOME(monoType t)

      fun monoTypes ts = List.map(ts, monoType)

      fun monoCon(c: Con.t, ts: Xtype.t list): Con.t * Stype.t list =
	 let val ts = monoTypes ts
	 in case peekCon c of
	    NONE => (c, ts)
	  | SOME f => (f ts, [])
	 end

      (* It is necessary to create new variables for monomorphic variables
       * because they still may have type variables in their type.
       *)
      fun renameMono(x, t) =
	 let
	    val x' = Var.new x
(* 	    val _ = Control.message let open Layout
 * 				    in seq[str "renameMono ",
 * 					   Var.layout x,
 * 					   str " = ", Var.layout x']
 * 				    end
 *)
	    val ve = SvarExp.mono x'
	    val inst =
	       fn [] => ve
		| _ => Error.error "monomorphise: expected monomorphic instance"
	 in setVar(x, inst) ; (x', monoType t)
	 end

(*       val renameMono =
 * 	 Trace.trace2("renameMono", Var.layout, Xtype.layout,
 * 		      Layout.tuple2(Var.layout, Stype.layout)) renameMono
 *)

      fun monoPat(Xpat.T{con, targs, arg}): Spat.t =
	 let val (con, targs) = monoCon(con, targs)
	 in Spat.T{con = con, targs = targs,
		   arg = (case arg of
			     NONE => NONE
			   | SOME x => SOME(renameMono x))}
	 end

(*       val monoPat = Trace.trace("monoPat", Xpat.layout, Spat.layout) monoPat
 *)
      fun finishDecs decs = decs []

(*       val traceMonoExp =
 * 	 Trace.trace("monoExp", Xexp.layout, Sexp.layout)
 * 
 *       val traceMonoDec =
 * 	 Trace.trace("monoDec", Xdec.layout,
 * 		     fn (_: unit -> Sdec.t list) => Layout.empty)
 *)

      (*------------------------------------*)
      (*             datatypes              *)
      (*------------------------------------*)

      val newDbs: {tyvars: Tyvar.t list,
		    types: Stype.t list,
		    tycon: Tycon.t,
		    cons: {con: Con.t,
			    typ: Xtype.t option,
			    used: bool} ref list} list ref = ref []
      val _ =
	 List.foreach
	 (datatypes, fn {tyvars, tycon, cons} =>
	  let val cache = Cache.new()
	     fun instantiate ts =
		Cache.getOrAdd
		(cache, ts, fn () =>
		 let
		    val (tycon, cons) =
		       if Tycon.equals(tycon, Tycon.bool)
			  then (tycon,
				List.map(cons, fn {con, ...} =>
					 ref{con = con, typ = NONE, used = true}))
		       else 
			  (Tycon.new tycon,
			   List.map(cons, fn {con, arg} =>
				    ref{con = con, typ = arg, used = false}))
		    val db =
		       {tyvars = tyvars, types = ts, tycon = tycon, cons = cons}
		 in ListRef.push(newDbs, db);
		    db
		 end)
	  in setTycon(tycon, SOME(fn ts => #tycon(instantiate ts))) ;
	     List.foreachi
	     (cons, fn (n, {con, ...}) =>
	      setCon(con,
		     SOME(fn ts =>
			  let
			     val r as ref{con, typ, used} =
				List.nth(#cons(instantiate ts), n)
			  in if used then con
			     else let val con = Con.new con
				  in r := {con = con, typ = typ, used = true};
				     con
				  end
			  end)))
	  end)

      val _ = (monoCon(Con.truee, []) ;
	       monoCon(Con.falsee, []))

      fun finishDbs accum =
	 let val dbs = !newDbs
	    val _ = newDbs := []
	 in case dbs of
	    [] => accum
	  | _ => finishDbs
	       (List.fold
		(dbs, accum,
		 fn ({tyvars, types, tycon, cons}, accum) =>
		 let
(* 		    val _ = Control.message let open Layout
 * 					    in seq[str "finishDbs ",
 * 						   Tycon.layout tycon]
 * 					    end
 *)
		    val cons =
		       List.fold(cons, [], fn (ref{con, typ, used}, cons) =>
				 if used
				    then (setTyvars(tyvars, types) ;
					  {con = con,
					   arg = monoTypeOpt typ} :: cons)
				 else cons)
		    val cons =
		       case cons of
			  [] => [{con = Con.newNoname(), arg = NONE}]
			| _ => cons
		 in {tycon = tycon, tyvars = [], cons = cons} :: accum
		 end))
	 end

      (*------------------------------------*)
      (*              monoExp               *)
      (*------------------------------------*)
      fun monoVarExp(XvarExp.T{var, targs}) =
	 getVar var (monoTypes targs)

(*       val monoVarExp =
 * 	 Trace.trace("monoVarExp", XvarExp.layout, SvarExp.layout) monoVarExp
 *)
      
      fun monoVarExps xs = List.map(xs, monoVarExp)
      
      fun monoExp(arg: Xexp.t): Sexp.t =
(*	 traceMonoExp *)
	 (fn (e: Xexp.t) =>
	  let val {decs, result} = Xexp.dest e
	     val thunks = List.fold(decs, [], fn (d, thunks) =>
				    monoDec d :: thunks)
	     val result = monoVarExp result
	     val decs = List.fold(thunks, [], fn (thunk, decs) => thunk() @ decs)
	  in Sexp.new{decs = decs,
		      result = result}
	  end) arg

      and monoPrimExp(e: XprimExp.t): SprimExp.t =
	 case e of
	    XprimExp.Const c => SprimExp.Const c
	  | XprimExp.Var x => SprimExp.Var(monoVarExp x)
	  | XprimExp.Tuple xs => SprimExp.Tuple(monoVarExps xs)
	  | XprimExp.Select{tuple, offset} =>
	       SprimExp.Select{tuple = monoVarExp tuple, offset = offset}
	  | XprimExp.Lambda l => SprimExp.Lambda(monoLambda l)
	  | XprimExp.ConApp{con, targs, arg} =>
	       let val (con, targs) = monoCon(con, targs)
	       in SprimExp.ConApp{con = con, targs = targs,
				  arg = Option.map(arg, monoVarExp)}
	       end
	  | XprimExp.PrimApp{prim, targs, args} =>
	       SprimExp.PrimApp{prim = prim,
				targs = monoTypes targs,
				args = monoVarExps args}
	  | XprimExp.App{func, arg} =>
	       SprimExp.App{func = monoVarExp func, arg = monoVarExp arg}
	  | XprimExp.Raise x => SprimExp.Raise(monoVarExp x)
	  | XprimExp.Handle{try, catch, handler} =>
	       SprimExp.Handle{try = monoExp try,
			       catch = renameMono catch,
			       handler = monoExp handler}
	  | XprimExp.Case{test, cases, default} =>
	       SprimExp.Case
	       {test = monoVarExp test,
		cases = List.revMap(cases, fn (pat, exp) =>
				    (monoPat pat, monoExp exp)),
		default = Option.map(default, monoExp)}

      and monoLambda l: Slambda.t =
	 let val {arg, argType, body} = Xlambda.dest l
	    val (arg, argType) = renameMono(arg, argType)
	 in Slambda.new{arg = arg,
			argType = argType,
			body = monoExp body}
	 end
	 
      (*------------------------------------*)
      (*              monoDec               *)
      (*------------------------------------*)

      and monoDec arg: unit -> Sdec.t list =
(*	 traceMonoDec *)
	 (fn Xdec.MonoVal{var, ty, exp} =>
	        let val (var, _) = renameMono(var, ty)
		in fn () => [Sdec.MonoVal{var = var,
					  ty = monoType ty,
					  exp = monoPrimExp exp}]
		end
          | Xdec.PolyVal{var, tyvars, ty, exp} =>
		let val cache = Cache.new()
		   fun instantiate ts =
		      (setTyvars(tyvars, ts) ;
		       Cache.getOrAdd(cache, ts, fn () =>
				      SvarExp.mono(Var.new var)))
		in setVar(var, instantiate) ;
		   fn () =>
		   List.fold
		   (Cache.toList cache, [], fn ((ts, ve), decs) =>
		    (setTyvars(tyvars, ts);
		     let val {decs = decs', result} = Sexp.dest(monoExp exp)
		     in decs' @ (Sdec.MonoVal{var = SvarExp.var ve,
					      ty = monoType ty,
					      exp = SprimExp.Var result} :: decs)
		     end))
		end
	   | Xdec.Fun{tyvars, decs} =>
		let
		   val cache = Cache.new()
		   val vars = List.map(decs, #var)
		   val _ =
		      List.foreachi
		      (vars, fn (n, x) =>
		       setVar
		       (x, fn ts =>
			(setTyvars(tyvars, ts);
			 List.nth(Cache.getOrAdd
				  (cache, ts, fn () =>
				   List.map(decs, SvarExp.mono o Var.new o #var)),
				  n))))
		in fn () =>
		   List.map
		   (Cache.toList cache, fn (ts, xs) =>
		    (setTyvars(tyvars, ts) ;
		     List.foreach2(decs, xs, fn({var, lambda, ...}, var') =>
				   setVar(var, fn _ => var'));
		     Sdec.Fun
		     {tyvars = [],
		      decs =
		      List.fold2
		      (decs, xs, [], fn ({ty, lambda, ...}, ve, decs) =>
		       {var = SvarExp.var ve,
			ty = monoType ty,
			lambda = monoLambda lambda} :: decs)}))
		end
	   | Xdec.Exception{con, arg} =>
		let
		   (* don't rename no-args so primitive excons aren't renamed *)
		   val con' = (case arg of
				  NONE => con 
				| SOME _ => Con.new con)
		in setCon(con, SOME(fn _ => con')) ;
		   fn () => [Sdec.Exception{con = con', arg = monoTypeOpt arg}]
		end) arg

      (*------------------------------------*)
      (*     main code for monomorphise     *)
      (*------------------------------------*)

      val body = monoExp body

      val datatypes = finishDbs []

      val program =
	 Sprogram.T{datatypes = datatypes,
		    body = body} 
   in Sprogram.clear program;
      
      program
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature IMPLEMENT_EXCEPTIONS_STRUCTS = 
   sig
      include SXML
   end

signature IMPLEMENT_EXCEPTIONS = 
   sig
      include IMPLEMENT_EXCEPTIONS_STRUCTS
      
      val doit: Program.t -> Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ImplementExceptions(S: IMPLEMENT_EXCEPTIONS_STRUCTS): IMPLEMENT_EXCEPTIONS = 
struct

open S
open Dec PrimExp
structure Dexp = DirectExp

fun doit(Program.T{datatypes, body}): Program.t =
   let
      val exnName = Var.newString "exnName"
      val exnNameType = Type.arrow(Type.exn, Type.string)
      val {get = exconInfo: Con.t -> {refVar: Var.t,
				      make: VarExp.t option -> Dexp.t} option,
	   set = setExconInfo} =
	 Property.new(Con.plist, Property.initConst NONE)

(*       val setExconInfo = Trace.trace2("setExconInfo", Con.layout,
 * 				      Layout.ignore, Unit.layout) setExconInfo
 * 
 *       val exconInfo =
 * 	 Trace.trace("exconInfo", Con.layout, Layout.ignore) exconInfo
 *)

      fun isExcon c =
	 case exconInfo c of
	    NONE => false
	  | SOME _ => true

      val exnValCons: {con: Con.t, arg: Type.t} list ref = ref []

      fun loopOpt e =
	 (case e of
	     NONE => NONE
	   | SOME e => SOME(loop e))

      and loops es = List.map(es, loop)

      and loop(e: Exp.t): Exp.t =
	 let val {decs, result} = Exp.dest e
	    val decs =
	       List.flatten
	       (List.reverse(List.fold(decs, [], fn (d, ds) => loopDec d :: ds)))
	 in Exp.new{decs = decs,
		    result = result}
	 end

      and loopDec(dec: Dec.t): Dec.t list =
	 case dec of
	    MonoVal b => loopMonoVal b
	  | Fun{decs, ...} =>
	       [Fun{tyvars = [],
		    decs = List.revMap(decs, fn {var, ty, lambda} =>
				       {var = var,
					ty = ty,
					lambda = loopLambda lambda})}]
	  | Exception{con, arg} =>
	       let open Dexp
		  val r = Var.newString "exnRef"
		  val uniq = monoVar(r, Type.unitRef)
		  val conApp =
		     fn arg =>
		     conApp{con = con, targs = [], ty = Type.exn, arg = SOME arg}
		  val (arg, decs, make) =
		     case arg of
			NONE =>
			   (* If the exception is not value carrying, then go
			    * ahead and make it now.
			    *)
			   let val exn = Var.newNoname()
			   in (Type.unitRef,
			       Dexp.vall{var = exn, exp = conApp uniq},
			       fn NONE => monoVar(exn, Type.exn)
				| _ => Error.bug "nullary excon applied to arg")
			   end
		      | SOME t =>
			   let val tupleType = Type.tuple[Type.unitRef, t]
			   in (tupleType,
			       [],
			       fn SOME x =>
			       conApp(tuple{exps = [uniq, varExp(x, t)],
					    ty = tupleType})
				| _ => Error.bug "unary excon not applied to arg")
			   end
	       in setExconInfo(con, SOME{refVar = r, make = make});
		  ListRef.push(exnValCons, {con = con, arg = arg});
		  vall{var = r, exp = reff(unit())} @ decs
	       end
	  | _ => Error.bug "implement exceptions saw unexpected dec"

      and loopMonoVal{var, ty, exp} : Dec.t list =
	 let fun primExp e = [MonoVal{var = var, ty = ty, exp = e}]
	    fun keep() = primExp exp
	    fun makeExp e = Dexp.vall{var = var, exp = e}
	 in case exp of
	    Lambda l => primExp(Lambda(loopLambda l))
	  | PrimApp{prim, targs, args} =>
	       primExp
	       (case (Prim.name prim, args) of
		   (Prim.Name.ExnName, [arg]) =>
		      App{func = VarExp.mono exnName, arg = arg}
		 | _ => PrimApp{prim = prim, targs = targs, args = args})
	  | ConApp{con, arg, ...} =>
	       (case exconInfo con of
		   NONE => keep()
		 | SOME{make, ...} => makeExp(make arg))
	  | Handle{try, catch = (catch, ty), handler} =>
	       primExp(Handle{try = loop try,
			      catch = (catch, ty),
			      handler = loop handler})
	  | Case{test, cases, default} =>
	       let
		  fun normal() =
		     primExp(Case
			     {test = test,
			      cases = List.revMap(cases, fn (pat, exp) =>
						  (pat, loop exp)),
			      default = loopOpt default})
	       in case cases of
		  [] => normal()
		| (Pat.T{con, ...}, _) :: _ =>
		     if not(isExcon con)
			then normal()
		     else (* convert to an exception match *)
			let
			   open Dexp
			   val defaultVar = Var.newString "default"
			   fun callDefault() =
			      app{func = monoVar(defaultVar,
						 Type.arrow(Type.unit, ty)),
				  arg = unit(),
				  ty = ty}
			   val unit = Var.newString "unit"
			in makeExp
			   (lett
			    {decs = vall{var = defaultVar,
					 exp =
					 lambda
					 {arg = unit,
					  argType = Type.unit,
					  bodyType = ty,
					  body =
					  case default of
					     SOME e => fromExp(loop e, ty)
					   | NONE =>
						Error.bug "no default for exception case"}},
			     
			     body = 
			     casee
			     {test = varExp(test, Type.exn),
			      ty = ty,
			      default = SOME(callDefault()),
			      cases =
			      List.revMap
			      (cases, fn (Pat.T{con, arg, ...}, e) =>
			       let
				  val refVar = Var.newNoname()
				  val body =
				     iff{test =
					 equal(monoVar(refVar, Type.unitRef),
					       monoVar
					       (#refVar(valOf(exconInfo con)),
						Type.unitRef)),
					 ty = ty,
					 thenn = fromExp(loop e, ty),
					 elsee = callDefault()}
				  fun make(arg, body) = 
				     (Pat.T{con = con, targs = [],
					    arg = SOME arg},
				      body)
			       in case arg of
				  NONE => make((refVar, Type.unitRef), body)
				| SOME(x, t) =>
				     let val tuple = (Var.newNoname(),
						      Type.tuple[Type.unitRef, t])
				     in make(tuple,
					     detupleBind
					     {tuple = monoVar tuple,
					      components = [refVar, x],
					      body = body})
				     end
			       end)}})
			end
	       end
	  | _ => keep()
	 end

      and loopLambda l =
	 let val {arg, argType, body} = Lambda.dest l
	 in Lambda.new{arg = arg,
		       argType = argType,
		       body = loop body}
	 end

      val body =
	 (* Wrap the body in a default handler that prints the exception name
	  * and exits with status 1.
	  *)
	 let val exn = (Var.newNoname(), Type.exn)
	 in Exp.fromPrimExp
	    (Handle
	     {try = loop body,
	      catch = exn,
	      handler =
	      let
		 open Dexp
		 fun string s = const(Const.fromString s)
		 fun print(e: Dexp.t): Dexp.t = 
		    primApp{prim = Prim.print,
			    targs = [],
			    args = [e],
			    ty = Type.unit}
	      in toExp
		 (sequence
		  [print(string "unhandled exception "),
		   print(app{func = monoVar(exnName, exnNameType),
			     arg = monoVar exn,
			     ty = Type.string}),
		   print(string "\n"),
		   primApp{prim = Prim.exit,
			   targs = [],
			   ty = Type.unit,
			   args = [const(Const.fromInt 1)]}])
	      end},
	     Type.unit)
	 end

      val (datatypes, body) =
	 case !exnValCons of
	    [] => (datatypes, body)
	  | cons =>
	       (* Add a datatype declaration for exceptions
		* and implement exnName.
		*)
	       ({tycon = Tycon.exn,
		 tyvars = [],
		 cons = List.revMap(cons, fn {con, arg} =>
				    {con = con, arg = SOME arg})}
		:: datatypes,
		Exp.prefix
		(body, 
		 MonoVal
		 {var = exnName,
		  ty = exnNameType,
		  exp =
		  Lambda
		  (Lambda.new
		   let val exn = Var.newNoname()
		   in {arg = exn,
		       argType = Type.exn,
		       body =
		       let open Dexp
		       in toExp
			  (casee
			   {test = monoVar(exn, Type.exn),
			    cases =
			    List.revMap
			    (cons, fn {con, arg} =>
			     (Pat.T{con = con,
				    targs = [],
				    arg = SOME(Var.newNoname(), arg)},
			      const
			      (Const.fromString(Con.originalName con)))),
			    default = NONE,
			    ty = Type.string})
		       end}
		   end)}))

      val program =
	 Program.T{datatypes = datatypes,
		   body = body}
	 
   in Program.clear program;
      program
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LAMBDA_FREE_STRUCTS = 
   sig
      include SXML
   end

signature LAMBDA_FREE = 
   sig
      include LAMBDA_FREE_STRUCTS
      
      (*
       * When called, descends the entire program and attaches a property
       * to each lambda primExp in the program.  Then, you can use
       * lambdaFree to get free variables of that lambda.
       * For lambdas bound in a Fun dec, lambdaFree gives the union of the
       * frees of the entire group of mutually recursive functions.  Hence,
       * lambdaFree for every lambda in a single Fun dec is the same.
       * Furthermore, for a lambda bound in a Fun dec, lambdaRec gives
       * the list of other funs bound in the same dec that the lambda refers
       * to.  For example:
       * 
       * val rec f = fn x => ... y ... g ... f ...
       * and g = fn z => ... f ... w ...
       * 
       * lambdaFree(fn x =>) = [y, w]
       * lambdaFree(fn z =>) = [y, w]
       * lambdaRec(fn x =>) = [g, f]
       * lambdaRec(fn z =>) = [f]
       *)
      val lambdaFree:
	 Program.t -> {
		       lambdaFree: Lambda.t -> Var.t list,
		       lambdaRec: Lambda.t -> Var.t list
		      }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor LambdaFree(S: LAMBDA_FREE_STRUCTS): LAMBDA_FREE = 
struct

open S

open Dec PrimExp

structure Status =
   struct
      datatype t = Unseen | Free | Bound

      val toString =
	 fn Unseen => "Unseen"
	  | Free => "Free"
	  | Bound => "Bound"

      val layout = Layout.str o toString
   end
datatype status = datatype Status.t

fun lambdaFree(Program.T{body, ...}) =
   let
      val {get = getFree, set = setFree} =
	 Property.new(Lambda.plist, Property.initRaise("free", Lambda.layout))

      val {get = getRec, set = setRec} =
	 Property.new(Lambda.plist, Property.initConst [])

      type scope = {frees: Var.t list ref,
		    get: Var.t -> Status.t,
		    set: Var.t * Status.t -> unit}

      fun bind(x: Var.t, {set, ...}: scope) = set(x, Bound)

      fun var(x: Var.t, {get, set, frees}: scope) =
	 case get x of
	    Unseen => (set(x, Free); ListRef.push(frees, x))
	  | _ => ()

      fun vars(xs, s) = List.foreach(xs, fn x => var(x, s))
	 
      fun varExp(x: VarExp.t, s: scope) = var(VarExp.var x, s)
	 
      fun varExpOpt(x, s) =
	 case x of
	    NONE => ()
	  | SOME x => varExp(x, s)

      fun varExps(xs, s) = List.foreach(xs, fn x => varExp(x, s))

      val {get = varInfo: Var.t -> {frees: Var.t list ref ref,
				    status: Status.t ref}, ...} =
	 Property.new(Var.plist,
		      Property.initFun(let val bogusFrees = ref []
				       in fn _ => {frees = ref bogusFrees,
						   status = ref Unseen}
				       end))

      fun newScope th = 
	 let
	    val frees = ref []
	    val all = ref []
	    fun statusRef x =
	       let val {frees = frees', status} = varInfo x
	       in if frees = !frees'
		     then ()
		  else (ListRef.push(all, (frees', !frees', status, !status))
			; frees' := frees; status := Unseen)
		  ; status
	       end
		    
	    fun get x = !(statusRef x)

	    fun set(x, s) = statusRef x := s

	 in th{frees = frees, get = get, set = set}
	    ; List.foreach(!all, fn (r, v, r', v') => (r := v; r' := v'))
	    ; !frees
	 end
      
      fun exp(e, s) =
	 let val {decs, result} = Exp.dest e
	 in List.foreach
	    (decs,
	     fn Exception _ => ()
	      | MonoVal{var, exp, ...} => (primExp(exp, s); bind(var, s))
	      | PolyVal{var, exp = e, ...} => (exp(e, s); bind(var, s))
	      | Fun{decs, ...} =>
		   let
		      val {get = isBound, set, destroy} =
			 Property.newDest(Var.plist, Property.initConst false)
		      val _ =
			 List.foreach(decs, fn {var, ...} => set(var, true))
		      val xs =
			 newScope
			 (fn s =>
			  List.foreach
			  (decs, fn {lambda = l, ...} =>
			   setRec(l, List.fold(lambda l, [], fn (x, xs) =>
					       if isBound x
						  then x :: xs
					       else (var(x, s); xs)))))
		   in List.foreach(decs, fn {var, lambda, ...} =>
				   (setFree(lambda, xs)
				    ; bind(var, s)))
		      ; vars(xs, s)
		   end)
	    ; varExp(result, s)
	 end

      and primExp(e, s) = 
	 case e of
	    Const _ => ()
	  | Var x => varExp(x, s)
	  | Tuple xs => varExps(xs, s)
	  | Select{tuple, ...} => varExp(tuple, s)
	  | Lambda l =>
	       let val xs = lambda l
	       in setFree(l, xs); vars(xs, s)
	       end
	  | ConApp{arg, ...} => varExpOpt(arg, s)
	  | PrimApp{args, ...} => varExps(args, s)
	  | App{func, arg} => (varExp(func, s); varExp(arg, s))
	  | Raise x => varExp(x, s)
	  | Handle{try, catch, handler} =>
	       (exp(try, s); bind(#1 catch, s); exp(handler, s))
	  | Case{test, cases, default} =>
	       (varExp(test, s);
		expOpt(default, s);
		List.foreach(cases, fn (Pat.T{arg, ...}, e) =>
			     (case arg of
				 NONE => ()
			       | SOME(x, _) => bind(x, s);
				    exp(e, s))))
      and expOpt(e, s) =
	 case e of
	    NONE => ()
	  | SOME e => exp(e, s)

      and lambda l =
	 let val {arg, body, ...} = Lambda.dest l
	 in newScope(fn s => (bind(arg, s); exp(body, s)))
	 end

      val frees = newScope(fn s => exp(body, s))
	 
   in
      case frees of
	 [] => ()
       | _ => Error.bug "program has free variables";
	    
      {
       lambdaFree = getFree,
       lambdaRec = getRec
      }
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature GLOBALIZE_STRUCTS = 
   sig
      include SXML
   end

signature GLOBALIZE = 
   sig
      include GLOBALIZE_STRUCTS
      
      val globalize: {
		      program: Program.t,
		      lambdaFree: Lambda.t -> Var.t list
		     } -> {
			   isGlobal: Var.t -> bool
			  }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Globalize(S: GLOBALIZE_STRUCTS): GLOBALIZE = 
struct

open S
open Dec PrimExp

fun globalize{program = Program.T{datatypes, body}, lambdaFree} =
   let
      local
	 val {get: Tycon.t -> bool, set} =
	    Property.new(Tycon.plist, Property.initConst false)
	 fun makeBig tycon = set(tycon, true)
	 val _ = (List.foreach(datatypes, makeBig o #tycon)
		  ; makeBig Tycon.array
		  ; makeBig Tycon.vector)
      in val tyconIsBig = get
      end

      fun typeIsSmall t =
	 let open Type
	 in case dest t of
	    Con(c, ts) =>
	       not(tyconIsBig c)
	       andalso if (Tycon.equals(c, Tycon.tuple)
			   orelse Tycon.equals(c, Tycon.reff))
			  then List.forall(ts, typeIsSmall)
		       else true
	  | _ => Error.bug "typeIsSmall saw type variable"
	 end
   
      val {get = varIsGlobal, set = setVarGlobal} =
	 Property.new(Var.plist, Property.initConst false)

      val isGlobal = varIsGlobal o VarExp.var
	 
      fun areGlobal xs = List.forall(xs, isGlobal)

      fun loopExp(e: Exp.t, once: bool): unit =
	 List.foreach
	 (Exp.decs e,
	  fn MonoVal{var, ty, exp} =>
	     let
		val global =
		   case exp of
		      Const _ => true
		    | Var x => isGlobal x
		    | Tuple xs => areGlobal xs
		    | Select{tuple, ...} => isGlobal tuple
		    | Lambda l => (loopLambda l;
				   List.forall(lambdaFree l, varIsGlobal))
		    | ConApp{arg, ...} =>
			 (case arg of
			     NONE => true
			   | SOME x => isGlobal x)
 		    | PrimApp{prim, args, ...} =>
 			 areGlobal args
 			 andalso (once
				  andalso (case Prim.name prim of
					      Prim.Name.Ref => typeIsSmall ty
					    | _ => false))
		    | Case{cases, default, ...} =>
			 (List.foreach(cases, fn (_, e) =>
				       loopExp(e, once));
			  case default of
			     NONE => ()
			   | SOME e => loopExp(e, once);
			  false)
		    | Handle{try, handler, ...} =>
			 (loopExp(try, once); loopExp(handler, once);
			  false)
		    | _ => false
	     in if global then setVarGlobal(var, true) else ()
	     end
	   | Fun{decs, ...} =>
		(case decs of
		    {lambda, ...} :: _ =>
		       (if List.forall(lambdaFree lambda, varIsGlobal)
			   then List.foreach(decs, fn {var, ...} => setVarGlobal(var, true))
			else ())
		  | _ => ();
		 List.foreach(decs, loopLambda o #lambda))
	   | _ => Error.bug "globalize saw strange dec")

      and loopLambda(l: Lambda.t): unit =
	 loopExp(Lambda.body l, false)

      val body = loopExp(body, true)

   in
      {isGlobal = varIsGlobal}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ABSTRACT_VALUE_STRUCTS = 
   sig
      structure Sxml: SXML
   end

signature ABSTRACT_VALUE = 
   sig
      include ABSTRACT_VALUE_STRUCTS

      structure Lambda:
	 sig
	    type t
	       
	    val dest: t -> Sxml.Lambda.t

	    val layout: t -> Layout.t
	 end

      structure Lambdas :
	 sig
	    type t

	    val equals: t * t -> bool
	    val toList: t -> Lambda.t list
	    val plist: t -> PropertyList.t
	 end

      type t

      (* constructors *)
      val fromType: Sxml.Type.t -> t
      val lambda: Sxml.Lambda.t -> t
	 
      (* operators *)
      val tuple: t list -> t (* there must be one argument that is not Type _ *)
      val select: t * int -> t
      val dearray: t -> t
      val deref: t -> t
      val primApply: {prim: Sxml.Prim.t,
		      args: t list,
		      resultTy: Sxml.Type.t} -> t
	 
      datatype dest =
	 Type of Sxml.Type.t (* type doesn't contain any arrows *)
       | Ref of t
       | Array of t
       | Vector of t
       | Tuple of t list
       | Lambdas of Lambdas.t
      val dest: t -> dest
	 
      (* no possible values correspond to me *) 
      val isEmpty: t -> bool

      val addHandler: t * (Lambda.t -> unit) -> unit

      val unify: t * t -> unit
      val coerce: {from: t, to: t} -> unit

      val typeIsFirstOrder: Sxml.Type.t -> bool

      (* destroy info associated with Sxml.Type used to keep track of arrows *)
      val destroy: unit -> unit

      val plist: t -> PropertyList.t
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature FLOW_ANALYSIS_STRUCTS = 
   sig
      structure Sxml: SXML
   end

signature FLOW_ANALYSIS = 
   sig
      include FLOW_ANALYSIS_STRUCTS

      structure Value: ABSTRACT_VALUE
      
      val analyze:
	 Sxml.Program.t
	 -> {
	     value: Sxml.Var.t -> Value.t,
	     conArg: Sxml.Con.t -> Value.t option
	    }
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor AbstractValue(S: ABSTRACT_VALUE_STRUCTS): ABSTRACT_VALUE = 
struct

open S
open Sxml

structure Dset = DisjointSet

structure Lambda =
   struct
      datatype t = Lambda of {lambda: Sxml.Lambda.t,
			      hash: word}
	 
      fun newHash() = Word.fromInt(Int.random())

      fun new lambda = Lambda{lambda = lambda,
			      hash = newHash()}

      fun hash(Lambda{hash, ...}) = hash
	 
      fun dest(Lambda{lambda, ...}) = lambda

      fun equals(Lambda r, Lambda r') =
	 #hash r = #hash r'
	 andalso Sxml.Lambda.equals(#lambda r, #lambda r')

      fun layout(Lambda{lambda, ...}) =
	 let open Layout
	 in seq[str "lambda ", Sxml.Var.layout(Sxml.Lambda.arg lambda)]
	 end
   end

structure Lambdas = HashSet(structure Element = Lambda
			    val cacheSize = 5
			    val bits = 13)

structure LambdaNode
  : sig
	type t
	val new: unit -> t
	val lambda: Sxml.Lambda.t -> t
	val addHandler: t * (Lambda.t -> unit) -> unit
	val toSet: t -> Lambdas.t
	val unify: t * t -> unit
	val coerce: {from: t, to: t} -> unit
	val layout: t -> Layout.t
     end =
   struct
      datatype t = LambdaNode of {me: Lambda.t HashSetRep.t ref,
				  handlers: (Lambda.t -> unit) list ref,
				  coercedTo: t list ref} Dset.t

      fun toSet(LambdaNode d) = !(#me(Dset.value d))

      val layout = Lambdas.layout o toSet
(*      fun layout(LambdaNode d) =
	 let val {id, me, ...} = Dset.value d
	    open Layout
	 in record[("id", Id.layout id),
		   ("me", Lambdas.layout(!me))]
	 end*)

      fun newSet s = LambdaNode(Dset.singleton{me = ref s,
(*				      id = Id.new(), *)
				      handlers = ref [],
				      coercedTo = ref []})

      fun new() = newSet Lambdas.empty

      fun lambda l = newSet(Lambdas.singleton(Lambda.new l))

      fun handles(h: Lambda.t -> unit, s: Lambdas.t): unit =
	 Lambdas.foreach(s, fn l => h l)
	 
      fun handless(hs: (Lambda.t -> unit) list, s: Lambdas.t): unit =
	 List.foreach(hs, fn h => handles(h, s))

      fun addHandler(LambdaNode d, h: Lambda.t -> unit) =
	 let val {me, handlers, ...} = Dset.value d
	 in ListRef.push(handlers, h)
	    ; handles(h, !me)
	 end

      fun send(LambdaNode d, s): unit =
	 let val {me, coercedTo, handlers, ...} = Dset.value d
	    val diff = Lambdas.-(s, !me)
	 in if Lambdas.isEmpty diff
	       then ()
	    else (me := Lambdas.+(diff, !me)
		  ; List.foreach(!coercedTo, fn to => send(to, diff))
		  ; handless(!handlers, diff))
	 end

(*       val send =
 * 	 Trace.trace2("LambdaNode.send", layout, Lambdas.layout, Unit.layout) send
 *)

      fun equals(LambdaNode d, LambdaNode d') = Dset.equals(d, d')

      fun coerce(arg as {from = from as LambdaNode d, to: t}): unit =
	 if equals(from, to)
	    then ()
	 else let val {me, coercedTo, ...} = Dset.value d
	      in
		 if List.exists(!coercedTo, fn ls => equals(ls, to))
		    then ()
		 else (ListRef.push(coercedTo, to)
		       ; send(to, !me))
	      end

	   		      
      fun update(c, h, diff) =
	 if Lambdas.isEmpty diff
	    then ()
	 else (List.foreach(c, fn to => send(to, diff))
	       ; handless(h, diff))

      fun unify(s as LambdaNode d, s' as LambdaNode d'): unit =
	 if Dset.equals(d, d')
	    then ()
	 else
	    let
	       val {me = ref m, coercedTo = ref c, handlers = ref h, ...} =
		  Dset.value d
	       val {me = ref m', coercedTo = ref c', handlers = ref h', ...} =
		  Dset.value d'
	       val diff = Lambdas.-(m, m')
	       val diff' = Lambdas.-(m', m)
	    in Dset.union(d, d')
	       ; (Dset.setValue
		  (d, {me = ref(if Lambdas.isEmpty diff
				   then m'
				else Lambdas.+(m', diff)),
		       coercedTo = ref(List.append
				       (List.removeAll(c', fn n' =>
						       List.exists(c, fn n =>
								   equals(n, n'))),
					c)),
		       handlers = ref(List.append(h, h'))}))
	       ; update(c, h, diff')
	       ; update(c', h', diff)
	    end

(*       val unify =
 * 	 Trace.trace2("LambdaNode.unify", layout, layout, Unit.layout) unify
 *)
   end

datatype tree =
   Type of Type.t
 | Unify of UnaryTycon.t * t
 | Tuple of t list
 | Lambdas of LambdaNode.t
withtype t = {tree: tree,
	      plist: PropertyList.t} Dset.t

fun new(tree: tree) = Dset.singleton{tree = tree,
				     plist = PropertyList.new()}
   
val tree: t -> tree = #tree o Dset.value
val plist: t -> PropertyList.t = #plist o Dset.value
   
fun layout v =
   let open Layout
   in case tree v of
      Type t => seq[str "Type ", Type.layout t]
    | Unify(t, v) => paren(seq[UnaryTycon.layout t, str " ", layout v])
    | Tuple vs => tuple(List.map(vs, layout))
    | Lambdas l => LambdaNode.layout l
   end
   
fun isType v =
   case tree v of
      Type _ => true
    | _ => false

fun isEmpty v =
   case tree v of
      Lambdas n => Lambdas.isEmpty(LambdaNode.toSet n)
    | Tuple vs => List.exists(vs, isEmpty)
    | Unify(UnaryTycon.Ref, v) => isEmpty v
    | _ => false

(* used in closure converter *)
fun equals(v, v') =
   Dset.equals(v, v')
   orelse
   (case (tree v,      tree v') of
       (Type t,        Type t')   =>
	  if Type.equals(t, t')
	     then true
	  else Error.bug "Value.equals called on different types"
     | (Unify(t, v), Unify(t', v'))     =>
	  UnaryTycon.equals(t, t') andalso equals(v, v')
     | (Tuple vs,  Tuple vs')  => List.forall2(vs, vs', equals)
     | (Lambdas n, Lambdas n') => Lambdas.equals(LambdaNode.toSet n,
						 LambdaNode.toSet n')
     | _ => Error.bug "Value.equals called on different kinds of values")

fun addHandler(v, h) =
   case tree v of
      Lambdas n => LambdaNode.addHandler(n, h)
    | _ => Error.bug "can't addHandler to non lambda"

local
   val {hom, destroy} =
      Type.makeMonoHom
      {con = fn (t, tycon, vs) =>
       if Tycon.equals(tycon, Tycon.arrow)
	  then {isFirstOrder = false,
		make = fn () => new(Lambdas(LambdaNode.new()))}
       else
	  if List.forall(vs, #isFirstOrder)
	     then {isFirstOrder = true,
		   make = let val v = new(Type t)
			  in fn () => v
			  end}
	  else
	     {isFirstOrder = false,
	      make = let fun mutable mt = let val make = #make(hd vs)
					  in fn () => new(Unify(mt, make()))
					  end
		     in if Tycon.equals(tycon, Tycon.reff)
			   then mutable UnaryTycon.Ref
		        else if Tycon.equals(tycon, Tycon.array)
				then mutable UnaryTycon.Array
			else if Tycon.equals(tycon, Tycon.vector)
				then mutable UnaryTycon.Vector
			else if Tycon.equals(tycon, Tycon.tuple)
				then (fn () =>
				      new(Tuple(List.map(vs, fn {make, ...} =>
							 make()))))
			     else Error.bug "fromType saw non-arrow type"
		     end}}
in
   val destroy = destroy
   val typeIsFirstOrder = #isFirstOrder o hom
   fun fromType t = #make (hom t) ()
end

(* val fromType = Trace.trace("Value.fromType", Type.layout, layout) fromType *)

val tuple = new o Tuple

fun select(v, i) =
   case tree v of
      Type t => fromType(List.nth(Type.detuple t, i))
    | Tuple vs => List.nth(vs, i)
    | _ => Error.bug "Value.select expected tuple"

fun deref v =
   case tree v of
      Type t => fromType(Type.deref t)
    | Unify(_, v) => v
    | _ => Error.bug "Value.deref"

fun dearray v =
   case tree v of
      Type t => fromType(Type.dearray t)
    | Unify(_, v) => v
    | _ => Error.bug "Value.dearray"

val lambda = new o Lambdas o LambdaNode.lambda

(* val traceUnify = Trace.trace2("Value.unify", layout, layout, Unit.layout) *)

fun unify(v, v') =
   if Dset.equals(v, v')
      then ()
   else let val t = tree v
	    val t' = tree v'
	in Dset.union(v, v')
	   ; (case (t,             t') of
		 (Type t,        Type t')        => if Type.equals(t, t')
						       then ()
						    else Error.bug "unify"
	       | (Unify(_, v), Unify(_, v')) => unify(v, v')
	       | (Tuple vs,      Tuple vs')      => List.foreach2(vs, vs', unify)
	       | (Lambdas l,     Lambdas l')     => LambdaNode.unify(l, l')
	       | _                               => Error.bug "impossible unify")
	end

(*val unify = Trace.trace2("Value.unify", layout, layout, Unit.layout) unify *)

fun coerce{from: t, to: t}: unit =
   if Dset.equals(from, to)
      then ()
   else (case (tree from, tree to) of
	    (Type t,    Type t')    => if Type.equals(t, t')
					  then ()
				       else Error.bug "coerce"
	  | (Unify _, Unify _) =>
	       (* Can't do a coercion for vectors, since that would imply
		* walking over the entire vector and coercing each element
		*)
	       unify(from, to)
	  | (Tuple vs,  Tuple vs')  => List.foreach2(vs, vs', fn (v, v') =>
						     coerce{from = v, to = v'})
	  | (Lambdas l, Lambdas l') => LambdaNode.coerce{from = l, to = l'}
	  | _ => Error.bug "impossible coerce")

(* val coerce = Trace.trace("Value.coerce",
 * 			 fn {from, to} =>
 * 			 let open Layout
 * 			 in record[("from", layout from),
 * 				   ("to" , layout to)]
 * 			 end, Unit.layout) coerce
 *)

structure Dest =
   struct
      datatype dest =
	 Type of Type.t
       | Ref of t
       | Array of t
       | Vector of t
       | Tuple of t list
       | Lambdas of Lambdas.t
   end

fun dest v =
   case tree v of
      Type t => Dest.Type t
    | Unify(mt, v) => (case mt of
			  UnaryTycon.Ref => Dest.Ref v
			| UnaryTycon.Array => Dest.Array v
			| UnaryTycon.Vector => Dest.Vector v)
    | Tuple vs => Dest.Tuple vs
    | Lambdas l => Dest.Lambdas(LambdaNode.toSet l)

open Dest

(*---------------------------------------------------*)
(*                     primApply                     *)
(*---------------------------------------------------*)
structure Name = Prim.Name

fun primApply{prim: Prim.t, args: t list, resultTy: Type.t}: t =
   let 
      fun result() = fromType resultTy
      fun typeError() =
	 (Control.message
	  (fn () =>
	   let open Layout
	   in align[seq[str "prim: ", Prim.layout prim],
		    seq[str "args: ", tuple(List.map(args, layout))]]
	   end)
	  ; Error.bug "Value.primApply: type error")
      fun oneArg f =
	 case args of
	    [n] => n
	  | _ => Error.bug "wrong number of args for primitive"
      fun twoArgs() =
	 case args of
	    [n1, n2] => (n1, n2)
	  | _ => Error.bug "wrong number of args for primitive"
      fun threeArgs() =
	 case args of
	    [n1, n2, n3] => (n1, n2, n3)
	  | _ => Error.bug "wrong number of args for primitive"
   in
      case Prim.name prim of
	 Name.VectorFromArray =>
	    let val r = result()
	    in (case (dest(oneArg()), dest r) of
		   (Type _, Type _) => ()
		 | (Array x, Vector y) =>
		      (* can't do a coercion here because that would imply
		       * walking over each element of the array and coercing it
		       *)
		      unify(x, y)
		 | _ => typeError())
	       ; r
	    end
       | Name.ArrayUpdate =>
	    let val (a, _, x) = threeArgs()
	    in (case dest a of
		   Array x' => coerce{from = x, to = x'} (* unify(x, x') *)
		 | Type _ => ()
		 | _ => typeError())
	       ; result()
	    end
       | Name.ArraySub =>
	    (case dest(#1(twoArgs())) of
		Array x => x
	      | Type _ => result()
	      | _ => typeError())
       | Name.VectorSub =>
	    (case dest(#1(twoArgs())) of
		Vector x => x
	      | Type _ => result()
	      | _ => typeError())
       | Name.Assign =>
	    let val (r, x) = twoArgs()
	    in (case dest r of
		   Ref x' => coerce{from = x, to = x'} (* unify(x, x') *)
		 | Type _ => ()
		 | _ => typeError())
	       ; result()
	    end
       | Name.Ref =>
	    let val r = result()
	    in (case dest r of
		  Ref x => coerce{from = oneArg(), to = x} (* unify(oneArg(), x) *)
		| Type _ => ()
		| _ => typeError())
	       ; r
	    end
       | Name.Deref => (case dest(oneArg()) of
			   Ref v => v
			 | Type _ => result()
			 | _ => typeError())
       | _ => result()
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Wherever there is a use of Value.coerce here,
 * there must be a corresponding coercion in closure-convert.fun.
 *)
functor FlowAnalysis(S: FLOW_ANALYSIS_STRUCTS): FLOW_ANALYSIS = 
struct

val wordEqual: word * word -> bool = op =
   
open S
open Sxml

structure Value = AbstractValue(structure Sxml = Sxml)

fun analyze(p as Program.T{datatypes, body}) =
   let
      val {get = conArg: Con.t -> Value.t option, set = setConArg} =
	 Property.new(Con.plist, Property.initRaise("conArg", Con.layout))

      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach
	  (cons, fn {con, arg} =>
	   setConArg(con, (case arg of
			      NONE => NONE
			    | SOME t => SOME(Value.fromType t)))))

      val {set = setValue, get = value} =
	 Property.new(Var.plist, Property.initRaise("value", Var.layout))

      val varExp = value o VarExp.var
      fun varExps xs = List.map(xs, varExp)
	 
      fun merge(vs: Value.t list, ty: Type.t): Value.t =
	 let val v = Value.fromType ty
	 in List.foreach(vs, fn v' => Value.coerce{from = v', to = v})
	    ; v
	 end

(*       val traceEvalPrimExp =
 * 	 Trace.trace2("evalPrimExp", PrimExp.layout, Var.layout, Value.layout)
 *)
      
      fun evalExp(e: Exp.t): Value.t =
	 let val {decs, result} = Exp.dest e
	 in evalDecs decs; varExp result
	 end
      and evalDecs(ds: Dec.t list): unit =
	 List.foreach
	 (ds,
	  fn Dec.Fun{decs, ...} =>
                (List.foreach(decs, fn {var, ty, ...} =>
			      setValue(var, Value.fromType ty));
		 List.foreach(decs, fn {var, lambda, ...} =>
			      Value.unify(value var, evalLambda lambda)))
	   | Dec.MonoVal{var, ty, exp} =>
		let fun set v = setValue(var, v)
		   fun new() = let val v = Value.fromType ty
			       in set v; v
			       end
		in case exp of
		   PrimExp.Var x => set(varExp x)
		 | PrimExp.Const c => (new(); ())
		 | PrimExp.ConApp{con, arg, ...} =>
		      (case (arg,    conArg con) of
			  (NONE,   NONE)       => ()
			| (SOME x, SOME v)     => Value.coerce{from = varExp x,
							       to = v}
			| _                    => Error.bug "constructor mismatch"
		       ; new(); ())
		 | PrimExp.Tuple xs =>
		      if Value.typeIsFirstOrder ty
			 then (new(); ())
		      else set(Value.tuple(varExps xs))
		 | PrimExp.Select{tuple, offset} =>
		      set(Value.select(varExp tuple, offset))
		 | PrimExp.Lambda lambda => set(evalLambda lambda)
		 | PrimExp.PrimApp{prim, args, ...} =>
		      set(Value.primApply{prim = prim,
					  args = varExps args,
					  resultTy = ty})
		 | PrimExp.App{func, arg} =>
		      let val argValue = varExp arg
			 val result = new()
		      in Value.addHandler
			 (varExp func, fn l =>
			  let val lambda = Value.Lambda.dest l
			     val {arg = formal, body, ...} =
				Lambda.dest lambda
			  in Value.coerce{from = argValue, to = value formal}
			     ; Value.coerce{from = varExp(Exp.result body),
					    to = result}
			  end)
		      end
		 | PrimExp.Case{test, cases, default, ...} =>
		      let val result = new()
			 fun branch e =
			    Value.coerce{from = evalExp e, to = result}
		      in List.foreach(cases, fn (Pat.T{con, arg, ...}, e) =>
				      ((case (arg,        conArg con) of
					   (NONE,       NONE)       => ()
					 | (SOME(x, _), SOME v)     =>
					      setValue(x, v)
					 | _ => Error.bug "constructor mismatch")
				       ; branch e))
			 ; (case default of
			       NONE => ()
			     | SOME e => branch e)
		      end
		 | PrimExp.Raise _ => (new(); ())
		 | PrimExp.Handle{try, catch = (x, t), handler} =>
		      let val result = new()
		      in Value.coerce{from = evalExp try, to = result}
			 ; setValue(x, Value.fromType t)
			 ; Value.coerce{from = evalExp handler, to = result}
		      end
		end
	   | _ => Error.bug "flow analysis saw bogus Dec")

      and evalLambda lambda : Value.t =
	 let val {arg, argType, body} = Lambda.dest lambda
	 in setValue(arg, Value.fromType argType)
	    ; evalExp body
	    ; Value.lambda lambda
	 end

   in evalExp body
      ; {
	 value = value,
	 conArg = conArg
	}
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CLOSURE_CONVERT_STRUCTS = 
   sig
      structure Sxml: SXML
      structure Cps: CPS
      sharing Sxml.Atoms = Cps.Atoms
   end

signature CLOSURE_CONVERT = 
   sig
      include CLOSURE_CONVERT_STRUCTS
      
      val closureConvert: Sxml.Program.t -> Cps.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 * All local variables in the Sxml are renamed to new variables in Cps,
 * unless they are global, as determined by the Globalization pass.
 * Renaming must happen because an Sxml variable will be bound in the Cps
 * once for each lambda it occurs in.  The main trickiness is caused because a
 * property is used to implement the renamer map.  Hence, a variable binding
 * must always be visited with "newVar" or "newScope" before it is looked up
 * with "getNewVar".  "newScope" also handles resetting the variable to its
 * old value once the processing of the lambda is done.
 *)
functor ClosureConvert(S: CLOSURE_CONVERT_STRUCTS): CLOSURE_CONVERT = 
struct

open S

structure Globalize = Globalize(open Sxml)

local open Sxml
in structure Sexp = Exp
   structure Sdec = Dec
   structure Slambda = Lambda
   structure Spat = Pat
   structure SprimExp = PrimExp
   structure SvarExp = VarExp
   structure Stype = Type
   open Atoms
end

local open Cps
in
   structure Cause = Cause
   structure Ctype = Type
   structure Func = Func
   structure Cdec = Dec
   structure Cexp = DirectExp
   structure Cprogram = Program
end

structure ExtractTargs = ExtractTargs(structure Prim = Prim
				      structure Type = Ctype)

structure FlowAnalysis = FlowAnalysis(structure Sxml = Sxml)

local open FlowAnalysis
in structure Value = Value
   local open Value
   in structure Lambdas = Lambdas
   end
end

(* This is one of the results returned internally by the converter -- an
 * accumulation of toplevel Cps globals and function declarations.
 *)
structure Accum =
   struct
      structure AL = AppendList

      datatype t = T of {globals: {var: Var.t, ty: Ctype.t, exp: Cexp.t} AL.t,
			 functions: {name: Func.t,
				     args: (Var.t * Ctype.t) list,
				     body: Cps.Exp.t,
				     returns: Ctype.t list} list}

      val empty = T{globals = AL.empty, functions = []}

      fun addGlobals(T{globals, functions}, gs) =
	 T{globals = AL.append(globals, AL.fromList gs),
	   functions = functions}

      fun addGlobal(ac, g) = addGlobals(ac, [g])

      fun addFunc(T{globals, functions}, f) =
	 T{globals = globals, functions = f :: functions}

      fun done(T{globals, functions}) =
	 {functions = functions,
	  globals =
	  let
	     val exp =
		(* Must shrink because coercions may be inserted at constructor
		 * applications.  I'm pretty sure the shrinking will eliminate
		 * any case expressions/local functions.
		 * The "NoDelete" is because the shrinker is just processing
		 * globals and hence cannot safely delete a variable that
		 * has no occurrences, since there may still be occurrences in
		 * functions.
		 *)
		Cps.shrinkExpNoDelete
		(Cexp.toExp(Cexp.lett
			    {decs = AL.toList globals,
			     body = Cexp.tuple{exps = [], ty = Ctype.unit}}))
	     val decs = Cps.Exp.decs exp
	  in 
	     List.map(decs,
		      fn Cdec.Bind b => b
		       | _ => Error.bug "globalization produced a non binding")
	  end}
   end

(* val traceConvertPrimExp =
 *    Trace.trace2("convertPrimExp",
 * 		Var.layout, SprimExp.layout,
 * 		Layout.ignore: Cexp.t -> Layout.t)
 *)

(* val traceConvertExp =
 *    Trace.trace2("convertExp", Sexp.layout, Instance.layout, Cexp.layout)
 *)

structure LambdaFree = LambdaFree(Sxml)

val incKnownCount = Prim.newNullary "MLTON_incCpsKnown"
val incUnknownCount = Prim.newNullary "MLTON_incCpsUnknown"

fun closureConvert(program as Sxml.Program.T{datatypes, body}): Cps.Program.t =
   let
      val {value, conArg} =
	 Control.pass
	 {name = "flow analysis",
	  suffix = "flow",
	  thunk = fn () => FlowAnalysis.analyze program,
	  style = Control.No,
	  display =
	  Control.Layouts
	  (fn ({value, ...}, display) =>
	   Sexp.foreachBoundVar
	   (body, fn (x, _, _) => display(let open Layout
					  in seq[Var.layout x,
						 str " ",
						 Value.layout(value x)]
					  end)))}

      val {lambdaFree, lambdaRec} =
	 Control.traceCall "free variables" LambdaFree.lambdaFree program

      val {isGlobal} =
	 Control.traceCall "globalize"
	 Globalize.globalize{program = program, lambdaFree = lambdaFree}

(*      val isGlobal = Trace.trace("isGlobal", Var.layout, Bool.layout) isGlobal *)

      local
	 fun filter f l = List.keepAll(f l, not o isGlobal)
      in
	 val lambdaFree = filter lambdaFree
	 val lambdaRec = filter lambdaRec
      end
	 
      (* For each lambda, keep track of
       * - the "original" name in the source program, so the closure
       *   conversion output has some readability.
       * - the type of its environment record
       *)
      local
	 val {get: Slambda.t -> {
				 name: Func.t,
				 ty: Ctype.t option ref
				},
	      set} = Property.new(Slambda.plist,
				  Property.initRaise("info", Layout.ignore))
      in
	 val lambdaType = #ty o get
	 val lambdaName = #name o get

	 val _ =
	    let
	       fun doit(l, x: Var.t) =
		  set(l, {name = Func.newString(Var.originalName x),
			  ty = ref NONE})
	    in Sexp.foreachExp
	       (body, fn e =>
		let val {decs, ...} = Sexp.dest e
		in List.foreach
		   (decs, fn d =>
		    case d of
		       Sdec.MonoVal{var, exp, ...} =>
			  (case exp of
			      SprimExp.Lambda l => doit(l, var)
			    | _ => ())
		     | Sdec.Fun{decs, ...} =>
			  List.foreach(decs, fn {var, lambda, ...} =>
				       doit(lambda, var))
		     | _ => ())
		end)
	    end
      end
   
      val {get = lambdasInfoOpt, set = setLambdasInfo} =
	 Property.new(Lambdas.plist, Property.initConst NONE)

      val {hom = convertType, destroy = destroyConvertType} =
	 Stype.makeMonoHom{con = fn (_, c, ts) => Ctype.con(c, ts)}

      fun convertTypes ts = List.map(ts, convertType)

      val {get = valueTypeOpt, set = setValueType} =
	 Property.new(Value.plist, Property.initConst NONE)

      (* accumulates the new datatypes built for sets of lambdas *)
      val newDatatypes = ref []

      fun valueType(v: Value.t): Ctype.t =
	 case valueTypeOpt v of
	    SOME t => t
	  | NONE =>
	       let
		  val t = 
		     case Value.dest v of
			Value.Type t => convertType t
		      | Value.Ref v => Ctype.reff(valueType v)
		      | Value.Array v => Ctype.array(valueType v)
		      | Value.Vector v => Ctype.vector(valueType v)
		      | Value.Tuple vs => Ctype.tuple(List.map(vs, valueType))
		      | Value.Lambdas ls => #ty(lambdasInfo ls)
	       in setValueType(v, SOME t); t
	       end
      and lambdasInfo ls =
	 case lambdasInfoOpt ls of
	    SOME info => info
	  | NONE => 
	       let val tycon = Tycon.newString "lambdas"
		  val cons = List.map(Lambdas.toList ls, fn l =>
				      {lambda = Value.Lambda.dest l,
				       con = Con.newString "Env"})
		  val ty = Ctype.con(tycon, [])
		  val info = {ty = ty, cons = cons}
	       in setLambdasInfo(ls, SOME info)
		  ; (ListRef.push
		     (newDatatypes,
		      {tycon = tycon,
		       cons = List.map(cons, fn {lambda, con} =>
				       {con = con, args = [envType lambda]})}))
		  ; info
	       end

      and varType x = valueType(value x)

      and envType l =
	 let val ty = lambdaType l
	 in case !ty of
	    NONE => let val t = Ctype.tuple(List.map(lambdaFree l, varType))
		    in ty := SOME t; t
		    end
	  | SOME t => t
	 end

      fun valueLambdasInfo v =
 	 case Value.dest v of
	    Value.Lambdas l => lambdasInfo l
	  | _ => Error.bug "valueLambdasInfo of non-lambda"

      val varLambdasInfo = valueLambdasInfo o value

      val varExp = value o SvarExp.var
      val expValue = varExp o #result o Sexp.dest
	 
      val expType = valueType o expValue

      val datatypes =
	 List.map(datatypes, fn {tycon, cons, ...} =>
		  {tycon = tycon,
		   cons = List.fold(cons, [], fn ({con, ...}, cons) =>
				    {con = con,
				     args = (case conArg con of
						NONE => []
					      | SOME v => [valueType v])}
				    :: cons)})

      (* Variable renaming *)
      local
	 val {get, set} = Property.new(Var.plist, Property.initFun ref)
      in
	 fun getNewVar x = !(get x)

	 fun newVar(x: Var.t): Var.t =
	    if isGlobal x
	       then x
	    else let val x' = Var.new x
		 in get x := x'; x'
		 end

	 fun newScope(xs: Var.t list, f: Var.t list -> 'a): 'a =
	    let val refs = List.map(xs, get)
	       val old = List.map(refs, !)
	       val new = List.map(xs, newVar)
	       val res = f new
	    in List.foreach2(refs, old, op :=)
	       ; res
	    end
      end

(*      val newVar = Trace.trace("newVar", Var.layout, Var.layout) newVar *)

      (*------------------------------------*)
      (*               coerce               *)
      (*------------------------------------*)

(*       val traceCoerce =
 * 	 Trace.trace3("coerce", Cexp.layout, Value.layout, Value.layout,
 * 		      Cexp.layout)
 *)

(*       val traceCoerceTuple =
 * 	 let val layoutValues = List.layout(", ", Value.layout)
 * 	 in Trace.trace3("coerceTuple", Cexp.layout,
 * 			 layoutValues, layoutValues, Cexp.layout)
 * 	 end
 *)

      fun coerce arg: Cexp.t =
(*	 traceCoerce *)
	 (fn (e: Cexp.t, from: Value.t, to: Value.t) =>
	  if Value.equals(from, to)
	     then e
	  else 
	     case (Value.dest from, Value.dest to) of
		(Value.Tuple vs, Value.Tuple vs') =>
		   coerceTuple(e, valueType from, vs, valueType to, vs')
	      | (Value.Lambdas ls, Value.Lambdas ls') =>
		   if Lambdas.equals(ls, ls') then e
		   else
		      let val {cons, ...} = lambdasInfo ls
			 val {cons = cons', ty, ...} = lambdasInfo ls'
			 val {get = lambdaCon, set = setLambdaCon, destroy}=
			    Property.newDest(Slambda.plist, Property.initRaise("lambda target", Slambda.layout))
		      in List.foreach(cons', fn {lambda, con, ...} =>
				      setLambdaCon(lambda, con));
			 Cexp.casee
			 {cause = Cause.Coerce,
			  test = e, default = NONE,
			  ty = ty,
			  cases =
			  List.map
			  (cons, fn {lambda, con} =>
			   let val tuple = (Var.newNoname(), envType lambda)
			   in {con = con,
			       args = [tuple],
			       body = Cexp.conApp{con = lambdaCon lambda,
						  ty = ty,
						  args = [Cexp.var tuple]}}
			   end)}
			 before destroy()
		      end
	   | _ => Error.bug "impossible coercion") arg

      and coerceTuple arg =
(*	 traceCoerceTuple *)
	 (fn (e: Cexp.t,
	      ty: Ctype.t, vs: Value.t list,
	      ty': Ctype.t, vs': Value.t list) =>
	  if Ctype.equals(ty, ty')
	     then e
	  else 
	     Cexp.detuple
	     {tuple = e,
	      body =
	      fn components => 
	      Cexp.tuple
	      {exps = List.map3(components, vs, vs', fn (x, v, v') =>
				coerce(Cexp.var(x, valueType v), v, v')),
	       ty = ty'}}) arg
	 
      fun convertVar x : Cexp.t = Cexp.var(getNewVar x, varType x)
      
      fun convertVarExp(x: SvarExp.t) = convertVar(SvarExp.var x)
	 
      fun convertVarExps xs = List.map(xs, convertVarExp)

      (*------------------------------------*)      	       
      (*               apply                *)
      (*------------------------------------*)

      fun apply{func, arg, resultVar, resultVal}: Cexp.t =
	 let val funcVal = varExp func
	    val argVal = varExp arg
	    val argExp = convertVarExp arg
	    val ty = valueType resultVal
	    val {cons, ...} = valueLambdasInfo funcVal
	 in Cexp.casee
	    {cause = Cause.Dispatch,
	     test = convertVarExp func,
	     ty = ty,
	     default = NONE,
	     cases =
	     List.map
	     (cons, fn {lambda, con} =>
	      let val {arg = param, body, ...} = Slambda.dest lambda
		 val result = expValue body
		 val env = (Var.newString "env", envType lambda)
	      in {con = con,
		  args = [env],
		  body = (coerce(Cexp.call
				 {func = lambdaName lambda,
				  args = [Cexp.var env,
					  coerce(argExp, argVal, value param)],
				  ty = valueType result},
				 result, resultVal))}
	      end)}
	 end

      (*------------------------------------*)
      (*             convertExp             *)
      (*------------------------------------*)
	 
      fun envTuple(l: Slambda.t): Cexp.t =
	 Cexp.tuple{exps = List.map(lambdaFree l, convertVar),
		    ty = envType l}

      fun recursives(old: Var.t list, new: Var.t list, env) =
	 List.fold2
	 (old, new, [], fn (old, new, ac) =>
	  let val {cons, ty, ...} = varLambdasInfo old
	  in case cons of
	     [{con, ...}] =>
		{var = new,
		 ty = ty,
		 exp = Cexp.conApp{con = con, ty = ty, args = [Cexp.var env]}}
		:: ac
	   | _ => Error.bug "val rec lambda must be singleton"
	  end)

      (* Closure convert an expression, returning:
       *   - the target cps expression
       *   - a list of global declarations (in order)
       *   - a list of function declarations
       * Accumulate the globals onto the end of the given ones.
       *)
      fun convertExp(e: Sexp.t, ac: Accum.t): Cexp.t * Accum.t =
	 let val {decs, result} = Sexp.dest e
	    (* process decs left to right, since bindings of variables
	     * must be visited before uses
	     *)
	    val (decs, ac) =
	       List.fold
	       (decs, ([], ac), fn (d, (binds, ac)) =>
		case d of
		   Sdec.MonoVal{var, ty, exp} =>
		      let val (exp, ac) = convertPrimExp(var, exp, ac)
			 val bind = {var = newVar var,
				     ty = varType var,
				     exp = exp}
		      in if isGlobal var
			    then (binds, Accum.addGlobal(ac, bind))
			 else (bind :: binds, ac)
		      end
		 | Sdec.Fun{decs, ...} =>
		      (case decs of
			  [] => (binds, ac)
			| {var, ty, lambda} :: _ => 
			     let val tupleVar = Var.newString "tuple"
				val tupleTy = envType lambda
				val binds' =
				   {var = tupleVar,
				    ty = tupleTy,
				    exp = envTuple lambda}
				   :: recursives(List.map(decs, #var),
						 List.map(decs, newVar o #var),
						 (tupleVar, tupleTy))
				val (binds, ac) =
				   if isGlobal var
				      then (binds, Accum.addGlobals(ac, binds'))
				   else (List.fold(binds', binds, op ::), ac)
			     in (binds,
				 List.fold(decs, ac, fn ({lambda, ...}, ac) =>
					   convertLambda(lambda, ac)))
			     end)
		 | _ => Error.bug "closure convert saw strange dec")
	 in (Cexp.lett{decs = List.reverse decs,
		       body = convertVarExp result},
	     ac)
	 end

      and convertPrimExp(x: Var.t, e: SprimExp.t, ac: Accum.t): Cexp.t * Accum.t =
	 let
	    val v = value x
	    val ty = valueType v
	    fun convertJoin(e, ac) =
	       let val (e', ac) = convertExp(e, ac)
	       in (coerce(e', expValue e, v), ac)
	       end
	    fun simple e = (e, ac)
	 in case e of
	    SprimExp.Var x => simple(convertVarExp x)
	  | SprimExp.Const c => simple(Cexp.const c)
	  | SprimExp.PrimApp{prim, targs, args} =>
	       let open Prim.Name
		  val (targs, args) =
		     case (Prim.name prim, args) of
			(ArrayUpdate, [a, i, y]) =>
			   let val v = Value.dearray(varExp a)
			   in ([valueType v],
			       [convertVarExp a,
				convertVarExp i,
				coerce(convertVarExp y, varExp y, v)])
			   end
		      | (Assign, [r, y]) =>
			   let val v = Value.deref(varExp r)
			   in ([valueType v],
			       [convertVarExp r,
				coerce(convertVarExp y, varExp y, v)])
			   end
		      | (Ref, [y]) =>
			   let val v = Value.deref v
			   in ([valueType v],
			       [coerce(convertVarExp y, varExp y, v)])
			   end
		      | _ =>
			   (ExtractTargs.extract
			    {prim = prim,
			     args = List.map(args, varType o SvarExp.var),
			     result = ty},
			    convertVarExps args)
	       in simple(Cexp.primApp{prim = prim,
				      ty = ty,
				      targs = targs,
				      args = args})
	       end
	  | SprimExp.Tuple xs =>
	       simple(Cexp.tuple{exps = convertVarExps xs, ty = ty})
	  | SprimExp.Select{tuple, offset} =>
	       simple(Cexp.select{tuple = convertVarExp tuple,
				  offset = offset,
				  ty = ty})
	  | SprimExp.ConApp{con = con, arg, ...} =>
	       simple
	       (Cexp.conApp
		{con = con,
		 ty = ty,
		 args = (case (arg, conArg con) of
			    (NONE, NONE) => []
			  | (SOME arg, SOME conArg) =>
			       let val argVal = varExp arg
				  val arg = convertVarExp arg
			       in if Value.equals(argVal, conArg)
				     then [arg]
				  else [coerce(arg, argVal, conArg)]
			       end
			  | _ => Error.bug "constructor mismatch")})
	  | SprimExp.Raise x => simple(Cexp.raisee(convertVarExp x))
	  | SprimExp.Handle{try, catch = (catch, _), handler} =>
	       let val (try, ac) = convertJoin(try, ac)
		  val catch = (newVar catch, varType catch)
		  val (handler, ac) = convertJoin(handler, ac)
	       in (Cexp.handlee{try = try, ty = ty,
				catch = catch, handler = handler},
		   ac)
	       end
	  | SprimExp.Case{test, cases, default} =>
	       let
		  val (default, ac) =
		     case default of
			NONE => (NONE, ac)
		      | SOME e => let val (e, ac) =  convertJoin(e, ac)
				  in (SOME e, ac)
				  end
		  val (cases, ac) =
		     List.fold
		     (cases, ([], ac),
		      fn ((p as Spat.T{con, arg, ...}, e), (cases, ac)) =>
		      let
			 val args =
			    case (conArg con, arg) of
			       (NONE, NONE) => []
			     | (SOME v, SOME(arg, _)) =>
				  [(newVar arg, valueType v)]
			     | _ => Error.bug "constructor mismatch"
			 val (body, ac) = convertJoin(e, ac)
		      in ({con = con, args = args, body = body} :: cases, ac)
		      end)
	       in (Cexp.casee
		   {cause = Cause.User,
		    test = convertVarExp test,
		    ty = ty, cases = cases, default = default},
		   ac)
	       end
	  | SprimExp.Lambda l =>
	       let val ac = convertLambda(l, ac)
		  val {cons, ...} = valueLambdasInfo v
	       in case List.peek(cons, fn {con, lambda = l'} =>
				 Slambda.equals(l, l')) of
		  NONE => Error.bug "lambda must exist in its own set"
		| SOME{con, ...} =>
		     (Cexp.conApp{con = con, ty = ty, args = [envTuple l]}, ac)
	       end
	  | SprimExp.App{func, arg} =>
	       (apply{func = func, arg = arg, resultVar = x, resultVal = v},
		ac)
	 end

      and convertLambda(lambda, ac: Accum.t): Accum.t =
	 let val {arg = argVar, body, ...} = Slambda.dest lambda
	    val env = (Var.newString "env", envType lambda)
	    val args = [env, (newVar argVar, varType argVar)]
	    val returns = [expType body]
	    val recs = lambdaRec lambda
	 in newScope
	    (lambdaFree lambda, fn components =>
	     newScope
	     (recs, fn recs' =>
	      let val decs = recursives(recs, recs', env)
		 val (body, ac) = convertExp(body, ac)
	      in Accum.addFunc
		 (ac, {name = lambdaName lambda,
		       args = args, returns = returns,
		       body =
		       Cps.shrinkExp []
		       (Cexp.toExp
			(Cexp.lett
			 {decs = decs,
			  body = Cexp.detupleBind{tuple = Cexp.var env,
						  components = components,
						  body = body}}))})
	      end))
	 end

      (*------------------------------------*)
      (*    main body of closure convert    *)
      (*------------------------------------*)

      val main = Func.newString "main"

      val {functions, globals} =
	 Control.traceCall "convert"
	 (fn () =>
	  let val (body, ac) = convertExp(body, Accum.empty)
	  in Accum.done
	     (Accum.addFunc(ac, {name = main,
				 args = [],
				 returns = [Ctype.unit],
				 body = Cps.shrinkExp [] (Cexp.toExpHalt body)}))
	  end) ()

      val datatypes = datatypes @ (!newDatatypes)

      val program =
	 Cprogram.T{datatypes = datatypes,
		    globals = globals,
		    functions = functions,
		    main = main}
	 
   in destroyConvertType();
      Value.destroy();
      Cprogram.clear program;
      program
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CORE_ML_STRUCTS = 
   sig
      include ATOMS
      structure Type: TYPE
      sharing Type = Prim.Type
   end

signature CORE_ML = 
   sig
      include CORE_ML_STRUCTS

      structure Pat :
	 sig
	    datatype t =
	       Wild
	     | Var of Var.t
	     | Const of Ast.Const.t
	     | Con of {
		       con: Con.t,
		       arg: t option
		      }
	     | Record of {
			  flexible: bool,
			  record: t Record.t
			 }
	     | Constraint of t * Type.t
	     | Layered of Var.t * t

	    (* constants *)
	    val unit: t
	       
	    (* constructors *)
	    val list: t list -> t
	    val tuple: t list -> t
	    val record: {
			 flexible: bool,
			 record: t Record.t
			} -> t

	    (* a list (without duplicates) of variables occurring in a pattern *)
	    val vars: t -> Var.t list

	    (* replace all variables with Wild *)
	    val removeVars: t -> t

	    (* removeOthersReplace(pat, old,new) replaces all variables in pat
	     * with Wild, except for old, which it replaces with new
	     *)
	    val removeOthersReplace: t * Var.t * Var.t -> t

	    (* true if the pattern contains a constant, constructor or variable *)
	    val isRefutable: t -> bool

	    val foreachVar: t * (Var.t -> unit) -> unit

	    (* conversion to Ast *)
	    val toAst: t -> Ast.Pat.t

	    val layout: t -> Layout.t
	 end

      structure Exp :
	 sig
	    type dec
	    datatype t =
	       Const of Ast.Const.t
	     | Var of Var.t
	     | Prim of Prim.t
	     | Con of Con.t
	     | Record of t Record.t
	     | Fn of (Pat.t * t) list
	     | App of t * t
	     | Let of dec list * t
	     | Constraint of t * Type.t
	     | Handle of t * (Pat.t * t) list
	     | Raise of t

	    (* macros *)
	    val selector: Record.Field.t -> t
	    val iff: t * t * t -> t
	    val andAlso: t * t -> t
	    val orElse: t * t -> t
	    val whilee: {test: t, expr: t} -> t
	    val casee: t * (Pat.t * t) list -> t

	    (* standard constructions *)
	    val unit: t
	    val seq: t list -> t
	    val list: t list -> t
	    val tuple: t list -> t
	    val delay: t -> t
	    val force: t -> t
	    val lambda: Var.t * t -> t
	    val compose: unit -> t

	    val foreachVar: t * (Var.t -> unit) -> unit
	       
	    (* true if the expression may side-effect. See p 19 of Definition *)
	    val isExpansive: t -> bool
	       
	    val layout: t -> Layout.t
	 end

      structure Dec :
	 sig
	    datatype t =
	       Val of {pat: Pat.t,
		       tyvars: Tyvar.t list,
		       exp: Exp.t}
	     | Fun of {
		       tyvars: Tyvar.t list,
		       decs: {
			      var: Var.t,
			      ty: Type.t option,
			      rules: (Pat.t * Exp.t) list
			     } list
		      }
	     | Datatype of {
			    tyvars: Tyvar.t list,
			    tycon: Tycon.t,
			    cons: {
				   con: Con.t,
				   arg: Type.t option
				  } list
			   } list
	     | Exception of {
			     con: Con.t,
			     arg: Type.t option
			    }
	     | Overload of {
			    var: Var.t,
			    scheme: Scheme.t,
			    ovlds: Var.t list
			   }

	    val toAst: t -> Ast.Dec.t
	    val layout: t -> Layout.t

	    val isExpansive: t -> bool
	 end
      where type t = Exp.dec

      structure Constructors:
	 sig
	    datatype t =
	       T of {
		     exceptions: {
				  con: Con.t,
				  arg: Type.t option
				  } list,
		     datatypes: {
				 tycon: Tycon.t,
				 tyvars: Tyvar.t list,
				 cons: {
					con: Con.t,
					arg: Type.t option
					} list
				 } list
		     }

	    val empty: t
	    val + : t * t -> t
	    val fromDecs: Dec.t list -> t
	    val toDecs: t -> Dec.t list
	    val layout: t -> Layout.t
	 end

      structure Program :
	 sig
	    datatype t = T of {
			       constructors: Constructors.t,
			       decs: Dec.t list
			      }

	    val layoutStats: t -> Layout.t

(*	    val empty: t *)
(*	    val append: t * t -> t *)

	    val layout: t -> Layout.t
	 end
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature DEAD_CODE_STRUCTS = 
   sig
      structure CoreML: CORE_ML
   end

signature DEAD_CODE = 
   sig
      include DEAD_CODE_STRUCTS
      
      val deadCode :
	 {basis: CoreML.Dec.t list,
	  user: CoreML.Dec.t list} -> CoreML.Dec.t list (* basis *)
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor DeadCode(S: DEAD_CODE_STRUCTS): DEAD_CODE = 
struct

open S
open CoreML
open Pat Dec Exp

fun deadCode{basis, user} =
   let
      val {get = varIsUsed, set = setVarUsed, destroy} =
	 Property.newDest(Var.plist, Property.initConst false)

      fun foreachDefinedVar(d, f) =
	 case d of
	    Val{pat, ...} => Pat.foreachVar(pat, f)
	  | Fun{decs, ...} => List.foreach(decs, f o #var)
	  | Overload{var, ...} => f var
	  | _ => ()
		

      fun patVarIsUsed p =
	 let
	    exception Used
	 in (Pat.foreachVar(p, fn x => if varIsUsed x
					 then raise Used
				      else ());
	     false) handle Used => true
	 end	    
	 
      fun decIsNeeded d =
	 case d of
	    Val{pat, ...} => patVarIsUsed pat
	  | Fun{decs, ...} => List.exists(decs, varIsUsed o #var)
	  | Datatype _ => true
	  | Exception _ => true
	  | Overload{var, ...} => varIsUsed var

      fun useVar x = setVarUsed(x, true)
      fun useExp e = Exp.foreachVar(e, useVar)
      fun useDec d = 
	 case d of
	    Val{exp, ...} => useExp exp
	  | Fun{decs, ...} => List.foreach(decs, fn {rules, ...} =>
					   List.foreach(rules, useExp o #2))
	  | Datatype _ => ()
	  | Exception _ => ()
	  | Overload{ovlds, ...} => List.foreach(ovlds, useVar)
	       
   in List.foreach(user, useDec);

      List.fold(List.reverse basis, [], fn (d, b) =>
		if decIsNeeded d
		   then (useDec d; d :: b)
		else b)
      before destroy()
   end

end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor CoreML(S: CORE_ML_STRUCTS): CORE_ML = 
struct

open S

local open Ast
in structure Adec = Dec
end

fun makeList (con, app, tuple) l =
   List.fold(List.reverse l, con Con.nill, fn (x1, x2) =>
	     app(Con.cons, tuple[x1, x2]))

structure Pat =
   struct
      datatype t =
	 Wild
       | Var of Var.t
       | Const of Ast.Const.t
       | Con of {con: Con.t, arg: t option}
       | Record of {flexible: bool, record: t Record.t}
       | Constraint of t * Type.t
       | Layered of Var.t * t

      fun isRefutable p =
	 case p of
	    Wild => false
	  | Var _ => false
	  | Const _ => true
	  | Con _ => true
	  | Record{record, ...} => Record.exists(record, isRefutable)
	  | Constraint(p, _) => isRefutable p
	  | Layered(_, p) => isRefutable p

      fun vars'(p, l) =
	 case p of
	    Wild => l
	  | Var x => x :: l
	  | Const _ => l
	  | Con{arg, ...} => (case arg of
				NONE => l
			      | SOME p => vars'(p, l))
	  | Record{record, ...} => Record.fold(record, l, vars')
	  | Constraint(p, _) => vars'(p, l)
	  | Layered(x, p) => vars'(p, x :: l)

      fun vars p = vars'(p, [])

      fun removeVarsPred(p: t, pred: Var.t -> bool): t =
	 let
	    fun loop p =
	       case p of
		  Wild => Wild
		| Const _ => p
		| Var x => if pred x then Wild else p
		| Record{flexible, record} =>
		     Record{flexible = flexible,
			    record = Record.map(record, loop)}
		| Con{con, arg} => Con{con = con, arg = (case arg of
							   NONE => NONE
							 | SOME p => SOME(loop p))}
		| Constraint(p, t) => Constraint(loop p, t)
		| Layered(_, p) => loop p
	 in loop p
	 end

      fun removeVars p = removeVarsPred(p, fn _ => true)

      fun removeOthersReplace(p, x, y) =
	 let
	    fun loop p =
	       case p of
		  Wild => Wild
		| Const _ => p
		| Var x' => if Var.equals(x, x') then Var y else Wild
		| Record{record, flexible} =>
		     Record{flexible = flexible,
			    record = Record.map(record, loop)}
		| Con{con, arg} => Con{con = con,
				      arg = (case arg of
						NONE => NONE
					      | SOME p => SOME(loop p))}
		| Constraint(p, t) => loop p (*Constraint(loop p, t)*)
		| Layered(_, p) => loop p
	 in loop p
	 end

      val tuple =
	 fn [p] => p
	  | ps => Record{flexible = false, record = Record.tuple ps}

      val unit = tuple[]
	       
      val list = makeList(fn c => Con{con = c, arg = NONE},
			  fn (c, p) => Con{con = c, arg = SOME p},
			  tuple)

      val record = Record

      fun tuple[p] = p
	| tuple ps = record{flexible = false, record = Record.tuple ps}

      val unit = tuple[]

      local
	 fun make c = Con{con = c, arg = NONE}
      in val truee = make Con.truee
	 val falsee = make Con.falsee
      end

      fun foreachVar(p, f) =
	 let
	    fun loop p =
	       case p of
		  Var x => f x
		| Con{arg = SOME p, ...} => loop p
		| Record{record, ...} => Record.foreach(record, loop)
		| Constraint(p, _) => loop p
		| Layered(x, p) => (f x; loop p)
		| _ => ()
	 in loop p
	 end
	 
      local structure Pat = Ast.Pat
      in
	 fun toAst p =
	    case p of
	       Wild => Pat.wild
	     | Var x => Pat.var(Var.toAst x)
	     | Const c => Pat.const c
	     | Record{record, flexible} =>
		  (case (flexible, Record.detupleOpt record) of
		      (false, SOME ps) => Pat.tuple(List.map(ps, toAst))
		    | _ =>
			 Pat.make
			 (Pat.Record
			  {flexible = flexible,
			   items =
			   List.map(Record.toList record, fn (field, p) =>
				    Pat.Item.Field(field, toAst p))}))
	     | Con{con, arg} =>
		  let val con = Con.toAst con
		  in case arg of
		     NONE => Pat.con con
		   | SOME p => Pat.app(con, toAst p)
		  end
	     | Constraint(p, t) => Pat.constraint(toAst p,  Type.toAst t)
	     | Layered(x, p) =>
		  Pat.layered{fixop = Ast.Fixop.None,
			      var = Var.toAst x,
			      constraint = NONE,
			      pat = toAst p}

	 val layout = Pat.layout o toAst
      end
   end

datatype dec =
   Val of {pat: Pat.t, tyvars: Tyvar.t list, exp: exp}
  | Fun of {tyvars: Tyvar.t list,
	    decs: {var: Var.t,
		    ty: Type.t option,
		    rules: rules} list}
  | Datatype of {
		 tyvars: Tyvar.t list,
		 tycon: Tycon.t,
		 cons: {
			con: Con.t,
			arg: Type.t option
			} list
		 } list
  | Exception of {
		  con: Con.t,
		  arg: Type.t option
		  }
  | Overload of {var: Var.t,
		 scheme: Scheme.t,
		 ovlds: Var.t list}
and exp =
   Var of Var.t
  | Prim of Prim.t
  | Const of Ast.Const.t
  | Con of Con.t
  | Record of exp Record.t
  | Fn of rules
  | App of exp * exp
  | Let of dec list * exp
  | Constraint of exp * Type.t
  | Handle of exp * rules
  | Raise of exp
withtype rules = (Pat.t * exp) list

local
   local open Ast
   in structure Dec = Dec
      structure Exp = Exp
      structure Longvar = Longvar
   end
in
   fun astDatatype ds =
      Dec.datatypee
      (List.map(ds, fn {tyvars, tycon, cons} =>
		{tyvars = tyvars,
		 tycon = Tycon.toAst tycon,
		 cons = List.map(cons, fn {con, arg} =>
				 (Con.toAst con, Type.optionToAst arg))}))
      
   fun decToAst d =
      case d of
	 Val{pat, tyvars, exp} =>
	    Dec.make(Dec.Val(tyvars, [(Pat.toAst pat, expToAst exp)]))
       | Fun{tyvars, decs} =>
	    Dec.funn(tyvars,
		     List.map(decs, fn {var, ty, rules} =>
			      (Var.toAst var,
			       rulesToAst rules,
			       Type.optionToAst ty)))
       | Datatype ds => astDatatype ds
       | Exception{con, arg} =>
	    Dec.exceptionn(Con.toAst con, Type.optionToAst arg)
       | Overload{var, scheme, ovlds} =>
	    Dec.make(Dec.Overload(Var.toAst var,
				  Type.toAst(Scheme.ty scheme),
				  List.map(ovlds, fn x =>
					   Longvar.short(Var.toAst x))))

   and expToAst e =
      case e of
	 Var x => Exp.var(Var.toAst x)
       | Prim p => Exp.longvid(Prim.longid p)
       | Const c => Exp.const c
       | Con c => Exp.con(Con.toAst c)
       | Record r => Exp.record(Record.map(r, expToAst))
       | Fn rs => Exp.fnn(rulesToAst rs)
       | App(e1, e2) => Exp.app(expToAst e1, expToAst e2)
       | Let(ds, e) => Exp.lett(List.map(ds, decToAst), expToAst e)
       | Constraint(e, t) => Exp.constraint(expToAst e, Type.toAst t)
       | Handle(e, rs) => Exp.handlee(expToAst e, rulesToAst rs)
       | Raise e => Exp.raisee(expToAst e)
   and rulesToAst rs =
      List.map(rs, fn (p, e) => (Pat.toAst p, expToAst e))
end

fun makeForeachVar f =
   let
      fun exp e =
	 case e of
	    Var x => f x
	  | Record r => Record.foreach(r, exp)
	  | Fn rs => rules rs
	  | App(e1, e2) => (exp e1; exp e2)
	  | Let(ds, e) => (List.foreach(ds, dec); exp e)
	  | Constraint(e, _) => exp e
	  | Handle(e, rs) => (exp e; rules rs)
	  | Raise e => exp e
	  | _ => ()
      and rules rs = List.foreach(rs, fn (_, e) => exp e)
      and dec d =
	 case d of
	    Val{exp = e, ...} => exp e
	  | Fun{decs, ...} => List.foreach(decs, fn {rules = rs, ...} =>
					  rules rs)
	  | Overload{ovlds, ...} => List.foreach(ovlds, f)
	  | _ => ()
   in {exp = exp, dec = dec}
   end

structure Exp =
   struct
      type dec = dec
      datatype t = datatype exp

      val toAst = expToAst

      fun foreachVar(e, f) = #exp(makeForeachVar f) e

      val fnn = Fn

      fun compose() =
	 let val f = Var.newNoname()
	    val g = Var.newNoname()
	    val x = Var.newNoname()
	 in fnn[(Pat.tuple[Pat.Var f, Pat.Var g],
		 fnn[(Pat.Var x, App(Var f, App(Var g, Var x)))])]
	 end
	 
      val rec isExpansive =
	 fn Var _ => false
	  | Const _ => false
	  | Con _ => false
	  | Fn _ => false
	  | Prim _ => false
	  | Record r => Record.exists(r, isExpansive)
	  | Constraint(e, _) => isExpansive e
	  | App(e1, e2) => (case e1 of
			      Con c => Con.equals(c, Con.reff) orelse isExpansive e2
			    | _ => true)
	  | _ => true

      val record = Record
	 
      fun seq es =
	 let val (es, e) = List.splitLast es
	 in case es of
	    [] => e
	  | _ => Let(List.map(es, fn e =>
			      Val{pat = Pat.Wild, tyvars = [], exp = e}),
		     e)
	 end

      fun lambda(x, e) = fnn[(Pat.Var x, e)]

      fun delay e = fnn[(Pat.unit, e)]

      fun casee(test, rules) = App(Fn rules, test)

      fun tuple[e] = e
	| tuple es = record(Record.tuple es)

      val unit = tuple[]

      fun force e = App(e, unit)
	 
      val list = makeList(Con, fn (c, e) => App(Con c, e), tuple)

      fun selector f =
	 let val x = Var.newNoname()
	 in fnn[(Pat.Record{flexible = true,
			    record = Record.fromList[(f, Pat.Var x)]},
		 Var x)]
	 end

      fun iff(test, thenCase, elseCase) =
	 casee(test, [(Pat.truee, thenCase),
		      (Pat.falsee, elseCase)])

      val truee = Con Con.truee
      val falsee = Con Con.falsee
	 
      fun andAlso(e1, e2) = iff(e1, e2, falsee)
      fun orElse(e1, e2) = iff(e1, truee, e2)

      fun whilee{test, expr} =
	 let val loop = Var.newNoname()
	 in Let([Fun{tyvars = [],
		     decs = [{var = loop, ty = NONE,
			      rules = [(Pat.tuple[],
					iff(test,
					    seq[expr, App(Var loop, unit)],
					    unit))]}]}],
		App(Var loop, unit))
	 end

      val layout = Ast.Exp.layout o toAst
   end

structure Dec =
   struct
      datatype t = datatype dec

      val isExpansive =
	 fn Val{exp, ...} => Exp.isExpansive exp
	  | _ => false

      val toAst = decToAst

      val layout = Adec.layout o toAst
   end

structure Constructors =
   struct
      datatype t =
	 T of {
	       exceptions: {
			    con: Con.t,
			    arg: Type.t option
			    } list,
	       datatypes: {
			   tycon: Tycon.t,
			   tyvars: Tyvar.t list,
			   cons: {
				  con: Con.t,
				  arg: Type.t option
				  } list
			   } list
	       }

      fun layout(T{exceptions, datatypes}) =
	 let open Layout
	 in align(Ast.Dec.layout(astDatatype datatypes)
		  ::
		  List.map(exceptions, fn {con, arg} =>
			   Ast.Dec.layout
			   (Ast.Dec.exceptionn
			    (Con.toAst con, Type.optionToAst arg))))
	 end

      val empty = T{exceptions = [], datatypes = []}

      val op + =
	 fn (T{exceptions = e, datatypes = d}, T{exceptions = e', datatypes = d'})
	 => T{exceptions = e @ e',
	      datatypes = d @ d'}

      local
	 fun loopDecs(ds, cs) =
	    List.fold
	    (ds, cs, fn (d, cs as T{datatypes, exceptions}) =>
	     case d of
		Val{exp, ...} => loopExp(exp, cs)
	      | Fun{decs, ...} => List.fold(decs, cs, fn ({rules, ...}, cs) =>
					    loopRules(rules, cs))
	      | Datatype ds => T{exceptions = exceptions,
				 datatypes = ds @ datatypes}
	      | Exception ca => T{exceptions = ca :: exceptions,
				  datatypes = datatypes}
	      | Overload _ => cs)
	 and loopExp(e, cs) =
	    case e of
	       Fn rs => loopRules(rs, cs)
	     | App(e1, e2) => loopExp(e1, loopExp(e2, cs))
	     | Let(ds, e) => loopDecs(ds, loopExp(e, cs))
	     | Constraint(e, _) => loopExp(e, cs)
	     | Handle(e, rs) => loopExp(e, loopRules(rs, cs))
	     | Raise e => loopExp(e, cs)
	     | _ => cs
	 and loopRules(rs, cs) =
	    List.fold(rs, cs, fn ((_, e), cs) => loopExp(e, cs))
      in
	 fun fromDecs ds = loopDecs(ds, empty)
      end

(*       val fromDecs =
 * 	 Trace.trace("fromDecs", List.layout Dec.layout, layout) fromDecs
 *)
   
      fun toDecs(T{datatypes, exceptions}) =
	 Datatype datatypes :: List.map(exceptions, Exception)
   end

structure Program =
   struct
      datatype t = T of {decs: Dec.t list,
			 constructors: Constructors.t}

(*      val empty = T[] *)

(*      fun append(T ds, T ds') = T(List.append(ds, ds')) *)

      fun toAst(T{decs, ...}) =
	 Adec.make
	 (Adec.Local(Adec.make(Adec.SeqDec(List.map(decs, Dec.toAst))),
		     Adec.empty))

      val layout = Adec.layout o toAst

      fun size(T{decs = ds, ...}) =
	 let
	    val n = ref 0
	    fun inc() = n := 1 + !n

	    fun exp e =
	       (inc()
		; (case e of
		      Fn rs => rules rs
		    | Record r => Record.foreach(r, exp)
		    | App(e, e') => (exp e; exp e')
		    | Let(ds, e) => (decs ds; exp e)
		    | Constraint(e, _) => exp e
		    | Handle(e, rs) => (exp e; rules rs)
		    | Raise e => exp e
		    | _ => ()))
	    and rules rs = List.foreach(rs, exp o #2)
	    and dec d =
	       case d of
		  Val{exp = e, ...} => exp e
		| Fun{decs, ...} => List.foreach(decs, rules o #rules)
		| Exception _ => inc()
		| _ => ()
	    and decs ds = List.foreach(ds, dec)
	 in decs ds
	    ; !n
	 end
	 
      fun layoutStats p =
	 let open Layout
	 in seq[str "size = ", Int.layout(size p)]
	 end
   end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_STR_STRUCTS = 
   sig
      structure Ast: AST
      structure Tycon :
	 sig
	    type t
	    val newNoname: unit -> t
	    val exn: t
	    val layout: t -> Layout.t
	 end
      structure Con :
	 sig
	    type t
	    val fromAst: Ast.Con.t -> t
	    val layout: t -> Layout.t
	 end
      structure Type :
	 sig
	    datatype t =
	       Var of Ast.Tyvar.t
	     | Con of Tycon.t * t list
	     | Record of t Ast.SortedRecord.t
	    val var: Ast.Tyvar.t -> t
	    val record: t Ast.SortedRecord.t -> t
	    val arrow: t * t -> t
(* 	    val hom: t * {var: Ast.Tyvar.t -> 'a,
 * 			  con: Tycon.t * 'a list -> 'a,
 * 			  record: 'a Ast.SortedRecord.t -> 'a} -> 'a
 *)
	 end
      structure Scheme :
	 sig
	    datatype t = T of {tyvars: Ast.Tyvar.t list, ty: Type.t}
	    val ty: t -> Type.t
	    val apply: t * Type.t list -> Type.t
	    val layout: t -> Layout.t
	 end
   end

signature TYPE_STR = 
   sig
      include TYPE_STR_STRUCTS
      
      structure TypeFcn :
	 sig
	    datatype t =
	       Tycon of Tycon.t
	     | Fun of Scheme.t
	       
	    val funn: Scheme.t -> t

	    val isFun: t -> bool
	    val isTycon: t -> bool

	    val apply: t * Type.t list -> Type.t

	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      structure Cons :
	 sig
	    datatype t = T of {name: Ast.Con.t,
			       con: Con.t,
			       scheme: Scheme.t} list
	       
	    val empty: t
	    val isEmpty: t -> bool
	    val toList: t -> {name: Ast.Con.t,
			      con: Con.t,
			      scheme: Scheme.t} list
	    val layout: t -> Layout.t
	 end

      type t

      (* constructors *)
      val bogus: t
      val make: TypeFcn.t * Cons.t -> t
      val def: Scheme.t -> t
      val fromTypeFcn: TypeFcn.t -> t
      val data: Tycon.t * Cons.t -> t
      val tycon: Tycon.t -> t
	 
      (* destructors *)
      val dest: t -> TypeFcn.t * Cons.t
      val typeFcn: t -> TypeFcn.t
      val cons: t -> Cons.t
	 
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeStr(S: TYPE_STR_STRUCTS): TYPE_STR = 
struct

open S

local open Ast
in structure Tyvar = Tyvar
end

structure TypeFcn =
   struct
      datatype t =
	 Tycon of Tycon.t
       | Fun of Scheme.t

      val isFun =
	 fn Fun _ => true
	  | _ => false

      val isTycon =
	 fn Tycon _ => true
	  | _ => false
	 
      val equals = fn _ => Error.unimplemented "TypeFcn.equals"
	 
      fun layout t =
	 let open Layout
	 in case t of
	    Tycon c => Tycon.layout c
	  | Fun s => Scheme.layout s
	 end

      fun funn(s as Scheme.T{tyvars, ty}): t =
	 case ty of
	    Type.Con(tycon, tys) =>
	       if List.equals(tyvars, tys,
			 fn (a, Type.Var a') => Tyvar.equals(a, a')
			  | _ => false)
		  then Tycon tycon
	       else Fun s
	  | _ => Fun s

      fun apply(f: t, ts) =
	 case f of
	    Tycon c => Type.Con(c, ts)
	  | Fun s => Scheme.apply(s, ts)
   end

structure Cons =
   struct
      datatype t = T of {name: Ast.Con.t,
			 con: Con.t,
			 scheme: Scheme.t} list
	 
      val empty = T []

      fun isEmpty(T l) = List.isEmpty l

      val isEmpty =
	 fn T[] => true
	  | _ => false
	 
      fun toList(T l) = l
	 
      fun layout(T l) =
	 let open Layout
	 in List.layout(fn {name, con, scheme} =>
			record[("name", Ast.Con.layout name),
			       ("con", Con.layout con),
			       ("scheme", Scheme.layout scheme)]) l
	 end
   end

datatype t =
   Fun of TypeFcn.t
 | Datatype of {tycon: Tycon.t, cons: Cons.t}

val bogus =
   Fun(TypeFcn.Fun
       (Scheme.T{tyvars = [],
		 ty = Type.Var(Ast.Tyvar.newNoname{equality = false})}))
   
fun make(typeFcn, cons) =
   if Cons.isEmpty cons
      then Fun typeFcn
   else case typeFcn of
      TypeFcn.Tycon tycon =>
	 Datatype{tycon = tycon, cons = cons}
    | TypeFcn.Fun _ => Error.error "TypeStr.make: non-well-formed typeStr"

val dest = 
   fn Fun typeFcn => (typeFcn, Cons.empty)
    | Datatype{tycon, cons} => (TypeFcn.Tycon tycon, cons)

val typeFcn = #1 o dest
val cons = #2 o dest
   
fun def s = Fun(TypeFcn.funn s)
val fromTypeFcn = Fun
fun data(tycon, cons) = Datatype{tycon = tycon, cons = cons}

val tycon = Fun o TypeFcn.Tycon

fun layout typeStr =
   let open Layout
   in case typeStr of
      Fun f => TypeFcn.layout f
    | Datatype{tycon, cons} =>
	 record[("tycon", Tycon.layout tycon),
		("cons", Cons.layout cons)]
   end

val equals = fn _ => Error.unimplemented "TypeStructure.equals"

end
signature UNIQUE_ID =
   sig
      type t
      val new: unit -> t
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
functor UniqueId() :> UNIQUE_ID =
   struct
      type t = unit ref

      fun new(): t = ref()

      val equals = Ref.equals

      fun layout _ = Layout.empty
   end

functor IntUniqueId() :> UNIQUE_ID =
   struct
      type t = int

      val cur = ref 0

      fun new(): t = (IntRef.inc cur ; !cur)

      val equals = Int.equals

      val layout = Int.layout
   end

functor UnitUniqueId() :> UNIQUE_ID =
   struct
      open Unit

      fun new() = ()
   end

functor Env(S: ENV_STRUCTS): ENV =
struct

open S

datatype 'a t = T of (Domain.t * 'a) List.t

fun size(T l) = List.length l
   
fun eq d d' = Domain.equals(d, d')
fun member(d, ds) = List.exists(ds, eq d)
   
fun invariant(T drs) =
   let fun loop([], _) = true
	 | loop((d, _) :: drs, seen) =
	      not(member(d, seen)) andalso loop(drs, d :: seen)
   in loop(drs, [])
   end

fun domain(T drs) = List.map(drs, #1)
   
val fromList = T
fun toList(T l) = l
   
fun empty() = T(List.empty())

fun single(d, r) = T[(d, r)]

fun isEmpty(T l) = List.isEmpty l

fun singleton dr = T(List.single dr)

fun new(ds, f) = T(List.map(ds, fn d => (d, f d)))

fun map(T drs, f) = T(List.map(drs, fn (d, r) => (d, f r)))
fun mapi(T drs, f) = T(List.map(drs, fn (d, r) => (d, f(d, r))))

fun fold(T drs, b, f) = List.fold(drs, b, fn ((_, r), b) => f(r, b))
fun foldi(T drs, b, f) = List.fold(drs, b, fn ((d, r), b) => f(d, r, b))
   
fun equal d (d', _) = Domain.equals(d, d')

fun remove(T drs, d) = T(List.remove(drs, equal d))
   
fun extend(T drs, d, r) =
   T(List.cons((d, r), List.remove(drs, equal d)))

fun env + (T l) = List.fold(l, env, fn ((d, r), env) => extend(env, d, r))

fun plus es = List.fold(es, empty(), fn (e, accum) => accum + e)
   
fun peek(T l, d) =
   case List.peek(l, equal d) of
      NONE => NONE
    | SOME (_, r) => SOME r
fun lookup(env, d) = case peek(env, d) of
   SOME r => r
 | NONE => (Layout.output(Domain.layout d, Out.error) ;
	    Out.newline Out.error ;
	    Error.error "lookup")

fun restrict(env, ds) = new(ds, fn d => lookup(env, d))
(* is this faster?
fun restrict(T srs, ss) =
	 T(let fun loop(_, []) = []
		 | loop([], _) = Error.error "didn't find everything"
		 | loop((s, r) :: srs, ss) =
		   let fun loop'([], _) = loop(srs, ss)
			 | loop'(s' :: ss', ss) =
			      if Domain.equals(s, s')
				 then (s, r) :: (loop(srs, ss'@ss))
			      else loop'(ss', s' :: ss)
		   in loop'(ss, [])
		   end
	   in loop(srs, ss)
	   end)
*)

fun multiExtend(env, [], []) = env
  | multiExtend(env, d :: ds, r :: rs) = multiExtend(extend(env, d, r), ds, rs)
  | multiExtend _ = Error.error "multiExtend"

fun foreach(e, f) = List.foreach(toList e, f o #2)
fun foreachi(e, f) = List.foreach(toList e, f)

fun forall(e, f) = List.forall(toList e, f o #2)
fun foralli(e, f) = List.forall(toList e, f)
   
fun equals rangeEqual (e1, e2) =
   Int.equals(size e1, size e2)
   andalso foralli(e1, fn (d, r) =>
		   case peek(e2, d) of
		      NONE => false
		    | SOME r' => rangeEqual(r, r'))
   
fun layout layoutR (T ps) =
   let open Layout
   in seq[str "[",
	  align(List.map(ps, fn (d, r) =>
			 seq[Domain.layout d, str " -> ", layoutR r])),
	  str"]"]
   end

fun maybeLayout (name, layoutR) env =
   if isEmpty env then Layout.empty
   else let open Layout
	in seq[str name, str " = ", layout layoutR env]
	end

end

functor PolyEnv(S: ENV_STRUCTS): ENV = Env(S)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INTERFACE_STRUCTS = 
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      structure TypeStr: TYPE_STR
      sharing Ast = CoreML.Ast = TypeStr.Ast
      sharing CoreML.Tycon = TypeStr.Tycon
      sharing CoreML.Con = TypeStr.Con
      sharing CoreML.Type = TypeStr.Type
      sharing CoreML.Scheme = TypeStr.Scheme
   end

signature INTERFACE = 
   sig
      include INTERFACE_STRUCTS

      structure ShapeId: UNIQUE_ID

      structure FlexibleTypeStr :
	 sig
	    include TYPE_STR
            val fromStrict: TypeStr.t -> t
	 end
      sharing Ast = FlexibleTypeStr.Ast

      structure Status :
	 sig
	    datatype t = Var | Con | Exn
	    val layout: t -> Layout.t
	 end

      type interface
      type t = interface
	 
      structure Element :
	 sig
	    datatype t =
	       Val of {name: Ast.Vid.t,
		       scheme: CoreML.Scheme.t,
		       status: Status.t}
	     | Type of {name: Ast.Tycon.t,
			typeStr: TypeStr.t}
	     | Str of {name: Ast.Strid.t,
		       interface: interface}
	 end
      
      (* constructors *)
      val empty: t
      val bogus: t
      val vals: {name: Ast.Vid.t,
		  scheme: FlexibleTypeStr.Scheme.t,
		  status: Status.t} list -> t
      val types: {name: Ast.Tycon.t,
		   typeStr: FlexibleTypeStr.t} list -> t
      val strs: {name: Ast.Strid.t,
		  interface: t} list -> t
      val extendTycon: t * Ast.Tycon.t * FlexibleTypeStr.t -> t
      val cons: FlexibleTypeStr.Cons.t -> t
      val excons: FlexibleTypeStr.Cons.t -> t

      val shapeId: t -> ShapeId.t
	 
      val + : t * t -> t
	 
      val lookupLongtycon: t * Ast.Longtycon.t -> FlexibleTypeStr.t
      val peekLongtycon: t * Ast.Longtycon.t -> FlexibleTypeStr.t option

      (* elements(I, f).
       f is called on the flexible tycons to get the realization *)
      val fold: t * 'a * (Element.t * 'a -> 'a) -> 'a
	 
      (* rename all flexible tycons. *)
      val copy: t -> t

      (* make a copy, and instantiate longtycons *)
      val realize: t * (Ast.Longtycon.t -> TypeStr.TypeFcn.t) -> t
	 
      val shareType: t * Ast.Longtycon.t * Ast.Longtycon.t -> unit
      val share: t * Ast.Longstrid.t * Ast.Longstrid.t -> unit
      val wheres: t * (Ast.Longtycon.t * FlexibleTypeStr.TypeFcn.t) list -> unit

      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Interface(S: INTERFACE_STRUCTS): INTERFACE =
struct

open S

local open Ast
in structure Longtycon = Longtycon
   structure Longstrid = Longstrid
   structure Record = SortedRecord
   structure Strid = Strid
   structure Tyvar = Tyvar
end

local open CoreML
in structure Cscheme = Scheme
   structure Ctycon = Tycon
   structure Ctype = Type
end

structure CtypeStr = TypeStr
local open TypeStr
in structure CtypeFcn = TypeFcn
   structure Ccons = Cons
end

structure Set = DisjointSet

structure ShapeId = UniqueId()

structure Status =
   struct
      datatype t = Var | Con | Exn
      fun layout s =
	 Layout.str(case s of
		       Var => "Var"
		     | Con => "Con"
		     | Exn => "Exn")
   end

(* only needed for debugging *)
structure TyconId = IntUniqueId()
       
structure FlexibleTycon =
   struct
      structure TypeFcn =
	 struct
	    datatype t =
	       Tycon
	     | Fun
	     | Forced of CtypeFcn.t

	    fun layout f =
	       let open Layout
	       in case f of
		  Tycon => str "<flexible tycon>"
		| Fun => str "<flexible def>"
		| Forced f => paren(seq[str "forced ", CtypeFcn.layout f])
	       end

	    val strict =
	       fn Forced f => f
		| _ => Error.bug "impossible force of FlexibleTycon"
	 end
      
      datatype t = T of {copy: copy,
			 hasCons: bool,
			 typeFcn: TypeFcn.t,
			 id: TyconId.t} Set.t
      withtype copy = t option ref

      val equals = fn (T s, T s') => Set.equals(s, s')
	 
      fun layout(T s) =
	 let open Layout
	    val {hasCons, typeFcn, id, ...} = Set.value s
	 in record[("id", TyconId.layout id),
		   ("hasCons", Bool.layout hasCons),
		   ("typeFcn", TypeFcn.layout typeFcn)]
	 end

      fun setTypeFcn(T s, f: CtypeFcn.t): unit =
	 let val {copy, id, hasCons, ...} = Set.value s
	 in Set.setValue(s, {copy = copy, id = id, hasCons = hasCons,
			     typeFcn = TypeFcn.Forced f})
	 end

      fun new{hasCons: bool, typeFcn: TypeFcn.t}: t =
	 T(Set.singleton{copy = ref NONE,
			 hasCons = hasCons,
			 typeFcn = typeFcn,
			 id = TyconId.new()})

      fun newNoname() = new{hasCons = false, typeFcn = TypeFcn.Tycon}

      val bogus = newNoname()

      fun toTypeFcn(T s) = #typeFcn(Set.value s)

      val copies: copy list ref = ref []
	 
      fun copy(T s): t =
	 let val {copy, typeFcn, hasCons, ...} = Set.value s
	 in case !copy of
	    SOME c => c
	  | NONE => 
	       let val c = new{hasCons = hasCons,
			       typeFcn = typeFcn}
	       in ListRef.push(copies, copy) ;
		  copy := SOME c ;
		  c
	       end
	 end

      fun share(T s, reg, T s', reg'): unit =
	 let val {typeFcn, hasCons, id, ...} = Set.value s
	    val {typeFcn=typeFcn', hasCons=hasCons', ...} = Set.value s'
	    fun error reg =
	       Control.errorStr(reg, "type sharing on definitional type")
	 in case (typeFcn, typeFcn') of
	    (TypeFcn.Tycon, TypeFcn.Tycon) =>
	       (Set.union(s, s') ;
		Set.setValue(s, {copy = ref NONE, id = id,
				 hasCons = hasCons orelse hasCons',
				 typeFcn = TypeFcn.Tycon}))
	  | (TypeFcn.Fun, _) => error reg
	  | (_, TypeFcn.Fun) => error reg'
	  | _ => Error.bug "type sharing on Forced typeFcn"
	 end

      fun setAbbrev(T s, isTycon: bool, reg): unit =
	 let val {copy, id, hasCons, typeFcn} = Set.value s
	 in if hasCons andalso not isTycon
	       then Control.errorStr(reg, "type abbreviation on datatype")
	    else 
	       case typeFcn of
		  TypeFcn.Tycon =>
		     Set.setValue(s, {copy = copy,
				      id = id,
				      hasCons = hasCons,
				      typeFcn = if isTycon
						   then TypeFcn.Tycon
						else TypeFcn.Fun})
		| TypeFcn.Fun =>
		     Control.errorStr
		     (reg, "type abbreviation on type abbreviation")
		| TypeFcn.Forced _ =>
		     Control.errorStr
		     (reg, "type abbreviation on forced flexible tycon")
	 end

      fun strict(T s): CtypeFcn.t =
	 TypeFcn.strict(#typeFcn(Set.value s))
   end

structure Tycon =
   struct
      datatype t =
	 Flexible of FlexibleTycon.t
       | Rigid of Ctycon.t

      val arrow = Rigid Ctycon.arrow

      val newNoname = Flexible o FlexibleTycon.newNoname

      val toAst =
	 fn Flexible c =>
	      Ast.Tycon.fromString(Layout.toString(FlexibleTycon.layout c))
	  | Rigid c => Ctycon.toAst c
   
      val layout =
	 fn Flexible c => FlexibleTycon.layout c
	  | Rigid c => Ctycon.layout c

      fun copy(t: t): t =
	 case t of
	    Flexible c => Flexible(FlexibleTycon.copy c)
	  | Rigid _ => t

      fun strict(t: t): CtypeFcn.t =
	 case t of
	    Flexible c => FlexibleTycon.strict c
	  | Rigid c => CtypeFcn.Tycon c

      val fromStrict = Rigid

      val exn = fromStrict Ctycon.exn
   end

structure Type =
   struct
      datatype t =
	 Var of Tyvar.t
       | Con of Tycon.t * t list
       | Record of t Record.t

      fun hom(t, {var, con, record}) =
	 let
	    val rec loop =
	       fn Var a => var a
		| Con(c, ts) => con(c, List.map(ts, loop))
		| Record r => record(Record.map(r, loop))
	 in loop t
	 end
	       
      fun toAst t =
	 hom(t, {var = Ast.Type.var,
		 con = fn (c, ts) => Ast.Type.con(Tycon.toAst c, ts),
	         record = Ast.Type.record})

      val layout = Ast.Type.layout o toAst

(*      val traceStrict = Trace.trace("strict", layout, Ctype.layout) *)
	 
      fun strict t =
(*	 traceStrict*)
	 (fn t =>
	 case t of
	    Var a => Ctype.Var a
	  | Con(c, ts) => CtypeFcn.apply(Tycon.strict c, List.map(ts, strict))
	  | Record r => Ctype.Record(Record.map(r, strict))) t

      fun fromStrict t =
	 case t of
	    Ctype.Var a => Var a
	  | Ctype.Con(c, ts) => Con(Tycon.fromStrict c, List.map(ts, fromStrict))
	  | Ctype.Record r => Record(Record.map(r, fromStrict))
		 
      fun copy(t: t): t =
	 let
	    fun loop t =
	       case t of
		  Var _ => t
		| Con(c, ts) => Con(Tycon.copy c, List.map(ts, loop))
		| Record r => Record(Record.map(r, loop))
	 in loop t
	 end
	 
      val var = Var
      val record = Record
      fun arrow(t1, t2) = Con(Tycon.arrow, [t1, t2])

      local
	 structure Env = Env(structure Domain = Tyvar)
      in
	 fun substitute(t: t, sub: (Tyvar.t * t) list): t =
	    let val env = Env.fromList sub
	       fun loop t =
		  case t of
		     Var a => Env.lookup(env, a)
		   | Con(c, ts) => Con(c, List.map(ts, loop))
		   | Record r => Record(Record.map(r, loop))
	    in loop t
	    end
      end
   end

structure Scheme = GenericScheme(structure Tyvar = Tyvar
				 structure Type = Type)
structure Scheme =
   struct
      fun strict(Scheme.T{tyvars, ty}) =
	 Cscheme.T{tyvars = tyvars,
		   ty = Type.strict ty}

(*      val strict = Trace.trace("strict", Scheme.layout, Cscheme.layout) strict *)

      fun fromStrict(Cscheme.T{tyvars, ty}) =
	 Scheme.T{tyvars = tyvars, ty = Type.fromStrict ty}

      open Scheme

      fun copy(T{tyvars, ty}): t =
	 T{tyvars = tyvars, ty = Type.copy ty}
   end

structure TypeStr = TypeStr(structure Ast = Ast
			    structure Tycon = Tycon
			    structure Con = CoreML.Con
			    structure Type = Type
			    structure Scheme = Scheme)

structure TypeFcn =
   struct
      open TypeStr.TypeFcn

      fun copy(f: t): t =
	 case f of
	    Tycon c => Tycon(Tycon.copy c)
	  | Fun s => Fun(Scheme.copy s)

      fun strict(f: t) =
	 case f of
	    Tycon c => Tycon.strict c
	  | Fun s => CtypeFcn.Fun(Scheme.strict s)

      fun fromStrict f: t =
	 case f of
	    CtypeFcn.Tycon c => Tycon(Tycon.fromStrict c)
	  | CtypeFcn.Fun s => Fun(Scheme.fromStrict s)
   end

structure Cons =
   struct
      open TypeStr.Cons

      fun copy(T l): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.copy scheme}))

      fun strict(T l): Ccons.t =
	 Ccons.T(List.map(l, fn {name, con, scheme} =>
			  {name = name, con = con,
			   scheme = Scheme.strict scheme}))

      fun fromStrict(Ccons.T l): t =
	 T(List.map(l, fn {name, con, scheme} =>
		    {name = name, con = con,
		     scheme = Scheme.fromStrict scheme}))
end

structure TypeStr =
   struct
      fun toFlexible(s: TypeStr.t, reg, msg): FlexibleTycon.t =
	 case TypeStr.typeFcn s of
	    TypeFcn.Tycon(Tycon.Flexible c) => c
	  | _ => (Control.errorStr(reg, msg) ;
		  FlexibleTycon.bogus)

      fun copy(s: TypeStr.t): TypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in TypeStr.make(TypeFcn.copy f, Cons.copy cs)
	 end

      fun strict(s: TypeStr.t): CtypeStr.t =
	 let val (f, cs) = TypeStr.dest s
	 in CtypeStr.make(TypeFcn.strict f, Cons.strict cs)
	 end

(*      val strict = Trace.trace("strict", TypeStr.layout, CtypeStr.layout) strict *)

      fun fromStrict(s: CtypeStr.t): TypeStr.t =
	 let val (f, cs) = CtypeStr.dest s
	 in TypeStr.make(TypeFcn.fromStrict f, Cons.fromStrict cs)
	 end
			 
      open TypeStr

      fun share(s: t, reg, s': t, reg'): unit =
	 let val msg = "type sharing on rigid tycon"
	 in FlexibleTycon.share(toFlexible(s, reg, msg), reg,
				toFlexible(s', reg', msg), reg')
	 end
   end
structure FlexibleTypeStr = TypeStr
	   
(*---------------------------------------------------*)
(*                   Main Datatype                   *)
(*---------------------------------------------------*)
(* Invariant: only ever union two envs if they have the same shape. *)
(* The shape of interface is the set of longtycons that are accessible in it. *)

datatype t = T of {copy: copy,
		   shapeId: ShapeId.t,
		   elements: element list,
		   wheres: (FlexibleTycon.t * TypeFcn.t) list ref} Set.t
and element =
   Val of {name: Ast.Vid.t,
	   scheme: Scheme.t,
	   status: Status.t}
  | Type of {name: Ast.Tycon.t,
	     typeStr: TypeStr.t}
  | Str of {name: Ast.Strid.t,
	    interface: t}
withtype copy = t option ref

type interface = t

fun equals(T s, T s') = Set.equals(s, s')

local open Layout
in
   fun layout(T s) =
      let val {elements, wheres, ...} = Set.value s
      in record[("wheres", list(List.map(!wheres, fn (c, f) =>
					 tuple[FlexibleTycon.layout c,
					       TypeFcn.layout f]))),
		("elements",
		 list(List.map(elements, layoutElement)))]
      end
   and layoutElement(e: element) =
      let
	 val (lhs, rhs) =
	    case e of
	       Val{name, scheme, status} =>
		  (Ast.Vid.layout name,
		   tuple[Status.layout status,
			 Scheme.layout scheme])
	     | Type{name, typeStr} =>
		  (Ast.Tycon.layout name,
		   TypeStr.layout typeStr)
	     | Str{name, interface} =>
		  (Ast.Strid.layout name, layout interface)
      in seq[lhs, str " -> ", rhs]
      end
end

(*val traceCheckType = Trace.trace("checkType", Type.layout, Unit.layout)
 *val traceCheckInterface = Trace.trace("checkInterface", layout, Unit.layout)
 *)
fun explicit elements: t =
   T(Set.singleton{copy = ref NONE,
		   shapeId = ShapeId.new(),
		   elements = elements,
		   wheres = ref []})

val empty = explicit []

val bogus = empty

fun vals l = explicit(List.map(l, Val))
fun strs l = explicit(List.map(l, Str))
fun types l = explicit(List.map(l, Type))
   
local
   fun make status (Cons.T l) =
      explicit(List.map(l, fn {name, scheme, ...} =>
			Val{name = Ast.Vid.fromCon name,
			    scheme = scheme,
			    status = status}))
in val cons = make Status.Con
   val excons = make Status.Exn
end

fun elements(T s): element list = #elements(Set.value s)
fun shapeId(T s) = #shapeId(Set.value s)
   
fun extendTycon(I, tycon, typeStr) =
   explicit(elements I @ [Type{name = tycon, typeStr = typeStr}])

val op + = fn (I, I') => explicit(elements I @ elements I')

fun peekTyconElements(elements: element list, tycon): TypeStr.t option =
   case List.peek(elements,
		  fn Type{name, ...} => Ast.Tycon.equals(tycon,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Type{typeStr, ...}) => SOME typeStr
    | _ => Error.bug "peekTyconElements"
	 
fun peekStridElements(elements, strid): t option =
   case List.peek(elements,
		  fn Str{name, ...} => Strid.equals(strid,name)
		   | _ => false) of
      NONE => NONE
    | SOME(Str{interface, ...}) => SOME interface
    | _ => Error.bug "peekStridElements"

fun peekStrid(I: t, strid: Ast.Strid.t): t option =
   peekStridElements(elements I, strid)

fun peekStrids(I: t, strids: Ast.Strid.t list): t option =
   case strids of
      [] => SOME I
    | strid :: strids =>
	 case peekStrid(I, strid) of
	    NONE => NONE
	  | SOME I => peekStrids(I, strids)

fun lookupLongstrid(I: t, s: Longstrid.t): t =
   let val (strids, strid) = Longstrid.split s
   in case peekStrids(I, strids @ [strid]) of
      NONE => (Longstrid.unbound s ; bogus)
    | SOME I => I
   end

fun peekLongtycon(I: t, c: Longtycon.t): TypeStr.t option =
   let val (strids, c) = Longtycon.split c
   in case peekStrids(I, strids) of
      NONE => NONE
    | SOME I => peekTyconElements(elements I, c)
   end

fun lookupLongtycon(I: t, c: Longtycon.t): TypeStr.t =
   case peekLongtycon(I, c) of
      SOME typeStr => typeStr
    | NONE => (Longtycon.unbound c ;
	       TypeStr.bogus)

fun shareType(I: t, c: Longtycon.t, c': Longtycon.t) =
   TypeStr.share(lookupLongtycon(I, c), Longtycon.region c,
		 lookupLongtycon(I, c'), Longtycon.region c')

fun sameShape(m, m') = ShapeId.equals(shapeId m, shapeId m')

fun share(I as T s, reg, I' as T s', reg'): unit = 
   if Set.equals(s, s') then ()
   else
      if sameShape(I, I')
	 then
	    let
	       fun loop(T s, T s'): unit =
		  if Set.equals(s, s') then ()
		  else 
		     let
			val {elements=es, ...} = Set.value s
			val {elements=es', ...} = Set.value s'
		     in Set.union(s, s') ;
			List.foreach2
			(es, es',
			 fn (Type{typeStr=s, ...}, Type{typeStr=s', ...}) =>
			    TypeStr.share(s, reg, s', reg')
			  | (Str{interface=I, ...}, Str{interface=I', ...}) =>
			       loop(I, I')
			  | _ => ())
		     end
	    in loop(I, I')
	    end
      else (* different shapes -- need to share pointwise *)
	 let
	    val es = elements I
	    val es' = elements I'
	 in List.foreach
	    (es,
	     fn Type{name, typeStr = s} =>
	           (case peekTyconElements(es',name) of
		       NONE => ()
		     | SOME s' => TypeStr.share(s, reg, s', reg'))
	      | Str{name, interface = I} =>
		   (case peekStridElements(es', name) of
		       NONE => ()
		     | SOME I' => share(I, reg, I', reg'))
	      | _ => ())
	 end

val share =
   fn (m, s: Longstrid.t, s': Longstrid.t) =>
   share(lookupLongstrid(m, s),
	 Longstrid.region s,
	 lookupLongstrid(m, s'),
	 Longstrid.region s')

fun wheres(I as T s, l: (Longtycon.t * TypeFcn.t) list): unit =
   let val {wheres, ...} = Set.value s
   in List.foreach(l, fn (c, f) =>
		   let
		      val reg = Longtycon.region c
		      val flex =
			 TypeStr.toFlexible
			 (lookupLongtycon(I, c), reg, "where type on rigid tycon")
		   in FlexibleTycon.setAbbrev(flex, TypeFcn.isTycon f, reg) ;
		      ListRef.push(wheres, (flex, f))
		   end)
   end

structure Element =
   struct
      type interface = t
      datatype t =
	 Val of {name: Ast.Vid.t,
		 scheme: Cscheme.t,
		 status: Status.t}
       | Type of {name: Ast.Tycon.t,
		  typeStr: CtypeStr.t}
       | Str of {name: Ast.Strid.t,
		 interface: interface}
   end

fun copyAndRealize(I: t, getTypeFcnOpt): t =
   let
      (* keep track of all nodes that have forward pointers to copies, so
       that we can gc them when done *)
      val copies: copy list ref = ref []
	 
      fun loop(T s, strids: Ast.Strid.t list): t =
	 let val {copy, shapeId, elements, wheres, ...} = Set.value s
	 in case !copy of
	    SOME I => I
	  | NONE =>
	       let
		  val elements =
		     List.map
		     (elements,
		      fn Val{name, scheme, status} =>
		      Val{name = name, status = status,
			  scheme = Scheme.copy scheme}
		       | Type{name, typeStr} =>
			    let val typeStr = TypeStr.copy typeStr
			    in (case (TypeStr.typeFcn typeStr, getTypeFcnOpt) of
				   (TypeFcn.Tycon(Tycon.Flexible c), SOME f) =>
				      FlexibleTycon.setTypeFcn
				      (c, f(Longtycon.long(strids,name)))
				 | _ => ()) ;
			       Type{name = name,
				    typeStr = typeStr}
			    end
		       | Str{name, interface} =>
			    Str{name = name,
				interface = loop(interface, strids @ [name])})
		     
		  val wheres =
		     List.map(!wheres, fn (c, f) =>
			      let val c = FlexibleTycon.copy c
				 val f = TypeFcn.copy f
			      in (case getTypeFcnOpt of
				     NONE => ()
				   | SOME _ =>
					FlexibleTycon.setTypeFcn
					(c, TypeFcn.strict f)) ;
				 (c, f)
			      end)

		  val I = T(Set.singleton{copy = ref NONE,
					  shapeId = shapeId,
					  elements = elements,
					  wheres = ref wheres})
				     
	       in ListRef.push(copies, copy) ;
		  copy := SOME I ;
		  I
	       end
	 end
      
      val I = loop(I, [])

      fun clear copies =
	 (List.foreach(!copies, fn copy => copy := NONE) ;
	  copies := [])
	 
   in clear copies ;
      clear FlexibleTycon.copies ;
      I
   end

fun copy I = copyAndRealize(I, NONE)
fun realize(I, f) = copyAndRealize(I, SOME f)

(*val realize = Trace.trace2("realize", layout, Layout.ignore, layout) realize *)

val 'a fold =
   fn (T s, b: 'a, f: Element.t * 'a -> 'a) =>
   let val {elements, ...} = Set.value s
   in List.fold
      (elements, b, fn (elt, b) =>
       let
	  val elt =
	     case elt of
		Val{name, scheme, status} =>
		   Element.Val{name = name, status = status,
			       scheme = Scheme.strict scheme}
	      | Str r => Element.Str r
	      | Type{name, typeStr} =>
		   Element.Type{name = name,
				typeStr = TypeStr.strict typeStr}
       in f(elt, b)
       end)
   end

structure TypeStr = CtypeStr

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_ATOMS_STRUCTS = 
   sig
      structure Ast: AST
      structure TypeStr: TYPE_STR
      sharing Ast = TypeStr.Ast
	 
      structure Decs:
	 sig
	    type t

	    val empty: t
	    val datatypes :
	       {
		tycon: TypeStr.Tycon.t,
		tyvars: Ast.Tyvar.t list,
		cons: {
		       con: TypeStr.Con.t,
		       arg: TypeStr.Type.t option
		      } list
	       } list -> t
	 end
      
      structure Env:
	 sig
	    type t
	    val empty: t
	    val extendTycon: t * Ast.Tycon.t * TypeStr.t -> t
	    val + : t * t -> t
	    val cons: TypeStr.Cons.t -> t
	    val lookupLongtycon: t * Ast.Longtycon.t -> TypeStr.t
	    val peekLongtycon: t * Ast.Longtycon.t -> TypeStr.t option
	 end
   end

signature ELABORATE_ATOMS = 
   sig
      include ELABORATE_ATOMS_STRUCTS

      structure Lookup:
	 sig
	    type t
	    val fromFun: (Ast.Longtycon.t -> TypeStr.t) -> t
	    val fromEnv: Env.t -> t
	    val plusEnv: t * Env.t -> t
	    val lookup: t * Ast.Longtycon.t -> TypeStr.t
	 end

      val elaborateType: Ast.Type.t * Lookup.t -> TypeStr.Scheme.t

      val elaborateTypeOpt: Ast.Type.t option * Env.t -> TypeStr.Type.t option

      val elaborateTypBind: Ast.TypBind.t * Lookup.t -> Env.t
	 
      val elaborateScheme:
 	 Ast.Tyvar.t list * Ast.Type.t * Lookup.t
	 -> TypeStr.Scheme.t * Ast.Tyvar.t list

      val elaborateClosedScheme:
	 Ast.Tyvar.t list * Ast.Type.t * Lookup.t -> TypeStr.Scheme.t

      val elaborateCons:
	 {cons: (Ast.Con.t * Ast.Type.t option) list,
	  tyvars: Ast.Tyvar.t list,
	  tycon: TypeStr.Tycon.t,
	  resultType: Ast.Type.t,
	  lookup: Lookup.t} -> TypeStr.Cons.t

      val elaborateDatBind:
	 {datBind: Ast.DatBind.t,
	  lookup: Lookup.t,
	  newTycon: Ast.Tycon.t -> TypeStr.Tycon.t}
	 -> {VE: Env.t, TE: Env.t, decs: Decs.t}

      val elaborateDatatypeRhs:
	 {rhs: Ast.DatatypeRhs.t,
	  lookup: Lookup.t,
	  newTycon: Ast.Tycon.t -> TypeStr.Tycon.t} -> Decs.t * Env.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateAtoms(S: ELABORATE_ATOMS_STRUCTS): ELABORATE_ATOMS = 
struct

open S

local open Ast
in structure Atype = Type
   structure DatatypeRhs = DatatypeRhs
   structure DatBind = DatBind
   structure Longtycon = Longtycon
   structure SortedRecord = SortedRecord
   structure TypBind = TypBind
   structure Tyvar = Tyvar
end

local open TypeStr
in structure Con = Con
   structure Cons = Cons
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
   structure TypeFcn = TypeFcn
end

structure Lookup =
   struct
      type t = Longtycon.t -> TypeStr.t

      fun fromFun f = f
      fun fromEnv(E: Env.t) longtycon = Env.lookupLongtycon(E, longtycon)
      fun plusEnv(lookup: t, E: Env.t): t =
	 fn longtycon =>
	 case Env.peekLongtycon(E, longtycon) of
	    SOME typeFcn => typeFcn
	  | NONE => lookup longtycon
      fun lookup(l: t, c: Longtycon.t) = l c
   end

(* make sure to replace all tyvars that have the same name with a single tyvar *)
fun elaborateType(ty: Atype.t, lookup: Lookup.t): Scheme.t =
   let 
      fun loop(ty: Atype.t, accum: Tyvar.t list) =
	 case Atype.node ty of
	    Atype.Var a => (* rule 44 *)
	       let
		  fun loop tyvars =
		     case tyvars of
			[] => (* new type variable, add it to the accum *)
			   (Type.var a, a :: accum)
		      | a' :: tyvars =>
			   if Tyvar.sameName(a, a')
			      then (Type.var a', accum)
			   else loop tyvars
	       in loop accum
	       end
	  | Atype.Con(c, ts) => (* rules 46, 47 *)
	       let val (ts, accum) = loops(ts, accum)
		  fun normal() = TypeFcn.apply(TypeStr.typeFcn(lookup c), ts)
	       in (case (Ast.Longtycon.split c, ts) of
		      (([], c), [t1, t2]) => if Ast.Tycon.equals(c, Ast.Tycon.arrow)
					     then Type.arrow(t1, t2)
					  else normal()
		    | _ => normal(),
		   accum)
	       end
	  | Atype.Record r => (* rules 45, 49 *)
	       let val (fs, ts) = SortedRecord.unzip r
		  val (ts, accum) = loops(ts, accum)
	       in (Type.record(SortedRecord.zip(fs, ts)), accum)
	       end

      and loops(ts, accum) =
	 List.fold(List.reverse ts, ([], accum), fn (t, (ts, accum)) =>
		   let val (t, accum) = loop(t, accum)
		   in (t :: ts, accum)
		   end)
	 
      val (ty, tyvars) = loop(ty, [])
	 
   in Scheme.T{tyvars = tyvars, ty = ty}
   end

(* Returns a scheme, plus the type variables that occured in the ty but
 * not in the tyvars.
 *)
fun elaborateScheme(tyvars, ty: Atype.t, lookup: Lookup.t)
  : Scheme.t * Tyvar.t list =
   let val Scheme.T{tyvars = tyvars', ty} = elaborateType(ty, lookup)
      (* Replace each tyvar with the corresponding tyvar'.
       * Keep track of any tyvars' that are left over.
       *)
      val (tyvars, tyvars') =
	 List.fold
	 (List.reverse tyvars, ([], tyvars'), fn (a, (tyvars, tyvars')) =>
	  let
	     fun loop(tyvars', remaining) =
		case tyvars' of
		   [] => (a, remaining)
		 | a' :: tyvars' =>
		      if Tyvar.sameName(a, a')
			 then (a', remaining @ tyvars')
		      else loop(tyvars', a' :: remaining)
	     val (a, tyvars') = loop(tyvars', [])
	  in (a :: tyvars, tyvars')
	  end)
   in (Scheme.T{tyvars = tyvars, ty = ty}, tyvars')
   end

fun elaborateClosedScheme arg: Scheme.t =
   let val (scheme, tyvars) = elaborateScheme arg
   in (case tyvars of
	  [] => ()
	| tyvar :: _ =>
	     Control.error
	     (Tyvar.region tyvar,
	      let open Layout
	      in seq[str "unbound type variables: ",
		     seq(separate(List.map(tyvars, Tyvar.layout), " "))]
	      end)) ;
      scheme
   end

fun elaborateTypeOpt(ty: Atype.t option, E: Env.t): Type.t option =
   case ty of
      NONE => NONE
    | SOME ty => SOME(Scheme.ty(elaborateType(ty, Lookup.fromEnv E)))

	    
fun elaborateTypBind(typBind, lookup: Lookup.t): Env.t =
   let val TypBind.T types = TypBind.node typBind
   in List.fold
      (types, Env.empty, fn ({tyvars, tycon, def}, E) =>
       let val tystr = TypeStr.def(elaborateClosedScheme(tyvars, def, lookup))
       in Env.extendTycon(E, tycon, tystr)
       end)
   end

fun elaborateCons{cons, tyvars, tycon: Tycon.t,
		  resultType, lookup}: Cons.t =
   let
      val cons =
	 List.map
	 (cons, fn (name, to) =>
	  let val con = Con.fromAst name
	     val scheme =
		elaborateClosedScheme(tyvars,
				      case to of
					 NONE => resultType
				       | SOME t => Atype.arrow(t, resultType),
				      lookup)
	  in (*CoreML.setConTycon(con, tycon) ;*)
	     ({name = name, con = con, scheme = scheme})
	  end)
	 
   in (*CoreML.setTyconCons(tycon, List.map(cons, #con)) ;*)
      Cons.T cons
   end
		
fun elaborateDatBind{datBind,
		     lookup: Lookup.t,
		     newTycon: Ast.Tycon.t -> Tycon.t}
  : {VE: Env.t, TE: Env.t, decs: Decs.t} =
   (* rules 28, 29, 81, 82 *)
   let
      val DatBind.T{datatypes, withtypes} = DatBind.node datBind
      (* build enough of an env so that that the withtypes
       * and the constructor argument types can be evaluated
       *)
      val (E, datatypes) =
	 List.fold
	 (datatypes, (Env.empty, []),
	  fn ({tyvars, tycon = name, cons}, (E, datatypes))=>
	  let val tycon = newTycon name
	  in (Env.extendTycon(E, name, TypeStr.tycon tycon),
	      {name = name, tycon = tycon: Tycon.t, tyvars = tyvars, cons = cons}
	      :: datatypes)
	  end)
      val lookup = Lookup.plusEnv(lookup, E)
      (* elaborate the withtypes *)
      val TE = elaborateTypBind(withtypes, lookup)
      val E = Env.+(E, TE)
      val lookup = Lookup.plusEnv(lookup, E)
      (* elaborate the datatypes, this time including the constructors *)
      val (VE, TE, datatypes) =
	 List.fold
	 (datatypes, (Env.empty, TE, []),
	  fn ({name, tyvars, tycon, cons}, (VE, TE, datatypes)) =>
	  let
	     val elabCons as Cons.T l =
		elaborateCons
		{cons = cons, tyvars = tyvars, tycon = tycon, lookup = lookup,
		 resultType = Atype.con(name, List.map(tyvars, Atype.var))}
	     val tyStr = TypeStr.data(tycon, elabCons)
	  in (Env.+(VE, Env.cons elabCons),
	      Env.extendTycon(TE, name, tyStr),
	      {tyvars = tyvars,
	       tycon = tycon,
	       cons =
	       List.map2(cons, l, fn ((_, to), {con, ...}) =>
			 {con = con,
			  arg = (case to of
				    NONE => NONE
				  | SOME t => SOME(Scheme.ty(elaborateType(t, lookup))))})}
	      :: datatypes)
	  end)
   in {VE = VE, TE = TE, decs = Decs.datatypes datatypes}
   end

fun elaborateDatatypeRhs{rhs, lookup: Lookup.t, newTycon}: Decs.t * Env.t =
   case DatatypeRhs.node rhs of
      DatatypeRhs.DatBind datBind => (* rule 17 *)
	 let val {VE, TE, decs} = elaborateDatBind{datBind = datBind,
						   lookup = lookup,
						   newTycon = newTycon}
	 in (decs, Env.+(TE, VE))
	 end
    | DatatypeRhs.Repl{lhs, rhs} => (* rule 18 *)
	 let val tyStr = lookup rhs
	    val E = Env.extendTycon(Env.empty, lhs, tyStr)
	 in (Decs.empty, Env.+(E, Env.cons(TypeStr.cons tyStr)))
	 end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_ENV_STRUCTS =
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      structure Interface: INTERFACE
      sharing Ast = CoreML.Ast
      sharing CoreML = Interface.CoreML
      structure TypeStr: TYPE_STR
      sharing TypeStr = Interface.TypeStr
   end

signature ELABORATE_ENV =
   sig
      include ELABORATE_ENV_STRUCTS

      (* the value of a vid.  This is used to distinguish between vids whose
       * status cannot be determined at parse time
       *)
      structure Vid :
	 sig
	    datatype t =
	       Var of CoreML.Var.t
	     | ConAsVar of CoreML.Con.t (* a constructor, but it has status
					 * of a variable
					 *)
	     | Con of CoreML.Con.t
	     | Exn of CoreML.Con.t
	     | Prim of CoreML.Prim.t

	    val deVar: t -> CoreML.Var.t option
	    val deCon: t -> CoreML.Con.t option
	       
	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      type t

      val empty: t
	 
      (* environment with all the primitive (data)types and constructors *)
      val prim: unit -> t * CoreML.Constructors.t
	 
      (* overlay the bindings in the second arg on top of the first *)
      val + : t * t -> t
	 
      (* overlay a sequence -- later ones have precedence *)
      val plus: t list -> t
	 
      (* section 4.9 p. 20.  remove constructors from tyStrs *)
      val abs: t -> t
	 
      (* extract only the fixity portion of the env and return true if
       the argument vid is in the env and is infix.  Used for Ast layout. *)
      val toIsInfix: t -> Ast.Vid.t -> bool
	 
      (* filters out only those bindings in the env that also appear in
       the interface.  proceeds recursively on substructures *)
      val cut: {env: t,
		interface: Interface.t,
		opaque: bool,
		region: Region.t} -> t
	 
      (* Change all the variables to primitives, and make sure
       * everything has a type scheme.
       * Also returns an Ast that should be prepended to any program
       * that uses the primitives in order to patch the record types.
       *)
      val identify: t -> t * Ast.Program.t * CoreML.Constructors.t
	 
      (* This is extensionally the identity function.  But consolidates all of
       * the env into a single hash table so that things are faster.
       *)
      val consolidate: t -> t

      (* add a binding for the "=" primitive *)
      val addEquals: t -> t
	 
      (* constructors *)
      val fixs: (Ast.Vid.t * Ast.Fixity.t) list -> t
      val strs: (Ast.Strid.t * t) list -> t
      val vars: (Ast.Var.t * CoreML.Var.t * CoreML.Scheme.t option) list -> t
      val cons: TypeStr.Cons.t -> t
      val types: (Ast.Tycon.t * TypeStr.t) list -> t
	 
      (* add a new binding to an env *)
      val extendVar: t * Ast.Var.t * CoreML.Var.t -> t
      val extendCon: t * Ast.Con.t * CoreML.Con.t -> t
      val extendExn: t * Ast.Con.t * CoreML.Con.t -> t
      val extendPrim: t * Ast.Var.t * CoreML.Prim.t -> t
      val extendTycon: t * Ast.Tycon.t * TypeStr.t -> t
      val extendStrid: t * Ast.Strid.t * t -> t
	 
      (* check if a binding is in an env *)
      val peekLongvid: t * Ast.Longvid.t -> Vid.t option
      val peekLongvar: t * Ast.Longvar.t -> CoreML.Var.t option
      val peekLongcon: t * Ast.Longcon.t -> CoreML.Con.t option
      val peekLongtycon: t * Ast.Longtycon.t -> TypeStr.t option
      val peekLongstrid: t * Ast.Longstrid.t -> t option
      val peekVar: t * Ast.Var.t -> CoreML.Var.t option

      (* lookup a binding in an env *)
      val lookupFixity: t * Ast.Vid.t -> Ast.Fixity.t
      val lookupLongvid: t * Ast.Longvid.t -> Vid.t
      val lookupLongvar: t * Ast.Longvar.t -> CoreML.Var.t
      val lookupLongcon: t * Ast.Longcon.t -> CoreML.Con.t
      val lookupLongtycon: t * Ast.Longtycon.t -> TypeStr.t
      val lookupLongstrid: t * Ast.Longstrid.t -> t

      val layout: t -> Layout.t
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature DECS_STRUCTS =
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      sharing Ast = CoreML.Ast
   end

signature DECS =
   sig
      include DECS_STRUCTS
      
      type dec = CoreML.Dec.t
	 
      type t

      (* empty list *)
      val empty: t

      (* primitive datatype and exception declarations *)
(*      val prim: t*)

      (* list with one dec *)
      val single: dec -> t

      (* add a dec to the end of the list *)
      val add: t * dec -> t

      (* append two lists *)
      val append: t * t -> t

      (* append a lot of lists *)
      val appends: t list -> t

      (* conversions from/to normal lists *)
      val fromList: dec list -> t
      val toList: t -> dec list

      val toAst: t -> Ast.Dec.t
      val layout: t -> Layout.t
   end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature BASIS_STRUCTS = 
   sig
      structure Ast: AST
      structure Decs: DECS
      structure Env: ELABORATE_ENV
      structure Interface: INTERFACE
      sharing Ast = Decs.Ast = Env.Ast = Interface.Ast
      sharing Decs.CoreML = Env.CoreML = Interface.CoreML
   end

signature BASIS = 
   sig
      include BASIS_STRUCTS

      (* the value of a functor *)
      structure FunctorClosure :
	 sig
	    type basis
	    datatype t = T of {arg: Ast.Strid.t,
			       argSig: Interface.t,
			       body: Ast.Strexp.t,
			       basis: basis,     (* for free variables *)
			       counter: int ref} (* the number of applications of me *)

	    val layout: t -> Layout.t
	 end

      type t = FunctorClosure.basis

      val empty: t
      val prim: unit -> t * Decs.CoreML.Constructors.t
	 
      val + : t * t -> t
      val toEnv: t -> Env.t
      val fromEnv: Env.t -> t
      val plusEnv: t * Env.t -> t
      val replaceEnv: t * Env.t -> t
	 
      val consolidate: t -> t
	 
      val extendStrid: t * Ast.Strid.t * Env.t -> t
      val extendSigid: t * Ast.Sigid.t * Interface.t -> t
      val extendFctid: t * Ast.Fctid.t * FunctorClosure.t -> t
	 
      val lookupLongstrid: t * Ast.Longstrid.t -> Env.t
      val lookupSigid: t * Ast.Sigid.t -> Interface.t
      val lookupFctid: t * Ast.Fctid.t -> FunctorClosure.t
	 
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_SIGEXP_STRUCTS = 
   sig
      structure Ast: AST
      structure Basis: BASIS
      sharing Ast = Basis.Ast
      structure Interface: INTERFACE
      sharing Interface = Basis.Interface
      sharing Basis.Env.TypeStr = Interface.TypeStr
   end

signature ELABORATE_SIGEXP = 
   sig
      include ELABORATE_SIGEXP_STRUCTS

      val elaborateSigexp: Ast.Sigexp.t * Basis.t -> Interface.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateSigexp(S: ELABORATE_SIGEXP_STRUCTS): ELABORATE_SIGEXP = 
struct

open S

local open Ast
in structure Atype = Type
   structure DatBind = DatBind
   structure DatatypeRhs = DatatypeRhs
   structure Equation = Equation
   structure Longstrid = Longstrid
   structure Longtycon = Longtycon
   structure Sigexp = Sigexp
   structure Spec = Spec
   structure Strid = Strid
end

local open Basis
in structure Env = Env
end

local open Interface
in structure Status = Status
   structure TypeStr = FlexibleTypeStr
end

local open TypeStr
in structure Tycon = Tycon
   structure TypeFcn = TypeFcn
end

structure ElaborateAtoms =
   ElaborateAtoms(structure Ast = Ast
		  structure Decs =
		     struct
			type t = unit
			val empty = ()
			fun datatypes _ = ()
		     end
		  structure Env = Interface
		  structure TypeStr = TypeStr)
   
local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
   val elaborateTypeOpt = elaborateTypeOpt
   val elaborateClosedScheme = elaborateClosedScheme
   val elaborateCons = elaborateCons
   val elaborateDatatypeRhs = elaborateDatatypeRhs
end

structure Set = DisjointSet

(* rule 65 *)
fun elaborateSigexp(sigexp: Sigexp.t, B: Basis.t): Interface.t =
   case Sigexp.node sigexp of
      Sigexp.Var s => Basis.lookupSigid(B, s)
    | _ =>
	 let
	    local val E = Basis.toEnv B
	       val lk =
		  Lookup.fromFun
		  (fn c => TypeStr.fromStrict(Env.lookupLongtycon(E, c)))
	    in fun lookup I = Lookup.plusEnv(lk, I)
	    end

	    fun elaborateTypedescs(typedescs, I: Interface.t): Interface.t =
	       (* rule 80 *)
	       Interface.types
	       (List.map(typedescs, fn {tycon, tyvars, ty} =>
			 {name = tycon,
			  typeStr = 
			  case ty of
			     NONE => TypeStr.tycon(Tycon.newNoname())
			   | SOME ty =>
				TypeStr.def
				(elaborateClosedScheme
				 (tyvars, ty, lookup I))}))

	    fun elaborateSigexp(sigexp: Sigexp.t, I: Interface.t)
	      : Interface.t =
	       case Sigexp.node sigexp of
		  Sigexp.Spec spec => (* rule 62 *)
		     elaborateSpec(spec, I)
		| Sigexp.Var x => (* rule 63 *)
		     Interface.copy(Basis.lookupSigid(B, x))
		| Sigexp.Where(sigexp, wheres) => (* rule 64 *)
		     let val I' = elaborateSigexp(sigexp, I)
		     in Interface.wheres
			(I', List.map(wheres, fn {tyvars, longtycon, ty} =>
				      (longtycon,
				       TypeFcn.funn
				       (elaborateClosedScheme
					(tyvars, ty, lookup I))))) ;
			I'
		     end
		  
	    and elaborateSpec(spec: Ast.Spec.t, I: Interface.t): Interface.t =
	       case Spec.node spec of
		  Spec.Val xts => (* rules 68, 79 *)
		     Interface.vals
		     (List.map(xts, fn (x, t) =>
			       {name = Ast.Vid.fromVar x,
				status = Status.Var,
				scheme = elaborateType(t, lookup I)}))
		| Spec.Type typedescs => (* rule 69 *)
		     elaborateTypedescs(typedescs, I)
		| Spec.Eqtype typedescs => (* rule 70 *)
		     elaborateTypedescs(typedescs, I)
		| Spec.Datatype rhs => (* rules 71, 72 *)
		     #2(elaborateDatatypeRhs{rhs = rhs,
					     lookup = lookup I,
					     newTycon = fn _ => Tycon.newNoname()})
		| Spec.Exception cons => (* rule 73 *)
		     Interface.excons(elaborateCons{cons = cons, tyvars = [],
						    tycon = Tycon.exn,
						    lookup = lookup I,
						    resultType = Atype.exn})
		| Spec.Structure ss => (* rules 74, 84 *)
		     Interface.strs
		     (List.map(ss, fn (strid, sigexp) =>
			       {name = strid,
				interface = elaborateSigexp(sigexp, I)}))
		| Spec.IncludeSigexp sigexp => (* rule 75 *)
		     elaborateSigexp(sigexp, I)
		| Spec.IncludeSigids sigids => (* Appendix A, p.59 *)
		     elaborateSpec
		     (List.fold(sigids, Spec.make Spec.Empty, fn (sigid, spec) =>
				Spec.make
				(Spec.Seq(spec,
					  Spec.make
					  (Spec.IncludeSigexp
					   (Sigexp.var sigid))))),
		      I)
		| Spec.Empty => (* rule 76 *)
		     Interface.empty
		| Spec.Seq(s, s') => (* rule 77 *)
		     let val I' = elaborateSpec(s, I)
			val I'' = elaborateSpec(s', Interface.+(I, I'))
		     in Interface.+(I', I'')
		     end
		| Spec.Sharing{spec, equations} => (* rule 78 and section G.3.3 *)
		     let val I' = elaborateSpec(spec, I)
			fun share eqn =
			   case Equation.node eqn of
			      Equation.Type cs =>
			      (case cs of
				  [] => ()
				| c :: cs =>
				     List.foreach
				     (cs, fn c' =>
				      Interface.shareType(I', c, c')))
			    | Equation.Structure ss =>
			      let
				 fun loop ss =
				    case ss of
				       [] => ()
				     | s :: ss =>
					  (List.foreach
					   (ss, fn s' =>
					    Interface.share(I', s, s')) ;
					   loop ss)
			      in loop ss
			      end
		     in List.foreach(equations, share) ;
			I'
		     end
	 in elaborateSigexp(sigexp, Interface.empty)
	 end

(* val elaborateSigexp = 
 *    Trace.trace2("elaborateSigexp",
 * 		Sigexp.layout,
 * 		Layout.ignore (*Basis.layout*),
 * 		(*Layout.ignore*) Interface.layout)
 *    elaborateSigexp
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRECEDENCE_PARSE_STRUCTS =
   sig
      structure Ast: AST
      structure Env: ELABORATE_ENV
      sharing Ast = Env.Ast
   end

signature PRECEDENCE_PARSE =
   sig
      include PRECEDENCE_PARSE_STRUCTS

      val parsePat: Ast.Pat.t list * Env.t -> Ast.Pat.t

      val parseExp: Ast.Exp.t list * Env.t -> Ast.Exp.t

      val parseClause: Ast.Pat.t list * Env.t -> {func: Ast.Var.t,
						   args: Ast.Pat.t list}
   end
(* Heavily modified from the SML/NJ sources by sweeks@research.nj.nec.com. *)

(* Copyright 1996 by AT&T Bell Laboratories *)
(* precedence.sml *)
 
functor PrecedenceParse(S: PRECEDENCE_PARSE_STRUCTS): PRECEDENCE_PARSE =
struct

open S
   
local open Ast
in structure Exp = Exp
   structure Fixity = Fixity
   structure Fixop = Fixop
   structure Longvid = Longvid
   structure Pat = Pat
   structure Vid = Vid
end
   
structure Fixval =
   struct
      datatype t = Nonfix | Infix of (int*int)

      fun eval(f: Fixity.t): t =
	 case f of
	    Fixity.Infix NONE => Infix(0, 1)
	  | Fixity.Infix(SOME n) => Infix(n+n, n+n+1)
	  | Fixity.Infixr NONE => Infix(1, 0)
	  | Fixity.Infixr(SOME n) => Infix(n+n+1, n+n)
	  | Fixity.Nonfix => Nonfix

      fun make({name: Longvid.t, fixop: Fixop.t}, E: Env.t): t =
	 case (fixop, Longvid.split name) of
	    (Fixop.None, ([], vid)) => eval(Env.lookupFixity(E, vid))
	  | _ => Nonfix

      fun makePat(p: Pat.t, E: Env.t): t =
	 case Pat.node p of
	    Pat.Var r => make(r, E)
	  | _ => Nonfix

      fun makeExp(e: Exp.t, E: Env.t): t =
	 case Exp.node e of
	    Exp.Var r => make(r, E)
	  | _ => Nonfix
   end

(*---------------------------------------------------*)
(*           from elaborate/precedence.sml           *)
(*---------------------------------------------------*)

datatype 'a precStack =
   INf of int * 'a * 'a precStack
 | NONf of 'a * 'a precStack
 | NILf
   
fun 'a parse{region: 'a -> Region.t,
	     items: 'a list,
	     fixval: 'a -> Fixval.t,
	     apply: 'a * 'a -> 'a,
	     tuple: 'a list -> 'a}: 'a =
   let
      fun error msg =
	 Control.errorStr(Region.list(items, region), msg)
	 
      fun ensureNONf((e, f), p) =
	 (case f of
	     Fixval.Nonfix => ()
	   | _ => Control.errorStr(region e,
				   "expression or pattern begins with infix identifier") ;
          NONf(e, p))
	  
      fun start token = ensureNONf(token, NILf)
	 
      (* parse an expression *)
      fun parse(stack: 'a precStack, (item: 'a, fixval: Fixval.t)) =
	 case (stack, (item, fixval)) of
	    (NONf(e, r), (e', Fixval.Nonfix)) => NONf(apply(e, e'), r)
	  | (p as INf _, token) => ensureNONf(token, p)
	  | (p as NONf(e1, INf(bp, e2, NONf(e3, r))),
	     (e4, f as Fixval.Infix(lbp, rbp))) =>
	    if lbp > bp then INf(rbp, e4, p)
	    else (if lbp = bp
		     then error "mixed left- and right-associative \
		      \operators of same precedence"
		  else ();
		  parse(NONf(apply(e2, tuple[e3, e1]), r), (e4, f)))
	   | (p as NONf _, (e', Fixval.Infix(_, rbp))) => INf(rbp, e', p)
	   | _ => Error.bug "Precedence.parse"

      (* clean up the stack *)
      fun finish stack =
	 case stack of
	    NONf(e1, INf(_, e2, NONf(e3, r))) =>
	       finish(NONf(apply(e2, tuple[e3, e1]), r))
	  | NONf(e1, NILf) => e1
	  | INf(_, e1, NONf(e2, p)) =>
	       (error "expression or pattern ends with infix identifier" ;
		finish(NONf(apply(e2, e1), p)))
	  | NILf => Error.bug "Corelang.finish NILf"
	  | _ => Error.bug "Corelang.finish"

      fun getfix x = (x, fixval x)
	 
   in case items of
      item :: items =>
	 finish(List.fold(items, start(getfix item), fn (item, state) =>
			  parse(state, getfix item)))
    | _ => Error.bug "parse"
   end

fun parsePat(ps, E) =
   parse{region = Pat.region,
	 items = ps,
	 fixval = fn p => Fixval.makePat(p, E),
	 tuple = Pat.tuple,
	 apply = fn (p1, p2) =>
	 case Pat.node p1 of
	    Pat.Var{name, ...} => Pat.make(Pat.App(Longvid.toLongcon name, p2))
	  | _ => Error.bug "applyPat got non-constructor"}

fun parseExp(es, E) =
   parse{region = Exp.region,
	 items = es,
	 fixval = fn e => Fixval.makeExp(e, E),
	 apply = Exp.app,
	 tuple = Exp.tuple}

(*---------------------------------------------------*)
(*                    parseClause                    *)
(*---------------------------------------------------*)

fun parseClause(pats: Pat.t list, E: Env.t) =
   let
      fun empty() = Error.bug "parseClause: empty clause"
      fun error msg =
	 (Control.errorStr(Region.list(pats, Pat.region), msg) ;
	  {func = Ast.Var.bogus,
	   args = []})
	    
      fun illegal() = error "illegal function symbol in clause"
      fun done(func: Pat.t, args: Pat.t list) =
	 case Pat.node func of
	    Pat.Var{name, ...} =>
	       (case Longvid.split name of
		   ([], x) => {func = Vid.toVar x, args = args}
		 | _ => illegal())
	  | _ => illegal()

      fun parse(ps : Pat.t list) =
	 case ps of
	    p :: rest =>
	       (case (Pat.node p, rest) of
		   (Pat.FlatApp[a, b, c], _) =>
		      done(b, Pat.tuple[a, c] :: rest)
		 | (_, []) => error "can't find function arguments in clause"
		 | _ => done(p, rest))
	  | _ => Error.bug "empty clause"
	     
   in case pats of
      [a, b, c] => (case Fixval.makePat(b, E) of
		     Fixval.Nonfix => parse pats
		   | _ => done(b, [Pat.tuple[a, c]]))
    | _ => parse pats
   end
    
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_PAT_STRUCTS = 
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      structure Decs: DECS
      structure ElaborateAtoms: ELABORATE_ATOMS
      structure Env: ELABORATE_ENV
      sharing Ast = CoreML.Ast
      sharing CoreML = Decs.CoreML = Env.CoreML
      sharing Decs = ElaborateAtoms.Decs
      sharing Env = ElaborateAtoms.Env
      sharing Env.TypeStr = ElaborateAtoms.TypeStr
   end

signature ELABORATE_PAT = 
   sig
      include ELABORATE_PAT_STRUCTS
      
      val elaboratePat :
	 {pat: Ast.Pat.t, env: Env.t, accum: Env.t} -> CoreML.Pat.t * Env.t

      val elaboratePats :
	 {pats: Ast.Pat.t list, env: Env.t, accum: Env.t}
	 -> CoreML.Pat.t list * Env.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaboratePat(S: ELABORATE_PAT_STRUCTS): ELABORATE_PAT = 
struct

open S

local open Ast
in structure Apat = Pat
   structure Field = Record.Field
   structure Fixop = Fixop
   structure Longvid = Longvid
   structure Record = Record
   structure Vid = Vid
end

local open CoreML
in structure Cpat = Pat
   structure Scheme = Scheme
end

local open Env
in structure TypeStr = TypeStr
end

structure Parse = PrecedenceParse(structure Ast = Ast
				  structure Env = Env)

local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
end

fun bind(x: Ast.Var.t, E': Env.t): CoreML.Var.t * Env.t = 
   let val x' = CoreML.Var.fromAst x
   in (x', Env.extendVar(E', x, x'))
   end

(* val traceElaboratePat =
 *    Trace.trace2("elaboratePat",
 * 		Apat.layout, Env.layout,
 * 		Layout.tuple2(Cpat.layout, Env.layout))
 *)

fun elaboratePat(p: Apat.t, E: Env.t, E'): Cpat.t * Env.t =
   let
      fun loop arg: Cpat.t * Env.t =
(*	 traceElaboratePat *)
	 (fn (p: Apat.t, E': Env.t) =>
	  case Apat.node p of
	    Apat.Wild => (Cpat.Wild, E')
	  | Apat.Var{name=x, ...} =>
	       (case Env.peekLongcon(E, Ast.Longvid.toLongcon x) of
		   SOME c => (Cpat.Con{con = c, arg = NONE}, E')
		 | NONE => (case Ast.Longvid.split x of
			       ([], x) => let val x = Ast.Vid.toVar x
					     val (x', E') = bind(x, E')
					 in (Cpat.Var x', E')
					 end
			     | _ => Error.error "longid in var pat"))
	  | Apat.Const c => (Cpat.Const c, E')
	  | Apat.Tuple ps => loopsCont(ps, E', Cpat.tuple)
	  | Apat.Record{items, flexible} =>
	       (* rules 36, 38, 39 and Appendix A, p.57 *)
	       let
		  val fps =
		     List.map
		     (items,
		      fn Apat.Item.Field fp => fp
		       | Apat.Item.Vid(vid, tyo, po) =>
			    (Field.String(Vid.toString vid),
			     let
				val p =
				   case po of
				      NONE => Apat.longvid(Longvid.short vid)
				    | SOME p =>
					 Apat.layered{fixop = Fixop.None,
						      var = Vid.toVar vid,
						      constraint = NONE,
						      pat = p}
			     in case tyo of
				NONE => p
			      | SOME ty => Apat.constraint(p, ty)
			     end))
		  val (fields, ps) = List.unzip fps
	       in loopsCont(ps, E', fn ps =>
			    Cpat.record
			    {record = Record.fromList(List.zip(fields, ps)),
			     flexible = flexible})
	       end
	  | Apat.List ps => loopsCont(ps, E', Cpat.list)
	  | Apat.FlatApp items => loop(Parse.parsePat(items, E), E')
	  | Apat.App(c, p) =>
	       let val (p, E') = loop(p, E')
	       in (Cpat.Con{con = Env.lookupLongcon(E, c),
			    arg = SOME p},
		   E')
	       end
	  | Apat.Constraint(p, t) =>
	       let val (p, E') = loop(p, E')
		  val t = Scheme.ty(elaborateType(t, Lookup.fromEnv E))
	       in (Cpat.Constraint(p, t), E')
	       end
	  | Apat.Layered{var=x, constraint, pat, ...} =>
	       let val (x', E') = bind(x, E')
		  val (pat, E') =
		     loop(case constraint of
			     NONE => pat
			   | SOME t => Apat.constraint(pat, t),
			  E')
	       in (Cpat.Layered(x', pat), E')
	       end) arg
      and loopsCont(ps: Apat.t list, E': Env.t, cont: Cpat.t list -> Cpat.t)
	: Cpat.t * Env.t =
	 let val (ps, E') = elaboratePats(ps, E, E')
	 in (cont ps, E')
	 end
   in loop(p, E')
   end
and elaboratePats(ps: Apat.t list, E, E'): Cpat.t list * Env.t =
   let val (ps, E') = List.fold(ps, ([], E'), fn (p, (ps, E')) =>
				let val (p, E') = elaboratePat(p, E, E')
				in (p :: ps, E')
				end)
   in (List.reverse ps, E')
   end

val elaboratePat =
   fn {pat, env, accum} => elaboratePat(pat, env, accum)

val elaboratePats =
   fn {pats, env, accum} => elaboratePats(pats, env, accum)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_CORE_STRUCTS = 
   sig
      include ELABORATE_PAT_STRUCTS
   end

signature ELABORATE_CORE = 
   sig
      include ELABORATE_CORE_STRUCTS

      (* elaborate dec in env, accumulating Core ML decs and bindings *)
      val elaborateDec:
	 Ast.Dec.t * Env.t * Decs.t * Env.t
	 -> Decs.t * Env.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateCore(S: ELABORATE_CORE_STRUCTS): ELABORATE_CORE = 
struct

open S

local open Env
in structure Vid = Vid
end

local open Ast
in structure Record = Record
   structure Adec = Dec
   structure Aexp = Exp
   structure Atype = Type
   structure Avar = Var
   structure EbRhs = EbRhs
   structure Longvid = Longvid
   structure Longtycon = Longtycon
   structure Strid = Strid
   structure TypBind = TypBind
end

local open CoreML
in structure Cdec = Dec
   structure Cexp =
      struct
	 open Exp
	 val bogus = unit
      end
   structure Cpat = Pat
   structure Cprim = Prim
   structure Ctype = Type
   structure Cvar = Var
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Tyvar = Tyvar
end
      
local open ElaborateAtoms
in structure Lookup = Lookup
   val elaborateType = elaborateType
   val elaborateTypeOpt = elaborateTypeOpt
end

local open Env
in structure TypeStr = TypeStr
end

structure Parse = PrecedenceParse(structure Ast = Ast
				  structure Env = Env)

structure ElaboratePat = ElaboratePat(S)
   
fun constrain(e, tyOpt) =
   case tyOpt of
      NONE => e
    | SOME ty => CoreML.Exp.Constraint(e, ty)

(* val traceElaborateExp =
 *    Trace.trace2("elaborateExp", Ast.Exp.layout, Env.layout, CoreML.Exp.layout)
 *)
   
fun elaborateEbs(ebs, E: Env.t, decs, E'): Decs.t * Env.t =
   List.fold(ebs, (decs, E'), fn ((exn, rhs), (decs, E')) =>
	     let
		val (decs, exn') =
		   case EbRhs.node rhs of
		      EbRhs.Def c => (decs, Env.lookupLongcon(E, c))
		    | EbRhs.Gen to =>
			 let val exn' = CoreML.Con.fromAst exn
			 in (Decs.add(decs,
				      Cdec.Exception{con = exn',
						     arg = elaborateTypeOpt(to, E)}),
			     exn')
			 end
	     in (decs, Env.extendExn(E', exn, exn'))
	     end)
   
(*---------------------------------------------------*)
(*                   Declarations                    *)
(*---------------------------------------------------*)

local
(*    val traceElaborateDec =
 *       Trace.trace4("elaborateDec",
 * 		   Ast.Dec.layout, Env.layout, Layout.ignore, Layout.ignore,
 * 		   Layout.tuple2(Decs.layout, Env.layout))
 *)
in
   
fun elaborateDec arg =
(*   traceElaborateDec *)
   (fn (d, E, decs, E') =>
    case Adec.node d of
       Adec.Val(tyvars, vbs) => elaborateVbs(tyvars, vbs, E, decs, E')
     | Adec.Valrec(tyvars, rvbs) => elaborateRvbs(tyvars, rvbs, E, decs, E')
     | Adec.Fun(tyvars, fbs) => elaborateFbs(tyvars, fbs, E, decs, E')
     | Adec.Type typBind =>
	  (decs,
	   Env.+(E', ElaborateAtoms.elaborateTypBind(typBind, Lookup.fromEnv E)))
     | Adec.Datatype rhs =>
	  let
	     val (decs', E'') =
		ElaborateAtoms.elaborateDatatypeRhs
		{rhs = rhs,
		 lookup = Lookup.fromEnv E,
		 newTycon = CoreML.Tycon.fromAst}
	  in (Decs.append(decs, decs'), Env.+(E', E''))
	  end
     | Adec.Abstype{datBind, body} => (* rule 19 and p.57 *)
	  let
	     val {VE, TE, decs = decs'} =
		ElaborateAtoms.elaborateDatBind{datBind = datBind,
						lookup = Lookup.fromEnv E,
						newTycon = CoreML.Tycon.fromAst}
	  in elaborateDec(body, Env.plus[E, VE, TE], Decs.append(decs, decs'),
			  Env.+(E', Env.abs TE))
	  end
     | Adec.Exception ebs => elaborateEbs(ebs, E, decs, E')
     | Adec.Local(d, d') =>
	  let val (decs, E) = elaborateDec(d, E, decs, E)
	  in elaborateDec(d', E, decs, E')
	  end
     | Adec.SeqDec ds =>
	  let val (decs, _, E') =
	     List.fold(ds, (decs, E, E'), fn (d, (decs, E, E')) =>
		       let val (decs, E'') = elaborateDec(d, E, decs, Env.empty)
		       in (decs, Env.+(E, E''), Env.+(E', E''))
		       end)
	  in (decs, E')
	  end
     | Adec.Fix{ops, fixity} =>
	  (decs, Env.+(E', Env.fixs(List.map(ops, fn op' => (op', fixity)))))
     | Adec.Open paths =>
	  (decs,
	   Env.plus(E' :: (List.map(paths, fn p => Env.lookupLongstrid(E, p)))))
     | Adec.Overload(x, t, xs) =>
	  let val x' = Cvar.fromAst x
	     val scheme = elaborateType(t, Lookup.fromEnv E)
	     val E' = Env.+(E', Env.vars[(x, x', SOME scheme)])
	  in (Decs.add(decs,
		       Cdec.Overload
		       {var = x',
			scheme = scheme,
			ovlds = List.map(xs, fn x => Env.lookupLongvar(E, x))}),
	      E')
	  end) arg

and elaborateVbs(tyvars, vbs, E: Env.t, decs, E'): Decs.t * Env.t = 
   List.fold
   (vbs, (decs, E'), fn ((pat, exp), (decs, E')) =>
    let val exp = elaborateExp(exp, E)
        val (pat, E') = ElaboratePat.elaboratePat{pat = pat, env = E, accum = E'}
    in (Decs.add(decs, Cdec.Val{pat = pat,
				tyvars = tyvars,
				exp = exp}),
	E')
    end)

and elaborateFbs(tyvars, fbs, E: Env.t, decs, E'): Decs.t * Env.t =
   let
      val clausess =
	 List.map
	 (fbs, fn clauses =>
	  List.map(clauses, fn {pats, resultType, body} =>
		   let val {func, args} = Parse.parseClause(pats, E)
		      val print =
			 Aexp.longvid
			 (Ast.Longvid.long
			  ([Strid.fromString "Primitive",
			    Strid.fromString "Stdio"],
			   Ast.Vid.fromString "print"))
		   in {func = func, args = args,
		       resultType = resultType,
		       body =
		       (if !Control.printAtFunEntry
			   then Aexp.make
			      (Aexp.Seq
			       [Aexp.app(print,
					 Aexp.const(Ast.Const.String
						    ("Entering "
						     ^ Avar.toString func
						     ^ "\n"))),
				body])
			else body)}
		   end))
      val funcs =
	 List.map(clausess, fn {func, ...} :: _ => func
                             | [] => Error.error "no clauses in fundec")
      val newFuncs = List.map(funcs, Cvar.fromAst)
      val E'' = Env.vars(List.map2(funcs, newFuncs, fn (name, var) =>
				   (name, var, NONE)))
      val E = Env.+(E, E'')
   in (Decs.add
       (decs,
	Cdec.Fun
	{tyvars = tyvars,
	 decs =
	 List.map2
	 (clausess, newFuncs, fn (clauses,newFunc) =>
	  case clauses of
	     [] => Error.error "empty clauses in fundec"
	   | {args, ...} :: _ =>
		let val numVars = List.length args
		in {var = newFunc, ty = NONE,
		    rules =
		    let
		       val rs =
			  List.map
			  (clauses, fn {args, resultType, body, ...} =>
			   let
			      val (pats, E') =
				 ElaboratePat.elaboratePats{pats = args,
							    env = E,
							    accum = E}
			      val body = elaborateExp(body, E')
			   in (Cpat.tuple pats,
			       constrain(body,
					 elaborateTypeOpt(resultType, E)))
			   end)
		       fun make(0, xs) =
			  Cexp.casee(Cexp.tuple(List.fold(xs, [], fn (x, xs) =>
							  (Cexp.Var x) :: xs)),
				     rs)
			 | make(n, xs) =
			   let val x = Cvar.newNoname()
			   in Cexp.lambda(x, make(n-1, x :: xs))
			   end
		    in if numVars = 1 then rs
		       else (case make(numVars, []) of
				Cexp.Fn rules => rules
			      | _ => Error.bug "elaborateFbs")
		    end}
		end)}),
       Env.+(E', E''))
   end

and elaborateRvbs(tyvars, rvbs, E: Env.t, decs, E'): Decs.t * Env.t =
   let val vars = List.map(rvbs, #var)
      val newvars = List.map(vars, Cvar.fromAst)
      val E'' = Env.vars(List.map2(vars, newvars, fn (name, var) =>
				   (name, var, NONE)))
      val E = Env.+(E, E'')
   in (Decs.add
       (decs,
	Cdec.Fun
	{tyvars = tyvars,
	 decs = List.map2(rvbs, newvars, fn ({var, fixity, rules, ty},newvar) =>
			  {var = newvar,
			   ty = elaborateTypeOpt(ty, E),
			   rules = elaborateRules(rules, E)})}),
       Env.+(E', E''))
   end

(*---------------------------------------------------*)
(*                    Expressions                    *)
(*---------------------------------------------------*)

and elaborateExp arg: CoreML.Exp.t =
(*   traceElaborateExp *)
   (fn (e: Aexp.t, E: Env.t) =>
   case Aexp.node e of
      Aexp.Var{name=id, ...} =>
	 (case Env.lookupLongvid(E, id) of
	     vid => (case vid of
			Vid.Var x => Cexp.Var x
		      | Vid.ConAsVar c => Cexp.Con c
		      | Vid.Con c => Cexp.Con c
		      | Vid.Exn c => Cexp.Con c
		      | Vid.Prim p => Cexp.Prim p))
    | Aexp.Fn rs => Cexp.Fn(elaborateRules(rs, E))
    | Aexp.FlatApp items => elaborateExp(Parse.parseExp(items, E), E)
    | Aexp.App(e1, e2) => Cexp.App(elaborateExp(e1, E), elaborateExp(e2, E))
    | Aexp.Case(e, rs) => Cexp.casee(elaborateExp(e, E), elaborateRules(rs, E))
    | Aexp.Let(d, e) => let val (d, E) = elaborateDec(d, E, Decs.empty, E)
		       in Cexp.Let(Decs.toList d, elaborateExp(e, E))
		       end
    | Aexp.Seq es => Cexp.seq(elaborateExps(es, E))
    | Aexp.Const c => Cexp.Const c
    | Aexp.Record r => Cexp.Record(Record.map(r, fn e => elaborateExp(e, E)))
    | Aexp.List es => Cexp.list(elaborateExps(es, E))
    | Aexp.Selector f => Cexp.selector f
    | Aexp.Constraint(e, t) =>
	 Cexp.Constraint(elaborateExp(e, E),
			 Scheme.ty(elaborateType(t, Lookup.fromEnv E)))
    | Aexp.Handle(e, rs) => Cexp.Handle(elaborateExp(e, E), elaborateRules(rs, E))
    | Aexp.Raise e => Cexp.Raise(elaborateExp(e, E))
    | Aexp.If(a, b, c) =>
	 Cexp.iff(elaborateExp(a, E), elaborateExp(b, E), elaborateExp(c, E))
    | Aexp.Andalso(e, e') => Cexp.andAlso(elaborateExp(e, E), elaborateExp(e', E))
    | Aexp.Orelse(e, e') => Cexp.orElse(elaborateExp(e, E), elaborateExp(e', E))
    | Aexp.While{test, expr} =>
	 Cexp.whilee{test = elaborateExp(test, E),
		     expr = elaborateExp(expr, E)}
    | Aexp.Prim{name, ty} =>
	 Cexp.Prim
	 (Cprim.new(name, ElaborateAtoms.elaborateType(ty, Lookup.fromEnv E)))
	 ) arg
   
and elaborateExps(es: Ast.Exp.t list, E: Env.t): CoreML.Exp.t list =
   List.map(es, fn e => elaborateExp(e, E))

and elaborateRules(rs, E: Env.t) =
   List.map(rs, fn (pat, exp) => 
	    let
	       val (pat, E) =
		  ElaboratePat.elaboratePat{pat = pat,
					    env = E,
					    accum = E}
	    in (pat, elaborateExp(exp, E))
	    end)

end (* local *)

end
functor MonoOption(X: T): T =
struct

type t = X.t option

val equals = Option.equals X.equals

val layout = Option.layout X.layout
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ElaborateEnv(S: ELABORATE_ENV_STRUCTS): ELABORATE_ENV =
struct

open S

local open Ast
in structure Fixity = Fixity
   structure Strid = Strid
   structure Longcon = Longcon
   structure Longvid = Longvid
   structure Longstrid = Longstrid
   structure Longtycon = Longtycon
end

local open CoreML
in structure Con = Con
   structure Var = Var
   structure Prim = Prim
   structure Record = Record
   structure Srecord = SortedRecord
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
   structure Tyvar = Tyvar
   structure Var = Var
end

local open Interface
in structure Element = Element
   structure ShapeId = ShapeId
   structure Status = Status
end

local open TypeStr
in structure Cons = Cons
   structure TypeFcn = TypeFcn
end

structure SchemeOpt = MonoOption(Scheme)

structure Vid =
   struct
      open CoreML
	 
      datatype t =
	 Var of Var.t
       | ConAsVar of CoreML.Con.t
       | Con of Con.t
       | Exn of Con.t
       | Prim of Prim.t

      val status =
	 fn Var _ => Status.Var
	  | Prim _ => Status.Var
	  | ConAsVar _ => Status.Var
	  | Con _ => Status.Con
	  | Exn _ => Status.Exn

      val bogus = Var Var.bogus

      fun layout vid =
	 let open Layout
	    val (name, l) =
	       case vid of
		  Var v => ("Var", Var.layout v)
		| Con c => ("Con", Con.layout c)
		| ConAsVar c => ("ConAsVar", Con.layout c)
		| Exn c => ("Exn", Con.layout c)
		| Prim p => ("Prim", Prim.layout p)
	 in l(* paren(seq[str name, str " ", l])*)
	 end

      val deVar =
	 fn Var v => SOME v
	  | _ => NONE
		
      val deCon =
	 fn Con c => SOME c
	  | Exn c => SOME c
	  | _ => NONE

      val dePrim =
	 fn Prim p => SOME p
	  | _ => NONE
	  
      fun output(r, out) = Layout.output(layout r, out)
	 
      val equals = 
	 fn (Var x, Var x') => Var.equals(x, x')
	  | (Con c, Con c') => Con.equals(c, c')
	  | (Exn c, Exn c') => Con.equals(c, c')
	  | (Prim p, Prim p') => Prim.equals(p, p')
	  | _ => false
   end

(*------------------------------------*)
(*               Domain               *)
(*------------------------------------*)

structure Domain =
   struct
      open Ast
	 
      datatype t =
	 Fixity of Vid.t
	| Vid of Vid.t
	| Tycon of Tycon.t
	| Strid of Strid.t

      val Var = Vid o Ast.Vid.fromVar
      val Prim = Var
      val Con = Vid o Ast.Vid.fromCon

      fun hash x =
	 case x of
	    Fixity x => Vid.hash x
	  | Vid x => Vid.hash x
	  | Tycon x => Tycon.hash x
	  | Strid x => Strid.hash x

      val equals =
	 fn (Fixity x, Fixity y) => Vid.equals(x, y)
	  | (Vid x, Vid y) => Vid.equals(x, y)
	  | (Tycon x, Tycon y) => Tycon.equals(x, y)
	  | (Strid x, Strid y) => Strid.equals(x, y)
	  | _ => false

      fun layout d =
	 case d of
	    Fixity x => Vid.layout x
	  | Vid x => Vid.layout x
	  | Tycon x => Tycon.layout x
	  | Strid x => Strid.layout x
   end

structure Table = HashTable

(*---------------------------------------------------*)
(*                 Main Env Datatype                 *)
(*---------------------------------------------------*)

datatype t = T of {elements: elements,
		   shapeId: ShapeId.t option}
and elements =
   Empty
  | Table of table
  | Extend of (Domain.t * range) list * elements
  | Plus of elements list (* newer bindings in front *)
and range =
   Fixity of Ast.Fixity.t
  | Vid of {vid: Vid.t, scheme: SchemeOpt.t}
  | Tycon of TypeStr.t
  | Strid of t  (* always consolidated *)
withtype table = (Domain.t, range) Table.hash_table

type env = t

fun stamp(T{elements, ...}, shapeId) =
   T{elements = elements, shapeId = SOME shapeId}

fun unconstrained(es: elements): t =
   T{elements = es, shapeId = NONE}

val bogus = unconstrained Empty
   
fun elements(T{elements, ...}) = elements

(*val equals = fn _ => Error.unimplemented "ElaborateEnv.="*)

exception NotFound
fun consolidate'(E: elements): table =
   case E of
      Table t => t
    | _ => 
	 let 
	    (* isn't really accurate - just used to get an approximation *)
	    fun size(E, accum) = 
	       case E of
		  Empty => accum
		| Table t => accum + Table.numItems t
		| Extend(bs, E) => size(E, accum + List.length bs)
		| Plus Es => List.fold(Es, accum, size)
	    val size = size(E, 0)
	    val table = Table.mkTable(Domain.hash, Domain.equals) (size, NotFound)
	    fun insert(d, r) =
	       case Table.find table d of
		  NONE => Table.insert table (d, r)
		| SOME _ => ()
	    fun loop E =
	       case E of
		  Empty => ()
		| Table t => Table.appi insert t
		| Extend(drs, E) => (List.foreach(drs, insert) ; loop E)
		| Plus Es => List.foreach(Es, loop)
	 in loop E ; table
	 end

fun consolidate(T{elements, shapeId}): t =
   T{elements = Table(consolidate' elements),
     shapeId = shapeId}

local
   open Layout
in
   fun layout(E:env): Layout.t =
      let val t = consolidate'(elements E)
      in list(List.map(Table.listItemsi t, fn (d, r) =>
		       seq[Domain.layout d,
			   str " -> ",
			   layoutRange r]))
      end
   and layoutRange r =
      case r of
	 Fixity x => Fixity.layout x
       | Vid{scheme, vid} => tuple[Vid.layout vid,
				  case scheme of
				     NONE => empty
				   | SOME s => Scheme.layout s]
       | Tycon x => TypeStr.layout x
       | Strid x => layout x
   val layoutEnv = layout
end

val deFixity =
   fn Fixity x => SOME x
    | _ => NONE
val deVar =
   fn Vid{vid, ...} => Vid.deVar vid
    | _ => NONE
val deCon =
   fn Vid{vid, ...} => Vid.deCon vid
    | _ => NONE
val dePrim =
   fn Vid{vid, ...} => Vid.dePrim vid
    | _ => NONE
val deVid =
   fn Vid{vid, ...} => SOME vid
    | _ => NONE
val deTycon =
   fn Tycon x => SOME x
    | _ => NONE
val deStrid =
   fn Strid x => SOME x
    | _ => NONE

val empty = unconstrained Empty

fun plus(Es: t list): t =
   unconstrained(Plus(List.fold(Es, [], fn (E, Es) =>
				elements E :: Es)))

fun extend(E, d, r) =
   unconstrained
   (case elements E of
       Extend(drs, E) => Extend((d, r) :: drs, E)
     | E => Extend([(d, r)], E))

fun extends(E, drs) =
   case E of
      Extend(drs', E) => Extend(List.append(drs, drs'), E)
    | E => Extend(drs, E)

local
   val extend = fn (D, R) => fn (E, d, r) => extend(E, D d, R r)
   fun noScheme make vid = Vid{scheme = NONE, vid = make vid}
in val extendVar = extend(Domain.Var, noScheme Vid.Var)
   val extendPrim = extend(Domain.Prim, noScheme Vid.Prim)
   val extendCon = extend(Domain.Con, noScheme Vid.Con)
   val extendExn = extend(Domain.Con, noScheme Vid.Exn)
   val extendTycon = extend(Domain.Tycon, Tycon)
   val extendStrid = extend(Domain.Strid, Strid o consolidate)
end

local
   fun list(D: 'a -> Domain.t, R: 'b -> range) (drs: ('a * 'b) list): t =
      unconstrained(Extend(List.map(drs, fn (d, r) => (D d, R r)), Empty))
in val fixs = list(Domain.Fixity, Fixity)
   val strs = list(Domain.Strid, Strid o consolidate)
   val types = list(Domain.Tycon, Tycon)
end

fun cons(Cons.T l) =
   unconstrained
   (Extend
    (List.map(l, fn {name, con, scheme} =>
	      (Domain.Con name,
	       Vid{scheme = SOME scheme, vid = Vid.Con con})),
     Empty))

val fromCons = cons

fun vars l =
   unconstrained
   (Extend(List.map(l, fn (name, var, scheme) =>
		    (Domain.Var name,
		     Vid{vid = Vid.Var var, scheme = scheme})),
	   Empty))

type ('a, 'b) peek = t * 'a -> 'b option

local
   fun peek(D: 'a -> Domain.t, de: range -> 'b option): ('a, 'b) peek =
      fn (E: t, d :'a) =>
      let val d = D d
	 fun loop E =
	    case E of
	       Empty => NONE
	     | Extend(drs, E) =>
		  (case List.peek(drs, fn (d', _) => Domain.equals(d, d')) of
		      NONE => loop E
		    | SOME(_, r) => de r)
	     | Table t => (case Table.find t d of
			      NONE => NONE
			    | SOME r => de r)
	     | Plus Es =>
		  let fun loop' [] = NONE
			| loop'(E :: Es) = (case loop E of
					     NONE => loop' Es
					   | x => x)
		  in loop' Es
		  end
      in loop(elements E)
      end
in
   val peekFixity = peek(Domain.Fixity, deFixity)
   val peekVar = peek(Domain.Var, deVar)
   val peekCon = peek(Domain.Con, deCon)
   val peekPrim = peek(Domain.Prim, dePrim)
   val peekVid = peek(Domain.Vid, deVid)
   val peekTycon = peek(Domain.Tycon, deTycon)
   val peekStrid = peek(Domain.Strid, deStrid)
   val peek = peek(fn x => x, SOME)
end

fun lookupFixity(E, vid) =
   case peekFixity(E, vid) of
      NONE => Fixity.Nonfix
    | SOME fixity => fixity

fun toIsInfix (E:t) (vid:Ast.Vid.t): bool =
   case peekFixity(E, vid) of
      NONE => false
    | SOME f => Fixity.isInfix f

fun peekStrids(E: t, strids): t option =
   case strids of
      [] => SOME E
    | strid :: strids =>
	 case peekStrid(E, strid) of
	    SOME E => peekStrids(E, strids)
	  | NONE => NONE

fun peekLong(split, peek) (E, x) =
   let val (strids, x) = split x
   in case peekStrids(E, strids) of
      SOME E => peek(E, x)
    | NONE => NONE
   end

val peekLongvid = peekLong(Ast.Longvid.split, peekVid)
val peekLongvar = peekLong(Ast.Longvar.split, peekVar)
val peekLongcon = peekLong(Ast.Longcon.split, peekCon)
val peekLongtycon = peekLong(Ast.Longtycon.split, peekTycon)
val peekLongstrid = peekLong(Ast.Longstrid.split, peekStrid)

local
   fun lookup(peek: ('a, 'b) peek,
	      bogus: 'b,
	      unbound: 'a -> unit) (E: t, x: 'a): 'b =
      case peek(E, x) of
	 SOME y => y
       | NONE => (unbound x ; bogus)
in
   val lookupTycon =
      lookup(peekTycon, TypeStr.bogus, Ast.Tycon.unbound)
   val lookupStrid =
      lookup(peekStrid, bogus, Ast.Strid.unbound)
   val lookupLongvid =
      lookup(peekLongvid, Vid.bogus, Ast.Longvid.unbound)
   val lookupLongvar =
      lookup(peekLongvar, Var.bogus, Ast.Longvar.unbound)
   val lookupLongcon =
      lookup(peekLongcon, Con.bogus, Ast.Longcon.unbound)
   val lookupLongtycon =
      lookup(peekLongtycon, TypeStr.bogus, Ast.Longtycon.unbound)
   val lookupLongstrid =
      lookup(peekLongstrid, bogus, Ast.Longstrid.unbound)
(*   val lookup =
      lookup(peek, Domain.layout) *)
end

(*---------------------------------------------------*)
(*                        cut                        *)
(*---------------------------------------------------*)

(* section 5.3, 5.5, 5.6 and rules 52, 53 *)
(* transparent:
     look at the stamp within the env *)
fun cut{env, interface, opaque, region}: t =
   let
      fun error(name, l) =
	 Control.error
	 (region, let open Layout
		  in seq[str name, str " ", l,
			 str " in signature but not in structure"]
		  end)
      val interface =
	 Interface.realize
	 (interface, fn c =>
	  if opaque
	     then TypeFcn.Tycon(Tycon.newNoname())
	  else
	     TypeStr.typeFcn
	     (case peekLongtycon(env, c) of
		 NONE => (error(Longtycon.className, Longtycon.layout c) ;
			  TypeStr.bogus)
	       | SOME typeStr => typeStr))
      fun cut(E as T{elements, shapeId}, I, strids) =
	 let val shapeId' = Interface.shapeId I
	    val cutoff =
	       if opaque then NONE
	       else case shapeId of
		  NONE => NONE
		| SOME shapeId =>
		     if ShapeId.equals(shapeId, shapeId')
			then SOME E
		     else NONE
	 in case cutoff of
	    SOME E => E
	  | NONE =>
	       let
		  val E' =
		     Interface.fold
		     (I, empty, fn (elt, E') =>
		      case elt of
			 Element.Str{name, interface=I} =>
			    (case peekStrid(E,name) of
				NONE =>
				   (error(Longstrid.className,
					  Longstrid.layout	
					  (Longstrid.long(List.reverse strids,
							  name))) ;
				    E')
			       | SOME E =>
				    extendStrid(E', name, cut(E, I, name :: strids)))
			| Element.Val{name, status, scheme} =>
			     (case peekVid(E, name) of
				 NONE =>
				    (error(Longvid.className,
					   Longvid.layout
					   (Longvid.long(List.reverse strids,
							 name))) ;
				     E')
			       | SOME vid =>
				    let
				       val vid =
					  case (vid, status) of
					     (Vid.Con c, Status.Var) =>
						Vid.ConAsVar c
					   | (Vid.Exn c, Status.Var) =>
						Vid.ConAsVar c
					   | (_, Status.Var) => vid
					   | (Vid.Con _, Status.Con) => vid
					   | (Vid.Exn _, Status.Exn) => vid
					   | _ =>
						(Control.error
						 (region,
						  let open Layout
						  in seq
						     [str Longvid.className,
						      str " ",
						      Longvid.layout
						      (Longvid.long(List.reverse strids,
								    name)),
						      str " has status ",
						      Status.layout(Vid.status vid),
						      str " in structure but status ",
						      Status.layout status,
						      str " in signature "]
						  end)
						 ; vid)
				    in extend(E', Domain.Vid name,
					      Vid{vid = vid,
						  scheme = SOME scheme})
				    end)
			| Element.Type{name, typeStr, ...} => 
			     case peekTycon(E,name) of
				NONE =>
				   (error(Longtycon.className,
					  Longtycon.layout
					  (Longtycon.long(List.reverse strids,
							  name))) ;
				    E')
			      | SOME typeStr' =>
				   let val (typeFcn, cons) = TypeStr.dest typeStr
				   in extendTycon
				      (E', name,
				       TypeStr.make
				       (typeFcn,
					let
					   val cons' =
					      Cons.toList(TypeStr.cons typeStr')
					   fun lookup name =
					      case List.peek(cons',
							     fn {name=name', ...} =>
							     Ast.Con.equals(name,name')) of
						 NONE =>
						    (error(Longcon.className,
							   Longcon.layout
							   (Longcon.long
							    (List.reverse strids,
							     name))) ;
						     Con.bogus)
					       | SOME{con, ...} => con
					in Cons.T
					   (List.map
					    (Cons.toList cons, fn {name, scheme, ...} =>
					     {name = name, con = lookup name,
					      scheme = scheme}))
					end))
				   end)

		in stamp(E', shapeId')
		end
	 end
	 
   in cut(env, interface, [])
   end

fun abs(E: t) =
   unconstrained
   (Table
    (Table.map
     (fn Tycon typeStr => Tycon(TypeStr.fromTypeFcn(TypeStr.typeFcn typeStr))
   | r => r)
     (consolidate'(elements E))))

(*---------------------------------------------------*)
(*                      primEnv                      *)
(*---------------------------------------------------*)
fun addEquals E = extendPrim(E, Ast.Var.fromString "=", Prim.equal)

fun prim() =
   let 
      val E = empty

      val E =
	 List.fold
	 (Tycon.prims, E, fn (tycon, E) =>
	  extendTycon(E,
		      Ast.Tycon.fromString(CoreML.Tycon.originalName tycon),
		      TypeStr.tycon tycon))

      local
	 open Scheme Type
	 val --> = arrow
	 infixr -->
      in
	 val datatypes =
	    [{tycon = Tycon.bool,
	      tyvars = [],
	      cons = [{con = Con.falsee, arg = NONE},
		      {con = Con.truee, arg = NONE}]},
	     let val a = Tyvar.newNoname{equality = false}
	     in {tycon = Tycon.list,
		 tyvars = [a],
		 cons = [{con = Con.nill, arg = NONE},
			 {con = Con.cons, arg = SOME(tuple[var a, list(var a)])}]}
	     end,
	     let val a = Tyvar.newNoname{equality = false}
	     in {tycon = Tycon.reff,
		 tyvars = [a],
		 cons = [{con = Con.reff, arg = SOME(var a)}]}
	     end]
      end
   
      val E =
	 List.fold
	 (datatypes, E, fn ({tyvars, tycon, cons}, E) =>
	  let
	     val cons as Cons.T cs =
		Cons.T(List.map
		       (cons, fn {con, arg} =>
			let
			   val resultType =
			      Type.con(tycon, List.map(tyvars, Type.var))
			   val scheme =
			      Scheme.T
			      {tyvars = tyvars,
			       ty = (case arg of
					NONE => resultType
				      | SOME t => Type.arrow(t, resultType))}
			in {name = Con.toAst con,
			    con = con,
			    scheme = scheme}
			end))
	  in plus[fromCons cons,
		  extendTycon(E, Tycon.toAst tycon, TypeStr.data(tycon, cons))]
	  end)
	 
      val E = extendTycon(E, Ast.Tycon.fromString "unit",
			  TypeStr.def(Scheme.make0 Type.unit))
	 
      val E = addEquals E

      val exceptions =
	 List.map(CoreML.Con.excons, fn c => {con = c, arg = NONE})

      val E =
	 List.fold(CoreML.Con.excons, E, fn (c, E) =>
		   (extendCon(E, Con.toAst c, c)))
	 
   in (E, CoreML.Constructors.T{datatypes = datatypes,
				exceptions = exceptions})
   end

(*------------------------------------*)
(*              identify              *)
(*------------------------------------*)

fun containsRecord(ty: Type.t): bool =
   case ty of
      Type.Var _ => false
    | Type.Con(_, ts) => List.exists(ts, containsRecord)
    | Type.Record r =>
	 case Srecord.detupleOpt r of
	    NONE => true
	  | SOME ts => List.exists(ts, containsRecord)

local
   local open Ast
   in structure Acon = Con
      structure Dec = Dec
      structure Exp = Exp
      structure Longcon = Longcon
      structure Longvid = Longvid
      structure Pat = Pat
      structure Strdec = Strdec
      structure Strexp = Strexp
      structure Avid = Vid
      structure Topdec = Topdec
   end
   structure Field = Srecord.Field
in
   fun error() = Error.bug "wrap/unwrap"
   fun longvidFromStr s = Longvid.short(Avid.fromString s)
   fun longconFromStr s = Longcon.short(Acon.fromString s)
   local val c = Counter.new 0
   in fun newLongvid() = longvidFromStr("temp" ^ Int.toString(Counter.next c))
   end
   val none = longvidFromStr "NONE"
   val someCon = longconFromStr "SOME"
   val someVid = longvidFromStr "SOME"
      
   fun bind(e: Exp.t, f: Exp.t -> Exp.t): Exp.t =
      case Exp.node e of
	 Exp.Var _ => f e
       | _ => let val x = newLongvid()
	      in Exp.make(Exp.Let(Dec.make(Dec.Val([], [(Pat.longvid x, e)])),
				  f(Exp.longvid x)))
	      end

   fun wrap(ty: Type.t, e: Exp.t,
	    record as (makeRecord: Type.t Srecord.t * Exp.t list -> Exp.t Record.t,
		       fields: Type.t Srecord.t -> Field.t list),
	    tuple): Exp.t =
      let val unwrap = fn (ty, e) => wrap(ty, e, tuple, record)
	  val wrap = fn (ty, e) => wrap(ty, e, record, tuple)
      in if not(containsRecord ty)
	    then e
	 else
	    (case ty of
		Type.Con(c, [t]) =>
		   if String.equals("option", Layout.toString(Tycon.layout c))
		      then
			 let val x = newLongvid()
			 in Exp.casee
			    (e, [(Pat.longvid none, Exp.longvid none),
				 (Pat.make(Pat.App(someCon, Pat.longvid x)),
				  Exp.app(Exp.longvid someVid,
					  wrap(t, Exp.longvid x)))])
			 end
		   else error()
	      | Type.Con(c, [t1, t2]) =>
		   if Tycon.equals(c, Tycon.arrow)
		      then
			 let val x = newLongvid()
			 in Exp.fnn
			    [(Pat.longvid x,
			      wrap(t2, Exp.app(e, unwrap(t1, Exp.longvid x))))]
			 end
		   else error()
	      | Type.Record r =>
		   let val fs = fields r
		      val temps = List.map(fs, fn _ => newLongvid())
		   in Exp.make
		      (Exp.Let
		       (Dec.make(Dec.Val
				 ([], [(Pat.make
					(Pat.Record
					 {flexible = false,
					  items = 
					  List.map2(fs, temps, fn (f, t) =>
						    Pat.Item.Field
						    (f, Pat.longvid t))}),
					e)])),
			Exp.record
			(makeRecord
			 (r, List.map2(Srecord.toList r, temps,
				       fn ((_, t), x) =>
				       wrap(t, Exp.longvid x))))))
		   end
	      | _ => error())
      end

   val wrap = fn (ty, e) =>
      wrap(ty, e,
	   (fn (_, es) => Record.tuple es,
	    fn r => List.map(Srecord.toList r, #1)),
	   (fn (r, es) => Record.fromList(List.map2
					  (Srecord.toList r, es, fn ((f, _), e) =>
					   (f, e))),
	    fn r => List.mapi(Srecord.toList r, fn (i, _) => Field.Int(i + 1))))

(*    val wrap =
 *       Trace.trace2("wrap", Type.layout, Exp.layout, Exp.layout) wrap
 *)

   fun makeWrapper(ty: Type.t, x: Avid.t): Dec.t =
      let val x = Longvid.short x
      in Dec.make(Dec.Val([], [(Pat.longvid x, wrap(ty, Exp.longvid x))]))
      end

fun identify(E: t) =
   let
(*       val traceLoop = Trace.trace("identify", fn (E, _) => layout E, layout)
 *)
      (* keep track of what has been renamed so we can give things the
       * shortest possible name
       *)

      local
	 fun 'a rename(toString: 'a -> string,
		       setName: 'a * string -> unit,
		       plist: 'a -> PropertyList.t) =
	    let
	       val {get, set, destroy} =
		  Property.newDest(plist, Property.initConst false)
	    in {destroy = destroy,
		rename = fn (x, name) =>
		if get x
		   then if String.size name < String.size(toString x)
			   then setName(x, name)
			else ()
		else (set(x, true) ; setName(x, name))}
	    end
      in
	 val {destroy = destroyTycon, rename = renameTycon} =
	    rename(Tycon.toString, Tycon.setName, Tycon.plist)
	 val {destroy = destroyCon, rename = renameCon} =
	    rename(Con.toString, Con.setName, Con.plist)
      end

      val datatypes = ref []
      val exceptions = ref []

      fun loop arg: t * Ast.Strdec.t list =
	 (fn (E, strids: Strid.t list) =>
	  let
	     val wrappers: Ast.Strdec.t list ref = ref []
	     fun addWrapper(d: Ast.Strdec.t): unit = ListRef.push(wrappers, d)
		
	     fun longvid x = Longvid.long(strids, x)
	     fun ident(d, r) =
		case (d, r) of
		   (Domain.Fixity _, Fixity _) => r
		 | (Domain.Vid x, Vid{scheme, vid}) =>
		      Vid
		      {scheme = scheme,
		       vid = 
		       let 
			  fun con c =
			     (renameCon(c, Longvid.toString(longvid x)) ;
			      vid)
		       in case vid of
			  Vid.Var y =>
			     (case scheme of
				 NONE =>
				    (let open Layout
				     in output(Var.layout y, Out.error)
				     end ;
				     Error.error "identify encountered var without scheme")
			       | SOME s =>
				    let
				       val p =
					  Prim.new(Longvid.toString(longvid x), s)
				       val ty = Scheme.ty s
				    in if containsRecord ty
					  then addWrapper(Strdec.make
							  (Strdec.Core
							   (makeWrapper(ty, x))))
				       else () ;
				       Vid.Prim p
				    end)
			| Vid.Prim p =>
 			     (Control.messageStr
			      ("identify encountered primitive "
			       ^ Layout.toString(Prim.layout p))
			      ; Vid.Prim p)
			| Vid.ConAsVar c => con c
			| Vid.Con c => con c
			| Vid.Exn c =>
			     (case scheme of
				 SOME s =>
				    ListRef.push
				    (exceptions,
				     {con = c,
				      arg = (case (Type.dearrowOpt
						   (Scheme.apply(s, []))) of
						NONE => NONE
					      | SOME(t, _) => SOME t)})
			       | NONE => ()
			      ; con c)
		       end}
		 | (Domain.Tycon c, Tycon typeStr) =>
		      let val (typeFcn, cons) = TypeStr.dest typeStr
			 val cons = Cons.toList cons
		      in (case typeFcn of
			     TypeFcn.Fun _ => ()
			   | TypeFcn.Tycon tycon =>
				(case cons of
				    [] => ()
				  | {scheme = Scheme.T{tyvars, ...}, ...} :: _ =>
				       ListRef.push
				       (datatypes,
					let
					   val tyvars =
					      List.map(tyvars, fn a =>
						       Tyvar.newNoname
						       {equality =
							Tyvar.isEquality a})
					   val tys = List.map(tyvars, Type.var)
					   fun patchScheme s =
					      case Type.dearrowOpt(Scheme.apply(s, tys)) of
						 NONE => NONE
					       | SOME(t, _) => SOME t
					in {tycon = tycon,
					    tyvars = tyvars,
					    cons =
					    List.map(cons, fn {con, scheme, ...} =>
						     {con = con,
						      arg = patchScheme scheme})}
					end);
				 renameTycon
				 (tycon, (Longtycon.toString
					  (Longtycon.long(strids, c))))));
			 r
		      end
		 | (Domain.Strid s, Strid E) =>
		      let val (E, decs) = loop(E, strids @ [s])
		      in (case decs of
			     [] => ()
			   | _ =>
				addWrapper
				(Ast.Strdec.structuree
				 [{name = s,
				   def = Strexp.structt
				   (Strdec.seq
				    (Strdec.openn[Ast.Longstrid.short s]
				     :: decs)),
				   constraint = Ast.SigConst.None}])) ;
			 Strid E
		      end
		 | _ => Error.bug "strange (domain, range)"
	     val table = Table.mapi ident (consolidate'(elements E))
	  in (unconstrained(Table table), !wrappers)
	  end) arg

      val (E, strdecs) = loop(E, [])
	 
   in destroyCon();
      destroyTycon();
      
      (addEquals E,
       Ast.Program.T(List.map(strdecs, Topdec.strdec)),
       CoreML.Constructors.T{datatypes = !datatypes,
			     exceptions = !exceptions})
   end
end (* local *)

fun E + E' = plus[E, E']

end (* functor ElaborateEnv *)
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_STRUCTS = 
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      sharing Ast = CoreML.Ast
   end

signature ELABORATE = 
   sig
      include ELABORATE_STRUCTS

      structure Basis: BASIS
      structure Decs: DECS
      sharing Decs = Basis.Decs
      sharing Ast = Basis.Ast
      sharing CoreML = Decs.CoreML

      (* if true, then opaque signature matching is used *)
      val useOpaque: bool ref

      structure Accum :
	 sig
	    datatype t = T of {decs: Decs.t,
			       basis: Basis.t}

	    val empty: t
	    val prim: unit -> t
	    val decs: t -> Decs.t
	    val basis: t -> Basis.t
	    val layoutDecs: t -> Layout.t
	 end
	 
      val elaborateProgram :
	 Ast.Program.t * Basis.t * Accum.t -> Basis.t * Accum.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Decs(S: DECS_STRUCTS): DECS =
struct

open S

type dec = CoreML.Dec.t

open AppendList

type t = dec t

val toList = toList
   
fun add(ds, d) = append(ds, single d)

val fromDec = single

fun toAsts ds = List.map(toList ds, CoreML.Dec.toAst)

fun toAst ds = Ast.Dec.make(Ast.Dec.SeqDec(toAsts ds))

local open Layout
in fun layout ds = align(List.map(toAsts ds, Ast.Dec.layout))
end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Basis(S: BASIS_STRUCTS): BASIS = 
struct

open S

local open Ast
in structure Strid = Strid
end
local open Env
in structure Interface = Interface
end

structure SigEnv = MonoEnv(structure Domain = Ast.Sigid
			   structure Range = Interface)
   
structure FctEnv = PolyEnv(structure Domain = Ast.Fctid)

datatype t = Basis of {E: Env.t,
		       G: SigEnv.t,
		       F: functorClosure FctEnv.t}
and functorClosure = T of {arg: Strid.t,
			   argSig: Interface.t,
			   body: Ast.Strexp.t,
			   basis: t,
			   counter: int ref}
fun layout(Basis{E, G, ...}) =
   let open Layout
   in record[("E", Env.layout E),
	     ("G", SigEnv.layout G)]
   end   

val empty = Basis{E = Env.empty,
		  G = SigEnv.empty,
		  F = FctEnv.empty()}
val bogus = empty

structure FunctorClosure =
   struct
      type basis = t
      datatype t = datatype functorClosure

      val bogus = T{arg = Strid.bogus,
		    argSig = Interface.bogus,
		    body = Ast.Strexp.var Ast.Longstrid.bogus,
		    basis = bogus,
		    counter = ref 0}
      val equals = fn _ => Error.unimplemented "FunctorClosure.equals"
      val layout = fn _ => Layout.str "<functor closure>"
   end

val op + =
   fn (Basis{E, F, G}, Basis{E=E', F=F', G=G'}) =>
   Basis{E = Env.+(E, E'),
	 G = SigEnv.+(G, G'),
	 F = FctEnv.+(F, F')}

fun toEnv(Basis{E, ...}) = E
fun fromEnv E = Basis{E = E, F = FctEnv.empty(), G = SigEnv.empty}

fun prim() =
   let val (E, cs) = Env.prim()
   in (fromEnv E, cs)
   end
   
fun consolidate(Basis{E, F, G}) =
   Basis{F = F, G = G, E = Env.consolidate E}

fun plusEnv(Basis{E, G, F}, E') = Basis{E = Env.+(E, E'), G = G, F = F}

fun replaceEnv(Basis{G, F, ...}, E) = Basis{E = E, G = G, F = F}

fun extendStrid(Basis{E, G, F}, x, y) =
   Basis{G = G, F = F, E = Env.extendStrid(E, x, y)}

fun extendFctid(Basis{E, F, G}, x, y) =
   Basis{E = E, G = G, F = FctEnv.extend(F, x, y)}

fun extendSigid(Basis{E, F, G}, x, y) =
   Basis{E = E, F = F, G = SigEnv.extend(G, x, y)}

fun lookupLongstrid(Basis{E, ...}, x) = Env.lookupLongstrid(E, x)
fun lookupFctid(Basis{F, ...}, x) =
   case FctEnv.peek(F, x) of
      SOME fct => fct
    | NONE => (Ast.Fctid.unbound x ;
	       FunctorClosure.bogus)
    
fun lookupSigid(Basis{G, ...}, x) =
   case SigEnv.peek(G, x) of
      SOME I => I
    | NONE => (Ast.Sigid.unbound x ;
	       Interface.bogus)

end

(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Elaborate(S: ELABORATE_STRUCTS): ELABORATE= 
struct

open S

local open Ast
in structure FctArg = FctArg
   structure Longstrid = Longstrid
   structure Topdec = Topdec
   structure SigConst = SigConst
   structure Sigexp = Sigexp
   structure Strdec = Strdec
   structure Strid = Strid
   structure Strexp = Strexp
end

local open CoreML
in structure Con = Con
   structure Prim = Prim
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
end

structure Decs = Decs(structure Ast = Ast
		      structure CoreML = CoreML)

structure TypeStr = TypeStr(structure Ast = Ast
			    structure Tycon = Tycon
			    structure Con = Con
			    structure Type =
			       struct
				  open Type
				  datatype t = datatype t'
			       end
			    structure Scheme = Scheme)

structure Interface = Interface(structure Ast = Ast
				structure CoreML = CoreML
				structure TypeStr = TypeStr)

structure Env = ElaborateEnv(structure Ast = Ast
			     structure CoreML = CoreML
			     structure Interface = Interface
			     structure TypeStr = TypeStr)

structure ElaborateAtoms =
   ElaborateAtoms(structure Ast = Ast
		  structure Decs =
		     struct
			open Decs
			val datatypes = single o CoreML.Dec.Datatype
		     end
		  structure Env = Env
		  structure TypeStr = TypeStr)
structure Lookup = ElaborateAtoms.Lookup

structure Basis = Basis(structure Ast = Ast
			structure Decs = Decs
			structure Env = Env
			structure Interface = Interface)
   
local open Basis
in structure FunctorClosure = FunctorClosure
end

structure ElaborateSigexp = ElaborateSigexp(structure Ast = Ast
					    structure Basis = Basis
					    structure Interface = Interface)
local open ElaborateSigexp
in val elaborateSigexp = elaborateSigexp
end

structure ElaborateCore = ElaborateCore(structure Ast = Ast
					structure CoreML = CoreML
					structure Decs = Decs
					structure ElaborateAtoms = ElaborateAtoms
					structure Env = Env)

val counters: int ref list ref = ref []
fun clearCounters() = counters := []
fun newCounter() = let val r = ref 0
		   in counters := r :: (!counters) ; r
		   end

val useOpaque = ref false
   
fun elaborateSigexpConstraint(cons: SigConst.t, B: Basis.t): Env.t -> Env.t =
   let
      fun s(sigexp, opaque) =
	 let val interface = elaborateSigexp(sigexp, B)
	 in fn E => Env.cut{env = E,
			    interface = interface,
			    opaque = opaque,
			    region = Sigexp.region sigexp}
	 end
   in case cons of
      SigConst.None => (fn E => E)
    | SigConst.Transparent sigexp => s(sigexp, false)
    | SigConst.Opaque sigexp => s(sigexp, !useOpaque)
   end	 
   
(* Elaborate d in basis B, accumulating core decs onto decs
 * and accumulating bindings into env.
 *)

(* val traceElaborateStrdec =
 *    Trace.trace4("elaborateStrdec" , Strdec.layout,
 * 		Layout.ignore: Basis.t -> Layout.t (*Basis.layout*),
 * 		Layout.ignore: Decs.t -> Layout.t (*Decs.layout*),
 * 		Layout.ignore: Env.t -> Layout.t (*Env.layout*),
 * 		Layout.ignore: Decs.t * Env.t -> Layout.t)
 * 		(*Layout.tuple2(Decs.layout, Env.layout))*)
 *)
   
fun elaborateStrdec arg: Decs.t * Env.t =
(*   traceElaborateStrdec *)
   (fn (d: Strdec.t, B: Basis.t, decs, E: Env.t) =>
    case Strdec.node d of
       Strdec.Core d => (* rule 56 *)
	  ElaborateCore.elaborateDec(d, Basis.toEnv B, decs, E)
     | Strdec.Structure strbinds => (* rules 57, 61 *)
	  List.fold
	  (strbinds, (decs, E), fn ({name, def, constraint}, (decs, E)) =>
	   let val (decs, E') = elaborateStrexp(def, B, decs)
	   in (decs,
	       Env.extendStrid
	       (E, name, elaborateSigexpConstraint(constraint, B) E'))
	   end)
     | Strdec.Local(d, d') => (* rule 58 *)
	  let val (decs, E') = elaborateStrdec(d, B, decs, Basis.toEnv B)
	  in elaborateStrdec(d', Basis.replaceEnv(B, E'), decs, E)
	  end
     | Strdec.Seq ds => (* rule 60 *)
	  let
	     val (_, decs, E) =
		List.fold
		(ds, (B, decs, E), fn (d, (B, decs, E)) =>
		 let val (decs, E') = elaborateStrdec(d, B, decs, Env.empty)
		 in (Basis.plusEnv(B, E'), decs, Env.+(E, E'))
		 end)
	  in (decs, E)
	  end
       ) arg

and elaborateStrexp(e: Strexp.t, B: Basis.t, decs): Decs.t * Env.t =
   case Strexp.node e of
      Strexp.Var p => (* rule 51 *)
	 (decs, Basis.lookupLongstrid(B, p))
    | Strexp.Struct d => (* rule 50 *)
	 elaborateStrdec(d, B, decs, Env.empty)
    | Strexp.Constrained(e, c) => (* rules 52, 53 *)
	 let val (decs, E) = elaborateStrexp(e, B, decs)
	 in (decs, elaborateSigexpConstraint(c, B) E)
	 end
    | Strexp.App(fctid, strexp) => (* rules 54, 154 *)
	 let
	    val (decs, E) = elaborateStrexp(strexp, B, decs)
	    val FunctorClosure.T{arg, argSig, body, basis=B', counter} =
	       Basis.lookupFctid(B, fctid)
	 in IntRef.inc counter ;
	    elaborateStrexp
	    (body, 
	     Basis.extendStrid(B', arg,
			       Env.cut{env = E,
				       interface = argSig,
				       opaque = false,
				       region = Strexp.region strexp}),
	     decs)
	 end
    | Strexp.Let(d, e) => (* rule 55 *)
	 let val (decs, E) = elaborateStrdec(d, B, decs, Basis.toEnv B)
	 in elaborateStrexp(e, Basis.replaceEnv(B, E), decs)
	 end

fun elaborateTopdec(d: Topdec.t, B: Basis.t, decs, B': Basis.t) =
   case Topdec.node d of
      Topdec.Strdec d =>
	 let val (decs, E) = elaborateStrdec(d, B, decs, Basis.toEnv B')
	 in (decs, Basis.replaceEnv(B', E))
	 end
    | Topdec.Signature sigbinds =>
	 (decs,
	  List.fold(sigbinds, B', fn ((sigid, sigexp), B') =>
		    Basis.extendSigid(B', sigid, elaborateSigexp(sigexp, B))))
    | Topdec.Functor funbinds =>
	 (* Appendix A, p.58 *)
	 (decs,
	  List.fold(funbinds, B', fn ({name, arg, result, body}, B') =>
		    let val body = Strexp.constrained(body, result)
		       val (arg, argSig, body) =
			  case FctArg.node arg of
			     FctArg.Structure(arg, argSig) => (arg, argSig, body)
			   | FctArg.Spec spec =>
				let val strid = Strid.fromString "ZZZNewStridZZZ"
				in (strid,
				    Sigexp.spec spec,
				    Strexp.lett
				    (Strdec.openn[Longstrid.short strid],
				     body))
				end
		    in Basis.extendFctid
		       (B', name,
			FunctorClosure.T{arg = arg,
					 argSig = elaborateSigexp(argSig, B),
					 body = body,
					 basis = B,
					 counter = newCounter()})
		    end))

(* val elaborateTopdec =
 *    Trace.trace4("elaborateTopdec",
 * 		Topdec.layout,
 * 		Layout.ignore (*Basis.layout*),
 * 		Layout.ignore,
 * 		Layout.ignore (*Basis.layout*),
 * 		Layout.ignore (*Layout.tuple2(Decs.layout, Layout.ignore)*))
 *    elaborateTopdec
 *)

structure Accum =
   struct
      datatype t = T of {decs: Decs.t,
			 basis: Basis.t}

      val empty = T{decs = Decs.empty,
		    basis = Basis.empty}

      fun prim() = T{decs = Decs.empty,
		     basis = #1(Basis.prim())}
	 
      fun decs(T{decs, ...}) = decs
      fun basis(T{basis, ...}) = basis

      val layoutDecs = Decs.layout o decs

      val layout = Basis.layout o basis
   end

	  
fun elaborateProgram(Ast.Program.T decs, B: Basis.t,
		     Accum.T{decs = decs', basis = B'}) =
   let
      val (B, decs, B') =
	 List.fold(decs, (B, decs', B'), fn (d, (B, decs, B')) =>
		   let val (decs, B'') = elaborateTopdec(d, B, decs, Basis.empty)
		   in (Basis.+(B, B''), decs, Basis.+(B', B''))
		   end)
   in (B,
       Accum.T{decs = decs,
	       basis = B'})
   end

end
signature ML_TOKENS =
sig
type ('a,'b) token
type svalue
val PRIM:  'a * 'a -> (svalue,'a) token
val FUNSIG:  'a * 'a -> (svalue,'a) token
val ANDALSO:  'a * 'a -> (svalue,'a) token
val ORELSE:  'a * 'a -> (svalue,'a) token
val RPAREN:  'a * 'a -> (svalue,'a) token
val RBRACKET:  'a * 'a -> (svalue,'a) token
val RBRACE:  'a * 'a -> (svalue,'a) token
val LPAREN:  'a * 'a -> (svalue,'a) token
val LBRACKET:  'a * 'a -> (svalue,'a) token
val LBRACE:  'a * 'a -> (svalue,'a) token
val COMMA:  'a * 'a -> (svalue,'a) token
val COLONGT:  'a * 'a -> (svalue,'a) token
val COLON:  'a * 'a -> (svalue,'a) token
val ASTERISK:  'a * 'a -> (svalue,'a) token
val WITHTYPE:  'a * 'a -> (svalue,'a) token
val WITH:  'a * 'a -> (svalue,'a) token
val WILD:  'a * 'a -> (svalue,'a) token
val WHILE:  'a * 'a -> (svalue,'a) token
val WHERE:  'a * 'a -> (svalue,'a) token
val VAL:  'a * 'a -> (svalue,'a) token
val TYPE:  'a * 'a -> (svalue,'a) token
val THEN:  'a * 'a -> (svalue,'a) token
val STRUCTURE:  'a * 'a -> (svalue,'a) token
val STRUCT:  'a * 'a -> (svalue,'a) token
val SIGNATURE:  'a * 'a -> (svalue,'a) token
val SIG:  'a * 'a -> (svalue,'a) token
val SHARING:  'a * 'a -> (svalue,'a) token
val REC:  'a * 'a -> (svalue,'a) token
val RAISE:  'a * 'a -> (svalue,'a) token
val OVERLOAD:  'a * 'a -> (svalue,'a) token
val OPEN:  'a * 'a -> (svalue,'a) token
val OP:  'a * 'a -> (svalue,'a) token
val OF:  'a * 'a -> (svalue,'a) token
val NONFIX:  'a * 'a -> (svalue,'a) token
val LOCAL:  'a * 'a -> (svalue,'a) token
val LET:  'a * 'a -> (svalue,'a) token
val INFIXR:  'a * 'a -> (svalue,'a) token
val INFIX:  'a * 'a -> (svalue,'a) token
val INCLUDE:  'a * 'a -> (svalue,'a) token
val IN:  'a * 'a -> (svalue,'a) token
val IF:  'a * 'a -> (svalue,'a) token
val HASH:  'a * 'a -> (svalue,'a) token
val HANDLE:  'a * 'a -> (svalue,'a) token
val FUNCTOR:  'a * 'a -> (svalue,'a) token
val FUN:  'a * 'a -> (svalue,'a) token
val FN:  'a * 'a -> (svalue,'a) token
val DARROW:  'a * 'a -> (svalue,'a) token
val DO:  'a * 'a -> (svalue,'a) token
val EXCEPTION:  'a * 'a -> (svalue,'a) token
val EQTYPE:  'a * 'a -> (svalue,'a) token
val EQUALOP:  'a * 'a -> (svalue,'a) token
val END:  'a * 'a -> (svalue,'a) token
val ELSE:  'a * 'a -> (svalue,'a) token
val DOTDOTDOT:  'a * 'a -> (svalue,'a) token
val DATATYPE:  'a * 'a -> (svalue,'a) token
val CASE:  'a * 'a -> (svalue,'a) token
val BAR:  'a * 'a -> (svalue,'a) token
val AS:  'a * 'a -> (svalue,'a) token
val ARROW:  'a * 'a -> (svalue,'a) token
val AND:  'a * 'a -> (svalue,'a) token
val ABSTYPE:  'a * 'a -> (svalue,'a) token
val CHAR: (char) *  'a * 'a -> (svalue,'a) token
val STRING: (string) *  'a * 'a -> (svalue,'a) token
val REAL: (string) *  'a * 'a -> (svalue,'a) token
val WORD: (Word32.word) *  'a * 'a -> (svalue,'a) token
val INT: (string) *  'a * 'a -> (svalue,'a) token
val TYVAR: (string) *  'a * 'a -> (svalue,'a) token
val LONGID: (string) *  'a * 'a -> (svalue,'a) token
val SEMICOLON:  'a * 'a -> (svalue,'a) token
val EOF:  'a * 'a -> (svalue,'a) token
end
signature ML_LRVALS=
sig
structure Tokens : ML_TOKENS
structure ParserData:PARSER_DATA
sharing type ParserData.Token.token = Tokens.token
sharing type ParserData.svalue = Tokens.svalue
end

functor MLLrValsFun(structure Token : TOKEN
			     structure Ast : AST) = 
struct
structure ParserData=
struct
structure Header = 
struct
(* Heavily modified from SML/NJ sources by sweeks@research.nj.nec.com *)

(* ml.grm
 *
 * Copyright 1989,1992 by AT&T Bell Laboratories
 *)

fun reg(left,right) = Region.T{left = left, right = right}
fun error(reg,msg) = Control.error(reg, Layout.str msg)

open Ast
structure Field = Record.Field
structure Srecord = SortedRecord

structure Type =
   struct
      open Type
      val tuple = Record o Srecord.tuple
      val unit = tuple[]

      fun arrow(t1,t2) = Con(Longtycon.arrow, [t1,t2])
   end

structure Pat =
   struct
      open Pat
      val tuple =
	 fn [p] => node p
	  | ps => Tuple ps
      val unit = tuple[]

      val bogus = unit

      fun var id =
	 Var{name = Longvid.short(Vid.fromString id),
	     fixop = Fixop.None}

      fun makeAs(p1 : t, p2 : t) : node =
	 let
	    fun err() =
	       (error(Pat.region p1, "must have variable to left in as pattern") ;
		bogus)

	    fun fixopVar(p : t) =
	       case node p of
		  FlatApp[p] =>
		     (case node p of
			 Var{fixop,name} =>
			    (case Longvid.split name of
				([],vid) =>
				   SOME(fixop, Vid.toVar vid)
			      | _ =>
				   (err() ;
				    SOME(Fixop.None,Var.bogus)))
		       | _ => NONE)
		| _ => NONE
		     
	 in case fixopVar p1 of
	    SOME(fixop,var) =>
	       Layered{fixop = fixop, var = var,
		       constraint = NONE,
		       pat = p2}
	  | NONE =>
	       case node p1 of
		  Pat.Constraint(p,t) =>
		     (case fixopVar p of
			 SOME(fixop,var) =>
			    Layered{fixop = fixop, var = var,
				    constraint = SOME t,
				    pat = p2}
		       | _ => err())
		| _ => err()
	 end
   end

structure Exp =
   struct
      open Exp
      val tuple =
	 fn [e] => node e
	  | es => Record(Record.tuple es)
      val unit = tuple[]
   end

structure Dec =
   struct
      open Dec
      fun sequence(d1 : t, d2 : t) : t =
	 makeRegion'(case (node d1, node d2) of
			(SeqDec d1, SeqDec d2) => SeqDec(d1@d2)
		      | (SeqDec d1, _) => SeqDec(d1@[d2])
		      | (_, SeqDec d2) => SeqDec(d1::d2)
		      | _ => SeqDec[d1,d2],
		     left d1, right d2)
   end

structure Spec =
   struct
      open Spec

      (* some of this mess is so that a sharing equation captures as
       many specs as possible in its scope *)
      fun seq(s: t, s': t): t =
	 let fun reg s'' = makeRegion'(s'', left s, right s')
	 in case (node s, node s') of
	    (Empty,  _)     => s'
	  | (_,      Empty) => s
	  | (_,      Seq(s1, s2)) => reg(Seq(seq(s, s1), s2))
	  | (_,      Sharing{spec, equations}) =>
	       reg(Sharing{spec = seq(s, spec), equations = equations})
	  | _               => reg(Seq(s, s'))
	 end

(*      val seq = Trace.trace2("Spec.seq", layout, layout, layout) seq *)
   end

type rule = Pat.t * Exp.t
type rules = rule list
type clause = {pats : Pat.t list,
	       resultType : Type.t option,
	       body : Exp.t}
type eb = Con.t * EbRhs.t
type db = {tyvars : Tyvar.t list,
	   tycon : Tycon.t,
	   cons : (Con.t * Type.t option) list}

type strdesc = Strid.t * Sigexp.t

type wherespec = {tyvars : Tyvar.t list,
		  longtycon : Longtycon.t,
		  ty : Type.t}

type typdesc =  {tyvars : Tyvar.t list,
		 tycon : Tycon.t,
		 ty : Type.t option}

type valdesc = Var.t * Type.t

type exndesc = Con.t * Type.t option

type strbind = {name: Strid.t,
		def: Strexp.t,
		constraint: SigConst.t}
   
type sigbind = Sigid.t * Sigexp.t

type funbind = {name : Fctid.t,
		arg : FctArg.t,
		result : SigConst.t,
		body : Strexp.t}

fun ensureNonqualified(s,region) =
   if String.contains(s, #".")
      then (error(region, "expected nonqualified id") ;
	    "<bogus>")
   else s

fun cons1(x,(l,r,y)) = (x::l,r,y)

fun augment(id, sigexp, (wherespecs,right,binds)) =
   (id, Sigexp.makeRegion'(Sigexp.Where(sigexp,wherespecs),
			   Sigexp.left sigexp, right))
   :: binds

fun 'a augment1((strexp : Strexp.t,
		 makesigconst : Sigexp.t -> SigConst.t,
		 sigexp : Sigexp.t),
		(wherespecs : wherespec list,
		 right : int,
		 z : 'a)) : Strexp.t * 'a =
   (Strexp.makeRegion'
    (Strexp.Constrained(strexp, makesigconst(Sigexp.makeRegion'
					     (Sigexp.Where(sigexp,wherespecs),
					      Sigexp.left sigexp,
					      right))),
     Strexp.left strexp,
     right),
    z)

type 'a whereAnd = wherespec list * int * 'a list
   
  
end
structure LrTable = Token.LrTable
structure Token = Token
local open LrTable in 
val table=let val actionRows =
"\
\\001\000\001\000\000\000\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\137\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\055\002\059\000\055\002\066\000\052\002\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\237\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\069\002\059\000\069\002\066\000\052\002\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\237\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\072\002\059\000\072\002\066\000\052\002\000\000\
\\001\000\001\000\060\002\002\000\060\002\010\000\060\002\011\000\060\002\
\\016\000\060\002\019\000\060\002\022\000\060\002\026\000\060\002\
\\027\000\060\002\031\000\060\002\033\000\060\002\034\000\060\002\
\\036\000\060\002\037\000\060\002\040\000\060\002\041\000\060\002\
\\046\000\060\002\048\000\060\002\050\000\060\002\051\000\060\002\
\\058\000\057\002\059\000\057\002\066\000\060\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\007\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\055\002\
\\059\000\055\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\022\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\069\002\
\\059\000\069\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\022\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\072\002\
\\059\000\072\002\000\000\
\\001\000\001\000\024\003\002\000\024\003\003\000\025\003\010\000\024\003\
\\011\000\024\003\012\000\024\003\013\000\024\003\014\000\024\003\
\\016\000\024\003\018\000\024\003\019\000\024\003\020\000\024\003\
\\021\000\024\003\022\000\024\003\023\000\024\003\024\000\024\003\
\\026\000\024\003\027\000\024\003\028\000\024\003\031\000\024\003\
\\032\000\024\003\033\000\024\003\034\000\024\003\036\000\024\003\
\\037\000\024\003\038\000\024\003\040\000\024\003\041\000\024\003\
\\044\000\024\003\046\000\024\003\048\000\024\003\049\000\024\003\
\\050\000\024\003\051\000\024\003\052\000\024\003\055\000\024\003\
\\056\000\024\003\057\000\025\003\058\000\024\003\059\000\024\003\
\\060\000\024\003\064\000\024\003\065\000\024\003\066\000\024\003\
\\067\000\024\003\068\000\024\003\000\000\
\\001\000\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\001\000\002\000\231\001\012\000\255\000\000\000\
\\001\000\003\000\034\000\000\000\
\\001\000\003\000\034\000\004\000\058\000\020\000\033\000\039\000\130\000\
\\057\000\032\000\063\000\129\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\017\000\125\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\017\000\125\000\057\000\032\000\
\\064\000\124\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\064\000\008\001\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\064\000\027\001\000\000\
\\001\000\003\000\034\000\016\000\095\001\057\000\032\000\000\000\
\\001\000\003\000\034\000\020\000\033\000\039\000\130\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\020\000\033\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\045\000\217\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\050\000\149\001\057\000\032\000\000\000\
\\001\000\003\000\034\000\057\000\032\000\000\000\
\\001\000\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\043\000\051\000\054\000\050\000\057\000\049\000\061\000\048\000\
\\062\000\047\000\063\000\046\000\000\000\
\\001\000\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\054\000\050\000\057\000\049\000\061\000\048\000\062\000\047\000\
\\063\000\046\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\065\000\023\001\
\\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\066\000\020\001\
\\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\001\000\003\000\059\000\020\000\193\000\057\000\049\000\000\000\
\\001\000\003\000\059\000\035\000\056\001\047\000\055\001\057\000\049\000\000\000\
\\001\000\003\000\059\000\050\000\078\001\057\000\049\000\000\000\
\\001\000\003\000\059\000\057\000\049\000\000\000\
\\001\000\003\000\173\000\000\000\
\\001\000\003\000\173\000\004\000\058\000\061\000\172\000\063\000\171\000\000\000\
\\001\000\003\000\173\000\057\000\001\001\000\000\
\\001\000\004\000\058\000\000\000\
\\001\000\008\000\016\001\000\000\
\\001\000\012\000\255\000\013\000\063\001\000\000\
\\001\000\012\000\255\000\060\000\106\001\066\000\105\001\000\000\
\\001\000\013\000\056\003\020\000\063\003\058\000\056\003\060\000\056\003\
\\064\000\056\003\000\000\
\\001\000\013\000\112\000\020\000\111\000\058\000\110\000\000\000\
\\001\000\013\000\112\000\024\000\127\001\058\000\110\000\000\000\
\\001\000\018\000\233\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\019\000\064\001\000\000\
\\001\000\019\000\096\001\000\000\
\\001\000\019\000\150\001\000\000\
\\001\000\019\000\205\001\000\000\
\\001\000\019\000\232\001\000\000\
\\001\000\019\000\254\001\000\000\
\\001\000\019\000\001\002\058\000\143\001\059\000\142\001\000\000\
\\001\000\020\000\139\000\000\000\
\\001\000\020\000\204\000\000\000\
\\001\000\020\000\210\000\000\000\
\\001\000\020\000\212\000\000\000\
\\001\000\020\000\239\000\000\000\
\\001\000\020\000\247\000\000\000\
\\001\000\020\000\254\000\000\000\
\\001\000\020\000\045\001\000\000\
\\001\000\020\000\123\001\000\000\
\\001\000\020\000\133\001\000\000\
\\001\000\020\000\163\001\000\000\
\\001\000\020\000\215\001\000\000\
\\001\000\020\000\224\001\000\000\
\\001\000\020\000\004\002\000\000\
\\001\000\023\000\124\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\025\000\132\001\000\000\
\\001\000\025\000\195\001\000\000\
\\001\000\028\000\014\001\038\000\129\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\028\000\014\001\049\000\126\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\031\000\143\000\000\000\
\\001\000\031\000\125\001\000\000\
\\001\000\031\000\174\001\000\000\
\\001\000\031\000\206\001\000\000\
\\001\000\050\000\149\001\000\000\
\\001\000\055\000\160\000\000\000\
\\001\000\058\000\140\000\000\000\
\\001\000\058\000\067\001\000\000\
\\001\000\058\000\109\001\000\000\
\\001\000\058\000\114\001\000\000\
\\001\000\058\000\143\001\059\000\142\001\000\000\
\\001\000\058\000\143\001\059\000\142\001\066\000\238\001\000\000\
\\001\000\058\000\159\001\000\000\
\\001\000\058\000\162\001\000\000\
\\001\000\063\000\145\000\000\000\
\\001\000\064\000\203\000\000\000\
\\001\000\064\000\107\001\000\000\
\\001\000\064\000\121\001\000\000\
\\001\000\065\000\201\000\000\000\
\\001\000\065\000\119\001\000\000\
\\001\000\066\000\196\000\000\000\
\\001\000\066\000\197\000\000\000\
\\001\000\066\000\211\000\000\000\
\\001\000\066\000\065\001\000\000\
\\001\000\066\000\104\001\000\000\
\\001\000\066\000\115\001\000\000\
\\001\000\066\000\118\001\000\000\
\\001\000\066\000\239\001\000\000\
\\028\002\000\000\
\\029\002\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\030\002\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\031\002\000\000\
\\032\002\000\000\
\\033\002\000\000\
\\034\002\000\000\
\\035\002\000\000\
\\036\002\000\000\
\\037\002\000\000\
\\038\002\002\000\078\000\003\000\059\000\010\000\024\000\016\000\023\000\
\\022\000\022\000\026\000\021\000\033\000\019\000\034\000\018\000\
\\035\000\056\001\036\000\017\000\037\000\016\000\040\000\015\000\
\\041\000\014\000\047\000\055\001\048\000\012\000\050\000\011\000\
\\051\000\010\000\057\000\049\000\000\000\
\\038\002\002\000\078\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\033\000\019\000\034\000\018\000\036\000\017\000\
\\037\000\016\000\040\000\015\000\041\000\014\000\048\000\012\000\
\\050\000\011\000\051\000\010\000\000\000\
\\039\002\000\000\
\\040\002\000\000\
\\041\002\000\000\
\\042\002\000\000\
\\043\002\000\000\
\\044\002\000\000\
\\045\002\000\000\
\\046\002\000\000\
\\047\002\000\000\
\\048\002\000\000\
\\049\002\000\000\
\\050\002\000\000\
\\051\002\000\000\
\\052\002\011\000\137\001\000\000\
\\053\002\000\000\
\\054\002\000\000\
\\055\002\052\000\058\001\000\000\
\\056\002\000\000\
\\057\002\000\000\
\\058\002\000\000\
\\059\002\000\000\
\\061\002\063\000\145\001\000\000\
\\062\002\000\000\
\\063\002\000\000\
\\064\002\000\000\
\\065\002\000\000\
\\066\002\000\000\
\\067\002\052\000\058\001\000\000\
\\068\002\000\000\
\\069\002\011\000\209\001\052\000\058\001\000\000\
\\070\002\000\000\
\\071\002\000\000\
\\072\002\011\000\209\001\052\000\058\001\000\000\
\\073\002\000\000\
\\074\002\000\000\
\\075\002\000\000\
\\076\002\000\000\
\\077\002\000\000\
\\077\002\003\000\034\000\057\000\032\000\000\000\
\\078\002\000\000\
\\079\002\011\000\062\001\052\000\061\001\000\000\
\\079\002\011\000\213\001\052\000\061\001\000\000\
\\080\002\000\000\
\\081\002\000\000\
\\082\002\000\000\
\\083\002\000\000\
\\084\002\012\000\255\000\000\000\
\\085\002\058\000\138\000\059\000\137\000\000\000\
\\086\002\000\000\
\\087\002\000\000\
\\088\002\002\000\236\000\003\000\034\000\016\000\235\000\021\000\234\000\
\\022\000\233\000\032\000\232\000\044\000\231\000\048\000\230\000\
\\050\000\229\000\051\000\228\000\057\000\032\000\000\000\
\\088\002\002\000\236\000\016\000\235\000\021\000\234\000\022\000\233\000\
\\032\000\232\000\044\000\231\000\048\000\230\000\050\000\229\000\
\\051\000\228\000\000\000\
\\089\002\000\000\
\\090\002\000\000\
\\091\002\000\000\
\\092\002\000\000\
\\093\002\000\000\
\\094\002\000\000\
\\095\002\000\000\
\\096\002\000\000\
\\097\002\000\000\
\\098\002\000\000\
\\099\002\000\000\
\\100\002\000\000\
\\101\002\000\000\
\\102\002\000\000\
\\103\002\020\000\163\001\000\000\
\\104\002\000\000\
\\105\002\020\000\224\001\000\000\
\\106\002\000\000\
\\107\002\000\000\
\\108\002\011\000\251\001\052\000\250\001\000\000\
\\108\002\011\000\019\002\052\000\250\001\000\000\
\\109\002\000\000\
\\110\002\000\000\
\\111\002\000\000\
\\112\002\000\000\
\\113\002\011\000\160\001\000\000\
\\114\002\000\000\
\\115\002\000\000\
\\116\002\020\000\220\001\000\000\
\\117\002\012\000\255\000\000\000\
\\118\002\011\000\158\001\000\000\
\\119\002\000\000\
\\120\002\012\000\255\000\000\000\
\\121\002\011\000\168\001\000\000\
\\122\002\000\000\
\\123\002\000\000\
\\124\002\038\000\170\001\000\000\
\\125\002\012\000\255\000\000\000\
\\126\002\000\000\
\\127\002\000\000\
\\128\002\000\000\
\\129\002\000\000\
\\130\002\000\000\
\\131\002\011\000\007\002\000\000\
\\132\002\000\000\
\\133\002\000\000\
\\134\002\000\000\
\\135\002\000\000\
\\136\002\000\000\
\\137\002\002\000\253\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\033\000\019\000\034\000\018\000\036\000\252\000\
\\037\000\016\000\040\000\015\000\041\000\014\000\050\000\011\000\
\\051\000\010\000\000\000\
\\138\002\000\000\
\\139\002\000\000\
\\140\002\000\000\
\\141\002\000\000\
\\142\002\000\000\
\\143\002\011\000\108\000\000\000\
\\144\002\011\000\108\000\000\000\
\\145\002\011\000\207\000\000\000\
\\146\002\011\000\207\000\000\000\
\\147\002\000\000\
\\148\002\000\000\
\\149\002\000\000\
\\150\002\000\000\
\\151\002\000\000\
\\152\002\000\000\
\\153\002\000\000\
\\154\002\000\000\
\\155\002\000\000\
\\156\002\011\000\108\000\000\000\
\\157\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\158\002\058\000\150\000\000\000\
\\159\002\012\000\255\000\000\000\
\\160\002\000\000\
\\161\002\000\000\
\\162\002\011\000\207\000\000\000\
\\163\002\011\000\148\000\000\000\
\\164\002\000\000\
\\165\002\014\000\147\000\000\000\
\\166\002\000\000\
\\167\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\168\002\000\000\
\\169\002\011\000\135\001\012\000\255\000\000\000\
\\170\002\000\000\
\\171\002\000\000\
\\172\002\000\000\
\\173\002\004\000\058\000\063\000\063\000\000\000\
\\174\002\000\000\
\\175\002\000\000\
\\176\002\060\000\200\000\000\000\
\\177\002\000\000\
\\178\002\014\000\171\001\000\000\
\\179\002\000\000\
\\180\002\038\000\172\001\000\000\
\\181\002\012\000\255\000\000\000\
\\182\002\011\000\151\000\000\000\
\\183\002\000\000\
\\184\002\000\000\
\\185\002\000\000\
\\186\002\020\000\155\000\038\000\154\000\000\000\
\\187\002\012\000\255\000\000\000\
\\188\002\000\000\
\\189\002\005\000\081\000\000\000\
\\190\002\000\000\
\\191\002\005\000\081\000\000\000\
\\192\002\000\000\
\\193\002\000\000\
\\194\002\000\000\
\\195\002\000\000\
\\196\002\000\000\
\\197\002\000\000\
\\198\002\000\000\
\\199\002\000\000\
\\200\002\000\000\
\\201\002\000\000\
\\202\002\000\000\
\\203\002\000\000\
\\204\002\000\000\
\\205\002\011\000\158\000\000\000\
\\206\002\000\000\
\\207\002\000\000\
\\208\002\056\000\157\000\000\000\
\\209\002\000\000\
\\210\002\011\000\214\001\000\000\
\\211\002\000\000\
\\212\002\014\000\128\001\000\000\
\\213\002\000\000\
\\214\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\215\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\216\002\000\000\
\\217\002\060\000\122\001\000\000\
\\218\002\002\000\117\001\028\000\014\001\058\000\013\001\060\000\116\001\
\\067\000\012\001\068\000\011\001\000\000\
\\218\002\002\000\117\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\219\002\000\000\
\\220\002\000\000\
\\221\002\000\000\
\\222\002\058\000\013\001\068\000\011\001\000\000\
\\223\002\058\000\013\001\000\000\
\\224\002\012\000\255\000\000\000\
\\225\002\000\000\
\\226\002\000\000\
\\227\002\000\000\
\\228\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\229\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\230\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\231\002\000\000\
\\232\002\000\000\
\\233\002\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\020\000\193\000\029\000\191\000\
\\035\000\189\000\039\000\188\000\057\000\049\000\061\000\185\000\
\\062\000\184\000\063\000\183\000\070\000\182\000\000\000\
\\234\002\000\000\
\\235\002\000\000\
\\236\002\000\000\
\\237\002\000\000\
\\238\002\000\000\
\\239\002\000\000\
\\240\002\000\000\
\\241\002\000\000\
\\242\002\000\000\
\\243\002\000\000\
\\244\002\000\000\
\\245\002\000\000\
\\246\002\000\000\
\\247\002\000\000\
\\248\002\028\000\014\001\058\000\013\001\060\000\116\001\067\000\012\001\
\\068\000\011\001\000\000\
\\249\002\028\000\014\001\058\000\013\001\060\000\120\001\067\000\012\001\
\\068\000\011\001\000\000\
\\250\002\000\000\
\\251\002\000\000\
\\252\002\013\000\112\000\058\000\110\000\000\000\
\\253\002\012\000\255\000\000\000\
\\254\002\000\000\
\\255\002\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\000\003\000\000\
\\001\003\000\000\
\\002\003\000\000\
\\003\003\000\000\
\\004\003\000\000\
\\005\003\000\000\
\\006\003\000\000\
\\007\003\000\000\
\\008\003\000\000\
\\009\003\000\000\
\\010\003\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\054\000\050\000\057\000\049\000\061\000\048\000\062\000\047\000\
\\063\000\107\000\000\000\
\\010\003\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\011\003\000\000\
\\012\003\013\000\112\000\058\000\110\000\060\000\199\000\000\000\
\\013\003\000\000\
\\014\003\000\000\
\\015\003\060\000\202\000\000\000\
\\016\003\000\000\
\\017\003\013\000\112\000\058\000\110\000\000\000\
\\018\003\000\000\
\\019\003\013\000\043\001\000\000\
\\020\003\013\000\112\000\058\000\110\000\000\000\
\\021\003\000\000\
\\022\003\000\000\
\\023\003\012\000\255\000\000\000\
\\025\003\000\000\
\\026\003\000\000\
\\027\003\000\000\
\\028\003\000\000\
\\029\003\000\000\
\\030\003\000\000\
\\031\003\000\000\
\\032\003\000\000\
\\033\003\012\000\255\000\000\000\
\\034\003\000\000\
\\035\003\060\000\108\001\000\000\
\\036\003\000\000\
\\037\003\003\000\173\000\057\000\001\001\000\000\
\\038\003\012\000\255\000\060\000\106\001\000\000\
\\039\003\000\000\
\\040\003\000\000\
\\041\003\000\000\
\\042\003\000\000\
\\043\003\000\000\
\\044\003\000\000\
\\045\003\000\000\
\\046\003\000\000\
\\047\003\000\000\
\\048\003\000\000\
\\049\003\000\000\
\\050\003\000\000\
\\051\003\000\000\
\\052\003\000\000\
\\053\003\000\000\
\\054\003\000\000\
\\055\003\000\000\
\\056\003\000\000\
\\057\003\003\000\034\000\020\000\033\000\057\000\032\000\000\000\
\\058\003\000\000\
\\059\003\000\000\
\\060\003\000\000\
\\061\003\000\000\
\\062\003\000\000\
\\063\003\000\000\
\\064\003\000\000\
\\065\003\000\000\
\\066\003\000\000\
\\067\003\003\000\034\000\057\000\032\000\000\000\
\\068\003\000\000\
\\069\003\000\000\
\\070\003\000\000\
\\071\003\000\000\
\\072\003\000\000\
\\073\003\000\000\
\\074\003\000\000\
\\075\003\000\000\
\\076\003\003\000\059\000\057\000\049\000\000\000\
\\077\003\000\000\
\"
val actionRowNumbers =
"\009\000\099\000\104\000\100\000\
\\113\000\105\000\116\000\020\000\
\\024\000\248\000\023\000\023\000\
\\020\000\033\000\012\001\110\000\
\\010\001\008\001\023\000\025\000\
\\023\000\248\000\248\000\101\000\
\\103\000\229\000\133\001\122\001\
\\131\001\124\001\123\001\125\001\
\\121\001\029\000\218\000\077\001\
\\075\001\074\001\071\001\042\000\
\\078\001\249\000\080\001\149\001\
\\086\001\087\001\014\000\127\001\
\\081\001\012\000\023\000\120\001\
\\119\001\118\001\117\001\116\001\
\\138\001\126\001\011\000\224\000\
\\246\000\037\000\114\000\158\000\
\\141\001\106\000\052\000\142\001\
\\077\000\135\001\228\000\152\001\
\\151\001\110\000\108\000\071\000\
\\110\000\014\001\011\001\013\001\
\\009\001\107\000\085\000\145\001\
\\029\000\240\000\238\000\222\000\
\\233\000\001\001\227\000\005\001\
\\136\001\027\001\024\001\018\001\
\\017\001\015\001\225\000\011\000\
\\076\000\011\000\102\000\134\001\
\\219\000\087\001\029\000\076\001\
\\035\000\028\000\029\000\091\000\
\\092\000\089\001\251\000\089\000\
\\092\001\086\000\053\000\233\000\
\\041\000\140\001\084\001\093\001\
\\019\000\220\000\233\000\037\000\
\\020\000\079\001\132\001\054\000\
\\137\001\093\000\055\000\021\000\
\\021\000\021\000\035\000\153\001\
\\112\000\110\000\111\000\161\000\
\\223\000\029\000\029\000\056\000\
\\035\000\023\000\004\001\003\001\
\\035\000\033\000\022\001\248\000\
\\248\000\057\000\212\000\058\000\
\\231\000\098\001\073\001\099\001\
\\008\000\036\000\108\001\102\001\
\\146\001\035\000\016\000\128\001\
\\053\001\045\001\040\001\232\000\
\\053\001\056\001\148\001\129\001\
\\038\000\027\000\026\000\017\000\
\\028\000\028\000\020\000\212\000\
\\028\000\015\000\029\000\130\001\
\\028\000\072\001\250\000\082\001\
\\088\001\029\000\037\000\083\001\
\\013\000\085\001\029\000\096\001\
\\221\000\019\000\059\000\233\000\
\\035\000\247\000\031\000\147\000\
\\138\000\160\000\148\000\162\000\
\\159\000\151\000\039\000\045\000\
\\094\000\174\000\165\000\162\000\
\\211\000\078\000\020\000\248\000\
\\023\000\032\000\021\000\023\000\
\\248\000\248\000\162\000\241\000\
\\239\000\028\000\234\000\002\001\
\\006\001\007\001\150\001\028\001\
\\025\001\018\000\046\000\216\000\
\\215\000\212\000\212\000\212\000\
\\023\000\035\000\107\001\035\000\
\\095\000\040\000\087\000\111\001\
\\079\000\139\001\104\001\051\001\
\\054\001\028\000\028\000\035\000\
\\029\000\052\001\080\000\096\000\
\\037\001\097\000\060\001\090\000\
\\069\001\064\001\088\000\036\001\
\\060\000\059\001\066\000\050\001\
\\055\001\072\000\070\000\057\001\
\\043\000\031\001\046\001\069\000\
\\089\001\252\000\091\001\094\001\
\\095\001\029\000\237\000\067\000\
\\061\000\244\000\004\000\124\000\
\\001\000\126\000\081\000\117\000\
\\132\000\110\000\110\000\139\000\
\\075\000\047\000\146\000\075\000\
\\023\000\030\000\115\000\158\000\
\\164\000\021\000\193\000\166\000\
\\083\000\188\000\167\000\011\000\
\\171\000\084\000\176\000\062\000\
\\034\000\172\000\149\000\196\000\
\\170\000\199\000\168\000\023\001\
\\020\001\019\001\016\001\169\000\
\\163\000\242\000\253\000\026\001\
\\255\000\034\000\226\000\213\000\
\\073\000\214\000\100\001\112\001\
\\101\001\113\001\034\000\106\001\
\\035\000\103\001\015\000\035\000\
\\043\001\042\001\044\001\041\001\
\\035\000\062\001\028\000\028\000\
\\061\001\063\001\028\000\058\001\
\\015\000\028\000\028\000\028\000\
\\028\000\028\000\029\000\029\000\
\\090\001\097\001\029\000\068\000\
\\243\000\248\000\119\000\023\000\
\\128\000\120\000\118\000\075\000\
\\021\000\021\000\134\000\109\000\
\\048\000\074\000\140\000\248\000\
\\150\000\152\000\153\000\230\000\
\\029\001\147\001\063\000\210\000\
\\020\000\035\000\248\000\191\000\
\\021\000\033\000\175\000\064\000\
\\173\000\143\001\023\000\198\000\
\\035\000\023\000\035\000\021\001\
\\212\000\105\001\115\001\114\001\
\\110\001\109\001\010\000\067\001\
\\068\001\038\001\039\001\070\001\
\\035\001\034\001\048\001\049\000\
\\044\000\033\001\032\001\047\001\
\\235\000\029\000\245\000\125\000\
\\002\000\131\000\130\000\129\000\
\\127\000\082\000\098\000\133\000\
\\031\000\142\000\141\000\075\000\
\\034\000\154\000\155\000\022\000\
\\030\000\031\000\194\000\195\000\
\\189\000\190\000\035\000\182\000\
\\178\000\177\000\034\000\144\001\
\\197\000\200\000\254\000\000\001\
\\050\000\066\001\065\001\028\000\
\\236\000\121\000\122\000\022\000\
\\136\000\137\000\051\000\143\000\
\\065\000\152\000\030\001\206\000\
\\005\000\201\000\192\000\181\000\
\\075\000\023\000\180\000\179\000\
\\217\000\049\001\003\000\135\000\
\\144\000\145\000\035\000\156\000\
\\203\000\023\000\204\000\202\000\
\\075\000\183\000\184\000\123\000\
\\157\000\207\000\006\000\185\000\
\\186\000\022\000\205\000\208\000\
\\022\000\183\000\007\000\187\000\
\\209\000\000\000"
val gotoT =
"\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\001\000\148\000\025\002\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\024\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\026\000\
\\012\000\025\000\000\000\
\\007\000\043\000\009\000\042\000\016\000\041\000\028\000\040\000\
\\054\000\039\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\035\000\066\000\034\000\075\000\033\000\000\000\
\\016\000\060\000\072\000\059\000\074\000\058\000\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\115\000\062\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\120\000\065\000\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\000\000\000\
\\007\000\072\000\029\000\071\000\030\000\070\000\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\075\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\001\000\078\000\002\000\077\000\000\000\
\\001\000\080\000\002\000\077\000\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\125\000\081\000\000\000\
\\007\000\043\000\009\000\042\000\016\000\041\000\028\000\040\000\
\\056\000\088\000\057\000\036\000\058\000\035\000\069\000\087\000\
\\070\000\086\000\071\000\085\000\075\000\084\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\091\000\079\000\090\000\
\\080\000\089\000\000\000\
\\016\000\060\000\074\000\099\000\131\000\098\000\133\000\097\000\
\\135\000\096\000\136\000\095\000\138\000\094\000\139\000\093\000\000\000\
\\016\000\060\000\074\000\101\000\136\000\100\000\138\000\094\000\
\\139\000\093\000\000\000\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\102\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\026\000\
\\012\000\103\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\066\000\104\000\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\107\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\016\000\114\000\028\000\040\000\
\\054\000\113\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\035\000\063\000\112\000\076\000\111\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\113\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\063\000\115\000\000\000\
\\002\000\121\000\003\000\120\000\005\000\027\000\011\000\119\000\
\\018\000\118\000\060\000\117\000\061\000\116\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\016\000\041\000\068\000\125\000\075\000\124\000\000\000\
\\003\000\130\000\005\000\027\000\011\000\129\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\131\000\000\000\
\\000\000\
\\000\000\
\\016\000\114\000\076\000\133\000\000\000\
\\000\000\
\\110\000\134\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\072\000\029\000\071\000\030\000\139\000\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\140\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\142\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\069\000\144\000\070\000\086\000\
\\071\000\085\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\067\000\147\000\000\000\
\\000\000\
\\000\000\
\\081\000\151\000\082\000\150\000\000\000\
\\000\000\
\\140\000\154\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\157\000\000\000\
\\000\000\
\\005\000\132\000\017\000\159\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\113\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\063\000\112\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\066\000\160\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\162\000\037\000\161\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\175\000\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\193\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\064\000\196\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\067\000\203\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\068\000\204\000\000\000\
\\000\000\
\\067\000\206\000\000\000\
\\016\000\114\000\076\000\111\000\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\207\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\213\000\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\216\000\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\217\000\
\\109\000\211\000\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\218\000\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\219\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\225\000\089\000\224\000\
\\090\000\223\000\091\000\222\000\092\000\221\000\130\000\220\000\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\070\000\235\000\071\000\085\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\069\000\236\000\070\000\086\000\
\\071\000\085\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\238\000\037\000\161\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\091\000\079\000\239\000\
\\080\000\089\000\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\240\000\037\000\161\000\000\000\
\\007\000\242\000\026\000\241\000\000\000\
\\000\000\
\\016\000\060\000\072\000\243\000\074\000\058\000\000\000\
\\016\000\060\000\074\000\101\000\138\000\094\000\139\000\244\000\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\246\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\254\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\001\001\037\000\161\000\
\\040\000\000\001\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\004\001\
\\031\000\003\001\032\000\002\001\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\048\000\008\001\049\000\007\001\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\048\000\008\001\049\000\013\001\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\017\001\046\000\016\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\052\000\015\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\020\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\053\000\019\001\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\024\001\
\\043\000\023\001\044\000\022\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\026\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\027\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\028\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\029\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\030\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\031\001\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\034\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\035\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\036\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\016\000\114\000\076\000\037\001\000\000\
\\000\000\
\\002\000\121\000\003\000\120\000\005\000\027\000\011\000\119\000\
\\018\000\118\000\060\000\038\001\061\000\116\000\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\065\000\040\001\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\068\000\042\001\000\000\
\\000\000\
\\067\000\044\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\045\001\037\000\161\000\000\000\
\\000\000\
\\007\000\052\001\116\000\051\001\149\000\050\001\150\000\049\001\
\\151\000\048\001\152\000\047\001\153\000\046\001\000\000\
\\000\000\
\\123\000\055\001\000\000\
\\000\000\
\\000\000\
\\089\000\057\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\121\000\058\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\089\000\064\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\001\
\\099\000\067\001\100\000\066\001\000\000\
\\016\000\060\000\074\000\071\001\096\000\070\001\097\000\069\001\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\072\001\000\000\
\\007\000\072\000\029\000\075\001\104\000\074\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\078\001\106\000\077\001\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\081\001\101\000\080\001\
\\102\000\079\001\000\000\
\\016\000\060\000\074\000\071\001\096\000\082\001\097\000\069\001\000\000\
\\016\000\060\000\074\000\099\000\132\000\087\001\134\000\086\001\
\\135\000\085\001\137\000\084\001\138\000\094\000\139\000\083\001\000\000\
\\089\000\088\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\089\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\091\001\
\\078\000\090\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\095\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\096\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\097\001\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\091\001\
\\078\000\090\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\098\001\037\000\161\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\101\001\
\\034\000\100\001\035\000\099\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\108\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\109\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\110\001\037\000\161\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\111\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\064\000\128\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\129\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\073\000\132\001\000\000\
\\000\000\
\\118\000\134\001\000\000\
\\117\000\138\001\118\000\137\001\123\000\136\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\111\000\142\001\000\000\
\\083\000\007\000\086\000\006\000\141\000\144\001\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\083\000\007\000\086\000\006\000\141\000\145\001\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\105\000\146\001\000\000\
\\000\000\
\\000\000\
\\105\000\149\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\120\000\150\001\000\000\
\\007\000\179\000\008\000\153\001\024\000\152\001\025\000\151\001\000\000\
\\000\000\
\\110\000\154\001\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\155\001\
\\108\000\212\000\109\000\211\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\159\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\163\001\103\000\162\001\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\165\001\021\000\164\001\000\000\
\\000\000\
\\000\000\
\\039\000\167\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\171\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\254\000\000\000\
\\006\000\168\000\023\000\173\001\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\175\001\037\000\161\000\
\\040\000\174\001\000\000\
\\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\004\001\
\\031\000\003\001\032\000\176\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\177\001\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\178\001\037\000\161\000\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\180\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\052\000\179\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\182\001\046\000\181\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\020\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\053\000\183\001\000\000\
\\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\024\001\
\\043\000\023\001\044\000\184\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\185\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\186\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\187\001\046\000\181\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\188\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\189\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\190\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\191\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\192\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\016\000\060\000\072\000\194\001\074\000\058\000\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\115\000\195\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\105\000\196\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\197\001\
\\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\198\001\
\\109\000\211\000\000\000\
\\000\000\
\\007\000\052\001\083\000\007\000\086\000\006\000\141\000\202\001\
\\142\000\074\000\143\000\073\000\144\000\004\000\149\000\201\001\
\\150\000\049\001\151\000\200\001\153\000\199\001\000\000\
\\000\000\
\\000\000\
\\123\000\206\001\124\000\205\001\000\000\
\\016\000\060\000\074\000\208\001\000\000\
\\000\000\
\\121\000\210\001\122\000\209\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\001\
\\099\000\214\001\100\000\066\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\215\001\037\000\161\000\000\000\
\\016\000\060\000\074\000\071\001\096\000\216\001\097\000\069\001\000\000\
\\098\000\217\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\219\001\
\\109\000\211\000\000\000\
\\007\000\072\000\029\000\221\001\104\000\220\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\165\001\021\000\223\001\000\000\
\\003\000\092\000\005\000\027\000\015\000\081\001\101\000\224\001\
\\102\000\079\001\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\225\001\037\000\161\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\226\001\
\\078\000\090\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\227\001\037\000\161\000\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\228\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\232\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\117\000\234\001\118\000\137\001\119\000\233\001\123\000\206\001\
\\124\000\205\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\136\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\052\001\149\000\238\001\150\000\049\001\151\000\200\001\
\\153\000\199\001\000\000\
\\000\000\
\\000\000\
\\105\000\239\001\000\000\
\\006\000\168\000\023\000\240\001\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\105\000\241\001\
\\120\000\150\001\000\000\
\\007\000\179\000\008\000\153\001\024\000\152\001\025\000\242\001\000\000\
\\007\000\052\001\126\000\245\001\149\000\050\001\150\000\049\001\
\\151\000\244\001\152\000\243\001\153\000\046\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\246\001\037\000\161\000\000\000\
\\094\000\247\001\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\251\001\103\000\250\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\253\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\105\000\254\001\
\\115\000\195\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\001\002\124\000\000\002\000\000\
\\000\000\
\\121\000\210\001\122\000\003\002\000\000\
\\000\000\
\\129\000\004\002\000\000\
\\123\000\136\001\127\000\007\002\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\105\000\009\002\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\010\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\117\000\234\001\118\000\137\001\119\000\011\002\123\000\001\002\
\\124\000\000\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\012\002\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\125\000\013\002\000\000\
\\000\000\
\\000\000\
\\105\000\014\002\000\000\
\\094\000\016\002\095\000\015\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\206\001\124\000\205\001\127\000\019\002\128\000\018\002\
\\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\010\002\
\\105\000\021\002\000\000\
\\000\000\
\\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\105\000\022\002\
\\125\000\013\002\000\000\
\\094\000\016\002\095\000\023\002\000\000\
\\123\000\001\002\124\000\000\002\127\000\019\002\128\000\024\002\
\\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\"
val numstates = 538
val numrules = 306
val s = ref "" and index = ref 0
val string_to_int = fn () => 
let val i = !index
in index := i+2; Char.ord(String.sub(!s,i)) + Char.ord(String.sub(!s,i+1)) * 256
end
val string_to_list = fn s' =>
    let val len = String.size s'
        fun f () =
           if !index < len then string_to_int() :: f()
           else nil
   in index := 0; s := s'; f ()
   end
val string_to_pairlist = fn (conv_key,conv_entry) =>
     let fun f () =
         case string_to_int()
         of 0 => EMPTY
          | n => PAIR(conv_key (n-1),conv_entry (string_to_int()),f())
     in f
     end
val string_to_pairlist_default = fn (conv_key,conv_entry) =>
    let val conv_row = string_to_pairlist(conv_key,conv_entry)
    in fn () =>
       let val default = conv_entry(string_to_int())
           val row = conv_row()
       in (row,default)
       end
   end
val string_to_table = fn (convert_row,s') =>
    let val len = String.size s'
        fun f ()=
           if !index < len then convert_row() :: f()
           else nil
     in (s := s'; index := 0; f ())
     end
local
  val memo = Array.array(numstates+numrules,ERROR)
  val _ =let fun g i=(Array.update(memo,i,REDUCE(i-numstates)); g(i+1))
       fun f i =
            if i=numstates then g i
            else (Array.update(memo,i,SHIFT (STATE i)); f (i+1))
          in f 0 handle Subscript => ()
          end
in
val entry_to_action = fn 0 => ACCEPT | 1 => ERROR | j => Array.sub(memo,(j-2))
end
val gotoT=Array.fromList(string_to_table(string_to_pairlist(NT,STATE),gotoT))
val actionRows=string_to_table(string_to_pairlist_default(T,entry_to_action),actionRows)
val actionRowNumbers = string_to_list actionRowNumbers
val actionT = let val actionRowLookUp=
let val a=Array.fromList(actionRows) in fn i=>Array.sub(a,i) end
in Array.fromList(map actionRowLookUp actionRowNumbers)
end
in LrTable.mkLrTable {actions=actionT,gotos=gotoT,numRules=numrules,
numStates=numstates,initialState=STATE 0}
end
end
local open Header in
type pos = int
type arg = unit
structure MlyValue = 
struct
datatype svalue = VOID | ntVOID of unit ->  unit
 | CHAR of unit ->  (char) | STRING of unit ->  (string)
 | REAL of unit ->  (string) | WORD of unit ->  (Word32.word)
 | INT of unit ->  (string) | TYVAR of unit ->  (string)
 | LONGID of unit ->  (string) | strexp2node of unit ->  (Strexp.node)
 | strexp2 of unit ->  (Strexp.t)
 | strexp1 of unit ->  (Strexp.t* ( Sigexp.t -> SigConst.t ) *Sigexp.t)
 | strexpnode of unit ->  (Strexp.node)
 | strexp of unit ->  (Strexp.t) | program of unit ->  (Program.t)
 | topdecs of unit ->  (Topdec.t list)
 | topdecnode of unit ->  (Topdec.node)
 | topdec of unit ->  (Topdec.t)
 | strdecnode of unit ->  (Strdec.node)
 | strdec of unit ->  (Strdec.t)
 | strdecsnode of unit ->  (Strdec.node)
 | strdecs of unit ->  (Strdec.t) | withtypes of unit ->  (TypBind.t)
 | dbs of unit ->  (db list) | db of unit ->  (db)
 | datBindNoWithtype of unit ->  (DatBind.t)
 | datBind of unit ->  (DatBind.t)
 | repl of unit ->  (DatatypeRhs.node)
 | datatypeRhsnodeNoWithtype of unit ->  (DatatypeRhs.node)
 | datatypeRhsnode of unit ->  (DatatypeRhs.node)
 | datatypeRhsNoWithtype of unit ->  (DatatypeRhs.t)
 | datatypeRhs of unit ->  (DatatypeRhs.t)
 | fctarg of unit ->  (FctArg.node)
 | funbinds'2 of unit ->  (funbind list)
 | funbinds'1' of unit ->  (funbind whereAnd)
 | funbinds'1 of unit ->  (funbind whereAnd)
 | funbinds' of unit ->  (Strexp.t*funbind list)
 | funbinds of unit ->  (funbind list)
 | andspecs of unit ->  (wherespec list)
 | wherespecs of unit ->  (wherespec list)
 | sigbinds'' of unit ->  (sigbind whereAnd)
 | sigbinds' of unit ->  (sigbind whereAnd)
 | sigbinds of unit ->  (sigbind list)
 | strbinds'1' of unit ->  (strbind whereAnd)
 | strbinds'2 of unit ->  (strbind list)
 | strbinds'1 of unit ->  (strbind whereAnd)
 | strbinds' of unit ->  (Strexp.t*strbind list)
 | strbinds of unit ->  (strbind list) | sdecsPlus of unit ->  (Dec.t)
 | sdec of unit ->  (Dec.t) | sdecs of unit ->  (Dec.t)
 | arg_fct of unit ->  (Strexp.t) | sigconst of unit ->  (SigConst.t)
 | sigexp'node of unit ->  (Sigexp.node)
 | sigexp' of unit ->  (Sigexp.t)
 | sigexpnode of unit ->  (Sigexp.node)
 | sigexp of unit ->  (Sigexp.t) | wherespec of unit ->  (wherespec)
 | longstrideqns of unit ->  (Longstrid.t list)
 | longtyconeqns of unit ->  (Longtycon.t list)
 | exndesc of unit ->  (exndesc) | exndescs of unit ->  (exndesc list)
 | valdesc of unit ->  (valdesc) | valdescs of unit ->  (valdesc list)
 | typdescRhs of unit ->  (Type.t option)
 | typdesc of unit ->  (typdesc) | typdescs of unit ->  (typdesc list)
 | strdescs'' of unit ->  (strdesc whereAnd)
 | strdescs' of unit ->  (strdesc whereAnd)
 | strdescs of unit ->  (strdesc list)
 | sharespec of unit ->  (Equation.node)
 | specnode of unit ->  (Spec.node) | spec of unit ->  (Spec.t)
 | specs of unit ->  (Spec.t) | decsnode of unit ->  (Dec.node)
 | decs of unit ->  (Dec.t) | decnolocal of unit ->  (Dec.node)
 | decnode of unit ->  (Dec.node) | dec of unit ->  (Dec.t)
 | fixity of unit ->  (Fixity.t) | ebrhsnode of unit ->  (EbRhs.node)
 | ebrhs of unit ->  (EbRhs.t) | eb of unit ->  (eb)
 | ebs of unit ->  (eb list)
 | constr of unit ->  (Con.t*Type.t option)
 | constrs of unit ->  ( ( Con.t * Type.t option )  list)
 | tyvar_pc of unit ->  (Tyvar.t list)
 | tyvarseq of unit ->  (Tyvar.t list)
 | tyvars of unit ->  (Tyvar.t list)
 | typBind' of unit ->  (TypBind.node)
 | typBind of unit ->  (TypBind.t) | clause of unit ->  (clause)
 | clauses of unit ->  (clause list)
 | funs of unit ->  (clause list list)
 | rvbs of unit ->  ({ var:Var.t,ty:Type.t option,fixity:Vid.t option,rules:rules }  list)
 | constraint of unit ->  (Type.t option)
 | vbs of unit ->  ( ( Pat.t * Exp.t )  list)
 | opaspat of unit ->  (Pat.t option)
 | commapats of unit ->  (Pat.t list) | pats of unit ->  (Pat.t list)
 | pat_2c of unit ->  (Pat.t list) | patitem of unit ->  (Pat.Item.t)
 | patitems of unit ->  ( ( Pat.Item.t list * bool ) )
 | apat' of unit ->  (Pat.t) | apatnode of unit ->  (Pat.node)
 | apat of unit ->  (Pat.t) | apats of unit ->  (Pat.t list)
 | patnode of unit ->  (Pat.node) | pat of unit ->  (Pat.t)
 | exp_list of unit ->  (Exp.t list) | exp_2c of unit ->  (Exp.t list)
 | ot_list of unit ->  (Exp.t list) | aexp of unit ->  (Exp.node)
 | app_exp1 of unit ->  (Exp.t list)
 | app_exp of unit ->  (Exp.t list) | expnode of unit ->  (Exp.node)
 | exp of unit ->  (Exp.t) | exp_ps of unit ->  (Exp.t list)
 | elabels of unit ->  ( ( Field.t * Exp.t )  list)
 | elabel of unit ->  ( ( Field.t * Exp.t ) )
 | rule of unit ->  (rule) | rules of unit ->  (rules)
 | ty0_pc of unit ->  (Type.t list)
 | tyOpt of unit ->  (Type.t option) | ty1 of unit ->  (Type.t)
 | tynode of unit ->  (Type.node) | ty of unit ->  (Type.t)
 | tuple_ty of unit ->  (Type.t list)
 | ty'node of unit ->  (Type.node) | ty' of unit ->  (Type.t)
 | tlabels of unit ->  ( ( Field.t * Type.t )  list)
 | tlabel of unit ->  ( ( Field.t * Type.t ) )
 | longstrids of unit ->  (Longstrid.t list)
 | longstrid of unit ->  (Longstrid.t)
 | longvidNoEqual of unit ->  (Longvid.t)
 | longvid of unit ->  (Longvid.t) | longcon of unit ->  (Longcon.t)
 | longvarands of unit ->  (Longvar.t list)
 | longvar of unit ->  (Longvar.t)
 | longtycon of unit ->  (Longtycon.t) | fctid of unit ->  (Fctid.t)
 | sigids of unit ->  (Sigid.t list) | sigid of unit ->  (Sigid.t)
 | strid of unit ->  (Strid.t) | field of unit ->  (Field.t)
 | tycon of unit ->  (Tycon.t) | tyvar of unit ->  (Tyvar.t)
 | con of unit ->  (Con.t) | varNoEqual of unit ->  (Var.t)
 | var of unit ->  (Var.t) | vids of unit ->  (Vid.t list)
 | vidNoEqual of unit ->  (Vid.t) | vid of unit ->  (Vid.t)
 | const of unit ->  (Const.t) | longidEqual of unit ->  (string)
 | longid of unit ->  (string) | longidNoAsterisk of unit ->  (string)
 | idNoAsterisk of unit ->  (string) | idEqual of unit ->  (string)
 | id of unit ->  (string) | int of unit ->  (int)
 | digit of unit ->  (int)
end
type svalue = MlyValue.svalue
type result = Program.t
end
structure EC=
struct
open LrTable
val is_keyword =
fn (T 9) => true | (T 10) => true | (T 12) => true | (T 14) => true | 
(T 15) => true | (T 16) => true | (T 17) => true | (T 18) => true | 
(T 20) => true | (T 21) => true | (T 22) => true | (T 23) => true | 
(T 24) => true | (T 25) => true | (T 26) => true | (T 27) => true | 
(T 29) => true | (T 30) => true | (T 31) => true | (T 32) => true | 
(T 33) => true | (T 34) => true | (T 35) => true | (T 36) => true | 
(T 37) => true | (T 38) => true | (T 39) => true | (T 40) => true | 
(T 41) => true | (T 42) => true | (T 43) => true | (T 44) => true | 
(T 45) => true | (T 46) => true | (T 47) => true | (T 48) => true | 
(T 49) => true | (T 50) => true | (T 52) => true | (T 51) => true | 
(T 54) => true | (T 55) => true | (T 66) => true | (T 67) => true | _ => false
val preferred_change = 
(nil
,(T 50) :: nil
)::
(nil
,(T 48) :: nil
)::
(nil
,(T 17) :: nil
)::
(nil
,(T 62) :: nil
)::
(nil
,(T 1) :: nil
)::
((T 23) :: nil
,(T 19) :: nil
)::
((T 19) :: nil
,(T 23) :: nil
)::
((T 10) :: nil
,(T 67) :: nil
)::
((T 57) :: nil
,(T 37) :: nil
)::
((T 1) :: nil
,(T 59) :: nil
)::
((T 59) :: nil
,(T 1) :: nil
)::
(nil
,(T 30) :: (T 2) :: (T 18) :: nil
)::
(nil
,(T 17) :: (T 2) :: nil
)::
nil
val noShift = 
fn (T 0) => true | _ => false
val showTerminal =
fn (T 0) => "EOF"
  | (T 1) => "SEMICOLON"
  | (T 2) => "LONGID"
  | (T 3) => "TYVAR"
  | (T 4) => "INT"
  | (T 5) => "WORD"
  | (T 6) => "REAL"
  | (T 7) => "STRING"
  | (T 8) => "CHAR"
  | (T 9) => "ABSTYPE"
  | (T 10) => "AND"
  | (T 11) => "ARROW"
  | (T 12) => "AS"
  | (T 13) => "BAR"
  | (T 14) => "CASE"
  | (T 15) => "DATATYPE"
  | (T 16) => "DOTDOTDOT"
  | (T 17) => "ELSE"
  | (T 18) => "END"
  | (T 19) => "EQUALOP"
  | (T 20) => "EQTYPE"
  | (T 21) => "EXCEPTION"
  | (T 22) => "DO"
  | (T 23) => "DARROW"
  | (T 24) => "FN"
  | (T 25) => "FUN"
  | (T 26) => "FUNCTOR"
  | (T 27) => "HANDLE"
  | (T 28) => "HASH"
  | (T 29) => "IF"
  | (T 30) => "IN"
  | (T 31) => "INCLUDE"
  | (T 32) => "INFIX"
  | (T 33) => "INFIXR"
  | (T 34) => "LET"
  | (T 35) => "LOCAL"
  | (T 36) => "NONFIX"
  | (T 37) => "OF"
  | (T 38) => "OP"
  | (T 39) => "OPEN"
  | (T 40) => "OVERLOAD"
  | (T 41) => "RAISE"
  | (T 42) => "REC"
  | (T 43) => "SHARING"
  | (T 44) => "SIG"
  | (T 45) => "SIGNATURE"
  | (T 46) => "STRUCT"
  | (T 47) => "STRUCTURE"
  | (T 48) => "THEN"
  | (T 49) => "TYPE"
  | (T 50) => "VAL"
  | (T 51) => "WHERE"
  | (T 52) => "WHILE"
  | (T 53) => "WILD"
  | (T 54) => "WITH"
  | (T 55) => "WITHTYPE"
  | (T 56) => "ASTERISK"
  | (T 57) => "COLON"
  | (T 58) => "COLONGT"
  | (T 59) => "COMMA"
  | (T 60) => "LBRACE"
  | (T 61) => "LBRACKET"
  | (T 62) => "LPAREN"
  | (T 63) => "RBRACE"
  | (T 64) => "RBRACKET"
  | (T 65) => "RPAREN"
  | (T 66) => "ORELSE"
  | (T 67) => "ANDALSO"
  | (T 68) => "FUNSIG"
  | (T 69) => "PRIM"
  | _ => "bogus-term"
local open Header in
val errtermvalue=
fn _ => MlyValue.VOID
end
val terms = (T 0) :: (T 1) :: (T 9) :: (T 10) :: (T 11) :: (T 12) :: 
(T 13) :: (T 14) :: (T 15) :: (T 16) :: (T 17) :: (T 18) :: (T 19) :: 
(T 20) :: (T 21) :: (T 22) :: (T 23) :: (T 24) :: (T 25) :: (T 26) :: 
(T 27) :: (T 28) :: (T 29) :: (T 30) :: (T 31) :: (T 32) :: (T 33) :: 
(T 34) :: (T 35) :: (T 36) :: (T 37) :: (T 38) :: (T 39) :: (T 40) :: 
(T 41) :: (T 42) :: (T 43) :: (T 44) :: (T 45) :: (T 46) :: (T 47) :: 
(T 48) :: (T 49) :: (T 50) :: (T 51) :: (T 52) :: (T 53) :: (T 54) :: 
(T 55) :: (T 56) :: (T 57) :: (T 58) :: (T 59) :: (T 60) :: (T 61) :: 
(T 62) :: (T 63) :: (T 64) :: (T 65) :: (T 66) :: (T 67) :: (T 68) :: 
(T 69) :: nil
end
structure Actions =
struct 
exception mlyAction of int
local open Header in
val actions = 
fn (i392,defaultPos,stack,
    (()):arg) =>
case (i392,stack)
of (0,(_,(MlyValue.topdecs topdecs1,topdecs1left,topdecs1right))::
rest671) => let val result=MlyValue.program(fn _ => let val topdecs
 as topdecs1=topdecs1 ()
 in (Program.T topdecs) end
)
 in (LrTable.NT 147,(result,topdecs1left,topdecs1right),rest671) end
| (1,(_,(MlyValue.topdec topdec1,topdec1left,topdec1right))::rest671)
 => let val result=MlyValue.topdecs(fn _ => let val topdec as topdec1=
topdec1 ()
 in ([topdec]) end
)
 in (LrTable.NT 146,(result,topdec1left,topdec1right),rest671) end
| (2,(_,(_,SEMICOLON1left,SEMICOLON1right))::rest671) => let val 
result=MlyValue.topdecs(fn _ => ([]))
 in (LrTable.NT 146,(result,SEMICOLON1left,SEMICOLON1right),rest671)
 end
| (3,(_,(MlyValue.topdecs topdecs1,_,topdecs1right))::(_,(_,
SEMICOLON1left,_))::rest671) => let val result=MlyValue.topdecs(fn _
 => let val topdecs as topdecs1=topdecs1 ()
 in (topdecs) end
)
 in (LrTable.NT 146,(result,SEMICOLON1left,topdecs1right),rest671) end
| (4,(_,(MlyValue.topdecs topdecs1,_,topdecs1right))::(_,(
MlyValue.topdec topdec1,topdec1left,_))::rest671) => let val result=
MlyValue.topdecs(fn _ => let val topdec as topdec1=topdec1 ()
val topdecs as topdecs1=topdecs1 ()
 in (topdec::topdecs) end
)
 in (LrTable.NT 146,(result,topdec1left,topdecs1right),rest671) end
| (5,(_,(MlyValue.topdecnode topdecnode1,topdecnodeleft as 
topdecnode1left,topdecnoderight as topdecnode1right))::rest671) => 
let val result=MlyValue.topdec(fn _ => let val topdecnode as 
topdecnode1=topdecnode1 ()
 in (
Topdec.makeRegion'(topdecnode,
					topdecnodeleft,
					topdecnoderight)
) end
)
 in (LrTable.NT 144,(result,topdecnode1left,topdecnode1right),rest671)
 end
| (6,(_,(MlyValue.strdec strdec1,strdec1left,strdec1right))::rest671)
 => let val result=MlyValue.topdecnode(fn _ => let val strdec as 
strdec1=strdec1 ()
 in (Topdec.Strdec strdec) end
)
 in (LrTable.NT 145,(result,strdec1left,strdec1right),rest671) end
| (7,(_,(MlyValue.sigbinds sigbinds1,_,sigbinds1right))::(_,(_,
SIGNATURE1left,_))::rest671) => let val result=MlyValue.topdecnode(fn 
_ => let val sigbinds as sigbinds1=sigbinds1 ()
 in (Topdec.Signature sigbinds) end
)
 in (LrTable.NT 145,(result,SIGNATURE1left,sigbinds1right),rest671)
 end
| (8,(_,(MlyValue.funbinds funbinds1,_,funbinds1right))::(_,(_,
FUNCTOR1left,_))::rest671) => let val result=MlyValue.topdecnode(fn _
 => let val funbinds as funbinds1=funbinds1 ()
 in (Topdec.Functor funbinds) end
)
 in (LrTable.NT 145,(result,FUNCTOR1left,funbinds1right),rest671) end
| (9,(_,(MlyValue.strdecsnode strdecsnode1,strdecsnodeleft as 
strdecsnode1left,strdecsnoderight as strdecsnode1right))::rest671) => 
let val result=MlyValue.strdecs(fn _ => let val strdecsnode as 
strdecsnode1=strdecsnode1 ()
 in (
Strdec.makeRegion'(strdecsnode,
					  strdecsnodeleft, strdecsnoderight)
) end
)
 in (LrTable.NT 140,(result,strdecsnode1left,strdecsnode1right),
rest671) end
| (10,rest671) => let val result=MlyValue.strdecsnode(fn _ => (
Strdec.Seq[]))
 in (LrTable.NT 141,(result,defaultPos,defaultPos),rest671) end
| (11,(_,(MlyValue.strdecs strdecs1,_,strdecs1right))::(_,(_,
SEMICOLON1left,_))::rest671) => let val result=MlyValue.strdecsnode(
fn _ => let val strdecs as strdecs1=strdecs1 ()
 in (Strdec.Seq[strdecs]) end
)
 in (LrTable.NT 141,(result,SEMICOLON1left,strdecs1right),rest671) end
| (12,(_,(MlyValue.strdecs strdecs1,_,strdecs1right))::(_,(
MlyValue.strdec strdec1,strdec1left,_))::rest671) => let val result=
MlyValue.strdecsnode(fn _ => let val strdec as strdec1=strdec1 ()
val strdecs as strdecs1=strdecs1 ()
 in (Strdec.Seq[strdec,strdecs]) end
)
 in (LrTable.NT 141,(result,strdec1left,strdecs1right),rest671) end
| (13,(_,(MlyValue.strdecnode strdecnode1,strdecnodeleft as 
strdecnode1left,strdecnoderight as strdecnode1right))::rest671) => 
let val result=MlyValue.strdec(fn _ => let val strdecnode as 
strdecnode1=strdecnode1 ()
 in (
Strdec.makeRegion'(strdecnode,
					strdecnodeleft, strdecnoderight))
 end
)
 in (LrTable.NT 142,(result,strdecnode1left,strdecnode1right),rest671)
 end
| (14,(_,(MlyValue.strbinds strbinds1,_,strbinds1right))::(_,(_,
STRUCTURE1left,_))::rest671) => let val result=MlyValue.strdecnode(fn 
_ => let val strbinds as strbinds1=strbinds1 ()
 in (Strdec.Structure strbinds) end
)
 in (LrTable.NT 143,(result,STRUCTURE1left,strbinds1right),rest671)
 end
| (15,(_,(_,_,END1right))::(_,(MlyValue.strdecs strdecs2,_,_))::_::(_,
(MlyValue.strdecs strdecs1,_,_))::(_,(_,LOCAL1left,_))::rest671) => 
let val result=MlyValue.strdecnode(fn _ => let val strdecs1=strdecs1 
()
val strdecs2=strdecs2 ()
 in (Strdec.Local(strdecs1,strdecs2)) end
)
 in (LrTable.NT 143,(result,LOCAL1left,END1right),rest671) end
| (16,(_,(MlyValue.decnolocal decnolocal1,decnolocalleft as 
decnolocal1left,decnolocalright as decnolocal1right))::rest671) => 
let val result=MlyValue.strdecnode(fn _ => let val decnolocal as 
decnolocal1=decnolocal1 ()
 in (
Strdec.Core
		  (Dec.makeRegion'(decnolocal, decnolocalleft, decnolocalright))
) end
)
 in (LrTable.NT 143,(result,decnolocal1left,decnolocal1right),rest671)
 end
| (17,(_,(MlyValue.strbinds' strbinds'1,_,strbinds'1right))::_::(_,(
MlyValue.sigconst sigconst1,_,_))::(_,(MlyValue.strid strid1,
strid1left,_))::rest671) => let val result=MlyValue.strbinds(fn _ => 
let val strid as strid1=strid1 ()
val sigconst as sigconst1=sigconst1 ()
val strbinds' as strbinds'1=strbinds'1 ()
 in (
let val (def,strbinds) = strbinds'
	    in {name = strid, def = def, constraint = sigconst}
	       :: strbinds
	    end
) end
)
 in (LrTable.NT 114,(result,strid1left,strbinds'1right),rest671) end
| (18,(_,(MlyValue.strbinds'1 strbinds'11,_,strbinds'11right))::(_,(
MlyValue.strexp1 strexp11,strexp11left,_))::rest671) => let val result
=MlyValue.strbinds'(fn _ => let val strexp1 as strexp11=strexp11 ()
val strbinds'1 as strbinds'11=strbinds'11 ()
 in (augment1(strexp1, strbinds'1)) end
)
 in (LrTable.NT 115,(result,strexp11left,strbinds'11right),rest671)
 end
| (19,(_,(MlyValue.strbinds'2 strbinds'21,_,strbinds'21right))::(_,(
MlyValue.strexp2 strexp21,strexp21left,_))::rest671) => let val result
=MlyValue.strbinds'(fn _ => let val strexp2 as strexp21=strexp21 ()
val strbinds'2 as strbinds'21=strbinds'21 ()
 in ((strexp2,strbinds'2)) end
)
 in (LrTable.NT 115,(result,strexp21left,strbinds'21right),rest671)
 end
| (20,(_,(MlyValue.strbinds'2 strbinds'21,strbinds'2left as 
strbinds'21left,strbinds'21right))::rest671) => let val result=
MlyValue.strbinds'1(fn _ => let val strbinds'2 as strbinds'21=
strbinds'21 ()
 in (([], strbinds'2left, strbinds'2)) end
)
 in (LrTable.NT 116,(result,strbinds'21left,strbinds'21right),rest671)
 end
| (21,(_,(MlyValue.strbinds'1' strbinds'1'1,_,strbinds'1'1right))::(_,
(MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.strbinds'1(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strbinds'1' as strbinds'1'1=strbinds'1'1 ()
 in (cons1(wherespec,strbinds'1')) end
)
 in (LrTable.NT 116,(result,WHERE1left,strbinds'1'1right),rest671) end
| (22,(_,(MlyValue.strbinds'1 strbinds'11,strbinds'11left,
strbinds'11right))::rest671) => let val result=MlyValue.strbinds'1'(
fn _ => let val strbinds'1 as strbinds'11=strbinds'11 ()
 in (strbinds'1) end
)
 in (LrTable.NT 118,(result,strbinds'11left,strbinds'11right),rest671)
 end
| (23,(_,(MlyValue.strbinds'1' strbinds'1'1,_,strbinds'1'1right))::(_,
(MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.strbinds'1'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strbinds'1' as strbinds'1'1=strbinds'1'1 ()
 in (cons1(wherespec,strbinds'1')) end
)
 in (LrTable.NT 118,(result,AND1left,strbinds'1'1right),rest671) end
| (24,rest671) => let val result=MlyValue.strbinds'2(fn _ => ([]))
 in (LrTable.NT 117,(result,defaultPos,defaultPos),rest671) end
| (25,(_,(MlyValue.strbinds strbinds1,_,strbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.strbinds'2(fn _ => 
let val strbinds as strbinds1=strbinds1 ()
 in (strbinds) end
)
 in (LrTable.NT 117,(result,AND1left,strbinds1right),rest671) end
| (26,(_,(MlyValue.strexpnode strexpnode1,strexpnodeleft as 
strexpnode1left,strexpnoderight as strexpnode1right))::rest671) => 
let val result=MlyValue.strexp(fn _ => let val strexpnode as 
strexpnode1=strexpnode1 ()
 in (
Strexp.makeRegion'(strexpnode,
					strexpnodeleft, strexpnoderight))
 end
)
 in (LrTable.NT 148,(result,strexpnode1left,strexpnode1right),rest671)
 end
| (27,(_,(MlyValue.strexp1 strexp11,strexp11left,strexp11right))::
rest671) => let val result=MlyValue.strexpnode(fn _ => let val strexp1
 as strexp11=strexp11 ()
 in (
let val (strexp,sigconst,sigexp) = strexp1
			      in Strexp.Constrained(strexp, sigconst sigexp)
			      end
) end
)
 in (LrTable.NT 149,(result,strexp11left,strexp11right),rest671) end
| (28,(_,(MlyValue.wherespecs wherespecs1,_,wherespecsright as 
wherespecs1right))::(_,(MlyValue.strexp1 strexp11,strexp11left,_))::
rest671) => let val result=MlyValue.strexpnode(fn _ => let val strexp1
 as strexp11=strexp11 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (
let val (strexp,sigconst,sigexp) = strexp1
			      in Strexp.Constrained
				 (strexp,
				  sigconst(Sigexp.makeRegion'
					   (Sigexp.Where(sigexp,wherespecs),
					    Sigexp.left sigexp,
					    wherespecsright)))
			      end
) end
)
 in (LrTable.NT 149,(result,strexp11left,wherespecs1right),rest671)
 end
| (29,(_,(MlyValue.strexp2node strexp2node1,strexp2node1left,
strexp2node1right))::rest671) => let val result=MlyValue.strexpnode(
fn _ => let val strexp2node as strexp2node1=strexp2node1 ()
 in (strexp2node) end
)
 in (LrTable.NT 149,(result,strexp2node1left,strexp2node1right),
rest671) end
| (30,(_,(MlyValue.sigexp' sigexp'1,_,sigexp'1right))::_::(_,(
MlyValue.strexp strexp1,strexp1left,_))::rest671) => let val result=
MlyValue.strexp1(fn _ => let val strexp as strexp1=strexp1 ()
val sigexp' as sigexp'1=sigexp'1 ()
 in ((strexp,SigConst.Transparent,sigexp')) end
)
 in (LrTable.NT 150,(result,strexp1left,sigexp'1right),rest671) end
| (31,(_,(MlyValue.sigexp' sigexp'1,_,sigexp'1right))::_::(_,(
MlyValue.strexp strexp1,strexp1left,_))::rest671) => let val result=
MlyValue.strexp1(fn _ => let val strexp as strexp1=strexp1 ()
val sigexp' as sigexp'1=sigexp'1 ()
 in ((strexp,SigConst.Opaque,sigexp')) end
)
 in (LrTable.NT 150,(result,strexp1left,sigexp'1right),rest671) end
| (32,(_,(MlyValue.strexp2node strexp2node1,strexp2nodeleft as 
strexp2node1left,strexp2noderight as strexp2node1right))::rest671) => 
let val result=MlyValue.strexp2(fn _ => let val strexp2node as 
strexp2node1=strexp2node1 ()
 in (
Strexp.makeRegion'(strexp2node,
					  strexp2nodeleft, strexp2noderight)
) end
)
 in (LrTable.NT 151,(result,strexp2node1left,strexp2node1right),
rest671) end
| (33,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.strexp2node(fn _ => let val longid as longid1=longid1 ()
 in (
Strexp.Var(Longstrid.fromString
						 (longid,
						  reg(longidleft,longidright)))
) end
)
 in (LrTable.NT 152,(result,longid1left,longid1right),rest671) end
| (34,(_,(_,_,END1right))::(_,(MlyValue.strdecs strdecs1,_,_))::(_,(_,
STRUCT1left,_))::rest671) => let val result=MlyValue.strexp2node(fn _
 => let val strdecs as strdecs1=strdecs1 ()
 in (Strexp.Struct strdecs) end
)
 in (LrTable.NT 152,(result,STRUCT1left,END1right),rest671) end
| (35,(_,(MlyValue.arg_fct arg_fct1,_,arg_fct1right))::(_,(
MlyValue.longid longid1,longidleft as longid1left,longidright))::
rest671) => let val result=MlyValue.strexp2node(fn _ => let val longid
 as longid1=longid1 ()
val arg_fct as arg_fct1=arg_fct1 ()
 in (
Strexp.App(Fctid.fromString
		      (ensureNonqualified(longid, reg(longidleft, longidright))),
		      arg_fct)
) end
)
 in (LrTable.NT 152,(result,longid1left,arg_fct1right),rest671) end
| (36,(_,(_,_,END1right))::(_,(MlyValue.strexp strexp1,_,_))::_::(_,(
MlyValue.strdecs strdecs1,_,_))::(_,(_,LET1left,_))::rest671) => let 
val result=MlyValue.strexp2node(fn _ => let val strdecs as strdecs1=
strdecs1 ()
val strexp as strexp1=strexp1 ()
 in (Strexp.Let(strdecs,strexp)) end
)
 in (LrTable.NT 152,(result,LET1left,END1right),rest671) end
| (37,(_,(_,_,RPAREN1right))::(_,(MlyValue.strexp strexp1,_,_))::(_,(_
,LPAREN1left,_))::rest671) => let val result=MlyValue.arg_fct(fn _ => 
let val strexp as strexp1=strexp1 ()
 in (strexp) end
)
 in (LrTable.NT 110,(result,LPAREN1left,RPAREN1right),rest671) end
| (38,(_,(_,_,RPAREN1right))::(_,(MlyValue.strdecs strdecs1,
strdecsleft,strdecsright))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.arg_fct(fn _ => let val strdecs as strdecs1=
strdecs1 ()
 in (
Strexp.makeRegion'
				  (Strexp.Struct strdecs,
				   strdecsleft, strdecsright)
) end
)
 in (LrTable.NT 110,(result,LPAREN1left,RPAREN1right),rest671) end
| (39,(_,(MlyValue.sigexp' sigexp'1,sigexp'1left,sigexp'1right))::
rest671) => let val result=MlyValue.sigexp(fn _ => let val sigexp' as 
sigexp'1=sigexp'1 ()
 in (sigexp') end
)
 in (LrTable.NT 105,(result,sigexp'1left,sigexp'1right),rest671) end
| (40,(_,(MlyValue.wherespecs wherespecs1,_,wherespecsright as 
wherespecs1right))::(_,(MlyValue.sigexp' sigexp'1,sigexp'left as 
sigexp'1left,_))::rest671) => let val result=MlyValue.sigexp(fn _ => 
let val sigexp' as sigexp'1=sigexp'1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (
Sigexp.makeRegion'
			      (Sigexp.Where(sigexp',wherespecs),
			       sigexp'left, wherespecsright)
) end
)
 in (LrTable.NT 105,(result,sigexp'1left,wherespecs1right),rest671)
 end
| (41,(_,(MlyValue.wherespec wherespec1,_,wherespec1right))::(_,(_,
WHERE1left,_))::rest671) => let val result=MlyValue.wherespecs(fn _
 => let val wherespec as wherespec1=wherespec1 ()
 in ([wherespec]) end
)
 in (LrTable.NT 122,(result,WHERE1left,wherespec1right),rest671) end
| (42,(_,(MlyValue.wherespecs wherespecs1,_,wherespecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.wherespecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (wherespec::wherespecs) end
)
 in (LrTable.NT 122,(result,WHERE1left,wherespecs1right),rest671) end
| (43,(_,(MlyValue.andspecs andspecs1,_,andspecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.wherespecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val andspecs as andspecs1=andspecs1 ()
 in (wherespec::andspecs) end
)
 in (LrTable.NT 122,(result,WHERE1left,andspecs1right),rest671) end
| (44,(_,(MlyValue.wherespec wherespec1,_,wherespec1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.andspecs(fn _ => 
let val wherespec as wherespec1=wherespec1 ()
 in ([wherespec]) end
)
 in (LrTable.NT 123,(result,AND1left,wherespec1right),rest671) end
| (45,(_,(MlyValue.andspecs andspecs1,_,andspecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.andspecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val andspecs as andspecs1=andspecs1 ()
 in (wherespec::andspecs) end
)
 in (LrTable.NT 123,(result,AND1left,andspecs1right),rest671) end
| (46,(_,(MlyValue.wherespecs wherespecs1,_,wherespecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.andspecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (wherespec::wherespecs) end
)
 in (LrTable.NT 123,(result,AND1left,wherespecs1right),rest671) end
| (47,(_,(MlyValue.sigbinds' sigbinds'1,_,sigbinds'1right))::(_,(
MlyValue.sigexp' sigexp'1,_,_))::_::(_,(MlyValue.sigid sigid1,
sigid1left,_))::rest671) => let val result=MlyValue.sigbinds(fn _ => 
let val sigid as sigid1=sigid1 ()
val sigexp' as sigexp'1=sigexp'1 ()
val sigbinds' as sigbinds'1=sigbinds'1 ()
 in (augment(sigid, sigexp', sigbinds')) end
)
 in (LrTable.NT 119,(result,sigid1left,sigbinds'1right),rest671) end
| (48,(_,(MlyValue.sigexp'node sigexp'node1,sigexp'nodeleft as 
sigexp'node1left,sigexp'noderight as sigexp'node1right))::rest671) => 
let val result=MlyValue.sigexp'(fn _ => let val sigexp'node as 
sigexp'node1=sigexp'node1 ()
 in (
Sigexp.makeRegion'(sigexp'node,
					  sigexp'nodeleft,
					  sigexp'noderight)
) end
)
 in (LrTable.NT 107,(result,sigexp'node1left,sigexp'node1right),
rest671) end
| (49,(_,(MlyValue.sigid sigid1,sigid1left,sigid1right))::rest671) => 
let val result=MlyValue.sigexp'node(fn _ => let val sigid as sigid1=
sigid1 ()
 in (Sigexp.Var sigid) end
)
 in (LrTable.NT 108,(result,sigid1left,sigid1right),rest671) end
| (50,(_,(_,_,END1right))::(_,(MlyValue.specs specs1,_,_))::(_,(_,
SIG1left,_))::rest671) => let val result=MlyValue.sigexp'node(fn _ => 
let val specs as specs1=specs1 ()
 in (Sigexp.Spec specs) end
)
 in (LrTable.NT 108,(result,SIG1left,END1right),rest671) end
| (51,rest671) => let val result=MlyValue.sigbinds'(fn _ => (
([], defaultPos, [])))
 in (LrTable.NT 120,(result,defaultPos,defaultPos),rest671) end
| (52,(_,(MlyValue.sigbinds sigbinds1,_,sigbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.sigbinds'(fn _ => 
let val sigbinds as sigbinds1=sigbinds1 ()
 in (([], defaultPos, sigbinds)) end
)
 in (LrTable.NT 120,(result,AND1left,sigbinds1right),rest671) end
| (53,(_,(MlyValue.sigbinds'' sigbinds''1,_,sigbinds''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.sigbinds'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val sigbinds'' as sigbinds''1=sigbinds''1 ()
 in (cons1(wherespec,sigbinds'')) end
)
 in (LrTable.NT 120,(result,WHERE1left,sigbinds''1right),rest671) end
| (54,(_,(MlyValue.sigbinds' sigbinds'1,sigbinds'1left,sigbinds'1right
))::rest671) => let val result=MlyValue.sigbinds''(fn _ => let val 
sigbinds' as sigbinds'1=sigbinds'1 ()
 in (sigbinds') end
)
 in (LrTable.NT 121,(result,sigbinds'1left,sigbinds'1right),rest671)
 end
| (55,(_,(MlyValue.sigbinds'' sigbinds''1,_,sigbinds''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.sigbinds''(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val sigbinds'' as sigbinds''1=sigbinds''1 ()
 in (cons1(wherespec,sigbinds'')) end
)
 in (LrTable.NT 121,(result,AND1left,sigbinds''1right),rest671) end
| (56,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.longtycon 
longtycon1,_,_))::(_,(MlyValue.tyvars tyvars1,_,_))::(_,(_,TYPE1left,_
))::rest671) => let val result=MlyValue.wherespec(fn _ => let val 
tyvars as tyvars1=tyvars1 ()
val longtycon as longtycon1=longtycon1 ()
val ty as ty1=ty1 ()
 in ({tyvars = tyvars,
						 longtycon = longtycon,
						 ty = ty})
 end
)
 in (LrTable.NT 104,(result,TYPE1left,ty1right),rest671) end
| (57,rest671) => let val result=MlyValue.sigconst(fn _ => (
SigConst.None))
 in (LrTable.NT 109,(result,defaultPos,defaultPos),rest671) end
| (58,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,COLON1left,_
))::rest671) => let val result=MlyValue.sigconst(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (SigConst.Transparent sigexp) end
)
 in (LrTable.NT 109,(result,COLON1left,sigexp1right),rest671) end
| (59,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,COLONGT1left
,_))::rest671) => let val result=MlyValue.sigconst(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (SigConst.Opaque sigexp) end
)
 in (LrTable.NT 109,(result,COLONGT1left,sigexp1right),rest671) end
| (60,rest671) => let val result=MlyValue.specs(fn _ => (
Spec.makeRegion(Spec.Empty, Region.bogus)))
 in (LrTable.NT 88,(result,defaultPos,defaultPos),rest671) end
| (61,(_,(MlyValue.specs specs1,_,specs1right))::(_,(_,SEMICOLON1left,
_))::rest671) => let val result=MlyValue.specs(fn _ => let val specs
 as specs1=specs1 ()
 in (specs) end
)
 in (LrTable.NT 88,(result,SEMICOLON1left,specs1right),rest671) end
| (62,(_,(MlyValue.specs specs1,_,specs1right))::(_,(MlyValue.spec 
spec1,spec1left,_))::rest671) => let val result=MlyValue.specs(fn _
 => let val spec as spec1=spec1 ()
val specs as specs1=specs1 ()
 in (Spec.seq(spec, specs)) end
)
 in (LrTable.NT 88,(result,spec1left,specs1right),rest671) end
| (63,(_,(MlyValue.specnode specnode1,specnodeleft as specnode1left,
specnoderight as specnode1right))::rest671) => let val result=
MlyValue.spec(fn _ => let val specnode as specnode1=specnode1 ()
 in (Spec.makeRegion'(specnode, specnodeleft, specnoderight)) end
)
 in (LrTable.NT 89,(result,specnode1left,specnode1right),rest671) end
| (64,(_,(MlyValue.valdescs valdescs1,_,valdescs1right))::(_,(_,
VAL1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val valdescs as valdescs1=valdescs1 ()
 in (Spec.Val valdescs) end
)
 in (LrTable.NT 90,(result,VAL1left,valdescs1right),rest671) end
| (65,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::(_,(_,
TYPE1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val typdescs as typdescs1=typdescs1 ()
 in (Spec.Type typdescs) end
)
 in (LrTable.NT 90,(result,TYPE1left,typdescs1right),rest671) end
| (66,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::(_,(_,
EQTYPE1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val typdescs as typdescs1=typdescs1 ()
 in (Spec.Eqtype typdescs) end
)
 in (LrTable.NT 90,(result,EQTYPE1left,typdescs1right),rest671) end
| (67,(_,(MlyValue.datatypeRhsNoWithtype datatypeRhsNoWithtype1,_,
datatypeRhsNoWithtype1right))::(_,(_,DATATYPE1left,_))::rest671) => 
let val result=MlyValue.specnode(fn _ => let val datatypeRhsNoWithtype
 as datatypeRhsNoWithtype1=datatypeRhsNoWithtype1 ()
 in (Spec.Datatype datatypeRhsNoWithtype) end
)
 in (LrTable.NT 90,(result,DATATYPE1left,datatypeRhsNoWithtype1right),
rest671) end
| (68,(_,(MlyValue.exndescs exndescs1,_,exndescs1right))::(_,(_,
EXCEPTION1left,_))::rest671) => let val result=MlyValue.specnode(fn _
 => let val exndescs as exndescs1=exndescs1 ()
 in (Spec.Exception exndescs) end
)
 in (LrTable.NT 90,(result,EXCEPTION1left,exndescs1right),rest671) end
| (69,(_,(MlyValue.strdescs strdescs1,_,strdescs1right))::(_,(_,
STRUCTURE1left,_))::rest671) => let val result=MlyValue.specnode(fn _
 => let val strdescs as strdescs1=strdescs1 ()
 in (Spec.Structure strdescs) end
)
 in (LrTable.NT 90,(result,STRUCTURE1left,strdescs1right),rest671) end
| (70,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,INCLUDE1left
,_))::rest671) => let val result=MlyValue.specnode(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (Spec.IncludeSigexp sigexp) end
)
 in (LrTable.NT 90,(result,INCLUDE1left,sigexp1right),rest671) end
| (71,(_,(MlyValue.sigids sigids1,_,sigids1right))::(_,(MlyValue.sigid
 sigid1,_,_))::(_,(_,INCLUDE1left,_))::rest671) => let val result=
MlyValue.specnode(fn _ => let val sigid as sigid1=sigid1 ()
val sigids as sigids1=sigids1 ()
 in (Spec.IncludeSigids(sigid::sigids)) end
)
 in (LrTable.NT 90,(result,INCLUDE1left,sigids1right),rest671) end
| (72,(_,(MlyValue.sharespec sharespec1,sharespecleft as 
sharespec1left,sharespecright as sharespec1right))::rest671) => let 
val result=MlyValue.specnode(fn _ => let val sharespec as sharespec1=
sharespec1 ()
 in (
Spec.Sharing
				 {spec = Spec.makeRegion'(Spec.Empty,
							  sharespecleft,
							  sharespecright),
				  equations =
				  [Equation.makeRegion'(sharespec,
							sharespecleft,
							sharespecright)]}
) end
)
 in (LrTable.NT 90,(result,sharespec1left,sharespec1right),rest671)
 end
| (73,(_,(MlyValue.longtyconeqns longtyconeqns1,_,longtyconeqns1right)
)::_::(_,(_,SHARING1left,_))::rest671) => let val result=
MlyValue.sharespec(fn _ => let val longtyconeqns as longtyconeqns1=
longtyconeqns1 ()
 in (Equation.Type longtyconeqns) end
)
 in (LrTable.NT 91,(result,SHARING1left,longtyconeqns1right),rest671)
 end
| (74,(_,(MlyValue.longstrideqns longstrideqns1,_,longstrideqns1right)
)::(_,(_,SHARING1left,_))::rest671) => let val result=
MlyValue.sharespec(fn _ => let val longstrideqns as longstrideqns1=
longstrideqns1 ()
 in (Equation.Structure longstrideqns) end
)
 in (LrTable.NT 91,(result,SHARING1left,longstrideqns1right),rest671)
 end
| (75,(_,(MlyValue.longstrid longstrid2,_,longstrid2right))::_::(_,(
MlyValue.longstrid longstrid1,longstrid1left,_))::rest671) => let val 
result=MlyValue.longstrideqns(fn _ => let val longstrid1=longstrid1 ()
val longstrid2=longstrid2 ()
 in ([longstrid1,longstrid2]) end
)
 in (LrTable.NT 103,(result,longstrid1left,longstrid2right),rest671)
 end
| (76,(_,(MlyValue.longstrideqns longstrideqns1,_,longstrideqns1right)
)::_::(_,(MlyValue.longstrid longstrid1,longstrid1left,_))::rest671)
 => let val result=MlyValue.longstrideqns(fn _ => let val longstrid
 as longstrid1=longstrid1 ()
val longstrideqns as longstrideqns1=longstrideqns1 ()
 in (longstrid :: longstrideqns) end
)
 in (LrTable.NT 103,(result,longstrid1left,longstrideqns1right),
rest671) end
| (77,(_,(MlyValue.longtycon longtycon2,_,longtycon2right))::_::(_,(
MlyValue.longtycon longtycon1,longtycon1left,_))::rest671) => let val 
result=MlyValue.longtyconeqns(fn _ => let val longtycon1=longtycon1 ()
val longtycon2=longtycon2 ()
 in ([longtycon1,longtycon2]) end
)
 in (LrTable.NT 102,(result,longtycon1left,longtycon2right),rest671)
 end
| (78,(_,(MlyValue.longtyconeqns longtyconeqns1,_,longtyconeqns1right)
)::_::(_,(MlyValue.longtycon longtycon1,longtycon1left,_))::rest671)
 => let val result=MlyValue.longtyconeqns(fn _ => let val longtycon
 as longtycon1=longtycon1 ()
val longtyconeqns as longtyconeqns1=longtyconeqns1 ()
 in (longtycon :: longtyconeqns) end
)
 in (LrTable.NT 102,(result,longtycon1left,longtyconeqns1right),
rest671) end
| (79,(_,(MlyValue.strdescs' strdescs'1,_,strdescs'1right))::(_,(
MlyValue.sigexp' sigexp'1,_,_))::_::(_,(MlyValue.strid strid1,
strid1left,_))::rest671) => let val result=MlyValue.strdescs(fn _ => 
let val strid as strid1=strid1 ()
val sigexp' as sigexp'1=sigexp'1 ()
val strdescs' as strdescs'1=strdescs'1 ()
 in (augment(strid, sigexp', strdescs')) end
)
 in (LrTable.NT 92,(result,strid1left,strdescs'1right),rest671) end
| (80,rest671) => let val result=MlyValue.strdescs'(fn _ => (
([], defaultPos, [])))
 in (LrTable.NT 93,(result,defaultPos,defaultPos),rest671) end
| (81,(_,(MlyValue.strdescs strdescs1,_,strdescs1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.strdescs'(fn _ => 
let val strdescs as strdescs1=strdescs1 ()
 in (([], defaultPos, strdescs)) end
)
 in (LrTable.NT 93,(result,AND1left,strdescs1right),rest671) end
| (82,(_,(MlyValue.strdescs'' strdescs''1,_,strdescs''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.strdescs'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strdescs'' as strdescs''1=strdescs''1 ()
 in (cons1(wherespec, strdescs'')) end
)
 in (LrTable.NT 93,(result,WHERE1left,strdescs''1right),rest671) end
| (83,(_,(MlyValue.strdescs' strdescs'1,strdescs'1left,strdescs'1right
))::rest671) => let val result=MlyValue.strdescs''(fn _ => let val 
strdescs' as strdescs'1=strdescs'1 ()
 in (strdescs') end
)
 in (LrTable.NT 94,(result,strdescs'1left,strdescs'1right),rest671)
 end
| (84,(_,(MlyValue.strdescs'' strdescs''1,_,strdescs''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.strdescs''(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strdescs'' as strdescs''1=strdescs''1 ()
 in (cons1(wherespec, strdescs'')) end
)
 in (LrTable.NT 94,(result,AND1left,strdescs''1right),rest671) end
| (85,(_,(MlyValue.typdesc typdesc1,typdesc1left,typdesc1right))::
rest671) => let val result=MlyValue.typdescs(fn _ => let val typdesc
 as typdesc1=typdesc1 ()
 in ([typdesc]) end
)
 in (LrTable.NT 95,(result,typdesc1left,typdesc1right),rest671) end
| (86,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::_::(_,(
MlyValue.typdesc typdesc1,typdesc1left,_))::rest671) => let val result
=MlyValue.typdescs(fn _ => let val typdesc as typdesc1=typdesc1 ()
val typdescs as typdescs1=typdescs1 ()
 in (typdesc :: typdescs) end
)
 in (LrTable.NT 95,(result,typdesc1left,typdescs1right),rest671) end
| (87,(_,(MlyValue.typdescRhs typdescRhs1,_,typdescRhs1right))::(_,(
MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvars1left,_
))::rest671) => let val result=MlyValue.typdesc(fn _ => let val tyvars
 as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val typdescRhs as typdescRhs1=typdescRhs1 ()
 in (
{tyvars = tyvars,
				     tycon = tycon,
				     ty = typdescRhs})
 end
)
 in (LrTable.NT 96,(result,tyvars1left,typdescRhs1right),rest671) end
| (88,rest671) => let val result=MlyValue.typdescRhs(fn _ => (NONE))
 in (LrTable.NT 97,(result,defaultPos,defaultPos),rest671) end
| (89,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,EQUALOP1left,_))::
rest671) => let val result=MlyValue.typdescRhs(fn _ => let val ty as 
ty1=ty1 ()
 in (SOME ty) end
)
 in (LrTable.NT 97,(result,EQUALOP1left,ty1right),rest671) end
| (90,(_,(MlyValue.valdesc valdesc1,valdesc1left,valdesc1right))::
rest671) => let val result=MlyValue.valdescs(fn _ => let val valdesc
 as valdesc1=valdesc1 ()
 in ([valdesc]) end
)
 in (LrTable.NT 98,(result,valdesc1left,valdesc1right),rest671) end
| (91,(_,(MlyValue.valdescs valdescs1,_,valdescs1right))::_::(_,(
MlyValue.valdesc valdesc1,valdesc1left,_))::rest671) => let val result
=MlyValue.valdescs(fn _ => let val valdesc as valdesc1=valdesc1 ()
val valdescs as valdescs1=valdescs1 ()
 in (valdesc :: valdescs) end
)
 in (LrTable.NT 98,(result,valdesc1left,valdescs1right),rest671) end
| (92,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.var var1,
var1left,_))::rest671) => let val result=MlyValue.valdesc(fn _ => let 
val var as var1=var1 ()
val ty as ty1=ty1 ()
 in ((var,ty)) end
)
 in (LrTable.NT 99,(result,var1left,ty1right),rest671) end
| (93,(_,(MlyValue.exndesc exndesc1,exndesc1left,exndesc1right))::
rest671) => let val result=MlyValue.exndescs(fn _ => let val exndesc
 as exndesc1=exndesc1 ()
 in ([exndesc]) end
)
 in (LrTable.NT 100,(result,exndesc1left,exndesc1right),rest671) end
| (94,(_,(MlyValue.exndescs exndescs1,_,exndescs1right))::_::(_,(
MlyValue.exndesc exndesc1,exndesc1left,_))::rest671) => let val result
=MlyValue.exndescs(fn _ => let val exndesc as exndesc1=exndesc1 ()
val exndescs as exndescs1=exndescs1 ()
 in (exndesc :: exndescs) end
)
 in (LrTable.NT 100,(result,exndesc1left,exndescs1right),rest671) end
| (95,(_,(MlyValue.tyOpt tyOpt1,_,tyOpt1right))::(_,(MlyValue.con con1
,con1left,_))::rest671) => let val result=MlyValue.exndesc(fn _ => 
let val con as con1=con1 ()
val tyOpt as tyOpt1=tyOpt1 ()
 in ((con, tyOpt)) end
)
 in (LrTable.NT 101,(result,con1left,tyOpt1right),rest671) end
| (96,rest671) => let val result=MlyValue.tyOpt(fn _ => (NONE))
 in (LrTable.NT 38,(result,defaultPos,defaultPos),rest671) end
| (97,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,OF1left,_))::rest671)
 => let val result=MlyValue.tyOpt(fn _ => let val ty as ty1=ty1 ()
 in (SOME ty) end
)
 in (LrTable.NT 38,(result,OF1left,ty1right),rest671) end
| (98,(_,(MlyValue.funbinds' funbinds'1,_,funbinds'1right))::_::(_,(
MlyValue.sigconst sigconst1,_,_))::_::(_,(MlyValue.fctarg fctarg1,
fctargleft,fctargright))::_::(_,(MlyValue.fctid fctid1,fctid1left,_))
::rest671) => let val result=MlyValue.funbinds(fn _ => let val fctid
 as fctid1=fctid1 ()
val fctarg as fctarg1=fctarg1 ()
val sigconst as sigconst1=sigconst1 ()
val funbinds' as funbinds'1=funbinds'1 ()
 in (
let val (strexp,funbinds) = funbinds'
	    in {name = fctid,
		arg = FctArg.makeRegion'(fctarg, fctargleft, fctargright),
		result = sigconst,
		body = strexp}
	       :: funbinds
	    end
) end
)
 in (LrTable.NT 124,(result,fctid1left,funbinds'1right),rest671) end
| (99,(_,(MlyValue.funbinds'1 funbinds'11,_,funbinds'11right))::(_,(
MlyValue.strexp1 strexp11,strexp11left,_))::rest671) => let val result
=MlyValue.funbinds'(fn _ => let val strexp1 as strexp11=strexp11 ()
val funbinds'1 as funbinds'11=funbinds'11 ()
 in (augment1(strexp1, funbinds'1)) end
)
 in (LrTable.NT 125,(result,strexp11left,funbinds'11right),rest671)
 end
| (100,(_,(MlyValue.funbinds'2 funbinds'21,_,funbinds'21right))::(_,(
MlyValue.strexp2 strexp21,strexp21left,_))::rest671) => let val result
=MlyValue.funbinds'(fn _ => let val strexp2 as strexp21=strexp21 ()
val funbinds'2 as funbinds'21=funbinds'21 ()
 in ((strexp2, funbinds'2)) end
)
 in (LrTable.NT 125,(result,strexp21left,funbinds'21right),rest671)
 end
| (101,(_,(MlyValue.funbinds'2 funbinds'21,funbinds'2left as 
funbinds'21left,funbinds'21right))::rest671) => let val result=
MlyValue.funbinds'1(fn _ => let val funbinds'2 as funbinds'21=
funbinds'21 ()
 in ([], funbinds'2left, funbinds'2) end
)
 in (LrTable.NT 126,(result,funbinds'21left,funbinds'21right),rest671)
 end
| (102,(_,(MlyValue.funbinds'1' funbinds'1'1,_,funbinds'1'1right))::(_
,(MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.funbinds'1(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val funbinds'1' as funbinds'1'1=funbinds'1'1 ()
 in (cons1(wherespec,funbinds'1')) end
)
 in (LrTable.NT 126,(result,WHERE1left,funbinds'1'1right),rest671) end
| (103,rest671) => let val result=MlyValue.funbinds'2(fn _ => ([]))
 in (LrTable.NT 128,(result,defaultPos,defaultPos),rest671) end
| (104,(_,(MlyValue.funbinds funbinds1,_,funbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.funbinds'2(fn _ => 
let val funbinds as funbinds1=funbinds1 ()
 in (funbinds) end
)
 in (LrTable.NT 128,(result,AND1left,funbinds1right),rest671) end
| (105,(_,(MlyValue.funbinds'1 funbinds'11,funbinds'11left,
funbinds'11right))::rest671) => let val result=MlyValue.funbinds'1'(
fn _ => let val funbinds'1 as funbinds'11=funbinds'11 ()
 in (funbinds'1) end
)
 in (LrTable.NT 127,(result,funbinds'11left,funbinds'11right),rest671)
 end
| (106,(_,(MlyValue.funbinds'1' funbinds'1'1,_,funbinds'1'1right))::(_
,(MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671)
 => let val result=MlyValue.funbinds'1'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val funbinds'1' as funbinds'1'1=funbinds'1'1 ()
 in (cons1(wherespec,funbinds'1')) end
)
 in (LrTable.NT 127,(result,AND1left,funbinds'1'1right),rest671) end
| (107,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::_::(_,(
MlyValue.strid strid1,strid1left,_))::rest671) => let val result=
MlyValue.fctarg(fn _ => let val strid as strid1=strid1 ()
val sigexp as sigexp1=sigexp1 ()
 in (FctArg.Structure(strid, sigexp)) end
)
 in (LrTable.NT 129,(result,strid1left,sigexp1right),rest671) end
| (108,(_,(MlyValue.specs specs1,specs1left,specs1right))::rest671)
 => let val result=MlyValue.fctarg(fn _ => let val specs as specs1=
specs1 ()
 in (FctArg.Spec specs) end
)
 in (LrTable.NT 129,(result,specs1left,specs1right),rest671) end
| (109,rest671) => let val result=MlyValue.decs(fn _ => (
Dec.makeRegion'(Dec.SeqDec nil,
					 defaultPos, defaultPos)))
 in (LrTable.NT 86,(result,defaultPos,defaultPos),rest671) end
| (110,(_,(MlyValue.decs decs1,_,decs1right))::(_,(MlyValue.dec dec1,
dec1left,_))::rest671) => let val result=MlyValue.decs(fn _ => let 
val dec as dec1=dec1 ()
val decs as decs1=decs1 ()
 in (Dec.sequence(dec,decs)) end
)
 in (LrTable.NT 86,(result,dec1left,decs1right),rest671) end
| (111,(_,(MlyValue.decs decs1,_,decs1right))::(_,(_,SEMICOLON1left,_)
)::rest671) => let val result=MlyValue.decs(fn _ => let val decs as 
decs1=decs1 ()
 in (decs) end
)
 in (LrTable.NT 86,(result,SEMICOLON1left,decs1right),rest671) end
| (112,(_,(MlyValue.decnode decnode1,decnodeleft as decnode1left,
decnoderight as decnode1right))::rest671) => let val result=
MlyValue.dec(fn _ => let val decnode as decnode1=decnode1 ()
 in (Dec.makeRegion'(decnode, decnodeleft, decnoderight)) end
)
 in (LrTable.NT 83,(result,decnode1left,decnode1right),rest671) end
| (113,(_,(MlyValue.decnolocal decnolocal1,decnolocal1left,
decnolocal1right))::rest671) => let val result=MlyValue.decnode(fn _
 => let val decnolocal as decnolocal1=decnolocal1 ()
 in (decnolocal) end
)
 in (LrTable.NT 84,(result,decnolocal1left,decnolocal1right),rest671)
 end
| (114,(_,(_,_,END1right))::(_,(MlyValue.decs decs2,_,_))::_::(_,(
MlyValue.decs decs1,_,_))::(_,(_,LOCAL1left,_))::rest671) => let val 
result=MlyValue.decnode(fn _ => let val decs1=decs1 ()
val decs2=decs2 ()
 in (Dec.Local(decs1,decs2)) end
)
 in (LrTable.NT 84,(result,LOCAL1left,END1right),rest671) end
| (115,(_,(MlyValue.vbs vbs1,_,vbs1right))::(_,(_,VAL1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val vbs as 
vbs1=vbs1 ()
 in (Dec.Val(nil,vbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,vbs1right),rest671) end
| (116,(_,(MlyValue.vbs vbs1,_,vbs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::(_,(_,VAL1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val vbs as vbs1=vbs1 ()
 in (Dec.Val(tyvarseq,vbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,vbs1right),rest671) end
| (117,(_,(MlyValue.rvbs rvbs1,_,rvbs1right))::_::(_,(_,VAL1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val rvbs
 as rvbs1=rvbs1 ()
 in (Dec.Valrec(nil,rvbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,rvbs1right),rest671) end
| (118,(_,(MlyValue.rvbs rvbs1,_,rvbs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::_::(_,(_,VAL1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val rvbs as rvbs1=rvbs1 ()
 in (Dec.Valrec(tyvarseq,rvbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,rvbs1right),rest671) end
| (119,(_,(MlyValue.funs funs1,_,funs1right))::(_,(_,FUN1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val funs
 as funs1=funs1 ()
 in (Dec.Fun(nil,funs)) end
)
 in (LrTable.NT 85,(result,FUN1left,funs1right),rest671) end
| (120,(_,(MlyValue.funs funs1,_,funs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::(_,(_,FUN1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val funs as funs1=funs1 ()
 in (Dec.Fun(tyvarseq,funs)) end
)
 in (LrTable.NT 85,(result,FUN1left,funs1right),rest671) end
| (121,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,
TYPE1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _ => 
let val typBind as typBind1=typBind1 ()
 in (Dec.Type typBind) end
)
 in (LrTable.NT 85,(result,TYPE1left,typBind1right),rest671) end
| (122,(_,(MlyValue.datatypeRhs datatypeRhs1,_,datatypeRhs1right))::(_
,(_,DATATYPE1left,_))::rest671) => let val result=MlyValue.decnolocal(
fn _ => let val datatypeRhs as datatypeRhs1=datatypeRhs1 ()
 in (Dec.Datatype datatypeRhs) end
)
 in (LrTable.NT 85,(result,DATATYPE1left,datatypeRhs1right),rest671)
 end
| (123,(_,(_,_,END1right))::(_,(MlyValue.decs decs1,_,_))::_::(_,(
MlyValue.datBind datBind1,_,_))::(_,(_,ABSTYPE1left,_))::rest671) => 
let val result=MlyValue.decnolocal(fn _ => let val datBind as datBind1
=datBind1 ()
val decs as decs1=decs1 ()
 in (Dec.Abstype{datBind = datBind,
						       body = decs}) end
)
 in (LrTable.NT 85,(result,ABSTYPE1left,END1right),rest671) end
| (124,(_,(MlyValue.ebs ebs1,_,ebs1right))::(_,(_,EXCEPTION1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val ebs as 
ebs1=ebs1 ()
 in (Dec.Exception ebs) end
)
 in (LrTable.NT 85,(result,EXCEPTION1left,ebs1right),rest671) end
| (125,(_,(MlyValue.longstrids longstrids1,_,longstrids1right))::(_,(_
,OPEN1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _
 => let val longstrids as longstrids1=longstrids1 ()
 in (Dec.Open longstrids) end
)
 in (LrTable.NT 85,(result,OPEN1left,longstrids1right),rest671) end
| (126,(_,(MlyValue.vids vids1,_,vids1right))::(_,(MlyValue.fixity 
fixity1,fixity1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val fixity as fixity1=fixity1 ()
val vids as vids1=vids1 ()
 in (Dec.Fix{fixity = fixity, ops = vids}) end
)
 in (LrTable.NT 85,(result,fixity1left,vids1right),rest671) end
| (127,(_,(MlyValue.longvarands longvarands1,_,longvarands1right))::_
::(_,(MlyValue.ty ty1,_,_))::_::(_,(MlyValue.var var1,_,_))::(_,(_,
OVERLOAD1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _
 => let val var as var1=var1 ()
val ty as ty1=ty1 ()
val longvarands as longvarands1=longvarands1 ()
 in (Dec.Overload(var, ty, longvarands)) end
)
 in (LrTable.NT 85,(result,OVERLOAD1left,longvarands1right),rest671)
 end
| (128,(_,(MlyValue.vbs vbs2,_,vbs2right))::_::(_,(MlyValue.vbs vbs1,
vbs1left,_))::rest671) => let val result=MlyValue.vbs(fn _ => let val 
vbs1=vbs1 ()
val vbs2=vbs2 ()
 in (vbs1 @ vbs2) end
)
 in (LrTable.NT 65,(result,vbs1left,vbs2right),rest671) end
| (129,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.vbs(fn _ => let val 
pat as pat1=pat1 ()
val exp as exp1=exp1 ()
 in ([(pat,exp)]) end
)
 in (LrTable.NT 65,(result,pat1left,exp1right),rest671) end
| (130,rest671) => let val result=MlyValue.constraint(fn _ => (NONE))
 in (LrTable.NT 66,(result,defaultPos,defaultPos),rest671) end
| (131,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,COLON1left,_))::rest671
) => let val result=MlyValue.constraint(fn _ => let val ty as ty1=ty1 
()
 in (SOME ty) end
)
 in (LrTable.NT 66,(result,COLON1left,ty1right),rest671) end
| (132,(_,(MlyValue.rules rules1,_,rules1right))::_::_::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.vid vid1,vid1left,
_))::rest671) => let val result=MlyValue.rvbs(fn _ => let val vid as 
vid1=vid1 ()
val constraint as constraint1=constraint1 ()
val rules as rules1=rules1 ()
 in (
[{var = Vid.toVar vid,
	     fixity = SOME vid,
	     ty = constraint,
	     rules = rules}]
) end
)
 in (LrTable.NT 67,(result,vid1left,rules1right),rest671) end
| (133,(_,(MlyValue.rules rules1,_,rules1right))::_::_::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.var var1,_,_))::(_
,(_,OP1left,_))::rest671) => let val result=MlyValue.rvbs(fn _ => let 
val var as var1=var1 ()
val constraint as constraint1=constraint1 ()
val rules as rules1=rules1 ()
 in (
[{var = var,
						 fixity = NONE,
						 ty = constraint,
						 rules = rules}]
) end
)
 in (LrTable.NT 67,(result,OP1left,rules1right),rest671) end
| (134,(_,(MlyValue.rvbs rvbs2,_,rvbs2right))::_::(_,(MlyValue.rvbs 
rvbs1,rvbs1left,_))::rest671) => let val result=MlyValue.rvbs(fn _ => 
let val rvbs1=rvbs1 ()
val rvbs2=rvbs2 ()
 in (rvbs1 @ rvbs2) end
)
 in (LrTable.NT 67,(result,rvbs1left,rvbs2right),rest671) end
| (135,(_,(MlyValue.clauses clauses1,clauses1left,clauses1right))::
rest671) => let val result=MlyValue.funs(fn _ => let val clauses as 
clauses1=clauses1 ()
 in ([clauses]) end
)
 in (LrTable.NT 68,(result,clauses1left,clauses1right),rest671) end
| (136,(_,(MlyValue.funs funs1,_,funs1right))::_::(_,(MlyValue.clauses
 clauses1,clauses1left,_))::rest671) => let val result=MlyValue.funs(
fn _ => let val clauses as clauses1=clauses1 ()
val funs as funs1=funs1 ()
 in (clauses :: funs) end
)
 in (LrTable.NT 68,(result,clauses1left,funs1right),rest671) end
| (137,(_,(MlyValue.clause clause1,clause1left,clause1right))::rest671
) => let val result=MlyValue.clauses(fn _ => let val clause as clause1
=clause1 ()
 in ([clause]) end
)
 in (LrTable.NT 69,(result,clause1left,clause1right),rest671) end
| (138,(_,(MlyValue.clauses clauses1,_,clauses1right))::_::(_,(
MlyValue.clause clause1,clause1left,_))::rest671) => let val result=
MlyValue.clauses(fn _ => let val clause as clause1=clause1 ()
val clauses as clauses1=clauses1 ()
 in (clause :: clauses) end
)
 in (LrTable.NT 69,(result,clause1left,clauses1right),rest671) end
| (139,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.constraint
 constraint1,_,_))::(_,(MlyValue.apats apats1,apats1left,_))::rest671)
 => let val result=MlyValue.clause(fn _ => let val apats as apats1=
apats1 ()
val constraint as constraint1=constraint1 ()
val exp as exp1=exp1 ()
 in ({pats = apats,
					  resultType = constraint,
					  body = exp}
) end
)
 in (LrTable.NT 70,(result,apats1left,exp1right),rest671) end
| (140,(_,(MlyValue.typBind' typBind'1,_,typBind'right as 
typBind'1right))::(_,(MlyValue.ty ty1,_,_))::_::(_,(MlyValue.tycon 
tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvarsleft as tyvars1left,_)
)::rest671) => let val result=MlyValue.typBind(fn _ => let val tyvars
 as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val ty as ty1=ty1 ()
val typBind' as typBind'1=typBind'1 ()
 in (
let val TypBind.T rest = typBind'
			 in TypBind.makeRegion'
			    (TypBind.T({tyvars = tyvars,
					tycon = tycon,
					def = ty} :: rest),
			     tyvarsleft, typBind'right)
			 end
) end
)
 in (LrTable.NT 71,(result,tyvars1left,typBind'1right),rest671) end
| (141,rest671) => let val result=MlyValue.typBind'(fn _ => (
TypBind.T[]))
 in (LrTable.NT 72,(result,defaultPos,defaultPos),rest671) end
| (142,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,AND1left
,_))::rest671) => let val result=MlyValue.typBind'(fn _ => let val 
typBind as typBind1=typBind1 ()
 in (TypBind.node typBind) end
)
 in (LrTable.NT 72,(result,AND1left,typBind1right),rest671) end
| (143,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvars(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 73,(result,tyvar1left,tyvar1right),rest671) end
| (144,(_,(_,_,RPAREN1right))::(_,(MlyValue.tyvar_pc tyvar_pc1,_,_))::
(_,(_,LPAREN1left,_))::rest671) => let val result=MlyValue.tyvars(fn _
 => let val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar_pc) end
)
 in (LrTable.NT 73,(result,LPAREN1left,RPAREN1right),rest671) end
| (145,rest671) => let val result=MlyValue.tyvars(fn _ => (nil))
 in (LrTable.NT 73,(result,defaultPos,defaultPos),rest671) end
| (146,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvarseq(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 74,(result,tyvar1left,tyvar1right),rest671) end
| (147,(_,(_,_,RPAREN1right))::(_,(MlyValue.tyvar_pc tyvar_pc1,_,_))::
(_,(_,LPAREN1left,_))::rest671) => let val result=MlyValue.tyvarseq(
fn _ => let val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar_pc) end
)
 in (LrTable.NT 74,(result,LPAREN1left,RPAREN1right),rest671) end
| (148,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvar_pc(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 75,(result,tyvar1left,tyvar1right),rest671) end
| (149,(_,(MlyValue.tyvar_pc tyvar_pc1,_,tyvar_pc1right))::_::(_,(
MlyValue.tyvar tyvar1,tyvar1left,_))::rest671) => let val result=
MlyValue.tyvar_pc(fn _ => let val tyvar as tyvar1=tyvar1 ()
val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar :: tyvar_pc) end
)
 in (LrTable.NT 75,(result,tyvar1left,tyvar_pc1right),rest671) end
| (150,(_,(MlyValue.constr constr1,constr1left,constr1right))::rest671
) => let val result=MlyValue.constrs(fn _ => let val constr as constr1
=constr1 ()
 in ([constr]) end
)
 in (LrTable.NT 76,(result,constr1left,constr1right),rest671) end
| (151,(_,(MlyValue.constrs constrs1,_,constrs1right))::_::(_,(
MlyValue.constr constr1,constr1left,_))::rest671) => let val result=
MlyValue.constrs(fn _ => let val constr as constr1=constr1 ()
val constrs as constrs1=constrs1 ()
 in (constr :: constrs) end
)
 in (LrTable.NT 76,(result,constr1left,constrs1right),rest671) end
| (152,(_,(MlyValue.con con1,con1left,con1right))::rest671) => let 
val result=MlyValue.constr(fn _ => let val con as con1=con1 ()
 in (con, NONE) end
)
 in (LrTable.NT 77,(result,con1left,con1right),rest671) end
| (153,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.con con1,
con1left,_))::rest671) => let val result=MlyValue.constr(fn _ => let 
val con as con1=con1 ()
val ty as ty1=ty1 ()
 in (con, SOME ty) end
)
 in (LrTable.NT 77,(result,con1left,ty1right),rest671) end
| (154,(_,(MlyValue.eb eb1,eb1left,eb1right))::rest671) => let val 
result=MlyValue.ebs(fn _ => let val eb as eb1=eb1 ()
 in ([eb]) end
)
 in (LrTable.NT 78,(result,eb1left,eb1right),rest671) end
| (155,(_,(MlyValue.ebs ebs1,_,ebs1right))::_::(_,(MlyValue.eb eb1,
eb1left,_))::rest671) => let val result=MlyValue.ebs(fn _ => let val 
eb as eb1=eb1 ()
val ebs as ebs1=ebs1 ()
 in (eb::ebs) end
)
 in (LrTable.NT 78,(result,eb1left,ebs1right),rest671) end
| (156,(_,(MlyValue.ebrhs ebrhs1,_,ebrhs1right))::(_,(MlyValue.con 
con1,con1left,_))::rest671) => let val result=MlyValue.eb(fn _ => let 
val con as con1=con1 ()
val ebrhs as ebrhs1=ebrhs1 ()
 in ((con,ebrhs)) end
)
 in (LrTable.NT 79,(result,con1left,ebrhs1right),rest671) end
| (157,(_,(MlyValue.ebrhsnode ebrhsnode1,ebrhsnodeleft as 
ebrhsnode1left,ebrhsnoderight as ebrhsnode1right))::rest671) => let 
val result=MlyValue.ebrhs(fn _ => let val ebrhsnode as ebrhsnode1=
ebrhsnode1 ()
 in (
EbRhs.makeRegion'(ebrhsnode,
				     ebrhsnodeleft, ebrhsnoderight))
 end
)
 in (LrTable.NT 80,(result,ebrhsnode1left,ebrhsnode1right),rest671)
 end
| (158,rest671) => let val result=MlyValue.ebrhsnode(fn _ => (
EbRhs.Gen NONE))
 in (LrTable.NT 81,(result,defaultPos,defaultPos),rest671) end
| (159,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,OF1left,_))::rest671)
 => let val result=MlyValue.ebrhsnode(fn _ => let val ty as ty1=ty1 ()
 in (EbRhs.Gen(SOME ty)) end
)
 in (LrTable.NT 81,(result,OF1left,ty1right),rest671) end
| (160,(_,(MlyValue.longcon longcon1,_,longcon1right))::(_,(_,
EQUALOP1left,_))::rest671) => let val result=MlyValue.ebrhsnode(fn _
 => let val longcon as longcon1=longcon1 ()
 in (EbRhs.Def longcon) end
)
 in (LrTable.NT 81,(result,EQUALOP1left,longcon1right),rest671) end
| (161,(_,(_,INFIX1left,INFIX1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Infix NONE))
 in (LrTable.NT 82,(result,INFIX1left,INFIX1right),rest671) end
| (162,(_,(MlyValue.digit digit1,_,digit1right))::(_,(_,INFIX1left,_))
::rest671) => let val result=MlyValue.fixity(fn _ => let val digit as 
digit1=digit1 ()
 in (Fixity.Infix(SOME digit)) end
)
 in (LrTable.NT 82,(result,INFIX1left,digit1right),rest671) end
| (163,(_,(_,INFIXR1left,INFIXR1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Infixr NONE))
 in (LrTable.NT 82,(result,INFIXR1left,INFIXR1right),rest671) end
| (164,(_,(MlyValue.digit digit1,_,digit1right))::(_,(_,INFIXR1left,_)
)::rest671) => let val result=MlyValue.fixity(fn _ => let val digit
 as digit1=digit1 ()
 in (Fixity.Infixr(SOME digit)) end
)
 in (LrTable.NT 82,(result,INFIXR1left,digit1right),rest671) end
| (165,(_,(_,NONFIX1left,NONFIX1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Nonfix))
 in (LrTable.NT 82,(result,NONFIX1left,NONFIX1right),rest671) end
| (166,(_,(MlyValue.INT INT1,INTleft as INT1left,INTright as INT1right
))::rest671) => let val result=MlyValue.int(fn _ => let val INT as 
INT1=INT1 ()
 in (
case Int.fromString INT of
	      NONE => (error(reg(INTleft, INTright), "expected integer");
		       ~1)
	    | SOME n => n
) end
)
 in (LrTable.NT 1,(result,INT1left,INT1right),rest671) end
| (167,(_,(MlyValue.int int1,intleft as int1left,intright as int1right
))::rest671) => let val result=MlyValue.digit(fn _ => let val int as 
int1=int1 ()
 in (
if int < 0 orelse int >= 10
		then (error(reg(intleft, intright), "expected single digit") ;
		      0)
	     else int
) end
)
 in (LrTable.NT 0,(result,int1left,int1right),rest671) end
| (168,(_,(MlyValue.datatypeRhsnode datatypeRhsnode1,
datatypeRhsnodeleft as datatypeRhsnode1left,datatypeRhsnoderight as 
datatypeRhsnode1right))::rest671) => let val result=
MlyValue.datatypeRhs(fn _ => let val datatypeRhsnode as 
datatypeRhsnode1=datatypeRhsnode1 ()
 in (
DatatypeRhs.makeRegion'(datatypeRhsnode,
						       datatypeRhsnodeleft,
						       datatypeRhsnoderight)
) end
)
 in (LrTable.NT 130,(result,datatypeRhsnode1left,datatypeRhsnode1right
),rest671) end
| (169,(_,(MlyValue.datatypeRhsnodeNoWithtype 
datatypeRhsnodeNoWithtype1,datatypeRhsnodeNoWithtypeleft as 
datatypeRhsnodeNoWithtype1left,datatypeRhsnodeNoWithtyperight as 
datatypeRhsnodeNoWithtype1right))::rest671) => let val result=
MlyValue.datatypeRhsNoWithtype(fn _ => let val 
datatypeRhsnodeNoWithtype as datatypeRhsnodeNoWithtype1=
datatypeRhsnodeNoWithtype1 ()
 in (
DatatypeRhs.makeRegion'(datatypeRhsnodeNoWithtype,
						 datatypeRhsnodeNoWithtypeleft,
						 datatypeRhsnodeNoWithtyperight)
) end
)
 in (LrTable.NT 131,(result,datatypeRhsnodeNoWithtype1left,
datatypeRhsnodeNoWithtype1right),rest671) end
| (170,(_,(MlyValue.repl repl1,repl1left,repl1right))::rest671) => 
let val result=MlyValue.datatypeRhsnode(fn _ => let val repl as repl1=
repl1 ()
 in (repl) end
)
 in (LrTable.NT 132,(result,repl1left,repl1right),rest671) end
| (171,(_,(MlyValue.datBind datBind1,datBind1left,datBind1right))::
rest671) => let val result=MlyValue.datatypeRhsnode(fn _ => let val 
datBind as datBind1=datBind1 ()
 in (DatatypeRhs.DatBind datBind) end
)
 in (LrTable.NT 132,(result,datBind1left,datBind1right),rest671) end
| (172,(_,(MlyValue.repl repl1,repl1left,repl1right))::rest671) => 
let val result=MlyValue.datatypeRhsnodeNoWithtype(fn _ => let val repl
 as repl1=repl1 ()
 in (repl) end
)
 in (LrTable.NT 133,(result,repl1left,repl1right),rest671) end
| (173,(_,(MlyValue.datBindNoWithtype datBindNoWithtype1,
datBindNoWithtype1left,datBindNoWithtype1right))::rest671) => let val 
result=MlyValue.datatypeRhsnodeNoWithtype(fn _ => let val 
datBindNoWithtype as datBindNoWithtype1=datBindNoWithtype1 ()
 in (DatatypeRhs.DatBind datBindNoWithtype) end
)
 in (LrTable.NT 133,(result,datBindNoWithtype1left,
datBindNoWithtype1right),rest671) end
| (174,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::_::_::(_
,(MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvarsleft
 as tyvars1left,tyvarsright))::rest671) => let val result=
MlyValue.repl(fn _ => let val tyvars as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val longtycon as longtycon1=longtycon1 ()
 in (
case tyvars of
	   _ :: _ => error(reg(tyvarsleft, tyvarsright),
			   "nonempty tyvars in datatype repl")
	 | _ => () ;
        DatatypeRhs.Repl{lhs = tycon, rhs = longtycon}
) end
)
 in (LrTable.NT 134,(result,tyvars1left,longtycon1right),rest671) end
| (175,(_,(MlyValue.withtypes withtypes1,_,withtypesright as 
withtypes1right))::(_,(MlyValue.dbs dbs1,dbsleft as dbs1left,_))::
rest671) => let val result=MlyValue.datBind(fn _ => let val dbs as 
dbs1=dbs1 ()
val withtypes as withtypes1=withtypes1 ()
 in (
DatBind.makeRegion'(DatBind.T{datatypes = dbs,
						       withtypes = withtypes},
					     dbsleft, withtypesright)
) end
)
 in (LrTable.NT 135,(result,dbs1left,withtypes1right),rest671) end
| (176,(_,(MlyValue.dbs dbs1,dbsleft as dbs1left,dbsright as dbs1right
))::rest671) => let val result=MlyValue.datBindNoWithtype(fn _ => let 
val dbs as dbs1=dbs1 ()
 in (
DatBind.makeRegion'(DatBind.T{datatypes = dbs,
					  withtypes = TypBind.empty},
				dbsleft, dbsright)
) end
)
 in (LrTable.NT 136,(result,dbs1left,dbs1right),rest671) end
| (177,(_,(MlyValue.db db1,db1left,db1right))::rest671) => let val 
result=MlyValue.dbs(fn _ => let val db as db1=db1 ()
 in ([db]) end
)
 in (LrTable.NT 138,(result,db1left,db1right),rest671) end
| (178,(_,(MlyValue.dbs dbs1,_,dbs1right))::_::(_,(MlyValue.db db1,
db1left,_))::rest671) => let val result=MlyValue.dbs(fn _ => let val 
db as db1=db1 ()
val dbs as dbs1=dbs1 ()
 in (db :: dbs) end
)
 in (LrTable.NT 138,(result,db1left,dbs1right),rest671) end
| (179,(_,(MlyValue.constrs constrs1,_,constrs1right))::_::(_,(
MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvars1left,_
))::rest671) => let val result=MlyValue.db(fn _ => let val tyvars as 
tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val constrs as constrs1=constrs1 ()
 in ({tycon = tycon,
					  tyvars = tyvars,
					  cons = constrs})
 end
)
 in (LrTable.NT 137,(result,tyvars1left,constrs1right),rest671) end
| (180,rest671) => let val result=MlyValue.withtypes(fn _ => (
TypBind.empty))
 in (LrTable.NT 139,(result,defaultPos,defaultPos),rest671) end
| (181,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,
WITHTYPE1left,_))::rest671) => let val result=MlyValue.withtypes(fn _
 => let val typBind as typBind1=typBind1 ()
 in (typBind) end
)
 in (LrTable.NT 139,(result,WITHTYPE1left,typBind1right),rest671) end
| (182,(_,(MlyValue.longvar longvar1,longvar1left,longvar1right))::
rest671) => let val result=MlyValue.longvarands(fn _ => let val 
longvar as longvar1=longvar1 ()
 in ([longvar]) end
)
 in (LrTable.NT 24,(result,longvar1left,longvar1right),rest671) end
| (183,(_,(MlyValue.longvarands longvarands1,_,longvarands1right))::_
::(_,(MlyValue.longvar longvar1,longvar1left,_))::rest671) => let val 
result=MlyValue.longvarands(fn _ => let val longvar as longvar1=
longvar1 ()
val longvarands as longvarands1=longvarands1 ()
 in (longvar :: longvarands) end
)
 in (LrTable.NT 24,(result,longvar1left,longvarands1right),rest671)
 end
| (184,(_,(MlyValue.rule rule1,rule1left,rule1right))::rest671) => 
let val result=MlyValue.rules(fn _ => let val rule as rule1=rule1 ()
 in ([rule]) end
)
 in (LrTable.NT 40,(result,rule1left,rule1right),rest671) end
| (185,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.rule
 rule1,rule1left,_))::rest671) => let val result=MlyValue.rules(fn _
 => let val rule as rule1=rule1 ()
val rules as rules1=rules1 ()
 in (rule::rules) end
)
 in (LrTable.NT 40,(result,rule1left,rules1right),rest671) end
| (186,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.rule(fn _ => let 
val pat as pat1=pat1 ()
val exp as exp1=exp1 ()
 in ((pat,exp)) end
)
 in (LrTable.NT 41,(result,pat1left,exp1right),rest671) end
| (187,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.field 
field1,field1left,_))::rest671) => let val result=MlyValue.elabel(fn _
 => let val field as field1=field1 ()
val exp as exp1=exp1 ()
 in (field,exp) end
)
 in (LrTable.NT 42,(result,field1left,exp1right),rest671) end
| (188,(_,(MlyValue.elabels elabels1,_,elabels1right))::_::(_,(
MlyValue.elabel elabel1,elabel1left,_))::rest671) => let val result=
MlyValue.elabels(fn _ => let val elabel as elabel1=elabel1 ()
val elabels as elabels1=elabels1 ()
 in (elabel :: elabels) end
)
 in (LrTable.NT 43,(result,elabel1left,elabels1right),rest671) end
| (189,(_,(MlyValue.elabel elabel1,elabel1left,elabel1right))::rest671
) => let val result=MlyValue.elabels(fn _ => let val elabel as elabel1
=elabel1 ()
 in ([elabel]) end
)
 in (LrTable.NT 43,(result,elabel1left,elabel1right),rest671) end
| (190,(_,(MlyValue.exp exp1,exp1left,exp1right))::rest671) => let 
val result=MlyValue.exp_ps(fn _ => let val exp as exp1=exp1 ()
 in ([exp]) end
)
 in (LrTable.NT 44,(result,exp1left,exp1right),rest671) end
| (191,(_,(MlyValue.exp_ps exp_ps1,_,exp_ps1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_ps(fn _ => let val exp as exp1=exp1 ()
val exp_ps as exp_ps1=exp_ps1 ()
 in (exp :: exp_ps) end
)
 in (LrTable.NT 44,(result,exp1left,exp_ps1right),rest671) end
| (192,(_,(MlyValue.expnode expnode1,expnodeleft as expnode1left,
expnoderight as expnode1right))::rest671) => let val result=
MlyValue.exp(fn _ => let val expnode as expnode1=expnode1 ()
 in (Exp.makeRegion'(expnode, expnodeleft, expnoderight)) end
)
 in (LrTable.NT 45,(result,expnode1left,expnode1right),rest671) end
| (193,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.exp 
exp1,exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _
 => let val exp as exp1=exp1 ()
val rules as rules1=rules1 ()
 in (Exp.Handle(exp,rules)) end
)
 in (LrTable.NT 46,(result,exp1left,rules1right),rest671) end
| (194,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.Orelse(exp1,exp2)) end
)
 in (LrTable.NT 46,(result,exp1left,exp2right),rest671) end
| (195,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.Andalso(exp1,exp2)) end
)
 in (LrTable.NT 46,(result,exp1left,exp2right),rest671) end
| (196,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp as exp1=exp1 ()
val ty as ty1=ty1 ()
 in (Exp.Constraint(exp,ty)) end
)
 in (LrTable.NT 46,(result,exp1left,ty1right),rest671) end
| (197,(_,(MlyValue.app_exp app_exp1,app_exp1left,app_exp1right))::
rest671) => let val result=MlyValue.expnode(fn _ => let val app_exp
 as app_exp1=app_exp1 ()
 in (Exp.FlatApp app_exp) end
)
 in (LrTable.NT 46,(result,app_exp1left,app_exp1right),rest671) end
| (198,(_,(MlyValue.rules rules1,_,rules1right))::(_,(_,FN1left,_))::
rest671) => let val result=MlyValue.expnode(fn _ => let val rules as 
rules1=rules1 ()
 in (Exp.Fn rules) end
)
 in (LrTable.NT 46,(result,FN1left,rules1right),rest671) end
| (199,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.exp 
exp1,_,_))::(_,(_,CASE1left,_))::rest671) => let val result=
MlyValue.expnode(fn _ => let val exp as exp1=exp1 ()
val rules as rules1=rules1 ()
 in (Exp.Case(exp,rules)) end
)
 in (LrTable.NT 46,(result,CASE1left,rules1right),rest671) end
| (200,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,_
,_))::(_,(_,WHILE1left,_))::rest671) => let val result=
MlyValue.expnode(fn _ => let val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.While{test = exp1,expr = exp2}) end
)
 in (LrTable.NT 46,(result,WHILE1left,exp2right),rest671) end
| (201,(_,(MlyValue.exp exp3,_,exp3right))::_::(_,(MlyValue.exp exp2,_
,_))::_::(_,(MlyValue.exp exp1,_,_))::(_,(_,IF1left,_))::rest671) => 
let val result=MlyValue.expnode(fn _ => let val exp1=exp1 ()
val exp2=exp2 ()
val exp3=exp3 ()
 in (Exp.If(exp1,exp2,exp3)) end
)
 in (LrTable.NT 46,(result,IF1left,exp3right),rest671) end
| (202,(_,(MlyValue.exp exp1,_,exp1right))::(_,(_,RAISE1left,_))::
rest671) => let val result=MlyValue.expnode(fn _ => let val exp as 
exp1=exp1 ()
 in (Exp.Raise exp) end
)
 in (LrTable.NT 46,(result,RAISE1left,exp1right),rest671) end
| (203,(_,(MlyValue.app_exp1 app_exp11,_,app_exp11right))::(_,(
MlyValue.aexp aexp1,aexpleft as aexp1left,aexpright))::rest671) => 
let val result=MlyValue.app_exp(fn _ => let val aexp as aexp1=aexp1 ()
val app_exp1 as app_exp11=app_exp11 ()
 in (Exp.makeRegion'(aexp,aexpleft,aexpright)
			     :: app_exp1) end
)
 in (LrTable.NT 47,(result,aexp1left,app_exp11right),rest671) end
| (204,(_,(MlyValue.app_exp1 app_exp11,_,app_exp11right))::(_,(
MlyValue.longvid longvid1,longvidleft as longvid1left,longvidright))::
rest671) => let val result=MlyValue.app_exp(fn _ => let val longvid
 as longvid1=longvid1 ()
val app_exp1 as app_exp11=app_exp11 ()
 in (
Exp.makeRegion'(Exp.Var{name = longvid,
						     fixop = Fixop.None},
					     longvidleft, longvidright)
			     :: app_exp1
) end
)
 in (LrTable.NT 47,(result,longvid1left,app_exp11right),rest671) end
| (205,rest671) => let val result=MlyValue.app_exp1(fn _ => ([]))
 in (LrTable.NT 48,(result,defaultPos,defaultPos),rest671) end
| (206,(_,(MlyValue.app_exp app_exp1,app_exp1left,app_exp1right))::
rest671) => let val result=MlyValue.app_exp1(fn _ => let val app_exp
 as app_exp1=app_exp1 ()
 in (app_exp) end
)
 in (LrTable.NT 48,(result,app_exp1left,app_exp1right),rest671) end
| (207,(_,(MlyValue.vid vid1,_,vid1right))::(_,(_,OP1left,_))::rest671
) => let val result=MlyValue.aexp(fn _ => let val vid as vid1=vid1 ()
 in (Exp.Var{name = Longvid.short vid,
					 fixop = Fixop.Op}) end
)
 in (LrTable.NT 49,(result,OP1left,vid1right),rest671) end
| (208,(_,(MlyValue.const const1,const1left,const1right))::rest671)
 => let val result=MlyValue.aexp(fn _ => let val const as const1=
const1 ()
 in (Exp.Const const) end
)
 in (LrTable.NT 49,(result,const1left,const1right),rest671) end
| (209,(_,(MlyValue.field field1,_,field1right))::(_,(_,HASH1left,_))
::rest671) => let val result=MlyValue.aexp(fn _ => let val field as 
field1=field1 ()
 in (Exp.Selector field) end
)
 in (LrTable.NT 49,(result,HASH1left,field1right),rest671) end
| (210,(_,(_,_,RBRACE1right))::(_,(MlyValue.elabels elabels1,_,_))::(_
,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val elabels as elabels1=elabels1 ()
 in (Exp.Record(Record.fromList elabels)) end
)
 in (LrTable.NT 49,(result,LBRACE1left,RBRACE1right),rest671) end
| (211,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => (Exp.unit))
 in (LrTable.NT 49,(result,LBRACE1left,RBRACE1right),rest671) end
| (212,(_,(_,_,RPAREN1right))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => (Exp.unit))
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (213,(_,(_,_,RPAREN1right))::(_,(MlyValue.exp_ps exp_ps1,_,_))::(_,(
_,LPAREN1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val exp_ps as exp_ps1=exp_ps1 ()
 in (Exp.Seq exp_ps) end
)
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (214,(_,(_,_,RPAREN1right))::(_,(MlyValue.exp_2c exp_2c1,_,_))::(_,(
_,LPAREN1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val exp_2c as exp_2c1=exp_2c1 ()
 in (Exp.tuple exp_2c) end
)
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (215,(_,(_,_,RBRACKET1right))::(_,(MlyValue.exp_list exp_list1,_,_))
::(_,(_,LBRACKET1left,_))::rest671) => let val result=MlyValue.aexp(
fn _ => let val exp_list as exp_list1=exp_list1 ()
 in (Exp.List exp_list) end
)
 in (LrTable.NT 49,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (216,(_,(_,_,RBRACKET1right))::(_,(_,LBRACKET1left,_))::rest671) => 
let val result=MlyValue.aexp(fn _ => (Exp.List nil))
 in (LrTable.NT 49,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (217,(_,(_,_,END1right))::(_,(MlyValue.exp_ps exp_ps1,exp_psleft,
exp_psright))::_::(_,(MlyValue.decs decs1,_,_))::(_,(_,LET1left,_))::
rest671) => let val result=MlyValue.aexp(fn _ => let val decs as decs1
=decs1 ()
val exp_ps as exp_ps1=exp_ps1 ()
 in (
Exp.Let(decs,
					       Exp.makeRegion'(Exp.Seq exp_ps,
							       exp_psleft,
							       exp_psright))
) end
)
 in (LrTable.NT 49,(result,LET1left,END1right),rest671) end
| (218,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.ty ty1,_,_))::_::(_,(
MlyValue.STRING STRING1,_,_))::(_,(_,PRIM1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => let val STRING as STRING1=STRING1 ()
val ty as ty1=ty1 ()
 in (Exp.Prim{name = STRING,
						   ty = ty}) end
)
 in (LrTable.NT 49,(result,PRIM1left,SEMICOLON1right),rest671) end
| (219,(_,(MlyValue.exp_2c exp_2c1,_,exp_2c1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_2c(fn _ => let val exp as exp1=exp1 ()
val exp_2c as exp_2c1=exp_2c1 ()
 in (exp :: exp_2c) end
)
 in (LrTable.NT 51,(result,exp1left,exp_2c1right),rest671) end
| (220,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.exp_2c(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in ([exp1, exp2]) end
)
 in (LrTable.NT 51,(result,exp1left,exp2right),rest671) end
| (221,(_,(MlyValue.exp exp1,exp1left,exp1right))::rest671) => let 
val result=MlyValue.exp_list(fn _ => let val exp as exp1=exp1 ()
 in ([exp]) end
)
 in (LrTable.NT 52,(result,exp1left,exp1right),rest671) end
| (222,(_,(MlyValue.exp_list exp_list1,_,exp_list1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_list(fn _ => let val exp as exp1=exp1 ()
val exp_list as exp_list1=exp_list1 ()
 in (exp :: exp_list) end
)
 in (LrTable.NT 52,(result,exp1left,exp_list1right),rest671) end
| (223,(_,(MlyValue.patnode patnode1,patnodeleft as patnode1left,
patnoderight as patnode1right))::rest671) => let val result=
MlyValue.pat(fn _ => let val patnode as patnode1=patnode1 ()
 in (Pat.makeRegion'(patnode, patnodeleft, patnoderight)) end
)
 in (LrTable.NT 53,(result,patnode1left,patnode1right),rest671) end
| (224,(_,(MlyValue.pat pat2,_,pat2right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.patnode(fn _ => let 
val pat1=pat1 ()
val pat2=pat2 ()
 in (Pat.makeAs(pat1,pat2)) end
)
 in (LrTable.NT 54,(result,pat1left,pat2right),rest671) end
| (225,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.patnode(fn _ => let 
val pat as pat1=pat1 ()
val ty as ty1=ty1 ()
 in (Pat.Constraint(pat,ty)) end
)
 in (LrTable.NT 54,(result,pat1left,ty1right),rest671) end
| (226,(_,(MlyValue.apats apats1,apats1left,apats1right))::rest671)
 => let val result=MlyValue.patnode(fn _ => let val apats as apats1=
apats1 ()
 in (Pat.FlatApp apats) end
)
 in (LrTable.NT 54,(result,apats1left,apats1right),rest671) end
| (227,(_,(MlyValue.apat apat1,apat1left,apat1right))::rest671) => 
let val result=MlyValue.apats(fn _ => let val apat as apat1=apat1 ()
 in ([apat]) end
)
 in (LrTable.NT 55,(result,apat1left,apat1right),rest671) end
| (228,(_,(MlyValue.apats apats1,_,apats1right))::(_,(MlyValue.apat 
apat1,apat1left,_))::rest671) => let val result=MlyValue.apats(fn _
 => let val apat as apat1=apat1 ()
val apats as apats1=apats1 ()
 in (apat :: apats) end
)
 in (LrTable.NT 55,(result,apat1left,apats1right),rest671) end
| (229,(_,(MlyValue.apatnode apatnode1,apatnodeleft as apatnode1left,
apatnoderight as apatnode1right))::rest671) => let val result=
MlyValue.apat(fn _ => let val apatnode as apatnode1=apatnode1 ()
 in (Pat.makeRegion'(apatnode,apatnodeleft,apatnoderight)) end
)
 in (LrTable.NT 56,(result,apatnode1left,apatnode1right),rest671) end
| (230,(_,(MlyValue.longvidNoEqual longvidNoEqual1,longvidNoEqual1left
,longvidNoEqual1right))::rest671) => let val result=MlyValue.apatnode(
fn _ => let val longvidNoEqual as longvidNoEqual1=longvidNoEqual1 ()
 in (Pat.Var{name = longvidNoEqual,
 					 fixop = Fixop.None}) end
)
 in (LrTable.NT 57,(result,longvidNoEqual1left,longvidNoEqual1right),
rest671) end
| (231,(_,(MlyValue.vidNoEqual vidNoEqual1,_,vidNoEqual1right))::(_,(_
,OP1left,_))::rest671) => let val result=MlyValue.apatnode(fn _ => 
let val vidNoEqual as vidNoEqual1=vidNoEqual1 ()
 in (Pat.Var{name = Longvid.short vidNoEqual,
					 fixop = Fixop.Op})
 end
)
 in (LrTable.NT 57,(result,OP1left,vidNoEqual1right),rest671) end
| (232,(_,(MlyValue.const const1,const1left,const1right))::rest671)
 => let val result=MlyValue.apatnode(fn _ => let val const as const1=
const1 ()
 in (Pat.Const const) end
)
 in (LrTable.NT 57,(result,const1left,const1right),rest671) end
| (233,(_,(_,WILD1left,WILD1right))::rest671) => let val result=
MlyValue.apatnode(fn _ => (Pat.Wild))
 in (LrTable.NT 57,(result,WILD1left,WILD1right),rest671) end
| (234,(_,(_,_,RPAREN1right))::(_,(MlyValue.pats pats1,_,_))::(_,(_,
LPAREN1left,_))::rest671) => let val result=MlyValue.apatnode(fn _ => 
let val pats as pats1=pats1 ()
 in (Pat.tuple pats) end
)
 in (LrTable.NT 57,(result,LPAREN1left,RPAREN1right),rest671) end
| (235,(_,(_,_,RBRACKET1right))::(_,(MlyValue.pats pats1,_,_))::(_,(_,
LBRACKET1left,_))::rest671) => let val result=MlyValue.apatnode(fn _
 => let val pats as pats1=pats1 ()
 in (Pat.List pats) end
)
 in (LrTable.NT 57,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (236,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.apatnode(fn _ => (Pat.unit))
 in (LrTable.NT 57,(result,LBRACE1left,RBRACE1right),rest671) end
| (237,(_,(_,_,RBRACE1right))::(_,(MlyValue.patitems patitems1,_,_))::
(_,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.apatnode(
fn _ => let val patitems as patitems1=patitems1 ()
 in (
let val (items,f) = patitems
				  in Pat.Record{items = items, flexible = f}
				  end
) end
)
 in (LrTable.NT 57,(result,LBRACE1left,RBRACE1right),rest671) end
| (238,rest671) => let val result=MlyValue.pats(fn _ => ([]))
 in (LrTable.NT 62,(result,defaultPos,defaultPos),rest671) end
| (239,(_,(MlyValue.commapats commapats1,_,commapats1right))::(_,(
MlyValue.pat pat1,pat1left,_))::rest671) => let val result=
MlyValue.pats(fn _ => let val pat as pat1=pat1 ()
val commapats as commapats1=commapats1 ()
 in (pat :: commapats) end
)
 in (LrTable.NT 62,(result,pat1left,commapats1right),rest671) end
| (240,rest671) => let val result=MlyValue.commapats(fn _ => ([]))
 in (LrTable.NT 63,(result,defaultPos,defaultPos),rest671) end
| (241,(_,(MlyValue.commapats commapats1,_,commapats1right))::(_,(
MlyValue.pat pat1,_,_))::(_,(_,COMMA1left,_))::rest671) => let val 
result=MlyValue.commapats(fn _ => let val pat as pat1=pat1 ()
val commapats as commapats1=commapats1 ()
 in (pat :: commapats) end
)
 in (LrTable.NT 63,(result,COMMA1left,commapats1right),rest671) end
| (242,(_,(MlyValue.patitems patitems1,_,patitems1right))::_::(_,(
MlyValue.patitem patitem1,patitem1left,_))::rest671) => let val result
=MlyValue.patitems(fn _ => let val patitem as patitem1=patitem1 ()
val patitems as patitems1=patitems1 ()
 in (
let val (items,f) = patitems
				    in (patitem::items, f)
				    end
) end
)
 in (LrTable.NT 59,(result,patitem1left,patitems1right),rest671) end
| (243,(_,(MlyValue.patitem patitem1,patitem1left,patitem1right))::
rest671) => let val result=MlyValue.patitems(fn _ => let val patitem
 as patitem1=patitem1 ()
 in ([patitem],false) end
)
 in (LrTable.NT 59,(result,patitem1left,patitem1right),rest671) end
| (244,(_,(_,DOTDOTDOT1left,DOTDOTDOT1right))::rest671) => let val 
result=MlyValue.patitems(fn _ => ([], true))
 in (LrTable.NT 59,(result,DOTDOTDOT1left,DOTDOTDOT1right),rest671)
 end
| (245,(_,(MlyValue.pat pat1,_,pat1right))::_::(_,(MlyValue.field 
field1,field1left,_))::rest671) => let val result=MlyValue.patitem(fn 
_ => let val field as field1=field1 ()
val pat as pat1=pat1 ()
 in (Pat.Item.Field(field,pat)) end
)
 in (LrTable.NT 60,(result,field1left,pat1right),rest671) end
| (246,(_,(MlyValue.opaspat opaspat1,_,opaspat1right))::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.vidNoEqual 
vidNoEqual1,vidNoEqual1left,_))::rest671) => let val result=
MlyValue.patitem(fn _ => let val vidNoEqual as vidNoEqual1=vidNoEqual1
 ()
val constraint as constraint1=constraint1 ()
val opaspat as opaspat1=opaspat1 ()
 in (Pat.Item.Vid(vidNoEqual,
						       constraint, opaspat)) end
)
 in (LrTable.NT 60,(result,vidNoEqual1left,opaspat1right),rest671) end
| (247,rest671) => let val result=MlyValue.opaspat(fn _ => (NONE))
 in (LrTable.NT 64,(result,defaultPos,defaultPos),rest671) end
| (248,(_,(MlyValue.pat pat1,_,pat1right))::(_,(_,AS1left,_))::rest671
) => let val result=MlyValue.opaspat(fn _ => let val pat as pat1=pat1 
()
 in (SOME pat) end
)
 in (LrTable.NT 64,(result,AS1left,pat1right),rest671) end
| (249,(_,(MlyValue.tynode tynode1,tynodeleft as tynode1left,
tynoderight as tynode1right))::rest671) => let val result=MlyValue.ty(
fn _ => let val tynode as tynode1=tynode1 ()
 in (Type.makeRegion'(tynode, tynodeleft, tynoderight)) end
)
 in (LrTable.NT 35,(result,tynode1left,tynode1right),rest671) end
| (250,(_,(MlyValue.tuple_ty tuple_ty1,tuple_ty1left,tuple_ty1right))
::rest671) => let val result=MlyValue.tynode(fn _ => let val tuple_ty
 as tuple_ty1=tuple_ty1 ()
 in (Type.tuple tuple_ty) end
)
 in (LrTable.NT 36,(result,tuple_ty1left,tuple_ty1right),rest671) end
| (251,(_,(MlyValue.ty ty2,_,ty2right))::_::(_,(MlyValue.ty ty1,
ty1left,_))::rest671) => let val result=MlyValue.tynode(fn _ => let 
val ty1=ty1 ()
val ty2=ty2 ()
 in (Type.arrow(ty1, ty2)) end
)
 in (LrTable.NT 36,(result,ty1left,ty2right),rest671) end
| (252,(_,(MlyValue.ty'node ty'node1,ty'node1left,ty'node1right))::
rest671) => let val result=MlyValue.tynode(fn _ => let val ty'node as 
ty'node1=ty'node1 ()
 in (ty'node) end
)
 in (LrTable.NT 36,(result,ty'node1left,ty'node1right),rest671) end
| (253,(_,(MlyValue.ty'node ty'node1,ty'nodeleft as ty'node1left,
ty'noderight as ty'node1right))::rest671) => let val result=
MlyValue.ty'(fn _ => let val ty'node as ty'node1=ty'node1 ()
 in (Type.makeRegion'(ty'node, ty'nodeleft, ty'noderight)) end
)
 in (LrTable.NT 32,(result,ty'node1left,ty'node1right),rest671) end
| (254,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.ty'node(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in (Type.Var tyvar) end
)
 in (LrTable.NT 33,(result,tyvar1left,tyvar1right),rest671) end
| (255,(_,(_,_,RBRACE1right))::(_,(MlyValue.tlabels tlabels1,_,_))::(_
,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.ty'node(fn _
 => let val tlabels as tlabels1=tlabels1 ()
 in (Type.Record(Srecord.fromList tlabels)) end
)
 in (LrTable.NT 33,(result,LBRACE1left,RBRACE1right),rest671) end
| (256,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.ty'node(fn _ => (Type.unit))
 in (LrTable.NT 33,(result,LBRACE1left,RBRACE1right),rest671) end
| (257,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::_::(_,(
MlyValue.ty0_pc ty0_pc1,_,_))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.ty'node(fn _ => let val ty0_pc as ty0_pc1=ty0_pc1 
()
val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, ty0_pc)) end
)
 in (LrTable.NT 33,(result,LPAREN1left,longtycon1right),rest671) end
| (258,(_,(_,_,RPAREN1right))::(_,(MlyValue.ty ty1,_,_))::(_,(_,
LPAREN1left,_))::rest671) => let val result=MlyValue.ty'node(fn _ => 
let val ty as ty1=ty1 ()
 in (Type.node ty) end
)
 in (LrTable.NT 33,(result,LPAREN1left,RPAREN1right),rest671) end
| (259,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::(_,(
MlyValue.ty' ty'1,ty'1left,_))::rest671) => let val result=
MlyValue.ty'node(fn _ => let val ty' as ty'1=ty'1 ()
val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, [ty'])) end
)
 in (LrTable.NT 33,(result,ty'1left,longtycon1right),rest671) end
| (260,(_,(MlyValue.longtycon longtycon1,longtycon1left,
longtycon1right))::rest671) => let val result=MlyValue.ty'node(fn _
 => let val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, [])) end
)
 in (LrTable.NT 33,(result,longtycon1left,longtycon1right),rest671)
 end
| (261,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.field field1,
field1left,_))::rest671) => let val result=MlyValue.tlabel(fn _ => 
let val field as field1=field1 ()
val ty as ty1=ty1 ()
 in (field, ty) end
)
 in (LrTable.NT 30,(result,field1left,ty1right),rest671) end
| (262,(_,(MlyValue.tlabels tlabels1,_,tlabels1right))::_::(_,(
MlyValue.tlabel tlabel1,tlabel1left,_))::rest671) => let val result=
MlyValue.tlabels(fn _ => let val tlabel as tlabel1=tlabel1 ()
val tlabels as tlabels1=tlabels1 ()
 in (tlabel :: tlabels) end
)
 in (LrTable.NT 31,(result,tlabel1left,tlabels1right),rest671) end
| (263,(_,(MlyValue.tlabel tlabel1,tlabel1left,tlabel1right))::rest671
) => let val result=MlyValue.tlabels(fn _ => let val tlabel as tlabel1
=tlabel1 ()
 in ([tlabel]) end
)
 in (LrTable.NT 31,(result,tlabel1left,tlabel1right),rest671) end
| (264,(_,(MlyValue.tuple_ty tuple_ty1,_,tuple_ty1right))::_::(_,(
MlyValue.ty' ty'1,ty'1left,_))::rest671) => let val result=
MlyValue.tuple_ty(fn _ => let val ty' as ty'1=ty'1 ()
val tuple_ty as tuple_ty1=tuple_ty1 ()
 in (ty' :: tuple_ty) end
)
 in (LrTable.NT 34,(result,ty'1left,tuple_ty1right),rest671) end
| (265,(_,(MlyValue.ty' ty'2,_,ty'2right))::_::(_,(MlyValue.ty' ty'1,
ty'1left,_))::rest671) => let val result=MlyValue.tuple_ty(fn _ => 
let val ty'1=ty'1 ()
val ty'2=ty'2 ()
 in ([ty'1, ty'2]) end
)
 in (LrTable.NT 34,(result,ty'1left,ty'2right),rest671) end
| (266,(_,(MlyValue.ty ty2,_,ty2right))::_::(_,(MlyValue.ty ty1,
ty1left,_))::rest671) => let val result=MlyValue.ty0_pc(fn _ => let 
val ty1=ty1 ()
val ty2=ty2 ()
 in ([ty1, ty2]) end
)
 in (LrTable.NT 39,(result,ty1left,ty2right),rest671) end
| (267,(_,(MlyValue.ty0_pc ty0_pc1,_,ty0_pc1right))::_::(_,(
MlyValue.ty ty1,ty1left,_))::rest671) => let val result=
MlyValue.ty0_pc(fn _ => let val ty as ty1=ty1 ()
val ty0_pc as ty0_pc1=ty0_pc1 ()
 in (ty :: ty0_pc) end
)
 in (LrTable.NT 39,(result,ty1left,ty0_pc1right),rest671) end
| (268,(_,(MlyValue.INT INT1,INT1left,INT1right))::rest671) => let 
val result=MlyValue.const(fn _ => let val INT as INT1=INT1 ()
 in (Const.Int INT) end
)
 in (LrTable.NT 8,(result,INT1left,INT1right),rest671) end
| (269,(_,(MlyValue.WORD WORD1,WORD1left,WORD1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val WORD as WORD1=WORD1 ()
 in (Const.Word WORD) end
)
 in (LrTable.NT 8,(result,WORD1left,WORD1right),rest671) end
| (270,(_,(MlyValue.REAL REAL1,REAL1left,REAL1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val REAL as REAL1=REAL1 ()
 in (Const.Real REAL) end
)
 in (LrTable.NT 8,(result,REAL1left,REAL1right),rest671) end
| (271,(_,(MlyValue.STRING STRING1,STRING1left,STRING1right))::rest671
) => let val result=MlyValue.const(fn _ => let val STRING as STRING1=
STRING1 ()
 in (Const.String STRING) end
)
 in (LrTable.NT 8,(result,STRING1left,STRING1right),rest671) end
| (272,(_,(MlyValue.CHAR CHAR1,CHAR1left,CHAR1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val CHAR as CHAR1=CHAR1 ()
 in (Const.Char CHAR) end
)
 in (LrTable.NT 8,(result,CHAR1left,CHAR1right),rest671) end
| (273,(_,(MlyValue.LONGID LONGID1,LONGIDleft as LONGID1left,
LONGIDright as LONGID1right))::rest671) => let val result=
MlyValue.idNoAsterisk(fn _ => let val LONGID as LONGID1=LONGID1 ()
 in (ensureNonqualified(LONGID, reg(LONGIDleft, LONGIDright))) end
)
 in (LrTable.NT 4,(result,LONGID1left,LONGID1right),rest671) end
| (274,(_,(MlyValue.idNoAsterisk idNoAsterisk1,idNoAsterisk1left,
idNoAsterisk1right))::rest671) => let val result=MlyValue.id(fn _ => 
let val idNoAsterisk as idNoAsterisk1=idNoAsterisk1 ()
 in (idNoAsterisk) end
)
 in (LrTable.NT 2,(result,idNoAsterisk1left,idNoAsterisk1right),
rest671) end
| (275,(_,(_,ASTERISK1left,ASTERISK1right))::rest671) => let val 
result=MlyValue.id(fn _ => ("*"))
 in (LrTable.NT 2,(result,ASTERISK1left,ASTERISK1right),rest671) end
| (276,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.idEqual(fn _ => let val id as id1=id1 ()
 in (id) end
)
 in (LrTable.NT 3,(result,id1left,id1right),rest671) end
| (277,(_,(_,EQUALOP1left,EQUALOP1right))::rest671) => let val result=
MlyValue.idEqual(fn _ => ("="))
 in (LrTable.NT 3,(result,EQUALOP1left,EQUALOP1right),rest671) end
| (278,(_,(MlyValue.LONGID LONGID1,LONGID1left,LONGID1right))::rest671
) => let val result=MlyValue.longid(fn _ => let val LONGID as LONGID1=
LONGID1 ()
 in (LONGID) end
)
 in (LrTable.NT 6,(result,LONGID1left,LONGID1right),rest671) end
| (279,(_,(_,ASTERISK1left,ASTERISK1right))::rest671) => let val 
result=MlyValue.longid(fn _ => ("*"))
 in (LrTable.NT 6,(result,ASTERISK1left,ASTERISK1right),rest671) end
| (280,(_,(MlyValue.LONGID LONGID1,LONGID1left,LONGID1right))::rest671
) => let val result=MlyValue.longidNoAsterisk(fn _ => let val LONGID
 as LONGID1=LONGID1 ()
 in (LONGID) end
)
 in (LrTable.NT 5,(result,LONGID1left,LONGID1right),rest671) end
| (281,(_,(MlyValue.longid longid1,longid1left,longid1right))::rest671
) => let val result=MlyValue.longidEqual(fn _ => let val longid as 
longid1=longid1 ()
 in (longid) end
)
 in (LrTable.NT 7,(result,longid1left,longid1right),rest671) end
| (282,(_,(_,EQUALOP1left,EQUALOP1right))::rest671) => let val result=
MlyValue.longidEqual(fn _ => ("="))
 in (LrTable.NT 7,(result,EQUALOP1left,EQUALOP1right),rest671) end
| (283,(_,(MlyValue.idEqual idEqual1,idEqual1left,idEqual1right))::
rest671) => let val result=MlyValue.vid(fn _ => let val idEqual as 
idEqual1=idEqual1 ()
 in (Vid.fromString idEqual) end
)
 in (LrTable.NT 9,(result,idEqual1left,idEqual1right),rest671) end
| (284,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.vidNoEqual(fn _ => let val id as id1=id1 ()
 in (Vid.fromString id) end
)
 in (LrTable.NT 10,(result,id1left,id1right),rest671) end
| (285,(_,(MlyValue.vid vid1,vid1left,vid1right))::rest671) => let 
val result=MlyValue.vids(fn _ => let val vid as vid1=vid1 ()
 in ([vid]) end
)
 in (LrTable.NT 11,(result,vid1left,vid1right),rest671) end
| (286,(_,(MlyValue.vids vids1,_,vids1right))::(_,(MlyValue.vid vid1,
vid1left,_))::rest671) => let val result=MlyValue.vids(fn _ => let 
val vid as vid1=vid1 ()
val vids as vids1=vids1 ()
 in (vid::vids) end
)
 in (LrTable.NT 11,(result,vid1left,vids1right),rest671) end
| (287,(_,(MlyValue.idEqual idEqual1,idEqual1left,idEqual1right))::
rest671) => let val result=MlyValue.var(fn _ => let val idEqual as 
idEqual1=idEqual1 ()
 in (Var.fromString idEqual) end
)
 in (LrTable.NT 12,(result,idEqual1left,idEqual1right),rest671) end
| (288,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.con(fn _ => let val id as id1=id1 ()
 in (Con.fromString id) end
)
 in (LrTable.NT 14,(result,id1left,id1right),rest671) end
| (289,(_,(MlyValue.idNoAsterisk idNoAsterisk1,idNoAsterisk1left,
idNoAsterisk1right))::rest671) => let val result=MlyValue.tycon(fn _
 => let val idNoAsterisk as idNoAsterisk1=idNoAsterisk1 ()
 in (Tycon.fromString idNoAsterisk) end
)
 in (LrTable.NT 16,(result,idNoAsterisk1left,idNoAsterisk1right),
rest671) end
| (290,(_,(MlyValue.TYVAR TYVAR1,TYVARleft as TYVAR1left,TYVARright
 as TYVAR1right))::rest671) => let val result=MlyValue.tyvar(fn _ => 
let val TYVAR as TYVAR1=TYVAR1 ()
 in (Tyvar.newString(TYVAR, TYVARleft, TYVARright)) end
)
 in (LrTable.NT 15,(result,TYVAR1left,TYVAR1right),rest671) end
| (291,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.field(fn _ => let val id as id1=id1 ()
 in (Field.String id) end
)
 in (LrTable.NT 17,(result,id1left,id1right),rest671) end
| (292,(_,(MlyValue.int int1,intleft as int1left,intright as int1right
))::rest671) => let val result=MlyValue.field(fn _ => let val int as 
int1=int1 ()
 in (
Field.Int
				(if int <= 0
				    then (error(reg(intleft, intright),
						"expected integer") ;
					  ~1)
				 else int)
) end
)
 in (LrTable.NT 17,(result,int1left,int1right),rest671) end
| (293,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.strid(fn _ => let val id as id1=id1 ()
 in (Strid.fromString id) end
)
 in (LrTable.NT 18,(result,id1left,id1right),rest671) end
| (294,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.sigid(fn _ => let val id as id1=id1 ()
 in (Sigid.fromString id) end
)
 in (LrTable.NT 19,(result,id1left,id1right),rest671) end
| (295,(_,(MlyValue.sigid sigid1,sigid1left,sigid1right))::rest671)
 => let val result=MlyValue.sigids(fn _ => let val sigid as sigid1=
sigid1 ()
 in ([sigid]) end
)
 in (LrTable.NT 20,(result,sigid1left,sigid1right),rest671) end
| (296,(_,(MlyValue.sigids sigids1,_,sigids1right))::(_,(
MlyValue.sigid sigid1,sigid1left,_))::rest671) => let val result=
MlyValue.sigids(fn _ => let val sigid as sigid1=sigid1 ()
val sigids as sigids1=sigids1 ()
 in (sigid::sigids) end
)
 in (LrTable.NT 20,(result,sigid1left,sigids1right),rest671) end
| (297,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.fctid(fn _ => let val id as id1=id1 ()
 in (Fctid.fromString id) end
)
 in (LrTable.NT 21,(result,id1left,id1right),rest671) end
| (298,(_,(MlyValue.longidNoAsterisk longidNoAsterisk1,
longidNoAsteriskleft as longidNoAsterisk1left,longidNoAsteriskright
 as longidNoAsterisk1right))::rest671) => let val result=
MlyValue.longtycon(fn _ => let val longidNoAsterisk as 
longidNoAsterisk1=longidNoAsterisk1 ()
 in (
Longtycon.fromString(longidNoAsterisk,
						   reg(longidNoAsteriskleft,
						       longidNoAsteriskright))
) end
)
 in (LrTable.NT 22,(result,longidNoAsterisk1left,
longidNoAsterisk1right),rest671) end
| (299,(_,(MlyValue.longidEqual longidEqual1,longidEqualleft as 
longidEqual1left,longidEqualright as longidEqual1right))::rest671) => 
let val result=MlyValue.longvar(fn _ => let val longidEqual as 
longidEqual1=longidEqual1 ()
 in (
Longvar.fromString(longidEqual,
					       reg(longidEqualleft,
						   longidEqualright))
) end
)
 in (LrTable.NT 23,(result,longidEqual1left,longidEqual1right),rest671
) end
| (300,(_,(MlyValue.longidEqual longidEqual1,longidEqualleft as 
longidEqual1left,longidEqualright as longidEqual1right))::rest671) => 
let val result=MlyValue.longvid(fn _ => let val longidEqual as 
longidEqual1=longidEqual1 ()
 in (
Longvid.fromString(longidEqual,
					       reg(longidEqualleft,
						   longidEqualright))
) end
)
 in (LrTable.NT 26,(result,longidEqual1left,longidEqual1right),rest671
) end
| (301,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longvidNoEqual(fn _ => let val longid as longid1=longid1 ()
 in (
Longvid.fromString(longid,
					       reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 27,(result,longid1left,longid1right),rest671) end
| (302,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longcon(fn _ => let val longid as longid1=longid1 ()
 in (
Longcon.fromString(longid,
					       reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 25,(result,longid1left,longid1right),rest671) end
| (303,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longstrid(fn _ => let val longid as longid1=longid1 ()
 in (Longstrid.fromString(longid,
						 reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 28,(result,longid1left,longid1right),rest671) end
| (304,(_,(MlyValue.longstrid longstrid1,longstrid1left,
longstrid1right))::rest671) => let val result=MlyValue.longstrids(fn _
 => let val longstrid as longstrid1=longstrid1 ()
 in ([longstrid]) end
)
 in (LrTable.NT 29,(result,longstrid1left,longstrid1right),rest671)
 end
| (305,(_,(MlyValue.longstrids longstrids1,_,longstrids1right))::(_,(
MlyValue.longstrid longstrid1,longstrid1left,_))::rest671) => let val 
result=MlyValue.longstrids(fn _ => let val longstrid as longstrid1=
longstrid1 ()
val longstrids as longstrids1=longstrids1 ()
 in (longstrid :: longstrids) end
)
 in (LrTable.NT 29,(result,longstrid1left,longstrids1right),rest671)
 end
| _ => raise (mlyAction i392)
end
val void = MlyValue.VOID
val extract = fn a => (fn MlyValue.program x => x
| _ => let exception ParseInternal
	in raise ParseInternal end) a ()
end
end
structure Tokens : ML_TOKENS =
struct
type svalue = ParserData.svalue
type ('a,'b) token = ('a,'b) Token.token
fun EOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 0,(
ParserData.MlyValue.VOID,p1,p2))
fun SEMICOLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 1,(
ParserData.MlyValue.VOID,p1,p2))
fun LONGID (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 2,(
ParserData.MlyValue.LONGID (fn () => i),p1,p2))
fun TYVAR (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 3,(
ParserData.MlyValue.TYVAR (fn () => i),p1,p2))
fun INT (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 4,(
ParserData.MlyValue.INT (fn () => i),p1,p2))
fun WORD (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 5,(
ParserData.MlyValue.WORD (fn () => i),p1,p2))
fun REAL (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 6,(
ParserData.MlyValue.REAL (fn () => i),p1,p2))
fun STRING (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 7,(
ParserData.MlyValue.STRING (fn () => i),p1,p2))
fun CHAR (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 8,(
ParserData.MlyValue.CHAR (fn () => i),p1,p2))
fun ABSTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 9,(
ParserData.MlyValue.VOID,p1,p2))
fun AND (p1,p2) = Token.TOKEN (ParserData.LrTable.T 10,(
ParserData.MlyValue.VOID,p1,p2))
fun ARROW (p1,p2) = Token.TOKEN (ParserData.LrTable.T 11,(
ParserData.MlyValue.VOID,p1,p2))
fun AS (p1,p2) = Token.TOKEN (ParserData.LrTable.T 12,(
ParserData.MlyValue.VOID,p1,p2))
fun BAR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 13,(
ParserData.MlyValue.VOID,p1,p2))
fun CASE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 14,(
ParserData.MlyValue.VOID,p1,p2))
fun DATATYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 15,(
ParserData.MlyValue.VOID,p1,p2))
fun DOTDOTDOT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 16,(
ParserData.MlyValue.VOID,p1,p2))
fun ELSE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 17,(
ParserData.MlyValue.VOID,p1,p2))
fun END (p1,p2) = Token.TOKEN (ParserData.LrTable.T 18,(
ParserData.MlyValue.VOID,p1,p2))
fun EQUALOP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 19,(
ParserData.MlyValue.VOID,p1,p2))
fun EQTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 20,(
ParserData.MlyValue.VOID,p1,p2))
fun EXCEPTION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 21,(
ParserData.MlyValue.VOID,p1,p2))
fun DO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 22,(
ParserData.MlyValue.VOID,p1,p2))
fun DARROW (p1,p2) = Token.TOKEN (ParserData.LrTable.T 23,(
ParserData.MlyValue.VOID,p1,p2))
fun FN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 24,(
ParserData.MlyValue.VOID,p1,p2))
fun FUN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 25,(
ParserData.MlyValue.VOID,p1,p2))
fun FUNCTOR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 26,(
ParserData.MlyValue.VOID,p1,p2))
fun HANDLE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 27,(
ParserData.MlyValue.VOID,p1,p2))
fun HASH (p1,p2) = Token.TOKEN (ParserData.LrTable.T 28,(
ParserData.MlyValue.VOID,p1,p2))
fun IF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 29,(
ParserData.MlyValue.VOID,p1,p2))
fun IN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 30,(
ParserData.MlyValue.VOID,p1,p2))
fun INCLUDE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 31,(
ParserData.MlyValue.VOID,p1,p2))
fun INFIX (p1,p2) = Token.TOKEN (ParserData.LrTable.T 32,(
ParserData.MlyValue.VOID,p1,p2))
fun INFIXR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 33,(
ParserData.MlyValue.VOID,p1,p2))
fun LET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 34,(
ParserData.MlyValue.VOID,p1,p2))
fun LOCAL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 35,(
ParserData.MlyValue.VOID,p1,p2))
fun NONFIX (p1,p2) = Token.TOKEN (ParserData.LrTable.T 36,(
ParserData.MlyValue.VOID,p1,p2))
fun OF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 37,(
ParserData.MlyValue.VOID,p1,p2))
fun OP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 38,(
ParserData.MlyValue.VOID,p1,p2))
fun OPEN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 39,(
ParserData.MlyValue.VOID,p1,p2))
fun OVERLOAD (p1,p2) = Token.TOKEN (ParserData.LrTable.T 40,(
ParserData.MlyValue.VOID,p1,p2))
fun RAISE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 41,(
ParserData.MlyValue.VOID,p1,p2))
fun REC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 42,(
ParserData.MlyValue.VOID,p1,p2))
fun SHARING (p1,p2) = Token.TOKEN (ParserData.LrTable.T 43,(
ParserData.MlyValue.VOID,p1,p2))
fun SIG (p1,p2) = Token.TOKEN (ParserData.LrTable.T 44,(
ParserData.MlyValue.VOID,p1,p2))
fun SIGNATURE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 45,(
ParserData.MlyValue.VOID,p1,p2))
fun STRUCT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 46,(
ParserData.MlyValue.VOID,p1,p2))
fun STRUCTURE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 47,(
ParserData.MlyValue.VOID,p1,p2))
fun THEN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 48,(
ParserData.MlyValue.VOID,p1,p2))
fun TYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 49,(
ParserData.MlyValue.VOID,p1,p2))
fun VAL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 50,(
ParserData.MlyValue.VOID,p1,p2))
fun WHERE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 51,(
ParserData.MlyValue.VOID,p1,p2))
fun WHILE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 52,(
ParserData.MlyValue.VOID,p1,p2))
fun WILD (p1,p2) = Token.TOKEN (ParserData.LrTable.T 53,(
ParserData.MlyValue.VOID,p1,p2))
fun WITH (p1,p2) = Token.TOKEN (ParserData.LrTable.T 54,(
ParserData.MlyValue.VOID,p1,p2))
fun WITHTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 55,(
ParserData.MlyValue.VOID,p1,p2))
fun ASTERISK (p1,p2) = Token.TOKEN (ParserData.LrTable.T 56,(
ParserData.MlyValue.VOID,p1,p2))
fun COLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 57,(
ParserData.MlyValue.VOID,p1,p2))
fun COLONGT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 58,(
ParserData.MlyValue.VOID,p1,p2))
fun COMMA (p1,p2) = Token.TOKEN (ParserData.LrTable.T 59,(
ParserData.MlyValue.VOID,p1,p2))
fun LBRACE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 60,(
ParserData.MlyValue.VOID,p1,p2))
fun LBRACKET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 61,(
ParserData.MlyValue.VOID,p1,p2))
fun LPAREN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 62,(
ParserData.MlyValue.VOID,p1,p2))
fun RBRACE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 63,(
ParserData.MlyValue.VOID,p1,p2))
fun RBRACKET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 64,(
ParserData.MlyValue.VOID,p1,p2))
fun RPAREN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 65,(
ParserData.MlyValue.VOID,p1,p2))
fun ORELSE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 66,(
ParserData.MlyValue.VOID,p1,p2))
fun ANDALSO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 67,(
ParserData.MlyValue.VOID,p1,p2))
fun FUNSIG (p1,p2) = Token.TOKEN (ParserData.LrTable.T 68,(
ParserData.MlyValue.VOID,p1,p2))
fun PRIM (p1,p2) = Token.TOKEN (ParserData.LrTable.T 69,(
ParserData.MlyValue.VOID,p1,p2))
end
end
functor MLLexFun(structure Tokens : ML_TOKENS)=
   struct
    structure UserDeclarations =
      struct
(* Heavily modified from the SML/NJ sources by sweeks@research.nj.nec.com. *)

(* ml.lex
 *
 * Copyright 1989 by AT&T Bell Laboratories
 *
 * $Log: ml.lex,v $
 * Revision 1.3  1997/05/22  20:17:22  jhr
 * Changed lexer to accept "1e1" style floating-point literals.
 *
 * Revision 1.2  1997/01/28  23:20:40  jhr
 * Integer and word literals are now represented by IntInf.int (instead of
 * as strings).
 *
 *)

val newline = Region.newline

fun error(left,right,msg) = 
   Control.error(Region.T{left = left, right = right},
		 Layout.str msg)

local 
   fun make (scan,token,default,msg) (radix,str,left,right) =
      token((case StringCvt.scanString(scan radix) str of
		NONE => (error(left, right, "invalid " ^ msg ^ " constant") ;
			 default)
	      | SOME n => n)
	    handle Overflow =>
	       (error(left, right, msg ^ " constant out of range") ;
		default),
	    left, right)
in 
   val makeWord = make(Word32.scan, Tokens.WORD, 0w0, "word")
end

type svalue = Tokens.svalue
type pos = int
type lexresult = (svalue,pos) Tokens.token
type lexarg = {comLevel : int ref,
	       stringtype : bool ref,
	       stringstart : int ref}
type arg = lexarg
type ('a,'b) token = ('a,'b) Tokens.token

val eof = fn ({comLevel,stringstart,stringtype}) => 
   (if !comLevel>0 then Control.errorStr(Region.bogus, "unclosed comment")
    else ();
    Tokens.EOF(~1,~1))

val charlist : string list ref = ref []
fun addString(s:string) = charlist := s :: (!charlist)
fun addChar(c:char) = addString(String.fromChar c)

fun has_quote s = let
      fun loop i = ((String.sub(s,i) = #"`") orelse loop (i + 1))
	    handle _ => false
      in
	loop 0
      end

fun inc (ri as ref i) = (ri := i + 1)
fun dec (ri as ref i) = (ri := i-1)
end (* end of user routines *)
exception LexError (* raised if illegal leaf action tried *)
structure Internal =
	struct

datatype yyfinstate = N of int
type statedata = {fin : yyfinstate list, trans: int Vector.vector}
(* transition & final state table *)
val tab = let
fun decode s k =
  let val k' = k + k
      val hi = Char.ord(String.sub(s, k'))
      val lo = Char.ord(String.sub(s, k' + 1))
  in hi * 256 + lo end
val s = [ 
 (0,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (1,129, 
"\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\231\000\233\000\009\000\231\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\231\000\185\000\230\000\228\000\185\000\185\000\185\000\226\
\\000\224\000\223\000\221\000\185\000\220\000\218\000\215\000\185\
\\000\208\000\206\000\206\000\206\000\206\000\206\000\206\000\206\
\\000\206\000\206\000\204\000\203\000\185\000\201\000\185\000\185\
\\000\185\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\200\000\185\000\199\000\185\000\186\
\\000\185\000\171\000\030\000\167\000\158\000\139\000\128\000\030\
\\000\122\000\110\000\030\000\030\000\103\000\030\000\097\000\087\
\\000\030\000\030\000\080\000\057\000\050\000\030\000\047\000\032\
\\000\030\000\030\000\030\000\029\000\028\000\027\000\010\000\009\
\\000\009"
),
 (3,129, 
"\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\239\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\237\000\234\000\235\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\
\\000\234"
),
 (5,129, 
"\000\240\000\240\000\240\000\240\000\240\000\240\000\240\000\240\
\\000\240\000\240\001\012\000\240\000\240\001\011\000\240\000\240\
\\000\240\000\240\000\240\000\240\000\240\000\240\000\240\000\240\
\\000\240\000\240\000\240\000\240\000\240\000\240\000\240\000\240\
\\000\241\000\241\001\010\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\242\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\
\\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\240\
\\000\240"
),
 (7,129, 
"\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\015\001\018\001\013\001\015\001\017\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\015\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\014\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013\001\013\001\013\001\013\001\013\001\013\001\013\001\013\
\\001\013"
),
 (10,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\024\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\011\000\000\000\011\000\011\000\011\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (11,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\011\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (12,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\013\000\013\000\000\000\013\000\000\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\013\000\000\000\013\000\013\000\013\000\013\
\\000\013\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\000\000\013\000\000\000\013\000\000\
\\000\013\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\000\000\013\000\000\000\013\000\000\
\\000\000"
),
 (13,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\013\000\013\000\000\000\013\000\012\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\013\000\000\000\013\000\013\000\013\000\013\
\\000\013\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\013\000\000\000\013\000\000\
\\000\013\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\013\000\000\000\013\000\000\
\\000\000"
),
 (14,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\000\000\000\000\000\000\000\000\014\
\\000\000\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\014\000\014\000\014\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (15,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (16,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\017\000\000\
\\000\000"
),
 (17,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (19,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (20,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\021\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\021\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (21,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\
\\000\023\000\023\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\022\000\000\
\\000\000"
),
 (22,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\023\000\023\000\023\000\023\000\023\000\023\000\023\000\023\
\\000\023\000\023\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (24,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\025\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (25,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\
\\000\026\000\026\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (30,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (32,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\040\000\033\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (33,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\034\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (34,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\035\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (35,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\036\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (36,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\037\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (37,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\038\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (38,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\039\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (40,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\044\000\031\000\031\
\\000\031\000\041\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (41,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\042\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (42,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\043\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (44,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\045\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (45,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\046\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (47,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\048\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (48,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\049\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (50,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\054\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\051\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (51,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\052\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (52,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\053\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (54,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\055\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (55,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\056\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (57,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\074\000\066\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\058\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (58,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\059\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (59,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\060\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (60,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\061\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (61,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\062\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (62,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\063\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (63,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\064\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (64,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\065\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (66,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\067\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (67,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\068\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (68,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\069\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (69,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\070\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (70,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\071\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (71,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\072\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (72,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\073\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (74,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\075\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (75,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\076\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (76,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\077\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (77,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\078\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (78,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\079\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (80,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\083\000\031\000\031\000\031\000\081\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (81,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\082\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (83,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\084\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (84,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\085\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (85,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\086\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (87,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\096\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\093\000\031\000\088\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (88,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\089\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (89,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\090\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (90,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\091\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (91,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\092\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (93,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\094\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (94,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\095\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (97,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\098\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (98,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\099\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (99,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\100\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (100,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\101\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (101,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\102\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (103,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\108\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\104\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (104,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\105\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (105,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\106\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (106,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\107\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (108,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\109\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (110,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\121\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\111\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (111,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\116\000\031\000\031\000\112\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (112,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\113\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (113,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\114\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (114,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\115\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (116,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\117\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (117,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\118\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (118,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\119\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (119,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\120\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (122,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\123\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (123,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\124\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (124,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\125\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (125,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\126\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (126,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\127\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (128,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\138\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\129\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (129,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\130\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (130,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\134\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\131\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (131,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\132\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (132,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\133\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (134,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\135\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (135,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\136\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (136,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\137\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (139,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\155\000\031\000\153\000\031\
\\000\031\000\148\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\140\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (140,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\141\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (141,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\142\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (142,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\143\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (143,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\144\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (144,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\145\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (145,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\146\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (146,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\147\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (148,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\149\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (149,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\150\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (150,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\151\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (151,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\152\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (153,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\154\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (155,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\156\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (156,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\157\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (158,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\160\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\159\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (160,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\161\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (161,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\162\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (162,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\163\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (163,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\164\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (164,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\165\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (165,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\166\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (167,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\168\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (168,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\169\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (169,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\170\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (171,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\179\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\173\000\031\
\\000\031\000\031\000\031\000\172\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (173,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\174\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (174,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\175\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (175,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\176\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (176,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\177\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (177,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\178\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (179,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\180\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (180,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\181\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (181,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\182\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (182,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\183\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (183,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\
\\000\000\000\031\000\031\000\031\000\031\000\184\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\
\\000\031\000\031\000\031\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (186,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\191\
\\000\187\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (187,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\188\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (188,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\189\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (189,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\190\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (191,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\192\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (192,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\193\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (193,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\194\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (194,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\195\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (195,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\196\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (196,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\197\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (197,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\198\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (201,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\202\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (204,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\205\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (206,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\
\\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\
\\000\207\000\207\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (208,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\
\\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\
\\000\207\000\207\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\
\\000\209\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (209,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\210\000\210\000\210\000\210\000\210\000\210\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\210\000\210\000\210\000\210\000\210\000\210\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (211,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\
\\000\214\000\214\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\212\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (212,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\213\000\213\000\213\000\213\000\213\000\213\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\213\000\213\000\213\000\213\000\213\000\213\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (214,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\
\\000\214\000\214\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (215,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\216\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (216,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\217\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (218,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\219\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (221,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\000\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\222\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\011\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (224,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\225\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (226,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\227\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\000\000\000\000\000\000\000\000\227\
\\000\000\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (228,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\011\000\229\000\011\000\011\000\011\000\011\000\000\
\\000\000\000\000\000\011\000\011\000\000\000\011\000\012\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\011\000\000\000\011\000\011\000\011\000\011\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\011\000\000\000\011\000\000\
\\000\000"
),
 (231,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\232\000\000\000\000\000\232\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\232\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (235,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\236\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (237,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\238\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (242,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\001\007\001\009\000\000\001\007\001\008\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\007\000\000\001\006\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\003\001\003\001\003\001\003\001\003\001\003\001\003\001\003\
\\001\003\001\003\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\001\002\000\000\000\255\000\000\
\\000\000\000\254\000\253\000\000\000\000\000\000\000\252\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\251\000\000\
\\000\000\000\000\000\250\000\000\000\249\000\244\000\243\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (244,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\245\000\245\000\245\000\245\000\245\000\245\000\245\000\245\
\\000\245\000\245\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\245\000\245\000\245\000\245\000\245\000\245\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\245\000\245\000\245\000\245\000\245\000\245\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (245,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\246\000\246\000\246\000\246\000\246\000\246\000\246\000\246\
\\000\246\000\246\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\246\000\246\000\246\000\246\000\246\000\246\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\246\000\246\000\246\000\246\000\246\000\246\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (246,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\247\000\247\000\247\000\247\000\247\000\247\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\247\000\247\000\247\000\247\000\247\000\247\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (247,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\248\000\248\000\248\000\248\000\248\000\248\000\248\000\248\
\\000\248\000\248\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\248\000\248\000\248\000\248\000\248\000\248\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\248\000\248\000\248\000\248\000\248\000\248\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (255,129, 
"\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\000\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
\\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
\\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
\\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\
\\001\000"
),
 (259,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\004\001\004\001\004\001\004\001\004\001\004\001\004\001\004\
\\001\004\001\004\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (260,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (263,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\001\007\000\000\000\000\001\007\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\007\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (264,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\009\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (267,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\012\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (271,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\001\016\000\000\000\000\001\016\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\001\016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
 (273,129, 
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\018\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000"
),
(0, 0, "")]
fun f(n, i, x) = (n, Vector.tabulate(i, decode x)) 
val s = map f (rev (tl (rev s))) 
exception LexHackingError 
fun look ((j,x)::r, i) = if i = j then x else look(r, i) 
  | look ([], i) = raise LexHackingError
fun g {fin=x, trans=i} = {fin=x, trans=look(s,i)} 
in Vector.fromList(map g 
[{fin = [], trans = 0},
{fin = [(N 2)], trans = 1},
{fin = [(N 2)], trans = 1},
{fin = [], trans = 3},
{fin = [], trans = 3},
{fin = [], trans = 5},
{fin = [], trans = 5},
{fin = [(N 473)], trans = 7},
{fin = [(N 473)], trans = 7},
{fin = [(N 384)], trans = 0},
{fin = [(N 322),(N 384)], trans = 10},
{fin = [(N 322)], trans = 11},
{fin = [], trans = 12},
{fin = [(N 322)], trans = 13},
{fin = [(N 322)], trans = 14},
{fin = [(N 349)], trans = 15},
{fin = [], trans = 16},
{fin = [], trans = 17},
{fin = [(N 342)], trans = 17},
{fin = [], trans = 19},
{fin = [(N 342)], trans = 20},
{fin = [], trans = 21},
{fin = [], trans = 22},
{fin = [(N 342)], trans = 22},
{fin = [(N 349)], trans = 24},
{fin = [], trans = 25},
{fin = [(N 360)], trans = 25},
{fin = [(N 28),(N 384)], trans = 0},
{fin = [(N 44),(N 322),(N 384)], trans = 11},
{fin = [(N 26),(N 384)], trans = 0},
{fin = [(N 322),(N 384)], trans = 30},
{fin = [(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 32},
{fin = [(N 322)], trans = 33},
{fin = [(N 322)], trans = 34},
{fin = [(N 281),(N 322)], trans = 35},
{fin = [(N 322)], trans = 36},
{fin = [(N 322)], trans = 37},
{fin = [(N 322)], trans = 38},
{fin = [(N 290),(N 322)], trans = 30},
{fin = [(N 322)], trans = 40},
{fin = [(N 322)], trans = 41},
{fin = [(N 322)], trans = 42},
{fin = [(N 276),(N 322)], trans = 30},
{fin = [(N 322)], trans = 44},
{fin = [(N 322)], trans = 45},
{fin = [(N 270),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 47},
{fin = [(N 322)], trans = 48},
{fin = [(N 264),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 50},
{fin = [(N 322)], trans = 51},
{fin = [(N 322)], trans = 52},
{fin = [(N 260),(N 322)], trans = 30},
{fin = [(N 322)], trans = 54},
{fin = [(N 322)], trans = 55},
{fin = [(N 255),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 57},
{fin = [(N 322)], trans = 58},
{fin = [(N 322)], trans = 59},
{fin = [(N 322)], trans = 60},
{fin = [(N 322)], trans = 61},
{fin = [(N 240),(N 322)], trans = 62},
{fin = [(N 322)], trans = 63},
{fin = [(N 322)], trans = 64},
{fin = [(N 250),(N 322)], trans = 30},
{fin = [(N 322)], trans = 66},
{fin = [(N 223),(N 322)], trans = 67},
{fin = [(N 322)], trans = 68},
{fin = [(N 322)], trans = 69},
{fin = [(N 322)], trans = 70},
{fin = [(N 322)], trans = 71},
{fin = [(N 322)], trans = 72},
{fin = [(N 233),(N 322)], trans = 30},
{fin = [(N 322)], trans = 74},
{fin = [(N 322)], trans = 75},
{fin = [(N 322)], trans = 76},
{fin = [(N 322)], trans = 77},
{fin = [(N 322)], trans = 78},
{fin = [(N 219),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 80},
{fin = [(N 322)], trans = 81},
{fin = [(N 211),(N 322)], trans = 30},
{fin = [(N 322)], trans = 83},
{fin = [(N 322)], trans = 84},
{fin = [(N 322)], trans = 85},
{fin = [(N 207),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 87},
{fin = [(N 322)], trans = 88},
{fin = [(N 322)], trans = 89},
{fin = [(N 322)], trans = 90},
{fin = [(N 322)], trans = 91},
{fin = [(N 297),(N 322)], trans = 30},
{fin = [(N 196),(N 322)], trans = 93},
{fin = [(N 322)], trans = 94},
{fin = [(N 201),(N 322)], trans = 30},
{fin = [(N 193),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 97},
{fin = [(N 322)], trans = 98},
{fin = [(N 322)], trans = 99},
{fin = [(N 322)], trans = 100},
{fin = [(N 322)], trans = 101},
{fin = [(N 190),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 103},
{fin = [(N 322)], trans = 104},
{fin = [(N 322)], trans = 105},
{fin = [(N 322)], trans = 106},
{fin = [(N 183),(N 322)], trans = 30},
{fin = [(N 322)], trans = 108},
{fin = [(N 177),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 110},
{fin = [(N 152),(N 322)], trans = 111},
{fin = [(N 322)], trans = 112},
{fin = [(N 322)], trans = 113},
{fin = [(N 166),(N 322)], trans = 114},
{fin = [(N 173),(N 322)], trans = 30},
{fin = [(N 322)], trans = 116},
{fin = [(N 322)], trans = 117},
{fin = [(N 322)], trans = 118},
{fin = [(N 322)], trans = 119},
{fin = [(N 160),(N 322)], trans = 30},
{fin = [(N 149),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 122},
{fin = [(N 322)], trans = 123},
{fin = [(N 322)], trans = 124},
{fin = [(N 322)], trans = 125},
{fin = [(N 322)], trans = 126},
{fin = [(N 146),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 128},
{fin = [(N 322)], trans = 129},
{fin = [(N 124),(N 322)], trans = 130},
{fin = [(N 322)], trans = 131},
{fin = [(N 322)], trans = 132},
{fin = [(N 139),(N 322)], trans = 30},
{fin = [(N 322)], trans = 134},
{fin = [(N 322)], trans = 135},
{fin = [(N 322)], trans = 136},
{fin = [(N 132),(N 322)], trans = 30},
{fin = [(N 120),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 139},
{fin = [(N 322)], trans = 140},
{fin = [(N 322)], trans = 141},
{fin = [(N 322)], trans = 142},
{fin = [(N 322)], trans = 143},
{fin = [(N 322)], trans = 144},
{fin = [(N 322)], trans = 145},
{fin = [(N 322)], trans = 146},
{fin = [(N 114),(N 322)], trans = 30},
{fin = [(N 322)], trans = 148},
{fin = [(N 322)], trans = 149},
{fin = [(N 322)], trans = 150},
{fin = [(N 322)], trans = 151},
{fin = [(N 104),(N 322)], trans = 30},
{fin = [(N 322)], trans = 153},
{fin = [(N 97),(N 322)], trans = 30},
{fin = [(N 322)], trans = 155},
{fin = [(N 322)], trans = 156},
{fin = [(N 93),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 158},
{fin = [(N 117),(N 322)], trans = 30},
{fin = [(N 322)], trans = 160},
{fin = [(N 322)], trans = 161},
{fin = [(N 322)], trans = 162},
{fin = [(N 322)], trans = 163},
{fin = [(N 322)], trans = 164},
{fin = [(N 322)], trans = 165},
{fin = [(N 88),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 167},
{fin = [(N 322)], trans = 168},
{fin = [(N 322)], trans = 169},
{fin = [(N 79),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 171},
{fin = [(N 74),(N 322)], trans = 30},
{fin = [(N 322)], trans = 173},
{fin = [(N 63),(N 322)], trans = 174},
{fin = [(N 322)], trans = 175},
{fin = [(N 322)], trans = 176},
{fin = [(N 322)], trans = 177},
{fin = [(N 305),(N 322)], trans = 30},
{fin = [(N 322)], trans = 179},
{fin = [(N 322)], trans = 180},
{fin = [(N 322)], trans = 181},
{fin = [(N 322)], trans = 182},
{fin = [(N 322)], trans = 183},
{fin = [(N 71),(N 322)], trans = 30},
{fin = [(N 322),(N 384)], trans = 11},
{fin = [(N 22),(N 384)], trans = 186},
{fin = [], trans = 187},
{fin = [], trans = 188},
{fin = [], trans = 189},
{fin = [(N 20)], trans = 0},
{fin = [], trans = 191},
{fin = [], trans = 192},
{fin = [], trans = 193},
{fin = [], trans = 194},
{fin = [], trans = 195},
{fin = [], trans = 196},
{fin = [], trans = 197},
{fin = [(N 14)], trans = 0},
{fin = [(N 32),(N 384)], trans = 0},
{fin = [(N 30),(N 384)], trans = 0},
{fin = [(N 51),(N 322),(N 384)], trans = 201},
{fin = [(N 59),(N 322)], trans = 11},
{fin = [(N 34),(N 384)], trans = 0},
{fin = [(N 46),(N 322),(N 384)], trans = 204},
{fin = [(N 49),(N 322)], trans = 11},
{fin = [(N 345),(N 384)], trans = 206},
{fin = [(N 345)], trans = 206},
{fin = [(N 345),(N 384)], trans = 208},
{fin = [], trans = 209},
{fin = [(N 354)], trans = 209},
{fin = [], trans = 211},
{fin = [], trans = 212},
{fin = [(N 371)], trans = 212},
{fin = [(N 365)], trans = 214},
{fin = [(N 384)], trans = 215},
{fin = [], trans = 216},
{fin = [(N 42)], trans = 0},
{fin = [(N 322),(N 384)], trans = 218},
{fin = [(N 56),(N 322)], trans = 11},
{fin = [(N 24),(N 384)], trans = 0},
{fin = [(N 322),(N 384)], trans = 221},
{fin = [(N 382)], trans = 0},
{fin = [(N 38),(N 384)], trans = 0},
{fin = [(N 36),(N 384)], trans = 224},
{fin = [(N 379)], trans = 0},
{fin = [(N 308),(N 384)], trans = 226},
{fin = [(N 308)], trans = 226},
{fin = [(N 53),(N 322),(N 384)], trans = 228},
{fin = [(N 376)], trans = 0},
{fin = [(N 373),(N 384)], trans = 0},
{fin = [(N 2),(N 384)], trans = 231},
{fin = [(N 2)], trans = 231},
{fin = [(N 4)], trans = 0},
{fin = [(N 394)], trans = 0},
{fin = [(N 394)], trans = 235},
{fin = [(N 392)], trans = 0},
{fin = [(N 394)], trans = 237},
{fin = [(N 387)], trans = 0},
{fin = [(N 389)], trans = 0},
{fin = [(N 465)], trans = 0},
{fin = [(N 463),(N 465)], trans = 0},
{fin = [(N 447),(N 455),(N 463),(N 465)], trans = 242},
{fin = [(N 417)], trans = 0},
{fin = [], trans = 244},
{fin = [], trans = 245},
{fin = [], trans = 246},
{fin = [], trans = 247},
{fin = [(N 437)], trans = 0},
{fin = [(N 414)], trans = 0},
{fin = [(N 411)], trans = 0},
{fin = [(N 408)], trans = 0},
{fin = [(N 405)], trans = 0},
{fin = [(N 402)], trans = 0},
{fin = [(N 399)], trans = 0},
{fin = [], trans = 255},
{fin = [(N 425)], trans = 0},
{fin = [(N 421),(N 425)], trans = 0},
{fin = [(N 443)], trans = 0},
{fin = [], trans = 259},
{fin = [], trans = 260},
{fin = [(N 430)], trans = 0},
{fin = [(N 440)], trans = 0},
{fin = [(N 447)], trans = 263},
{fin = [(N 453)], trans = 264},
{fin = [(N 453)], trans = 0},
{fin = [(N 396),(N 463),(N 465)], trans = 0},
{fin = [(N 460),(N 465)], trans = 267},
{fin = [(N 460)], trans = 0},
{fin = [(N 477)], trans = 0},
{fin = [(N 475),(N 477)], trans = 0},
{fin = [(N 473),(N 477)], trans = 271},
{fin = [(N 473)], trans = 271},
{fin = [(N 470),(N 477)], trans = 273},
{fin = [(N 470)], trans = 0}])
end
structure StartStates =
	struct
	datatype yystartstate = STARTSTATE of int

(* start state definitions *)

val A = STARTSTATE 3;
val F = STARTSTATE 7;
val INITIAL = STARTSTATE 1;
val S = STARTSTATE 5;

end
type result = UserDeclarations.lexresult
	exception LexerError (* raised if illegal leaf action tried *)
end

fun makeLexer yyinput =
let	val yygone0=1
	val yyb = ref "\n" 		(* buffer *)
	val yybl = ref 1		(*buffer length *)
	val yybufpos = ref 1		(* location of next character to use *)
	val yygone = ref yygone0	(* position in file of beginning of buffer *)
	val yydone = ref false		(* eof found yet? *)
	val yybegin = ref 1		(*Current 'start state' for lexer *)

	val YYBEGIN = fn (Internal.StartStates.STARTSTATE x) =>
		 yybegin := x

fun lex (yyarg as ({comLevel,stringstart,stringtype})) =
let fun continue() : Internal.result = 
  let fun scan (s,AcceptingLeaves : Internal.yyfinstate list list,l,i0) =
	let fun action (i,nil) = raise LexError
	| action (i,nil::l) = action (i-1,l)
	| action (i,(node::acts)::l) =
		case node of
		    Internal.N yyk => 
			(let fun yymktext() = substring(!yyb,i0,i-i0)
			     val yypos = i0+ !yygone
			fun REJECT() = action(i,acts::l)
			open UserDeclarations Internal.StartStates
 in (yybufpos := i; case yyk of 

			(* Application actions *)

  104 => (Tokens.EQTYPE(yypos, yypos + 6))
| 114 => (Tokens.EXCEPTION(yypos, yypos + 9))
| 117 => (Tokens.DO(yypos, yypos + 2))
| 120 => (Tokens.FN(yypos, yypos + 2))
| 124 => (Tokens.FUN(yypos, yypos + 3))
| 132 => (Tokens.FUNCTOR(yypos, yypos + 7))
| 139 => (Tokens.FUNSIG(yypos, yypos + 7))
| 14 => (Tokens.OVERLOAD(yypos, yypos + 9))
| 146 => (Tokens.HANDLE(yypos, yypos + 6))
| 149 => (Tokens.IF(yypos, yypos + 2))
| 152 => (Tokens.IN(yypos, yypos + 2))
| 160 => (Tokens.INCLUDE(yypos, yypos + 7))
| 166 => (Tokens.INFIX(yypos, yypos + 5))
| 173 => (Tokens.INFIXR(yypos, yypos + 6))
| 177 => (Tokens.LET(yypos, yypos + 3))
| 183 => (Tokens.LOCAL(yypos, yypos + 5))
| 190 => (Tokens.NONFIX(yypos, yypos + 6))
| 193 => (Tokens.OF(yypos, yypos + 2))
| 196 => (Tokens.OP(yypos, yypos + 2))
| 2 => (continue())
| 20 => (Tokens.PRIM(yypos, yypos + 5))
| 201 => (Tokens.OPEN(yypos, yypos + 4))
| 207 => (Tokens.RAISE(yypos, yypos + 5))
| 211 => (Tokens.REC(yypos, yypos + 3))
| 219 => (Tokens.SHARING(yypos, yypos + 7))
| 22 => (Tokens.WILD(yypos, yypos + 1))
| 223 => (Tokens.SIG(yypos, yypos + 3))
| 233 => (Tokens.SIGNATURE(yypos, yypos + 9))
| 24 => (Tokens.COMMA(yypos, yypos + 1))
| 240 => (Tokens.STRUCT(yypos, yypos + 6))
| 250 => (Tokens.STRUCTURE(yypos, yypos + 9))
| 255 => (Tokens.THEN(yypos, yypos + 4))
| 26 => (Tokens.LBRACE(yypos, yypos + 1))
| 260 => (Tokens.TYPE(yypos, yypos + 4))
| 264 => (Tokens.VAL(yypos, yypos + 3))
| 270 => (Tokens.WHERE(yypos, yypos + 5))
| 276 => (Tokens.WHILE(yypos, yypos + 5))
| 28 => (Tokens.RBRACE(yypos, yypos + 1))
| 281 => (Tokens.WITH(yypos, yypos + 4))
| 290 => (Tokens.WITHTYPE(yypos, yypos + 8))
| 297 => (Tokens.ORELSE(yypos, yypos + 6))
| 30 => (Tokens.LBRACKET(yypos, yypos + 1))
| 305 => (Tokens.ANDALSO(yypos, yypos + 7))
| 308 => let val yytext=yymktext() in Tokens.TYVAR(yytext, yypos,
					yypos + String.size yytext) end
| 32 => (Tokens.RBRACKET(yypos, yypos + 1))
| 322 => let val yytext=yymktext() in case yytext of
			 "*" => Tokens.ASTERISK(yypos, yypos + 1)
		       | _ => Tokens.LONGID(yytext, yypos, yypos + size yytext) end
| 34 => (Tokens.SEMICOLON(yypos, yypos + 1))
| 342 => let val yytext=yymktext() in Tokens.REAL(yytext, yypos, yypos+size yytext) end
| 345 => let val yytext=yymktext() in Tokens.INT(yytext, yypos, yypos + size yytext) end
| 349 => let val yytext=yymktext() in Tokens.INT(yytext, yypos, yypos + size yytext) end
| 354 => let val yytext=yymktext() in Tokens.INT(yytext, yypos, yypos + size yytext) end
| 36 => (Tokens.LPAREN(yypos, yypos + 1))
| 360 => let val yytext=yymktext() in Tokens.INT(yytext, yypos, yypos + size yytext) end
| 365 => let val yytext=yymktext() in makeWord(StringCvt.DEC, yytext, yypos,
				yypos + size yytext) end
| 371 => let val yytext=yymktext() in makeWord(StringCvt.HEX, yytext,
				    yypos, yypos + size yytext) end
| 373 => (charlist := [""]; stringstart := yypos;
                    stringtype := true; YYBEGIN S; continue())
| 376 => (charlist := [""]; stringstart := yypos;
                    stringtype := false; YYBEGIN S; continue())
| 379 => (YYBEGIN A; stringstart := yypos; comLevel := 1; continue())
| 38 => (Tokens.RPAREN(yypos, yypos + 1))
| 382 => (error(yypos, yypos + 2, "unmatched close comment") ;
		    continue())
| 384 => (error(yypos, yypos + 1, "illegal token") ;
		    continue())
| 387 => (inc comLevel; continue())
| 389 => (newline yypos ; continue())
| 392 => (dec comLevel; if !comLevel=0 then YYBEGIN INITIAL else (); continue())
| 394 => (continue())
| 396 => (let val s = (concat(rev(!charlist)) before charlist := nil)
                        fun make(tok, v) = tok(v, !stringstart, yypos + 1)
                    in YYBEGIN INITIAL;
		       if !stringtype
			  then make(Tokens.STRING, s)
		       else
			  make(Tokens.CHAR,
			       if size s <> 1 
				  then (error(yypos, yypos + 1,
					      "character constant not length 1") ;
					#"\000")
			       else String.sub(s, 0))
                    end)
| 399 => (addChar #"\a"; continue())
| 4 => (newline yypos; continue())
| 402 => (addChar #"\b"; continue())
| 405 => (addChar #"\f"; continue())
| 408 => (addChar #"\n"; continue())
| 411 => (addChar #"\r"; continue())
| 414 => (addChar #"\t"; continue())
| 417 => (addChar #"\v"; continue())
| 42 => (Tokens.DOTDOTDOT(yypos, yypos + 3))
| 421 => let val yytext=yymktext() in addChar (Char.chr(Char.ord(String.sub(yytext, 2))
				      -Char.ord #"@"));
		    continue() end
| 425 => (error(yypos, yypos + 2,
	       "illegal control escape; must be one of @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_");
	continue())
| 430 => let val yytext=yymktext() in let
		       val x =
			  Char.ord(String.sub(yytext, 1)) * 100
			  + Char.ord(String.sub(yytext, 2)) * 10
			  + Char.ord(String.sub(yytext, 3))
			  - (Char.ord #"0") *111
		    in (if x>255
			   then error(!stringstart, yypos,
				      "illegal ascii escape")
			else addChar(Char.chr x);
			   continue())
		    end end
| 437 => let val yytext=yymktext() in let
			   val x = 
			      StringCvt.scanString
			      (Pervasive.Int.scan StringCvt.HEX)
			      (Substring.string(Substring.substring(yytext, 2, 4)))
			   fun err() =
			      error(!stringstart, yypos, "illegal unicode escape")
			in (case x of
			       SOME x => if x > 255
					    then err()
					 else addChar(Char.chr x)
			     | _ => err())
			   ; continue()
			end end
| 44 => (Tokens.BAR(yypos, yypos + 1))
| 440 => (addString "\""; continue())
| 443 => (addString "\\"; continue())
| 447 => (YYBEGIN F; continue())
| 453 => (newline yypos ; YYBEGIN F ; continue())
| 455 => (error(!stringstart, yypos, "illegal string escape") ;
		    continue())
| 46 => (Tokens.COLON(yypos, yypos + 1))
| 460 => (newline yypos ;
		    error(!stringstart, yypos, "unclosed string");
		    continue())
| 463 => let val yytext=yymktext() in addString yytext; continue() end
| 465 => (error(yypos, yypos + 1, "illegal character in string");
	  continue())
| 470 => (newline yypos ; continue())
| 473 => (continue())
| 475 => (YYBEGIN S; stringstart := yypos; continue())
| 477 => (error(!stringstart, yypos, "unclosed string");
		    continue())
| 49 => (Tokens.COLONGT(yypos, yypos + 1))
| 51 => (Tokens.EQUALOP(yypos, yypos + 1))
| 53 => (Tokens.HASH(yypos, yypos + 1))
| 56 => (Tokens.ARROW(yypos, yypos + 2))
| 59 => (Tokens.DARROW(yypos, yypos + 2))
| 63 => (Tokens.AND(yypos, yypos + 3))
| 71 => (Tokens.ABSTYPE(yypos, yypos + 7))
| 74 => (Tokens.AS(yypos, yypos + 2))
| 79 => (Tokens.CASE(yypos, yypos + 4))
| 88 => (Tokens.DATATYPE(yypos, yypos + 8))
| 93 => (Tokens.ELSE(yypos, yypos + 4))
| 97 => (Tokens.END(yypos, yypos + 3))
| _ => raise Internal.LexerError

		) end )

	val {fin,trans} = Unsafe.Vector.sub(Internal.tab, s)
	val NewAcceptingLeaves = fin::AcceptingLeaves
	in if l = !yybl then
	     if trans = #trans(Vector.sub(Internal.tab,0))
	       then action(l,NewAcceptingLeaves
) else	    let val newchars= if !yydone then "" else yyinput 1024
	    in if (size newchars)=0
		  then (yydone := true;
		        if (l=i0) then UserDeclarations.eof yyarg
		                  else action(l,NewAcceptingLeaves))
		  else (if i0=l then yyb := newchars
		     else yyb := substring(!yyb,i0,l-i0)^newchars;
		     yygone := !yygone+i0;
		     yybl := size (!yyb);
		     scan (s,AcceptingLeaves,l-i0,0))
	    end
	  else let val NewChar = Char.ord(Unsafe.CharVector.sub(!yyb,l))
		val NewChar = if NewChar<128 then NewChar else 128
		val NewState = Unsafe.Vector.sub(trans, NewChar)
		in if NewState=0 then action(l,NewAcceptingLeaves)
		else scan(NewState,NewAcceptingLeaves,l+1,i0)
	end
	end
(*
	val start= if substring(!yyb,!yybufpos-1,1)="\n"
then !yybegin+1 else !yybegin
*)
	in scan(!yybegin (* start *),nil,!yybufpos,!yybufpos)
    end
in continue end
  in lex
  end
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature FRONT_END_STRUCTS = 
   sig
      structure Ast: AST
   end

signature FRONT_END = 
   sig
      include FRONT_END_STRUCTS
	 
      val lexAndParse: In.t -> Ast.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor FrontEnd(S: FRONT_END_STRUCTS): FRONT_END = 
struct

open S

structure LrVals = MLLrValsFun(structure Token = LrParser.Token
			       structure Ast = Ast)
structure Lex = MLLexFun(structure Tokens = LrVals.Tokens)
structure Parse = JoinWithArg(structure ParserData = LrVals.ParserData
			      structure Lex = Lex
			      structure LrParser = LrParser)

fun lexAndParse ins =
   let
      val stream =
	 Parse.makeLexer(fn n => In.inputN(ins,n))
	 {comLevel = ref 0,
	  stringtype = ref false,
	  stringstart = ref 0}
      val lookahead = 0
      val result =
	 (#1(Parse.parse(lookahead, stream,
			 fn (s, left, right) =>
			 Control.errorStr(Region.T{left = left,
						   right = right},
					  s),
			 ())))
	 handle _ =>
	    (Control.errorStr(Region.current(),
			      "unhandled exception in parser") ;
	     Ast.Program.T[])
	     
   in result
   end

(* val lexAndParse =
 *    Trace.trace("lexAndParse", Layout.ignore, Ast.Program.layout)
 *    lexAndParse
 *)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INFER_TYPE_STRUCTS =
   sig
      include ATOMS
      structure XmlType: XML_TYPE
      sharing Atoms = XmlType.Atoms
   end

signature INFER_TYPE = 
   sig
      include INFER_TYPE_STRUCTS

      structure Frees :
	 sig
	    include SET
	    val fromTyvars: Tyvar.t list -> t
	 end
      
      include TYPE_OPS sharing type tycon = Tycon.t

      (* constructors *)
      val new: {equality: bool, canGeneralize: bool} -> t
      val var: Tyvar.t -> t
      val record: {flexible: bool, record: t SortedRecord.t} -> t
      val ofConst: Ast.Const.t -> t

      (* destructors *)
      val derecord: t -> {flexible: bool, record: t SortedRecord.t}

      (* predicates *)
      val isUnit: t -> bool

      (* the set of free variables of a type *)
      val frees: t -> Frees.t

      (* close(t, f)
       * Make a list of type variables that are free in t, but do not appear in f.
       * Turn any unknown types into new type variables.
       * The second argument expressed as a thunk for performance reasons --
       * it is not necessary to know f if the type is closed.
       *)
      val close: t * (unit -> Frees.t) -> Tyvar.t list

      (* substitute(t, [(a1, t1), ..., (an, tn)]) performs simultaneous
       * substitution of the ti for ai in t.
       *)
      val substitute: t * (Tyvar.t * t) list -> t

      (* make two types identical (recursively).  side-effecting. *)
      val unify: t * t -> unit
	 
      (* make all types in the list identical.  return new type if empty list. *)
      val unifys: t list -> t 

      (* can two types be unified?  not side-effecting. *)
      val canUnify: t * t -> bool

      (* cached for speed *)
      val toXml: t -> XmlType.t

      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INFER_SCHEME_STRUCTS = 
   sig
      structure Type: INFER_TYPE
   end

signature INFER_SCHEME = 
   sig
      include INFER_SCHEME_STRUCTS
      include GENERIC_SCHEME
      sharing type tyvar = Type.Tyvar.t
      sharing type ty = Type.t

      structure Frees: SET sharing Frees = Type.Frees
      
      (* the set of free variables of a scheme, i.e. the frees of the body
       minus the bound tyvars *)
      val frees: t -> Frees.t
	 
      (* create a new copy of of the body of the scheme, replacing
       the bound tyvars with new unknown types of the same equality
       as the tyvar and the given canGeneralize. *)
      val instantiate :
	 {scheme: t,
	  canGeneralize: bool} -> {args: Type.t list,
				    instance: Type.t}
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_ENV_STRUCTS = 
   sig
      include ATOMS
      structure XmlType: XML_TYPE
      sharing Atoms = XmlType.Atoms
   end

signature TYPE_ENV = 
   sig
      include TYPE_ENV_STRUCTS
      
      structure InferScheme: INFER_SCHEME
      structure InferType: INFER_TYPE
      sharing Atoms = InferType.Atoms
      sharing InferType = InferScheme.Type
	 
      structure VarRange :
	 sig
	    datatype kind =
	       Normal
	     | Delayed
	     | Recursive of Tyvar.t list ref
	     | Overload of (Var.t * InferType.t) list
	       
	    datatype t = T of {scheme: InferScheme.t,
			       kind: kind}

	    val scheme: t -> InferScheme.t
	    val layout: t -> Layout.t
	 end

      type t

      (* empty environment *)
      val empty: t

      (* extend adds a new binding *)
      val extendVarRange: t * Var.t * VarRange.t -> t
      val extendVar: t * Var.t * InferScheme.t -> t

      (* lookup a binding *)
      val lookupVarRange: t * Var.t -> VarRange.t
      val lookupVar: t * Var.t -> InferScheme.t

      (* the union of the frees of all of the range elements *)
      val frees: t -> InferType.Frees.t

      val layout: t -> Layout.t
   end

signature INFER_TYPE_ENV = TYPE_ENV
signature SUM =
   sig
      structure X: T
      structure Y: T
	 
      datatype t =
	 X of X.t
       | Y of Y.t

      val outX: t -> X.t
      val outY: t -> Y.t
      val out: t * (X.t -> 'a) * (Y.t -> 'a) -> 'a
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
functor Sum(structure X: T
	    structure Y: T): SUM =
struct

structure X = X
structure Y = Y

datatype t =
   X of X.t
 | Y of Y.t

fun outX(X x) = x
  | outX _ = Error.error "outX"

fun outY(Y y) = y
  | outY _ = Error.error "outY"

fun out(X x, f, _) = f x
  | out(Y y, _, f) = f y

val equals =
   fn (X l, X l') => X.equals(l, l')
    | (Y l, Y l') => Y.equals(l, l')
    | _ => false

fun layout(X l) = X.layout l
  | layout(Y l) = Y.layout l

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor InferType(S: INFER_TYPE_STRUCTS): INFER_TYPE =
struct

open S
local open Ast
in structure Aconst = Const
end
structure Field = Record.Field
structure Srecord = SortedRecord

structure Type =
   struct
      structure Set = DisjointSet

      (* tuples of length <> 1 are always represented as records.
       * there will never be tuples of length one
       *)
      datatype ty =
	 Unknown of unknown
       | Var of Tyvar.t
       | Con of Tycon.t * t list
       | Word (* an unresolved word type *)
       | Int (* an unresolved int type *)
       | Record of {flexible: bool, record: t Srecord.t}
      withtype t = {ty: ty,
		    plist: PropertyList.t} Set.t
      and unknown = {equality: bool, canGeneralize: bool}

      fun joinUnknown(r: unknown, r': unknown): unknown =
	 {equality = #equality r orelse #equality r',
	  canGeneralize = #canGeneralize r andalso #canGeneralize r'}

      val toType: t -> ty = #ty o Set.value
      val plist: t -> PropertyList.t = #plist o Set.value

      (* unknown is a bit strange because frees needs the reference to the
       * whole type (Set.t) and not just the unknown info
       *)
      fun makeHom{unknown, var, con, word, int, record} =
	 let
	    val {get, set, destroy} =
	       Property.newDest(plist, Property.initConst NONE)

	    fun loop t =
	       case get t of
		  SOME x => x
		| NONE =>
		     case toType t of
			Word => word
		      | Int => int
		      | Unknown u => unknown(u, t)
		      | Var a => var a
		      | Con(c, ts) => con(c, List.map(ts, loop))
		      | Record{flexible, record = r} =>
			   record{flexible = flexible,
				  record = Srecord.map(r, loop)}
	 in {hom = loop, destroy = destroy}
	 end

      fun hom{ty, unknown, var, con, word, int, record} =
	 let val {hom, destroy} = makeHom{unknown = unknown,
					  var = var,
					  con = con,
					  word = word,
					  int = int,
					  record = record}
	 in hom ty before destroy()
	 end

      fun deconOpt t =
	 case toType t of
	    Con x => SOME x
	  | _ => NONE

      fun layoutDetailed ty =
	 let open Layout
	 in hom{ty = ty,
		unknown = 
		fn ({equality, canGeneralize}, _) =>
		namedRecord("Unknown",
			    [("equality", Bool.layout equality),
			     ("canGeneralize", Bool.layout canGeneralize)]),
		var = fn a => paren(seq[str "Var ", Tyvar.layout a]),
		word = str "Word",
		int = str "Int",
		con = fn (c, ts) => paren(align[seq[str "Con ", Tycon.layout c],
						tuple ts]),
		record = fn {flexible, record} =>
		paren(seq([str "Record ",
			   Srecord.layout
			   {record = record,
			    separator = "=",
			    extra = if flexible then "..." else "",
			    layoutTuple = tuple,
			    layoutElt = fn x => x}]))}
      end

      local
	 structure T = Ast.Type
      in
	 fun toAst t =
	    case toType t of
	       Unknown _ => T.unit
	     | Var a => T.var a
	     | Word => T.con(Ast.Tycon.word, [])
	     | Int => T.con(Ast.Tycon.int, [])
	     | Con(c, ts) => T.con(Tycon.toAst c, List.map(ts, toAst))
	     | Record{flexible, record} => T.record(Srecord.map(record, toAst))
      end

      fun newTy(ty: ty): t =
	 Set.singleton{ty = ty,
		       plist = PropertyList.new()}

      val new = newTy o Unknown

      val record = newTy o Record

      val tuple =
	 fn [t] => t
	  | ts => record{flexible = false,
			 record = Srecord.tuple ts}

      val unit = tuple[]

      fun isUnit s =
	 case toType s of
	    Record{flexible = false, record} =>
	       (case Srecord.detupleOpt record of
		   SOME [] => true
		 | _ => false)
	  | _ => false

      val _ = if isUnit unit then () else Error.bug "unit isn't unit"
	  
      fun con(tycon, ts) =
	 if Tycon.equals(tycon, Tycon.tuple) then tuple ts
	 else newTy(Con(tycon, ts))

      val layout = Layout.switch{detailed = layoutDetailed,
				 normal = Ast.Type.layout o toAst}
   
      val equals: t * t -> bool = Set.equals
   end

structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
			   
open Ops Type (* order is important, since want specialized definitions
	       in Type to override general definitions in Ops *)

structure Free =
   struct
      structure F = Sum(structure X = Tyvar
			structure Y = Type)
      open F
      val fromTyvar = X
      val fromType = Y
   end

structure Frees =
   struct
      structure S = UnorderedSet(Free)
      open S

      fun fromTyvars tyvs = fromList(List.map(tyvs, Free.fromTyvar))
   end

local
   structure Env = MonoEnv(structure Domain = Tyvar
			   structure Range = Type)
in
   fun substitute(t: t, tyvarsAndTypes: (Tyvar.t * Type.t) list): t =
      let val env = Env.fromList tyvarsAndTypes
	 fun loop t =
	    case toType t of
	       Unknown _ => t
	     | Word => t
	     | Int => t
	     | Var a => (case Env.peek(env, a) of
				  NONE => t
				| SOME t => t)
	     | Con(c, ts) => newTy(Con(c, List.map(ts, loop)))
	     | Record{flexible, record} =>
		  if flexible
		     then Error.bug "substituting in flexible record type"
		  else newTy(Record{flexible = false,
				    record = Srecord.map(record, loop)})
      in loop t
      end
end

(* val substitute =
 *    Trace.trace2("substitute", layout, Layout.ignore, layout) substitute		
 *)

val var = newTy o Var

(*val con = Trace.trace2("con", Tycon.layout,
		       List.layout(", ", layout),
		       layout) con*)
fun ofConst c =
   case c of
      Aconst.Int _ => newTy Type.Int
    | Aconst.Word _ => newTy Type.Word
    | Aconst.Real _ => real
    | Aconst.String _ => string
    | Aconst.Char _ => char

fun derecord t =
   case toType t of
      Record r => r
    | _ => Error.error "Type.deRecord"

(* val traceFrees = Trace.trace("frees", layoutDetailed, Frees.layout) *)

local open Frees
in
   fun frees ty =
      hom{ty = ty,
	  unknown = (fn ({canGeneralize, ...}, t) =>
		     if canGeneralize
			then singleton(Free.fromType t)
		     else empty),
	  var = fn a => singleton(Free.fromTyvar a),
	  word = empty,
	  int = empty,
	  con = fn (_, fs) => unions fs,
	  record = fn {record, ...} => unions(Srecord.range record)}
end

fun close(t, getFs) =
   let val fs = frees t
   in if Frees.isEmpty fs
      then []
      else List.map(Frees.toList(Frees.-(fs, getFs())), fn free =>
		    case free of
		       Free.X a => a
		     | Free.Y s => 
			  case toType s of
			     Unknown{equality, ...} =>
				let val a = Tyvar.newNoname{equality = equality}
				in Set.setValue(s, {ty = Var a,
						    plist = PropertyList.new()});
				   a
				end
			   | _ => Error.bug "Type.close: not unknown")
   end

(* val close =
 *    Trace.trace2("close", layout, Function.layout, List.layout Tyvar.layout)
 *    close
 *)

(*val traceCanUnify = Trace.trace2("canUnify", layout, layout, Bool.layout) *)

fun canUnify arg = 
(*   traceCanUnify *)
   (fn (t, t') =>
    case (toType t,   toType t') of
         (Unknown _,  _)            => true
       | (_,          Unknown _)    => true
       | (Con(c, ts), Con(c', ts')) => (Tycon.equals(c, c')
					andalso
					List.forall2(ts, ts', canUnify))
       | (Con(c, []), Word)         => Tycon.isWordX c
       | (Word,       Con(c, []))   => Tycon.isWordX c
       | (Con(c, []), Int)          => Tycon.isIntX c
       | (Int,        Con(c, []))   => Tycon.isIntX c
       | (Var a,      Var a')       => Tyvar.equals(a, a')
       | (Word,       Word)         => true
       | (Int,        Int)          => true
       | (Record{flexible=f, record=r}, Record{flexible=f', record=r'}) =>
	    if f orelse f' then Error.bug "canUnify on flexible record"
	    else let val fs = Srecord.toList r
		     val fs' = Srecord.toList r'
		 in Int.equals(List.length fs, List.length fs')
		    andalso List.forall2(fs, fs', fn ((f, t), (f', t')) =>
					 Field.equals(f, f') andalso canUnify(t, t'))
		 end
       | _                          => false) arg

(*val traceUnify = Trace.trace2("unify", layout, layout, Unit.layout) *)

fun unify arg =
(*   traceUnify *)
   (fn (s, s') =>
    if Set.equals(s, s') then ()
    else
       let
	  val t = toType s
	  val t' = toType s'
	  val t =
	     case (t,          t')           of
		  (Unknown r,  Unknown r')   => Unknown(joinUnknown(r, r'))
		| (t,          Unknown _)    => t
		| (Unknown _,  t)            => t
		| (Var a,      Var a')       => if Tyvar.equals(a, a')
						   then t
						else Error.error "unify Var"
		| (Con(c, ts), Con(c', ts')) => if Tycon.equals(c, c')
						   then (unifys(ts, ts'); t)
						else Error.error "unify Con"
	        | (Con(c, []), Word)         => if Tycon.isWordX c
						   then t
						else Error.error "unify Word"
		| (Word,       Con(c, []))   => if Tycon.isWordX c
						 then t'
						else Error.error "unify Word"
		| (Con(c, []), Int)          => if Tycon.isIntX c
						   then t
						else Error.error "unify Int"
		| (Int,        Con(c, []))    => if Tycon.isIntX c
						   then t'
						else Error.error "unify Int"
		| (Word,       Word)         => t
		| (Int,        Int)          => t
		| (Record{flexible=f, record=r}, Record{flexible=f', record=r'}) =>
		     let val fs = Srecord.toList r
			val fs' = Srecord.toList r'
			val rec oneFlex =
			   fn ([], _) => ()
			    | (fs0 as (f, t) :: fs, (f', t') :: fs') =>
				 if Field.equals(f, f')
				    then (unify(t, t');
					  oneFlex(fs, fs'))
				 else oneFlex(fs0, fs')
			    | _ => Error.error "impossible unification of flexible record pattern"
				 
		     in if f
			   then
			      if f'
				 then
				    let
				       val rec loop =
					  fn ([], []) => []
					   | (fs, []) => fs
					   | ([], fs) => fs
					   | ((f, t) :: fs, (f', t') :: fs') =>
						if Field.equals(f, f')
						   then (unify(t, t');
							 (f, t) :: loop(fs, fs'))
						else if Field.<(f, f')
							then (f, t) :: loop(fs, (f', t') :: fs')
						     else (f', t') :: loop((f, t) :: fs, fs')
				    in Record{flexible = true,
					      record = Srecord.fromList(loop(fs, fs'))}
				    end
			      else (oneFlex(fs, fs'); t')
			      else if f' then (oneFlex(fs', fs); t)
			      else if Int.equals(List.length fs, List.length fs')
				      then (List.foreach2(fs, fs', fn ((_, t), (_, t')) =>
							  unify(t, t'));
					    t)
				   else Error.error "unify mismatch: different length records"
		     end
		| _ => Error.error "unify mismatch"
       in Set.union(s, s')
	  ; Set.setValue(s, {ty = t, plist = PropertyList.new()})
       end) arg

and unifys(ts: t list, ts': t list): unit =
   let
      val rec loop =
	 fn ([], []) => ()
	  | (t :: ts, t' :: ts') => (unify(t, t'); loop(ts, ts'))
	  | _ => Error.error "unify Con: different number of args"
   in loop(ts, ts')
   end

fun unifys ts =
   case ts of
      t :: ts => (List.foreach(ts, fn t' => unify(t, t')) ; t)
    | [] => new{equality = false, canGeneralize = true}

local
   structure X = XmlType
   val con = X.con
   val unknown = con(Tycon.tuple, [])
   val {hom, ...} =
      makeHom{unknown = fn _ => unknown,
	      var = X.var,
	      word = con(Tycon.defaultWord, []),
	      int = con(Tycon.defaultInt, []),
	      con = con,
	      record = fn {flexible, record} =>
	      if flexible
		 then Error.error "unresolved ... in flexible pattern"
	      else
		 case Srecord.range record of
		    [t] => t
		  | ts => con(Tycon.tuple, ts)}
in
   val toXml = hom

(*   val toXml = Trace.trace("toXml", layout, XmlType.layout) toXml *)
end

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor InferScheme(S: INFER_SCHEME_STRUCTS): INFER_SCHEME = 
struct

open S

local open Type
in structure Frees = Frees
   structure Tyvar = Tyvar
end

structure Scheme = GenericScheme(structure Tyvar = Tyvar
				 structure Type = Type)
open Scheme

fun frees(T{tyvars, ty}) =
   Frees.-(Type.frees ty, Frees.fromTyvars tyvars)

fun instantiate{scheme, canGeneralize} =
   let
      val ts = List.map(tyvars scheme, fn v =>
			Type.new{equality = Tyvar.isEquality v,
				 canGeneralize = canGeneralize})
   in {instance = apply(scheme, ts), args = ts}
   end
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeEnv(S: TYPE_ENV_STRUCTS): TYPE_ENV =
struct

open S

structure InferType = InferType(S)
   
structure InferScheme = InferScheme(structure Type = InferType)
   
structure Frees = InferType.Frees

structure VarRange =
   struct
      datatype kind =
	 Normal
       | Delayed
       | Recursive of Tyvar.t list ref
       | Overload of (Var.t * InferType.t) list

      datatype t = T of {scheme: InferScheme.t,
			 kind: kind}

      fun scheme(T{scheme, ...}) = scheme

      fun layout(T{scheme, ...}) = InferScheme.layout scheme
   end

structure Opens(* :
   sig
      type t
      val empty: t
      val maybeAdd: t * InferScheme.t -> t
      val append: t * t -> t
      val frees: t -> Frees.t
      val layout: t -> Layout.t
   end*) =
   struct
      datatype list =
	 Empty
       | Cons of InferScheme.t * t
      withtype t = list ref

      fun toList opens =
	 case !opens of
	    Empty => []
	  | Cons(s, opens) => s :: (toList opens)

      local open Layout
      in fun layout opens = list(List.map(toList opens, InferScheme.layout))
      end

      fun append(opens, opens') =
	 let
	    fun loop opens =
	       case !opens of
		  Empty => opens'
		| Cons(s, opens) => ref(Cons(s, loop opens))
	 in loop opens 
	 end

      val empty = ref Empty

      fun maybeAdd(opens, s) =
	 if Frees.isEmpty(InferScheme.frees s)
	    then opens
	 else ref(Cons(s, opens))
	    
(*       val maybeAdd =
 * 	 Trace.trace2("maybeAdd", layout, InferScheme.layout, layout)
 * 	 maybeAdd
 *)

      fun frees opens =
	 let
	    fun loop(r, r', fs) =
	       case !r' of
		  Empty => (r := Empty ; fs)
		| pair as Cons(s, r') =>
		     let val fs' = InferScheme.frees s
		     in if Frees.isEmpty fs'
			   then loop(r, r', fs)
			else (r := pair ; loop(r', r', Frees.+(fs', fs)))
		     end
	 in loop(opens, opens, Frees.empty)
	 end
   end

datatype t = T of {opens: Opens.t}

fun layout(T{opens}) =
   let open Layout
   in record[("opens", Opens.layout opens)]
   end	     

val empty = T{opens = Opens.empty}

(* this property is destroyed in infer.fun after type inference is done *)
val {get = getVarRange: Var.t -> VarRange.t, set = setVarRange} =
   Property.new(Var.plist, Property.initRaise("range", Var.layout))
   
fun extendVarRange(T{opens}, x, r) =
   (setVarRange(x, r) ;
    T{opens = Opens.maybeAdd(opens, VarRange.scheme r)})
fun lookupVarRange(_, x) = getVarRange x
fun lookupVar(e, x) = VarRange.scheme(lookupVarRange(e, x))
fun extendVar(e, x, s) = extendVarRange(e, x, VarRange.T{scheme = s,
						       kind = VarRange.Normal})   

fun frees(T{opens, ...}) = Opens.frees opens

(*val frees = Trace.trace("frees", layout, Frees.layout) frees *)
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature SCOPE_STRUCTS =
   sig
      structure CoreML: CORE_ML
   end

signature SCOPE =
   sig
      include SCOPE_STRUCTS

      (* add free type variables to the val or fun declaration
       where they are implicitly scoped *)
      val scope: CoreML.Program.t -> CoreML.Program.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 renameDec walks down the tree renaming all explicitly bound tyvars, and on the
 way back up, tries to bind implicitly scoped tyvars at each possible point.
 
 removeDec walks down and binds a tyvar as soon as it sees it, removing all
 lower binding occurrences of the tyvar. 

 removeDec also renames all lower free occurrences of the tyvar to be the "same"
 as the binding occurrence (so that they can share info)
 *)

functor Scope(S: SCOPE_STRUCTS): SCOPE =
struct

open S
open CoreML

structure Env = TyvarEnv
structure Tyvars = UnorderedSet(UseName(Tyvar))

fun renames(xs, renameX) =
   let
      val (xs, unguarded) =
	 List.fold(xs, ([], Tyvars.empty),
		    fn (x, (xs, tyvars)) =>
		    let val (x, unguarded) = renameX x
		    in (x :: xs,
			Tyvars.+(tyvars, unguarded))
		    end)
   in (List.reverse xs, unguarded)
   end

fun renameTy(t, env) =
   let open Type
      fun loop t =
	 case t of
	    Var a =>
	       let val a = (case Env.peek(env, a) of
			       NONE => a
			     | SOME a => a)
	       in (Var a, Tyvars.singleton a)
	       end
	  | Con(c, ts) => let val (ts, unguarded) = loops ts
			 in (Con(c, ts), unguarded)
			 end
	  | Record r => let val (r, u) = Record.change(r, loops)
			in (Record r, u)
			end
      and loops ts = renames(ts, loop)
   in loop t
   end

fun renameTyOpt(to, env) =
   case to of
      NONE => (NONE, Tyvars.empty)
    | SOME t => let val (t, unguarded) = renameTy(t, env)
		in (SOME t, unguarded)
		end

fun renamePat(p, env) =
   let open Pat
      fun loopOpt opt =
	 case opt of
	    NONE => (NONE, Tyvars.empty)
	  | SOME p => let val (p, u) = loop p
		      in (SOME p, u)
		      end
      and loop p =
	 case p of
	    Wild => (p, Tyvars.empty)
	  | Var _ => (p, Tyvars.empty)
	  | Const _ => (p, Tyvars.empty)
	  | Con{con, arg} => let val (arg, unguarded) = loopOpt arg
			    in (Con{con = con, arg = arg}, unguarded)
			    end
	  | Record{flexible, record} =>
	       let val (r, u) = Record.change(record, loops)
	       in (Record{flexible = flexible,
			  record = r},
		   u)
	       end
	  | Constraint(p, t) =>
	       let val (p, unguarded) = loop p
		  val (t, unguarded') = renameTy(t, env)
	       in (Constraint(p, t),
		   Tyvars.+(unguarded, unguarded'))
	       end
	  | Layered(x, p) =>
	       let val (p, unguarded) = loop p
	       in (Layered(x, p), unguarded)
	       end
      and loops ps = renames(ps, loop)
   in loop p
   end

open Dec Exp
   
fun renameDec(d, env) =
   case d of
      Val{tyvars, pat, exp} =>
	 let val (env, tyvars) = TyvarEnv.rename(env, tyvars)
	    val (pat, u1) = renamePat(pat, env)
	    val (exp, u2) = renameExp(exp, env)
	 in (Val{tyvars =
		 Tyvars.toList(Tyvars.unions[u1, u2, Tyvars.fromList tyvars]),
		 pat = pat, exp = exp},
	     Tyvars.empty)
	 end
    | Fun{tyvars, decs} =>
	 let val (env, tyvars) = TyvarEnv.rename(env, tyvars)
	    val (decs, unguarded) =
	       renames(decs, fn {var, ty, rules} =>
		       let val (ty, u1) = renameTyOpt(ty, env)
			  val (rules, u2) = renameRules(rules, env)
		       in ({var = var, ty = ty, rules = rules},
			   Tyvars.+(u1, u2))
		       end)
	 in (Fun{tyvars = Tyvars.toList(Tyvars.+(unguarded,
						     Tyvars.fromList tyvars)),
		 decs = decs},
	     Tyvars.empty)
	 end
    | Datatype dbs =>
	 let
	    val (dbs, unguarded) =
	       renames
	       (dbs, fn {tyvars, tycon, cons} =>
		let val (env, tyvars) = TyvarEnv.rename(env, tyvars)
		   val (cons, unguarded) =
		      renames(cons, fn {con, arg} =>
			      let val (arg, unguarded) = renameTyOpt(arg, env)
			      in ({con = con, arg = arg}, unguarded)
			      end)
		in ({tyvars = tyvars, tycon = tycon, cons = cons},
		    Tyvars.-(unguarded, Tyvars.fromList tyvars))
		end)
	 in (Datatype dbs, unguarded)
	 end
    | Exception{con, arg} =>
	 let val (arg, unguarded) = renameTyOpt(arg, env)
	 in (Exception{con = con, arg = arg}, unguarded)
	 end
    | Overload _ => (d, Tyvars.empty)
and renameDecs(ds, env) = renames(ds, fn d => renameDec(d, env))
and renameExp(e, env) =
   let val empty = (e, Tyvars.empty)
   in case e of
      Var _ => empty
    | Prim _ => empty
    | Const _ => empty
    | Con _ => empty
    | Record r =>
	 let val (r, u) = Record.change(r, fn es => renameExps(es, env))
	 in (Record r, u)
	 end
    | Fn rs =>
	 let val (rs, unguarded) = renameRules(rs, env)
	 in (Fn rs, unguarded)
	 end
    | App(e1, e2) =>
	 let val (e1, u1) = renameExp(e1, env)
	    val (e2, u2) = renameExp(e2, env)
	 in (App(e1, e2), Tyvars.+(u1, u2))
	 end
    | Let(ds, e) =>
	 let val (ds, u1) = renameDecs(ds, env)
	    val (e, u2) = renameExp(e, env)
	 in (Let(ds, e), Tyvars.+(u1, u2))
	 end
    | Constraint(e, t) =>
	 let val (e, u1) = renameExp(e, env)
	    val (t, u2) = renameTy(t, env)
	 in (Constraint(e, t), Tyvars.+(u1, u2))
	 end
    | Handle(e, rs) =>
	 let val (e, u1) = renameExp(e, env)
	    val (rs, u2) = renameRules(rs, env)
	 in (Handle(e, rs), Tyvars.+(u1, u2))
	 end
    | Raise e =>
	 let val (e, unguarded) = renameExp(e, env)
	 in (Raise e, unguarded)
	 end
   end
and renameExps(es, env) = renames(es, fn e => renameExp(e, env))
and renameRules(rs, env) =
   let val (rs, unguarded) = renames(rs, fn r => renameRule(r, env))
   in (rs, unguarded)
   end
and renameRule((p, e), env) =
   let val (p, u1) = renamePat(p, env)
      val (e, u2) = renameExp(e, env)
   in ((p, e), Tyvars.+(u1, u2))
   end      

(*---------------------------------------------------*)
(*                      Remove                       *)
(*---------------------------------------------------*)
    
fun bindNew(bound: Env.t, tyvars: Tyvar.t list): Env.t * Tyvar.t list =
   TyvarEnv.rename(bound,
	  Tyvars.toList(Tyvars.-(Tyvars.fromList tyvars,
				 Tyvars.fromList(Env.domain bound))))

fun removeTy(ty: Type.t, env: Env.t): Type.t =
   let open Type
      fun loop ty =
	 case ty of
	    Var a => Var(Env.lookup(env, a))
	  | Con(c, tys) => Con(c, List.map(tys, loop))
	  | Record r => Record(Record.map(r, loop))
   in loop ty
   end

fun removeTyOpt(ty: Type.t option, env: Env.t): Type.t option =
   case ty of
      NONE => NONE
    | SOME ty => SOME(removeTy(ty, env))

fun removePat(p: Pat.t, env: Env.t): Pat.t =
   let open Pat
      fun loopOpt opt =
	 case opt of
	    NONE => NONE
	  | SOME p => SOME(loop p)
      and loops ps = List.map(ps, loop)
      and loop p =
	 case p of
	    Wild => p
	  | Var _ => p
	  | Const _ => p
	  | Con{con, arg} => Con{con = con, arg = loopOpt arg}
	  | Record{flexible, record} => Record{flexible = flexible,
					      record = Record.map(record, loop)}
	  | Constraint(p, t) => Constraint(loop p, removeTy(t, env))
	  | Layered(x, p) => Layered(x, loop p)
   in loop p
   end
      
fun removes(xs, scope, removeX) =
   List.map(xs, fn x => removeX(x, scope))

fun removeDec(d: Dec.t, scope: Env.t): Dec.t =
   case d of
      Val{tyvars, pat, exp} =>
	 let val (scope, tyvars) = bindNew(scope, tyvars)
	 in Val{tyvars = tyvars,
		pat = removePat(pat, scope), exp = removeExp(exp, scope)}
	 end
    | Fun{tyvars, decs} =>
	 let val (scope, tyvars) = bindNew(scope, tyvars)
	 in Fun{tyvars = tyvars,
		decs = List.map(decs, fn {var, ty, rules} =>
				{var = var, ty = removeTyOpt(ty, scope),
				 rules = removeRules(rules, scope)})}
	 end
    | Exception{con, arg} => Exception{con = con,
				       arg = removeTyOpt(arg, scope)}
    | Datatype dbs =>
	 Datatype
	 (List.map(dbs, fn {tyvars, tycon, cons} =>
		   let val (scope, tyvars) = TyvarEnv.rename(scope, tyvars)
		   in {tyvars = tyvars,
		       tycon = tycon,
		       cons = List.map(cons, fn {con, arg} =>
				       {con = con,
					arg = removeTyOpt(arg, scope)})}
		   end))
    | Overload _ => d

and removeDecs(ds: Dec.t list, scope: Env.t): Dec.t list =
   removes(ds, scope, removeDec)

and removeExp(e: Exp.t, scope: Env.t): Exp.t =
   case e of
      Var _ => e
    | Prim _ => e
    | Const _ => e
    | Con _ => e
    | Record r => Record(Record.map(r, fn e => removeExp(e, scope)))
    | Fn rules => Fn(removeRules(rules, scope))
    | App(e1, e2) => App(removeExp(e1, scope), removeExp(e2, scope))
    | Let(ds, e) => Let(removeDecs(ds, scope), removeExp(e, scope))
    | Constraint(e, t) => Constraint(removeExp(e, scope), removeTy(t, scope))
    | Handle(e, rs) => Handle(removeExp(e, scope), removeRules(rs, scope))
    | Raise e => Raise(removeExp(e, scope))
and removeExps(es, scope) = removes(es, scope, removeExp)
and removeRules(rs, scope) = removes(rs, scope, removeRule)
and removeRule((p, e), scope) = (removePat(p, scope), removeExp(e, scope))
    
fun scope d =
   let val (d, unguarded) = renameDec(d, Env.empty)
   in if Tyvars.isEmpty unguarded
	 then removeDec(d, Env.empty)
      else Error.error "scope: free type variable"
   end

val scope =
   fn (Program.T{constructors, decs}) =>
   Program.T{constructors = constructors,
	     decs = List.map(decs, scope)}

(*val scope = Trace.trace("scope", Program.layout, Program.layout) scope *)
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature NESTED_PAT_STRUCTS = 
   sig
      include ATOMS
      structure Type :
	 sig
	    type t
	    val tuple: t list -> t
	 end
   end

signature NESTED_PAT = 
   sig
      include NESTED_PAT_STRUCTS
      
      datatype t = T of {pat: node, ty: Type.t}
      and node =
	 Wild
       | Var of Var.t
       | Const of Const.t
       | Con of {con: Con.t, targs: Type.t list, arg: t option}
       | Tuple of t list
       | Layered of Var.t * t

      (* constants *)
      val unit: t

      (* constructors *)
      val new: node * Type.t -> t
      val tuple: t list -> t
      val wild: Type.t -> t

      (* destructors *)
      val node: t -> node
      val ty: t -> Type.t

      (* true if the pattern contains a constant, constructor or variable *)
      val isRefutable: t -> bool

      (* a list of the variables in the pattern, along with their types.
       Used for match compilation in order to build a function that abstracts
       over the expression of a case rule p => e.
       See infer.fun *)
      val varsAndTypes: t -> (Var.t * Type.t) list
	 
      val layout: t -> Layout.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor NestedPat(S: NESTED_PAT_STRUCTS): NESTED_PAT = 
struct

open S

datatype t = T of {pat: node, ty: Type.t}
and node =
   Wild
  | Var of Var.t
  | Const of Const.t
  | Con of {con: Con.t, targs: Type.t list, arg: t option}
  | Tuple of t list
  | Layered of Var.t * t

fun node(T{pat, ...}) = pat
fun ty(T{ty, ...}) = ty
   
fun tuple[p] = p
  | tuple ps = T{pat = Tuple ps, ty = Type.tuple(List.map(ps, ty))}

local structure Pat = Ast.Pat
in fun toAst p =
   case node p of
      Wild => Pat.wild
    | Var x => Pat.var(Var.toAst x)
    | Const c => Const.toAstPat c
    | Con{con, arg, ...} =>
	 let val con = Con.toAst con
	 in case arg of
	    NONE => Pat.con con
	  | SOME p => Pat.app(con, toAst p)
	 end
    | Tuple ps => Pat.tuple(List.map(ps, toAst))
    | Layered(x, p) => Pat.layered{fixop = Ast.Fixop.None,
				  var = Var.toAst x,
				  constraint = NONE,
				  pat = toAst p}
end

val layout = Ast.Pat.layout o toAst

fun new(p, t) =
   case p of
      Tuple[p] => p
    | _ => T{pat = p, ty = t}

fun wild t = new(Wild, t)
   
fun isRefutable p =
   case node p of
      Wild => false
    | Var _ => false
    | Const _ => true
    | Con _ => true
    | Tuple ps => List.exists(ps, isRefutable)
    | Layered(_, p) => isRefutable p

val unit = T{pat = Tuple[], ty = Type.tuple[]}
   
fun varsAndTypes(p: t): (Var.t * Type.t) list =
   let
      fun loop(p: t, accum: (Var.t * Type.t) list) =
	 case node p of
	    Wild => accum
	  | Const _ => accum
	  | Var x => (x, ty p) :: accum
	  | Tuple ps => List.fold(ps, accum, loop)
	  | Con{arg, ...} => (case arg of
				NONE => accum
			      | SOME p => loop(p, accum))
	  | Layered(x, p) => loop(p, (x, ty p) :: accum)
   in loop(p, [])
   end

fun vars(p: t): Var.t list =
   List.map(varsAndTypes p, #1)

end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MATCH_COMPILE_STRUCTS =
   sig
      include ATOMS
      structure Type :
	 sig
	    type t
	    val detuple: t -> t list
	    val layout: t -> Layout.t
	 end
      structure Exp :
	 sig
	    type t
	    val const: Const.t -> t
	    val var: Var.t * Type.t -> t
	    val detuple: {tuple: t,
			  body: (Var.t * Type.t) list -> t} -> t
	    val casee: {test: t,
			default: t option,
			cases: {
				con: Con.t,
				targs: Type.t list,
				arg: (Var.t * Type.t) option,
				rhs: t
				} list,
			ty: Type.t (* type of entire case expression *)
			} -> t
	    val lett: {var: Var.t, exp: t, body: t} -> t
	    val iff: {test: t, thenn: t, elsee: t, ty: Type.t} -> t
	    val equal: t * t -> t
	    val deref: t -> t
	    val layout: t -> Layout.t
	 end
      structure NestedPat: NESTED_PAT
      sharing Atoms = NestedPat.Atoms
      sharing Type = NestedPat.Type
   end

signature MATCH_COMPILE =
   sig
      include MATCH_COMPILE_STRUCTS

      val matchCompile :
	 {conTycon: Con.t -> Tycon.t,
	  tyconCons: Tycon.t -> Con.t list,
	  test: Var.t,
	  testType: Type.t,
	  caseType: Type.t, (* type of entire expression *)
	  cases: (NestedPat.t * ((Var.t -> Var.t) -> Exp.t)) list}
	 -> Exp.t
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor MatchCompile(S: MATCH_COMPILE_STRUCTS): MATCH_COMPILE =
struct

open S

structure Env = MonoEnv(structure Domain = Var
			structure Range = Var)
   
structure FlatPat =
   struct
      datatype t =
	 Any
       | Const of Const.t
       | Con of {con: Con.t,
		 targs: Type.t list,
		 arg: NestedPat.t option}
       | Tuple of NestedPat.t list

      val isRefutable =
	 fn Any => false
	  | Const _ => true
	  | Con _ => true
	  | Tuple ps => List.exists(ps, NestedPat.isRefutable)

      val isAny =
	 fn Any => true
	  | _ => false

      (* get rid of Wild, Var, Layered, Or - also remove unary tuples *)
      fun flatten(var: Var.t, pat: NestedPat.t, env: Env.t): t list * Env.t =
	 let fun done p = ([p], env)
	    fun extend x = Env.extend(env, x, var)
	 in case NestedPat.node pat of
	    NestedPat.Wild => done Any
	  | NestedPat.Var x => ([Any], extend x)
	  | NestedPat.Layered(x, p) => flatten(var, p, extend x)
	  | NestedPat.Const c => done(Const c)
	  | NestedPat.Con x => done(Con x)
	  | NestedPat.Tuple[p] => flatten(var, p, env)
	  | NestedPat.Tuple ps => done(Tuple ps)
	 end

      fun flattens(vars, pats: NestedPat.t list, env): t list list * Env.t =
	 let
	    val (patss, env) =
	       List.fold2
	       (vars, pats, ([], env), fn (var, pat, (patss, env)) =>
		let val (pats, env) = flatten(var, pat, env)
		in (pats :: patss, env)
		end)
	 in (List.cross(List.reverse patss), env)
	 end
   end   

datatype continue =
   Finish of (Var.t -> Var.t) -> Exp.t
 | Matches of FlatPat.t list option * continue
type info = {accum: Env.t, continue: continue}
type rule = {pat: NestedPat.t, info: info}
type rules = rule list
type flatrule = {pat: FlatPat.t, info: info}
type flatrules = flatrule list
fun layoutFlatrules(_: flatrules) = Layout.str "<flatrules>"
type finish = info list -> Exp.t
fun layoutFinish(_: finish) = Layout.str "<finish>"

(* val traceMatchFlat =
 *    Trace.trace4
 *    ("matchFlat", Var.layout, Type.layout, layoutFlatrules, layoutFinish,
 *     Exp.layout)
 *)

(*---------------------------------------------------*)
(*                   matchCompile                    *)
(*---------------------------------------------------*)

fun matchCompile{test: Var.t,
		 conTycon: Con.t -> Tycon.t,
		 tyconCons: Tycon.t -> Con.t list,
		 testType: Type.t,
		 caseType: Type.t,
		 cases: (NestedPat.t * ((Var.t -> Var.t) -> Exp.t)) list}
  : Exp.t =
   let
      fun match(var: Var.t, ty: Type.t, rules: rules, finish: finish)
	: Exp.t =
	 let
	    val rules =
	       List.fold(List.reverse rules, [], fn (rule, rules) =>
			  flatten(var, rule, rules))
	 in matchFlat(var, ty, rules, finish)
	 end
      
      and flatten(var: Var.t,
		  {pat, info as {accum, continue}}: rule,
		  rules: flatrules): flatrules =
	 let val (pats, accum) = FlatPat.flatten(var, pat, accum)
	 in List.fold(List.reverse pats, rules,
		       fn (pat: FlatPat.t, rules: flatrules) =>
		       {pat = pat, info = {accum = accum,
					   continue = continue}}
		       :: rules)
	 end
      
      and matchFlat arg: Exp.t =
(*	 traceMatchFlat *)
	 (fn (var: Var.t, ty: Type.t, rules: flatrules, finish: finish) =>
	 let
	    val test = Exp.var(var, ty)
	    fun loop rules' =
	       case rules' of
		  [] => finish(List.map(rules, fn {pat, info} =>
					case pat of
					   FlatPat.Any => info
					 | _ => Error.bug "matchFlat"))
		| {pat, info} :: rs =>
		     case pat of
			FlatPat.Any => loop rs
		      | FlatPat.Const _ => const(test, rules, finish)
		      | FlatPat.Con _ => sum(test, rules, finish)
(*		      | FlatPat.Tuple[] =>
			   finish
			   (List.map(rules, fn {pat, info} =>
				     case pat of
					(FlatPat.Any | FlatPat.Tuple[]) =>
					   info
				      | _ => Error.bug "expected unit pattern"))*)
		      | FlatPat.Tuple ps => tuple(test, ty, rules, finish)
	 in loop rules
	 end) arg
	 
      and matches(vars: (Var.t * Type.t) list,
		  rules: {pats: NestedPat.t list option, info: info} list,
		  finish: finish): Exp.t =
	 let
	    val rules =
	       List.fold
	       (List.reverse rules, [],
		fn ({pats, info as {accum, continue}}, rules) =>
		case pats of
		   NONE => {pats = NONE, info = info} :: rules
		 | SOME pats =>
		      let
			 val (patss, accum) =
			    FlatPat.flattens(List.map(vars, #1), pats, accum)
		      in List.fold
			 (List.reverse patss, rules, fn (pats, rules) =>
			  {pats = SOME pats,
			   info = {accum = accum, continue = continue}}
			  :: rules)
		      end)
	 in matchesFlat(vars, rules, finish)
	 end
      
      and matchesFlat(vars: (Var.t * Type.t) list,
		      rules: {pats: FlatPat.t list option, info: info} list,
		      finish: finish): Exp.t =
	 let
	    (* bring the i'th element of the list to the front *)
	    fun rotate(l, i) =
	       let
		  fun loop(prev, i, rest) =
		     case (i, rest) of
			(0, x :: rest) => x :: (List.reverseAppend(prev, rest))
		      | (_, x :: rest) => loop(x :: prev, i-1, rest)
		      | _ => Error.error "rotate"
	       in loop([], i, l)
	       end
	    val normal = (vars, rules)
	    val (vars, rules) =
	       case (!Control.match, rules) of
		  (Control.LeftToRight, _) => normal
		| (_, []) => normal
		| (_, [_]) => normal
		| (Control.Rotate, {pats, info} :: _) =>
		  case pats of
		     NONE => normal
		   | SOME pats =>
			case List.index(pats, FlatPat.isRefutable) of
			   NONE => normal
			 | SOME 0 => normal
			 | SOME i =>
			      (rotate(vars, i),
			       List.map(rules, fn {pats, info} =>
					{pats = (case pats of
						    NONE => NONE
						  | SOME ps => SOME(rotate(ps, i))),
					 info = info}))
	 in case vars of
	    [] =>
	       finish(List.map(rules, fn {pats, info} =>
			       (case pats of
				   SOME(_ :: _) => Error.bug "matchesFlat1"
				 | _ => ()
			        ; info)))
	  | (var, ty) :: vars =>
	       let
		  val rules =
		     List.map
		     (rules, fn {pats, info as {accum, continue}} =>
		      case pats of
			 NONE => {pat = FlatPat.Any,
				  info = {accum = accum,
					  continue = Matches(NONE, continue)}}
		       | SOME(pat :: pats) =>
			    {pat = pat,
			     info = {accum = accum,
				     continue = Matches(SOME pats, continue)}}
		       | _ => Error.bug "wrong number of patterns")
	       in matchFlat
		  (var, ty, rules, fn rules =>
		   matchesFlat
		   (vars,
		    List.map(rules, fn {accum, continue} =>
			     case continue of
				Matches(pats, continue) =>
				   {pats = pats,
				    info = {accum = accum,
					    continue = continue}}
			      | _ => Error.bug "matchesFlat:"),
		    finish))
	       end
	 end

      (*------------------------------------*)
      (*               tuple                *)
      (*------------------------------------*)
      and tuple(test: Exp.t, ty: Type.t, rules: flatrules, finish: finish) =
	 let
	    val rules =
	       List.map
	       (rules, fn rule as {pat, info} =>
		case pat of
		   FlatPat.Any => {pats = NONE, info = info}
		 | FlatPat.Tuple pats => {pats = SOME pats, info = info}
		 | _ => Error.error "expected tuple pattern")
	 in Exp.detuple
	    {tuple = test,
	     body = fn vars => matches(vars, rules, finish)}
	 end

      (*------------------------------------*)
      (*                sum                 *)
      (*------------------------------------*)
      and sum(test, rules: flatrules, finish: finish) =
	 let
	    datatype arg = 
	       NoArg of info list
	     | Arg of {var: Var.t, ty: Type.t, rules: rules}
	    val (cases, defaults) =
	       List.fold
	       (List.reverse rules, ([], []), fn ({pat, info}, (cases, defaults)) =>
		case pat of
		   FlatPat.Any =>
		      (List.map
		       (cases, fn {con, tys, arg} =>
			{con = con, tys = tys,
			 arg = (case arg of
				   NoArg infos => NoArg(info :: infos)
				 | Arg{var, ty, rules} =>
				      Arg{var = var,
					  ty = ty,
					  rules = {pat = NestedPat.wild ty,
						   info = info} :: rules})}),
		       info :: defaults)
		 | FlatPat.Con{con=c, targs=tys, arg} => 
		      let
			 fun insert cases =
			    case cases of
			       [] =>
				  [{con = c, tys = tys,
				    arg =
				    (case arg of
					NONE => NoArg(info :: defaults)
				      | SOME p =>
					   let val ty = NestedPat.ty p
					   in Arg{var = Var.newNoname(),
						  ty = ty,
						  rules =
						  {pat = p, info = info}
						  :: (List.map
						      (defaults, fn info =>
						       {pat = NestedPat.wild ty,
							info = info}))}
					   end)}]
			     | (cas as {con, tys, arg=a}) :: cases =>
				  if Con.equals(c, con)
				     then {con = con, tys = tys,
					   arg = (case (a, arg) of
						     (NoArg infos, NONE) =>
							NoArg(info :: infos)
						   | (Arg{var, ty, rules},
						      SOME p) =>
						     Arg{var = var,
							 ty = ty,
							 rules = 
							 {pat = p, info = info}
							 :: rules}
						   | _ => Error.bug "use of constructor with and without arg in pattern match")}
					:: cases
				  else cas :: (insert cases)
		      in (insert cases, defaults)
		      end
		 | _ => Error.error "expected constructor pat")
	       
	    val default =
	       case cases of
		  [] => SOME(finish defaults)
		| {con, ...} :: _ =>
		     let val tycon = conTycon con
		     in if Tycon.equals(tycon, Tycon.exn)
			orelse List.length cases <> List.length(tyconCons tycon)
			   then SOME(finish defaults)
			else NONE
		     end
		  
	    fun normal() =
	       Exp.casee{test = test, default = default,
			 ty = caseType,
			 cases = 
			 List.map(cases, fn {con, tys, arg} =>
				  let
				     val (arg, rhs) =
					case arg of
					   NoArg infos => (NONE, finish infos)
					 | Arg{var, ty, rules} =>
					      (SOME(var, ty),
					       match(var, ty, rules, finish))
				  in {con = con, targs = tys, arg = arg,
				      rhs = rhs}
				  end)}
	 in case cases of
	    [{con, arg = Arg{var, ty, rules}, ...}] =>
	       if Con.equals(con, Con.reff)
		  then Exp.lett{var = var,
				exp = Exp.deref test,
				body = match(var, ty, rules, finish)}
	       else normal()
	  | _ => normal()
	 end

      (*------------------------------------*)
      (*               const                *)
      (*------------------------------------*)
      and const(test: Exp.t, rules: flatrules, finish: finish) =
	 let
	    val (cases, defaults) =
	       List.fold
	       (List.reverse rules, ([], []),
		fn ({pat, info}, (cases, defaults)) =>
		case pat of
		   FlatPat.Any =>
		      (List.map(cases, fn {const, infos} =>
				{const = const, infos = info :: infos}),
		       info :: defaults)
		 | FlatPat.Const c =>
		      let
			 fun insert cases =
			    case cases of
			       [] => [{const = c, infos = info :: defaults}]
			     | (casee as {const, infos}) :: cases =>
				  if Const.equals(c, const)
				     then {const = c, infos = info :: infos}
					    :: cases
				  else casee :: insert cases
		      in (insert cases, defaults)
		      end
		 | _ => Error.error "expected Const pat")
	 in List.fold(cases, finish defaults, fn ({const, infos}, rest) =>
		      Exp.iff{test = Exp.equal(test, Exp.const const),
			      thenn = finish infos,
			      elsee = rest,
			      ty = caseType})
	 end

   (*------------------------------------*)
   (*    main code for match compile     *)
   (*------------------------------------*)
      
   in match(test, testType,
	    List.map(cases, fn (p, f) => {pat = p, info = {accum = Env.empty,
							  continue = Finish f}}),
	    fn [] => Error.bug "matchRules: no default"
	     | {accum=env, continue} :: _ =>
		  (case continue of
		      Finish f => f (fn x => Env.lookup(env, x))
		    | _ => Error.bug "matchRules: expecting Finish"))
   end

(*val matchCase =
 *  Trace.trace3("matchCase", Var.layout, Type.layout,
 *		List.layout(", ", Layout.tuple2(NestedPat.layout, Exp.layout)),
 *		Exp.layout) matchCase
 *)
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INFER_STRUCTS = 
   sig
      structure CoreML: CORE_ML
      structure Xml: XML
      sharing CoreML.Atoms = Xml.Atoms
   end

signature INFER = 
   sig
      include INFER_STRUCTS
      
      val infer: CoreML.Program.t -> Xml.Program.t
   end



(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Changes all uses of ref into primapps.
 *)

functor Infer(S: INFER_STRUCTS): INFER =
struct
      
open S

open CoreML.Atoms

structure Srecord = SortedRecord
structure Field = Record.Field
structure Scope = Scope(structure CoreML = CoreML)
structure Env = TypeEnv(open CoreML
			structure XmlType = Xml.Type)
structure Scheme = Env.InferScheme
structure Type = Env.InferType

structure PScheme = Prim.Scheme

local open Ast
in structure Aconst = Const
end
   
local open CoreML
in structure Cdec = Dec
   structure Cexp = Exp
   structure Ctype = Type
   structure Cpat = Pat
   structure Cprogram = Program
end

local open Xml
in structure Xdec = Dec
   structure Xexp = DirectExp
   structure Xtype = Type
   structure Xconst = Const
   structure Xlambda = Lambda
   structure Xpat = Pat
   structure XprimExp = PrimExp
   structure XvarExp = VarExp
end

structure Type =
   struct
      open Type

      fun fromCoreML t =
	 case t of
	    Ctype.Var a => var a
	  | Ctype.Con(c, ts) => con(c, List.map(ts, fromCoreML))
	  | Ctype.Record r => record{flexible = false,
				     record = Srecord.map(r, fromCoreML)}
   end

fun instantiate arg =
   let val {instance, args} = Scheme.instantiate arg
   in {instance = instance,
       args = fn () => List.map(args, Type.toXml)}
   end

fun instantiatePrim(PScheme.T{tyvars, ty}) =
   instantiate{scheme = Scheme.T{tyvars = tyvars, ty = Type.fromCoreML ty},
	       canGeneralize = true}

structure Frees = Type.Frees
structure VarRange = Env.VarRange

structure NestedPat = NestedPat(open Xml)
structure MatchCompile =
   MatchCompile(open CoreML
		structure Type = Xtype
		structure NestedPat = NestedPat
		structure Exp =
		   struct
		      open Xexp
		      val lett = let1
		      val var = monoVar

		      fun detuple{tuple, body} =
			 Xexp.detuple
			 {tuple = tuple,
			  body = fn xts => body(List.map(xts, fn (x, t) =>
							 (XvarExp.var x, t)))}
			 
		      fun casee{test, default, cases, ty} =
			 Xexp.casee
			 {test = test,
			  ty = ty,
			  default = default,
			  cases = List.map(cases, fn {con, targs, arg, rhs} =>
					   (Xpat.T{con = con,
						   targs = targs,
						   arg = arg},
					    rhs))}
		   end)

type patCode = (unit -> NestedPat.node) * Type.t
type expCode = (unit -> Xexp.t) * Type.t

(* decCode can't quite be of type  unit -> Xdec.t list
 * because expanding val p = e requires nesting the remaining declarations
 * inside a case expression
 *)
type decCode = Xexp.t * Xtype.t -> Xexp.t * Xtype.t

type ruleCode = patCode * expCode
type rulesCode = ruleCode list

fun finishPat((p, ty): patCode): NestedPat.t =
   NestedPat.new(p(), Type.toXml ty)
   
fun finishExp((e, ty): expCode): Xexp.t = e()
   
local open Layout
in
   fun layoutPatCode(_: patCode) = str "<patcode>"
   val layoutExpCode: expCode -> t = tuple2(fn _ => str"<exp>", Type.layout)
   fun layoutDecCode (_: decCode) = str "<dec>"
end

fun newType() = Type.new{equality = false,
			 canGeneralize = true}

val minInt31 = Int32.fromInt(valOf Int31.minInt)
val maxInt31 = Int32.fromInt(valOf Int31.maxInt)
   
fun makeXconst(c: Aconst.t, ty: Type.t): Xconst.t =
   let val tycon = Xtype.detycon(Type.toXml ty)
   in Xconst.make
      (case c of
	  Aconst.Char c => Xconst.Char c
	| Aconst.Int s =>
	     if Tycon.equals(tycon, Tycon.intInf)
		then (case SmallIntInf.fromString s of
			 NONE => Xconst.IntInf s
		       | SOME i => Xconst.SmallIntInf i)
	     else
		Xconst.Int
		(let
		    val radix =
		       if String.isPrefix{string = s, prefix = "0x"}
			  orelse String.isPrefix{string = s, prefix = "~0x"}
			  then StringCvt.HEX
		       else StringCvt.DEC
		 in case StringCvt.scanString(Pervasive.Int32.scan radix) s of
		    SOME n => if Tycon.equals(tycon, Tycon.int)
				 then n
			      else if Tycon.equals(tycon, Tycon.int31)
				      then if minInt31 <= n andalso n <= maxInt31
					      then n
					   else Error.error "int31 too big"
				   else Error.bug "strange int"
		  | NONE => Error.error "invalid integer constant"
		 end handle Overflow => Error.error "integer constant too big")
	| Aconst.Real r => Xconst.Real r
	| Aconst.String s => Xconst.String s
	| Aconst.Word w =>
	     Xconst.Word(if Tycon.equals(tycon, Tycon.word)
			    then w
			 else if Tycon.equals(tycon, Tycon.word8)
				 then if w = Word32.andb(w, 0wxFF)
					 then w
				      else Error.error "word8 too big"
			 else if Tycon.equals(tycon, Tycon.word31)
				 then if w = Word32.andb(w, 0wx7FFFFFFF)
					 then w
				      else Error.error "word31 too big"
			 else Error.bug("strange word " ^ Tycon.toString tycon)),
       tycon)
   end

fun 'a sortByField(l: (Field.t * 'a) list): (Field.t * 'a) list =
   #insertionSort
   (List.ordered{< = fn ((f, _), (f', _)) =>
		 Field.<(f, f')})
   l

(*---------------------------------------------------*)
(*                       infer                       *)
(*---------------------------------------------------*)

(* local
 *    structure Env =
 *       struct
 * 	 open Env
 * 	 val layout: t -> Layout.t = Layout.ignore
 *       end
 * in 
 *    val traceInferExp =
 *       Trace.trace2("inferExp", Cexp.layout, Env.layout, layoutExpCode)
 * 
 *    val traceInferDec =
 *       Trace.trace2("inferDec", Cdec.layout, Env.layout,
 * 		   Layout.tuple2(layoutDecCode, Env.layout))
 * 	 
 *    val traceInferPat =
 *       Trace.trace2("inferPat", Cpat.layout, Env.layout,
 * 		   Layout.tuple2(layoutPatCode, Env.layout))
 * end
 *)

fun infer(p: CoreML.Program.t): Xml.Program.t =
   let
      val {get = tyconCons: Tycon.t -> Con.t list,
	   set = setTyconCons} =
	 Property.new(Tycon.plist, Property.initRaise("tyconCons", Tycon.layout))
	 
      val {get = conInfo: Con.t -> {tycon: Tycon.t,
				    scheme: Prim.Scheme.t},
	   set = setConInfo} =
	 Property.new(Con.plist, Property.initRaise("conInfo", Con.layout))

      val conTycon = #tycon o conInfo
      val conScheme = #scheme o conInfo

(*      val conTycon = Trace.trace("conTycon", Con.layout, Tycon.layout) conTycon *)

(*       val conScheme =
 * 	 Trace.trace("conScheme", Con.layout, Prim.Scheme.layout) conScheme
 *)

      fun instCon c  = instantiatePrim(conScheme c)

      (*---------------------------------------------------*)
      (*                     inferPat                      *)
      (*---------------------------------------------------*)
      (*
       * inferPat yields
       * 1. thunk that can generate a typed nested pattern
       * 2. the type of the pattern
       * 3. a type environment extended with the variables declared in the pattern
       *)

      fun inferPat arg: patCode * Env.t =
(*	 traceInferPat *)
	 (fn (p: Cpat.t, env: Env.t) =>
	  case p of
	     Cpat.Wild => ((fn () => NestedPat.Wild, newType()), env)
	   | Cpat.Var x => let val t = newType()
			   in ((fn () => NestedPat.Var x, t),
			       Env.extendVar(env, x, Scheme.fromType t))
			   end
	   | Cpat.Const c =>
		let val ty = Type.ofConst c
		in ((fn () => NestedPat.Const(makeXconst(c, ty)), ty), env)
		end
	   | Cpat.Con{con, arg} =>
		let val {instance, args} = instCon con
		in case arg of
		   NONE => ((fn () => NestedPat.Con{con = con, targs = args(),
						    arg = NONE},
			     instance),
			    env)
		 | SOME p => 
		      let val (t1, t2) = Type.dearrow instance
			 val (p as (_, tp), env) = inferPat(p, env)
		      in Type.unify(t1, tp) ;
			 ((fn () => NestedPat.Con{con = con,
						  targs = args(),
						  arg = SOME(finishPat p)},
			   t2),
			  env)
		      end
		end
	   | Cpat.Record{flexible, record} =>
		let val (fs, ps) = Record.unzip record
		   val (ps, ts, env) = inferPats(ps, env)
		   val t = Type.record{flexible = flexible,
				       record = Srecord.zip(fs, ts)}
		in ((fn () =>
		     NestedPat.Tuple
		     (if not flexible
			 then List.map(sortByField(List.map2(fs, ps, fn (f, p) =>
							     (f, finishPat p))),
				       #2)
		      else let val {flexible, record=record'} = Type.derecord t
			   in if flexible
				 then (Control.messageLay(Cpat.layout p)
				       ; Error.error "unresolved flexible pattern")
			      else
				 let val record = Record.zip(fs, ps)
				 in List.map
				    (Srecord.toList record', fn (f, t) =>
				     case Record.peek(record, f) of
					NONE => NestedPat.new(NestedPat.Wild,
							      Type.toXml t)
				      | SOME p => finishPat p)
				 end
			   end),
			 t),
		    env)
		end
	   | Cpat.Constraint(p, t') => let val ((p, t), env) = inferPat(p, env)
				       in Type.unify(t, Type.fromCoreML t') ;
					  ((p, t), env)
				       end
	   | Cpat.Layered(x, p) => let val (p as (_, t), env) = inferPat(p, env)
				   in ((fn () => NestedPat.Layered(x, finishPat p),
					t),
				       Env.extendVar(env, x, Scheme.fromType t))
				   end) arg

      and inferPats(ps, env): patCode list * Type.t list * Env.t =
	 let
	    val (ps, ts, env) = 
	       List.fold(List.reverse ps, ([], [], env), fn (p, (ps, ts, env)) =>
			 let val (p as (_, t), env) = inferPat(p, env)
			 in (p :: ps, t :: ts, env)
			 end)
	 in (ps, ts, env)
	 end

      fun delayExp(e as (_, t): expCode): expCode =
	 (fn () => Xexp.lambda{arg = Var.newNoname(),
			       argType = Xtype.unit,
			       body = finishExp e,
			       bodyType = Type.toXml t},
	  Type.arrow(Type.unit, t))
	 
      fun letExp(d: decCode, e as (_, t): expCode): expCode =
	 (fn () => #1(d(finishExp e, Type.toXml t)), t)

      local
	 fun make(c: Con.t)(): Xexp.t =
	    Xexp.conApp{con = c, targs = [], arg = NONE, ty = Xtype.exn}
      in val match = make Con.match
	 val bind = make Con.bind
      end

      (* store all the overloads as promises in a list, so that we can
       * force them after performing the unifications.  This is necessary because
       * forcing an overload can cause further unifications to be done, and all
       * unifications must be done before converting to Xml
       *)
      val overloads: Xexp.t Promise.t list ref = ref []
	 
      fun varExp(x: Var.t, env: Env.t): expCode =
	 let val vr as VarRange.T{scheme, kind} = Env.lookupVarRange(env, x)
	 in case kind of
	    VarRange.Normal =>
	       let val {instance, args} = instantiate{scheme = scheme,
						      canGeneralize = true}
	       in (fn () => Xexp.var{var = x, targs = args(), ty = Type.toXml instance},
		   instance)
	       end
	  | VarRange.Recursive tyvars =>
	       let val ty = Scheme.ty scheme
	       in (fn () => Xexp.var{var = x,
				     targs = List.map(!tyvars, Xtype.var),
				     ty = Type.toXml ty},
		   ty)
	       end
	  | VarRange.Delayed =>
	       let val {instance, args} = instantiate{scheme = scheme,
						      canGeneralize = true}
		  val (_, t') = Type.dearrow instance
	       in (fn () =>
		   Xexp.app{func = Xexp.var{var = x,
					    targs = args(),
					    ty = Xtype.arrow(Xtype.unit,
							     Type.toXml t')},
			    arg = Xexp.unit(),
			    ty = Type.toXml t'},
		   t')
	       end
	  | VarRange.Overload yts =>
	       let
		  val {instance, args} = instantiate{scheme = scheme,
						     canGeneralize = false}
		  val promise =
		     Promise.delay
		     (fn () =>
		      let
			 val rec loop =
			    fn (y, t) :: yts =>
			    if Type.canUnify(instance, t)
			       then (Type.unify(instance, t) ;
				     Xexp.monoVar(y, Type.toXml t))
			    else loop yts
			     | [] => 
				  (Control.messageLay
				   let open Layout
				   in align[Var.layout x,
					    VarRange.layout vr]
				   end;
				   Error.error "impossible use of overloaded var")
		      in loop yts
		      end)
	       in ListRef.push(overloads, promise) ;
		  (fn () => Promise.force promise, instance)
	       end
	 end

      fun tyvarsDifference(t1, t2) =
	 List.keepAll(t1, fn a => not(List.exists(t2, fn a' => Tyvar.equals(a, a'))))

      (*------------------------------------*)
      (*               casee                *)
      (*------------------------------------*)

      fun casee{test: Xexp.t,
		testType: Xtype.t,
		caseType: Xtype.t,
		cases: (NestedPat.t * Xexp.t) list}: Xexp.t =
	 case cases of
	    (p, e) :: _ => 
	       let fun lett x = Xexp.let1{var = x, exp = test, body = e}
	       in case NestedPat.node p of
		  NestedPat.Wild => lett(Var.newNoname())
		| NestedPat.Var x => lett x
		(* for flat tuples, could do special case *)
		| _ => 
		     let
			val (decs, cases) =
			   List.unzip
			   (List.map
			    (cases, fn (p: NestedPat.t, e: Xexp.t) =>
			     let val args = NestedPat.varsAndTypes p
				val (vars, tys) = List.unzip args
				val func = Var.newNoname()
				val arg = Var.newNoname()
				val argType = Xtype.tuple tys
				val funcType = Xtype.arrow(argType, caseType)
				val dec =
				   Xdec.MonoVal
				   {var = func,
				    ty = funcType,
				    exp =
				    XprimExp.Lambda
				    (Xlambda.new
				     {arg = arg,
				      argType = argType,
				      body =
				      Xexp.toExp
				      (Xexp.detupleBind
				       {tuple = Xexp.monoVar(arg, argType),
					components = vars,
					body = e})})}
				fun finish rename =
				   Xexp.app
				   {func = Xexp.monoVar(func, funcType),
				    arg =
				    Xexp.tuple{exps = (List.map
						       (args, fn (x, t) =>
							Xexp.monoVar(rename x, t))),
					       ty = argType},
				    ty = caseType}
			     in (dec, (p, finish))
			     end))
			val testVar = Var.newNoname()
		     in Xexp.let1
			{var = testVar,
			 exp = test,
			 body = 
			 Xexp.lett
			 {decs = decs,
			  body = MatchCompile.matchCompile{conTycon = conTycon,
							   tyconCons = tyconCons,
							   test = testVar,
							   testType = testType,
							   caseType = caseType,
							   cases = cases}}}
		     end
	       end
	  | _ => Error.bug "case with no patterns"

      fun forceRules({argType, resultType, rules},
		     p: NestedPat.node,
		     e: Xexp.t): Xml.Lambda.t =
	 let val arg = Var.newNoname()
	    val argType = Type.toXml argType
	    val resultType = Type.toXml resultType
	 in Xml.Lambda.new
	    {arg = arg, argType = argType,
	     body = (Xexp.toExp
		     (casee{testType = argType,
			    test = Xexp.monoVar(arg, argType),
			    cases = (List.map(rules, fn (p, e) =>
					      (finishPat p, finishExp e))
				     @ [(NestedPat.new(p, argType),
					 Xexp.raisee(e, resultType))]),
			    caseType = resultType}))}
	 end

      fun forceRulesMatch rs =
	 forceRules(rs, NestedPat.Wild, match())

      fun forceRulesReraise rs =
	 let val x = Var.newNoname()
	 in forceRules(rs, NestedPat.Var x, Xexp.monoVar(x, Xtype.exn))
	 end

      val emptyDec: decCode = fn e => e

      fun appendDec(d: decCode, d': decCode): decCode = d o d'

      fun cons(thunk: unit -> Xdec.t list): decCode = 
	 fn (e, t) => (Xexp.lett{decs = thunk(), body = e}, t)

      fun valDec(tyvars: Tyvar.t list, var: Var.t, exp as (_, t): expCode,
		 kind: VarRange.kind, env: Env.t): decCode * Env.t =
	 (cons(fn () => [Xdec.PolyVal{tyvars = tyvars,
				      var = var,
				      ty = Type.toXml t,
				      exp = Xexp.toExp(finishExp exp)}]),
	  Env.extendVarRange(env, var,
			     VarRange.T
			     {scheme = Scheme.T{tyvars = tyvars, ty = t},
			      kind = kind}))

      fun patDec(p as (_, tp): patCode, e as (_, te): expCode): decCode =
	 (Type.unify(tp, te) ;
	  fn (body, ty) =>
	  let val p = finishPat p
	  in (casee{test = finishExp e,
		    testType = Type.toXml te,
		    caseType = ty,
		    cases = [(p, body),
			     (NestedPat.new(NestedPat.Wild, NestedPat.ty p),
			      Xexp.raisee(bind(), ty))]},
	      ty)
	  end)

      fun checkTyvars(tyvars, tyvarsClose, frees) =
	 case tyvarsDifference(tyvars, tyvarsClose) of
	    _ :: _ =>
	       if Frees.areDisjoint(Frees.fromTyvars tyvars, frees())
		  then ()
	       else Error.error "can't generalize tyvars"
	  | _ => ()

      fun inferValDec(tyvars: Tyvar.t list,
		      pat: Cpat.t,
		      exp: Cexp.t,
		      e: expCode as (_, te),
		      env: Env.t): decCode * Env.t =
	 let val (p as (_, tp), env') = inferPat(pat, env)
	 in Type.unify(te, tp) ;
	    if Cexp.isExpansive exp
	       then (case tyvars of
			[] => (patDec(p, e), env')
		      | _ => (Layout.output(CoreML.Exp.layout exp, Out.error) ;
			      Out.newline Out.error ;
			      Error.error "can't generalize an expansive exp"))
	    else
	       let val freesPromise = Promise.delay(fn () => Env.frees env)
		  fun frees() = Promise.force freesPromise
		  val tyvarsClose = Type.close(te, frees)
		  val _ = checkTyvars(tyvars, tyvarsClose, frees)
		  fun vd x = valDec(tyvarsClose, x, e, VarRange.Normal, env)
	       in case (tyvarsClose, pat) of
		  ([], _) => (patDec(p, e), env')
		| (_, Cpat.Wild) => vd(Var.newNoname())
		| (_, Cpat.Var x) => vd x
		| (_, Cpat.Constraint(Cpat.Var x, _)) => vd x
		| _ =>
		     let
			val _ =
			   Control.messageLay
			   (let open Layout
			    in align[str "expansion of polymorphic pattern",
				     Cdec.layout
				     (Cdec.Val{pat = pat, exp = exp,
					       tyvars = tyvars})]
			    end)
			val x = Var.newNoname()
			val (d, env) = vd x
			val d =
			   if Cpat.isRefutable pat
			      then appendDec
				 (d, patDec(#1(inferPat(Cpat.removeVars pat, env)),
					    varExp(x, env)))
			   else d
		     in List.fold
			(Cpat.vars pat, (d, env), fn (y, (d, env)) =>
			 let val y' = Var.new y
			    val p = Cpat.removeOthersReplace(pat, y, y')
			    val (p, env') = inferPat(p, env)
			    val e as (_, te) =
			       delayExp(letExp(patDec(p, varExp(x, env)),
					       varExp(y', env')))
			    val (d', env) =
			       valDec(Type.close(te, fn () =>
						 Env.frees env),
				      y, e, VarRange.Delayed, env)
			 in (appendDec(d, d'), env)
			 end)
		     end
	       end
	 end
(* accumulate the datatype declarations *)
      val dbsRef = ref []

      fun processConArg(tycon, tyvars, {con, arg}) =
	 let val result = Ctype.con(tycon, List.map(tyvars, Ctype.var))
	    val (ty, arg) =
	       case arg of
		  NONE => (result, NONE)
		| SOME arg => (Ctype.arrow(arg, result),
			       SOME(Type.toXml(Type.fromCoreML arg)))
	 in setConInfo(con, {tycon = tycon,
			     scheme = PScheme.T{tyvars = tyvars, ty = ty}});
	    {con = con, arg = arg}
	 end

      fun processException ca = processConArg(Tycon.exn, [], ca)
   
      fun processDatatypes(dbs, f): unit =
	 List.foreach
	 (dbs, fn {tyvars, tycon, cons} =>
	  let val cons = List.map(cons, fn ca => processConArg(tycon, tyvars, ca))
	  in setTyconCons(tycon, List.map(cons, #con));
	     f{tyvars = tyvars,
	       tycon = tycon,
	       cons = cons}
	  end)

      (*------------------------------------*)
      (*              inferDec              *)
      (*------------------------------------*)
      fun inferDec arg: decCode * Env.t =
(*	 traceInferDec *)
	 (fn (d: Cdec.t, env: Env.t) =>
	  case d of
	     Cdec.Val{tyvars, pat, exp} =>
		inferValDec(tyvars, pat, exp, inferExp(exp, env), env)
	   | Cdec.Datatype dbs =>
		(processDatatypes(dbs, fn db as {tycon, ...} =>
				  if Tycon.equals(tycon, Tycon.reff)
				     then ()
				  else ListRef.push(dbsRef, db));
		 (emptyDec, env))
	   | Cdec.Exception(e as {con, ...}) =>
		let val ca = processException e
		in (cons(fn () => [Xdec.Exception ca]),
		    env)
		end
	   | Cdec.Overload{var, scheme=CoreML.Scheme.T{tyvars, ty}, ovlds} =>
		(emptyDec,
		 let val ty = Type.fromCoreML ty
		 in Env.extendVarRange
		    (env, var,
		     VarRange.T{scheme = Scheme.T{tyvars = tyvars, ty = ty},
				kind =
				VarRange.Overload
				(List.map(ovlds, fn y =>
					  (y, Scheme.ty(Env.lookupVar(env, y)))))})
		 end)
	   | Cdec.Fun{tyvars, decs} =>
		let
		   (* type args to recursive calls *)
		   val args: Tyvar.t list ref = ref []
		   val (env', decs) =
		      List.fold
		      (decs, (env, []), fn ({var, ty, rules}, (env, decs)) =>
		       let val argType = newType()
			  val resultType = newType()
			  val t = Type.arrow(argType, resultType)
		       in (case ty of
			      NONE => ()
			    | SOME t' => Type.unify(t, Type.fromCoreML t')) ;
			  (Env.extendVarRange
			   (env, var, VarRange.T{scheme = Scheme.fromType t,
						 kind = VarRange.Recursive args}),
			   {var = var,
			    argType = argType,
			    resultType = resultType,
			    rules = rules} :: decs)
		       end)
		   val decs =
		      List.map(decs, fn {var, rules, argType, resultType} =>
			       let
				  val rs = inferRulesUnify(rules, env',
							   argType, resultType)
			       in {var = var,
				   ty = Type.arrow(argType, resultType),
				   rules = rs}
			       end)
		   val freesPromise = Promise.delay(fn () => Env.frees env)
		   fun frees() = Promise.force freesPromise
		   val tyvarsClose =
		      Tyvars.toList
		      (List.fold(decs, Tyvars.empty, fn ({ty, ...}, tyvars) =>
				 Tyvars.+(tyvars,
					  Tyvars.fromList(Type.close(ty, frees)))))
		   val _ = args := tyvarsClose
		   val env =
		      List.fold(decs, env, fn ({var, ty, ...}, env) =>
				Env.extendVar(env, var,
					      Scheme.T{tyvars = tyvarsClose,
						       ty = ty}))
		in checkTyvars(tyvars, tyvarsClose, frees) ;
		   (cons(fn () =>
			 [Xdec.Fun
			  {tyvars = tyvarsClose,
			   decs =
			   List.map(decs, fn {var, ty, rules} =>
				    {var = var,
				     ty = Type.toXml ty,
				     lambda = forceRulesMatch rules})}]),
		    env)
		end) arg

      and inferDecs(ds: Cdec.t list, env: Env.t): decCode * Env.t =
	 List.fold(ds, (emptyDec, env), fn (d, (d', env)) =>
		    let val (d, env) = inferDec(d, env)
		    in (appendDec(d', d), env)
		    end)

      (*------------------------------------*)
      (*              inferExp              *)
      (*------------------------------------*)

      and inferExps(es, env) =
	 List.map(es, fn e => inferExp(e, env))

      and inferExp arg: expCode =
(*	 traceInferExp *)
	 (fn (e, env) => 
	 case e of
	    Cexp.Var x => varExp(x, env)
	  | Cexp.Prim _ => apply(e, env, NONE)
	  | Cexp.Con _ => apply(e, env, NONE)
	  | Cexp.App(e1, e2) => apply(e1, env, SOME(inferExp(e2, env)))
	  | Cexp.Const c => let val ty = Type.ofConst c
			    in (fn () => Xexp.const(makeXconst(c, ty)),
				ty)
			    end
	  | Cexp.Record r =>
	       (* This code is messy because the components of the record have
		* to be evaluated left to right as they appeared in the source
		* program, but then ordered according to sorted field name within
		* the tuple
		*)
	       let
		  val fes = Record.toList r
		  val es = inferExps(List.map(fes, #2), env)
		  val ty = Type.record{flexible = false,
				       record =
				       Srecord.fromList
				       (List.map2(fes, es,
						  fn ((f, _), (_, t)) => (f, t)))}
	       in (fn () =>
		   Xexp.seq
		   (List.map(es, finishExp), fn es =>
		    Xexp.tuple{exps = (List.map
				       (sortByField
					(List.map2(fes, es, fn ((f, _), e) =>
						   (f, e))),
					#2)),
			       ty = Type.toXml ty}),
		   ty)
	       end
	  | Cexp.Fn rules =>
	       let val rs as {argType, resultType, rules} = inferRules(rules, env)
	       in (fn () => let
			       val {arg, argType, body} =
				  Xlambda.dest(forceRulesMatch rs)
			       val resultType = Type.toXml resultType
			    in Xexp.lambda
			       {arg = arg,
				argType = argType,
				body = Xexp.fromExp(body, resultType),
				bodyType = resultType}
			    end,
		   Type.arrow(argType, resultType))
	       end
	  | Cexp.Let(ds, e) =>
	       let val (ds, env) = inferDecs(ds, env)
	       in letExp(ds, inferExp(e, env))
	       end
	  | Cexp.Constraint(e, t') =>
	       let val ans as (_, t) = inferExp(e, env)
	       in Type.unify(t, Type.fromCoreML t') ;
		  ans
	       end
	  | Cexp.Handle(e, rs) =>
	       let val (e, t) = inferExp(e, env)
		  val rs as {argType, resultType, rules} = inferRules(rs, env)
	       in Type.unify(t, resultType) ;
		  Type.unify(argType, Type.exn) ;
		  (fn () => let
			       val {arg, body, ...} =
				  Xlambda.dest(forceRulesReraise rs)
			       val t' = Type.toXml t
			    in Xexp.handlee{try = finishExp(e, t),
					    catch = (arg, Type.toXml Type.exn),
					    handler = Xexp.fromExp(body, t'),
					    ty = t'}
			    end,
		   t)
	       end
	  | Cexp.Raise e =>
	       let val e as (_, t) = inferExp(e, env)
		  val resultType = newType()
	       in Type.unify(t, Type.exn) ;
		  (fn () => Xexp.raisee(finishExp e,
					Type.toXml resultType),
		   resultType)
	       end) arg

      and applyOne(e as (_, t): expCode, e' as (_, t'): expCode): expCode =
	 let val resultType = newType()
	 in Type.unify(t, Type.arrow(t', resultType)) ;
	    (fn () => Xexp.app{func = finishExp e,
			       arg = finishExp e',
			       ty = Type.toXml resultType},
	     resultType)
	 end

      and apply(e1: Cexp.t, env: Env.t, arg: expCode option): expCode =
	 let
	    fun eta(ty: Type.t,
		    make: (Xexp.t * Xtype.t) option * Xtype.t -> Xexp.t)
	       : expCode =
	       case (Type.dearrowOpt ty, arg)                 of
		    (NONE,               NONE)                =>
		       (fn () => make(NONE, Type.toXml ty), ty)
		  | (NONE,               SOME _)              =>
		       Error.error "application mismatch"
		  | (SOME(t1, t2),       SOME(argExp as (_, argTy))) =>
		       (Type.unify(t1, argTy) ;
			(fn () => make(SOME(finishExp argExp, Type.toXml t1),
				       Type.toXml t2),
			 t2))
		  | (SOME(t1, t2),       NONE)                =>
		       let val x = Var.newNoname()
		       in (fn () =>
			   let val t1 = Type.toXml t1
			      val t2 = Type.toXml t2
			   in Xexp.lambda
			      {arg = x,
			       argType = t1,
			       body = make(SOME(Xexp.monoVar(x, t1), t1), t2),
			       bodyType = t2}
			   end,
			   ty)
		       end
	 in
	    case e1 of
	       Cexp.Con con =>
		  let val {instance, args} = instCon con
		  in eta(instance, fn (arg, resultType) =>
			 if Con.equals(con, Con.reff)
			    then (case arg of
				     NONE => Error.bug "bad ref"
				   | SOME(arg, _) =>
					Xexp.primApp{prim = Prim.reff,
						     targs = args(),
						     args = [arg],
						     ty = resultType})
			 else Xexp.conApp{con = con,
					  targs = args(),
					  arg = (case arg of
						    NONE => NONE
						  | SOME(x, _) => SOME x),
					  ty = resultType})
		  end
	     | Cexp.Prim prim =>
		  let
		     val {instance, args = targs} =
			instantiatePrim(Prim.scheme prim)
		  in eta(instance, fn (arg, resultType) =>
			 let
			    fun make(args: Xexp.t list): Xexp.t =
			       Xexp.primApp{prim = prim,
					    targs = targs(),
					    args = args,
					    ty = resultType}
			 in
			    case (Prim.numArgs prim, arg) of
			         (NONE,              NONE)     => make[]
			       | (SOME n,            SOME(arg, argType)) =>
				    (case n of
					0 => make[]
				      | 1 => make[arg]
				      | _ => 
					   Xexp.detuple
					   {tuple = arg,
					    body =
					    fn components =>
					    make(List.map(components, Xexp.varExp))})
			       | _                             =>
				    Error.error "primApp mismatch"
			 end)
		  end
	     | Cexp.App(e1, e2) =>
		  let val e = apply(e1, env, SOME(inferExp(e2, env)))
		  in case arg of
		     NONE => e
		   | SOME e' => applyOne(e, e')
		  end
	     | _ => let val e1 = inferExp(e1, env)
		    in case arg of
		       NONE => e1
		     | SOME e2 => applyOne(e1, e2)
		    end
	 end
      and inferRules(rs, env): {argType: Type.t, resultType: Type.t,
				rules: (patCode * expCode) list} =
	 inferRulesUnify(rs, env, newType(), newType())
   
      and inferRulesUnify(rs, env, argType, resultType) =
	 {argType = argType,
	  resultType = resultType,
	  rules = List.map(rs, fn (p, e) =>
			   let val (p as (_, tp), env) = inferPat(p, env)
			      val e as (_, te)  = inferExp(e, env)
			   in Type.unify(tp, argType) ;
			      Type.unify(te, resultType) ;
			      (p, e)
			   end)}

      (*------------------------------------*)
      (*    main code for type inference    *)
      (*------------------------------------*)

      val Cprogram.T{decs,
		     constructors =
		     cs as CoreML.Constructors.T{datatypes, exceptions}}
	 = Scope.scope p

      val _ = (processDatatypes(datatypes, fn _ => ());
	       List.foreach(exceptions, fn ca => (processException ca; ())))
	    
      val (ds, env) = Control.traceCall "unification" inferDecs(decs, Env.empty)

      val _ = (List.foreach(!overloads, fn p => (Promise.force p ; ())) ;
	       overloads := [])
	 
      val (body, _) =
	 Control.traceCall "finish infer" ds (Xexp.unit(), Xtype.unit)

      val xml = Xml.Program.T{datatypes = !dbsRef,
			      body = Xexp.toExp body}

   in Xml.Program.clear xml;
      xml
   end

(*val infer = Trace.trace("infer", Layout.ignore, Xml.Program.layout) infer *)

structure Scheme = PScheme
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRIM_ENV_STRUCTS = 
   sig
      include ELABORATE

      val preprocess: In.t * Basis.t * Accum.t -> Basis.t * Accum.t
   end

signature PRIM_ENV = 
   sig
      include PRIM_ENV_STRUCTS

      (* all primEnv's are delayed *)
      val defunctorize: unit -> Accum.t * CoreML.Constructors.t
(*      val monomorphise: unit -> Accum.t * CoreML.Constructors.t *)
      val compile: unit -> Accum.t * Decs.t * CoreML.Constructors.t

      (* ML code that must be prefixed to any monomorphised program.  This
       * is necessary because the monomorphic internal language (Sxml) does not
       * deal with record types, but some of the basis library functions do.
       * This code mediates between the record types used in the basis library
       * and the tuple types used in the monomorphised output
       *)
(*      val getPatchRecordTypes: unit -> Layout.t *)

      (* force all primEnv's to be computed now *)
      val forceAll: Dir.t -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PrimEnv(S: PRIM_ENV_STRUCTS): PRIM_ENV = 
struct

open S

local open CoreML
in structure Constructors = Constructors
end
structure Env = Basis.Env

fun preprocessFile(f: File.t, B, A: Accum.t): Basis.t * Accum.t =
   (Control.inputFile := f;
    File.withIn(f, fn ins => preprocess(ins, B, A)))

fun preprocessFiles(fs: File.t list, B, A: Accum.t): Basis.t * Accum.t =
   List.fold(fs, (B, A), fn (f, (B, A)) => preprocessFile(f, B, A))

val root = ref(Dir.fromString "..")
   
fun lazy(name: string, thunk: unit -> 'a): unit -> 'a =
   Promise.lazy
   (fn () =>
    let val a = thunk()
    in Control.message(fn () => Control.sizeMessage(name, a));
       a
    end)

fun basisFile f =
   File.relative(Dir.relative(!root, "basis-library"), f)

fun basis() =
   Ref.fluidLet
   (Control.aux, false, fn () =>
    let
       fun files(f, B, A) =
	  preprocessFiles(List.map(List.keepAll(File.lines(basisFile f),
						fn s =>
						(s <> ""
						 andalso
						 #"#" <> String.sub(s, 0))),
				   basisFile),
			  B, A)
       val (B, cs) = Basis.prim()
       val (_, A) = files("build-basis", B, Accum.prim())
       val (_, A) =
	  files("bind-basis", Accum.basis A,
		Accum.T{decs = Accum.decs A,
			basis = Basis.fromEnv(Env.addEquals Env.empty)})
    in (A, cs)
    end)

(* val patchRecordTypes: Layout.t option ref = ref NONE
 * 
 * fun getPatchRecordTypes() =
 *    case !patchRecordTypes of
 *       NONE => Error.bug "getPatchRecordTypes"
 *     | SOME l => l
 *)

val defunctorize =
   lazy
   ("defunctorize", fn () =>
    let
       val (A, cs) =  Ref.fluidLet(useOpaque, true, basis)
       val B = Accum.basis A
	val (E, P, cs') = Env.identify(Basis.toEnv B)
(*	val _ = patchRecordTypes := SOME(Ast.Program.layout P) *)
	(* Uncomment the following to see the basis *)
(* 	val _ = File.withOut(File.fromString "/tmp/z",
 * 			     fn out => Layout.output(Env.layout E, out))
 *)
	val B = Basis.replaceEnv(B, E)
     in (Accum.T{decs = Decs.empty, basis = B},
	 Constructors.+(cs, cs'))
     end)

(* val monomorphise =
 *    lazy
 *    ("monomorphise",
 *     fn () =>
 *     let val (A, cs) = defunctorize()
 *     in (#2(preprocessFile(basisFile "top-level/overloads.sml", Accum.basis A, A)),
 * 	cs)
 *     end)
 *)

val compile =
   lazy
   ("compile env",
    fn () =>
    let val (A, cs) = basis()
       (* concatenated onto the end of the program for cleanup *)
       val (_, suffix) =
	  preprocessFiles([basisFile "misc/suffix.sml"],
			  Accum.basis A,
			  Accum.empty)
    in (A, Accum.decs suffix, cs)
    end)

fun forceAll d = (root := d
		  ; let open Control
		    in
		       Control.setMode Defunctorize; defunctorize()
(*		       ; Control.setMode Monomorphise; monomorphise() *)
		       ; Control.setMode Compile; compile()
		    end
		  ; ())
			       
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature COMPILE =
   sig
      structure PrimEnv: PRIM_ENV
      val compileFile: {input: string,
			output: string} -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Compile: COMPILE =
struct

(*---------------------------------------------------*)
(*              Intermediate Languages               *)
(*---------------------------------------------------*)

structure Ast = Ast()
structure Atoms = Atoms(structure Ast = Ast)
structure CoreML = CoreML(open Atoms
			  structure Type = Prim.Type)
structure Xml = Xml(open Atoms)
structure Sxml = Xml
structure Cps = Cps(open Atoms)
structure Machine = Machine(structure Label =
			       struct
				  open Cps.Func
				  fun newNoname() = newString "L"
			       end
			    structure Prim = Atoms.Prim
			    structure SmallIntInf = Atoms.SmallIntInf)

(*---------------------------------------------------*)
(*                  Compiler Passes                  *)
(*---------------------------------------------------*)

structure FrontEnd = FrontEnd(structure Ast = Ast)
structure DeadCode = DeadCode(structure CoreML = CoreML)
structure Elaborate = Elaborate(structure Ast = Ast
				structure CoreML = CoreML)
structure Infer = Infer(structure CoreML = CoreML
			structure Xml = Xml)
(*structure CallCount = CallCount(open Xml) *)
structure Monomorphise = Monomorphise(structure Xml = Xml
				      structure Sxml = Sxml)
structure ImplementExceptions = ImplementExceptions(open Sxml)
structure Polyvariance = Polyvariance(open Sxml)
structure ClosureConvert = ClosureConvert(structure Cps = Cps
					  structure Sxml = Sxml)
structure Backend = Backend(structure Cps = Cps
			    structure Machine = Machine
			    fun funcToLabel f = f
			    fun jumpToLabel j = j)

local open CoreML
in structure Constructors = Constructors
end

local open Elaborate
in structure Accum = Accum
   structure Decs = Decs
   structure Basis = Basis
   structure Env = Basis.Env
end

(* The first declaration declares the boolean variable "safe",
 * which we need to set according to Control.safe.
 * This is used to implement the -u option.
 * This is only used when Control.getMode() = Compile.
 *)

fun patchDecsSafe(ds: Decs.t): CoreML.Dec.t list =
   let fun error() = Error.bug "declarations don't start with \"val safe = true\""
   in case Decs.toList ds of
      d :: ds =>
	 let open CoreML
	 in
	    case d of
	       Dec.Val{pat = pat as Pat.Var x, tyvars = [], ...} =>
		  if Var.originalName x = "safe"
		     then
			Dec.Val{pat = pat, tyvars = [],
				exp = Exp.Con(if !Control.safe
						 then Con.truee
					      else Con.falsee)}
			:: ds
		  else error()
	     | _ => error()
	 end
    | _ => error()
   end

fun preprocess(ins: In.t, B: Basis.t, accum: Accum.t): Basis.t * Accum.t =
   let val _ = Control.reset()
      val ast =
	 Ref.fluidLet
	 (Ast.isInfix, fn _ => false,
	  fn () => Control.pass{name = "parse",
				suffix = "ast",
				style = Control.ML,
				display = Control.Layout Ast.Program.layout,
				thunk = fn () => FrontEnd.lexAndParse ins})
   in Control.message(fn () =>
		      let open Layout
		      in seq[str "size = ",
			     Int.layout(Ast.Program.size ast)]
		      end)
      ; (Control.pass
	 {name = "elaborate",
	  suffix = "core-ml",
	  style = Control.ML,
	  thunk = fn () => Elaborate.elaborateProgram(ast, B, accum),
	  display =
	  Control.Layout
	  (fn (_, A) =>
	   CoreML.Program.layout
	   (CoreML.Program.T{constructors = Constructors.empty,
			     decs = Decs.toList(Accum.decs A)}))})
   end

structure PrimEnv = PrimEnv(open Elaborate
			    val preprocess = preprocess)

(* raised to stop the compilation at some intermediate pass *)   
exception Done

fun process(ins: In.t, suffix: Decs.t, A: Accum.t, constructors,
	    out: Out.t): unit =
   let
      val decs = Accum.decs A
      fun partial() =
	 CoreML.Program.T
	 {decs = Decs.toList(Accum.decs(#2(preprocess(ins, Accum.basis A, A)))),
	  constructors = constructors}
      val coreml =
	 case Control.getMode() of
	    Control.Defunctorize => partial()
(*	  | Control.Monomorphise => partial() *)
	  | Control.Compile =>
	       let val basis = patchDecsSafe decs
		  val user =
		     Decs.toList
		     (Decs.append(Accum.decs
				  (#2(preprocess(ins, Accum.basis A, Accum.empty))),
				  suffix))
		  val basis = DeadCode.deadCode{basis = basis, user = user}
	       in CoreML.Program.T
		  {constructors = Constructors.empty,
		   decs = Constructors.toDecs constructors @ basis @ user}
	       end
      val _ = (Control.message(fn () => CoreML.Program.layoutStats coreml);
	       (case Control.getMode() of
		   Control.Defunctorize =>
		      (Layout.output(Layout.align
				     [Control.header Control.ML,
				      CoreML.Program.layout coreml],
				     out);
		       raise Done)
		 | _ => ()))
      val xml =
	 Control.passSimplify{name = "infer",
			      suffix = "xml",
			      style = Control.ML,
			      thunk = fn () => Infer.infer coreml,
			      display = Control.Layout Xml.Program.layout,
			      typeCheck = Xml.typeCheck,
			      simplify = Xml.simplify}
      val _ = Control.message(fn () => Xml.Program.layoutStats xml)
      val sxml =
	 Control.passSimplify{name = "mono",
			      suffix = "",
			      style = Control.ML,
			      thunk = fn () => Monomorphise.monomorphise xml,
			      display = Control.NoDisplay,
			      typeCheck = Sxml.typeCheck,
			      simplify = Sxml.simplify}
      val Sxml.Program.T{datatypes, ...} = sxml
   in
      Control.message(fn () => Sxml.Program.layoutStats sxml);
      case Control.getMode() of
	 Control.Defunctorize => Error.bug "shouldn't be here"
(*        | Control.Monomorphise =>
 * 	    (let open Layout
 * 	     in output(align
 * 		       [Control.header Control.ML,
 * 			str "local",
 * 			PrimEnv.getPatchRecordTypes(),
 * 			str "in val _ = ",
 * 			Sxml.Program.layout sxml,
 * 			str "end"], out)
 * 	     end;
 * 	     raise Done)
 *)
       | _ =>
	    let
	       val sxml =
		  Control.passSimplify
		  {name = "implement exceptions",
		   suffix = "sxml",
		   style = Control.ML,
		   thunk = fn () => ImplementExceptions.doit sxml,
		   typeCheck = Sxml.typeCheck,
		   display = Control.Layout Sxml.Program.layout,
		   simplify = Sxml.simplify}
	       val sxml =
		  Control.pass (* polyvariance has simplify built in *)
		  {name = "polyvariance",
		   suffix = "sxml.poly",
		   style = Control.ML,
		   thunk = fn () => Polyvariance.duplicate sxml,
		   display = Control.Layout Sxml.Program.layout}
 	       val _ = Control.message(fn () => Sxml.Program.layoutStats sxml)
	       val _ = Cps.Type.initialize()
	       val cps =
		  Control.passSimplify
		  {name = "closure convert",
		   suffix = "cps",
		   style = Control.No,
		   thunk = fn () => ClosureConvert.closureConvert sxml,
		   typeCheck = Cps.typeCheck,
		   simplify = Cps.simplify,
		   display = Control.Layouts Cps.Program.layouts}

	       val _ =
		  let open Control
		  in if !keepCps andalso not(!aux)
		     then Ref.fluidLet
			(aux, true, fn () =>
			 displays("cps", fn disp =>
				  Cps.Program.layouts(cps, disp)))
		     else ()
		  end

(* 	       val _ =
 * 		  Ref.fluidLet
 * 		  (Control.aux, true, fn () =>
 * 		   Control.displays("cps", fn disp =>
 * 				    Cps.Program.layouts(cps, disp)))
 *)

	       val machine =
		  Control.pass
		  {name = "backend",
		   suffix = "c",
		   style = Control.C,
		   display = Control.NoDisplay,
		   thunk = fn () => Backend.generate cps}
		  
	       val _ =
		  Control.traceCall
		  "emit C"
		  (fn () =>
		   (Layout.output(Control.header Control.C, out)
		    ; Out.newline out
		    ; Machine.Program.output{program = machine,
					     includes = !Control.includes,
					     print = fn s => Out.output(out, s)}
		    ))
		  ()
	    in Machine.reset()
	       ; Atoms.clearPrimitive()
	    end
   end

(*---------------------------------------------------*)
(*                    compileFile                    *)
(*---------------------------------------------------*)

fun compileFile{input: string, output: string}: unit =
   let val inputFile = File.fromString input
      fun partial accum =
	 let val (A, cs) = accum()
	 in Ast.isInfix := Env.toIsInfix(Basis.toEnv(Accum.basis A));
	    (A, Decs.empty, cs)
	 end
      (* need to do this before the Control.inputFile := *)
      val (accum, suffix, constructors) =
	 case Control.getMode() of
	    Control.Defunctorize => partial PrimEnv.defunctorize
(*	  | Control.Monomorphise => partial PrimEnv.monomorphise *)
	  | Control.Compile => PrimEnv.compile()
   in Control.inputFile := inputFile
      ; Control.messageLay(Control.header Control.No)
      ; (File.withIn
	 (inputFile, fn ins =>
	  (File.withOut
	   (File.fromString output, fn out =>
	    (process(ins, suffix, accum, constructors, out);
	     Control.message PropertyList.stats)
	    handle Done => ()))))
   end

val compileFile = Control.traceCall "compile" compileFile
   
end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature MAIN =
   sig
      val exportNJ: Dir.t * File.t -> unit
      val exportMLton: unit -> unit
   end
(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Main: MAIN =
struct

structure PrimEnv = Compile.PrimEnv

fun message s = Out.output(Out.error, s)
   
exception Usage
fun 'a usage(s: string): 'a =
   (message s
    ; message "\nUsage: mlton [-cCdgmpSv] [-o file] [-k{a|c|s|o}*] [-O{0-9}]\n\
                \\t[-DDEFINE] [-iinclude] [-Idir] [-llibrary] [-Ldir]\n\
                \\t[-flatten n] [-inline n] [-no-polyvariance]\n\
		\\t[-unsafe] [-no-type-check]\n\
		\\t[-h heapSize[{k|m}]] [-s stackSize[{k|m}]]\n\
		\\tfile.{sml|c|s|o}\n"
    ; raise Usage)

structure Place =
   struct
      datatype t = SML | CoreML | C | S | O | OUT

      val toInt =
	 fn SML => 1
	  | CoreML => 2
	  | C => 3
	  | S => 4
	  | O => 5
	  | OUT => 6

      val toString =
	 fn SML => "sml"
	  | CoreML => "core-ml"
	  | C => "c"
	  | S => "s"
	  | O => "o"
	  | OUT => "out"

      val layout = Layout.str o toString

      fun compare(p, p') = Int.compare(toInt p, toInt p')

      fun start(s: string): t * string =
	 let
	    val rec loop =
	       fn [] => usage "invalid file suffix"
		| (suf, start) :: sufs =>
		     if String.isSuffix{string = s, suffix = suf}
			then (start, String.dropSuffix(s, size suf))
		     else loop sufs
	 in loop[(".sml", SML), (".c", C), (".s", S), (".o", O)]
	 end
   end

fun system(ss: string list): unit =
   let val s = concat(List.separate(ss, " "))
   in Control.messageStr s
      ; (case OS.Process.system s of
	    0 => ()
	  | _ => Error.error("call to system failed: " ^ s))
   end

fun commandLine(args: string list): unit =
   let
      fun error() = Error.bug "incorrect args from shell script"
      val (root, gcc, gccSwitches, args) =
	 case args of
	    root :: gcc :: args =>
	       let
		  fun loop(args, ac) =
		     case args of
			[] => error()
		      | arg :: args =>
			   if arg = "END"
			      then (rev ac, args)
			   else loop(args, arg :: ac)
		  val (gccSwitches, args) = loop(args, [])
	       in (Dir.fromString root, gcc, gccSwitches, args)
	       end
	  | _ => error()

      val debug = ref false
      val defines: string list ref = ref []
      val keepC = ref false
      val keepS = ref false
      val keepO = ref false
      val includes = ref []
      val includeDirs = ref []
      val lib = ref "mlton"
      val libs = ref []
      val libDirs = ref []
      val output = ref NONE
      val optimization = ref 1
      val stop = ref Place.OUT


      fun dooKeeps s =
	 Int.foreach(0, size s - 1, fn i =>
		     (case String.sub(s, i) of
			 #"a" => Control.aux
		       | #"c" => keepC
		       | #"s" => keepS
		       | #"o" => keepO
		       | _ => usage "invalid -k flag") := true)

      val rest =
	 let open Popt
	    fun push(r) = String(fn s => ListRef.push(r, s))
	 in parse{switches = args,
		  opts =
		  [
		   ("c", None(fn () => stop := Place.O)),
		   ("C", None(fn () => stop := Place.C)),
		   ("d", None(fn () => (Control.setMode Control.Defunctorize
					; stop := Place.CoreML))),
		   ("D", String(fn s => (ListRef.push(defines, s)
					 ; if s = "INSTRUMENT"
					      then Control.instrument := true
					   else ()))),
		   ("flatten", Int(fn n => Control.flattenRounds := n)),
		   ("g", None(fn () => (debug := true; lib := "mlton-gdb"))),
		   ("h", Mem(fn n => Control.fixedHeap := SOME n)),
		   ("i", push includes),
		   ("inline", Int Control.setInlineSize),
		   ("I", push includeDirs),
		   ("k", String dooKeeps),
		   ("l", push libs),
		   ("L", push libDirs),
		   ("no-polyvariance", None(fn () =>
					    Control.polyvariance := NONE)),
		   ("no-type-check", None(fn () => Control.setTypeCheck false)),
		   ("o", SpaceString(fn s => output := SOME(File.fromString s))),
		   ("O", Digit(fn i => optimization := i)),
		   ("p", None(fn () => (Control.profile := true
					; keepC := true
					; Control.keepCps := true))),
		   ("S", None(fn () => stop := Place.S)),
		   ("unsafe", None(fn () => Control.safe := false)),
		   ("v", None Control.messagesOn),
		   ("trace", String(fn s => let open Trace.Immediate
					    in flagged()
					       ; on[s]
					    end))
		   ]}
	 end

     in
	Control.includes := List.reverse(!includes)
	;
	case rest of
	   NONE => usage "error while parsing switches"
	 | SOME [] =>
	      let open Control
	      in if messagesAreOn()
		    then messageLay(header No)
		 else usage "must supply a file"
	      end
	 | SOME(input :: rest) =>
	      let val (start, base) = Place.start input
		 val stop = !stop
	      in case Place.compare(start, stop) of
		 GREATER => usage(concat["cannot go from ",
					 Place.toString start,
					 " to ",
					 Place.toString stop])
	       | EQUAL => usage "nothing to do"
	       | LESS =>
		    let
		       val tempFiles: File.t list ref = ref []
		       fun temp(suf: string): File.t =
			  let val f = File.fromString(OS.FileSys.tmpName() ^ suf)
			  in ListRef.push(tempFiles, f)
			     ; f
			  end
		       val input = File.fromString input
		       fun suffix s = File.fromString(base ^ s)
		       fun file(b, suf) =
			  if b then suffix suf
			  else temp suf
		       fun maybeOut suf =
			  case !output of
			     NONE => suffix suf
			   | SOME f => f
		       fun list(prefix: string, l: string list): string list =
			  List.map(l, fn s => prefix ^ s)
		       fun docc(input, output, switches, libs) =
			  system
			  (gcc :: switches
			   @ list("-D", !defines)
			   @ list("-I", rev(!includeDirs))
			   @ list("-L", rev(!libDirs))
			   @ [if !debug then "-g" else "",
				 "-O" ^ Int.toString(!optimization)]
			   @ gccSwitches
			   @ ["-o", File.toString output,
			      File.toString input]
			   @ libs)
		       fun compileO(input: File.t) =
			  case stop of
			     Place.O => ()
			   | Place.OUT =>
				docc(input, maybeOut "",
				     if !Control.profile
					then ["-p"]
				     else [],
				     list("-l", !lib :: !libs)
				     @ rest)
			   | _ => Error.bug "strange stop in compileO"
		       fun makeO(input: File.t) =
			  let
			     val output =
				case stop of
				   Place.O => maybeOut ".o"
				 | _ => if !keepO then suffix ".o"
					else temp ".o"
			  in docc(input, output,
				  "-c" :: (if !Control.profile
					      then ["-g"]
					   else []),
				  [])
			     ; compileO output
			  end
		       fun compileC(input: File.t) =
			  let
			     fun makeS output = docc(input, output, ["-S"], [])
			  in case stop of
			     Place.S => makeS(maybeOut ".s")
			   | _ => if !keepS
				     then let val output = suffix ".s"
					  in makeS output; makeO output
					  end
				  else makeO input
			  end
		       fun compile() =
			  case start of
			     Place.SML =>
				let
				   val output =
				      case stop of
					 Place.CoreML => maybeOut ".core-ml"
				       | Place.C => maybeOut ".c"
				       | _ => if !keepC then suffix ".c"
					      else temp ".c"
				in Compile.compileFile
				   {input = File.toString input,
				    output = File.toString output}
				   (* These collects are so that the heap
				    * is shrunk as much as possible before
				    * calling gcc
				    *)
				   ; MLton.GC.collect()
				   ; MLton.GC.collect()
				   ; (case stop of
					 Place.CoreML => ()
				       | Place.C => ()
				       | _ => compileC output)
				end
			   | Place.C => compileC input
			   | Place.S => makeO input
			   | Place.O => compileO input
			   | _ => Error.bug "invalid start"
		    in DynamicWind.dynamicWind
		       (compile, fn () => List.foreach(!tempFiles, File.remove))
		    end
	      end
   end 
   
fun exportNJ(root: Dir.t, file: File.t): unit =
   (PrimEnv.forceAll root
    ; Control.defaults()
    ; (SMLofNJ.exportFn
       (File.toString file,
	fn (_, args) => ((commandLine args; OS.Process.success)
			 handle _ => OS.Process.failure))))
   
fun exportMLton(): unit =
   case CommandLine.arguments() of
      [root, file] => exportNJ(Dir.fromString root, File.fromString file)
    | _ => Error.bug "usage: exportMLton root file"

end
val _ = Main.exportMLton()
